@@ -1,9 +1,6 @@
 /**
- * @overview Schema for validating JSDoc doclets.
- *
- * @author Michael Mathews <micmath@gmail.com>
- * @author Jeff Williams <jeffrey.l.williams@gmail.com>
- * @license Apache License 2.0 - See file 'LICENSE.md' in this project.
+ * Schema for validating JSDoc doclets.
+ * @module jsdoc/schema
  * @see <http://tools.ietf.org/html/draft-zyp-json-schema-03>
  */
 'use strict';
@@ -11,7 +8,6 @@
 // JSON schema types
 var ARRAY = 'array';
 var BOOLEAN = 'boolean';
-var INTEGER = 'integer';
 var NULL = 'null';
 var NUMBER = 'number';
 var OBJECT = 'object';
@@ -21,8 +17,8 @@ var UNDEFINED = 'undefined';
 var BOOLEAN_OPTIONAL = [BOOLEAN, NULL, UNDEFINED];
 var STRING_OPTIONAL = [STRING, NULL, UNDEFINED];
 
-var EVENT_REGEXP = /event\:[\S]+/;
-var PACKAGE_REGEXP = /package\:[\S]+/;
+var EVENT_REGEXP = /event:[\S]+/;
+var PACKAGE_REGEXP = /package:[\S]+/;
 
 // information about the code associated with a doclet
 var META_SCHEMA = exports.META_SCHEMA = {
@@ -66,6 +62,11 @@ var META_SCHEMA = exports.META_SCHEMA = {
                 }
             }
         },
+        columnno: {
+            title: 'The column number of the code associated with this doclet.',
+            type: NUMBER,
+            optional: true
+        },
         filename: {
             title: 'The name of the file that contains the code associated with this doclet.',
             type: STRING,
@@ -112,8 +113,8 @@ var TYPE_PROPERTY_SCHEMA = exports.TYPE_PROPERTY_SCHEMA = {
         },
         // type parser output
         parsedType: {
-           type: OBJECT,
-           additionalProperties: true
+            type: OBJECT,
+            additionalProperties: true
         }
     }
 };
@@ -218,6 +219,7 @@ var DOCLET_SCHEMA = exports.DOCLET_SCHEMA = {
             optional: true,
             // TODO: define this as an enumeration elsewhere
             enum: [
+                'package',
                 'private',
                 'protected',
                 'public'
@@ -227,6 +229,10 @@ var DOCLET_SCHEMA = exports.DOCLET_SCHEMA = {
             type: STRING,
             optional: true
         },
+        async: {
+            type: BOOLEAN,
+            optional: true
+        },
         augments: {
             type: ARRAY,
             optional: true,
@@ -328,6 +334,14 @@ var DOCLET_SCHEMA = exports.DOCLET_SCHEMA = {
             type: BOOLEAN_OPTIONAL,
             optional: true
         },
+        generator: {
+            type: BOOLEAN,
+            optional: true
+        },
+        hideconstructor: {
+            type: BOOLEAN,
+            optional: true
+        },
         ignore: {
             type: BOOLEAN,
             optional: true
@@ -583,6 +597,12 @@ var DOCLET_SCHEMA = exports.DOCLET_SCHEMA = {
         virtual: {
             type: BOOLEAN,
             optional: true
+        },
+        yields: {
+            type: ARRAY,
+            optional: true,
+            minItems: 1,
+            items: PARAM_SCHEMA
         }
     }
 };
@@ -740,7 +760,7 @@ var PACKAGE_SCHEMA = exports.PACKAGE_SCHEMA = {
     }
 };
 
-var DOCLETS_SCHEMA = exports.DOCLETS_SCHEMA = {
+exports.DOCLETS_SCHEMA = {
     type: ARRAY,
     items: {
         anyOf: [DOCLET_SCHEMA, PACKAGE_SCHEMA]
