@@ -6,6 +6,7 @@
  * LICENSE file in the root directory of this source tree. An additional grant
  * of patent rights can be found in the PATENTS file in the same directory.
  *
+ * @providesModule monitorCodeUse
  */
 
 'use strict';
@@ -20,7 +21,7 @@ var invariant = require('./invariant');
  */
 
 function monitorCodeUse(eventName, data) {
-  !(eventName && !/[^a-z0-9_]/.test(eventName)) ? process.env.NODE_ENV !== 'production' ? invariant(false, 'You must provide an eventName using only the characters [a-z0-9_]') : invariant(false) : void 0;
+  !(eventName && !/[^a-z0-9_]/.test(eventName)) ? process.env.NODE_ENV !== 'production' ? invariant(false, 'You must provide an eventName using only the characters [a-z0-9_]') : invariant(false) : undefined;
 }
 
 module.exports = monitorCodeUse;
\ No newline at end of file
