@@ -1,16 +1,17 @@
-'use strict';
-
 /**
- * Copyright (c) 2013-present, Facebook, Inc.
+ * Copyright 2013-2015, Facebook, Inc.
  * All rights reserved.
  *
  * This source code is licensed under the BSD-style license found in the
  * LICENSE file in the root directory of this source tree. An additional grant
  * of patent rights can be found in the PATENTS file in the same directory.
  *
+ * @providesModule performanceNow
  * @typechecks
  */
 
+'use strict';
+
 var performance = require('./performance');
 
 var performanceNow;
@@ -21,11 +22,11 @@ var performanceNow;
  * because of Facebook's testing infrastructure.
  */
 if (performance.now) {
-  performanceNow = function performanceNow() {
+  performanceNow = function () {
     return performance.now();
   };
 } else {
-  performanceNow = function performanceNow() {
+  performanceNow = function () {
     return Date.now();
   };
 }
