@@ -86,6 +86,11 @@ DBStatus ProcessDeltaKey(Getter* base, rocksdb::WBWIIterator* delta, rocksdb::Sl
       }
       break;
     }
+    case rocksdb::kSingleDeleteRecord:
+      // Treating a SingleDelete operation as a standard Delete doesn't
+      // quite mirror SingleDelete's expected implementation, but it
+      // doesn't violate the operation's contract:
+      // https://github.com/facebook/rocksdb/wiki/Single-Delete.
     case rocksdb::kDeleteRecord:
       if (value->data != NULL) {
         free(value->data);
@@ -417,20 +422,38 @@ class DBBatchInserter : public rocksdb::WriteBatch::Handler {
  public:
   DBBatchInserter(rocksdb::WriteBatchBase* batch) : batch_(batch) {}
 
-  virtual void Put(const rocksdb::Slice& key, const rocksdb::Slice& value) {
-    batch_->Put(key, value);
+  virtual rocksdb::Status PutCF(uint32_t column_family_id, const rocksdb::Slice& key,
+                                const rocksdb::Slice& value) {
+    if (column_family_id != 0) {
+      return rocksdb::Status::InvalidArgument("DBBatchInserter: column families not supported");
+    }
+    return batch_->Put(key, value);
   }
-  virtual void Delete(const rocksdb::Slice& key) { batch_->Delete(key); }
-  virtual void Merge(const rocksdb::Slice& key, const rocksdb::Slice& value) {
-    batch_->Merge(key, value);
+  virtual rocksdb::Status DeleteCF(uint32_t column_family_id, const rocksdb::Slice& key) {
+    if (column_family_id != 0) {
+      return rocksdb::Status::InvalidArgument("DBBatchInserter: column families not supported");
+    }
+    return batch_->Delete(key);
+  }
+  virtual rocksdb::Status SingleDeleteCF(uint32_t column_family_id, const rocksdb::Slice& key) {
+    if (column_family_id != 0) {
+      return rocksdb::Status::InvalidArgument("DBBatchInserter: column families not supported");
+    }
+    return batch_->SingleDelete(key);
+  }
+  virtual rocksdb::Status MergeCF(uint32_t column_family_id, const rocksdb::Slice& key,
+                                  const rocksdb::Slice& value) {
+    if (column_family_id != 0) {
+      return rocksdb::Status::InvalidArgument("DBBatchInserter: column families not supported");
+    }
+    return batch_->Merge(key, value);
   }
   virtual rocksdb::Status DeleteRangeCF(uint32_t column_family_id, const rocksdb::Slice& begin_key,
                                         const rocksdb::Slice& end_key) {
-    if (column_family_id == 0) {
-      batch_->DeleteRange(begin_key, end_key);
-      return rocksdb::Status::OK();
+    if (column_family_id != 0) {
+      return rocksdb::Status::InvalidArgument("DBBatchInserter: column families not supported");
     }
-    return rocksdb::Status::InvalidArgument("DeleteRangeCF not implemented");
+    return batch_->DeleteRange(begin_key, end_key);
   }
   virtual void LogData(const rocksdb::Slice& blob) { batch_->PutLogData(blob); }
 
@@ -479,6 +502,12 @@ DBStatus DBBatch::Delete(DBKey key) {
   return kSuccess;
 }
 
+DBStatus DBBatch::SingleDelete(DBKey key) {
+  ++updates;
+  batch.SingleDelete(EncodeKey(key));
+  return kSuccess;
+}
+
 DBStatus DBBatch::DeleteRange(DBKey start, DBKey end) {
   ++updates;
   has_delete_range = true;
@@ -528,6 +557,8 @@ DBIterator* DBBatch::NewIter(DBIterOptions iter_options) {
 
 DBStatus DBBatch::GetStats(DBStatsResult* stats) { return FmtStatus("unsupported"); }
 
+DBStatus DBBatch::GetTickersAndHistograms(DBTickersAndHistogramsResult* stats) { return FmtStatus("unsupported"); }
+
 DBString DBBatch::GetCompactionStats() { return ToDBString("unsupported"); }
 
 DBStatus DBBatch::GetEnvStats(DBEnvStatsResult* stats) { return FmtStatus("unsupported"); }
@@ -582,6 +613,12 @@ DBStatus DBWriteOnlyBatch::Delete(DBKey key) {
   return kSuccess;
 }
 
+DBStatus DBWriteOnlyBatch::SingleDelete(DBKey key) {
+  ++updates;
+  batch.SingleDelete(EncodeKey(key));
+  return kSuccess;
+}
+
 DBStatus DBWriteOnlyBatch::DeleteRange(DBKey start, DBKey end) {
   ++updates;
   batch.DeleteRange(EncodeKey(start), EncodeKey(end));
@@ -619,6 +656,8 @@ DBIterator* DBWriteOnlyBatch::NewIter(DBIterOptions) { return NULL; }
 
 DBStatus DBWriteOnlyBatch::GetStats(DBStatsResult* stats) { return FmtStatus("unsupported"); }
 
+DBStatus DBWriteOnlyBatch::GetTickersAndHistograms(DBTickersAndHistogramsResult* stats) { return FmtStatus("unsupported"); }
+
 DBString DBWriteOnlyBatch::GetCompactionStats() { return ToDBString("unsupported"); }
 
 DBStatus DBWriteOnlyBatch::GetEnvStats(DBEnvStatsResult* stats) { return FmtStatus("unsupported"); }
