@@ -18,6 +18,7 @@
 #include <rocksdb/perf_context.h>
 #include <rocksdb/sst_file_writer.h>
 #include <rocksdb/table.h>
+#include <rocksdb/utilities/checkpoint.h>
 #include <stdarg.h>
 #include "batch.h"
 #include "cache.h"
@@ -35,7 +36,7 @@
 #include "options.h"
 #include "snapshot.h"
 #include "status.h"
-#include "timebound.h"
+#include "table_props.h"
 
 using namespace cockroach;
 
@@ -179,6 +180,9 @@ DBStatus DBOpen(DBEngine** db, DBSlice dir, DBOptions db_opts) {
     auto memenv = rocksdb::NewMemEnv(rocksdb::Env::Default());
     // Register it for deletion.
     env_mgr->TakeEnvOwnership(memenv);
+    // Create a root directory to suppress error messages that RocksDB would
+    // print if it had to create the DB directory itself.
+    memenv->CreateDir("/");
     // Make it the env that all other Envs must wrap.
     env_mgr->base_env = memenv;
     // Make it the env for rocksdb.
@@ -254,6 +258,22 @@ DBStatus DBOpen(DBEngine** db, DBSlice dir, DBOptions db_opts) {
   return kSuccess;
 }
 
+DBStatus DBCreateCheckpoint(DBEngine* db, DBSlice dir) {
+  const std::string cp_dir = ToString(dir);
+
+  rocksdb::Checkpoint* cp_ptr;
+  auto status = rocksdb::Checkpoint::Create(db->rep, &cp_ptr);
+  if (!status.ok()) {
+    return ToDBStatus(status);
+  }
+  // NB: passing 0 for log_size_for_flush forces a WAL sync, i.e. makes sure
+  // that the checkpoint is up to date.
+  status = cp_ptr->CreateCheckpoint(cp_dir, 0 /* log_size_for_flush */);
+  delete(cp_ptr);
+  return ToDBStatus(status);
+}
+
+
 DBStatus DBDestroy(DBSlice dir) {
   rocksdb::Options options;
   return ToDBStatus(rocksdb::DestroyDB(ToString(dir), options));
@@ -291,7 +311,7 @@ DBStatus DBSyncWAL(DBEngine* db) {
   options.sync = true;
   return ToDBStatus(db->rep->Write(options, &batch));
 #else
-  return ToDBStatus(db->rep->SyncWAL());
+  return ToDBStatus(db->rep->FlushWAL(true /* sync */));
 #endif
 }
 
@@ -411,6 +431,16 @@ DBStatus DBCompactRange(DBEngine* db, DBSlice start, DBSlice end, bool force_bot
   return kSuccess;
 }
 
+DBStatus DBDisableAutoCompaction(DBEngine* db) {
+  auto status = db->rep->SetOptions({{"disable_auto_compactions", "true"}});
+  return ToDBStatus(status);
+}
+
+DBStatus DBEnableAutoCompaction(DBEngine* db) {
+  auto status = db->rep->EnableAutoCompaction({db->rep->DefaultColumnFamily()});
+  return ToDBStatus(status);
+}
+
 DBStatus DBApproximateDiskBytes(DBEngine* db, DBKey start, DBKey end, uint64_t* size) {
   const std::string start_key(EncodeKey(start));
   const std::string end_key(EncodeKey(end));
@@ -429,6 +459,8 @@ DBStatus DBGet(DBEngine* db, DBKey key, DBString* value) { return db->Get(key, v
 
 DBStatus DBDelete(DBEngine* db, DBKey key) { return db->Delete(key); }
 
+DBStatus DBSingleDelete(DBEngine* db, DBKey key) { return db->SingleDelete(key); }
+
 DBStatus DBDeleteRange(DBEngine* db, DBKey start, DBKey end) { return db->DeleteRange(start, end); }
 
 DBStatus DBDeleteIterRange(DBEngine* db, DBIterator* iter, DBKey start, DBKey end) {
@@ -612,6 +644,7 @@ DBIterState DBIterPrev(DBIterator* iter, bool skip_current_key_versions) {
   return DBIterGetState(iter);
 }
 
+void DBIterSetLowerBound(DBIterator* iter, DBKey key) { iter->SetLowerBound(key); }
 void DBIterSetUpperBound(DBIterator* iter, DBKey key) { iter->SetUpperBound(key); }
 
 DBStatus DBMerge(DBSlice existing, DBSlice update, DBString* new_value, bool full_merge) {
@@ -645,6 +678,17 @@ DBStatus DBPartialMergeOne(DBSlice existing, DBSlice update, DBString* new_value
 // write them to the provided DBStatsResult instance.
 DBStatus DBGetStats(DBEngine* db, DBStatsResult* stats) { return db->GetStats(stats); }
 
+// `DBGetTickersAndHistograms` retrieves maps of all RocksDB tickers and histograms.
+// It differs from `DBGetStats` by getting _every_ ticker and histogram, and by not
+// getting anything else (DB properties, for example).
+//
+// In addition to freeing the `DBString`s in the result, the caller is also
+// responsible for freeing `DBTickersAndHistogramsResult::tickers` and
+// `DBTickersAndHistogramsResult::histograms`.
+DBStatus DBGetTickersAndHistograms(DBEngine* db, DBTickersAndHistogramsResult* stats) {
+  return db->GetTickersAndHistograms(stats);
+}
+
 DBString DBGetCompactionStats(DBEngine* db) { return db->GetCompactionStats(); }
 
 DBStatus DBGetEnvStats(DBEngine* db, DBEnvStatsResult* stats) { return db->GetEnvStats(stats); }
@@ -662,7 +706,7 @@ DBStatus DBGetSortedWALFiles(DBEngine* db, DBWALFile** files, int* n) {
 DBString DBGetUserProperties(DBEngine* db) { return db->GetUserProperties(); }
 
 DBStatus DBIngestExternalFiles(DBEngine* db, char** paths, size_t len, bool move_files,
-                               bool allow_file_modifications) {
+                               bool write_global_seqno, bool allow_file_modifications) {
   std::vector<std::string> paths_vec;
   for (size_t i = 0; i < len; i++) {
     paths_vec.push_back(paths[i]);
@@ -678,15 +722,60 @@ DBStatus DBIngestExternalFiles(DBEngine* db, char** paths, size_t len, bool move
   ingest_options.snapshot_consistency = true;
   // If a file is ingested over existing data (including the range tombstones
   // used by range snapshots) or if a RocksDB snapshot is outstanding when this
-  // ingest runs, then after moving/copying the file, RocksDB will edit it
-  // (overwrite some of the bytes) to have a global sequence number. If this is
-  // false, it will error in these cases instead.
-  ingest_options.allow_global_seqno = allow_file_modifications;
+  // ingest runs, then after moving/copying the file, historically RocksDB would
+  // edit it (overwrite some of the bytes) to have a global sequence number.
+  // After https://github.com/facebook/rocksdb/pull/4172 this can be disabled
+  // (with the mutable manifest/metadata tracking that instead). However it is
+  // only safe to disable the seqno write if older versions of RocksDB (<5.16)
+  // will not be used to read these SSTs.
+  //
+  // RocksDB checks the option allow_global_seqno and, if it is false, returns
+  // an error instead of ingesting a file that would require one. However it
+  // does this check *even if it is not planning on writing seqno* at all, so we
+  // need to set allow_global_seqno to true for !write_global_seqno to have the
+  // desired effect.
+  ingest_options.write_global_seqno = write_global_seqno;
+  ingest_options.allow_global_seqno = allow_file_modifications || !write_global_seqno;
   // If there are mutations in the memtable for the keyrange covered by the file
   // being ingested, this option is checked. If true, the memtable is flushed
-  // and the ingest run. If false, an error is returned.
-  ingest_options.allow_blocking_flush = true;
+  // using a blocking, write-stalling flush and the ingest run. If false, an
+  // error is returned.
+  //
+  // We want to ingest, but we do not want a write-stall, so we initially set it
+  // to false -- if our ingest fails, we'll do a manual, no-stall flush and wait
+  // for it to finish before trying the ingest again.
+  ingest_options.allow_blocking_flush = false;
+
   rocksdb::Status status = db->rep->IngestExternalFile(paths_vec, ingest_options);
+  if (status.IsInvalidArgument()) {
+    // TODO(dt): inspect status to see if it has the message
+    //          `External file requires flush`
+    //           since the move_file and other errors also use kInvalidArgument.
+
+    // It is possible we failed because the memtable required a flush but in the
+    // options above, we set "allow_blocking_flush = false" preventing ingest
+    // from running flush with allow_write_stall = true and halting foreground
+    // traffic. Now that we know we need to flush, let's do one ourselves, with
+    // allow_write_stall = false and wait for it. After it finishes we can retry
+    // the ingest.
+    rocksdb::FlushOptions flush_options;
+    flush_options.allow_write_stall = false;
+    flush_options.wait = true;
+
+    rocksdb::Status flush_status = db->rep->Flush(flush_options);
+    if (!flush_status.ok()) {
+      return ToDBStatus(flush_status);
+    }
+
+    // Hopefully on this second attempt we will not need to flush at all, but
+    // just in case we do, we'll allow the write stall this time -- that way we
+    // can ensure we actually get the ingestion done and move on. A stalling
+    // flush is be less than ideal, but since we just flushed, a) this shouldn't
+    // happen often and b) if it does, it should be small and quick.
+    ingest_options.allow_blocking_flush = true;
+    status = db->rep->IngestExternalFile(paths_vec, ingest_options);
+  }
+
   if (!status.ok()) {
     return ToDBStatus(status);
   }
@@ -727,6 +816,9 @@ DBSstFileWriter* DBSstFileWriterNew() {
   // timestamps present in each sstable in the metadata for that sstable. Used
   // by the time bounded iterator optimization.
   options->table_properties_collector_factories.emplace_back(DBMakeTimeBoundCollector());
+  // Automatically request compactions whenever an SST contains too many range
+  // deletions.
+  options->table_properties_collector_factories.emplace_back(DBMakeDeleteRangeCollector());
 
   std::unique_ptr<rocksdb::Env> memenv;
   memenv.reset(rocksdb::NewMemEnv(rocksdb::Env::Default()));
