@@ -53,6 +53,7 @@ typedef struct {
 
 typedef struct {
   bool prefix;
+  DBKey lower_bound;
   DBKey upper_bound;
   bool with_stats;
   DBTimestamp min_timestamp_hint;
@@ -99,6 +100,12 @@ void DBReleaseCache(DBCache* cache);
 // exist.
 DBStatus DBOpen(DBEngine** db, DBSlice dir, DBOptions options);
 
+// Creates a RocksDB checkpoint in the specified directory (which must not exist).
+// A checkpoint is a logical copy of the database, though it will hardlink the
+// SSTs references by it (when possible), thus avoiding duplication of any of
+// the actual data.
+DBStatus DBCreateCheckpoint(DBEngine* db, DBSlice dir);
+
 // Set a callback to be invoked during DBOpen that can make changes to RocksDB
 // initialization. Used by CCL code to install additional features.
 //
@@ -130,6 +137,12 @@ DBStatus DBCompact(DBEngine* db);
 // If end is empty, it indicates the end of the database.
 DBStatus DBCompactRange(DBEngine* db, DBSlice start, DBSlice end, bool force_bottommost);
 
+// Disable/enable automatic compactions. Automatic compactions are
+// enabled by default. Disabling is provided for testing purposes so
+// that automatic compactions do not interfere with test expectations.
+DBStatus DBDisableAutoCompaction(DBEngine* db);
+DBStatus DBEnableAutoCompaction(DBEngine* db);
+
 // Stores the approximate on-disk size of the given key range into the
 // supplied uint64.
 DBStatus DBApproximateDiskBytes(DBEngine* db, DBKey start, DBKey end, uint64_t* size);
@@ -146,6 +159,11 @@ DBStatus DBGet(DBEngine* db, DBKey key, DBString* value);
 // Deletes the database entry for "key".
 DBStatus DBDelete(DBEngine* db, DBKey key);
 
+// Deletes the most recent database entry for "key". See the following
+// documentation for details on the subtleties of this operation:
+// https://github.com/facebook/rocksdb/wiki/Single-Delete.
+DBStatus DBSingleDelete(DBEngine* db, DBKey key);
+
 // Deletes a range of keys from start (inclusive) to end (exclusive).
 DBStatus DBDeleteRange(DBEngine* db, DBKey start, DBKey end);
 
@@ -243,6 +261,9 @@ DBIterState DBIterNext(DBIterator* iter, bool skip_current_key_versions);
 // iff the iterator was not positioned at the first key.
 DBIterState DBIterPrev(DBIterator* iter, bool skip_current_key_versions);
 
+// DBIterSetLowerBound updates this iterator's lower bound.
+void DBIterSetLowerBound(DBIterator* iter, DBKey key);
+
 // DBIterSetUpperBound updates this iterator's upper bound.
 void DBIterSetUpperBound(DBIterator* iter, DBKey key);
 
@@ -292,6 +313,7 @@ DBStatus MVCCFindSplitKey(DBIterator* iter, DBKey start, DBKey end, DBKey min_sp
 typedef struct {
   DBSlice id;
   uint32_t epoch;
+  int32_t sequence;
   DBTimestamp max_timestamp;
 } DBTxn;
 
@@ -310,12 +332,14 @@ typedef struct {
   DBChunkedBuffer data;
   DBSlice intents;
   DBTimestamp uncertainty_timestamp;
+  DBSlice resume_key;
 } DBScanResults;
 
 DBScanResults MVCCGet(DBIterator* iter, DBSlice key, DBTimestamp timestamp, DBTxn txn,
-                      bool consistent, bool tombstones);
+                      bool inconsistent, bool tombstones, bool ignore_sequence);
 DBScanResults MVCCScan(DBIterator* iter, DBSlice start, DBSlice end, DBTimestamp timestamp,
-                       int64_t max_keys, DBTxn txn, bool consistent, bool reverse, bool tombstones);
+                       int64_t max_keys, DBTxn txn, bool inconsistent, bool reverse,
+                       bool tombstones, bool ignore_sequence);
 
 // DBStatsResult contains various runtime stats for RocksDB.
 typedef struct {
@@ -330,8 +354,32 @@ typedef struct {
   int64_t compactions;
   int64_t table_readers_mem_estimate;
   int64_t pending_compaction_bytes_estimate;
+  int64_t l0_file_count;
 } DBStatsResult;
 
+typedef struct {
+  DBString name;
+  uint64_t value;
+} TickerInfo;
+
+typedef struct {
+  DBString name;
+  double mean;
+  double p50;
+  double p95;
+  double p99;
+  double max;
+  uint64_t count;
+  uint64_t sum;
+} HistogramInfo;
+
+typedef struct {
+  TickerInfo* tickers;
+  size_t tickers_len;
+  HistogramInfo* histograms;
+  size_t histograms_len;
+} DBTickersAndHistogramsResult;
+
 // DBEnvStatsResult contains Env stats (filesystem layer).
 typedef struct {
   // Basic file encryption stats:
@@ -341,6 +389,8 @@ typedef struct {
   // Files/bytes using the active data key.
   uint64_t active_key_files;
   uint64_t active_key_bytes;
+  // Enum of the encryption algorithm in use.
+  int32_t encryption_type;
   // encryption status (CCL only).
   // This is a serialized enginepbccl/stats.proto:EncryptionStatus
   DBString encryption_status;
@@ -355,6 +405,7 @@ typedef struct {
 } DBEncryptionRegistries;
 
 DBStatus DBGetStats(DBEngine* db, DBStatsResult* stats);
+DBStatus DBGetTickersAndHistograms(DBEngine* db, DBTickersAndHistogramsResult* stats);
 DBString DBGetCompactionStats(DBEngine* db);
 DBStatus DBGetEnvStats(DBEngine* db, DBEnvStatsResult* stats);
 DBStatus DBGetEncryptionRegistries(DBEngine* db, DBEncryptionRegistries* result);
@@ -390,9 +441,12 @@ DBString DBGetUserProperties(DBEngine* db);
 // what can be added. If move_files is true, the files will be moved instead of
 // copied. If allow_file_modifications is false, RocksDB will return an error if
 // it would have tried to modify any of the files' sequence numbers rather than
-// editing the files in place.
+// editing the files in place. If write_global_seqno is false, it will skip
+// writing the seq_no to the SSTs -- this is only safe if this will only ever be
+// read by Rocks version >= 5.16 as older versions would still be looking for
+// the seqno.
 DBStatus DBIngestExternalFiles(DBEngine* db, char** paths, size_t len, bool move_files,
-                               bool allow_file_modifications);
+                               bool write_global_seqno, bool allow_file_modifications);
 
 typedef struct DBSstFileWriter DBSstFileWriter;
 
