@@ -23,6 +23,7 @@
 struct DBIterator {
   DBIterator(std::atomic<int64_t>* iters, DBIterOptions iter_options);
   ~DBIterator();
+  void SetLowerBound(DBKey key);
   void SetUpperBound(DBKey key);
 
   std::atomic<int64_t>* const iters_count;
@@ -30,8 +31,11 @@ struct DBIterator {
   std::unique_ptr<cockroach::chunkedBuffer> kvs;
   std::unique_ptr<rocksdb::WriteBatch> intents;
   std::unique_ptr<IteratorStats> stats;
+  std::string rev_resume_key;
 
   rocksdb::ReadOptions read_opts;
+  std::string lower_bound_str;
   std::string upper_bound_str;
+  rocksdb::Slice lower_bound;
   rocksdb::Slice upper_bound;
 };
