@@ -101,6 +101,9 @@ class UncollapsedRangeDelMap : public RangeDelMap {
     // TODO(peter): This is called from
     // Version::OverlapWithLevelIterator. That method needs to be tested in
     // the presence of range tombstones.
+    (void)start;
+    (void)end;
+    (void)seqno;
     return false;
   }
 
@@ -504,12 +507,7 @@ class CollapsedRangeDelMap : public RangeDelMap {
     }
   }
 
-  size_t Size() const override {
-    if (rep_.empty()) {
-      return 0;
-    }
-    return rep_.size() - 1;
-  }
+  size_t Size() const override { return rep_.empty() ? 0 : rep_.size() - 1; }
 
   void InvalidatePosition() override { iter_ = rep_.end(); }
 
