@@ -188,8 +188,9 @@ bool ShouldDeleteRange(const AddTombstonesArgs& range_dels,
 
 void VerifyGetTombstone(const AddTombstonesArgs& range_dels,
                         const ExpectedPoint& expected_point,
-                        const PartialRangeTombstone& expected_tombstone) {
-  RangeDelAggregator range_del_agg(bytewise_icmp, {} /* snapshots */, true);
+                        const PartialRangeTombstone& expected_tombstone,
+                        const InternalKeyComparator& icmp = bytewise_icmp) {
+  RangeDelAggregator range_del_agg(icmp, {} /* snapshots */, true);
   ASSERT_TRUE(range_del_agg.IsEmpty());
   AddTombstones(&range_del_agg, range_dels.tombstones,
                 range_dels.smallest, range_dels.largest);
@@ -506,25 +507,6 @@ TEST_F(RangeDelAggregatorTest, TruncateTombstones) {
       {{"b", "c", 10}, {"d", "e", 10}});
 }
 
-TEST_F(RangeDelAggregatorTest, IsEmpty) {
-  const std::vector<SequenceNumber> snapshots;
-  RangeDelAggregator range_del_agg1(
-      bytewise_icmp, snapshots, false /* collapse_deletions */);
-  ASSERT_TRUE(range_del_agg1.IsEmpty());
-
-  RangeDelAggregator range_del_agg2(
-      bytewise_icmp, snapshots, true /* collapse_deletions */);
-  ASSERT_TRUE(range_del_agg2.IsEmpty());
-
-  RangeDelAggregator range_del_agg3(
-      bytewise_icmp, kMaxSequenceNumber, false /* collapse_deletions */);
-  ASSERT_TRUE(range_del_agg3.IsEmpty());
-
-  RangeDelAggregator range_del_agg4(
-      bytewise_icmp, kMaxSequenceNumber, true /* collapse_deletions */);
-  ASSERT_TRUE(range_del_agg4.IsEmpty());
-}
-
 TEST_F(RangeDelAggregatorTest, OverlappingLargestKeyTruncateBelowTombstone) {
   const InternalKey smallest("b", kMaxSequenceNumber, kTypeRangeDeletion);
   const InternalKey largest(
@@ -642,6 +624,26 @@ TEST_F(RangeDelAggregatorTest, FileCoversOneKeyAndTombstoneBelow) {
       {{"a", "a", 15}}); // empty tombstone but can't occur during a compaction
 }
 
+TEST_F(RangeDelAggregatorTest, IsEmpty) {
+  auto& icmp = bytewise_icmp;
+  const std::vector<SequenceNumber> snapshots;
+  RangeDelAggregator range_del_agg1(
+      icmp, snapshots, false /* collapse_deletions */);
+  ASSERT_TRUE(range_del_agg1.IsEmpty());
+
+  RangeDelAggregator range_del_agg2(
+      icmp, snapshots, true /* collapse_deletions */);
+  ASSERT_TRUE(range_del_agg2.IsEmpty());
+
+  RangeDelAggregator range_del_agg3(
+      icmp, kMaxSequenceNumber, false /* collapse_deletions */);
+  ASSERT_TRUE(range_del_agg3.IsEmpty());
+
+  RangeDelAggregator range_del_agg4(
+      icmp, kMaxSequenceNumber, true /* collapse_deletions */);
+  ASSERT_TRUE(range_del_agg4.IsEmpty());
+}
+
 }  // namespace rocksdb
 
 int main(int argc, char** argv) {
