@@ -14,9 +14,7 @@
 
 package acceptance
 
-import (
-	"testing"
-)
+import "testing"
 
 func TestMain(m *testing.M) {
 	MainTest(m)
