@@ -14,9 +14,7 @@
 
 package base
 
-import (
-	"github.com/cockroachdb/cockroach/pkg/build"
-)
+import "github.com/cockroachdb/cockroach/pkg/build"
 
 // DocsURLBase is the root URL for the version of the docs associated with this
 // binary.
