@@ -15,8 +15,6 @@ import (
 	"sort"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/ccl/storageccl"
 	"github.com/cockroachdb/cockroach/pkg/ccl/utilccl"
@@ -43,6 +41,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 const (
@@ -669,73 +668,81 @@ func backup(
 			return progressLogger.Loop(ctx, requestFinishedCh)
 		})
 	}
+	g.GoCtx(func(ctx context.Context) error {
+		for i := range allSpans {
+			{
+				select {
+				case exportsSem <- struct{}{}:
+				case <-ctx.Done():
+					// Break the for loop to avoid creating more work - the backup
+					// has failed because either the context has been canceled or an
+					// error has been returned. Either way, Wait() is guaranteed to
+					// return an error now.
+					return ctx.Err()
+				}
+			}
 
-	for i := range allSpans {
-		select {
-		case exportsSem <- struct{}{}:
-		case <-g.Done:
-			return mu.exported, g.Err()
+			span := allSpans[i]
+			g.GoCtx(func(ctx context.Context) error {
+				defer func() { <-exportsSem }()
+				header := roachpb.Header{Timestamp: span.end}
+				req := &roachpb.ExportRequest{
+					RequestHeader: roachpb.RequestHeaderFromSpan(span.span),
+					Storage:       exportStore.Conf(),
+					StartTime:     span.start,
+					MVCCFilter:    roachpb.MVCCFilter(backupDesc.MVCCFilter),
+				}
+				rawRes, pErr := client.SendWrappedWith(ctx, db.NonTransactionalSender(), header, req)
+				if pErr != nil {
+					return pErr.GoError()
+				}
+				res := rawRes.(*roachpb.ExportResponse)
+
+				mu.Lock()
+				if backupDesc.RevisionStartTime.Less(res.StartTime) {
+					backupDesc.RevisionStartTime = res.StartTime
+				}
+				for _, file := range res.Files {
+					f := BackupDescriptor_File{
+						Span:        file.Span,
+						Path:        file.Path,
+						Sha512:      file.Sha512,
+						EntryCounts: file.Exported,
+					}
+					if span.start != backupDesc.StartTime {
+						f.StartTime = span.start
+						f.EndTime = span.end
+					}
+					mu.files = append(mu.files, f)
+					mu.exported.Add(file.Exported)
+				}
+				var checkpointFiles BackupFileDescriptors
+				if timeutil.Since(mu.lastCheckpoint) > BackupCheckpointInterval {
+					// We optimistically assume the checkpoint will succeed to prevent
+					// multiple threads from attempting to checkpoint.
+					mu.lastCheckpoint = timeutil.Now()
+					checkpointFiles = append(checkpointFiles, mu.files...)
+				}
+				mu.Unlock()
+
+				requestFinishedCh <- struct{}{}
+
+				if checkpointFiles != nil {
+					checkpointMu.Lock()
+					backupDesc.Files = checkpointFiles
+					err := writeBackupDescriptor(
+						ctx, exportStore, BackupDescriptorCheckpointName, backupDesc,
+					)
+					checkpointMu.Unlock()
+					if err != nil {
+						log.Errorf(ctx, "unable to checkpoint backup descriptor: %+v", err)
+					}
+				}
+				return nil
+			})
 		}
-
-		span := allSpans[i]
-		g.GoCtx(func(ctx context.Context) error {
-			defer func() { <-exportsSem }()
-			header := roachpb.Header{Timestamp: span.end}
-			req := &roachpb.ExportRequest{
-				RequestHeader: roachpb.RequestHeaderFromSpan(span.span),
-				Storage:       exportStore.Conf(),
-				StartTime:     span.start,
-				MVCCFilter:    roachpb.MVCCFilter(backupDesc.MVCCFilter),
-			}
-			rawRes, pErr := client.SendWrappedWith(ctx, db.NonTransactionalSender(), header, req)
-			if pErr != nil {
-				return pErr.GoError()
-			}
-			res := rawRes.(*roachpb.ExportResponse)
-
-			mu.Lock()
-			if backupDesc.RevisionStartTime.Less(res.StartTime) {
-				backupDesc.RevisionStartTime = res.StartTime
-			}
-			for _, file := range res.Files {
-				f := BackupDescriptor_File{
-					Span:        file.Span,
-					Path:        file.Path,
-					Sha512:      file.Sha512,
-					EntryCounts: file.Exported,
-				}
-				if span.start != backupDesc.StartTime {
-					f.StartTime = span.start
-					f.EndTime = span.end
-				}
-				mu.files = append(mu.files, f)
-				mu.exported.Add(file.Exported)
-			}
-			var checkpointFiles BackupFileDescriptors
-			if timeutil.Since(mu.lastCheckpoint) > BackupCheckpointInterval {
-				// We optimistically assume the checkpoint will succeed to prevent
-				// multiple threads from attempting to checkpoint.
-				mu.lastCheckpoint = timeutil.Now()
-				checkpointFiles = append(checkpointFiles, mu.files...)
-			}
-			mu.Unlock()
-
-			requestFinishedCh <- struct{}{}
-
-			if checkpointFiles != nil {
-				checkpointMu.Lock()
-				backupDesc.Files = checkpointFiles
-				err := writeBackupDescriptor(
-					ctx, exportStore, BackupDescriptorCheckpointName, backupDesc,
-				)
-				checkpointMu.Unlock()
-				if err != nil {
-					log.Errorf(ctx, "unable to checkpoint backup descriptor: %+v", err)
-				}
-			}
-			return nil
-		})
-	}
+		return nil
+	})
 
 	if err := g.Wait(); err != nil {
 		return mu.exported, errors.Wrapf(err, "exporting %d ranges", len(spans))
@@ -785,23 +792,23 @@ func VerifyUsableExportTarget(
 // backupPlanHook implements PlanHookFn.
 func backupPlanHook(
 	_ context.Context, stmt tree.Statement, p sql.PlanHookState,
-) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, error) {
+) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, bool, error) {
 	backupStmt, ok := stmt.(*tree.Backup)
 	if !ok {
-		return nil, nil, nil, nil
+		return nil, nil, nil, false, nil
 	}
 
 	toFn, err := p.TypeAsString(backupStmt.To, "BACKUP")
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 	incrementalFromFn, err := p.TypeAsStringArray(backupStmt.IncrementalFrom, "BACKUP")
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 	optsFn, err := p.TypeAsStringOpts(backupStmt.Options, backupOptionExpectValues)
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 
 	header := sqlbase.ResultColumns{
@@ -847,7 +854,7 @@ func backupPlanHook(
 		endTime := p.ExecCfg().Clock.Now()
 		if backupStmt.AsOf.Expr != nil {
 			var err error
-			if endTime, err = p.EvalAsOfTimestamp(backupStmt.AsOf, endTime); err != nil {
+			if endTime, err = p.EvalAsOfTimestamp(backupStmt.AsOf); err != nil {
 				return err
 			}
 		}
@@ -999,7 +1006,7 @@ func backupPlanHook(
 
 		// older nodes don't know about many new fields, e.g. MVCCAll and may
 		// incorrectly evaluate either an export RPC, or a resumed backup job.
-		if requireVersion2 && !p.ExecCfg().Settings.Version.IsMinSupported(cluster.Version2_0) {
+		if requireVersion2 && !p.ExecCfg().Settings.Version.IsActive(cluster.Version2_0) {
 			return errors.Errorf(
 				"BACKUP features introduced in 2.0 requires cluster version >= %s (",
 				cluster.VersionByKey(cluster.Version2_0).String(),
@@ -1075,7 +1082,7 @@ func backupPlanHook(
 		}
 		return <-errCh
 	}
-	return fn, header, nil, nil
+	return fn, header, nil, false, nil
 }
 
 type backupResumer struct {
