@@ -16,6 +16,7 @@ import (
 	"sort"
 	"sync/atomic"
 
+	"github.com/cockroachdb/cockroach/pkg/ccl/gossipccl"
 	"github.com/cockroachdb/cockroach/pkg/ccl/storageccl"
 	"github.com/cockroachdb/cockroach/pkg/ccl/utilccl"
 	"github.com/cockroachdb/cockroach/pkg/ccl/utilccl/intervalccl"
@@ -32,6 +33,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/sql/stats"
 	"github.com/cockroachdb/cockroach/pkg/util/ctxgroup"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/interval"
@@ -165,20 +167,17 @@ func rewriteViewQueryDBNames(table *sqlbase.TableDescriptor, newDB string) error
 		return errors.Wrapf(err, "failed to parse underlying query from view %q", table.Name)
 	}
 	// Re-format to change all DB names to `newDB`.
-	f := tree.NewFmtCtxWithBuf(tree.FmtParsable)
-	f.WithReformatTableNames(
-		func(ctx *tree.FmtCtx, t *tree.NormalizableTableName) {
-			tn, err := t.Normalize()
-			if err != nil {
-				return
-			}
-			// empty catalog e.g. ``"".information_schema.tables` should stay empty.
-			if tn.CatalogName != "" {
-				tn.CatalogName = tree.Name(newDB)
-			}
+	f := tree.NewFmtCtx(tree.FmtParsable)
+	f.SetReformatTableNames(func(ctx *tree.FmtCtx, tn *tree.TableName) {
+		// empty catalog e.g. ``"".information_schema.tables` should stay empty.
+		if tn.CatalogName != "" {
+			tn.CatalogName = tree.Name(newDB)
+		}
+		ctx.WithReformatTableNames(nil, func() {
 			ctx.FormatNode(tn)
 		})
-	f.FormatNode(stmt)
+	})
+	f.FormatNode(stmt.AST)
 	table.ViewQuery = f.CloseAndGetString()
 	return nil
 }
@@ -808,8 +807,8 @@ func splitAndScatter(
 			}
 
 			select {
-			case <-g.Done:
-				return g.Err()
+			case <-ctx.Done():
+				return ctx.Err()
 			case importSpanChunksCh <- importSpanChunk:
 			}
 		}
@@ -852,8 +851,8 @@ func splitAndScatter(
 					}
 
 					select {
-					case <-g.Done:
-						return g.Err()
+					case <-ctx.Done():
+						return ctx.Err()
 					case readyForImportCh <- importSpan:
 					}
 				}
@@ -1046,6 +1045,18 @@ func restore(
 		return mu.res, nil, nil, err
 	}
 
+	{
+		// Disable merging for the table IDs being restored into. We don't want the
+		// merge queue undoing the splits performed during RESTORE.
+		tableIDs := make([]uint32, 0, len(tables))
+		for _, t := range tables {
+			tableIDs = append(tableIDs, uint32(t.ID))
+		}
+		disableCtx, cancel := context.WithCancel(restoreCtx)
+		defer cancel()
+		gossipccl.DisableMerges(disableCtx, gossip, tableIDs)
+	}
+
 	// Get TableRekeys to use when importing raw data.
 	var rekeys []roachpb.ImportRequest_TableRekey
 	for i := range tables {
@@ -1136,68 +1147,71 @@ func restore(
 		defer tracing.FinishSpan(progressSpan)
 		return progressLogger.Loop(ctx, requestFinishedCh)
 	})
+	g.GoCtx(func(ctx context.Context) error {
+		log.Eventf(restoreCtx, "commencing import of data with concurrency %d", maxConcurrentImports)
+		for readyForImportSpan := range readyForImportCh {
+			newSpanKey, err := rewriteBackupSpanKey(kr, readyForImportSpan.Span.Key)
+			if err != nil {
+				return err
+			}
+			idx := readyForImportSpan.progressIdx
 
-	log.Eventf(restoreCtx, "commencing import of data with concurrency %d", maxConcurrentImports)
-	for readyForImportSpan := range readyForImportCh {
-		newSpanKey, err := rewriteBackupSpanKey(kr, readyForImportSpan.Span.Key)
-		if err != nil {
-			return mu.res, nil, nil, err
-		}
-		idx := readyForImportSpan.progressIdx
-
-		importRequest := &roachpb.ImportRequest{
-			// Import is a point request because we don't want DistSender to split
-			// it. Assume (but don't require) the entire post-rewrite span is on the
-			// same range.
-			RequestHeader: roachpb.RequestHeader{Key: newSpanKey},
-			DataSpan:      readyForImportSpan.Span,
-			Files:         readyForImportSpan.files,
-			EndTime:       endTime,
-			Rekeys:        rekeys,
-		}
-
-		log.VEventf(restoreCtx, 1, "importing %d of %d", idx, len(importSpans))
-
-		select {
-		case importsSem <- struct{}{}:
-		case <-g.Done:
-			return mu.res, nil, nil, errors.Wrapf(g.Wait(), "importing %d ranges", len(importSpans))
-		}
-
-		g.GoCtx(func(ctx context.Context) error {
-			ctx, importSpan := tracing.ChildSpan(ctx, "import")
-			log.Event(ctx, "acquired semaphore")
-			defer tracing.FinishSpan(importSpan)
-			defer func() { <-importsSem }()
-
-			importRes, pErr := client.SendWrapped(ctx, db.NonTransactionalSender(), importRequest)
-			if pErr != nil {
-				return pErr.GoError()
+			importRequest := &roachpb.ImportRequest{
+				// Import is a point request because we don't want DistSender to split
+				// it. Assume (but don't require) the entire post-rewrite span is on the
+				// same range.
+				RequestHeader: roachpb.RequestHeader{Key: newSpanKey},
+				DataSpan:      readyForImportSpan.Span,
+				Files:         readyForImportSpan.files,
+				EndTime:       endTime,
+				Rekeys:        rekeys,
 			}
 
-			mu.Lock()
-			mu.res.Add(importRes.(*roachpb.ImportResponse).Imported)
+			log.VEventf(restoreCtx, 1, "importing %d of %d", idx, len(importSpans))
 
-			// Assert that we're actually marking the correct span done. See #23977.
-			if !importSpans[idx].Key.Equal(importRequest.DataSpan.Key) {
+			select {
+			case importsSem <- struct{}{}:
+			case <-ctx.Done():
+				return ctx.Err()
+			}
+
+			g.GoCtx(func(ctx context.Context) error {
+				ctx, importSpan := tracing.ChildSpan(ctx, "import")
+				log.Event(ctx, "acquired semaphore")
+				defer tracing.FinishSpan(importSpan)
+				defer func() { <-importsSem }()
+
+				importRes, pErr := client.SendWrapped(ctx, db.NonTransactionalSender(), importRequest)
+				if pErr != nil {
+					return errors.Wrapf(pErr.GoError(), "importing span %v", importRequest.DataSpan)
+
+				}
+
+				mu.Lock()
+				mu.res.Add(importRes.(*roachpb.ImportResponse).Imported)
+
+				// Assert that we're actually marking the correct span done. See #23977.
+				if !importSpans[idx].Key.Equal(importRequest.DataSpan.Key) {
+					mu.Unlock()
+					return errors.Errorf(
+						"request %d for span %v (to %v) does not match import span for same idx: %v",
+						idx, importRequest.DataSpan, newSpanKey, importSpans[idx],
+					)
+				}
+				mu.requestsCompleted[idx] = true
+				for j := mu.highWaterMark + 1; j < len(mu.requestsCompleted) && mu.requestsCompleted[j]; j++ {
+					mu.highWaterMark = j
+				}
 				mu.Unlock()
-				return errors.Errorf(
-					"request %d for span %v (to %v) does not match import span for same idx: %v",
-					idx, importRequest.DataSpan, newSpanKey, importSpans[idx],
-				)
-			}
-			mu.requestsCompleted[idx] = true
-			for j := mu.highWaterMark + 1; j < len(mu.requestsCompleted) && mu.requestsCompleted[j]; j++ {
-				mu.highWaterMark = j
-			}
-			mu.Unlock()
 
-			requestFinishedCh <- struct{}{}
-			return nil
-		})
-	}
+				requestFinishedCh <- struct{}{}
+				return nil
+			})
+		}
+		log.Event(restoreCtx, "wait for outstanding imports to finish")
+		return nil
+	})
 
-	log.Event(restoreCtx, "wait for outstanding imports to finish")
 	if err := g.Wait(); err != nil {
 		// This leaves the data that did get imported in case the user wants to
 		// retry.
@@ -1222,20 +1236,20 @@ var RestoreHeader = sqlbase.ResultColumns{
 // restorePlanHook implements sql.PlanHookFn.
 func restorePlanHook(
 	_ context.Context, stmt tree.Statement, p sql.PlanHookState,
-) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, error) {
+) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, bool, error) {
 	restoreStmt, ok := stmt.(*tree.Restore)
 	if !ok {
-		return nil, nil, nil, nil
+		return nil, nil, nil, false, nil
 	}
 
 	fromFn, err := p.TypeAsStringArray(restoreStmt.From, "RESTORE")
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 
 	optsFn, err := p.TypeAsStringOpts(restoreStmt.Options, restoreOptionExpectValues)
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 
 	fn := func(ctx context.Context, _ []sql.PlanNode, resultsCh chan<- tree.Datums) error {
@@ -1266,7 +1280,7 @@ func restorePlanHook(
 		// nodes if needed. VersionClearRange was introduced after many of these
 		// fields and the refactors to how jobs were saved and resumed, though we may
 		// want to bump this to 2.0 for simplicity.
-		if !p.ExecCfg().Settings.Version.IsMinSupported(cluster.VersionClearRange) {
+		if !p.ExecCfg().Settings.Version.IsActive(cluster.VersionClearRange) {
 			return errors.Errorf(
 				"running RESTORE on a 2.x node requires cluster version >= %s",
 				cluster.VersionByKey(cluster.VersionClearRange).String(),
@@ -1279,10 +1293,8 @@ func restorePlanHook(
 		}
 		var endTime hlc.Timestamp
 		if restoreStmt.AsOf.Expr != nil {
-			// Use Now() for the max timestamp because Restore does its own
-			// (more restrictive) check.
 			var err error
-			endTime, err = p.EvalAsOfTimestamp(restoreStmt.AsOf, p.ExecCfg().Clock.Now())
+			endTime, err = p.EvalAsOfTimestamp(restoreStmt.AsOf)
 			if err != nil {
 				return err
 			}
@@ -1294,7 +1306,7 @@ func restorePlanHook(
 		}
 		return doRestorePlan(ctx, restoreStmt, p, from, endTime, opts, resultsCh)
 	}
-	return fn, RestoreHeader, nil, nil
+	return fn, RestoreHeader, nil, false, nil
 }
 
 func doRestorePlan(
@@ -1410,10 +1422,11 @@ func loadBackupSQLDescs(
 }
 
 type restoreResumer struct {
-	settings  *cluster.Settings
-	res       roachpb.BulkOpSummary
-	databases []*sqlbase.DatabaseDescriptor
-	tables    []*sqlbase.TableDescriptor
+	settings       *cluster.Settings
+	res            roachpb.BulkOpSummary
+	databases      []*sqlbase.DatabaseDescriptor
+	tables         []*sqlbase.TableDescriptor
+	statsRefresher *stats.Refresher
 }
 
 func (r *restoreResumer) Resume(
@@ -1442,6 +1455,7 @@ func (r *restoreResumer) Resume(
 	r.res = res
 	r.databases = databases
 	r.tables = tables
+	r.statsRefresher = p.ExecCfg().StatsRefresher
 	return err
 }
 
@@ -1474,6 +1488,13 @@ func (r *restoreResumer) OnSuccess(ctx context.Context, txn *client.Txn, job *jo
 		return errors.Wrapf(err, "restoring %d TableDescriptors", len(r.tables))
 	}
 
+	// Initiate a run of CREATE STATISTICS. We don't know the actual number of
+	// rows affected per table, so we use a large number because we want to make
+	// sure that stats always get created/refreshed here.
+	for i := range r.tables {
+		r.statsRefresher.NotifyMutation(r.tables[i].ID, math.MaxInt32 /* rowsAffected */)
+	}
+
 	return nil
 }
 
