@@ -8,9 +8,7 @@
 
 package buildccl
 
-import (
-	"github.com/cockroachdb/cockroach/pkg/build"
-)
+import "github.com/cockroachdb/cockroach/pkg/build"
 
 func init() {
 	build.Distribution = "CCL"
