@@ -10,18 +10,18 @@ package changefeedccl
 
 import (
 	"context"
-	"regexp"
+	"net/url"
 	"sort"
 	"time"
 
-	"github.com/cockroachdb/cockroach/pkg/util/retry"
-
 	"github.com/cockroachdb/cockroach/pkg/ccl/backupccl"
+	"github.com/cockroachdb/cockroach/pkg/ccl/storageccl"
 	"github.com/cockroachdb/cockroach/pkg/ccl/utilccl"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/jobs"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
 	"github.com/cockroachdb/cockroach/pkg/keys"
+	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
@@ -29,6 +29,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/pkg/errors"
 )
@@ -46,21 +47,30 @@ const (
 	optCursor                  = `cursor`
 	optEnvelope                = `envelope`
 	optFormat                  = `format`
+	optKeyInValue              = `key_in_value`
 	optResolvedTimestamps      = `resolved`
 	optUpdatedTimestamps       = `updated`
 
-	optEnvelopeKeyOnly envelopeType = `key_only`
-	optEnvelopeRow     envelopeType = `row`
-	optEnvelopeDiff    envelopeType = `diff`
+	optEnvelopeKeyOnly       envelopeType = `key_only`
+	optEnvelopeRow           envelopeType = `row`
+	optEnvelopeDeprecatedRow envelopeType = `deprecated_row`
+	optEnvelopeWrapped       envelopeType = `wrapped`
 
 	optFormatJSON formatType = `json`
 	optFormatAvro formatType = `experimental_avro`
 
-	sinkParamTopicPrefix      = `topic_prefix`
+	sinkParamCACert           = `ca_cert`
+	sinkParamFileSize         = `file_size`
 	sinkParamSchemaTopic      = `schema_topic`
+	sinkParamTLSEnabled       = `tls_enabled`
+	sinkParamTopicPrefix      = `topic_prefix`
 	sinkSchemeBuffer          = ``
 	sinkSchemeExperimentalSQL = `experimental-sql`
 	sinkSchemeKafka           = `kafka`
+	sinkParamSASLEnabled      = `sasl_enabled`
+	sinkParamSASLHandshake    = `sasl_handshake`
+	sinkParamSASLUser         = `sasl_user`
+	sinkParamSASLPassword     = `sasl_password`
 )
 
 var changefeedOptionExpectValues = map[string]sql.KVStringOptValidate{
@@ -68,6 +78,7 @@ var changefeedOptionExpectValues = map[string]sql.KVStringOptValidate{
 	optCursor:                  sql.KVStringOptRequireValue,
 	optEnvelope:                sql.KVStringOptRequireValue,
 	optFormat:                  sql.KVStringOptRequireValue,
+	optKeyInValue:              sql.KVStringOptRequireNoValue,
 	optResolvedTimestamps:      sql.KVStringOptAny,
 	optUpdatedTimestamps:       sql.KVStringOptRequireNoValue,
 }
@@ -75,15 +86,16 @@ var changefeedOptionExpectValues = map[string]sql.KVStringOptValidate{
 // changefeedPlanHook implements sql.PlanHookFn.
 func changefeedPlanHook(
 	_ context.Context, stmt tree.Statement, p sql.PlanHookState,
-) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, error) {
+) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, bool, error) {
 	changefeedStmt, ok := stmt.(*tree.CreateChangefeed)
 	if !ok {
-		return nil, nil, nil, nil
+		return nil, nil, nil, false, nil
 	}
 
 	var sinkURIFn func() (string, error)
 	var header sqlbase.ResultColumns
 	unspecifiedSink := changefeedStmt.SinkURI == nil
+	avoidBuffering := false
 	if unspecifiedSink {
 		// An unspecified sink triggers a fairly radical change in behavior.
 		// Instead of setting up a system.job to emit to a sink in the
@@ -98,11 +110,12 @@ func changefeedPlanHook(
 			{Name: "key", Typ: types.Bytes},
 			{Name: "value", Typ: types.Bytes},
 		}
+		avoidBuffering = true
 	} else {
 		var err error
 		sinkURIFn, err = p.TypeAsString(changefeedStmt.SinkURI, `CREATE CHANGEFEED`)
 		if err != nil {
-			return nil, nil, nil, err
+			return nil, nil, nil, false, err
 		}
 		header = sqlbase.ResultColumns{
 			{Name: "job_id", Typ: types.Int},
@@ -111,14 +124,14 @@ func changefeedPlanHook(
 
 	optsFn, err := p.TypeAsStringOpts(changefeedStmt.Options, changefeedOptionExpectValues)
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 
 	fn := func(ctx context.Context, _ []sql.PlanNode, resultsCh chan<- tree.Datums) error {
 		ctx, span := tracing.ChildSpan(ctx, stmt.StatementTag())
 		defer tracing.FinishSpan(span)
 
-		if !p.ExecCfg().Settings.Version.IsMinSupported(cluster.VersionCreateChangefeed) {
+		if !p.ExecCfg().Settings.Version.IsActive(cluster.VersionCreateChangefeed) {
 			return errors.Errorf(`CREATE CHANGEFEED requires all nodes to be upgraded to %s`,
 				cluster.VersionByKey(cluster.VersionCreateChangefeed),
 			)
@@ -143,12 +156,19 @@ func changefeedPlanHook(
 			return err
 		}
 
-		statementTime := p.ExecCfg().Clock.Now()
+		jobDescription, err := changefeedJobDescription(changefeedStmt, sinkURI, opts)
+		if err != nil {
+			return err
+		}
+
+		statementTime := hlc.Timestamp{
+			WallTime: p.ExtendedEvalContext().GetStmtTimestamp().UnixNano(),
+		}
 		var initialHighWater hlc.Timestamp
 		if cursor, ok := opts[optCursor]; ok {
 			asOf := tree.AsOfClause{Expr: tree.NewStrVal(cursor)}
 			var err error
-			if initialHighWater, err = p.EvalAsOfTimestamp(asOf, statementTime); err != nil {
+			if initialHighWater, err = p.EvalAsOfTimestamp(asOf); err != nil {
 				return err
 			}
 			statementTime = initialHighWater
@@ -202,12 +222,63 @@ func changefeedPlanHook(
 			},
 		}
 
-		if details.SinkURI == `` {
-			return distChangefeedFlow(ctx, p, 0 /* jobID */, details, progress, resultsCh)
+		// TODO(dan): In an attempt to present the most helpful error message to the
+		// user, the ordering requirements between all these usage validations have
+		// become extremely fragile and non-obvious.
+		//
+		// - `validateDetails` has to run first to fill in defaults for `envelope`
+		//   and `format` if the user didn't specify them.
+		// - Then `getEncoder` is run to return any configuration errors.
+		// - Then the changefeed is opted in to `optKeyInValue` for any cloud
+		//   storage sink. Kafka etc have a key and value field in each message but
+		//   cloud storage sinks don't have anywhere to put the key. So if the key
+		//   is not in the value, then for DELETEs there is no way to recover which
+		//   key was deleted. We could make the user explicitly pass this option for
+		//   every cloud storage sink and error if they don't, but that seems
+		//   user-hostile for insufficient reason. We can't do this any earlier,
+		//   because we might return errors about `key_in_value` being incompatible
+		//   which is confusing when the user didn't type that option.
+		// - Finally, we create a "canary" sink to test sink configuration and
+		//   connectivity. This has to go last because it is strange to return sink
+		//   connectivity errors before we've finished validating all the other
+		//   options. We should probably split sink configuration checking and sink
+		//   connectivity checking into separate methods.
+		//
+		// The only upside in all this nonsense is the tests are decent. I've tuned
+		// this particular order simply by rearranging stuff until the changefeedccl
+		// tests all pass.
+		parsedSink, err := url.Parse(sinkURI)
+		if err != nil {
+			return err
+		}
+		if details, err = validateDetails(details); err != nil {
+			return err
 		}
 
+		if _, err := getEncoder(details.Opts); err != nil {
+			return err
+		}
+		if isCloudStorageSink(parsedSink) {
+			details.Opts[optKeyInValue] = ``
+		}
+
+		// Feature telemetry
+		telemetrySink := parsedSink.Scheme
+		if telemetrySink == `` {
+			telemetrySink = `sinkless`
+		}
+		telemetry.Count(`changefeed.create.sink.` + telemetrySink)
+		telemetry.Count(`changefeed.create.format.` + details.Opts[optFormat])
+		telemetry.CountBucketed(`changefeed.create.num_tables`, int64(len(targets)))
+
+		if details.SinkURI == `` {
+			err := distChangefeedFlow(ctx, p, 0 /* jobID */, details, progress, resultsCh)
+			return MaybeStripRetryableErrorMarker(err)
+		}
+
+		settings := p.ExecCfg().Settings
 		if err := utilccl.CheckEnterpriseEnabled(
-			p.ExecCfg().Settings, p.ExecCfg().ClusterID(), p.ExecCfg().Organization(), "CHANGEFEED",
+			settings, p.ExecCfg().ClusterID(), p.ExecCfg().Organization(), "CHANGEFEED",
 		); err != nil {
 			return err
 		}
@@ -218,14 +289,10 @@ func changefeedPlanHook(
 		// the CREATE CHANGEFEED statement. To do this, we create a "canary" sink,
 		// which will be immediately closed, only to check for errors.
 		{
-			canarySink, err := getSink(sinkURI, opts, targets)
+			nodeID := p.ExtendedEvalContext().NodeID
+			canarySink, err := getSink(details.SinkURI, nodeID, details.Opts, details.Targets, settings)
 			if err != nil {
-				// In this context, we don't want to retry even retryable errors from the
-				// sync. Unwrap any retryable errors encountered.
-				if rErr, ok := err.(*retryableSinkError); ok {
-					return rErr.cause
-				}
-				return err
+				return MaybeStripRetryableErrorMarker(err)
 			}
 			if err := canarySink.Close(); err != nil {
 				return err
@@ -237,7 +304,7 @@ func changefeedPlanHook(
 		// hooked up to resultsCh to avoid a bunch of extra plumbing.
 		startedCh := make(chan tree.Datums)
 		job, errCh, err := p.ExecCfg().JobRegistry.StartJob(ctx, startedCh, jobs.Record{
-			Description: changefeedJobDescription(changefeedStmt, sinkURI, opts),
+			Description: jobDescription,
 			Username:    p.User(),
 			DescriptorIDs: func() (sqlDescIDs []sqlbase.ID) {
 				for _, desc := range targetDescs {
@@ -265,17 +332,19 @@ func changefeedPlanHook(
 		}
 		return nil
 	}
-	return fn, header, nil, nil
+	return fn, header, nil, avoidBuffering, nil
 }
 
 func changefeedJobDescription(
 	changefeed *tree.CreateChangefeed, sinkURI string, opts map[string]string,
-) string {
+) (string, error) {
+	cleanedSinkURI, err := storageccl.SanitizeExportStorageURI(sinkURI)
+	if err != nil {
+		return "", err
+	}
 	c := &tree.CreateChangefeed{
 		Targets: changefeed.Targets,
-		// If/when we start accepting export storage uris (or ones with
-		// secrets), we'll need to sanitize sinkURI.
-		SinkURI: tree.NewDString(sinkURI),
+		SinkURI: tree.NewDString(cleanedSinkURI),
 	}
 	for k, v := range opts {
 		opt := tree.KVOption{Key: tree.Name(k)}
@@ -285,7 +354,7 @@ func changefeedJobDescription(
 		c.Options = append(c.Options, opt)
 	}
 	sort.Slice(c.Options, func(i, j int) bool { return c.Options[i].Key < c.Options[j].Key })
-	return tree.AsStringWithFlags(c, tree.FmtAlwaysQualifyTableNames)
+	return tree.AsStringWithFlags(c, tree.FmtAlwaysQualifyTableNames), nil
 }
 
 func validateDetails(details jobspb.ChangefeedDetails) (jobspb.ChangefeedDetails, error) {
@@ -307,13 +376,12 @@ func validateDetails(details jobspb.ChangefeedDetails) (jobspb.ChangefeedDetails
 	}
 
 	switch envelopeType(details.Opts[optEnvelope]) {
-	case ``, optEnvelopeRow:
+	case optEnvelopeRow, optEnvelopeDeprecatedRow:
 		details.Opts[optEnvelope] = string(optEnvelopeRow)
 	case optEnvelopeKeyOnly:
 		details.Opts[optEnvelope] = string(optEnvelopeKeyOnly)
-	case optEnvelopeDiff:
-		return jobspb.ChangefeedDetails{}, errors.Errorf(
-			`%s=%s is not yet supported`, optEnvelope, optEnvelopeDiff)
+	case ``, optEnvelopeWrapped:
+		details.Opts[optEnvelope] = string(optEnvelopeWrapped)
 	default:
 		return jobspb.ChangefeedDetails{}, errors.Errorf(
 			`unknown %s: %s`, optEnvelope, details.Opts[optEnvelope])
@@ -384,12 +452,25 @@ func (b *changefeedResumer) Resume(
 	ctx context.Context, job *jobs.Job, planHookState interface{}, startedCh chan<- tree.Datums,
 ) error {
 	phs := planHookState.(sql.PlanHookState)
+	execCfg := phs.ExecCfg()
+	jobID := *job.ID()
 	details := job.Details().(jobspb.ChangefeedDetails)
 	progress := job.Progress()
 
-	// Errors encountered while emitting changes to the Sink may be transient; for
-	// example, a temporary network outage. When one of these errors occurs, we do
-	// not fail the job but rather restart the distSQL flow after a short backoff.
+	// TODO(dan): This is a workaround for not being able to set an initial
+	// progress high-water when creating a job (currently only the progress
+	// details can be set). I didn't want to pick off the refactor to get this
+	// fix in, but it'd be nice to remove this hack.
+	if _, ok := details.Opts[optCursor]; ok {
+		if h := progress.GetHighWater(); h == nil || *h == (hlc.Timestamp{}) {
+			progress.Progress = &jobspb.Progress_HighWater{HighWater: &details.StatementTime}
+		}
+	}
+
+	// We'd like to avoid failing a changefeed unnecessarily, so when an error
+	// bubbles up to this level, we'd like to "retry" the flow if possible. This
+	// could be because the sink is down or because a cockroach node has crashed
+	// or for many other reasons.
 	opts := retry.Options{
 		InitialBackoff: 5 * time.Millisecond,
 		Multiplier:     2,
@@ -397,43 +478,38 @@ func (b *changefeedResumer) Resume(
 	}
 	var err error
 	for r := retry.StartWithCtx(ctx, opts); r.Next(); {
-		// TODO(dan): This is a workaround for not being able to set an initial
-		// progress high-water when creating a job (currently only the progress
-		// details can be set). I didn't want to pick off the refactor to get this
-		// fix in, but it'd be nice to remove this hack.
-		if _, ok := details.Opts[optCursor]; ok {
-			if h := progress.GetHighWater(); h == nil || *h == (hlc.Timestamp{}) {
-				progress.Progress = &jobspb.Progress_HighWater{HighWater: &details.StatementTime}
-			}
+		if err = distChangefeedFlow(ctx, phs, jobID, details, progress, startedCh); err == nil {
+			return nil
+		}
+		if !IsRetryableError(err) {
+			log.Warningf(ctx, `CHANGEFEED job %d returning with error: %+v`, jobID, err)
+			return err
 		}
 
-		err = distChangefeedFlow(ctx, phs, *job.ID(), details, progress, startedCh)
-		if !isRetryableSinkError(err) && !isRetryableRPCError(err) {
-			break
+		log.Warningf(ctx, `CHANGEFEED job %d encountered retryable error: %v`, jobID, err)
+		if metrics, ok := execCfg.JobRegistry.MetricsStruct().Changefeed.(*Metrics); ok {
+			metrics.ErrorRetries.Inc(1)
 		}
-		log.Infof(ctx, `CHANGEFEED job %d encountered retryable error: %v`, *job.ID(), err)
 		// Re-load the job in order to update our progress object, which may have
 		// been updated by the changeFrontier processor since the flow started.
-		reloadedJob, phsErr := phs.ExecCfg().JobRegistry.LoadJob(ctx, *job.ID())
-		if phsErr != nil {
-			err = phsErr
-			break
+		reloadedJob, reloadErr := execCfg.JobRegistry.LoadJob(ctx, jobID)
+		if reloadErr != nil {
+			log.Warningf(ctx, `CHANGEFEED job %d could not reload job progress; `+
+				`continuing from last known high-water of %s: %v`,
+				jobID, progress.GetHighWater(), reloadErr)
+		} else {
+			progress = reloadedJob.Progress()
 		}
-		progress = reloadedJob.Progress()
+
 		// startedCh is normally used to signal back to the creator of the job that
 		// the job has started; however, in this case nothing will ever receive
 		// on the channel, causing the changefeed flow to block. Replace it with
 		// a dummy channel.
 		startedCh = make(chan tree.Datums, 1)
-		if metrics, ok := phs.ExecCfg().JobRegistry.MetricsStruct().Changefeed.(*Metrics); ok {
-			metrics.SinkErrorRetries.Inc(1)
-		}
-		continue
 	}
-	if err != nil {
-		log.Infof(ctx, `CHANGEFEED job %d returning with error: %v`, *job.ID(), err)
-	}
-	return err
+	// We only hit this if `r.Next()` returns false, which right now only happens
+	// on context cancellation.
+	return errors.Wrap(err, `ran out of retries`)
 }
 
 func (b *changefeedResumer) OnFailOrCancel(context.Context, *client.Txn, *jobs.Job) error { return nil }
@@ -449,14 +525,3 @@ func changefeedResumeHook(typ jobspb.Type, _ *cluster.Settings) jobs.Resumer {
 	}
 	return &changefeedResumer{}
 }
-
-// Retryable RPC Error represents a gRPC error which indicates a retryable
-// situation such as a connected node going down. In this case the DistSQL flow
-// should be retried.
-const retryableErrorStr = "rpc error|node unavailable"
-
-var retryableErrorRegex = regexp.MustCompile(retryableErrorStr)
-
-func isRetryableRPCError(err error) bool {
-	return retryableErrorRegex.MatchString(err.Error())
-}
