@@ -10,11 +10,15 @@ package changefeedccl
 
 import (
 	"context"
+	"crypto/tls"
+	"crypto/x509"
 	gosql "database/sql"
+	"encoding/base64"
 	"fmt"
 	"hash"
 	"hash/fnv"
 	"net/url"
+	"strconv"
 	"strings"
 	"sync"
 	"time"
@@ -22,11 +26,15 @@ import (
 	"github.com/Shopify/sarama"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
-	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/builtins"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/util/bufalloc"
+	"github.com/cockroachdb/cockroach/pkg/util/hlc"
+	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/cockroachdb/cockroach/pkg/util/log/logtags"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/pkg/errors"
@@ -36,12 +44,16 @@ import (
 type Sink interface {
 	// EmitRow enqueues a row message for asynchronous delivery on the sink. An
 	// error may be returned if a previously enqueued message has failed.
-	EmitRow(ctx context.Context, topic string, key, value []byte) error
+	EmitRow(
+		ctx context.Context,
+		table *sqlbase.TableDescriptor,
+		key, value []byte,
+		updated hlc.Timestamp,
+	) error
 	// EmitResolvedTimestamp enqueues a resolved timestamp message for
-	// asynchronous delivery on every partition of every topic that has been
-	// seen by EmitRow. The list of partitions used may be stale. An error may
-	// be returned if a previously enqueued message has failed.
-	EmitResolvedTimestamp(ctx context.Context, payload []byte) error
+	// asynchronous delivery on every topic that has been seen by EmitRow. An
+	// error may be returned if a previously enqueued message has failed.
+	EmitResolvedTimestamp(ctx context.Context, encoder Encoder, resolved hlc.Timestamp) error
 	// Flush blocks until every message enqueued by EmitRow and
 	// EmitResolvedTimestamp has been acknowledged by the sink. If an error is
 	// returned, no guarantees are given about which messages have been
@@ -52,7 +64,11 @@ type Sink interface {
 }
 
 func getSink(
-	sinkURI string, opts map[string]string, targets jobspb.ChangefeedTargets,
+	sinkURI string,
+	nodeID roachpb.NodeID,
+	opts map[string]string,
+	targets jobspb.ChangefeedTargets,
+	settings *cluster.Settings,
 ) (Sink, error) {
 	u, err := url.Parse(sinkURI)
 	if err != nil {
@@ -63,21 +79,100 @@ func getSink(
 	// Use a function here to delay creation of the sink until after we've done
 	// all the parameter verification.
 	var makeSink func() (Sink, error)
-	switch u.Scheme {
-	case sinkSchemeBuffer:
+	switch {
+	case u.Scheme == sinkSchemeBuffer:
 		makeSink = func() (Sink, error) { return &bufferSink{}, nil }
-	case sinkSchemeKafka:
-		kafkaTopicPrefix := q.Get(sinkParamTopicPrefix)
+	case u.Scheme == sinkSchemeKafka:
+		var cfg kafkaSinkConfig
+		cfg.kafkaTopicPrefix = q.Get(sinkParamTopicPrefix)
 		q.Del(sinkParamTopicPrefix)
-		schemaTopic := q.Get(sinkParamSchemaTopic)
-		q.Del(sinkParamSchemaTopic)
-		if schemaTopic != `` {
+		if schemaTopic := q.Get(sinkParamSchemaTopic); schemaTopic != `` {
 			return nil, errors.Errorf(`%s is not yet supported`, sinkParamSchemaTopic)
 		}
-		makeSink = func() (Sink, error) {
-			return getKafkaSink(kafkaTopicPrefix, u.Host, targets)
+		q.Del(sinkParamSchemaTopic)
+		if tlsBool := q.Get(sinkParamTLSEnabled); tlsBool != `` {
+			var err error
+			if cfg.tlsEnabled, err = strconv.ParseBool(tlsBool); err != nil {
+				return nil, errors.Errorf(`param %s must be a bool: %s`, sinkParamTLSEnabled, err)
+			}
 		}
-	case sinkSchemeExperimentalSQL:
+		q.Del(sinkParamTLSEnabled)
+		if caCertHex := q.Get(sinkParamCACert); caCertHex != `` {
+			// TODO(dan): There's a straightforward and unambiguous transformation
+			// between the base 64 encoding defined in RFC 4648 and the URL variant
+			// defined in the same RFC: simply replace all `+` with `-` and `/` with
+			// `_`. Consider always doing this for the user and accepting either
+			// variant.
+			if cfg.caCert, err = base64.StdEncoding.DecodeString(caCertHex); err != nil {
+				return nil, errors.Errorf(`param %s must be base 64 encoded: %s`, sinkParamCACert, err)
+			}
+		}
+		q.Del(sinkParamCACert)
+
+		saslParam := q.Get(sinkParamSASLEnabled)
+		q.Del(sinkParamSASLEnabled)
+		if saslParam != `` {
+			b, err := strconv.ParseBool(saslParam)
+			if err != nil {
+				return nil, errors.Wrapf(err, `param %s must be a bool:`, sinkParamSASLEnabled)
+			}
+			cfg.saslEnabled = b
+		}
+		handshakeParam := q.Get(sinkParamSASLHandshake)
+		q.Del(sinkParamSASLHandshake)
+		if handshakeParam == `` {
+			cfg.saslHandshake = true
+		} else {
+			if !cfg.saslEnabled {
+				return nil, errors.Errorf(`%s must be enabled to configure SASL handshake behavior`, sinkParamSASLEnabled)
+			}
+			b, err := strconv.ParseBool(handshakeParam)
+			if err != nil {
+				return nil, errors.Wrapf(err, `param %s must be a bool:`, sinkParamSASLHandshake)
+			}
+			cfg.saslHandshake = b
+		}
+		cfg.saslUser = q.Get(sinkParamSASLUser)
+		q.Del(sinkParamSASLUser)
+		cfg.saslPassword = q.Get(sinkParamSASLPassword)
+		q.Del(sinkParamSASLPassword)
+		if cfg.saslEnabled {
+			if cfg.saslUser == `` {
+				return nil, errors.Errorf(`%s must be provided when SASL is enabled`, sinkParamSASLUser)
+			}
+			if cfg.saslPassword == `` {
+				return nil, errors.Errorf(`%s must be provided when SASL is enabled`, sinkParamSASLPassword)
+			}
+		} else {
+			if cfg.saslUser != `` {
+				return nil, errors.Errorf(`%s must be enabled if a SASL user is provided`, sinkParamSASLEnabled)
+			}
+			if cfg.saslPassword != `` {
+				return nil, errors.Errorf(`%s must be enabled if a SASL password is provided`, sinkParamSASLEnabled)
+			}
+		}
+
+		makeSink = func() (Sink, error) {
+			return makeKafkaSink(cfg, u.Host, targets)
+		}
+	case isCloudStorageSink(u):
+		fileSizeParam := q.Get(sinkParamFileSize)
+		q.Del(sinkParamFileSize)
+		var fileSize int64 = 16 << 20 // 16MB
+		if fileSizeParam != `` {
+			if fileSize, err = humanizeutil.ParseBytes(fileSizeParam); err != nil {
+				return nil, errors.Wrapf(err, `parsing %s`, fileSizeParam)
+			}
+		}
+		u.Scheme = strings.TrimPrefix(u.Scheme, `experimental-`)
+		// Transfer "ownership" of validating all remaining query parameters to
+		// ExportStorage.
+		u.RawQuery = q.Encode()
+		q = url.Values{}
+		makeSink = func() (Sink, error) {
+			return makeCloudStorageSink(u.String(), nodeID, fileSize, settings, opts)
+		}
+	case u.Scheme == sinkSchemeExperimentalSQL:
 		// Swap the changefeed prefix for the sql connection one that sqlSink
 		// expects.
 		u.Scheme = `postgres`
@@ -107,22 +202,93 @@ func getSink(
 	return s, nil
 }
 
+// errorWrapperSink delegates to another sink and marks all returned errors as
+// retryable. During changefeed setup, we use the sink once without this to
+// verify configuration, but in the steady state, no sink error should be
+// terminal.
+type errorWrapperSink struct {
+	wrapped Sink
+}
+
+func (s errorWrapperSink) EmitRow(
+	ctx context.Context, table *sqlbase.TableDescriptor, key, value []byte, updated hlc.Timestamp,
+) error {
+	if err := s.wrapped.EmitRow(ctx, table, key, value, updated); err != nil {
+		return MarkRetryableError(err)
+	}
+	return nil
+}
+
+func (s errorWrapperSink) EmitResolvedTimestamp(
+	ctx context.Context, encoder Encoder, resolved hlc.Timestamp,
+) error {
+	if err := s.wrapped.EmitResolvedTimestamp(ctx, encoder, resolved); err != nil {
+		return MarkRetryableError(err)
+	}
+	return nil
+}
+
+func (s errorWrapperSink) Flush(ctx context.Context) error {
+	if err := s.wrapped.Flush(ctx); err != nil {
+		return MarkRetryableError(err)
+	}
+	return nil
+}
+
+func (s errorWrapperSink) Close() error {
+	if err := s.wrapped.Close(); err != nil {
+		return MarkRetryableError(err)
+	}
+	return nil
+}
+
+type kafkaLogAdapter struct {
+	ctx context.Context
+}
+
+var _ sarama.StdLogger = (*kafkaLogAdapter)(nil)
+
+func (l *kafkaLogAdapter) Print(v ...interface{}) {
+	log.InfoDepth(l.ctx, 1, v...)
+}
+func (l *kafkaLogAdapter) Printf(format string, v ...interface{}) {
+	log.InfofDepth(l.ctx, 1, format, v...)
+}
+func (l *kafkaLogAdapter) Println(v ...interface{}) {
+	log.InfoDepth(l.ctx, 1, v...)
+}
+
+func init() {
+	// We'd much prefer to make one of these per sink, so we can use the real
+	// context, but quite unfortunately, sarama only has a global logger hook.
+	ctx := context.Background()
+	ctx = logtags.AddTag(ctx, "kafka-producer", nil)
+	sarama.Logger = &kafkaLogAdapter{ctx: ctx}
+}
+
+type kafkaSinkConfig struct {
+	kafkaTopicPrefix string
+	tlsEnabled       bool
+	caCert           []byte
+	saslEnabled      bool
+	saslHandshake    bool
+	saslUser         string
+	saslPassword     string
+}
+
 // kafkaSink emits to Kafka asynchronously. It is not concurrency-safe; all
 // calls to Emit and Flush should be from the same goroutine.
 type kafkaSink struct {
-	// TODO(dan): This uses the shopify kafka producer library because the
-	// official confluent one depends on librdkafka and it didn't seem worth it
-	// to add a new c dep for the prototype. Revisit before 2.1 and check
-	// stability, performance, etc.
-	kafkaTopicPrefix string
-	client           sarama.Client
-	producer         sarama.AsyncProducer
-	topics           map[string]struct{}
+	cfg      kafkaSinkConfig
+	client   sarama.Client
+	producer sarama.AsyncProducer
+	topics   map[string]struct{}
 
 	lastMetadataRefresh time.Time
 
 	stopWorkerCh chan struct{}
 	worker       sync.WaitGroup
+	scratch      bufalloc.ByteAllocator
 
 	// Only synchronized between the client goroutine and the worker goroutine.
 	mu struct {
@@ -133,21 +299,41 @@ type kafkaSink struct {
 	}
 }
 
-func getKafkaSink(
-	kafkaTopicPrefix string, bootstrapServers string, targets jobspb.ChangefeedTargets,
+func makeKafkaSink(
+	cfg kafkaSinkConfig, bootstrapServers string, targets jobspb.ChangefeedTargets,
 ) (Sink, error) {
-	sink := &kafkaSink{
-		kafkaTopicPrefix: kafkaTopicPrefix,
-	}
+	sink := &kafkaSink{cfg: cfg}
 	sink.topics = make(map[string]struct{})
 	for _, t := range targets {
-		sink.topics[kafkaTopicPrefix+SQLNameToKafkaName(t.StatementTimeName)] = struct{}{}
+		sink.topics[cfg.kafkaTopicPrefix+SQLNameToKafkaName(t.StatementTimeName)] = struct{}{}
 	}
 
 	config := sarama.NewConfig()
+	config.ClientID = `CockroachDB`
 	config.Producer.Return.Successes = true
 	config.Producer.Partitioner = newChangefeedPartitioner
 
+	if cfg.caCert != nil {
+		if !cfg.tlsEnabled {
+			return nil, errors.Errorf(`%s requires %s=true`, sinkParamCACert, sinkParamTLSEnabled)
+		}
+		caCertPool := x509.NewCertPool()
+		caCertPool.AppendCertsFromPEM(cfg.caCert)
+		config.Net.TLS.Config = &tls.Config{
+			RootCAs: caCertPool,
+		}
+		config.Net.TLS.Enable = true
+	} else if cfg.tlsEnabled {
+		config.Net.TLS.Enable = true
+	}
+
+	if cfg.saslEnabled {
+		config.Net.SASL.Enable = true
+		config.Net.SASL.Handshake = cfg.saslHandshake
+		config.Net.SASL.User = cfg.saslUser
+		config.Net.SASL.Password = cfg.saslPassword
+	}
+
 	// When we emit messages to sarama, they're placed in a queue (as does any
 	// reasonable kafka producer client). When our sink's Flush is called, we
 	// have to wait for all buffered and inflight requests to be sent and then
@@ -179,16 +365,20 @@ func getKafkaSink(
 	// to test this one more before changing it.
 	config.Producer.Flush.MaxMessages = 1000
 
+	// config.Producer.Flush.Messages is set to 1 so we don't need this, but
+	// sarama prints scary things to the logs if we don't.
+	config.Producer.Flush.Frequency = time.Hour
+
 	var err error
 	sink.client, err = sarama.NewClient(strings.Split(bootstrapServers, `,`), config)
 	if err != nil {
 		err = errors.Wrapf(err, `connecting to kafka: %s`, bootstrapServers)
-		return nil, &retryableSinkError{cause: err}
+		return nil, err
 	}
 	sink.producer, err = sarama.NewAsyncProducerFromClient(sink.client)
 	if err != nil {
 		err = errors.Wrapf(err, `connecting to kafka: %s`, bootstrapServers)
-		return nil, &retryableSinkError{cause: err}
+		return nil, err
 	}
 
 	sink.start()
@@ -217,8 +407,10 @@ func (s *kafkaSink) Close() error {
 }
 
 // EmitRow implements the Sink interface.
-func (s *kafkaSink) EmitRow(ctx context.Context, tableName string, key, value []byte) error {
-	topic := s.kafkaTopicPrefix + SQLNameToKafkaName(tableName)
+func (s *kafkaSink) EmitRow(
+	ctx context.Context, table *sqlbase.TableDescriptor, key, value []byte, _ hlc.Timestamp,
+) error {
+	topic := s.cfg.kafkaTopicPrefix + SQLNameToKafkaName(table.Name)
 	if _, ok := s.topics[topic]; !ok {
 		return errors.Errorf(`cannot emit to undeclared topic: %s`, topic)
 	}
@@ -232,7 +424,9 @@ func (s *kafkaSink) EmitRow(ctx context.Context, tableName string, key, value []
 }
 
 // EmitResolvedTimestamp implements the Sink interface.
-func (s *kafkaSink) EmitResolvedTimestamp(ctx context.Context, payload []byte) error {
+func (s *kafkaSink) EmitResolvedTimestamp(
+	ctx context.Context, encoder Encoder, resolved hlc.Timestamp,
+) error {
 	// Periodically ping sarama to refresh its metadata. This means talking to
 	// zookeeper, so it shouldn't be done too often, but beyond that this
 	// constant was picked pretty arbitrarily.
@@ -253,7 +447,13 @@ func (s *kafkaSink) EmitResolvedTimestamp(ctx context.Context, payload []byte) e
 	}
 
 	for topic := range s.topics {
-		// sarama caches this, which is why we have to periodically refresh. the
+		payload, err := encoder.EncodeResolvedTimestamp(topic, resolved)
+		if err != nil {
+			return err
+		}
+		s.scratch, payload = s.scratch.Copy(payload, 0 /* extraCap */)
+
+		// sarama caches this, which is why we have to periodically refresh the
 		// metadata above. Staleness here does not impact correctness. Some new
 		// partitions will miss this resolved timestamp, but they'll eventually
 		// be picked up and get later ones.
@@ -291,9 +491,6 @@ func (s *kafkaSink) Flush(ctx context.Context) error {
 	s.mu.Unlock()
 
 	if immediateFlush {
-		if _, ok := flushErr.(*sarama.ProducerError); ok {
-			flushErr = &retryableSinkError{cause: flushErr}
-		}
 		return flushErr
 	}
 
@@ -308,9 +505,6 @@ func (s *kafkaSink) Flush(ctx context.Context) error {
 		flushErr := s.mu.flushErr
 		s.mu.flushErr = nil
 		s.mu.Unlock()
-		if _, ok := flushErr.(*sarama.ProducerError); ok {
-			flushErr = &retryableSinkError{cause: flushErr}
-		}
 		return flushErr
 	}
 }
@@ -415,7 +609,8 @@ type sqlSink struct {
 	topics    map[string]struct{}
 	hasher    hash.Hash32
 
-	rowBuf []interface{}
+	rowBuf  []interface{}
+	scratch bufalloc.ByteAllocator
 }
 
 func makeSQLSink(uri, tableName string, targets jobspb.ChangefeedTargets) (*sqlSink, error) {
@@ -446,7 +641,10 @@ func makeSQLSink(uri, tableName string, targets jobspb.ChangefeedTargets) (*sqlS
 }
 
 // EmitRow implements the Sink interface.
-func (s *sqlSink) EmitRow(ctx context.Context, topic string, key, value []byte) error {
+func (s *sqlSink) EmitRow(
+	ctx context.Context, table *sqlbase.TableDescriptor, key, value []byte, _ hlc.Timestamp,
+) error {
+	topic := table.Name
 	if _, ok := s.topics[topic]; !ok {
 		return errors.Errorf(`cannot emit to undeclared topic: %s`, topic)
 	}
@@ -466,9 +664,16 @@ func (s *sqlSink) EmitRow(ctx context.Context, topic string, key, value []byte)
 }
 
 // EmitResolvedTimestamp implements the Sink interface.
-func (s *sqlSink) EmitResolvedTimestamp(ctx context.Context, payload []byte) error {
+func (s *sqlSink) EmitResolvedTimestamp(
+	ctx context.Context, encoder Encoder, resolved hlc.Timestamp,
+) error {
 	var noKey, noValue []byte
 	for topic := range s.topics {
+		payload, err := encoder.EncodeResolvedTimestamp(topic, resolved)
+		if err != nil {
+			return err
+		}
+		s.scratch, payload = s.scratch.Copy(payload, 0 /* extraCap */)
 		for partition := int32(0); partition < sqlSinkNumPartitions; partition++ {
 			if err := s.emit(ctx, topic, partition, noKey, noValue, payload); err != nil {
 				return err
@@ -543,18 +748,22 @@ func (b *encDatumRowBuffer) Pop() sqlbase.EncDatumRow {
 }
 
 type bufferSink struct {
-	buf    encDatumRowBuffer
-	alloc  sqlbase.DatumAlloc
-	closed bool
+	buf     encDatumRowBuffer
+	alloc   sqlbase.DatumAlloc
+	scratch bufalloc.ByteAllocator
+	closed  bool
 }
 
 // EmitRow implements the Sink interface.
-func (s *bufferSink) EmitRow(_ context.Context, topic string, key, value []byte) error {
+func (s *bufferSink) EmitRow(
+	_ context.Context, table *sqlbase.TableDescriptor, key, value []byte, _ hlc.Timestamp,
+) error {
 	if s.closed {
 		return errors.New(`cannot EmitRow on a closed sink`)
 	}
+	topic := table.Name
 	s.buf.Push(sqlbase.EncDatumRow{
-		{Datum: tree.DNull},                              // resolved span
+		{Datum: tree.DNull}, // resolved span
 		{Datum: s.alloc.NewDString(tree.DString(topic))}, // topic
 		{Datum: s.alloc.NewDBytes(tree.DBytes(key))},     // key
 		{Datum: s.alloc.NewDBytes(tree.DBytes(value))},   //value
@@ -563,14 +772,22 @@ func (s *bufferSink) EmitRow(_ context.Context, topic string, key, value []byte)
 }
 
 // EmitResolvedTimestamp implements the Sink interface.
-func (s *bufferSink) EmitResolvedTimestamp(_ context.Context, payload []byte) error {
+func (s *bufferSink) EmitResolvedTimestamp(
+	_ context.Context, encoder Encoder, resolved hlc.Timestamp,
+) error {
 	if s.closed {
-		return errors.New(`cannot EmitRow on a closed sink`)
+		return errors.New(`cannot EmitResolvedTimestamp on a closed sink`)
 	}
+	var noTopic string
+	payload, err := encoder.EncodeResolvedTimestamp(noTopic, resolved)
+	if err != nil {
+		return err
+	}
+	s.scratch, payload = s.scratch.Copy(payload, 0 /* extraCap */)
 	s.buf.Push(sqlbase.EncDatumRow{
-		{Datum: tree.DNull},                              // resolved span
-		{Datum: tree.DNull},                              // topic
-		{Datum: tree.DNull},                              // key
+		{Datum: tree.DNull}, // resolved span
+		{Datum: tree.DNull}, // topic
+		{Datum: tree.DNull}, // key
 		{Datum: s.alloc.NewDBytes(tree.DBytes(payload))}, // value
 	})
 	return nil
@@ -586,46 +803,3 @@ func (s *bufferSink) Close() error {
 	s.closed = true
 	return nil
 }
-
-// causer matches the (unexported) interface used by Go to allow errors to wrap
-// their parent cause.
-type causer interface {
-	Cause() error
-}
-
-// String and regex used to match retryable sink errors when they have been
-// "flattened" into a pgerror.
-const retryableSinkErrorString = "retryable sink error"
-
-// retryableSinkError should be used by sinks to wrap any error which may
-// be retried.
-type retryableSinkError struct {
-	cause error
-}
-
-func (e retryableSinkError) Error() string {
-	return fmt.Sprintf(retryableSinkErrorString+": %s", e.cause.Error())
-}
-func (e retryableSinkError) Cause() error { return e.cause }
-
-// isRetryableSinkError returns true if the supplied error, or any of its parent
-// causes, is a retryableSinkError.
-func isRetryableSinkError(err error) bool {
-	for {
-		if _, ok := err.(*retryableSinkError); ok {
-			return true
-		}
-		// TODO(mrtracy): This pathway, which occurs when the retryable error is
-		// detected on a non-local node of the distsql flow, is only currently
-		// being tested with a roachtest, which is expensive. See if it can be
-		// tested via a unit test,
-		if _, ok := err.(*pgerror.Error); ok {
-			return strings.Contains(err.Error(), retryableSinkErrorString)
-		}
-		if e, ok := err.(causer); ok {
-			err = e.Cause()
-			continue
-		}
-		return false
-	}
-}
