@@ -8,6 +8,4 @@
 
 package cliccl
 
-import (
-	_ "github.com/cockroachdb/cockroach/pkg/ccl/workloadccl/cliccl" // init hooks
-)
+import _ "github.com/cockroachdb/cockroach/pkg/ccl/workloadccl/cliccl" // init hooks
