@@ -18,12 +18,9 @@ import (
 	"strconv"
 	"strings"
 
-	"github.com/pkg/errors"
-	"google.golang.org/grpc/codes"
-	"google.golang.org/grpc/status"
-
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/ccl/backupccl"
+	"github.com/cockroachdb/cockroach/pkg/ccl/gossipccl"
 	"github.com/cockroachdb/cockroach/pkg/ccl/storageccl"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/jobs"
@@ -35,15 +32,21 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql"
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/sql/stats"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
+	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/status"
 )
 
 const (
@@ -63,6 +66,8 @@ const (
 	importOptionOversample = "oversample"
 	importOptionSkipFKs    = "skip_foreign_keys"
 
+	importOptionDirectIngest = "experimental_direct_ingestion"
+
 	pgCopyDelimiter = "delimiter"
 	pgCopyNull      = "nullif"
 
@@ -87,6 +92,8 @@ var importOptionExpectValues = map[string]sql.KVStringOptValidate{
 
 	importOptionSkipFKs: sql.KVStringOptRequireNoValue,
 
+	importOptionDirectIngest: sql.KVStringOptRequireNoValue,
+
 	pgMaxRowSize: sql.KVStringOptRequireValue,
 }
 
@@ -119,7 +126,7 @@ func readCreateTableFromStore(
 	if err != nil {
 		return nil, err
 	}
-	create, ok := stmt.(*tree.CreateTable)
+	create, ok := stmt.AST.(*tree.CreateTable)
 	if !ok {
 		return nil, errors.New("expected CREATE TABLE statement in table file")
 	}
@@ -135,7 +142,7 @@ type fkHandler struct {
 // NoFKs is used by formats that do not support FKs.
 var NoFKs = fkHandler{resolver: make(fkResolver)}
 
-// MakeSimpleTableDescriptor creates a TableDescriptor from a CreateTable parse
+// MakeSimpleTableDescriptor creates a MutableTableDescriptor from a CreateTable parse
 // node without the full machinery. Many parts of the syntax are unsupported
 // (see the implementation and TestMakeSimpleTableDescriptorErrors for details),
 // but this is enough for our csv IMPORT and for some unit tests.
@@ -147,11 +154,10 @@ func MakeSimpleTableDescriptor(
 	ctx context.Context,
 	st *cluster.Settings,
 	create *tree.CreateTable,
-	parentID,
-	tableID sqlbase.ID,
+	parentID, tableID sqlbase.ID,
 	fks fkHandler,
 	walltime int64,
-) (*sqlbase.TableDescriptor, error) {
+) (*sqlbase.MutableTableDescriptor, error) {
 	create.HoistConstraints()
 	if create.IfNotExists {
 		return nil, pgerror.Unimplemented("import.if-no-exists", "unsupported IF NOT EXISTS")
@@ -163,11 +169,6 @@ func MakeSimpleTableDescriptor(
 		return nil, pgerror.Unimplemented("import.create-as", "CREATE AS not supported")
 	}
 
-	tableName, err := create.Table.Normalize()
-	if err != nil {
-		return nil, err
-	}
-
 	filteredDefs := create.Defs[:0]
 	for i := range create.Defs {
 		switch def := create.Defs[i].(type) {
@@ -181,7 +182,7 @@ func MakeSimpleTableDescriptor(
 				return nil, pgerror.Unimplemented("import.computed", "computed columns not supported: %s", tree.AsString(def))
 			}
 
-			if err := sql.SimplifySerialInColumnDefWithRowID(ctx, def, tableName); err != nil {
+			if err := sql.SimplifySerialInColumnDefWithRowID(ctx, def, &create.Table); err != nil {
 				return nil, err
 			}
 
@@ -193,12 +194,7 @@ func MakeSimpleTableDescriptor(
 				continue
 			}
 			// Strip the schema/db prefix.
-			refTable, err := def.Table.Normalize()
-			if err != nil {
-				return nil, err
-			}
-			*refTable = tree.MakeUnqualifiedTableName(refTable.TableName)
-			def.Table.TableNameReference = refTable
+			def.Table = tree.MakeUnqualifiedTableName(def.Table.TableName)
 
 		default:
 			return nil, pgerror.Unimplemented(fmt.Sprintf("import.%T", def), "unsupported table definition: %s", tree.AsString(def))
@@ -213,7 +209,7 @@ func MakeSimpleTableDescriptor(
 		Context:  ctx,
 		Sequence: &importSequenceOperators{},
 	}
-	affected := make(map[sqlbase.ID]*sqlbase.TableDescriptor)
+	affected := make(map[sqlbase.ID]*sqlbase.MutableTableDescriptor)
 
 	tableDesc, err := sql.MakeTableDesc(
 		ctx,
@@ -232,7 +228,7 @@ func MakeSimpleTableDescriptor(
 	if err != nil {
 		return nil, err
 	}
-	if err := fixDescriptorFKState(&tableDesc); err != nil {
+	if err := fixDescriptorFKState(tableDesc.TableDesc()); err != nil {
 		return nil, err
 	}
 
@@ -302,7 +298,7 @@ func (so *importSequenceOperators) SetSequenceValue(
 	return errSequenceOperators
 }
 
-type fkResolver map[string]*sqlbase.TableDescriptor
+type fkResolver map[string]*sqlbase.MutableTableDescriptor
 
 var _ sql.SchemaResolver = fkResolver{}
 
@@ -327,18 +323,18 @@ func (r fkResolver) CurrentSearchPath() sessiondata.SearchPath {
 }
 
 // Implements the sql.SchemaResolver interface.
-func (r fkResolver) CommonLookupFlags(ctx context.Context, required bool) sql.CommonLookupFlags {
+func (r fkResolver) CommonLookupFlags(required bool) sql.CommonLookupFlags {
 	return sql.CommonLookupFlags{}
 }
 
 // Implements the sql.SchemaResolver interface.
-func (r fkResolver) ObjectLookupFlags(ctx context.Context, required bool) sql.ObjectLookupFlags {
+func (r fkResolver) ObjectLookupFlags(required bool, requireMutable bool) sql.ObjectLookupFlags {
 	return sql.ObjectLookupFlags{}
 }
 
 // Implements the tree.TableNameExistingResolver interface.
 func (r fkResolver) LookupObject(
-	ctx context.Context, dbName, scName, obName string,
+	ctx context.Context, requireMutable bool, dbName, scName, obName string,
 ) (found bool, objMeta tree.NameResolutionResult, err error) {
 	if scName != "" {
 		obName = strings.TrimPrefix(obName, scName+".")
@@ -362,6 +358,11 @@ func (r fkResolver) LookupSchema(
 	return false, nil, errSchemaResolver
 }
 
+// Implements the sql.SchemaResolver interface.
+func (r fkResolver) LookupTableByID(ctx context.Context, id sqlbase.ID) (row.TableEntry, error) {
+	return row.TableEntry{}, errSchemaResolver
+}
+
 const csvDatabaseName = "csv"
 
 func finalizeCSVBackup(
@@ -437,28 +438,28 @@ func importJobDescription(
 // importPlanHook implements sql.PlanHookFn.
 func importPlanHook(
 	_ context.Context, stmt tree.Statement, p sql.PlanHookState,
-) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, error) {
+) (sql.PlanHookRowFn, sqlbase.ResultColumns, []sql.PlanNode, bool, error) {
 	importStmt, ok := stmt.(*tree.Import)
 	if !ok {
-		return nil, nil, nil, nil
+		return nil, nil, nil, false, nil
 	}
 
 	filesFn, err := p.TypeAsStringArray(importStmt.Files, "IMPORT")
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 
 	var createFileFn func() (string, error)
 	if !importStmt.Bundle && importStmt.CreateDefs == nil {
 		createFileFn, err = p.TypeAsString(importStmt.CreateFile, "IMPORT")
 		if err != nil {
-			return nil, nil, nil, err
+			return nil, nil, nil, false, err
 		}
 	}
 
 	optsFn, err := p.TypeAsStringOpts(importStmt.Options, importOptionExpectValues)
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, nil, false, err
 	}
 
 	fn := func(ctx context.Context, _ []sql.PlanNode, resultsCh chan<- tree.Datums) error {
@@ -486,18 +487,7 @@ func importPlanHook(
 			return err
 		}
 
-		var table *tree.TableName
-		if importStmt.Table.TableNameReference != nil {
-			// Normalize must be called regardles of whether there is a
-			// transform because it prepares a TableName with the right
-			// structure and stores it back into the statement AST, which we
-			// need later when computing the job title.
-			table, err = importStmt.Table.Normalize()
-			if err != nil {
-				return errors.Wrap(err, "normalize create table")
-			}
-		}
-
+		table := importStmt.Table
 		transform := opts[importOptionTransform]
 
 		var parentID sqlbase.ID
@@ -563,7 +553,7 @@ func importPlanHook(
 				}
 				// We need to handle the case where the user wants to skip records and the node
 				// interpreting the statement might be newer than other nodes in the cluster.
-				if !p.ExecCfg().Settings.Version.IsMinSupported(cluster.VersionImportSkipRecords) {
+				if !p.ExecCfg().Settings.Version.IsActive(cluster.VersionImportSkipRecords) {
 					return errors.Errorf("Using non-CSV import format requires all nodes to be upgraded to %s",
 						cluster.VersionByKey(cluster.VersionImportSkipRecords))
 				}
@@ -661,7 +651,7 @@ func importPlanHook(
 		}
 
 		if format.Format != roachpb.IOFileFormat_CSV {
-			if !p.ExecCfg().Settings.Version.IsMinSupported(cluster.VersionImportFormats) {
+			if !p.ExecCfg().Settings.Version.IsActive(cluster.VersionImportFormats) {
 				return errors.Errorf("Using %s requires all nodes to be upgraded to %s",
 					csvSkip, cluster.VersionByKey(cluster.VersionImportFormats))
 			}
@@ -706,6 +696,17 @@ func importPlanHook(
 			}
 		}
 
+		_, ingestDirectly := opts[importOptionDirectIngest]
+		if ingestDirectly {
+			if !p.ExecCfg().Settings.Version.IsActive(cluster.VersionDirectImport) {
+				return errors.Errorf("Using %q requires all nodes to be upgraded to %s",
+					importOptionDirectIngest, cluster.VersionByKey(cluster.VersionDirectImport))
+			}
+			if transform != "" {
+				return errors.Errorf("cannot use %q and %q options together", importOptionDirectIngest, importOptionTransform)
+			}
+		}
+
 		var tableDescs []*sqlbase.TableDescriptor
 		var jobDesc string
 		var names []string
@@ -760,7 +761,10 @@ func importPlanHook(
 			}
 			var create *tree.CreateTable
 			if importStmt.CreateDefs != nil {
-				create = &tree.CreateTable{Table: importStmt.Table, Defs: importStmt.CreateDefs}
+				create = &tree.CreateTable{
+					Table: *importStmt.Table,
+					Defs:  importStmt.CreateDefs,
+				}
 			} else {
 				filename, err := createFileFn()
 				if err != nil {
@@ -771,10 +775,11 @@ func importPlanHook(
 					return err
 				}
 
-				if parsed, err := create.Table.Normalize(); err != nil {
-					return errors.Wrap(err, "normalize create table")
-				} else if table.TableName != parsed.TableName {
-					return errors.Errorf("importing table %s, but file specifies a schema for table %s", table.TableName, parsed.TableName)
+				if table.TableName != create.Table.TableName {
+					return errors.Errorf(
+						"importing table %s, but file specifies a schema for table %s",
+						table.TableName, create.Table.TableName,
+					)
 				}
 			}
 
@@ -783,7 +788,7 @@ func importPlanHook(
 			if err != nil {
 				return err
 			}
-			tableDescs = []*sqlbase.TableDescriptor{tbl}
+			tableDescs = []*sqlbase.TableDescriptor{tbl.TableDesc()}
 			descStr, err := importJobDescription(importStmt, create.Defs, files, opts)
 			if err != nil {
 				return err
@@ -846,15 +851,16 @@ func importPlanHook(
 			Description: jobDesc,
 			Username:    p.User(),
 			Details: jobspb.ImportDetails{
-				URIs:       files,
-				Format:     format,
-				ParentID:   parentID,
-				Tables:     tableDetails,
-				BackupPath: transform,
-				SSTSize:    sstSize,
-				Oversample: oversample,
-				Walltime:   walltime,
-				SkipFKs:    skipFKs,
+				URIs:           files,
+				Format:         format,
+				ParentID:       parentID,
+				Tables:         tableDetails,
+				BackupPath:     transform,
+				SSTSize:        sstSize,
+				Oversample:     oversample,
+				Walltime:       walltime,
+				SkipFKs:        skipFKs,
+				IngestDirectly: ingestDirectly,
 			},
 			Progress: jobspb.ImportProgress{},
 		})
@@ -863,7 +869,7 @@ func importPlanHook(
 		}
 		return <-errCh
 	}
-	return fn, backupccl.RestoreHeader, nil, nil
+	return fn, backupccl.RestoreHeader, nil, false, nil
 }
 
 func doDistributedCSVTransform(
@@ -878,7 +884,13 @@ func doDistributedCSVTransform(
 	walltime int64,
 	sstSize int64,
 	oversample int64,
+	ingestDirectly bool,
 ) (roachpb.BulkOpSummary, error) {
+	if ingestDirectly {
+		return sql.DistIngest(ctx, p, job, tables, files, format, walltime)
+		// TODO(dt): check for errors in job records as is done below.
+	}
+
 	evalCtx := p.ExtendedEvalContext()
 
 	ci := sqlbase.ColTypeInfoFromColTypes([]sqlbase.ColumnType{
@@ -888,7 +900,7 @@ func doDistributedCSVTransform(
 		{SemanticType: sqlbase.ColumnType_BYTES},
 		{SemanticType: sqlbase.ColumnType_BYTES},
 	})
-	rows := sqlbase.NewRowContainer(evalCtx.Mon.MakeBoundAccount(), ci, 0)
+	rows := rowcontainer.NewRowContainer(evalCtx.Mon.MakeBoundAccount(), ci, 0)
 	defer func() {
 		if rows != nil {
 			rows.Close(ctx)
@@ -1005,8 +1017,9 @@ func doDistributedCSVTransform(
 }
 
 type importResumer struct {
-	settings *cluster.Settings
-	res      roachpb.BulkOpSummary
+	settings       *cluster.Settings
+	res            roachpb.BulkOpSummary
+	statsRefresher *stats.Refresher
 }
 
 func (r *importResumer) Resume(
@@ -1024,6 +1037,7 @@ func (r *importResumer) Resume(
 	sstSize := details.SSTSize
 	format := details.Format
 	oversample := details.Oversample
+	ingestDirectly := details.IngestDirectly
 
 	if sstSize == 0 {
 		// The distributed importer will correctly chunk up large ranges into
@@ -1050,13 +1064,26 @@ func (r *importResumer) Resume(
 		}
 	}
 
+	{
+		// Disable merging for the table IDs being imported into. We don't want the
+		// merge queue undoing the splits performed during IMPORT.
+		tableIDs := make([]uint32, 0, len(tables))
+		for _, t := range tables {
+			tableIDs = append(tableIDs, uint32(t.ID))
+		}
+		disableCtx, cancel := context.WithCancel(ctx)
+		defer cancel()
+		gossipccl.DisableMerges(disableCtx, p.ExecCfg().Gossip, tableIDs)
+	}
+
 	res, err := doDistributedCSVTransform(
-		ctx, job, files, p, parentID, tables, transform, format, walltime, sstSize, oversample,
+		ctx, job, files, p, parentID, tables, transform, format, walltime, sstSize, oversample, ingestDirectly,
 	)
 	if err != nil {
 		return err
 	}
 	r.res = res
+	r.statsRefresher = p.ExecCfg().StatsRefresher
 	return nil
 }
 
@@ -1121,6 +1148,13 @@ func (r *importResumer) OnSuccess(ctx context.Context, txn *client.Txn, job *job
 		return errors.Wrapf(err, "creating tables")
 	}
 
+	// Initiate a run of CREATE STATISTICS. We don't know the actual number of
+	// rows affected per table, so we use a large number because we want to make
+	// sure that stats always get created/refreshed here.
+	for i := range toWrite {
+		r.statsRefresher.NotifyMutation(toWrite[i].ID, math.MaxInt32 /* rowsAffected */)
+	}
+
 	return nil
 }
 
