@@ -22,18 +22,15 @@ import (
 	"path/filepath"
 	"runtime"
 	"strings"
+	"sync"
 	"testing"
 	"time"
 
-	"github.com/cockroachdb/cockroach/pkg/keys"
-	"github.com/cockroachdb/cockroach/pkg/roachpb"
-	"github.com/cockroachdb/cockroach/pkg/util/hlc"
-	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
-	"golang.org/x/sync/errgroup"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/jobs"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
+	"github.com/cockroachdb/cockroach/pkg/keys"
+	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql"
@@ -41,6 +38,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/builtins"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/sql/stats"
 	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/jobutils"
@@ -48,8 +46,11 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/testcluster"
 	"github.com/cockroachdb/cockroach/pkg/util"
+	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
+	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/pkg/errors"
+	"golang.org/x/sync/errgroup"
 )
 
 func TestImportData(t *testing.T) {
@@ -72,8 +73,8 @@ func TestImportData(t *testing.T) {
 		{
 			name: "duplicate unique index key",
 			create: `
-				a int primary key,
-				i int,
+				a int8 primary key,
+				i int8,
 				unique index idx_f (i)
 			`,
 			typ: "CSV",
@@ -87,7 +88,7 @@ func TestImportData(t *testing.T) {
 		{
 			name: "duplicate PK",
 			create: `
-				i int primary key
+				i int8 primary key
 			`,
 			typ: "CSV",
 			data: `1
@@ -114,7 +115,7 @@ d
 		{
 			name: "duplicate PK at sst boundary",
 			create: `
-				i int primary key,
+				i int8 primary key,
 				s string
 			`,
 			with: `WITH sstsize = '10B'`,
@@ -126,10 +127,10 @@ d
 		{
 			name: "verify no splits mid row",
 			create: `
-				i int primary key,
+				i int8 primary key,
 				s string,
-				b int,
-				c int,
+				b int8,
+				c int8,
 				index (s),
 				index (i, s),
 				family (i, b),
@@ -168,7 +169,7 @@ d
 		},
 		{
 			name:   "oversample",
-			create: `i int`,
+			create: `i int8`,
 			with:   `WITH oversample = '100'`,
 			typ:    "CSV",
 			data:   "1",
@@ -183,8 +184,8 @@ d
 			},
 		},
 		{
-			name:   "CR in int, 2 cols",
-			create: `a int, b int`,
+			name:   "CR in int8, 2 cols",
+			create: `a int8, b int8`,
 			typ:    "CSV",
 			data:   "1,2\r\n3,4\n5,6",
 			query: map[string][][]string{
@@ -192,8 +193,8 @@ d
 			},
 		},
 		{
-			name:   "CR in int, 1 col",
-			create: `a int`,
+			name:   "CR in int8, 1 col",
+			create: `a int8`,
 			typ:    "CSV",
 			data:   "1\r\n3\n5",
 			query: map[string][][]string{
@@ -213,14 +214,14 @@ d
 		// MySQL OUTFILE
 		{
 			name:   "unexpected number of columns",
-			create: `i int`,
+			create: `i int8`,
 			typ:    "MYSQLOUTFILE",
 			data:   "1\t2",
 			err:    "row 1: too many columns, expected 1",
 		},
 		{
 			name:   "unmatched field enclosure",
-			create: `i int`,
+			create: `i int8`,
 			with:   `WITH fields_enclosed_by = '"'`,
 			typ:    "MYSQLOUTFILE",
 			data:   "\"1",
@@ -228,7 +229,7 @@ d
 		},
 		{
 			name:   "unmatched literal",
-			create: `i int`,
+			create: `i int8`,
 			with:   `WITH fields_escaped_by = '\'`,
 			typ:    "MYSQLOUTFILE",
 			data:   `\`,
@@ -315,7 +316,7 @@ d
 		},
 		{
 			name:   "normal",
-			create: `i int, s string`,
+			create: `i int8, s string`,
 			typ:    "PGCOPY",
 			data:   "1\tSTR\n2\t\\N\n\\N\t\\t",
 			query: map[string][][]string{
@@ -324,7 +325,7 @@ d
 		},
 		{
 			name:   "comma delim",
-			create: `i int, s string`,
+			create: `i int8, s string`,
 			typ:    "PGCOPY",
 			with:   `WITH delimiter = ','`,
 			data:   "1,STR\n2,\\N\n\\N,\\,",
@@ -334,14 +335,14 @@ d
 		},
 		{
 			name:   "size out of range",
-			create: `i int`,
+			create: `i int8`,
 			typ:    "PGCOPY",
 			with:   `WITH max_row_size = '10GB'`,
 			err:    "max_row_size out of range",
 		},
 		{
 			name:   "line too long",
-			create: `i int`,
+			create: `i int8`,
 			typ:    "PGCOPY",
 			data:   "123456",
 			with:   `WITH max_row_size = '5B'`,
@@ -350,14 +351,14 @@ d
 		{
 			name:   "not enough values",
 			typ:    "PGCOPY",
-			create: "a INT, b INT",
+			create: "a INT8, b INT8",
 			data:   `1`,
 			err:    "expected 2 values, got 1",
 		},
 		{
 			name:   "too many values",
 			typ:    "PGCOPY",
-			create: "a INT, b INT",
+			create: "a INT8, b INT8",
 			data:   "1\t2\t3",
 			err:    "expected 2 values, got 3",
 		},
@@ -367,7 +368,7 @@ d
 			name: "mismatch cols",
 			typ:  "PGDUMP",
 			data: `
-				CREATE TABLE t (i int);
+				CREATE TABLE t (i int8);
 				COPY t (s) FROM stdin;
 				0
 				\.
@@ -378,7 +379,7 @@ d
 			name: "missing COPY done",
 			typ:  "PGDUMP",
 			data: `
-				CREATE TABLE t (i int);
+				CREATE TABLE t (i int8);
 				COPY t (i) FROM stdin;
 0
 `,
@@ -388,7 +389,7 @@ d
 			name: "semicolons and comments",
 			typ:  "PGDUMP",
 			data: `
-				CREATE TABLE t (i int);
+				CREATE TABLE t (i int8);
 				;;;
 				-- nothing ;
 				;
@@ -407,7 +408,7 @@ d
 		{
 			name: "line too long",
 			typ:  "PGDUMP",
-			data: "CREATE TABLE t (i INT);",
+			data: "CREATE TABLE t (i INT8);",
 			with: `WITH max_row_size = '5B'`,
 			err:  "line too long",
 		},
@@ -415,7 +416,7 @@ d
 			name: "not enough values",
 			typ:  "PGDUMP",
 			data: `
-CREATE TABLE t (a INT, b INT);
+CREATE TABLE t (a INT8, b INT8);
 
 COPY t (a, b) FROM stdin;
 1
@@ -427,7 +428,7 @@ COPY t (a, b) FROM stdin;
 			name: "too many values",
 			typ:  "PGDUMP",
 			data: `
-CREATE TABLE t (a INT, b INT);
+CREATE TABLE t (a INT8, b INT8);
 
 COPY t (a, b) FROM stdin;
 1	2	3
@@ -439,7 +440,7 @@ COPY t (a, b) FROM stdin;
 			name: "too many cols",
 			typ:  "PGDUMP",
 			data: `
-CREATE TABLE t (a INT, b INT);
+CREATE TABLE t (a INT8, b INT8);
 
 COPY t (a, b, c) FROM stdin;
 1	2	3
@@ -493,8 +494,8 @@ COPY t (a, b, c) FROM stdin;
 				ORDER BY descriptor_name
 				`: {{
 					`CREATE TABLE a (
-	i INT NOT NULL,
-	k INT NULL,
+	i INT8 NOT NULL,
+	k INT8 NULL,
 	CONSTRAINT a_pkey PRIMARY KEY (i ASC),
 	CONSTRAINT a_k_fkey FOREIGN KEY (k) REFERENCES a (i),
 	INDEX a_auto_index_a_k_fkey (k ASC),
@@ -502,7 +503,7 @@ COPY t (a, b, c) FROM stdin;
 	FAMILY "primary" (i, k)
 )`}, {
 					`CREATE TABLE b (
-	j INT NOT NULL,
+	j INT8 NOT NULL,
 	CONSTRAINT b_pkey PRIMARY KEY (j ASC),
 	CONSTRAINT b_j_fkey FOREIGN KEY (j) REFERENCES a (i),
 	FAMILY "primary" (j)
@@ -551,7 +552,7 @@ COPY t (a, b, c) FROM stdin;
 			name: "sequence",
 			typ:  "PGDUMP",
 			data: `
-					CREATE TABLE t (a INT);
+					CREATE TABLE t (a INT8);
 					CREATE SEQUENCE public.i_seq
 						START WITH 1
 						INCREMENT BY 1
@@ -570,7 +571,7 @@ COPY t (a, b, c) FROM stdin;
 		{
 			name: "non-public schema",
 			typ:  "PGDUMP",
-			data: "create table s.t (i INT)",
+			data: "create table s.t (i INT8)",
 			err:  `non-public schemas unsupported: s`,
 		},
 		{
@@ -597,7 +598,7 @@ COPY t (a, b, c) FROM stdin;
 				$_$;
 				ALTER FUNCTION public.isnumeric(text) OWNER TO roland;
 
-				CREATE TABLE t (i INT);
+				CREATE TABLE t (i INT8);
 			`,
 			query: map[string][][]string{
 				`SHOW TABLES`: {{"t"}},
@@ -624,7 +625,7 @@ COPY t (a, b, c) FROM stdin;
 		},
 		{
 			name:   "sequences",
-			create: `i int default nextval('s')`,
+			create: `i int8 default nextval('s')`,
 			typ:    "CSV",
 			err:    `"s" not found`,
 		},
@@ -640,30 +641,41 @@ COPY t (a, b, c) FROM stdin;
 
 	// Create and drop a table to make sure a descriptor ID gets used to verify
 	// ID rewrites happen correctly. Useful when running just a single test.
-	sqlDB.Exec(t, `CREATE TABLE blah (i int)`)
+	sqlDB.Exec(t, `CREATE TABLE blah (i int8)`)
 	sqlDB.Exec(t, `DROP TABLE blah`)
 
-	for i, tc := range tests {
-		t.Run(fmt.Sprintf("%s: %s", tc.typ, tc.name), func(t *testing.T) {
-			dbName := fmt.Sprintf("d%d", i)
-			sqlDB.Exec(t, fmt.Sprintf(`CREATE DATABASE %s; USE %[1]s`, dbName))
-			defer sqlDB.Exec(t, fmt.Sprintf(`DROP DATABASE %s`, dbName))
-			var q string
-			if tc.create != "" {
-				q = fmt.Sprintf(`IMPORT TABLE t (%s) %s DATA ($1) %s`, tc.create, tc.typ, tc.with)
-			} else {
-				q = fmt.Sprintf(`IMPORT %s ($1) %s`, tc.typ, tc.with)
+	for _, direct := range []bool{false, true} {
+		// this test is big and slow as is, so we can't afford to double it in race.
+		if util.RaceEnabled && direct {
+			continue
+		}
+
+		for i, tc := range tests {
+			if direct {
+				if tc.with == "" {
+					tc.with = "WITH experimental_direct_ingestion"
+				} else {
+					tc.with += ", experimental_direct_ingestion"
+				}
 			}
-			t.Log(q)
-			dataString = tc.data
-			_, err := db.Exec(q, srv.URL)
-			if !testutils.IsError(err, tc.err) {
-				t.Fatalf("unexpected: %v", err)
-			}
-			for query, res := range tc.query {
-				sqlDB.CheckQueryResults(t, query, res)
-			}
-		})
+			t.Run(fmt.Sprintf("%s: %s direct=%v", tc.typ, tc.name, direct), func(t *testing.T) {
+				dbName := fmt.Sprintf("d%d", i)
+				sqlDB.Exec(t, fmt.Sprintf(`CREATE DATABASE %s; USE %[1]s`, dbName))
+				defer sqlDB.Exec(t, fmt.Sprintf(`DROP DATABASE %s`, dbName))
+				var q string
+				if tc.create != "" {
+					q = fmt.Sprintf(`IMPORT TABLE t (%s) %s DATA ($1) %s`, tc.create, tc.typ, tc.with)
+				} else {
+					q = fmt.Sprintf(`IMPORT %s ($1) %s`, tc.typ, tc.with)
+				}
+				t.Log(q)
+				dataString = tc.data
+				sqlDB.ExpectErr(t, tc.err, q, srv.URL)
+				for query, res := range tc.query {
+					sqlDB.CheckQueryResults(t, query, res)
+				}
+			})
+		}
 	}
 
 	t.Run("mysqlout multiple", func(t *testing.T) {
@@ -682,8 +694,8 @@ const (
 )`
 	testPgdumpCreateWeather = `CREATE TABLE weather (
 	city VARCHAR(80) NULL,
-	temp_lo INT NULL,
-	temp_hi INT NULL,
+	temp_lo INT8 NULL,
+	temp_hi INT8 NULL,
 	prcp FLOAT4 NULL,
 	date DATE NULL,
 	CONSTRAINT weather_city_fkey FOREIGN KEY (city) REFERENCES cities (city),
@@ -699,8 +711,8 @@ ALTER TABLE public.cities OWNER TO postgres;
 
 CREATE TABLE public.weather (
     city character varying(80),
-    temp_lo integer,
-    temp_hi integer,
+    temp_lo int8,
+    temp_hi int8,
     prcp real,
     date date
 );
@@ -724,12 +736,12 @@ ALTER TABLE ONLY public.weather
 
 	testPgdumpFkCircular = `
 CREATE TABLE public.a (
-    i integer NOT NULL,
-    k integer
+    i int8 NOT NULL,
+    k int8
 );
 
 CREATE TABLE public.b (
-    j integer NOT NULL
+    j int8 NOT NULL
 );
 
 COPY public.a (i, k) FROM stdin;
@@ -870,11 +882,17 @@ func TestImportCSVStmt(t *testing.T) {
 		t.Skip("short")
 	}
 
-	const (
-		nodes       = 3
-		numFiles    = nodes + 2
-		rowsPerFile = 1000
-	)
+	const nodes = 3
+
+	numFiles := nodes + 2
+	rowsPerFile := 1000
+	if util.RaceEnabled {
+		// This test takes a while with the race detector, so reduce the number of
+		// files and rows per file in an attempt to speed it up.
+		numFiles = nodes
+		rowsPerFile = 16
+	}
+
 	ctx := context.Background()
 	dir, cleanup := testutils.TempDir(t)
 	defer cleanup()
@@ -889,7 +907,7 @@ func TestImportCSVStmt(t *testing.T) {
 	tablePath := filepath.Join(dir, "table")
 	if err := ioutil.WriteFile(tablePath, []byte(`
 		CREATE TABLE t (
-			a int primary key,
+			a int8 primary key,
 			b string,
 			index (b),
 			index (a, b)
@@ -898,9 +916,6 @@ func TestImportCSVStmt(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	// Get the number of existing jobs.
-	baseNumJobs := jobutils.GetSystemJobsCount(t, sqlDB)
-
 	if err := ioutil.WriteFile(filepath.Join(dir, "empty.csv"), nil, 0666); err != nil {
 		t.Fatal(err)
 	}
@@ -960,7 +975,7 @@ func TestImportCSVStmt(t *testing.T) {
 		},
 		{
 			"schema-in-query",
-			`IMPORT TABLE t (a INT PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s)`,
+			`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s)`,
 			nil,
 			files,
 			``,
@@ -968,7 +983,7 @@ func TestImportCSVStmt(t *testing.T) {
 		},
 		{
 			"schema-in-query-opts",
-			`IMPORT TABLE t (a INT PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH delimiter = '|', comment = '#', nullif='', skip = '2'`,
+			`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH delimiter = '|', comment = '#', nullif='', skip = '2'`,
 			nil,
 			filesWithOpts,
 			` WITH comment = '#', delimiter = '|', "nullif" = '', skip = '2'`,
@@ -985,7 +1000,7 @@ func TestImportCSVStmt(t *testing.T) {
 		},
 		{
 			"schema-in-query-transform-only",
-			`IMPORT TABLE t (a INT PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH delimiter = '|', comment = '#', nullif='', skip = '2', transform = $1`,
+			`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH delimiter = '|', comment = '#', nullif='', skip = '2', transform = $1`,
 			nil,
 			filesWithOpts,
 			` WITH comment = '#', delimiter = '|', "nullif" = '', skip = '2', transform = 'nodelocal:///5'`,
@@ -1074,7 +1089,7 @@ func TestImportCSVStmt(t *testing.T) {
 		// NB: successes above, failures below, because we check the i-th job.
 		{
 			"bad-opt-name",
-			`IMPORT TABLE t (a INT PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH foo = 'bar'`,
+			`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH foo = 'bar'`,
 			nil,
 			files,
 			``,
@@ -1082,7 +1097,7 @@ func TestImportCSVStmt(t *testing.T) {
 		},
 		{
 			"bad-opt-no-arg",
-			`IMPORT TABLE t (a INT PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH transform`,
+			`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH transform`,
 			nil,
 			files,
 			``,
@@ -1090,7 +1105,7 @@ func TestImportCSVStmt(t *testing.T) {
 		},
 		{
 			"bad-computed-column",
-			`IMPORT TABLE t (a INT PRIMARY KEY, b STRING AS ('hello') STORED, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH skip = '2', transform = $1`,
+			`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING AS ('hello') STORED, INDEX (b), INDEX (a, b)) CSV DATA (%s) WITH skip = '2', transform = $1`,
 			nil,
 			filesWithOpts,
 			``,
@@ -1159,14 +1174,13 @@ func TestImportCSVStmt(t *testing.T) {
 			var result int
 			query := fmt.Sprintf(tc.query, strings.Join(tc.files, ", "))
 			testNum++
-			if err := sqlDB.DB.QueryRow(query, tc.args...).Scan(
-				&unused, &unused, &unused, &restored.rows, &restored.idx, &restored.sys, &restored.bytes,
-			); err != nil {
-				if !testutils.IsError(err, tc.err) {
-					t.Fatalf("%s: %+v (%#v)", query, err, tc.args)
-				}
+			if tc.err != "" {
+				sqlDB.ExpectErr(t, tc.err, query, tc.args...)
 				return
 			}
+			sqlDB.QueryRow(t, query, tc.args...).Scan(
+				&unused, &unused, &unused, &restored.rows, &restored.idx, &restored.sys, &restored.bytes,
+			)
 
 			jobPrefix := `IMPORT TABLE `
 			if !hasTransform {
@@ -1174,9 +1188,9 @@ func TestImportCSVStmt(t *testing.T) {
 			} else {
 				jobPrefix += `""."".`
 			}
-			jobPrefix += `t (a INT PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s)`
+			jobPrefix += `t (a INT8 PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b)) CSV DATA (%s)`
 
-			if err := jobutils.VerifySystemJob(t, sqlDB, baseNumJobs+testNum, jobspb.TypeImport, jobs.StatusSucceeded, jobs.Record{
+			if err := jobutils.VerifySystemJob(t, sqlDB, testNum, jobspb.TypeImport, jobs.StatusSucceeded, jobs.Record{
 				Username:    security.RootUser,
 				Description: fmt.Sprintf(jobPrefix+tc.jobOpts, strings.Join(tc.files, ", ")),
 			}); err != nil {
@@ -1186,19 +1200,15 @@ func TestImportCSVStmt(t *testing.T) {
 			isEmpty := len(tc.files) == 1 && tc.files[0] == empty[0]
 
 			if hasTransform {
-				if err := sqlDB.DB.QueryRow(`SELECT count(*) FROM t`).Scan(&unused); !testutils.IsError(
-					err, "does not exist",
-				) {
-					t.Fatal(err)
-				}
-				testNum++
-				if err := sqlDB.DB.QueryRow(
-					`RESTORE csv.* FROM $1 WITH into_db = $2`, backupPath, intodb,
+				sqlDB.ExpectErr(
+					t, "does not exist",
+					`SELECT count(*) FROM t`,
+				)
+				sqlDB.QueryRow(
+					t, `RESTORE csv.* FROM $1 WITH into_db = $2`, backupPath, intodb,
 				).Scan(
 					&unused, &unused, &unused, &restored.rows, &restored.idx, &restored.sys, &restored.bytes,
-				); err != nil {
-					t.Fatal(err)
-				}
+				)
 				if expected, actual := expectedRows, restored.rows; expected != actual && !isEmpty {
 					t.Fatalf("expected %d rows, got %d", expected, actual)
 				}
@@ -1258,7 +1268,7 @@ func TestImportCSVStmt(t *testing.T) {
 	// Verify unique_rowid is replaced for tables without primary keys.
 	t.Run("unique_rowid", func(t *testing.T) {
 		sqlDB.Exec(t, "CREATE DATABASE pk")
-		sqlDB.Exec(t, fmt.Sprintf(`IMPORT TABLE pk.t (a INT, b STRING) CSV DATA (%s)`, strings.Join(files, ", ")))
+		sqlDB.Exec(t, fmt.Sprintf(`IMPORT TABLE pk.t (a INT8, b STRING) CSV DATA (%s)`, strings.Join(files, ", ")))
 		// Verify the rowids are being generated as expected.
 		sqlDB.CheckQueryResults(t,
 			`SELECT count(*), sum(rowid) FROM pk.t`,
@@ -1277,26 +1287,26 @@ func TestImportCSVStmt(t *testing.T) {
 		sqlDB.Exec(t, "CREATE DATABASE checkpoint; USE checkpoint")
 
 		// Specify wrong number of columns.
-		_, err := conn.Exec(fmt.Sprintf(`IMPORT TABLE t (a INT PRIMARY KEY) CSV DATA (%s)`, files[0]))
-		if !testutils.IsError(err, "expected 1 fields, got 2") {
-			t.Fatalf("unexpected: %v", err)
-		}
+		sqlDB.ExpectErr(
+			t, "expected 1 fields, got 2",
+			fmt.Sprintf(`IMPORT TABLE t (a INT8 PRIMARY KEY) CSV DATA (%s)`, files[0]),
+		)
 
 		// Specify wrong table name; still shouldn't leave behind a checkpoint file.
-		_, err = conn.Exec(fmt.Sprintf(`IMPORT TABLE bad CREATE USING $1 CSV DATA (%s)`, files[0]), schema[0])
-		if !testutils.IsError(err, `file specifies a schema for table t`) {
-			t.Fatalf("unexpected: %v", err)
-		}
+		sqlDB.ExpectErr(
+			t, `file specifies a schema for table t`,
+			fmt.Sprintf(`IMPORT TABLE bad CREATE USING $1 CSV DATA (%s)`, files[0]), schema[0],
+		)
 
 		// Expect it to succeed with correct columns.
-		sqlDB.Exec(t, fmt.Sprintf(`IMPORT TABLE t (a INT PRIMARY KEY, b STRING) CSV DATA (%s)`, files[0]))
+		sqlDB.Exec(t, fmt.Sprintf(`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING) CSV DATA (%s)`, files[0]))
 
 		// A second attempt should fail fast. A "slow fail" is the error message
 		// "restoring table desc and namespace entries: table already exists".
-		_, err = conn.Exec(fmt.Sprintf(`IMPORT TABLE t (a INT PRIMARY KEY, b STRING) CSV DATA (%s)`, files[0]))
-		if !testutils.IsError(err, `relation "t" already exists`) {
-			t.Fatalf("unexpected: %v", err)
-		}
+		sqlDB.ExpectErr(
+			t, `relation "t" already exists`,
+			fmt.Sprintf(`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING) CSV DATA (%s)`, files[0]),
+		)
 	})
 
 	// Verify DEFAULT columns and SERIAL are allowed but not evaluated.
@@ -1314,11 +1324,11 @@ func TestImportCSVStmt(t *testing.T) {
 
 		const (
 			query = `IMPORT TABLE t (
-			a SERIAL,
-			b INT DEFAULT unique_rowid(),
+			a SERIAL8,
+			b INT8 DEFAULT unique_rowid(),
 			c STRING DEFAULT 's',
-			d SERIAL,
-			e INT DEFAULT unique_rowid(),
+			d SERIAL8,
+			e INT8 DEFAULT unique_rowid(),
 			f STRING DEFAULT 's',
 			PRIMARY KEY (a, b, c)
 		) CSV DATA ($1)`
@@ -1327,31 +1337,36 @@ func TestImportCSVStmt(t *testing.T) {
 
 		data = ",5,e,7,,"
 		t.Run(data, func(t *testing.T) {
-			if _, err := conn.Exec(query, srv.URL); !testutils.IsError(err, `row 1: parse "a" as INT: could not parse ""`) {
-				t.Fatalf("unexpected: %v", err)
-			}
-			if _, err := conn.Exec(query+nullif, srv.URL); !testutils.IsError(err, `row 1: generate insert row: null value in column "a" violates not-null constraint`) {
-				t.Fatalf("unexpected: %v", err)
-			}
+			sqlDB.ExpectErr(
+				t, `row 1: parse "a" as INT8: could not parse ""`,
+				query, srv.URL,
+			)
+			sqlDB.ExpectErr(
+				t, `row 1: generate insert row: null value in column "a" violates not-null constraint`,
+				query+nullif, srv.URL,
+			)
 		})
 		data = "2,5,e,,,"
 		t.Run(data, func(t *testing.T) {
-			if _, err := conn.Exec(query+nullif, srv.URL); !testutils.IsError(err, `row 1: generate insert row: null value in column "d" violates not-null constraint`) {
-				t.Fatalf("unexpected: %v", err)
-			}
+			sqlDB.ExpectErr(
+				t, `row 1: generate insert row: null value in column "d" violates not-null constraint`,
+				query+nullif, srv.URL,
+			)
 		})
 		data = "2,,e,,,"
 		t.Run(data, func(t *testing.T) {
-			if _, err := conn.Exec(query+nullif, srv.URL); !testutils.IsError(err, `"b" violates not-null constraint`) {
-				t.Fatalf("unexpected: %v", err)
-			}
+			sqlDB.ExpectErr(
+				t, `"b" violates not-null constraint`,
+				query+nullif, srv.URL,
+			)
 		})
 
 		data = "2,5,,,,"
 		t.Run(data, func(t *testing.T) {
-			if _, err := conn.Exec(query+nullif, srv.URL); !testutils.IsError(err, `"c" violates not-null constraint`) {
-				t.Fatalf("unexpected: %v", err)
-			}
+			sqlDB.ExpectErr(
+				t, `"c" violates not-null constraint`,
+				query+nullif, srv.URL,
+			)
 		})
 
 		data = "2,5,e,-1,,"
@@ -1385,7 +1400,7 @@ func BenchmarkImport(b *testing.B) {
 
 	sqlDB.Exec(b,
 		fmt.Sprintf(
-			`IMPORT TABLE t (a INT PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b))
+			`IMPORT TABLE t (a INT8 PRIMARY KEY, b STRING, INDEX (b), INDEX (a, b))
 			CSV DATA (%s) WITH transform = $1`,
 			strings.Join(files, ","),
 		),
@@ -1414,10 +1429,10 @@ func BenchmarkConvertRecord(b *testing.B) {
 	b.SetBytes(120) // Raw input size. With 8 indexes, expect more on output side.
 
 	stmt, err := parser.ParseOne(`CREATE TABLE lineitem (
-		l_orderkey      INTEGER NOT NULL,
-		l_partkey       INTEGER NOT NULL,
-		l_suppkey       INTEGER NOT NULL,
-		l_linenumber    INTEGER NOT NULL,
+		l_orderkey      INT8 NOT NULL,
+		l_partkey       INT8 NOT NULL,
+		l_suppkey       INT8 NOT NULL,
+		l_linenumber    INT8 NOT NULL,
 		l_quantity      DECIMAL(15,2) NOT NULL,
 		l_extendedprice DECIMAL(15,2) NOT NULL,
 		l_discount      DECIMAL(15,2) NOT NULL,
@@ -1443,7 +1458,7 @@ func BenchmarkConvertRecord(b *testing.B) {
 	if err != nil {
 		b.Fatal(err)
 	}
-	create := stmt.(*tree.CreateTable)
+	create := stmt.AST.(*tree.CreateTable)
 	st := cluster.MakeTestingClusterSettings()
 
 	tableDesc, err := MakeSimpleTableDescriptor(ctx, st, create, sqlbase.ID(100), sqlbase.ID(100), NoFKs, 1)
@@ -1460,7 +1475,7 @@ func BenchmarkConvertRecord(b *testing.B) {
 		}
 	}()
 
-	c := &csvInputReader{recordCh: recordCh, tableDesc: tableDesc}
+	c := &csvInputReader{recordCh: recordCh, tableDesc: tableDesc.TableDesc()}
 	// start up workers.
 	for i := 0; i < runtime.NumCPU(); i++ {
 		group.Go(func() error {
@@ -1498,7 +1513,6 @@ func BenchmarkConvertRecord(b *testing.B) {
 // work as intended on import jobs.
 func TestImportControlJob(t *testing.T) {
 	defer leaktest.AfterTest(t)()
-	t.Skipf("#24658")
 
 	defer func(oldInterval time.Duration) {
 		jobs.DefaultAdoptInterval = oldInterval
@@ -1527,15 +1541,33 @@ func TestImportControlJob(t *testing.T) {
 	sqlDB := sqlutils.MakeSQLRunner(tc.Conns[0])
 	sqlDB.Exec(t, `CREATE DATABASE data`)
 
-	t.Run("cancel", func(t *testing.T) {
-		sqlDB.Exec(t, `CREATE DATABASE cancelimport`)
-
-		srv := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+	makeSrv := func() *httptest.Server {
+		var once sync.Once
+		return httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 			if r.Method == "GET" {
-				<-allowResponse
+				// The following code correctly handles both the case where, after the
+				// CANCEL JOB is issued, the second stage of the IMPORT (the shuffle,
+				// after the sampling) may or may not be started. If it was started, then a
+				// second GET request is done. The once here will cause that request to not
+				// block. The draining for loop below will cause jobutils.RunJob's second send
+				// on allowResponse to succeed (which it does after issuing the CANCEL JOB).
+				once.Do(func() {
+					<-allowResponse
+					go func() {
+						for range allowResponse {
+						}
+					}()
+				})
+
 				_, _ = w.Write([]byte(r.URL.Path[1:]))
 			}
 		}))
+	}
+
+	t.Run("cancel", func(t *testing.T) {
+		sqlDB.Exec(t, `CREATE DATABASE cancelimport`)
+
+		srv := makeSrv()
 		defer srv.Close()
 
 		var urls []string
@@ -1544,7 +1576,7 @@ func TestImportControlJob(t *testing.T) {
 		}
 		csvURLs := strings.Join(urls, ", ")
 
-		query := fmt.Sprintf(`IMPORT TABLE cancelimport.t (i INT PRIMARY KEY) CSV DATA (%s)`, csvURLs)
+		query := fmt.Sprintf(`IMPORT TABLE cancelimport.t (i INT8 PRIMARY KEY) CSV DATA (%s)`, csvURLs)
 
 		if _, err := jobutils.RunJob(
 			t, sqlDB, &allowResponse, []string{"cancel"}, query,
@@ -1567,11 +1599,7 @@ func TestImportControlJob(t *testing.T) {
 
 		sqlDB.Exec(t, `CREATE DATABASE pauseimport`)
 
-		srv := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
-			if r.Method == "GET" {
-				_, _ = w.Write([]byte(r.URL.Path[1:]))
-			}
-		}))
+		srv := makeSrv()
 		defer srv.Close()
 
 		count := 100
@@ -1586,7 +1614,7 @@ func TestImportControlJob(t *testing.T) {
 			urls[i] = fmt.Sprintf("'%s/%d'", srv.URL, i)
 		}
 		csvURLs := strings.Join(urls, ", ")
-		query := fmt.Sprintf(`IMPORT TABLE pauseimport.t (i INT PRIMARY KEY) CSV DATA (%s) WITH sstsize = '50B'`, csvURLs)
+		query := fmt.Sprintf(`IMPORT TABLE pauseimport.t (i INT8 PRIMARY KEY) CSV DATA (%s) WITH sstsize = '50B'`, csvURLs)
 
 		jobID, err := jobutils.RunJob(
 			t, sqlDB, &allowResponse, []string{"PAUSE"}, query,
@@ -1595,9 +1623,7 @@ func TestImportControlJob(t *testing.T) {
 			t.Fatalf("unexpected: %v", err)
 		}
 		sqlDB.Exec(t, fmt.Sprintf(`RESUME JOB %d`, jobID))
-		if err := jobutils.WaitForJob(sqlDB.DB, jobID); err != nil {
-			t.Fatal(err)
-		}
+		jobutils.WaitForJob(t, sqlDB, jobID)
 		sqlDB.CheckQueryResults(t,
 			`SELECT * FROM pauseimport.t ORDER BY i`,
 			sqlDB.QueryStr(t, `SELECT * FROM generate_series(0, $1)`, count-1),
@@ -1629,7 +1655,8 @@ func TestImportWorkerFailure(t *testing.T) {
 	ctx := context.Background()
 	tc := testcluster.StartTestCluster(t, 3, params)
 	defer tc.Stopper().Stop(ctx)
-	sqlDB := sqlutils.MakeSQLRunner(tc.Conns[0])
+	conn := tc.Conns[0]
+	sqlDB := sqlutils.MakeSQLRunner(conn)
 
 	srv := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 		if r.Method == "GET" {
@@ -1644,11 +1671,11 @@ func TestImportWorkerFailure(t *testing.T) {
 		urls[i] = fmt.Sprintf("'%s/%d'", srv.URL, i)
 	}
 	csvURLs := strings.Join(urls, ", ")
-	query := fmt.Sprintf(`IMPORT TABLE t (i INT PRIMARY KEY) CSV DATA (%s) WITH sstsize = '1B'`, csvURLs)
+	query := fmt.Sprintf(`IMPORT TABLE t (i INT8 PRIMARY KEY) CSV DATA (%s) WITH sstsize = '1B'`, csvURLs)
 
 	errCh := make(chan error)
 	go func() {
-		_, err := sqlDB.DB.Exec(query)
+		_, err := conn.Exec(query)
 		errCh <- err
 	}()
 	select {
@@ -1670,9 +1697,7 @@ func TestImportWorkerFailure(t *testing.T) {
 	}
 
 	// But the job should be restarted and succeed eventually.
-	if err := jobutils.WaitForJob(sqlDB.DB, jobID); err != nil {
-		t.Fatal(err)
-	}
+	jobutils.WaitForJob(t, sqlDB, jobID)
 	sqlDB.CheckQueryResults(t,
 		`SELECT * FROM t ORDER BY i`,
 		sqlDB.QueryStr(t, `SELECT * FROM generate_series(0, $1)`, count-1),
@@ -1716,7 +1741,8 @@ func TestImportLivenessWithRestart(t *testing.T) {
 	ctx := context.Background()
 	tc := testcluster.StartTestCluster(t, nodes, params)
 	defer tc.Stopper().Stop(ctx)
-	sqlDB := sqlutils.MakeSQLRunner(tc.Conns[0])
+	conn := tc.Conns[0]
+	sqlDB := sqlutils.MakeSQLRunner(conn)
 
 	// Prevent hung HTTP connections in leaktest.
 	sqlDB.Exec(t, `SET CLUSTER SETTING cloudstorage.timeout = '3s'`)
@@ -1734,13 +1760,13 @@ func TestImportLivenessWithRestart(t *testing.T) {
 	}))
 	defer srv.Close()
 
-	const query = `IMPORT TABLE liveness.t (i INT PRIMARY KEY) CSV DATA ($1) WITH sstsize = '500B'`
+	const query = `IMPORT TABLE liveness.t (i INT8 PRIMARY KEY) CSV DATA ($1) WITH sstsize = '500B'`
 
 	// Start an IMPORT and wait until it's done one addsstable.
 	allowResponse = make(chan struct{})
 	errCh := make(chan error)
 	go func() {
-		_, err := sqlDB.DB.Exec(query, srv.URL)
+		_, err := conn.Exec(query, srv.URL)
 		errCh <- err
 	}()
 	// Allow many, but not all, addsstables to complete.
@@ -1784,9 +1810,7 @@ func TestImportLivenessWithRestart(t *testing.T) {
 	// Make the node live again
 	nl.FakeSetExpiration(1, hlc.MaxTimestamp)
 	// The registry should now adopt the job and resume it.
-	if err := jobutils.WaitForJob(sqlDB.DB, jobID); err != nil {
-		t.Fatal(err)
-	}
+	jobutils.WaitForJob(t, sqlDB, jobID)
 	// Verify that the job lease was updated
 	rescheduledProgress := jobutils.GetJobProgress(t, sqlDB, jobID)
 	if rescheduledProgress.ModifiedMicros <= originalLease.ModifiedMicros {
@@ -1847,7 +1871,8 @@ func TestImportLivenessWithLeniency(t *testing.T) {
 	ctx := context.Background()
 	tc := testcluster.StartTestCluster(t, nodes, params)
 	defer tc.Stopper().Stop(ctx)
-	sqlDB := sqlutils.MakeSQLRunner(tc.Conns[0])
+	conn := tc.Conns[0]
+	sqlDB := sqlutils.MakeSQLRunner(conn)
 
 	// Prevent hung HTTP connections in leaktest.
 	sqlDB.Exec(t, `SET CLUSTER SETTING cloudstorage.timeout = '3s'`)
@@ -1866,13 +1891,13 @@ func TestImportLivenessWithLeniency(t *testing.T) {
 	}))
 	defer srv.Close()
 
-	const query = `IMPORT TABLE liveness.t (i INT PRIMARY KEY) CSV DATA ($1) WITH sstsize = '500B'`
+	const query = `IMPORT TABLE liveness.t (i INT8 PRIMARY KEY) CSV DATA ($1) WITH sstsize = '500B'`
 
 	// Start an IMPORT and wait until it's done one addsstable.
 	allowResponse = make(chan struct{})
 	errCh := make(chan error)
 	go func() {
-		_, err := sqlDB.DB.Exec(query, srv.URL)
+		_, err := conn.Exec(query, srv.URL)
 		errCh <- err
 	}()
 	// Allow many, but not all, addsstables to complete.
@@ -1917,9 +1942,7 @@ func TestImportLivenessWithLeniency(t *testing.T) {
 	}
 
 	// The job should have completed normally.
-	if err := jobutils.WaitForJob(sqlDB.DB, jobID); err != nil {
-		t.Fatal(err)
-	}
+	jobutils.WaitForJob(t, sqlDB, jobID)
 }
 
 // TestImportMVCCChecksums verifies that MVCC checksums are correctly
@@ -1943,9 +1966,9 @@ func TestImportMVCCChecksums(t *testing.T) {
 	defer srv.Close()
 
 	sqlDB.Exec(t, `IMPORT TABLE d.t (
-		a INT PRIMARY KEY,
-		b INT,
-		c INT,
+		a INT8 PRIMARY KEY,
+		b INT8,
+		c INT8,
 		INDEX (b) STORING (c)
 	) CSV DATA ($1)`, srv.URL)
 	sqlDB.Exec(t, `UPDATE d.t SET c = 2 WHERE a = 1`)
@@ -1962,8 +1985,7 @@ func TestImportMysql(t *testing.T) {
 	args := base.TestServerArgs{ExternalIODir: baseDir}
 	tc := testcluster.StartTestCluster(t, nodes, base.TestClusterArgs{ServerArgs: args})
 	defer tc.Stopper().Stop(ctx)
-	conn := tc.Conns[0]
-	sqlDB := sqlutils.MakeSQLRunner(conn)
+	sqlDB := sqlutils.MakeSQLRunner(tc.Conns[0])
 
 	sqlDB.Exec(t, `SET CLUSTER SETTING kv.import.batch_size = '10KB'`)
 	sqlDB.Exec(t, `CREATE DATABASE foo; SET DATABASE = foo`)
@@ -1983,7 +2005,7 @@ func TestImportMysql(t *testing.T) {
 		query    string
 		args     []interface{}
 	}{
-		{`read data only`, expectSimple, `IMPORT TABLE simple (i INT PRIMARY KEY, s text, b bytea) MYSQLDUMP DATA ($1)`, simple},
+		{`read data only`, expectSimple, `IMPORT TABLE simple (i INT8 PRIMARY KEY, s text, b bytea) MYSQLDUMP DATA ($1)`, simple},
 		{`single table dump`, expectSimple, `IMPORT TABLE simple FROM MYSQLDUMP ($1)`, simple},
 		{`second table dump`, expectSecond, `IMPORT TABLE second FROM MYSQLDUMP ($1) WITH skip_foreign_keys`, second},
 		{`simple from multi`, expectSimple, `IMPORT TABLE simple FROM MYSQLDUMP ($1)`, multitable},
@@ -2030,11 +2052,7 @@ func TestImportMysql(t *testing.T) {
 					}
 				}
 			} else {
-				_, err := sqlDB.DB.Exec(`SELECT 1 FROM simple LIMIT 1`)
-				expected := "does not exist"
-				if !testutils.IsError(err, expected) {
-					t.Fatalf("expected %s, got %v", expected, err)
-				}
+				sqlDB.ExpectErr(t, "does not exist", `SELECT 1 FROM simple LIMIT 1`)
 			}
 
 			if c.expected&expectSecond != 0 {
@@ -2048,11 +2066,7 @@ func TestImportMysql(t *testing.T) {
 					}
 				}
 			} else {
-				_, err := sqlDB.DB.Exec(`SELECT 1 FROM second LIMIT 1`)
-				expected := "does not exist"
-				if !testutils.IsError(err, expected) {
-					t.Fatalf("expected %s, got %v", expected, err)
-				}
+				sqlDB.ExpectErr(t, "does not exist", `SELECT 1 FROM second LIMIT 1`)
 			}
 			if c.expected&expectEverything != 0 {
 				res := sqlDB.QueryStr(t, "SELECT i, c, iw, fl, d53, j FROM everything ORDER BY i")
@@ -2080,11 +2094,7 @@ func TestImportMysql(t *testing.T) {
 					}
 				}
 			} else {
-				_, err := sqlDB.DB.Exec(`SELECT 1 FROM everything LIMIT 1`)
-				expected := "does not exist"
-				if !testutils.IsError(err, expected) {
-					t.Fatalf("expected %s, got %v", expected, err)
-				}
+				sqlDB.ExpectErr(t, "does not exist", `SELECT 1 FROM everything LIMIT 1`)
 			}
 		})
 	}
@@ -2113,7 +2123,7 @@ func TestImportMysqlOutfile(t *testing.T) {
 		t.Run(cfg.name, func(t *testing.T) {
 			var opts []interface{}
 
-			cmd := fmt.Sprintf(`IMPORT TABLE test%d (i INT PRIMARY KEY, s text, b bytea) MYSQLOUTFILE DATA ($1)`, i)
+			cmd := fmt.Sprintf(`IMPORT TABLE test%d (i INT8 PRIMARY KEY, s text, b bytea) MYSQLOUTFILE DATA ($1)`, i)
 			opts = append(opts, fmt.Sprintf("nodelocal://%s", strings.TrimPrefix(cfg.filename, baseDir)))
 
 			var flags []string
@@ -2174,7 +2184,7 @@ func TestImportPgCopy(t *testing.T) {
 		t.Run(cfg.name, func(t *testing.T) {
 			var opts []interface{}
 
-			cmd := fmt.Sprintf(`IMPORT TABLE test%d (i INT PRIMARY KEY, s text, b bytea) PGCOPY DATA ($1)`, i)
+			cmd := fmt.Sprintf(`IMPORT TABLE test%d (i INT8 PRIMARY KEY, s text, b bytea) PGCOPY DATA ($1)`, i)
 			opts = append(opts, fmt.Sprintf("nodelocal://%s", strings.TrimPrefix(cfg.filename, baseDir)))
 
 			var flags []string
@@ -2253,7 +2263,7 @@ func TestImportPgDump(t *testing.T) {
 			`read data only`,
 			expectSimple,
 			`IMPORT TABLE simple (
-				i INTEGER,
+				i INT8,
 				s text,
 				b bytea,
 				CONSTRAINT simple_pkey PRIMARY KEY (i),
@@ -2276,7 +2286,7 @@ func TestImportPgDump(t *testing.T) {
 				// Verify table schema because PKs and indexes are at the bottom of pg_dump.
 				sqlDB.CheckQueryResults(t, `SHOW CREATE TABLE simple`, [][]string{{
 					"simple", `CREATE TABLE simple (
-	i INT NOT NULL,
+	i INT8 NOT NULL,
 	s STRING NULL,
 	b BYTES NULL,
 	CONSTRAINT simple_pkey PRIMARY KEY (i ASC),
@@ -2318,7 +2328,7 @@ func TestImportPgDump(t *testing.T) {
 				// Verify table schema because PKs and indexes are at the bottom of pg_dump.
 				sqlDB.CheckQueryResults(t, `SHOW CREATE TABLE second`, [][]string{{
 					"second", `CREATE TABLE second (
-	i INT NOT NULL,
+	i INT8 NOT NULL,
 	s STRING NULL,
 	CONSTRAINT second_pkey PRIMARY KEY (i ASC),
 	FAMILY "primary" (i, s)
@@ -2336,24 +2346,16 @@ func TestImportPgDump(t *testing.T) {
 			}
 
 			if c.expected == expectSecond {
-				_, err := sqlDB.DB.Exec(`SELECT 1 FROM simple LIMIT 1`)
-				expected := "does not exist"
-				if !testutils.IsError(err, expected) {
-					t.Fatalf("expected %s, got %v", expected, err)
-				}
+				sqlDB.ExpectErr(t, "does not exist", `SELECT 1 FROM simple LIMIT 1`)
 			}
 			if c.expected == expectSimple {
-				_, err := sqlDB.DB.Exec(`SELECT 1 FROM second LIMIT 1`)
-				expected := "does not exist"
-				if !testutils.IsError(err, expected) {
-					t.Fatalf("expected %s, got %v", expected, err)
-				}
+				sqlDB.ExpectErr(t, "does not exist", `SELECT 1 FROM second LIMIT 1`)
 			}
 			if c.expected == expectAll {
 				sqlDB.CheckQueryResults(t, `SHOW CREATE TABLE seqtable`, [][]string{{
 					"seqtable", `CREATE TABLE seqtable (
-	a INT NULL DEFAULT nextval('public.a_seq':::STRING),
-	b INT NULL,
+	a INT8 NULL DEFAULT nextval('public.a_seq':::STRING),
+	b INT8 NULL,
 	FAMILY "primary" (a, b, rowid)
 )`,
 				}})
@@ -2361,12 +2363,18 @@ func TestImportPgDump(t *testing.T) {
 					"a_seq", `CREATE SEQUENCE a_seq MINVALUE 1 MAXVALUE 9223372036854775807 INCREMENT 1 START 1`,
 				}})
 				sqlDB.CheckQueryResults(t, `select last_value from a_seq`, [][]string{{"7"}})
-				sqlDB.Exec(t, `INSERT INTO seqtable (b) VALUES (70)`)
 				sqlDB.CheckQueryResults(t,
 					`SELECT * FROM seqtable ORDER BY a`,
-					sqlDB.QueryStr(t, `select a+1, a*10 from generate_series(0, 7) a`),
+					sqlDB.QueryStr(t, `select a+1, a*10 from generate_series(0, 6) a`),
+				)
+				sqlDB.CheckQueryResults(t, `select last_value from a_seq`, [][]string{{"7"}})
+				// This can sometimes retry, so the next value might not be 8.
+				sqlDB.Exec(t, `INSERT INTO seqtable (b) VALUES (70)`)
+				sqlDB.CheckQueryResults(t, `select last_value >= 8 from a_seq`, [][]string{{"true"}})
+				sqlDB.CheckQueryResults(t,
+					`SELECT b FROM seqtable WHERE a = (SELECT last_value FROM a_seq)`,
+					[][]string{{"70"}},
 				)
-				sqlDB.CheckQueryResults(t, `select last_value from a_seq`, [][]string{{"8"}})
 			}
 		})
 	}
@@ -2403,7 +2411,7 @@ func TestImportCockroachDump(t *testing.T) {
 	})
 	sqlDB.CheckQueryResults(t, "SHOW CREATE TABLE t", [][]string{
 		{"t", `CREATE TABLE t (
-	i INT NOT NULL,
+	i INT8 NOT NULL,
 	t STRING NULL,
 	CONSTRAINT "primary" PRIMARY KEY (i ASC),
 	INDEX t_t_idx (t ASC),
@@ -2412,10 +2420,49 @@ func TestImportCockroachDump(t *testing.T) {
 	})
 	sqlDB.CheckQueryResults(t, "SHOW CREATE TABLE a", [][]string{
 		{"a", `CREATE TABLE a (
-	i INT NOT NULL,
+	i INT8 NOT NULL,
 	CONSTRAINT "primary" PRIMARY KEY (i ASC),
 	CONSTRAINT fk_i_ref_t FOREIGN KEY (i) REFERENCES t (i),
 	FAMILY "primary" (i)
 )`},
 	})
 }
+
+func TestCreateStatsAfterImport(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	defer func(oldRefreshInterval, oldAsOf time.Duration) {
+		stats.DefaultRefreshInterval = oldRefreshInterval
+		stats.DefaultAsOfTime = oldAsOf
+	}(stats.DefaultRefreshInterval, stats.DefaultAsOfTime)
+	stats.DefaultRefreshInterval = time.Millisecond
+	stats.DefaultAsOfTime = time.Microsecond
+
+	const nodes = 1
+	ctx := context.Background()
+	baseDir := filepath.Join("testdata")
+	args := base.TestServerArgs{ExternalIODir: baseDir}
+	tc := testcluster.StartTestCluster(t, nodes, base.TestClusterArgs{ServerArgs: args})
+	defer tc.Stopper().Stop(ctx)
+	conn := tc.Conns[0]
+	sqlDB := sqlutils.MakeSQLRunner(conn)
+
+	sqlDB.Exec(t, `SET CLUSTER SETTING sql.stats.automatic_collection.enabled=true`)
+
+	sqlDB.Exec(t, "IMPORT PGDUMP ($1)", "nodelocal:///cockroachdump/dump.sql")
+
+	// Verify that statistics have been created.
+	sqlDB.CheckQueryResultsRetry(t,
+		`SELECT statistics_name, column_names, row_count, distinct_count, null_count
+	  FROM [SHOW STATISTICS FOR TABLE t]`,
+		[][]string{
+			{"__auto__", "{i}", "2", "2", "0"},
+			{"__auto__", "{t}", "2", "2", "0"},
+		})
+	sqlDB.CheckQueryResultsRetry(t,
+		`SELECT statistics_name, column_names, row_count, distinct_count, null_count
+	  FROM [SHOW STATISTICS FOR TABLE a]`,
+		[][]string{
+			{"__auto__", "{i}", "1", "1", "0"},
+		})
+}
