@@ -17,10 +17,8 @@ import (
 	"strconv"
 	"strings"
 
-	"github.com/pkg/errors"
-	mysqltypes "vitess.io/vitess/go/sqltypes"
-	mysql "vitess.io/vitess/go/vt/sqlparser"
-
+	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
@@ -32,6 +30,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/ctxgroup"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/pkg/errors"
+	mysqltypes "vitess.io/vitess/go/sqltypes"
+	mysql "vitess.io/vitess/go/vt/sqlparser"
 )
 
 // mysqldumpReader reads the default output of `mysqldump`, which consists of
@@ -77,6 +78,16 @@ func (m *mysqldumpReader) inputFinished(ctx context.Context) {
 	close(m.kvCh)
 }
 
+func (m *mysqldumpReader) readFiles(
+	ctx context.Context,
+	dataFiles map[int32]string,
+	format roachpb.IOFileFormat,
+	progressFn func(float32) error,
+	settings *cluster.Settings,
+) error {
+	return readInputFiles(ctx, dataFiles, format, m.readFile, progressFn, settings)
+}
+
 func (m *mysqldumpReader) readFile(
 	ctx context.Context, input io.Reader, inputIdx int32, inputName string, progressFn progressFn,
 ) error {
@@ -370,15 +381,12 @@ func mysqlTableToCockroach(
 		if err != nil {
 			return nil, nil, err
 		}
-		seqDesc = &desc
-		fks.resolver[seqName] = seqDesc
+		seqDesc = desc.TableDesc()
+		fks.resolver[seqName] = &desc
 		id++
 	}
 
-	tblName := tree.MakeUnqualifiedTableName(tree.Name(name))
-	stmt := &tree.CreateTable{
-		Table: tree.NormalizableTableName{TableNameReference: &tblName},
-	}
+	stmt := &tree.CreateTable{Table: tree.MakeUnqualifiedTableName(tree.Name(name))}
 
 	checks := make(map[string]*tree.CheckConstraintTableDef)
 
@@ -451,15 +459,15 @@ func mysqlTableToCockroach(
 				d.Actions.Update = mysqlActionToCockroach(i.OnUpdate)
 			}
 
-			d.Table.TableNameReference = &toTable
+			d.Table = toTable
 			fkDefs = append(fkDefs, delayedFK{desc, d})
 		}
 	}
 	fks.resolver[desc.Name] = desc
 	if seqDesc != nil {
-		return []*sqlbase.TableDescriptor{seqDesc, desc}, fkDefs, nil
+		return []*sqlbase.TableDescriptor{seqDesc, desc.TableDesc()}, fkDefs, nil
 	}
-	return []*sqlbase.TableDescriptor{desc}, fkDefs, nil
+	return []*sqlbase.TableDescriptor{desc.TableDesc()}, fkDefs, nil
 }
 
 func mysqlActionToCockroach(action mysql.ReferenceAction) tree.ReferenceAction {
@@ -477,18 +485,18 @@ func mysqlActionToCockroach(action mysql.ReferenceAction) tree.ReferenceAction {
 }
 
 type delayedFK struct {
-	tbl *sqlbase.TableDescriptor
+	tbl *sqlbase.MutableTableDescriptor
 	def *tree.ForeignKeyConstraintTableDef
 }
 
 func addDelayedFKs(ctx context.Context, defs []delayedFK, resolver fkResolver) error {
 	for _, def := range defs {
 		if err := sql.ResolveFK(
-			ctx, nil, resolver, def.tbl, def.def, map[sqlbase.ID]*sqlbase.TableDescriptor{}, sqlbase.ConstraintValidity_Validated,
+			ctx, nil, resolver, def.tbl, def.def, map[sqlbase.ID]*sqlbase.MutableTableDescriptor{}, sql.NewTable,
 		); err != nil {
 			return err
 		}
-		if err := fixDescriptorFKState(def.tbl); err != nil {
+		if err := fixDescriptorFKState(def.tbl.TableDesc()); err != nil {
 			return err
 		}
 		if err := def.tbl.AllocateIDs(); err != nil {
@@ -558,7 +566,7 @@ func mysqlColToCockroach(
 	case mysqltypes.Int16:
 		def.Type = coltypes.Int2
 	case mysqltypes.Uint16:
-		def.Type = coltypes.Int2
+		def.Type = coltypes.Int4
 	case mysqltypes.Int24:
 		def.Type = coltypes.Int4
 	case mysqltypes.Uint24:
@@ -566,7 +574,7 @@ func mysqlColToCockroach(
 	case mysqltypes.Int32:
 		def.Type = coltypes.Int4
 	case mysqltypes.Uint32:
-		def.Type = coltypes.Int4
+		def.Type = coltypes.Int8
 	case mysqltypes.Int64:
 		def.Type = coltypes.Int8
 	case mysqltypes.Uint64:
