@@ -58,8 +58,7 @@ func splitSQLSemicolon(data []byte, atEOF bool) (advance int, token []byte, err
 		return 0, nil, nil
 	}
 
-	sc := parser.MakeScanner(string(data))
-	if pos := sc.Until(';'); pos > 0 {
+	if pos, ok := parser.SplitFirstStatement(string(data)); ok {
 		return pos, data[:pos], nil
 	}
 	// If we're at EOF, we have a final, non-terminated line. Return it.
@@ -101,7 +100,7 @@ func (p *postgreStream) Next() (interface{}, error) {
 			// If the statement is COPY ... FROM STDIN, set p.copy so the next call to
 			// this function will read copy data. We still return this COPY statement
 			// for this invocation.
-			if cf, ok := stmts[0].(*tree.CopyFrom); ok && cf.Stdin {
+			if cf, ok := stmts[0].AST.(*tree.CopyFrom); ok && cf.Stdin {
 				// Set p.copy which reconfigures the scanner's split func.
 				p.copy = newPostgreStreamCopy(p.s, copyDefaultDelimiter, copyDefaultNull)
 
@@ -117,7 +116,7 @@ func (p *postgreStream) Next() (interface{}, error) {
 					return nil, errors.Errorf("expected empty line")
 				}
 			}
-			return stmts[0], nil
+			return stmts[0].AST, nil
 		default:
 			return nil, errors.Errorf("unexpected: got %d statements", len(stmts))
 		}
@@ -245,9 +244,9 @@ func readPostgresCreateTable(
 					return nil, err
 				}
 				fks.resolver[desc.Name] = &desc
-				ret = append(ret, &desc)
+				ret = append(ret, desc.TableDesc())
 			}
-			backrefs := make(map[sqlbase.ID]*sqlbase.TableDescriptor)
+			backrefs := make(map[sqlbase.ID]*sqlbase.MutableTableDescriptor)
 			for _, create := range createTbl {
 				if create == nil {
 					continue
@@ -260,7 +259,7 @@ func readPostgresCreateTable(
 				}
 				fks.resolver[desc.Name] = desc
 				backrefs[desc.ID] = desc
-				ret = append(ret, desc)
+				ret = append(ret, desc.TableDesc())
 			}
 			for name, constraints := range tableFKs {
 				desc := fks.resolver[name]
@@ -268,14 +267,11 @@ func readPostgresCreateTable(
 					continue
 				}
 				for _, constraint := range constraints {
-					if _, err := constraint.Table.Normalize(); err != nil {
-						return nil, err
-					}
-					if err := sql.ResolveFK(evalCtx.Ctx(), nil /* txn */, fks.resolver, desc, constraint, backrefs, sqlbase.ConstraintValidity_Validated); err != nil {
+					if err := sql.ResolveFK(evalCtx.Ctx(), nil /* txn */, fks.resolver, desc, constraint, backrefs, sql.NewTable); err != nil {
 						return nil, err
 					}
 				}
-				if err := fixDescriptorFKState(desc); err != nil {
+				if err := fixDescriptorFKState(desc.TableDesc()); err != nil {
 					return nil, err
 				}
 			}
@@ -299,7 +295,7 @@ func readPostgresCreateTable(
 		}
 		switch stmt := stmt.(type) {
 		case *tree.CreateTable:
-			name, err := getTableName(stmt.Table)
+			name, err := getTableName(&stmt.Table)
 			if err != nil {
 				return nil, err
 			}
@@ -309,7 +305,7 @@ func readPostgresCreateTable(
 				createTbl[name] = stmt
 			}
 		case *tree.CreateIndex:
-			name, err := getTableName(stmt.Table)
+			name, err := getTableName(&stmt.Table)
 			if err != nil {
 				return nil, err
 			}
@@ -330,7 +326,7 @@ func readPostgresCreateTable(
 			}
 			create.Defs = append(create.Defs, idx)
 		case *tree.AlterTable:
-			name, err := getTableName(stmt.Table)
+			name, err := getTableName(&stmt.Table)
 			if err != nil {
 				return nil, err
 			}
@@ -365,7 +361,7 @@ func readPostgresCreateTable(
 				}
 			}
 		case *tree.CreateSequence:
-			name, err := getTableName(stmt.Name)
+			name, err := getTableName(&stmt.Name)
 			if err != nil {
 				return nil, err
 			}
@@ -376,11 +372,7 @@ func readPostgresCreateTable(
 	}
 }
 
-func getTableName(n tree.NormalizableTableName) (string, error) {
-	tn, err := n.Normalize()
-	if err != nil {
-		return "", err
-	}
+func getTableName(tn *tree.TableName) (string, error) {
 	if sc := tn.Schema(); sc != "" && sc != "public" {
 		return "", pgerror.Unimplemented(
 			"import non-public schema",
@@ -431,6 +423,16 @@ func (m *pgDumpReader) inputFinished(ctx context.Context) {
 	close(m.kvCh)
 }
 
+func (m *pgDumpReader) readFiles(
+	ctx context.Context,
+	dataFiles map[int32]string,
+	format roachpb.IOFileFormat,
+	progressFn func(float32) error,
+	settings *cluster.Settings,
+) error {
+	return readInputFiles(ctx, dataFiles, format, m.readFile, progressFn, settings)
+}
+
 func (m *pgDumpReader) readFile(
 	ctx context.Context, input io.Reader, inputIdx int32, inputName string, progressFn progressFn,
 ) error {
@@ -447,11 +449,11 @@ func (m *pgDumpReader) readFile(
 		}
 		switch i := stmt.(type) {
 		case *tree.Insert:
-			n, ok := i.Table.(*tree.NormalizableTableName)
+			n, ok := i.Table.(*tree.TableName)
 			if !ok {
 				return errors.Errorf("unexpected: %T", i.Table)
 			}
-			name, err := getTableName(*n)
+			name, err := getTableName(n)
 			if err != nil {
 				return errors.Wrapf(err, "%s", i)
 			}
@@ -495,7 +497,7 @@ func (m *pgDumpReader) readFile(
 			if !i.Stdin {
 				return errors.New("expected STDIN option on COPY FROM")
 			}
-			name, err := getTableName(i.Table)
+			name, err := getTableName(&i.Table)
 			if err != nil {
 				return errors.Wrapf(err, "%s", i)
 			}
