@@ -16,27 +16,35 @@ import (
 	"io/ioutil"
 	"math/rand"
 	"strings"
-	"sync"
-
-	"github.com/cockroachdb/cockroach/pkg/keys"
-	"github.com/cockroachdb/cockroach/pkg/sql"
-	"github.com/cockroachdb/cockroach/pkg/sql/sem/builtins"
-	"github.com/cockroachdb/cockroach/pkg/sql/sem/transform"
-	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 
 	"github.com/cockroachdb/cockroach/pkg/ccl/storageccl"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
+	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/sql"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/distsqlrun"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
+	"github.com/cockroachdb/cockroach/pkg/sql/sem/builtins"
+	"github.com/cockroachdb/cockroach/pkg/sql/sem/transform"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
+	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
 	"github.com/cockroachdb/cockroach/pkg/util/ctxgroup"
+	"github.com/cockroachdb/cockroach/pkg/util/encoding"
+	"github.com/cockroachdb/cockroach/pkg/util/envutil"
+	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/pkg/errors"
 )
 
+var useWorkloadFastpath = envutil.EnvOrDefaultBool("COCKROACH_IMPORT_WORKLOAD_FASTER", false)
+
 type readFileFunc func(context.Context, io.Reader, int32, string, progressFn) error
 
 // readInputFile reads each of the passed dataFiles using the passed func. The
@@ -196,11 +204,11 @@ type rowConverter struct {
 	kvBatch  kvBatch
 	batchCap int
 
-	tableDesc *sqlbase.TableDescriptor
+	tableDesc *sqlbase.ImmutableTableDescriptor
 
 	// The rest of these are derived from tableDesc, just cached here.
 	hidden                int
-	ri                    sqlbase.RowInserter
+	ri                    row.Inserter
 	evalCtx               *tree.EvalContext
 	cols                  []sqlbase.ColumnDescriptor
 	visibleCols           []sqlbase.ColumnDescriptor
@@ -209,19 +217,20 @@ type rowConverter struct {
 	computedIVarContainer sqlbase.RowIndexedVarContainer
 }
 
-const kvBatchSize = 1000
+const kvBatchSize = 5000
 
 func newRowConverter(
 	tableDesc *sqlbase.TableDescriptor, evalCtx *tree.EvalContext, kvCh chan<- kvBatch,
 ) (*rowConverter, error) {
+	immutDesc := sqlbase.NewImmutableTableDescriptor(*tableDesc)
 	c := &rowConverter{
-		tableDesc: tableDesc,
+		tableDesc: immutDesc,
 		kvCh:      kvCh,
 		evalCtx:   evalCtx,
 	}
 
-	ri, err := sqlbase.MakeRowInserter(nil /* txn */, tableDesc, nil, /* fkTables */
-		tableDesc.Columns, false /* checkFKs */, &sqlbase.DatumAlloc{})
+	ri, err := row.MakeInserter(nil /* txn */, immutDesc, nil, /* fkTables */
+		immutDesc.Columns, false /* checkFKs */, &sqlbase.DatumAlloc{})
 	if err != nil {
 		return nil, errors.Wrap(err, "make row inserter")
 	}
@@ -231,14 +240,14 @@ func newRowConverter(
 	// Although we don't yet support DEFAULT expressions on visible columns,
 	// we do on hidden columns (which is only the default _rowid one). This
 	// allows those expressions to run.
-	cols, defaultExprs, err := sqlbase.ProcessDefaultColumns(tableDesc.Columns, tableDesc, &txCtx, c.evalCtx)
+	cols, defaultExprs, err := sqlbase.ProcessDefaultColumns(immutDesc.Columns, immutDesc, &txCtx, c.evalCtx)
 	if err != nil {
 		return nil, errors.Wrap(err, "process default columns")
 	}
 	c.cols = cols
 	c.defaultExprs = defaultExprs
 
-	c.visibleCols = tableDesc.VisibleColumns()
+	c.visibleCols = immutDesc.VisibleColumns()
 	c.visibleColTypes = make([]types.T, len(c.visibleCols))
 	for i := range c.visibleCols {
 		c.visibleColTypes[i] = c.visibleCols[i].DatumType()
@@ -260,13 +269,13 @@ func newRowConverter(
 		return nil, errors.New("unexpected hidden column")
 	}
 
-	padding := 2 * (len(tableDesc.Indexes) + len(tableDesc.Families))
+	padding := 2 * (len(immutDesc.Indexes) + len(immutDesc.Families))
 	c.batchCap = kvBatchSize + padding
 	c.kvBatch = make(kvBatch, 0, c.batchCap)
 
 	c.computedIVarContainer = sqlbase.RowIndexedVarContainer{
 		Mapping: ri.InsertColIDtoRowIndex,
-		Cols:    tableDesc.Columns,
+		Cols:    immutDesc.Columns,
 	}
 	return c, nil
 }
@@ -293,7 +302,7 @@ func (c *rowConverter) row(ctx context.Context, fileIndex int32, rowIndex int64)
 	var computeExprs []tree.TypedExpr
 	var computedCols []sqlbase.ColumnDescriptor
 
-	row, err := sql.GenerateInsertRow(
+	insertRow, err := sql.GenerateInsertRow(
 		c.defaultExprs, computeExprs, c.cols, computedCols, *c.evalCtx, c.tableDesc, c.datums, &c.computedIVarContainer)
 	if err != nil {
 		return errors.Wrapf(err, "generate insert row")
@@ -304,9 +313,9 @@ func (c *rowConverter) row(ctx context.Context, fileIndex int32, rowIndex int64)
 			kv.Value.InitChecksum(kv.Key)
 			c.kvBatch = append(c.kvBatch, kv)
 		}),
-		row,
+		insertRow,
 		true, /* ignoreConflicts */
-		sqlbase.SkipFKs,
+		row.SkipFKs,
 		false, /* traceKV */
 	); err != nil {
 		return errors.Wrapf(err, "insert row")
@@ -341,7 +350,7 @@ var csvOutputTypes = []sqlbase.ColumnType{
 func newReadImportDataProcessor(
 	flowCtx *distsqlrun.FlowCtx,
 	processorID int32,
-	spec distsqlrun.ReadImportDataSpec,
+	spec distsqlpb.ReadImportDataSpec,
 	output distsqlrun.RowReceiver,
 ) (distsqlrun.Processor, error) {
 	cp := &readImportDataProcessor{
@@ -351,7 +360,7 @@ func newReadImportDataProcessor(
 		output:      output,
 	}
 
-	if err := cp.out.Init(&distsqlrun.PostProcessSpec{}, csvOutputTypes, flowCtx.NewEvalCtx(), output); err != nil {
+	if err := cp.out.Init(&distsqlpb.PostProcessSpec{}, csvOutputTypes, flowCtx.NewEvalCtx(), output); err != nil {
 		return nil, err
 	}
 	return cp, nil
@@ -361,14 +370,14 @@ type progressFn func(finished bool) error
 
 type inputConverter interface {
 	start(group ctxgroup.Group)
-	readFile(ctx context.Context, input io.Reader, fileIdx int32, filename string, progress progressFn) error
+	readFiles(ctx context.Context, dataFiles map[int32]string, format roachpb.IOFileFormat, progressFn func(float32) error, settings *cluster.Settings) error
 	inputFinished(ctx context.Context)
 }
 
 type readImportDataProcessor struct {
 	flowCtx     *distsqlrun.FlowCtx
 	processorID int32
-	spec        distsqlrun.ReadImportDataSpec
+	spec        distsqlpb.ReadImportDataSpec
 	out         distsqlrun.ProcOutputHelper
 	output      distsqlrun.RowReceiver
 }
@@ -388,15 +397,11 @@ func isMultiTableFormat(format roachpb.IOFileFormat_FileFormat) bool {
 	return false
 }
 
-func (cp *readImportDataProcessor) Run(ctx context.Context, wg *sync.WaitGroup) {
+func (cp *readImportDataProcessor) Run(ctx context.Context) {
 	ctx, span := tracing.ChildSpan(ctx, "readImportDataProcessor")
 	defer tracing.FinishSpan(span)
 
-	if wg != nil {
-		defer wg.Done()
-	}
-
-	if err := cp.doRun(ctx, wg); err != nil {
+	if err := cp.doRun(ctx); err != nil {
 		distsqlrun.DrainAndClose(ctx, cp.output, err, func(context.Context) {} /* pushTrailingMeta */)
 	} else {
 		cp.out.Close()
@@ -406,9 +411,9 @@ func (cp *readImportDataProcessor) Run(ctx context.Context, wg *sync.WaitGroup)
 // doRun uses a more familiar error return API, allowing concise early returns
 // on errors in setup that all then are handled by the actual DistSQL Run method
 // wrapper, doing the correct DrainAndClose error handling logic.
-func (cp *readImportDataProcessor) doRun(ctx context.Context, wg *sync.WaitGroup) error {
+func (cp *readImportDataProcessor) doRun(ctx context.Context) error {
 	group := ctxgroup.WithContext(ctx)
-	kvCh := make(chan kvBatch)
+	kvCh := make(chan kvBatch, 10)
 	evalCtx := cp.flowCtx.NewEvalCtx()
 
 	var singleTable *sqlbase.TableDescriptor
@@ -418,6 +423,8 @@ func (cp *readImportDataProcessor) doRun(ctx context.Context, wg *sync.WaitGroup
 		}
 	}
 
+	typeBytes := sqlbase.ColumnType{SemanticType: sqlbase.ColumnType_BYTES}
+
 	if format := cp.spec.Format.Format; singleTable == nil && !isMultiTableFormat(format) {
 		return errors.Errorf("%s only supports reading a single, pre-specified table", format.String())
 	}
@@ -426,7 +433,18 @@ func (cp *readImportDataProcessor) doRun(ctx context.Context, wg *sync.WaitGroup
 	var err error
 	switch cp.spec.Format.Format {
 	case roachpb.IOFileFormat_CSV:
-		conv = newCSVInputReader(kvCh, cp.spec.Format.Csv, singleTable, cp.flowCtx)
+		isWorkload := useWorkloadFastpath
+		for _, file := range cp.spec.Uri {
+			if conf, err := storageccl.ExportStorageConfFromURI(file); err != nil || conf.Provider != roachpb.ExportStorageProvider_Workload {
+				isWorkload = false
+				break
+			}
+		}
+		if isWorkload {
+			conv, err = newWorkloadReader(kvCh, singleTable, evalCtx)
+		} else {
+			conv = newCSVInputReader(kvCh, cp.spec.Format.Csv, singleTable, cp.flowCtx)
+		}
 	case roachpb.IOFileFormat_MysqlOutfile:
 		conv, err = newMysqloutfileReader(kvCh, cp.spec.Format.MysqlOut, singleTable, evalCtx)
 	case roachpb.IOFileFormat_Mysqldump:
@@ -467,75 +485,115 @@ func (cp *readImportDataProcessor) doRun(ctx context.Context, wg *sync.WaitGroup
 			})
 		}
 
-		return readInputFiles(ctx, cp.spec.Uri, cp.spec.Format, conv.readFile, progFn, cp.flowCtx.Settings)
+		return conv.readFiles(ctx, cp.spec.Uri, cp.spec.Format, progFn, cp.flowCtx.Settings)
 	})
 
-	// Sample KVs
-	group.GoCtx(func(ctx context.Context) error {
-		ctx, span := tracing.ChildSpan(ctx, "sendImportKVs")
-		defer tracing.FinishSpan(span)
+	if cp.spec.IngestDirectly {
+		// IngestDirectly means this reader will just ingest the KVs that the
+		// producer emitted to the chan, and the only result we push into distsql at
+		// the end is one row containing an encoded BulkOpSummary.
+		group.GoCtx(func(ctx context.Context) error {
+			ctx, span := tracing.ChildSpan(ctx, "ingestKVs")
+			defer tracing.FinishSpan(span)
 
-		var fn sampleFunc
-		var sampleAll bool
-		if cp.spec.SampleSize == 0 {
-			sampleAll = true
-		} else {
-			sr := sampleRate{
-				rnd:        rand.New(rand.NewSource(rand.Int63())),
-				sampleSize: float64(cp.spec.SampleSize),
+			writeTS := hlc.Timestamp{WallTime: cp.spec.WalltimeNanos}
+			const bufferSize, flushSize = 64 << 20, 16 << 20
+			adder, err := cp.flowCtx.BulkAdder(ctx, cp.flowCtx.ClientDB, bufferSize, flushSize, writeTS)
+			if err != nil {
+				return err
 			}
-			fn = sr.sample
-		}
+			defer adder.Close(ctx)
 
-		// Populate the split-point spans which have already been imported.
-		var completedSpans roachpb.SpanGroup
-		job, err := cp.flowCtx.JobRegistry.LoadJob(ctx, cp.spec.Progress.JobID)
-		if err != nil {
-			return err
-		}
-		progress := job.Progress()
-		if details, ok := progress.Details.(*jobspb.Progress_Import); ok {
-			completedSpans.Add(details.Import.SpanProgress...)
-		} else {
-			return errors.Errorf("unexpected progress type %T", progress)
-		}
+			// Drain the kvCh using the BulkAdder until it closes.
+			if err := ingestKvs(ctx, adder, kvCh); err != nil {
+				return err
+			}
 
-		typeBytes := sqlbase.ColumnType{SemanticType: sqlbase.ColumnType_BYTES}
-		for kvBatch := range kvCh {
-			for _, kv := range kvBatch {
-				// Allow KV pairs to be dropped if they belong to a completed span.
-				if completedSpans.Contains(kv.Key) {
-					continue
+			added := adder.GetSummary()
+			countsBytes, err := protoutil.Marshal(&added)
+			if err != nil {
+				return err
+			}
+			cs, err := cp.out.EmitRow(ctx, sqlbase.EncDatumRow{
+				sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes(countsBytes))),
+				sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes([]byte{}))),
+			})
+			if err != nil {
+				return err
+			}
+			if cs != distsqlrun.NeedMoreRows {
+				return errors.New("unexpected closure of consumer")
+			}
+			return nil
+		})
+	} else {
+		// Sample KVs
+		group.GoCtx(func(ctx context.Context) error {
+
+			ctx, span := tracing.ChildSpan(ctx, "sendImportKVs")
+			defer tracing.FinishSpan(span)
+
+			var fn sampleFunc
+			var sampleAll bool
+			if cp.spec.SampleSize == 0 {
+				sampleAll = true
+			} else {
+				sr := sampleRate{
+					rnd:        rand.New(rand.NewSource(rand.Int63())),
+					sampleSize: float64(cp.spec.SampleSize),
 				}
+				fn = sr.sample
+			}
 
-				rowRequired := sampleAll || keys.IsDescriptorKey(kv.Key)
-				if rowRequired || fn(kv) {
-					var row sqlbase.EncDatumRow
-					if rowRequired {
-						row = sqlbase.EncDatumRow{
-							sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes(kv.Key))),
-							sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes(kv.Value.RawBytes))),
-						}
-					} else {
-						// Don't send the value for rows returned for sampling
-						row = sqlbase.EncDatumRow{
-							sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes(kv.Key))),
-							sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes([]byte{}))),
-						}
+			// Populate the split-point spans which have already been imported.
+			var completedSpans roachpb.SpanGroup
+			job, err := cp.flowCtx.JobRegistry.LoadJob(ctx, cp.spec.Progress.JobID)
+			if err != nil {
+				return err
+			}
+			progress := job.Progress()
+			if details, ok := progress.Details.(*jobspb.Progress_Import); ok {
+				completedSpans.Add(details.Import.SpanProgress...)
+			} else {
+				return errors.Errorf("unexpected progress type %T", progress)
+			}
+
+			for kvBatch := range kvCh {
+				for _, kv := range kvBatch {
+					// Allow KV pairs to be dropped if they belong to a completed span.
+					if completedSpans.Contains(kv.Key) {
+						continue
 					}
 
-					cs, err := cp.out.EmitRow(ctx, row)
-					if err != nil {
-						return err
-					}
-					if cs != distsqlrun.NeedMoreRows {
-						return errors.New("unexpected closure of consumer")
+					rowRequired := sampleAll || keys.IsDescriptorKey(kv.Key)
+					if rowRequired || fn(kv) {
+						var row sqlbase.EncDatumRow
+						if rowRequired {
+							row = sqlbase.EncDatumRow{
+								sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes(kv.Key))),
+								sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes(kv.Value.RawBytes))),
+							}
+						} else {
+							// Don't send the value for rows returned for sampling
+							row = sqlbase.EncDatumRow{
+								sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes(kv.Key))),
+								sqlbase.DatumToEncDatum(typeBytes, tree.NewDBytes(tree.DBytes([]byte{}))),
+							}
+						}
+
+						cs, err := cp.out.EmitRow(ctx, row)
+						if err != nil {
+							return err
+						}
+						if cs != distsqlrun.NeedMoreRows {
+							return errors.New("unexpected closure of consumer")
+						}
 					}
 				}
 			}
-		}
-		return nil
-	})
+			return nil
+		})
+	}
 
 	return group.Wait()
 }
@@ -559,6 +617,84 @@ func makeRowErr(file string, row int64, format string, args ...interface{}) erro
 	return errors.Errorf("%q: row %d: "+format, append([]interface{}{file, row}, args...)...)
 }
 
+// ingestKvs drains kvs from the channel until it closes, ingesting them using
+// the BulkAdder. It handles the required buffering/sorting/etc.
+func ingestKvs(ctx context.Context, adder storagebase.BulkAdder, kvCh <-chan kvBatch) error {
+	const sortBatchSize = 48 << 20 // 48MB
+
+	// TODO(dt): buffer to disk instead of all in-mem.
+
+	// Batching all kvs together leads to worst case overlap behavior in the
+	// resulting AddSSTable calls, leading to compactions and potentially L0
+	// stalls. Instead maintain a separate buffer for each table's primary data.
+	// This optimizes for the case when the data arriving to IMPORT is already
+	// sorted by primary key, leading to no overlapping AddSSTable requests. Given
+	// that many workloads (and actual imported data) will be sorted by primary
+	// key, it makes sense to try to exploit this.
+	//
+	// TODO(dan): This was merged because it stabilized direct ingest IMPORT, but
+	// we may be able to do something simpler (such as chunking along index
+	// boundaries in flush) or more general (such as chunking based on the common
+	// prefix of the last N kvs).
+	kvsByTableIDIndexID := make(map[string]roachpb.KeyValueByKey)
+	sizeByTableIDIndexID := make(map[string]int64)
+
+	flush := func(ctx context.Context, buf roachpb.KeyValueByKey) error {
+		if len(buf) == 0 {
+			return nil
+		}
+		for i := range buf {
+			if err := adder.Add(ctx, buf[i].Key, buf[i].Value.RawBytes); err != nil {
+				if _, ok := err.(storagebase.DuplicateKeyError); ok {
+					return pgerror.Wrap(err, pgerror.CodeDataExceptionError, "")
+				}
+				return err
+			}
+		}
+		return nil
+	}
+
+	for kvBatch := range kvCh {
+		for _, kv := range kvBatch {
+			tableLen, err := encoding.PeekLength(kv.Key)
+			if err != nil {
+				return err
+			}
+			indexLen, err := encoding.PeekLength(kv.Key[tableLen:])
+			if err != nil {
+				return err
+			}
+			bufKey := kv.Key[:tableLen+indexLen]
+			kvsByTableIDIndexID[string(bufKey)] = append(kvsByTableIDIndexID[string(bufKey)], kv)
+			sizeByTableIDIndexID[string(bufKey)] += int64(len(kv.Key) + len(kv.Value.RawBytes))
+
+			// TODO(dan): Prevent unbounded memory usage by flushing the largest
+			// buffer when the total size of all buffers exceeds some threshold.
+			if s := sizeByTableIDIndexID[string(bufKey)]; s > sortBatchSize {
+				buf := kvsByTableIDIndexID[string(bufKey)]
+				if err := flush(ctx, buf); err != nil {
+					return err
+				}
+				kvsByTableIDIndexID[string(bufKey)] = buf[:0]
+				sizeByTableIDIndexID[string(bufKey)] = 0
+			}
+		}
+	}
+	for _, buf := range kvsByTableIDIndexID {
+		if err := flush(ctx, buf); err != nil {
+			return err
+		}
+	}
+
+	if err := adder.Flush(ctx); err != nil {
+		if err, ok := err.(storagebase.DuplicateKeyError); ok {
+			return pgerror.Wrap(err, pgerror.CodeDataExceptionError, "")
+		}
+		return err
+	}
+	return nil
+}
+
 func init() {
 	distsqlrun.NewReadImportDataProcessor = newReadImportDataProcessor
 }
