@@ -18,6 +18,7 @@ import (
 	"testing"
 	"time"
 
+	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/ccl/importccl"
 	"github.com/cockroachdb/cockroach/pkg/ccl/utilccl"
 	"github.com/cockroachdb/cockroach/pkg/config"
@@ -28,18 +29,19 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/testcluster"
-	"github.com/pkg/errors"
-	yaml "gopkg.in/yaml.v2"
-
-	"github.com/cockroachdb/cockroach/pkg/base"
+	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/randutil"
+	"github.com/gogo/protobuf/proto"
+	"github.com/pkg/errors"
+	yaml "gopkg.in/yaml.v2"
 )
 
 // partitioningTest represents a single test case used in the various
@@ -118,17 +120,18 @@ func (t *partitioningTest) parse() error {
 		if err != nil {
 			return errors.Wrapf(err, `parsing %s`, t.parsed.createStmt)
 		}
-		createTable, ok := stmt.(*tree.CreateTable)
+		createTable, ok := stmt.AST.(*tree.CreateTable)
 		if !ok {
 			return errors.Errorf("expected *tree.CreateTable got %T", stmt)
 		}
 		st := cluster.MakeTestingClusterSettings()
 		const parentID, tableID = keys.MinUserDescID, keys.MinUserDescID + 1
-		t.parsed.tableDesc, err = importccl.MakeSimpleTableDescriptor(
+		mutDesc, err := importccl.MakeSimpleTableDescriptor(
 			ctx, st, createTable, parentID, tableID, importccl.NoFKs, hlc.UnixNano())
 		if err != nil {
 			return err
 		}
+		t.parsed.tableDesc = mutDesc.TableDesc()
 		if err := t.parsed.tableDesc.ValidateTable(st); err != nil {
 			return err
 		}
@@ -180,7 +183,8 @@ func (t *partitioningTest) parse() error {
 		if err := yaml.UnmarshalStrict([]byte("["+constraints+"]"), &parsedConstraints); err != nil {
 			return errors.Wrapf(err, "parsing constraints: %s", constraints)
 		}
-		subzone.Config.Constraints = ([]config.Constraints)(parsedConstraints)
+		subzone.Config.Constraints = parsedConstraints.Constraints
+		subzone.Config.InheritedConstraints = parsedConstraints.Inherited
 
 		t.parsed.subzones = append(t.parsed.subzones, subzone)
 	}
@@ -843,19 +847,21 @@ func allPartitioningTests(rng *rand.Rand) []partitioningTest {
 
 	const schemaFmt = `CREATE TABLE %%s (a %s PRIMARY KEY) PARTITION BY LIST (a) (PARTITION p VALUES IN (%s))`
 	for semTypeID, semTypeName := range sqlbase.ColumnType_SemanticType_name {
-		// Tuples are not valid types for table creation
-		if semTypeID == int32(sqlbase.ColumnType_TUPLE) {
+		semType := sqlbase.ColumnType_SemanticType(semTypeID)
+		switch semType {
+		case sqlbase.ColumnType_ARRAY,
+			sqlbase.ColumnType_TUPLE,
+			sqlbase.ColumnType_JSONB:
+			// Not indexable.
 			continue
 		}
-		typ := sqlbase.ColumnType{SemanticType: sqlbase.ColumnType_SemanticType(semTypeID)}
+
+		typ := sqlbase.ColumnType{SemanticType: semType}
 		colType := semTypeName
 		switch typ.SemanticType {
 		case sqlbase.ColumnType_COLLATEDSTRING:
 			typ.Locale = sqlbase.RandCollationLocale(rng)
 			colType = fmt.Sprintf(`STRING COLLATE %s`, *typ.Locale)
-		case sqlbase.ColumnType_JSONB:
-			// Not indexable.
-			continue
 		}
 		datum := sqlbase.RandDatum(rng, typ, false /* nullOk */)
 		if datum == tree.DNull {
@@ -1078,8 +1084,16 @@ func verifyScansOnNode(db *gosql.DB, query string, node string) error {
 			return err
 		}
 		traceLines = append(traceLines, traceLine.String)
-		if strings.Contains(traceLine.String, "read completed") && !strings.Contains(traceLine.String, node) {
-			scansWrongNode = append(scansWrongNode, traceLine.String)
+		if strings.Contains(traceLine.String, "read completed") {
+			if strings.Contains(traceLine.String, "SystemCon") {
+				// Ignore trace lines for the system config range (abbreviated as
+				// "SystemCon" in pretty printing of the range descriptor). A read might
+				// be performed to the system config range to update the table lease.
+				continue
+			}
+			if !strings.Contains(traceLine.String, node) {
+				scansWrongNode = append(scansWrongNode, traceLine.String)
+			}
 		}
 	}
 	if len(scansWrongNode) > 0 {
@@ -1095,13 +1109,22 @@ func verifyScansOnNode(db *gosql.DB, query string, node string) error {
 	return nil
 }
 
-func setupPartitioningTestCluster(ctx context.Context, t testing.TB) (*sqlutils.SQLRunner, func()) {
+func setupPartitioningTestCluster(
+	ctx context.Context, t testing.TB,
+) (*gosql.DB, *sqlutils.SQLRunner, func()) {
 	cfg := config.DefaultZoneConfig()
-	cfg.NumReplicas = 1
+	cfg.NumReplicas = proto.Int32(1)
 	resetZoneConfig := config.TestingSetDefaultZoneConfig(cfg)
 
 	tsArgs := func(attr string) base.TestServerArgs {
 		return base.TestServerArgs{
+			Knobs: base.TestingKnobs{
+				Store: &storage.StoreTestingKnobs{
+					// Disable LBS because when the scan is happening at the rate it's happening
+					// below, it's possible that one of the system ranges trigger a split.
+					DisableLoadBasedSplitting: true,
+				},
+			},
 			ScanInterval: 100 * time.Millisecond,
 			StoreSpecs: []base.StoreSpec{
 				{InMemory: true, Attributes: roachpb.Attributes{Attrs: []string{attr}}},
@@ -1123,11 +1146,7 @@ func setupPartitioningTestCluster(ctx context.Context, t testing.TB) (*sqlutils.
 	sqlDB.Exec(t, `SET CLUSTER SETTING server.declined_reservation_timeout = '0s'`)
 	sqlDB.Exec(t, `SET CLUSTER SETTING server.failed_reservation_timeout = '0s'`)
 
-	// Make sure all stores are present in the NodeStatus endpoint or else zone
-	// config changes may flake (#25488).
-	tc.WaitForNodeStatuses(t)
-
-	return sqlDB, func() {
+	return tc.Conns[0], sqlDB, func() {
 		tc.Stopper().Stop(context.Background())
 		resetZoneConfig()
 	}
@@ -1136,13 +1155,20 @@ func setupPartitioningTestCluster(ctx context.Context, t testing.TB) (*sqlutils.
 func TestInitialPartitioning(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	t.Skip("#28789")
+	// This test configures many sub-tests and is too slow to run under nightly
+	// race stress.
+	if testutils.NightlyStress() && util.RaceEnabled {
+		t.Skip("too big for nightly stress race")
+	}
+	if testing.Short() {
+		t.Skip("short")
+	}
 
 	rng, _ := randutil.NewPseudoRand()
 	testCases := allPartitioningTests(rng)
 
 	ctx := context.Background()
-	sqlDB, cleanup := setupPartitioningTestCluster(ctx, t)
+	db, sqlDB, cleanup := setupPartitioningTestCluster(ctx, t)
 	defer cleanup()
 
 	for _, test := range testCases {
@@ -1156,7 +1182,7 @@ func TestInitialPartitioning(t *testing.T) {
 			sqlDB.Exec(t, test.parsed.createStmt)
 			sqlDB.Exec(t, test.parsed.zoneConfigStmts)
 
-			testutils.SucceedsSoon(t, test.verifyScansFn(ctx, sqlDB.DB))
+			testutils.SucceedsSoon(t, test.verifyScansFn(ctx, db))
 		})
 	}
 }
@@ -1239,7 +1265,11 @@ func TestSelectPartitionExprs(t *testing.T) {
 func TestRepartitioning(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	t.Skip("#28786")
+	// This test configures many sub-tests and is too slow to run under nightly
+	// race stress.
+	if testutils.NightlyStress() && util.RaceEnabled {
+		t.Skip()
+	}
 
 	rng, _ := randutil.NewPseudoRand()
 	testCases, err := allRepartitioningTests(allPartitioningTests(rng))
@@ -1248,7 +1278,7 @@ func TestRepartitioning(t *testing.T) {
 	}
 
 	ctx := context.Background()
-	sqlDB, cleanup := setupPartitioningTestCluster(ctx, t)
+	db, sqlDB, cleanup := setupPartitioningTestCluster(ctx, t)
 	defer cleanup()
 
 	for _, test := range testCases {
@@ -1263,7 +1293,7 @@ func TestRepartitioning(t *testing.T) {
 				sqlDB.Exec(t, test.old.parsed.createStmt)
 				sqlDB.Exec(t, test.old.parsed.zoneConfigStmts)
 
-				testutils.SucceedsSoon(t, test.old.verifyScansFn(ctx, sqlDB.DB))
+				testutils.SucceedsSoon(t, test.old.verifyScansFn(ctx, db))
 			}
 
 			{
@@ -1287,15 +1317,13 @@ func TestRepartitioning(t *testing.T) {
 					repartition.WriteString(`PARTITION BY NOTHING`)
 				} else {
 					if err := sql.ShowCreatePartitioning(
-						&sqlbase.DatumAlloc{}, test.new.parsed.tableDesc, &testIndex,
+						&sqlbase.DatumAlloc{}, test.new.parsed.tableDesc, testIndex,
 						&testIndex.Partitioning, &repartition, 0 /* indent */, 0, /* colOffset */
 					); err != nil {
 						t.Fatalf("%+v", err)
 					}
 				}
-				if _, err := sqlDB.DB.Exec(repartition.String()); err != nil {
-					t.Fatalf("%s: %+v", repartition.String(), err)
-				}
+				sqlDB.Exec(t, repartition.String())
 
 				// Verify that repartitioning removes zone configs for partitions that
 				// have been removed.
@@ -1313,11 +1341,7 @@ func TestRepartitioning(t *testing.T) {
 						// Ignore zone configs that target databases or system ranges.
 						continue
 					}
-					tn, err := zs.TableOrIndex.Table.Normalize()
-					if err != nil {
-						t.Fatal(err)
-					}
-					if tn.Table() != test.new.parsed.tableDesc.Name || zs.Partition == "" {
+					if zs.TableOrIndex.Table.Table() != test.new.parsed.tableDesc.Name || zs.Partition == "" {
 						// Ignore zone configs that do not target a partition of this table.
 						continue
 					}
@@ -1333,7 +1357,7 @@ func TestRepartitioning(t *testing.T) {
 				// does not apply a new zone config). This is fine.
 				sqlDB.Exec(t, test.new.parsed.zoneConfigStmts)
 
-				testutils.SucceedsSoon(t, test.new.verifyScansFn(ctx, sqlDB.DB))
+				testutils.SucceedsSoon(t, test.new.verifyScansFn(ctx, db))
 			}
 		})
 	}
@@ -1370,9 +1394,7 @@ func TestRemovePartitioningExpiredLicense(t *testing.T) {
 	const zoneErr = "use of replication zones on indexes or partitions requires an enterprise license"
 	expectErr := func(q string, expErr string) {
 		t.Helper()
-		if _, err := sqlDB.DB.Exec(q); !testutils.IsError(err, expErr) {
-			t.Fatalf("expected error %q, but got %+v", expErr, err)
-		}
+		sqlDB.ExpectErr(t, expErr, q)
 	}
 
 	// Partitions and zone configs cannot be modified without a valid license.
