@@ -12,6 +12,7 @@ import (
 	"context"
 	"crypto/tls"
 	"crypto/x509"
+	"encoding/base64"
 	"fmt"
 	"hash/fnv"
 	"io"
@@ -21,26 +22,27 @@ import (
 	"os"
 	"path"
 	"path/filepath"
+	"strconv"
 	"strings"
 	"time"
 
-	"google.golang.org/api/option"
-
 	gcs "cloud.google.com/go/storage"
 	"github.com/Azure/azure-storage-blob-go/azblob"
 	"github.com/aws/aws-sdk-go/aws"
 	"github.com/aws/aws-sdk-go/aws/session"
 	"github.com/aws/aws-sdk-go/service/s3"
 	"github.com/aws/aws-sdk-go/service/s3/s3manager"
-	"github.com/pkg/errors"
-	"golang.org/x/oauth2/google"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
+	"github.com/cockroachdb/cockroach/pkg/workload"
+	"github.com/pkg/errors"
+	"golang.org/x/oauth2/google"
+	"google.golang.org/api/option"
 )
 
 const (
@@ -66,9 +68,14 @@ const (
 
 	// AuthParam is the query parameter for the cluster settings named
 	// key in a URI.
-	AuthParam         = "AUTH"
-	authParamImplicit = "implicit"
-	authParamDefault  = "default"
+	AuthParam          = "AUTH"
+	authParamImplicit  = "implicit"
+	authParamDefault   = "default"
+	authParamSpecified = "specified"
+
+	// CredentialsParam is the query parameter for the base64-encoded contents of
+	// the Google Application Credentials JSON file.
+	CredentialsParam = "CREDENTIALS"
 
 	cloudstoragePrefix = "cloudstorage"
 	cloudstorageGS     = cloudstoragePrefix + ".gs"
@@ -126,6 +133,7 @@ func ExportStorageConfFromURI(path string) (roachpb.ExportStorage, error) {
 			Prefix:         uri.Path,
 			Auth:           uri.Query().Get(AuthParam),
 			BillingProject: uri.Query().Get(GoogleBillingProjectParam),
+			Credentials:    uri.Query().Get(CredentialsParam),
 		}
 		conf.GoogleCloudConfig.Prefix = strings.TrimLeft(conf.GoogleCloudConfig.Prefix, "/")
 	case "azure":
@@ -147,11 +155,18 @@ func ExportStorageConfFromURI(path string) (roachpb.ExportStorage, error) {
 		conf.Provider = roachpb.ExportStorageProvider_Http
 		conf.HttpPath.BaseUri = path
 	case "nodelocal":
-		if uri.Host != "" {
-			return conf, errors.Errorf("nodelocal does not support hosts: %s", path)
+		nodeID, err := strconv.Atoi(uri.Host)
+		if err != nil && uri.Host != "" {
+			return conf, errors.Errorf("host component of nodelocal URI must be a node ID: %s", path)
 		}
 		conf.Provider = roachpb.ExportStorageProvider_LocalFile
 		conf.LocalFile.Path = uri.Path
+		conf.LocalFile.NodeID = roachpb.NodeID(nodeID)
+	case "experimental-workload":
+		conf.Provider = roachpb.ExportStorageProvider_Workload
+		if conf.WorkloadConfig, err = ParseWorkloadConfig(uri); err != nil {
+			return conf, err
+		}
 	default:
 		return conf, errors.Errorf("unsupported storage scheme: %q", uri.Scheme)
 	}
@@ -176,6 +191,9 @@ func SanitizeExportStorageURI(path string) (string, error) {
 	if err != nil {
 		return "", err
 	}
+	if uri.Scheme == "experimental-workload" {
+		return path, nil
+	}
 	// All current export storage providers store credentials in the query string,
 	// if they store it in the URI at all.
 	uri.RawQuery = ""
@@ -189,7 +207,7 @@ func MakeExportStorage(
 	switch dest.Provider {
 	case roachpb.ExportStorageProvider_LocalFile:
 		telemetry.Count("external-io.nodelocal")
-		return makeLocalStorage(dest.LocalFile.Path, settings)
+		return makeLocalStorage(dest.LocalFile, settings)
 	case roachpb.ExportStorageProvider_Http:
 		telemetry.Count("external-io.http")
 		return makeHTTPStorage(dest.HttpPath.BaseUri, settings)
@@ -202,6 +220,14 @@ func MakeExportStorage(
 	case roachpb.ExportStorageProvider_Azure:
 		telemetry.Count("external-io.azure")
 		return makeAzureStorage(dest.AzureConfig, settings)
+	case roachpb.ExportStorageProvider_Workload:
+		if err := settings.Version.CheckVersion(
+			cluster.VersionExportStorageWorkload, "experimental-workload",
+		); err != nil {
+			return nil, err
+		}
+		telemetry.Count("external-io.workload")
+		return makeWorkloadStorage(dest.WorkloadConfig)
 	}
 	return nil, errors.Errorf("unsupported export destination type: %s", dest.Provider.String())
 }
@@ -255,8 +281,8 @@ var (
 )
 
 type localFileStorage struct {
-	rawBase string // un-prefixed base -- DO NOT use for I/O ops.
-	base    string // the prefixed base, for I/O ops on this node.
+	cfg  roachpb.ExportStorage_LocalFilePath // constains un-prefixed base -- DO NOT use for I/O ops.
+	base string                              // the prefixed base, for I/O ops on this node.
 }
 
 var _ ExportStorage = &localFileStorage{}
@@ -271,12 +297,15 @@ func MakeLocalStorageURI(path string) (string, error) {
 	return fmt.Sprintf("nodelocal://%s", path), nil
 }
 
-func makeLocalStorage(base string, settings *cluster.Settings) (ExportStorage, error) {
-	if base == "" {
+func makeLocalStorage(
+	cfg roachpb.ExportStorage_LocalFilePath, settings *cluster.Settings,
+) (ExportStorage, error) {
+	if cfg.Path == "" {
 		return nil, errors.Errorf("Local storage requested but path not provided")
 	}
+	// TODO(dt): check that this node is cfg.NodeID if non-zero.
 
-	localBase := base
+	localBase := cfg.Path
 	// In non-server execution we have no settings and no restriction on local IO.
 	if settings != nil {
 		if settings.ExternalIODir == "" {
@@ -289,15 +318,13 @@ func makeLocalStorage(base string, settings *cluster.Settings) (ExportStorage, e
 			return nil, errors.Errorf("local file access to paths outside of external-io-dir is not allowed")
 		}
 	}
-	return &localFileStorage{base: localBase, rawBase: base}, nil
+	return &localFileStorage{base: localBase, cfg: cfg}, nil
 }
 
 func (l *localFileStorage) Conf() roachpb.ExportStorage {
 	return roachpb.ExportStorage{
-		Provider: roachpb.ExportStorageProvider_LocalFile,
-		LocalFile: roachpb.ExportStorage_LocalFilePath{
-			Path: l.rawBase,
-		},
+		Provider:  roachpb.ExportStorageProvider_LocalFile,
+		LocalFile: l.cfg,
 	}
 }
 
@@ -308,16 +335,20 @@ func (l *localFileStorage) WriteFile(
 	if err := os.MkdirAll(filepath.Dir(p), 0755); err != nil {
 		return errors.Wrap(err, "creating local export storage path")
 	}
-	f, err := os.Create(p)
+	tmpP := p + `.tmp`
+	f, err := os.Create(tmpP)
 	if err != nil {
-		return errors.Wrapf(err, "creating local export file %q", p)
+		return errors.Wrapf(err, "creating local export tmp file %q", tmpP)
 	}
 	defer f.Close()
 	_, err = io.Copy(f, content)
 	if err != nil {
-		return errors.Wrapf(err, "writing to local export file %q", p)
+		return errors.Wrapf(err, "writing to local export tmp file %q", tmpP)
 	}
-	return f.Sync()
+	if err := f.Sync(); err != nil {
+		return errors.Wrapf(err, "syncing to local export tmp file %q", tmpP)
+	}
+	return errors.Wrapf(os.Rename(tmpP, p), "renaming to local export file %q", p)
 }
 
 func (l *localFileStorage) ReadFile(_ context.Context, basename string) (io.ReadCloser, error) {
@@ -419,24 +450,29 @@ func (h *httpStorage) ReadFile(ctx context.Context, basename string) (io.ReadClo
 }
 
 func (h *httpStorage) WriteFile(ctx context.Context, basename string, content io.ReadSeeker) error {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&h.settings.SV))
-	defer cancel()
-	_, err := h.reqNoBody(ctx, "PUT", basename, content)
-	return err
+	return contextutil.RunWithTimeout(ctx, fmt.Sprintf("PUT %s", basename),
+		timeoutSetting.Get(&h.settings.SV), func(ctx context.Context) error {
+			_, err := h.reqNoBody(ctx, "PUT", basename, content)
+			return err
+		})
 }
 
 func (h *httpStorage) Delete(ctx context.Context, basename string) error {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&h.settings.SV))
-	defer cancel()
-	_, err := h.reqNoBody(ctx, "DELETE", basename, nil)
-	return err
+	return contextutil.RunWithTimeout(ctx, fmt.Sprintf("DELETE %s", basename),
+		timeoutSetting.Get(&h.settings.SV), func(ctx context.Context) error {
+			_, err := h.reqNoBody(ctx, "DELETE", basename, nil)
+			return err
+		})
 }
 
 func (h *httpStorage) Size(ctx context.Context, basename string) (int64, error) {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&h.settings.SV))
-	defer cancel()
-	resp, err := h.reqNoBody(ctx, "HEAD", basename, nil)
-	if err != nil {
+	var resp *http.Response
+	if err := contextutil.RunWithTimeout(ctx, fmt.Sprintf("HEAD %s", basename),
+		timeoutSetting.Get(&h.settings.SV), func(ctx context.Context) error {
+			var err error
+			resp, err = h.reqNoBody(ctx, "HEAD", basename, nil)
+			return err
+		}); err != nil {
 		return 0, err
 	}
 	if resp.ContentLength < 0 {
@@ -483,7 +519,7 @@ func (h *httpStorage) req(
 	}
 	resp, err := h.client.Do(req)
 	if err != nil {
-		return nil, errors.Wrapf(err, "error exeucting request %s %q", method, url)
+		return nil, errors.Wrapf(err, "error executing request %s %q", method, url)
 	}
 	switch resp.StatusCode {
 	case 200, 201, 204:
@@ -594,13 +630,16 @@ func (s *s3Storage) Conf() roachpb.ExportStorage {
 }
 
 func (s *s3Storage) WriteFile(ctx context.Context, basename string, content io.ReadSeeker) error {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&s.settings.SV))
-	defer cancel()
-	_, err := s.s3.PutObjectWithContext(ctx, &s3.PutObjectInput{
-		Bucket: s.bucket,
-		Key:    aws.String(path.Join(s.prefix, basename)),
-		Body:   content,
-	})
+	err := contextutil.RunWithTimeout(ctx, "put s3 object",
+		timeoutSetting.Get(&s.settings.SV),
+		func(ctx context.Context) error {
+			_, err := s.s3.PutObjectWithContext(ctx, &s3.PutObjectInput{
+				Bucket: s.bucket,
+				Key:    aws.String(path.Join(s.prefix, basename)),
+				Body:   content,
+			})
+			return err
+		})
 	return errors.Wrap(err, "failed to put s3 object")
 }
 
@@ -617,22 +656,29 @@ func (s *s3Storage) ReadFile(ctx context.Context, basename string) (io.ReadClose
 }
 
 func (s *s3Storage) Delete(ctx context.Context, basename string) error {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&s.settings.SV))
-	defer cancel()
-	_, err := s.s3.DeleteObjectWithContext(ctx, &s3.DeleteObjectInput{
-		Bucket: s.bucket,
-		Key:    aws.String(path.Join(s.prefix, basename)),
-	})
-	return err
+	return contextutil.RunWithTimeout(ctx, "delete s3 object",
+		timeoutSetting.Get(&s.settings.SV),
+		func(ctx context.Context) error {
+			_, err := s.s3.DeleteObjectWithContext(ctx, &s3.DeleteObjectInput{
+				Bucket: s.bucket,
+				Key:    aws.String(path.Join(s.prefix, basename)),
+			})
+			return err
+		})
 }
 
 func (s *s3Storage) Size(ctx context.Context, basename string) (int64, error) {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&s.settings.SV))
-	defer cancel()
-	out, err := s.s3.HeadObjectWithContext(ctx, &s3.HeadObjectInput{
-		Bucket: s.bucket,
-		Key:    aws.String(path.Join(s.prefix, basename)),
-	})
+	var out *s3.HeadObjectOutput
+	err := contextutil.RunWithTimeout(ctx, "get s3 object header",
+		timeoutSetting.Get(&s.settings.SV),
+		func(ctx context.Context) error {
+			var err error
+			out, err = s.s3.HeadObjectWithContext(ctx, &s3.HeadObjectInput{
+				Bucket: s.bucket,
+				Key:    aws.String(path.Join(s.prefix, basename)),
+			})
+			return err
+		})
 	if err != nil {
 		return 0, errors.Wrap(err, "failed to get s3 object headers")
 	}
@@ -670,6 +716,7 @@ func makeGCSStorage(
 	opts := []option.ClientOption{option.WithScopes(scope)}
 
 	// "default": only use the key in the settings; error if not present.
+	// "specified": the JSON object for authentication is given by the CREDENTIALS param.
 	// "implicit": only use the environment data.
 	// "": if default key is in the settings use it; otherwise use environment data.
 	switch conf.Auth {
@@ -689,6 +736,24 @@ func makeGCSStorage(
 			}
 			opts = append(opts, option.WithTokenSource(source.TokenSource(ctx)))
 		}
+	case authParamSpecified:
+		if conf.Credentials == "" {
+			return nil, errors.Errorf(
+				"%s is set to '%s', but %s is not set",
+				AuthParam,
+				authParamSpecified,
+				CredentialsParam,
+			)
+		}
+		decodedKey, err := base64.StdEncoding.DecodeString(conf.Credentials)
+		if err != nil {
+			return nil, errors.Wrap(err, fmt.Sprintf("decoding value of %s", CredentialsParam))
+		}
+		source, err := google.JWTConfigFromJSON(decodedKey, scope)
+		if err != nil {
+			return nil, errors.Wrap(err, "creating GCS oauth token source from specified credentials")
+		}
+		opts = append(opts, option.WithTokenSource(source.TokenSource(ctx)))
 	case authParamImplicit:
 		// Do nothing; use implicit params:
 		// https://godoc.org/golang.org/x/oauth2/google#FindDefaultCredentials
@@ -715,18 +780,19 @@ func makeGCSStorage(
 func (g *gcsStorage) WriteFile(ctx context.Context, basename string, content io.ReadSeeker) error {
 	const maxAttempts = 3
 	err := retry.WithMaxAttempts(ctx, base.DefaultRetryOptions(), maxAttempts, func() error {
-		// Set the timeout within the retry loop.
-		deadlineCtx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&g.settings.SV))
-		defer cancel()
 		if _, err := content.Seek(0, io.SeekStart); err != nil {
 			return err
 		}
-		w := g.bucket.Object(path.Join(g.prefix, basename)).NewWriter(deadlineCtx)
-		if _, err := io.Copy(w, content); err != nil {
-			_ = w.Close()
-			return err
-		}
-		return w.Close()
+		// Set the timeout within the retry loop.
+		return contextutil.RunWithTimeout(ctx, "put gcs file", timeoutSetting.Get(&g.settings.SV),
+			func(ctx context.Context) error {
+				w := g.bucket.Object(path.Join(g.prefix, basename)).NewWriter(ctx)
+				if _, err := io.Copy(w, content); err != nil {
+					_ = w.Close()
+					return err
+				}
+				return w.Close()
+			})
 	})
 	return errors.Wrap(err, "write to google cloud")
 }
@@ -744,19 +810,25 @@ func (g *gcsStorage) ReadFile(ctx context.Context, basename string) (io.ReadClos
 }
 
 func (g *gcsStorage) Delete(ctx context.Context, basename string) error {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&g.settings.SV))
-	defer cancel()
-	return g.bucket.Object(path.Join(g.prefix, basename)).Delete(ctx)
+	return contextutil.RunWithTimeout(ctx, "delete gcs file",
+		timeoutSetting.Get(&g.settings.SV),
+		func(ctx context.Context) error {
+			return g.bucket.Object(path.Join(g.prefix, basename)).Delete(ctx)
+		})
 }
 
 func (g *gcsStorage) Size(ctx context.Context, basename string) (int64, error) {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&g.settings.SV))
-	defer cancel()
-	r, err := g.bucket.Object(path.Join(g.prefix, basename)).NewReader(ctx)
-	if err != nil {
+	var r *gcs.Reader
+	if err := contextutil.RunWithTimeout(ctx, "size gcs file",
+		timeoutSetting.Get(&g.settings.SV),
+		func(ctx context.Context) error {
+			var err error
+			r, err = g.bucket.Object(path.Join(g.prefix, basename)).NewReader(ctx)
+			return err
+		}); err != nil {
 		return 0, err
 	}
-	sz := r.Size()
+	sz := r.Attrs.Size
 	_ = r.Close()
 	return sz, nil
 }
@@ -813,10 +885,12 @@ func (s *azureStorage) Conf() roachpb.ExportStorage {
 func (s *azureStorage) WriteFile(
 	ctx context.Context, basename string, content io.ReadSeeker,
 ) error {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&s.settings.SV))
-	defer cancel()
-	blob := s.getBlob(basename)
-	_, err := blob.Upload(ctx, content, azblob.BlobHTTPHeaders{}, azblob.Metadata{}, azblob.BlobAccessConditions{})
+	err := contextutil.RunWithTimeout(ctx, "write azure file", timeoutSetting.Get(&s.settings.SV),
+		func(ctx context.Context) error {
+			blob := s.getBlob(basename)
+			_, err := blob.Upload(ctx, content, azblob.BlobHTTPHeaders{}, azblob.Metadata{}, azblob.BlobAccessConditions{})
+			return err
+		})
 	return errors.Wrapf(err, "write file: %s", basename)
 }
 
@@ -832,18 +906,24 @@ func (s *azureStorage) ReadFile(ctx context.Context, basename string) (io.ReadCl
 }
 
 func (s *azureStorage) Delete(ctx context.Context, basename string) error {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&s.settings.SV))
-	defer cancel()
-	blob := s.getBlob(basename)
-	_, err := blob.Delete(ctx, azblob.DeleteSnapshotsOptionNone, azblob.BlobAccessConditions{})
+	err := contextutil.RunWithTimeout(ctx, "delete azure file", timeoutSetting.Get(&s.settings.SV),
+		func(ctx context.Context) error {
+			blob := s.getBlob(basename)
+			_, err := blob.Delete(ctx, azblob.DeleteSnapshotsOptionNone, azblob.BlobAccessConditions{})
+			return err
+		})
 	return errors.Wrap(err, "delete file")
 }
 
 func (s *azureStorage) Size(ctx context.Context, basename string) (int64, error) {
-	ctx, cancel := context.WithTimeout(ctx, timeoutSetting.Get(&s.settings.SV))
-	defer cancel()
-	blob := s.getBlob(basename)
-	props, err := blob.GetProperties(ctx, azblob.BlobAccessConditions{})
+	var props *azblob.BlobGetPropertiesResponse
+	err := contextutil.RunWithTimeout(ctx, "size azure file", timeoutSetting.Get(&s.settings.SV),
+		func(ctx context.Context) error {
+			blob := s.getBlob(basename)
+			var err error
+			props, err = blob.GetProperties(ctx, azblob.BlobAccessConditions{})
+			return err
+		})
 	if err != nil {
 		return 0, errors.Wrap(err, "get file properties")
 	}
@@ -853,3 +933,115 @@ func (s *azureStorage) Size(ctx context.Context, basename string) (int64, error)
 func (s *azureStorage) Close() error {
 	return nil
 }
+
+// ParseWorkloadConfig parses a workload config URI to a proto config.
+func ParseWorkloadConfig(uri *url.URL) (*roachpb.ExportStorage_Workload, error) {
+	c := &roachpb.ExportStorage_Workload{}
+	pathParts := strings.Split(strings.Trim(uri.Path, `/`), `/`)
+	if len(pathParts) != 3 {
+		return nil, errors.Errorf(
+			`path must be of the form /<format>/<generator>/<table>: %s`, uri.Path)
+	}
+	c.Format, c.Generator, c.Table = pathParts[0], pathParts[1], pathParts[2]
+	q := uri.Query()
+	if _, ok := q[`version`]; !ok {
+		return nil, errors.New(`parameter version is required`)
+	}
+	c.Version = q.Get(`version`)
+	q.Del(`version`)
+	if s := q.Get(`row-start`); len(s) > 0 {
+		q.Del(`row-start`)
+		var err error
+		if c.BatchBegin, err = strconv.ParseInt(s, 10, 64); err != nil {
+			return nil, err
+		}
+	}
+	if e := q.Get(`row-end`); len(e) > 0 {
+		q.Del(`row-end`)
+		var err error
+		if c.BatchEnd, err = strconv.ParseInt(e, 10, 64); err != nil {
+			return nil, err
+		}
+	}
+	for k, vs := range q {
+		for _, v := range vs {
+			c.Flags = append(c.Flags, `--`+k+`=`+v)
+		}
+	}
+	return c, nil
+}
+
+type workloadStorage struct {
+	conf  *roachpb.ExportStorage_Workload
+	gen   workload.Generator
+	table workload.Table
+}
+
+var _ ExportStorage = &workloadStorage{}
+
+func makeWorkloadStorage(conf *roachpb.ExportStorage_Workload) (ExportStorage, error) {
+	if conf == nil {
+		return nil, errors.Errorf("workload upload requested but info missing")
+	}
+	if strings.ToLower(conf.Format) != `csv` {
+		return nil, errors.Errorf(`unsupported format: %s`, conf.Format)
+	}
+	meta, err := workload.Get(conf.Generator)
+	if err != nil {
+		return nil, err
+	}
+	// Different versions of the workload could generate different data, so
+	// disallow this.
+	if meta.Version != conf.Version {
+		return nil, errors.Errorf(
+			`expected %s version "%s" but got "%s"`, meta.Name, conf.Version, meta.Version)
+	}
+	gen := meta.New()
+	if f, ok := gen.(workload.Flagser); ok {
+		if err := f.Flags().Parse(conf.Flags); err != nil {
+			return nil, errors.Wrapf(err, `parsing parameters %s`, strings.Join(conf.Flags, ` `))
+		}
+	}
+	s := &workloadStorage{
+		conf: conf,
+		gen:  gen,
+	}
+	for _, t := range gen.Tables() {
+		if t.Name == conf.Table {
+			s.table = t
+			break
+		}
+	}
+	if s.table.Name == `` {
+		return nil, errors.Wrapf(err, `unknown table %s for generator %s`, conf.Table, meta.Name)
+	}
+	return s, nil
+}
+
+func (s *workloadStorage) Conf() roachpb.ExportStorage {
+	return roachpb.ExportStorage{
+		Provider:       roachpb.ExportStorageProvider_Workload,
+		WorkloadConfig: s.conf,
+	}
+}
+
+func (s *workloadStorage) ReadFile(_ context.Context, basename string) (io.ReadCloser, error) {
+	if basename != `` {
+		return nil, errors.New(`basenames are not supported by workload storage`)
+	}
+	r := workload.NewCSVRowsReader(s.table, int(s.conf.BatchBegin), int(s.conf.BatchEnd))
+	return ioutil.NopCloser(r), nil
+}
+
+func (s *workloadStorage) WriteFile(_ context.Context, _ string, _ io.ReadSeeker) error {
+	return errors.Errorf(`workload storage does not support writes`)
+}
+func (s *workloadStorage) Delete(_ context.Context, _ string) error {
+	return errors.Errorf(`workload storage does not support writes`)
+}
+func (s *workloadStorage) Size(_ context.Context, _ string) (int64, error) {
+	return 0, errors.Errorf(`workload storage does not support sizing`)
+}
+func (s *workloadStorage) Close() error {
+	return nil
+}
