@@ -26,13 +26,12 @@ import (
 	"path/filepath"
 	"reflect"
 	"regexp"
+	"sort"
 	"strconv"
 	"strings"
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
@@ -46,9 +45,10 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
-
 	// register some workloads for TestWorkload
 	_ "github.com/cockroachdb/cockroach/pkg/workload/examples"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/assert"
 )
 
 type cliTest struct {
@@ -160,7 +160,7 @@ func newCLITest(params cliTestParams) cliTest {
 func setCLIDefaultsForTests() {
 	initCLIDefaults()
 	cliCtx.terminalOutput = false
-	cliCtx.showTimes = false
+	sqlCtx.showTimes = false
 	// Even though we pretend there is no terminal, most tests want
 	// pretty tables.
 	cliCtx.tableDisplayFormat = tableDisplayTable
@@ -513,7 +513,7 @@ func Example_zone() {
 	// system
 	// system.jobs
 	// zone set system --file=./testdata/zone_attrs.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 67108864
 	// gc:
 	//   ttlseconds: 90000
@@ -530,7 +530,7 @@ func Example_zone() {
 	// system.jobs
 	// zone get .liveness
 	// .liveness
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 67108864
 	// gc:
 	//   ttlseconds: 600
@@ -539,7 +539,7 @@ func Example_zone() {
 	// lease_preferences: []
 	// zone get .meta
 	// .meta
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 67108864
 	// gc:
 	//   ttlseconds: 3600
@@ -550,7 +550,7 @@ func Example_zone() {
 	// pq: relation "system.public.nonexistent" does not exist
 	// zone get system.descriptor
 	// system
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 67108864
 	// gc:
 	//   ttlseconds: 90000
@@ -564,7 +564,7 @@ func Example_zone() {
 	// zone set system.nonexistent --file=./testdata/zone_attrs.yaml
 	// pq: relation "system.public.nonexistent" does not exist
 	// zone set system --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -574,7 +574,7 @@ func Example_zone() {
 	//
 	// zone get system
 	// system
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -591,7 +591,7 @@ func Example_zone() {
 	// zone rm .default
 	// pq: cannot remove default zone
 	// zone set .liveness --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 600
@@ -600,7 +600,7 @@ func Example_zone() {
 	// lease_preferences: []
 	//
 	// zone set .meta --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 3600
@@ -609,7 +609,7 @@ func Example_zone() {
 	// lease_preferences: []
 	//
 	// zone set .system --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -618,7 +618,7 @@ func Example_zone() {
 	// lease_preferences: []
 	//
 	// zone set .timeseries --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -628,7 +628,7 @@ func Example_zone() {
 	//
 	// zone get .system
 	// .system
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -643,7 +643,7 @@ func Example_zone() {
 	// .timeseries
 	// system.jobs
 	// zone set .default --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -653,7 +653,7 @@ func Example_zone() {
 	//
 	// zone get system
 	// .default
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -661,7 +661,7 @@ func Example_zone() {
 	// constraints: []
 	// lease_preferences: []
 	// zone set .default --disable-replication
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -671,7 +671,7 @@ func Example_zone() {
 	//
 	// zone get system
 	// .default
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -696,7 +696,7 @@ func Example_zone() {
 	// zone set system.jobs@primary --file=./testdata/zone_attrs.yaml
 	// pq: setting zone configs on indexes or partitions requires a CCL binary
 	// zone set system --file=./testdata/zone_attrs_advanced.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -705,7 +705,7 @@ func Example_zone() {
 	// lease_preferences: [[+region=us-east-1], [+zone=us-east-1a]]
 	//
 	// zone set system --file=./testdata/zone_attrs_experimental.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -716,7 +716,7 @@ func Example_zone() {
 	// sql -e create database t; create table t.f (x int, y int)
 	// CREATE TABLE
 	// zone set t --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -730,7 +730,7 @@ func Example_zone() {
 	// system.jobs
 	// t
 	// zone set t.f --file=./testdata/zone_range_max_bytes.yaml
-	// range_min_bytes: 1048576
+	// range_min_bytes: 16777216
 	// range_max_bytes: 134217728
 	// gc:
 	//   ttlseconds: 90000
@@ -1383,7 +1383,7 @@ func Example_sql_table() {
 	// sql -e insert into t.t values (e'a\tb\tc\n12\t123123213\t12313', 'tabs')
 	// INSERT 1
 	// sql -e insert into t.t values (e'\xc3\x28', 'non-UTF8 string')
-	// pq: invalid UTF-8 byte sequence
+	// pq: lexical error: invalid UTF-8 byte sequence
 	// DETAIL: source SQL:
 	// insert into t.t values (e'\xc3\x28', 'non-UTF8 string')
 	//                         ^
@@ -1621,7 +1621,9 @@ func TestRenderHTML(t *testing.T) {
 		name := fmt.Sprintf("escape=%v/rowStats=%v", tc.reporter.escape, tc.reporter.rowStats)
 		t.Run(name, func(t *testing.T) {
 			var buf bytes.Buffer
-			err := render(&tc.reporter, &buf, cols, newRowSliceIter(rows, align), nil /* noRowsHook */)
+			err := render(&tc.reporter, &buf,
+				cols, newRowSliceIter(rows, align),
+				nil /* completedHook */, nil /* noRowsHook */)
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -1817,6 +1819,7 @@ Available Commands:
   debug       debugging commands
   sqlfmt      format SQL statements
   workload    [experimental] generators for data and query loads
+  systembench Run systembench
   help        Help about any command
 
 Flags:
@@ -1878,9 +1881,7 @@ Use "cockroach [command] --help" for more information about a command.
 			}
 			got := strings.Join(final, "\n")
 
-			if got != test.expected {
-				t.Errorf("got:\n%s\n----\nexpected:\n%s", got, test.expected)
-			}
+			assert.Equal(t, test.expected, got)
 		})
 	}
 }
@@ -2068,7 +2069,7 @@ func checkNodeStatus(t *testing.T, c cliTest, output string, start time.Time) {
 	// address. They don't need closer checks.
 	baseIdx := len(baseNodeColumnHeaders)
 
-	// Adding fields that need verification for --range flag.
+	// Adding fields that need verification for --ranges flag.
 	// We have to allow up to 1 unavailable/underreplicated range because
 	// sometimes we run the `node status` command before the server has fully
 	// initialized itself and it doesn't consider itself live yet. In such cases,
@@ -2076,8 +2077,8 @@ func checkNodeStatus(t *testing.T, c cliTest, output string, start time.Time) {
 	// have been able to do any splits yet.
 	if nodeCtx.statusShowRanges || nodeCtx.statusShowAll {
 		testcases = append(testcases,
-			testCase{"leader_ranges", baseIdx, 3},
-			testCase{"leaseholder_ranges", baseIdx + 1, 3},
+			testCase{"leader_ranges", baseIdx, 22},
+			testCase{"leaseholder_ranges", baseIdx + 1, 22},
 			testCase{"ranges", baseIdx + 2, 22},
 			testCase{"unavailable_ranges", baseIdx + 3, 1},
 			testCase{"underreplicated_ranges", baseIdx + 4, 1},
@@ -2287,10 +2288,65 @@ func TestJunkPositionalArguments(t *testing.T) {
 	}
 }
 
+// TestZipContainsAllInternalTables verifies that we don't add new internal tables
+// without also taking them into account in a `debug zip`. If this test fails,
+// add your table to either of the []string slices referenced in the test (which
+// are used by `debug zip`) or add it as an exception after having verified that
+// it indeed should not be collected (this is rare).
+// NB: if you're adding a new one, you'll also have to update TestZip.
+func TestZipContainsAllInternalTables(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	s, db, _ := serverutils.StartServer(t, base.TestServerArgs{})
+	defer s.Stopper().Stop(context.Background())
+
+	rows, err := db.Query(`
+SELECT concat('crdb_internal.', table_name) as name FROM [ SHOW TABLES FROM crdb_internal ] WHERE
+    table_name NOT IN (
+-- whitelisted tables that don't need to be in debug zip
+'backward_dependencies',
+'builtin_functions',
+'create_statements',
+'forward_dependencies',
+'index_columns',
+'table_columns',
+'table_indexes',
+'ranges',
+'ranges_no_leases',
+'predefined_comments',
+'session_trace',
+'session_variables',
+'tables'
+)
+ORDER BY name ASC`)
+	assert.NoError(t, err)
+
+	var tables []string
+	for rows.Next() {
+		var table string
+		assert.NoError(t, rows.Scan(&table))
+		tables = append(tables, table)
+	}
+
+	var exp []string
+	exp = append(exp, debugZipTablesPerNode...)
+	exp = append(exp, debugZipTablesPerCluster...)
+	sort.Strings(exp)
+
+	assert.Equal(t, exp, tables)
+}
+
 func TestZip(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	c := newCLITest(cliTestParams{})
+	dir, cleanupFn := testutils.TempDir(t)
+	defer cleanupFn()
+
+	c := newCLITest(cliTestParams{
+		storeSpecs: []base.StoreSpec{{
+			Path: dir,
+		}},
+	})
 	defer c.cleanup()
 
 	out, err := c.RunWithCapture("debug zip " + os.DevNull)
@@ -2300,64 +2356,79 @@ func TestZip(t *testing.T) {
 
 	const expected = `debug zip ` + os.DevNull + `
 writing ` + os.DevNull + `
-  debug/events
-  debug/rangelog
-  debug/liveness
-  debug/settings
-  debug/gossip/liveness
-  debug/gossip/network
-  debug/gossip/nodes
-  debug/metrics
-  debug/alerts
-  debug/nodes/1/status
-  debug/nodes/1/gossip
-  debug/nodes/1/stacks
-  debug/nodes/1/heap
-  debug/nodes/1/ranges/1
-  debug/nodes/1/ranges/2
-  debug/nodes/1/ranges/3
-  debug/nodes/1/ranges/4
-  debug/nodes/1/ranges/5
-  debug/nodes/1/ranges/6
-  debug/nodes/1/ranges/7
-  debug/nodes/1/ranges/8
-  debug/nodes/1/ranges/9
-  debug/nodes/1/ranges/10
-  debug/nodes/1/ranges/11
-  debug/nodes/1/ranges/12
-  debug/nodes/1/ranges/13
-  debug/nodes/1/ranges/14
-  debug/nodes/1/ranges/15
-  debug/nodes/1/ranges/16
-  debug/nodes/1/ranges/17
-  debug/nodes/1/ranges/18
-  debug/nodes/1/ranges/19
-  debug/nodes/1/ranges/20
-  debug/nodes/1/ranges/21
-  debug/nodes/1/ranges/22
-  debug/reports/problemranges
-  debug/schema/defaultdb@details
-  debug/schema/postgres@details
-  debug/schema/system@details
-  debug/schema/system/descriptor
-  debug/schema/system/eventlog
-  debug/schema/system/jobs
-  debug/schema/system/lease
-  debug/schema/system/locations
-  debug/schema/system/namespace
-  debug/schema/system/rangelog
-  debug/schema/system/role_members
-  debug/schema/system/settings
-  debug/schema/system/table_statistics
-  debug/schema/system/ui
-  debug/schema/system/users
-  debug/schema/system/web_sessions
-  debug/schema/system/zones
+  debug/events.json
+  debug/rangelog.json
+  debug/liveness.json
+  debug/settings.json
+  debug/reports/problemranges.json
+  debug/crdb_internal.cluster_queries.txt
+  debug/crdb_internal.cluster_sessions.txt
+  debug/crdb_internal.cluster_settings.txt
+  debug/crdb_internal.jobs.txt
+  debug/crdb_internal.kv_node_status.txt
+  debug/crdb_internal.kv_store_status.txt
+  debug/crdb_internal.schema_changes.txt
+  debug/crdb_internal.partitions.txt
+  debug/crdb_internal.zones.txt
+  debug/nodes/1/status.json
+  debug/nodes/1/crdb_internal.feature_usage.txt
+  debug/nodes/1/crdb_internal.gossip_alerts.txt
+  debug/nodes/1/crdb_internal.gossip_liveness.txt
+  debug/nodes/1/crdb_internal.gossip_network.txt
+  debug/nodes/1/crdb_internal.gossip_nodes.txt
+  debug/nodes/1/crdb_internal.leases.txt
+  debug/nodes/1/crdb_internal.node_statement_statistics.txt
+  debug/nodes/1/crdb_internal.node_build_info.txt
+  debug/nodes/1/crdb_internal.node_metrics.txt
+  debug/nodes/1/crdb_internal.node_queries.txt
+  debug/nodes/1/crdb_internal.node_runtime_info.txt
+  debug/nodes/1/crdb_internal.node_sessions.txt
+  debug/nodes/1/details.json
+  debug/nodes/1/gossip.json
+  debug/nodes/1/enginestats.json
+  debug/nodes/1/stacks.txt
+  debug/nodes/1/heap.pprof
+  debug/nodes/1/ranges/1.json
+  debug/nodes/1/ranges/2.json
+  debug/nodes/1/ranges/3.json
+  debug/nodes/1/ranges/4.json
+  debug/nodes/1/ranges/5.json
+  debug/nodes/1/ranges/6.json
+  debug/nodes/1/ranges/7.json
+  debug/nodes/1/ranges/8.json
+  debug/nodes/1/ranges/9.json
+  debug/nodes/1/ranges/10.json
+  debug/nodes/1/ranges/11.json
+  debug/nodes/1/ranges/12.json
+  debug/nodes/1/ranges/13.json
+  debug/nodes/1/ranges/14.json
+  debug/nodes/1/ranges/15.json
+  debug/nodes/1/ranges/16.json
+  debug/nodes/1/ranges/17.json
+  debug/nodes/1/ranges/18.json
+  debug/nodes/1/ranges/19.json
+  debug/nodes/1/ranges/20.json
+  debug/schema/defaultdb@details.json
+  debug/schema/postgres@details.json
+  debug/schema/system@details.json
+  debug/schema/system/comments.json
+  debug/schema/system/descriptor.json
+  debug/schema/system/eventlog.json
+  debug/schema/system/jobs.json
+  debug/schema/system/lease.json
+  debug/schema/system/locations.json
+  debug/schema/system/namespace.json
+  debug/schema/system/rangelog.json
+  debug/schema/system/role_members.json
+  debug/schema/system/settings.json
+  debug/schema/system/table_statistics.json
+  debug/schema/system/ui.json
+  debug/schema/system/users.json
+  debug/schema/system/web_sessions.json
+  debug/schema/system/zones.json
 `
 
-	if out != expected {
-		t.Errorf("expected:\n%s\ngot:\n%s", expected, out)
-	}
+	assert.Equal(t, expected, out)
 }
 
 func TestWorkload(t *testing.T) {
