@@ -262,6 +262,16 @@ results of each SQL statement are printed on the standard output.`,
 Reveal the SQL statements sent implicitly by the command-line utility.`,
 	}
 
+	CliDebugMode = FlagInfo{
+		Name: "debug-sql-cli",
+		Description: `
+Simplify the SQL CLI to ease troubleshooting of CockroachDB
+issues. This echoes sent SQL, removes the database name and txn status
+from the prompt, and forces behavior to become independent on current
+transaction state. Equivalent to --echo-sql, \unset check_syntax,
+\unset smart_prompt, and \set prompt1 %n@%M>.`,
+	}
+
 	SafeUpdates = FlagInfo{
 		Name: "safe-updates",
 		Description: `
@@ -473,6 +483,11 @@ a public network without combining it with --listen-addr.`,
 		Description: `Certificate and key files are overwritten if they exist.`,
 	}
 
+	GeneratePKCS8Key = FlagInfo{
+		Name:        "also-generate-pkcs8-key",
+		Description: `Also write the key in pkcs8 format to <certs-dir>/client.<username>.key.pk8.`,
+	}
+
 	Password = FlagInfo{
 		Name:        "password",
 		Description: `Prompt for the new user's password.`,
@@ -754,8 +769,9 @@ in the history of the cluster.`,
 	}
 
 	SQLFmtLen = FlagInfo{
-		Name:        "print-width",
-		Description: `The line length where sqlfmt will try to wrap.`,
+		Name: "print-width",
+		Description: `
+The line length where sqlfmt will try to wrap.`,
 	}
 
 	SQLFmtSpaces = FlagInfo{
@@ -806,4 +822,51 @@ Maximum size of each log file.
 Minimum verbosity of messages written to the log file.
 `,
 	}
+
+	WriteSize = FlagInfo{
+		Name: "write-size",
+		Description: `
+Size of blocks to write to storage device.`,
+	}
+
+	SyncInterval = FlagInfo{
+		Name: "sync-interval",
+		Description: `
+Number of bytes to write before running fsync.`,
+	}
+
+	BenchConcurrency = FlagInfo{
+		Name: "concurrency",
+		Description: `
+Number of workers for benchmarking.`,
+	}
+
+	BenchDuration = FlagInfo{
+		Name: "duration",
+		Description: `
+Amount of time to run workers.`,
+	}
+
+	BenchServer = FlagInfo{
+		Name: "server",
+		Description: `
+Run as server for network benchmark.`,
+	}
+
+	BenchPort = FlagInfo{
+		Name: "port",
+		Description: `
+Port for network benchmark.`,
+	}
+
+	BenchAddresses = FlagInfo{
+		Name: "addresses",
+		Description: `
+Addresses for network benchmark.`,
+	}
+	BenchLatency = FlagInfo{
+		Name: "latency",
+		Description: `
+Latency or throughput mode.`,
+	}
 )
