@@ -25,29 +25,32 @@ import (
 	"math"
 	"os"
 	"path/filepath"
+	"regexp"
 	"sort"
 	"strconv"
 	"strings"
 	"time"
 
 	"github.com/cockroachdb/cockroach/pkg/cli/debug"
-	"github.com/cockroachdb/cockroach/pkg/cli/synctest"
+	"github.com/cockroachdb/cockroach/pkg/cli/syncbench"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/server"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
-	"github.com/cockroachdb/cockroach/pkg/server/status"
-	"github.com/cockroachdb/cockroach/pkg/sql/distsqlrun"
+	"github.com/cockroachdb/cockroach/pkg/server/status/statuspb"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/storage/rditer"
 	"github.com/cockroachdb/cockroach/pkg/storage/stateloader"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/ts/tspb"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
+	"github.com/cockroachdb/cockroach/pkg/util/flagutil"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
@@ -88,20 +91,42 @@ Create a ballast file to fill the store directory up to a given amount
 // It must do nothing in OSS code.
 var PopulateRocksDBConfigHook func(*engine.RocksDBConfig) error
 
-func parseRangeID(arg string) (roachpb.RangeID, error) {
-	rangeIDInt, err := strconv.ParseInt(arg, 10, 64)
+func parsePositiveInt(arg string) (int64, error) {
+	i, err := strconv.ParseInt(arg, 10, 64)
 	if err != nil {
 		return 0, err
 	}
-	if rangeIDInt < 1 {
-		return 0, fmt.Errorf("illegal RangeID: %d", rangeIDInt)
+	if i < 1 {
+		return 0, fmt.Errorf("illegal val: %d < 1", i)
+	}
+	return i, nil
+}
+
+func parseRangeID(arg string) (roachpb.RangeID, error) {
+	rangeIDInt, err := parsePositiveInt(arg)
+	if err != nil {
+		return 0, err
 	}
 	return roachpb.RangeID(rangeIDInt), nil
 }
 
+// OpenEngineOptions tunes the behavior of OpenEngine.
+type OpenEngineOptions struct {
+	ReadOnly  bool
+	MustExist bool
+}
+
 // OpenExistingStore opens the rocksdb engine rooted at 'dir'.
 // If 'readOnly' is true, opens the store in read-only mode.
 func OpenExistingStore(dir string, stopper *stop.Stopper, readOnly bool) (*engine.RocksDB, error) {
+	return OpenEngine(dir, stopper, OpenEngineOptions{ReadOnly: readOnly, MustExist: true})
+}
+
+// OpenEngine opens the RocksDB engine at 'dir'. Depending on the supplied options,
+// an empty engine might be initialized.
+func OpenEngine(
+	dir string, stopper *stop.Stopper, opts OpenEngineOptions,
+) (*engine.RocksDB, error) {
 	cache := engine.NewRocksDBCache(server.DefaultCacheSize)
 	defer cache.Release()
 	maxOpenFiles, err := server.SetOpenFileLimitForOneStore()
@@ -113,8 +138,8 @@ func OpenExistingStore(dir string, stopper *stop.Stopper, readOnly bool) (*engin
 		Settings:     serverCfg.Settings,
 		Dir:          dir,
 		MaxOpenFiles: maxOpenFiles,
-		MustExist:    true,
-		ReadOnly:     readOnly,
+		MustExist:    opts.MustExist,
+		ReadOnly:     opts.ReadOnly,
 	}
 
 	if PopulateRocksDBConfigHook != nil {
@@ -153,7 +178,7 @@ func runDebugKeys(cmd *cobra.Command, args []string) error {
 	printer := printKey
 	if debugCtx.values {
 		printer = func(kv engine.MVCCKeyValue) (bool, error) {
-			debug.PrintKeyValue(kv, debugCtx.sizes)
+			debug.PrintKeyValue(kv)
 			return false, nil
 		}
 	}
@@ -257,7 +282,7 @@ func runDebugRangeData(cmd *cobra.Command, args []string) error {
 		debug.PrintKeyValue(engine.MVCCKeyValue{
 			Key:   iter.Key(),
 			Value: iter.Value(),
-		}, debugCtx.sizes)
+		})
 	}
 	return nil
 }
@@ -327,7 +352,7 @@ func runDebugRangeDescriptors(cmd *cobra.Command, args []string) error {
 		if debug.IsRangeDescriptorKey(kv.Key) != nil {
 			return false, nil
 		}
-		debug.PrintKeyValue(kv, debugCtx.sizes)
+		debug.PrintKeyValue(kv)
 		return false, nil
 	})
 }
@@ -348,7 +373,7 @@ Decode a hexadecimal-encoded key and pretty-print it. For example:
 			if err != nil {
 				return err
 			}
-			k, err := engine.DecodeKey(b)
+			k, err := engine.DecodeMVCCKey(b)
 			if err != nil {
 				return err
 			}
@@ -378,12 +403,15 @@ Decode and print a hexadecimal-encoded key-value pair.
 			bs = append(bs, b)
 		}
 
-		k, err := engine.DecodeKey(bs[0])
+		isTS := bytes.HasPrefix(bs[0], keys.TimeseriesPrefix)
+		k, err := engine.DecodeMVCCKey(bs[0])
 		if err != nil {
 			// Older versions of the consistency checker give you diffs with a raw_key that
 			// is already a roachpb.Key, so make a half-assed attempt to support both.
-			fmt.Printf("unable to decode key: %v, assuming it's a roachpb.Key with fake timestamp;\n"+
-				"if the result below looks like garbage, then it likely is:\n\n", err)
+			if !isTS {
+				fmt.Printf("unable to decode key: %v, assuming it's a roachpb.Key with fake timestamp;\n"+
+					"if the result below looks like garbage, then it likely is:\n\n", err)
+			}
 			k = engine.MVCCKey{
 				Key:       bs[0],
 				Timestamp: hlc.Timestamp{WallTime: 987654321},
@@ -393,7 +421,7 @@ Decode and print a hexadecimal-encoded key-value pair.
 		debug.PrintKeyValue(engine.MVCCKeyValue{
 			Key:   k,
 			Value: bs[1],
-		}, debugCtx.sizes)
+		})
 		return nil
 	},
 }
@@ -424,15 +452,17 @@ func runDebugRaftLog(cmd *cobra.Command, args []string) error {
 
 	start := engine.MakeMVCCMetadataKey(keys.RaftLogPrefix(rangeID))
 	end := engine.MakeMVCCMetadataKey(keys.RaftLogPrefix(rangeID).PrefixEnd())
+	fmt.Printf("Printing keys %s -> %s (RocksDB keys: %#x - %#x )\n",
+		start, end, string(engine.EncodeKey(start)), string(engine.EncodeKey(end)))
 
 	return db.Iterate(start, end, func(kv engine.MVCCKeyValue) (bool, error) {
-		debug.PrintKeyValue(kv, debugCtx.sizes)
+		debug.PrintKeyValue(kv)
 		return false, nil
 	})
 }
 
 var debugGCCmd = &cobra.Command{
-	Use:   "estimate-gc <directory> [range id]",
+	Use:   "estimate-gc <directory> [range id] [ttl-in-seconds]",
 	Short: "find out what a GC run would do",
 	Long: `
 Sets up (but does not run) a GC collection cycle, giving insight into how much
@@ -441,7 +471,7 @@ work would be done (assuming all intent resolution and pushes succeed).
 Without a RangeID specified on the command line, runs the analysis for all
 ranges individually.
 
-Uses a hard-coded GC policy with a 24 hour TTL for old versions.
+Uses a configurable GC policy, with a default 24 hour TTL, for old versions.
 `,
 	Args: cobra.RangeArgs(1, 2),
 	RunE: MaybeDecorateGRPCError(runDebugGCCmd),
@@ -452,7 +482,18 @@ func runDebugGCCmd(cmd *cobra.Command, args []string) error {
 	defer stopper.Stop(context.Background())
 
 	var rangeID roachpb.RangeID
-	if len(args) == 2 {
+	gcTTLInSeconds := int64((24 * time.Hour).Seconds())
+	switch len(args) {
+	case 3:
+		var err error
+		if rangeID, err = parseRangeID(args[1]); err != nil {
+			return errors.Wrapf(err, "unable to parse %v as range ID", args[1])
+		}
+		if gcTTLInSeconds, err = parsePositiveInt(args[2]); err != nil {
+			return errors.Wrapf(err, "unable to parse %v as TTL", args[2])
+		}
+
+	case 2:
 		var err error
 		if rangeID, err = parseRangeID(args[1]); err != nil {
 			return err
@@ -470,8 +511,7 @@ func runDebugGCCmd(cmd *cobra.Command, args []string) error {
 	var descs []roachpb.RangeDescriptor
 
 	if _, err := engine.MVCCIterate(context.Background(), db, start, end, hlc.MaxTimestamp,
-		false /* consistent */, false /* tombstones */, nil, /* txn */
-		false /* reverse */, func(kv roachpb.KeyValue) (bool, error) {
+		engine.MVCCScanOptions{Inconsistent: true}, func(kv roachpb.KeyValue) (bool, error) {
 			var desc roachpb.RangeDescriptor
 			_, suffix, _, err := keys.DecodeRangeKey(kv.Key)
 			if err != nil {
@@ -503,7 +543,7 @@ func runDebugGCCmd(cmd *cobra.Command, args []string) error {
 			&desc,
 			snap,
 			hlc.Timestamp{WallTime: timeutil.Now().UnixNano()},
-			config.GCPolicy{TTLSeconds: 24 * 60 * 60 /* 1 day */},
+			config.GCPolicy{TTLSeconds: int32(gcTTLInSeconds)},
 			storage.NoopGCer{},
 			func(_ context.Context, _ []roachpb.Intent) error { return nil },
 			func(_ context.Context, _ *roachpb.Transaction, _ []roachpb.Intent) error { return nil },
@@ -571,7 +611,7 @@ func runDebugCheckStoreDescriptors(ctx context.Context, db *engine.RocksDB) erro
 	var failed bool
 	if err := storage.IterateRangeDescriptors(ctx, db,
 		func(desc roachpb.RangeDescriptor) (bool, error) {
-			claimedMS, err := stateloader.Make(serverCfg.Settings, desc.RangeID).LoadMVCCStats(ctx, db)
+			claimedMS, err := stateloader.Make(desc.RangeID).LoadMVCCStats(ctx, db)
 			if err != nil {
 				return false, err
 			}
@@ -617,8 +657,7 @@ func runDebugCheckStoreRaft(ctx context.Context, db *engine.RocksDB) error {
 	var hasError bool
 
 	if _, err := engine.MVCCIterate(ctx, db, start, end, hlc.MaxTimestamp,
-		false /* consistent */, false /* tombstones */, nil, /* txn */
-		false /* reverse */, func(kv roachpb.KeyValue) (bool, error) {
+		engine.MVCCScanOptions{Inconsistent: true}, func(kv roachpb.KeyValue) (bool, error) {
 			rangeID, _, suffix, detail, err := keys.DecodeRangeIDKey(kv.Key)
 			if err != nil {
 				return false, err
@@ -631,7 +670,7 @@ func runDebugCheckStoreRaft(ctx context.Context, db *engine.RocksDB) error {
 					return false, err
 				}
 				getReplicaInfo(rangeID).committedIndex = hs.Commit
-			case bytes.Equal(suffix, keys.LocalRaftTruncatedStateSuffix):
+			case bytes.Equal(suffix, keys.LocalRaftTruncatedStateLegacySuffix):
 				var trunc roachpb.RaftTruncatedState
 				if err := kv.Value.GetProto(&trunc); err != nil {
 					return false, err
@@ -899,7 +938,7 @@ func parseGossipValues(gossipInfo *gossip.InfoStatus) (string, error) {
 			output = append(output, fmt.Sprintf("%q: %v", key, clusterID))
 		} else if key == gossip.KeySystemConfig {
 			if debugCtx.printSystemConfig {
-				var config config.SystemConfig
+				var config config.SystemConfigEntries
 				if err := protoutil.Unmarshal(bytes, &config); err != nil {
 					return "", errors.Wrapf(err, "failed to parse value for key %q", key)
 				}
@@ -926,31 +965,25 @@ func parseGossipValues(gossipInfo *gossip.InfoStatus) (string, error) {
 			}
 			output = append(output, fmt.Sprintf("%q: %+v", key, desc))
 		} else if strings.HasPrefix(key, gossip.KeyNodeLivenessPrefix) {
-			var liveness storage.Liveness
+			var liveness storagepb.Liveness
 			if err := protoutil.Unmarshal(bytes, &liveness); err != nil {
 				return "", errors.Wrapf(err, "failed to parse value for key %q", key)
 			}
 			output = append(output, fmt.Sprintf("%q: %+v", key, liveness))
-		} else if strings.HasPrefix(key, gossip.KeyDeadReplicasPrefix) {
-			var deadReplicas roachpb.StoreDeadReplicas
-			if err := protoutil.Unmarshal(bytes, &deadReplicas); err != nil {
-				return "", errors.Wrapf(err, "failed to parse value for key %q", key)
-			}
-			output = append(output, fmt.Sprintf("%q: %+v", key, deadReplicas))
 		} else if strings.HasPrefix(key, gossip.KeyNodeHealthAlertPrefix) {
-			var healthAlert status.HealthCheckResult
+			var healthAlert statuspb.HealthCheckResult
 			if err := protoutil.Unmarshal(bytes, &healthAlert); err != nil {
 				return "", errors.Wrapf(err, "failed to parse value for key %q", key)
 			}
 			output = append(output, fmt.Sprintf("%q: %+v", key, healthAlert))
 		} else if strings.HasPrefix(key, gossip.KeyDistSQLNodeVersionKeyPrefix) {
-			var version distsqlrun.DistSQLVersionGossipInfo
+			var version distsqlpb.DistSQLVersionGossipInfo
 			if err := protoutil.Unmarshal(bytes, &version); err != nil {
 				return "", errors.Wrapf(err, "failed to parse value for key %q", key)
 			}
 			output = append(output, fmt.Sprintf("%q: %+v", key, version))
 		} else if strings.HasPrefix(key, gossip.KeyDistSQLDrainingPrefix) {
-			var drainingInfo distsqlrun.DistSQLDrainingInfo
+			var drainingInfo distsqlpb.DistSQLDrainingInfo
 			if err := protoutil.Unmarshal(bytes, &drainingInfo); err != nil {
 				return "", errors.Wrapf(err, "failed to parse value for key %q", key)
 			}
@@ -1011,28 +1044,28 @@ func runTimeSeriesDump(cmd *cobra.Command, args []string) error {
 	}
 }
 
-var debugSyncTestCmd = &cobra.Command{
-	Use:   "synctest [directory]",
+var debugSyncBenchCmd = &cobra.Command{
+	Use:   "syncbench [directory]",
 	Short: "Run a performance test for WAL sync speed",
 	Long: `
 `,
 	Args:   cobra.MaximumNArgs(1),
 	Hidden: true,
-	RunE:   MaybeDecorateGRPCError(runDebugSyncTest),
+	RunE:   MaybeDecorateGRPCError(runDebugSyncBench),
 }
 
-var syncTestOpts = synctest.Options{
+var syncBenchOpts = syncbench.Options{
 	Concurrency: 1,
 	Duration:    10 * time.Second,
 	LogOnly:     true,
 }
 
-func runDebugSyncTest(cmd *cobra.Command, args []string) error {
-	syncTestOpts.Dir = "./testdb"
+func runDebugSyncBench(cmd *cobra.Command, args []string) error {
+	syncBenchOpts.Dir = "./testdb"
 	if len(args) == 1 {
-		syncTestOpts.Dir = args[0]
+		syncBenchOpts.Dir = args[0]
 	}
-	return synctest.Run(syncTestOpts)
+	return syncbench.Run(syncBenchOpts)
 }
 
 var debugUnsafeRemoveDeadReplicasCmd = &cobra.Command{
@@ -1124,7 +1157,7 @@ func removeDeadReplicas(
 	if err != nil {
 		return nil, err
 	}
-	fmt.Printf("Scanning replicas on store %s for dead peers %v\n", storeIdent,
+	fmt.Printf("Scanning replicas on store %s for dead peers %v\n", storeIdent.String(),
 		removeDeadReplicasOpts.deadStoreIDs)
 
 	if _, ok := deadStoreIDs[storeIdent.StoreID]; ok {
@@ -1204,20 +1237,41 @@ func removeDeadReplicas(
 	return batch, nil
 }
 
-func init() {
-	DebugCmd.AddCommand(debugCmds...)
+var debugMergeLogsCommand = &cobra.Command{
+	Use:   "merge-logs <log file globs>",
+	Short: "merge multiple log files from different machines into a single stream",
+	Long: `
+Takes a list of glob patterns (not left exclusively to the shell because of
+MAX_ARG_STRLEN, usually 128kB) pointing to log files and merges them into a 
+single stream printed to stdout. Files not matching the log file name pattern
+are ignored. If log lines appear out of order within a file (which happens), the
+timestamp is ratcheted to the highest value seen so far. The command supports
+efficient time filtering as well as multiline regexp pattern matching via flags.
+`,
+	Args: cobra.MinimumNArgs(1),
+	RunE: runDebugMergeLogs,
+}
 
-	f := debugSyncTestCmd.Flags()
-	f.IntVarP(&syncTestOpts.Concurrency, "concurrency", "c", syncTestOpts.Concurrency,
-		"number of concurrent writers")
-	f.DurationVarP(&syncTestOpts.Duration, "duration", "d", syncTestOpts.Duration,
-		"duration to run the test for")
-	f.BoolVarP(&syncTestOpts.LogOnly, "log-only", "l", syncTestOpts.LogOnly,
-		"only write to the WAL, not to sstables")
+var debugMergeLogsOpts = struct {
+	from    time.Time
+	to      time.Time
+	filter  *regexp.Regexp
+	program *regexp.Regexp
+	file    *regexp.Regexp
+	prefix  string
+}{
+	program: regexp.MustCompile("^cockroach.*$"),
+	file:    regexp.MustCompile(log.FilePattern),
+}
 
-	f = debugUnsafeRemoveDeadReplicasCmd.Flags()
-	f.IntSliceVar(&removeDeadReplicasOpts.deadStoreIDs, "dead-store-ids", nil,
-		"list of dead store IDs")
+func runDebugMergeLogs(cmd *cobra.Command, args []string) error {
+	o := debugMergeLogsOpts
+	s, err := newMergedStreamFromPatterns(context.Background(),
+		args, o.file, o.program, o.from, o.to)
+	if err != nil {
+		return err
+	}
+	return writeLogStream(s, cmd.OutOrStdout(), o.filter, o.prefix)
 }
 
 // DebugCmdsForRocksDB lists debug commands that access rocksdb through the engine
@@ -1243,10 +1297,12 @@ var debugCmds = append(DebugCmdsForRocksDB,
 	debugSSTDumpCmd,
 	debugGossipValuesCmd,
 	debugTimeSeriesDumpCmd,
+	debugSyncBenchCmd,
 	debugSyncTestCmd,
 	debugUnsafeRemoveDeadReplicasCmd,
 	debugEnvCmd,
 	debugZipCmd,
+	debugMergeLogsCommand,
 )
 
 // DebugCmd is the root of all debug commands. Exported to allow modification by CCL code.
@@ -1260,3 +1316,34 @@ process that has failed and cannot restart.
 `,
 	RunE: usageAndErr,
 }
+
+func init() {
+	DebugCmd.AddCommand(debugCmds...)
+
+	f := debugSyncBenchCmd.Flags()
+	f.IntVarP(&syncBenchOpts.Concurrency, "concurrency", "c", syncBenchOpts.Concurrency,
+		"number of concurrent writers")
+	f.DurationVarP(&syncBenchOpts.Duration, "duration", "d", syncBenchOpts.Duration,
+		"duration to run the test for")
+	f.BoolVarP(&syncBenchOpts.LogOnly, "log-only", "l", syncBenchOpts.LogOnly,
+		"only write to the WAL, not to sstables")
+
+	f = debugUnsafeRemoveDeadReplicasCmd.Flags()
+	f.IntSliceVar(&removeDeadReplicasOpts.deadStoreIDs, "dead-store-ids", nil,
+		"list of dead store IDs")
+
+	f = debugMergeLogsCommand.Flags()
+	f.Var(flagutil.Time(&debugMergeLogsOpts.from), "from",
+		"time before which messages should be filtered")
+	f.Var(flagutil.Time(&debugMergeLogsOpts.to), "to",
+		"time after which messages should be filtered")
+	f.Var(flagutil.Regexp(&debugMergeLogsOpts.filter), "filter",
+		"re which filters log messages")
+	f.Var(flagutil.Regexp(&debugMergeLogsOpts.file), "file-pattern",
+		"re which filters log files based on path, also used with prefix and program-filter")
+	f.Var(flagutil.Regexp(&debugMergeLogsOpts.program), "program-filter",
+		"re which filter log files that operates on the capture group named \"program\" in file-pattern, "+
+			"if no such group exists, program-filter is ignored")
+	f.StringVar(&debugMergeLogsOpts.prefix, "prefix", "${host}> ",
+		"expansion template (see regexp.Expand) used as prefix to merged log messages evaluated on file-pattern")
+}
