@@ -24,15 +24,15 @@ import (
 	"strings"
 	"time"
 
-	"github.com/pkg/errors"
-	"github.com/spf13/cobra"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/util/ctxgroup"
 	"github.com/cockroachdb/cockroach/pkg/util/timeofday"
+	"github.com/cockroachdb/cockroach/pkg/util/version"
+	"github.com/pkg/errors"
+	"github.com/spf13/cobra"
 )
 
 // dumpCmd dumps SQL tables.
@@ -47,6 +47,10 @@ is omitted, dump all tables in the database.
 	RunE: MaybeDecorateGRPCError(runDump),
 }
 
+// We accept versions that are strictly newer than v2.1.0-alpha.20180416
+// (hence the "-0" at the end).
+var verDump = version.MustParse("v2.1.0-alpha.20180416-0")
+
 // runDumps performs a dump of a table or database.
 //
 // The approach here and its current flaws are summarized
@@ -58,7 +62,7 @@ func runDump(cmd *cobra.Command, args []string) error {
 	}
 	defer conn.Close()
 
-	if err := conn.requireServerVersion(">v2.1.0-alpha.20180416"); err != nil {
+	if err := conn.requireServerVersion(verDump); err != nil {
 		return err
 	}
 
@@ -209,7 +213,7 @@ func getDumpMetadata(
 		clusterTS = string(vals[0].([]byte))
 	} else {
 		// Validate the timestamp. This prevents SQL injection.
-		if _, err := tree.ParseDTimestamp(asOf, time.Nanosecond); err != nil {
+		if _, err := tree.ParseDTimestamp(nil, asOf, time.Nanosecond); err != nil {
 			return nil, "", err
 		}
 		clusterTS = asOf
@@ -422,7 +426,7 @@ func getMetadataForTable(conn *sqlConn, md basicMetadata, ts string) (tableMetad
 	}
 	vals := make([]driver.Value, 2)
 	coltypes := make(map[string]coltypes.T)
-	colnames := tree.NewFmtCtxWithBuf(tree.FmtSimple)
+	colnames := tree.NewFmtCtx(tree.FmtSimple)
 	defer colnames.Close()
 	for {
 		if err := rows.Next(vals); err == io.EOF {
@@ -446,7 +450,7 @@ func getMetadataForTable(conn *sqlConn, md basicMetadata, ts string) (tableMetad
 		if err != nil {
 			return tableMetadata{}, fmt.Errorf("type %s is not a valid CockroachDB type", typ)
 		}
-		coltyp := stmt.(*tree.AlterTable).Cmds[0].(*tree.AlterTableAlterColumnType).ToType
+		coltyp := stmt.AST.(*tree.AlterTable).Cmds[0].(*tree.AlterTableAlterColumnType).ToType
 
 		coltypes[name] = coltyp
 		if colnames.Len() > 0 {
@@ -565,6 +569,7 @@ func dumpTableData(w io.Writer, conn *sqlConn, clusterTS string, bmd basicMetada
 	g.GoCtx(func(ctx context.Context) error {
 		// Fetch SQL rows and put them onto valsCh.
 		defer close(valsCh)
+		done := ctx.Done()
 		for i := 0; ; i++ {
 			vals := valArray[i%len(valArray)]
 			if err := rows.Next(vals); err == io.EOF {
@@ -573,8 +578,8 @@ func dumpTableData(w io.Writer, conn *sqlConn, clusterTS string, bmd basicMetada
 				return err
 			}
 			select {
-			case <-g.Done:
-				return g.Err()
+			case <-done:
+				return ctx.Err()
 			case valsCh <- vals:
 			}
 		}
@@ -582,8 +587,9 @@ func dumpTableData(w io.Writer, conn *sqlConn, clusterTS string, bmd basicMetada
 	g.GoCtx(func(ctx context.Context) error {
 		// Convert SQL rows into VALUE strings.
 		defer close(stringsCh)
-		f := tree.NewFmtCtxWithBuf(tree.FmtParsableNumerics)
+		f := tree.NewFmtCtx(tree.FmtParsableNumerics)
 		defer f.Close()
+		done := ctx.Done()
 		for vals := range valsCh {
 			f.Reset()
 			// Values need to be correctly encoded for INSERT statements in a text file.
@@ -668,11 +674,11 @@ func dumpTableData(w io.Writer, conn *sqlConn, clusterTS string, bmd basicMetada
 				default:
 					return errors.Errorf("unknown field type: %T (%s)", t, cols[si])
 				}
-				d.Format(&f.FmtCtx)
+				d.Format(f)
 			}
 			select {
-			case <-g.Done:
-				return g.Err()
+			case <-done:
+				return ctx.Err()
 			case stringsCh <- f.String():
 			}
 		}
