@@ -21,15 +21,15 @@ import (
 	"strings"
 	"time"
 
-	"github.com/spf13/cobra"
-	"github.com/spf13/pflag"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/cli/cliflags"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
+	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/log/logflags"
+	"github.com/spf13/cobra"
+	"github.com/spf13/pflag"
 )
 
 // special global variables used by flag variable definitions below.
@@ -131,6 +131,14 @@ func VarFlag(f *pflag.FlagSet, value pflag.Value, flagInfo cliflags.FlagInfo) {
 	setFlagFromEnv(f, flagInfo)
 }
 
+// StringSlice creates a string slice flag and registers it with the FlagSet.
+func StringSlice(
+	f *pflag.FlagSet, valPtr *[]string, flagInfo cliflags.FlagInfo, defaultVal []string,
+) {
+	f.StringSliceVar(valPtr, flagInfo.Name, defaultVal, flagInfo.Usage())
+	setFlagFromEnv(f, flagInfo)
+}
+
 // aliasStrVar wraps a string configuration option and is meant
 // to be used in addition to / next to another flag that targets the
 // same option. It does not implement "default values" so that the
@@ -223,6 +231,12 @@ func init() {
 		if strings.HasPrefix(flag.Name, "lightstep_") {
 			flag.Hidden = true
 		}
+		if strings.HasPrefix(flag.Name, "httptest.") {
+			// If we test the cli commands in tests, we may end up transitively
+			// importing httptest, for example via `testify/assert`. Make sure
+			// it doesn't show up in the output or it will confuse tests.
+			flag.Hidden = true
+		}
 		switch flag.Name {
 		case logflags.NoRedirectStderrName:
 			flag.Hidden = true
@@ -380,6 +394,8 @@ func init() {
 		IntFlag(f, &keySize, cliflags.KeySize, defaultKeySize)
 		BoolFlag(f, &overwriteFiles, cliflags.OverwriteFiles, false)
 	}
+	// PKCS8 key format is only available for the client cert command.
+	BoolFlag(createClientCertCmd.Flags(), &generatePKCS8Key, cliflags.GeneratePKCS8Key, false)
 
 	BoolFlag(setUserCmd.Flags(), &password, cliflags.Password, false)
 
@@ -396,6 +412,7 @@ func init() {
 	clientCmds = append(clientCmds, userCmds...)
 	clientCmds = append(clientCmds, zoneCmds...)
 	clientCmds = append(clientCmds, nodeCmds...)
+	clientCmds = append(clientCmds, systemBenchCmds...)
 	clientCmds = append(clientCmds, initCmd)
 	for _, cmd := range clientCmds {
 		f := cmd.PersistentFlags()
@@ -431,6 +448,32 @@ func init() {
 		BoolFlag(f, &nodeCtx.statusShowDecommission, cliflags.NodeDecommission, nodeCtx.statusShowDecommission)
 	}
 
+	// HDD Bench command.
+	{
+		f := seqWriteBench.Flags()
+		VarFlag(f, humanizeutil.NewBytesValue(&systemBenchCtx.writeSize), cliflags.WriteSize)
+		VarFlag(f, humanizeutil.NewBytesValue(&systemBenchCtx.syncInterval), cliflags.SyncInterval)
+	}
+
+	// Network Bench command.
+	{
+		f := networkBench.Flags()
+		BoolFlag(f, &networkBenchCtx.server, cliflags.BenchServer, networkBenchCtx.server)
+		IntFlag(f, &networkBenchCtx.port, cliflags.BenchPort, networkBenchCtx.port)
+		StringSlice(f, &networkBenchCtx.addresses, cliflags.BenchAddresses, networkBenchCtx.addresses)
+		BoolFlag(f, &networkBenchCtx.latency, cliflags.BenchLatency, networkBenchCtx.latency)
+	}
+
+	// Bench command.
+	{
+		for _, cmd := range systemBenchCmds {
+			f := cmd.Flags()
+			IntFlag(f, &systemBenchCtx.concurrency, cliflags.BenchConcurrency, systemBenchCtx.concurrency)
+			DurationFlag(f, &systemBenchCtx.duration, cliflags.BenchDuration, systemBenchCtx.duration)
+			StringFlag(f, &systemBenchCtx.tempDir, cliflags.TempDir, systemBenchCtx.tempDir)
+		}
+	}
+
 	// Decommission command.
 	VarFlag(decommissionNodeCmd.Flags(), &nodeCtx.nodeDecommissionWait, cliflags.Wait)
 
@@ -446,6 +489,7 @@ func init() {
 		VarFlag(f, &sqlCtx.setStmts, cliflags.Set)
 		VarFlag(f, &sqlCtx.execStmts, cliflags.Execute)
 		BoolFlag(f, &sqlCtx.safeUpdates, cliflags.SafeUpdates, sqlCtx.safeUpdates)
+		BoolFlag(f, &sqlCtx.debugMode, cliflags.CliDebugMode, sqlCtx.debugMode)
 	}
 
 	VarFlag(dumpCmd.Flags(), &dumpCtx.dumpMode, cliflags.DumpMode)
@@ -556,6 +600,13 @@ func extraClientFlagInit() {
 		serverHTTPAddr = startCtx.serverListenAddr
 	}
 	serverCfg.HTTPAddr = net.JoinHostPort(serverHTTPAddr, serverHTTPPort)
+
+	// If CLI/SQL debug mode is requested, override the echo mode here,
+	// so that the initial client/server handshake reveals the SQL being
+	// sent.
+	if sqlCtx.debugMode {
+		sqlCtx.echo = true
+	}
 }
 
 func setDefaultStderrVerbosity(cmd *cobra.Command, defaultSeverity log.Severity) error {
