@@ -24,8 +24,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/spf13/cobra"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/cli/cliflags"
 	"github.com/cockroachdb/cockroach/pkg/server/status"
@@ -33,6 +31,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/testutils/buildutil"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/log/logflags"
+	"github.com/spf13/cobra"
 )
 
 func TestStdFlagToPflag(t *testing.T) {
@@ -428,9 +427,21 @@ func TestServerConnSettings(t *testing.T) {
 				t.Errorf("%d. serverCfg.AdvertiseAddr expected '%s', but got '%s'. td.args was '%#v'.",
 					i, td.expectedAdvertiseAddr, serverCfg.AdvertiseAddr, td.args)
 			}
-			if td.expLocalityAdvertiseAddr != fmt.Sprintf("%s", serverCfg.LocalityAddresses) {
+			var locAddrStr strings.Builder
+			locAddrStr.WriteString("[")
+			for i, a := range serverCfg.LocalityAddresses {
+				if i > 0 {
+					locAddrStr.WriteString(" ")
+				}
+				fmt.Fprintf(
+					&locAddrStr, "{{%s %s} %s}",
+					a.Address.NetworkField, a.Address.AddressField, a.LocalityTier.String(),
+				)
+			}
+			locAddrStr.WriteString("]")
+			if td.expLocalityAdvertiseAddr != locAddrStr.String() {
 				t.Errorf("%d. serverCfg.expLocalityAdvertiseAddr expected '%s', but got '%s'. td.args was '%#v'.",
-					i, td.expLocalityAdvertiseAddr, serverCfg.LocalityAddresses, td.args)
+					i, td.expLocalityAdvertiseAddr, locAddrStr.String(), td.args)
 			}
 		})
 	}
