@@ -28,10 +28,6 @@ import (
 	"unicode/utf8"
 
 	"github.com/cockroachdb/cockroach-go/crdb"
-	version "github.com/hashicorp/go-version"
-	"github.com/lib/pq"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/sql"
@@ -41,6 +37,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/cockroachdb/cockroach/pkg/util/version"
+	"github.com/lib/pq"
+	"github.com/pkg/errors"
 )
 
 type sqlConnI interface {
@@ -215,9 +214,9 @@ func (c *sqlConn) checkServerMetadata() error {
 		}
 		fmt.Printf("# Server version: %s%s\n", c.serverBuild, isSame)
 
-		sv, err := version.NewVersion(c.serverVersion)
+		sv, err := version.Parse(c.serverVersion)
 		if err == nil {
-			cv, err := version.NewVersion(client.Tag)
+			cv, err := version.Parse(client.Tag)
 			if err == nil {
 				if sv.Compare(cv) == -1 { // server ver < client ver
 					fmt.Fprintln(stderr, "\nwarning: server version older than client! "+
@@ -246,23 +245,19 @@ func (c *sqlConn) checkServerMetadata() error {
 }
 
 // requireServerVersion returns an error if the version of the connected server
-// does not match the constraints in constraintString.
-func (c *sqlConn) requireServerVersion(constraintString string) error {
+// is not at least the given version.
+func (c *sqlConn) requireServerVersion(required *version.Version) error {
 	versionString, _, err := c.getServerMetadata()
 	if err != nil {
 		return err
 	}
-	constraints, err := version.NewConstraint(constraintString)
+	vers, err := version.Parse(versionString)
 	if err != nil {
-		return err
+		return fmt.Errorf("unable to parse server version %q", versionString)
 	}
-	vers, err := version.NewVersion(versionString)
-	if err != nil {
-		return fmt.Errorf("unable to parse server version %q", c.serverVersion)
-	}
-	if !constraints.Check(vers) {
+	if !vers.AtLeast(required) {
 		return fmt.Errorf("incompatible client and server versions (detected server version: %s, required: %s)",
-			vers, constraints)
+			vers, required)
 	}
 	return nil
 }
@@ -537,7 +532,7 @@ func makeSQLClient(appName string) (*sqlConn, error) {
 	// Load the application name. It's not a command-line flag, so
 	// anything already in the URL should take priority.
 	if options.Get("application_name") == "" && appName != "" {
-		options.Set("application_name", sql.InternalAppNamePrefix+appName)
+		options.Set("application_name", sql.ReportableAppNamePrefix+appName)
 	}
 
 	// Set a connection timeout if none is provided already. This
@@ -744,24 +739,37 @@ func runQueryAndFormatResults(conn *sqlConn, w io.Writer, fn queryFunc) error {
 		if err != nil {
 			return err
 		}
+
+		var queryCompleteTime time.Time
+		completedHook := func() { queryCompleteTime = timeutil.Now() }
+
 		if err := func() error {
 			if cleanup != nil {
 				defer cleanup()
 			}
-			return render(reporter, w, cols, newRowIter(rows, true), noRowsHook)
+			return render(reporter, w, cols, newRowIter(rows, true), completedHook, noRowsHook)
 		}(); err != nil {
 			return err
 		}
 
-		if cliCtx.showTimes {
+		if sqlCtx.showTimes {
 			// Present the time since the last result, or since the
 			// beginning of execution. Currently the execution engine makes
 			// all the work upfront so most of the time is accounted for by
 			// the 1st result; this is subject to change once CockroachDB
 			// evolves to stream results as statements are executed.
-			newNow := timeutil.Now()
-			fmt.Fprintf(w, "\nTime: %s\n\n", newNow.Sub(startTime))
-			startTime = newNow
+			fmt.Fprintf(w, "\nTime: %s\n", queryCompleteTime.Sub(startTime))
+			// Make users better understand any discrepancy they observe.
+			renderDelay := timeutil.Now().Sub(queryCompleteTime)
+			if renderDelay >= 1*time.Second {
+				fmt.Fprintf(w,
+					"Note: an additional delay of %s was spent formatting the results.\n"+
+						"You can use \\set display_format to change the formatting.\n",
+					renderDelay)
+			}
+			fmt.Fprintln(w)
+			// Reset the clock. We ignore the rendering time.
+			startTime = timeutil.Now()
 		}
 
 		if more, err := rows.NextResultSet(); err != nil {
