@@ -32,11 +32,6 @@ import (
 	"text/tabwriter"
 	"time"
 
-	opentracing "github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-	"github.com/spf13/cobra"
-	"google.golang.org/grpc"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/cli/cliflags"
@@ -58,6 +53,11 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/sysutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
+	"github.com/spf13/cobra"
+	"google.golang.org/grpc"
 )
 
 // jemallocHeapDump is an optional function to be called at heap dump time.
@@ -414,15 +414,21 @@ func runStart(cmd *cobra.Command, args []string) error {
 	signalCh := make(chan os.Signal, 1)
 	signal.Notify(signalCh, drainSignals...)
 
+	// Set up a cancellable context for the entire start command.
+	// The context will be canceled at the end.
+	ctx, cancel := context.WithCancel(context.Background())
+	defer cancel()
+
 	// Set up a tracing span for the start process.  We want any logging
 	// happening beyond this point to be accounted to this start
 	// context, including logging related to the initialization of
 	// the logging infrastructure below.
 	// This span concludes when the startup goroutine started below
 	// has completed.
+	// TODO(andrei): we don't close the span on the early returns below.
 	tracer := serverCfg.Settings.Tracer
-	sp := tracer.StartSpan("server start")
-	ctx := opentracing.ContextWithSpan(context.Background(), sp)
+	sp := tracer.StartRootSpan("server start", nil /* logTags */, tracing.NonRecordableSpan)
+	ctx = opentracing.ContextWithSpan(ctx, sp)
 
 	// Set up the logging and profiling output.
 	//
@@ -440,6 +446,8 @@ func runStart(cmd *cobra.Command, args []string) error {
 	if err != nil {
 		return err
 	}
+
+	serverCfg.GoroutineDumpDirName = logOutputDirectory()
 	serverCfg.HeapProfileDirName = logOutputDirectory()
 	// We don't care about GRPCs fairly verbose logs in most client commands,
 	// but when actually starting a server, we enable them.
@@ -503,9 +511,10 @@ func runStart(cmd *cobra.Command, args []string) error {
 
 		if waitForInit {
 			log.Shout(ctx, log.Severity_INFO,
-				"initial startup completed, will now wait for `cockroach init`\n"+
-					"or a join to a running cluster to start accepting clients.\n"+
-					"Check the log file(s) for progress.")
+				"initial startup completed.\n"+
+					"Node will now attempt to join a running cluster, or wait for `cockroach init`.\n"+
+					"Client connections will be accepted after this completes successfully.\n"+
+					"Check the log file(s) for progress. ")
 		}
 
 		// Ensure the configuration logging is written to disk in case a
@@ -1003,6 +1012,11 @@ func setupAndInitializeLoggingAndProfiling(ctx context.Context) (*stop.Stopper,
 			}
 		}
 
+		// Make sure the path exists.
+		if err := os.MkdirAll(logDir, 0755); err != nil {
+			return nil, errors.Wrap(err, "unable to create log directory")
+		}
+
 		// Note that we configured the --log-dir flag to set
 		// startContext.logDir. This is the point at which we set log-dir for the
 		// util/log package. We don't want to set it earlier to avoid spuriously
@@ -1013,15 +1027,14 @@ func setupAndInitializeLoggingAndProfiling(ctx context.Context) (*stop.Stopper,
 			return nil, err
 		}
 
-		// Make sure the path exists.
-		if err := os.MkdirAll(logDir, 0755); err != nil {
-			return nil, err
-		}
-		log.Eventf(ctx, "created log directory %s", logDir)
+		// NB: this message is a crutch until #33458 is addressed. Without it,
+		// the calls to log.Shout below can be the first use of logging, hitting
+		// the bug described in the issue.
+		log.Infof(ctx, "logging to directory %s", logDir)
 
 		// Start the log file GC daemon to remove files that make the log
 		// directory too large.
-		log.StartGCDaemon()
+		log.StartGCDaemon(ctx)
 	}
 
 	outputDirectory := "."
