@@ -23,12 +23,13 @@ import (
 	"encoding/json"
 	"fmt"
 	"io"
+	"io/ioutil"
+	"log"
 	"math/rand"
 	"net"
 	"net/url"
 	"os"
 	"os/exec"
-	"os/user"
 	"path/filepath"
 	"regexp"
 	"sort"
@@ -38,34 +39,64 @@ import (
 	"sync/atomic"
 	"time"
 
-	// "postgres" gosql driver
-	_ "github.com/lib/pq"
-
 	"github.com/armon/circbuf"
-	"github.com/pkg/errors"
-	"golang.org/x/sync/errgroup"
-
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	_ "github.com/lib/pq"
+	"github.com/pkg/errors"
+	"golang.org/x/sync/errgroup"
 )
 
 var (
-	local                 bool
-	artifacts             string
-	cockroach             string
-	encrypt               bool
-	workload              string
-	roachprod             string
-	buildTag              string
-	clusterName           string
-	clusterID             string
-	clusterWipe           bool
-	username              = os.Getenv("ROACHPROD_USER")
-	zonesF                string
-	teamCity              bool
-	testingSkipValidation bool
+	local       bool
+	cockroach   string
+	cloud                    = "gce"
+	encrypt     encryptValue = "false"
+	workload    string
+	roachprod   string
+	buildTag    string
+	clusterName string
+	clusterID   string
+	clusterWipe bool
+	zonesF      string
+	teamCity    bool
 )
 
+type encryptValue string
+
+func (v *encryptValue) String() string {
+	return string(*v)
+}
+
+func (v *encryptValue) Set(s string) error {
+	if s == "random" {
+		*v = encryptValue(s)
+		return nil
+	}
+	t, err := strconv.ParseBool(s)
+	if err != nil {
+		return err
+	}
+	*v = encryptValue(fmt.Sprint(t))
+	return nil
+}
+
+func (v *encryptValue) asBool() bool {
+	if *v == "random" {
+		return rand.Intn(2) == 0
+	}
+	t, err := strconv.ParseBool(string(*v))
+	if err != nil {
+		return false
+	}
+	return t
+}
+
+func (v *encryptValue) Type() string {
+	return "string"
+}
+
 func ifLocal(trueVal, falseVal string) string {
 	if local {
 		return trueVal
@@ -76,7 +107,7 @@ func ifLocal(trueVal, falseVal string) string {
 func filepathAbs(path string) (string, error) {
 	path, err := filepath.Abs(path)
 	if err != nil {
-		return "", errors.Wrap(err, "")
+		return "", errors.WithStack(err)
 	}
 	return path, nil
 }
@@ -96,13 +127,13 @@ func findBinary(binary, defValue string) (string, error) {
 	path, err := exec.LookPath(binary)
 	if err != nil {
 		if strings.HasPrefix(binary, "/") {
-			return "", errors.Wrap(err, "")
+			return "", errors.WithStack(err)
 		}
 		// We're unable to find the binary in PATH and "binary" is a relative path:
 		// look in the cockroach repo.
 		gopath := os.Getenv("GOPATH")
 		if gopath == "" {
-			return "", errors.Wrap(err, "")
+			gopath = filepath.Join(os.Getenv("HOME"), "go")
 		}
 
 		var binSuffix string
@@ -110,19 +141,19 @@ func findBinary(binary, defValue string) (string, error) {
 			binSuffix = ".docker_amd64"
 		}
 		dirs := []string{
-			"/src/github.com/cockroachdb/cockroach/",
-			"/src/github.com/cockroachdb/cockroach/bin" + binSuffix,
-			filepath.Join(os.ExpandEnv("PWD"), "bin"+binSuffix),
+			filepath.Join(gopath, "/src/github.com/cockroachdb/cockroach/"),
+			filepath.Join(gopath, "/src/github.com/cockroachdb/cockroach/bin"+binSuffix),
+			filepath.Join(os.ExpandEnv("$PWD"), "bin"+binSuffix),
 		}
 		for _, dir := range dirs {
-			path = filepath.Join(gopath, dir, binary)
+			path = filepath.Join(dir, binary)
 			var err2 error
 			path, err2 = exec.LookPath(path)
 			if err2 == nil {
 				return filepathAbs(path)
 			}
 		}
-		return "", errors.Wrap(err, "")
+		return "", fmt.Errorf("failed to find %q in $PATH or any of %s", binary, dirs)
 	}
 	return filepathAbs(path)
 }
@@ -302,33 +333,95 @@ func execCmdWithBuffer(ctx context.Context, l *logger, args ...string) ([]byte,
 	return out, nil
 }
 
-func makeGCEClusterName(testName, id, username string) string {
-	name := fmt.Sprintf("%s-%s-%s", username, id, testName)
+func makeGCEClusterName(name string) string {
 	name = strings.ToLower(name)
 	name = regexp.MustCompile(`[^-a-z0-9]+`).ReplaceAllString(name, "-")
 	name = regexp.MustCompile(`-+`).ReplaceAllString(name, "-")
 	return name
 }
 
-func makeClusterName(t testI) string {
-	if clusterName != "" {
-		return clusterName
-	}
-	if local {
-		return "local"
-	}
-	if username == "" {
-		usr, err := user.Current()
-		if err != nil {
-			panic(fmt.Sprintf("user.Current: %s", err))
+func makeClusterName(name string) string {
+	return makeGCEClusterName(name)
+}
+
+// MachineTypeToCPUs returns a CPU count for either a GCE or AWS
+// machine type.
+func MachineTypeToCPUs(s string) int {
+	{
+		// GCE machine types.
+		var v int
+		if _, err := fmt.Sscanf(s, "n1-standard-%d", &v); err == nil {
+			return v
+		}
+		if _, err := fmt.Sscanf(s, "n1-highcpu-%d", &v); err == nil {
+			return v
+		}
+		if _, err := fmt.Sscanf(s, "n1-highmem-%d", &v); err == nil {
+			return v
 		}
-		username = usr.Username
 	}
-	id := clusterID
-	if id == "" {
-		id = fmt.Sprintf("%d", timeutil.Now().Unix())
+
+	typeAndSize := strings.Split(s, ".")
+
+	if len(typeAndSize) == 2 {
+		size := typeAndSize[1]
+
+		switch size {
+		case "large":
+			return 2
+		case "xlarge":
+			return 4
+		case "2xlarge":
+			return 8
+		case "4xlarge":
+			return 16
+		case "9xlarge":
+			return 36
+		case "12xlarge":
+			return 48
+		case "18xlarge":
+			return 72
+		case "24xlarge":
+			return 96
+		}
 	}
-	return makeGCEClusterName(t.Name(), id, username)
+
+	fmt.Fprintf(os.Stderr, "unknown machine type: %s\n", s)
+	os.Exit(1)
+	return -1
+}
+
+func awsMachineType(cpus int) string {
+	switch {
+	case cpus <= 2:
+		return "c5d.large"
+	case cpus <= 4:
+		return "c5d.xlarge"
+	case cpus <= 8:
+		return "c5d.2xlarge"
+	case cpus <= 16:
+		return "c5d.4xlarge"
+	case cpus <= 36:
+		return "c5d.9xlarge"
+	case cpus <= 72:
+		return "c5d.18xlarge"
+	case cpus <= 96:
+		// There is no c5d.24xlarge.
+		return "m5d.24xlarge"
+	default:
+		panic(fmt.Sprintf("no aws machine type with %d cpus", cpus))
+	}
+}
+
+func gceMachineType(cpus int) string {
+	// TODO(peter): This is awkward: below 16 cpus, use n1-standard so that the
+	// machines have a decent amount of RAM. We could use customer machine
+	// configurations, but the rules for the amount of RAM per CPU need to be
+	// determined (you can't request any arbitrary amount of RAM).
+	if cpus < 16 {
+		return fmt.Sprintf("n1-standard-%d", cpus)
+	}
+	return fmt.Sprintf("n1-highcpu-%d", cpus)
 }
 
 type testI interface {
@@ -336,6 +429,10 @@ type testI interface {
 	Fatal(args ...interface{})
 	Fatalf(format string, args ...interface{})
 	Failed() bool
+	// Path to a directory where the test is supposed to store its log and other
+	// artifacts.
+	ArtifactsDir() string
+	logger() *logger
 }
 
 // TODO(tschottdorf): Consider using a more idiomatic approach in which options
@@ -408,19 +505,60 @@ func (n nodeListOption) String() string {
 	return buf.String()
 }
 
-type nodeSpec struct {
-	Count       int
-	CPUs        int
-	MachineType string
-	Zones       string
-	Geo         bool
-	Lifetime    time.Duration
+// clusterSpec represents a test's description of what its cluster needs to
+// look like. It becomes part of a clusterConfig when the cluster is created.
+type clusterSpec struct {
+	NodeCount int
+	// CPUs is the number of CPUs per node.
+	CPUs     int
+	Zones    string
+	Geo      bool
+	Lifetime time.Duration
 }
 
-func (s *nodeSpec) args() []string {
+func makeClusterSpec(nodeCount int, opts ...createOption) clusterSpec {
+	spec := clusterSpec{
+		NodeCount: nodeCount,
+	}
+	cpu(4).apply(&spec)
+	for _, o := range opts {
+		o.apply(&spec)
+	}
+	return spec
+}
+
+func (s *clusterSpec) String() string {
+	str := fmt.Sprintf("n%dcpu%d", s.NodeCount, s.CPUs)
+	if s.Geo {
+		str += "-geo"
+	}
+	return str
+}
+
+func (s *clusterSpec) args() []string {
 	var args []string
-	if s.MachineType != "" {
-		args = append(args, "--gce-machine-type="+s.MachineType)
+
+	switch cloud {
+	case "aws":
+		if s.Zones != "" {
+			fmt.Fprintf(os.Stderr, "zones spec not yet supported on AWS: %s\n", s.Zones)
+			os.Exit(1)
+		}
+		if s.Geo {
+			fmt.Fprintf(os.Stderr, "geo-distributed clusters not yet supported on AWS\n")
+			os.Exit(1)
+		}
+
+		args = append(args, "--clouds=aws")
+	}
+
+	if !local && s.CPUs != 0 {
+		switch cloud {
+		case "aws":
+			args = append(args, "--aws-machine-type-ssd="+awsMachineType(s.CPUs))
+		case "gce":
+			args = append(args, "--gce-machine-type="+gceMachineType(s.CPUs))
+		}
 	}
 	if s.Zones != "" {
 		args = append(args, "--gce-zones="+s.Zones)
@@ -434,7 +572,7 @@ func (s *nodeSpec) args() []string {
 	return args
 }
 
-func (s *nodeSpec) expiration() time.Time {
+func (s *clusterSpec) expiration() time.Time {
 	l := s.Lifetime
 	if l == 0 {
 		l = 12 * time.Hour
@@ -443,24 +581,13 @@ func (s *nodeSpec) expiration() time.Time {
 }
 
 type createOption interface {
-	apply(spec *nodeSpec)
+	apply(spec *clusterSpec)
 }
 
 type nodeCPUOption int
 
-func (o nodeCPUOption) apply(spec *nodeSpec) {
+func (o nodeCPUOption) apply(spec *clusterSpec) {
 	spec.CPUs = int(o)
-	if !local {
-		// TODO(peter): This is awkward: below 16 cpus, use n1-standard so that the
-		// machines have a decent amount of RAM. We could use customer machine
-		// configurations, but the rules for the amount of RAM per CPU need to be
-		// determined (you can't request any arbitrary amount of RAM).
-		if spec.CPUs < 16 {
-			spec.MachineType = fmt.Sprintf("n1-standard-%d", spec.CPUs)
-		} else {
-			spec.MachineType = fmt.Sprintf("n1-highcpu-%d", spec.CPUs)
-		}
-	}
 }
 
 // cpu is a node option which requests nodes with the specified number of CPUs.
@@ -470,7 +597,7 @@ func cpu(n int) nodeCPUOption {
 
 type nodeGeoOption struct{}
 
-func (o nodeGeoOption) apply(spec *nodeSpec) {
+func (o nodeGeoOption) apply(spec *clusterSpec) {
 	spec.Geo = true
 }
 
@@ -481,7 +608,7 @@ func geo() nodeGeoOption {
 
 type nodeZonesOption string
 
-func (o nodeZonesOption) apply(spec *nodeSpec) {
+func (o nodeZonesOption) apply(spec *clusterSpec) {
 	spec.Zones = string(o)
 }
 
@@ -494,40 +621,57 @@ func zones(s string) nodeZonesOption {
 
 type nodeLifetimeOption time.Duration
 
-func (o nodeLifetimeOption) apply(spec *nodeSpec) {
+func (o nodeLifetimeOption) apply(spec *clusterSpec) {
 	spec.Lifetime = time.Duration(o)
 }
 
-// nodes is a helper method for creating a []nodeSpec given a node count and
-// options.
-func nodes(count int, opts ...createOption) []nodeSpec {
-	spec := nodeSpec{
-		Count: count,
-	}
-	cpu(4).apply(&spec)
-	for _, o := range opts {
-		o.apply(&spec)
-	}
-	return []nodeSpec{spec}
-}
-
 // cluster provides an interface for interacting with a set of machines,
 // starting and stopping a cockroach cluster on a subset of those machines, and
 // running load generators and other operations on the machines.
 //
-// A cluster is intended to be used only by a single test. Sharing of a cluster
-// between a test and a subtest is current disallowed (see cluster.assertT). A
-// cluster is safe for concurrent use by multiple goroutines.
+// A cluster is intended to be shared between all the subtests under a root.
+// A cluster is safe for concurrent use by multiple goroutines.
 type cluster struct {
-	name       string
-	nodes      int
-	status     func(...interface{})
-	t          testI
-	l          *logger
+	name   string
+	nodes  int
+	status func(...interface{})
+	t      testI
+	// l is the logger used to log various cluster operations.
+	// DEPRECATED for use outside of cluster methods: Use a test's t.l instead.
+	// This is generally set to the current test's logger.
+	l *logger
+	// destroyed is used to coordinate between different goroutines that want to
+	// destroy a cluster. It is nil when the cluster should not be destroyed (i.e.
+	// when Destroy() should not be called).
 	destroyed  chan struct{}
 	expiration time.Time
+	// owned is set if this instance is responsible for `roachprod destroy`ing the
+	// cluster. It is set when a new cluster is created, but not when one is
+	// cloned or when we attach to an existing roachprod cluster.
+	// If not set, Destroy() only wipes the cluster.
+	owned bool
+	// encryptDefault is true if the cluster should default to having encryption
+	// at rest enabled. The default only applies if encryption is not explicitly
+	// enabled or disabled by options passed to Start.
+	encryptDefault bool
 }
 
+type clusterConfig struct {
+	// name must be empty if localCluster is specified.
+	name  string
+	nodes clusterSpec
+	// artifactsDir is the path where log file will be stored.
+	artifactsDir string
+	localCluster bool
+	teeOpt       teeOptType
+	user         string
+	useIOBarrier bool
+}
+
+// newCluster creates a new roachprod cluster.
+//
+// NOTE: setTest() needs to be called before a test can use this cluster.
+//
 // TODO(peter): Should set the lifetime of clusters to 2x the expected test
 // duration. The default lifetime of 12h is too long for some tests and will be
 // too short for others.
@@ -536,120 +680,181 @@ type cluster struct {
 // to figure out how to make that work with `roachprod create`. Perhaps one
 // invocation of `roachprod create` per unique node-spec. Are there guarantees
 // we're making here about the mapping of nodeSpecs to node IDs?
-func newCluster(ctx context.Context, t testI, nodes []nodeSpec) *cluster {
+func newCluster(ctx context.Context, l *logger, cfg clusterConfig) (*cluster, error) {
 	if atomic.LoadInt32(&interrupted) == 1 {
-		t.Fatal("interrupted")
+		return nil, fmt.Errorf("newCluster interrupted")
 	}
 
-	switch {
-	case len(nodes) == 0:
-		return nil
-	case len(nodes) > 1:
-		// TODO(peter): Need a motivating test that has different specs per node.
-		t.Fatalf("TODO(peter): unsupported nodes spec: %v", nodes)
+	var name string
+	if cfg.localCluster {
+		if cfg.name != "" {
+			log.Fatal(ctx, "can't specify name %q with local flag", cfg.name)
+		}
+		name = "local" // The roachprod tool understands this magic name.
+	} else {
+		name = makeClusterName(cfg.user + "-" + cfg.name)
 	}
 
-	l, err := rootLogger(t.Name())
-	if err != nil {
-		t.Fatal(err)
+	if cfg.nodes.NodeCount == 0 {
+		// For tests. Return the minimum that makes them happy.
+		return &cluster{
+			expiration: timeutil.Now().Add(24 * time.Hour),
+		}, nil
 	}
 
 	c := &cluster{
-		name:       makeClusterName(t),
-		nodes:      nodes[0].Count,
-		status:     func(...interface{}) {},
-		t:          t,
-		l:          l,
-		destroyed:  make(chan struct{}),
-		expiration: nodes[0].expiration(),
-	}
-	if impl, ok := t.(*test); ok {
-		c.status = impl.Status
+		name:           name,
+		nodes:          cfg.nodes.NodeCount,
+		status:         func(...interface{}) {},
+		l:              l,
+		destroyed:      make(chan struct{}),
+		expiration:     cfg.nodes.expiration(),
+		owned:          true,
+		encryptDefault: encrypt.asBool(),
 	}
 	registerCluster(c)
 
-	if c.name != clusterName {
-		sargs := []string{roachprod, "create", c.name, "-n", fmt.Sprint(c.nodes)}
-		sargs = append(sargs, nodes[0].args()...)
-		if !local && zonesF != "" && nodes[0].Zones == "" {
-			sargs = append(sargs, "--gce-zones="+zonesF)
-		}
-
-		c.status("creating cluster")
-		if err := execCmd(ctx, l, sargs...); err != nil {
-			t.Fatal(err)
-			return nil
-		}
-	} else if !testingSkipValidation {
-		// Perform validation on the existing cluster.
-		c.status("checking that existing cluster matches spec")
-		sargs := []string{roachprod, "list", c.name, "--json"}
-		out, err := execCmdWithBuffer(ctx, l, sargs...)
-		if err != nil {
-			t.Fatal(err)
-			return nil
-		}
-
-		// jsonOutput matches the structure of the output from `roachprod list`
-		// when in json mode.
-		type jsonOutput struct {
-			Clusters map[string]struct {
-				VMs []struct {
-					MachineType string `json:"machine_type"`
-				} `json:"vms"`
-			} `json:"clusters"`
-		}
-		var details jsonOutput
-		if err := json.Unmarshal(out, &details); err != nil {
-			t.Fatal(err)
-			return nil
-		}
-
-		cDetails, ok := details.Clusters[c.name]
-		if !ok {
-			t.Fatalf("cluster %q not found", c.name)
-			return nil
-		}
-		if len(cDetails.VMs) < c.nodes {
-			t.Fatalf("cluster has %d nodes, test requires at least %d", len(cDetails.VMs), c.nodes)
-			return nil
-		}
-		if typ := nodes[0].MachineType; typ != "" {
-			for i, vm := range cDetails.VMs {
-				if vm.MachineType != typ {
-					t.Fatalf("node %d has machine type %s, test requires %s", i, vm.MachineType, typ)
-					return nil
-				}
-			}
-		}
-
-		c.status("stopping cluster")
-		c.Stop(ctx, c.All())
-		if clusterWipe {
-			c.Wipe(ctx, c.All())
-		} else {
-			l.Printf("skipping cluster wipe\n")
-		}
+	sargs := []string{roachprod, "create", c.name, "-n", fmt.Sprint(c.nodes)}
+	sargs = append(sargs, cfg.nodes.args()...)
+	if !local && zonesF != "" && cfg.nodes.Zones == "" {
+		sargs = append(sargs, "--gce-zones="+zonesF)
 	}
+	if !cfg.useIOBarrier {
+		sargs = append(sargs, "--local-ssd-no-ext4-barrier")
+	}
+
+	c.status("creating cluster")
+	if err := execCmd(ctx, l, sargs...); err != nil {
+		return nil, err
+	}
+
 	c.status("running test")
-	return c
+	return c, nil
 }
 
-// clone creates a new cluster object that refers to the same cluster as the
-// receiver, but is associated with the specified test.
-func (c *cluster) clone(t *test) *cluster {
-	l, err := rootLogger(t.Name())
-	if err != nil {
-		t.Fatal(err)
-	}
-	return &cluster{
-		name:       c.name,
-		nodes:      c.nodes,
-		status:     t.Status,
-		t:          t,
+type attachOpt struct {
+	skipValidation bool
+	// Implies skipWipe.
+	skipStop bool
+	skipWipe bool
+}
+
+// attachToExistingCluster creates a cluster object based on machines that have
+// already been already allocated by roachprod.
+//
+// NOTE: setTest() needs to be called before a test can use this cluster.
+func attachToExistingCluster(
+	ctx context.Context, name string, l *logger, nodes clusterSpec, opt attachOpt,
+) (*cluster, error) {
+	c := &cluster{
+		name:       name,
+		nodes:      nodes.NodeCount,
+		status:     func(...interface{}) {},
 		l:          l,
-		expiration: c.expiration,
+		destroyed:  make(chan struct{}),
+		expiration: nodes.expiration(),
+		// If we're attaching to an existing cluster, we're not going to destoy it.
+		owned:          false,
+		encryptDefault: encrypt.asBool(),
 	}
+	registerCluster(c)
+
+	if !opt.skipValidation {
+		if err := c.validate(ctx, nodes, l); err != nil {
+			return nil, err
+		}
+	}
+
+	if !opt.skipStop {
+		c.status("stopping cluster")
+		if err := c.StopE(ctx, c.All()); err != nil {
+			return nil, err
+		}
+		if !opt.skipWipe {
+			if clusterWipe {
+				if err := c.WipeE(ctx, c.All()); err != nil {
+					return nil, err
+				}
+			} else {
+				l.Printf("skipping cluster wipe\n")
+			}
+		}
+	}
+
+	c.status("running test")
+	return c, nil
+}
+
+// setTest prepares c for being used on behalf of t.
+//
+// TODO(andrei): Get rid of c.t, c.l and of this method.
+func (c *cluster) setTest(t testI) {
+	c.t = t
+	c.l = t.logger()
+	if impl, ok := t.(*test); ok {
+		c.status = impl.Status
+	}
+}
+
+// validateCluster takes a cluster and checks that the reality corresponds to
+// the cluster's spec. It's intended to be used with clusters created by
+// attachToExistingCluster(); otherwise, clusters create with newCluster() are
+// know to be up to spec.
+func (c *cluster) validate(ctx context.Context, nodes clusterSpec, l *logger) error {
+	// Perform validation on the existing cluster.
+	c.status("checking that existing cluster matches spec")
+	sargs := []string{roachprod, "list", c.name, "--json"}
+	out, err := execCmdWithBuffer(ctx, l, sargs...)
+	if err != nil {
+		return err
+	}
+
+	// jsonOutput matches the structure of the output from `roachprod list`
+	// when in json mode.
+	type jsonOutput struct {
+		Clusters map[string]struct {
+			VMs []struct {
+				MachineType string `json:"machine_type"`
+			} `json:"vms"`
+		} `json:"clusters"`
+	}
+	var details jsonOutput
+	if err := json.Unmarshal(out, &details); err != nil {
+		return err
+	}
+
+	cDetails, ok := details.Clusters[c.name]
+	if !ok {
+		return fmt.Errorf("cluster %q not found", c.name)
+	}
+	if len(cDetails.VMs) < c.nodes {
+		return fmt.Errorf("cluster has %d nodes, test requires at least %d", len(cDetails.VMs), c.nodes)
+	}
+	if cpus := nodes.CPUs; cpus != 0 {
+		for i, vm := range cDetails.VMs {
+			vmCPUs := MachineTypeToCPUs(vm.MachineType)
+			if vmCPUs < cpus {
+				return fmt.Errorf("node %d has %d CPUs, test requires %d", i, vmCPUs, cpus)
+			}
+		}
+	}
+	return nil
+}
+
+// clone creates a non-owned handle on the same cluster.
+//
+// NOTE: If the cloning has been done for a subtest, setTest() needs to be
+// called on the returned cluster.
+//
+// TODO(andrei): Get rid of the funky concept of cloning once we implement a
+// more principled aproach to wiping and destroying cluster.
+func (c *cluster) clone() *cluster {
+	cpy := *c
+	// This cloned cluster is not taking ownership. The parent retains it.
+	cpy.owned = false
+	cpy.encryptDefault = encrypt.asBool()
+	cpy.destroyed = nil
+	return &cpy
 }
 
 // All returns a node list containing all of the nodes in the cluster.
@@ -674,18 +879,110 @@ func (c *cluster) Node(i int) nodeListOption {
 	return c.Range(i, i)
 }
 
-func (c *cluster) FetchLogs(ctx context.Context) {
-	// Don't hang forever if we can't fetch the logs.
-	execCtx, cancel := context.WithTimeout(ctx, 2*time.Minute)
-	defer cancel()
+// FetchLogs downloads the logs from the cluster using `roachprod get`.
+// The logs will be placed in the test's artifacts dir.
+func (c *cluster) FetchLogs(ctx context.Context) error {
+	if c.nodes == 0 {
+		// No nodes can happen during unit tests and implies nothing to do.
+		return nil
+	}
 
-	c.status("retrieving logs")
-	_ = execCmd(execCtx, c.l, roachprod, "get", c.name, "logs",
-		filepath.Join(artifacts, teamCityNameEscape(c.t.Name()), "logs"))
+	c.l.Printf("fetching logs\n")
+	c.status("fetching logs")
+
+	// Don't hang forever if we can't fetch the logs.
+	return contextutil.RunWithTimeout(ctx, "fetch logs", 2*time.Minute, func(ctx context.Context) error {
+		path := filepath.Join(c.t.ArtifactsDir(), "logs")
+		if err := os.MkdirAll(filepath.Dir(path), 0755); err != nil {
+			return err
+		}
+
+		return execCmd(ctx, c.l, roachprod, "get", c.name, "logs" /* src */, path /* dest */)
+	})
+}
+
+// FetchDebugZip downloads the debug zip from the cluster using `roachprod ssh`.
+// The logs will be placed in the test's artifacts dir.
+func (c *cluster) FetchDebugZip(ctx context.Context) error {
+	if c.nodes == 0 {
+		// No nodes can happen during unit tests and implies nothing to do.
+		return nil
+	}
+
+	c.l.Printf("fetching debug zip\n")
+	c.status("fetching debug zip")
+
+	// Don't hang forever if we can't fetch the debug zip.
+	return contextutil.RunWithTimeout(ctx, "debug zip", 5*time.Minute, func(ctx context.Context) error {
+		const zipName = "debug.zip"
+		path := filepath.Join(c.t.ArtifactsDir(), zipName)
+		if err := os.MkdirAll(filepath.Dir(path), 0755); err != nil {
+			return err
+		}
+		// `./cockroach debug zip` is noisy. Suppress the output unless it fails.
+		output, err := execCmdWithBuffer(ctx, c.l, roachprod, "ssh", c.name+":1", "--",
+			"./cockroach", "debug", "zip", "--url", "{pgurl:1}", zipName)
+		if err != nil {
+			c.l.Printf("./cockroach debug zip failed: %s", output)
+			return err
+		}
+		return execCmd(ctx, c.l, roachprod, "get", c.name+":1", zipName /* src */, path /* dest */)
+	})
+}
+
+// FetchDmesg grabs the dmesg logs if possible. This requires being able to run
+// `sudo dmesg` on the remote nodes.
+func (c *cluster) FetchDmesg(ctx context.Context) error {
+	if c.nodes == 0 || c.isLocal() {
+		// No nodes can happen during unit tests and implies nothing to do.
+		// Also, don't grab dmesg on local runs.
+		return nil
+	}
+
+	c.l.Printf("fetching dmesg\n")
+	c.status("fetching dmesg")
+
+	// Don't hang forever.
+	return contextutil.RunWithTimeout(ctx, "dmesg", 20*time.Second, func(ctx context.Context) error {
+		const name = "dmesg.txt"
+		path := filepath.Join(c.t.ArtifactsDir(), name)
+		if err := os.MkdirAll(filepath.Dir(path), 0755); err != nil {
+			return err
+		}
+		if err := execCmd(
+			ctx, c.l, roachprod, "ssh", c.name, "--",
+			"/bin/bash", "-c", "'sudo dmesg > "+name+"'", /* src */
+		); err != nil {
+			// Don't error out because it might've worked on some nodes. Fetching will
+			// error out below but will get everything it can first.
+			c.l.Printf("during dmesg fetching: %s", err)
+		}
+		return execCmd(ctx, c.l, roachprod, "get", c.name, name /* src */, path /* dest */)
+	})
+}
+
+// FetchCores fetches any core files on the cluster.
+func (c *cluster) FetchCores(ctx context.Context) error {
+	if c.nodes == 0 || c.isLocal() {
+		// No nodes can happen during unit tests and implies nothing to do.
+		// Also, don't grab dmesg on local runs.
+		return nil
+	}
+
+	c.l.Printf("fetching cores\n")
+	c.status("fetching cores")
+
+	// Don't hang forever. The core files can be large, so we give a generous
+	// timeout.
+	return contextutil.RunWithTimeout(ctx, "cores", 60*time.Second, func(ctx context.Context) error {
+		path := filepath.Join(c.t.ArtifactsDir(), "cores")
+		return execCmd(ctx, c.l, roachprod, "get", c.name, "/tmp/cores" /* src */, path /* dest */)
+	})
 }
 
 func (c *cluster) Destroy(ctx context.Context) {
-	if c == nil {
+	if c.nodes == 0 {
+		// No nodes can happen during unit tests and implies nothing to do.
 		return
 	}
 
@@ -707,7 +1004,7 @@ func (c *cluster) destroy(ctx context.Context) {
 	defer close(c.destroyed)
 
 	if clusterWipe {
-		if c.name != clusterName {
+		if c.owned {
 			c.status("destroying cluster")
 			if err := execCmd(ctx, c.l, roachprod, "destroy", c.name); err != nil {
 				c.l.Errorf("%s", err)
@@ -749,11 +1046,64 @@ func (c *cluster) Put(ctx context.Context, src, dest string, opts ...option) {
 	}
 }
 
-// GitClone clones a git repo from src into dest and checks out
-// origin's version of the given branch. The src, dest, and branch
-// arguments must not contain shell special characters.
-func (c *cluster) GitClone(ctx context.Context, src, dest, branch string, node nodeListOption) {
-	c.Run(ctx, node, "bash", "-e", "-c", fmt.Sprintf(`'
+// Get gets files from remote hosts.
+func (c *cluster) Get(ctx context.Context, src, dest string, opts ...option) {
+	if c.t.Failed() {
+		// If the test has failed, don't try to limp along.
+		return
+	}
+	if atomic.LoadInt32(&interrupted) == 1 {
+		c.t.Fatal("interrupted")
+	}
+	c.status(fmt.Sprintf("getting %v", src))
+	err := execCmd(ctx, c.l, roachprod, "get", c.makeNodes(opts...), src, dest)
+	if err != nil {
+		c.t.Fatal(err)
+	}
+}
+
+// Put a string into the specified file on the remote(s).
+func (c *cluster) PutString(
+	ctx context.Context, content, dest string, mode os.FileMode, opts ...option,
+) {
+	if c.t.Failed() {
+		// If the test has failed, don't try to limp along.
+		return
+	}
+	if atomic.LoadInt32(&interrupted) == 1 {
+		c.t.Fatal("interrupted")
+	}
+	c.status("uploading string")
+
+	temp, err := ioutil.TempFile("", filepath.Base(dest))
+	if err != nil {
+		c.t.Fatal(err)
+	}
+	if _, err := temp.WriteString(content); err != nil {
+		c.t.Fatal(err)
+	}
+	temp.Close()
+	src := temp.Name()
+
+	if err := os.Chmod(src, mode); err != nil {
+		c.t.Fatal(err)
+	}
+	// NB: we intentionally don't remove the temp files. This is because roachprod
+	// will symlink them when running locally.
+
+	if err := execCmd(ctx, c.l, roachprod, "put", c.makeNodes(opts...), src, dest); err != nil {
+		c.t.Fatal(err)
+	}
+}
+
+// GitCloneE clones a git repo from src into dest and checks out origin's
+// version of the given branch. The src, dest, and branch arguments must not
+// contain shell special characters. GitCloneE unlike GitClone returns an
+// error.
+func (c *cluster) GitCloneE(
+	ctx context.Context, src, dest, branch string, node nodeListOption,
+) error {
+	return c.RunE(ctx, node, "bash", "-e", "-c", fmt.Sprintf(`'
 if ! test -d %s; then
   git clone -b %s --depth 1 %s %s
 else
@@ -767,6 +1117,15 @@ fi
 		branch))
 }
 
+// GitClone clones a git repo from src into dest and checks out origin's
+// version of the given branch. The src, dest, and branch arguments must not
+// contain shell special characters.
+func (c *cluster) GitClone(ctx context.Context, src, dest, branch string, node nodeListOption) {
+	if err := c.GitCloneE(ctx, src, dest, branch, node); err != nil {
+		c.t.Fatal(err)
+	}
+}
+
 // startArgs specifies extra arguments that are passed to `roachprod` during `c.Start`.
 func startArgs(extraArgs ...string) option {
 	return roachprodArgOption(extraArgs)
@@ -804,16 +1163,16 @@ func roachprodArgs(opts []option) []string {
 	return args
 }
 
-// Start cockroach nodes on a subset of the cluster. The nodes parameter can
-// either be a specific node, empty (to indicate all nodes), or a pair of nodes
-// indicating a range.
-func (c *cluster) Start(ctx context.Context, opts ...option) {
-	if c.t.Failed() {
-		// If the test has failed, don't try to limp along.
-		return
-	}
+// StartE starts cockroach nodes on a subset of the cluster. The nodes parameter
+// can either be a specific node, empty (to indicate all nodes), or a pair of
+// nodes indicating a range.
+func (c *cluster) StartE(ctx context.Context, opts ...option) error {
 	if atomic.LoadInt32(&interrupted) == 1 {
-		c.t.Fatal("interrupted")
+		return fmt.Errorf("cluster.Start() interrupted")
+	}
+	// If the test failed (indicated by a canceled ctx), short-circuit.
+	if ctx.Err() != nil {
+		return ctx.Err()
 	}
 	c.status("starting cluster")
 	defer c.status()
@@ -823,19 +1182,15 @@ func (c *cluster) Start(ctx context.Context, opts ...option) {
 	}
 	args = append(args, roachprodArgs(opts)...)
 	args = append(args, c.makeNodes(opts...))
-	if encrypt && !argExists(args, "--encrypt") {
+	if !argExists(args, "--encrypt") && c.encryptDefault {
 		args = append(args, "--encrypt")
 	}
-	if local {
-		// This avoids annoying firewall prompts on macos.
-		// NB: we have to use the deprecated --host flag here
-		// (and not --listen-addr) until we don't run any mixed
-		// version tests involving v2.0 any more.
-		args = append(args, "--args", "--host=127.0.0.1")
-	}
-	if err := execCmd(ctx, c.l, args...); err != nil {
-		c.t.Fatal(err)
-	}
+	return execCmd(ctx, c.l, args...)
+}
+
+// Start is like StartE() except it takes a test and, on error, calls t.Fatal().
+func (c *cluster) Start(ctx context.Context, t *test, opts ...option) {
+	FatalIfErr(t, c.StartE(ctx, opts...))
 }
 
 func argExists(args []string, target string) bool {
@@ -847,14 +1202,12 @@ func argExists(args []string, target string) bool {
 	return false
 }
 
-// Stop cockroach nodes running on a subset of the cluster. See cluster.Start()
+// StopE cockroach nodes running on a subset of the cluster. See cluster.Start()
 // for a description of the nodes parameter.
-func (c *cluster) Stop(ctx context.Context, opts ...option) {
-	if c.t.Failed() {
-		// If the test has failed, don't try to limp along.
-		return
+func (c *cluster) StopE(ctx context.Context, opts ...option) error {
+	if ctx.Err() != nil {
+		return ctx.Err()
 	}
-
 	args := []string{
 		roachprod,
 		"stop",
@@ -862,30 +1215,47 @@ func (c *cluster) Stop(ctx context.Context, opts ...option) {
 	args = append(args, roachprodArgs(opts)...)
 	args = append(args, c.makeNodes(opts...))
 	if atomic.LoadInt32(&interrupted) == 1 {
-		c.t.Fatal("interrupted")
+		return fmt.Errorf("interrupted")
 	}
 	c.status("stopping cluster")
 	defer c.status()
-	err := execCmd(ctx, c.l, args...)
-	if err != nil {
+	return execCmd(ctx, c.l, args...)
+}
+
+// Stop is like StopE, except instead of returning an error, it does
+// c.t.Fatal(). c.t needs to be set.
+func (c *cluster) Stop(ctx context.Context, opts ...option) {
+	if c.t.Failed() {
+		// If the test has failed, don't try to limp along.
+		return
+	}
+	if err := c.StopE(ctx, opts...); err != nil {
 		c.t.Fatal(err)
 	}
 }
 
-// Wipe a subset of the nodes in a cluster. See cluster.Start() for a
+// WipeE wipes a subset of the nodes in a cluster. See cluster.Start() for a
 // description of the nodes parameter.
+func (c *cluster) WipeE(ctx context.Context, opts ...option) error {
+	if ctx.Err() != nil {
+		return ctx.Err()
+	}
+	if atomic.LoadInt32(&interrupted) == 1 {
+		return fmt.Errorf("interrupted")
+	}
+	c.status("wiping cluster")
+	defer c.status()
+	return execCmd(ctx, c.l, roachprod, "wipe", c.makeNodes(opts...))
+}
+
+// Wipe is like WipeE, except instead of returning an error, it does
+// c.t.Fatal(). c.t needs to be set.
 func (c *cluster) Wipe(ctx context.Context, opts ...option) {
 	if c.t.Failed() {
 		// If the test has failed, don't try to limp along.
 		return
 	}
-	if atomic.LoadInt32(&interrupted) == 1 {
-		c.t.Fatal("interrupted")
-	}
-	c.status("wiping cluster")
-	defer c.status()
-	err := execCmd(ctx, c.l, roachprod, "wipe", c.makeNodes(opts...))
-	if err != nil {
+	if err := c.WipeE(ctx, opts...); err != nil {
 		c.t.Fatal(err)
 	}
 }
@@ -955,14 +1325,6 @@ func (c *cluster) RunWithBuffer(
 		append([]string{roachprod, "run", c.makeNodes(node), "--"}, args...)...)
 }
 
-// RemountNoBarrier remounts the cluster's local SSDs with the nobarrier option.
-func (c *cluster) RemountNoBarrier(ctx context.Context) {
-	c.Run(ctx, c.All(),
-		"sudo", "umount", "/mnt/data1", ";",
-		"sudo", "mount", "-o", "discard,defaults,nobarrier",
-		"/dev/disk/by-id/google-local-ssd-0", "/mnt/data1")
-}
-
 // pgURL returns the Postgres endpoint for the specified node. It accepts a flag
 // specifying whether the URL should include the node's internal or external IP
 // address. In general, inter-cluster communication and should use internal IPs
@@ -1021,12 +1383,21 @@ func urlToAddr(c *cluster, pgURL string) string {
 	return u.Host
 }
 
-func addrToIP(c *cluster, addr string) string {
-	host, _, err := net.SplitHostPort(addr)
+func addrToHost(c *cluster, addr string) string {
+	host, _ := addrToHostPort(c, addr)
+	return host
+}
+
+func addrToHostPort(c *cluster, addr string) (string, int) {
+	host, portStr, err := net.SplitHostPort(addr)
 	if err != nil {
 		c.t.Fatal(err)
 	}
-	return host
+	port, err := strconv.Atoi(portStr)
+	if err != nil {
+		c.t.Fatal(err)
+	}
+	return host, port
 }
 
 // InternalAdminUIAddr returns the internal Admin UI address in the form host:port
@@ -1063,7 +1434,7 @@ func (c *cluster) InternalAddr(ctx context.Context, node nodeListOption) []strin
 func (c *cluster) InternalIP(ctx context.Context, node nodeListOption) []string {
 	var ips []string
 	for _, addr := range c.InternalAddr(ctx, node) {
-		ips = append(ips, addrToIP(c, addr))
+		ips = append(ips, addrToHost(c, addr))
 	}
 	return ips
 }
@@ -1082,7 +1453,7 @@ func (c *cluster) ExternalAddr(ctx context.Context, node nodeListOption) []strin
 func (c *cluster) ExternalIP(ctx context.Context, node nodeListOption) []string {
 	var ips []string
 	for _, addr := range c.ExternalAddr(ctx, node) {
-		ips = append(ips, addrToIP(c, addr))
+		ips = append(ips, addrToHost(c, addr))
 	}
 	return ips
 }
@@ -1100,6 +1471,16 @@ func (c *cluster) Conn(ctx context.Context, node int) *gosql.DB {
 	return db
 }
 
+// ConnE returns a SQL connection to the specified node.
+func (c *cluster) ConnE(ctx context.Context, node int) (*gosql.DB, error) {
+	url := c.ExternalPGUrl(ctx, c.Node(node))[0]
+	db, err := gosql.Open("postgres", url)
+	if err != nil {
+		return nil, err
+	}
+	return db, nil
+}
+
 func (c *cluster) makeNodes(opts ...option) string {
 	var r nodeListOption
 	for _, o := range opts {
@@ -1114,10 +1495,24 @@ func (c *cluster) isLocal() bool {
 	return c.name == "local"
 }
 
-func getDiskUsageInByte(ctx context.Context, c *cluster, nodeIdx int) (int, error) {
-	out, err := c.RunWithBuffer(ctx, c.l, c.Node(nodeIdx), fmt.Sprint("du -sk {store-dir} | grep -oE '^[0-9]+'"))
-	if err != nil {
-		return 0, err
+// getDiskUsageInBytes does what's on the tin. nodeIdx starts at one.
+func getDiskUsageInBytes(
+	ctx context.Context, c *cluster, logger *logger, nodeIdx int,
+) (int, error) {
+	var out []byte
+	for {
+		if c.t.Failed() {
+			return 0, errors.New("already failed")
+		}
+		var err error
+		out, err = c.RunWithBuffer(ctx, logger, c.Node(nodeIdx), fmt.Sprint("du -sk {store-dir} | grep -oE '^[0-9]+'"))
+		if err != nil {
+			// `du` can fail if files get removed out from under it. RocksDB likes to do that
+			// during compactions and such. It's rare enough to just retry.
+			logger.Printf("retrying disk usage computation after spurious error: %s", err)
+			continue
+		}
+		break
 	}
 
 	str := string(out)
@@ -1168,6 +1563,10 @@ func (m *monitor) ExpectDeaths(count int32) {
 	atomic.AddInt32(&m.expDeaths, count)
 }
 
+func (m *monitor) ResetDeaths() {
+	atomic.StoreInt32(&m.expDeaths, 0)
+}
+
 var errGoexit = errors.New("Goexit() was called")
 
 func (m *monitor) Go(fn func(context.Context) error) {
@@ -1229,7 +1628,7 @@ func (m *monitor) Wait() {
 }
 
 func (m *monitor) wait(args ...string) error {
-	// It is surprisingly difficult to get the cancelation semantics exactly
+	// It is surprisingly difficult to get the cancellation semantics exactly
 	// right. We need to watch for the "workers" group (m.g) to finish, or for
 	// the monitor command to emit an unexpected node failure, or for the monitor
 	// command itself to exit. We want to capture whichever error happens first
@@ -1243,12 +1642,12 @@ func (m *monitor) wait(args ...string) error {
 	//   })
 	//
 	// Now consider what happens when an error is returned. Before the error
-	// reaches the errgroup, we invoke the cancelation closure which can cause
+	// reaches the errgroup, we invoke the cancellation closure which can cause
 	// the other goroutines to wake up and perhaps race and set the errgroup
 	// error first.
 	//
 	// The solution is to implement our own errgroup mechanism here which allows
-	// us to set the error before performing the cancelation.
+	// us to set the error before performing the cancellation.
 
 	var errOnce sync.Once
 	var err error
