@@ -28,7 +28,6 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
-
 	"github.com/kr/pretty"
 	_ "github.com/lib/pq"
 	"github.com/pkg/errors"
@@ -55,15 +54,17 @@ func runDecommission(t *test, c *cluster, nodes int, duration time.Duration) {
 	c.Put(ctx, cockroach, "./cockroach", c.All())
 
 	for i := 1; i <= numDecom; i++ {
-		c.Start(ctx, c.Node(i), startArgs(fmt.Sprintf("-a=--attrs=node%d", i)))
+		c.Start(ctx, t, c.Node(i), startArgs(fmt.Sprintf("-a=--attrs=node%d", i)))
 	}
 
-	c.Start(ctx, c.Range(numDecom+1, nodes))
+	c.Start(ctx, t, c.Range(numDecom+1, nodes))
 	c.Run(ctx, c.Node(nodes), `./workload init kv --drop`)
 
 	waitReplicatedAwayFrom := func(downNodeID string) error {
 		db := c.Conn(ctx, nodes)
-		defer db.Close()
+		defer func() {
+			_ = db.Close()
+		}()
 
 		for {
 			var count int
@@ -93,7 +94,9 @@ func runDecommission(t *test, c *cluster, nodes int, duration time.Duration) {
 
 	waitUpReplicated := func(targetNodeID string) error {
 		db := c.Conn(ctx, nodes)
-		defer db.Close()
+		defer func() {
+			_ = db.Close()
+		}()
 
 		for ok := false; !ok; {
 			stmtReplicaCount := fmt.Sprintf(
@@ -138,7 +141,7 @@ func runDecommission(t *test, c *cluster, nodes int, duration time.Duration) {
 		if err != nil {
 			t.Fatal(err)
 		}
-		c.l.Printf("run: %s\n", stmt)
+		t.l.Printf("run: %s\n", stmt)
 	}
 
 	var m *errgroup.Group // see comment in version.go
@@ -149,7 +152,7 @@ func runDecommission(t *test, c *cluster, nodes int, duration time.Duration) {
 
 		cmd = fmt.Sprintf(cmd, nodes)
 		m.Go(func() error {
-			quietL, err := c.l.ChildLogger("kv-"+strconv.Itoa(i), quietStdout)
+			quietL, err := t.l.ChildLogger("kv-"+strconv.Itoa(i), quietStdout)
 			if err != nil {
 				return err
 			}
@@ -223,7 +226,7 @@ func runDecommission(t *test, c *cluster, nodes int, duration time.Duration) {
 			db := c.Conn(ctx, 1)
 			defer db.Close()
 
-			c.Start(ctx, c.Node(node), startArgs(fmt.Sprintf("-a=--join %s --attrs=node%d",
+			c.Start(ctx, t, c.Node(node), startArgs(fmt.Sprintf("-a=--join %s --attrs=node%d",
 				c.InternalAddr(ctx, c.Node(nodes))[0], node)))
 		}
 		// TODO(tschottdorf): run some ui sanity checks about decommissioned nodes
@@ -241,37 +244,34 @@ func registerDecommission(r *registry) {
 	duration := time.Hour
 
 	r.Add(testSpec{
-		Name:   fmt.Sprintf("decommission/nodes=%d/duration=%s", numNodes, duration),
-		Nodes:  nodes(numNodes),
-		Stable: true, // DO NOT COPY to new tests
+		Name:    fmt.Sprintf("decommission/nodes=%d/duration=%s", numNodes, duration),
+		Cluster: makeClusterSpec(numNodes),
 		Run: func(ctx context.Context, t *test, c *cluster) {
 			if local {
 				duration = 3 * time.Minute
-				c.l.Printf("running with duration=%s in local mode\n", duration)
+				t.l.Printf("running with duration=%s in local mode\n", duration)
 			}
 			runDecommission(t, c, numNodes, duration)
 		},
 	})
 }
 
-func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
-	args := startArgs("--sequential", "--env=COCKROACH_SCAN_MAX_IDLE_TIME=5ms")
-	c.Put(ctx, cockroach, "./cockroach")
-	c.Start(ctx, args)
+func execCLI(
+	ctx context.Context, t *test, c *cluster, runNode int, extraArgs ...string,
+) (string, error) {
+	args := []string{"./cockroach"}
+	args = append(args, extraArgs...)
+	args = append(args, "--insecure")
+	args = append(args, fmt.Sprintf("--port={pgport:%d}", runNode))
+	buf, err := c.RunWithBuffer(ctx, t.l, c.Node(runNode), args...)
+	t.l.Printf("%s\n", buf)
+	return string(buf), err
+}
 
-	execCLI := func(
-		ctx context.Context,
-		runNode int,
-		extraArgs ...string,
-	) (string, error) {
-		args := []string{"./cockroach"}
-		args = append(args, extraArgs...)
-		args = append(args, "--insecure")
-		args = append(args, fmt.Sprintf("--port={pgport:%d}", runNode))
-		buf, err := c.RunWithBuffer(ctx, c.l, c.Node(runNode), args...)
-		c.l.Printf("%s\n", buf)
-		return string(buf), err
-	}
+func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
+	args := startArgs("--env=COCKROACH_SCAN_MAX_IDLE_TIME=5ms")
+	c.Put(ctx, cockroach, "./cockroach")
+	c.Start(ctx, t, args)
 
 	decommission := func(
 		ctx context.Context,
@@ -284,7 +284,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 		for _, target := range targetNodes {
 			args = append(args, strconv.Itoa(target))
 		}
-		return execCLI(ctx, runNode, args...)
+		return execCLI(ctx, t, c, runNode, args...)
 	}
 
 	matchCSV := func(csvStr string, matchColRow [][]string) (err error) {
@@ -340,7 +340,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 	}
 	waitLiveDeprecated := "--wait=live is deprecated and is treated as --wait=all"
 
-	c.l.Printf("decommissioning first node from the second, polling the status manually\n")
+	t.l.Printf("decommissioning first node from the second, polling the status manually\n")
 	retryOpts := retry.Options{
 		InitialBackoff: time.Second,
 		MaxBackoff:     5 * time.Second,
@@ -367,7 +367,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 	// Check that even though the node is decommissioned, we still see it (since
 	// it remains live) in `node ls`.
 	{
-		o, err := execCLI(ctx, 2, "node", "ls", "--format", "csv")
+		o, err := execCLI(ctx, t, c, 2, "node", "ls", "--format", "csv")
 		if err != nil {
 			t.Fatalf("node-ls failed: %v", err)
 		}
@@ -384,7 +384,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 	}
 	// Ditto `node status`.
 	{
-		o, err := execCLI(ctx, 2, "node", "status", "--format", "csv")
+		o, err := execCLI(ctx, t, c, 2, "node", "status", "--format", "csv")
 		if err != nil {
 			t.Fatalf("node-status failed: %v", err)
 		}
@@ -400,12 +400,12 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 		}
 	}
 
-	c.l.Printf("recommissioning first node (from third node)\n")
+	t.l.Printf("recommissioning first node (from third node)\n")
 	if _, err := decommission(ctx, 3, c.Node(1), "recommission"); err != nil {
 		t.Fatalf("recommission failed: %v", err)
 	}
 
-	c.l.Printf("decommissioning second node from third, using --wait=all\n")
+	t.l.Printf("decommissioning second node from third, using --wait=all\n")
 	{
 		o, err := decommission(ctx, 3, c.Node(2),
 			"decommission", "--wait", "all", "--format", "csv")
@@ -423,30 +423,30 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 		}
 	}
 
-	c.l.Printf("recommissioning second node from itself\n")
+	t.l.Printf("recommissioning second node from itself\n")
 	if _, err := decommission(ctx, 2, c.Node(2), "recommission"); err != nil {
 		t.Fatalf("recommission failed: %v", err)
 	}
 
-	c.l.Printf("decommissioning third node via `quit --decommission`\n")
+	t.l.Printf("decommissioning third node via `quit --decommission`\n")
 	func() {
 		// This should not take longer than five minutes, and if it does, it's
 		// likely stuck forever and we want to see the output.
 		timeoutCtx, cancel := context.WithTimeout(ctx, 5*time.Minute)
 		defer cancel()
-		if _, err := execCLI(timeoutCtx, 3, "quit", "--decommission"); err != nil {
+		if _, err := execCLI(timeoutCtx, t, c, 3, "quit", "--decommission"); err != nil {
 			if timeoutCtx.Err() != nil {
 				t.Fatalf("quit --decommission failed: %s", err)
 			}
 			// TODO(tschottdorf): grep the process output for the string announcing success?
-			c.l.Errorf("WARNING: ignoring error on quit --decommission: %s\n", err)
+			t.l.Errorf("WARNING: ignoring error on quit --decommission: %s\n", err)
 		}
 	}()
 
 	// Now that the third node is down and decommissioned, decommissioning it
 	// again should be a no-op. We do it from node one but as always it doesn't
 	// matter.
-	c.l.Printf("checking that other nodes see node three as successfully decommissioned\n")
+	t.l.Printf("checking that other nodes see node three as successfully decommissioned\n")
 	{
 		o, err := decommission(ctx, 2, c.Node(3),
 			"decommission", "--format", "csv") // wait=all is implied
@@ -467,7 +467,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 
 		// Bring the node back up. It's still decommissioned, so it won't be of much use.
 		c.Stop(ctx, c.Node(3))
-		c.Start(ctx, c.Node(3), args)
+		c.Start(ctx, t, c.Node(3), args)
 
 		// Recommission. Welcome back!
 		if _, err = decommission(ctx, 2, c.Node(3), "recommission"); err != nil {
@@ -477,9 +477,9 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 
 	// Kill the first node and verify that we can decommission it while it's down,
 	// bringing it back up to verify that its replicas still get removed.
-	c.l.Printf("intentionally killing first node\n")
+	t.l.Printf("intentionally killing first node\n")
 	c.Stop(ctx, c.Node(1))
-	c.l.Printf("decommission first node, starting with it down but restarting it for verification\n")
+	t.l.Printf("decommission first node, starting with it down but restarting it for verification\n")
 	{
 		o, err := decommission(ctx, 2, c.Node(1),
 			"decommission", "--wait", "live")
@@ -489,7 +489,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 		if strings.Split(o, "\n")[1] != waitLiveDeprecated {
 			t.Fatal("missing deprecate message for --wait=live")
 		}
-		c.Start(ctx, c.Node(1), args)
+		c.Start(ctx, t, c.Node(1), args)
 		// Run a second time to wait until the replicas have all been GC'ed.
 		// Note that we specify "all" because even though the first node is
 		// now running, it may not be live by the time the command runs.
@@ -522,13 +522,13 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 		}
 	}()
 
-	c.l.Printf("intentionally killing first node\n")
+	t.l.Printf("intentionally killing first node\n")
 	c.Stop(ctx, c.Node(1))
 	// It is being decommissioned in absentia, meaning that its replicas are
 	// being removed due to deadness. We can't see that reflected in the output
 	// since the current mechanism gets its replica counts from what the node
 	// reports about itself, so our assertion here is somewhat weak.
-	c.l.Printf("decommission first node in absentia using --wait=live\n")
+	t.l.Printf("decommission first node in absentia using --wait=live\n")
 	{
 		o, err := decommission(ctx, 3, c.Node(1),
 			"decommission", "--wait", "live", "--format", "csv")
@@ -558,7 +558,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 	// Check that (at least after a bit) the node disappears from `node ls`
 	// because it is decommissioned and not live.
 	for {
-		o, err := execCLI(ctx, 2, "node", "ls", "--format", "csv")
+		o, err := execCLI(ctx, t, c, 2, "node", "ls", "--format", "csv")
 		if err != nil {
 			t.Fatalf("node-ls failed: %v", err)
 		}
@@ -577,7 +577,7 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 		break
 	}
 	for {
-		o, err := execCLI(ctx, 2, "node", "status", "--format", "csv")
+		o, err := execCLI(ctx, t, c, 2, "node", "status", "--format", "csv")
 		if err != nil {
 			t.Fatalf("node-status failed: %v", err)
 		}
@@ -601,12 +601,12 @@ func runDecommissionAcceptance(ctx context.Context, t *test, c *cluster) {
 	// with an address belonging to an old decommissioned node.
 	{
 		c.Wipe(ctx, c.Node(1))
-		c.Start(ctx, c.Node(1), startArgs(fmt.Sprintf("-a=--join %s",
+		c.Start(ctx, t, c.Node(1), startArgs(fmt.Sprintf("-a=--join %s",
 			c.InternalAddr(ctx, c.Node(2))[0])))
 	}
 
 	if err := retry.WithMaxAttempts(ctx, retryOpts, 20, func() error {
-		o, err := execCLI(ctx, 2, "node", "status", "--format", "csv")
+		o, err := execCLI(ctx, t, c, 2, "node", "status", "--format", "csv")
 		if err != nil {
 			t.Fatalf("node-status failed: %v", err)
 		}
@@ -643,7 +643,7 @@ WHERE "eventType" IN ($1, $2) ORDER BY timestamp`,
 			"node_decommissioned", "node_recommissioned",
 		)
 		if err != nil {
-			c.l.Printf("retrying: %v\n", err)
+			t.l.Printf("retrying: %v\n", err)
 			return err
 		}
 		defer rows.Close()
