@@ -19,66 +19,128 @@ import (
 	"bytes"
 	"context"
 	"fmt"
+	"io/ioutil"
 	"math"
-	"strconv"
+	"math/rand"
+	"os"
+	"path/filepath"
 	"strings"
 	"time"
 
-	"github.com/lib/pq"
-	"github.com/pkg/errors"
-	"golang.org/x/sync/errgroup"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/util/search"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/cockroachdb/cockroach/pkg/util/version"
+	"github.com/cockroachdb/cockroach/pkg/workload/histogram"
+	"github.com/cockroachdb/cockroach/pkg/workload/tpcc"
 	"github.com/cockroachdb/ttycolor"
+	"github.com/lib/pq"
+	"github.com/pkg/errors"
+	"golang.org/x/sync/errgroup"
 )
 
-func runTPCC(
-	ctx context.Context, t *test, c *cluster, warehouses int, duration string, extraArgs ...string,
-) {
+type tpccOptions struct {
+	Warehouses int
+	Extra      string
+	Chaos      func() Chaos                // for late binding of stopper
+	During     func(context.Context) error // for running a function during the test
+	Duration   time.Duration
+	ZFS        bool
+}
+
+// tpccFixturesCmd generates the command string to load tpcc data for the
+// specified warehouse count into a cluster using either `fixtures import`
+// or `fixtures load` depending on the cloud.
+func tpccFixturesCmd(t *test, cloud string, warehouses int, checks bool) string {
+	var action string
+	switch cloud {
+	case "gce":
+		// TODO(nvanbenschoten): We could switch to import for both clouds.
+		// At the moment, import is still a little unstable and load is still
+		// marginally faster.
+		action = "load"
+		fixtureWarehouses := -1
+		for _, w := range []int{1, 10, 100, 1000, 2000, 5000, 10000} {
+			if w >= warehouses {
+				fixtureWarehouses = w
+				break
+			}
+		}
+		if fixtureWarehouses == -1 {
+			t.Fatalf("could not find fixture big enough for %d warehouses", warehouses)
+		}
+		warehouses = fixtureWarehouses
+	case "aws":
+		action = "import"
+	default:
+		t.Fatalf("unknown cloud: %q", cloud)
+	}
+	return fmt.Sprintf("./workload fixtures %s tpcc --checks=%v --warehouses=%d {pgurl:1}",
+		action, checks, warehouses)
+}
+
+func runTPCC(ctx context.Context, t *test, c *cluster, opts tpccOptions) {
 	crdbNodes := c.Range(1, c.nodes-1)
 	workloadNode := c.Node(c.nodes)
+	rampDuration := 5 * time.Minute
 	if c.isLocal() {
-		warehouses = 1
-		duration = `10s`
-	} else {
-		c.RemountNoBarrier(ctx)
+		opts.Warehouses = 1
+		opts.Duration = 1 * time.Minute
+		rampDuration = 30 * time.Second
 	}
 
 	c.Put(ctx, cockroach, "./cockroach", crdbNodes)
 	c.Put(ctx, workload, "./workload", workloadNode)
-	c.Start(ctx, crdbNodes)
 
 	t.Status("loading fixture")
-	fixtureWarehouses := -1
-	for _, w := range []int{1, 10, 100, 1000, 2000, 5000, 10000} {
-		if w >= warehouses {
-			fixtureWarehouses = w
-			break
-		}
-	}
-	if fixtureWarehouses == -1 {
-		t.Fatalf("could not find fixture big enough for %d warehouses", warehouses)
-	}
-	c.Run(ctx, workloadNode, fmt.Sprintf(
-		`./workload fixtures load tpcc --warehouses=%d {pgurl:1}`, fixtureWarehouses,
-	))
+	func() {
+		db := c.Conn(ctx, 1)
+		defer db.Close()
+		if opts.ZFS {
+			if err := c.RunE(ctx, c.Node(1), "test -d /mnt/data1/.zfs/snapshot/pristine"); err != nil {
+				// Use ZFS so the initial store dumps can be instantly rolled back to their
+				// pristine state. Useful for iterating quickly on the test, especially when
+				// used in a repro.
+				c.Reformat(ctx, crdbNodes, "zfs")
 
-	t.Status("running tpcc")
+				t.Status("loading dataset")
+				c.Start(ctx, t, crdbNodes)
+
+				c.Run(ctx, workloadNode, tpccFixturesCmd(t, cloud, opts.Warehouses, true /* checks */))
+				c.Stop(ctx, crdbNodes)
+
+				c.Run(ctx, crdbNodes, "test -e /sbin/zfs && sudo zfs snapshot data1@pristine")
+			}
+			t.Status(`restoring store dumps`)
+			c.Run(ctx, crdbNodes, "sudo zfs rollback data1@pristine")
+			c.Start(ctx, t, crdbNodes)
+		} else {
+			c.Start(ctx, t, crdbNodes)
+			c.Run(ctx, workloadNode, tpccFixturesCmd(t, cloud, opts.Warehouses, true /* checks */))
+		}
+	}()
+	t.Status("waiting")
 	m := newMonitor(ctx, c, crdbNodes)
 	m.Go(func(ctx context.Context) error {
+		t.WorkerStatus("running tpcc")
 		cmd := fmt.Sprintf(
-			"./workload run tpcc --warehouses=%d --duration=%s {pgurl%s}",
-			warehouses, duration, crdbNodes)
-		cmd += ` --histograms=logs/stats.json`
-		for _, extraArg := range extraArgs {
-			cmd += extraArg
-		}
+			"./workload run tpcc --warehouses=%d --histograms=logs/stats.json "+
+				opts.Extra+" --ramp=%s --duration=%s {pgurl:1-%d}",
+			opts.Warehouses, rampDuration, opts.Duration, c.nodes-1)
 		c.Run(ctx, workloadNode, cmd)
 		return nil
 	})
+	if opts.Chaos != nil {
+		chaos := opts.Chaos()
+		m.Go(chaos.Runner(c, m))
+	}
+	if opts.During != nil {
+		m.Go(opts.During)
+	}
 	m.Wait()
+
+	c.Run(ctx, workloadNode, fmt.Sprintf(
+		"./workload check tpcc --warehouses=%d {pgurl:1}", opts.Warehouses))
 }
 
 func registerTPCC(r *registry) {
@@ -87,34 +149,126 @@ func registerTPCC(r *registry) {
 		// TODO(dan): Instead of MinVersion, adjust the warehouses below to
 		// match our expectation for the max tpcc warehouses that previous
 		// releases will support on this hardware.
-		MinVersion: "2.1.0",
-		Nodes:      nodes(4, cpu(16)),
-		Stable:     false,
+		MinVersion: maxVersion("v2.1.0", maybeMinVersionForFixturesImport(cloud)),
+		Cluster:    makeClusterSpec(4, cpu(16)),
 		Run: func(ctx context.Context, t *test, c *cluster) {
-			warehouses := 1400
-			runTPCC(ctx, t, c, warehouses, `120m`)
+			var warehouses int
+			switch cloud {
+			case "gce":
+				warehouses = 1250
+			case "aws":
+				warehouses = 2100
+			default:
+				t.Fatalf("unknown cloud: %q", cloud)
+			}
+			runTPCC(ctx, t, c, tpccOptions{
+				Warehouses: warehouses,
+				Duration:   120 * time.Minute,
+			})
 		},
 	})
 	r.Add(testSpec{
-		Name:   "tpcc-nowait/nodes=3/w=1",
-		Nodes:  nodes(4, cpu(16)),
-		Stable: false,
+		Name:       "tpcc-nowait/nodes=3/w=1",
+		MinVersion: maybeMinVersionForFixturesImport(cloud),
+		Cluster:    makeClusterSpec(4, cpu(16)),
 		Run: func(ctx context.Context, t *test, c *cluster) {
-			runTPCC(ctx, t, c, 1, `10m`, ` --wait=false`)
+			runTPCC(ctx, t, c, tpccOptions{
+				Warehouses: 1,
+				Duration:   10 * time.Minute,
+				Extra:      "--wait=false",
+			})
+		},
+	})
+	r.Add(testSpec{
+		Name:       "weekly/tpcc-max",
+		MinVersion: maybeMinVersionForFixturesImport(cloud),
+		Tags:       []string{`weekly`},
+		Cluster:    makeClusterSpec(4, cpu(16)),
+		Run: func(ctx context.Context, t *test, c *cluster) {
+			warehouses := 1350
+			runTPCC(ctx, t, c, tpccOptions{
+				Warehouses: warehouses,
+				Duration:   6 * 24 * time.Hour,
+			})
 		},
 	})
 
-	// Run a single tpccbench spec in CI.
+	r.Add(testSpec{
+		Name:       "tpcc/w=100/nodes=3/chaos=true",
+		Cluster:    makeClusterSpec(4),
+		MinVersion: maybeMinVersionForFixturesImport(cloud),
+		Run: func(ctx context.Context, t *test, c *cluster) {
+			duration := 30 * time.Minute
+			runTPCC(ctx, t, c, tpccOptions{
+				Warehouses: 100,
+				Duration:   duration,
+				Extra:      "--wait=false --tolerate-errors",
+				Chaos: func() Chaos {
+					return Chaos{
+						Timer: Periodic{
+							Period:   45 * time.Second,
+							DownTime: 10 * time.Second,
+						},
+						Target:       func() nodeListOption { return c.Node(1 + rand.Intn(c.nodes-1)) },
+						Stopper:      time.After(duration),
+						DrainAndQuit: false,
+					}
+				},
+				ZFS: false, // change to true during debugging/development
+			})
+		},
+	})
+
+	// Run a few representative tpccbench specs in CI.
 	registerTPCCBenchSpec(r, tpccBenchSpec{
 		Nodes: 3,
 		CPUs:  4,
-		// TODO(m-schneider): enable when geo-distributed benchmarking is supported.
-		// Chaos: true,
-		// Geo: true,
 
 		LoadWarehouses: 1000,
 		EstimatedMax:   300,
 	})
+	registerTPCCBenchSpec(r, tpccBenchSpec{
+		Nodes:        6,
+		CPUs:         16,
+		Distribution: multiZone,
+
+		LoadWarehouses: 5000,
+		EstimatedMax:   2500,
+	})
+	registerTPCCBenchSpec(r, tpccBenchSpec{
+		Nodes:      9,
+		CPUs:       4,
+		Chaos:      true,
+		LoadConfig: singlePartitionedLoadgen,
+
+		LoadWarehouses: 2000,
+		EstimatedMax:   600,
+	})
+}
+
+func maxVersion(vers ...string) string {
+	var max *version.Version
+	for _, v := range vers {
+		v, err := version.Parse(v)
+		if err != nil {
+			continue
+		}
+		if max == nil || v.AtLeast(max) {
+			max = v
+		}
+	}
+	if max == nil {
+		return ""
+	}
+	return max.String()
+}
+
+func maybeMinVersionForFixturesImport(cloud string) string {
+	const minVersionForFixturesImport = "v2.2.0"
+	if cloud == "aws" {
+		return minVersionForFixturesImport
+	}
+	return ""
 }
 
 // tpccBenchDistribution represents a distribution of nodes in a tpccbench
@@ -133,9 +287,9 @@ const (
 func (d tpccBenchDistribution) zones() []string {
 	switch d {
 	case singleZone:
-		return []string{"us-east1-b"}
+		return []string{"us-central1-b"}
 	case multiZone:
-		return []string{"us-east1-b", "us-east1-c", "us-east1-d"}
+		return []string{"us-central1-a", "us-central1-b", "us-central1-c"}
 	case multiRegion:
 		return []string{"us-east1-b", "us-west1-b", "europe-west2-b"}
 	default:
@@ -255,12 +409,12 @@ func registerTPCCBenchSpec(r *registry, b tpccBenchSpec) {
 	name := strings.Join(nameParts, "/")
 
 	numNodes := b.Nodes + b.LoadConfig.numLoadNodes(b.Distribution)
-	nodes := nodes(numNodes, opts...)
+	nodes := makeClusterSpec(numNodes, opts...)
 
 	r.Add(testSpec{
-		Name:   name,
-		Nodes:  nodes,
-		Stable: true, // DO NOT COPY to new tests
+		Name:       name,
+		Cluster:    nodes,
+		MinVersion: maybeMinVersionForFixturesImport(cloud),
 		Run: func(ctx context.Context, t *test, c *cluster) {
 			runTPCCBench(ctx, t, c, b)
 		},
@@ -271,7 +425,7 @@ func registerTPCCBenchSpec(r *registry, b tpccBenchSpec) {
 // function is idempotent and first checks whether a compatible dataset exists,
 // performing an expensive dataset restore only if it doesn't.
 func loadTPCCBench(
-	ctx context.Context, c *cluster, b tpccBenchSpec, roachNodes, loadNode nodeListOption,
+	ctx context.Context, t *test, c *cluster, b tpccBenchSpec, roachNodes, loadNode nodeListOption,
 ) error {
 	db := c.Conn(ctx, 1)
 	defer db.Close()
@@ -279,7 +433,7 @@ func loadTPCCBench(
 	// Check if the dataset already exists and is already large enough to
 	// accommodate this benchmarking. If so, we can skip the fixture RESTORE.
 	if _, err := db.ExecContext(ctx, `USE tpcc`); err == nil {
-		c.l.Printf("found existing tpcc database\n")
+		t.l.Printf("found existing tpcc database\n")
 
 		var curWarehouses int
 		if err := db.QueryRowContext(ctx,
@@ -295,7 +449,7 @@ func loadTPCCBench(
 		// If the dataset exists but is not large enough, wipe the cluster
 		// before restoring.
 		c.Wipe(ctx, roachNodes)
-		c.Start(ctx, append(b.startOpts(), roachNodes)...)
+		c.Start(ctx, t, append(b.startOpts(), roachNodes)...)
 	} else if pqErr, ok := err.(*pq.Error); !ok ||
 		string(pqErr.Code) != pgerror.CodeInvalidCatalogNameError {
 		return err
@@ -303,9 +457,9 @@ func loadTPCCBench(
 
 	// If the fixture has a corresponding store dump, use it.
 	if b.StoreDirVersion != "" {
-		c.l.Printf("ingesting existing tpcc store dump\n")
+		t.l.Printf("ingesting existing tpcc store dump\n")
 
-		urlBase, err := c.RunWithBuffer(ctx, c.l, c.Node(loadNode[0]),
+		urlBase, err := c.RunWithBuffer(ctx, t.l, c.Node(loadNode[0]),
 			fmt.Sprintf(`./workload fixtures url tpcc --warehouses=%d`, b.LoadWarehouses))
 		if err != nil {
 			return err
@@ -317,24 +471,23 @@ func loadTPCCBench(
 	}
 
 	// Load the corresponding fixture.
-	c.l.Printf("restoring tpcc fixture\n")
-	cmd := fmt.Sprintf(
-		"./workload fixtures load tpcc --checks=false --warehouses=%d {pgurl:1}", b.LoadWarehouses)
+	t.l.Printf("restoring tpcc fixture\n")
+	cmd := tpccFixturesCmd(t, cloud, b.LoadWarehouses, false /* checks */)
 	if err := c.RunE(ctx, loadNode, cmd); err != nil {
 		return err
 	}
 
 	partArgs := ""
-	rebalanceWait := time.Duration(b.LoadWarehouses/150) * time.Minute
+	rebalanceWait := time.Duration(b.LoadWarehouses/100) * time.Minute
 	switch b.LoadConfig {
 	case singleLoadgen:
-		c.l.Printf("splitting and scattering\n")
+		t.l.Printf("splitting and scattering\n")
 	case singlePartitionedLoadgen:
-		c.l.Printf("splitting, scattering, and partitioning\n")
+		t.l.Printf("splitting, scattering, and partitioning\n")
 		partArgs = fmt.Sprintf(`--partitions=%d`, b.partitions())
 		rebalanceWait = time.Duration(b.LoadWarehouses/50) * time.Minute
 	case multiLoadgen:
-		c.l.Printf("splitting, scattering, and partitioning\n")
+		t.l.Printf("splitting, scattering, and partitioning\n")
 		partArgs = fmt.Sprintf(`--partitions=%d --zones="%s" --partition-affinity=0`,
 			b.partitions(), strings.Join(b.Distribution.zones(), ","))
 		rebalanceWait = time.Duration(b.LoadWarehouses/20) * time.Minute
@@ -342,27 +495,20 @@ func loadTPCCBench(
 		panic("unexpected")
 	}
 
-	// Split and scatter the tables. Set duration to 1ms so that the load
-	// generation doesn't actually run.
-	cmd = fmt.Sprintf("ulimit -n 32768; "+
-		"./workload run tpcc --warehouses=%d --split --scatter %s "+
-		"--duration=3m --tolerate-errors {pgurl:1}", b.LoadWarehouses, partArgs)
-	if out, err := c.RunWithBuffer(ctx, c.l, loadNode, cmd); err != nil {
-		return errors.Wrapf(err, "failed with output %q", string(out))
-	}
-
-	c.l.Printf("waiting %v for rebalancing\n", rebalanceWait)
+	t.l.Printf("waiting %v for rebalancing\n", rebalanceWait)
 	_, err := db.ExecContext(ctx, `SET CLUSTER SETTING kv.snapshot_rebalance.max_rate='64MiB'`)
 	if err != nil {
 		return err
 	}
 
-	// TODO(nvanbenschoten): we should find a way to make this reactive and
-	// wait until no zone constraints are being violated.
-	select {
-	case <-time.After(rebalanceWait):
-	case <-ctx.Done():
-		return ctx.Err()
+	// Split and scatter the tables. Ramp up to the expected load in the desired
+	// distribution. This should allow for load-based rebalancing to help
+	// distribute load. Optionally pass some load configuration-specific flags.
+	cmd = fmt.Sprintf("./workload run tpcc --warehouses=%d --workers=%d --split --scatter "+
+		"--wait=false --duration=%s --tolerate-errors %s {pgurl%s}",
+		b.LoadWarehouses, b.LoadWarehouses, rebalanceWait, partArgs, roachNodes)
+	if out, err := c.RunWithBuffer(ctx, c.l, loadNode, cmd); err != nil {
+		return errors.Wrapf(err, "failed with output %q", string(out))
 	}
 
 	_, err = db.ExecContext(ctx, `SET CLUSTER SETTING kv.snapshot_rebalance.max_rate='2MiB'`)
@@ -390,31 +536,39 @@ func loadTPCCBench(
 func runTPCCBench(ctx context.Context, t *test, c *cluster, b tpccBenchSpec) {
 	// Determine the nodes in each load group. A load group consists of a set of
 	// Cockroach nodes and a single load generator.
+	numRoachNodes := b.Nodes
 	numLoadGroups := b.LoadConfig.numLoadNodes(b.Distribution)
-	nodesPerGroup := c.nodes / numLoadGroups
-	loadGroup := make([]struct{ roachNodes, loadNodes nodeListOption }, numLoadGroups)
-	for i, j := 1, 0; i <= c.nodes; i += nodesPerGroup {
-		loadGroup[j].roachNodes = c.Range(i, i+nodesPerGroup-2)
-		loadGroup[j].loadNodes = c.Node(i + nodesPerGroup - 1)
-		j++
+	numZones := len(b.Distribution.zones())
+	roachNodesPerGroup := numRoachNodes / numLoadGroups
+	nodesPerGroup := roachNodesPerGroup + 1 // nodesPerGroup * numLoadGroups == c.nodes
+	zonesPerGroup := numZones / numLoadGroups
+	// Roachprod round-robins the allocation of nodes across zones. When running
+	// a single load group across zones, the first zone will always have one more
+	// node than the others. This extra node is the load node.
+	loadNodePerGroup := roachNodesPerGroup / zonesPerGroup
+	loadGroups := make([]struct{ roachNodes, loadNodes nodeListOption }, numLoadGroups)
+	for i := range loadGroups {
+		for j := 0; j < nodesPerGroup; j++ {
+			n := c.Node(i*nodesPerGroup + j + 1)
+			if j == loadNodePerGroup {
+				loadGroups[i].loadNodes = loadGroups[i].loadNodes.merge(n)
+			} else {
+				loadGroups[i].roachNodes = loadGroups[i].roachNodes.merge(n)
+			}
+		}
 	}
 
 	// Aggregate nodes across load groups.
 	var roachNodes nodeListOption
 	var loadNodes nodeListOption
-	for _, g := range loadGroup {
+	for _, g := range loadGroups {
 		roachNodes = roachNodes.merge(g.roachNodes)
 		loadNodes = loadNodes.merge(g.loadNodes)
 	}
 
-	// Disable write barrier on mounted SSDs.
-	if !c.isLocal() {
-		c.RemountNoBarrier(ctx)
-	}
-
 	c.Put(ctx, cockroach, "./cockroach", roachNodes)
 	c.Put(ctx, workload, "./workload", loadNodes)
-	c.Start(ctx, append(b.startOpts(), roachNodes)...)
+	c.Start(ctx, t, append(b.startOpts(), roachNodes)...)
 
 	useHAProxy := b.Chaos
 	if useHAProxy {
@@ -423,15 +577,15 @@ func runTPCCBench(ctx context.Context, t *test, c *cluster, b tpccBenchSpec) {
 		}
 		t.Status("installing haproxy")
 		c.Install(ctx, loadNodes, "haproxy")
-		c.Run(ctx, loadNodes, fmt.Sprintf("./cockroach gen haproxy --insecure --host %s",
-			c.InternalIP(ctx, c.Node(1))[0]))
+		c.Put(ctx, cockroach, "./cockroach", loadNodes)
+		c.Run(ctx, loadNodes, "./cockroach gen haproxy --insecure --url {pgurl:1}")
 		c.Run(ctx, loadNodes, "haproxy -f haproxy.cfg -D")
 	}
 
 	m := newMonitor(ctx, c, roachNodes)
 	m.Go(func(ctx context.Context) error {
 		t.Status("setting up dataset")
-		err := loadTPCCBench(ctx, c, b, roachNodes, c.Node(loadNodes[0]))
+		err := loadTPCCBench(ctx, t, c, b, roachNodes, c.Node(loadNodes[0]))
 		if err != nil {
 			return err
 		}
@@ -441,13 +595,21 @@ func runTPCCBench(ctx context.Context, t *test, c *cluster, b tpccBenchSpec) {
 		// threshold, set to a fraction of max tpmC.
 		precision := int(math.Max(1.0, float64(b.LoadWarehouses/200)))
 		initStepSize := precision
+
+		// Create a temp directory to store the local copy of results from the
+		// workloads.
+		resultsDir, err := ioutil.TempDir("", "roachtest-tpcc")
+		if err != nil {
+			return errors.Wrap(err, "failed to create temp dir")
+		}
+		defer func() { _ = os.RemoveAll(resultsDir) }()
 		s := search.NewLineSearcher(1, b.LoadWarehouses, b.EstimatedMax, initStepSize, precision)
 		res, err := s.Search(func(warehouses int) (bool, error) {
 			// Restart the cluster before each iteration to help eliminate
 			// inter-trial interactions.
 			m.ExpectDeaths(int32(len(roachNodes)))
 			c.Stop(ctx, roachNodes)
-			c.Start(ctx, append(b.startOpts(), roachNodes)...)
+			c.Start(ctx, t, append(b.startOpts(), roachNodes)...)
 			time.Sleep(10 * time.Second)
 
 			// Set up the load generation configuration.
@@ -462,10 +624,9 @@ func runTPCCBench(ctx context.Context, t *test, c *cluster, b tpccBenchSpec) {
 
 				// Kill one node at a time.
 				ch := Chaos{
-					Timer:        Periodic{Period: 90 * time.Second, DownTime: 1 * time.Second},
-					Target:       roachNodes.randNode,
-					Stopper:      loadDone,
-					DrainAndQuit: true,
+					Timer:   Periodic{Period: 90 * time.Second, DownTime: 5 * time.Second},
+					Target:  roachNodes.randNode,
+					Stopper: loadDone,
 				}
 				m.Go(ch.Runner(c, m))
 			}
@@ -475,10 +636,12 @@ func runTPCCBench(ctx context.Context, t *test, c *cluster, b tpccBenchSpec) {
 			}
 
 			// If we're running multiple load generators, run them in parallel and then
-			// aggregate tpmCChan.
+			// aggregate resultChan. In order to process the results we need to copy
+			// over the histograms. Create a temp dir which will contain the fetched
+			// data.
 			var eg errgroup.Group
-			tpmCChan := make(chan float64, len(loadGroup))
-			for groupIdx, group := range loadGroup {
+			resultChan := make(chan *tpcc.Result, numLoadGroups)
+			for groupIdx, group := range loadGroups {
 				// Copy for goroutine
 				groupIdx := groupIdx
 				group := group
@@ -504,90 +667,57 @@ func runTPCCBench(ctx context.Context, t *test, c *cluster, b tpccBenchSpec) {
 					}
 
 					t.Status(fmt.Sprintf("running benchmark, warehouses=%d", warehouses))
-
-					cmd := fmt.Sprintf("ulimit -n 32768; "+
-						"./workload run tpcc --warehouses=%d --active-warehouses=%d "+
-						"--tolerate-errors --ramp=%s --duration=%s%s {pgurl%s}",
+					histogramsPath := fmt.Sprintf("logs/warehouses=%d/stats.json", activeWarehouses)
+					cmd := fmt.Sprintf("./workload run tpcc --warehouses=%d --active-warehouses=%d "+
+						"--tolerate-errors --ramp=%s --duration=%s%s {pgurl%s} "+
+						"--histograms=%s",
 						b.LoadWarehouses, activeWarehouses, rampDur,
-						loadDur, extraFlags, sqlGateways)
-					out, err := c.RunWithBuffer(ctx, c.l, group.loadNodes, cmd)
+						loadDur, extraFlags, sqlGateways, histogramsPath)
+					err := c.RunE(ctx, group.loadNodes, cmd)
 					loadDone <- timeutil.Now()
 					if err != nil {
-						return errors.Wrapf(err, "error running tpcc load generator:\n\n%s\n", out)
+						return errors.Wrapf(err, "error running tpcc load generator")
 					}
-
-					// Parse the stats header and stats lines from the output.
-					str := string(out)
-					lines := strings.Split(str, "\n")
-					for i, line := range lines {
-						if strings.Contains(line, "tpmC") {
-							lines = lines[i:]
-						}
-						if i == len(lines)-1 {
-							return errors.Errorf("tpmC not found in output:\n\n%s\n", out)
-						}
-					}
-					headerLine, statsLine := lines[0], lines[1]
-					c.l.Printf("%s\n%s\n", headerLine, statsLine)
-
-					// Parse tpmC value from stats line.
-					fields := strings.Fields(statsLine)
-					tpmC, err := strconv.ParseFloat(fields[1], 64)
+					roachtestHistogramsPath := filepath.Join(resultsDir, fmt.Sprintf("%d.%d-stats.json", warehouses, groupIdx))
+					c.Get(ctx, histogramsPath, roachtestHistogramsPath, group.loadNodes)
+					snapshots, err := histogram.DecodeSnapshots(roachtestHistogramsPath)
 					if err != nil {
-						return err
+						return errors.Wrapf(err, "failed to decode histogram snapshots")
 					}
-					tpmCChan <- tpmC
+					result := tpcc.NewResultWithSnapshots(activeWarehouses, 0, snapshots)
+					resultChan <- result
 					return nil
 				})
 			}
 			if err = eg.Wait(); err != nil {
 				return false, err
 			}
-			close(tpmCChan)
-			var tpmC float64
-			for partialTpMc := range tpmCChan {
-				tpmC += partialTpMc
+			close(resultChan)
+			var results []*tpcc.Result
+			for partial := range resultChan {
+				results = append(results, partial)
 			}
-			// Determine the fraction of the maximum possible tpmC realized.
-			//
-			// The 12.605 is computed from the operation mix and the number of secs
-			// it takes to cycle through a deck:
-			//
-			//   10*(18+12) + 10*(3+12) + 1*(2+10) + 1*(2+5) + 1*(2+5) = 476
-			//
-			// 10 workers per warehouse times 10 newOrder ops per deck results in:
-			//
-			//   (10*10)/(476/60) = 12.605...
-			maxTpmC := 12.605 * float64(warehouses)
-			tpmCRatio := tpmC / maxTpmC
-
-			// Determine whether this means the test passed or not. We use a
-			// threshold of 85% of max tpmC as the "passing" criteria for a
-			// given number of warehouses. This does not take response latencies
-			// for different op types into account directly as required by the
-			// formal TPC-C spec, but in practice it results in stable results.
-			passRatio := 0.85
-			pass := tpmCRatio > passRatio
-
+			res := tpcc.MergeResults(results...)
+			failErr := res.FailureError()
 			// Print the result.
-			ttycolor.Stdout(ttycolor.Green)
-			passStr := "PASS"
-			if !pass {
+			if failErr == nil {
+				ttycolor.Stdout(ttycolor.Green)
+				t.l.Printf("--- PASS: tpcc %d resulted in %.1f tpmC (%.1f%% of max tpmC)\n\n",
+					warehouses, res.TpmC(), res.Efficiency())
+			} else {
 				ttycolor.Stdout(ttycolor.Red)
-				passStr = "FAIL"
+				t.l.Printf("--- FAIL: tpcc %d resulted in %.1f tpmC and failed due to %v",
+					warehouses, res.TpmC(), failErr)
 			}
-			c.l.Printf("--- %s: tpcc %d resulted in %.1f tpmC (%.1f%% of max tpmC)\n\n",
-				passStr, warehouses, tpmC, tpmCRatio*100)
 			ttycolor.Stdout(ttycolor.Reset)
-
-			return pass, nil
+			return failErr == nil, nil
 		})
 		if err != nil {
 			return err
 		}
 
 		ttycolor.Stdout(ttycolor.Green)
-		c.l.Printf("------\nMAX WAREHOUSES = %d\n------\n\n", res)
+		t.l.Printf("------\nMAX WAREHOUSES = %d\n------\n\n", res)
 		ttycolor.Stdout(ttycolor.Reset)
 		return nil
 	})
@@ -637,7 +767,7 @@ func registerTPCCBench(r *registry) {
 			Chaos: true,
 
 			LoadWarehouses: 5000,
-			EstimatedMax:   500,
+			EstimatedMax:   2000,
 			// TODO(nvanbenschoten): Need to regenerate.
 			// StoreDirVersion: "2.0-5",
 		},
@@ -668,6 +798,41 @@ func registerTPCCBench(r *registry) {
 			LoadWarehouses: 50000,
 			EstimatedMax:   40000,
 		},
+
+		// See https://github.com/cockroachdb/cockroach/issues/31409 for the next three specs.
+		{
+			Nodes: 6,
+			CPUs:  16,
+
+			LoadWarehouses: 5000,
+			EstimatedMax:   3000,
+			LoadConfig:     singlePartitionedLoadgen,
+		},
+		{
+			Nodes: 12,
+			CPUs:  16,
+
+			LoadWarehouses: 10000,
+			EstimatedMax:   6000,
+			LoadConfig:     singlePartitionedLoadgen,
+		},
+		{
+			Nodes: 24,
+			CPUs:  16,
+
+			LoadWarehouses: 20000,
+			EstimatedMax:   12000,
+			LoadConfig:     singlePartitionedLoadgen,
+		},
+
+		// Requested by @awoods87.
+		{
+			Nodes: 11,
+			CPUs:  32,
+
+			LoadWarehouses: 10000,
+			EstimatedMax:   8000,
+		},
 	}
 
 	for _, b := range specs {
