@@ -22,12 +22,12 @@ import (
 	"runtime"
 	"time"
 
-	_ "github.com/lib/pq"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/util/binfetcher"
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
+	_ "github.com/lib/pq"
+	"github.com/pkg/errors"
 )
 
 func registerUpgrade(r *registry) {
@@ -37,7 +37,7 @@ func registerUpgrade(r *registry) {
 
 		b, err := binfetcher.Download(ctx, binfetcher.Options{
 			Binary:  "cockroach",
-			Version: oldVersion,
+			Version: "v" + oldVersion,
 			GOOS:    goos,
 			GOARCH:  "amd64",
 		})
@@ -46,9 +46,10 @@ func registerUpgrade(r *registry) {
 		}
 
 		c.Put(ctx, b, "./cockroach", c.Range(1, nodes))
-		// Force disable encryption.
-		// TODO(mberhault): allow it once oldVersion >= 2.1.
-		c.Start(ctx, c.Range(1, nodes), startArgsDontEncrypt)
+
+		// NB: remove startArgsDontEncrypt across this file once we're not running
+		// roachtest against v2.1 any more (which would start a v2.0 cluster here).
+		c.Start(ctx, t, c.Range(1, nodes), startArgsDontEncrypt)
 
 		const stageDuration = 30 * time.Second
 		const timeUntilStoreDead = 90 * time.Second
@@ -66,6 +67,11 @@ func registerUpgrade(r *registry) {
 
 		db := c.Conn(ctx, 1)
 		defer db.Close()
+		// Without this line, the test reliably fails (at least on OSX), presumably
+		// because a connection to a node that gets restarted somehow sticks around
+		// in the pool and throws an error at the next client using it (instead of
+		// transparently reconnecting).
+		db.SetMaxIdleConns(0)
 
 		if _, err := db.ExecContext(ctx,
 			"SET CLUSTER SETTING server.time_until_store_dead = $1", timeUntilStoreDead.String(),
@@ -79,9 +85,9 @@ func registerUpgrade(r *registry) {
 
 		stop := func(node int) error {
 			port := fmt.Sprintf("{pgport:%d}", node)
-			// Note that the following command line needs to run against both v2.0
+			// Note that the following command line needs to run against both v2.1
 			// and the current branch. Do not change it in a manner that is
-			// incompatible with 2.0.
+			// incompatible with 2.1.
 			if err := c.RunE(ctx, c.Node(node), "./cockroach quit --insecure --port="+port); err != nil {
 				return err
 			}
@@ -91,9 +97,9 @@ func registerUpgrade(r *registry) {
 
 		decommissionAndStop := func(node int) error {
 			port := fmt.Sprintf("{pgport:%d}", node)
-			// Note that the following command line needs to run against both v2.0
+			// Note that the following command line needs to run against both v2.1
 			// and the current branch. Do not change it in a manner that is
-			// incompatible with 2.0.
+			// incompatible with 2.1.
 			if err := c.RunE(ctx, c.Node(node), "./cockroach quit --decommission --insecure --port="+port); err != nil {
 				return err
 			}
@@ -104,7 +110,7 @@ func registerUpgrade(r *registry) {
 		clusterVersion := func() (string, error) {
 			var version string
 			if err := db.QueryRowContext(ctx, `SHOW CLUSTER SETTING version`).Scan(&version); err != nil {
-				return "", err
+				return "", errors.Wrap(err, "determining cluster version")
 			}
 			return version, nil
 		}
@@ -126,7 +132,7 @@ func registerUpgrade(r *registry) {
 			if _, err := db.ExecContext(ctx,
 				"SET CLUSTER SETTING cluster.preserve_downgrade_option = $1;", version,
 			); err == nil {
-				return fmt.Errorf("cluster.preserve_downgrade_option shouldn't be set to any other values besides current cluster version")
+				return fmt.Errorf("cluster.preserve_downgrade_option shouldn't be set to any other values besides current cluster version; was able to set it to %s", version)
 			} else if !testutils.IsError(err, "cannot set cluster.preserve_downgrade_option") {
 				return err
 			}
@@ -140,7 +146,7 @@ func registerUpgrade(r *registry) {
 				t.Fatal(err)
 			}
 			c.Put(ctx, cockroach, "./cockroach", c.Node(i))
-			c.Start(ctx, c.Node(i), startArgsDontEncrypt)
+			c.Start(ctx, t, c.Node(i), startArgsDontEncrypt)
 			if err := sleep(stageDuration); err != nil {
 				t.Fatal(err)
 			}
@@ -162,7 +168,7 @@ func registerUpgrade(r *registry) {
 			t.Fatal(err)
 		}
 		c.Put(ctx, cockroach, "./cockroach", c.Node(nodes))
-		c.Start(ctx, c.Node(nodes), startArgsDontEncrypt)
+		c.Start(ctx, t, c.Node(nodes), startArgsDontEncrypt)
 		if err := sleep(stageDuration); err != nil {
 			t.Fatal(err)
 		}
@@ -184,10 +190,10 @@ func registerUpgrade(r *registry) {
 
 		// Check cannot set cluster setting cluster.preserve_downgrade_option to any
 		// value besides the old cluster version.
-		if err := checkDowngradeOption("1.1"); err != nil {
+		if err := checkDowngradeOption("1.9"); err != nil {
 			t.Fatal(err)
 		}
-		if err := checkDowngradeOption("2.2"); err != nil {
+		if err := checkDowngradeOption("99.9"); err != nil {
 			t.Fatal(err)
 		}
 
@@ -203,7 +209,7 @@ func registerUpgrade(r *registry) {
 		}
 
 		// Restart the previously stopped node.
-		c.Start(ctx, c.Node(nodes-1), startArgsDontEncrypt)
+		c.Start(ctx, t, c.Node(nodes-1), startArgsDontEncrypt)
 		if err := sleep(stageDuration); err != nil {
 			t.Fatal(err)
 		}
@@ -256,27 +262,37 @@ func registerUpgrade(r *registry) {
 		}
 	}
 
-	const oldVersion = "v2.0.5"
+	mixedWithVersion := r.PredecessorVersion()
+	var skip string
+	if mixedWithVersion == "" {
+		skip = "unable to determine predecessor version"
+	}
+
 	for _, n := range []int{5} {
 		r.Add(testSpec{
-			Name:       fmt.Sprintf("upgrade/oldVersion=%s/nodes=%d", oldVersion, n),
+			Name:       fmt.Sprintf("upgrade/mixedWith=%s/nodes=%d", mixedWithVersion, n),
 			MinVersion: "v2.1.0",
-			Nodes:      nodes(n),
+			Cluster:    makeClusterSpec(n),
 			Run: func(ctx context.Context, t *test, c *cluster) {
-				runUpgrade(ctx, t, c, oldVersion)
+				runUpgrade(ctx, t, c, mixedWithVersion)
 			},
+			Skip: skip,
 		})
 	}
 }
 
 func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
+	// This is ugly, but we can't pass `--encrypt=false` to old versions of
+	// Cockroach.
+	c.encryptDefault = false
+
 	nodes := c.Range(1, 3)
 	goos := ifLocal(runtime.GOOS, "linux")
 	const headVersion = "HEAD"
 
 	// versionStep is an isolated version migration on a running cluster.
 	type versionStep struct {
-		clusterVersion string
+		clusterVersion string // if empty, use crdb_internal.node_executable_version
 		run            func()
 	}
 
@@ -329,7 +345,7 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 	binaryVersionUpgrade := func(newVersion string, nodes nodeListOption) versionStep {
 		return versionStep{
 			run: func() {
-				c.l.Printf("%s: binary\n", newVersion)
+				t.l.Printf("%s: binary\n", newVersion)
 				args := uploadVersion(newVersion)
 
 				// Restart nodes in a random order; otherwise node 1 would be running all
@@ -338,9 +354,9 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 					nodes[i], nodes[j] = nodes[j], nodes[i]
 				})
 				for _, node := range nodes {
-					c.l.Printf("%s: upgrading node %d\n", newVersion, node)
+					t.l.Printf("%s: upgrading node %d\n", newVersion, node)
 					c.Stop(ctx, c.Node(node))
-					c.Start(ctx, c.Node(node), args)
+					c.Start(ctx, t, c.Node(node), args, startArgsDontEncrypt)
 
 					checkNode(node, newVersion)
 
@@ -355,12 +371,26 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 
 	// clusterVersionUpgrade performs a cluster version upgrade to its version.
 	// It waits until all nodes have seen the upgraded cluster version.
+	// If manual is set, we'll performe a SET CLUSTER SETTING version =
+	// <newVersion>. If it's not, we'll rely on the automatic cluster version
+	// upgrade mechanism (which is not inhibited by the
+	// cluster.preserve_downgrade_option cluster setting in this test.
 	var currentVersion string
 	clusterVersionUpgrade := func(newVersion string, manual bool) versionStep {
 		return versionStep{
 			clusterVersion: newVersion,
 			run: func() {
-				c.l.Printf("%s: cluster\n", newVersion)
+				func() {
+					if newVersion != "" {
+						return
+					}
+					db1 := c.Conn(ctx, 1)
+					defer db1.Close()
+					if err := db1.QueryRow(`SELECT crdb_internal.node_executable_version()`).Scan(&newVersion); err != nil {
+						t.Fatal(err)
+					}
+				}()
+				t.l.Printf("%s: cluster\n", newVersion)
 
 				// hasShowSettingBug is true when we're working around
 				// https://github.com/cockroachdb/cockroach/issues/22796.
@@ -384,7 +414,7 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 						db := c.Conn(ctx, node)
 						defer db.Close()
 
-						c.l.Printf("%s: upgrading cluster version (node %d)\n", newVersion, node)
+						t.l.Printf("%s: upgrading cluster version (node %d)\n", newVersion, node)
 						if _, err := db.Exec(fmt.Sprintf(`SET CLUSTER SETTING version = '%s'`, newVersion)); err != nil {
 							t.Fatal(err)
 						}
@@ -392,7 +422,7 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 				}
 
 				if hasShowSettingBug {
-					c.l.Printf("%s: using workaround for upgrade\n", newVersion)
+					t.l.Printf("%s: using workaround for upgrade\n", newVersion)
 				}
 
 				for i := 1; i < c.nodes; i++ {
@@ -423,7 +453,7 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 					}
 				}
 
-				c.l.Printf("%s: cluster is upgraded\n", newVersion)
+				t.l.Printf("%s: cluster is upgraded\n", newVersion)
 
 				// TODO(nvanbenschoten): add upgrade qualification step.
 				time.Sleep(1 * time.Second)
@@ -433,20 +463,18 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 
 	const baseVersion = "v1.0.6"
 	steps := []versionStep{
-		// v1.1.0 is the first binary version that knows about cluster versions,
-		// but thinks it can only support up to 1.0-3.
-		binaryVersionUpgrade("v1.1.0", nodes),
-		clusterVersionUpgrade("1.0", true /* manual */),
-		clusterVersionUpgrade("1.0-3", true /* manual */),
-
-		binaryVersionUpgrade("v1.1.1", nodes),
+		// v1.1.0 is the first binary version that knows about cluster versions.
+		binaryVersionUpgrade("v1.1.9", nodes),
 		clusterVersionUpgrade("1.1", true /* manual */),
 
-		binaryVersionUpgrade("v2.0.0", nodes),
-		clusterVersionUpgrade("1.1-6", true /* manual */),
+		binaryVersionUpgrade("v2.0.7", nodes),
 		clusterVersionUpgrade("2.0", true /* manual */),
 
+		binaryVersionUpgrade("v2.1.2", nodes),
+		clusterVersionUpgrade("2.1", true /* manual */),
+
 		binaryVersionUpgrade("HEAD", nodes),
+		clusterVersionUpgrade("19.1", false /* manual */),
 	}
 
 	type feature struct {
@@ -501,12 +529,12 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 			if err == nil {
 				t.Fatalf("expected %s to fail on cluster version %s", f.name, cv)
 			}
-			c.l.Printf("%s: %s fails expected\n", cv, f.name)
+			t.l.Printf("%s: %s fails expected\n", cv, f.name)
 		} else {
 			if err != nil {
 				t.Fatalf("expected %s to succeed on cluster version %s, got %s", f.name, cv, err)
 			}
-			c.l.Printf("%s: %s works as expected\n", cv, f.name)
+			t.l.Printf("%s: %s works as expected\n", cv, f.name)
 		}
 	}
 
@@ -514,7 +542,7 @@ func runVersionUpgrade(ctx context.Context, t *test, c *cluster) {
 	// Hack to skip initializing settings which doesn't work on very old versions
 	// of cockroach.
 	c.Run(ctx, c.Node(1), "mkdir -p {store-dir} && touch {store-dir}/settings-initialized")
-	c.Start(ctx, nodes, args)
+	c.Start(ctx, t, nodes, args, startArgsDontEncrypt)
 
 	func() {
 		// Create a bunch of tables, over the batch size on which some migrations
