@@ -21,9 +21,12 @@ import (
 
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/opt/cat"
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
+	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
+	"github.com/gogo/protobuf/proto"
 	"github.com/pkg/errors"
 )
 
@@ -76,10 +79,9 @@ func ZoneSpecifierFromID(
 	if err != nil {
 		return tree.ZoneSpecifier{}, err
 	}
-	tn := tree.NewTableName(tree.Name(db), tree.Name(name))
 	return tree.ZoneSpecifier{
-		TableOrIndex: tree.TableNameWithIndex{
-			Table: tree.NormalizableTableName{TableNameReference: tn},
+		TableOrIndex: tree.TableIndexName{
+			Table: tree.MakeTableName(tree.Name(db), tree.Name(name)),
 		},
 	}, nil
 }
@@ -97,50 +99,59 @@ func ParseCLIZoneSpecifier(s string) (tree.ZoneSpecifier, error) {
 		}
 		return tree.ZoneSpecifier{NamedZone: tree.UnrestrictedName(name)}, nil
 	}
-	// ParseTableNameWithIndex is not vulnerable to SQL injection, so passing s
+	// ParseTableIndexName is not vulnerable to SQL injection, so passing s
 	// directly is safe. See #8389 for details.
-	parsed, err := parser.ParseTableNameWithIndex(s)
+	parsed, err := parser.ParseTableIndexName(s)
 	if err != nil {
 		return tree.ZoneSpecifier{}, fmt.Errorf("malformed name: %q", s)
 	}
-	parsed.SearchTable = false
-	var partition tree.Name
-	if un := parsed.Table.TableNameReference.(*tree.UnresolvedName); un.NumParts == 1 {
-		// Unlike in SQL, where a name with one part indicates a table in the
-		// current database, if a CLI specifier has just one name part, it indicates
-		// a database.
-		return tree.ZoneSpecifier{Database: tree.Name(un.Parts[0])}, nil
-	} else if un.NumParts == 3 {
-		// If a CLI specifier has three name parts, the last name is a partition.
-		// Pop it off so TableNameReference.Normalize sees only the table name
-		// below.
-		partition = tree.Name(un.Parts[0])
-		un.Parts[0], un.Parts[1], un.Parts[2] = un.Parts[1], un.Parts[2], un.Parts[3]
-		un.NumParts--
+
+	tn := parsed.Table
+	if tn.Table() == "" {
+		// In this case, the input was parsed as [[CATALOG.]SCHEMA.]INDEX
+		// (this is the syntax when specifying an index, e.g. for ALTER INDEX).
+		tn.TableName = tree.Name(parsed.Index)
+		parsed.Index = ""
 	}
-	// We've handled the special cases for named zones, databases and partitions;
-	// have TableNameReference.Normalize tell us whether what remains is a valid
-	// table or index name.
-	tn, err := parsed.Table.Normalize()
-	if err != nil {
-		return tree.ZoneSpecifier{}, err
+
+	// To reuse the SQL parsing code, we unfortunately have to abuse TableIndexName
+	// here. A table index name is of the form:
+	//   [[CATALOG.]SCHEMA.]TABLE[@INDEX]
+	// and we want to reinterpret this as
+	//   DATABASE[.TABLE[.PARTITION|@INDEX]]
+	//
+	// To make this conversion we have three cases:
+	//   1) one part:    TABLE -> DATABASE
+	//   2) two parts:   SCHEMA.TABLE -> DATABASE.TABLE
+	//   3) three parts: CATALOG.SCHEMA.TABLE -> DATABASE.TABLE.PARTITION
+
+	if !tn.ExplicitSchema {
+		// Case 1: TABLE -> DATABASE.
+		return tree.ZoneSpecifier{Database: tn.TableName}, nil
 	}
-	if parsed.Index != "" && partition != "" {
+
+	if !tn.ExplicitCatalog {
+		// Case 2: SCHEMA.TABLE -> DATABASE.TABLE
+		database, table := tn.SchemaName, tn.TableName
+		return tree.ZoneSpecifier{
+			TableOrIndex: tree.TableIndexName{
+				Table: tree.MakeTableName(database, table),
+				Index: parsed.Index,
+			},
+		}, nil
+	}
+
+	// Case 3: CATALOG.SCHEMA.TABLE -> DATABASE.TABLE.PARTITION
+	if parsed.Index != "" {
 		return tree.ZoneSpecifier{}, fmt.Errorf(
 			"index and partition cannot be specified simultaneously: %q", s)
 	}
-	// Table prefixes in CLI zone specifiers always have the form
-	// "table" or "db.table" and do not know about schemas. They never
-	// refer to virtual schemas. So migrate the db part to the catalog
-	// position.
-	if tn.ExplicitSchema {
-		tn.ExplicitCatalog = true
-		tn.CatalogName = tn.SchemaName
-		tn.SchemaName = tree.PublicSchemaName
-	}
+	database, table, partition := tn.CatalogName, tn.SchemaName, tn.TableName
 	return tree.ZoneSpecifier{
-		TableOrIndex: parsed,
-		Partition:    partition,
+		TableOrIndex: tree.TableIndexName{
+			Table: tree.MakeTableName(database, table),
+		},
+		Partition: partition,
 	}, nil
 }
 
@@ -157,15 +168,22 @@ func CLIZoneSpecifier(zs *tree.ZoneSpecifier) string {
 
 	// The table name may have a schema specifier. CLI zone specifiers
 	// do not support this, so strip it.
-	tn := ti.Table.TableName()
-	if zs.Partition == "" {
-		ti.Table.TableNameReference = tree.NewUnresolvedName(tn.Catalog(), tn.Table())
-	} else {
-		// The index is redundant when the partition is specified, so omit it.
-		ti.Table.TableNameReference = tree.NewUnresolvedName(tn.Catalog(), tn.Table(), string(zs.Partition))
-		ti.Index = ""
+	ctx := tree.NewFmtCtx(tree.FmtSimple)
+	catalog := tree.Name(ti.Table.Catalog())
+	ctx.FormatNode(&catalog)
+	ctx.WriteByte('.')
+	table := tree.UnrestrictedName(ti.Table.Table())
+	ctx.FormatNode(&table)
+
+	if zs.Partition != "" {
+		ctx.WriteByte('.')
+		partition := tree.UnrestrictedName(zs.Partition)
+		ctx.FormatNode(&partition)
+	} else if ti.Index != "" {
+		ctx.WriteByte('@')
+		ctx.FormatNode(&ti.Index)
 	}
-	return tree.AsStringWithFlags(&ti, tree.FmtAlwaysQualifyTableNames)
+	return ctx.CloseAndGetString()
 }
 
 // ResolveZoneSpecifier converts a zone specifier to the ID of most specific
@@ -193,10 +211,7 @@ func ResolveZoneSpecifier(
 
 	// Third case: a table or index name. We look up the table part here.
 
-	tn, err := zs.TableOrIndex.Table.Normalize()
-	if err != nil {
-		return 0, err
-	}
+	tn := &zs.TableOrIndex.Table
 	if tn.SchemaName != tree.PublicSchemaName {
 		return 0, pgerror.NewErrorf(pgerror.CodeReservedNameError,
 			"only schema \"public\" is supported: %q", tree.ErrString(tn))
@@ -255,11 +270,11 @@ const minRangeMaxBytes = 64 << 10 // 64 KB
 
 // defaultZoneConfig is the default zone configuration used when no custom
 // config has been specified.
-var defaultZoneConfig = ZoneConfig{
-	NumReplicas:   3,
-	RangeMinBytes: 1 << 20,  // 1 MB
-	RangeMaxBytes: 64 << 20, // 64 MB
-	GC: GCPolicy{
+var defaultZoneConfig = &ZoneConfig{
+	NumReplicas:   proto.Int32(3),
+	RangeMinBytes: proto.Int64(16 << 20), // 16 MB
+	RangeMaxBytes: proto.Int64(64 << 20), // 64 MB
+	GC: &GCPolicy{
 		// Use 25 hours instead of the previous 24 to make users successful by
 		// default. Users desiring to take incremental backups every 24h may
 		// incorrectly assume that the previous default 24h was sufficient to do
@@ -274,11 +289,11 @@ var defaultZoneConfig = ZoneConfig{
 
 // defaultSystemZoneConfig is the default zone configuration used when no custom
 // config has been specified for system ranges.
-var defaultSystemZoneConfig = ZoneConfig{
-	NumReplicas:   5,
-	RangeMinBytes: 1 << 20,  // 1 MB
-	RangeMaxBytes: 64 << 20, // 64 MB
-	GC: GCPolicy{
+var defaultSystemZoneConfig = &ZoneConfig{
+	NumReplicas:   proto.Int32(5),
+	RangeMinBytes: proto.Int64(16 << 20), // 16 MB
+	RangeMaxBytes: proto.Int64(64 << 20), // 64 MB
+	GC: &GCPolicy{
 		// Use 25 hours instead of the previous 24 to make users successful by
 		// default. Users desiring to take incremental backups every 24h may
 		// incorrectly assume that the previous default 24h was sufficient to do
@@ -291,9 +306,38 @@ var defaultSystemZoneConfig = ZoneConfig{
 	},
 }
 
+// NewZoneConfig is the zone configuration used when no custom
+// config has been specified.
+func NewZoneConfig() *ZoneConfig {
+	return &ZoneConfig{
+		InheritedConstraints:      true,
+		InheritedLeasePreferences: true,
+	}
+}
+
+// EmptyCompleteZoneConfig is the zone configuration where
+// all fields are set but set to their respective zero values.
+func EmptyCompleteZoneConfig() *ZoneConfig {
+	return &ZoneConfig{
+		NumReplicas:               proto.Int32(0),
+		RangeMinBytes:             proto.Int64(0),
+		RangeMaxBytes:             proto.Int64(0),
+		GC:                        &GCPolicy{TTLSeconds: 0},
+		InheritedConstraints:      true,
+		InheritedLeasePreferences: true,
+	}
+}
+
 // DefaultZoneConfig is the default zone configuration used when no custom
 // config has been specified.
 func DefaultZoneConfig() ZoneConfig {
+	testingLock.Lock()
+	defer testingLock.Unlock()
+	return *protoutil.Clone(defaultZoneConfig).(*ZoneConfig)
+}
+
+// DefaultZoneConfigRef returns a reference to the default zone config.
+func DefaultZoneConfigRef() *ZoneConfig {
 	testingLock.Lock()
 	defer testingLock.Unlock()
 	return defaultZoneConfig
@@ -304,7 +348,7 @@ func DefaultZoneConfig() ZoneConfig {
 func DefaultSystemZoneConfig() ZoneConfig {
 	testingLock.Lock()
 	defer testingLock.Unlock()
-	return defaultSystemZoneConfig
+	return *protoutil.Clone(defaultSystemZoneConfig).(*ZoneConfig)
 }
 
 // TestingSetDefaultZoneConfig is a testing-only function that changes the
@@ -312,7 +356,7 @@ func DefaultSystemZoneConfig() ZoneConfig {
 func TestingSetDefaultZoneConfig(cfg ZoneConfig) func() {
 	testingLock.Lock()
 	oldConfig := defaultZoneConfig
-	defaultZoneConfig = cfg
+	defaultZoneConfig = &cfg
 	testingLock.Unlock()
 
 	return func() {
@@ -326,8 +370,8 @@ func TestingSetDefaultZoneConfig(cfg ZoneConfig) func() {
 // default zone config and returns a function that reverts the change.
 func TestingSetDefaultSystemZoneConfig(cfg ZoneConfig) func() {
 	testingLock.Lock()
-	oldConfig := defaultZoneConfig
-	defaultSystemZoneConfig = cfg
+	oldConfig := defaultSystemZoneConfig
+	defaultSystemZoneConfig = &cfg
 	testingLock.Unlock()
 
 	return func() {
@@ -337,6 +381,35 @@ func TestingSetDefaultSystemZoneConfig(cfg ZoneConfig) func() {
 	}
 }
 
+// IsComplete returns whether all the fields are set.
+func (z *ZoneConfig) IsComplete() bool {
+	return ((z.NumReplicas != nil) && (z.RangeMinBytes != nil) &&
+		(z.RangeMaxBytes != nil) && (z.GC != nil) &&
+		(!z.InheritedConstraints) && (!z.InheritedLeasePreferences))
+}
+
+// ValidateTandemFields returns an error if the ZoneConfig to be written
+// specifies a configuration that could cause problems with the introduction
+// of cascading zone configs.
+func (z *ZoneConfig) ValidateTandemFields() error {
+	var numConstrainedRepls int32
+	for _, constraint := range z.Constraints {
+		numConstrainedRepls += constraint.NumReplicas
+	}
+
+	if numConstrainedRepls > 0 && z.NumReplicas == nil {
+		return fmt.Errorf("when per-replica constraints are set, num_replicas must be set as well")
+	}
+	if (z.RangeMinBytes != nil || z.RangeMaxBytes != nil) &&
+		(z.RangeMinBytes == nil || z.RangeMaxBytes == nil) {
+		return fmt.Errorf("range_min_bytes and range_max_bytes must be set together")
+	}
+	if !z.InheritedLeasePreferences && z.InheritedConstraints {
+		return fmt.Errorf("lease preferences can not be set unless the constraints are explicitly set as well")
+	}
+	return nil
+}
+
 // Validate returns an error if the ZoneConfig specifies a known-dangerous or
 // disallowed configuration.
 func (z *ZoneConfig) Validate() error {
@@ -346,36 +419,38 @@ func (z *ZoneConfig) Validate() error {
 		}
 	}
 
-	switch {
-	case z.NumReplicas < 0:
-		return fmt.Errorf("at least one replica is required")
-	case z.NumReplicas == 0:
-		if len(z.Subzones) > 0 {
-			// NumReplicas == 0 is allowed when this ZoneConfig is a subzone
-			// placeholder. See IsSubzonePlaceholder.
-			return nil
+	if z.NumReplicas != nil {
+		switch {
+		case *z.NumReplicas < 0:
+			return fmt.Errorf("at least one replica is required")
+		case *z.NumReplicas == 0:
+			if len(z.Subzones) > 0 {
+				// NumReplicas == 0 is allowed when this ZoneConfig is a subzone
+				// placeholder. See IsSubzonePlaceholder.
+				return nil
+			}
+			return fmt.Errorf("at least one replica is required")
+		case *z.NumReplicas == 2:
+			return fmt.Errorf("at least 3 replicas are required for multi-replica configurations")
 		}
-		return fmt.Errorf("at least one replica is required")
-	case z.NumReplicas == 2:
-		return fmt.Errorf("at least 3 replicas are required for multi-replica configurations")
 	}
 
-	if z.RangeMaxBytes < minRangeMaxBytes {
+	if z.RangeMaxBytes != nil && *z.RangeMaxBytes < minRangeMaxBytes {
 		return fmt.Errorf("RangeMaxBytes %d less than minimum allowed %d",
-			z.RangeMaxBytes, minRangeMaxBytes)
+			*z.RangeMaxBytes, minRangeMaxBytes)
 	}
 
-	if z.RangeMinBytes < 0 {
-		return fmt.Errorf("RangeMinBytes %d less than minimum allowed 0", z.RangeMinBytes)
+	if z.RangeMinBytes != nil && *z.RangeMinBytes < 0 {
+		return fmt.Errorf("RangeMinBytes %d less than minimum allowed 0", *z.RangeMinBytes)
 	}
-	if z.RangeMinBytes >= z.RangeMaxBytes {
+	if z.RangeMinBytes != nil && z.RangeMaxBytes != nil && *z.RangeMinBytes >= *z.RangeMaxBytes {
 		return fmt.Errorf("RangeMinBytes %d is greater than or equal to RangeMaxBytes %d",
-			z.RangeMinBytes, z.RangeMaxBytes)
+			*z.RangeMinBytes, *z.RangeMaxBytes)
 	}
 
 	// Reserve the value 0 to potentially have some special meaning in the future,
 	// such as to disable GC.
-	if z.GC.TTLSeconds < 1 {
+	if z.GC != nil && z.GC.TTLSeconds < 1 {
 		return fmt.Errorf("GC.TTLSeconds %d less than minimum allowed 1", z.GC.TTLSeconds)
 	}
 
@@ -401,16 +476,16 @@ func (z *ZoneConfig) Validate() error {
 			for _, constraint := range constraints.Constraints {
 				// TODO(a-robinson): Relax this constraint to allow prohibited replicas,
 				// as discussed on #23014.
-				if constraint.Type != Constraint_REQUIRED && constraints.NumReplicas != z.NumReplicas {
+				if constraint.Type != Constraint_REQUIRED && z.NumReplicas != nil && constraints.NumReplicas != *z.NumReplicas {
 					return fmt.Errorf(
 						"only required constraints (prefixed with a '+') can be applied to a subset of replicas")
 				}
 			}
 		}
-		if numConstrainedRepls > int64(z.NumReplicas) {
+		if z.NumReplicas != nil && numConstrainedRepls > int64(*z.NumReplicas) {
 			return fmt.Errorf("the number of replicas specified in constraints (%d) cannot be greater "+
 				"than the number of replicas configured for the zone (%d)",
-				numConstrainedRepls, z.NumReplicas)
+				numConstrainedRepls, *z.NumReplicas)
 		}
 	}
 
@@ -429,6 +504,82 @@ func (z *ZoneConfig) Validate() error {
 	return nil
 }
 
+// InheritFromParent hydrates a zones missing fields from its parent.
+func (z *ZoneConfig) InheritFromParent(parent *ZoneConfig) {
+	if z.NumReplicas == nil {
+		if parent.NumReplicas != nil {
+			z.NumReplicas = proto.Int32(*parent.NumReplicas)
+		}
+	}
+	if z.RangeMinBytes == nil {
+		if parent.RangeMinBytes != nil {
+			z.RangeMinBytes = proto.Int64(*parent.RangeMinBytes)
+		}
+	}
+	if z.RangeMaxBytes == nil {
+		if parent.RangeMaxBytes != nil {
+			z.RangeMaxBytes = proto.Int64(*parent.RangeMaxBytes)
+		}
+	}
+	if z.GC == nil {
+		if parent.GC != nil {
+			tempGC := *parent.GC
+			z.GC = &tempGC
+		}
+	}
+	if z.InheritedConstraints {
+		if !parent.InheritedConstraints {
+			z.Constraints = parent.Constraints
+			z.InheritedConstraints = false
+		}
+	}
+	if z.InheritedLeasePreferences {
+		if !parent.InheritedLeasePreferences {
+			z.LeasePreferences = parent.LeasePreferences
+			z.InheritedLeasePreferences = false
+		}
+	}
+}
+
+// CopyFromZone copies over the specified fields from the other zone.
+func (z *ZoneConfig) CopyFromZone(other ZoneConfig, fieldList []tree.Name) {
+	for _, fieldName := range fieldList {
+		if fieldName == "num_replicas" {
+			z.NumReplicas = nil
+			if other.NumReplicas != nil {
+				z.NumReplicas = proto.Int32(*other.NumReplicas)
+			}
+		}
+		if fieldName == "range_min_bytes" {
+			z.RangeMinBytes = nil
+			if other.RangeMinBytes != nil {
+				z.RangeMinBytes = proto.Int64(*other.RangeMinBytes)
+			}
+		}
+		if fieldName == "range_max_bytes" {
+			z.RangeMaxBytes = nil
+			if other.RangeMaxBytes != nil {
+				z.RangeMaxBytes = proto.Int64(*other.RangeMaxBytes)
+			}
+		}
+		if fieldName == "gc.ttlseconds" {
+			z.GC = nil
+			if other.GC != nil {
+				tempGC := *other.GC
+				z.GC = &tempGC
+			}
+		}
+		if fieldName == "constraints" {
+			z.Constraints = other.Constraints
+			z.InheritedConstraints = other.InheritedConstraints
+		}
+		if fieldName == "lease_preferences" {
+			z.LeasePreferences = other.LeasePreferences
+			z.InheritedLeasePreferences = other.InheritedLeasePreferences
+		}
+	}
+}
+
 // StoreMatchesConstraint returns whether a store matches the given constraint.
 func StoreMatchesConstraint(store roachpb.StoreDescriptor, constraint Constraint) bool {
 	hasConstraint := storeHasConstraint(store, constraint)
@@ -469,6 +620,8 @@ func storeHasConstraint(store roachpb.StoreDescriptor, c Constraint) bool {
 // method on non-table ZoneConfigs.
 func (z *ZoneConfig) DeleteTableConfig() {
 	*z = ZoneConfig{
+		// Have to set NumReplicas to 0 so it is recognized as a placeholder.
+		NumReplicas:  proto.Int32(0),
 		Subzones:     z.Subzones,
 		SubzoneSpans: z.SubzoneSpans,
 	}
@@ -482,7 +635,7 @@ func (z *ZoneConfig) IsSubzonePlaceholder() bool {
 	// A ZoneConfig with zero replicas is otherwise invalid, so we repurpose it to
 	// indicate that a ZoneConfig is a placeholder for subzones rather than
 	// introducing a dedicated IsPlaceholder flag.
-	return z.NumReplicas == 0
+	return z.NumReplicas != nil && *z.NumReplicas == 0
 }
 
 // GetSubzone returns the most specific Subzone that applies to the specified
@@ -492,7 +645,8 @@ func (z *ZoneConfig) IsSubzonePlaceholder() bool {
 func (z *ZoneConfig) GetSubzone(indexID uint32, partition string) *Subzone {
 	for _, s := range z.Subzones {
 		if s.IndexID == indexID && s.PartitionName == partition {
-			return &s
+			copySubzone := s
+			return &copySubzone
 		}
 	}
 	if partition != "" {
@@ -510,7 +664,8 @@ func (z ZoneConfig) GetSubzoneForKeySuffix(keySuffix []byte) *Subzone {
 		// directly to keySuffix. An unset EndKey implies Key.PrefixEnd().
 		if (s.Key.Compare(keySuffix) <= 0) &&
 			((s.EndKey == nil && bytes.HasPrefix(keySuffix, s.Key)) || s.EndKey.Compare(keySuffix) > 0) {
-			return &z.Subzones[s.SubzoneIndex]
+			copySubzone := z.Subzones[s.SubzoneIndex]
+			return &copySubzone
 		}
 	}
 	return nil
@@ -572,3 +727,63 @@ func (z ZoneConfig) subzoneSplits() []roachpb.RKey {
 	}
 	return out
 }
+
+// ReplicaConstraintsCount is part of the cat.Zone interface.
+func (z *ZoneConfig) ReplicaConstraintsCount() int {
+	return len(z.Constraints)
+}
+
+// ReplicaConstraints is part of the cat.Zone interface.
+func (z *ZoneConfig) ReplicaConstraints(i int) cat.ReplicaConstraints {
+	return &z.Constraints[i]
+}
+
+// LeasePreferenceCount is part of the cat.Zone interface.
+func (z *ZoneConfig) LeasePreferenceCount() int {
+	return len(z.LeasePreferences)
+}
+
+// LeasePreference is part of the cat.Zone interface.
+func (z *ZoneConfig) LeasePreference(i int) cat.ConstraintSet {
+	return &z.LeasePreferences[i]
+}
+
+// ConstraintCount is part of the cat.LeasePreference interface.
+func (l *LeasePreference) ConstraintCount() int {
+	return len(l.Constraints)
+}
+
+// Constraint is part of the cat.LeasePreference interface.
+func (l *LeasePreference) Constraint(i int) cat.Constraint {
+	return &l.Constraints[i]
+}
+
+// ReplicaCount is part of the cat.ReplicaConstraints interface.
+func (c *Constraints) ReplicaCount() int32 {
+	return c.NumReplicas
+}
+
+// ConstraintCount is part of the cat.ReplicaConstraints interface.
+func (c *Constraints) ConstraintCount() int {
+	return len(c.Constraints)
+}
+
+// Constraint is part of the cat.ReplicaConstraints interface.
+func (c *Constraints) Constraint(i int) cat.Constraint {
+	return &c.Constraints[i]
+}
+
+// IsRequired is part of the cat.Constraint interface.
+func (c *Constraint) IsRequired() bool {
+	return c.Type == Constraint_REQUIRED
+}
+
+// GetKey is part of the cat.Constraint interface.
+func (c *Constraint) GetKey() string {
+	return c.Key
+}
+
+// GetValue is part of the cat.Constraint interface.
+func (c *Constraint) GetValue() string {
+	return c.Value
+}
