@@ -34,66 +34,70 @@ func TestZoneConfigValidate(t *testing.T) {
 		cfg      ZoneConfig
 		expected string
 	}{
-		{
-			ZoneConfig{},
-			"at least one replica is required",
-		},
 		{
 			ZoneConfig{
-				NumReplicas: -1,
+				NumReplicas: proto.Int32(0),
 			},
 			"at least one replica is required",
 		},
 		{
 			ZoneConfig{
-				NumReplicas: 2,
+				NumReplicas: proto.Int32(-1),
+			},
+			"at least one replica is required",
+		},
+		{
+			ZoneConfig{
+				NumReplicas: proto.Int32(2),
 			},
 			"at least 3 replicas are required for multi-replica configurations",
 		},
 		{
 			ZoneConfig{
-				NumReplicas: 1,
+				NumReplicas:   proto.Int32(1),
+				RangeMaxBytes: proto.Int64(0),
 			},
 			"RangeMaxBytes 0 less than minimum allowed",
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
+				GC:            &GCPolicy{TTLSeconds: 0},
 			},
 			"GC.TTLSeconds 0 less than minimum allowed",
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 			},
 			"",
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
-				RangeMinBytes: -1,
+				NumReplicas:   proto.Int32(1),
+				RangeMinBytes: proto.Int64(-1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 			},
 			"RangeMinBytes -1 less than minimum allowed",
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMinBytes: DefaultZoneConfig().RangeMaxBytes,
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 			},
 			"is greater than or equal to RangeMaxBytes",
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				Constraints: []Constraints{
 					{Constraints: []Constraint{{Value: "a", Type: Constraint_DEPRECATED_POSITIVE}}},
 				},
@@ -102,9 +106,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				Constraints: []Constraints{
 					{Constraints: []Constraint{{Value: "a", Type: Constraint_PROHIBITED}}},
 				},
@@ -113,9 +117,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				Constraints: []Constraints{
 					{
 						Constraints: []Constraint{{Value: "a", Type: Constraint_PROHIBITED}},
@@ -127,9 +131,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				Constraints: []Constraints{
 					{
 						Constraints: []Constraint{{Value: "a", Type: Constraint_REQUIRED}},
@@ -141,9 +145,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   3,
+				NumReplicas:   proto.Int32(3),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				Constraints: []Constraints{
 					{
 						Constraints: []Constraint{{Value: "a", Type: Constraint_REQUIRED}},
@@ -155,9 +159,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				Constraints: []Constraints{
 					{
 						Constraints: []Constraint{{Value: "a", Type: Constraint_REQUIRED}},
@@ -173,9 +177,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   3,
+				NumReplicas:   proto.Int32(3),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				Constraints: []Constraints{
 					{
 						Constraints: []Constraint{{Value: "a", Type: Constraint_REQUIRED}},
@@ -191,9 +195,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				LeasePreferences: []LeasePreference{
 					{
 						Constraints: []Constraint{},
@@ -204,9 +208,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				LeasePreferences: []LeasePreference{
 					{
 						Constraints: []Constraint{{Value: "a", Type: Constraint_DEPRECATED_POSITIVE}},
@@ -217,9 +221,9 @@ func TestZoneConfigValidate(t *testing.T) {
 		},
 		{
 			ZoneConfig{
-				NumReplicas:   1,
+				NumReplicas:   proto.Int32(1),
 				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
-				GC:            GCPolicy{TTLSeconds: 1},
+				GC:            &GCPolicy{TTLSeconds: 1},
 				LeasePreferences: []LeasePreference{
 					{
 						Constraints: []Constraint{{Value: "a", Type: Constraint_REQUIRED}},
@@ -241,13 +245,65 @@ func TestZoneConfigValidate(t *testing.T) {
 	}
 }
 
+func TestZoneConfigValidateTandemFields(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	testCases := []struct {
+		cfg      ZoneConfig
+		expected string
+	}{
+		{
+			ZoneConfig{
+				RangeMaxBytes: DefaultZoneConfig().RangeMaxBytes,
+			},
+			"range_min_bytes and range_max_bytes must be set together",
+		},
+		{
+			ZoneConfig{
+				RangeMinBytes: DefaultZoneConfig().RangeMinBytes,
+			},
+			"range_min_bytes and range_max_bytes must be set together",
+		},
+		{
+			ZoneConfig{
+				Constraints: []Constraints{
+					{
+						Constraints: []Constraint{{Value: "a", Type: Constraint_REQUIRED}},
+						NumReplicas: 2,
+					},
+				},
+			},
+			"when per-replica constraints are set, num_replicas must be set as well",
+		},
+		{
+			ZoneConfig{
+				InheritedConstraints:      true,
+				InheritedLeasePreferences: false,
+				LeasePreferences: []LeasePreference{
+					{
+						Constraints: []Constraint{},
+					},
+				},
+			},
+			"lease preferences can not be set unless the constraints are explicitly set as well",
+		},
+	}
+
+	for i, c := range testCases {
+		err := c.cfg.ValidateTandemFields()
+		if !testutils.IsError(err, c.expected) {
+			t.Errorf("%d: expected %q, got %v", i, c.expected, err)
+		}
+	}
+}
+
 func TestZoneConfigSubzones(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
 	zone := DefaultZoneConfig()
-	subzoneAInvalid := Subzone{IndexID: 1, PartitionName: "a", Config: ZoneConfig{}}
-	subzoneA := Subzone{IndexID: 1, PartitionName: "a", Config: DefaultZoneConfig()}
-	subzoneB := Subzone{IndexID: 1, PartitionName: "b", Config: DefaultZoneConfig()}
+	subzoneAInvalid := Subzone{IndexID: 1, PartitionName: "a", Config: ZoneConfig{NumReplicas: proto.Int32(-1)}}
+	subzoneA := Subzone{IndexID: 1, PartitionName: "a", Config: zone}
+	subzoneB := Subzone{IndexID: 1, PartitionName: "b", Config: zone}
 
 	if zone.IsSubzonePlaceholder() {
 		t.Errorf("default zone config should not be considered a subzone placeholder")
@@ -281,7 +337,8 @@ func TestZoneConfigSubzones(t *testing.T) {
 
 	zone.DeleteTableConfig()
 	if e := (ZoneConfig{
-		Subzones: []Subzone{subzoneA, subzoneB},
+		NumReplicas: proto.Int32(0),
+		Subzones:    []Subzone{subzoneA, subzoneB},
 	}); !e.Equal(zone) {
 		t.Errorf("expected zone after clearing to equal %+v, but got %+v", e, zone)
 	}
@@ -333,12 +390,12 @@ func TestZoneConfigMarshalYAML(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
 	original := ZoneConfig{
-		RangeMinBytes: 1,
-		RangeMaxBytes: 1,
-		GC: GCPolicy{
+		RangeMinBytes: proto.Int64(1),
+		RangeMaxBytes: proto.Int64(1),
+		GC: &GCPolicy{
 			TTLSeconds: 1,
 		},
-		NumReplicas: 1,
+		NumReplicas: proto.Int32(1),
 	}
 
 	testCases := []struct {
@@ -720,7 +777,7 @@ func TestMarshalableZoneConfigRoundTrip(t *testing.T) {
 	original := NewPopulatedZoneConfig(
 		rand.New(rand.NewSource(timeutil.Now().UnixNano())), false /* easy */)
 	marshalable := zoneConfigToMarshalable(*original)
-	roundTripped := zoneConfigFromMarshalable(marshalable)
+	roundTripped := zoneConfigFromMarshalable(marshalable, *original)
 
 	if !reflect.DeepEqual(roundTripped, *original) {
 		t.Errorf("round-tripping a ZoneConfig through a marshalableZoneConfig failed:\noriginal:\n%+v\nmarshable:\n%+v\ngot:\n%+v", original, marshalable, roundTripped)
