@@ -63,19 +63,13 @@ import (
 	"sync"
 	"time"
 
-	"google.golang.org/grpc"
-
-	"github.com/gogo/protobuf/proto"
-	"github.com/pkg/errors"
-	circuit "github.com/rubyist/circuitbreaker"
-
+	circuit "github.com/cockroachdb/circuitbreaker"
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip/resolver"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/rpc"
 	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
-	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/log/logtags"
@@ -84,6 +78,10 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/gogo/protobuf/proto"
+	"github.com/pkg/errors"
+	"google.golang.org/grpc"
 )
 
 const (
@@ -256,8 +254,7 @@ type Gossip struct {
 	// here and its own set of callbacks.
 	// We do not use the infostore to avoid unmarshalling under the
 	// main gossip lock.
-	systemConfig         config.SystemConfig
-	systemConfigSet      bool
+	systemConfig         *config.SystemConfig
 	systemConfigMu       syncutil.RWMutex
 	systemConfigChannels []chan<- struct{}
 
@@ -997,6 +994,26 @@ func (g *Gossip) AddInfoProto(key string, msg protoutil.Message, ttl time.Durati
 	return g.AddInfo(key, bytes, ttl)
 }
 
+// AddClusterID is a convenience method for gossipping the cluster ID. There's
+// no TTL - the record lives forever.
+func (g *Gossip) AddClusterID(val uuid.UUID) error {
+	return g.AddInfo(KeyClusterID, val.GetBytes(), 0 /* ttl */)
+}
+
+// GetClusterID returns the cluster ID if it has been gossipped. If it hasn't,
+// (so if this gossip instance is not "connected"), an error is returned.
+func (g *Gossip) GetClusterID() (uuid.UUID, error) {
+	uuidBytes, err := g.GetInfo(KeyClusterID)
+	if err != nil {
+		return uuid.Nil, errors.Wrap(err, "unable to ascertain cluster ID from gossip network")
+	}
+	clusterID, err := uuid.FromBytes(uuidBytes)
+	if err != nil {
+		return uuid.Nil, errors.Wrap(err, "unable to parse cluster ID from gossip network")
+	}
+	return clusterID, nil
+}
+
 // GetInfo returns an info value by key or an KeyNotPresentError if specified
 // key does not exist or has expired.
 func (g *Gossip) GetInfo(key string) ([]byte, error) {
@@ -1112,11 +1129,11 @@ func (g *Gossip) RegisterCallback(pattern string, method Callback, opts ...Callb
 }
 
 // GetSystemConfig returns the local unmarshaled version of the system config.
-// The second return value indicates whether the system config has been set yet.
-func (g *Gossip) GetSystemConfig() (config.SystemConfig, bool) {
+// Returns nil if the system config hasn't been set yet.
+func (g *Gossip) GetSystemConfig() *config.SystemConfig {
 	g.systemConfigMu.RLock()
 	defer g.systemConfigMu.RUnlock()
-	return g.systemConfig, g.systemConfigSet
+	return g.systemConfig
 }
 
 // RegisterSystemConfigChannel registers a channel to signify updates for the
@@ -1132,23 +1149,23 @@ func (g *Gossip) RegisterSystemConfigChannel() <-chan struct{} {
 	g.systemConfigChannels = append(g.systemConfigChannels, c)
 
 	// Notify the channel right away if we have a config.
-	if g.systemConfigSet {
+	if g.systemConfig != nil {
 		c <- struct{}{}
 	}
 
 	return c
 }
 
-// updateSystemConfig is the raw gossip info callback.
-// Unmarshal the system config, and if successfully, update out
-// copy and run the callbacks.
+// updateSystemConfig is the raw gossip info callback. Unmarshal the
+// system config, and if successful, send on each system config
+// channel.
 func (g *Gossip) updateSystemConfig(key string, content roachpb.Value) {
 	ctx := g.AnnotateCtx(context.TODO())
 	if key != KeySystemConfig {
 		log.Fatalf(ctx, "wrong key received on SystemConfig callback: %s", key)
 	}
-	cfg := config.SystemConfig{}
-	if err := content.GetProto(&cfg); err != nil {
+	cfg := config.NewSystemConfig()
+	if err := content.GetProto(&cfg.SystemConfigEntries); err != nil {
 		log.Errorf(ctx, "could not unmarshal system config on callback: %s", err)
 		return
 	}
@@ -1156,7 +1173,6 @@ func (g *Gossip) updateSystemConfig(key string, content roachpb.Value) {
 	g.systemConfigMu.Lock()
 	defer g.systemConfigMu.Unlock()
 	g.systemConfig = cfg
-	g.systemConfigSet = true
 	for _, c := range g.systemConfigChannels {
 		select {
 		case c <- struct{}{}:
@@ -1212,6 +1228,11 @@ func (g *Gossip) Start(advertAddr net.Addr, resolvers []resolver.Resolver) {
 	g.manage()                 // manage gossip clients
 }
 
+// Stopper returns the stopper for this gossip instance.
+func (g *Gossip) Stopper() *stop.Stopper {
+	return g.server.stopper
+}
+
 // hasIncomingLocked returns whether the server has an incoming gossip
 // client matching the provided node ID. Mutex should be held by
 // caller.
@@ -1527,7 +1548,8 @@ func (g *Gossip) startClientLocked(addr net.Addr) {
 	defer g.clientsMu.Unlock()
 	breaker, ok := g.clientsMu.breakers[addr.String()]
 	if !ok {
-		breaker = g.rpcContext.NewBreaker()
+		name := fmt.Sprintf("gossip %v->%v", g.rpcContext.Addr, addr)
+		breaker = g.rpcContext.NewBreaker(name)
 		g.clientsMu.breakers[addr.String()] = breaker
 	}
 	ctx := g.AnnotateCtx(context.TODO())
@@ -1564,11 +1586,3 @@ func (g *Gossip) findClient(match func(*client) bool) *client {
 	}
 	return nil
 }
-
-var _ security.RequestWithUser = &Request{}
-
-// GetUser implements security.RequestWithUser.
-// Gossip messages are always sent by the node user.
-func (*Request) GetUser() string {
-	return security.NodeUser
-}
