@@ -17,9 +17,8 @@ package client
 import (
 	"context"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/pkg/errors"
 )
 
 const (
@@ -416,6 +415,17 @@ func (b *Batch) PutInline(key, value interface{}) {
 // key can be either a byte slice or a string. value can be any key type, a
 // protoutil.Message or any Go primitive type (bool, int, etc).
 func (b *Batch) CPut(key, value, expValue interface{}) {
+	b.cputInternal(key, value, expValue, false)
+}
+
+// CPutAllowingIfNotExists is like CPut except it also allows the Put when the
+// existing entry does not exist -- i.e. it succeeds if there is no existing
+// entry or the existing entry has the expected value.
+func (b *Batch) CPutAllowingIfNotExists(key, value, expValue interface{}) {
+	b.cputInternal(key, value, expValue, true)
+}
+
+func (b *Batch) cputInternal(key, value, expValue interface{}, allowNotExist bool) {
 	k, err := marshalKey(key)
 	if err != nil {
 		b.initResult(0, 1, notRaw, err)
@@ -431,7 +441,7 @@ func (b *Batch) CPut(key, value, expValue interface{}) {
 		b.initResult(0, 1, notRaw, err)
 		return
 	}
-	b.appendReqs(roachpb.NewConditionalPut(k, v, ev))
+	b.appendReqs(roachpb.NewConditionalPut(k, v, ev, allowNotExist))
 	b.initResult(1, 1, notRaw, nil)
 }
 
@@ -619,7 +629,10 @@ func (b *Batch) adminTransferLease(key interface{}, target roachpb.StoreID) {
 // adminChangeReplicas is only exported on DB. It is here for symmetry with the
 // other operations.
 func (b *Batch) adminChangeReplicas(
-	key interface{}, changeType roachpb.ReplicaChangeType, targets []roachpb.ReplicationTarget,
+	key interface{},
+	changeType roachpb.ReplicaChangeType,
+	targets []roachpb.ReplicationTarget,
+	expDesc roachpb.RangeDescriptor,
 ) {
 	k, err := marshalKey(key)
 	if err != nil {
@@ -632,6 +645,7 @@ func (b *Batch) adminChangeReplicas(
 		},
 		ChangeType: changeType,
 		Targets:    targets,
+		ExpDesc:    &expDesc,
 	}
 	b.appendReqs(req)
 	b.initResult(1, 0, notRaw, nil)
