@@ -19,8 +19,6 @@ import (
 	"context"
 	"fmt"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
@@ -29,6 +27,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 // KeyValue represents a single key/value pair. This is similar to
@@ -500,10 +499,23 @@ func (db *DB) AdminChangeReplicas(
 	key interface{},
 	changeType roachpb.ReplicaChangeType,
 	targets []roachpb.ReplicationTarget,
-) error {
+	expDesc roachpb.RangeDescriptor,
+) (*roachpb.RangeDescriptor, error) {
 	b := &Batch{}
-	b.adminChangeReplicas(key, changeType, targets)
-	return getOneErr(db.Run(ctx, b), b)
+	b.adminChangeReplicas(key, changeType, targets, expDesc)
+	if err := getOneErr(db.Run(ctx, b), b); err != nil {
+		return nil, err
+	}
+	responses := b.response.Responses
+	if len(responses) == 0 {
+		return nil, errors.Errorf("unexpected empty responses for AdminChangeReplicas")
+	}
+	resp, ok := responses[0].GetInner().(*roachpb.AdminChangeReplicasResponse)
+	if !ok {
+		return nil, errors.Errorf("unexpected response of type %T for AdminChangeReplicas",
+			responses[0].GetInner())
+	}
+	return resp.Desc, nil
 }
 
 // AdminRelocateRange relocates the replicas for a range onto the specified
@@ -588,10 +600,10 @@ func (db *DB) Txn(ctx context.Context, retryable func(context.Context, *Txn) err
 	if err != nil {
 		txn.CleanupOnError(ctx, err)
 	}
-	// Terminate HandledRetryableTxnError here, so it doesn't cause a higher-level
+	// Terminate TransactionRetryWithProtoRefreshError here, so it doesn't cause a higher-level
 	// txn to be retried. We don't do this in any of the other functions in DB; I
 	// guess we should.
-	if _, ok := err.(*roachpb.HandledRetryableTxnError); ok {
+	if _, ok := err.(*roachpb.TransactionRetryWithProtoRefreshError); ok {
 		return errors.Wrapf(err, "terminated retryable error")
 	}
 	return err
