@@ -20,7 +20,7 @@ import (
 	"time"
 
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
@@ -58,7 +58,6 @@ type Txn struct {
 		syncutil.Mutex
 		ID        uuid.UUID
 		debugName string
-		isolation enginepb.IsolationType
 
 		// userPriority is the transaction's priority. If not set,
 		// NormalUserPriority will be used.
@@ -104,7 +103,6 @@ func NewTxn(ctx context.Context, db *DB, gatewayNodeID roachpb.NodeID, typ TxnTy
 		"unnamed",
 		nil, // baseKey
 		roachpb.NormalUserPriority,
-		enginepb.SERIALIZABLE,
 		now,
 		db.clock.MaxOffset().Nanoseconds(),
 	)
@@ -242,27 +240,6 @@ func (txn *Txn) debugNameLocked() string {
 	return fmt.Sprintf("%s (id: %s)", txn.mu.debugName, txn.mu.ID)
 }
 
-// SetIsolation sets the transaction's isolation type. Transactions default to
-// serializable isolation. The isolation must be set before any operations are
-// performed on the transaction.
-func (txn *Txn) SetIsolation(isolation enginepb.IsolationType) error {
-	txn.mu.Lock()
-	defer txn.mu.Unlock()
-
-	if err := txn.mu.sender.SetIsolation(isolation); err != nil {
-		return err
-	}
-	txn.mu.isolation = isolation
-	return nil
-}
-
-// Isolation returns the transaction's isolation type.
-func (txn *Txn) Isolation() enginepb.IsolationType {
-	txn.mu.Lock()
-	defer txn.mu.Unlock()
-	return txn.mu.isolation
-}
-
 // OrigTimestamp returns the transaction's starting timestamp.
 // Note a transaction can be internally pushed forward in time before
 // committing so this is not guaranteed to be the commit timestamp.
@@ -623,21 +600,28 @@ func (txn *Txn) rollback(ctx context.Context) *roachpb.Error {
 	// long this request is going to be around or it could leak a goroutine (in case of a
 	// long-lived network partition).
 	stopper := txn.db.ctx.Stopper
-	ctx, cancel1 := stopper.WithCancelOnQuiesce(txn.db.AnnotateCtx(context.Background()))
-	// NB: cancel2 makes cancel1 obsolete, but the linters don't know that.
-	ctx, cancel2 := context.WithTimeout(ctx, 3*time.Second)
-
+	ctx, cancel := stopper.WithCancelOnQuiesce(txn.db.AnnotateCtx(context.Background()))
 	if err := stopper.RunAsyncTask(ctx, "async-rollback", func(ctx context.Context) {
-		defer cancel1()
-		defer cancel2()
+		defer cancel()
 		var ba roachpb.BatchRequest
 		ba.Add(endTxnReq(false /* commit */, nil /* deadline */, false /* systemConfigTrigger */))
-		if _, pErr := txn.Send(ctx, ba); pErr != nil {
-			log.Infof(ctx, "async rollback failed: %s", pErr)
-		}
+		_ = contextutil.RunWithTimeout(ctx, "async txn rollback", 3*time.Second, func(ctx context.Context) error {
+			if _, pErr := txn.Send(ctx, ba); pErr != nil {
+				if statusErr, ok := pErr.GetDetail().(*roachpb.TransactionStatusError); ok &&
+					statusErr.Reason == roachpb.TransactionStatusError_REASON_TXN_COMMITTED {
+					// A common cause of these async rollbacks failing is when they're
+					// triggered by a ctx canceled while a commit is in-flight (and it's too
+					// late for it to be canceled), and so the rollback finds the txn to be
+					// already committed. We don't spam the logs with those.
+					log.VEventf(ctx, 2, "async rollback failed: %s", pErr)
+				} else {
+					log.Infof(ctx, "async rollback failed: %s", pErr)
+				}
+			}
+			return nil
+		})
 	}); err != nil {
-		cancel1()
-		cancel2()
+		cancel()
 		return roachpb.NewError(err)
 	}
 	return nil
@@ -710,7 +694,7 @@ func (txn *Txn) exec(ctx context.Context, fn func(context.Context, *Txn) error)
 				err = txn.Commit(ctx)
 				log.Eventf(ctx, "client.Txn did AutoCommit. err: %v\n", err)
 				if err != nil {
-					if _, retryable := err.(*roachpb.HandledRetryableTxnError); !retryable {
+					if _, retryable := err.(*roachpb.TransactionRetryWithProtoRefreshError); !retryable {
 						// We can't retry, so let the caller know we tried to
 						// autocommit.
 						err = &AutoCommitError{cause: err}
@@ -726,16 +710,16 @@ func (txn *Txn) exec(ctx context.Context, fn func(context.Context, *Txn) error)
 		case *roachpb.UnhandledRetryableError:
 			if txn.typ == RootTxn {
 				// We sent transactional requests, so the TxnCoordSender was supposed to
-				// turn retryable errors into HandledRetryableTxnError. Note that this
+				// turn retryable errors into TransactionRetryWithProtoRefreshError. Note that this
 				// applies only in the case where this is the root transaction.
 				log.Fatalf(ctx, "unexpected UnhandledRetryableError at the txn.exec() level: %s", err)
 			}
 
-		case *roachpb.HandledRetryableTxnError:
+		case *roachpb.TransactionRetryWithProtoRefreshError:
 			if !txn.IsRetryableErrMeantForTxn(*t) {
 				// Make sure the txn record that err carries is for this txn.
 				// If it's not, we terminate the "retryable" character of the error. We
-				// might get a HandledRetryableTxnError if the closure ran another
+				// might get a TransactionRetryWithProtoRefreshError if the closure ran another
 				// transaction internally and let the error propagate upwards.
 				return errors.Wrapf(err, "retryable error from another txn")
 			}
@@ -764,7 +748,9 @@ func (txn *Txn) PrepareForRetry(ctx context.Context, err error) {
 
 // IsRetryableErrMeantForTxn returns true if err is a retryable
 // error meant to restart this client transaction.
-func (txn *Txn) IsRetryableErrMeantForTxn(retryErr roachpb.HandledRetryableTxnError) bool {
+func (txn *Txn) IsRetryableErrMeantForTxn(
+	retryErr roachpb.TransactionRetryWithProtoRefreshError,
+) bool {
 	txn.mu.Lock()
 	defer txn.mu.Unlock()
 
@@ -807,7 +793,7 @@ func (txn *Txn) Send(
 		return br, nil
 	}
 
-	if retryErr, ok := pErr.GetDetail().(*roachpb.HandledRetryableTxnError); ok {
+	if retryErr, ok := pErr.GetDetail().(*roachpb.TransactionRetryWithProtoRefreshError); ok {
 		if requestTxnID != retryErr.TxnID {
 			// KV should not return errors for transactions other than the one that sent
 			// the request.
@@ -823,7 +809,7 @@ func (txn *Txn) Send(
 }
 
 func (txn *Txn) handleErrIfRetryableLocked(ctx context.Context, err error) {
-	retryErr, ok := err.(*roachpb.HandledRetryableTxnError)
+	retryErr, ok := err.(*roachpb.TransactionRetryWithProtoRefreshError)
 	if !ok {
 		return
 	}
@@ -871,7 +857,7 @@ func (txn *Txn) AugmentTxnCoordMeta(ctx context.Context, meta roachpb.TxnCoordMe
 
 // UpdateStateOnRemoteRetryableErr updates the txn in response to an error
 // encountered when running a request through the txn. Returns a
-// HandledRetryableError on success or another error on failure.
+// TransactionRetryWithProtoRefreshError on success or another error on failure.
 func (txn *Txn) UpdateStateOnRemoteRetryableErr(ctx context.Context, pErr *roachpb.Error) error {
 	txn.mu.Lock()
 	defer txn.mu.Unlock()
@@ -891,7 +877,7 @@ func (txn *Txn) UpdateStateOnRemoteRetryableErr(ctx context.Context, pErr *roach
 	}
 
 	pErr = txn.mu.sender.UpdateStateOnRemoteRetryableErr(ctx, pErr)
-	txn.replaceSenderIfTxnAbortedLocked(ctx, pErr.GetDetail().(*roachpb.HandledRetryableTxnError), origTxnID)
+	txn.replaceSenderIfTxnAbortedLocked(ctx, pErr.GetDetail().(*roachpb.TransactionRetryWithProtoRefreshError), origTxnID)
 
 	return pErr.GoError()
 }
@@ -902,23 +888,23 @@ func (txn *Txn) UpdateStateOnRemoteRetryableErr(ctx context.Context, pErr *roach
 // origTxnID is the id of the txn that generated retryErr. Note that this can be
 // different from retryErr.Transaction - the latter might be a new transaction.
 func (txn *Txn) replaceSenderIfTxnAbortedLocked(
-	ctx context.Context, retryErr *roachpb.HandledRetryableTxnError, origTxnID uuid.UUID,
+	ctx context.Context, retryErr *roachpb.TransactionRetryWithProtoRefreshError, origTxnID uuid.UUID,
 ) {
 	// The proto inside the error has been prepared for use by the next
 	// transaction attempt.
 	newTxn := &retryErr.Transaction
 
-	if txn.mu.ID == newTxn.ID {
-		// We don't need a new transaction as a result of this error. Nothing more
-		// to do.
-		return
-	}
 	if txn.mu.ID != origTxnID {
 		// The transaction has changed since the request that generated the error
 		// was sent. Nothing more to do.
 		log.VEventf(ctx, 2, "retriable error for old incarnation of the transaction")
 		return
 	}
+	if !retryErr.PrevTxnAborted() {
+		// We don't need a new transaction as a result of this error. Nothing more
+		// to do.
+		return
+	}
 
 	// The ID changed, which means that the cause was a TransactionAbortedError;
 	// we've created a new Transaction that we're about to start using, so we save
@@ -951,7 +937,7 @@ func (txn *Txn) SetFixedTimestamp(ctx context.Context, ts hlc.Timestamp) {
 	txn.mu.sender.SetFixedTimestamp(ctx, ts)
 }
 
-// GenerateForcedRetryableError returns a HandledRetryableTxnError that will
+// GenerateForcedRetryableError returns a TransactionRetryWithProtoRefreshError that will
 // cause the txn to be retried.
 //
 // The transaction's epoch is bumped, simulating to an extent what the
@@ -965,14 +951,13 @@ func (txn *Txn) GenerateForcedRetryableError(ctx context.Context, msg string) er
 	now := txn.db.clock.Now()
 	txn.mu.sender.ManualRestart(ctx, txn.mu.userPriority, now)
 	txn.resetDeadlineLocked()
-	return roachpb.NewHandledRetryableTxnError(
+	return roachpb.NewTransactionRetryWithProtoRefreshError(
 		msg,
 		txn.mu.ID,
 		roachpb.MakeTransaction(
 			txn.debugNameLocked(),
 			nil, // baseKey
 			txn.mu.userPriority,
-			txn.mu.isolation,
 			now,
 			txn.db.clock.MaxOffset().Nanoseconds(),
 		))
