@@ -19,8 +19,6 @@ import (
 	gosql "database/sql"
 	"fmt"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
@@ -31,6 +29,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/pkg/errors"
 )
 
 // Job manages logging the progress of long-running system processes, like
@@ -53,15 +52,21 @@ type Job struct {
 // Record bundles together the user-managed fields in jobspb.Payload.
 type Record struct {
 	Description   string
+	Statement     string
 	Username      string
 	DescriptorIDs sqlbase.IDs
 	Details       jobspb.Details
 	Progress      jobspb.ProgressDetails
+	RunningStatus RunningStatus
 }
 
 // Status represents the status of a job in the system.jobs table.
 type Status string
 
+// RunningStatus represents the more detailed status of a running job in
+// the system.jobs table.
+type RunningStatus string
+
 const (
 	// StatusPending is for jobs that have been created but on which work has
 	// not yet started.
@@ -78,6 +83,15 @@ const (
 	// StatusCanceled is for jobs that were explicitly canceled by the user and
 	// cannot be resumed.
 	StatusCanceled Status = "canceled"
+	// RunningStatusDrainingNames is for jobs that are currently in progress and
+	// are draining names.
+	RunningStatusDrainingNames RunningStatus = "draining names"
+	// RunningStatusWaitingGC is for jobs that are currently in progress and
+	// are waiting for the GC interval to expire
+	RunningStatusWaitingGC RunningStatus = "waiting for GC TTL"
+	// RunningStatusCompaction is for jobs that are currently in progress and
+	// undergoing RocksDB compaction
+	RunningStatusCompaction RunningStatus = "RocksDB compaction"
 )
 
 // Terminal returns whether this status represents a "terminal" state: a state
@@ -138,6 +152,67 @@ func (j *Job) Started(ctx context.Context) error {
 	})
 }
 
+// CheckStatus verifies the status of the job and returns an error if the job's
+// status isn't Running.
+func (j *Job) CheckStatus(ctx context.Context) error {
+	return j.updateRow(
+		ctx, updateProgressOnly,
+		func(
+			_ *client.Txn, status *Status, payload *jobspb.Payload, _ *jobspb.Progress,
+		) (doUpdate bool, _ error) {
+			if *status != StatusRunning {
+				return false, &InvalidStatusError{*j.id, *status, "update progress on", payload.Error}
+			}
+			return false, nil
+		},
+	)
+}
+
+// RunningStatus updates the detailed status of a job currently in progress.
+// It sets the job's RunningStatus field to the value returned by runningStatusFn
+// and persists runningStatusFn's modifications to the job's details, if any.
+func (j *Job) RunningStatus(ctx context.Context, runningStatusFn RunningStatusFn) error {
+	return j.updateRow(ctx, updateProgressAndDetails,
+		func(
+			_ *client.Txn, status *Status, payload *jobspb.Payload, progress *jobspb.Progress,
+		) (doUpdate bool, _ error) {
+			if *status != StatusRunning {
+				return false, &InvalidStatusError{*j.id, *status, "update progress on", payload.Error}
+			}
+			runningStatus, err := runningStatusFn(ctx, progress.Details)
+			if err != nil {
+				return false, err
+			}
+			progress.RunningStatus = string(runningStatus)
+			return true, nil
+		},
+	)
+}
+
+// SetDescription updates the description of a created job.
+func (j *Job) SetDescription(ctx context.Context, updateFn DescriptionUpdateFn) error {
+	return j.updateRow(ctx, updateProgressAndDetails,
+		func(_ *client.Txn, status *Status, payload *jobspb.Payload, _ *jobspb.Progress) (bool, error) {
+			prev := payload.Description
+			desc, err := updateFn(ctx, prev)
+			if err != nil {
+				return false, err
+			}
+			payload.Description = desc
+			return prev != desc, nil
+		},
+	)
+}
+
+// RunningStatusFn is a callback that computes a job's running status
+// given its details. It is safe to modify details in the callback; those
+// modifications will be automatically persisted to the database record.
+type RunningStatusFn func(ctx context.Context, details jobspb.Details) (RunningStatus, error)
+
+// DescriptionUpdateFn is a callback that computes a job's description
+// given its current one.
+type DescriptionUpdateFn func(ctx context.Context, description string) (string, error)
+
 // FractionProgressedFn is a callback that computes a job's completion fraction
 // given its details. It is safe to modify details in the callback; those
 // modifications will be automatically persisted to the database record.
@@ -646,8 +721,8 @@ func RunAndWaitForTerminalState(
 	}()
 
 	var jobID int64
-	var execErr error
 	for r := retry.StartWithCtx(ctx, retry.Options{}); ; {
+		var execErr error
 		select {
 		case <-ctx.Done():
 			return 0, "", ctx.Err()
@@ -675,7 +750,7 @@ func RunAndWaitForTerminalState(
 		select {
 		case <-ctx.Done():
 			return jobID, "", ctx.Err()
-		case execErr = <-execErrCh:
+		case <-execErrCh:
 			// The closure finished, this is a nice hint to wake up, but it only
 			// works once. Close and nil out execErrCh so it blocks from now on.
 			close(execErrCh)
