@@ -23,8 +23,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/jobs"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
@@ -41,6 +39,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/kr/pretty"
+	"github.com/pkg/errors"
 )
 
 // expectation defines the information necessary to determine the validity of
@@ -300,7 +299,7 @@ func TestRegistryLifecycle(t *testing.T) {
 		check(t)
 		sqlDB.Exec(t, "CANCEL JOB $1", *job.ID())
 		// Test for a canceled error message.
-		if err := job.FractionProgressed(ctx, jobs.FractionUpdater(0)); !testutils.IsError(err, "cannot update progress on canceled job") {
+		if err := job.CheckStatus(ctx); !testutils.IsError(err, "cannot update progress on canceled job") {
 			t.Fatalf("unexpected %v", err)
 		}
 		resumeCheckCh <- struct{}{}
@@ -324,7 +323,7 @@ func TestRegistryLifecycle(t *testing.T) {
 		check(t)
 		// Rollback a CANCEL.
 		{
-			txn, err := sqlDB.DB.Begin()
+			txn, err := outerDB.Begin()
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -342,7 +341,7 @@ func TestRegistryLifecycle(t *testing.T) {
 		}
 		// Rollback a PAUSE.
 		{
-			txn, err := sqlDB.DB.Begin()
+			txn, err := outerDB.Begin()
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -358,7 +357,7 @@ func TestRegistryLifecycle(t *testing.T) {
 		}
 		// Now pause it for reals.
 		{
-			txn, err := sqlDB.DB.Begin()
+			txn, err := outerDB.Begin()
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -371,7 +370,7 @@ func TestRegistryLifecycle(t *testing.T) {
 				t.Fatal(err)
 			}
 			// Test for a paused error message.
-			if err := job.FractionProgressed(ctx, jobs.FractionUpdater(0)); !testutils.IsError(err, "cannot update progress on paused job") {
+			if err := job.CheckStatus(ctx); !testutils.IsError(err, "cannot update progress on paused job") {
 				t.Fatalf("unexpected %v", err)
 			}
 		}
@@ -380,7 +379,7 @@ func TestRegistryLifecycle(t *testing.T) {
 		check(t)
 		// Rollback a RESUME.
 		{
-			txn, err := sqlDB.DB.Begin()
+			txn, err := outerDB.Begin()
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -394,7 +393,7 @@ func TestRegistryLifecycle(t *testing.T) {
 		}
 		// Commit a RESUME.
 		{
-			txn, err := sqlDB.DB.Begin()
+			txn, err := outerDB.Begin()
 			if err != nil {
 				t.Fatal(err)
 			}
