@@ -20,9 +20,6 @@ import (
 	"strings"
 	"time"
 
-	opentracing "github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
@@ -31,13 +28,17 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/builtins"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
+	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlutil"
-	"github.com/cockroachdb/cockroach/pkg/storage"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
 const defaultLeniencySetting = 60 * time.Second
@@ -50,13 +51,17 @@ var (
 		"jobs.registry.leniency",
 		"the amount of time to defer any attempts to reschedule a job",
 		defaultLeniencySetting)
+	gcSetting = settings.RegisterDurationSetting(
+		"jobs.retention_time",
+		"the amount of time to retain records for completed jobs before",
+		time.Hour*24*14)
 )
 
 // NodeLiveness is the subset of storage.NodeLiveness's interface needed
 // by Registry.
 type NodeLiveness interface {
-	Self() (*storage.Liveness, error)
-	GetLivenesses() []storage.Liveness
+	Self() (*storagepb.Liveness, error)
+	GetLivenesses() []storagepb.Liveness
 }
 
 // Registry creates Jobs and manages their leases and cancelation.
@@ -230,12 +235,14 @@ func (r *Registry) NewJob(record Record) *Job {
 	}
 	job.mu.payload = jobspb.Payload{
 		Description:   record.Description,
+		Statement:     record.Statement,
 		Username:      record.Username,
 		DescriptorIDs: record.DescriptorIDs,
 		Details:       jobspb.WrapPayloadDetails(record.Details),
 	}
 	job.mu.progress = jobspb.Progress{
-		Details: jobspb.WrapProgressDetails(record.Progress),
+		Details:       jobspb.WrapProgressDetails(record.Progress),
+		RunningStatus: string(record.RunningStatus),
 	}
 	return job
 }
@@ -271,6 +278,10 @@ var DefaultCancelInterval = base.DefaultHeartbeatInterval
 // Registry.Start has been called will not have any effect.
 var DefaultAdoptInterval = 30 * time.Second
 
+// gcInterval is how often we check for and delete job records older than the
+// retention limit.
+const gcInterval = 1 * time.Hour
+
 // Start polls the current node for liveness failures and cancels all registered
 // jobs if it observes a failure.
 func (r *Registry) Start(
@@ -294,6 +305,20 @@ func (r *Registry) Start(
 		}
 	})
 
+	stopper.RunWorker(context.Background(), func(ctx context.Context) {
+		for {
+			select {
+			case <-time.After(gcInterval):
+				old := timeutil.Now().Add(-1 * gcSetting.Get(&r.settings.SV))
+				if err := r.cleanupOldJobs(ctx, old); err != nil {
+					log.Warningf(ctx, "error cleaning up old job records: %v", err)
+				}
+			case <-stopper.ShouldStop():
+				return
+			}
+		}
+	})
+
 	stopper.RunWorker(context.Background(), func(ctx context.Context) {
 		for {
 			select {
@@ -340,6 +365,40 @@ func (r *Registry) maybeCancelJobs(ctx context.Context, nl NodeLiveness) {
 	}
 }
 
+func (r *Registry) cleanupOldJobs(ctx context.Context, olderThan time.Time) error {
+	const stmt = `SELECT id, payload FROM system.jobs WHERE status IN ($1, $2, $3) AND created < $4 ORDER BY created LIMIT 1000`
+	rows, err := r.ex.Query(
+		ctx, "gc-jobs", nil /* txn */, stmt, StatusFailed, StatusSucceeded, StatusCanceled, olderThan,
+	)
+	if err != nil {
+		return err
+	}
+
+	toDelete := tree.NewDArray(types.Int)
+	toDelete.Array = make(tree.Datums, 0, len(rows))
+	oldMicros := timeutil.ToUnixMicros(olderThan)
+	for _, row := range rows {
+		payload, err := UnmarshalPayload(row[1])
+		if err != nil {
+			return err
+		}
+		if payload.FinishedMicros < oldMicros {
+			toDelete.Array = append(toDelete.Array, row[0])
+		}
+	}
+	if len(toDelete.Array) > 0 {
+
+		log.Infof(ctx, "cleaning up %d expired job records", len(toDelete.Array))
+		const stmt = `DELETE FROM system.jobs WHERE id = ANY($1)`
+		if _ /* cols */, err := r.ex.Exec(
+			ctx, "gc-jobs", nil /* txn */, stmt, toDelete,
+		); err != nil {
+			return errors.Wrap(err, "deleting old jobs")
+		}
+	}
+	return nil
+}
+
 // getJobFn attempts to get a resumer from the given job id. If the job id
 // does not have a resumer then it returns an error message suitable for users.
 func (r *Registry) getJobFn(ctx context.Context, txn *client.Txn, id int64) (*Job, Resumer, error) {
@@ -542,7 +601,7 @@ func AddResumeHook(fn ResumeHookFn) {
 
 func (r *Registry) maybeAdoptJob(ctx context.Context, nl NodeLiveness) error {
 	const stmt = `SELECT id, payload, progress IS NULL FROM system.jobs WHERE status IN ($1, $2) ORDER BY created DESC`
-	rows, _ /* cols */, err := r.ex.Query(
+	rows, err := r.ex.Query(
 		ctx, "adopt-job", nil /* txn */, stmt, StatusPending, StatusRunning,
 	)
 	if err != nil {
@@ -605,26 +664,23 @@ func (r *Registry) maybeAdoptJob(ctx context.Context, nl NodeLiveness) error {
 		// has been upgraded to 2.1 then we know nothing is running the job and it
 		// can be safely failed.
 		if nullProgress, ok := row[2].(*tree.DBool); ok && bool(*nullProgress) {
-			// TODO(mjibson): set this to cluster.Version_2_1 when it exists.
-			if r.settings.Version.IsMinSupported(cluster.VersionRangeMerges) {
-				payload.Error = "job predates cluster upgrade and must be re-run"
-				payloadBytes, err := protoutil.Marshal(payload)
-				if err != nil {
-					return err
-				}
+			payload.Error = "job predates cluster upgrade and must be re-run"
+			payloadBytes, err := protoutil.Marshal(payload)
+			if err != nil {
+				return err
+			}
 
-				// We can't use job.update here because it fails while attempting to unmarshal
-				// the progress. Setting the status to failed is idempotent so we don't care
-				// if multiple nodes execute this.
-				const updateStmt = `UPDATE system.jobs SET status = $1, payload = $2 WHERE id = $3`
-				updateArgs := []interface{}{StatusFailed, payloadBytes, *id}
-				err = r.db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
-					_, err := r.ex.Exec(ctx, "job-update", txn, updateStmt, updateArgs...)
-					return err
-				})
-				if err != nil {
-					log.Warningf(ctx, "job %d: has no progress but unable to mark failed: %s", id, err)
-				}
+			// We can't use job.update here because it fails while attempting to unmarshal
+			// the progress. Setting the status to failed is idempotent so we don't care
+			// if multiple nodes execute this.
+			const updateStmt = `UPDATE system.jobs SET status = $1, payload = $2 WHERE id = $3`
+			updateArgs := []interface{}{StatusFailed, payloadBytes, *id}
+			err = r.db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
+				_, err := r.ex.Exec(ctx, "job-update", txn, updateStmt, updateArgs...)
+				return err
+			})
+			if err != nil {
+				log.Warningf(ctx, "job %d: has no progress but unable to mark failed: %s", id, err)
 			}
 			continue
 		}
