@@ -20,11 +20,10 @@ import (
 	"strconv"
 	"strings"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
 )
 
 // PrettyPrintTimeseriesKey is a hook for pretty printing a timeseries key. The
@@ -151,7 +150,7 @@ var (
 			ppFunc: raftLogKeyPrint,
 			psFunc: raftLogKeyParse,
 		},
-		{name: "RaftTruncatedState", suffix: LocalRaftTruncatedStateSuffix},
+		{name: "RaftTruncatedState", suffix: LocalRaftTruncatedStateLegacySuffix},
 		{name: "RaftLastIndex", suffix: LocalRaftLastIndexSuffix},
 		{name: "RangeLastReplicaGCTimestamp", suffix: LocalRangeLastReplicaGCTimestampSuffix},
 		{name: "RangeLastVerificationTimestamp", suffix: LocalRangeLastVerificationTimestampSuffixDeprecated},
@@ -513,13 +512,13 @@ func prettyPrintInternal(valDirs []encoding.Direction, key roachpb.Key, quoteRaw
 	}
 
 	helper := func(key roachpb.Key) (string, bool) {
-		var buf bytes.Buffer
+		var b strings.Builder
 		for _, k := range keyDict {
 			if key.Compare(k.start) >= 0 && (k.end == nil || key.Compare(k.end) <= 0) {
-				buf.WriteString(k.name)
+				b.WriteString(k.name)
 				if k.end != nil && k.end.Compare(key) == 0 {
-					buf.WriteString("/Max")
-					return buf.String(), true
+					b.WriteString("/Max")
+					return b.String(), true
 				}
 
 				hasPrefix := false
@@ -527,20 +526,24 @@ func prettyPrintInternal(valDirs []encoding.Direction, key roachpb.Key, quoteRaw
 					if bytes.HasPrefix(key, e.prefix) {
 						hasPrefix = true
 						key = key[len(e.prefix):]
-						fmt.Fprintf(&buf, "%s%s", e.name, e.ppFunc(valDirs, key))
+						b.WriteString(e.name)
+						b.WriteString(e.ppFunc(valDirs, key))
 						break
 					}
 				}
 				if !hasPrefix {
 					key = key[len(k.start):]
 					if quoteRawKeys {
-						fmt.Fprintf(&buf, "/%q", []byte(key))
-					} else {
-						fmt.Fprintf(&buf, "/%s", []byte(key))
+						b.WriteByte('/')
+						b.WriteByte('"')
+					}
+					b.Write([]byte(key))
+					if quoteRawKeys {
+						b.WriteByte('"')
 					}
 				}
 
-				return buf.String(), true
+				return b.String(), true
 			}
 		}
 
@@ -564,40 +567,9 @@ func prettyPrintInternal(valDirs []encoding.Direction, key roachpb.Key, quoteRaw
 	return str
 }
 
-// PrettyPrint prints the key in a human readable format:
-//
-//   Key format                                        Key value
-//   /Local/...                                        "\x01"+...
-//      /Store/...                                     "\x01s"+...
-//      /RangeID/...                                   "\x01s"+[rangeid]
-//        /[rangeid]/AbortSpan/[id]                    "\x01s"+[rangeid]+"abc-"+[id]
-//        /[rangeid]/Lease                             "\x01s"+[rangeid]+"rfll"
-//        /[rangeid]/RaftTombstone                     "\x01s"+[rangeid]+"rftb"
-//        /[rangeid]/RaftHardState                     "\x01s"+[rangeid]+"rfth"
-//        /[rangeid]/RaftAppliedIndex                  "\x01s"+[rangeid]+"rfta"
-//        /[rangeid]/RaftLog/logIndex:[logIndex]       "\x01s"+[rangeid]+"rftl"+[logIndex]
-//        /[rangeid]/RaftTruncatedState                "\x01s"+[rangeid]+"rftt"
-//        /[rangeid]/RaftLastIndex                     "\x01s"+[rangeid]+"rfti"
-//        /[rangeid]/RangeLastReplicaGCTimestamp       "\x01s"+[rangeid]+"rlrt"
-//        /[rangeid]/RangeLastVerificationTimestamp    "\x01s"+[rangeid]+"rlvt"
-//        /[rangeid]/RangeStats                        "\x01s"+[rangeid]+"stat"
-//      /Range/...                                     "\x01k"+...
-//        [key]/RangeDescriptor                        "\x01k"+[key]+"rdsc"
-//        [key]/Transaction/[id]                       "\x01k"+[key]+"txn-"+[txn-id]
-//        [key]/QueueLastProcessed/[queue]             "\x01k"+[key]+"qlpt"+[queue]
-//   /Local/Max                                        "\x02"
-//
-//   /Meta1/[key]                                      "\x02"+[key]
-//   /Meta2/[key]                                      "\x03"+[key]
-//   /System/...                                       "\x04"
-//      /NodeLiveness/[key]                            "\x04\0x00liveness-"+[key]
-//      /StatusNode/[key]                              "\x04status-node-"+[key]
-//   /System/Max                                       "\x05"
-//
-//   /Table/[key]                                      [key]
-//
-//   /Min                                              ""
-//   /Max                                              "\xff\xff"
+// PrettyPrint prints the key in a human readable format, see TestPrettyPrint.
+// The output does not indicate whether a key is part of the replicated or un-
+// replicated keyspace.
 //
 // valDirs correspond to the encoding direction of each encoded value in key.
 // For example, table keys could have column values encoded in ascending or
@@ -693,28 +665,29 @@ func MassagePrettyPrintedSpanForTest(span string, dirs []encoding.Direction) str
 	var r string
 	colIdx := -1
 	for i := 0; i < len(span); i++ {
-		d := -789
-		fmt.Sscanf(span[i:], "%d", &d)
-		if (dirs != nil) && (d != -789) {
-			// We've managed to consume an int.
-			dir := dirs[colIdx]
-			i += len(strconv.Itoa(d)) - 1
-			x := d
-			if dir == encoding.Descending {
-				x = ^x
-			}
-			r += strconv.Itoa(x)
-		} else {
-			r += string(span[i])
-			switch span[i] {
-			case '/':
-				colIdx++
-			case '-', ' ':
-				// We're switching from the start constraints to the end constraints,
-				// or starting another span.
-				colIdx = -1
+		if dirs != nil {
+			var d int
+			if _, err := fmt.Sscanf(span[i:], "%d", &d); err != nil {
+				// We've managed to consume an int.
+				dir := dirs[colIdx]
+				i += len(strconv.Itoa(d)) - 1
+				x := d
+				if dir == encoding.Descending {
+					x = ^x
+				}
+				r += strconv.Itoa(x)
+				continue
 			}
 		}
+		r += string(span[i])
+		switch span[i] {
+		case '/':
+			colIdx++
+		case '-', ' ':
+			// We're switching from the start constraints to the end constraints,
+			// or starting another span.
+			colIdx = -1
+		}
 	}
 	return r
 }
