@@ -22,9 +22,8 @@ import (
 	"sync"
 
 	"github.com/biogo/store/llrb"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/keys"
+	"github.com/cockroachdb/cockroach/pkg/kv/kvbase"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/util/cache"
@@ -33,6 +32,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil/singleflight"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 // rangeCacheKey is the key type used to store and sort values in the
@@ -89,6 +89,9 @@ type RangeDescriptorCache struct {
 	lookupRequests singleflight.Group
 }
 
+// RangeDescriptorCache implements the kvbase interface.
+var _ kvbase.RangeDescriptorCache = (*RangeDescriptorCache)(nil)
+
 type lookupResult struct {
 	desc       *roachpb.RangeDescriptor
 	evictToken *EvictionToken
@@ -218,7 +221,7 @@ func (et *EvictionToken) EvictAndReplace(
 	return err
 }
 
-// LookupRangeDescriptor attempts to locate a descriptor for the range
+// LookupRangeDescriptorWithEvictionToken attempts to locate a descriptor for the range
 // containing the given Key. This is done by first trying the cache, and then
 // querying the two-level lookup table of range descriptors which cockroach
 // maintains. The function should be provided with an EvictionToken if one was
@@ -235,12 +238,23 @@ func (et *EvictionToken) EvictAndReplace(
 // This method returns the RangeDescriptor for the range containing
 // the key's data and a token to manage evicting the RangeDescriptor
 // if it is found to be stale, or an error if any occurred.
-func (rdc *RangeDescriptorCache) LookupRangeDescriptor(
+func (rdc *RangeDescriptorCache) LookupRangeDescriptorWithEvictionToken(
 	ctx context.Context, key roachpb.RKey, evictToken *EvictionToken, useReverseScan bool,
 ) (*roachpb.RangeDescriptor, *EvictionToken, error) {
 	return rdc.lookupRangeDescriptorInternal(ctx, key, evictToken, useReverseScan, nil)
 }
 
+// LookupRangeDescriptor presents a simpler interface for looking up a
+// RangeDescriptor for a key without the eviction tokens or scan direction
+// control of LookupRangeDescriptorWithEvictionToken. This method is exported
+// to lower level clients through the kvbase.RangeDescriptorCache interface.
+func (rdc *RangeDescriptorCache) LookupRangeDescriptor(
+	ctx context.Context, key roachpb.RKey,
+) (*roachpb.RangeDescriptor, error) {
+	rd, _, err := rdc.lookupRangeDescriptorInternal(ctx, key, nil, false, nil)
+	return rd, err
+}
+
 // lookupRangeDescriptorInternal is called from LookupRangeDescriptor or from tests.
 //
 // If a WaitGroup is supplied, it is signaled when the request is
@@ -356,13 +370,19 @@ func (rdc *RangeDescriptorCache) lookupRangeDescriptorInternal(
 	select {
 	case res = <-resC:
 	case <-ctxDone:
-		return nil, nil, ctx.Err()
+		return nil, nil, errors.Wrap(ctx.Err(), "aborted during range descriptor lookup")
 	}
 
-	if res.Shared {
-		log.Event(ctx, "looked up range descriptor with shared request")
+	var s string
+	if res.Err != nil {
+		s = res.Err.Error()
 	} else {
-		log.Event(ctx, "looked up range descriptor")
+		s = res.Val.(lookupResult).desc.String()
+	}
+	if res.Shared {
+		log.Eventf(ctx, "looked up range descriptor with shared request: %s", s)
+	} else {
+		log.Eventf(ctx, "looked up range descriptor: %s", s)
 	}
 	if res.Err != nil {
 		return nil, nil, res.Err
