@@ -16,17 +16,17 @@ package kv
 
 import (
 	"context"
+	"fmt"
+	"runtime/debug"
 	"sync"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
@@ -37,6 +37,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
 )
 
 const (
@@ -107,6 +108,9 @@ type TxnCoordSender struct {
 		syncutil.Mutex
 
 		txnState txnState
+		// storedErr is set when txnState == txnError. This storedErr is returned to
+		// clients on Send().
+		storedErr *roachpb.Error
 
 		// active is set whenever the transaction has sent any requests.
 		active bool
@@ -139,18 +143,20 @@ type TxnCoordSender struct {
 
 	// An ordered stack of pluggable request interceptors that can transform
 	// batch requests and responses while each maintaining targeted state.
-	// The stack is stored in an array and each txnInterceptor implementation
-	// is embedded in the interceptorAlloc struct, so the entire stack is
-	// allocated together with TxnCoordSender without any additional heap
-	// allocations necessary.
-	interceptorStack [6]txnInterceptor
+	// The stack is stored in a slice backed by the interceptorAlloc.arr and each
+	// txnInterceptor implementation is embedded in the interceptorAlloc struct,
+	// so the entire stack is allocated together with TxnCoordSender without any
+	// additional heap allocations necessary.
+	interceptorStack []txnInterceptor
 	interceptorAlloc struct {
-		txnHeartbeat
+		arr [7]txnInterceptor
+		txnHeartbeater
+		txnSeqNumAllocator
 		txnIntentCollector
 		txnPipeliner
 		txnSpanRefresher
-		txnSeqNumAllocator
-		txnMetrics
+		txnCommitter
+		txnMetricRecorder
 		txnLockGatekeeper // not in interceptorStack array.
 	}
 
@@ -202,14 +208,8 @@ type txnInterceptor interface {
 	epochBumpedLocked()
 
 	// closeLocked closes the interceptor. It is called when the TxnCoordSender
-	// shuts down due to either a txn commit or a txn abort.
-	//
-	// This method can be called multiple times (e.g. if the txn is aborted by the
-	// heartbeat loop and then upon a client rollback); implementers beware.
-	//
-	// Note that EndTransaction(commit=false) requests can still be sent (via the
-	// lockedSender interface) after this is called, and they're expected to be
-	// forwarded along. The idea for this method is to stop background tasks.
+	// shuts down due to either a txn commit or a txn abort. The method will
+	// be called exactly once from cleanupTxnLocked.
 	closeLocked()
 }
 
@@ -246,11 +246,15 @@ type TxnMetrics struct {
 	Restarts *metric.Histogram
 
 	// Counts of restart types.
-	RestartsWriteTooOld       *metric.Counter
-	RestartsDeleteRange       *metric.Counter
-	RestartsSerializable      *metric.Counter
-	RestartsPossibleReplay    *metric.Counter
-	RestartsAsyncWriteFailure *metric.Counter
+	RestartsWriteTooOld           telemetry.CounterWithMetric
+	RestartsWriteTooOldMulti      telemetry.CounterWithMetric
+	RestartsSerializable          telemetry.CounterWithMetric
+	RestartsPossibleReplay        telemetry.CounterWithMetric
+	RestartsAsyncWriteFailure     telemetry.CounterWithMetric
+	RestartsReadWithinUncertainty telemetry.CounterWithMetric
+	RestartsTxnAborted            telemetry.CounterWithMetric
+	RestartsTxnPush               telemetry.CounterWithMetric
+	RestartsUnknown               telemetry.CounterWithMetric
 }
 
 var (
@@ -293,15 +297,32 @@ var (
 		Measurement: "KV Transactions",
 		Unit:        metric.Unit_COUNT,
 	}
+	// There are two ways we can get "write too old" restarts. In both
+	// cases, a WriteTooOldError is generated in the MVCC layer. This is
+	// intercepted on the way out by the Store, which performs a single
+	// retry at a pushed timestamp. If the retry succeeds, the immediate
+	// operation succeeds but the WriteTooOld flag is set on the
+	// Transaction, which causes EndTransaction to return a
+	// TransactionRetryError with RETRY_WRITE_TOO_OLD. These are
+	// captured as txn.restarts.writetooold.
+	//
+	// If the Store's retried operation generates a second
+	// WriteTooOldError (indicating a conflict with a third transaction
+	// with a higher timestamp than the one that caused the first
+	// WriteTooOldError), the store doesn't retry again, and the
+	// WriteTooOldError will be returned up the stack to be retried at
+	// this level. These are captured as txn.restarts.writetoooldmulti.
+	// This path is inefficient, and if it turns out to be common we may
+	// want to do something about it.
 	metaRestartsWriteTooOld = metric.Metadata{
 		Name:        "txn.restarts.writetooold",
 		Help:        "Number of restarts due to a concurrent writer committing first",
 		Measurement: "Restarted Transactions",
 		Unit:        metric.Unit_COUNT,
 	}
-	metaRestartsDeleteRange = metric.Metadata{
-		Name:        "txn.restarts.deleterange",
-		Help:        "Number of restarts due to a forwarded commit timestamp and a DeleteRange command",
+	metaRestartsWriteTooOldMulti = metric.Metadata{
+		Name:        "txn.restarts.writetoooldmulti",
+		Help:        "Number of restarts due to multiple concurrent writers committing first",
 		Measurement: "Restarted Transactions",
 		Unit:        metric.Unit_COUNT,
 	}
@@ -323,23 +344,57 @@ var (
 		Measurement: "Restarted Transactions",
 		Unit:        metric.Unit_COUNT,
 	}
+	metaRestartsReadWithinUncertainty = metric.Metadata{
+		Name:        "txn.restarts.readwithinuncertainty",
+		Help:        "Number of restarts due to reading a new value within the uncertainty interval",
+		Measurement: "Restarted Transactions",
+		Unit:        metric.Unit_COUNT,
+	}
+	metaRestartsTxnAborted = metric.Metadata{
+		Name:        "txn.restarts.txnaborted",
+		Help:        "Number of restarts due to an abort by a concurrent transaction (usually due to deadlock)",
+		Measurement: "Restarted Transactions",
+		Unit:        metric.Unit_COUNT,
+	}
+	// TransactionPushErrors at this level are unusual. They are
+	// normally handled at the Store level with the txnwait and
+	// contention queues. However, they can reach this level and be
+	// retried in tests that disable the store-level retries, and
+	// there may be edge cases that allow them to reach this point in
+	// production.
+	metaRestartsTxnPush = metric.Metadata{
+		Name:        "txn.restarts.txnpush",
+		Help:        "Number of restarts due to a transaction push failure",
+		Measurement: "Restarted Transactions",
+		Unit:        metric.Unit_COUNT,
+	}
+	metaRestartsUnknown = metric.Metadata{
+		Name:        "txn.restarts.unknown",
+		Help:        "Number of restarts due to a unknown reasons",
+		Measurement: "Restarted Transactions",
+		Unit:        metric.Unit_COUNT,
+	}
 )
 
 // MakeTxnMetrics returns a TxnMetrics struct that contains metrics whose
 // windowed portions retain data for approximately histogramWindow.
 func MakeTxnMetrics(histogramWindow time.Duration) TxnMetrics {
 	return TxnMetrics{
-		Aborts:                    metric.NewCounter(metaAbortsRates),
-		Commits:                   metric.NewCounter(metaCommitsRates),
-		Commits1PC:                metric.NewCounter(metaCommits1PCRates),
-		AutoRetries:               metric.NewCounter(metaAutoRetriesRates),
-		Durations:                 metric.NewLatency(metaDurationsHistograms, histogramWindow),
-		Restarts:                  metric.NewHistogram(metaRestartsHistogram, histogramWindow, 100, 3),
-		RestartsWriteTooOld:       metric.NewCounter(metaRestartsWriteTooOld),
-		RestartsDeleteRange:       metric.NewCounter(metaRestartsDeleteRange),
-		RestartsSerializable:      metric.NewCounter(metaRestartsSerializable),
-		RestartsPossibleReplay:    metric.NewCounter(metaRestartsPossibleReplay),
-		RestartsAsyncWriteFailure: metric.NewCounter(metaRestartsAsyncWriteFailure),
+		Aborts:                        metric.NewCounter(metaAbortsRates),
+		Commits:                       metric.NewCounter(metaCommitsRates),
+		Commits1PC:                    metric.NewCounter(metaCommits1PCRates),
+		AutoRetries:                   metric.NewCounter(metaAutoRetriesRates),
+		Durations:                     metric.NewLatency(metaDurationsHistograms, histogramWindow),
+		Restarts:                      metric.NewHistogram(metaRestartsHistogram, histogramWindow, 100, 3),
+		RestartsWriteTooOld:           telemetry.NewCounterWithMetric(metaRestartsWriteTooOld),
+		RestartsWriteTooOldMulti:      telemetry.NewCounterWithMetric(metaRestartsWriteTooOldMulti),
+		RestartsSerializable:          telemetry.NewCounterWithMetric(metaRestartsSerializable),
+		RestartsPossibleReplay:        telemetry.NewCounterWithMetric(metaRestartsPossibleReplay),
+		RestartsAsyncWriteFailure:     telemetry.NewCounterWithMetric(metaRestartsAsyncWriteFailure),
+		RestartsReadWithinUncertainty: telemetry.NewCounterWithMetric(metaRestartsReadWithinUncertainty),
+		RestartsTxnAborted:            telemetry.NewCounterWithMetric(metaRestartsTxnAborted),
+		RestartsTxnPush:               telemetry.NewCounterWithMetric(metaRestartsTxnPush),
+		RestartsUnknown:               telemetry.NewCounterWithMetric(metaRestartsUnknown),
 	}
 }
 
@@ -410,7 +465,7 @@ func (tcf *TxnCoordSenderFactory) TransactionalSender(
 ) client.TxnSender {
 	meta.Txn.AssertInitialized(context.TODO())
 	tcs := &TxnCoordSender{
-		typ: typ,
+		typ:                   typ,
 		TxnCoordSenderFactory: tcf,
 	}
 	tcs.mu.txnState = txnPending
@@ -419,25 +474,34 @@ func (tcf *TxnCoordSenderFactory) TransactionalSender(
 	// this stack are pre-allocated on the TxnCoordSender struct, so this just
 	// initializes the interceptors and pieces them together. It then adds a
 	// txnLockGatekeeper at the bottom of the stack to connect it with the
-	// TxnCoordSender's wrapped sender.
+	// TxnCoordSender's wrapped sender. First, each of the interceptor objects
+	// is initialized.
 	var ri *RangeIterator
 	if ds, ok := tcf.wrapped.(*DistSender); ok {
 		ri = NewRangeIterator(ds)
 	}
-	tcs.interceptorAlloc.txnHeartbeat.init(
-		&tcs.mu.Mutex,
-		&tcs.mu.txn,
-		tcs.clock,
-		tcs.heartbeatInterval,
-		&tcs.interceptorAlloc.txnLockGatekeeper,
-		&tcs.metrics,
-		tcs.stopper,
-		tcs.cleanupTxnLocked,
-	)
-	tcs.interceptorAlloc.txnMetrics.init(&tcs.mu.txn, tcs.clock, &tcs.metrics)
-	tcs.interceptorAlloc.txnIntentCollector = txnIntentCollector{
-		st: tcf.st,
-		ri: ri,
+	// Some interceptors are only needed by roots.
+	if typ == client.RootTxn {
+		tcs.interceptorAlloc.txnHeartbeater.init(
+			&tcs.mu.Mutex,
+			&tcs.mu.txn,
+			tcf.st,
+			tcs.clock,
+			tcs.heartbeatInterval,
+			&tcs.interceptorAlloc.txnLockGatekeeper,
+			&tcs.metrics,
+			tcs.stopper,
+			tcs.cleanupTxnLocked,
+		)
+		tcs.interceptorAlloc.txnIntentCollector = txnIntentCollector{
+			st: tcf.st,
+			ri: ri,
+		}
+		tcs.interceptorAlloc.txnMetricRecorder = txnMetricRecorder{
+			metrics: &tcs.metrics,
+			clock:   tcs.clock,
+			txn:     &tcs.mu.txn,
+		}
 	}
 	tcs.interceptorAlloc.txnPipeliner = txnPipeliner{
 		st: tcf.st,
@@ -454,19 +518,61 @@ func (tcf *TxnCoordSenderFactory) TransactionalSender(
 	}
 	tcs.interceptorAlloc.txnLockGatekeeper = txnLockGatekeeper{
 		wrapped: tcs.wrapped,
-		mu:      &tcs.mu,
+		mu:      &tcs.mu.Mutex,
 	}
-	tcs.interceptorStack = [...]txnInterceptor{
-		&tcs.interceptorAlloc.txnHeartbeat,
-		// The seq num allocator is the below the txnHeartbeat so that it sees the
-		// BeginTransaction prepended by that interceptor. (An alternative would be
-		// to not assign seq nums to BeginTransaction; it doesn't need it.)
-		// Note though that it skips assigning seq nums to heartbeats.
-		&tcs.interceptorAlloc.txnSeqNumAllocator,
-		&tcs.interceptorAlloc.txnIntentCollector,
-		&tcs.interceptorAlloc.txnPipeliner,
-		&tcs.interceptorAlloc.txnSpanRefresher,
-		&tcs.interceptorAlloc.txnMetrics,
+
+	// Once the interceptors are initialized, piece them all together in the
+	// correct order.
+	switch typ {
+	case client.RootTxn:
+		tcs.interceptorAlloc.arr = [...]txnInterceptor{
+			&tcs.interceptorAlloc.txnHeartbeater,
+			// Various interceptors below rely on sequence number allocation,
+			// so the sequence number allocator is near the top of the stack.
+			&tcs.interceptorAlloc.txnSeqNumAllocator,
+			&tcs.interceptorAlloc.txnIntentCollector,
+			// The pipelinger sits above the span refresher because it will
+			// never generate transaction retry errors that could be avoided
+			// with a refresh.
+			&tcs.interceptorAlloc.txnPipeliner,
+			// The span refresher may resend entire batches to avoid transaction
+			// retries. Because of that, we need to be careful which interceptors
+			// sit below it in the stack.
+			&tcs.interceptorAlloc.txnSpanRefresher,
+			// The committer sits beneath the span refresher so that any
+			// retryable errors that it generates have a chance of being
+			// "refreshed away" without the need for a txn restart. Because the
+			// span refresher can re-issue batches, it needs to be careful about
+			// what parts of the batch it mutates. Any mutation needs to be
+			// idempotent and should avoid writing to memory when not changing
+			// it to avoid looking like a data race.
+			&tcs.interceptorAlloc.txnCommitter,
+			// The metrics recorder sits at the bottom of the stack so that it
+			// can observe all transformations performed by other interceptors.
+			&tcs.interceptorAlloc.txnMetricRecorder,
+		}
+		tcs.interceptorStack = tcs.interceptorAlloc.arr[:]
+	case client.LeafTxn:
+		tcs.interceptorAlloc.arr = [cap(tcs.interceptorAlloc.arr)]txnInterceptor{
+			// LeafTxns never perform writes so the sequence number allocator
+			// should never increment its sequence number counter over its
+			// lifetime, but it still plays the important role of assigning each
+			// read request the latest sequence number.
+			&tcs.interceptorAlloc.txnSeqNumAllocator,
+			// The pipeliner is needed on leaves to ensure that in-flight writes
+			// are chained onto by reads that should see them.
+			&tcs.interceptorAlloc.txnPipeliner,
+			// The span refresher was configured above to not actually perform
+			// refreshes for leaves. It is still needed for accumulating the
+			// spans to be reported to the Root. But the gateway doesn't do much
+			// with them; see #24798.
+			&tcs.interceptorAlloc.txnSpanRefresher,
+		}
+		// All other interceptors are absent from a LeafTxn's interceptor stack
+		// because they do not serve a role on leaves.
+		tcs.interceptorStack = tcs.interceptorAlloc.arr[:3]
+	default:
+		panic(fmt.Sprintf("unknown TxnType %v", typ))
 	}
 	for i, reqInt := range tcs.interceptorStack {
 		if i < len(tcs.interceptorStack)-1 {
@@ -476,7 +582,7 @@ func (tcf *TxnCoordSenderFactory) TransactionalSender(
 		}
 	}
 
-	tcs.augmentMetaLocked(meta)
+	tcs.augmentMetaLocked(context.TODO(), meta)
 	return tcs
 }
 
@@ -498,7 +604,7 @@ func (tc *TxnCoordSender) GetMeta(
 	defer tc.mu.Unlock()
 	// Copy mutable state so access is safe for the caller.
 	var meta roachpb.TxnCoordMeta
-	meta.Txn = tc.mu.txn.Clone()
+	meta.Txn = tc.mu.txn
 	for _, reqInt := range tc.interceptorStack {
 		reqInt.populateMetaLocked(&meta)
 	}
@@ -525,10 +631,32 @@ func (tc *TxnCoordSender) AugmentMeta(ctx context.Context, meta roachpb.TxnCoord
 	if tc.mu.txn.ID != meta.Txn.ID {
 		return
 	}
-	tc.augmentMetaLocked(meta)
+
+	// If the TxnCoordMeta is telling us the transaction has been aborted, it's
+	// better if we don't ingest it. Ingesting it would possibly put us in an
+	// inconsistent state, with an ABORTED proto but with the heartbeat loop still
+	// running. When this TxnCoordMeta was received from DistSQL, it presumably
+	// followed a TxnAbortedError that was also received. If that error was also
+	// passed to us, then we've already aborted the txn and ingesting the meta
+	// would be OK. However, as it stands, if the TxnAbortedError followed a
+	// non-retriable error, than we don't get the aborted error (in fact, we don't
+	// get either of the errors; the client is responsible for rolling back).
+	// TODO(andrei): A better design would be to abort the txn as soon as any
+	// error is received from DistSQL, which would eliminate qualms about what
+	// error comes first.
+	if meta.Txn.Status != roachpb.PENDING {
+		return
+	}
+
+	tc.augmentMetaLocked(ctx, meta)
 }
 
-func (tc *TxnCoordSender) augmentMetaLocked(meta roachpb.TxnCoordMeta) {
+func (tc *TxnCoordSender) augmentMetaLocked(ctx context.Context, meta roachpb.TxnCoordMeta) {
+	if meta.Txn.Status != roachpb.PENDING {
+		// Non-pending transactions should only come in errors, which are not
+		// handled by this method.
+		log.Fatalf(ctx, "unexpected non-pending txn in augmentMetaLocked: %s", meta.Txn)
+	}
 	tc.mu.txn.Update(&meta.Txn)
 	for _, reqInt := range tc.interceptorStack {
 		reqInt.augmentMetaLocked(meta)
@@ -553,18 +681,38 @@ func (tc *TxnCoordSender) DisablePipelining() error {
 	return nil
 }
 
+func generateTxnDeadlineExceededErr(
+	txn *roachpb.Transaction, deadline hlc.Timestamp,
+) *roachpb.Error {
+	exceededBy := txn.Timestamp.GoTime().Sub(deadline.GoTime())
+	fromStart := txn.Timestamp.GoTime().Sub(txn.OrigTimestamp.GoTime())
+	extraMsg := fmt.Sprintf(
+		"txn timestamp pushed too much; deadline exceeded by %s (%s > %s), "+
+			"original timestamp %s ago (%s)",
+		exceededBy, txn.Timestamp, deadline, fromStart, txn.OrigTimestamp)
+	return roachpb.NewErrorWithTxn(
+		roachpb.NewTransactionRetryError(roachpb.RETRY_COMMIT_DEADLINE_EXCEEDED, extraMsg), txn)
+}
+
 // commitReadOnlyTxnLocked "commits" a read-only txn. It is equivalent, but
 // cheaper than, sending an EndTransactionRequest. A read-only txn doesn't have
 // a transaction record, so there's no need to send any request to the server.
-// An EndTransactionRequest for a read-only txn is elided by the txnHeartbeat
+// An EndTransactionRequest for a read-only txn is elided by the txnCommitter
 // interceptor. However, calling this and short-circuting even earlier is
 // even more efficient (and shows in benchmarks).
+// TODO(nvanbenschoten): we could have this call into txnCommitter's
+// sendLockedWithElidedEndTransaction method, but we would want to confirm
+// that doing so doesn't cut into the speed-up we see from this fast-path.
 func (tc *TxnCoordSender) commitReadOnlyTxnLocked(
-	ctx context.Context, deadline *hlc.Timestamp,
+	ctx context.Context, ba roachpb.BatchRequest,
 ) *roachpb.Error {
-	if deadline != nil && deadline.Less(tc.mu.txn.Timestamp) {
-		return roachpb.NewError(
-			roachpb.NewTransactionStatusError("deadline exceeded before transaction finalization"))
+	deadline := ba.Requests[0].GetEndTransaction().Deadline
+	txn := tc.mu.txn
+	if deadline != nil && !txn.Timestamp.Less(*deadline) {
+		pErr := generateTxnDeadlineExceededErr(&txn, *deadline)
+		// We need to bump the epoch and transform this retriable error.
+		ba.Txn = &txn
+		return tc.updateStateLocked(ctx, ba, nil /* br */, pErr)
 	}
 	tc.mu.txnState = txnFinalized
 	// Mark the transaction as committed so that, in case this commit is done by
@@ -591,8 +739,8 @@ func (tc *TxnCoordSender) Send(
 		return nil, pErr
 	}
 
-	if ba.IsSingleEndTransactionRequest() && !tc.interceptorAlloc.txnHeartbeat.mu.everSentBeginTxn {
-		return nil, tc.commitReadOnlyTxnLocked(ctx, ba.Requests[0].GetEndTransaction().Deadline)
+	if ba.IsSingleEndTransactionRequest() && !tc.interceptorAlloc.txnIntentCollector.haveIntents() {
+		return nil, tc.commitReadOnlyTxnLocked(ctx, ba)
 	}
 
 	startNs := tc.clock.PhysicalNow()
@@ -641,13 +789,12 @@ func (tc *TxnCoordSender) Send(
 	}
 	// Clone the Txn's Proto so that future modifications can be made without
 	// worrying about synchronization.
-	newTxn := tc.mu.txn.Clone()
-	ba.Txn = &newTxn
+	ba.Txn = tc.mu.txn.Clone()
 
 	// Send the command through the txnInterceptor stack.
 	br, pErr := tc.interceptorStack[0].SendLocked(ctx, ba)
 
-	pErr = tc.updateStateLocked(ctx, startNs, ba, br, pErr)
+	pErr = tc.updateStateLocked(ctx, ba, br, pErr)
 
 	// If we succeeded to commit, or we attempted to rollback, we move to
 	// txnFinalized.
@@ -666,28 +813,7 @@ func (tc *TxnCoordSender) Send(
 		}
 	}
 
-	// Move to the error state on non-retriable errors.
 	if pErr != nil {
-		log.VEventf(ctx, 2, "failed batch: %s", pErr)
-		var retriable bool
-		// Note that unhandled retryable txn errors are allowed from leaf
-		// transactions. We pass them up through distributed SQL flows to
-		// the root transactions, at the receiver.
-		if pErr.TransactionRestart != roachpb.TransactionRestart_NONE {
-			retriable = true
-			if tc.typ == client.RootTxn {
-				log.Fatalf(ctx,
-					"unexpected retryable error at the client.Txn level: (%T) %s",
-					pErr.GetDetail(), pErr)
-			}
-		} else if _, ok := pErr.GetDetail().(*roachpb.HandledRetryableTxnError); ok {
-			retriable = true
-		}
-
-		if !retriable {
-			tc.mu.txnState = txnError
-		}
-
 		return nil, pErr
 	}
 
@@ -747,20 +873,21 @@ func (tc *TxnCoordSender) maybeRejectClientLocked(
 	}
 
 	if tc.mu.txnState == txnFinalized {
-		return roachpb.NewErrorWithTxn(
-			roachpb.NewTransactionStatusError(
-				"client already committed or rolled back the transaction"),
-			&tc.mu.txn)
+		msg := fmt.Sprintf("client already committed or rolled back the transaction. "+
+			"Trying to execute: %s", ba)
+		stack := string(debug.Stack())
+		log.Errorf(ctx, "%s. stack:\n%s", msg, stack)
+		return roachpb.NewErrorWithTxn(roachpb.NewTransactionStatusError(msg), &tc.mu.txn)
 	}
 	if tc.mu.txnState == txnError {
-		return roachpb.NewError(&roachpb.TxnAlreadyEncounteredErrorError{})
+		return tc.mu.storedErr
 	}
 	if tc.mu.txn.Status == roachpb.ABORTED {
 		abortedErr := roachpb.NewErrorWithTxn(
 			roachpb.NewTransactionAbortedError(roachpb.ABORT_REASON_CLIENT_REJECT), &tc.mu.txn)
 		if tc.typ == client.LeafTxn {
 			// Leaf txns return raw retriable errors (which get handled by the
-			// root) rather than HandledRetryableTxnError.
+			// root) rather than TransactionRetryWithProtoRefreshError.
 			return abortedErr
 		}
 		newTxn := roachpb.PrepareTransactionForRetry(
@@ -768,7 +895,7 @@ func (tc *TxnCoordSender) maybeRejectClientLocked(
 			// priority is not used for aborted errors
 			roachpb.NormalUserPriority,
 			tc.clock)
-		return roachpb.NewError(roachpb.NewHandledRetryableTxnError(
+		return roachpb.NewError(roachpb.NewTransactionRetryWithProtoRefreshError(
 			abortedErr.Message, tc.mu.txn.ID, newTxn))
 	}
 
@@ -804,42 +931,65 @@ func (tc *TxnCoordSender) UpdateStateOnRemoteRetryableErr(
 ) *roachpb.Error {
 	tc.mu.Lock()
 	defer tc.mu.Unlock()
+	txnID := tc.mu.txn.ID
 	err := tc.handleRetryableErrLocked(ctx, pErr)
-	tc.mu.txn.Update(&err.Transaction)
+	// We'll update our txn, unless this was an abort error. If it was an abort
+	// error, the transaction has been rolled back and the state was updated in
+	// handleRetryableErrLocked().
+	if err.Transaction.ID == txnID {
+		// This is where we get a new epoch.
+		tc.mu.txn.Update(&err.Transaction)
+	}
 	return roachpb.NewError(err)
 }
 
 // handleRetryableErrLocked takes a retriable error and creates a
-// HandledRetryableError containing the transaction that needs to be used by the
+// TransactionRetryWithProtoRefreshError containing the transaction that needs to be used by the
 // next attempt. It also handles various aspects of updating the
 // TxnCoordSender's state, but notably it does not update its proto: the caller
 // needs to call tc.mu.txn.Update(pErr.GetTxn()).
 func (tc *TxnCoordSender) handleRetryableErrLocked(
 	ctx context.Context, pErr *roachpb.Error,
-) *roachpb.HandledRetryableTxnError {
+) *roachpb.TransactionRetryWithProtoRefreshError {
 	// If the error is a transaction retry error, update metrics to
-	// reflect the reason for the restart.
-	// TODO(spencer): this code path does not account for retry errors
-	//   experienced by dist sql (see internal/client/txn.go).
-	if tErr, ok := pErr.GetDetail().(*roachpb.TransactionRetryError); ok {
+	// reflect the reason for the restart. More details about the
+	// different error types are documented above on the metaRestart
+	// variables.
+	switch tErr := pErr.GetDetail().(type) {
+	case *roachpb.TransactionRetryError:
 		switch tErr.Reason {
 		case roachpb.RETRY_WRITE_TOO_OLD:
-			tc.metrics.RestartsWriteTooOld.Inc(1)
-		case roachpb.RETRY_DELETE_RANGE:
-			tc.metrics.RestartsDeleteRange.Inc(1)
+			tc.metrics.RestartsWriteTooOld.Inc()
 		case roachpb.RETRY_SERIALIZABLE:
-			tc.metrics.RestartsSerializable.Inc(1)
+			tc.metrics.RestartsSerializable.Inc()
 		case roachpb.RETRY_POSSIBLE_REPLAY:
-			tc.metrics.RestartsPossibleReplay.Inc(1)
+			tc.metrics.RestartsPossibleReplay.Inc()
 		case roachpb.RETRY_ASYNC_WRITE_FAILURE:
-			tc.metrics.RestartsAsyncWriteFailure.Inc(1)
+			tc.metrics.RestartsAsyncWriteFailure.Inc()
+		default:
+			tc.metrics.RestartsUnknown.Inc()
 		}
+
+	case *roachpb.WriteTooOldError:
+		tc.metrics.RestartsWriteTooOldMulti.Inc()
+
+	case *roachpb.ReadWithinUncertaintyIntervalError:
+		tc.metrics.RestartsReadWithinUncertainty.Inc()
+
+	case *roachpb.TransactionAbortedError:
+		tc.metrics.RestartsTxnAborted.Inc()
+
+	case *roachpb.TransactionPushError:
+		tc.metrics.RestartsTxnPush.Inc()
+
+	default:
+		tc.metrics.RestartsUnknown.Inc()
 	}
 	errTxnID := pErr.GetTxn().ID
 	newTxn := roachpb.PrepareTransactionForRetry(ctx, pErr, tc.mu.userPriority, tc.clock)
 
-	// We'll pass a HandledRetryableTxnError up to the next layer.
-	retErr := roachpb.NewHandledRetryableTxnError(
+	// We'll pass a TransactionRetryWithProtoRefreshError up to the next layer.
+	retErr := roachpb.NewTransactionRetryWithProtoRefreshError(
 		pErr.Message,
 		errTxnID, // the id of the transaction that encountered the error
 		newTxn)
@@ -853,7 +1003,7 @@ func (tc *TxnCoordSender) handleRetryableErrLocked(
 		tc.mu.txn.Status = roachpb.ABORTED
 		// Abort the old txn. The client is not supposed to use use this
 		// TxnCoordSender any more.
-		tc.interceptorAlloc.txnHeartbeat.abortTxnAsyncLocked(ctx)
+		tc.interceptorAlloc.txnHeartbeater.abortTxnAsyncLocked(ctx)
 		return retErr
 	}
 
@@ -869,59 +1019,79 @@ func (tc *TxnCoordSender) handleRetryableErrLocked(
 // updateStateLocked updates the transaction state in both the success and error
 // cases. It also updates retryable errors with the updated transaction for use
 // by client restarts.
-//
-// startNS is the time when the request that's updating the state has been sent.
-// This is not used if the request is known to not be the one in charge of
-// starting tracking the transaction - i.e. this is the case for DistSQL, which
-// just does reads and passes 0.
 func (tc *TxnCoordSender) updateStateLocked(
-	ctx context.Context,
-	startNS int64,
-	ba roachpb.BatchRequest,
-	br *roachpb.BatchResponse,
-	pErr *roachpb.Error,
+	ctx context.Context, ba roachpb.BatchRequest, br *roachpb.BatchResponse, pErr *roachpb.Error,
 ) *roachpb.Error {
-	txnID := ba.Txn.ID
-	var responseTxn *roachpb.Transaction
+
+	// We handle a couple of different cases:
+	// 1) A successful response. If that response carries a transaction proto,
+	// we'll use it to update our proto.
+	// 2) A non-retriable error. We move to the txnError state and we cleanup. If
+	// the error carries a transaction in it, we update our proto with it
+	// (although Andrei doesn't know if that serves any purpose).
+	// 3) A retriable error. We "handle" it, in the sense that we call
+	// handleRetryableErrLocked() to transform the error. If the error instructs
+	// the client to start a new transaction (i.e. TransactionAbortedError), then
+	// the current transaction is automatically rolled-back. Otherwise, we update
+	// our proto for a new epoch.
+	// NOTE: We'd love to move to state txnError in case of new error but alas
+	// with the current interface we can't: there's no way for the client to ack
+	// the receipt of the error and control the switching to the new epoch. This
+	// is a major problem of the current txn interface - it means that concurrent
+	// users of a txn might operate at the wrong epoch if they race with the
+	// receipt of such an error.
+
 	if pErr == nil {
-		responseTxn = br.Txn
-	} else {
-		// Only handle transaction retry errors if this is a root transaction.
-		if pErr.TransactionRestart != roachpb.TransactionRestart_NONE &&
-			tc.typ == client.RootTxn {
-
-			errTxnID := pErr.GetTxn().ID // The ID of the txn that needs to be restarted.
-			if errTxnID != txnID {
-				// KV should not return errors for transactions other than the one in
-				// the BatchRequest.
-				log.Fatalf(ctx, "retryable error for the wrong txn. ba.Txn: %s. pErr: %s",
-					ba.Txn, pErr)
-			}
-
-			err := tc.handleRetryableErrLocked(ctx, pErr)
-			if err.Transaction.ID == ba.Txn.ID {
-				// We'll update our txn, unless this was an abort error.
-				cp := err.Transaction.Clone()
-				responseTxn = &cp
-			}
-			pErr = roachpb.NewError(err)
-		} else {
-			// We got a non-retryable error, or a retryable error at a leaf
-			// transaction, and need to pass responsibility for handling it
-			// up to the root transaction.
-
-			if errTxn := pErr.GetTxn(); errTxn != nil {
-				responseTxn = errTxn
-			}
-		}
+		tc.mu.txn.Update(br.Txn)
+		return nil
 	}
 
-	// Update our record of this transaction, even on error.
-	// Note that multiple retriable errors for the same epoch might arrive; also
-	// we might get retriable errors for old epochs. We rely on the associativity
-	// of Transaction.Update to sort out this lack of ordering guarantee.
-	if responseTxn != nil {
-		tc.mu.txn.Update(responseTxn)
+	if pErr.TransactionRestart != roachpb.TransactionRestart_NONE {
+		if tc.typ == client.LeafTxn {
+			// Leaves handle retriable errors differently than roots. The leaf
+			// transaction is not supposed to be used any more after a retriable
+			// error. Separately, the error needs to make its way back to the root.
+
+			// From now on, clients will get this error whenever they Send(). We want
+			// clients to get the same retriable error so we don't wrap it in
+			// TxnAlreadyEncounteredErrorError as we do elsewhere.
+			tc.mu.txnState = txnError
+			tc.mu.storedErr = pErr
+
+			// Cleanup.
+			tc.mu.txn.Update(pErr.GetTxn())
+			tc.cleanupTxnLocked(ctx)
+			return pErr
+		}
+
+		txnID := ba.Txn.ID
+		errTxnID := pErr.GetTxn().ID // The ID of the txn that needs to be restarted.
+		if errTxnID != txnID {
+			// KV should not return errors for transactions other than the one in
+			// the BatchRequest.
+			log.Fatalf(ctx, "retryable error for the wrong txn. ba.Txn: %s. pErr: %s",
+				ba.Txn, pErr)
+		}
+		err := tc.handleRetryableErrLocked(ctx, pErr)
+		// We'll update our txn, unless this was an abort error. If it was an abort
+		// error, the transaction has been rolled back and the state was updated in
+		// handleRetryableErrLocked().
+		if err.Transaction.ID == ba.Txn.ID {
+			// This is where we get a new epoch.
+			tc.mu.txn.Update(&err.Transaction)
+		}
+		return roachpb.NewError(err)
+	}
+
+	// This is the non-retriable error case.
+	if errTxn := pErr.GetTxn(); errTxn != nil {
+		tc.mu.txnState = txnError
+		tc.mu.storedErr = roachpb.NewError(&roachpb.TxnAlreadyEncounteredErrorError{
+			PrevError: pErr.String(),
+		})
+		// Cleanup.
+		tc.mu.txn.Update(errTxn)
+		tc.cleanupTxnLocked(ctx)
 	}
 	return pErr
 }
@@ -986,21 +1156,6 @@ func (tc *TxnCoordSender) SetDebugName(name string) {
 	tc.mu.txn.Name = name
 }
 
-// SetIsolation is part of the client.TxnSender interface.
-func (tc *TxnCoordSender) SetIsolation(isolation enginepb.IsolationType) error {
-	tc.mu.Lock()
-	defer tc.mu.Unlock()
-
-	if tc.mu.txn.Isolation == isolation {
-		return nil
-	}
-	if tc.mu.active {
-		return errors.Errorf("cannot change the isolation level of a running transaction")
-	}
-	tc.mu.txn.Isolation = isolation
-	return nil
-}
-
 // OrigTimestamp is part of the client.TxnSender interface.
 func (tc *TxnCoordSender) OrigTimestamp() hlc.Timestamp {
 	tc.mu.Lock()
@@ -1070,8 +1225,7 @@ func (tc *TxnCoordSender) IsSerializablePushAndRefreshNotPossible() bool {
 		tc.mu.txn.OrigTimestampWasObserved
 	// We check OrigTimestampWasObserved here because, if that's set, refreshing
 	// of reads is not performed.
-	return tc.mu.txn.Isolation == enginepb.SERIALIZABLE &&
-		isTxnPushed && refreshAttemptNotPossible
+	return isTxnPushed && refreshAttemptNotPossible
 }
 
 // Epoch is part of the client.TxnSender interface.
@@ -1083,6 +1237,5 @@ func (tc *TxnCoordSender) Epoch() uint32 {
 func (tc *TxnCoordSender) SerializeTxn() *roachpb.Transaction {
 	tc.mu.Lock()
 	defer tc.mu.Unlock()
-	cpy := tc.mu.txn.Clone()
-	return &cpy
+	return tc.mu.txn.Clone()
 }
