@@ -24,12 +24,9 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
-	"github.com/stretchr/testify/require"
-	"golang.org/x/sync/errgroup"
-
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
@@ -42,7 +39,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
-	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/require"
+	"golang.org/x/sync/errgroup"
 )
 
 // createTestDB creates a local test server and starts it. The caller
@@ -74,10 +73,12 @@ func makeTS(walltime int64, logical int32) hlc.Timestamp {
 func (tc *TxnCoordSender) isTracking() bool {
 	tc.mu.Lock()
 	defer tc.mu.Unlock()
-	return tc.interceptorAlloc.txnHeartbeat.mu.txnEnd != nil
+	return tc.interceptorAlloc.txnHeartbeater.mu.txnEnd != nil
 }
 
-// Test that the Transaction.Writing flag is set after performing any writes.
+// Test that the Transaction.DeprecatedWriting flag is set after performing any
+// writes.
+// TODO(nvanbenschoten): Remove this in 19.2.
 func TestTxnCoordSenderSetWritingFlag(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	s := createTestDB(t)
@@ -88,7 +89,7 @@ func TestTxnCoordSenderSetWritingFlag(t *testing.T) {
 	if err := txn.Put(ctx, roachpb.Key("a"), []byte("value")); err != nil {
 		t.Fatal(err)
 	}
-	if !txn.Serialize().Writing {
+	if !txn.Serialize().DeprecatedWriting {
 		t.Fatal("txn is not marked as writing")
 	}
 }
@@ -107,9 +108,6 @@ func TestTxnCoordSenderBeginTransaction(t *testing.T) {
 	key := roachpb.Key("key")
 	txn.InternalSetPriority(10)
 	txn.SetDebugName("test txn")
-	if err := txn.SetIsolation(enginepb.SNAPSHOT); err != nil {
-		t.Fatal(err)
-	}
 	if err := txn.Put(ctx, key, []byte("value")); err != nil {
 		t.Fatal(err)
 	}
@@ -123,9 +121,6 @@ func TestTxnCoordSenderBeginTransaction(t *testing.T) {
 	if !bytes.Equal(proto.Key, key) {
 		t.Errorf("expected txn Key to match %q != %q", key, proto.Key)
 	}
-	if proto.Isolation != enginepb.SNAPSHOT {
-		t.Errorf("expected txn isolation to be SNAPSHOT; got %s", proto.Isolation)
-	}
 }
 
 // TestTxnCoordSenderKeyRanges verifies that multiple requests to same or
@@ -454,16 +449,13 @@ func TestTxnCoordSenderEndTxn(t *testing.T) {
 	for i := 0; i < 4; i++ {
 		key := roachpb.Key("key: " + strconv.Itoa(i))
 		txn := client.NewTxn(ctx, s.DB, 0 /* gatewayNodeID */, client.RootTxn)
-		// Set to SNAPSHOT so that it can be pushed without restarting.
-		if err := txn.SetIsolation(enginepb.SNAPSHOT); err != nil {
-			t.Fatal(err)
-		}
 		// Initialize the transaction.
 		if pErr := txn.Put(ctx, key, []byte("value")); pErr != nil {
 			t.Fatal(pErr)
 		}
 		// Conflicting transaction that pushes the above transaction.
 		conflictTxn := client.NewTxn(ctx, s.DB, 0 /* gatewayNodeID */, client.RootTxn)
+		conflictTxn.InternalSetPriority(roachpb.MaxTxnPriority)
 		if _, pErr := conflictTxn.Get(ctx, key); pErr != nil {
 			t.Fatal(pErr)
 		}
@@ -510,17 +502,15 @@ func TestTxnCoordSenderEndTxn(t *testing.T) {
 				if err != nil {
 					t.Fatal(err)
 				}
+
 			case 1:
 				// Past deadline.
-				if statusError, ok := err.(*roachpb.TransactionStatusError); !ok {
-					t.Fatalf("expected TransactionStatusError but got %T: %s", err, err)
-				} else if expected := "transaction deadline exceeded"; statusError.Msg != expected {
-					t.Fatalf("expected %s, got %s", expected, statusError.Msg)
-				}
+				fallthrough
 			case 2:
 				// Equal deadline.
-				if err != nil {
-					t.Fatal(err)
+				assertTransactionRetryError(t, err)
+				if !testutils.IsError(err, "RETRY_COMMIT_DEADLINE_EXCEEDED") {
+					t.Fatalf("expected deadline exceeded, got: %s", err)
 				}
 			case 3:
 				// Future deadline.
@@ -573,7 +563,7 @@ func TestTxnCoordSenderAddIntentOnError(t *testing.T) {
 }
 
 func assertTransactionRetryError(t *testing.T, e error) {
-	if retErr, ok := e.(*roachpb.HandledRetryableTxnError); ok {
+	if retErr, ok := e.(*roachpb.TransactionRetryWithProtoRefreshError); ok {
 		if !testutils.IsError(retErr, "TransactionRetryError") {
 			t.Fatalf("expected the cause to be TransactionRetryError, but got %s",
 				retErr)
@@ -584,7 +574,7 @@ func assertTransactionRetryError(t *testing.T, e error) {
 }
 
 func assertTransactionAbortedError(t *testing.T, e error) {
-	if retErr, ok := e.(*roachpb.HandledRetryableTxnError); ok {
+	if retErr, ok := e.(*roachpb.TransactionRetryWithProtoRefreshError); ok {
 		if !testutils.IsError(retErr, "TransactionAbortedError") {
 			t.Fatalf("expected the cause to be TransactionAbortedError, but got %s",
 				retErr)
@@ -687,7 +677,7 @@ func TestTxnCoordSenderTxnUpdatedOnError(t *testing.T) {
 	ctx := context.Background()
 	origTS := makeTS(123, 0)
 	plus10 := origTS.Add(10, 10)
-	plus20 := plus10.Add(10, 0)
+	plus20 := origTS.Add(20, 0)
 	testCases := []struct {
 		// The test's name.
 		name             string
@@ -789,6 +779,10 @@ func TestTxnCoordSenderTxnUpdatedOnError(t *testing.T) {
 				if pErr == nil {
 					reply = ba.CreateReply()
 					reply.Txn = ba.Txn
+				} else if txn := pErr.GetTxn(); txn != nil {
+					// Update the manual clock to simulate an
+					// error updating a local hlc clock.
+					manual.Set(txn.Timestamp.WallTime)
 				}
 				return reply, pErr
 			}
@@ -807,7 +801,6 @@ func TestTxnCoordSenderTxnUpdatedOnError(t *testing.T) {
 				"test txn",
 				key,
 				roachpb.UserPriority(0),
-				enginepb.SERIALIZABLE,
 				clock.Now(),
 				clock.MaxOffset().Nanoseconds(),
 			)
@@ -901,67 +894,6 @@ func TestTxnMultipleCoord(t *testing.T) {
 	}
 }
 
-// TestTxnCoordSenderErrorWithIntent validates that if a transactional request
-// returns an error but also indicates a Writing transaction, the coordinator
-// tracks it just like a successful request.
-//
-// Note(andrei): This test was written at a time when the Writing status
-// returned by the server mattered for the client. As of June 2018, that's no
-// longer the case. The test doesn't hurt, though.
-func TestTxnCoordSenderErrorWithIntent(t *testing.T) {
-	defer leaktest.AfterTest(t)()
-	stopper := stop.NewStopper()
-	defer stopper.Stop(context.TODO())
-	manual := hlc.NewManualClock(123)
-	clock := hlc.NewClock(manual.UnixNano, 20*time.Nanosecond)
-
-	testCases := []struct {
-		roachpb.Error
-		errMsg string
-	}{
-		{*roachpb.NewError(roachpb.NewTransactionRetryError(roachpb.RETRY_REASON_UNKNOWN)), "retry txn"},
-		{
-			*roachpb.NewError(roachpb.NewTransactionPushError(roachpb.Transaction{
-				TxnMeta: enginepb.TxnMeta{ID: uuid.MakeV4()}}),
-			), "failed to push",
-		},
-		{*roachpb.NewErrorf("testError"), "testError"},
-	}
-	for i, test := range testCases {
-		t.Run("", func(t *testing.T) {
-			var senderFn client.SenderFunc = func(_ context.Context, ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
-				txn := ba.Txn.Clone()
-				txn.Writing = true
-				pErr := &roachpb.Error{}
-				*pErr = test.Error
-				pErr.SetTxn(&txn)
-				return nil, pErr
-			}
-			factory := NewTxnCoordSenderFactory(
-				TxnCoordSenderFactoryConfig{
-					AmbientCtx: log.AmbientContext{Tracer: tracing.NewTracer()},
-					Clock:      clock,
-					Stopper:    stopper,
-				},
-				senderFn,
-			)
-
-			var ba roachpb.BatchRequest
-			key := roachpb.Key("test")
-			ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: key}})
-			ba.Add(&roachpb.EndTransactionRequest{})
-			txn := roachpb.MakeTransaction("test", key, 0, 0, clock.Now(), 0)
-			meta := roachpb.MakeTxnCoordMeta(txn)
-			tc := factory.TransactionalSender(client.RootTxn, meta)
-			ba.Txn = &txn
-			_, pErr := tc.Send(context.Background(), ba)
-			if !testutils.IsPError(pErr, test.errMsg) {
-				t.Errorf("%d: error did not match %s: %v", i, test.errMsg, pErr)
-			}
-		})
-	}
-}
-
 // TestTxnCoordSenderNoDuplicateIntents verifies that TxnCoordSender does not
 // generate duplicate intents and that it merges intents for overlapping ranges.
 func TestTxnCoordSenderNoDuplicateIntents(t *testing.T) {
@@ -982,9 +914,7 @@ func TestTxnCoordSenderNoDuplicateIntents(t *testing.T) {
 			}
 		}
 		br := ba.CreateReply()
-		txnClone := ba.Txn.Clone()
-		br.Txn = &txnClone
-		br.Txn.Writing = true
+		br.Txn = ba.Txn.Clone()
 		return br, nil
 	}
 	ambient := log.AmbientContext{Tracer: tracing.NewTracer()}
@@ -1040,21 +970,15 @@ func TestTxnCoordSenderNoDuplicateIntents(t *testing.T) {
 // values. This is done through a series of retries with increasing backoffs, to work around
 // the TxnCoordSender's asynchronous updating of metrics after a transaction ends.
 func checkTxnMetrics(
-	t *testing.T,
-	metrics TxnMetrics,
-	name string,
-	commits, commits1PC, abandons, aborts, restarts int64,
+	t *testing.T, metrics TxnMetrics, name string, commits, commits1PC, aborts, restarts int64,
 ) {
 	testutils.SucceedsSoon(t, func() error {
-		return checkTxnMetricsOnce(t, metrics, name, commits, commits1PC, abandons, aborts, restarts)
+		return checkTxnMetricsOnce(t, metrics, name, commits, commits1PC, aborts, restarts)
 	})
 }
 
 func checkTxnMetricsOnce(
-	t *testing.T,
-	metrics TxnMetrics,
-	name string,
-	commits, commits1PC, abandons, aborts, restarts int64,
+	t *testing.T, metrics TxnMetrics, name string, commits, commits1PC, aborts, restarts int64,
 ) error {
 	testcases := []struct {
 		name string
@@ -1063,8 +987,7 @@ func checkTxnMetricsOnce(
 		{"commits", metrics.Commits.Count(), commits},
 		{"commits1PC", metrics.Commits1PC.Count(), commits1PC},
 		{"aborts", metrics.Aborts.Count(), aborts},
-		{"durations", metrics.Durations.TotalCount(),
-			commits + abandons + aborts},
+		{"durations", metrics.Durations.TotalCount(), commits + aborts},
 	}
 
 	for _, tc := range testcases {
@@ -1096,7 +1019,15 @@ func checkTxnMetricsOnce(
 // have a faster sample interval and returns a cleanup function to be
 // executed by callers.
 func setupMetricsTest(t *testing.T) (*localtestcluster.LocalTestCluster, TxnMetrics, func()) {
-	s := createTestDB(t)
+	dbCtx := client.DefaultDBContext()
+	s := &localtestcluster.LocalTestCluster{
+		DBContext: &dbCtx,
+		// Liveness heartbeat txns mess up the metrics.
+		DisableLivenessHeartbeat: true,
+		DontCreateSystemRanges:   true,
+	}
+	s.Start(t, testutils.NewNodeTestBaseContext(), InitFactoryForLocalTestCluster)
+
 	metrics := MakeTxnMetrics(metric.TestSampleInterval)
 	s.DB.GetFactory().(*TxnCoordSenderFactory).metrics = metrics
 	return s, metrics, func() {
@@ -1120,7 +1051,7 @@ func TestTxnCommit(t *testing.T) {
 	}); err != nil {
 		t.Fatal(err)
 	}
-	checkTxnMetrics(t, metrics, "commit txn", 1 /* commits */, 0 /* commits1PC */, 0, 0, 0)
+	checkTxnMetrics(t, metrics, "commit txn", 1 /* commits */, 0 /* commits1PC */, 0, 0)
 
 	// Test a read-only txn.
 	if err := s.DB.Txn(context.TODO(), func(ctx context.Context, txn *client.Txn) error {
@@ -1131,7 +1062,7 @@ func TestTxnCommit(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	checkTxnMetrics(t, metrics, "commit txn", 2 /* commits */, 0 /* commits1PC */, 0, 0, 0)
+	checkTxnMetrics(t, metrics, "commit txn", 2 /* commits */, 0 /* commits1PC */, 0, 0)
 }
 
 // TestTxnOnePhaseCommit verifies that 1PC metric tracking works.
@@ -1139,6 +1070,7 @@ func TestTxnOnePhaseCommit(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	s, metrics, cleanupFn := setupMetricsTest(t)
 	defer cleanupFn()
+
 	value := []byte("value")
 
 	ctx := context.TODO()
@@ -1164,7 +1096,7 @@ func TestTxnOnePhaseCommit(t *testing.T) {
 	if !bytes.Equal(val, value) {
 		t.Fatalf("expected: %s, got: %s", value, val)
 	}
-	checkTxnMetrics(t, metrics, "commit 1PC txn", 1 /* commits */, 1 /* 1PC */, 0, 0, 0)
+	checkTxnMetrics(t, metrics, "commit 1PC txn", 1 /* commits */, 1 /* 1PC */, 0, 0)
 }
 
 func TestTxnAbortCount(t *testing.T) {
@@ -1179,10 +1111,6 @@ func TestTxnAbortCount(t *testing.T) {
 	if err := s.DB.Txn(context.TODO(), func(ctx context.Context, txn *client.Txn) error {
 		key := []byte("key-abort")
 
-		if err := txn.SetIsolation(enginepb.SNAPSHOT); err != nil {
-			return err
-		}
-
 		if err := txn.Put(ctx, key, value); err != nil {
 			t.Fatal(err)
 		}
@@ -1191,7 +1119,7 @@ func TestTxnAbortCount(t *testing.T) {
 	}); !testutils.IsError(err, intentionalErrText) {
 		t.Fatalf("unexpected error: %v", err)
 	}
-	checkTxnMetrics(t, metrics, "abort txn", 0, 0, 0, 1 /* aborts */, 0)
+	checkTxnMetrics(t, metrics, "abort txn", 0, 0, 1 /* aborts */, 0)
 }
 
 func TestTxnRestartCount(t *testing.T) {
@@ -1248,7 +1176,7 @@ func TestTxnRestartCount(t *testing.T) {
 	// Commit (should cause restart metric to increase).
 	err := txn.CommitOrCleanup(ctx)
 	assertTransactionRetryError(t, err)
-	checkTxnMetrics(t, metrics, "restart txn", 0, 0, 0, 1 /* aborts */, 1 /* restarts */)
+	checkTxnMetrics(t, metrics, "restart txn", 0, 0, 1 /* aborts */, 1 /* restarts */)
 }
 
 func TestTxnDurations(t *testing.T) {
@@ -1262,9 +1190,6 @@ func TestTxnDurations(t *testing.T) {
 	for i := 0; i < puts; i++ {
 		key := roachpb.Key(fmt.Sprintf("key-txn-durations-%d", i))
 		if err := s.DB.Txn(context.TODO(), func(ctx context.Context, txn *client.Txn) error {
-			if err := txn.SetIsolation(enginepb.SNAPSHOT); err != nil {
-				return err
-			}
 			if err := txn.Put(ctx, key, []byte("val")); err != nil {
 				return err
 			}
@@ -1275,7 +1200,7 @@ func TestTxnDurations(t *testing.T) {
 		}
 	}
 
-	checkTxnMetrics(t, metrics, "txn durations", puts, 0, 0, 0, 0)
+	checkTxnMetrics(t, metrics, "txn durations", puts, 0, 0, 0)
 
 	hist := metrics.Durations
 	// The clock is a bit odd in these tests, so I can't test the mean without
@@ -1335,21 +1260,16 @@ func TestAbortTransactionOnCommitErrors(t *testing.T) {
 				_ context.Context, ba roachpb.BatchRequest,
 			) (*roachpb.BatchResponse, *roachpb.Error) {
 				br := ba.CreateReply()
-				txn := ba.Txn.Clone()
-				br.Txn = &txn
+				br.Txn = ba.Txn.Clone()
 
-				if _, hasBT := ba.GetArg(roachpb.BeginTransaction); hasBT {
-					if _, ok := ba.Requests[1].GetInner().(*roachpb.PutRequest); !ok {
+				if _, hasPut := ba.GetArg(roachpb.Put); hasPut {
+					if _, ok := ba.Requests[0].GetInner().(*roachpb.PutRequest); !ok {
 						t.Fatalf("expected Put")
 					}
 					union := &br.Responses[0] // avoid operating on copy
-					union.MustSetInner(&roachpb.BeginTransactionResponse{})
-					union = &br.Responses[1] // avoid operating on copy
 					union.MustSetInner(&roachpb.PutResponse{})
 					if ba.Txn != nil && br.Txn == nil {
-						txnClone := ba.Txn.Clone()
-						br.Txn = &txnClone
-						br.Txn.Writing = true
+						br.Txn = ba.Txn.Clone()
 						br.Txn.Status = roachpb.PENDING
 					}
 				} else if et, hasET := ba.GetArg(roachpb.EndTransaction); hasET {
@@ -1430,8 +1350,7 @@ func (s *mockSender) Send(
 	}
 	// If none of the matchers triggered, just create an empty reply.
 	br := ba.CreateReply()
-	txn := ba.Txn.Clone()
-	br.Txn = &txn
+	br.Txn = ba.Txn.Clone()
 	return br, nil
 }
 
@@ -1550,8 +1469,7 @@ func TestOnePCErrorTracking(t *testing.T) {
 		resp := ba.CreateReply()
 		// Set the response's txn to the Aborted status (as the server would). This
 		// will make the TxnCoordSender stop the heartbeat loop.
-		txnCopy := ba.Txn.Clone()
-		resp.Txn = &txnCopy
+		resp.Txn = ba.Txn.Clone()
 		resp.Txn.Status = roachpb.ABORTED
 		return resp, nil
 	})
@@ -1611,7 +1529,6 @@ func TestIntentTrackingBeforeBeginTransaction(t *testing.T) {
 		"test txn",
 		key,
 		roachpb.UserPriority(0),
-		enginepb.SERIALIZABLE,
 		clock.Now(),
 		clock.MaxOffset().Nanoseconds(),
 	)
@@ -1706,8 +1623,8 @@ func TestCommitMutatingTransaction(t *testing.T) {
 	var calls []roachpb.Method
 	sender.match(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
 		calls = append(calls, ba.Methods()...)
-		if bt, ok := ba.GetArg(roachpb.BeginTransaction); ok && !bt.Header().Key.Equal(roachpb.Key("a")) {
-			t.Errorf("expected begin transaction key to be \"a\"; got %s", bt.Header().Key)
+		if !bytes.Equal(ba.Txn.Key, roachpb.Key("a")) {
+			t.Errorf("expected transaction key to be \"a\"; got %s", ba.Txn.Key)
 		}
 		if et, ok := ba.GetArg(roachpb.EndTransaction); ok && !et.(*roachpb.EndTransactionRequest).Commit {
 			t.Errorf("expected commit to be true")
@@ -1769,7 +1686,7 @@ func TestCommitMutatingTransaction(t *testing.T) {
 			if err := db.Txn(ctx, test.f); err != nil {
 				t.Fatalf("%d: unexpected error on commit: %s", i, err)
 			}
-			expectedCalls := []roachpb.Method{roachpb.BeginTransaction, test.expMethod}
+			expectedCalls := []roachpb.Method{test.expMethod}
 			if test.pointWrite {
 				expectedCalls = append(expectedCalls, roachpb.QueryIntent)
 			}
@@ -1783,6 +1700,7 @@ func TestCommitMutatingTransaction(t *testing.T) {
 
 // TestTxnInsertBeginTransaction verifies that a begin transaction
 // request is inserted just before the first mutating command.
+// TODO(nvanbenschoten): Remove in 2.3.
 func TestTxnInsertBeginTransaction(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	ctx := context.Background()
@@ -1804,9 +1722,12 @@ func TestTxnInsertBeginTransaction(t *testing.T) {
 		return nil, nil
 	})
 
+	v := cluster.VersionByKey(cluster.Version2_1)
+	st := cluster.MakeTestingClusterSettingsWithVersion(v, v)
 	factory := NewTxnCoordSenderFactory(
 		TxnCoordSenderFactoryConfig{
 			AmbientCtx: ambient,
+			Settings:   st,
 			Clock:      clock,
 			Stopper:    stopper,
 		},
@@ -1835,6 +1756,7 @@ func TestTxnInsertBeginTransaction(t *testing.T) {
 
 // TestBeginTransactionErrorIndex verifies that the error index is cleared
 // when a BeginTransaction command causes an error.
+// TODO(nvanbenschoten): Remove in 2.3.
 func TestBeginTransactionErrorIndex(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	ctx := context.Background()
@@ -1850,9 +1772,12 @@ func TestBeginTransactionErrorIndex(t *testing.T) {
 		return nil, pErr
 	})
 
+	v := cluster.VersionByKey(cluster.Version2_1)
+	st := cluster.MakeTestingClusterSettingsWithVersion(v, v)
 	factory := NewTxnCoordSenderFactory(
 		TxnCoordSenderFactoryConfig{
 			AmbientCtx: ambient,
+			Settings:   st,
 			Clock:      clock,
 			Stopper:    stopper,
 		},
@@ -1943,7 +1868,9 @@ func TestEndWriteRestartReadOnlyTransaction(t *testing.T) {
 	sender.match(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
 		calls = append(calls, ba.Methods()...)
 		if _, ok := ba.GetArg(roachpb.Put); ok {
-			return nil, roachpb.NewErrorWithTxn(roachpb.NewTransactionRetryError(roachpb.RETRY_SERIALIZABLE), ba.Txn)
+			return nil, roachpb.NewErrorWithTxn(
+				roachpb.NewTransactionRetryError(roachpb.RETRY_SERIALIZABLE, "test err"),
+				ba.Txn)
 		}
 		return nil, nil
 	})
@@ -1962,7 +1889,7 @@ func TestEndWriteRestartReadOnlyTransaction(t *testing.T) {
 		sender,
 	)
 	db := client.NewDB(testutils.MakeAmbientCtx(), factory, clock)
-	expCalls := []roachpb.Method{roachpb.BeginTransaction, roachpb.Put, roachpb.EndTransaction}
+	expCalls := []roachpb.Method{roachpb.Put, roachpb.EndTransaction}
 
 	testutils.RunTrueAndFalse(t, "success", func(t *testing.T, success bool) {
 		calls = nil
@@ -1988,8 +1915,8 @@ func TestEndWriteRestartReadOnlyTransaction(t *testing.T) {
 }
 
 // TestTransactionKeyNotChangedInRestart verifies that if the transaction
-// already has a key (we're in a restart), the key in the begin transaction
-// request is not changed.
+// already has a key (we're in a restart), the key in the transaction request is
+// not changed.
 func TestTransactionKeyNotChangedInRestart(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	ctx := context.Background()
@@ -2005,18 +1932,11 @@ func TestTransactionKeyNotChangedInRestart(t *testing.T) {
 		// Ignore the final EndTxnRequest.
 		if _, ok := ba.GetArg(roachpb.EndTransaction); ok {
 			br := ba.CreateReply()
-			txn := ba.Txn.Clone()
-			br.Txn = &txn
+			br.Txn = ba.Txn.Clone()
 			return nil, nil
 		}
 
-		// Attempt 0 should have a BeginTxnRequest, and a PutRequest.
-		// Attempt 1 should have a PutRequest.
-		if attempt == 0 {
-			if _, ok := ba.GetArg(roachpb.BeginTransaction); !ok {
-				t.Fatalf("failed to find a begin transaction request: %v", ba)
-			}
-		}
+		// Both attempts should have a PutRequest.
 		if _, ok := ba.GetArg(roachpb.Put); !ok {
 			t.Fatalf("failed to find a put request: %v", ba)
 		}
@@ -2028,7 +1948,9 @@ func TestTransactionKeyNotChangedInRestart(t *testing.T) {
 		}
 
 		if attempt == 0 {
-			return nil, roachpb.NewErrorWithTxn(roachpb.NewTransactionRetryError(roachpb.RETRY_SERIALIZABLE), ba.Txn)
+			return nil, roachpb.NewErrorWithTxn(
+				roachpb.NewTransactionRetryError(roachpb.RETRY_SERIALIZABLE, "test err"),
+				ba.Txn)
 		}
 		return nil, nil
 	})
@@ -2081,9 +2003,6 @@ func TestSequenceNumbers(t *testing.T) {
 					expSequence, seq, args)
 			}
 		}
-		if expSequence != ba.Txn.Sequence {
-			t.Errorf("expected header sequence %d; got %d", expSequence, ba.Txn.Sequence)
-		}
 		br := ba.CreateReply()
 		br.Txn = ba.Txn
 		return br, nil
@@ -2158,10 +2077,9 @@ func TestConcurrentTxnRequests(t *testing.T) {
 	}
 
 	expectedCallCounts := map[roachpb.Method]int{
-		roachpb.BeginTransaction: 1,
-		roachpb.Put:              26,
-		roachpb.QueryIntent:      26,
-		roachpb.EndTransaction:   1,
+		roachpb.Put:            26,
+		roachpb.QueryIntent:    26,
+		roachpb.EndTransaction: 1,
 	}
 	if !reflect.DeepEqual(expectedCallCounts, callCounts) {
 		t.Errorf("expected %v, got %v", expectedCallCounts, callCounts)
@@ -2210,9 +2128,8 @@ func TestTxnRequestTxnTimestamp(t *testing.T) {
 				curReq, req.expRequestTS, ba.Txn.Timestamp)
 		}
 
-		txnClone := ba.Txn.Clone()
 		br := ba.CreateReply()
-		br.Txn = &txnClone
+		br.Txn = ba.Txn.Clone()
 		br.Txn.Timestamp.Forward(requests[curReq].responseTS)
 		return br, nil
 	})
@@ -2247,9 +2164,8 @@ func TestReadOnlyTxnObeysDeadline(t *testing.T) {
 	sender.match(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
 		if _, ok := ba.GetArg(roachpb.Get); ok {
 			manual.Increment(100)
-			txnClone := ba.Txn.Clone()
 			br := ba.CreateReply()
-			br.Txn = &txnClone
+			br.Txn = ba.Txn.Clone()
 			br.Txn.Timestamp.Forward(clock.Now())
 			return br, nil
 		}
@@ -2276,9 +2192,10 @@ func TestReadOnlyTxnObeysDeadline(t *testing.T) {
 		if _, err := txn.Get(ctx, "k"); err != nil {
 			t.Fatal(err)
 		}
-		if err := txn.Commit(ctx); !testutils.IsError(
-			err, "deadline exceeded before transaction finalization") {
-			t.Fatal(err)
+		err := txn.Commit(ctx)
+		assertTransactionRetryError(t, err)
+		if !testutils.IsError(err, "RETRY_COMMIT_DEADLINE_EXCEEDED") {
+			t.Fatalf("expected deadline exceeded, got: %s", err)
 		}
 	})
 
@@ -2288,9 +2205,10 @@ func TestReadOnlyTxnObeysDeadline(t *testing.T) {
 		txn.UpdateDeadlineMaybe(ctx, clock.Now())
 		b := txn.NewBatch()
 		b.Get("k")
-		if err := txn.CommitInBatch(ctx, b); !testutils.IsError(
-			err, "deadline exceeded before transaction finalization") {
-			t.Fatal(err)
+		err := txn.CommitInBatch(ctx, b)
+		assertTransactionRetryError(t, err)
+		if !testutils.IsError(err, "RETRY_COMMIT_DEADLINE_EXCEEDED") {
+			t.Fatalf("expected deadline exceeded, got: %s", err)
 		}
 	})
 }
@@ -2342,8 +2260,8 @@ func TestTxnCoordSenderPipelining(t *testing.T) {
 	}
 
 	require.Equal(t, []roachpb.Method{
-		roachpb.BeginTransaction, roachpb.Put, roachpb.QueryIntent, roachpb.EndTransaction,
-		roachpb.BeginTransaction, roachpb.Put, roachpb.EndTransaction,
+		roachpb.Put, roachpb.QueryIntent, roachpb.EndTransaction,
+		roachpb.Put, roachpb.EndTransaction,
 	}, calls)
 
 	for _, action := range []func(ctx context.Context, txn *client.Txn) error{
@@ -2384,8 +2302,7 @@ func TestAnchorKey(t *testing.T) {
 			t.Fatalf("expected anchor %q, got %q", key2, ba.Txn.Key)
 		}
 		br := ba.CreateReply()
-		txn := ba.Txn.Clone()
-		br.Txn = &txn
+		br.Txn = ba.Txn.Clone()
 		return br, nil
 	}
 
@@ -2409,92 +2326,32 @@ func TestAnchorKey(t *testing.T) {
 	}
 }
 
-// TestCommitTurnedToRollback tests that the TxnCoordSender (or, rather, one of
-// the interceptors) turns a commit into a rollback in situations where a txn
-// has performed writes at old epochs but no writes at the current epoch. See
-// the comment in txnHeartbeat about why this is needed.
-// We check that the transformation happened (by looking at a trace) and that
-// other things (e.g. the proto status) look like a committed transaction even
-// though we technically performed a rollback.
-func TestCommitTurnedToRollback(t *testing.T) {
-	defer leaktest.AfterTest(t)()
-
-	ctx := context.Background()
-	s, metrics, cleanupFn := setupMetricsTest(t)
-	defer cleanupFn()
-
-	readKey := []byte("read")
-	writeKey := []byte("write")
-	value := []byte("value")
-
-	// Start a transaction and read a key that we're going to modify outside the
-	// txn. This ensures that refreshing the txn will not succeed, so a restart
-	// will be necessary.
-	txn := client.NewTxn(ctx, s.DB, 0 /* gatewayNodeID */, client.RootTxn)
-	if _, err := txn.Get(ctx, readKey); err != nil {
-		t.Fatal(err)
-	}
-
-	// Write the read key outside of the
-	// transaction, at a higher timestamp, which will necessitate a
-	// txn restart when the original read key span is updated.
-	if err := s.DB.Put(ctx, readKey, value); err != nil {
-		t.Fatal(err)
-	}
-
-	// Outside of the transaction, read the same key as will be
-	// written within the transaction. This means that future
-	// attempts to write will forward the txn timestamp.
-	if _, err := s.DB.Get(ctx, writeKey); err != nil {
-		t.Fatal(err)
-	}
-
-	// This put will lay down an intent, txn timestamp will be bumped.
-	if err := txn.Put(ctx, writeKey, value); err != nil {
-		t.Fatal(err)
-	}
-	proto := txn.Serialize()
-	if !proto.OrigTimestamp.Less(proto.Timestamp) {
-		t.Errorf("expected timestamp to increase: %s", proto)
-	}
-
-	// Attempt to commit, expect a retriable error.
-	err := txn.Commit(ctx)
-	assertTransactionRetryError(t, err)
-
-	// Attempt to commit again, at the next epoch. This should succeed, and the
-	// commit should be turned to a rollback.
-	commitCtx, getRec, cancel := tracing.ContextWithRecordingSpan(ctx, "test")
-	defer cancel()
-	err = txn.Commit(commitCtx)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	// Check that the commit metric has been incremented.
-	checkTxnMetrics(t, metrics, "commit turned to rollback",
-		1 /* commits */, 0, 0, 0, 1 /* restarts */)
-
-	// Check that the proto's status is the expected one for a commit.
-	if s := txn.Serialize().Status; s != roachpb.COMMITTED {
-		t.Fatalf("expected COMMITTED status, got: %s", s)
-	}
-
-	// Look for a specific log message indicating that this test is not fooling
-	// itself and indeed we transformed a commit to a rollback.
-	if tracing.FindMsgInRecording(getRec(), turningCommitToRollbackMsg) == -1 {
-		t.Fatalf("didn't find trace message: %s", turningCommitToRollbackMsg)
-	}
-}
-
 // Test that a leaf txn returns a raw error when "rejecting a client" (a client
 // sending something after the txn is known to be aborted), not a
-// HandledRetryableError. This is important as leaves are not supposed to create
+// TransactionRetryWithProtoRefreshError. This is important as leaves are not supposed to create
 // "handled" errors; instead the DistSQL infra knows to recognize raw retryable
 // errors and feed them to the root txn.
 func TestLeafTxnClientRejectError(t *testing.T) {
 	defer leaktest.AfterTest(t)()
-	s := createTestDB(t)
+
+	// We're going to inject an error so that a leaf txn is "poisoned". This can
+	// happen, for example, if the leaf is used concurrently by multiple requests,
+	// where the first one gets a TransactionAbortedError.
+	errKey := roachpb.Key("a")
+	knobs := &storage.StoreTestingKnobs{
+		TestingRequestFilter: func(ba roachpb.BatchRequest) *roachpb.Error {
+			if g, ok := ba.GetArg(roachpb.Get); ok && g.(*roachpb.GetRequest).Key.Equal(errKey) {
+				txn := ba.Txn.Clone()
+				txn.Status = roachpb.ABORTED
+				return roachpb.NewErrorWithTxn(
+					roachpb.NewTransactionAbortedError(roachpb.ABORT_REASON_UNKNOWN), txn,
+				)
+			}
+			return nil
+		},
+	}
+
+	s := createTestDBWithContextAndKnobs(t, client.DefaultDBContext(), knobs)
 	defer s.Stop()
 
 	ctx := context.Background()
@@ -2504,20 +2361,37 @@ func TestLeafTxnClientRejectError(t *testing.T) {
 	leafTxn := client.NewTxnWithCoordMeta(
 		ctx, s.DB, 0 /* gatewayNodeID */, client.LeafTxn, rootTxn.GetTxnCoordMeta(ctx))
 
-	// Poison the leaf. This can happen, for example, if the leaf is used
-	// concurrently by multiple requests, where the first one gets a
-	// TransactionAbortedError.
-	meta := rootTxn.GetTxnCoordMeta(ctx)
-	meta.Txn.Status = roachpb.ABORTED
-	leafTxn.AugmentTxnCoordMeta(ctx, meta)
+	if _, err := leafTxn.Get(ctx, errKey); !testutils.IsError(err, "TransactionAbortedError") {
+		t.Fatalf("expected injected err, got: %v", err)
+	}
 
 	// Now use the leaf and check the error. At the TxnCoordSender level, the
 	// pErr will be TransactionAbortedError. When pErr.GoError() is called, that's
 	// transformed into an UnhandledRetryableError. For our purposes, what this
 	// test is interested in demonstrating is that it's not a
-	// HandledRetryableError.
+	// TransactionRetryWithProtoRefreshError.
 	_, err := leafTxn.Get(ctx, roachpb.Key("a"))
 	if _, ok := err.(*roachpb.UnhandledRetryableError); !ok {
 		t.Fatalf("expected UnhandledRetryableError(TransactionAbortedError), got: (%T) %v", err, err)
 	}
 }
+
+// Check that ingesting an Aborted txn record is a no-op. The TxnCoordSender is
+// supposed to reject such updates because they risk putting it into an
+// inconsistent state. See comments in TxnCoordSender.AugmentMeta().
+func TestAugmentTxnCoordMetaAbortedTxn(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	s := createTestDBWithContextAndKnobs(t, client.DefaultDBContext(), nil /* knobs */)
+	defer s.Stop()
+	ctx := context.Background()
+
+	txn := client.NewTxn(ctx, s.DB, 0 /* gatewayNodeID */, client.RootTxn)
+	txnProto := txn.Serialize()
+	txnProto.Status = roachpb.ABORTED
+	txn.AugmentTxnCoordMeta(ctx, roachpb.TxnCoordMeta{Txn: *txnProto})
+	// Check that the transaction was not updated.
+	txnProto = txn.Serialize()
+	if txnProto.Status != roachpb.PENDING {
+		t.Fatalf("expected PENDING txn, got: %s", txnProto.Status)
+	}
+}
