@@ -16,13 +16,13 @@ package kv
 
 import (
 	"context"
-
-	"github.com/google/btree"
+	"fmt"
 
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/google/btree"
 )
 
 // The degree of the outstandingWrites btree.
@@ -33,6 +33,16 @@ var pipelinedWritesEnabled = settings.RegisterBoolSetting(
 	"if enabled, transactional writes are pipelined through Raft consensus",
 	true,
 )
+var pipelinedWritesMaxOutstandingSize = settings.RegisterByteSizeSetting(
+	// TODO(nvanbenschoten): The need for this extra setting alongside
+	// kv.transaction.max_intents_bytes indicates that we should explore
+	// the unification of intent tracking and in-flight write tracking.
+	// The two mechanisms track subtly different information, but there's
+	// no fundamental reason why they can't be unified.
+	"kv.transaction.write_pipelining_max_outstanding_size",
+	"maximum number of bytes used to track in-flight pipelined writes before disabling pipelining",
+	1<<18, /* 256 KB */
+)
 var pipelinedWritesMaxBatchSize = settings.RegisterNonNegativeIntSetting(
 	"kv.transaction.write_pipelining_max_batch_size",
 	"if non-zero, defines that maximum size batch that will be pipelined through Raft consensus",
@@ -59,6 +69,7 @@ var pipelinedWritesMaxBatchSize = settings.RegisterNonNegativeIntSetting(
 //
 // Chaining on to in-flight async writes is important for two main reasons to
 // txnPipeliner:
+//
 // 1. requests proposed to Raft will not necessarily succeed. For any number of
 //    reasons, the request may make it through Raft and be discarded or fail to
 //    ever even be replicated. A transaction must check that all async writes
@@ -70,19 +81,19 @@ var pipelinedWritesMaxBatchSize = settings.RegisterNonNegativeIntSetting(
 //    discovered during write *evaluation*, which an async write will perform
 //    synchronously before consensus. Any error during consensus is outside of the
 //    Transaction-domain and can always trigger a transaction retry.
+//
 // 2. transport layers beneath the txnPipeliner do not provide strong enough
 //    ordering guarantees between concurrent requests in the same transaction to
 //    avoid needing explicit chaining. For instance, DistSender uses unary gRPC
 //    requests instead of gRPC streams, so it can't natively expose strong ordering
-//    guarantees. Perhaps more importantly, even when a command has entered the
-//    command queue and evaluated on a Replica, it is not guaranteed to be applied
-//    before interfering commands. This is because the command may be retried
-//    outside of the serialization of the command queue for any number of reasons,
-//    such as leaseholder changes. When the command re-enters the command queue,
-//    it's possible that interfering commands may jump ahead of it. To combat
-//    this, the txnPipeliner uses chaining to throw an error when these
-//    re-orderings would have affected the order that transactional requests
-//    evaluate in.
+//    guarantees. Perhaps more importantly, even when a command has acquired latches
+//    and evaluated on a Replica, it is not guaranteed to be applied before
+//    interfering commands. This is because the command may be retried outside of
+//    the serialization of the spanlatch manager for any number of reasons, such as
+//    leaseholder changes. When the command re-acquired its latches, it's possible
+//    that interfering commands may jump ahead of it. To combat this, the
+//    txnPipeliner uses chaining to throw an error when these re-orderings would
+//    have affected the order that transactional requests evaluate in.
 //
 // The interceptor proves all outstanding writes before committing a transaction
 // by tacking on a QueryIntent request for each one to the front of an
@@ -100,12 +111,14 @@ var pipelinedWritesMaxBatchSize = settings.RegisterNonNegativeIntSetting(
 // would be in the txnPipeliner's best interest to prove outstanding writes as
 // early as possible, even if no other overlapping requests force them to be
 // proven. The approaches are:
+//
 // 1. launch a background process after each successful async write to query its
 //    intents and wait for it to succeed. This would effectively solve the issue,
 //    but at the cost of many more goroutines and many more QueryIntent requests,
 //    most of which would be redundant because their corresponding write wouldn't
 //    complete until after an EndTransaction synchronously needed to prove them
 //    anyway.
+//
 // 2. to address the issue of an unbounded number of background goroutines
 //    proving writes in approach 1, a single background goroutine could be run
 //    that repeatedly loops over all outstanding writes and attempts to prove
@@ -115,12 +128,14 @@ var pipelinedWritesMaxBatchSize = settings.RegisterNonNegativeIntSetting(
 //    reason as approach 1: most of its QueryIntent requests will be useless
 //    because a transaction will send an EndTransaction immediately after sending
 //    all of its writes.
+//
 // 3. turn the KV interface into a streaming protocol (#8360) that could support
 //    returning multiple results. This would allow clients to return immediately
 //    after a writes "evaluation" phase completed but hold onto a handle to the
 //    request and be notified immediately after its "replication" phase completes.
 //    This would allow txnPipeliner to prove outstanding writes immediately after
 //    they finish consensus without any extra RPCs.
+//
 // So far, none of these approaches have been integrated.
 //
 // [1] A proposal called "parallel commits" (#24194) exists that would allow all
@@ -135,6 +150,7 @@ type txnPipeliner struct {
 	disabled bool
 
 	outstandingWrites *btree.BTree
+	owSizeBytes       int64 // byte size of all keys in outstandingWrites
 	owAlloc           outstandingWriteAlloc
 	tmpOW1, tmpOW2    outstandingWrite // avoid allocs
 }
@@ -156,9 +172,7 @@ func (tp *txnPipeliner) SendLocked(
 	ctx context.Context, ba roachpb.BatchRequest,
 ) (*roachpb.BatchResponse, *roachpb.Error) {
 	// Fast-path for 1PC transactions.
-	_, hasBT := ba.GetArg(roachpb.BeginTransaction)
-	_, hasET := ba.GetArg(roachpb.EndTransaction)
-	if hasBT && hasET {
+	if ba.IsCompleteTransaction() {
 		return tp.wrapped.SendLocked(ctx, ba)
 	}
 
@@ -194,8 +208,13 @@ func (tp *txnPipeliner) SendLocked(
 // outstandingWrites tree. Once this limit is hit, we'll either need to
 // proactively prove outstanding writes or stop pipelining new writes.
 func (tp *txnPipeliner) chainToOutstandingWrites(ba roachpb.BatchRequest) roachpb.BatchRequest {
-	asyncConsensus := tp.st.Version.IsActive(cluster.VersionAsyncConsensus) &&
-		pipelinedWritesEnabled.Get(&tp.st.SV) && !tp.disabled
+	asyncConsensus := pipelinedWritesEnabled.Get(&tp.st.SV) && !tp.disabled
+
+	// We provide a setting to bound the size of outstanding writes that the
+	// pipeliner is tracking. If this batch would push us over this setting,
+	// don't allow it to perform async consensus.
+	owSizeBytes := tp.owSizeBytes
+	maxOWSizeBytes := pipelinedWritesMaxOutstandingSize.Get(&tp.st.SV)
 
 	// We provide a setting to bound the number of writes we permit in a batch
 	// that uses async consensus. This is useful because we'll have to prove
@@ -233,15 +252,33 @@ func (tp *txnPipeliner) chainToOutstandingWrites(ba roachpb.BatchRequest) roachp
 			// writes.
 			continue
 		}
-		if !roachpb.IsTransactionWrite(req) || roachpb.IsRange(req) {
-			// Only allow batches consisting of solely transactional point
-			// writes to perform consensus asynchronously.
-			// TODO(nvanbenschoten): We could allow batches with reads and point
-			// writes to perform async consensus, but this would be a bit
-			// tricky. Any read would need to chain on to any write that came
-			// before it in the batch and overlaps. For now, it doesn't seem
-			// worth it.
-			asyncConsensus = false
+
+		if asyncConsensus {
+			// If we're currently planning on performing the batch with
+			// performing async consensus, determine whether this request
+			// changes that.
+			if !roachpb.IsTransactionWrite(req) || roachpb.IsRange(req) {
+				// Only allow batches consisting of solely transactional point
+				// writes to perform consensus asynchronously.
+				// TODO(nvanbenschoten): We could allow batches with reads and point
+				// writes to perform async consensus, but this would be a bit
+				// tricky. Any read would need to chain on to any write that came
+				// before it in the batch and overlaps. For now, it doesn't seem
+				// worth it.
+				asyncConsensus = false
+			} else {
+				// Only allow batches that would not push us over the maximum
+				// outstanding write size limit to perform consensus asynchronously.
+				//
+				// NB: this estimation is conservative because it doesn't factor
+				// in that some writes may be proven by this batch and removed
+				// from the outstanding write set. The real accounting in
+				// maybe{InsertOutstanding/RemoveProven}WriteLocked gets this
+				// right.
+				owSizeBytes += int64(len(req.Header().Key))
+				asyncConsensus = owSizeBytes <= maxOWSizeBytes
+			}
+
 		}
 
 		if tp.outstandingWritesLen() > len(chainedKeys) {
@@ -322,7 +359,7 @@ func (tp *txnPipeliner) chainToOutstandingWrites(ba roachpb.BatchRequest) roachp
 // two actions:
 // 1. it removes all outstanding writes that the request proved to exist from
 //    the outstanding writes set.
-// 2. it adds all async writes the the request performed to the outstanding
+// 2. it adds all async writes that the request performed to the outstanding
 //    write set.
 //
 // While doing so, the method also strips all QueryIntent responses from the
@@ -334,6 +371,7 @@ func (tp *txnPipeliner) updateOutstandingWrites(
 	// tree. This will turn maybeRemoveProvenWriteLocked into a quick no-op.
 	if br.Txn != nil && br.Txn.Status != roachpb.PENDING && tp.outstandingWrites != nil {
 		tp.outstandingWrites.Clear(false /* addNodesToFreelist */)
+		tp.owSizeBytes = 0
 	}
 
 	j := 0
@@ -389,7 +427,8 @@ func (tp *txnPipeliner) adjustError(
 	// Turn an IntentMissingError into a transactional retry error.
 	if ime, ok := pErr.GetDetail().(*roachpb.IntentMissingError); ok {
 		log.VEventf(ctx, 2, "transforming intent missing error into retry: %v", ime)
-		err := roachpb.NewTransactionRetryError(roachpb.RETRY_ASYNC_WRITE_FAILURE)
+		err := roachpb.NewTransactionRetryError(
+			roachpb.RETRY_ASYNC_WRITE_FAILURE, fmt.Sprintf("missing intent on: %s", ime.Key))
 		retryErr := roachpb.NewErrorWithTxn(err, pErr.GetTxn())
 		retryErr.Index = pErr.Index
 		return retryErr
@@ -426,6 +465,7 @@ func (tp *txnPipeliner) epochBumpedLocked() {
 	if tp.outstandingWrites != nil {
 		// Add nodes to freelist so that next epoch can reuse btree memory.
 		tp.outstandingWrites.Clear(true /* addNodesToFreelist */)
+		tp.owSizeBytes = 0
 		tp.owAlloc.Clear()
 	}
 }
@@ -463,6 +503,7 @@ func (tp *txnPipeliner) maybeInsertOutstandingWriteLocked(key roachpb.Key, seq i
 
 	w := tp.owAlloc.Alloc(key, seq)
 	tp.outstandingWrites.ReplaceOrInsert(w)
+	tp.owSizeBytes += int64(len(key))
 }
 
 // maybeRemoveProvenWriteLocked attempts to remove an outstanding write that
@@ -490,6 +531,14 @@ func (tp *txnPipeliner) maybeRemoveProvenWriteLocked(key roachpb.Key, seq int32)
 	if delItem != nil {
 		*delItem.(*outstandingWrite) = outstandingWrite{} // for GC
 	}
+	tp.owSizeBytes -= int64(len(key))
+
+	// Assert that the byte accounting is believable.
+	if tp.owSizeBytes < 0 {
+		panic("negative outstanding write size")
+	} else if tp.outstandingWrites.Len() == 0 && tp.owSizeBytes != 0 {
+		panic("non-zero outstanding write size with 0 outstanding writes")
+	}
 }
 
 // outstandingWriteAlloc provides chunk allocation of outstandingWrites,
