@@ -17,16 +17,16 @@ package kv
 import (
 	"context"
 	"fmt"
+	"math"
 	"strings"
 	"testing"
 
-	"github.com/google/btree"
-	"github.com/stretchr/testify/require"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
+	"github.com/google/btree"
+	"github.com/stretchr/testify/require"
 )
 
 // mockLockedSender implements the lockedSender interface and provides a way to
@@ -56,7 +56,7 @@ func makeMockTxnPipeliner() (txnPipeliner, *mockLockedSender) {
 }
 
 func makeTxnProto() roachpb.Transaction {
-	return roachpb.MakeTransaction("test", []byte("key"), 0, 0, hlc.Timestamp{}, 0)
+	return roachpb.MakeTransaction("test", []byte("key"), 0, hlc.Timestamp{}, 0)
 }
 
 // TestTxnPipeliner1PCTransaction tests that 1PC transactions pass through the
@@ -71,16 +71,14 @@ func TestTxnPipeliner1PCTransaction(t *testing.T) {
 
 	var ba roachpb.BatchRequest
 	ba.Header = roachpb.Header{Txn: &txn}
-	ba.Add(&roachpb.BeginTransactionRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
 	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
 	ba.Add(&roachpb.EndTransactionRequest{Commit: true})
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
-		require.Equal(t, 3, len(ba.Requests))
+		require.Equal(t, 2, len(ba.Requests))
 		require.False(t, ba.AsyncConsensus)
-		require.IsType(t, &roachpb.BeginTransactionRequest{}, ba.Requests[0].GetInner())
-		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[1].GetInner())
-		require.IsType(t, &roachpb.EndTransactionRequest{}, ba.Requests[2].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[0].GetInner())
+		require.IsType(t, &roachpb.EndTransactionRequest{}, ba.Requests[1].GetInner())
 
 		br := ba.CreateReply()
 		br.Txn = ba.Txn
@@ -108,16 +106,14 @@ func TestTxnPipelinerTrackOutstandingWrites(t *testing.T) {
 
 	var ba roachpb.BatchRequest
 	ba.Header = roachpb.Header{Txn: &txn}
-	ba.Add(&roachpb.BeginTransactionRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
 	putArgs := roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}}
-	putArgs.Sequence = 2
+	putArgs.Sequence = 1
 	ba.Add(&putArgs)
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
-		require.Equal(t, 2, len(ba.Requests))
+		require.Equal(t, 1, len(ba.Requests))
 		require.True(t, ba.AsyncConsensus)
-		require.IsType(t, &roachpb.BeginTransactionRequest{}, ba.Requests[0].GetInner())
-		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[1].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[0].GetInner())
 
 		br := ba.CreateReply()
 		br.Txn = ba.Txn
@@ -137,18 +133,18 @@ func TestTxnPipelinerTrackOutstandingWrites(t *testing.T) {
 	keyB, keyC := roachpb.Key("b"), roachpb.Key("c")
 	ba.Requests = nil
 	cputArgs := roachpb.ConditionalPutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}}
-	cputArgs.Sequence = 3
+	cputArgs.Sequence = 2
 	ba.Add(&cputArgs)
 	initPutArgs := roachpb.InitPutRequest{RequestHeader: roachpb.RequestHeader{Key: keyB}}
-	initPutArgs.Sequence = 4
+	initPutArgs.Sequence = 3
 	ba.Add(&initPutArgs)
 	incArgs := roachpb.IncrementRequest{RequestHeader: roachpb.RequestHeader{Key: keyC}}
-	incArgs.Sequence = 5
+	incArgs.Sequence = 4
 	ba.Add(&incArgs)
 	// Write at the same key as another write in the same batch. Will only
 	// result in a single outstanding write, at the larger sequence number.
 	delArgs := roachpb.DeleteRequest{RequestHeader: roachpb.RequestHeader{Key: keyC}}
-	delArgs.Sequence = 6
+	delArgs.Sequence = 5
 	ba.Add(&delArgs)
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
@@ -164,7 +160,7 @@ func TestTxnPipelinerTrackOutstandingWrites(t *testing.T) {
 		require.Equal(t, keyA, qiReq.Key)
 		require.Equal(t, txn.ID, qiReq.Txn.ID)
 		require.Equal(t, txn.Timestamp, qiReq.Txn.Timestamp)
-		require.Equal(t, int32(2), qiReq.Txn.Sequence)
+		require.Equal(t, int32(1), qiReq.Txn.Sequence)
 		require.Equal(t, roachpb.QueryIntentRequest_RETURN_ERROR, qiReq.IfMissing)
 
 		// No outstanding writes have been proved yet.
@@ -198,9 +194,11 @@ func TestTxnPipelinerTrackOutstandingWrites(t *testing.T) {
 	keyD := roachpb.Key("d")
 	ba.Requests = nil
 	putArgs2 := roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyD}}
-	putArgs2.Sequence = 7
+	putArgs2.Sequence = 6
 	ba.Add(&putArgs2)
-	ba.Add(&roachpb.EndTransactionRequest{Commit: true})
+	etArgs := roachpb.EndTransactionRequest{Commit: true}
+	etArgs.Sequence = 7
+	ba.Add(&etArgs)
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
 		require.Equal(t, 5, len(ba.Requests))
@@ -217,9 +215,9 @@ func TestTxnPipelinerTrackOutstandingWrites(t *testing.T) {
 		require.Equal(t, keyA, qiReq1.Key)
 		require.Equal(t, keyB, qiReq2.Key)
 		require.Equal(t, keyC, qiReq3.Key)
-		require.Equal(t, int32(3), qiReq1.Txn.Sequence)
-		require.Equal(t, int32(4), qiReq2.Txn.Sequence)
-		require.Equal(t, int32(6), qiReq3.Txn.Sequence)
+		require.Equal(t, int32(2), qiReq1.Txn.Sequence)
+		require.Equal(t, int32(3), qiReq2.Txn.Sequence)
+		require.Equal(t, int32(5), qiReq3.Txn.Sequence)
 
 		br = ba.CreateReply()
 		br.Txn = ba.Txn
@@ -495,7 +493,8 @@ func TestTxnPipelinerManyWrites(t *testing.T) {
 	ctx := context.Background()
 	tp, mockSender := makeMockTxnPipeliner()
 
-	// Disable maxBatchSize limit.
+	// Disable maxOutstandingSize and maxBatchSize limits/
+	pipelinedWritesMaxOutstandingSize.Override(&tp.st.SV, math.MaxInt64)
 	pipelinedWritesMaxBatchSize.Override(&tp.st.SV, 0)
 
 	const writes = 2048
@@ -507,21 +506,16 @@ func TestTxnPipelinerManyWrites(t *testing.T) {
 	var ba roachpb.BatchRequest
 	ba.Header = roachpb.Header{Txn: &txn}
 	for i := 0; i < writes; i++ {
-		key := makeKey(i)
-		if i == 0 {
-			ba.Add(&roachpb.BeginTransactionRequest{RequestHeader: roachpb.RequestHeader{Key: key}})
-		}
-		putArgs := roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: key}}
+		putArgs := roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: makeKey(i)}}
 		putArgs.Sequence = makeSeq(i)
 		ba.Add(&putArgs)
 	}
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
-		require.Equal(t, writes+1, len(ba.Requests))
+		require.Equal(t, writes, len(ba.Requests))
 		require.True(t, ba.AsyncConsensus)
-		require.IsType(t, &roachpb.BeginTransactionRequest{}, ba.Requests[0].GetInner())
 		for i := 0; i < writes; i++ {
-			require.IsType(t, &roachpb.PutRequest{}, ba.Requests[i+1].GetInner())
+			require.IsType(t, &roachpb.PutRequest{}, ba.Requests[i].GetInner())
 		}
 
 		br := ba.CreateReply()
@@ -601,16 +595,14 @@ func TestTxnPipelinerTransactionAbort(t *testing.T) {
 
 	var ba roachpb.BatchRequest
 	ba.Header = roachpb.Header{Txn: &txn}
-	ba.Add(&roachpb.BeginTransactionRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
 	putArgs := roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}}
-	putArgs.Sequence = 2
+	putArgs.Sequence = 1
 	ba.Add(&putArgs)
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
-		require.Equal(t, 2, len(ba.Requests))
+		require.Equal(t, 1, len(ba.Requests))
 		require.True(t, ba.AsyncConsensus)
-		require.IsType(t, &roachpb.BeginTransactionRequest{}, ba.Requests[0].GetInner())
-		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[1].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[0].GetInner())
 
 		br := ba.CreateReply()
 		br.Txn = ba.Txn
@@ -629,7 +621,9 @@ func TestTxnPipelinerTransactionAbort(t *testing.T) {
 	// We'll unrealistically return a PENDING transaction, which won't allow
 	// the txnPipeliner to clean up.
 	ba.Requests = nil
-	ba.Add(&roachpb.EndTransactionRequest{Commit: false})
+	etArgs := roachpb.EndTransactionRequest{Commit: false}
+	etArgs.Sequence = 2
+	ba.Add(&etArgs)
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
 		require.Equal(t, 1, len(ba.Requests))
@@ -724,7 +718,7 @@ func TestTxnPipelinerIntentMissingError(t *testing.T) {
 				require.IsType(t, &roachpb.QueryIntentRequest{}, ba.Requests[5].GetInner())
 				require.IsType(t, &roachpb.PutRequest{}, ba.Requests[6].GetInner())
 
-				err := roachpb.NewIntentMissingError(nil)
+				err := roachpb.NewIntentMissingError(nil /* key */, nil /* intent */)
 				pErr := roachpb.NewErrorWithTxn(err, &txn)
 				pErr.SetErrorIndex(errIdx)
 				return nil, pErr
@@ -756,16 +750,14 @@ func TestTxnPipelinerEnableDisableMixTxn(t *testing.T) {
 
 	var ba roachpb.BatchRequest
 	ba.Header = roachpb.Header{Txn: &txn}
-	ba.Add(&roachpb.BeginTransactionRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
 	putArgs := roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}}
 	putArgs.Sequence = 1
 	ba.Add(&putArgs)
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
-		require.Equal(t, 2, len(ba.Requests))
+		require.Equal(t, 1, len(ba.Requests))
 		require.False(t, ba.AsyncConsensus)
-		require.IsType(t, &roachpb.BeginTransactionRequest{}, ba.Requests[0].GetInner())
-		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[1].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[0].GetInner())
 
 		br := ba.CreateReply()
 		br.Txn = ba.Txn
@@ -837,7 +829,9 @@ func TestTxnPipelinerEnableDisableMixTxn(t *testing.T) {
 	// Commit the txn. Again with pipeling disabled. Again, outstanding writes
 	// should be proven first.
 	ba.Requests = nil
-	ba.Add(&roachpb.EndTransactionRequest{Commit: true})
+	etArgs := roachpb.EndTransactionRequest{Commit: true}
+	etArgs.Sequence = 5
+	ba.Add(&etArgs)
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
 		require.Equal(t, 2, len(ba.Requests))
@@ -862,8 +856,186 @@ func TestTxnPipelinerEnableDisableMixTxn(t *testing.T) {
 	require.Equal(t, 0, tp.outstandingWritesLen())
 }
 
+// TestTxnPipelinerMaxOutstandingSize tests that batches are not pipelined if
+// doing so would push the memory used to track outstanding writes over the
+// limit allowed by the kv.transaction.write_pipelining_max_outstanding_size
+// setting.
+func TestTxnPipelinerMaxOutstandingSize(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
+	tp, mockSender := makeMockTxnPipeliner()
+
+	// Set maxOutstandingSize limit to 3 bytes.
+	pipelinedWritesMaxOutstandingSize.Override(&tp.st.SV, 3)
+
+	txn := makeTxnProto()
+	keyA, keyB := roachpb.Key("a"), roachpb.Key("b")
+	keyC, keyD := roachpb.Key("c"), roachpb.Key("d")
+
+	// Send a batch that would exceed the limit.
+	var ba roachpb.BatchRequest
+	ba.Header = roachpb.Header{Txn: &txn}
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyB}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyC}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyD}})
+
+	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
+		require.Equal(t, 4, len(ba.Requests))
+		require.False(t, ba.AsyncConsensus)
+
+		br := ba.CreateReply()
+		br.Txn = ba.Txn
+		return br, nil
+	})
+
+	br, pErr := tp.SendLocked(ctx, ba)
+	require.NotNil(t, br)
+	require.Nil(t, pErr)
+	require.Equal(t, int64(0), tp.owSizeBytes)
+
+	// Send a batch that is equal to the limit.
+	ba.Requests = nil
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyB}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyC}})
+
+	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
+		require.Equal(t, 3, len(ba.Requests))
+		require.True(t, ba.AsyncConsensus)
+
+		br = ba.CreateReply()
+		br.Txn = ba.Txn
+		return br, nil
+	})
+
+	br, pErr = tp.SendLocked(ctx, ba)
+	require.NotNil(t, br)
+	require.Nil(t, pErr)
+	require.Equal(t, int64(3), tp.owSizeBytes)
+
+	// Send a batch that would be under the limit if we weren't already at it.
+	ba.Requests = nil
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyD}})
+
+	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
+		require.Equal(t, 1, len(ba.Requests))
+		require.False(t, ba.AsyncConsensus)
+
+		br = ba.CreateReply()
+		br.Txn = ba.Txn
+		return br, nil
+	})
+
+	br, pErr = tp.SendLocked(ctx, ba)
+	require.NotNil(t, br)
+	require.Nil(t, pErr)
+	require.Equal(t, int64(3), tp.owSizeBytes)
+
+	// Send a batch that proves two of the outstanding writes.
+	ba.Requests = nil
+	ba.Add(&roachpb.GetRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
+	ba.Add(&roachpb.GetRequest{RequestHeader: roachpb.RequestHeader{Key: keyB}})
+
+	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
+		require.Equal(t, 4, len(ba.Requests))
+		require.False(t, ba.AsyncConsensus)
+		require.IsType(t, &roachpb.QueryIntentRequest{}, ba.Requests[0].GetInner())
+		require.IsType(t, &roachpb.GetRequest{}, ba.Requests[1].GetInner())
+		require.IsType(t, &roachpb.QueryIntentRequest{}, ba.Requests[2].GetInner())
+		require.IsType(t, &roachpb.GetRequest{}, ba.Requests[3].GetInner())
+
+		br = ba.CreateReply()
+		br.Txn = ba.Txn
+		br.Responses[0].GetInner().(*roachpb.QueryIntentResponse).FoundIntent = true
+		br.Responses[2].GetInner().(*roachpb.QueryIntentResponse).FoundIntent = true
+		return br, nil
+	})
+
+	br, pErr = tp.SendLocked(ctx, ba)
+	require.NotNil(t, br)
+	require.Nil(t, pErr)
+	require.Equal(t, int64(1), tp.owSizeBytes)
+
+	// Now that we're not up against the limit, send a batch that proves one
+	// write and immediately writes it again, along with a second write.
+	ba.Requests = nil
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyB}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyC}})
+
+	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
+		require.Equal(t, 3, len(ba.Requests))
+		require.True(t, ba.AsyncConsensus)
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[0].GetInner())
+		require.IsType(t, &roachpb.QueryIntentRequest{}, ba.Requests[1].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[2].GetInner())
+
+		br = ba.CreateReply()
+		br.Txn = ba.Txn
+		br.Responses[1].GetInner().(*roachpb.QueryIntentResponse).FoundIntent = true
+		return br, nil
+	})
+
+	br, pErr = tp.SendLocked(ctx, ba)
+	require.NotNil(t, br)
+	require.Nil(t, pErr)
+	require.Equal(t, int64(2), tp.owSizeBytes)
+
+	// Send the same batch again. Even though it would prove two outstanding
+	// writes while performing two others, we won't allow it to perform async
+	// consensus because the estimation is conservative.
+	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
+		require.Equal(t, 4, len(ba.Requests))
+		require.False(t, ba.AsyncConsensus)
+		require.IsType(t, &roachpb.QueryIntentRequest{}, ba.Requests[0].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[1].GetInner())
+		require.IsType(t, &roachpb.QueryIntentRequest{}, ba.Requests[2].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[3].GetInner())
+
+		br = ba.CreateReply()
+		br.Txn = ba.Txn
+		br.Responses[0].GetInner().(*roachpb.QueryIntentResponse).FoundIntent = true
+		br.Responses[2].GetInner().(*roachpb.QueryIntentResponse).FoundIntent = true
+		return br, nil
+	})
+
+	br, pErr = tp.SendLocked(ctx, ba)
+	require.NotNil(t, br)
+	require.Nil(t, pErr)
+	require.Equal(t, int64(0), tp.owSizeBytes)
+
+	// Increase maxOutstandingSize limit to 5 bytes.
+	pipelinedWritesMaxOutstandingSize.Override(&tp.st.SV, 5)
+
+	// The original batch with 4 writes should succeed.
+	ba.Requests = nil
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyB}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyC}})
+	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyD}})
+
+	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
+		require.Equal(t, 4, len(ba.Requests))
+		require.True(t, ba.AsyncConsensus)
+
+		br = ba.CreateReply()
+		br.Txn = ba.Txn
+		return br, nil
+	})
+
+	br, pErr = tp.SendLocked(ctx, ba)
+	require.NotNil(t, br)
+	require.Nil(t, pErr)
+	require.Equal(t, int64(4), tp.owSizeBytes)
+
+	// Bump the txn epoch. The outstanding bytes counter should reset.
+	tp.epochBumpedLocked()
+	require.Equal(t, int64(0), tp.owSizeBytes)
+}
+
 // TestTxnPipelinerMaxBatchSize tests that batches that contain more requests
-// than allowed by the maxBatchSize setting will not be pipelined.
+// than allowed by the kv.transaction.write_pipelining_max_batch_size setting
+// will not be pipelined.
 func TestTxnPipelinerMaxBatchSize(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	ctx := context.Background()
@@ -878,14 +1050,12 @@ func TestTxnPipelinerMaxBatchSize(t *testing.T) {
 	// Batch below limit.
 	var ba roachpb.BatchRequest
 	ba.Header = roachpb.Header{Txn: &txn}
-	ba.Add(&roachpb.BeginTransactionRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
 	ba.Add(&roachpb.PutRequest{RequestHeader: roachpb.RequestHeader{Key: keyA}})
 
 	mockSender.MockSend(func(ba roachpb.BatchRequest) (*roachpb.BatchResponse, *roachpb.Error) {
-		require.Equal(t, 2, len(ba.Requests))
+		require.Equal(t, 1, len(ba.Requests))
 		require.True(t, ba.AsyncConsensus)
-		require.IsType(t, &roachpb.BeginTransactionRequest{}, ba.Requests[0].GetInner())
-		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[1].GetInner())
+		require.IsType(t, &roachpb.PutRequest{}, ba.Requests[0].GetInner())
 
 		br := ba.CreateReply()
 		br.Txn = ba.Txn
