@@ -241,7 +241,7 @@ func (rh *ResponseHeader) combine(otherRH ResponseHeader) error {
 		rh.Txn = nil
 	}
 	if rh.ResumeSpan != nil {
-		panic(fmt.Sprintf("combining %+v with %+v", rh.ResumeSpan, otherRH.ResumeSpan))
+		return errors.Errorf("combining %+v with %+v", rh.ResumeSpan, otherRH.ResumeSpan)
 	}
 	rh.ResumeSpan = otherRH.ResumeSpan
 	rh.ResumeReason = otherRH.ResumeReason
@@ -256,7 +256,7 @@ func (sr *ScanResponse) combine(c combinable) error {
 	if sr != nil {
 		sr.Rows = append(sr.Rows, otherSR.Rows...)
 		sr.IntentRows = append(sr.IntentRows, otherSR.IntentRows...)
-		sr.BatchResponse = append(sr.BatchResponse, otherSR.BatchResponse...)
+		sr.BatchResponses = append(sr.BatchResponses, otherSR.BatchResponses...)
 		if err := sr.ResponseHeader.combine(otherSR.Header()); err != nil {
 			return err
 		}
@@ -272,7 +272,7 @@ func (sr *ReverseScanResponse) combine(c combinable) error {
 	if sr != nil {
 		sr.Rows = append(sr.Rows, otherSR.Rows...)
 		sr.IntentRows = append(sr.IntentRows, otherSR.IntentRows...)
-		sr.BatchResponse = append(sr.BatchResponse, otherSR.BatchResponse...)
+		sr.BatchResponses = append(sr.BatchResponses, otherSR.BatchResponses...)
 		if err := sr.ResponseHeader.combine(otherSR.Header()); err != nil {
 			return err
 		}
@@ -311,6 +311,7 @@ var _ combinable = &ResolveIntentRangeResponse{}
 func (cc *CheckConsistencyResponse) combine(c combinable) error {
 	if cc != nil {
 		otherCC := c.(*CheckConsistencyResponse)
+		cc.Result = append(cc.Result, otherCC.Result...)
 		if err := cc.ResponseHeader.combine(otherCC.Header()); err != nil {
 			return err
 		}
@@ -383,10 +384,12 @@ func (h *BatchResponse_Header) combine(o BatchResponse_Header) error {
 		)
 	}
 	h.Timestamp.Forward(o.Timestamp)
-	if h.Txn == nil {
-		h.Txn = o.Txn
-	} else {
-		h.Txn.Update(o.Txn)
+	if txn := o.Txn; txn != nil {
+		if h.Txn == nil {
+			h.Txn = txn.Clone()
+		} else {
+			h.Txn.Update(txn)
+		}
 	}
 	h.Now.Forward(o.Now)
 	h.CollectedSpans = append(h.CollectedSpans, o.CollectedSpans...)
@@ -873,7 +876,7 @@ func NewPutInline(key Key, value Value) Request {
 
 // NewConditionalPut returns a Request initialized to put value as a byte
 // slice at key if the existing value at key equals expValueBytes.
-func NewConditionalPut(key Key, value, expValue Value) Request {
+func NewConditionalPut(key Key, value, expValue Value, allowNotExist bool) Request {
 	value.InitChecksum(key)
 	var expValuePtr *Value
 	if expValue.RawBytes != nil {
@@ -886,8 +889,9 @@ func NewConditionalPut(key Key, value, expValue Value) Request {
 		RequestHeader: RequestHeader{
 			Key: key,
 		},
-		Value:    value,
-		ExpValue: expValuePtr,
+		Value:               value,
+		ExpValue:            expValuePtr,
+		AllowIfDoesNotExist: allowNotExist,
 	}
 }
 
@@ -965,19 +969,19 @@ func (*ConditionalPutRequest) flags() int {
 
 // InitPut, like ConditionalPut, effectively reads and may not write.
 // It also may return the actual data read on ConditionFailedErrors,
-// so must update the timestamp cache on errors. Unlike CPut, InitPuts
-// require a refresh because they may execute successfully without
-// leaving an intent (i.e., when the existing value is equal to the
-// proposed value).
+// so must update the timestamp cache on errors. InitPuts do not require
+// a refresh because on write-too-old errors, they return an error
+// immediately instead of continuing a serializable transaction to be
+// retried at end transaction.
 func (*InitPutRequest) flags() int {
-	return isRead | isWrite | isTxn | isTxnWrite | updatesReadTSCache | updatesTSCacheOnError | needsRefresh | consultsTSCache
+	return isRead | isWrite | isTxn | isTxnWrite | updatesReadTSCache | updatesTSCacheOnError | consultsTSCache
 }
 
 // Increment reads the existing value, but always leaves an intent so
-// it does not need to update the timestamp cache. It also does not
-// require refresh because on write-too-old errors experienced during
-// serializable transactions, increment returns the error instead of
-// continuing the transaction.
+// it does not need to update the timestamp cache. Increments do not
+// require a refresh because on write-too-old errors, they return an
+// error immediately instead of continuing a serializable transaction
+// to be retried at end transaction.
 func (*IncrementRequest) flags() int {
 	return isRead | isWrite | isTxn | isTxnWrite | consultsTSCache
 }
@@ -1015,7 +1019,7 @@ func (*ScanRequest) flags() int       { return isRead | isRange | isTxn | update
 func (*ReverseScanRequest) flags() int {
 	return isRead | isRange | isReverse | isTxn | updatesReadTSCache | needsRefresh
 }
-func (*BeginTransactionRequest) flags() int { return isWrite | isTxn | consultsTSCache }
+func (*BeginTransactionRequest) flags() int { return isWrite | isTxn }
 
 // EndTransaction updates the write timestamp cache to prevent
 // replays. Replays for the same transaction key and timestamp will
@@ -1028,8 +1032,13 @@ func (*AdminChangeReplicasRequest) flags() int { return isAdmin | isAlone }
 func (*AdminRelocateRangeRequest) flags() int  { return isAdmin | isAlone }
 func (*HeartbeatTxnRequest) flags() int        { return isWrite | isTxn }
 func (*GCRequest) flags() int                  { return isWrite | isRange }
-func (*PushTxnRequest) flags() int             { return isWrite | isAlone }
-func (*QueryTxnRequest) flags() int            { return isRead | isAlone }
+
+// PushTxnRequest updates the read timestamp cache when pushing a transaction's
+// timestamp and updates the write timestamp cache when aborting a transaction.
+func (*PushTxnRequest) flags() int {
+	return isWrite | isAlone | updatesReadTSCache | updatesWriteTSCache
+}
+func (*QueryTxnRequest) flags() int { return isRead | isAlone }
 
 // QueryIntent only updates the read timestamp cache when attempting
 // to prevent an intent that is found missing from ever being written
@@ -1039,10 +1048,7 @@ func (*ResolveIntentRequest) flags() int      { return isWrite }
 func (*ResolveIntentRangeRequest) flags() int { return isWrite | isRange }
 func (*TruncateLogRequest) flags() int        { return isWrite }
 func (*MergeRequest) flags() int              { return isWrite }
-
-func (*RequestLeaseRequest) flags() int {
-	return isWrite | isAlone | skipLeaseCheck
-}
+func (*RequestLeaseRequest) flags() int       { return isWrite | isAlone | skipLeaseCheck }
 
 // LeaseInfoRequest is usually executed in an INCONSISTENT batch, which has the
 // effect of the `skipLeaseCheck` flag that lease write operations have.
@@ -1071,14 +1077,22 @@ func (*ImportRequest) flags() int           { return isAdmin | isAlone }
 func (*AdminScatterRequest) flags() int     { return isAdmin | isAlone | isRange }
 func (*AddSSTableRequest) flags() int       { return isWrite | isAlone | isRange | isUnsplittable }
 
-// RefreshRequest and RefreshRangeRequest both list
-// updates(Read)TSCache, though they actually update the read or write
-// timestamp cache depending on the write parameter in the request.
-func (*RefreshRequest) flags() int      { return isRead | isTxn | updatesReadTSCache }
-func (*RefreshRangeRequest) flags() int { return isRead | isTxn | isRange | updatesReadTSCache }
-
-func (*SubsumeRequest) flags() int { return isRead | isAlone | updatesReadTSCache }
+// RefreshRequest and RefreshRangeRequest both determine which timestamp cache
+// they update based on their Write parameter.
+func (r *RefreshRequest) flags() int {
+	if r.Write {
+		return isRead | isTxn | updatesWriteTSCache
+	}
+	return isRead | isTxn | updatesReadTSCache
+}
+func (r *RefreshRangeRequest) flags() int {
+	if r.Write {
+		return isRead | isTxn | isRange | updatesWriteTSCache
+	}
+	return isRead | isTxn | isRange | updatesReadTSCache
+}
 
+func (*SubsumeRequest) flags() int    { return isRead | isAlone | updatesReadTSCache }
 func (*RangeStatsRequest) flags() int { return isRead }
 
 // Keys returns credentials in an aws.Config.
