@@ -20,9 +20,8 @@ import (
 	"fmt"
 	"strings"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
+	"github.com/pkg/errors"
 )
 
 //go:generate go run -tags gen-batch gen_batch.go
@@ -39,9 +38,22 @@ func (ba *BatchRequest) SetActiveTimestamp(nowFn func() hlc.Timestamp) error {
 			return errors.New("transactional request must not set batch timestamp")
 		}
 
-		// Always use the original timestamp for reads and writes, even
-		// though some intents may be written at higher timestamps in the
-		// event of a WriteTooOldError.
+		// The batch timestamp is the timestamp at which reads are performed. We set
+		// this to the txn's original timestamp, even if the txn's provisional
+		// commit timestamp has been forwarded, so that all reads within a txn
+		// observe the same snapshot of the database.
+		//
+		// In other words, we want to preserve the invariant that reading the same
+		// key multiple times in the same transaction will always return the same
+		// value. If we were to read at the latest provisional commit timestamp,
+		// txn.Timestamp, instead, reading the same key twice in the same txn might
+		// yield different results, e.g., if an intervening write caused the
+		// provisional commit timestamp to be advanced. Such a txn would fail to
+		// commit, as its reads would not successfully be refreshed, but only after
+		// confusing the client with spurious data.
+		//
+		// Note that writes will be performed at the provisional commit timestamp,
+		// txn.Timestamp, regardless of the batch timestamp.
 		ba.Timestamp = txn.OrigTimestamp
 		// If a refreshed timestamp is set for the transaction, forward
 		// the batch timestamp to it. The refreshed timestamp indicates a
@@ -60,18 +72,18 @@ func (ba *BatchRequest) SetActiveTimestamp(nowFn func() hlc.Timestamp) error {
 // UpdateTxn updates the batch transaction from the supplied one in
 // a copy-on-write fashion, i.e. without mutating an existing
 // Transaction struct.
-func (ba *BatchRequest) UpdateTxn(otherTxn *Transaction) {
-	if otherTxn == nil {
+func (ba *BatchRequest) UpdateTxn(o *Transaction) {
+	if o == nil {
 		return
 	}
-	otherTxn.AssertInitialized(context.TODO())
+	o.AssertInitialized(context.TODO())
 	if ba.Txn == nil {
-		ba.Txn = otherTxn
+		ba.Txn = o
 		return
 	}
 	clonedTxn := ba.Txn.Clone()
-	clonedTxn.Update(otherTxn)
-	ba.Txn = &clonedTxn
+	clonedTxn.Update(o)
+	ba.Txn = clonedTxn
 }
 
 // IsLeaseRequest returns whether the batch consists of a single RequestLease
@@ -105,7 +117,6 @@ func (ba *BatchRequest) IsReadOnly() bool {
 // leaseholders of the ranges it addresses.
 func (ba *BatchRequest) RequiresLeaseHolder() bool {
 	return !ba.IsReadOnly() || ba.Header.ReadConsistency.RequiresReadLease()
-
 }
 
 // IsReverse returns true iff the BatchRequest contains a reverse request.
@@ -113,22 +124,23 @@ func (ba *BatchRequest) IsReverse() bool {
 	return ba.hasFlag(isReverse)
 }
 
-// IsPossibleTransaction returns true iff the BatchRequest contains
-// requests that can be part of a transaction.
-func (ba *BatchRequest) IsPossibleTransaction() bool {
+// IsTransactional returns true iff the BatchRequest contains requests that can
+// be part of a transaction.
+func (ba *BatchRequest) IsTransactional() bool {
 	return ba.hasFlag(isTxn)
 }
 
+// IsAllTransactional returns true iff the BatchRequest contains only requests
+// that can be part of a transaction.
+func (ba *BatchRequest) IsAllTransactional() bool {
+	return ba.hasFlagForAll(isTxn)
+}
+
 // IsTransactionWrite returns true iff the BatchRequest contains a txn write.
 func (ba *BatchRequest) IsTransactionWrite() bool {
 	return ba.hasFlag(isTxnWrite)
 }
 
-// IsRange returns true iff the BatchRequest contains range-based requests.
-func (ba *BatchRequest) IsRange() bool {
-	return ba.hasFlag(isRange)
-}
-
 // IsUnsplittable returns true iff the BatchRequest an un-splittable request.
 func (ba *BatchRequest) IsUnsplittable() bool {
 	return ba.hasFlag(isUnsplittable)
@@ -205,6 +217,66 @@ func (ba *BatchRequest) IsSingleComputeChecksumRequest() bool {
 	return false
 }
 
+// IsSingleAddSSTableRequest returns true iff the batch contains a single
+// request, and that request is an AddSSTableRequest.
+func (ba *BatchRequest) IsSingleAddSSTableRequest() bool {
+	if ba.IsSingleRequest() {
+		_, ok := ba.Requests[0].GetInner().(*AddSSTableRequest)
+		return ok
+	}
+	return false
+}
+
+// IsCompleteTransaction determines whether a batch contains every write in a
+// transactions.
+func (ba *BatchRequest) IsCompleteTransaction() bool {
+	et, hasET := ba.GetArg(EndTransaction)
+	if !hasET || !et.(*EndTransactionRequest).Commit {
+		return false
+	}
+	if _, hasBegin := ba.GetArg(BeginTransaction); hasBegin {
+		// TODO(nvanbenschoten): Remove this condition in 2.3. It can be removed
+		// in 2.3 once we're sure that all nodes will properly set sequence
+		// numbers (i.e. on writes only).
+		return true
+	}
+	maxSeq := et.Header().Sequence
+	switch maxSeq {
+	case 0:
+		// If the batch isn't using sequence numbers,
+		// assume that it is not a complete transaction.
+		return false
+	case 1:
+		// The transaction performed no writes.
+		return true
+	}
+	if int(maxSeq) > len(ba.Requests) {
+		// Fast-path.
+		return false
+	}
+	// Check whether any sequence numbers were skipped between 1 and the
+	// EndTransaction's sequence number. A Batch is only a complete transaction
+	// if it contains every write that the transaction performed.
+	nextSeq := int32(1)
+	for _, args := range ba.Requests {
+		req := args.GetInner()
+		seq := req.Header().Sequence
+		if seq > nextSeq {
+			return false
+		}
+		if seq == nextSeq {
+			if !IsTransactionWrite(req) {
+				return false
+			}
+			nextSeq++
+			if nextSeq == maxSeq {
+				return true
+			}
+		}
+	}
+	panic("unreachable")
+}
+
 // GetPrevLeaseForLeaseRequest returns the previous lease, at the time
 // of proposal, for a request lease or transfer lease request. If the
 // batch does not contain a single lease request, this method will panic.
@@ -223,6 +295,20 @@ func (ba *BatchRequest) hasFlag(flag int) bool {
 	return false
 }
 
+// hasFlagForAll returns true iff all of the requests within the batch contains
+// the specified flag.
+func (ba *BatchRequest) hasFlagForAll(flag int) bool {
+	if len(ba.Requests) == 0 {
+		return false
+	}
+	for _, union := range ba.Requests {
+		if (union.GetInner().flags() & flag) == 0 {
+			return false
+		}
+	}
+	return true
+}
+
 // GetArg returns a request of the given type if one is contained in the
 // Batch. The request returned is the first of its kind, with the exception
 // of EndTransaction, where it examines the very last request only.
@@ -337,11 +423,13 @@ func actualSpan(req Request, resp Response) (Span, bool) {
 	return h.Span(), true
 }
 
-// Combine implements the Combinable interface. It combines each slot of the
-// given request into the corresponding slot of the base response. The number
-// of slots must be equal and the respective slots must be combinable.
+// Combine combines each slot of the given request into the corresponding slot
+// of the base response. The number of slots must be equal and the respective
+// slots must be combinable.
 // On error, the receiver BatchResponse is in an invalid state. In either case,
 // the supplied BatchResponse must not be used any more.
+// It is an error to call Combine on responses with errors in them. The
+// DistSender strips the errors from any responses that it combines.
 func (br *BatchResponse) Combine(otherBatch *BatchResponse, positions []int) error {
 	if err := br.BatchResponse_Header.combine(otherBatch.BatchResponse_Header); err != nil {
 		return err
@@ -365,7 +453,6 @@ func (br *BatchResponse) Combine(otherBatch *BatchResponse, positions []int) err
 			return errors.Errorf("can not combine %T and %T", valLeft, valRight)
 		}
 	}
-	br.Txn.Update(otherBatch.Txn)
 	return nil
 }
 
@@ -501,19 +588,15 @@ func (ba BatchRequest) String() string {
 			str = append(str, fmt.Sprintf("%s(commit:%t) [%s]", req.Method(), et.Commit, h.Key))
 		} else {
 			h := req.Header()
-			str = append(str, fmt.Sprintf("%s [%s,%s)", req.Method(), h.Key, h.EndKey))
+			var s string
+			if req.Method() == PushTxn {
+				pushReq := req.(*PushTxnRequest)
+				s = fmt.Sprintf("PushTxn(%s->%s)", pushReq.PusherTxn.Short(), pushReq.PusheeTxn.Short())
+			} else {
+				s = req.Method().String()
+			}
+			str = append(str, fmt.Sprintf("%s [%s,%s)", s, h.Key, h.EndKey))
 		}
 	}
 	return strings.Join(str, ", ")
 }
-
-// TODO(marc): we should assert
-// var _ security.RequestWithUser = &BatchRequest{}
-// here, but we need to break cycles first.
-
-// GetUser implements security.RequestWithUser.
-// KV messages are always sent by the node user.
-func (*BatchRequest) GetUser() string {
-	// TODO(marc): we should use security.NodeUser here, but we need to break cycles first.
-	return "node"
-}
