@@ -29,10 +29,9 @@ import (
 	"sync"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
+	"github.com/cockroachdb/cockroach/pkg/util/bitarray"
 	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
@@ -41,6 +40,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
 )
 
 var (
@@ -115,6 +115,11 @@ func (rk RKey) String() string {
 	return Key(rk).String()
 }
 
+// StringWithDirs - see Key.String.WithDirs.
+func (rk RKey) StringWithDirs(valDirs []encoding.Direction, maxLen int) string {
+	return Key(rk).StringWithDirs(valDirs, maxLen)
+}
+
 // Key is a custom type for a byte string in proto
 // messages which refer to Cockroach keys.
 type Key []byte
@@ -187,18 +192,28 @@ func (k Key) Compare(b Key) int {
 
 // String returns a string-formatted version of the key.
 func (k Key) String() string {
-	// Leave valDirs unspecified such that values are pretty-printed
-	// with default encoding direction.
-	return k.StringWithDirs(nil /* valDirs */)
+	return k.StringWithDirs(nil /* valDirs */, 0 /* maxLen */)
 }
 
 // StringWithDirs is the value encoding direction-aware version of String.
-func (k Key) StringWithDirs(valDirs []encoding.Direction) string {
+//
+// Args:
+// valDirs: The direction for the key's components, generally needed for correct
+// 	decoding. If nil, the values are pretty-printed with default encoding
+// 	direction.
+// maxLen: If not 0, only the first maxLen chars from the decoded key are
+//   returned, plus a "..." suffix.
+func (k Key) StringWithDirs(valDirs []encoding.Direction, maxLen int) string {
+	var s string
 	if PrettyPrintKey != nil {
-		return PrettyPrintKey(valDirs, k)
+		s = PrettyPrintKey(valDirs, k)
+	} else {
+		s = fmt.Sprintf("%q", []byte(k))
 	}
-
-	return fmt.Sprintf("%q", []byte(k))
+	if maxLen != 0 && len(s) > maxLen {
+		return s[0:maxLen] + "..."
+	}
+	return s
 }
 
 // Format implements the fmt.Formatter interface.
@@ -430,6 +445,15 @@ func (v *Value) SetDuration(t duration.Duration) error {
 	return nil
 }
 
+// SetBitArray encodes the specified bit array value into the bytes field of the
+// receiver, sets the tag and clears the checksum.
+func (v *Value) SetBitArray(t bitarray.BitArray) {
+	words, _ := t.EncodingParts()
+	v.RawBytes = make([]byte, headerSize, headerSize+encoding.NonsortingUvarintMaxLen+8*len(words))
+	v.RawBytes = encoding.EncodeUntaggedBitArrayValue(v.RawBytes, t)
+	v.setTag(ValueType_BITARRAY)
+}
+
 // SetDecimal encodes the specified decimal value into the bytes field of
 // the receiver using Gob encoding, sets the tag and clears the checksum.
 func (v *Value) SetDecimal(dec *apd.Decimal) error {
@@ -544,6 +568,16 @@ func (v Value) GetDuration() (duration.Duration, error) {
 	return t, err
 }
 
+// GetBitArray decodes a bit array value from the bytes field of the receiver. If
+// the tag is not BITARRAY an error will be returned.
+func (v Value) GetBitArray() (bitarray.BitArray, error) {
+	if tag := v.GetTag(); tag != ValueType_BITARRAY {
+		return bitarray.BitArray{}, fmt.Errorf("value type is not %s: %s", ValueType_BITARRAY, tag)
+	}
+	_, t, err := encoding.DecodeUntaggedBitArrayValue(v.dataBytes())
+	return t, err
+}
+
 // GetDecimal decodes a decimal value from the bytes of the receiver. If the
 // tag is not DECIMAL an error will be returned.
 func (v Value) GetDecimal() (apd.Decimal, error) {
@@ -553,6 +587,16 @@ func (v Value) GetDecimal() (apd.Decimal, error) {
 	return encoding.DecodeNonsortingDecimal(v.dataBytes(), nil)
 }
 
+// GetDecimalInto decodes a decimal value from the bytes of the receiver,
+// writing it directly into the provided non-null apd.Decimal. If the
+// tag is not DECIMAL an error will be returned.
+func (v Value) GetDecimalInto(d *apd.Decimal) error {
+	if tag := v.GetTag(); tag != ValueType_DECIMAL {
+		return fmt.Errorf("value type is not %s: %s", ValueType_DECIMAL, tag)
+	}
+	return encoding.DecodeIntoNonsortingDecimal(d, v.dataBytes(), nil)
+}
+
 // GetTimeseries decodes an InternalTimeSeriesData value from the bytes
 // field of the receiver. An error will be returned if the tag is not
 // TIMESERIES or if decoding fails.
@@ -661,6 +705,11 @@ func (v Value) PrettyPrint() string {
 			buf.WriteString("0x")
 			buf.WriteString(hex.EncodeToString(data))
 		}
+	case ValueType_BITARRAY:
+		var data bitarray.BitArray
+		data, err = v.GetBitArray()
+		buf.WriteByte('B')
+		data.Format(&buf)
 	case ValueType_TIME:
 		var t time.Time
 		t, err = v.GetTime()
@@ -672,7 +721,7 @@ func (v Value) PrettyPrint() string {
 	case ValueType_DURATION:
 		var d duration.Duration
 		d, err = v.GetDuration()
-		buf.WriteString(d.String())
+		buf.WriteString(d.StringNanos())
 	default:
 		err = errors.Errorf("unknown tag: %s", t)
 	}
@@ -703,14 +752,9 @@ const (
 // baseKey can be nil, in which case it will be set when sending the first
 // write.
 func MakeTransaction(
-	name string,
-	baseKey Key,
-	userPriority UserPriority,
-	isolation enginepb.IsolationType,
-	now hlc.Timestamp,
-	maxOffsetNs int64,
+	name string, baseKey Key, userPriority UserPriority, now hlc.Timestamp, maxOffsetNs int64,
 ) Transaction {
-	u := uuid.MakeV4()
+	u := uuid.FastMakeV4()
 	var maxTS hlc.Timestamp
 	if maxOffsetNs == timeutil.ClocklessMaxOffset {
 		// For clockless reads, use the largest possible maxTS. This means we'll
@@ -725,7 +769,6 @@ func MakeTransaction(
 		TxnMeta: enginepb.TxnMeta{
 			Key:       baseKey,
 			ID:        u,
-			Isolation: isolation,
 			Timestamp: now,
 			Priority:  MakePriority(userPriority),
 			Sequence:  0, // 1-indexed, incremented before each Request
@@ -740,7 +783,7 @@ func MakeTransaction(
 // MakeTxnCoordMeta creates a new transaction coordinator meta for the given
 // transaction.
 func MakeTxnCoordMeta(txn Transaction) TxnCoordMeta {
-	return TxnCoordMeta{Txn: txn, DeprecatedRefreshValid: true}
+	return TxnCoordMeta{Txn: txn}
 }
 
 // StripRootToLeaf strips out all information that is unnecessary to communicate
@@ -770,26 +813,15 @@ func (t Transaction) LastActive() hlc.Timestamp {
 	return ts
 }
 
-// Clone creates a copy of the given transaction. The copy is "mostly" deep,
-// but does share pieces of memory with the original such as Key, ID and the
-// keys with the intent spans.
-func (t Transaction) Clone() Transaction {
-	mt := t.ObservedTimestamps
-	if mt != nil {
-		t.ObservedTimestamps = make([]ObservedTimestamp, len(mt))
-		copy(t.ObservedTimestamps, mt)
-	}
-	// Note that we're not cloning the span keys under the assumption that the
-	// keys themselves are not mutable.
-	t.Intents = append([]Span(nil), t.Intents...)
-	return t
+// Clone creates a copy of the given transaction. The copy is shallow because
+// none of the references held by a transaction allow interior mutability.
+func (t Transaction) Clone() *Transaction {
+	return &t
 }
 
 // AssertInitialized crashes if the transaction is not initialized.
 func (t *Transaction) AssertInitialized(ctx context.Context) {
-	if t.ID == (uuid.UUID{}) ||
-		t.OrigTimestamp == (hlc.Timestamp{}) ||
-		t.Timestamp == (hlc.Timestamp{}) {
+	if t.ID == (uuid.UUID{}) || t.Timestamp == (hlc.Timestamp{}) {
 		log.Fatalf(ctx, "uninitialized txn: %s", t)
 	}
 }
@@ -871,9 +903,9 @@ func MakePriority(userPriority UserPriority) int32 {
 	// For userPriority=MaxUserPriority, the probability of overflow is 0.7%.
 	// For userPriority=(MaxUserPriority/2), the probability of overflow is 0.005%.
 	val = (val / (5 * float64(MaxUserPriority))) * math.MaxInt32
-	if val <= MinTxnPriority {
+	if val < MinTxnPriority+1 {
 		return MinTxnPriority + 1
-	} else if val >= MaxTxnPriority {
+	} else if val > MaxTxnPriority-1 {
 		return MaxTxnPriority - 1
 	}
 	return int32(val)
@@ -899,11 +931,10 @@ func (t *Transaction) Restart(
 	t.UpgradePriority(MakePriority(userPriority))
 	t.UpgradePriority(upgradePriority)
 	t.WriteTooOld = false
-	t.RetryOnPush = false
 	t.Sequence = 0
 	// Reset Writing. Since we're using a new epoch, we don't care about the abort
 	// cache.
-	t.Writing = false
+	t.DeprecatedWriting = false
 }
 
 // BumpEpoch increments the transaction's epoch, allowing for an in-place
@@ -921,7 +952,7 @@ func (t *Transaction) BumpEpoch() {
 func (t *Transaction) InclusiveTimeBounds() (hlc.Timestamp, hlc.Timestamp) {
 	min := t.OrigTimestamp
 	max := t.Timestamp
-	if t.Epoch != 0 {
+	if t.Epoch != 0 && t.EpochZeroTimestamp != (hlc.Timestamp{}) {
 		if min.Less(t.EpochZeroTimestamp) {
 			panic(fmt.Sprintf("orig timestamp %s less than epoch zero %s", min, t.EpochZeroTimestamp))
 		}
@@ -939,7 +970,7 @@ func (t *Transaction) Update(o *Transaction) {
 	}
 	o.AssertInitialized(context.TODO())
 	if t.ID == (uuid.UUID{}) {
-		*t = o.Clone()
+		*t = *o
 		return
 	}
 	if len(t.Key) == 0 {
@@ -950,14 +981,12 @@ func (t *Transaction) Update(o *Transaction) {
 	}
 
 	// If the epoch or refreshed timestamp move forward, overwrite
-	// WriteTooOld and RetryOnPush, otherwise the flags are cumulative.
+	// WriteTooOld, otherwise the flags are cumulative.
 	if t.Epoch < o.Epoch || t.RefreshedTimestamp.Less(o.RefreshedTimestamp) {
 		t.WriteTooOld = o.WriteTooOld
-		t.RetryOnPush = o.RetryOnPush
 		t.OrigTimestampWasObserved = o.OrigTimestampWasObserved
 	} else {
 		t.WriteTooOld = t.WriteTooOld || o.WriteTooOld
-		t.RetryOnPush = t.RetryOnPush || o.RetryOnPush
 		t.OrigTimestampWasObserved = t.OrigTimestampWasObserved || o.OrigTimestampWasObserved
 	}
 
@@ -979,7 +1008,7 @@ func (t *Transaction) Update(o *Transaction) {
 
 	// We can't assert against regression here since it can actually happen
 	// that we update from a transaction which isn't Writing.
-	t.Writing = t.Writing || o.Writing
+	t.DeprecatedWriting = t.DeprecatedWriting || o.DeprecatedWriting
 
 	if t.Sequence < o.Sequence {
 		t.Sequence = o.Sequence
@@ -1005,10 +1034,10 @@ func (t *Transaction) UpgradePriority(minPriority int32) {
 	}
 }
 
-// IsSerializable returns whether this transaction uses serializable
-// isolation.
-func (t *Transaction) IsSerializable() bool {
-	return t != nil && t.Isolation == enginepb.SERIALIZABLE
+// IsWriting returns whether the transaction has begun writing intents.
+// This method will never return true for a read-only transaction.
+func (t *Transaction) IsWriting() bool {
+	return t.Key != nil
 }
 
 // String formats transaction into human readable string.
@@ -1021,10 +1050,10 @@ func (t Transaction) String() string {
 	if len(t.Name) > 0 {
 		fmt.Fprintf(&buf, "%q ", t.Name)
 	}
-	fmt.Fprintf(&buf, "id=%s key=%s rw=%t pri=%.8f iso=%s stat=%s epo=%d "+
-		"ts=%s orig=%s max=%s wto=%t rop=%t seq=%d",
-		t.Short(), Key(t.Key), t.Writing, floatPri, t.Isolation, t.Status, t.Epoch, t.Timestamp,
-		t.OrigTimestamp, t.MaxTimestamp, t.WriteTooOld, t.RetryOnPush, t.Sequence)
+	fmt.Fprintf(&buf, "id=%s key=%s rw=%t pri=%.8f stat=%s epo=%d "+
+		"ts=%s orig=%s max=%s wto=%t seq=%d",
+		t.Short(), Key(t.Key), t.IsWriting(), floatPri, t.Status, t.Epoch, t.Timestamp,
+		t.OrigTimestamp, t.MaxTimestamp, t.WriteTooOld, t.Sequence)
 	if ni := len(t.Intents); t.Status != PENDING && ni > 0 {
 		fmt.Fprintf(&buf, " int=%d", ni)
 	}
@@ -1044,7 +1073,7 @@ func (t Transaction) SafeMessage() string {
 	}
 	fmt.Fprintf(&buf, "id=%s rw=%t pri=%.8f stat=%s epo=%d "+
 		"ts=%s orig=%s max=%s wto=%t seq=%d",
-		t.Short(), t.Writing, floatPri, t.Status, t.Epoch, t.Timestamp,
+		t.Short(), t.IsWriting(), floatPri, t.Status, t.Epoch, t.Timestamp,
 		t.OrigTimestamp, t.MaxTimestamp, t.WriteTooOld, t.Sequence)
 	if ni := len(t.Intents); t.Status != PENDING && ni > 0 {
 		fmt.Fprintf(&buf, " int=%d", ni)
@@ -1081,11 +1110,36 @@ func (t *Transaction) UpdateObservedTimestamp(nodeID NodeID, maxTS hlc.Timestamp
 // given node's clock during the transaction. The returned boolean is false if
 // no observation about the requested node was found. Otherwise, MaxTimestamp
 // can be lowered to the returned timestamp when reading from nodeID.
-func (t Transaction) GetObservedTimestamp(nodeID NodeID) (hlc.Timestamp, bool) {
+func (t *Transaction) GetObservedTimestamp(nodeID NodeID) (hlc.Timestamp, bool) {
 	s := observedTimestampSlice(t.ObservedTimestamps)
 	return s.get(nodeID)
 }
 
+// AsRecord returns a TransactionRecord object containing only the subset of
+// fields from the receiver that must be persisted in the transaction record.
+func (t *Transaction) AsRecord() TransactionRecord {
+	var tr TransactionRecord
+	tr.TxnMeta = t.TxnMeta
+	tr.Status = t.Status
+	tr.LastHeartbeat = t.LastHeartbeat
+	tr.OrigTimestamp = t.OrigTimestamp
+	tr.Intents = t.Intents
+	return tr
+}
+
+// AsTransaction returns a Transaction object containing populated fields for
+// state in the transaction record and empty fields for state omitted from the
+// transaction record.
+func (tr *TransactionRecord) AsTransaction() Transaction {
+	var t Transaction
+	t.TxnMeta = tr.TxnMeta
+	t.Status = tr.Status
+	t.LastHeartbeat = tr.LastHeartbeat
+	t.OrigTimestamp = tr.OrigTimestamp
+	t.Intents = tr.Intents
+	return t
+}
+
 // PrepareTransactionForRetry returns a new Transaction to be used for retrying
 // the original Transaction. Depending on the error, this might return an
 // already-existing Transaction with an incremented epoch, or a completely new
@@ -1111,7 +1165,7 @@ func PrepareTransactionForRetry(
 		log.Fatalf(ctx, "missing txn for retryable error: %s", pErr)
 	}
 
-	txn := pErr.GetTxn().Clone()
+	txn := *pErr.GetTxn()
 	aborted := false
 	switch tErr := pErr.GetDetail().(type) {
 	case *TransactionAbortedError:
@@ -1121,21 +1175,17 @@ func PrepareTransactionForRetry(
 		// TODO(andrei): Should we preserve the ObservedTimestamps across the
 		// restart?
 		errTxnPri := txn.Priority
-		// The OrigTimestamp of the new transaction is going to be the greater of
-		// two the current clock and the timestamp received in the error.
-		// TODO(andrei): Can we just use the clock since it has already been
-		// advanced to at least the error's timestamp?
+		// Start the new transaction at the current time from the local clock.
+		// The local hlc should have been advanced to at least the error's
+		// timestamp already.
 		now := clock.Now()
-		newTxnTimestamp := now
-		newTxnTimestamp.Forward(txn.Timestamp)
 		txn = MakeTransaction(
 			txn.Name,
 			nil, // baseKey
 			// We have errTxnPri, but this wants a UserPriority. So we're going to
 			// overwrite the priority below.
 			NormalUserPriority,
-			txn.Isolation,
-			newTxnTimestamp,
+			now,
 			clock.MaxOffset().Nanoseconds(),
 		)
 		// Use the priority communicated back by the server.
@@ -1172,7 +1222,7 @@ func CanTransactionRetryAtRefreshedTimestamp(
 	ctx context.Context, pErr *Error,
 ) (bool, *Transaction) {
 	txn := pErr.GetTxn()
-	if !txn.IsSerializable() || txn.OrigTimestampWasObserved {
+	if txn == nil || txn.OrigTimestampWasObserved {
 		return false, nil
 	}
 	timestamp := txn.Timestamp
@@ -1200,7 +1250,7 @@ func CanTransactionRetryAtRefreshedTimestamp(
 	newTxn.RefreshedTimestamp.Forward(newTxn.Timestamp)
 	newTxn.WriteTooOld = false
 
-	return true, &newTxn
+	return true, newTxn
 }
 
 func readWithinUncertaintyIntervalRetryTimestamp(
@@ -1213,7 +1263,7 @@ func readWithinUncertaintyIntervalRetryTimestamp(
 	if !ok {
 		log.Fatalf(ctx,
 			"missing observed timestamp for node %d found on uncertainty restart. "+
-				"err: %s. txn: %s. Observed timestamps: %s",
+				"err: %s. txn: %s. Observed timestamps: %v",
 			origin, err, txn, txn.ObservedTimestamps)
 	}
 	// Also forward by the existing timestamp.
@@ -1502,6 +1552,11 @@ func (s Span) ContainsKey(key Key) bool {
 	return bytes.Compare(key, s.Key) >= 0 && bytes.Compare(key, s.EndKey) < 0
 }
 
+// ProperlyContainsKey returns whether the span properly contains the given key.
+func (s Span) ProperlyContainsKey(key Key) bool {
+	return bytes.Compare(key, s.Key) > 0 && bytes.Compare(key, s.EndKey) < 0
+}
+
 // AsRange returns the Span as an interval.Range.
 func (s Span) AsRange() interval.Range {
 	startKey := s.Key
@@ -1667,7 +1722,8 @@ func (kv KeyValueByKey) Swap(i, j int) {
 
 var _ sort.Interface = KeyValueByKey{}
 
-// observedTimestampSlice maintains a sorted list of observed timestamps.
+// observedTimestampSlice maintains an immutable sorted list of observed
+// timestamps.
 type observedTimestampSlice []ObservedTimestamp
 
 func (s observedTimestampSlice) index(nodeID NodeID) int {
@@ -1689,19 +1745,27 @@ func (s observedTimestampSlice) get(nodeID NodeID) (hlc.Timestamp, bool) {
 }
 
 // update the timestamp for the specified node, or add a new entry in the
-// correct (sorted) location.
+// correct (sorted) location. The receiver is not mutated.
 func (s observedTimestampSlice) update(
 	nodeID NodeID, timestamp hlc.Timestamp,
 ) observedTimestampSlice {
 	i := s.index(nodeID)
 	if i < len(s) && s[i].NodeID == nodeID {
 		if timestamp.Less(s[i].Timestamp) {
-			s[i].Timestamp = timestamp
+			// The input slice is immutable, so copy and update.
+			cpy := make(observedTimestampSlice, len(s))
+			copy(cpy, s)
+			cpy[i].Timestamp = timestamp
+			return cpy
 		}
 		return s
 	}
-	s = append(s, ObservedTimestamp{})
-	copy(s[i+1:], s[i:])
-	s[i] = ObservedTimestamp{NodeID: nodeID, Timestamp: timestamp}
-	return s
+	// The input slice is immutable, so copy and update. Don't append to
+	// avoid an allocation. Doing so could invalidate a previous update
+	// to this receiver.
+	cpy := make(observedTimestampSlice, len(s)+1)
+	copy(cpy[:i], s[:i])
+	cpy[i] = ObservedTimestamp{NodeID: nodeID, Timestamp: timestamp}
+	copy(cpy[i+1:], s[i:])
+	return cpy
 }
