@@ -24,18 +24,18 @@ import (
 	"testing"
 	"time"
 
-	"github.com/kr/pretty"
-
 	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils/zerofields"
 	"github.com/cockroachdb/cockroach/pkg/util"
+	"github.com/cockroachdb/cockroach/pkg/util/bitarray"
 	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/randutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/kr/pretty"
 )
 
 func makeTS(walltime int64, logical int32) hlc.Timestamp {
@@ -386,7 +386,7 @@ func TestSetGetChecked(t *testing.T) {
 
 func TestTransactionBumpEpoch(t *testing.T) {
 	origNow := makeTS(10, 1)
-	txn := MakeTransaction("test", Key("a"), 1, enginepb.SERIALIZABLE, origNow, 0)
+	txn := MakeTransaction("test", Key("a"), 1, origNow, 0)
 	// Advance the txn timestamp.
 	txn.Timestamp.Add(10, 2)
 	txn.BumpEpoch()
@@ -407,7 +407,7 @@ func TestTransactionInclusiveTimeBounds(t *testing.T) {
 		}
 	}
 	origNow := makeTS(1, 1)
-	txn := MakeTransaction("test", Key("a"), 1, enginepb.SERIALIZABLE, origNow, 0)
+	txn := MakeTransaction("test", Key("a"), 1, origNow, 0)
 	verify(txn, origNow, origNow)
 	txn.Timestamp.Forward(makeTS(1, 2))
 	verify(txn, origNow, makeTS(1, 2))
@@ -475,7 +475,6 @@ func TestFastPathObservedTimestamp(t *testing.T) {
 
 var nonZeroTxn = Transaction{
 	TxnMeta: enginepb.TxnMeta{
-		Isolation: enginepb.SNAPSHOT,
 		Key:       Key("foo"),
 		ID:        uuid.MakeV4(),
 		Epoch:     2,
@@ -490,9 +489,8 @@ var nonZeroTxn = Transaction{
 	RefreshedTimestamp:       makeTS(20, 22),
 	MaxTimestamp:             makeTS(40, 41),
 	ObservedTimestamps:       []ObservedTimestamp{{NodeID: 1, Timestamp: makeTS(1, 2)}},
-	Writing:                  true,
+	DeprecatedWriting:        true,
 	WriteTooOld:              true,
-	RetryOnPush:              true,
 	Intents:                  []Span{{Key: []byte("a"), EndKey: []byte("b")}},
 	EpochZeroTimestamp:       makeTS(1, 1),
 	OrigTimestampWasObserved: true,
@@ -514,7 +512,6 @@ func TestTransactionUpdate(t *testing.T) {
 	var txn3 Transaction
 	txn3.ID = uuid.MakeV4()
 	txn3.Name = "carl"
-	txn3.Isolation = enginepb.SNAPSHOT
 	txn3.Update(&txn)
 
 	if err := zerofields.NoZeroField(txn3); err != nil {
@@ -541,7 +538,8 @@ func TestTransactionUpdateEpochZero(t *testing.T) {
 }
 
 func TestTransactionClone(t *testing.T) {
-	txn := nonZeroTxn.Clone()
+	txnPtr := nonZeroTxn.Clone()
+	txn := *txnPtr
 
 	fields := util.EqualPtrFields(reflect.ValueOf(nonZeroTxn), reflect.ValueOf(txn), "")
 	sort.Strings(fields)
@@ -550,8 +548,10 @@ func TestTransactionClone(t *testing.T) {
 	// listed below. If this test fails, please update the list below and/or
 	// Transaction.Clone().
 	expFields := []string{
+		"Intents",
 		"Intents.EndKey",
 		"Intents.Key",
+		"ObservedTimestamps",
 		"TxnMeta.Key",
 	}
 	if !reflect.DeepEqual(expFields, fields) {
@@ -562,6 +562,80 @@ func TestTransactionClone(t *testing.T) {
 	}
 }
 
+// TestTransactionRecordRoundtrips tests a few properties about Transaction
+// and TransactionRecord protos. Remember that the latter is wire compatible
+// with the former and contains a subset of its protos.
+//
+// Assertions:
+// 1. Transaction->TransactionRecord->Transaction is lossless for the fields
+//    in TransactionRecord. It drops all other fields.
+// 2. TransactionRecord->Transaction->TransactionRecord is lossless.
+//    Fields not in TransactionRecord are set as zero values.
+// 3. Transaction messages can be decoded as TransactionRecord messages.
+//    Fields not in TransactionRecord are dropped.
+// 4. TransactionRecord messages can be decoded as Transaction messages.
+//    Fields not in TransactionRecord are decoded as zero values.
+func TestTransactionRecordRoundtrips(t *testing.T) {
+	// Verify that converting from a Transaction to a TransactionRecord
+	// strips out fields but is lossless for the desired fields.
+	txn := nonZeroTxn
+	txnRecord := txn.AsRecord()
+	if err := zerofields.NoZeroField(txnRecord); err != nil {
+		t.Fatal(err)
+	}
+	if !reflect.DeepEqual(txnRecord.TxnMeta, txn.TxnMeta) {
+		t.Fatalf("txnRecord.TxnMeta = %v, txn.TxnMeta = %v", txnRecord.TxnMeta, txn.TxnMeta)
+	}
+	if !reflect.DeepEqual(txnRecord.Status, txn.Status) {
+		t.Fatalf("txnRecord.Status = %v, txn.Status = %v", txnRecord.Status, txn.Status)
+	}
+	if !reflect.DeepEqual(txnRecord.LastHeartbeat, txn.LastHeartbeat) {
+		t.Fatalf("txnRecord.LastHeartbeat = %v, txn.LastHeartbeat = %v", txnRecord.LastHeartbeat, txn.LastHeartbeat)
+	}
+	if !reflect.DeepEqual(txnRecord.OrigTimestamp, txn.OrigTimestamp) {
+		t.Fatalf("txnRecord.OrigTimestamp = %v, txn.OrigTimestamp = %v", txnRecord.OrigTimestamp, txn.OrigTimestamp)
+	}
+	if !reflect.DeepEqual(txnRecord.Intents, txn.Intents) {
+		t.Fatalf("txnRecord.Intents = %v, txn.Intents = %v", txnRecord.Intents, txn.Intents)
+	}
+
+	// Verify that converting through a Transaction message and back
+	// to a TransactionRecord is a lossless round trip.
+	txn2 := txnRecord.AsTransaction()
+	txnRecord2 := txn2.AsRecord()
+	if !reflect.DeepEqual(txnRecord, txnRecord2) {
+		t.Fatalf("txnRecord = %v, txnRecord2 = %v", txnRecord, txnRecord2)
+	}
+
+	// Verify that encoded Transaction messages can be decoded as
+	// TransactionRecord messages.
+	txnBytes, err := protoutil.Marshal(&txn)
+	if err != nil {
+		t.Fatal(err)
+	}
+	var txnRecord3 TransactionRecord
+	if err := protoutil.Unmarshal(txnBytes, &txnRecord3); err != nil {
+		t.Fatal(err)
+	}
+	if !reflect.DeepEqual(txnRecord, txnRecord3) {
+		t.Fatalf("txnRecord = %v, txnRecord3 = %v", txnRecord, txnRecord3)
+	}
+
+	// Verify that encoded TransactionRecord messages can be decoded
+	// as Transaction messages.
+	txnRecordBytes, err := protoutil.Marshal(&txnRecord)
+	if err != nil {
+		t.Fatal(err)
+	}
+	var txn3 Transaction
+	if err := protoutil.Unmarshal(txnRecordBytes, &txn3); err != nil {
+		t.Fatal(err)
+	}
+	if !reflect.DeepEqual(txn2, txn3) {
+		t.Fatalf("txn2 = %v, txn3 = %v", txn2, txn3)
+	}
+}
+
 // checkVal verifies if a value is close to an expected value, within a fraction (e.g. if
 // fraction=0.1, it checks if val is within 10% of expected).
 func checkVal(val, expected, errFraction float64) bool {
@@ -597,8 +671,15 @@ func TestMakePriority(t *testing.T) {
 	const trials = 100000
 	values := make([][trials]int32, len(userPs))
 	for i, userPri := range userPs {
-		for t := 0; t < trials; t++ {
-			values[i][t] = MakePriority(userPri)
+		for tr := 0; tr < trials; tr++ {
+			p := MakePriority(userPri)
+			if p == MinTxnPriority {
+				t.Fatalf("unexpected min txn priority")
+			}
+			if p == MaxTxnPriority {
+				t.Fatalf("unexpected max txn priority")
+			}
+			values[i][tr] = p
 		}
 	}
 
@@ -760,11 +841,16 @@ func TestLeaseEqual(t *testing.T) {
 		ProposedTS            *hlc.Timestamp
 		Epoch                 int64
 		Sequence              LeaseSequence
+		XXX_NoUnkeyedLiteral  struct{}
+		XXX_sizecache         int32
 	}
 	// Verify that the lease structure does not change unexpectedly. If a compile
 	// error occurs on the following line of code, update the expectedLease
 	// structure AND update Lease.Equal.
 	var _ = expectedLease(Lease{})
+	// Appease the linter.
+	var _ = expectedLease{}.XXX_NoUnkeyedLiteral
+	var _ = expectedLease{}.XXX_sizecache
 
 	// Verify that nil == &hlc.Timestamp{} for the Expiration and
 	// DeprecatedStartStasis fields. See #19843.
@@ -906,28 +992,28 @@ func TestSpanContains(t *testing.T) {
 	s := Span{Key: []byte("a"), EndKey: []byte("b")}
 
 	testData := []struct {
-		start, end []byte
+		start, end string
 		contains   bool
 	}{
 		// Single keys.
-		{[]byte("a"), nil, true},
-		{[]byte("aa"), nil, true},
-		{[]byte("`"), nil, false},
-		{[]byte("b"), nil, false},
-		{[]byte("c"), nil, false},
+		{"a", "", true},
+		{"aa", "", true},
+		{"`", "", false},
+		{"b", "", false},
+		{"c", "", false},
 		// Key ranges.
-		{[]byte("a"), []byte("b"), true},
-		{[]byte("a"), []byte("aa"), true},
-		{[]byte("aa"), []byte("b"), true},
-		{[]byte("0"), []byte("9"), false},
-		{[]byte("`"), []byte("a"), false},
-		{[]byte("b"), []byte("bb"), false},
-		{[]byte("0"), []byte("bb"), false},
-		{[]byte("aa"), []byte("bb"), false},
-		{[]byte("b"), []byte("a"), false},
+		{"a", "b", true},
+		{"a", "aa", true},
+		{"aa", "b", true},
+		{"0", "9", false},
+		{"`", "a", false},
+		{"b", "bb", false},
+		{"0", "bb", false},
+		{"aa", "bb", false},
+		{"b", "a", false},
 	}
 	for i, test := range testData {
-		if s.Contains(Span{test.start, test.end}) != test.contains {
+		if s.Contains(sp(test.start, test.end)) != test.contains {
 			t.Errorf("%d: expected span %q-%q within range to be %v",
 				i, test.start, test.end, test.contains)
 		}
@@ -970,8 +1056,8 @@ func TestSpanSplitOnKey(t *testing.T) {
 		// Simple split.
 		{
 			[]byte("bb"),
-			Span{[]byte("b"), []byte("bb")},
-			Span{[]byte("bb"), []byte("c")},
+			sp("b", "bb"),
+			sp("bb", "c"),
 		},
 	}
 	for testIdx, test := range testData {
@@ -1001,7 +1087,7 @@ func TestSpanValid(t *testing.T) {
 		{[]byte("b"), []byte("aa"), false},
 	}
 	for i, test := range testData {
-		s := Span{test.start, test.end}
+		s := Span{Key: test.start, EndKey: test.end}
 		if test.valid != s.Valid() {
 			t.Errorf("%d: expected span %q-%q to return %t for Valid, instead got %t",
 				i, test.start, test.end, test.valid, s.Valid())
@@ -1317,7 +1403,7 @@ func TestValuePrettyPrint(t *testing.T) {
 	_ = decimalValue.SetDecimal(apd.New(628, -2))
 
 	var durationValue Value
-	_ = durationValue.SetDuration(duration.Duration{Months: 1, Days: 2, Nanos: 3})
+	_ = durationValue.SetDuration(duration.DecodeDuration(1, 2, 3))
 
 	var tupleValue Value
 	tupleBytes := encoding.EncodeBytesValue(encoding.EncodeIntValue(nil, 1, 8), 2, []byte("foo"))
@@ -1327,6 +1413,10 @@ func TestValuePrettyPrint(t *testing.T) {
 	bytesValuePrintable.SetBytes([]byte("abc"))
 	bytesValueNonPrintable.SetBytes([]byte{0x89})
 
+	var bitArrayValue Value
+	ba := bitarray.MakeBitArrayFromInt64(8, 58, 7)
+	bitArrayValue.SetBitArray(ba)
+
 	var errValue Value
 	errValue.SetInt(7)
 	errValue.setTag(ValueType_FLOAT)
@@ -1344,12 +1434,13 @@ func TestValuePrettyPrint(t *testing.T) {
 		{floatValue, "/FLOAT/6.28"},
 		{timeValue, "/TIME/2016-06-29T16:02:50.000000005Z"},
 		{decimalValue, "/DECIMAL/6.28"},
-		{durationValue, "/DURATION/1mon2d3ns"},
+		{durationValue, "/DURATION/1 mon 2 days 00:00:00+3ns"},
 		{MakeValueFromBytes([]byte{0x1, 0x2, 0xF, 0xFF}), "/BYTES/0x01020fff"},
 		{MakeValueFromString("foo"), "/BYTES/foo"},
 		{tupleValue, "/TUPLE/1:1:Int/8/2:3:Bytes/foo"},
 		{bytesValuePrintable, "/BYTES/abc"},
 		{bytesValueNonPrintable, "/BYTES/0x89"},
+		{bitArrayValue, "/BITARRAY/B00111010"},
 		{errValue, "/<err: float64 value should be exactly 8 bytes: 1>"},
 		{errTagValue, "/<err: unknown tag: 99>"},
 	}
