@@ -17,16 +17,33 @@ package roachpb
 import (
 	"bytes"
 	"fmt"
+	"strings"
 
 	"github.com/cockroachdb/cockroach/pkg/util/caller"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
 )
 
+// ClientVisibleRetryError is to be implemented by errors visible by
+// layers above and that can be handled by retrying the transaction.
+type ClientVisibleRetryError interface {
+	ClientVisibleRetryError()
+}
+
+// ClientVisibleAmbiguousError is to be implemented by errors visible
+// by layers above and that indicate uncertainty.
+type ClientVisibleAmbiguousError interface {
+	ClientVisibleAmbiguousError()
+}
+
 func (e *UnhandledRetryableError) Error() string {
 	return e.PErr.Message
 }
 
+// ClientVisibleRetryError implements the ClientVisibleRetryError interface.
+func (e *UnhandledRetryableError) ClientVisibleRetryError() {}
+
+var _ ClientVisibleRetryError = &UnhandledRetryableError{}
 var _ error = &UnhandledRetryableError{}
 
 // ErrorUnexpectedlySet creates a string to panic with when a response (typically
@@ -149,12 +166,10 @@ func (e *Error) GoError() error {
 
 // SetTxn sets the txn and resets the error message. txn is cloned before being
 // stored in the Error.
-// TODO(kaneda): Unexpose this method and make callers use NewErrorWithTxn.
 func (e *Error) SetTxn(txn *Transaction) {
 	e.UnexposedTxn = txn
 	if txn != nil {
-		txnClone := txn.Clone()
-		e.UnexposedTxn = &txnClone
+		e.UnexposedTxn = txn.Clone()
 	}
 	if sErr, ok := e.Detail.GetInner().(ErrorDetailInterface); ok {
 		// Refresh the message as the txn is updated.
@@ -170,9 +185,9 @@ func (e *Error) GetTxn() *Transaction {
 	return e.UnexposedTxn
 }
 
-// UpdateTxn updates the txn.
+// UpdateTxn updates the error transaction.
 func (e *Error) UpdateTxn(o *Transaction) {
-	if e == nil {
+	if o == nil {
 		return
 	}
 	if e.UnexposedTxn == nil {
@@ -306,7 +321,11 @@ func (e *AmbiguousResultError) message(_ *Error) string {
 	return fmt.Sprintf("result is ambiguous (%s)", e.Message)
 }
 
+// ClientVisibleAmbiguousError implements the ClientVisibleAmbiguousError interface.
+func (e *AmbiguousResultError) ClientVisibleAmbiguousError() {}
+
 var _ ErrorDetailInterface = &AmbiguousResultError{}
+var _ ClientVisibleAmbiguousError = &AmbiguousResultError{}
 
 func (e *TransactionAbortedError) Error() string {
 	return fmt.Sprintf("TransactionAbortedError(%s)", e.Reason)
@@ -323,15 +342,19 @@ func (*TransactionAbortedError) canRestartTransaction() TransactionRestart {
 var _ ErrorDetailInterface = &TransactionAbortedError{}
 var _ transactionRestartError = &TransactionAbortedError{}
 
-func (e *HandledRetryableTxnError) Error() string {
+// ClientVisibleRetryError implements the ClientVisibleRetryError interface.
+func (e *TransactionRetryWithProtoRefreshError) ClientVisibleRetryError() {}
+
+func (e *TransactionRetryWithProtoRefreshError) Error() string {
 	return e.message(nil)
 }
 
-func (e *HandledRetryableTxnError) message(_ *Error) string {
-	return fmt.Sprintf("HandledRetryableTxnError: %s", e.Msg)
+func (e *TransactionRetryWithProtoRefreshError) message(_ *Error) string {
+	return fmt.Sprintf("TransactionRetryWithProtoRefreshError: %s", e.Msg)
 }
 
-var _ ErrorDetailInterface = &HandledRetryableTxnError{}
+var _ ClientVisibleRetryError = &TransactionRetryWithProtoRefreshError{}
+var _ ErrorDetailInterface = &TransactionRetryWithProtoRefreshError{}
 
 // NewTransactionAbortedError initializes a new TransactionAbortedError.
 func NewTransactionAbortedError(reason TransactionAbortedReason) *TransactionAbortedError {
@@ -340,14 +363,26 @@ func NewTransactionAbortedError(reason TransactionAbortedReason) *TransactionAbo
 	}
 }
 
-// NewHandledRetryableTxnError initializes a new HandledRetryableTxnError.
+// NewTransactionRetryWithProtoRefreshError initializes a new TransactionRetryWithProtoRefreshError.
 //
 // txnID is the ID of the transaction being restarted.
 // txn is the transaction that the client should use for the next attempts.
-func NewHandledRetryableTxnError(
+func NewTransactionRetryWithProtoRefreshError(
 	msg string, txnID uuid.UUID, txn Transaction,
-) *HandledRetryableTxnError {
-	return &HandledRetryableTxnError{Msg: msg, TxnID: txnID, Transaction: txn}
+) *TransactionRetryWithProtoRefreshError {
+	return &TransactionRetryWithProtoRefreshError{
+		Msg:         msg,
+		TxnID:       txnID,
+		Transaction: txn,
+	}
+}
+
+// PrevTxnAborted returns true if this error originated from a
+// TransactionAbortedError. If true, the client will need to create a new
+// transaction, as opposed to continuing with the existing one at a bumped
+// epoch.
+func (e *TransactionRetryWithProtoRefreshError) PrevTxnAborted() bool {
+	return !e.TxnID.Equal(e.Transaction.ID)
 }
 
 // NewTransactionPushError initializes a new TransactionPushError.
@@ -355,7 +390,7 @@ func NewHandledRetryableTxnError(
 func NewTransactionPushError(pusheeTxn Transaction) *TransactionPushError {
 	// Note: this error will cause a txn restart. The error that the client
 	// receives contains a txn that might have a modified priority.
-	return &TransactionPushError{PusheeTxn: pusheeTxn.Clone()}
+	return &TransactionPushError{PusheeTxn: pusheeTxn}
 }
 
 func (e *TransactionPushError) Error() string {
@@ -377,9 +412,12 @@ func (*TransactionPushError) canRestartTransaction() TransactionRestart {
 }
 
 // NewTransactionRetryError initializes a new TransactionRetryError.
-func NewTransactionRetryError(reason TransactionRetryReason) *TransactionRetryError {
+func NewTransactionRetryError(
+	reason TransactionRetryReason, extraMsg string,
+) *TransactionRetryError {
 	return &TransactionRetryError{
-		Reason: reason,
+		Reason:   reason,
+		ExtraMsg: extraMsg,
 	}
 }
 
@@ -407,12 +445,12 @@ func NewTransactionStatusError(msg string) *TransactionStatusError {
 	}
 }
 
-// NewTransactionNotFoundStatusError initializes a new TransactionStatusError with
-// a REASON_TXN_NOT_FOUND reason.
-func NewTransactionNotFoundStatusError() *TransactionStatusError {
+// NewTransactionCommittedStatusError initializes a new TransactionStatusError
+// with a REASON_TXN_COMMITTED.
+func NewTransactionCommittedStatusError() *TransactionStatusError {
 	return &TransactionStatusError{
-		Msg:    "txn record not found",
-		Reason: TransactionStatusError_REASON_TXN_NOT_FOUND,
+		Msg:    "already committed",
+		Reason: TransactionStatusError_REASON_TXN_COMMITTED,
 	}
 }
 
@@ -494,7 +532,7 @@ func NewReadWithinUncertaintyIntervalError(
 	if txn != nil {
 		maxTS := txn.MaxTimestamp
 		rwue.MaxTimestamp = &maxTS
-		rwue.ObservedTimestamps = append([]ObservedTimestamp(nil), txn.ObservedTimestamps...)
+		rwue.ObservedTimestamps = txn.ObservedTimestamps
 	}
 	return rwue
 }
@@ -504,10 +542,20 @@ func (e *ReadWithinUncertaintyIntervalError) Error() string {
 }
 
 func (e *ReadWithinUncertaintyIntervalError) message(_ *Error) string {
+	var ts strings.Builder
+	ts.WriteByte('[')
+	for i, ot := range observedTimestampSlice(e.ObservedTimestamps) {
+		if i > 0 {
+			ts.WriteByte(' ')
+		}
+		fmt.Fprintf(&ts, "{%d %v}", ot.NodeID, ot.Timestamp)
+	}
+	ts.WriteByte(']')
+
 	return fmt.Sprintf("ReadWithinUncertaintyIntervalError: read at time %s encountered "+
 		"previous write with future timestamp %s within uncertainty interval `t <= %v`; "+
-		"observed timestamps: %v",
-		e.ReadTimestamp, e.ExistingTimestamp, e.MaxTimestamp, observedTimestampSlice(e.ObservedTimestamps))
+		"observed timestamps: %s",
+		e.ReadTimestamp, e.ExistingTimestamp, e.MaxTimestamp, ts.String())
 }
 
 var _ ErrorDetailInterface = &ReadWithinUncertaintyIntervalError{}
@@ -605,8 +653,11 @@ func (e *TxnAlreadyEncounteredErrorError) Error() string {
 	return e.message(nil)
 }
 
-func (*TxnAlreadyEncounteredErrorError) message(_ *Error) string {
-	return "txn already encountered an error; cannot be used anymore"
+func (e *TxnAlreadyEncounteredErrorError) message(_ *Error) string {
+	return fmt.Sprintf(
+		"txn already encountered an error; cannot be used anymore (previous err: %s)",
+		e.PrevError,
+	)
 }
 
 var _ ErrorDetailInterface = &TxnAlreadyEncounteredErrorError{}
@@ -638,7 +689,7 @@ func (e *MixedSuccessError) Error() string {
 }
 
 func (e *MixedSuccessError) message(_ *Error) string {
-	return "the batch experienced mixed success and failure"
+	return fmt.Sprintf("the batch experienced mixed success and failure: %s", e.Wrapped)
 }
 
 var _ ErrorDetailInterface = &MixedSuccessError{}
@@ -648,14 +699,15 @@ func (e *BatchTimestampBeforeGCError) Error() string {
 }
 
 func (e *BatchTimestampBeforeGCError) message(_ *Error) string {
-	return fmt.Sprintf("batch timestamp %v must be after GC threshold %v", e.Timestamp, e.Threshold)
+	return fmt.Sprintf("batch timestamp %v must be after replica GC threshold %v", e.Timestamp, e.Threshold)
 }
 
 var _ ErrorDetailInterface = &BatchTimestampBeforeGCError{}
 
 // NewIntentMissingError creates a new IntentMissingError.
-func NewIntentMissingError(wrongIntent *Intent) *IntentMissingError {
+func NewIntentMissingError(key Key, wrongIntent *Intent) *IntentMissingError {
 	return &IntentMissingError{
+		Key:         key,
 		WrongIntent: wrongIntent,
 	}
 }
