@@ -24,12 +24,6 @@ import (
 	"strings"
 	"time"
 
-	gwruntime "github.com/grpc-ecosystem/grpc-gateway/runtime"
-	"github.com/pkg/errors"
-	"google.golang.org/grpc"
-	"google.golang.org/grpc/codes"
-	"google.golang.org/grpc/status"
-
 	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
@@ -46,11 +40,18 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/storage"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	gwruntime "github.com/grpc-ecosystem/grpc-gateway/runtime"
+	"github.com/pkg/errors"
+	"google.golang.org/grpc"
+	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/status"
 )
 
 const (
@@ -61,6 +62,11 @@ const (
 	// defaultAPIEventLimit is the default maximum number of events returned by any
 	// endpoints returning events.
 	defaultAPIEventLimit = 1000
+
+	// Number of empty ranges for table descriptors that aren't actually tables,
+	// e.g. descriptors 17, 18, 19, and 22 which correspond to MetaRangesID,
+	// SystemRangesID, TimeseriesRangesID, and LivenessRangesID in pkg/keys.
+	nonTableDescriptorRangeCount = 4
 )
 
 // apiServerMessage is the standard body for all HTTP 500 responses.
@@ -567,15 +573,15 @@ func (s *adminServer) TableStats(
 	tableID := path[2]
 	tableSpan := generateTableSpan(tableID)
 
-	return s.tableStatsForSpan(ctx, tableSpan)
+	return s.statsForSpan(ctx, tableSpan)
 }
 
 // NonTableStats is an endpoint that returns disk usage and replication
-// statistics for the time series system.
+// statistics for non-table parts of the system.
 func (s *adminServer) NonTableStats(
 	ctx context.Context, req *serverpb.NonTableStatsRequest,
 ) (*serverpb.NonTableStatsResponse, error) {
-	timeSeriesStats, err := s.tableStatsForSpan(ctx, roachpb.Span{
+	timeSeriesStats, err := s.statsForSpan(ctx, roachpb.Span{
 		Key:    keys.TimeseriesPrefix,
 		EndKey: keys.TimeseriesPrefix.PrefixEnd(),
 	})
@@ -585,29 +591,75 @@ func (s *adminServer) NonTableStats(
 	response := serverpb.NonTableStatsResponse{
 		TimeSeriesStats: timeSeriesStats,
 	}
+
+	spansForInternalUse := []roachpb.Span{
+		{
+			Key:    keys.LocalMax,
+			EndKey: keys.TimeseriesPrefix,
+		},
+		{
+			Key:    keys.TimeseriesKeyMax,
+			EndKey: keys.TableDataMin,
+		},
+	}
+	for _, span := range spansForInternalUse {
+		nonTableStats, err := s.statsForSpan(ctx, span)
+		if err != nil {
+			return nil, err
+		}
+		if response.InternalUseStats == nil {
+			response.InternalUseStats = nonTableStats
+		} else {
+			response.InternalUseStats.Add(nonTableStats)
+		}
+	}
+
+	// There are four empty ranges for table descriptors 17, 18, 19, and 22 that
+	// aren't actually tables (a.k.a. MetaRangesID, SystemRangesID,
+	// TimeseriesRangesID, and LivenessRangesID in pkg/keys).
+	// No data is ever really written to them since they don't have actual
+	// tables. Some backend work could probably be done to eliminate these empty
+	// ranges, but it may be more trouble than it's worth. In the meantime,
+	// sweeping them under the general-purpose "Internal use" label in
+	// the "Non-Table" section of the Databases page.
+	response.InternalUseStats.RangeCount += nonTableDescriptorRangeCount
+
 	return &response, nil
 }
 
-func (s *adminServer) tableStatsForSpan(
-	ctx context.Context, tableSpan roachpb.Span,
+func (s *adminServer) statsForSpan(
+	ctx context.Context, span roachpb.Span,
 ) (*serverpb.TableStatsResponse, error) {
-	startKey, err := keys.Addr(tableSpan.Key)
+	startKey, err := keys.Addr(span.Key)
 	if err != nil {
 		return nil, s.serverError(err)
 	}
-	endKey, err := keys.Addr(tableSpan.EndKey)
+	endKey, err := keys.Addr(span.EndKey)
 	if err != nil {
 		return nil, s.serverError(err)
 	}
 
 	// Get current range descriptors for table. This is done by scanning over
-	// meta2 keys for the range.
-	rangeDescKVs, err := s.server.db.Scan(ctx, keys.RangeMetaKey(startKey), keys.RangeMetaKey(endKey), 0)
+	// meta2 keys for the range. A special case occurs if we wish to include
+	// the meta1 key range itself, in which case we'll get KeyMin back and that
+	// cannot be scanned (due to range-local addressing confusion). This is
+	// handled appropriately by adjusting the bounds to grab the descriptors
+	// for all ranges (including range1, which is not only gossiped but also
+	// persisted in meta1).
+	startMetaKey := keys.RangeMetaKey(startKey)
+	if bytes.Equal(startMetaKey, roachpb.RKeyMin) {
+		// This is the special case described above. The following key instructs
+		// the code below to scan all of the addressing, i.e. grab all of the
+		// descriptors including that for r1.
+		startMetaKey = keys.RangeMetaKey(keys.MustAddr(keys.Meta2Prefix))
+	}
+
+	rangeDescKVs, err := s.server.db.Scan(ctx, startMetaKey, keys.RangeMetaKey(endKey), 0)
 	if err != nil {
 		return nil, s.serverError(err)
 	}
 
-	// Extract a list of node IDs from the response.
+	// This map will store the nodes we need to fan out to.
 	nodeIDs := make(map[roachpb.NodeID]struct{})
 	for _, kv := range rangeDescKVs {
 		var rng roachpb.RangeDescriptor
@@ -652,19 +704,20 @@ func (s *adminServer) tableStatsForSpan(
 			ctx, "server.adminServer: requesting remote stats",
 			func(ctx context.Context) {
 				// Set a generous timeout on the context for each individual query.
-				ctx, cancel := context.WithTimeout(ctx, 5*base.NetworkTimeout)
-				defer cancel()
-
 				var spanResponse *serverpb.SpanStatsResponse
-				client, err := s.server.status.dialNode(ctx, nodeID)
-				if err == nil {
-					req := serverpb.SpanStatsRequest{
-						StartKey: startKey,
-						EndKey:   endKey,
-						NodeID:   nodeID.String(),
-					}
-					spanResponse, err = client.SpanStats(ctx, &req)
-				}
+				err := contextutil.RunWithTimeout(ctx, "request remote stats", 5*base.NetworkTimeout,
+					func(ctx context.Context) error {
+						client, err := s.server.status.dialNode(ctx, nodeID)
+						if err == nil {
+							req := serverpb.SpanStatsRequest{
+								StartKey: startKey,
+								EndKey:   endKey,
+								NodeID:   nodeID.String(),
+							}
+							spanResponse, err = client.SpanStats(ctx, &req)
+						}
+						return err
+					})
 
 				// Channel is buffered, can always write.
 				responses <- nodeResponse{
@@ -856,7 +909,7 @@ func (s *adminServer) RangeLog(
 	}
 	scanner := makeResultScanner(cols)
 	for _, row := range rows {
-		var event storage.RangeLogEvent
+		var event storagepb.RangeLogEvent
 		var ts time.Time
 		if err := scanner.ScanIndex(row, 0, &ts); err != nil {
 			return nil, errors.Wrap(err, fmt.Sprintf("Timestamp didn't parse correctly: %s", row[0].String()))
@@ -876,8 +929,8 @@ func (s *adminServer) RangeLog(
 		if err := scanner.ScanIndex(row, 3, &eventTypeString); err != nil {
 			return nil, errors.Wrap(err, fmt.Sprintf("EventType didn't parse correctly: %s", row[3].String()))
 		}
-		if eventType, ok := storage.RangeLogEventType_value[eventTypeString]; ok {
-			event.EventType = storage.RangeLogEventType(eventType)
+		if eventType, ok := storagepb.RangeLogEventType_value[eventTypeString]; ok {
+			event.EventType = storagepb.RangeLogEventType(eventType)
 		} else {
 			return nil, errors.Errorf("EventType didn't parse correctly: %s", eventTypeString)
 		}
@@ -1117,9 +1170,9 @@ func (s *adminServer) Jobs(
 
 	q := makeSQLQuery()
 	q.Append(`
-      SELECT job_id, job_type, description, user_name, descriptor_ids, status,
-						 created, started, finished, modified, fraction_completed,
-						 high_water_timestamp, error
+      SELECT job_id, job_type, description, statement, user_name, descriptor_ids, status,
+						 running_status, created, started, finished, modified,
+						 fraction_completed, high_water_timestamp, error
         FROM crdb_internal.jobs
        WHERE true
 	`)
@@ -1128,6 +1181,9 @@ func (s *adminServer) Jobs(
 	}
 	if req.Type != jobspb.TypeUnspecified {
 		q.Append(" AND job_type = $", req.Type.String())
+	} else {
+		// Don't show auto stats jobs in the overview page.
+		q.Append(" AND (job_type != $ OR job_type IS NULL)", jobspb.TypeAutoCreateStats.String())
 	}
 	q.Append("ORDER BY created DESC")
 	if req.Limit > 0 {
@@ -1148,14 +1204,17 @@ func (s *adminServer) Jobs(
 		job := &resp.Jobs[i]
 		var fractionCompletedOrNil *float32
 		var highwaterOrNil *apd.Decimal
+		var runningStatusOrNil *string
 		if err := scanner.ScanAll(
 			row,
 			&job.ID,
 			&job.Type,
 			&job.Description,
+			&job.Statement,
 			&job.Username,
 			&job.DescriptorIDs,
 			&job.Status,
+			&runningStatusOrNil,
 			&job.Created,
 			&job.Started,
 			&job.Finished,
@@ -1178,6 +1237,9 @@ func (s *adminServer) Jobs(
 		if fractionCompletedOrNil != nil {
 			job.FractionCompleted = *fractionCompletedOrNil
 		}
+		if runningStatusOrNil != nil {
+			job.RunningStatus = *runningStatusOrNil
+		}
 	}
 
 	return &resp, nil
@@ -1417,7 +1479,7 @@ func (s *adminServer) DataDistribution(
 	// and deleted tables (as opposed to e.g. information_schema) because we are interested
 	// in the data for all ranges, not just ranges for visible tables.
 	userName := s.getUser(req)
-	tablesQuery := `SELECT name, table_id, database_name, drop_time FROM "".crdb_internal.tables`
+	tablesQuery := `SELECT name, table_id, database_name, drop_time FROM "".crdb_internal.tables WHERE schema_name = 'public'`
 	rows1, _ /* cols */, err := s.server.internalExecutor.QueryWithUser(
 		ctx, "admin-replica-matrix", nil /* txn */, userName, tablesQuery,
 	)
@@ -1618,13 +1680,12 @@ func (s *adminServer) EnqueueRange(
 		response.Details = append(response.Details, errDetail)
 	}
 
-	nodeCtx, cancel := context.WithTimeout(ctx, time.Minute)
-	defer cancel()
-
-	if err := s.server.status.iterateNodes(
-		nodeCtx, fmt.Sprintf("enqueue r%d in queue %s", req.RangeID, req.Queue),
-		dialFn, nodeFn, responseFn, errorFn,
-	); err != nil {
+	if err := contextutil.RunWithTimeout(ctx, "enqueue range", time.Minute, func(ctx context.Context) error {
+		return s.server.status.iterateNodes(
+			ctx, fmt.Sprintf("enqueue r%d in queue %s", req.RangeID, req.Queue),
+			dialFn, nodeFn, responseFn, errorFn,
+		)
+	}); err != nil {
 		if len(response.Details) == 0 {
 			return nil, err
 		}
@@ -1793,6 +1854,18 @@ func (rs resultScanner) ScanIndex(row tree.Datums, index int, dst interface{}) e
 		}
 		*d = string(s)
 
+	case **string:
+		s, ok := tree.AsDString(src)
+		if !ok {
+			if src != tree.DNull {
+				return errors.Errorf("source type assertion failed")
+			}
+			*d = nil
+			break
+		}
+		val := string(s)
+		*d = &val
+
 	case *bool:
 		s, ok := src.(*tree.DBool)
 		if !ok {
@@ -1931,23 +2004,23 @@ func (s *adminServer) queryZone(
 		ctx, "admin-query-zone", nil /* txn */, userName, query, id,
 	)
 	if err != nil {
-		return config.ZoneConfig{}, false, err
+		return *config.NewZoneConfig(), false, err
 	}
 
 	if len(rows) == 0 {
-		return config.ZoneConfig{}, false, nil
+		return *config.NewZoneConfig(), false, nil
 	}
 
 	var zoneBytes []byte
 	scanner := resultScanner{}
 	err = scanner.ScanIndex(rows[0], 0, &zoneBytes)
 	if err != nil {
-		return config.ZoneConfig{}, false, err
+		return *config.NewZoneConfig(), false, err
 	}
 
 	var zone config.ZoneConfig
 	if err := protoutil.Unmarshal(zoneBytes, &zone); err != nil {
-		return config.ZoneConfig{}, false, err
+		return *config.NewZoneConfig(), false, err
 	}
 	return zone, true, nil
 }
@@ -1964,7 +2037,7 @@ func (s *adminServer) queryZonePath(
 			return path[i], zone, true, err
 		}
 	}
-	return 0, config.ZoneConfig{}, false, nil
+	return 0, *config.NewZoneConfig(), false, nil
 }
 
 // queryNamespaceID queries for the ID of the namespace with the given name and
