@@ -30,13 +30,11 @@ import (
 	"testing"
 	"time"
 
-	"github.com/gogo/protobuf/proto"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/jobs"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
+	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/server/debug"
@@ -46,6 +44,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/storage"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
@@ -55,6 +55,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/gogo/protobuf/proto"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/assert"
 )
 
 func getAdminJSONProto(
@@ -346,6 +349,111 @@ func TestAdminAPIDatabaseSQLInjection(t *testing.T) {
 	}
 }
 
+func TestAdminAPINonTableStats(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	testCluster := serverutils.StartTestCluster(t, 3, base.TestClusterArgs{})
+	defer testCluster.Stopper().Stop(context.Background())
+	s := testCluster.Server(0)
+
+	// Skip TableStatsResponse.Stats comparison, since it includes data which
+	// aren't consistent (time, bytes).
+	expectedResponse := serverpb.NonTableStatsResponse{
+		TimeSeriesStats: &serverpb.TableStatsResponse{
+			RangeCount:   1,
+			ReplicaCount: 3,
+			NodeCount:    3,
+		},
+		InternalUseStats: &serverpb.TableStatsResponse{
+			RangeCount:   8,
+			ReplicaCount: 12,
+			NodeCount:    3,
+		},
+	}
+
+	var resp serverpb.NonTableStatsResponse
+	if err := getAdminJSONProto(s, "nontablestats", &resp); err != nil {
+		t.Fatal(err)
+	}
+
+	assertExpectedStatsResponse := func(expected, actual *serverpb.TableStatsResponse) {
+		assert.Equal(t, expected.RangeCount, actual.RangeCount)
+		assert.Equal(t, expected.ReplicaCount, actual.ReplicaCount)
+		assert.Equal(t, expected.NodeCount, actual.NodeCount)
+	}
+
+	assertExpectedStatsResponse(expectedResponse.TimeSeriesStats, resp.TimeSeriesStats)
+	assertExpectedStatsResponse(expectedResponse.InternalUseStats, resp.InternalUseStats)
+}
+
+// TODO(celia): I expect all the ranges listed on the Database page to equal
+// the RangeCount returned from doing a span on [LocalMax, MaxKey). For a cluster
+// with no user data, all the ranges on the Databases page consist of:
+// 1) the total ranges listed for the system database
+// 2) the total ranges listed for the Non-Table data
+func TestRangeCount_MissingOneRange(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	testCluster := serverutils.StartTestCluster(t, 3, base.TestClusterArgs{})
+	defer testCluster.Stopper().Stop(context.Background())
+	s := testCluster.Server(0)
+
+	// Sum up ranges for non-table parts of the system returned
+	// from the "nontablestats" enpoint.
+	getNonTableRangeCount := func() int64 {
+		var resp serverpb.NonTableStatsResponse
+		if err := getAdminJSONProto(s, "nontablestats", &resp); err != nil {
+			t.Fatal(err)
+		}
+		return resp.TimeSeriesStats.RangeCount + resp.InternalUseStats.RangeCount
+	}
+
+	// Sum up ranges from system database tables returned
+	// from the "databases/system/tables/{table}" endpoints.
+	getSystemTableRangeCount := func() int64 {
+		var systemTableRangeCount int64
+		var dbResp serverpb.DatabaseDetailsResponse
+		if err := getAdminJSONProto(s, "databases/system", &dbResp); err != nil {
+			t.Fatal(err)
+		}
+		for _, tableName := range dbResp.TableNames {
+			var tblResp serverpb.TableStatsResponse
+			path := "databases/system/tables/" + tableName + "/stats"
+			if err := getAdminJSONProto(s, path, &tblResp); err != nil {
+				t.Fatal(err)
+			}
+			systemTableRangeCount += tblResp.RangeCount
+		}
+		return systemTableRangeCount
+	}
+
+	getRangeCountFromFullSpan := func() int64 {
+		adminServer := s.(*TestServer).Server.admin
+		stats, err := adminServer.statsForSpan(context.Background(), roachpb.Span{
+			Key:    keys.LocalMax,
+			EndKey: keys.MaxKey,
+		})
+		if err != nil {
+			t.Fatal(err)
+		}
+		return stats.RangeCount
+	}
+
+	totalRangeCount := getRangeCountFromFullSpan()
+	nonTableRangeCount := getNonTableRangeCount()
+	systemTableRangeCount := getSystemTableRangeCount()
+
+	// expected: expectedRangeCount == nonTableRangeCount+systemTableRangeCount
+	// actual: expectedRangeCount > nonTableRangeCount+systemTableRangeCount
+	if totalRangeCount == nonTableRangeCount+systemTableRangeCount {
+		t.Fail()
+	}
+
+	// TODO(celia): We're missing 1 range -- where is it?
+	expectedMissingRangeCount := int64(1)
+	assert.Equal(t,
+		totalRangeCount,
+		nonTableRangeCount+systemTableRangeCount+expectedMissingRangeCount)
+}
+
 func TestAdminAPITableDoesNotExist(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	s, _, _ := serverutils.StartServer(t, base.TestServerArgs{})
@@ -403,9 +511,9 @@ func TestAdminAPITableDetails(t *testing.T) {
 			setupQueries := []string{
 				fmt.Sprintf("CREATE DATABASE %s", escDBName),
 				fmt.Sprintf(`CREATE TABLE %s.%s (
-							nulls_allowed INT,
-							nulls_not_allowed INT NOT NULL DEFAULT 1000,
-							default2 INT DEFAULT 2,
+							nulls_allowed INT8,
+							nulls_not_allowed INT8 NOT NULL DEFAULT 1000,
+							default2 INT8 DEFAULT 2,
 							string_default STRING DEFAULT 'default_string'
 						)`, escDBName, escTblName),
 				fmt.Sprintf("CREATE USER readonly"),
@@ -430,9 +538,9 @@ func TestAdminAPITableDetails(t *testing.T) {
 
 			// Verify columns.
 			expColumns := []serverpb.TableDetailsResponse_Column{
-				{Name: "nulls_allowed", Type: "INT", Nullable: true, DefaultValue: ""},
-				{Name: "nulls_not_allowed", Type: "INT", Nullable: false, DefaultValue: "1000:::INT"},
-				{Name: "default2", Type: "INT", Nullable: true, DefaultValue: "2:::INT"},
+				{Name: "nulls_allowed", Type: "INT8", Nullable: true, DefaultValue: ""},
+				{Name: "nulls_not_allowed", Type: "INT8", Nullable: false, DefaultValue: "1000:::INT8"},
+				{Name: "default2", Type: "INT8", Nullable: true, DefaultValue: "2:::INT8"},
 				{Name: "string_default", Type: "STRING", Nullable: true, DefaultValue: "'default_string':::STRING"},
 				{Name: "rowid", Type: "INT", Nullable: false, DefaultValue: "unique_rowid()", Hidden: true},
 			}
@@ -607,7 +715,7 @@ func TestAdminAPIZoneDetails(t *testing.T) {
 
 	// Apply zone configuration to database and check again.
 	dbZone := config.ZoneConfig{
-		RangeMinBytes: 456,
+		RangeMinBytes: proto.Int64(456),
 	}
 	setZone(dbZone, idPath[1])
 	verifyDbZone(dbZone, serverpb.ZoneConfigurationLevel_DATABASE)
@@ -615,7 +723,7 @@ func TestAdminAPIZoneDetails(t *testing.T) {
 
 	// Apply zone configuration to table and check again.
 	tblZone := config.ZoneConfig{
-		RangeMinBytes: 789,
+		RangeMinBytes: proto.Int64(789),
 	}
 	setZone(tblZone, idPath[2])
 	verifyDbZone(dbZone, serverpb.ZoneConfigurationLevel_DATABASE)
@@ -1227,29 +1335,46 @@ func TestAdminAPIQueryPlan(t *testing.T) {
 
 		for _, exp := range testCase.exp {
 			if !strings.Contains(res.DistSQLPhysicalQueryPlan, exp) {
-				t.Errorf("%d: expected response %s to contain %s", i, res, exp)
+				t.Errorf("%d: expected response %v to contain %s", i, res, exp)
 			}
 		}
 	}
 
 }
 
-func TestAdminAPIRangeLog(t *testing.T) {
+func TestAdminAPIRangeLogByRangeID(t *testing.T) {
 	defer leaktest.AfterTest(t)()
-	s, _, _ := serverutils.StartServer(t, base.TestServerArgs{})
+	s, db, _ := serverutils.StartServer(t, base.TestServerArgs{})
 	defer s.Stopper().Stop(context.Background())
 
+	rangeID := 654321
 	testCases := []struct {
 		rangeID  int
 		hasLimit bool
 		limit    int
 		expected int
 	}{
-		{1, false, 0, 1},
-		{2, false, 0, 2},
-		{2, true, 0, 2},
-		{2, true, -1, 2},
-		{2, true, 1, 1},
+		{rangeID, true, 0, 2},
+		{rangeID, true, -1, 2},
+		{rangeID, true, 1, 1},
+		{rangeID, false, 0, 2},
+		// We'll create one event that has rangeID+1 as the otherRangeID.
+		{rangeID + 1, false, 0, 1},
+	}
+
+	for _, otherRangeID := range []int{rangeID + 1, rangeID + 2} {
+		if _, err := db.Exec(
+			`INSERT INTO system.rangelog (
+             timestamp, "rangeID", "otherRangeID", "storeID", "eventType"
+           ) VALUES (
+             now(), $1, $2, $3, $4
+          )`,
+			rangeID, otherRangeID,
+			1, // storeID
+			storagepb.RangeLogEventType_add.String(),
+		); err != nil {
+			t.Fatal(err)
+		}
 	}
 
 	for _, tc := range testCases {
@@ -1268,26 +1393,62 @@ func TestAdminAPIRangeLog(t *testing.T) {
 			}
 
 			for _, event := range resp.Events {
-				if event.Event.RangeID != roachpb.RangeID(tc.rangeID) &&
-					event.Event.OtherRangeID != roachpb.RangeID(tc.rangeID) {
-					t.Errorf("expected rangeID or otherRangeID to be r%d, got r%d and r%d",
-						tc.rangeID, event.Event.RangeID, event.Event.OtherRangeID)
+				expID := roachpb.RangeID(tc.rangeID)
+				if event.Event.RangeID != expID && event.Event.OtherRangeID != expID {
+					t.Errorf("expected rangeID or otherRangeID to be %d, got %d and r%d",
+						expID, event.Event.RangeID, event.Event.OtherRangeID)
 				}
 			}
 		})
 	}
 }
 
+// Test the range log API when queries are not filtered by a range ID (like in
+// TestAdminAPIRangeLogByRangeID).
 func TestAdminAPIFullRangeLog(t *testing.T) {
 	defer leaktest.AfterTest(t)()
-	s, _, kvDB := serverutils.StartServer(t, base.TestServerArgs{})
+	s, db, _ := serverutils.StartServer(t,
+		base.TestServerArgs{
+			Knobs: base.TestingKnobs{
+				Store: &storage.StoreTestingKnobs{
+					DisableSplitQueue: true,
+				},
+			},
+		})
 	defer s.Stopper().Stop(context.Background())
 
-	expectedRanges, err := ExpectedInitialRangeCount(kvDB)
+	// Insert something in the rangelog table, otherwise it's empty for new
+	// clusters.
+	rows, err := db.Query(`SELECT count(1) FROM system.rangelog`)
 	if err != nil {
 		t.Fatal(err)
 	}
-	expectedEvents := expectedRanges - 1 // one for each split
+	if !rows.Next() {
+		t.Fatal("missing row")
+	}
+	var cnt int
+	if err := rows.Scan(&cnt); err != nil {
+		t.Fatal(err)
+	}
+	if err := rows.Close(); err != nil {
+		t.Fatal(err)
+	}
+	if cnt != 0 {
+		t.Fatalf("expected 0 rows in system.rangelog, found: %d", cnt)
+	}
+	const rangeID = 100
+	for i := 0; i < 10; i++ {
+		if _, err := db.Exec(
+			`INSERT INTO system.rangelog (
+             timestamp, "rangeID", "storeID", "eventType"
+           ) VALUES (now(), $1, 1, $2)`,
+			rangeID,
+			storagepb.RangeLogEventType_add.String(),
+		); err != nil {
+			t.Fatal(err)
+		}
+	}
+	expectedEvents := 10
 
 	testCases := []struct {
 		hasLimit bool
@@ -1310,8 +1471,13 @@ func TestAdminAPIFullRangeLog(t *testing.T) {
 			if err := getAdminJSONProto(s, url, &resp); err != nil {
 				t.Fatal(err)
 			}
-			if e, a := tc.expected, len(resp.Events); e != a {
-				t.Fatalf("expected %d events, got %d", e, a)
+			events := resp.Events
+			if e, a := tc.expected, len(events); e != a {
+				var sb strings.Builder
+				for _, ev := range events {
+					sb.WriteString(ev.String() + "\n")
+				}
+				t.Fatalf("expected %d events, got %d:\n%s", e, a, sb.String())
 			}
 		})
 	}
@@ -1537,3 +1703,21 @@ func TestEnqueueRange(t *testing.T) {
 		})
 	}
 }
+
+func TestStatsforSpanOnLocalMax(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	testCluster := serverutils.StartTestCluster(t, 3, base.TestClusterArgs{})
+	defer testCluster.Stopper().Stop(context.Background())
+	firstServer := testCluster.Server(0)
+	adminServer := firstServer.(*TestServer).Server.admin
+
+	underTest := roachpb.Span{
+		Key:    keys.LocalMax,
+		EndKey: keys.SystemPrefix,
+	}
+
+	_, err := adminServer.statsForSpan(context.Background(), underTest)
+	if err != nil {
+		t.Fatal(err)
+	}
+}
