@@ -23,9 +23,6 @@ import (
 	"text/tabwriter"
 	"time"
 
-	"github.com/elastic/gosigar"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/gossip/resolver"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
@@ -40,6 +37,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/elastic/gosigar"
+	"github.com/pkg/errors"
 )
 
 // Context defaults.
@@ -73,9 +72,10 @@ const (
 	minimumNetworkFileDescriptors     = 256
 	recommendedNetworkFileDescriptors = 5000
 
-	defaultConnResultsBufferBytes = 16 << 10 // 16 KiB
-
 	defaultSQLTableStatCacheSize = 256
+
+	// This comes out to 1024 cache entries.
+	defaultSQLQueryCacheSize = 8 * 1024 * 1024
 )
 
 var productionSettingsWebpage = fmt.Sprintf(
@@ -170,6 +170,13 @@ type Config struct {
 	// statistics cache.
 	SQLTableStatCacheSize int
 
+	// SQLQueryCacheSize is the memory size (in bytes) of the query plan cache.
+	SQLQueryCacheSize int64
+
+	// GoroutineDumpDirName is the directory name for goroutine dumps using
+	// goroutinedumper.
+	GoroutineDumpDirName string
+
 	// HeapProfileDirName is the directory name for heap profiles using
 	// heapprofiler.
 	HeapProfileDirName string
@@ -240,10 +247,11 @@ type Config struct {
 	EventLogEnabled bool
 
 	// ReadyFn is called when the server has started listening on its
-	// sockets. The boolean argument indicates (iff true) that the
+	// sockets.
+	// The argument waitForInit indicates (iff true) that the
 	// server is not bootstrapped yet, will not bootstrap itself and
-	// will be waiting for an `init` command. This can be used to inform
-	// the user.
+	// will be waiting for an `init` command or accept bootstrapping
+	// from a joined node.
 	ReadyFn func(waitForInit bool)
 
 	// DelayedBootstrapFn is called if the boostrap process does not complete
@@ -254,11 +262,6 @@ type Config struct {
 	// the Admin API's HTTP endpoints.
 	EnableWebSessionAuthentication bool
 
-	// ConnResultsBufferBytes is the size of the buffer in which each connection
-	// accumulates results set. Results are flushed to the network when this
-	// buffer overflows.
-	ConnResultsBufferBytes int
-
 	enginesCreated bool
 }
 
@@ -333,6 +336,7 @@ func MakeConfig(ctx context.Context, st *cluster.Settings) Config {
 		CacheSize:                      DefaultCacheSize,
 		SQLMemoryPoolSize:              defaultSQLMemoryPoolSize,
 		SQLTableStatCacheSize:          defaultSQLTableStatCacheSize,
+		SQLQueryCacheSize:              defaultSQLQueryCacheSize,
 		ScanInterval:                   defaultScanInterval,
 		ScanMinIdleTime:                defaultScanMinIdleTime,
 		ScanMaxIdleTime:                defaultScanMaxIdleTime,
@@ -343,10 +347,6 @@ func MakeConfig(ctx context.Context, st *cluster.Settings) Config {
 		},
 		TempStorageConfig: base.TempStorageConfigFromEnv(
 			ctx, st, storeSpec, "" /* parentDir */, base.DefaultTempStorageMaxSizeBytes, 0),
-		// TODO(dan): Hack. Remove this env override once changefeeds have
-		// control over buffering.
-		ConnResultsBufferBytes: envutil.EnvOrDefaultInt(
-			"COCKROACH_CONN_RESULTS_BUFFER_BYTES", defaultConnResultsBufferBytes),
 	}
 	cfg.AmbientCtx.Tracer = st.Tracer
 
