@@ -14,9 +14,7 @@
 
 package server
 
-import (
-	"github.com/cockroachdb/cockroach/pkg/storage/engine"
-)
+import "github.com/cockroachdb/cockroach/pkg/storage/engine"
 
 func setOpenFileLimitInner(physicalStoreCount int) (uint64, error) {
 	return engine.RecommendedMaxOpenFiles, nil
