@@ -14,7 +14,10 @@
 
 package pprofui
 
-import "github.com/spf13/pflag"
+import (
+	"github.com/google/pprof/driver"
+	"github.com/spf13/pflag"
+)
 
 // pprofFlags is a wrapper to satisfy pprof's client flag interface.
 // That interface is satisfied by what the standard flag package
@@ -28,10 +31,17 @@ type pprofFlags struct {
 	*pflag.FlagSet
 }
 
+var _ driver.FlagSet = &pprofFlags{}
+
+// ExtraUsage is part of the driver.FlagSet interface.
 func (pprofFlags) ExtraUsage() string {
 	return ""
 }
 
+// AddExtraUsage is part of the driver.FlagSet interface.
+func (pprofFlags) AddExtraUsage(eu string) {
+}
+
 func (f pprofFlags) StringList(o, d, c string) *[]*string {
 	return &[]*string{f.String(o, d, c)}
 }
