@@ -19,13 +19,12 @@ import (
 	"fmt"
 	"math"
 	"net"
+	"sort"
 	"time"
 
-	opentracing "github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/build"
+	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/keys"
@@ -50,6 +49,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
 const (
@@ -191,10 +192,13 @@ func GetBootstrapSchema() sqlbase.MetadataSchema {
 	return sqlbase.MakeMetadataSchema()
 }
 
-// bootstrapCluster bootstraps a multiple stores using the provided
-// engines and cluster ID. The first bootstrapped store contains a
-// single range spanning all keys. Initial range lookup metadata is
-// populated for the range. Returns the cluster ID.
+// bootstrapCluster bootstraps multiple stores using the provided engines.
+// Returns the cluster ID.
+//
+// Ranges for various static split points are created (i.e. various system
+// ranges and system tables). Note however that many of these ranges cannot be
+// accessed by KV in regular means until the node liveness is written, since
+// epoch-based leases cannot be granted until then.
 func bootstrapCluster(
 	ctx context.Context,
 	cfg storage.StoreConfig,
@@ -262,11 +266,16 @@ func bootstrapCluster(
 		// not create the range, just its data. Only do this if this is the
 		// first store.
 		if i == 0 {
-			initialValues := GetBootstrapSchema().GetInitialValues()
-			// The MinimumVersion is the ServerVersion when we are bootstrapping
-			// a cluster (except in some tests that specifically want to set up
-			// an "old-looking" cluster).
-			if err := s.BootstrapRange(initialValues, bootstrapVersion.MinimumVersion); err != nil {
+			schema := GetBootstrapSchema()
+			initialValues, tableSplits := schema.GetInitialValues()
+			splits := append(config.StaticSplits(), tableSplits...)
+			sort.Slice(splits, func(i, j int) bool {
+				return splits[i].Less(splits[j])
+			})
+
+			if err := s.WriteInitialData(
+				ctx, initialValues, bootstrapVersion.Version, len(engines), splits,
+			); err != nil {
 				return uuid.UUID{}, err
 			}
 		}
@@ -275,31 +284,10 @@ func bootstrapCluster(
 		}
 
 		stores.AddStore(s)
-
-		// Initialize node and store ids.  Only initialize the node once.
-		if i == 0 {
-			if nodeID, err := allocateNodeID(ctx, cfg.DB); nodeID != sIdent.NodeID || err != nil {
-				return uuid.UUID{}, errors.Errorf("expected to initialize node id allocator to %d, got %d: %s",
-					sIdent.NodeID, nodeID, err)
-			}
-		}
-		if storeID, err := allocateStoreIDs(ctx, sIdent.NodeID, 1, cfg.DB); storeID != sIdent.StoreID || err != nil {
-			return uuid.UUID{}, errors.Errorf("expected to initialize store id allocator to %d, got %d: %s",
-				sIdent.StoreID, storeID, err)
-		}
 	}
 	return clusterID, nil
 }
 
-// duplicateBootstrapError is returned by Node.bootstrap when the node is already initialized.
-type duplicateBootstrapError struct {
-	ClusterID uuid.UUID
-}
-
-func (e *duplicateBootstrapError) Error() string {
-	return fmt.Sprintf("cluster has already been initialized with ID %s", e.ClusterID)
-}
-
 // NewNode returns a new instance of Node.
 //
 // execCfg can be nil to help bootstrapping of a Server (the Node is created
@@ -358,7 +346,7 @@ func (n *Node) bootstrap(
 	ctx context.Context, engines []engine.Engine, bootstrapVersion cluster.ClusterVersion,
 ) error {
 	if n.initialBoot || n.clusterID.Get() != uuid.Nil {
-		return &duplicateBootstrapError{ClusterID: n.clusterID.Get()}
+		return fmt.Errorf("cluster has already been initialized with ID %s", n.clusterID.Get())
 	}
 	n.initialBoot = true
 	clusterID, err := bootstrapCluster(ctx, n.storeCfg, engines, bootstrapVersion, n.txnMetrics)
@@ -374,7 +362,7 @@ func (n *Node) bootstrap(
 func (n *Node) onClusterVersionChange(cv cluster.ClusterVersion) {
 	ctx := n.AnnotateCtx(context.Background())
 	if err := n.stores.OnClusterVersionChange(ctx, cv); err != nil {
-		log.Fatal(ctx, errors.Wrapf(err, "updating cluster version to %s", cv))
+		log.Fatal(ctx, errors.Wrapf(err, "updating cluster version to %v", cv))
 	}
 }
 
@@ -393,7 +381,6 @@ func (n *Node) start(
 	localityAddress []roachpb.LocalityAddress,
 	nodeDescriptorCallback func(descriptor roachpb.NodeDescriptor),
 ) error {
-	n.storeCfg.Settings.Version.OnChange(n.onClusterVersionChange)
 	if err := n.storeCfg.Settings.InitializeVersion(cv); err != nil {
 		return errors.Wrap(err, "while initializing cluster version")
 	}
@@ -534,6 +521,15 @@ func (n *Node) start(
 	}
 
 	n.startComputePeriodicMetrics(n.stopper, DefaultMetricsSampleInterval)
+
+	// Now that we've created all our stores, install the gossip version update
+	// handler to write version updates to them.
+	n.storeCfg.Settings.Version.OnChange(n.onClusterVersionChange)
+	// Invoke the callback manually once so that we persist the updated value that
+	// gossip might have already received.
+	clusterVersion := n.storeCfg.Settings.Version.Version()
+	n.onClusterVersionChange(clusterVersion)
+
 	// Be careful about moving this line above `startStores`; store migrations rely
 	// on the fact that the cluster version has not been updated via Gossip (we
 	// have migrations that want to run only if the server starts with a given
@@ -596,6 +592,15 @@ func (n *Node) initStores(
 }
 
 func (n *Node) addStore(store *storage.Store) {
+	cv, err := store.GetClusterVersion(context.TODO())
+	if err != nil {
+		log.Fatal(context.TODO(), err)
+	}
+	if cv == (cluster.ClusterVersion{}) {
+		// The store should have had a version written to it during the store
+		// bootstrap process.
+		log.Fatal(context.TODO(), "attempting to add a store without a version")
+	}
 	n.stores.AddStore(store)
 	n.recorder.AddStore(store)
 }
@@ -615,7 +620,7 @@ func (n *Node) validateStores(ctx context.Context) error {
 // bootstrapStores bootstraps uninitialized stores once the cluster
 // and node IDs have been established for this node. Store IDs are
 // allocated via a sequence id generator stored at a system key per
-// node.
+// node. The new stores are added to n.stores.
 func (n *Node) bootstrapStores(
 	ctx context.Context, emptyEngines []engine.Engine, stopper *stop.Stopper,
 ) error {
@@ -627,12 +632,11 @@ func (n *Node) bootstrapStores(
 	// is joining an existing cluster for the first time, it doesn't have any engines
 	// set up yet, and cv below will be the MinSupportedVersion. At the same time,
 	// the Gossip update which notifies us about the real cluster version won't
-	// persist it to any engines (because none of them are bootstrapped). The correct
-	// version is likely in Settings.Version.Version(), but what if the callback fires
-	// too late? In that case that too is the MinSupportedVersion. So we just accept
-	// that we won't use the correct version here, but post-bootstrapping will invoke
-	// the callback manually, which will disseminate the correct version to all engines
-	// that still need it.
+	// persist it to any engines (because we haven't installed the gossip update
+	// handler yet and also because none of the stores are bootstrapped). So we
+	// just accept that we won't use the correct version here, but
+	// post-bootstrapping will invoke the callback manually, which will
+	// disseminate the correct version to all engines.
 	cv, err := n.stores.SynthesizeClusterVersion(ctx)
 	if err != nil {
 		return errors.Errorf("error retrieving cluster version for bootstrap: %s", err)
@@ -640,7 +644,8 @@ func (n *Node) bootstrapStores(
 
 	{
 		// Bootstrap all waiting stores by allocating a new store id for
-		// each and invoking store.Bootstrap() to persist.
+		// each and invoking storage.Bootstrap() to persist it and the cluster
+		// version.
 		inc := int64(len(emptyEngines))
 		firstID, err := allocateStoreIDs(ctx, n.Descriptor.NodeID, inc, n.storeCfg.DB)
 		if err != nil {
@@ -677,9 +682,6 @@ func (n *Node) bootstrapStores(
 		}
 	}
 
-	clusterVersion := n.storeCfg.Settings.Version.Version()
-	n.onClusterVersionChange(clusterVersion)
-
 	// write a new status summary after all stores have been bootstrapped; this
 	// helps the UI remain responsive when new nodes are added.
 	if err := n.writeNodeStatus(ctx, 0 /* alertTTL */); err != nil {
@@ -703,15 +705,10 @@ func (n *Node) connectGossip(ctx context.Context) error {
 	case <-n.storeCfg.Gossip.Connected:
 	}
 
-	uuidBytes, err := n.storeCfg.Gossip.GetInfo(gossip.KeyClusterID)
+	gossipClusterID, err := n.storeCfg.Gossip.GetClusterID()
 	if err != nil {
-		return errors.Wrap(err, "unable to ascertain cluster ID from gossip network")
+		return err
 	}
-	gossipClusterID, err := uuid.FromBytes(uuidBytes)
-	if err != nil {
-		return errors.Wrap(err, "unable to parse cluster ID from gossip network")
-	}
-
 	clusterID := n.clusterID.Get()
 	if clusterID == uuid.Nil {
 		n.clusterID.Set(ctx, gossipClusterID)
@@ -766,10 +763,7 @@ func (n *Node) startGossip(ctx context.Context, stopper *stop.Stopper) {
 // gossipStores broadcasts each store and dead replica to the gossip network.
 func (n *Node) gossipStores(ctx context.Context) {
 	if err := n.stores.VisitStores(func(s *storage.Store) error {
-		if err := s.GossipStore(ctx, false /* useCached */); err != nil {
-			return err
-		}
-		return s.GossipDeadReplicas(ctx)
+		return s.GossipStore(ctx, false /* useCached */)
 	}); err != nil {
 		log.Warning(ctx, err)
 	}
@@ -1061,9 +1055,8 @@ func (n *Node) setupSpanForIncomingRPC(
 			// If tracing information was passed via gRPC metadata, the gRPC interceptor
 			// should have opened a span for us. If not, open a span now (if tracing is
 			// disabled, this will be a noop span).
-			newSpan = n.storeCfg.AmbientCtx.Tracer.StartSpan(
-				opName,
-				tracing.LogTags(n.storeCfg.AmbientCtx.LogTags()),
+			newSpan = n.storeCfg.AmbientCtx.Tracer.(*tracing.Tracer).StartRootSpan(
+				opName, n.storeCfg.AmbientCtx.LogTags(), tracing.NonRecordableSpan,
 			)
 			ctx = opentracing.ContextWithSpan(ctx, newSpan)
 		}
@@ -1093,8 +1086,7 @@ func (n *Node) setupSpanForIncomingRPC(
 func (n *Node) RangeFeed(
 	args *roachpb.RangeFeedRequest, stream roachpb.Internal_RangeFeedServer,
 ) error {
-	ctx := n.storeCfg.AmbientCtx.ResetAndAnnotateCtx(stream.Context())
-	pErr := n.stores.RangeFeed(ctx, args, stream)
+	pErr := n.stores.RangeFeed(args, stream)
 	if pErr != nil {
 		var event roachpb.RangeFeedEvent
 		event.SetValue(&roachpb.RangeFeedError{
