@@ -25,10 +25,8 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
-	"google.golang.org/grpc"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
+	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/gossip/resolver"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
@@ -36,7 +34,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/kv"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/rpc"
+	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
 	"github.com/cockroachdb/cockroach/pkg/server/status"
+	"github.com/cockroachdb/cockroach/pkg/server/status/statuspb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/storage/closedts/container"
@@ -51,6 +51,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/netutil"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
+	"google.golang.org/grpc"
 )
 
 // createTestNode creates an rpc server using the specified address,
@@ -85,6 +87,7 @@ func createTestNode(
 		Clock:           cfg.Clock,
 		RPCContext:      nodeRPCContext,
 		RPCRetryOptions: &retryOpts,
+		NodeDialer:      nodedialer.New(nodeRPCContext, gossip.AddressResolver(cfg.Gossip)),
 	}, cfg.Gossip)
 	tsf := kv.NewTxnCoordSenderFactory(
 		kv.TxnCoordSenderFactoryConfig{
@@ -153,6 +156,7 @@ func createTestNode(
 
 // createAndStartTestNode creates a new test node and starts it. The server and node are returned.
 func createAndStartTestNode(
+	ctx context.Context,
 	addr net.Addr,
 	engines []engine.Engine,
 	gossipBS net.Addr,
@@ -161,20 +165,21 @@ func createAndStartTestNode(
 ) (*grpc.Server, net.Addr, *Node, *stop.Stopper) {
 	grpcServer, addr, cfg, node, stopper := createTestNode(addr, engines, gossipBS, t)
 	bootstrappedEngines, newEngines, cv, err := inspectEngines(
-		context.TODO(), engines, cfg.Settings.Version.MinSupportedVersion,
+		ctx, engines, cfg.Settings.Version.MinSupportedVersion,
 		cfg.Settings.Version.ServerVersion, node.clusterID)
 	if err != nil {
 		t.Fatal(err)
 	}
-	if err := node.start(context.Background(), addr, bootstrappedEngines, newEngines,
+	// Starting the heartbeat is usually done by the server. This test needs it
+	// because otherwise some of the initial ranges cannot be accessed (since
+	// they need an epoch-based lease).
+	cfg.NodeLiveness.StartHeartbeat(ctx, stopper, nil /* alive */)
+	if err := node.start(ctx, addr, bootstrappedEngines, newEngines,
 		roachpb.Attributes{}, locality, cv, []roachpb.LocalityAddress{},
 		nil, /*nodeDescriptorCallback */
 	); err != nil {
 		t.Fatal(err)
 	}
-	if err := WaitForInitialSplits(node.storeCfg.DB); err != nil {
-		t.Fatal(err)
-	}
 	return grpcServer, addr, node, stopper
 }
 
@@ -209,7 +214,7 @@ func TestBootstrapCluster(t *testing.T) {
 	}
 
 	// Scan the complete contents of the local database directly from the engine.
-	rows, _, _, err := engine.MVCCScan(context.Background(), e, keys.LocalMax, roachpb.KeyMax, math.MaxInt64, hlc.MaxTimestamp, true, nil)
+	rows, _, _, err := engine.MVCCScan(context.Background(), e, keys.LocalMax, roachpb.KeyMax, math.MaxInt64, hlc.MaxTimestamp, engine.MVCCScanOptions{})
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -222,17 +227,29 @@ func TestBootstrapCluster(t *testing.T) {
 		testutils.MakeKey(roachpb.Key("\x03"), roachpb.KeyMax),
 		roachpb.Key("\x04bootstrap-version"),
 		roachpb.Key("\x04node-idgen"),
+		roachpb.Key("\x04range-idgen"),
 		roachpb.Key("\x04store-idgen"),
 	}
+	for _, splitKey := range config.StaticSplits() {
+		meta2Key := keys.RangeMetaKey(splitKey)
+		expectedKeys = append(expectedKeys, meta2Key.AsRawKey())
+	}
+
 	// Add the initial keys for sql.
-	for _, kv := range GetBootstrapSchema().GetInitialValues() {
+	kvs, tableSplits := GetBootstrapSchema().GetInitialValues()
+	for _, kv := range kvs {
 		expectedKeys = append(expectedKeys, kv.Key)
 	}
+	for _, splitKey := range tableSplits {
+		meta2Key := keys.RangeMetaKey(splitKey)
+		expectedKeys = append(expectedKeys, meta2Key.AsRawKey())
+	}
+
 	// Resort the list. The sql values are not sorted.
 	sort.Sort(expectedKeys)
 
 	if !reflect.DeepEqual(foundKeys, expectedKeys) {
-		t.Errorf("expected keys mismatch:\n%s\n  -- vs. -- \n\n%s",
+		t.Errorf("expected keys mismatch (found vs expected):\n%s\n  -- vs. -- \n\n%s",
 			formatKeys(foundKeys), formatKeys(expectedKeys))
 	}
 
@@ -243,10 +260,11 @@ func TestBootstrapCluster(t *testing.T) {
 // stores and verifies both stores are added and started.
 func TestBootstrapNewStore(t *testing.T) {
 	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
 	e := engine.NewInMem(roachpb.Attributes{}, 1<<20)
 	cfg := bootstrapNodeConfig()
 	if _, err := bootstrapCluster(
-		context.TODO(), cfg, []engine.Engine{e}, cfg.Settings.Version.BootstrapVersion(),
+		ctx, cfg, []engine.Engine{e}, cfg.Settings.Version.BootstrapVersion(),
 		kv.MakeTxnMetrics(metric.TestSampleInterval),
 	); err != nil {
 		t.Fatal(err)
@@ -260,13 +278,14 @@ func TestBootstrapNewStore(t *testing.T) {
 	})
 	defer engines.Close()
 	_, _, node, stopper := createAndStartTestNode(
+		ctx,
 		util.TestAddr,
 		engines,
 		util.TestAddr,
 		roachpb.Locality{},
 		t,
 	)
-	defer stopper.Stop(context.TODO())
+	defer stopper.Stop(ctx)
 
 	// Non-initialized stores (in this case the new in-memory-based
 	// store) will be bootstrapped by the node upon start. This happens
@@ -308,14 +327,17 @@ func bootstrapNodeConfig() storage.StoreConfig {
 // cluster consisting of one node.
 func TestNodeJoin(t *testing.T) {
 	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
 	engineStopper := stop.NewStopper()
-	defer engineStopper.Stop(context.TODO())
+	defer engineStopper.Stop(ctx)
 	e := engine.NewInMem(roachpb.Attributes{}, 1<<20)
 	engineStopper.AddCloser(e)
 
 	cfg := bootstrapNodeConfig()
 	if _, err := bootstrapCluster(
-		context.TODO(), cfg, []engine.Engine{e}, cfg.Settings.Version.BootstrapVersion(), kv.MakeTxnMetrics(metric.TestSampleInterval),
+		ctx, cfg, []engine.Engine{e},
+		cfg.Settings.Version.BootstrapVersion(),
+		kv.MakeTxnMetrics(metric.TestSampleInterval),
 	); err != nil {
 		t.Fatal(err)
 	}
@@ -323,26 +345,28 @@ func TestNodeJoin(t *testing.T) {
 	// Start the bootstrap node.
 	engines1 := []engine.Engine{e}
 	_, server1Addr, node1, stopper1 := createAndStartTestNode(
+		ctx,
 		util.TestAddr,
 		engines1,
 		util.TestAddr,
 		roachpb.Locality{},
 		t,
 	)
-	defer stopper1.Stop(context.TODO())
+	defer stopper1.Stop(ctx)
 
 	// Create a new node.
 	e2 := engine.NewInMem(roachpb.Attributes{}, 1<<20)
 	engineStopper.AddCloser(e2)
 	engines2 := []engine.Engine{e2}
 	_, server2Addr, node2, stopper2 := createAndStartTestNode(
+		ctx,
 		util.TestAddr,
 		engines2,
 		server1Addr,
 		roachpb.Locality{},
 		t,
 	)
-	defer stopper2.Stop(context.TODO())
+	defer stopper2.Stop(ctx)
 
 	// Verify new node is able to bootstrap its store.
 	testutils.SucceedsSoon(t, func() error {
@@ -425,13 +449,13 @@ func TestCorruptedClusterID(t *testing.T) {
 // And that UpdatedAt has increased.
 // The latest actual stats are returned.
 func compareNodeStatus(
-	t *testing.T, ts *TestServer, expectedNodeStatus *status.NodeStatus, testNumber int,
-) *status.NodeStatus {
+	t *testing.T, ts *TestServer, expectedNodeStatus *statuspb.NodeStatus, testNumber int,
+) *statuspb.NodeStatus {
 	// ========================================
 	// Read NodeStatus from server and validate top-level fields.
 	// ========================================
 	nodeStatusKey := keys.NodeStatusKey(ts.node.Descriptor.NodeID)
-	nodeStatus := &status.NodeStatus{}
+	nodeStatus := &statuspb.NodeStatus{}
 	if err := ts.db.GetProto(context.TODO(), nodeStatusKey, nodeStatus); err != nil {
 		t.Fatalf("%d: failure getting node status: %s", testNumber, err)
 	}
@@ -444,8 +468,8 @@ func compareNodeStatus(
 	// ========================================
 	// Ensure all expected stores are represented in the node status.
 	// ========================================
-	storesToMap := func(ns *status.NodeStatus) map[roachpb.StoreID]status.StoreStatus {
-		strMap := make(map[roachpb.StoreID]status.StoreStatus, len(ns.StoreStatuses))
+	storesToMap := func(ns *statuspb.NodeStatus) map[roachpb.StoreID]statuspb.StoreStatus {
+		strMap := make(map[roachpb.StoreID]statuspb.StoreStatus, len(ns.StoreStatuses))
 		for _, str := range ns.StoreStatuses {
 			strMap[str.Desc.StoreID] = str
 		}
@@ -590,7 +614,7 @@ func TestNodeStatusWritten(t *testing.T) {
 	// Construct an initial expectation for NodeStatus to compare to the first
 	// status produced by the server.
 	// ========================================
-	expectedNodeStatus := &status.NodeStatus{
+	expectedNodeStatus := &statuspb.NodeStatus{
 		Desc:      ts.node.Descriptor,
 		StartedAt: 0,
 		UpdatedAt: 0,
@@ -600,7 +624,7 @@ func TestNodeStatusWritten(t *testing.T) {
 		},
 	}
 
-	expectedStoreStatuses := make(map[roachpb.StoreID]status.StoreStatus)
+	expectedStoreStatuses := make(map[roachpb.StoreID]statuspb.StoreStatus)
 	if err := ts.node.stores.VisitStores(func(s *storage.Store) error {
 		desc, err := s.Descriptor(false /* useCached */)
 		if err != nil {
@@ -610,7 +634,7 @@ func TestNodeStatusWritten(t *testing.T) {
 		if s.StoreID() == roachpb.StoreID(1) {
 			expectedReplicas = initialRanges
 		}
-		stat := status.StoreStatus{
+		stat := statuspb.StoreStatus{
 			Desc: *desc,
 			Metrics: map[string]float64{
 				"replicas":              float64(expectedReplicas),
@@ -724,24 +748,28 @@ func TestNodeStatusWritten(t *testing.T) {
 // collection of different localities.
 func TestStartNodeWithLocality(t *testing.T) {
 	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
 
 	testLocalityWithNewNode := func(locality roachpb.Locality) {
 		e := engine.NewInMem(roachpb.Attributes{}, 1<<20)
 		defer e.Close()
 		cfg := bootstrapNodeConfig()
 		if _, err := bootstrapCluster(
-			context.TODO(), cfg, []engine.Engine{e}, cfg.Settings.Version.BootstrapVersion(), kv.MakeTxnMetrics(metric.TestSampleInterval),
+			ctx, cfg, []engine.Engine{e},
+			cfg.Settings.Version.BootstrapVersion(),
+			kv.MakeTxnMetrics(metric.TestSampleInterval),
 		); err != nil {
 			t.Fatal(err)
 		}
 		_, _, node, stopper := createAndStartTestNode(
+			ctx,
 			util.TestAddr,
 			[]engine.Engine{e},
 			util.TestAddr,
 			locality,
 			t,
 		)
-		defer stopper.Stop(context.TODO())
+		defer stopper.Stop(ctx)
 
 		// Check the node to make sure the locality was propagated to its
 		// nodeDescriptor.
