@@ -31,12 +31,6 @@ import (
 	"sync/atomic"
 	"time"
 
-	raven "github.com/getsentry/raven-go"
-	gwruntime "github.com/grpc-ecosystem/grpc-gateway/runtime"
-	opentracing "github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-	"google.golang.org/grpc"
-
 	"github.com/cockroachdb/cmux"
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
@@ -47,22 +41,27 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/rpc"
 	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
 	"github.com/cockroachdb/cockroach/pkg/server/debug"
+	"github.com/cockroachdb/cockroach/pkg/server/goroutinedumper"
 	"github.com/cockroachdb/cockroach/pkg/server/heapprofiler"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
 	"github.com/cockroachdb/cockroach/pkg/server/status"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/distsqlrun"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire"
+	"github.com/cockroachdb/cockroach/pkg/sql/querycache"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlutil"
 	"github.com/cockroachdb/cockroach/pkg/sql/stats"
 	"github.com/cockroachdb/cockroach/pkg/sqlmigrations"
 	"github.com/cockroachdb/cockroach/pkg/storage"
+	"github.com/cockroachdb/cockroach/pkg/storage/bulk"
 	"github.com/cockroachdb/cockroach/pkg/storage/closedts/container"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
 	"github.com/cockroachdb/cockroach/pkg/ts"
 	"github.com/cockroachdb/cockroach/pkg/ui"
 	"github.com/cockroachdb/cockroach/pkg/util"
@@ -81,6 +80,11 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	raven "github.com/getsentry/raven-go"
+	gwruntime "github.com/grpc-ecosystem/grpc-gateway/runtime"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
+	"google.golang.org/grpc"
 )
 
 var (
@@ -106,7 +110,7 @@ var (
 
 	forwardClockJumpCheckEnabled = settings.RegisterBoolSetting(
 		"server.clock.forward_jump_check_enabled",
-		"if enabled, forward clock jumps > max_offset/2 will cause a panic.",
+		"if enabled, forward clock jumps > max_offset/2 will cause a panic",
 		false,
 	)
 
@@ -148,12 +152,14 @@ func (mux *safeServeMux) ServeHTTP(w http.ResponseWriter, r *http.Request) {
 type Server struct {
 	nodeIDContainer base.NodeIDContainer
 
-	cfg              Config
-	st               *cluster.Settings
-	mux              safeServeMux
-	clock            *hlc.Clock
-	rpcContext       *rpc.Context
-	grpc             *grpc.Server
+	cfg        Config
+	st         *cluster.Settings
+	mux        safeServeMux
+	clock      *hlc.Clock
+	startTime  time.Time
+	rpcContext *rpc.Context
+	// The gRPC server on which the different RPC handlers will be registered.
+	grpc             *grpcServer
 	gossip           *gossip.Gossip
 	nodeDialer       *nodedialer.Dialer
 	nodeLiveness     *storage.NodeLiveness
@@ -182,12 +188,12 @@ type Server struct {
 	// shared between the sql.Server and the statusServer.
 	sessionRegistry    *sql.SessionRegistry
 	jobRegistry        *jobs.Registry
+	statsRefresher     *stats.Refresher
 	engines            Engines
 	internalMemMetrics sql.MemoryMetrics
 	adminMemMetrics    sql.MemoryMetrics
 	// sqlMemMetrics are used to track memory usage of sql sessions.
 	sqlMemMetrics sql.MemoryMetrics
-	serveMode
 }
 
 // NewServer creates a Server from a server.Config.
@@ -210,7 +216,6 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 		cfg:      cfg,
 		registry: metric.NewRegistry(),
 	}
-	s.serveMode.set(modeInitializing)
 
 	// If the tracer has a Close function, call it after the server stops.
 	if tr, ok := cfg.AmbientCtx.Tracer.(stop.Closer); ok {
@@ -249,14 +254,14 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 		}
 	}
 
-	s.grpc = rpc.NewServerWithInterceptor(s.rpcContext, s.Intercept())
+	s.grpc = newGRPCServer(s.rpcContext)
 
 	s.gossip = gossip.New(
 		s.cfg.AmbientCtx,
 		&s.rpcContext.ClusterID,
 		&s.nodeIDContainer,
 		s.rpcContext,
-		s.grpc,
+		s.grpc.Server,
 		s.stopper,
 		s.registry,
 		s.cfg.Locality,
@@ -339,7 +344,7 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 	)
 
 	s.raftTransport = storage.NewRaftTransport(
-		s.cfg.AmbientCtx, st, s.nodeDialer, s.grpc, s.stopper,
+		s.cfg.AmbientCtx, st, s.nodeDialer, s.grpc.Server, s.stopper,
 	)
 
 	// Set up internal memory metrics for use by internal SQL executors.
@@ -442,6 +447,7 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 		StorePool:               s.storePool,
 		SQLExecutor:             internalExecutor,
 		LogRangeEvents:          s.cfg.EventLogEnabled,
+		RangeDescriptorCache:    s.distSender.RangeDescriptorCache(),
 		TimeSeriesDataStore:     s.tsDB,
 
 		// Initialize the closed timestamp subsystem. Note that it won't
@@ -480,9 +486,9 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 	s.node = NewNode(
 		storeCfg, s.recorder, s.registry, s.stopper,
 		txnMetrics, nil /* execCfg */, &s.rpcContext.ClusterID)
-	roachpb.RegisterInternalServer(s.grpc, s.node)
-	storage.RegisterPerReplicaServer(s.grpc, s.node.perReplicaServer)
-	s.node.storeCfg.ClosedTimestamp.RegisterClosedTimestampServer(s.grpc)
+	roachpb.RegisterInternalServer(s.grpc.Server, s.node)
+	storage.RegisterPerReplicaServer(s.grpc.Server, s.node.perReplicaServer)
+	s.node.storeCfg.ClosedTimestamp.RegisterClosedTimestampServer(s.grpc.Server)
 
 	s.sessionRegistry = sql.NewSessionRegistry()
 	s.jobRegistry = jobs.MakeRegistry(
@@ -509,6 +515,7 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 	distSQLCfg := distsqlrun.ServerConfig{
 		AmbientContext: s.cfg.AmbientCtx,
 		Settings:       st,
+		RuntimeStats:   s.runtime,
 		DB:             s.db,
 		Executor:       internalExecutor,
 		FlowDB:         client.NewDB(s.cfg.AmbientCtx, s.tcsFactory, s.clock),
@@ -518,6 +525,9 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 		ClusterID:      &s.rpcContext.ClusterID,
 
 		TempStorage: tempEngine,
+		BulkAdder: func(ctx context.Context, db *client.DB, bufferSize, flushSize int64, ts hlc.Timestamp) (storagebase.BulkAdder, error) {
+			return bulk.MakeBulkAdder(db, s.distSender.RangeDescriptorCache(), bufferSize, flushSize, ts)
+		},
 		DiskMonitor: s.cfg.TempStorageConfig.Mon,
 
 		ParentMemoryMonitor: &rootSQLMemoryMonitor,
@@ -534,7 +544,7 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 	}
 
 	s.distSQLServer = distsqlrun.NewServer(ctx, distSQLCfg)
-	distsqlrun.RegisterDistSQLServer(s.grpc, s.distSQLServer)
+	distsqlpb.RegisterDistSQLServer(s.grpc.Server, s.distSQLServer)
 
 	s.admin = newAdminServer(s)
 	s.status = newStatusServer(
@@ -554,13 +564,17 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 	)
 	s.authentication = newAuthenticationServer(s)
 	for _, gw := range []grpcGatewayServer{s.admin, s.status, s.authentication, &s.tsServer} {
-		gw.RegisterService(s.grpc)
+		gw.RegisterService(s.grpc.Server)
 	}
 
-	s.initServer = newInitServer(s)
-	s.initServer.semaphore.acquire()
-
-	serverpb.RegisterInitServer(s.grpc, s.initServer)
+	// TODO(andrei): We're creating an initServer even through the inspection of
+	// our engines in Server.Start() might reveal that we're already bootstrapped
+	// and so we don't need to accept a Bootstrap RPC. The creation of this server
+	// early means that a Bootstrap RPC might erroneously succeed. We should
+	// figure out early if our engines are bootstrapped and, if they are, create a
+	// dummy implementation of the InitServer that rejects all RPCs.
+	s.initServer = newInitServer(s.gossip.Connected, s.stopper.ShouldStop())
+	serverpb.RegisterInitServer(s.grpc.Server, s.initServer)
 
 	nodeInfo := sql.NodeInfo{
 		AdminURL:  cfg.AdminURL,
@@ -576,16 +590,19 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 
 	// Set up Executor
 
-	var sqlExecutorTestingKnobs *sql.ExecutorTestingKnobs
+	var sqlExecutorTestingKnobs sql.ExecutorTestingKnobs
 	if k := s.cfg.TestingKnobs.SQLExecutor; k != nil {
-		sqlExecutorTestingKnobs = k.(*sql.ExecutorTestingKnobs)
+		sqlExecutorTestingKnobs = *k.(*sql.ExecutorTestingKnobs)
 	} else {
-		sqlExecutorTestingKnobs = new(sql.ExecutorTestingKnobs)
+		sqlExecutorTestingKnobs = sql.ExecutorTestingKnobs{}
 	}
 
+	loggerCtx, _ := s.stopper.WithCancelOnStop(ctx)
+
 	execCfg = sql.ExecutorConfig{
 		Settings:                s.st,
 		NodeInfo:                nodeInfo,
+		Locality:                s.cfg.Locality,
 		AmbientCtx:              s.cfg.AmbientCtx,
 		DB:                      s.db,
 		Gossip:                  s.gossip,
@@ -627,14 +644,14 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 		),
 
 		ExecLogger: log.NewSecondaryLogger(
-			nil /* dirName */, "sql-exec", true /* enableGc */, false, /*forceSyncWrites*/
+			loggerCtx, nil /* dirName */, "sql-exec", true /* enableGc */, false, /*forceSyncWrites*/
 		),
 
 		AuditLogger: log.NewSecondaryLogger(
-			s.cfg.SQLAuditLogDirName, "sql-audit", true /*enableGc*/, true, /*forceSyncWrites*/
+			loggerCtx, s.cfg.SQLAuditLogDirName, "sql-audit", true /*enableGc*/, true, /*forceSyncWrites*/
 		),
 
-		ConnResultsBufferBytes: s.cfg.ConnResultsBufferBytes,
+		QueryCache: querycache.New(s.cfg.SQLQueryCacheSize),
 	}
 
 	if sqlSchemaChangerTestingKnobs := s.cfg.TestingKnobs.SQLSchemaChanger; sqlSchemaChangerTestingKnobs != nil {
@@ -650,6 +667,17 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 	if sqlEvalContext := s.cfg.TestingKnobs.SQLEvalContext; sqlEvalContext != nil {
 		execCfg.EvalContextTestingKnobs = *sqlEvalContext.(*tree.EvalContextTestingKnobs)
 	}
+	if pgwireKnobs := s.cfg.TestingKnobs.PGWireTestingKnobs; pgwireKnobs != nil {
+		execCfg.PGWireTestingKnobs = pgwireKnobs.(*sql.PGWireTestingKnobs)
+	}
+
+	s.statsRefresher = stats.MakeRefresher(
+		s.st,
+		internalExecutor,
+		execCfg.TableStatsCache,
+		stats.DefaultAsOfTime,
+	)
+	execCfg.StatsRefresher = s.statsRefresher
 
 	// Set up internal memory metrics for use by internal SQL executors.
 	s.sqlMemMetrics = sql.MakeMemMetrics("sql", cfg.HistogramWindowInterval())
@@ -666,24 +694,24 @@ func NewServer(cfg Config, stopper *stop.Stopper) (*Server, error) {
 
 	// Now that we have a pgwire.Server (which has a sql.Server), we can close a
 	// circular dependency between the distsqlrun.Server and sql.Server and set
-	// SessionBoundInternalExecutorCtor.
+	// SessionBoundInternalExecutorFactory.
 	s.distSQLServer.ServerConfig.SessionBoundInternalExecutorFactory =
 		func(
 			ctx context.Context, sessionData *sessiondata.SessionData,
 		) sqlutil.InternalExecutor {
-			ie := sql.MakeSessionBoundInternalExecutor(
+			ie := sql.NewSessionBoundInternalExecutor(
 				ctx,
 				sessionData,
 				s.pgServer.SQLServer,
 				s.sqlMemMetrics,
 				s.st,
 			)
-			return &ie
+			return ie
 		}
 
-	s.registry.AddMetricStruct(s.pgServer.Metrics())
-	s.registry.AddMetricStruct(s.pgServer.StatementCounters())
-	s.registry.AddMetricStruct(s.pgServer.EngineMetrics())
+	for _, m := range s.pgServer.Metrics() {
+		s.registry.AddMetricStruct(m)
+	}
 	*internalExecutor = sql.MakeInternalExecutor(
 		ctx, s.pgServer.SQLServer, s.internalMemMetrics, s.ClusterSettings(),
 	)
@@ -756,8 +784,7 @@ type ListenError struct {
 func inspectEngines(
 	ctx context.Context,
 	engines []engine.Engine,
-	minVersion,
-	serverVersion roachpb.Version,
+	minVersion, serverVersion roachpb.Version,
 	clusterIDContainer *base.ClusterIDContainer,
 ) (
 	bootstrappedEngines []engine.Engine,
@@ -1062,7 +1089,7 @@ func (s *Server) Start(ctx context.Context) error {
 	}
 	ctx = s.AnnotateCtx(ctx)
 
-	startTime := timeutil.Now()
+	s.startTime = timeutil.Now()
 	s.startMonitoringForwardClockJumps(ctx)
 
 	uiTLSConfig, err := s.cfg.GetUIServerTLSConfig()
@@ -1310,6 +1337,8 @@ func (s *Server) Start(ctx context.Context) error {
 	}
 	s.stopper.AddCloser(&s.engines)
 
+	startAssertEngineHealth(ctx, s.stopper, s.engines)
+
 	// Write listener info files early in the startup sequence. `listenerInfo` has a comment.
 	listenerFiles := listenerInfo{
 		advertise: s.cfg.AdvertiseAddr,
@@ -1349,8 +1378,11 @@ func (s *Server) Start(ctx context.Context) error {
 	}
 
 	var hlcUpperBoundExists bool
+	// doBootstrap is set if we're the ones who bootstrapped the cluster.
+	var doBootstrap bool
 	if len(bootstrappedEngines) > 0 {
 		// The cluster was already initialized.
+		doBootstrap = false
 		if s.cfg.ReadyFn != nil {
 			s.cfg.ReadyFn(false /*waitForInit*/)
 		}
@@ -1367,17 +1399,17 @@ func (s *Server) Start(ctx context.Context) error {
 		ensureClockMonotonicity(
 			ctx,
 			s.clock,
-			startTime,
+			s.startTime,
 			hlcUpperBound,
 			timeutil.SleepUntil,
 		)
-
 	} else if len(s.cfg.GossipBootstrapResolvers) == 0 {
 		// If the _unfiltered_ list of hosts from the --join flag is
 		// empty, then this node can bootstrap a new cluster. We disallow
 		// this if this node is being started with itself specified as a
 		// --join host, because that's too likely to be operator error.
 		//
+		doBootstrap = true
 		if s.cfg.ReadyFn != nil {
 			// TODO(knz): when CockroachDB stops auto-initializing when --join
 			// is not specified, this needs to be adjusted as well. See issue
@@ -1386,25 +1418,17 @@ func (s *Server) Start(ctx context.Context) error {
 			s.cfg.ReadyFn(false /*waitForInit*/)
 		}
 
-		bootstrapVersion := s.cfg.Settings.Version.BootstrapVersion()
-		if s.cfg.TestingKnobs.Store != nil {
-			if storeKnobs, ok := s.cfg.TestingKnobs.Store.(*storage.StoreTestingKnobs); ok && storeKnobs.BootstrapVersion != nil {
-				bootstrapVersion = *storeKnobs.BootstrapVersion
-			}
-		}
-		if err := s.node.bootstrap(ctx, s.engines, bootstrapVersion); err != nil {
+		if err := s.bootstrapCluster(ctx); err != nil {
 			return err
 		}
+
 		log.Infof(ctx, "**** add additional nodes by specifying --join=%s", s.cfg.AdvertiseAddr)
 	} else {
-		if s.cfg.ReadyFn != nil {
-			s.cfg.ReadyFn(true /*waitForInit*/)
-		}
-		log.Info(ctx, "no stores bootstrapped and --join flag specified, awaiting init command.")
-
-		// Note that when we created the init server, we acquired its semaphore
-		// (to stop anyone from rushing in).
-		s.initServer.semaphore.release()
+		// We have no existing stores and we've been told to join a cluster. Wait
+		// for the initServer to bootstrap the cluster or connect to an existing
+		// one.
+		//
+		// TODO(knz): This may need tweaking when #24118 is addressed.
 
 		s.stopper.RunWorker(workersCtx, func(context.Context) {
 			serveOnMux.Do(func() {
@@ -1412,20 +1436,46 @@ func (s *Server) Start(ctx context.Context) error {
 			})
 		})
 
-		if err := s.initServer.awaitBootstrap(); err != nil {
+		ready := make(chan struct{})
+		if s.cfg.ReadyFn != nil {
+			// s.cfg.ReadyFn must be called in any case because the `start`
+			// command requires it to signal readiness to a process manager.
+			//
+			// However we want to be somewhat precisely informative to the user
+			// about whether the node is waiting on init / join, or whether
+			// the join was successful straight away. So we spawn this gogoroutine
+			// and either:
+			// - its timer will fire after 2 seconds and we call ReadyFn(true)
+			// - bootstrap completes earlier and the ready chan gets closed,
+			//   then we call ReadyFn(false).
+			go func() {
+				waitForInit := false
+				tm := time.After(2 * time.Second)
+				select {
+				case <-tm:
+					waitForInit = true
+				case <-ready:
+				}
+				s.cfg.ReadyFn(waitForInit)
+			}()
+		}
+
+		log.Info(ctx, "no stores bootstrapped and --join flag specified, awaiting init command or join with an already initialized node.")
+
+		initRes, err := s.initServer.awaitBootstrap()
+		close(ready)
+		if err != nil {
 			return err
 		}
 
-		// Reacquire the semaphore, allowing the code below to be oblivious to
-		// the fact that this branch was taken.
-		s.initServer.semaphore.acquire()
+		doBootstrap = initRes == needBootstrap
+		if doBootstrap {
+			if err := s.bootstrapCluster(ctx); err != nil {
+				return err
+			}
+		}
 	}
 
-	// Release the semaphore of the init server. Anyone still managing to talk
-	// to it may do so, but will be greeted with an error telling them that the
-	// cluster is already initialized.
-	s.initServer.semaphore.release()
-
 	// This opens the main listener.
 	s.stopper.RunWorker(workersCtx, func(context.Context) {
 		serveOnMux.Do(func() {
@@ -1523,12 +1573,14 @@ func (s *Server) Start(ctx context.Context) error {
 		*s.db,
 		s.node.Descriptor,
 		s.execCfg.DistSQLPlanner,
+		// We're reusing the ieFactory from the distSQLServer.
+		s.distSQLServer.ServerConfig.SessionBoundInternalExecutorFactory,
 	).Start(s.stopper)
 
 	s.distSQLServer.Start()
 	s.pgServer.Start(ctx, s.stopper)
 
-	s.serveMode.set(modeOperational)
+	s.grpc.setMode(modeOperational)
 
 	log.Infof(ctx, "starting %s server at %s (use: %s)",
 		s.cfg.HTTPRequestScheme(), s.cfg.HTTPAddr, s.cfg.HTTPAdvertiseAddr)
@@ -1561,6 +1613,11 @@ func (s *Server) Start(ctx context.Context) error {
 		}
 	}
 
+	// Start the background thread for periodically refreshing table statistics.
+	if err := s.statsRefresher.Start(ctx, s.stopper, stats.DefaultRefreshInterval); err != nil {
+		return err
+	}
+
 	// Before serving SQL requests, we have to make sure the database is
 	// in an acceptable form for this version of the software.
 	// We have to do this after actually starting up the server to be able to
@@ -1577,7 +1634,14 @@ func (s *Server) Start(ctx context.Context) error {
 		mmKnobs,
 		s.NodeID().String(),
 	)
-	if err := migMgr.EnsureMigrations(ctx); err != nil {
+	migrationFilter := sqlmigrations.AllMigrations
+	if doBootstrap {
+		// If we've just bootstrapped, some migrations are unnecessary because
+		// they're included in the metadata schema that we've written to the store
+		// by hand.
+		migrationFilter = sqlmigrations.ExcludeMigrationsIncludedInBootstrap
+	}
+	if err := migMgr.EnsureMigrations(ctx, migrationFilter); err != nil {
 		select {
 		case <-s.stopper.ShouldQuiesce():
 			// Avoid turning an early shutdown into a fatal error. See #19579.
@@ -1603,6 +1667,7 @@ func (s *Server) Start(ctx context.Context) error {
 		ui.Handler(ui.Config{
 			ExperimentalUseLogin: s.cfg.EnableWebSessionAuthentication,
 			LoginEnabled:         s.cfg.RequireWebSession(),
+			NodeID:               &s.nodeIDContainer,
 			GetUser: func(ctx context.Context) *string {
 				if u, ok := ctx.Value(webSessionUserKey{}).(string); ok {
 					return &u
@@ -1643,12 +1708,7 @@ func (s *Server) Start(ctx context.Context) error {
 			connCtx := logtags.AddTag(pgCtx, "client", conn.RemoteAddr().String())
 			setTCPKeepAlive(connCtx, conn)
 
-			// Unless this is a simple disconnect or context timeout, report the error on
-			// this connection's context, so that we know which remote client caused the
-			// error when looking at the logs. Note that we pass a non-cancelable context
-			// in, but the callee eventually wraps the context so that it can get
-			// canceled and it may return the error here.
-			if err := errors.Cause(s.pgServer.ServeConn(connCtx, conn)); err != nil && !netutil.IsClosedConnection(err) && err != context.Canceled && err != context.DeadlineExceeded {
+			if err := s.pgServer.ServeConn(connCtx, conn); err != nil {
 				log.Error(connCtx, err)
 			}
 		}))
@@ -1677,17 +1737,15 @@ func (s *Server) Start(ctx context.Context) error {
 			}
 			netutil.FatalIfUnexpected(httpServer.ServeWith(pgCtx, s.stopper, unixLn, func(conn net.Conn) {
 				connCtx := logtags.AddTag(pgCtx, "client", conn.RemoteAddr().String())
-				if err := s.pgServer.ServeConn(connCtx, conn); err != nil &&
-					!netutil.IsClosedConnection(err) {
-					// Report the error on this connection's context, so that we
-					// know which remote client caused the error when looking at
-					// the logs.
+				if err := s.pgServer.ServeConn(connCtx, conn); err != nil {
 					log.Error(connCtx, err)
 				}
 			}))
 		})
 	}
 
+	s.startSystemLogsGC(ctx)
+
 	// Record that this node joined the cluster in the event log. Since this
 	// executes a SQL query, this must be done after the SQL layer is ready.
 	s.node.recordJoinEvent()
@@ -1701,6 +1759,31 @@ func (s *Server) Start(ctx context.Context) error {
 	return nil
 }
 
+func (s *Server) bootstrapCluster(ctx context.Context) error {
+	bootstrapVersion := s.cfg.Settings.Version.BootstrapVersion()
+	if s.cfg.TestingKnobs.Store != nil {
+		if storeKnobs, ok := s.cfg.TestingKnobs.Store.(*storage.StoreTestingKnobs); ok && storeKnobs.BootstrapVersion != nil {
+			bootstrapVersion = *storeKnobs.BootstrapVersion
+		}
+	}
+
+	if err := s.node.bootstrap(ctx, s.engines, bootstrapVersion); err != nil {
+		return err
+	}
+	// Force all the system ranges through the replication queue so they
+	// upreplicate as quickly as possible when a new node joins. Without this
+	// code, the upreplication would be up to the whim of the scanner, which
+	// might be too slow for new clusters.
+	done := false
+	return s.node.stores.VisitStores(func(store *storage.Store) error {
+		if !done {
+			done = true
+			return store.ForceReplicationScanAndProcess()
+		}
+		return nil
+	})
+}
+
 func (s *Server) doDrain(
 	ctx context.Context, modes []serverpb.DrainMode, setTo bool,
 ) ([]serverpb.DrainMode, error) {
@@ -1708,7 +1791,7 @@ func (s *Server) doDrain(
 		switch mode {
 		case serverpb.DrainMode_CLIENT:
 			if setTo {
-				s.serveMode.set(modeDraining)
+				s.grpc.setMode(modeDraining)
 				// Wait for drainUnreadyWait. This will fail load balancer checks and
 				// delay draining so that client traffic can move off this node.
 				time.Sleep(drainWait.Get(&s.st.SV))
@@ -1716,7 +1799,7 @@ func (s *Server) doDrain(
 			if err := func() error {
 				if !setTo {
 					// Execute this last.
-					defer func() { s.serveMode.set(modeOperational) }()
+					defer func() { s.grpc.setMode(modeOperational) }()
 				}
 				// Since enabling the lease manager's draining mode will prevent
 				// the acquisition of new leases, the switch must be made after
@@ -1766,7 +1849,7 @@ func (s *Server) doDrain(
 // On failure, the system may be in a partially drained state and should be
 // recovered by calling Undrain() with the same (or a larger) slice of modes.
 func (s *Server) Drain(ctx context.Context, on []serverpb.DrainMode) ([]serverpb.DrainMode, error) {
-	return s.doDrain(ctx, on, true)
+	return s.doDrain(ctx, on, true /* setTo */)
 }
 
 // Undrain idempotently deactivates the given DrainModes on the Server in the
@@ -1816,6 +1899,10 @@ func (s *Server) Decommission(ctx context.Context, setTo bool, nodeIDs []roachpb
 func (s *Server) startSampleEnvironment(frequency time.Duration) {
 	// Immediately record summaries once on server startup.
 	ctx := s.AnnotateCtx(context.Background())
+	goroutineDumper, err := goroutinedumper.NewGoroutineDumper(s.cfg.GoroutineDumpDirName)
+	if err != nil {
+		log.Infof(ctx, "Could not start goroutine dumper worker due to: %s", err)
+	}
 	var heapProfiler *heapprofiler.HeapProfiler
 
 	{
@@ -1860,6 +1947,9 @@ func (s *Server) startSampleEnvironment(frequency time.Duration) {
 			case <-timer.C:
 				timer.Read = true
 				s.runtime.SampleEnvironment(ctx)
+				if goroutineDumper != nil {
+					goroutineDumper.MaybeDump(ctx, s.ClusterSettings(), s.runtime.Goroutines.Value())
+				}
 				if heapProfiler != nil {
 					heapProfiler.MaybeTakeProfile(ctx, s.ClusterSettings(), s.runtime.Rss.Value())
 				}
