@@ -26,6 +26,7 @@ import (
 	"io/ioutil"
 	"net/http"
 	"os"
+	"os/exec"
 	"path/filepath"
 	"reflect"
 	"regexp"
@@ -35,16 +36,6 @@ import (
 	"strings"
 	"sync"
 
-	"github.com/cockroachdb/cockroach/pkg/config"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine"
-	gwruntime "github.com/grpc-ecosystem/grpc-gateway/runtime"
-	"github.com/pkg/errors"
-	"go.etcd.io/etcd/raft"
-	"google.golang.org/grpc"
-	"google.golang.org/grpc/codes"
-	"google.golang.org/grpc/metadata"
-	grpcstatus "google.golang.org/grpc/status"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
@@ -56,16 +47,27 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/server/debug"
 	"github.com/cockroachdb/cockroach/pkg/server/diagnosticspb"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
-	"github.com/cockroachdb/cockroach/pkg/server/status"
+	"github.com/cockroachdb/cockroach/pkg/server/status/statuspb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql"
 	"github.com/cockroachdb/cockroach/pkg/storage"
+	"github.com/cockroachdb/cockroach/pkg/storage/engine"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/httputil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	gwruntime "github.com/grpc-ecosystem/grpc-gateway/runtime"
+	"github.com/pkg/errors"
+	"go.etcd.io/etcd/raft"
+	"google.golang.org/grpc"
+	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/metadata"
+	"google.golang.org/grpc/status"
+	grpcstatus "google.golang.org/grpc/status"
 )
 
 const (
@@ -135,6 +137,7 @@ type statusServer struct {
 	stores          *storage.Stores
 	stopper         *stop.Stopper
 	sessionRegistry *sql.SessionRegistry
+	si              systemInfoOnce
 }
 
 // newStatusServer allocates and returns a statusServer.
@@ -241,6 +244,49 @@ func (s *statusServer) Gossip(
 	return status.Gossip(ctx, req)
 }
 
+func (s *statusServer) EngineStats(
+	ctx context.Context, req *serverpb.EngineStatsRequest,
+) (*serverpb.EngineStatsResponse, error) {
+	ctx = propagateGatewayMetadata(ctx)
+	ctx = s.AnnotateCtx(ctx)
+	nodeID, local, err := s.parseNodeID(req.NodeId)
+	if err != nil {
+		return nil, grpcstatus.Errorf(codes.InvalidArgument, err.Error())
+	}
+
+	if !local {
+		status, err := s.dialNode(ctx, nodeID)
+		if err != nil {
+			return nil, err
+		}
+		return status.EngineStats(ctx, req)
+	}
+
+	resp := new(serverpb.EngineStatsResponse)
+	err = s.stores.VisitStores(func(store *storage.Store) error {
+		if rocksdb, ok := store.Engine().(*engine.RocksDB); ok {
+			tickersAndHistograms, err := rocksdb.GetTickersAndHistograms()
+			if err != nil {
+				return grpcstatus.Errorf(codes.Internal, err.Error())
+			}
+			resp.Stats = append(
+				resp.Stats,
+				serverpb.EngineStatsInfo{
+					StoreID:              store.Ident.StoreID,
+					TickersAndHistograms: tickersAndHistograms,
+				},
+			)
+		} else {
+			return grpcstatus.Errorf(codes.Internal, "engine is not rocksdb")
+		}
+		return nil
+	})
+	if err != nil {
+		return nil, err
+	}
+	return resp, nil
+}
+
 // Allocator returns simulated allocator info for the ranges on the given node.
 func (s *statusServer) Allocator(
 	ctx context.Context, req *serverpb.AllocatorRequest,
@@ -276,6 +322,9 @@ func (s *statusServer) Allocator(
 				func(desc roachpb.RangeDescriptor) (bool, error) {
 					rep, err := store.GetReplica(desc.RangeID)
 					if err != nil {
+						if _, skip := err.(*roachpb.RangeNotFoundError); skip {
+							return true, nil // continue
+						}
 						return true, err
 					}
 					if !rep.OwnsValidLease(store.Clock().Now()) {
@@ -359,44 +408,43 @@ func (s *statusServer) AllocatorRange(
 
 	ctx = propagateGatewayMetadata(ctx)
 	ctx = s.AnnotateCtx(ctx)
-	nodeCtx, cancel := context.WithTimeout(ctx, base.NetworkTimeout)
-	defer cancel()
-
 	isLiveMap := s.nodeLiveness.GetIsLiveMap()
 	type nodeResponse struct {
 		nodeID roachpb.NodeID
 		resp   *serverpb.AllocatorResponse
 		err    error
 	}
-
 	responses := make(chan nodeResponse)
 	// TODO(bram): consider abstracting out this repeated pattern.
 	for nodeID := range isLiveMap {
 		nodeID := nodeID
 		if err := s.stopper.RunAsyncTask(
-			nodeCtx,
+			ctx,
 			"server.statusServer: requesting remote Allocator simulation",
 			func(ctx context.Context) {
-				status, err := s.dialNode(ctx, nodeID)
-				var allocatorResponse *serverpb.AllocatorResponse
-				if err == nil {
-					allocatorRequest := &serverpb.AllocatorRequest{
-						RangeIDs: []roachpb.RangeID{roachpb.RangeID(req.RangeId)},
+				_ = contextutil.RunWithTimeout(ctx, "allocator range", base.NetworkTimeout, func(ctx context.Context) error {
+					status, err := s.dialNode(ctx, nodeID)
+					var allocatorResponse *serverpb.AllocatorResponse
+					if err == nil {
+						allocatorRequest := &serverpb.AllocatorRequest{
+							RangeIDs: []roachpb.RangeID{roachpb.RangeID(req.RangeId)},
+						}
+						allocatorResponse, err = status.Allocator(ctx, allocatorRequest)
+					}
+					response := nodeResponse{
+						nodeID: nodeID,
+						resp:   allocatorResponse,
+						err:    err,
 					}
-					allocatorResponse, err = status.Allocator(ctx, allocatorRequest)
-				}
-				response := nodeResponse{
-					nodeID: nodeID,
-					resp:   allocatorResponse,
-					err:    err,
-				}
 
-				select {
-				case responses <- response:
-					// Response processed.
-				case <-ctx.Done():
-					// Context completed, response no longer needed.
-				}
+					select {
+					case responses <- response:
+						// Response processed.
+					case <-ctx.Done():
+						// Context completed, response no longer needed.
+					}
+					return nil
+				})
 			}); err != nil {
 			return nil, grpcstatus.Errorf(codes.Internal, err.Error())
 		}
@@ -571,8 +619,9 @@ func (s *statusServer) Details(
 	}
 
 	resp := &serverpb.DetailsResponse{
-		NodeID:    s.gossip.NodeID.Get(),
-		BuildInfo: build.GetInfo(),
+		NodeID:     s.gossip.NodeID.Get(),
+		BuildInfo:  build.GetInfo(),
+		SystemInfo: s.si.systemInfo(ctx),
 	}
 	if addr, err := s.gossip.GetNodeIDAddress(s.gossip.NodeID.Get()); err == nil {
 		resp.Address = *addr
@@ -584,7 +633,7 @@ func (s *statusServer) Details(
 		return resp, nil
 	}
 
-	serveMode := s.admin.server.serveMode.get()
+	serveMode := s.admin.server.grpc.mode.get()
 	if serveMode != modeOperational {
 		return nil, grpcstatus.Error(codes.Unavailable, "node is not ready")
 	}
@@ -920,7 +969,7 @@ func (s *statusServer) Nodes(
 	rows := b.Results[0].Rows
 
 	resp := serverpb.NodesResponse{
-		Nodes: make([]status.NodeStatus, len(rows)),
+		Nodes: make([]statuspb.NodeStatus, len(rows)),
 	}
 	for i, row := range rows {
 		if err := row.ValueProto(&resp.Nodes[i]); err != nil {
@@ -947,7 +996,7 @@ func (s *statusServer) NodesWithLiveness(
 	for _, node := range nodes.Nodes {
 		nodeID := node.Desc.NodeID
 		livenessStatus := statusMap[nodeID]
-		if livenessStatus == storage.NodeLivenessStatus_DECOMMISSIONED {
+		if livenessStatus == storagepb.NodeLivenessStatus_DECOMMISSIONED {
 			// Skip over removed nodes.
 			continue
 		}
@@ -961,14 +1010,14 @@ func (s *statusServer) NodesWithLiveness(
 
 // NodeStatusWithLiveness combines a NodeStatus with a NodeLivenessStatus.
 type NodeStatusWithLiveness struct {
-	status.NodeStatus
-	LivenessStatus storage.NodeLivenessStatus
+	statuspb.NodeStatus
+	LivenessStatus storagepb.NodeLivenessStatus
 }
 
 // handleNodeStatus handles GET requests for a single node's status.
 func (s *statusServer) Node(
 	ctx context.Context, req *serverpb.NodeRequest,
-) (*status.NodeStatus, error) {
+) (*statuspb.NodeStatus, error) {
 	ctx = propagateGatewayMetadata(ctx)
 	ctx = s.AnnotateCtx(ctx)
 	nodeID, _, err := s.parseNodeID(req.NodeId)
@@ -984,7 +1033,7 @@ func (s *statusServer) Node(
 		return nil, grpcstatus.Errorf(codes.Internal, err.Error())
 	}
 
-	var nodeStatus status.NodeStatus
+	var nodeStatus statuspb.NodeStatus
 	if err := b.Results[0].Rows[0].ValueProto(&nodeStatus); err != nil {
 		err = errors.Errorf("could not unmarshal NodeStatus from %s: %s", key, err)
 		log.Error(ctx, err)
@@ -1034,14 +1083,6 @@ func (s *statusServer) RaftDebug(
 		},
 	}
 
-	// Subtract base.NetworkTimeout from the deadline so we have time to process
-	// the results and return them.
-	if deadline, ok := ctx.Deadline(); ok {
-		var cancel context.CancelFunc
-		ctx, cancel = context.WithDeadline(ctx, deadline.Add(-base.NetworkTimeout))
-		defer cancel()
-	}
-
 	// Parallelize fetching of ranges to minimize total time.
 	var wg sync.WaitGroup
 	for _, node := range nodes.Nodes {
@@ -1211,27 +1252,19 @@ func (s *statusServer) Ranges(
 				LeaderNotLeaseHolder:   metrics.Leader && metrics.LeaseValid && !metrics.Leaseholder,
 				NoRaftLeader:           !storage.HasRaftLeader(raftStatus) && !metrics.Quiescent,
 				Underreplicated:        metrics.Underreplicated,
+				Overreplicated:         metrics.Overreplicated,
 				NoLease:                metrics.Leader && !metrics.LeaseValid && !metrics.Quiescent,
 				QuiescentEqualsTicking: raftStatus != nil && metrics.Quiescent == metrics.Ticking,
 				RaftLogTooLarge:        metrics.RaftLogTooLarge,
 			},
-			CmdQLocal:   serverpb.CommandQueueMetrics(metrics.CmdQMetricsLocal),
-			CmdQGlobal:  serverpb.CommandQueueMetrics(metrics.CmdQMetricsGlobal),
-			LeaseStatus: metrics.LeaseStatus,
-			Quiescent:   metrics.Quiescent,
-			Ticking:     metrics.Ticking,
+			LatchesLocal:  metrics.LatchInfoLocal,
+			LatchesGlobal: metrics.LatchInfoGlobal,
+			LeaseStatus:   metrics.LeaseStatus,
+			Quiescent:     metrics.Quiescent,
+			Ticking:       metrics.Ticking,
 		}
 	}
 
-	cfg, ok := s.gossip.GetSystemConfig()
-	if !ok {
-		// Very little on the status pages requires the system config -- as of June
-		// 2017, only the underreplicated range metric does. Refusing to return a
-		// status page (that may help debug why the config isn't available) due to
-		// such a small piece of missing information is overly harsh.
-		log.Error(ctx, "system config not yet available, serving status page without it")
-		cfg = config.SystemConfig{}
-	}
 	isLiveMap := s.nodeLiveness.GetIsLiveMap()
 	clusterNodes := s.storePool.ClusterNodeCount()
 
@@ -1245,6 +1278,9 @@ func (s *statusServer) Ranges(
 			err := storage.IterateRangeDescriptors(ctx, store.Engine(),
 				func(desc roachpb.RangeDescriptor) (bool, error) {
 					rep, err := store.GetReplica(desc.RangeID)
+					if _, skip := err.(*roachpb.RangeNotFoundError); skip {
+						return true, nil // continue
+					}
 					if err != nil {
 						return true, err
 					}
@@ -1253,7 +1289,7 @@ func (s *statusServer) Ranges(
 							desc,
 							rep,
 							store.Ident.StoreID,
-							rep.Metrics(ctx, timestamp, cfg, isLiveMap, clusterNodes),
+							rep.Metrics(ctx, timestamp, isLiveMap, clusterNodes),
 						))
 					return false, nil
 				})
@@ -1273,7 +1309,7 @@ func (s *statusServer) Ranges(
 					*desc,
 					rep,
 					store.Ident.StoreID,
-					rep.Metrics(ctx, timestamp, cfg, isLiveMap, clusterNodes),
+					rep.Metrics(ctx, timestamp, isLiveMap, clusterNodes),
 				))
 		}
 		return nil
@@ -1284,6 +1320,91 @@ func (s *statusServer) Ranges(
 	return &output, nil
 }
 
+// HotRanges returns the hottest ranges on each store on the requested node(s).
+func (s *statusServer) HotRanges(
+	ctx context.Context, req *serverpb.HotRangesRequest,
+) (*serverpb.HotRangesResponse, error) {
+	ctx = propagateGatewayMetadata(ctx)
+	ctx = s.AnnotateCtx(ctx)
+
+	response := &serverpb.HotRangesResponse{
+		NodeID:            s.gossip.NodeID.Get(),
+		HotRangesByNodeID: make(map[roachpb.NodeID]serverpb.HotRangesResponse_NodeResponse),
+	}
+
+	if len(req.NodeID) > 0 {
+		requestedNodeID, local, err := s.parseNodeID(req.NodeID)
+		if err != nil {
+			return nil, status.Errorf(codes.InvalidArgument, err.Error())
+		}
+
+		// Only hot ranges from the local node.
+		if local {
+			response.HotRangesByNodeID[requestedNodeID] = s.localHotRanges(ctx)
+			return response, nil
+		}
+
+		// Only hot ranges from one non-local node.
+		status, err := s.dialNode(ctx, requestedNodeID)
+		if err != nil {
+			return nil, err
+		}
+		return status.HotRanges(ctx, req)
+	}
+
+	// Hot ranges from all nodes.
+	dialFn := func(ctx context.Context, nodeID roachpb.NodeID) (interface{}, error) {
+		client, err := s.dialNode(ctx, nodeID)
+		return client, err
+	}
+	remoteRequest := serverpb.HotRangesRequest{NodeID: "local"}
+	nodeFn := func(ctx context.Context, client interface{}, _ roachpb.NodeID) (interface{}, error) {
+		status := client.(serverpb.StatusClient)
+		return status.HotRanges(ctx, &remoteRequest)
+	}
+	responseFn := func(nodeID roachpb.NodeID, resp interface{}) {
+		hotRangesResp := resp.(*serverpb.HotRangesResponse)
+		response.HotRangesByNodeID[nodeID] = hotRangesResp.HotRangesByNodeID[nodeID]
+	}
+	errorFn := func(nodeID roachpb.NodeID, err error) {
+		response.HotRangesByNodeID[nodeID] = serverpb.HotRangesResponse_NodeResponse{
+			ErrorMessage: err.Error(),
+		}
+	}
+
+	if err := s.iterateNodes(ctx, "hot ranges", dialFn, nodeFn, responseFn, errorFn); err != nil {
+		return nil, err
+	}
+
+	return response, nil
+}
+
+func (s *statusServer) localHotRanges(ctx context.Context) serverpb.HotRangesResponse_NodeResponse {
+	var resp serverpb.HotRangesResponse_NodeResponse
+	includeRawKeys := debug.GatewayRemoteAllowed(ctx, s.st)
+	err := s.stores.VisitStores(func(store *storage.Store) error {
+		ranges := store.HottestReplicas()
+		storeResp := &serverpb.HotRangesResponse_StoreResponse{
+			StoreID:   store.StoreID(),
+			HotRanges: make([]serverpb.HotRangesResponse_HotRange, len(ranges)),
+		}
+		for i, r := range ranges {
+			storeResp.HotRanges[i].Desc = *r.Desc
+			if !includeRawKeys {
+				storeResp.HotRanges[i].Desc.StartKey = nil
+				storeResp.HotRanges[i].Desc.EndKey = nil
+			}
+			storeResp.HotRanges[i].QueriesPerSecond = r.QPS
+		}
+		resp.Stores = append(resp.Stores, storeResp)
+		return nil
+	})
+	if err != nil {
+		return serverpb.HotRangesResponse_NodeResponse{ErrorMessage: err.Error()}
+	}
+	return resp
+}
+
 // Range returns rangeInfos for all nodes in the cluster about a specific
 // range. It also returns the range history for that range as well.
 func (s *statusServer) Range(
@@ -1331,26 +1452,6 @@ func (s *statusServer) Range(
 	return response, nil
 }
 
-// CommandQueue returns a snapshot of the command queue state for the
-// specified range.
-func (s *statusServer) CommandQueue(
-	ctx context.Context, req *serverpb.CommandQueueRequest,
-) (*serverpb.CommandQueueResponse, error) {
-	rangeID := roachpb.RangeID(req.RangeId)
-	replica, err := s.stores.GetReplicaForRangeID(rangeID)
-	if err != nil {
-		return nil, err
-	}
-
-	if replica == nil {
-		return nil, roachpb.NewRangeNotFoundError(rangeID, 0)
-	}
-
-	return &serverpb.CommandQueueResponse{
-		Snapshot: replica.GetCommandQueueSnapshot(),
-	}, nil
-}
-
 // ListLocalSessions returns a list of SQL sessions on this node.
 func (s *statusServer) ListLocalSessions(
 	ctx context.Context, req *serverpb.ListSessionsRequest,
@@ -1428,10 +1529,12 @@ func (s *statusServer) iterateNodes(
 	responseChan := make(chan nodeResponse, numNodes)
 
 	nodeQuery := func(ctx context.Context, nodeID roachpb.NodeID) {
-		rpcCtx, cancel := context.WithTimeout(ctx, base.NetworkTimeout)
-		defer cancel()
-
-		client, err := dialFn(rpcCtx, nodeID)
+		var client interface{}
+		err := contextutil.RunWithTimeout(ctx, "dial node", base.NetworkTimeout, func(ctx context.Context) error {
+			var err error
+			client, err = dialFn(ctx, nodeID)
+			return err
+		})
 		if err != nil {
 			err = errors.Wrapf(err, "failed to dial into node %d (%s)",
 				nodeID, nodeStatuses[nodeID].LivenessStatus)
@@ -1767,3 +1870,40 @@ func (s *statusServer) isSuperUser(ctx context.Context, username string) bool {
 	}
 	return true
 }
+
+type systemInfoOnce struct {
+	once sync.Once
+	info serverpb.SystemInfo
+}
+
+func (si *systemInfoOnce) systemInfo(ctx context.Context) serverpb.SystemInfo {
+	// We only want to attempt to populate the si.info once. If an error occurs
+	// it is logged but the corresponding field in the returned struct is just
+	// left empty as there isn't anything to do with an error from this function.
+	si.once.Do(func() {
+		// Don't use CommandContext because uname ought to not run for too long and
+		// if the status request were canceled, we don't want future requests to
+		// get blank information because we bailed out of this Once.
+		cmd := exec.Command("uname", "-a")
+		var errBuf bytes.Buffer
+		cmd.Stderr = &errBuf
+		output, err := cmd.Output()
+		if err != nil {
+			log.Warningf(ctx, "failed to get system information: %v\nstderr: %v",
+				err, errBuf.String())
+			return
+		}
+		si.info.SystemInfo = string(bytes.TrimSpace(output))
+		cmd = exec.Command("uname", "-r")
+		errBuf.Reset()
+		cmd.Stderr = &errBuf
+		output, err = cmd.Output()
+		if err != nil {
+			log.Warningf(ctx, "failed to get kernel information: %v\nstderr: %v",
+				err, errBuf.String())
+			return
+		}
+		si.info.KernelInfo = string(bytes.TrimSpace(output))
+	})
+	return si.info
+}
