@@ -33,8 +33,12 @@ func getDiskCounters(context.Context) ([]diskStats, error) {
 		output[i] = diskStats{
 			readBytes:      counters.BytesRead,
 			readCount:      counters.NumRead,
+			readTime:       counters.TotalReadTime,
 			writeBytes:     counters.BytesWritten,
 			writeCount:     counters.NumWrite,
+			writeTime:      counters.TotalWriteTime,
+			ioTime:         0, // Not reported by this library.
+			weightedIOTime: 0, // Not reported by this library.
 			iopsInProgress: 0, // Not reported by this library. (#27927)
 		}
 	}
