@@ -18,6 +18,7 @@ import (
 	"bytes"
 	"context"
 	"fmt"
+	"math"
 	"path/filepath"
 	"reflect"
 	"regexp"
@@ -27,9 +28,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/gogo/protobuf/proto"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
@@ -40,7 +38,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/security/securitytest"
 	"github.com/cockroachdb/cockroach/pkg/server/diagnosticspb"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
-	"github.com/cockroachdb/cockroach/pkg/server/status"
+	"github.com/cockroachdb/cockroach/pkg/server/status/statuspb"
+	"github.com/cockroachdb/cockroach/pkg/sql"
 	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
@@ -52,6 +51,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/gogo/protobuf/proto"
+	"github.com/kr/pretty"
+	"github.com/pkg/errors"
 )
 
 func getStatusJSONProto(
@@ -153,6 +155,61 @@ func TestStatusGossipJson(t *testing.T) {
 	}
 }
 
+// TestStatusEngineStatsJson ensures that the output response for the engine
+// stats contains the required fields.
+func TestStatusEngineStatsJson(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	dir, cleanupFn := testutils.TempDir(t)
+	defer cleanupFn()
+
+	s, err := serverutils.StartServerRaw(base.TestServerArgs{
+		StoreSpecs: []base.StoreSpec{{
+			Path: dir,
+		}},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer s.Stopper().Stop(context.TODO())
+
+	var engineStats serverpb.EngineStatsResponse
+	if err := getStatusJSONProto(s, "enginestats/local", &engineStats); err != nil {
+		t.Fatal(err)
+	}
+	if len(engineStats.Stats) != 1 {
+		t.Fatal(errors.Errorf("expected one engine stats, got: %v", engineStats))
+	}
+
+	tickers := engineStats.Stats[0].TickersAndHistograms.Tickers
+	if len(tickers) == 0 {
+		t.Fatal(errors.Errorf("expected non-empty tickers list, got: %v", tickers))
+	}
+	allTickersZero := true
+	for _, ticker := range tickers {
+		if ticker != 0 {
+			allTickersZero = false
+		}
+	}
+	if allTickersZero {
+		t.Fatal(errors.Errorf("expected some tickers nonzero, got: %v", tickers))
+	}
+
+	histograms := engineStats.Stats[0].TickersAndHistograms.Histograms
+	if len(histograms) == 0 {
+		t.Fatal(errors.Errorf("expected non-empty histograms list, got: %v", histograms))
+	}
+	allHistogramsZero := true
+	for _, histogram := range histograms {
+		if histogram.Max == 0 {
+			allHistogramsZero = false
+		}
+	}
+	if allHistogramsZero {
+		t.Fatal(errors.Errorf("expected some histograms nonzero, got: %v", histograms))
+	}
+}
+
 // startServer will start a server with a short scan interval, wait for
 // the scan to complete, and return the server. The caller is
 // responsible for stopping the server.
@@ -408,7 +465,7 @@ func TestNodeStatusResponse(t *testing.T) {
 	// Now fetch each one individually. Loop through the nodeStatuses to use the
 	// ids only.
 	for _, oldNodeStatus := range nodeStatuses {
-		nodeStatus := status.NodeStatus{}
+		nodeStatus := statuspb.NodeStatus{}
 		if err := getStatusJSONProto(s, "nodes/"+oldNodeStatus.Desc.NodeID.String(), &nodeStatus); err != nil {
 			t.Fatal(err)
 		}
@@ -490,6 +547,48 @@ func TestMetricsMetadata(t *testing.T) {
 	}
 }
 
+func TestHotRangesResponse(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	ts := startServer(t)
+	defer ts.Stopper().Stop(context.TODO())
+
+	var hotRangesResp serverpb.HotRangesResponse
+	if err := getStatusJSONProto(ts, "hotranges", &hotRangesResp); err != nil {
+		t.Fatal(err)
+	}
+	if len(hotRangesResp.HotRangesByNodeID) == 0 {
+		t.Fatalf("didn't get hot range responses from any nodes")
+	}
+
+	for nodeID, nodeResp := range hotRangesResp.HotRangesByNodeID {
+		if len(nodeResp.Stores) == 0 {
+			t.Errorf("didn't get any stores in hot range response from n%d: %v",
+				nodeID, nodeResp.ErrorMessage)
+		}
+		for _, storeResp := range nodeResp.Stores {
+			// Only the first store will actually have any ranges on it.
+			if storeResp.StoreID != roachpb.StoreID(1) {
+				continue
+			}
+			lastQPS := math.MaxFloat64
+			if len(storeResp.HotRanges) == 0 {
+				t.Errorf("didn't get any hot ranges in response from n%d,s%d: %v",
+					nodeID, storeResp.StoreID, nodeResp.ErrorMessage)
+			}
+			for _, r := range storeResp.HotRanges {
+				if r.Desc.RangeID == 0 || (len(r.Desc.StartKey) == 0 && len(r.Desc.EndKey) == 0) {
+					t.Errorf("unexpected empty/unpopulated range descriptor: %+v", r.Desc)
+				}
+				if r.QueriesPerSecond > lastQPS {
+					t.Errorf("unexpected increase in qps between ranges; prev=%.2f, current=%.2f, desc=%v",
+						lastQPS, r.QueriesPerSecond, r.Desc)
+				}
+				lastQPS = r.QueriesPerSecond
+			}
+		}
+	}
+}
+
 func TestRangesResponse(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	defer storage.EnableLeaseHistory(100)()
@@ -629,11 +728,12 @@ func TestSpanStatsResponse(t *testing.T) {
 
 func TestSpanStatsGRPCResponse(t *testing.T) {
 	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
 	ts := startServer(t)
-	defer ts.Stopper().Stop(context.TODO())
+	defer ts.Stopper().Stop(ctx)
 
 	rpcStopper := stop.NewStopper()
-	defer rpcStopper.Stop(context.TODO())
+	defer rpcStopper.Stop(ctx)
 	rpcContext := rpc.NewContext(
 		log.AmbientContext{Tracer: ts.ClusterSettings().Tracer}, ts.RPCContext().Config, ts.Clock(),
 		rpcStopper, &ts.ClusterSettings().Version)
@@ -644,13 +744,13 @@ func TestSpanStatsGRPCResponse(t *testing.T) {
 	}
 
 	url := ts.ServingAddr()
-	conn, err := rpcContext.GRPCDial(url).Connect(context.Background())
+	conn, err := rpcContext.GRPCDial(url).Connect(ctx)
 	if err != nil {
 		t.Fatal(err)
 	}
 	client := serverpb.NewStatusClient(conn)
 
-	response, err := client.SpanStats(context.Background(), &request)
+	response, err := client.SpanStats(ctx, &request)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -659,7 +759,7 @@ func TestSpanStatsGRPCResponse(t *testing.T) {
 		t.Fatal(err)
 	}
 	if a, e := int(response.RangeCount), initialRanges; a != e {
-		t.Errorf("expected %d ranges, found %d", e, a)
+		t.Fatalf("expected %d ranges, found %d", e, a)
 	}
 }
 
@@ -818,6 +918,7 @@ func TestRangeResponse(t *testing.T) {
 
 func TestRemoteDebugModeSetting(t *testing.T) {
 	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
 	s, db, _ := serverutils.StartServer(t, base.TestServerArgs{
 		StoreSpecs: []base.StoreSpec{
 			base.DefaultTestStoreSpec,
@@ -832,6 +933,16 @@ func TestRemoteDebugModeSetting(t *testing.T) {
 		t.Fatal(err)
 	}
 
+	// Create a split so that there's some records in the system.rangelog table.
+	// The test needs them.
+	if _, err := db.Exec(
+		`set experimental_force_split_at = true;
+		create table t(x int primary key);
+		alter table t split at values(1);`,
+	); err != nil {
+		t.Fatal(err)
+	}
+
 	// Verify that the remote debugging mode is respected for HTTP requests.
 	// This needs to be wrapped in SucceedsSoon because settings changes have to
 	// propagate through gossip and thus don't always take effect immediately.
@@ -871,7 +982,6 @@ func TestRemoteDebugModeSetting(t *testing.T) {
 		t.Fatal(err)
 	}
 	client := serverpb.NewStatusClient(conn)
-	ctx := context.Background()
 	if _, err := client.Gossip(ctx, &serverpb.GossipRequest{}); err != nil {
 		t.Error(err)
 	}
@@ -894,7 +1004,8 @@ func TestRemoteDebugModeSetting(t *testing.T) {
 		t.Error(err)
 	}
 
-	// Check that keys are properly omitted from the Ranges and RangeLog endpoints.
+	// Check that keys are properly omitted from the Ranges, HotRanges, and
+	// RangeLog endpoints.
 	var rangesResp serverpb.RangesResponse
 	if err := getStatusJSONProto(ts, "ranges/local", &rangesResp); err != nil {
 		t.Fatal(err)
@@ -909,6 +1020,35 @@ func TestRemoteDebugModeSetting(t *testing.T) {
 		}
 	}
 
+	var hotRangesResp serverpb.HotRangesResponse
+	if err := getStatusJSONProto(ts, "hotranges", &hotRangesResp); err != nil {
+		t.Fatal(err)
+	}
+	if len(hotRangesResp.HotRangesByNodeID) == 0 {
+		t.Errorf("didn't get hot range responses from any nodes")
+	}
+	for nodeID, nodeResp := range hotRangesResp.HotRangesByNodeID {
+		if len(nodeResp.Stores) == 0 {
+			t.Errorf("didn't get any stores in hot range response from n%d: %v",
+				nodeID, nodeResp.ErrorMessage)
+		}
+		for _, storeResp := range nodeResp.Stores {
+			// Only the first store will actually have any ranges on it.
+			if storeResp.StoreID != roachpb.StoreID(1) {
+				continue
+			}
+			if len(storeResp.HotRanges) == 0 {
+				t.Errorf("didn't get any hot ranges in response from n%d,s%d: %v",
+					nodeID, storeResp.StoreID, nodeResp.ErrorMessage)
+			}
+			for _, r := range storeResp.HotRanges {
+				if r.Desc.StartKey != nil || r.Desc.EndKey != nil {
+					t.Errorf("unexpected key value found in hot ranges range descriptor: %+v", r.Desc)
+				}
+			}
+		}
+	}
+
 	var rangelogResp serverpb.RangeLogResponse
 	if err := getAdminJSONProto(ts, "rangelog", &rangelogResp); err != nil {
 		t.Fatal(err)
@@ -917,9 +1057,11 @@ func TestRemoteDebugModeSetting(t *testing.T) {
 		t.Errorf("didn't get any Events")
 	}
 	for _, event := range rangelogResp.Events {
-		if event.Event.Info.NewDesc.StartKey != nil || event.Event.Info.NewDesc.EndKey != nil ||
-			event.Event.Info.UpdatedDesc.StartKey != nil || event.Event.Info.UpdatedDesc.EndKey != nil {
-			t.Errorf("unexpected key value found in rangelog event: %+v", event)
+		if event.Event.Info.NewDesc != nil {
+			if event.Event.Info.NewDesc.StartKey != nil || event.Event.Info.NewDesc.EndKey != nil ||
+				event.Event.Info.UpdatedDesc.StartKey != nil || event.Event.Info.UpdatedDesc.EndKey != nil {
+				t.Errorf("unexpected key value found in rangelog event: %+v", event)
+			}
 		}
 		if strings.Contains(event.PrettyInfo.NewDesc, "Min-System") ||
 			strings.Contains(event.PrettyInfo.UpdatedDesc, "Min-System") {
@@ -931,8 +1073,6 @@ func TestRemoteDebugModeSetting(t *testing.T) {
 func TestStatusAPIStatements(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	t.Skip("https://github.com/cockroachdb/cockroach/issues/27272")
-
 	testCluster := serverutils.StartTestCluster(t, 3, base.TestClusterArgs{})
 	defer testCluster.Stopper().Stop(context.Background())
 
@@ -945,7 +1085,7 @@ func TestStatusAPIStatements(t *testing.T) {
 	}{
 		{stmt: `CREATE DATABASE roachblog`},
 		{stmt: `SET database = roachblog`},
-		{stmt: `CREATE TABLE posts (id INT PRIMARY KEY, body TEXT)`},
+		{stmt: `CREATE TABLE posts (id INT8 PRIMARY KEY, body STRING)`},
 		{
 			stmt:          `INSERT INTO posts VALUES (1, 'foo')`,
 			fingerprinted: `INSERT INTO posts VALUES (_, _)`,
@@ -975,6 +1115,16 @@ func TestStatusAPIStatements(t *testing.T) {
 
 	var statementsInResponse []string
 	for _, respStatement := range resp.Statements {
+		if respStatement.Key.KeyData.Failed {
+			// We ignore failed statements here as the INSERT statement can fail and
+			// be automatically retried, confusing the test success check.
+			continue
+		}
+		if strings.HasPrefix(respStatement.Key.KeyData.App, sql.InternalAppNamePrefix) {
+			// We ignore internal queries, these are not relevant for the
+			// validity of this test.
+			continue
+		}
 		statementsInResponse = append(statementsInResponse, respStatement.Key.KeyData.Query)
 	}
 
@@ -982,7 +1132,8 @@ func TestStatusAPIStatements(t *testing.T) {
 	sort.Strings(statementsInResponse)
 
 	if !reflect.DeepEqual(expectedStatements, statementsInResponse) {
-		t.Fatalf("expected queries\n\n%v\n\ngot queries\n\n%v", expectedStatements, statementsInResponse)
+		t.Fatalf("expected queries\n\n%v\n\ngot queries\n\n%v\n%s",
+			expectedStatements, statementsInResponse, pretty.Sprint(resp))
 	}
 }
 
