@@ -20,14 +20,11 @@ import (
 	"net/http"
 	"net/http/cookiejar"
 	"net/url"
-	"runtime"
 	"sync"
 	"time"
 
-	"github.com/cenk/backoff"
-	"github.com/pkg/errors"
-	circuit "github.com/rubyist/circuitbreaker"
-
+	"github.com/cenkalti/backoff"
+	circuit "github.com/cockroachdb/circuitbreaker"
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
@@ -48,20 +45,16 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/ts"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
-	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
-	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
+	"github.com/gogo/protobuf/proto"
+	"github.com/pkg/errors"
 )
 
 const (
 	// TestUser is a fixed user used in unittests.
 	// It has valid embedded client certs.
 	TestUser = "testuser"
-	// initialSplitsTimeout is the amount of time to wait for initial splits to
-	// occur on a freshly started server.
-	// Note: this needs to be fairly high or tests become flaky.
-	initialSplitsTimeout = 10 * time.Second
 )
 
 // makeTestConfig returns a config for testing. It overrides the
@@ -204,11 +197,6 @@ func makeTestConfigFromParams(params base.TestServerArgs) Config {
 		cfg.TestingKnobs.Store = &storage.StoreTestingKnobs{}
 	}
 	cfg.TestingKnobs.Store.(*storage.StoreTestingKnobs).SkipMinSizeCheck = true
-
-	if params.ConnResultsBufferBytes != 0 {
-		cfg.ConnResultsBufferBytes = params.ConnResultsBufferBytes
-	}
-
 	return cfg
 }
 
@@ -322,7 +310,7 @@ func (ts *TestServer) Start(params base.TestServerArgs) error {
 		// Change the replication requirements so we don't get log spam about ranges
 		// not being replicated enough.
 		cfg := config.DefaultZoneConfig()
-		cfg.NumReplicas = 1
+		cfg.NumReplicas = proto.Int32(1)
 		fn := config.TestingSetDefaultZoneConfig(cfg)
 		params.Stopper.AddCloser(stop.CloserFn(fn))
 	}
@@ -349,23 +337,7 @@ func (ts *TestServer) Start(params base.TestServerArgs) error {
 	// Our context must be shared with our server.
 	ts.Cfg = &ts.Server.cfg
 
-	if err := ts.Server.Start(context.Background()); err != nil {
-		return err
-	}
-
-	// If enabled, wait for initial splits to complete before returning control.
-	// If initial splits do not complete, the server is stopped before
-	// returning.
-	if stk, ok := ts.cfg.TestingKnobs.Store.(*storage.StoreTestingKnobs); ok &&
-		stk.DisableSplitQueue {
-		return nil
-	}
-	if err := ts.WaitForInitialSplits(); err != nil {
-		ts.Stop()
-		return err
-	}
-
-	return nil
+	return ts.Server.Start(context.Background())
 }
 
 // ExpectedInitialRangeCount returns the expected number of ranges that should
@@ -376,18 +348,8 @@ func (ts *TestServer) ExpectedInitialRangeCount() (int, error) {
 	return ExpectedInitialRangeCount(ts.DB())
 }
 
-// ExpectedInitialUserRangeCount returns the expected number of ranges that should
-// be on the server after initial (asynchronous) splits have been completed,
-// assuming no additional information is added outside of the normal bootstrap
-// process.
-func (ts *TestServer) ExpectedInitialUserRangeCount() (int, error) {
-	return ExpectedInitialUserRangeCount(ts.DB())
-}
-
 // ExpectedInitialRangeCount returns the expected number of ranges that should
-// be on the server after initial (asynchronous) splits have been completed,
-// assuming no additional information is added outside of the normal bootstrap
-// process.
+// be on the server after bootstrap.
 func ExpectedInitialRangeCount(db *client.DB) (int, error) {
 	descriptorIDs, err := sqlmigrations.ExpectedDescriptorIDs(context.Background(), db)
 	if err != nil {
@@ -406,84 +368,8 @@ func ExpectedInitialRangeCount(db *client.DB) (int, error) {
 	}
 	systemTableSplits := int(maxSystemDescriptorID - keys.MaxSystemConfigDescID)
 
-	// User table splits are analogous to system table splits: they occur at every
-	// possible table boundary between the end of the system ID space
-	// (keys.MaxReservedDescID) and the user table with the maximum ID
-	// (maxUserDescriptorID), even when an ID within the span does not have an
-	// associated descriptor.
-	maxUserDescriptorID := descriptorIDs[len(descriptorIDs)-1]
-	userTableSplits := 0
-	if maxUserDescriptorID >= keys.MaxReservedDescID {
-		userTableSplits = int(maxUserDescriptorID - keys.MaxReservedDescID)
-	}
-
 	// `n` splits create `n+1` ranges.
-	return len(config.StaticSplits()) + systemTableSplits + userTableSplits + 1, nil
-}
-
-// ExpectedInitialUserRangeCount returns the expected number of user ranges that should
-// be on the server after initial (asynchronous) splits have been completed,
-// assuming no additional information is added outside of the normal bootstrap
-// process.
-func ExpectedInitialUserRangeCount(db *client.DB) (int, error) {
-	descriptorIDs, err := sqlmigrations.ExpectedDescriptorIDs(context.Background(), db)
-	if err != nil {
-		return 0, err
-	}
-
-	maxUserDescriptorID := descriptorIDs[len(descriptorIDs)-1]
-	userTableSplits := 0
-	if maxUserDescriptorID >= keys.MaxReservedDescID {
-		userTableSplits = int(maxUserDescriptorID - keys.MaxReservedDescID)
-	}
-	return userTableSplits + 1, nil
-}
-
-// WaitForInitialSplits waits for the server to complete its expected initial
-// splits at startup. If the expected range count is not reached within a
-// configured timeout, an error is returned.
-func (ts *TestServer) WaitForInitialSplits() error {
-	return WaitForInitialSplits(ts.DB())
-}
-
-// WaitForInitialSplits waits for the expected number of initial ranges to be
-// populated in the meta2 table. If the expected range count is not reached
-// within a configured timeout, an error is returned.
-func WaitForInitialSplits(db *client.DB) error {
-	expectedRanges, err := ExpectedInitialRangeCount(db)
-	if err != nil {
-		return err
-	}
-	err = retry.ForDuration(initialSplitsTimeout, func() error {
-		// Scan all keys in the Meta2Prefix; we only need a count.
-		rows, err := db.Scan(context.TODO(), keys.Meta2Prefix, keys.MetaMax, 0)
-		if err != nil {
-			return err
-		}
-		if a, e := len(rows), expectedRanges; a != e {
-			err := errors.Errorf("had %d ranges at startup, expected %d", a, e)
-			log.InfoDepth(context.Background(), 3, err)
-			return err
-		}
-		return nil
-	})
-	if err == nil {
-		return nil
-	}
-
-	// TODO(peter): This is a debugging aid to track down the difficult to
-	// reproduce failures with the initial splits not finishing promptly.
-	for bufSize := 1 << 20; ; bufSize *= 2 {
-		buf := make([]byte, bufSize)
-		length := runtime.Stack(buf, true)
-		// If this wasn't large enough to accommodate the full set of
-		// stack traces, increase by 2 and try again.
-		if length == bufSize {
-			continue
-		}
-		log.Infof(context.TODO(), "%s\n%s", err, buf[:length])
-		return err
-	}
+	return len(config.StaticSplits()) + systemTableSplits + 1, nil
 }
 
 // Stores returns the collection of stores from this TestServer's node.
@@ -544,7 +430,11 @@ func (ts *TestServer) GetAuthenticatedHTTPClient() (http.Client, error) {
 	return httpClient, err
 }
 
-func (ts *TestServer) getAuthenticatedHTTPClientAndCookie() (http.Client, *serverpb.SessionCookie, error) {
+func (ts *TestServer) getAuthenticatedHTTPClientAndCookie() (
+	http.Client,
+	*serverpb.SessionCookie,
+	error,
+) {
 	ts.authClient.once.Do(func() {
 		// Create an authentication session for an arbitrary user. We do not
 		// currently have an authorization mechanism, so a specific user is not
@@ -615,7 +505,9 @@ func (ts *TestServer) MustGetSQLNetworkCounter(name string) int64 {
 	var found bool
 
 	reg := metric.NewRegistry()
-	reg.AddMetricStruct(ts.pgServer.Metrics())
+	for _, m := range ts.pgServer.Metrics() {
+		reg.AddMetricStruct(m)
+	}
 	reg.Each(func(n string, v interface{}) {
 		if name == n {
 			switch t := v.(type) {
@@ -695,6 +587,24 @@ func (ts *TestServer) LookupRange(key roachpb.Key) (roachpb.RangeDescriptor, err
 	return rs[0], nil
 }
 
+// MergeRanges merges the range containing leftKey with the range to its right.
+func (ts *TestServer) MergeRanges(leftKey roachpb.Key) (roachpb.RangeDescriptor, error) {
+
+	ctx := context.Background()
+	mergeReq := roachpb.AdminMergeRequest{
+		RequestHeader: roachpb.RequestHeader{
+			Key: leftKey,
+		},
+	}
+	_, pErr := client.SendWrapped(ctx, ts.DB().NonTransactionalSender(), &mergeReq)
+	if pErr != nil {
+		return roachpb.RangeDescriptor{},
+			errors.Errorf(
+				"%q: merge unexpected error: %s", leftKey, pErr)
+	}
+	return ts.LookupRange(leftKey)
+}
+
 // SplitRange splits the range containing splitKey.
 // The right range created by the split starts at the split key and extends to the
 // original range's end key.
@@ -823,6 +733,20 @@ func (ts *TestServer) ExecutorConfig() interface{} {
 	return *ts.execCfg
 }
 
+// GCSystemLog deletes entries in the given system log table between
+// timestamp and timestampUpperBound if the server is the lease holder
+// for range 1.
+// Leaseholder constraint is present so that only one node in the cluster
+// performs gc.
+// The system log table is expected to have a "timestamp" column.
+// It returns the timestampLowerBound to be used in the next iteration, number
+// of rows affected and error (if any).
+func (ts *TestServer) GCSystemLog(
+	ctx context.Context, table string, timestampLowerBound, timestampUpperBound time.Time,
+) (time.Time, int64, error) {
+	return ts.gcSystemLog(ctx, table, timestampLowerBound, timestampUpperBound)
+}
+
 type testServerFactoryImpl struct{}
 
 // TestServerFactory can be passed to serverutils.InitTestServerFactory
