@@ -17,22 +17,27 @@ package server_test
 import (
 	"context"
 	gosql "database/sql"
+	"fmt"
 	"path/filepath"
 	"strconv"
 	"sync/atomic"
 	"testing"
 
 	"github.com/cockroachdb/cockroach/pkg/base"
+	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/server"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/storage"
+	"github.com/cockroachdb/cockroach/pkg/storage/engine"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/testcluster"
+	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/pkg/errors"
+	"github.com/stretchr/testify/assert"
 )
 
 type testClusterWithHelpers struct {
@@ -61,7 +66,7 @@ func (th *testClusterWithHelpers) getVersionFromSelect(i int) string {
 	if err := protoutil.Unmarshal([]byte(version), &v); err != nil {
 		th.Fatalf("%d: %s", i, err)
 	}
-	return v.MinimumVersion.String()
+	return v.Version.String()
 }
 
 func (th *testClusterWithHelpers) getVersionFromSetting(i int) *cluster.ExposedClusterVersion {
@@ -157,17 +162,14 @@ func TestClusterVersionPersistedOnJoin(t *testing.T) {
 	// new version (and not the old one).
 	versions := [][2]string{{oldVersion.String(), newVersion.String()}, {oldVersion.String(), newVersion.String()}, {oldVersion.String(), newVersion.String()}}
 
-	bootstrapVersion := cluster.ClusterVersion{
-		UseVersion:     newVersion,
-		MinimumVersion: newVersion,
-	}
+	bootstrapVersion := cluster.ClusterVersion{Version: newVersion}
 
 	knobs := base.TestingKnobs{
 		Store: &storage.StoreTestingKnobs{
 			BootstrapVersion: &bootstrapVersion,
 		},
-		Upgrade: &server.UpgradeTestingKnobs{
-			DisableUpgrade: 1,
+		Server: &server.TestingKnobs{
+			DisableAutomaticVersionUpgrade: 1,
 		},
 	}
 
@@ -184,7 +186,7 @@ func TestClusterVersionPersistedOnJoin(t *testing.T) {
 				if err != nil {
 					t.Fatal(err)
 				}
-				if cv.MinimumVersion != newVersion {
+				if cv.Version != newVersion {
 					return errors.Errorf("n%d: expected version %v, got %v", i+1, newVersion, cv)
 				}
 			}
@@ -193,6 +195,174 @@ func TestClusterVersionPersistedOnJoin(t *testing.T) {
 	}
 }
 
+// TestClusterVersionUnreplicatedRaftTruncatedState exercises the
+// VersionUnreplicatedRaftTruncatedState migration in as much detail as possible
+// in a unit test.
+//
+// It starts a four node cluster with a pre-migration version and upgrades into
+// the new version while traffic and scattering are active, verifying that the
+// truncated states are rewritten.
+func TestClusterVersionUnreplicatedRaftTruncatedState(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
+
+	dir, finish := testutils.TempDir(t)
+	defer finish()
+
+	oldVersion := cluster.VersionByKey(cluster.VersionUnreplicatedRaftTruncatedState - 1)
+	oldVersionS := oldVersion.String()
+	newVersionS := cluster.VersionByKey(cluster.VersionUnreplicatedRaftTruncatedState).String()
+
+	// Four node cluster in which all versions support newVersion (i.e. would in
+	// principle upgrade to it) but are bootstrapped at oldVersion.
+	versions := [][2]string{
+		{oldVersionS, newVersionS},
+		{oldVersionS, newVersionS},
+		{oldVersionS, newVersionS},
+		{oldVersionS, newVersionS},
+	}
+
+	bootstrapVersion := cluster.ClusterVersion{Version: oldVersion}
+
+	knobs := base.TestingKnobs{
+		Store: &storage.StoreTestingKnobs{
+			BootstrapVersion: &bootstrapVersion,
+		},
+		Server: &server.TestingKnobs{
+			DisableAutomaticVersionUpgrade: 1,
+		},
+	}
+
+	tc := setupMixedCluster(t, knobs, versions, dir)
+	defer tc.TestCluster.Stopper().Stop(ctx)
+
+	if _, err := tc.ServerConn(0).Exec(`
+CREATE TABLE kv (id INT PRIMARY KEY, v INT);
+ALTER TABLE kv SPLIT AT SELECT i FROM generate_series(1, 9) AS g(i);
+`); err != nil {
+		t.Fatal(err)
+	}
+
+	scatter := func() {
+		t.Helper()
+		if _, err := tc.ServerConn(0).Exec(
+			`ALTER TABLE kv EXPERIMENTAL_RELOCATE SELECT ARRAY[i%$1+1], i FROM generate_series(0, 9) AS g(i)`, len(versions),
+		); err != nil {
+			t.Log(err)
+		}
+	}
+
+	var n int
+	insert := func() {
+		t.Helper()
+		n++
+		// Write only to a subset of our ranges to guarantee log truncations there.
+		_, err := tc.ServerConn(0).Exec(`UPSERT INTO kv VALUES($1, $2)`, n%2, n)
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+
+	for i := 0; i < 500; i++ {
+		insert()
+	}
+	scatter()
+
+	for _, server := range tc.Servers {
+		assert.NoError(t, server.GetStores().(*storage.Stores).VisitStores(func(s *storage.Store) error {
+			s.VisitReplicas(func(r *storage.Replica) bool {
+				key := keys.RaftTruncatedStateKey(r.RangeID)
+				var truncState roachpb.RaftTruncatedState
+				found, err := engine.MVCCGetProto(
+					context.Background(), s.Engine(), key,
+					hlc.Timestamp{}, &truncState, engine.MVCCGetOptions{},
+				)
+				if err != nil {
+					t.Fatal(err)
+				}
+				if found {
+					t.Errorf("unexpectedly found unreplicated TruncatedState at %s", key)
+				}
+				return true // want more
+			})
+			return nil
+		}))
+	}
+
+	if v := tc.getVersionFromSelect(0); v != oldVersionS {
+		t.Fatalf("running %s, wanted %s", v, oldVersionS)
+	}
+
+	assert.NoError(t, tc.setVersion(0, newVersionS))
+	for i := 0; i < 500; i++ {
+		insert()
+	}
+	scatter()
+
+	for _, server := range tc.Servers {
+		testutils.SucceedsSoon(t, func() error {
+			err := server.GetStores().(*storage.Stores).VisitStores(func(s *storage.Store) error {
+				// We scattered and so old copies of replicas may be laying around.
+				// If we're not proactive about removing them, the test gets pretty
+				// slow because those replicas aren't caught up any more.
+				s.MustForceReplicaGCScanAndProcess()
+				var err error
+				s.VisitReplicas(func(r *storage.Replica) bool {
+					snap := s.Engine().NewSnapshot()
+					defer snap.Close()
+
+					keyLegacy := keys.RaftTruncatedStateLegacyKey(r.RangeID)
+					keyUnreplicated := keys.RaftTruncatedStateKey(r.RangeID)
+
+					if found, innerErr := engine.MVCCGetProto(
+						context.Background(), snap, keyLegacy,
+						hlc.Timestamp{}, nil, engine.MVCCGetOptions{},
+					); innerErr != nil {
+						t.Fatal(innerErr)
+					} else if found {
+						if err == nil {
+							err = errors.New("found legacy TruncatedState")
+						}
+						err = errors.Wrap(err, r.String())
+
+						// Force a log truncation to prove that this rectifies
+						// the situation.
+						status := r.RaftStatus()
+						if status != nil {
+							desc := r.Desc()
+							truncate := &roachpb.TruncateLogRequest{}
+							truncate.Key = desc.StartKey.AsRawKey()
+							truncate.RangeID = desc.RangeID
+							truncate.Index = status.HardState.Commit
+							var ba roachpb.BatchRequest
+							ba.RangeID = r.RangeID
+							ba.Add(truncate)
+							if _, err := s.DB().NonTransactionalSender().Send(ctx, ba); err != nil {
+								t.Fatal(err)
+							}
+						}
+						return true // want more
+					}
+
+					if found, err := engine.MVCCGetProto(
+						context.Background(), snap, keyUnreplicated,
+						hlc.Timestamp{}, nil, engine.MVCCGetOptions{},
+					); err != nil {
+						t.Fatal(err)
+					} else if !found {
+						// We can't have neither of the keys present.
+						t.Fatalf("%s: unexpectedly did not find unreplicated TruncatedState at %s", r, keyUnreplicated)
+					}
+
+					return true // want more
+				})
+				return err
+			})
+			return err
+		})
+	}
+}
+
 func TestClusterVersionUpgrade(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	ctx := context.Background()
@@ -208,17 +378,14 @@ func TestClusterVersionUpgrade(t *testing.T) {
 	// version to upgrade automatically from oldVersion to newVersion.
 	versions := [][2]string{{oldVersion.String(), newVersion.String()}, {oldVersion.String(), newVersion.String()}, {oldVersion.String(), newVersion.String()}}
 
-	bootstrapVersion := cluster.ClusterVersion{
-		UseVersion:     oldVersion,
-		MinimumVersion: oldVersion,
-	}
+	bootstrapVersion := cluster.ClusterVersion{Version: oldVersion}
 
 	knobs := base.TestingKnobs{
 		Store: &storage.StoreTestingKnobs{
 			BootstrapVersion: &bootstrapVersion,
 		},
-		Upgrade: &server.UpgradeTestingKnobs{
-			DisableUpgrade: 1,
+		Server: &server.TestingKnobs{
+			DisableAutomaticVersionUpgrade: 1,
 		},
 	}
 	tc := setupMixedCluster(t, knobs, versions, dir)
@@ -228,7 +395,7 @@ func TestClusterVersionUpgrade(t *testing.T) {
 	if err := tc.setDowngrade(0, oldVersion.String()); err != nil {
 		t.Fatalf("error setting CLUSTER SETTING cluster.preserve_downgrade_option: %s", err)
 	}
-	atomic.StoreInt32(&knobs.Upgrade.(*server.UpgradeTestingKnobs).DisableUpgrade, 0)
+	atomic.StoreInt32(&knobs.Server.(*server.TestingKnobs).DisableAutomaticVersionUpgrade, 0)
 
 	// Check the cluster version is still oldVersion.
 	curVersion := tc.getVersionFromSelect(0)
@@ -259,7 +426,7 @@ func TestClusterVersionUpgrade(t *testing.T) {
 				return errors.Errorf("%d: v%s active=%t (wanted %t)", i, newVersion, isActive, wantActive)
 			}
 
-			if tableV, curV := tc.getVersionFromSelect(i), v.Version().MinimumVersion.String(); tableV != curV {
+			if tableV, curV := tc.getVersionFromSelect(i), v.Version().Version.String(); tableV != curV {
 				return errors.Errorf("%d: read v%s from table, v%s from setting", i, tableV, curV)
 			}
 		}
@@ -290,7 +457,7 @@ func TestClusterVersionUpgrade(t *testing.T) {
 	testutils.SucceedsSoon(t, func() error {
 		for i := 0; i < tc.NumServers(); i++ {
 			vers := tc.getVersionFromSetting(i)
-			if v := vers.Version().MinimumVersion.String(); v == curVersion {
+			if v := vers.Version().Version.String(); v == curVersion {
 				if isNoopUpdate {
 					continue
 				}
@@ -309,7 +476,7 @@ func TestClusterVersionUpgrade(t *testing.T) {
 		if err != nil {
 			return err
 		}
-		if act := cv.MinimumVersion.String(); act != exp {
+		if act := cv.Version.String(); act != exp {
 			t.Fatalf("%s: %s persisted, but should be %s", s, act, exp)
 		}
 		return nil
@@ -318,45 +485,38 @@ func TestClusterVersionUpgrade(t *testing.T) {
 	}
 }
 
-func TestClusterVersionBootstrapStrict(t *testing.T) {
+// Test that, after cluster bootstrap, the different ways of getting the cluster
+// version all agree.
+func TestAllVersionsAgree(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	ctx := context.Background()
 
-	// Four nodes that are all strictly version X without accepting anything else.
-	for _, versions := range [][][2]string{
-		{{"1.1", "1.1"}, {"1.1", "1.1"}, {"1.1", "1.1"}, {"1.1", "1.1"}},
-		{{"4.7", "4.7"}, {"4.7", "4.7"}, {"4.7", "4.7"}, {"4.7", "4.7"}},
-	} {
-		func() {
-			bootstrapVersion := cluster.ClusterVersion{
-				UseVersion:     roachpb.MustParseVersion(versions[0][0]),
-				MinimumVersion: roachpb.MustParseVersion(versions[0][0]),
-			}
-
-			knobs := base.TestingKnobs{
-				Store: &storage.StoreTestingKnobs{
-					BootstrapVersion: &bootstrapVersion,
-				},
-			}
-			tc := setupMixedCluster(t, knobs, versions, "")
-			defer tc.Stopper().Stop(ctx)
-
-			exp := versions[0][0]
-
-			for i := 0; i < tc.NumServers(); i++ {
-				if version := tc.getVersionFromSetting(i).Version().MinimumVersion.String(); version != exp {
-					t.Fatalf("%d: incorrect version %s (wanted %s)", i, version, exp)
-				}
-				if version := tc.getVersionFromShow(i); version != exp {
-					t.Fatalf("%d: incorrect version %s (wanted %s)", i, version, exp)
-				}
-
-				if version := tc.getVersionFromSelect(i); version != exp {
-					t.Fatalf("%d: incorrect version %q (wanted %s)", i, version, exp)
-				}
-			}
-		}()
+	tcRaw := testcluster.StartTestCluster(t, 3, base.TestClusterArgs{})
+	defer tcRaw.Stopper().Stop(ctx)
+	tc := testClusterWithHelpers{
+		T:           t,
+		TestCluster: tcRaw,
 	}
+
+	exp := cluster.BinaryServerVersion.String()
+
+	// The node bootstrapping the cluster starts at BinaryServerVersion, the
+	// others start at MinimumSupportedVersion and it takes them a gossip update
+	// to get to BinaryServerVersion. Hence, we loop until that gossip comes.
+	testutils.SucceedsSoon(tc, func() error {
+		for i := 0; i < tc.NumServers(); i++ {
+			if version := tc.getVersionFromSetting(i).Version().Version.String(); version != exp {
+				return fmt.Errorf("%d: incorrect version %s (wanted %s)", i, version, exp)
+			}
+			if version := tc.getVersionFromShow(i); version != exp {
+				return fmt.Errorf("%d: incorrect version %s (wanted %s)", i, version, exp)
+			}
+			if version := tc.getVersionFromSelect(i); version != exp {
+				return fmt.Errorf("%d: incorrect version %q (wanted %s)", i, version, exp)
+			}
+		}
+		return nil
+	})
 }
 
 func TestClusterVersionMixedVersionTooOld(t *testing.T) {
@@ -377,8 +537,7 @@ func TestClusterVersionMixedVersionTooOld(t *testing.T) {
 
 	// Start by running 1.0.
 	bootstrapVersion := cluster.ClusterVersion{
-		UseVersion:     cluster.VersionByKey(cluster.VersionBase),
-		MinimumVersion: cluster.VersionByKey(cluster.VersionBase),
+		Version: cluster.VersionByKey(cluster.VersionBase),
 	}
 
 	knobs := base.TestingKnobs{
@@ -404,7 +563,7 @@ func TestClusterVersionMixedVersionTooOld(t *testing.T) {
 	// Check that we can still talk to the first three nodes.
 	for i := 0; i < tc.NumServers()-1; i++ {
 		testutils.SucceedsSoon(tc, func() error {
-			if version := tc.getVersionFromSetting(i).Version().MinimumVersion.String(); version != exp {
+			if version := tc.getVersionFromSetting(i).Version().Version.String(); version != exp {
 				return errors.Errorf("%d: incorrect version %s (wanted %s)", i, version, exp)
 			}
 			if version := tc.getVersionFromShow(i); version != exp {
@@ -434,8 +593,7 @@ func TestClusterVersionMixedVersionTooNew(t *testing.T) {
 
 	// Try running 1.1.
 	bootstrapVersion := cluster.ClusterVersion{
-		UseVersion:     roachpb.Version{Major: 1, Minor: 1},
-		MinimumVersion: roachpb.Version{Major: 1, Minor: 1},
+		Version: roachpb.Version{Major: 1, Minor: 1},
 	}
 
 	knobs := base.TestingKnobs{
@@ -474,7 +632,7 @@ func TestClusterVersionMixedVersionTooNew(t *testing.T) {
 	// Check that we can still talk to the first three nodes.
 	for i := 0; i < tc.NumServers()-1; i++ {
 		testutils.SucceedsSoon(tc, func() error {
-			if version := tc.getVersionFromSetting(i).Version().MinimumVersion.String(); version != exp {
+			if version := tc.getVersionFromSetting(i).Version().Version.String(); version != exp {
 				return errors.Errorf("%d: incorrect version %s (wanted %s)", i, version, exp)
 			}
 			if version := tc.getVersionFromShow(i); version != exp {
