@@ -20,6 +20,7 @@ import (
 	gojson "encoding/json"
 	"fmt"
 
+	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/privilege"
@@ -36,7 +37,7 @@ import (
 
 type alterTableNode struct {
 	n         *tree.AlterTable
-	tableDesc *sqlbase.TableDescriptor
+	tableDesc *MutableTableDescriptor
 	// statsData is populated with data for "alter table inject statistics"
 	// commands - the JSON stats expressions.
 	// It is parallel with n.Cmds (for the inject stats commands).
@@ -48,12 +49,7 @@ type alterTableNode struct {
 //   notes: postgres requires CREATE on the table.
 //          mysql requires ALTER, CREATE, INSERT on the table.
 func (p *planner) AlterTable(ctx context.Context, n *tree.AlterTable) (planNode, error) {
-	tn, err := n.Table.Normalize()
-	if err != nil {
-		return nil, err
-	}
-
-	tableDesc, err := p.ResolveMutableTableDescriptor(ctx, tn, !n.IfExists, requireTableDesc)
+	tableDesc, err := p.ResolveMutableTableDescriptor(ctx, &n.Table, !n.IfExists, requireTableDesc)
 	if err != nil {
 		return nil, err
 	}
@@ -65,6 +61,8 @@ func (p *planner) AlterTable(ctx context.Context, n *tree.AlterTable) (planNode,
 		return nil, err
 	}
 
+	n.HoistAddColumnConstraints()
+
 	// See if there's any "inject statistics" in the query and type check the
 	// expressions.
 	statsData := make(map[int]tree.TypedExpr)
@@ -95,19 +93,15 @@ func (n *alterTableNode) startExec(params runParams) error {
 	descriptorChanged := false
 	origNumMutations := len(n.tableDesc.Mutations)
 	var droppedViews []string
-	tn := n.n.Table.TableName()
+	tn := &n.n.Table
 
 	for i, cmd := range n.n.Cmds {
 		switch t := cmd.(type) {
 		case *tree.AlterTableAddColumn:
 			d := t.ColumnDef
-			if len(d.CheckExprs) > 0 {
-				return pgerror.UnimplementedWithIssueError(29639,
-					"adding a CHECK constraint via ALTER not supported")
-			}
 			if d.HasFKConstraint() {
-				return pgerror.UnimplementedWithIssueError(8855,
-					"adding a REFERENCES constraint via ALTER not supported")
+				return pgerror.UnimplementedWithIssueError(32917,
+					"adding a REFERENCES constraint while also adding a column via ALTER not supported")
 			}
 
 			newDef, seqDbDesc, seqName, seqOpts, err := params.p.processSerialInColumnDef(params.ctx, d, tn)
@@ -162,7 +156,7 @@ func (n *alterTableNode) startExec(params runParams) error {
 
 			n.tableDesc.AddColumnMutation(*col, sqlbase.DescriptorMutation_ADD)
 			if idx != nil {
-				if err := n.tableDesc.AddIndexMutation(*idx, sqlbase.DescriptorMutation_ADD); err != nil {
+				if err := n.tableDesc.AddIndexMutation(idx, sqlbase.DescriptorMutation_ADD); err != nil {
 					return err
 				}
 			}
@@ -175,6 +169,12 @@ func (n *alterTableNode) startExec(params runParams) error {
 				}
 			}
 
+			if d.IsComputed() {
+				if err := validateComputedColumn(n.tableDesc, d, &params.p.semaCtx); err != nil {
+					return err
+				}
+			}
+
 		case *tree.AlterTableAddConstraint:
 			info, err := n.tableDesc.GetConstraintInfo(params.ctx, nil)
 			if err != nil {
@@ -212,38 +212,35 @@ func (n *alterTableNode) startExec(params runParams) error {
 						return fmt.Errorf("index %q being dropped, try again later", d.Name)
 					}
 				}
-				if err := n.tableDesc.AddIndexMutation(idx, sqlbase.DescriptorMutation_ADD); err != nil {
+				if err := n.tableDesc.AddIndexMutation(&idx, sqlbase.DescriptorMutation_ADD); err != nil {
 					return err
 				}
 
 			case *tree.CheckConstraintTableDef:
-				tableName, err := n.n.Table.Normalize()
-				if err != nil {
-					return err
-				}
 				ck, err := MakeCheckConstraint(params.ctx,
-					*n.tableDesc, d, inuseNames, &params.p.semaCtx, *tableName)
+					n.tableDesc, d, inuseNames, &params.p.semaCtx, n.n.Table)
 				if err != nil {
 					return err
 				}
-				ck.Validity = sqlbase.ConstraintValidity_Unvalidated
-				n.tableDesc.Checks = append(n.tableDesc.Checks, ck)
-				descriptorChanged = true
+				ck.Validity = sqlbase.ConstraintValidity_Validating
+				n.tableDesc.AddCheckValidationMutation(ck)
 
 			case *tree.ForeignKeyConstraintTableDef:
-				if _, err := d.Table.Normalize(); err != nil {
-					return err
-				}
 				for _, colName := range d.FromCols {
-					col, _, err := n.tableDesc.FindColumnByName(colName)
+					col, err := n.tableDesc.FindActiveColumnByName(string(colName))
 					if err != nil {
+						if _, dropped, inactiveErr := n.tableDesc.FindColumnByName(colName); inactiveErr == nil && !dropped {
+							return pgerror.UnimplementedWithIssueError(32917,
+								"adding a REFERENCES constraint while the column is being added not supported")
+						}
 						return err
 					}
+
 					if err := col.CheckCanBeFKRef(); err != nil {
 						return err
 					}
 				}
-				affected := make(map[sqlbase.ID]*sqlbase.TableDescriptor)
+				affected := make(map[sqlbase.ID]*sqlbase.MutableTableDescriptor)
 
 				// If there are any FKs, we will need to update the table descriptor of the
 				// depended-on table (to register this table against its DependedOnBy field).
@@ -252,7 +249,21 @@ func (n *alterTableNode) startExec(params runParams) error {
 				// the global-scope resolveFK().
 				// TODO(vivek): check if the cache can be used.
 				params.p.runWithOptions(resolveFlags{skipCache: true}, func() {
-					err = params.p.resolveFK(params.ctx, n.tableDesc, d, affected, sqlbase.ConstraintValidity_Unvalidated)
+					// Check whether the table is empty, and pass the result to resolveFK(). If
+					// the table is empty, then resolveFK will automatically add the necessary
+					// index for a fk constraint if the index does not exist.
+					kvs, scanErr := params.p.txn.Scan(params.ctx, n.tableDesc.PrimaryIndexSpan().Key, n.tableDesc.PrimaryIndexSpan().EndKey, 1)
+					if scanErr != nil {
+						err = scanErr
+						return
+					}
+					var tableState FKTableState
+					if len(kvs) == 0 {
+						tableState = EmptyTable
+					} else {
+						tableState = NonEmptyTable
+					}
+					err = params.p.resolveFK(params.ctx, n.tableDesc, d, affected, tableState)
 				})
 				if err != nil {
 					return err
@@ -396,10 +407,14 @@ func (n *alterTableNode) startExec(params runParams) error {
 
 			// Drop check constraints which reference the column.
 			validChecks := n.tableDesc.Checks[:0]
-			for _, check := range n.tableDesc.Checks {
-				if used, err := check.UsesColumn(n.tableDesc, col.ID); err != nil {
+			for _, check := range n.tableDesc.AllActiveAndInactiveChecks() {
+				if used, err := check.UsesColumn(n.tableDesc.TableDesc(), col.ID); err != nil {
 					return err
-				} else if !used {
+				} else if used {
+					if check.Validity == sqlbase.ConstraintValidity_Validating {
+						return fmt.Errorf("referencing constraint %q in the middle of being added, try again later", check.Name)
+					}
+				} else {
 					validChecks = append(validChecks, check)
 				}
 			}
@@ -409,6 +424,13 @@ func (n *alterTableNode) startExec(params runParams) error {
 				descriptorChanged = true
 			}
 
+			if err != nil {
+				return err
+			}
+			if err := params.p.removeColumnComment(params.ctx, n.tableDesc.ID, col.ID); err != nil {
+				return err
+			}
+
 			found := false
 			for i := range n.tableDesc.Columns {
 				if n.tableDesc.Columns[i].ID == col.ID {
@@ -435,32 +457,14 @@ func (n *alterTableNode) startExec(params runParams) error {
 				}
 				return fmt.Errorf("constraint %q does not exist", t.Constraint)
 			}
-			switch details.Kind {
-			case sqlbase.ConstraintTypePK:
-				return fmt.Errorf("cannot drop primary key")
-			case sqlbase.ConstraintTypeUnique:
-				return fmt.Errorf("UNIQUE constraint depends on index %q, use DROP INDEX with CASCADE if you really want to drop it", t.Constraint)
-			case sqlbase.ConstraintTypeCheck:
-				for i := range n.tableDesc.Checks {
-					if n.tableDesc.Checks[i].Name == name {
-						n.tableDesc.Checks = append(n.tableDesc.Checks[:i], n.tableDesc.Checks[i+1:]...)
-						descriptorChanged = true
-						break
-					}
-				}
-			case sqlbase.ConstraintTypeFK:
-				idx, err := n.tableDesc.FindIndexByID(details.Index.ID)
-				if err != nil {
-					return err
-				}
-				if err := params.p.removeFKBackReference(params.ctx, n.tableDesc, *idx); err != nil {
-					return err
-				}
-				idx.ForeignKey = sqlbase.ForeignKeyReference{}
-				descriptorChanged = true
-			default:
-				return errors.Errorf("dropping %s constraint %q unsupported", details.Kind, t.Constraint)
+			if err := n.tableDesc.DropConstraint(
+				name, details,
+				func(desc *sqlbase.MutableTableDescriptor, idx *sqlbase.IndexDescriptor) error {
+					return params.p.removeFKBackReference(params.ctx, desc, idx)
+				}); err != nil {
+				return err
 			}
+			descriptorChanged = true
 
 		case *tree.AlterTableValidateConstraint:
 			info, err := n.tableDesc.GetConstraintInfo(params.ctx, nil)
@@ -486,11 +490,12 @@ func (n *alterTableNode) startExec(params runParams) error {
 					}
 				}
 				if !found {
-					panic("constraint returned by GetConstraintInfo not found")
+					return fmt.Errorf("constraint %q in the middle of being added, try again later", t.Constraint)
 				}
+
 				ck := n.tableDesc.Checks[idx]
-				if err := params.p.validateCheckExpr(
-					params.ctx, ck.Expr, &n.n.Table, n.tableDesc,
+				if err := validateCheckExpr(
+					params.ctx, ck.Expr, n.tableDesc.TableDesc(), params.EvalContext().InternalExecutor, params.EvalContext().Txn,
 				); err != nil {
 					return err
 				}
@@ -514,7 +519,7 @@ func (n *alterTableNode) startExec(params runParams) error {
 				if err != nil {
 					panic(err)
 				}
-				if err := params.p.validateForeignKey(params.ctx, n.tableDesc, idx); err != nil {
+				if err := params.p.validateForeignKey(params.ctx, n.tableDesc.TableDesc(), idx); err != nil {
 					return err
 				}
 				idx.ForeignKey.Validity = sqlbase.ConstraintValidity_Validated
@@ -553,7 +558,7 @@ func (n *alterTableNode) startExec(params runParams) error {
 			)
 			err = deleteRemovedPartitionZoneConfigs(
 				params.ctx, params.p.txn,
-				n.tableDesc, &n.tableDesc.PrimaryIndex, &n.tableDesc.PrimaryIndex.Partitioning,
+				n.tableDesc.TableDesc(), &n.tableDesc.PrimaryIndex, &n.tableDesc.PrimaryIndex.Partitioning,
 				&partitioning, params.extendedEvalCtx.ExecCfg,
 			)
 			if err != nil {
@@ -563,7 +568,7 @@ func (n *alterTableNode) startExec(params runParams) error {
 
 		case *tree.AlterTableSetAudit:
 			var err error
-			descriptorChanged, err = params.p.setAuditMode(params.ctx, n.tableDesc, t.Mode)
+			descriptorChanged, err = params.p.setAuditMode(params.ctx, n.tableDesc.TableDesc(), t.Mode)
 			if err != nil {
 				return err
 			}
@@ -573,13 +578,58 @@ func (n *alterTableNode) startExec(params runParams) error {
 			if !ok {
 				return pgerror.NewAssertionErrorf("missing stats data")
 			}
-			if err := injectTableStats(params, n.tableDesc, sd); err != nil {
+			if err := injectTableStats(params, n.tableDesc.TableDesc(), sd); err != nil {
 				return err
 			}
 
+		case *tree.AlterTableRenameColumn:
+			descChanged, err := params.p.renameColumn(params.ctx, n.tableDesc, &t.Column, &t.NewName)
+			if err != nil {
+				return err
+			}
+			descriptorChanged = descChanged
+
+		case *tree.AlterTableRenameConstraint:
+			info, err := n.tableDesc.GetConstraintInfo(params.ctx, nil)
+			if err != nil {
+				return err
+			}
+			details, ok := info[string(t.Constraint)]
+			if !ok {
+				return fmt.Errorf("constraint %q does not exist", tree.ErrString(&t.Constraint))
+			}
+			if t.Constraint == t.NewName {
+				// Nothing to do.
+				break
+			}
+
+			if _, ok := info[string(t.NewName)]; ok {
+				return errors.Errorf("duplicate constraint name: %q", tree.ErrString(&t.NewName))
+			}
+
+			if err := params.p.CheckPrivilege(params.ctx, n.tableDesc, privilege.CREATE); err != nil {
+				return err
+			}
+
+			depViewRenameError := func(objType string, refTableID sqlbase.ID) error {
+				return params.p.dependentViewRenameError(params.ctx,
+					objType, tree.ErrString(&t.NewName), n.tableDesc.ParentID, refTableID)
+			}
+
+			if err := n.tableDesc.RenameConstraint(
+				details, string(t.Constraint), string(t.NewName), depViewRenameError); err != nil {
+				return err
+			}
+			descriptorChanged = true
+
 		default:
 			return pgerror.NewAssertionErrorf("unsupported alter command: %T", cmd)
 		}
+
+		// Allocate IDs now, so new IDs are available to subsequent commands
+		if err := n.tableDesc.AllocateIDs(); err != nil {
+			return err
+		}
 	}
 	// Were some changes made?
 	//
@@ -593,14 +643,10 @@ func (n *alterTableNode) startExec(params runParams) error {
 		return nil
 	}
 
-	if err := n.tableDesc.AllocateIDs(); err != nil {
-		return err
-	}
-
 	mutationID := sqlbase.InvalidMutationID
 	if addedMutations {
 		var err error
-		mutationID, err = params.p.createSchemaChangeJob(params.ctx, n.tableDesc,
+		mutationID, err = params.p.createOrUpdateSchemaChangeJob(params.ctx, n.tableDesc,
 			tree.AsStringWithFlags(n.n, tree.FmtAlwaysQualifyTableNames))
 		if err != nil {
 			return err
@@ -626,7 +672,7 @@ func (n *alterTableNode) startExec(params runParams) error {
 			User                string
 			MutationID          uint32
 			CascadeDroppedViews []string
-		}{n.n.Table.TableName().FQString(), n.n.String(),
+		}{n.n.Table.FQString(), n.n.String(),
 			params.SessionData().User, uint32(mutationID), droppedViews},
 	)
 }
@@ -656,7 +702,7 @@ func (n *alterTableNode) Close(context.Context)        {}
 // columnDescriptor, and saves the containing table descriptor. If the column's
 // dependencies on sequences change, it updates them as well.
 func applyColumnMutation(
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.MutableTableDescriptor,
 	col *sqlbase.ColumnDescriptor,
 	mut tree.ColumnMutationCmd,
 	params runParams,
@@ -753,6 +799,10 @@ func applyColumnMutation(
 		col.Nullable = true
 
 	case *tree.AlterTableDropStored:
+		if !col.IsComputed() {
+			return pgerror.NewErrorf(pgerror.CodeInvalidColumnDefinitionError,
+				"column %q is not a computed column", col.Name)
+		}
 		col.ComputeExpr = nil
 	}
 	return nil
@@ -867,3 +917,18 @@ func injectTableStats(
 
 	return stats.GossipTableStatAdded(params.extendedEvalCtx.ExecCfg.Gossip, desc.ID)
 }
+
+func (p *planner) removeColumnComment(
+	ctx context.Context, tableID sqlbase.ID, columnID sqlbase.ColumnID,
+) error {
+	_, err := p.ExtendedEvalContext().ExecCfg.InternalExecutor.Exec(
+		ctx,
+		"delete-column-comment",
+		p.txn,
+		"DELETE FROM system.comments WHERE type=$1 AND object_id=$2 AND sub_id=$3",
+		keys.ColumnCommentType,
+		tableID,
+		columnID)
+
+	return err
+}
