@@ -26,8 +26,6 @@ import (
 
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
-
 	// We dot-import fsm to use common names such as fsm.True/False. State machine
 	// implementations using that library are weird beasts intimately inter-twined
 	// with that package; therefor this file should stay as small as possible.
@@ -38,12 +36,12 @@ import (
 // Constants for the String() representation of the session states. Shared with
 // the CLI code which needs to recognize them.
 const (
-	NoTxnStr              string = "NoTxn"
-	OpenStateStr                 = "Open"
-	AbortedStateStr              = "Aborted"
-	CommitWaitStateStr           = "CommitWait"
-	RestartWaitStateStr          = "RestartWait"
-	InternalErrorStateStr        = "InternalError"
+	NoTxnStr              = "NoTxn"
+	OpenStateStr          = "Open"
+	AbortedStateStr       = "Aborted"
+	CommitWaitStateStr    = "CommitWait"
+	RestartWaitStateStr   = "RestartWait"
+	InternalErrorStateStr = "InternalError"
 )
 
 /// States.
@@ -126,28 +124,28 @@ type eventTxnStart struct {
 type eventTxnStartPayload struct {
 	tranCtx transitionCtx
 
-	iso enginepb.IsolationType
 	pri roachpb.UserPriority
 	// txnSQLTimestamp is the timestamp that statements executed in the
 	// transaction that is started by this event will report for now(),
 	// current_timestamp(), transaction_timestamp().
-	txnSQLTimestamp time.Time
-	readOnly        tree.ReadWriteMode
+	txnSQLTimestamp     time.Time
+	readOnly            tree.ReadWriteMode
+	historicalTimestamp *hlc.Timestamp
 }
 
 func makeEventTxnStartPayload(
-	iso enginepb.IsolationType,
 	pri roachpb.UserPriority,
 	readOnly tree.ReadWriteMode,
 	txnSQLTimestamp time.Time,
+	historicalTimestamp *hlc.Timestamp,
 	tranCtx transitionCtx,
 ) eventTxnStartPayload {
 	return eventTxnStartPayload{
-		iso:             iso,
-		pri:             pri,
-		readOnly:        readOnly,
-		txnSQLTimestamp: txnSQLTimestamp,
-		tranCtx:         tranCtx,
+		pri:                 pri,
+		readOnly:            readOnly,
+		txnSQLTimestamp:     txnSQLTimestamp,
+		historicalTimestamp: historicalTimestamp,
+		tranCtx:             tranCtx,
 	}
 }
 
@@ -447,7 +445,9 @@ var TxnStateTransitions = Compile(Pattern{
 				ts.resetForNewSQLTxn(
 					ts.connCtx,
 					explicitTxn,
-					payload.txnSQLTimestamp, payload.iso, payload.pri, payload.readOnly,
+					payload.txnSQLTimestamp,
+					payload.historicalTimestamp,
+					payload.pri, payload.readOnly,
 					nil, /* txn */
 					args.Payload.(eventTxnStartPayload).tranCtx,
 				)
@@ -545,7 +545,7 @@ func (ts *txnState) noTxnToOpen(
 		connCtx,
 		txnTyp,
 		payload.txnSQLTimestamp,
-		payload.iso,
+		payload.historicalTimestamp,
 		payload.pri,
 		payload.readOnly,
 		nil, /* txn */
