@@ -21,15 +21,15 @@ import (
 	"sync"
 	"time"
 
-	"github.com/lib/pq/oid"
-	"github.com/pkg/errors"
-
+	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgwirebase"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/lib/pq/oid"
+	"github.com/pkg/errors"
 )
 
 // This file contains utils and interfaces used by a connExecutor to communicate
@@ -97,10 +97,6 @@ type StmtBuf struct {
 		// cond is signaled when new commands are pushed.
 		cond *sync.Cond
 
-		// readerBlocked is set while the reader is blocked on waiting for a command
-		// to be pushed into the buffer.
-		readerBlocked bool
-
 		// data contains the elements of the buffer.
 		data []Command
 
@@ -127,15 +123,10 @@ type Command interface {
 // ExecStmt is the command for running a query sent through the "simple" pgwire
 // protocol.
 type ExecStmt struct {
-	// Stmt can be nil, in which case a "empty query response" message should be
-	// produced.
-	//
-	// TODO(andrei): Many places call Stmt.String() (e.g. ExecStmt.String()),
-	// which seems inneficient. We should find a way to memo-ize this. Another
-	// option is to keep track of the query as we got it from the client, except
-	// that we might have gotten a batch of them at once, in which case only the
-	// parser can do the splitting.
-	Stmt tree.Statement
+	// Information returned from parsing: AST, SQL, NumPlaceholders.
+	// Note that AST can be nil, in which case executing it should produce an
+	// "empty query response" message.
+	parser.Statement
 
 	// TimeReceived is the time at which the exec message was received
 	// from the client. Used to compute the service latency.
@@ -150,7 +141,9 @@ type ExecStmt struct {
 func (ExecStmt) command() {}
 
 func (e ExecStmt) String() string {
-	return fmt.Sprintf("ExecStmt: %s", e.Stmt.String())
+	// We have the original SQL, but we still use String() because it obfuscates
+	// passwords.
+	return fmt.Sprintf("ExecStmt: %s", e.AST.String())
 }
 
 var _ Command = ExecStmt{}
@@ -177,10 +170,14 @@ var _ Command = ExecPortal{}
 
 // PrepareStmt is the command for creating a prepared statement.
 type PrepareStmt struct {
+	// Name of the prepared statement (optional).
 	Name string
-	// Stmt can be nil, in which case executing it should produce an "empty query
-	// response" message.
-	Stmt      tree.Statement
+
+	// Information returned from parsing: AST, SQL, NumPlaceholders.
+	// Note that AST can be nil, in which case executing it should produce an
+	// "empty query response" message.
+	parser.Statement
+
 	TypeHints tree.PlaceholderTypes
 	// RawTypeHints is the representation of type hints exactly as specified by
 	// the client.
@@ -193,7 +190,9 @@ type PrepareStmt struct {
 func (PrepareStmt) command() {}
 
 func (p PrepareStmt) String() string {
-	return fmt.Sprintf("PrepareStmt: %s", p.Stmt.String())
+	// We have the original SQL, but we still use String() because it obfuscates
+	// passwords.
+	return fmt.Sprintf("PrepareStmt: %s", p.AST.String())
 }
 
 var _ Command = PrepareStmt{}
@@ -427,9 +426,7 @@ func (buf *StmtBuf) curCmd() (Command, CmdPos, error) {
 				"can only wait for next command; corrupt cursor: %d", curPos)
 		}
 		// Wait for the next Command to arrive to the buffer.
-		buf.mu.readerBlocked = true
 		buf.mu.cond.Wait()
-		buf.mu.readerBlocked = false
 	}
 }
 
@@ -630,19 +627,14 @@ type CommandResult interface {
 // query execution error.
 type CommandResultErrBase interface {
 	// SetError accumulates an execution error that needs to be reported to the
-	// client. No further calls other than OverwriteError(), Close() and Discard()
-	// are allowed. In particular, CloseWithErr() is not allowed.
+	// client. No further calls other than SetError(), Close()/CloseWithError()
+	// and Discard() are allowed.
+	//
+	// Calling SetError() a second time overwrites the previously set error.
 	SetError(error)
 
 	// Err returns the error previously set with SetError(), if any.
 	Err() error
-
-	// OverwriteError is like SetError(), except it can be called after SetError()
-	// has already been called and it will overwrite the error. Used by high-level
-	// code when it has a strong opinion about what the error that should be
-	// returned to the client is and doesn't much care about whether an error has
-	// already been set on the result.
-	OverwriteError(error)
 }
 
 // ResultBase is the common interface implemented by all the different command
@@ -714,6 +706,12 @@ type RestrictedCommandResult interface {
 	// RowsAffected returns either the number of times AddRow was called, or the
 	// sum of all n passed into IncrementRowsAffected.
 	RowsAffected() int
+
+	// DisableBuffering can be called during execution to ensure that
+	// the results accumulated so far, and all subsequent rows added
+	// to this CommandResult, will be flushed immediately to the client.
+	// This is currently used for sinkless changefeeds.
+	DisableBuffering()
 }
 
 // DescribeResult represents the result of a Describe command (for either
@@ -721,8 +719,8 @@ type RestrictedCommandResult interface {
 type DescribeResult interface {
 	ResultBase
 
-	// SetInTypes tells the client about the inferred placeholder types.
-	SetInTypes([]oid.Oid)
+	// SetInferredTypes tells the client about the inferred placeholder types.
+	SetInferredTypes([]oid.Oid)
 	// SetNoDataDescription is used to tell the client that the prepared statement
 	// or portal produces no rows.
 	SetNoDataRowDescription()
@@ -882,13 +880,12 @@ func (r *bufferedCommandResult) AddRow(ctx context.Context, row tree.Datums) err
 	return nil
 }
 
-// SetError is part of the RestrictedCommandResult interface.
-func (r *bufferedCommandResult) SetError(err error) {
-	r.err = err
+func (r *bufferedCommandResult) DisableBuffering() {
+	panic("cannot disable buffering here")
 }
 
-// OverwriteError is part of the RestrictedCommandResult interface.
-func (r *bufferedCommandResult) OverwriteError(err error) {
+// SetError is part of the RestrictedCommandResult interface.
+func (r *bufferedCommandResult) SetError(err error) {
 	r.err = err
 }
 
@@ -935,3 +932,18 @@ func (r *bufferedCommandResult) Discard() {
 		r.closeCallback(r, discarded, nil /* err */)
 	}
 }
+
+// SetInferredTypes is part of the DescribeResult interface.
+func (r *bufferedCommandResult) SetInferredTypes([]oid.Oid) {}
+
+// SetNoDataRowDescription is part of the DescribeResult interface.
+func (r *bufferedCommandResult) SetNoDataRowDescription() {}
+
+// SetPrepStmtOutput is part of the DescribeResult interface.
+func (r *bufferedCommandResult) SetPrepStmtOutput(context.Context, sqlbase.ResultColumns) {}
+
+// SetPortalOutput is part of the DescribeResult interface.
+func (r *bufferedCommandResult) SetPortalOutput(
+	context.Context, sqlbase.ResultColumns, []pgwirebase.FormatCode,
+) {
+}
