@@ -28,6 +28,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/privilege"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
@@ -48,12 +49,7 @@ type createTableNode struct {
 // Privileges: CREATE on database.
 //   Notes: postgres/mysql require CREATE on database.
 func (p *planner) CreateTable(ctx context.Context, n *tree.CreateTable) (planNode, error) {
-	tn, err := n.Table.Normalize()
-	if err != nil {
-		return nil, err
-	}
-
-	dbDesc, err := p.ResolveUncachedDatabase(ctx, tn)
+	dbDesc, err := p.ResolveUncachedDatabase(ctx, &n.Table)
 	if err != nil {
 		return nil, err
 	}
@@ -63,20 +59,9 @@ func (p *planner) CreateTable(ctx context.Context, n *tree.CreateTable) (planNod
 	}
 
 	n.HoistConstraints()
-	for _, def := range n.Defs {
-		switch t := def.(type) {
-		case *tree.ForeignKeyConstraintTableDef:
-			// Just check the target name has the right structure. We'll do
-			// name resolution later, after the descriptor has been
-			// allocated.
-			_, err := t.Table.Normalize()
-			if err != nil {
-				return nil, err
-			}
-		}
-	}
 
 	var sourcePlan planNode
+	var synthRowID bool
 	if n.As() {
 		// The sourcePlan is needed to determine the set of columns to use
 		// to populate the new table descriptor in Start() below.
@@ -94,9 +79,15 @@ func (p *planner) CreateTable(ctx context.Context, n *tree.CreateTable) (planNod
 				numColNames, util.Pluralize(int64(numColNames)),
 				numColumns, util.Pluralize(int64(numColumns))))
 		}
+
+		// Synthesize an input column that provides the default value for the
+		// hidden rowid column.
+		synthRowID = true
 	}
 
-	return &createTableNode{n: n, dbDesc: dbDesc, sourcePlan: sourcePlan}, nil
+	ct := &createTableNode{n: n, dbDesc: dbDesc, sourcePlan: sourcePlan}
+	ct.run.synthRowID = synthRowID
+	return ct, nil
 }
 
 // createTableRun contains the run-time state of createTableNode
@@ -104,10 +95,16 @@ func (p *planner) CreateTable(ctx context.Context, n *tree.CreateTable) (planNod
 type createTableRun struct {
 	autoCommit   autoCommitOpt
 	rowsAffected int
+
+	// synthRowID indicates whether an input column needs to be synthesized to
+	// provide the default value for the hidden rowid column. The optimizer's
+	// plan already includes this column (so synthRowID is false), whereas the
+	// heuristic planner's plan does not (so synthRowID is true).
+	synthRowID bool
 }
 
 func (n *createTableNode) startExec(params runParams) error {
-	tKey := tableKey{parentID: n.dbDesc.ID, name: n.n.Table.TableName().Table()}
+	tKey := tableKey{parentID: n.dbDesc.ID, name: n.n.Table.Table()}
 	key := tKey.Key()
 	if exists, err := descExists(params.ctx, params.p.txn, key); err == nil && exists {
 		if n.n.IfNotExists {
@@ -130,15 +127,23 @@ func (n *createTableNode) startExec(params runParams) error {
 		privs = sqlbase.NewDefaultPrivilegeDescriptor()
 	}
 
-	var desc sqlbase.TableDescriptor
-	var affected map[sqlbase.ID]*sqlbase.TableDescriptor
+	var asCols sqlbase.ResultColumns
+	var desc sqlbase.MutableTableDescriptor
+	var affected map[sqlbase.ID]*sqlbase.MutableTableDescriptor
 	creationTime := params.p.txn.CommitTimestamp()
 	if n.n.As() {
+		asCols = planColumns(n.sourcePlan)
+		if !n.run.synthRowID {
+			// rowID column is already present in the input as the last column, so
+			// ignore it for the purpose of creating column metadata (because
+			// makeTableDescIfAs does it automatically).
+			asCols = asCols[:len(asCols)-1]
+		}
 		desc, err = makeTableDescIfAs(
-			n.n, n.dbDesc.ID, id, creationTime, planColumns(n.sourcePlan),
+			n.n, n.dbDesc.ID, id, creationTime, asCols,
 			privs, &params.p.semaCtx)
 	} else {
-		affected = make(map[sqlbase.ID]*sqlbase.TableDescriptor)
+		affected = make(map[sqlbase.ID]*sqlbase.MutableTableDescriptor)
 		desc, err = makeTableDesc(params, n.n, n.dbDesc.ID, id, creationTime, privs, affected)
 	}
 	if err != nil {
@@ -154,7 +159,7 @@ func (n *createTableNode) startExec(params runParams) error {
 		}
 		var foundExternalReference bool
 		for id := range refs {
-			if !params.p.Tables().isCreatedTable(id) {
+			if t := params.p.Tables().getUncommittedTableByID(id).MutableTableDescriptor; t == nil || !t.IsNewTable() {
 				foundExternalReference = true
 				break
 			}
@@ -176,8 +181,6 @@ func (n *createTableNode) startExec(params runParams) error {
 		}
 	}
 
-	params.p.Tables().addCreatedTable(id)
-
 	for _, index := range desc.AllNonDropIndexes() {
 		if len(index.Interleave.Ancestors) > 0 {
 			if err := params.p.finalizeInterleave(params.ctx, &desc, index); err != nil {
@@ -202,7 +205,7 @@ func (n *createTableNode) startExec(params runParams) error {
 			TableName string
 			Statement string
 			User      string
-		}{n.n.Table.TableName().FQString(), n.n.String(), params.SessionData().User},
+		}{n.n.Table.FQString(), n.n.String(), params.SessionData().User},
 	); err != nil {
 		return err
 	}
@@ -214,14 +217,22 @@ func (n *createTableNode) startExec(params runParams) error {
 
 		// Instantiate a row inserter and table writer. It has a 1-1
 		// mapping to the definitions in the descriptor.
-		ri, err := sqlbase.MakeRowInserter(
-			params.p.txn, &desc, nil, desc.Columns, sqlbase.SkipFKs, &params.p.alloc)
+		ri, err := row.MakeInserter(
+			params.p.txn,
+			sqlbase.NewImmutableTableDescriptor(*desc.TableDesc()),
+			nil,
+			desc.Columns,
+			row.SkipFKs,
+			&params.p.alloc)
 		if err != nil {
 			return err
 		}
 		ti := tableInserterPool.Get().(*tableInserter)
 		*ti = tableInserter{ri: ri}
 		tw := tableWriter(ti)
+		if n.run.autoCommit == autoCommitEnabled {
+			tw.enableAutoCommit()
+		}
 		defer func() {
 			tw.close(params.ctx)
 			*ti = tableInserter{}
@@ -237,22 +248,28 @@ func (n *createTableNode) startExec(params runParams) error {
 		rowBuffer := make(tree.Datums, len(desc.Columns))
 		pkColIdx := len(desc.Columns) - 1
 
-		// Prepare the rowID expression.
-		defExprSQL := *desc.Columns[pkColIdx].DefaultExpr
-		defExpr, err := parser.ParseExpr(defExprSQL)
-		if err != nil {
-			return err
-		}
-		defTypedExpr, err := params.p.analyzeExpr(
-			params.ctx,
-			defExpr,
-			nil, /*sources*/
-			tree.IndexedVarHelper{},
-			types.Any,
-			false, /*requireType*/
-			"CREATE TABLE AS")
-		if err != nil {
-			return err
+		// The optimizer includes the rowID expression as part of the input
+		// expression. But the heuristic planner does not do this, so construct
+		// a rowID expression to be evaluated separately.
+		var defTypedExpr tree.TypedExpr
+		if n.run.synthRowID {
+			// Prepare the rowID expression.
+			defExprSQL := *desc.Columns[pkColIdx].DefaultExpr
+			defExpr, err := parser.ParseExpr(defExprSQL)
+			if err != nil {
+				return err
+			}
+			defTypedExpr, err = params.p.analyzeExpr(
+				params.ctx,
+				defExpr,
+				nil, /*sources*/
+				tree.IndexedVarHelper{},
+				types.Any,
+				false, /*requireType*/
+				"CREATE TABLE AS")
+			if err != nil {
+				return err
+			}
 		}
 
 		for {
@@ -264,7 +281,7 @@ func (n *createTableNode) startExec(params runParams) error {
 					return err
 				}
 				_, err := tw.finalize(
-					params.ctx, n.run.autoCommit, params.extendedEvalCtx.Tracing.KVTracingEnabled())
+					params.ctx, params.extendedEvalCtx.Tracing.KVTracingEnabled())
 				if err != nil {
 					return err
 				}
@@ -273,17 +290,22 @@ func (n *createTableNode) startExec(params runParams) error {
 
 			// Populate the buffer and generate the PK value.
 			copy(rowBuffer, n.sourcePlan.Values())
-			rowBuffer[pkColIdx], err = defTypedExpr.Eval(params.p.EvalContext())
-			if err != nil {
-				return err
+			if n.run.synthRowID {
+				rowBuffer[pkColIdx], err = defTypedExpr.Eval(params.p.EvalContext())
+				if err != nil {
+					return err
+				}
 			}
 
-			_, err := tw.row(params.ctx, rowBuffer, params.extendedEvalCtx.Tracing.KVTracingEnabled())
+			err := tw.row(params.ctx, rowBuffer, params.extendedEvalCtx.Tracing.KVTracingEnabled())
 			if err != nil {
 				return err
 			}
 			n.run.rowsAffected++
 		}
+
+		// Initiate a run of CREATE STATISTICS.
+		params.ExecCfg().StatsRefresher.NotifyMutation(desc.ID, n.run.rowsAffected)
 	}
 	return nil
 }
@@ -341,12 +363,12 @@ func matchesIndex(
 // descriptors without caching. See the comment on resolveFK().
 func (p *planner) resolveFK(
 	ctx context.Context,
-	tbl *sqlbase.TableDescriptor,
+	tbl *sqlbase.MutableTableDescriptor,
 	d *tree.ForeignKeyConstraintTableDef,
-	backrefs map[sqlbase.ID]*sqlbase.TableDescriptor,
-	mode sqlbase.ConstraintValidity,
+	backrefs map[sqlbase.ID]*sqlbase.MutableTableDescriptor,
+	ts FKTableState,
 ) error {
-	return ResolveFK(ctx, p.txn, p, tbl, d, backrefs, mode)
+	return ResolveFK(ctx, p.txn, p, tbl, d, backrefs, ts)
 }
 
 func qualifyFKColErrorWithDB(
@@ -364,6 +386,19 @@ func qualifyFKColErrorWithDB(
 	return tree.ErrString(tree.NewUnresolvedName(db.Name, tree.PublicSchema, tbl.Name, col))
 }
 
+// FKTableState is the state of the referencing table resolveFK() is called on.
+type FKTableState int
+
+const (
+	// NewTable represents a new table, where the FK constraint is specified in the
+	// CREATE TABLE
+	NewTable FKTableState = iota
+	// EmptyTable represents an existing table that is empty
+	EmptyTable
+	// NonEmptyTable represents an existing non-empty table
+	NonEmptyTable
+)
+
 // ResolveFK looks up the tables and columns mentioned in a `REFERENCES`
 // constraint and adds metadata representing that constraint to the descriptor.
 // It may, in doing so, add to or alter descriptors in the passed in `backrefs`
@@ -389,10 +424,10 @@ func ResolveFK(
 	ctx context.Context,
 	txn *client.Txn,
 	sc SchemaResolver,
-	tbl *sqlbase.TableDescriptor,
+	tbl *sqlbase.MutableTableDescriptor,
 	d *tree.ForeignKeyConstraintTableDef,
-	backrefs map[sqlbase.ID]*sqlbase.TableDescriptor,
-	mode sqlbase.ConstraintValidity,
+	backrefs map[sqlbase.ID]*sqlbase.MutableTableDescriptor,
+	ts FKTableState,
 ) error {
 	for _, col := range d.FromCols {
 		col, _, err := tbl.FindColumnByName(col)
@@ -404,9 +439,7 @@ func ResolveFK(
 		}
 	}
 
-	targetTable := d.Table.TableName()
-
-	target, err := ResolveExistingObject(ctx, sc, targetTable, true /*required*/, requireTableDesc)
+	target, err := ResolveMutableExistingObject(ctx, sc, &d.Table, true /*required*/, requireTableDesc)
 	if err != nil {
 		return err
 	}
@@ -417,8 +450,9 @@ func ResolveFK(
 	} else {
 		// Since this FK is referencing another table, this table must be created in
 		// a non-public "ADD" state and made public only after all leases on the
-		// other table are updated to include the backref.
-		if mode == sqlbase.ConstraintValidity_Validated {
+		// other table are updated to include the backref, if it does not already
+		// exist.
+		if ts == NewTable {
 			tbl.State = sqlbase.TableDescriptor_ADD
 		}
 
@@ -501,7 +535,7 @@ func ResolveFK(
 	if d.Actions.Delete == tree.SetNull || d.Actions.Update == tree.SetNull {
 		for _, sourceColumn := range srcCols {
 			if !sourceColumn.Nullable {
-				col := qualifyFKColErrorWithDB(ctx, txn, tbl, sourceColumn.Name)
+				col := qualifyFKColErrorWithDB(ctx, txn, tbl.TableDesc(), sourceColumn.Name)
 				return pgerror.NewErrorf(pgerror.CodeInvalidForeignKeyError,
 					"cannot add a SET NULL cascading action on column %q which has a NOT NULL constraint", col,
 				)
@@ -514,7 +548,7 @@ func ResolveFK(
 	if d.Actions.Delete == tree.SetDefault || d.Actions.Update == tree.SetDefault {
 		for _, sourceColumn := range srcCols {
 			if sourceColumn.DefaultExpr == nil {
-				col := qualifyFKColErrorWithDB(ctx, txn, tbl, sourceColumn.Name)
+				col := qualifyFKColErrorWithDB(ctx, txn, tbl.TableDesc(), sourceColumn.Name)
 				return pgerror.NewErrorf(pgerror.CodeInvalidForeignKeyError,
 					"cannot add a SET DEFAULT cascading action on column %q which has no DEFAULT expression", col,
 				)
@@ -529,9 +563,10 @@ func ResolveFK(
 		SharedPrefixLen: int32(len(srcCols)),
 		OnDelete:        sqlbase.ForeignKeyReferenceActionValue[d.Actions.Delete],
 		OnUpdate:        sqlbase.ForeignKeyReferenceActionValue[d.Actions.Update],
+		Match:           sqlbase.CompositeKeyMatchMethodValue[d.Match],
 	}
 
-	if mode == sqlbase.ConstraintValidity_Unvalidated {
+	if ts != NewTable {
 		ref.Validity = sqlbase.ConstraintValidity_Unvalidated
 	}
 	backref := sqlbase.ForeignKeyReference{Table: tbl.ID}
@@ -559,11 +594,11 @@ func ResolveFK(
 		}
 		if !found {
 			// Avoid unexpected index builds from ALTER TABLE ADD CONSTRAINT.
-			if mode == sqlbase.ConstraintValidity_Unvalidated {
+			if ts == NonEmptyTable {
 				return pgerror.NewErrorf(pgerror.CodeInvalidForeignKeyError,
 					"foreign key requires an existing index on columns %s", colNames(srcCols))
 			}
-			added, err := addIndexForFK(tbl, srcCols, constraintName, ref)
+			added, err := addIndexForFK(tbl, srcCols, constraintName, ref, ts)
 			if err != nil {
 				return err
 			}
@@ -601,10 +636,11 @@ func ResolveFK(
 // Adds an index to a table descriptor (that is in the process of being created)
 // that will support using `srcCols` as the referencing (src) side of an FK.
 func addIndexForFK(
-	tbl *sqlbase.TableDescriptor,
+	tbl *sqlbase.MutableTableDescriptor,
 	srcCols []sqlbase.ColumnDescriptor,
 	constraintName string,
 	ref sqlbase.ForeignKeyReference,
+	ts FKTableState,
 ) (sqlbase.IndexID, error) {
 	// No existing index for the referencing columns found, so we add one.
 	idx := sqlbase.IndexDescriptor{
@@ -617,23 +653,32 @@ func addIndexForFK(
 		idx.ColumnDirections[i] = sqlbase.IndexDescriptor_ASC
 		idx.ColumnNames[i] = c.Name
 	}
-	if err := tbl.AddIndex(idx, false); err != nil {
+
+	if ts == NewTable {
+		if err := tbl.AddIndex(idx, false); err != nil {
+			return 0, err
+		}
+		if err := tbl.AllocateIDs(); err != nil {
+			return 0, err
+		}
+		added := tbl.Indexes[len(tbl.Indexes)-1]
+
+		// Since we just added the index, we can assume it is the last one rather than
+		// searching all the indexes again. That said, we sanity check that it matches
+		// in case a refactor ever violates that assumption.
+		if !matchesIndex(srcCols, added, matchPrefix) {
+			panic("no matching index and auto-generated index failed to match")
+		}
+		return added.ID, nil
+	}
+
+	if err := tbl.AddIndexMutation(&idx, sqlbase.DescriptorMutation_ADD); err != nil {
 		return 0, err
 	}
 	if err := tbl.AllocateIDs(); err != nil {
 		return 0, err
 	}
-
-	added := tbl.Indexes[len(tbl.Indexes)-1]
-
-	// Since we just added the index, we can assume it is the last one rather than
-	// searching all the indexes again. That said, we sanity check that it matches
-	// in case a refactor ever violates that assumption.
-	if !matchesIndex(srcCols, added, matchPrefix) {
-		panic("no matching index and auto-generated index failed to match")
-	}
-
-	return added.ID, nil
+	return tbl.Mutations[len(tbl.Mutations)-1].GetIndex().ID, nil
 }
 
 // colNames converts a []colDesc to a human-readable string for use in error messages.
@@ -652,7 +697,7 @@ func colNames(cols []sqlbase.ColumnDescriptor) string {
 
 func (p *planner) addInterleave(
 	ctx context.Context,
-	desc *sqlbase.TableDescriptor,
+	desc *sqlbase.MutableTableDescriptor,
 	index *sqlbase.IndexDescriptor,
 	interleave *tree.InterleaveDef,
 ) error {
@@ -665,7 +710,7 @@ func addInterleave(
 	ctx context.Context,
 	txn *client.Txn,
 	vt SchemaResolver,
-	desc *sqlbase.TableDescriptor,
+	desc *sqlbase.MutableTableDescriptor,
 	index *sqlbase.IndexDescriptor,
 	interleave *tree.InterleaveDef,
 ) error {
@@ -674,12 +719,9 @@ func addInterleave(
 			7854, "unsupported shorthand %s", interleave.DropBehavior)
 	}
 
-	tn, err := interleave.Parent.Normalize()
-	if err != nil {
-		return err
-	}
-
-	parentTable, err := ResolveExistingObject(ctx, vt, tn, true /*required*/, requireTableDesc)
+	parentTable, err := ResolveExistingObject(
+		ctx, vt, &interleave.Parent, true /*required*/, requireTableDesc,
+	)
 	if err != nil {
 		return err
 	}
@@ -758,7 +800,7 @@ func addInterleave(
 // finalizeInterleave creates backreferences from an interleaving parent to the
 // child data being interleaved.
 func (p *planner) finalizeInterleave(
-	ctx context.Context, desc *sqlbase.TableDescriptor, index sqlbase.IndexDescriptor,
+	ctx context.Context, desc *sqlbase.MutableTableDescriptor, index *sqlbase.IndexDescriptor,
 ) error {
 	// TODO(dan): This is similar to finalizeFKs. Consolidate them
 	if len(index.Interleave.Ancestors) == 0 {
@@ -766,12 +808,12 @@ func (p *planner) finalizeInterleave(
 	}
 	// Only the last ancestor needs the backreference.
 	ancestor := index.Interleave.Ancestors[len(index.Interleave.Ancestors)-1]
-	var ancestorTable *sqlbase.TableDescriptor
+	var ancestorTable *sqlbase.MutableTableDescriptor
 	if ancestor.TableID == desc.ID {
 		ancestorTable = desc
 	} else {
 		var err error
-		ancestorTable, err = sqlbase.GetTableDescFromID(ctx, p.txn, ancestor.TableID)
+		ancestorTable, err = p.Tables().getMutableTableVersionByID(ctx, ancestor.TableID, p.txn)
 		if err != nil {
 			return err
 		}
@@ -804,7 +846,7 @@ func CreatePartitioning(
 	ctx context.Context,
 	st *cluster.Settings,
 	evalCtx *tree.EvalContext,
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.MutableTableDescriptor,
 	indexDesc *sqlbase.IndexDescriptor,
 	partBy *tree.PartitionBy,
 ) (sqlbase.PartitioningDescriptor, error) {
@@ -821,7 +863,7 @@ var CreatePartitioningCCL = func(
 	ctx context.Context,
 	st *cluster.Settings,
 	evalCtx *tree.EvalContext,
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.MutableTableDescriptor,
 	indexDesc *sqlbase.IndexDescriptor,
 	partBy *tree.PartitionBy,
 ) (sqlbase.PartitioningDescriptor, error) {
@@ -835,8 +877,8 @@ func InitTableDescriptor(
 	name string,
 	creationTime hlc.Timestamp,
 	privileges *sqlbase.PrivilegeDescriptor,
-) sqlbase.TableDescriptor {
-	return sqlbase.TableDescriptor{
+) sqlbase.MutableTableDescriptor {
+	return *sqlbase.NewMutableCreatedTableDescriptor(sqlbase.TableDescriptor{
 		ID:               id,
 		Name:             name,
 		ParentID:         parentID,
@@ -844,7 +886,7 @@ func InitTableDescriptor(
 		Version:          1,
 		ModificationTime: creationTime,
 		Privileges:       privileges,
-	}
+	})
 }
 
 // makeTableDescIfAs is the MakeTableDesc method for when we have a table
@@ -856,12 +898,8 @@ func makeTableDescIfAs(
 	resultColumns []sqlbase.ResultColumn,
 	privileges *sqlbase.PrivilegeDescriptor,
 	semaCtx *tree.SemaContext,
-) (desc sqlbase.TableDescriptor, err error) {
-	tableName, err := p.Table.Normalize()
-	if err != nil {
-		return desc, err
-	}
-	desc = InitTableDescriptor(id, parentID, tableName.Table(), creationTime, privileges)
+) (desc sqlbase.MutableTableDescriptor, err error) {
+	desc = InitTableDescriptor(id, parentID, p.Table.Table(), creationTime, privileges)
 	for i, colRes := range resultColumns {
 		colType, err := coltypes.DatumTypeToColumnType(colRes.Typ)
 		if err != nil {
@@ -948,15 +986,11 @@ func MakeTableDesc(
 	parentID, id sqlbase.ID,
 	creationTime hlc.Timestamp,
 	privileges *sqlbase.PrivilegeDescriptor,
-	affected map[sqlbase.ID]*sqlbase.TableDescriptor,
+	affected map[sqlbase.ID]*sqlbase.MutableTableDescriptor,
 	semaCtx *tree.SemaContext,
 	evalCtx *tree.EvalContext,
-) (sqlbase.TableDescriptor, error) {
-	tableName, err := n.Table.Normalize()
-	if err != nil {
-		return sqlbase.TableDescriptor{}, err
-	}
-	desc := InitTableDescriptor(id, parentID, tableName.Table(), creationTime, privileges)
+) (sqlbase.MutableTableDescriptor, error) {
+	desc := InitTableDescriptor(id, parentID, n.Table.Table(), creationTime, privileges)
 
 	for _, def := range n.Defs {
 		if d, ok := def.(*tree.ColumnTableDef); ok {
@@ -1004,7 +1038,7 @@ func MakeTableDesc(
 	// Now that we've constructed our columns, we pop into any of our computed
 	// columns so that we can dequalify any column references.
 	sourceInfo := sqlbase.NewSourceInfoForSingleTable(
-		*tableName, sqlbase.ResultColumnsFromColDescs(desc.Columns),
+		n.Table, sqlbase.ResultColumnsFromColDescs(desc.Columns),
 	)
 	sources := sqlbase.MultiSourceInfo{sourceInfo}
 
@@ -1143,46 +1177,53 @@ func MakeTableDesc(
 	// table.
 	fkResolver := &fkSelfResolver{
 		SchemaResolver: vt,
-		newTableDesc:   &desc,
-		newTableName:   n.Table.TableName(),
+		newTableDesc:   desc.TableDesc(),
+		newTableName:   &n.Table,
 	}
 
 	generatedNames := map[string]struct{}{}
 	for _, def := range n.Defs {
 		switch d := def.(type) {
 		case *tree.ColumnTableDef:
-			// Now that we have all the other columns set up, we can validate any
-			// computed columns.
-			if d.IsComputed() {
-				if err := validateComputedColumn(desc, n, d, semaCtx); err != nil {
-					return desc, err
-				}
-			}
+			// Check after all ResolveFK calls.
 
 		case *tree.IndexTableDef, *tree.UniqueConstraintTableDef, *tree.FamilyTableDef:
 			// Pass, handled above.
 
 		case *tree.CheckConstraintTableDef:
-			ck, err := MakeCheckConstraint(ctx, desc, d, generatedNames, semaCtx, *tableName)
+			ck, err := MakeCheckConstraint(ctx, &desc, d, generatedNames, semaCtx, n.Table)
 			if err != nil {
 				return desc, err
 			}
 			desc.Checks = append(desc.Checks, ck)
 
 		case *tree.ForeignKeyConstraintTableDef:
-			if err := ResolveFK(ctx, txn, fkResolver, &desc, d, affected, sqlbase.ConstraintValidity_Validated); err != nil {
+			if err := ResolveFK(ctx, txn, fkResolver, &desc, d, affected, NewTable); err != nil {
 				return desc, err
 			}
+
 		default:
 			return desc, errors.Errorf("unsupported table def: %T", def)
 		}
 	}
+	// Now that we have all the other columns set up, we can validate
+	// any computed columns.
+	for _, def := range n.Defs {
+		switch d := def.(type) {
+		case *tree.ColumnTableDef:
+			if d.IsComputed() {
+				if err := validateComputedColumn(&desc, d, semaCtx); err != nil {
+					return desc, err
+				}
+			}
+		}
+	}
 
 	// AllocateIDs mutates its receiver. `return desc, desc.AllocateIDs()`
 	// happens to work in gc, but does not work in gccgo.
 	//
 	// See https://github.com/golang/go/issues/23188.
-	err = desc.AllocateIDs()
+	err := desc.AllocateIDs()
 	return desc, err
 }
 
@@ -1193,12 +1234,8 @@ func makeTableDesc(
 	parentID, id sqlbase.ID,
 	creationTime hlc.Timestamp,
 	privileges *sqlbase.PrivilegeDescriptor,
-	affected map[sqlbase.ID]*sqlbase.TableDescriptor,
-) (ret sqlbase.TableDescriptor, err error) {
-	tableName, err := n.Table.Normalize()
-	if err != nil {
-		return ret, err
-	}
+	affected map[sqlbase.ID]*sqlbase.MutableTableDescriptor,
+) (ret sqlbase.MutableTableDescriptor, err error) {
 	// Process any SERIAL columns to remove the SERIAL type,
 	// as required by MakeTableDesc.
 	createStmt := n
@@ -1214,7 +1251,7 @@ func makeTableDesc(
 		if !ok {
 			continue
 		}
-		newDef, seqDbDesc, seqName, seqOpts, err := params.p.processSerialInColumnDef(params.ctx, d, tableName)
+		newDef, seqDbDesc, seqName, seqOpts, err := params.p.processSerialInColumnDef(params.ctx, d, &n.Table)
 		if err != nil {
 			return ret, err
 		}
@@ -1291,7 +1328,7 @@ func (d *dummyColumnItem) ResolvedType() types.T {
 }
 
 func generateNameForCheckConstraint(
-	desc sqlbase.TableDescriptor, expr tree.Expr, inuseNames map[string]struct{},
+	desc *sqlbase.MutableTableDescriptor, expr tree.Expr, inuseNames map[string]struct{},
 ) (string, error) {
 	var nameBuf bytes.Buffer
 	nameBuf.WriteString("check")
@@ -1326,7 +1363,7 @@ func generateNameForCheckConstraint(
 }
 
 func iterColDescriptorsInExpr(
-	desc sqlbase.TableDescriptor, rootExpr tree.Expr, f func(sqlbase.ColumnDescriptor) error,
+	desc *sqlbase.MutableTableDescriptor, rootExpr tree.Expr, f func(sqlbase.ColumnDescriptor) error,
 ) error {
 	_, err := tree.SimpleVisit(rootExpr, func(expr tree.Expr) (err error, recurse bool, newExpr tree.Expr) {
 		vBase, ok := expr.(tree.VarName)
@@ -1345,8 +1382,8 @@ func iterColDescriptorsInExpr(
 			return nil, true, expr
 		}
 
-		col, err := desc.FindActiveColumnByName(string(c.ColumnName))
-		if err != nil {
+		col, dropped, err := desc.FindColumnByName(c.ColumnName)
+		if err != nil || dropped {
 			return pgerror.NewErrorf(pgerror.CodeInvalidTableDefinitionError,
 				"column %q not found, referenced in %q",
 				c.ColumnName, rootExpr), false, nil
@@ -1364,10 +1401,7 @@ func iterColDescriptorsInExpr(
 // validateComputedColumn checks that a computed column satisfies a number of
 // validity constraints, for instance, that it typechecks.
 func validateComputedColumn(
-	desc sqlbase.TableDescriptor,
-	t *tree.CreateTable,
-	d *tree.ColumnTableDef,
-	semaCtx *tree.SemaContext,
+	desc *sqlbase.MutableTableDescriptor, d *tree.ColumnTableDef, semaCtx *tree.SemaContext,
 ) error {
 	if d.HasDefaultExpr() {
 		return pgerror.NewError(
@@ -1393,24 +1427,28 @@ func validateComputedColumn(
 	// TODO(justin,bram): allow depending on columns like this. We disallow it
 	// for now because cascading changes must hook into the computed column
 	// update path.
-	for _, def := range t.Defs {
-		switch c := def.(type) {
-		case *tree.ColumnTableDef:
-			if _, ok := dependencies[string(c.Name)]; !ok {
+	if err := desc.ForeachNonDropIndex(func(idx *sqlbase.IndexDescriptor) error {
+		for _, name := range idx.ColumnNames {
+			if _, ok := dependencies[name]; !ok {
 				// We don't depend on this column.
 				continue
 			}
-			for _, action := range []tree.ReferenceAction{
-				c.References.Actions.Update,
-				c.References.Actions.Delete,
+			for _, action := range []sqlbase.ForeignKeyReference_Action{
+				idx.ForeignKey.OnDelete,
+				idx.ForeignKey.OnUpdate,
 			} {
 				switch action {
-				case tree.Cascade, tree.SetNull, tree.SetDefault:
+				case sqlbase.ForeignKeyReference_CASCADE,
+					sqlbase.ForeignKeyReference_SET_NULL,
+					sqlbase.ForeignKeyReference_SET_DEFAULT:
 					return pgerror.NewError(pgerror.CodeInvalidTableDefinitionError,
 						"computed columns cannot reference non-restricted FK columns")
 				}
 			}
 		}
+		return nil
+	}); err != nil {
+		return err
 	}
 
 	// Replace column references with typed dummies to allow typechecking.
@@ -1433,7 +1471,7 @@ func validateComputedColumn(
 // this new expression tree alongside a set containing the ColumnID of each
 // column seen in the expression.
 func replaceVars(
-	desc sqlbase.TableDescriptor, expr tree.Expr,
+	desc *sqlbase.MutableTableDescriptor, expr tree.Expr,
 ) (tree.Expr, map[sqlbase.ColumnID]struct{}, error) {
 	colIDs := make(map[sqlbase.ColumnID]struct{})
 	newExpr, err := tree.SimpleVisit(expr, func(expr tree.Expr) (err error, recurse bool, newExpr tree.Expr) {
@@ -1453,8 +1491,8 @@ func replaceVars(
 			return nil, true, expr
 		}
 
-		col, err := desc.FindActiveColumnByName(string(c.ColumnName))
-		if err != nil {
+		col, dropped, err := desc.FindColumnByName(c.ColumnName)
+		if err != nil || dropped {
 			return fmt.Errorf("column %q not found for constraint %q",
 				c.ColumnName, expr.String()), false, nil
 		}
@@ -1468,7 +1506,7 @@ func replaceVars(
 // MakeCheckConstraint makes a descriptor representation of a check from a def.
 func MakeCheckConstraint(
 	ctx context.Context,
-	desc sqlbase.TableDescriptor,
+	desc *sqlbase.MutableTableDescriptor,
 	d *tree.CheckConstraintTableDef,
 	inuseNames map[string]struct{},
 	semaCtx *tree.SemaContext,
@@ -1502,7 +1540,7 @@ func MakeCheckConstraint(
 	sort.Sort(sqlbase.ColumnIDs(colIDs))
 
 	sourceInfo := sqlbase.NewSourceInfoForSingleTable(
-		tableName, sqlbase.ResultColumnsFromColDescs(desc.Columns),
+		tableName, sqlbase.ResultColumnsFromColDescs(desc.TableDesc().AllNonDropColumns()),
 	)
 	sources := sqlbase.MultiSourceInfo{sourceInfo}
 
