@@ -17,11 +17,10 @@ package sql_test
 import (
 	"context"
 	gosql "database/sql"
+	"fmt"
 	"math/rand"
 	"testing"
 
-	"fmt"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/keys"
@@ -87,6 +86,7 @@ func (mt mutationTest) makeMutationsActive() {
 		}
 	}
 	mt.tableDesc.Mutations = nil
+	mt.tableDesc.Version++
 	if err := mt.tableDesc.ValidateTable(cluster.MakeTestingClusterSettings()); err != nil {
 		mt.Fatal(err)
 	}
@@ -142,6 +142,7 @@ func (mt mutationTest) writeMutation(m sqlbase.DescriptorMutation) {
 		}
 	}
 	mt.tableDesc.Mutations = append(mt.tableDesc.Mutations, m)
+	mt.tableDesc.Version++
 	if err := mt.tableDesc.ValidateTable(cluster.MakeTestingClusterSettings()); err != nil {
 		mt.Fatal(err)
 	}
@@ -286,7 +287,8 @@ CREATE INDEX allidx ON t.test (k, v);
 					} else {
 						_, err = sqlDB.Exec(`INSERT INTO t.test (k, v, i) VALUES ('b', 'y', 'i')`)
 					}
-					if !testutils.IsError(err, `column "i" does not exist`) {
+					if !testutils.IsError(err, `column "i" does not exist`) &&
+						!testutils.IsError(err, `column "i" is being backfilled`) {
 						t.Fatal(err)
 					}
 					if useUpsert {
@@ -304,7 +306,8 @@ CREATE INDEX allidx ON t.test (k, v);
 					} else {
 						_, err = sqlDB.Exec(`INSERT INTO t.test VALUES ('b', 'y', 'i')`)
 					}
-					if !testutils.IsError(err, "(IN|UP)SERT has more expressions than target columns, 3 expressions for 2 targets") {
+					if !testutils.IsError(err, "(IN|UP)SERT has more expressions than target columns, 3 expressions for 2 targets") &&
+						!testutils.IsError(err, `column "i" is being backfilled`) {
 						t.Fatal(err)
 					}
 
@@ -333,10 +336,9 @@ CREATE INDEX allidx ON t.test (k, v);
 						// The default value of "i" for column "i" is written.
 						afterInsert = [][]string{{"a", "z", "q"}, {"c", "x", "i"}}
 						if useUpsert {
-							// Update is not a noop for column "i". Column "i" gets updated
-							// with its default value (#9474).
-							afterUpdate = [][]string{{"a", "u", "i"}, {"c", "x", "i"}}
-							afterPKUpdate = [][]string{{"a", "u", "i"}, {"d", "x", "i"}}
+							// Update is not a noop for column "i".
+							afterUpdate = [][]string{{"a", "u", "q"}, {"c", "x", "i"}}
+							afterPKUpdate = [][]string{{"a", "u", "q"}, {"d", "x", "i"}}
 						} else {
 							// Update is a noop for column "i".
 							afterUpdate = [][]string{{"a", "u", "q"}, {"c", "x", "i"}}
@@ -390,7 +392,8 @@ CREATE INDEX allidx ON t.test (k, v);
 						}
 					} else {
 						_, err := sqlDB.Exec(`UPDATE t.test SET (v, i) = ('u', 'u') WHERE k = 'a'`)
-						if !testutils.IsError(err, `column "i" does not exist`) {
+						if !testutils.IsError(err, `column "i" does not exist`) &&
+							!testutils.IsError(err, `column "i" is being backfilled`) {
 							t.Fatal(err)
 						}
 					}
@@ -465,14 +468,17 @@ func (mt mutationTest) writeIndexMutation(index string, m sqlbase.DescriptorMuta
 	if err != nil {
 		mt.Fatal(err)
 	}
+	// The rewrite below potentially invalidates the original object with an overwrite.
+	// Clarify what's going on.
+	idxCopy := *idx
 	for i := range tableDesc.Indexes {
-		if idx.ID == tableDesc.Indexes[i].ID {
+		if idxCopy.ID == tableDesc.Indexes[i].ID {
 			tableDesc.Indexes = append(tableDesc.Indexes[:i], tableDesc.Indexes[i+1:]...)
 			break
 		}
 	}
 
-	m.Descriptor_ = &sqlbase.DescriptorMutation_Index{Index: &idx}
+	m.Descriptor_ = &sqlbase.DescriptorMutation_Index{Index: &idxCopy}
 	mt.writeMutation(m)
 }
 
@@ -726,7 +732,8 @@ CREATE INDEX allidx ON t.test (k, v);
 					}
 				} else {
 					_, err := sqlDB.Exec(`UPDATE t.test SET (v, i) = ('u', 'u') WHERE k = 'a'`)
-					if !testutils.IsError(err, `column "i" does not exist`) {
+					if !testutils.IsError(err, `column "i" does not exist`) &&
+						!testutils.IsError(err, `column "i" is being backfilled`) {
 						t.Error(err)
 					}
 				}
@@ -739,7 +746,8 @@ CREATE INDEX allidx ON t.test (k, v);
 				// TODO(vivek): Fix this error to return the same is being
 				// backfilled error.
 				_, err = sqlDB.Exec(`UPDATE t.test SET i = 'u' WHERE k = 'a'`)
-				if !testutils.IsError(err, `column "i" does not exist`) {
+				if !testutils.IsError(err, `column "i" does not exist`) &&
+					!testutils.IsError(err, `column "i" is being backfilled`) {
 					t.Error(err)
 				}
 				_, err = sqlDB.Exec(`DELETE FROM t.test WHERE i < 'a'`)
@@ -987,8 +995,8 @@ CREATE TABLE t.test (a STRING PRIMARY KEY, b STRING, c STRING, INDEX foo (c));
 	mt.CheckQueryResults(t,
 		"SHOW COLUMNS FROM t.test",
 		[][]string{
-			{"a", "STRING", "false", "NULL", "", "{\"primary\",\"ufo\"}", "false"},
-			{"d", "STRING", "true", "NULL", "", "{\"ufo\"}", "false"},
+			{"a", "STRING", "false", "NULL", "", "{primary,ufo}", "false"},
+			{"d", "STRING", "true", "NULL", "", "{ufo}", "false"},
 			{"e", "STRING", "true", "NULL", "", "{}", "false"},
 		},
 	)
@@ -1079,9 +1087,9 @@ CREATE TABLE t.test (k CHAR PRIMARY KEY, v CHAR UNIQUE);
 		// Third.
 		{"idx_g", 3, sqlbase.DescriptorMutation_DELETE_ONLY},
 		// Drop mutations start off in the DELETE_AND_WRITE_ONLY state.
-		// UNIQUE column deletion gets split into two mutation ids.
+		// UNIQUE column deletion gets split into two mutations with the same ID.
 		{"test_v_key", 4, sqlbase.DescriptorMutation_DELETE_AND_WRITE_ONLY},
-		{"v", 5, sqlbase.DescriptorMutation_DELETE_AND_WRITE_ONLY},
+		{"v", 4, sqlbase.DescriptorMutation_DELETE_AND_WRITE_ONLY},
 	}
 
 	if len(tableDesc.Mutations) != len(expected) {
