@@ -28,14 +28,16 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/distsqlplan"
-	"github.com/cockroachdb/cockroach/pkg/sql/distsqlrun"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/log/logtags"
+	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/pkg/errors"
 )
 
@@ -56,7 +58,7 @@ func PlanAndRunExport(
 	txn *client.Txn,
 	evalCtx *extendedEvalContext,
 	in planNode,
-	out distsqlrun.ProcessorCoreUnion,
+	out distsqlpb.ProcessorCoreUnion,
 	resultRows *RowResultWriter,
 ) error {
 	planCtx := dsp.NewPlanningCtx(ctx, evalCtx, txn)
@@ -64,13 +66,13 @@ func PlanAndRunExport(
 	rec, err := dsp.checkSupportForNode(in)
 	planCtx.isLocal = err != nil || rec == cannotDistribute
 
-	p, err := dsp.createPlanForNode(&planCtx, in)
+	p, err := dsp.createPlanForNode(planCtx, in)
 	if err != nil {
 		return errors.Wrap(err, "constructing distSQL plan")
 	}
 
 	p.AddNoGroupingStage(
-		out, distsqlrun.PostProcessSpec{}, ExportPlanResultTypes, distsqlrun.Ordering{},
+		out, distsqlpb.PostProcessSpec{}, ExportPlanResultTypes, distsqlpb.Ordering{},
 	)
 
 	// Overwrite PlanToStreamColMap (used by recv below) to reflect the output of
@@ -78,7 +80,7 @@ func PlanAndRunExport(
 	// columns filename/rows/bytes.
 	p.PlanToStreamColMap = identityMap(p.PlanToStreamColMap, len(ExportPlanResultTypes))
 
-	dsp.FinalizePlan(&planCtx, &p)
+	dsp.FinalizePlan(planCtx, &p)
 
 	recv := MakeDistSQLReceiver(
 		ctx, resultRows, tree.Rows,
@@ -86,13 +88,15 @@ func PlanAndRunExport(
 		evalCtx.Tracing,
 	)
 
-	dsp.Run(&planCtx, txn, &p, recv, evalCtx, nil /* finishedSetupFn */)
+	// Copy the evalCtx, as dsp.Run() might change it.
+	evalCtxCopy := *evalCtx
+	dsp.Run(planCtx, txn, &p, recv, &evalCtxCopy, nil /* finishedSetupFn */)
 	return resultRows.Err()
 }
 
 // RowResultWriter is a thin wrapper around a RowContainer.
 type RowResultWriter struct {
-	rowContainer *sqlbase.RowContainer
+	rowContainer *rowcontainer.RowContainer
 	rowsAffected int
 	err          error
 }
@@ -100,7 +104,7 @@ type RowResultWriter struct {
 var _ rowResultWriter = &RowResultWriter{}
 
 // NewRowResultWriter creates a new RowResultWriter.
-func NewRowResultWriter(rowContainer *sqlbase.RowContainer) *RowResultWriter {
+func NewRowResultWriter(rowContainer *rowcontainer.RowContainer) *RowResultWriter {
 	return &RowResultWriter{rowContainer: rowContainer}
 }
 
@@ -182,66 +186,19 @@ func LoadCSV(
 	makeRewriter func(map[sqlbase.ID]*sqlbase.TableDescriptor) (KeyRewriter, error),
 ) error {
 	ctx = logtags.AddTag(ctx, "import-distsql", nil)
-
 	dsp := phs.DistSQLPlanner()
 	evalCtx := phs.ExtendedEvalContext()
-	planCtx := dsp.NewPlanningCtx(ctx, evalCtx, nil /* txn */)
 
-	resp, err := phs.ExecCfg().StatusServer.Nodes(ctx, &serverpb.NodesRequest{})
+	planCtx, nodes, err := dsp.setupAllNodesPlanning(ctx, evalCtx, phs.ExecCfg())
 	if err != nil {
 		return err
 	}
-	// Because we're not going through the normal pathways, we have to set up
-	// the nodeID -> nodeAddress map ourselves.
-	for _, node := range resp.Nodes {
-		if err := dsp.CheckNodeHealthAndVersion(&planCtx, &node.Desc); err != nil {
-			continue
-		}
-	}
-	nodes := make([]roachpb.NodeID, 0, len(planCtx.NodeAddresses))
-	for nodeID := range planCtx.NodeAddresses {
-		nodes = append(nodes, nodeID)
-	}
-	// Shuffle node order so that multiple IMPORTs done in parallel will not
-	// identically schedule CSV reading. For example, if there are 3 nodes and 4
-	// files, the first node will get 2 files while the other nodes will each get 1
-	// file. Shuffling will make that first node random instead of always the same.
-	rand.Shuffle(len(nodes), func(i, j int) {
-		nodes[i], nodes[j] = nodes[j], nodes[i]
-	})
 
-	// Setup common to both stages.
+	inputSpecs := makeImportReaderSpecs(job, tables, from, format, nodes, walltime)
 
-	// For each input file, assign it to a node.
-	inputSpecs := make([]*distsqlrun.ReadImportDataSpec, 0, len(nodes))
-	for i, input := range from {
-		// Round robin assign CSV files to nodes. Files 0 through len(nodes)-1
-		// creates the spec. Future files just add themselves to the Uris.
-		if i < len(nodes) {
-			spec := &distsqlrun.ReadImportDataSpec{
-				Tables: tables,
-				Format: format,
-				Progress: distsqlrun.JobProgress{
-					JobID: *job.ID(),
-					Slot:  int32(i),
-				},
-				Uri: make(map[int32]string),
-			}
-			inputSpecs = append(inputSpecs, spec)
-		}
-		n := i % len(nodes)
-		inputSpecs[n].Uri[int32(i)] = input
-	}
-
-	for i := range inputSpecs {
-		// TODO(mjibson): using the actual file sizes here would improve progress
-		// accuracy.
-		inputSpecs[i].Progress.Contribution = float32(len(inputSpecs[i].Uri)) / float32(len(from))
-	}
-
-	sstSpecs := make([]distsqlrun.SSTWriterSpec, len(nodes))
+	sstSpecs := make([]distsqlpb.SSTWriterSpec, len(nodes))
 	for i := range nodes {
-		sstSpecs[i] = distsqlrun.SSTWriterSpec{
+		sstSpecs[i] = distsqlpb.SSTWriterSpec{
 			Destination:   to,
 			WalltimeNanos: walltime,
 		}
@@ -256,7 +213,7 @@ func LoadCSV(
 	samples := details.Samples
 	if samples == nil {
 		var err error
-		samples, err = dsp.loadCSVSamplingPlan(ctx, job, db, evalCtx, thisNode, nodes, from, splitSize, oversample, &planCtx, inputSpecs, sstSpecs)
+		samples, err = dsp.loadCSVSamplingPlan(ctx, job, db, evalCtx, thisNode, nodes, from, splitSize, oversample, planCtx, inputSpecs, sstSpecs)
 		if err != nil {
 			return err
 		}
@@ -339,7 +296,7 @@ func LoadCSV(
 	// jobSpans is a slice of split points, including table start and end keys
 	// for the table. We create router range spans then from taking each pair
 	// of adjacent keys.
-	spans := make([]distsqlrun.OutputRouterSpec_RangeRouterSpec_Span, len(splits)-1)
+	spans := make([]distsqlpb.OutputRouterSpec_RangeRouterSpec_Span, len(splits)-1)
 	encFn := func(b []byte) []byte {
 		return encoding.EncodeBytesAscending(nil, b)
 	}
@@ -347,12 +304,12 @@ func LoadCSV(
 		start := splits[i-1]
 		end := splits[i]
 		stream := int32((i - 1) % len(nodes))
-		spans[i-1] = distsqlrun.OutputRouterSpec_RangeRouterSpec_Span{
+		spans[i-1] = distsqlpb.OutputRouterSpec_RangeRouterSpec_Span{
 			Start:  encFn(start),
 			End:    encFn(end),
 			Stream: stream,
 		}
-		sstSpecs[stream].Spans = append(sstSpecs[stream].Spans, distsqlrun.SSTWriterSpec_SpanName{
+		sstSpecs[stream].Spans = append(sstSpecs[stream].Spans, distsqlpb.SSTWriterSpec_SpanName{
 			Name: fmt.Sprintf("%d.sst", i),
 			End:  end,
 		})
@@ -367,9 +324,9 @@ func LoadCSV(
 	firstStageRouters := make([]distsqlplan.ProcessorIdx, len(inputSpecs))
 	firstStageTypes := []sqlbase.ColumnType{colTypeBytes, colTypeBytes}
 
-	routerSpec := distsqlrun.OutputRouterSpec_RangeRouterSpec{
+	routerSpec := distsqlpb.OutputRouterSpec_RangeRouterSpec{
 		Spans: spans,
-		Encodings: []distsqlrun.OutputRouterSpec_RangeRouterSpec_ColumnEncoding{
+		Encodings: []distsqlpb.OutputRouterSpec_RangeRouterSpec_ColumnEncoding{
 			{
 				Column:   0,
 				Encoding: sqlbase.DatumEncoding_ASCENDING_KEY,
@@ -383,10 +340,10 @@ func LoadCSV(
 		rcs.SampleSize = 0
 		proc := distsqlplan.Processor{
 			Node: nodes[i],
-			Spec: distsqlrun.ProcessorSpec{
-				Core: distsqlrun.ProcessorCoreUnion{ReadImport: rcs},
-				Output: []distsqlrun.OutputRouterSpec{{
-					Type:             distsqlrun.OutputRouterSpec_BY_RANGE,
+			Spec: distsqlpb.ProcessorSpec{
+				Core: distsqlpb.ProcessorCoreUnion{ReadImport: rcs},
+				Output: []distsqlpb.OutputRouterSpec{{
+					Type:             distsqlpb.OutputRouterSpec_BY_RANGE,
 					RangeRouterSpec:  routerSpec,
 					DisableBuffering: true,
 				}},
@@ -415,19 +372,19 @@ func LoadCSV(
 		if len(swSpec.Spans) == 0 {
 			continue
 		}
-		swSpec.Progress = distsqlrun.JobProgress{
+		swSpec.Progress = distsqlpb.JobProgress{
 			JobID:        *job.ID(),
 			Slot:         int32(len(p.ResultRouters)),
 			Contribution: float32(len(swSpec.Spans)) / float32(len(spans)),
 		}
 		proc := distsqlplan.Processor{
 			Node: node,
-			Spec: distsqlrun.ProcessorSpec{
-				Input: []distsqlrun.InputSyncSpec{{
+			Spec: distsqlpb.ProcessorSpec{
+				Input: []distsqlpb.InputSyncSpec{{
 					ColumnTypes: firstStageTypes,
 				}},
-				Core:    distsqlrun.ProcessorCoreUnion{SSTWriter: &swSpec},
-				Output:  []distsqlrun.OutputRouterSpec{{Type: distsqlrun.OutputRouterSpec_PASS_THROUGH}},
+				Core:    distsqlpb.ProcessorCoreUnion{SSTWriter: &swSpec},
+				Output:  []distsqlpb.OutputRouterSpec{{Type: distsqlpb.OutputRouterSpec_PASS_THROUGH}},
 				StageID: stageID,
 			},
 		}
@@ -462,7 +419,7 @@ func LoadCSV(
 		return err
 	}
 
-	dsp.FinalizePlan(&planCtx, &p)
+	dsp.FinalizePlan(planCtx, &p)
 
 	recv := MakeDistSQLReceiver(
 		ctx,
@@ -474,14 +431,86 @@ func LoadCSV(
 		func(ts hlc.Timestamp) {},
 		evalCtx.Tracing,
 	)
+	defer recv.Release()
 
 	defer log.VEventf(ctx, 1, "finished job %s", job.Payload().Description)
+	// Copy the evalCtx, as dsp.Run() might change it.
+	evalCtxCopy := *evalCtx
 	return db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
-		dsp.Run(&planCtx, txn, &p, recv, evalCtx, nil /* finishedSetupFn */)
+		dsp.Run(planCtx, txn, &p, recv, &evalCtxCopy, nil /* finishedSetupFn */)
 		return resultRows.Err()
 	})
 }
 
+func (dsp *DistSQLPlanner) setupAllNodesPlanning(
+	ctx context.Context, evalCtx *extendedEvalContext, execCfg *ExecutorConfig,
+) (*PlanningCtx, []roachpb.NodeID, error) {
+	planCtx := dsp.NewPlanningCtx(ctx, evalCtx, nil /* txn */)
+
+	resp, err := execCfg.StatusServer.Nodes(ctx, &serverpb.NodesRequest{})
+	if err != nil {
+		return nil, nil, err
+	}
+	// Because we're not going through the normal pathways, we have to set up
+	// the nodeID -> nodeAddress map ourselves.
+	for _, node := range resp.Nodes {
+		if err := dsp.CheckNodeHealthAndVersion(planCtx, &node.Desc); err != nil {
+			continue
+		}
+	}
+	nodes := make([]roachpb.NodeID, 0, len(planCtx.NodeAddresses))
+	for nodeID := range planCtx.NodeAddresses {
+		nodes = append(nodes, nodeID)
+	}
+	// Shuffle node order so that multiple IMPORTs done in parallel will not
+	// identically schedule CSV reading. For example, if there are 3 nodes and 4
+	// files, the first node will get 2 files while the other nodes will each get 1
+	// file. Shuffling will make that first node random instead of always the same.
+	rand.Shuffle(len(nodes), func(i, j int) {
+		nodes[i], nodes[j] = nodes[j], nodes[i]
+	})
+	return planCtx, nodes, nil
+}
+
+func makeImportReaderSpecs(
+	job *jobs.Job,
+	tables map[string]*sqlbase.TableDescriptor,
+	from []string,
+	format roachpb.IOFileFormat,
+	nodes []roachpb.NodeID,
+	walltime int64,
+) []*distsqlpb.ReadImportDataSpec {
+
+	// For each input file, assign it to a node.
+	inputSpecs := make([]*distsqlpb.ReadImportDataSpec, 0, len(nodes))
+	for i, input := range from {
+		// Round robin assign CSV files to nodes. Files 0 through len(nodes)-1
+		// creates the spec. Future files just add themselves to the Uris.
+		if i < len(nodes) {
+			spec := &distsqlpb.ReadImportDataSpec{
+				Tables: tables,
+				Format: format,
+				Progress: distsqlpb.JobProgress{
+					JobID: *job.ID(),
+					Slot:  int32(i),
+				},
+				WalltimeNanos: walltime,
+				Uri:           make(map[int32]string),
+			}
+			inputSpecs = append(inputSpecs, spec)
+		}
+		n := i % len(nodes)
+		inputSpecs[n].Uri[int32(i)] = input
+	}
+
+	for i := range inputSpecs {
+		// TODO(mjibson): using the actual file sizes here would improve progress
+		// accuracy.
+		inputSpecs[i].Progress.Contribution = float32(len(inputSpecs[i].Uri)) / float32(len(from))
+	}
+	return inputSpecs
+}
+
 func (dsp *DistSQLPlanner) loadCSVSamplingPlan(
 	ctx context.Context,
 	job *jobs.Job,
@@ -493,8 +522,8 @@ func (dsp *DistSQLPlanner) loadCSVSamplingPlan(
 	splitSize int64,
 	oversample int64,
 	planCtx *PlanningCtx,
-	csvSpecs []*distsqlrun.ReadImportDataSpec,
-	sstSpecs []distsqlrun.SSTWriterSpec,
+	csvSpecs []*distsqlpb.ReadImportDataSpec,
+	sstSpecs []distsqlpb.SSTWriterSpec,
 ) ([][]byte, error) {
 	// splitSize is the target number of bytes at which to create SST files. We
 	// attempt to do this by sampling, which is what the first DistSQL plan of this
@@ -527,9 +556,9 @@ func (dsp *DistSQLPlanner) loadCSVSamplingPlan(
 	for i, rcs := range csvSpecs {
 		proc := distsqlplan.Processor{
 			Node: nodes[i],
-			Spec: distsqlrun.ProcessorSpec{
-				Core:    distsqlrun.ProcessorCoreUnion{ReadImport: rcs},
-				Output:  []distsqlrun.OutputRouterSpec{{Type: distsqlrun.OutputRouterSpec_PASS_THROUGH}},
+			Spec: distsqlpb.ProcessorSpec{
+				Core:    distsqlpb.ProcessorCoreUnion{ReadImport: rcs},
+				Output:  []distsqlpb.OutputRouterSpec{{Type: distsqlpb.OutputRouterSpec_PASS_THROUGH}},
 				StageID: stageID,
 			},
 		}
@@ -549,23 +578,23 @@ func (dsp *DistSQLPlanner) loadCSVSamplingPlan(
 	p.PlanToStreamColMap = []int{0, 1}
 	p.ResultTypes = []sqlbase.ColumnType{colTypeBytes, colTypeBytes}
 
-	kvOrdering := distsqlrun.Ordering{
-		Columns: []distsqlrun.Ordering_Column{{
+	kvOrdering := distsqlpb.Ordering{
+		Columns: []distsqlpb.Ordering_Column{{
 			ColIdx:    0,
-			Direction: distsqlrun.Ordering_Column_ASC,
+			Direction: distsqlpb.Ordering_Column_ASC,
 		}, {
 			ColIdx:    1,
-			Direction: distsqlrun.Ordering_Column_ASC,
+			Direction: distsqlpb.Ordering_Column_ASC,
 		}},
 	}
 
-	sorterSpec := distsqlrun.SorterSpec{
+	sorterSpec := distsqlpb.SorterSpec{
 		OutputOrdering: kvOrdering,
 	}
 
 	p.AddSingleGroupStage(thisNode,
-		distsqlrun.ProcessorCoreUnion{Sorter: &sorterSpec},
-		distsqlrun.PostProcessSpec{},
+		distsqlpb.ProcessorCoreUnion{Sorter: &sorterSpec},
+		distsqlpb.PostProcessSpec{},
 		[]sqlbase.ColumnType{colTypeBytes, colTypeBytes},
 	)
 
@@ -614,10 +643,13 @@ func (dsp *DistSQLPlanner) loadCSVSamplingPlan(
 		func(ts hlc.Timestamp) {},
 		evalCtx.Tracing,
 	)
+	defer recv.Release()
 	log.VEventf(ctx, 1, "begin sampling phase of job %s", job.Payload().Description)
 	// Clear the stage 2 data in case this function is ever restarted (it shouldn't be).
 	samples = nil
-	dsp.Run(planCtx, nil, &p, recv, evalCtx, nil /* finishedSetupFn */)
+	// Copy the evalCtx, as dsp.Run() might change it.
+	evalCtxCopy := *evalCtx
+	dsp.Run(planCtx, nil, &p, recv, &evalCtxCopy, nil /* finishedSetupFn */)
 	if err := rowResultWriter.Err(); err != nil {
 		return nil, err
 	}
@@ -626,3 +658,98 @@ func (dsp *DistSQLPlanner) loadCSVSamplingPlan(
 
 	return samples, nil
 }
+
+// DistIngest is used by IMPORT to run a DistSQL flow to ingest data by starting
+// reader processes on many nodes that each read and ingest their assigned files
+// and then send back a summary of what they ingested. The combined summary is
+// returned.
+func DistIngest(
+	ctx context.Context,
+	phs PlanHookState,
+	job *jobs.Job,
+	tables map[string]*sqlbase.TableDescriptor,
+	from []string,
+	format roachpb.IOFileFormat,
+	walltime int64,
+) (roachpb.BulkOpSummary, error) {
+	ctx = logtags.AddTag(ctx, "import-distsql-ingest", nil)
+
+	dsp := phs.DistSQLPlanner()
+	evalCtx := phs.ExtendedEvalContext()
+
+	planCtx, nodes, err := dsp.setupAllNodesPlanning(ctx, evalCtx, phs.ExecCfg())
+	if err != nil {
+		return roachpb.BulkOpSummary{}, err
+	}
+
+	inputSpecs := makeImportReaderSpecs(job, tables, from, format, nodes, walltime)
+
+	for i := range inputSpecs {
+		inputSpecs[i].IngestDirectly = true
+	}
+
+	var p PhysicalPlan
+
+	// Setup a one-stage plan with one proc per input spec.
+	stageID := p.NewStageID()
+	p.ResultRouters = make([]distsqlplan.ProcessorIdx, len(inputSpecs))
+	for i, rcs := range inputSpecs {
+		proc := distsqlplan.Processor{
+			Node: nodes[i],
+			Spec: distsqlpb.ProcessorSpec{
+				Core:    distsqlpb.ProcessorCoreUnion{ReadImport: rcs},
+				Output:  []distsqlpb.OutputRouterSpec{{Type: distsqlpb.OutputRouterSpec_PASS_THROUGH}},
+				StageID: stageID,
+			},
+		}
+		pIdx := p.AddProcessor(proc)
+		p.ResultRouters[i] = pIdx
+	}
+
+	var res roachpb.BulkOpSummary
+
+	// The direct-ingest readers will emit a binary encoded BulkOpSummary.
+	p.PlanToStreamColMap = []int{0}
+	p.ResultTypes = []sqlbase.ColumnType{colTypeBytes}
+
+	rowResultWriter := newCallbackResultWriter(func(ctx context.Context, row tree.Datums) error {
+		var counts roachpb.BulkOpSummary
+		if err := protoutil.Unmarshal([]byte(*row[0].(*tree.DBytes)), &counts); err != nil {
+			return err
+		}
+		res.Add(counts)
+		return nil
+	})
+
+	dsp.FinalizePlan(planCtx, &p)
+
+	if err := job.FractionDetailProgressed(ctx,
+		func(ctx context.Context, details jobspb.Details, progress jobspb.ProgressDetails) float32 {
+			prog := progress.(*jobspb.Progress_Import).Import
+			prog.ReadProgress = make([]float32, len(inputSpecs))
+			return prog.Completed()
+		},
+	); err != nil {
+		return roachpb.BulkOpSummary{}, err
+	}
+
+	recv := MakeDistSQLReceiver(
+		ctx,
+		rowResultWriter,
+		tree.Rows,
+		nil, /* rangeCache */
+		nil, /* leaseCache */
+		nil, /* txn - the flow does not read or write the database */
+		func(ts hlc.Timestamp) {},
+		evalCtx.Tracing,
+	)
+	defer recv.Release()
+	// Copy the evalCtx, as dsp.Run() might change it.
+	evalCtxCopy := *evalCtx
+	dsp.Run(planCtx, nil, &p, recv, &evalCtxCopy, nil /* finishedSetupFn */)
+	if err := rowResultWriter.Err(); err != nil {
+		return roachpb.BulkOpSummary{}, err
+	}
+
+	return res, nil
+}
