@@ -16,23 +16,26 @@ package sql
 
 import (
 	"context"
-	"sync/atomic"
-
 	"fmt"
+	"sync"
+	"sync/atomic"
 
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/kv"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
-	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlplan"
 	"github.com/cockroachdb/cockroach/pkg/sql/distsqlrun"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
-	"github.com/opentracing/opentracing-go"
+	opentracing "github.com/opentracing/opentracing-go"
 	"github.com/pkg/errors"
 )
 
@@ -47,7 +50,7 @@ const clientRejectedMsg string = "client rejected when attempting to run DistSQL
 type runnerRequest struct {
 	ctx        context.Context
 	nodeDialer *nodedialer.Dialer
-	flowReq    *distsqlrun.SetupFlowRequest
+	flowReq    *distsqlpb.SetupFlowRequest
 	nodeID     roachpb.NodeID
 	resultChan chan<- runnerResult
 }
@@ -60,13 +63,15 @@ type runnerResult struct {
 }
 
 func (req runnerRequest) run() {
+	defer distsqlplan.ReleaseSetupFlowRequest(req.flowReq)
+
 	res := runnerResult{nodeID: req.nodeID}
 
 	conn, err := req.nodeDialer.Dial(req.ctx, req.nodeID)
 	if err != nil {
 		res.err = err
 	} else {
-		client := distsqlrun.NewDistSQLClient(conn)
+		client := distsqlpb.NewDistSQLClient(conn)
 		// TODO(radu): do we want a timeout here?
 		resp, err := client.SetupFlow(req.ctx, req.flowReq)
 		if err != nil {
@@ -102,17 +107,19 @@ func (dsp *DistSQLPlanner) initRunners() {
 // Run executes a physical plan. The plan should have been finalized using
 // FinalizePlan.
 //
-// txn is the transaction in which the plan will run. If nil, the different
-// processors are expected to manage their own internal transactions.
-//
-// `finishedSetupFn`, if non-nil, is called synchronously after all the
-// processors have successfully started up.
-//
 // All errors encountered are reported to the DistSQLReceiver's resultWriter.
 // Additionally, if the error is a "communication error" (an error encountered
 // while using that resultWriter), the error is also stored in
 // DistSQLReceiver.commErr. That can be tested to see if a client session needs
 // to be closed.
+//
+// Args:
+// - txn is the transaction in which the plan will run. If nil, the different
+// processors are expected to manage their own internal transactions.
+// - evalCtx is the evaluation context in which the plan will run. It might be
+// mutated.
+// - finishedSetupFn, if non-nil, is called synchronously after all the
+// processors have successfully started up.
 func (dsp *DistSQLPlanner) Run(
 	planCtx *PlanningCtx,
 	txn *client.Txn,
@@ -127,9 +134,9 @@ func (dsp *DistSQLPlanner) Run(
 		localState   distsqlrun.LocalState
 		txnCoordMeta *roachpb.TxnCoordMeta
 	)
-	if planCtx.validExtendedEvalCtx {
-		localState.EvalContext = planCtx.EvalContext()
-	}
+	// NB: putting part of evalCtx in localState means it might be mutated down
+	// the line.
+	localState.EvalContext = &evalCtx.EvalContext
 	if planCtx.isLocal {
 		localState.IsLocal = true
 		localState.LocalProcs = plan.LocalProcessors
@@ -156,7 +163,7 @@ func (dsp *DistSQLPlanner) Run(
 
 	if logPlanDiagram {
 		log.VEvent(ctx, 1, "creating plan diagram")
-		json, url, err := distsqlrun.GeneratePlanDiagramURL(flows)
+		json, url, err := distsqlpb.GeneratePlanDiagramURL(flows)
 		if err != nil {
 			log.Infof(ctx, "Error generating diagram: %s", err)
 		} else {
@@ -174,21 +181,12 @@ func (dsp *DistSQLPlanner) Run(
 	recv.resultToStreamColMap = plan.PlanToStreamColMap
 	thisNodeID := dsp.nodeDesc.NodeID
 
-	evalCtxProto := distsqlrun.MakeEvalContext(evalCtx.EvalContext)
-	setupReq := distsqlrun.SetupFlowRequest{
-		Version:     distsqlrun.Version,
-		EvalContext: evalCtxProto,
-		TraceKV:     evalCtx.Tracing.KVTracingEnabled(),
-	}
-	if txnCoordMeta != nil {
-		// The receiver may not know about the TxnCoordMeta field if it is
-		// an old binary. In this case, set the DeprecatedTxn field as well.
-		// TODO(nvanbenschoten): remove in 2.2.
-		if !dsp.st.Version.IsActive(cluster.VersionTxnCoordMetaInvalidField) {
-			setupReq.DeprecatedTxn = &txnCoordMeta.Txn
-		} else {
-			setupReq.TxnCoordMeta = txnCoordMeta
-		}
+	evalCtxProto := distsqlpb.MakeEvalContext(evalCtx.EvalContext)
+	setupReq := distsqlpb.SetupFlowRequest{
+		TxnCoordMeta: txnCoordMeta,
+		Version:      distsqlrun.Version,
+		EvalContext:  evalCtxProto,
+		TraceKV:      evalCtx.Tracing.KVTracingEnabled(),
 	}
 
 	// Start all the flows except the flow on this node (there is always a flow on
@@ -203,7 +201,7 @@ func (dsp *DistSQLPlanner) Run(
 			continue
 		}
 		req := setupReq
-		req.Flow = flowSpec
+		req.Flow = *flowSpec
 		runReq := runnerRequest{
 			ctx:        ctx,
 			nodeDialer: dsp.nodeDialer,
@@ -238,7 +236,8 @@ func (dsp *DistSQLPlanner) Run(
 
 	// Set up the flow on this node.
 	localReq := setupReq
-	localReq.Flow = flows[thisNodeID]
+	localReq.Flow = *flows[thisNodeID]
+	defer distsqlplan.ReleaseSetupFlowRequest(&localReq)
 	ctx, flow, err := dsp.distSQLSrv.SetupLocalSyncFlow(ctx, evalCtx.Mon, &localReq, recv, localState)
 	if err != nil {
 		recv.SetError(err)
@@ -250,20 +249,30 @@ func (dsp *DistSQLPlanner) Run(
 	}
 
 	// TODO(radu): this should go through the flow scheduler.
-	if err := flow.StartSync(ctx, func() {}); err != nil {
+	if err := flow.Run(ctx, func() {}); err != nil {
 		log.Fatalf(ctx, "unexpected error from syncFlow.Start(): %s "+
 			"The error should have gone to the consumer.", err)
 	}
 	// We need to close the planNode tree we translated into a DistSQL plan before
 	// flow.Cleanup, which closes memory accounts that expect to be emptied.
 	if planCtx.planner != nil && !planCtx.ignoreClose {
+		planCtx.planner.curPlan.execErr = recv.resultWriter.Err()
 		planCtx.planner.curPlan.close(ctx)
 	}
-
-	flow.Wait()
 	flow.Cleanup(ctx)
 }
 
+// errorPriority is used to rank errors such that the "best" one is chosen to be
+// presented as the query result.
+type errorPriority int
+
+const (
+	scoreNoError errorPriority = iota
+	scoreTxnRestart
+	scoreTxnAbort
+	scoreNonRetriable
+)
+
 // DistSQLReceiver is a RowReceiver that writes results to a rowResultWriter.
 // This is where the DistSQL execution meets the SQL Session - the RowContainer
 // comes from a client Session.
@@ -289,6 +298,10 @@ type DistSQLReceiver struct {
 	// existence of a single row. Used by subqueries in EXISTS mode.
 	noColsRequired bool
 
+	// discardRows is set when we want to discard rows (for testing/benchmarks).
+	// See EXECUTE .. DISCARD ROWS.
+	discardRows bool
+
 	// commErr keeps track of the error received from interacting with the
 	// resultWriter. This represents a "communication error" and as such is unlike
 	// query execution errors: when the DistSQLReceiver is used within a SQL
@@ -369,6 +382,12 @@ func (w *errOnlyResultWriter) IncrementRowsAffected(n int) {
 
 var _ distsqlrun.RowReceiver = &DistSQLReceiver{}
 
+var receiverSyncPool = sync.Pool{
+	New: func() interface{} {
+		return &DistSQLReceiver{}
+	},
+}
+
 // MakeDistSQLReceiver creates a DistSQLReceiver.
 //
 // ctx is the Context that the receiver will use throughput its
@@ -388,7 +407,8 @@ func MakeDistSQLReceiver(
 	tracing *SessionTracing,
 ) *DistSQLReceiver {
 	consumeCtx, cleanup := tracing.TraceExecConsume(ctx)
-	r := &DistSQLReceiver{
+	r := receiverSyncPool.Get().(*DistSQLReceiver)
+	*r = DistSQLReceiver{
 		ctx:          consumeCtx,
 		cleanup:      cleanup,
 		resultWriter: resultWriter,
@@ -417,10 +437,17 @@ func MakeDistSQLReceiver(
 	return r
 }
 
+// Release releases this DistSQLReceiver back to the pool.
+func (r *DistSQLReceiver) Release() {
+	*r = DistSQLReceiver{}
+	receiverSyncPool.Put(r)
+}
+
 // clone clones the receiver for running subqueries. Not all fields are cloned,
 // only those required for running subqueries.
 func (r *DistSQLReceiver) clone() *DistSQLReceiver {
-	return &DistSQLReceiver{
+	ret := receiverSyncPool.Get().(*DistSQLReceiver)
+	*ret = DistSQLReceiver{
 		ctx:         r.ctx,
 		cleanup:     func() {},
 		rangeCache:  r.rangeCache,
@@ -430,6 +457,7 @@ func (r *DistSQLReceiver) clone() *DistSQLReceiver {
 		stmtType:    tree.Rows,
 		tracing:     r.tracing,
 	}
+	return ret
 }
 
 // SetError provides a convenient way for a client to pass in an error, thus
@@ -454,23 +482,27 @@ func (r *DistSQLReceiver) Push(
 					errors.Errorf("received a leaf TxnCoordMeta (%s); but have no root", meta.TxnCoordMeta))
 			}
 		}
-		if meta.Err != nil && r.resultWriter.Err() == nil {
-			if r.txn != nil {
-				if retryErr, ok := meta.Err.(*roachpb.UnhandledRetryableError); ok {
-					// Update the txn in response to remote errors. In the non-DistSQL
-					// world, the TxnCoordSender handles "unhandled" retryable errors,
-					// but this one is coming from a distributed SQL node, which has
-					// left the handling up to the root transaction.
-					meta.Err = r.txn.UpdateStateOnRemoteRetryableErr(r.ctx, &retryErr.PErr)
-					// Update the clock with information from the error. On non-DistSQL
-					// code paths, the DistSender does this.
-					// TODO(andrei): We don't propagate clock signals on success cases
-					// through DistSQL; we should. We also don't propagate them through
-					// non-retryable errors; we also should.
-					r.updateClock(retryErr.PErr.Now)
+		if meta.Err != nil {
+			// Check if the error we just received should take precedence over a
+			// previous error (if any).
+			if errPriority(meta.Err) > errPriority(r.resultWriter.Err()) {
+				if r.txn != nil {
+					if retryErr, ok := meta.Err.(*roachpb.UnhandledRetryableError); ok {
+						// Update the txn in response to remote errors. In the non-DistSQL
+						// world, the TxnCoordSender handles "unhandled" retryable errors,
+						// but this one is coming from a distributed SQL node, which has
+						// left the handling up to the root transaction.
+						meta.Err = r.txn.UpdateStateOnRemoteRetryableErr(r.ctx, &retryErr.PErr)
+						// Update the clock with information from the error. On non-DistSQL
+						// code paths, the DistSender does this.
+						// TODO(andrei): We don't propagate clock signals on success cases
+						// through DistSQL; we should. We also don't propagate them through
+						// non-retryable errors; we also should.
+						r.updateClock(retryErr.PErr.Now)
+					}
 				}
+				r.resultWriter.SetError(meta.Err)
 			}
-			r.resultWriter.SetError(meta.Err)
 		}
 		if len(meta.Ranges) > 0 {
 			if err := r.updateCaches(r.ctx, meta.Ranges); err != nil && r.resultWriter.Err() == nil {
@@ -509,6 +541,12 @@ func (r *DistSQLReceiver) Push(
 		r.resultWriter.IncrementRowsAffected(int(tree.MustBeDInt(row[0].Datum)))
 		return r.status
 	}
+
+	if r.discardRows {
+		// Discard rows.
+		return r.status
+	}
+
 	// If no columns are needed by the output, the consumer is only looking for
 	// whether a single row is pushed or not, so the contents do not matter, and
 	// planNodeToRowSource is not set up to handle decoding the row.
@@ -549,6 +587,29 @@ func (r *DistSQLReceiver) Push(
 	return r.status
 }
 
+// errPriority computes the priority of err.
+func errPriority(err error) errorPriority {
+	if err == nil {
+		return scoreNoError
+	}
+	if retryErr, ok := err.(*roachpb.UnhandledRetryableError); ok {
+		pErr := retryErr.PErr
+		switch pErr.GetDetail().(type) {
+		case *roachpb.TransactionAbortedError:
+			return scoreTxnAbort
+		default:
+			return scoreTxnRestart
+		}
+	}
+	if retryErr, ok := err.(*roachpb.TransactionRetryWithProtoRefreshError); ok {
+		if retryErr.PrevTxnAborted() {
+			return scoreTxnAbort
+		}
+		return scoreTxnRestart
+	}
+	return scoreNonRetriable
+}
+
 // ProducerDone is part of the RowReceiver interface.
 func (r *DistSQLReceiver) ProducerDone() {
 	if r.txn != nil {
@@ -561,6 +622,11 @@ func (r *DistSQLReceiver) ProducerDone() {
 	r.cleanup()
 }
 
+// Types is part of the RowReceiver interface.
+func (r *DistSQLReceiver) Types() []sqlbase.ColumnType {
+	return r.outputTypes
+}
+
 // updateCaches takes information about some ranges that were mis-planned and
 // updates the range descriptor and lease-holder caches accordingly.
 //
@@ -595,142 +661,157 @@ func (dsp *DistSQLPlanner) PlanAndRunSubqueries(
 	recv *DistSQLReceiver,
 	maybeDistribute bool,
 ) bool {
-	var subqueryMemAccount mon.BoundAccount
-
 	for planIdx, subqueryPlan := range subqueryPlans {
-		evalCtx := evalCtxFactory()
-
-		subqueryMonitor := mon.MakeMonitor(
-			"subquery",
-			mon.MemoryResource,
-			dsp.distSQLSrv.Metrics.CurBytesCount,
-			dsp.distSQLSrv.Metrics.MaxBytesHist,
-			-1, /* use default block size */
-			noteworthyMemoryUsageBytes,
-			dsp.distSQLSrv.Settings,
-		)
-		subqueryMonitor.Start(ctx, evalCtx.Mon, mon.BoundAccount{})
-		defer subqueryMonitor.Stop(ctx)
-
-		subqueryMemAccount = subqueryMonitor.MakeBoundAccount()
-		defer subqueryMemAccount.Close(ctx)
-
-		evalCtx.ActiveMemAcc = &subqueryMemAccount
-
-		var subqueryPlanCtx PlanningCtx
-		var distributeSubquery bool
-		if maybeDistribute {
-			distributeSubquery = shouldDistributePlan(
-				ctx, planner.SessionData().DistSQLMode, dsp, subqueryPlan.plan)
-		}
-		if distributeSubquery {
-			subqueryPlanCtx = dsp.NewPlanningCtx(ctx, evalCtx, planner.txn)
-		} else {
-			subqueryPlanCtx = dsp.newLocalPlanningCtx(ctx, evalCtx)
-		}
-
-		subqueryPlanCtx.isLocal = !distributeSubquery
-		subqueryPlanCtx.planner = planner
-		subqueryPlanCtx.stmtType = tree.Rows
-		subqueryPlanCtx.validExtendedEvalCtx = true
-		// Don't close the top-level plan from subqueries - someone else will handle
-		// that.
-		subqueryPlanCtx.ignoreClose = true
-
-		subqueryPhysPlan, err := dsp.createPlanForNode(&subqueryPlanCtx, subqueryPlan.plan)
-		if err != nil {
+		if err := dsp.planAndRunSubquery(
+			ctx,
+			planIdx,
+			subqueryPlan,
+			planner,
+			evalCtxFactory(),
+			subqueryPlans,
+			recv,
+			maybeDistribute,
+		); err != nil {
 			recv.SetError(err)
 			return false
 		}
-		dsp.FinalizePlan(&subqueryPlanCtx, &subqueryPhysPlan)
-
-		// TODO(arjun): #28264: We set up a row container, wrap it in a row
-		// receiver, and use it and serialize the results of the subquery. The type
-		// of the results stored in the container depends on the type of the subquery.
-		subqueryRecv := recv.clone()
-		var typ sqlbase.ColTypeInfo
-		var rows *sqlbase.RowContainer
-		if subqueryPlan.execMode == distsqlrun.SubqueryExecModeExists {
-			subqueryRecv.noColsRequired = true
-			typ = sqlbase.ColTypeInfoFromColTypes([]sqlbase.ColumnType{})
-		} else {
-			// Apply the PlanToStreamColMap projection to the ResultTypes to get the
-			// final set of output types for the subquery. The reason this is necessary
-			// is that the output schema of a query sometimes contains columns necessary
-			// to merge the streams, but that aren't required by the final output of the
-			// query. These get projected out, so we need to similarly adjust the
-			// expected result types of the subquery here.
-			colTypes := make([]sqlbase.ColumnType, len(subqueryPhysPlan.PlanToStreamColMap))
-			for i, resIdx := range subqueryPhysPlan.PlanToStreamColMap {
-				colTypes[i] = subqueryPhysPlan.ResultTypes[resIdx]
-			}
-			typ = sqlbase.ColTypeInfoFromColTypes(colTypes)
-		}
-		rows = sqlbase.NewRowContainer(subqueryMemAccount, typ, 0)
-		defer rows.Close(evalCtx.Ctx())
-
-		subqueryRowReceiver := NewRowResultWriter(rows)
-		subqueryRecv.resultWriter = subqueryRowReceiver
-		subqueryPlans[planIdx].started = true
-		dsp.Run(&subqueryPlanCtx, planner.txn, &subqueryPhysPlan, subqueryRecv, evalCtx, nil /* finishedSetupFn */)
-		if subqueryRecv.commErr != nil {
-			recv.SetError(subqueryRecv.commErr)
-			return false
-		}
-		if subqueryRowReceiver.Err() != nil {
-			recv.SetError(subqueryRowReceiver.Err())
-			return false
-		}
-		switch subqueryPlan.execMode {
-		case distsqlrun.SubqueryExecModeExists:
-			// For EXISTS expressions, all we want to know if there is at least one row.
-			hasRows := rows.Len() != 0
-			subqueryPlans[planIdx].result = tree.MakeDBool(tree.DBool(hasRows))
-		case distsqlrun.SubqueryExecModeAllRows, distsqlrun.SubqueryExecModeAllRowsNormalized:
-			var result tree.DTuple
-			for rows.Len() > 0 {
-				row := rows.At(0)
-				rows.PopFirst()
-				if row.Len() == 1 {
-					// This seems hokey, but if we don't do this then the subquery expands
-					// to a tuple of tuples instead of a tuple of values and an expression
-					// like "k IN (SELECT foo FROM bar)" will fail because we're comparing
-					// a single value against a tuple.
-					result.D = append(result.D, row[0])
-				} else {
-					result.D = append(result.D, &tree.DTuple{D: row})
-				}
-			}
-
-			if subqueryPlan.execMode == distsqlrun.SubqueryExecModeAllRowsNormalized {
-				result.Normalize(&evalCtx.EvalContext)
-			}
-			subqueryPlans[planIdx].result = &result
-		case distsqlrun.SubqueryExecModeOneRow:
-			switch rows.Len() {
-			case 0:
-				subqueryPlans[planIdx].result = tree.DNull
-			case 1:
-				row := rows.At(0)
-				switch row.Len() {
-				case 1:
-					subqueryPlans[planIdx].result = row[0]
-				default:
-					subqueryPlans[planIdx].result = &tree.DTuple{D: rows.At(0)}
-				}
-			default:
-				recv.SetError(fmt.Errorf("more than one row returned by a subquery used as an expression"))
-				return false
-			}
-		default:
-			recv.SetError(fmt.Errorf("unexpected subqueryExecMode: %d", subqueryPlan.execMode))
-			return false
-		}
 	}
 
 	return true
 }
 
+func (dsp *DistSQLPlanner) planAndRunSubquery(
+	ctx context.Context,
+	planIdx int,
+	subqueryPlan subquery,
+	planner *planner,
+	evalCtx *extendedEvalContext,
+	subqueryPlans []subquery,
+	recv *DistSQLReceiver,
+	maybeDistribute bool,
+) error {
+	subqueryMonitor := mon.MakeMonitor(
+		"subquery",
+		mon.MemoryResource,
+		dsp.distSQLSrv.Metrics.CurBytesCount,
+		dsp.distSQLSrv.Metrics.MaxBytesHist,
+		-1, /* use default block size */
+		noteworthyMemoryUsageBytes,
+		dsp.distSQLSrv.Settings,
+	)
+	subqueryMonitor.Start(ctx, evalCtx.Mon, mon.BoundAccount{})
+	defer subqueryMonitor.Stop(ctx)
+
+	subqueryMemAccount := subqueryMonitor.MakeBoundAccount()
+	defer subqueryMemAccount.Close(ctx)
+
+	var subqueryPlanCtx *PlanningCtx
+	var distributeSubquery bool
+	if maybeDistribute {
+		distributeSubquery = shouldDistributePlan(
+			ctx, planner.SessionData().DistSQLMode, dsp, subqueryPlan.plan)
+	}
+	if distributeSubquery {
+		subqueryPlanCtx = dsp.NewPlanningCtx(ctx, evalCtx, planner.txn)
+	} else {
+		subqueryPlanCtx = dsp.newLocalPlanningCtx(ctx, evalCtx)
+	}
+
+	subqueryPlanCtx.isLocal = !distributeSubquery
+	subqueryPlanCtx.planner = planner
+	subqueryPlanCtx.stmtType = tree.Rows
+	// Don't close the top-level plan from subqueries - someone else will handle
+	// that.
+	subqueryPlanCtx.ignoreClose = true
+
+	subqueryPhysPlan, err := dsp.createPlanForNode(subqueryPlanCtx, subqueryPlan.plan)
+	if err != nil {
+		return err
+	}
+	dsp.FinalizePlan(subqueryPlanCtx, &subqueryPhysPlan)
+
+	// TODO(arjun): #28264: We set up a row container, wrap it in a row
+	// receiver, and use it and serialize the results of the subquery. The type
+	// of the results stored in the container depends on the type of the subquery.
+	subqueryRecv := recv.clone()
+	var typ sqlbase.ColTypeInfo
+	var rows *rowcontainer.RowContainer
+	if subqueryPlan.execMode == distsqlrun.SubqueryExecModeExists {
+		subqueryRecv.noColsRequired = true
+		typ = sqlbase.ColTypeInfoFromColTypes([]sqlbase.ColumnType{})
+	} else {
+		// Apply the PlanToStreamColMap projection to the ResultTypes to get the
+		// final set of output types for the subquery. The reason this is necessary
+		// is that the output schema of a query sometimes contains columns necessary
+		// to merge the streams, but that aren't required by the final output of the
+		// query. These get projected out, so we need to similarly adjust the
+		// expected result types of the subquery here.
+		colTypes := make([]sqlbase.ColumnType, len(subqueryPhysPlan.PlanToStreamColMap))
+		for i, resIdx := range subqueryPhysPlan.PlanToStreamColMap {
+			colTypes[i] = subqueryPhysPlan.ResultTypes[resIdx]
+		}
+		typ = sqlbase.ColTypeInfoFromColTypes(colTypes)
+	}
+	rows = rowcontainer.NewRowContainer(subqueryMemAccount, typ, 0)
+	defer rows.Close(ctx)
+
+	subqueryRowReceiver := NewRowResultWriter(rows)
+	subqueryRecv.resultWriter = subqueryRowReceiver
+	subqueryPlans[planIdx].started = true
+	dsp.Run(subqueryPlanCtx, planner.txn, &subqueryPhysPlan, subqueryRecv, evalCtx, nil /* finishedSetupFn */)
+	if subqueryRecv.commErr != nil {
+		return subqueryRecv.commErr
+	}
+	if err := subqueryRowReceiver.Err(); err != nil {
+		return err
+	}
+	switch subqueryPlan.execMode {
+	case distsqlrun.SubqueryExecModeExists:
+		// For EXISTS expressions, all we want to know if there is at least one row.
+		hasRows := rows.Len() != 0
+		subqueryPlans[planIdx].result = tree.MakeDBool(tree.DBool(hasRows))
+	case distsqlrun.SubqueryExecModeAllRows, distsqlrun.SubqueryExecModeAllRowsNormalized:
+		var result tree.DTuple
+		for rows.Len() > 0 {
+			row := rows.At(0)
+			rows.PopFirst()
+			if row.Len() == 1 {
+				// This seems hokey, but if we don't do this then the subquery expands
+				// to a tuple of tuples instead of a tuple of values and an expression
+				// like "k IN (SELECT foo FROM bar)" will fail because we're comparing
+				// a single value against a tuple.
+				result.D = append(result.D, row[0])
+			} else {
+				result.D = append(result.D, &tree.DTuple{D: row})
+			}
+		}
+
+		if subqueryPlan.execMode == distsqlrun.SubqueryExecModeAllRowsNormalized {
+			result.Normalize(&evalCtx.EvalContext)
+		}
+		subqueryPlans[planIdx].result = &result
+	case distsqlrun.SubqueryExecModeOneRow:
+		switch rows.Len() {
+		case 0:
+			subqueryPlans[planIdx].result = tree.DNull
+		case 1:
+			row := rows.At(0)
+			switch row.Len() {
+			case 1:
+				subqueryPlans[planIdx].result = row[0]
+			default:
+				subqueryPlans[planIdx].result = &tree.DTuple{D: rows.At(0)}
+			}
+		default:
+			return pgerror.NewErrorf(pgerror.CodeCardinalityViolationError,
+				"more than one row returned by a subquery used as an expression")
+		}
+	default:
+		return fmt.Errorf("unexpected subqueryExecMode: %d", subqueryPlan.execMode)
+	}
+	return nil
+}
+
 // PlanAndRun generates a physical plan from a planNode tree and executes it. It
 // assumes that the tree is supported (see CheckSupport).
 //
