@@ -23,6 +23,7 @@ import (
 	"math"
 
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/distsqlrun"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
@@ -38,7 +39,7 @@ type Processor struct {
 	// Spec for the processor; note that the StreamEndpointSpecs in the input
 	// synchronizers and output routers are not set until the end of the planning
 	// process.
-	Spec distsqlrun.ProcessorSpec
+	Spec distsqlpb.ProcessorSpec
 }
 
 // ProcessorIdx identifies a processor by its index in PhysicalPlan.Processors.
@@ -114,7 +115,7 @@ type PhysicalPlan struct {
 	// an ordering sometimes requires to add columns to streams for the sole
 	// reason of correctly merging the streams later (see AddProjection); we don't
 	// want to pay this cost if we don't have multiple streams to merge.
-	MergeOrdering distsqlrun.Ordering
+	MergeOrdering distsqlpb.Ordering
 
 	// Used internally for numbering stages.
 	stageCounter int32
@@ -139,11 +140,11 @@ func (p *PhysicalPlan) AddProcessor(proc Processor) ProcessorIdx {
 }
 
 // SetMergeOrdering sets p.MergeOrdering.
-func (p *PhysicalPlan) SetMergeOrdering(o distsqlrun.Ordering) {
+func (p *PhysicalPlan) SetMergeOrdering(o distsqlpb.Ordering) {
 	if len(p.ResultRouters) > 1 {
 		p.MergeOrdering = o
 	} else {
-		p.MergeOrdering = distsqlrun.Ordering{}
+		p.MergeOrdering = distsqlpb.Ordering{}
 	}
 }
 
@@ -152,13 +153,13 @@ func (p *PhysicalPlan) SetMergeOrdering(o distsqlrun.Ordering) {
 // stages that correspond to logical blocks that don't require any grouping
 // (e.g. evaluator, sorting, etc).
 func (p *PhysicalPlan) AddNoGroupingStage(
-	core distsqlrun.ProcessorCoreUnion,
-	post distsqlrun.PostProcessSpec,
+	core distsqlpb.ProcessorCoreUnion,
+	post distsqlpb.PostProcessSpec,
 	outputTypes []sqlbase.ColumnType,
-	newOrdering distsqlrun.Ordering,
+	newOrdering distsqlpb.Ordering,
 ) {
 	p.AddNoGroupingStageWithCoreFunc(
-		func(_ int, _ *Processor) distsqlrun.ProcessorCoreUnion { return core },
+		func(_ int, _ *Processor) distsqlpb.ProcessorCoreUnion { return core },
 		post,
 		outputTypes,
 		newOrdering,
@@ -168,10 +169,10 @@ func (p *PhysicalPlan) AddNoGroupingStage(
 // AddNoGroupingStageWithCoreFunc is like AddNoGroupingStage, but creates a core
 // spec based on the input processor's spec.
 func (p *PhysicalPlan) AddNoGroupingStageWithCoreFunc(
-	coreFunc func(int, *Processor) distsqlrun.ProcessorCoreUnion,
-	post distsqlrun.PostProcessSpec,
+	coreFunc func(int, *Processor) distsqlpb.ProcessorCoreUnion,
+	post distsqlpb.PostProcessSpec,
 	outputTypes []sqlbase.ColumnType,
-	newOrdering distsqlrun.Ordering,
+	newOrdering distsqlpb.Ordering,
 ) {
 	stageID := p.NewStageID()
 	for i, resultProc := range p.ResultRouters {
@@ -179,15 +180,15 @@ func (p *PhysicalPlan) AddNoGroupingStageWithCoreFunc(
 
 		proc := Processor{
 			Node: prevProc.Node,
-			Spec: distsqlrun.ProcessorSpec{
-				Input: []distsqlrun.InputSyncSpec{{
-					Type:        distsqlrun.InputSyncSpec_UNORDERED,
+			Spec: distsqlpb.ProcessorSpec{
+				Input: []distsqlpb.InputSyncSpec{{
+					Type:        distsqlpb.InputSyncSpec_UNORDERED,
 					ColumnTypes: p.ResultTypes,
 				}},
 				Core: coreFunc(int(resultProc), prevProc),
 				Post: post,
-				Output: []distsqlrun.OutputRouterSpec{{
-					Type: distsqlrun.OutputRouterSpec_PASS_THROUGH,
+				Output: []distsqlpb.OutputRouterSpec{{
+					Type: distsqlpb.OutputRouterSpec_PASS_THROUGH,
 				}},
 				StageID: stageID,
 			},
@@ -213,15 +214,15 @@ func (p *PhysicalPlan) AddNoGroupingStageWithCoreFunc(
 func (p *PhysicalPlan) MergeResultStreams(
 	resultRouters []ProcessorIdx,
 	sourceRouterSlot int,
-	ordering distsqlrun.Ordering,
+	ordering distsqlpb.Ordering,
 	destProcessor ProcessorIdx,
 	destInput int,
 ) {
 	proc := &p.Processors[destProcessor]
 	if len(ordering.Columns) == 0 || len(resultRouters) == 1 {
-		proc.Spec.Input[destInput].Type = distsqlrun.InputSyncSpec_UNORDERED
+		proc.Spec.Input[destInput].Type = distsqlpb.InputSyncSpec_UNORDERED
 	} else {
-		proc.Spec.Input[destInput].Type = distsqlrun.InputSyncSpec_ORDERED
+		proc.Spec.Input[destInput].Type = distsqlpb.InputSyncSpec_ORDERED
 		proc.Spec.Input[destInput].Ordering = ordering
 	}
 
@@ -240,21 +241,21 @@ func (p *PhysicalPlan) MergeResultStreams(
 // previous stage (ResultRouters) are all connected to this processor.
 func (p *PhysicalPlan) AddSingleGroupStage(
 	nodeID roachpb.NodeID,
-	core distsqlrun.ProcessorCoreUnion,
-	post distsqlrun.PostProcessSpec,
+	core distsqlpb.ProcessorCoreUnion,
+	post distsqlpb.PostProcessSpec,
 	outputTypes []sqlbase.ColumnType,
 ) {
 	proc := Processor{
 		Node: nodeID,
-		Spec: distsqlrun.ProcessorSpec{
-			Input: []distsqlrun.InputSyncSpec{{
+		Spec: distsqlpb.ProcessorSpec{
+			Input: []distsqlpb.InputSyncSpec{{
 				// The other fields will be filled in by mergeResultStreams.
 				ColumnTypes: p.ResultTypes,
 			}},
 			Core: core,
 			Post: post,
-			Output: []distsqlrun.OutputRouterSpec{{
-				Type: distsqlrun.OutputRouterSpec_PASS_THROUGH,
+			Output: []distsqlpb.OutputRouterSpec{{
+				Type: distsqlpb.OutputRouterSpec_PASS_THROUGH,
 			}},
 			StageID: p.NewStageID(),
 		},
@@ -270,7 +271,7 @@ func (p *PhysicalPlan) AddSingleGroupStage(
 	p.ResultRouters[0] = pIdx
 
 	p.ResultTypes = outputTypes
-	p.MergeOrdering = distsqlrun.Ordering{}
+	p.MergeOrdering = distsqlpb.Ordering{}
 }
 
 // CheckLastStagePost checks that the processors of the last stage of the
@@ -282,8 +283,10 @@ func (p *PhysicalPlan) CheckLastStagePost() error {
 	// verify this assumption.
 	for i := 1; i < len(p.ResultRouters); i++ {
 		pi := &p.Processors[p.ResultRouters[i]].Spec.Post
-		if pi.Filter != post.Filter || pi.Projection != post.Projection ||
-			len(pi.OutputColumns) != len(post.OutputColumns) {
+		if pi.Filter != post.Filter ||
+			pi.Projection != post.Projection ||
+			len(pi.OutputColumns) != len(post.OutputColumns) ||
+			len(pi.RenderExprs) != len(post.RenderExprs) {
 			return errors.Errorf("inconsistent post-processing: %v vs %v", post, pi)
 		}
 		for j, col := range pi.OutputColumns {
@@ -291,6 +294,11 @@ func (p *PhysicalPlan) CheckLastStagePost() error {
 				return errors.Errorf("inconsistent post-processing: %v vs %v", post, pi)
 			}
 		}
+		for j, col := range pi.RenderExprs {
+			if col != post.RenderExprs[j] {
+				return errors.Errorf("inconsistent post-processing: %v vs %v", post, pi)
+			}
+		}
 	}
 
 	return nil
@@ -298,7 +306,7 @@ func (p *PhysicalPlan) CheckLastStagePost() error {
 
 // GetLastStagePost returns the PostProcessSpec for the processors in the last
 // stage (ResultRouters).
-func (p *PhysicalPlan) GetLastStagePost() distsqlrun.PostProcessSpec {
+func (p *PhysicalPlan) GetLastStagePost() distsqlpb.PostProcessSpec {
 	if err := p.CheckLastStagePost(); err != nil {
 		panic(err)
 	}
@@ -309,7 +317,7 @@ func (p *PhysicalPlan) GetLastStagePost() distsqlrun.PostProcessSpec {
 // stage (ResultRouters).
 // The caller must update the ordering via SetOrdering.
 func (p *PhysicalPlan) SetLastStagePost(
-	post distsqlrun.PostProcessSpec, outputTypes []sqlbase.ColumnType,
+	post distsqlpb.PostProcessSpec, outputTypes []sqlbase.ColumnType,
 ) {
 	for _, pIdx := range p.ResultRouters {
 		p.Processors[pIdx].Spec.Post = post
@@ -347,7 +355,7 @@ func (p *PhysicalPlan) AddProjection(columns []uint32) {
 
 	// Update the ordering.
 	if len(p.MergeOrdering.Columns) > 0 {
-		newOrdering := make([]distsqlrun.Ordering_Column, len(p.MergeOrdering.Columns))
+		newOrdering := make([]distsqlpb.Ordering_Column, len(p.MergeOrdering.Columns))
 		for i, c := range p.MergeOrdering.Columns {
 			// Look for the column in the new projection.
 			found := -1
@@ -380,7 +388,7 @@ func (p *PhysicalPlan) AddProjection(columns []uint32) {
 		// only the renders needed by the new output columns, and reorder them
 		// accordingly.
 		oldRenders := post.RenderExprs
-		post.RenderExprs = make([]distsqlrun.Expression, len(columns))
+		post.RenderExprs = make([]distsqlpb.Expression, len(columns))
 		for i, c := range columns {
 			post.RenderExprs[i] = oldRenders[c]
 		}
@@ -421,10 +429,7 @@ func exprColumn(expr tree.TypedExpr, indexVarMap []int) (int, bool) {
 //
 // See MakeExpression for a description of indexVarMap.
 func (p *PhysicalPlan) AddRendering(
-	exprs []tree.TypedExpr,
-	evalCtx *tree.EvalContext,
-	indexVarMap []int,
-	outTypes []sqlbase.ColumnType,
+	exprs []tree.TypedExpr, exprCtx ExprContext, indexVarMap []int, outTypes []sqlbase.ColumnType,
 ) error {
 	// First check if we need an Evaluator, or we are just shuffling values. We
 	// also check if the rendering is a no-op ("identity").
@@ -461,12 +466,12 @@ func (p *PhysicalPlan) AddRendering(
 
 	post := p.GetLastStagePost()
 	if len(post.RenderExprs) > 0 {
-		post = distsqlrun.PostProcessSpec{}
+		post = distsqlpb.PostProcessSpec{}
 		// The last stage contains render expressions. The new renders refer to
 		// the output of these, so we need to add another "no-op" stage to which
 		// to attach the new rendering.
 		p.AddNoGroupingStage(
-			distsqlrun.ProcessorCoreUnion{Noop: &distsqlrun.NoopCoreSpec{}},
+			distsqlpb.ProcessorCoreUnion{Noop: &distsqlpb.NoopCoreSpec{}},
 			post,
 			p.ResultTypes,
 			p.MergeOrdering,
@@ -477,10 +482,10 @@ func (p *PhysicalPlan) AddRendering(
 	if post.Projection {
 		compositeMap = reverseProjection(post.OutputColumns, indexVarMap)
 	}
-	post.RenderExprs = make([]distsqlrun.Expression, len(exprs))
+	post.RenderExprs = make([]distsqlpb.Expression, len(exprs))
 	for i, e := range exprs {
 		var err error
-		post.RenderExprs[i], err = MakeExpression(e, evalCtx, compositeMap)
+		post.RenderExprs[i], err = MakeExpression(e, exprCtx, compositeMap)
 		if err != nil {
 			return err
 		}
@@ -488,7 +493,7 @@ func (p *PhysicalPlan) AddRendering(
 
 	if len(p.MergeOrdering.Columns) > 0 {
 		outTypes = outTypes[:len(outTypes):len(outTypes)]
-		newOrdering := make([]distsqlrun.Ordering_Column, len(p.MergeOrdering.Columns))
+		newOrdering := make([]distsqlpb.Ordering_Column, len(p.MergeOrdering.Columns))
 		for i, c := range p.MergeOrdering.Columns {
 			found := -1
 			// Look for the column in the new projection.
@@ -507,7 +512,10 @@ func (p *PhysicalPlan) AddRendering(
 				if post.Projection {
 					internalColIdx = post.OutputColumns[internalColIdx]
 				}
-				newExpr, err := MakeExpression(&tree.IndexedVar{Idx: int(internalColIdx)}, evalCtx, nil)
+				newExpr, err := MakeExpression(tree.NewTypedOrdinalReference(
+					int(internalColIdx),
+					p.ResultTypes[c.ColIdx].ToDatumType()),
+					exprCtx, nil /* indexVarMap */)
 				if err != nil {
 					return err
 				}
@@ -591,7 +599,7 @@ func reverseProjection(outputColumns []uint32, indexVarMap []int) []int {
 //
 // See MakeExpression for a description of indexVarMap.
 func (p *PhysicalPlan) AddFilter(
-	expr tree.TypedExpr, evalCtx *tree.EvalContext, indexVarMap []int,
+	expr tree.TypedExpr, exprCtx ExprContext, indexVarMap []int,
 ) error {
 	post := p.GetLastStagePost()
 	if len(post.RenderExprs) > 0 || post.Offset != 0 || post.Limit != 0 {
@@ -602,9 +610,9 @@ func (p *PhysicalPlan) AddFilter(
 		// In general, we might be able to push the filter "through" the rendering;
 		// but the higher level planning code should figure this out when
 		// propagating filters.
-		post = distsqlrun.PostProcessSpec{}
+		post = distsqlpb.PostProcessSpec{}
 		p.AddNoGroupingStage(
-			distsqlrun.ProcessorCoreUnion{Noop: &distsqlrun.NoopCoreSpec{}},
+			distsqlpb.ProcessorCoreUnion{Noop: &distsqlpb.NoopCoreSpec{}},
 			post,
 			p.ResultTypes,
 			p.MergeOrdering,
@@ -615,12 +623,19 @@ func (p *PhysicalPlan) AddFilter(
 	if post.Projection {
 		compositeMap = reverseProjection(post.OutputColumns, indexVarMap)
 	}
-	filter, err := MakeExpression(expr, evalCtx, compositeMap)
+	filter, err := MakeExpression(expr, exprCtx, compositeMap)
 	if err != nil {
 		return err
 	}
-	if post.Filter.Expr != "" {
-		filter.Expr = fmt.Sprintf("(%s) AND (%s)", post.Filter.Expr, filter.Expr)
+	if !post.Filter.Empty() {
+		if filter.Expr != "" {
+			filter.Expr = fmt.Sprintf("(%s) AND (%s)", post.Filter.Expr, filter.Expr)
+		} else if filter.LocalExpr != nil {
+			filter.LocalExpr = tree.NewTypedAndExpr(
+				post.Filter.LocalExpr,
+				filter.LocalExpr,
+			)
+		}
 	}
 	for _, pIdx := range p.ResultRouters {
 		p.Processors[pIdx].Spec.Post.Filter = filter
@@ -631,8 +646,8 @@ func (p *PhysicalPlan) AddFilter(
 // emptyPlan creates a plan with a single processor that generates no rows; the
 // output stream has the given types.
 func emptyPlan(types []sqlbase.ColumnType, node roachpb.NodeID) PhysicalPlan {
-	s := distsqlrun.ValuesCoreSpec{
-		Columns: make([]distsqlrun.DatumInfo, len(types)),
+	s := distsqlpb.ValuesCoreSpec{
+		Columns: make([]distsqlpb.DatumInfo, len(types)),
 	}
 	for i, t := range types {
 		s.Columns[i].Encoding = sqlbase.DatumEncoding_VALUE
@@ -642,9 +657,9 @@ func emptyPlan(types []sqlbase.ColumnType, node roachpb.NodeID) PhysicalPlan {
 	return PhysicalPlan{
 		Processors: []Processor{{
 			Node: node,
-			Spec: distsqlrun.ProcessorSpec{
-				Core:   distsqlrun.ProcessorCoreUnion{Values: &s},
-				Output: make([]distsqlrun.OutputRouterSpec, 1),
+			Spec: distsqlpb.ProcessorSpec{
+				Core:   distsqlpb.ProcessorCoreUnion{Values: &s},
+				Output: make([]distsqlpb.OutputRouterSpec, 1),
 			},
 		}},
 		ResultRouters: []ProcessorIdx{0},
@@ -658,7 +673,7 @@ func emptyPlan(types []sqlbase.ColumnType, node roachpb.NodeID) PhysicalPlan {
 //
 // For no limit, count should be MaxInt64.
 func (p *PhysicalPlan) AddLimit(
-	count int64, offset int64, evalCtx *tree.EvalContext, node roachpb.NodeID,
+	count int64, offset int64, exprCtx ExprContext, node roachpb.NodeID,
 ) error {
 	if count < 0 {
 		return errors.Errorf("negative limit")
@@ -712,7 +727,7 @@ func (p *PhysicalPlan) AddLimit(
 		}
 		p.SetLastStagePost(post, p.ResultTypes)
 		if limitZero {
-			if err := p.AddFilter(tree.DBoolFalse, evalCtx, nil); err != nil {
+			if err := p.AddFilter(tree.DBoolFalse, exprCtx, nil); err != nil {
 				return err
 			}
 		}
@@ -733,7 +748,7 @@ func (p *PhysicalPlan) AddLimit(
 		}
 	}
 
-	post := distsqlrun.PostProcessSpec{
+	post := distsqlpb.PostProcessSpec{
 		Offset: uint64(offset),
 	}
 	if count != math.MaxInt64 {
@@ -741,12 +756,12 @@ func (p *PhysicalPlan) AddLimit(
 	}
 	p.AddSingleGroupStage(
 		node,
-		distsqlrun.ProcessorCoreUnion{Noop: &distsqlrun.NoopCoreSpec{}},
+		distsqlpb.ProcessorCoreUnion{Noop: &distsqlpb.NoopCoreSpec{}},
 		post,
 		p.ResultTypes,
 	)
 	if limitZero {
-		if err := p.AddFilter(tree.DBoolFalse, evalCtx, nil); err != nil {
+		if err := p.AddFilter(tree.DBoolFalse, exprCtx, nil); err != nil {
 			return err
 		}
 	}
@@ -766,23 +781,18 @@ func (p *PhysicalPlan) PopulateEndpoints(nodeAddresses map[roachpb.NodeID]string
 	for sIdx, s := range p.Streams {
 		p1 := &p.Processors[s.SourceProcessor]
 		p2 := &p.Processors[s.DestProcessor]
-		endpoint := distsqlrun.StreamEndpointSpec{StreamID: distsqlrun.StreamID(sIdx)}
+		endpoint := distsqlpb.StreamEndpointSpec{StreamID: distsqlpb.StreamID(sIdx)}
 		if p1.Node == p2.Node {
-			endpoint.Type = distsqlrun.StreamEndpointSpec_LOCAL
+			endpoint.Type = distsqlpb.StreamEndpointSpec_LOCAL
 		} else {
-			endpoint.Type = distsqlrun.StreamEndpointSpec_REMOTE
+			endpoint.Type = distsqlpb.StreamEndpointSpec_REMOTE
 		}
 		p2.Spec.Input[s.DestInput].Streams = append(p2.Spec.Input[s.DestInput].Streams, endpoint)
-		if endpoint.Type == distsqlrun.StreamEndpointSpec_REMOTE {
+		if endpoint.Type == distsqlpb.StreamEndpointSpec_REMOTE {
 			if !p.remotePlan {
 				p.remotePlan = true
 			}
-			var ok bool
 			endpoint.TargetNodeID = p2.Node
-			endpoint.DeprecatedTargetAddr, ok = nodeAddresses[p2.Node]
-			if !ok {
-				panic(fmt.Sprintf("node %d not in nodeAddresses map", p2.Node))
-			}
 		}
 
 		router := &p1.Spec.Output[0]
@@ -805,23 +815,23 @@ func (p *PhysicalPlan) PopulateEndpoints(nodeAddresses map[roachpb.NodeID]string
 // gateway is the current node's NodeID.
 func (p *PhysicalPlan) GenerateFlowSpecs(
 	gateway roachpb.NodeID,
-) map[roachpb.NodeID]distsqlrun.FlowSpec {
+) map[roachpb.NodeID]*distsqlpb.FlowSpec {
 	// Only generate a flow ID for a remote plan because it will need to be
 	// referenced by remote nodes when connecting streams. This id generation is
 	// skipped for performance reasons on local flows.
-	flowID := distsqlrun.FlowID{}
+	flowID := distsqlpb.FlowID{}
 	if p.remotePlan {
 		flowID.UUID = uuid.MakeV4()
 	}
-	flows := make(map[roachpb.NodeID]distsqlrun.FlowSpec)
+	flows := make(map[roachpb.NodeID]*distsqlpb.FlowSpec, 1)
 
 	for _, proc := range p.Processors {
 		flowSpec, ok := flows[proc.Node]
 		if !ok {
-			flowSpec = distsqlrun.FlowSpec{FlowID: flowID, Gateway: gateway}
+			flowSpec = NewFlowSpec(flowID, gateway)
+			flows[proc.Node] = flowSpec
 		}
 		flowSpec.Processors = append(flowSpec.Processors, proc.Spec)
-		flows[proc.Node] = flowSpec
 	}
 	return flows
 }
@@ -895,11 +905,13 @@ func MergeResultTypes(left, right []sqlbase.ColumnType) ([]sqlbase.ColumnType, e
 
 // equivalentType checks whether a column type is equivalent to
 // another for the purpose of UNION. This excludes its VisibleType
-// type alias, which doesn't effect the merging of values.
+// type alias, which doesn't effect the merging of values. There is
+// also special handling for "naked" int types of no defined size.
 func equivalentTypes(c, other *sqlbase.ColumnType) bool {
-	// Convert pre-2.1 INTs to INT8.
+	// Convert pre-2.1 and pre-2.2 INTs to INT8.
 	lhs := *c
 	if lhs.SemanticType == sqlbase.ColumnType_INT {
+		// Pre-2.2 INT without size was assigned width 0.
 		// Pre-2.1 BIT was assigned arbitrary width, and is mapped to INT8 post-2.1. See #34161.
 		if lhs.Width != 64 && lhs.Width != 32 && lhs.Width != 16 {
 			lhs.Width = 64
@@ -914,38 +926,35 @@ func equivalentTypes(c, other *sqlbase.ColumnType) bool {
 		}
 	}
 	rhs.VisibleType = c.VisibleType
-	return rhs.Equal(lhs)
+	return lhs.Equal(rhs)
 }
 
 // AddJoinStage adds join processors at each of the specified nodes, and wires
 // the left and right-side outputs to these processors.
 func (p *PhysicalPlan) AddJoinStage(
 	nodes []roachpb.NodeID,
-	core distsqlrun.ProcessorCoreUnion,
-	post distsqlrun.PostProcessSpec,
+	core distsqlpb.ProcessorCoreUnion,
+	post distsqlpb.PostProcessSpec,
 	leftEqCols, rightEqCols []uint32,
 	leftTypes, rightTypes []sqlbase.ColumnType,
-	leftMergeOrd, rightMergeOrd distsqlrun.Ordering,
+	leftMergeOrd, rightMergeOrd distsqlpb.Ordering,
 	leftRouters, rightRouters []ProcessorIdx,
-	includeRight bool,
 ) {
 	pIdxStart := ProcessorIdx(len(p.Processors))
 	stageID := p.NewStageID()
 
 	for _, n := range nodes {
-		inputs := make([]distsqlrun.InputSyncSpec, 0, 2)
-		inputs = append(inputs, distsqlrun.InputSyncSpec{ColumnTypes: leftTypes})
-		if includeRight {
-			inputs = append(inputs, distsqlrun.InputSyncSpec{ColumnTypes: rightTypes})
-		}
+		inputs := make([]distsqlpb.InputSyncSpec, 0, 2)
+		inputs = append(inputs, distsqlpb.InputSyncSpec{ColumnTypes: leftTypes})
+		inputs = append(inputs, distsqlpb.InputSyncSpec{ColumnTypes: rightTypes})
 
 		proc := Processor{
 			Node: n,
-			Spec: distsqlrun.ProcessorSpec{
+			Spec: distsqlpb.ProcessorSpec{
 				Input:   inputs,
 				Core:    core,
 				Post:    post,
-				Output:  []distsqlrun.OutputRouterSpec{{Type: distsqlrun.OutputRouterSpec_PASS_THROUGH}},
+				Output:  []distsqlpb.OutputRouterSpec{{Type: distsqlpb.OutputRouterSpec_PASS_THROUGH}},
 				StageID: stageID,
 			},
 		}
@@ -956,22 +965,17 @@ func (p *PhysicalPlan) AddJoinStage(
 		// Parallel hash or merge join: we distribute rows (by hash of
 		// equality columns) to len(nodes) join processors.
 
-		routerType := distsqlrun.OutputRouterSpec_BY_HASH
-		if !includeRight {
-			routerType = distsqlrun.OutputRouterSpec_PASS_THROUGH
-		}
-
 		// Set up the left routers.
 		for _, resultProc := range leftRouters {
-			p.Processors[resultProc].Spec.Output[0] = distsqlrun.OutputRouterSpec{
-				Type:        routerType,
+			p.Processors[resultProc].Spec.Output[0] = distsqlpb.OutputRouterSpec{
+				Type:        distsqlpb.OutputRouterSpec_BY_HASH,
 				HashColumns: leftEqCols,
 			}
 		}
 		// Set up the right routers.
 		for _, resultProc := range rightRouters {
-			p.Processors[resultProc].Spec.Output[0] = distsqlrun.OutputRouterSpec{
-				Type:        routerType,
+			p.Processors[resultProc].Spec.Output[0] = distsqlpb.OutputRouterSpec{
+				Type:        distsqlpb.OutputRouterSpec_BY_HASH,
 				HashColumns: rightEqCols,
 			}
 		}
@@ -983,15 +987,11 @@ func (p *PhysicalPlan) AddJoinStage(
 	for bucket := 0; bucket < len(nodes); bucket++ {
 		pIdx := pIdxStart + ProcessorIdx(bucket)
 
-		if includeRight {
-			// Connect left routers to the processor's first input. Currently the join
-			// node doesn't care about the orderings of the left and right results.
-			p.MergeResultStreams(leftRouters, bucket, leftMergeOrd, pIdx, 0)
-			// Connect right routers to the processor's second input if it has one.
-			p.MergeResultStreams(rightRouters, bucket, rightMergeOrd, pIdx, 1)
-		} else {
-			p.MergeResultStreams(leftRouters[bucket:bucket+1], 0, leftMergeOrd, pIdx, 0)
-		}
+		// Connect left routers to the processor's first input. Currently the join
+		// node doesn't care about the orderings of the left and right results.
+		p.MergeResultStreams(leftRouters, bucket, leftMergeOrd, pIdx, 0)
+		// Connect right routers to the processor's second input if it has one.
+		p.MergeResultStreams(rightRouters, bucket, rightMergeOrd, pIdx, 1)
 
 		p.ResultRouters = append(p.ResultRouters, pIdx)
 	}
@@ -1004,17 +1004,17 @@ func (p *PhysicalPlan) AddJoinStage(
 // can elide the distinct stage on that side.
 func (p *PhysicalPlan) AddDistinctSetOpStage(
 	nodes []roachpb.NodeID,
-	joinCore distsqlrun.ProcessorCoreUnion,
-	distinctCores []distsqlrun.ProcessorCoreUnion,
-	post distsqlrun.PostProcessSpec,
+	joinCore distsqlpb.ProcessorCoreUnion,
+	distinctCores []distsqlpb.ProcessorCoreUnion,
+	post distsqlpb.PostProcessSpec,
 	eqCols []uint32,
 	leftTypes, rightTypes []sqlbase.ColumnType,
-	leftMergeOrd, rightMergeOrd distsqlrun.Ordering,
+	leftMergeOrd, rightMergeOrd distsqlpb.Ordering,
 	leftRouters, rightRouters []ProcessorIdx,
 ) {
 	const numSides = 2
 	inputResultTypes := [numSides][]sqlbase.ColumnType{leftTypes, rightTypes}
-	inputMergeOrderings := [numSides]distsqlrun.Ordering{leftMergeOrd, rightMergeOrd}
+	inputMergeOrderings := [numSides]distsqlpb.Ordering{leftMergeOrd, rightMergeOrd}
 	inputResultRouters := [numSides][]ProcessorIdx{leftRouters, rightRouters}
 
 	// Create distinct stages for the left and right sides, where left and right
@@ -1031,13 +1031,13 @@ func (p *PhysicalPlan) AddDistinctSetOpStage(
 		for _, n := range nodes {
 			proc := Processor{
 				Node: n,
-				Spec: distsqlrun.ProcessorSpec{
-					Input: []distsqlrun.InputSyncSpec{
+				Spec: distsqlpb.ProcessorSpec{
+					Input: []distsqlpb.InputSyncSpec{
 						{ColumnTypes: types},
 					},
 					Core:    distinctCores[side],
-					Post:    distsqlrun.PostProcessSpec{},
-					Output:  []distsqlrun.OutputRouterSpec{{Type: distsqlrun.OutputRouterSpec_PASS_THROUGH}},
+					Post:    distsqlpb.PostProcessSpec{},
+					Output:  []distsqlpb.OutputRouterSpec{{Type: distsqlpb.OutputRouterSpec_PASS_THROUGH}},
 					StageID: distinctStageID,
 				},
 			}
@@ -1049,15 +1049,15 @@ func (p *PhysicalPlan) AddDistinctSetOpStage(
 	if len(nodes) > 1 {
 		// Set up the left routers.
 		for _, resultProc := range leftRouters {
-			p.Processors[resultProc].Spec.Output[0] = distsqlrun.OutputRouterSpec{
-				Type:        distsqlrun.OutputRouterSpec_BY_HASH,
+			p.Processors[resultProc].Spec.Output[0] = distsqlpb.OutputRouterSpec{
+				Type:        distsqlpb.OutputRouterSpec_BY_HASH,
 				HashColumns: eqCols,
 			}
 		}
 		// Set up the right routers.
 		for _, resultProc := range rightRouters {
-			p.Processors[resultProc].Spec.Output[0] = distsqlrun.OutputRouterSpec{
-				Type:        distsqlrun.OutputRouterSpec_BY_HASH,
+			p.Processors[resultProc].Spec.Output[0] = distsqlpb.OutputRouterSpec{
+				Type:        distsqlpb.OutputRouterSpec_BY_HASH,
 				HashColumns: eqCols,
 			}
 		}
@@ -1081,14 +1081,14 @@ func (p *PhysicalPlan) AddDistinctSetOpStage(
 	for _, n := range nodes {
 		proc := Processor{
 			Node: n,
-			Spec: distsqlrun.ProcessorSpec{
-				Input: []distsqlrun.InputSyncSpec{
+			Spec: distsqlpb.ProcessorSpec{
+				Input: []distsqlpb.InputSyncSpec{
 					{ColumnTypes: leftTypes},
 					{ColumnTypes: rightTypes},
 				},
 				Core:    joinCore,
 				Post:    post,
-				Output:  []distsqlrun.OutputRouterSpec{{Type: distsqlrun.OutputRouterSpec_PASS_THROUGH}},
+				Output:  []distsqlpb.OutputRouterSpec{{Type: distsqlpb.OutputRouterSpec_PASS_THROUGH}},
 				StageID: joinStageID,
 			},
 		}
