@@ -16,12 +16,11 @@ package distsqlrun
 
 import (
 	"context"
+	"fmt"
 	"strings"
 	"unsafe"
 
-	"fmt"
-
-	"github.com/cockroachdb/cockroach/pkg/sql/parser"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/builtins"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
@@ -38,13 +37,13 @@ import (
 // GetAggregateInfo returns the aggregate constructor and the return type for
 // the given aggregate function when applied on the given type.
 func GetAggregateInfo(
-	fn AggregatorSpec_Func, inputTypes ...sqlbase.ColumnType,
+	fn distsqlpb.AggregatorSpec_Func, inputTypes ...sqlbase.ColumnType,
 ) (
 	aggregateConstructor func(*tree.EvalContext, tree.Datums) tree.AggregateFunc,
 	returnType sqlbase.ColumnType,
 	err error,
 ) {
-	if fn == AggregatorSpec_ANY_NOT_NULL {
+	if fn == distsqlpb.AggregatorSpec_ANY_NOT_NULL {
 		// The ANY_NOT_NULL builtin does not have a fixed return type;
 		// handle it separately.
 		if len(inputTypes) != 1 {
@@ -130,7 +129,7 @@ type aggregatorBase struct {
 	isScalar         bool
 	groupCols        columns
 	orderedGroupCols columns
-	aggregations     []AggregatorSpec_Aggregation
+	aggregations     []distsqlpb.AggregatorSpec_Aggregation
 
 	lastOrdGroupCols sqlbase.EncDatumRow
 	arena            stringarena.Arena
@@ -148,9 +147,9 @@ func (ag *aggregatorBase) init(
 	self RowSource,
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *AggregatorSpec,
+	spec *distsqlpb.AggregatorSpec,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	output RowReceiver,
 	trailingMetaCallback func(context.Context) []ProducerMetadata,
 ) error {
@@ -162,9 +161,9 @@ func (ag *aggregatorBase) init(
 	}
 	ag.input = input
 	switch spec.Type {
-	case AggregatorSpec_SCALAR:
+	case distsqlpb.AggregatorSpec_SCALAR:
 		ag.isScalar = true
-	case AggregatorSpec_NON_SCALAR:
+	case distsqlpb.AggregatorSpec_NON_SCALAR:
 		ag.isScalar = false
 	default:
 		// This case exists for backward compatibility.
@@ -208,22 +207,20 @@ func (ag *aggregatorBase) init(
 
 		arguments := make(tree.Datums, len(aggInfo.Arguments))
 		for j, argument := range aggInfo.Arguments {
-			expr, err := parser.ParseExpr(argument.Expr)
-			if err != nil {
-				return err
+			h := exprHelper{}
+			// Pass nil types and row - there are no variables in these expressions.
+			if err := h.init(argument, nil /* types */, flowCtx.EvalCtx); err != nil {
+				return errors.Wrap(err, argument.String())
 			}
-			typedExpr, err := tree.TypeCheck(expr, &tree.SemaContext{}, types.Any)
+			d, err := h.eval(nil /* row */)
 			if err != nil {
-				return errors.Wrap(err, expr.String())
+				return errors.Wrap(err, argument.String())
 			}
-			argTypes[len(aggInfo.ColIdx)+j], err = sqlbase.DatumTypeToColumnType(typedExpr.ResolvedType())
+			argTypes[len(aggInfo.ColIdx)+j], err = sqlbase.DatumTypeToColumnType(d.ResolvedType())
 			if err != nil {
-				return errors.Wrap(err, expr.String())
-			}
-			arguments[j], err = typedExpr.Eval(ag.evalCtx)
-			if err != nil {
-				return errors.Wrap(err, expr.String())
+				return errors.Wrap(err, argument.String())
 			}
+			arguments[j] = d
 		}
 
 		aggConstructor, retType, err := GetAggregateInfo(aggInfo.Func, argTypes...)
@@ -248,7 +245,7 @@ func (ag *aggregatorBase) init(
 	)
 }
 
-var _ DistSQLSpanStats = &AggregatorStats{}
+var _ distsqlpb.DistSQLSpanStats = &AggregatorStats{}
 
 const aggregatorTagPrefix = "aggregator."
 
@@ -330,11 +327,18 @@ const (
 func newAggregator(
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *AggregatorSpec,
+	spec *distsqlpb.AggregatorSpec,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	output RowReceiver,
 ) (Processor, error) {
+	if len(spec.GroupCols) == 0 &&
+		len(spec.Aggregations) == 1 &&
+		spec.Aggregations[0].FilterColIdx == nil &&
+		spec.Aggregations[0].Func == distsqlpb.AggregatorSpec_COUNT_ROWS &&
+		!spec.Aggregations[0].Distinct {
+		return newCountAggregator(flowCtx, processorID, input, post, output)
+	}
 	if len(spec.OrderedGroupCols) == len(spec.GroupCols) {
 		return newOrderedAggregator(flowCtx, processorID, spec, input, post, output)
 	}
@@ -363,9 +367,9 @@ func newAggregator(
 func newOrderedAggregator(
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *AggregatorSpec,
+	spec *distsqlpb.AggregatorSpec,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	output RowReceiver,
 ) (*orderedAggregator, error) {
 	ag := &orderedAggregator{}
@@ -457,7 +461,11 @@ func (ag *aggregatorBase) matchLastOrdGroupCols(row sqlbase.EncDatumRow) (bool,
 // into intermediary aggregate results. If it encounters metadata, the metadata
 // is immediately returned. Subsequent calls of this function will resume row
 // accumulation.
-func (ag *hashAggregator) accumulateRows() (aggregatorState, sqlbase.EncDatumRow, *ProducerMetadata) {
+func (ag *hashAggregator) accumulateRows() (
+	aggregatorState,
+	sqlbase.EncDatumRow,
+	*ProducerMetadata,
+) {
 	for {
 		row, meta := ag.input.Next()
 		if meta != nil {
@@ -516,7 +524,11 @@ func (ag *hashAggregator) accumulateRows() (aggregatorState, sqlbase.EncDatumRow
 // into intermediary aggregate results. If it encounters metadata, the metadata
 // is immediately returned. Subsequent calls of this function will resume row
 // accumulation.
-func (ag *orderedAggregator) accumulateRows() (aggregatorState, sqlbase.EncDatumRow, *ProducerMetadata) {
+func (ag *orderedAggregator) accumulateRows() (
+	aggregatorState,
+	sqlbase.EncDatumRow,
+	*ProducerMetadata,
+) {
 	for {
 		row, meta := ag.input.Next()
 		if meta != nil {
@@ -721,11 +733,6 @@ func (ag *orderedAggregator) Next() (sqlbase.EncDatumRow, *ProducerMetadata) {
 	return nil, ag.DrainHelper()
 }
 
-// ConsumerDone is part of the RowSource interface.
-func (ag *aggregatorBase) ConsumerDone() {
-	ag.MoveToDraining(nil /* err */)
-}
-
 // ConsumerClosed is part of the RowSource interface.
 func (ag *hashAggregator) ConsumerClosed() {
 	// The consumer is done, Next() will not be called again.
@@ -840,11 +847,15 @@ func (ag *orderedAggregator) accumulateRow(row sqlbase.EncDatumRow) error {
 }
 
 type aggregateFuncHolder struct {
-	create    func(*tree.EvalContext, tree.Datums) tree.AggregateFunc
+	create func(*tree.EvalContext, tree.Datums) tree.AggregateFunc
+
+	// arguments is the list of constant (non-aggregated) arguments to the
+	// aggregate, for instance, the separator in string_agg.
 	arguments tree.Datums
-	group     *aggregatorBase
-	seen      map[string]struct{}
-	arena     *stringarena.Arena
+
+	group *aggregatorBase
+	seen  map[string]struct{}
+	arena *stringarena.Arena
 }
 
 const sizeOfAggregateFunc = int64(unsafe.Sizeof(tree.AggregateFunc(nil)))
