@@ -16,20 +16,18 @@ package distsqlrun
 
 import (
 	"context"
-	"fmt"
 	"sync"
 	"sync/atomic"
 
-	opentracing "github.com/opentracing/opentracing-go"
-
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
-	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/opentracing/opentracing-go"
 )
 
 const rowChannelBufSize = 16
@@ -80,6 +78,10 @@ type RowReceiver interface {
 	// Implementations of Push() must be thread-safe.
 	Push(row sqlbase.EncDatumRow, meta *ProducerMetadata) ConsumerStatus
 
+	// Types returns the types of the EncDatumRow that this RowReceiver expects
+	// to be pushed.
+	Types() []sqlbase.ColumnType
+
 	// ProducerDone is called when the producer has pushed all the rows and
 	// metadata; it causes the RowReceiver to process all rows and clean up.
 	//
@@ -160,7 +162,7 @@ type RowSource interface {
 // RowSourcedProcessor is the union of RowSource and Processor.
 type RowSourcedProcessor interface {
 	RowSource
-	Run(_ context.Context, wg *sync.WaitGroup)
+	Run(context.Context)
 }
 
 // Run reads records from the source and outputs them to the receiver, properly
@@ -374,7 +376,10 @@ type ProducerMetadata struct {
 	// RowNum corresponds to a row produced by a "source" processor that takes no
 	// inputs. It is used in tests to verify that all metadata is forwarded
 	// exactly once to the receiver on the gateway node.
-	RowNum *RemoteProducerMetadata_RowNum
+	RowNum *distsqlpb.RemoteProducerMetadata_RowNum
+	// SamplerProgress contains incremental progress information from the sampler
+	// processor.
+	SamplerProgress *distsqlpb.RemoteProducerMetadata_SamplerProgress
 }
 
 // RowChannel is a thin layer over a RowChannelMsg channel, which can be used to
@@ -486,6 +491,11 @@ func (rc *RowChannel) ConsumerClosed() {
 	}
 }
 
+// Types is part of the RowReceiver interface.
+func (rc *RowChannel) Types() []sqlbase.ColumnType {
+	return rc.types
+}
+
 // BufferedRecord represents a row or metadata record that has been buffered
 // inside a RowBuffer.
 type BufferedRecord struct {
@@ -612,6 +622,11 @@ func (rb *RowBuffer) ProducerDone() {
 	rb.mu.producerClosed = true
 }
 
+// Types is part of the RowReceiver interface.
+func (rb *RowBuffer) Types() []sqlbase.ColumnType {
+	return rb.types
+}
+
 // OutputTypes is part of the RowSource interface.
 func (rb *RowBuffer) OutputTypes() []sqlbase.ColumnType {
 	if rb.types == nil {
@@ -676,46 +691,3 @@ func (r *copyingRowReceiver) Push(row sqlbase.EncDatumRow, meta *ProducerMetadat
 	}
 	return r.RowReceiver.Push(row, meta)
 }
-
-// String implements fmt.Stringer.
-func (e *Error) String() string {
-	if err := e.ErrorDetail(); err != nil {
-		return err.Error()
-	}
-	return "<nil>"
-}
-
-// NewError creates an Error from an error, to be sent on the wire. It will
-// recognize certain errors and marshall them accordingly, and everything
-// unrecognized is turned into a PGError with code "internal".
-func NewError(err error) *Error {
-	if pgErr, ok := pgerror.GetPGCause(err); ok {
-		return &Error{Detail: &Error_PGError{PGError: pgErr}}
-	} else if retryErr, ok := err.(*roachpb.UnhandledRetryableError); ok {
-		return &Error{
-			Detail: &Error_RetryableTxnError{
-				RetryableTxnError: retryErr,
-			}}
-	} else {
-		// Anything unrecognized is an "internal error".
-		return &Error{
-			Detail: &Error_PGError{
-				PGError: pgerror.NewError(
-					pgerror.CodeInternalError, err.Error())}}
-	}
-}
-
-// ErrorDetail returns the payload as a Go error.
-func (e *Error) ErrorDetail() error {
-	if e == nil {
-		return nil
-	}
-	switch t := e.Detail.(type) {
-	case *Error_PGError:
-		return t.PGError
-	case *Error_RetryableTxnError:
-		return t.RetryableTxnError
-	default:
-		panic(fmt.Sprintf("bad error detail: %+v", t))
-	}
-}
