@@ -18,43 +18,36 @@ import (
 	"context"
 	"sync"
 
-	"github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/jobs"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
-	"github.com/cockroachdb/cockroach/pkg/rpc"
 	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
+	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlutil"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine"
+	"github.com/cockroachdb/cockroach/pkg/storage/diskmap"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
 	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
-	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
-// StreamID identifies a stream; it may be local to a flow or it may cross
-// machine boundaries. The identifier can only be used in the context of a
-// specific flow.
-type StreamID int
-
-// FlowID identifies a flow. It is most importantly used when setting up streams
-// between nodes.
-type FlowID struct {
-	uuid.UUID
-}
-
 // FlowCtx encompasses the contexts needed for various flow components.
 type FlowCtx struct {
 	log.AmbientContext
 
-	Settings *cluster.Settings
+	// TODO(radu): FlowCtx should store a pointer to the server's ServerConfig
+	// instead of having copies of most of its fields.
+	Settings     *cluster.Settings
+	RuntimeStats RuntimeStats
 
 	stopper *stop.Stopper
 
@@ -63,7 +56,7 @@ type FlowCtx struct {
 	// registry (no inbound stream connections need to be performed), they are not
 	// assigned ids. This is done for performance reasons, as local flows are
 	// more likely to be dominated by setup time.
-	id FlowID
+	id distsqlpb.FlowID
 	// EvalCtx is used by all the processors in the flow to evaluate expressions.
 	// Processors that intend to evaluate expressions with this EvalCtx should
 	// get a copy with NewEvalCtx instead of storing a pointer to this one
@@ -72,12 +65,8 @@ type FlowCtx struct {
 	// TODO(andrei): Get rid of this field and pass a non-shared EvalContext to
 	// cores of the processors that need it.
 	EvalCtx *tree.EvalContext
-	// rpcCtx and nodeDialer are used by the Outboxes that may be
-	// present in the flow for connecting to other nodes (rpcCtx for
-	// flows initiated by 2.0 nodes that specify addresses, and
-	// nodeDialer for 2.1 nodes that specify node IDs).
-	// TODO(bdarnell): remove rpcCtx after 2.1.
-	rpcCtx     *rpc.Context
+	// nodeDialer is used by the Outboxes that may be present in the
+	// flow for connecting to other nodes.
 	nodeDialer *nodedialer.Dialer
 	// Gossip is used by the sample aggregator to notify nodes of a new statistic.
 	Gossip *gossip.Gossip
@@ -108,7 +97,12 @@ type FlowCtx struct {
 	// working set is larger than can be stored in memory.
 	// This is not supposed to be used as a general engine.Engine and thus
 	// one should sparingly use the set of features offered by it.
-	TempStorage engine.Engine
+	TempStorage diskmap.Factory
+
+	// BulkAdder is used by backfill/bulk-ingestion processors to ingest large
+	// amounts of data in bulk as SSTs.
+	BulkAdder storagebase.BulkAdderFactory
+
 	// diskMonitor is used to monitor temporary storage disk usage.
 	diskMonitor *mon.BytesMonitor
 
@@ -130,8 +124,7 @@ type FlowCtx struct {
 // them at runtime to ensure expressions are evaluated with the correct indexed
 // var context.
 func (ctx *FlowCtx) NewEvalCtx() *tree.EvalContext {
-	evalCtx := *ctx.EvalCtx
-	return &evalCtx
+	return ctx.EvalCtx.Copy()
 }
 
 // TestingKnobs returns the distsql testing knobs for this flow context.
@@ -181,11 +174,11 @@ type Flow struct {
 	// goroutines.
 	startedGoroutines bool
 
-	localStreams map[StreamID]RowReceiver
+	localStreams map[distsqlpb.StreamID]RowReceiver
 
 	// inboundStreams are streams that receive data from other hosts; this map
 	// is to be passed to flowRegistry.RegisterFlow.
-	inboundStreams map[StreamID]*inboundStreamInfo
+	inboundStreams map[distsqlpb.StreamID]*inboundStreamInfo
 
 	// waitGroup is used to wait for async components of the flow:
 	//  - processors
@@ -203,7 +196,7 @@ type Flow struct {
 	ctxDone   <-chan struct{}
 
 	// spec is the request that produced this flow. Only used for debugging.
-	spec *FlowSpec
+	spec *distsqlpb.FlowSpec
 }
 
 func newFlow(
@@ -225,34 +218,31 @@ func newFlow(
 // setupInboundStream adds a stream to the stream map (inboundStreams or
 // localStreams).
 func (f *Flow) setupInboundStream(
-	ctx context.Context, spec StreamEndpointSpec, receiver RowReceiver,
+	ctx context.Context, spec distsqlpb.StreamEndpointSpec, receiver RowReceiver,
 ) error {
-	if spec.DeprecatedTargetAddr != "" {
-		return errors.Errorf("inbound stream has target address set: %s", spec.DeprecatedTargetAddr)
-	}
 	sid := spec.StreamID
 	switch spec.Type {
-	case StreamEndpointSpec_SYNC_RESPONSE:
+	case distsqlpb.StreamEndpointSpec_SYNC_RESPONSE:
 		return errors.Errorf("inbound stream of type SYNC_RESPONSE")
 
-	case StreamEndpointSpec_REMOTE:
+	case distsqlpb.StreamEndpointSpec_REMOTE:
 		if _, found := f.inboundStreams[sid]; found {
 			return errors.Errorf("inbound stream %d has multiple consumers", sid)
 		}
 		if f.inboundStreams == nil {
-			f.inboundStreams = make(map[StreamID]*inboundStreamInfo)
+			f.inboundStreams = make(map[distsqlpb.StreamID]*inboundStreamInfo)
 		}
 		if log.V(2) {
 			log.Infof(ctx, "set up inbound stream %d", sid)
 		}
 		f.inboundStreams[sid] = &inboundStreamInfo{receiver: receiver, waitGroup: &f.waitGroup}
 
-	case StreamEndpointSpec_LOCAL:
+	case distsqlpb.StreamEndpointSpec_LOCAL:
 		if _, found := f.localStreams[sid]; found {
 			return errors.Errorf("local stream %d has multiple consumers", sid)
 		}
 		if f.localStreams == nil {
-			f.localStreams = make(map[StreamID]RowReceiver)
+			f.localStreams = make(map[distsqlpb.StreamID]RowReceiver)
 		}
 		f.localStreams[sid] = receiver
 
@@ -263,43 +253,21 @@ func (f *Flow) setupInboundStream(
 	return nil
 }
 
-// This RowReceiver clears its BoundAccount on every input row. This is useful
-// for clearing the per-row memory account that's used for expression
-// evaluation.
-type accountClearingRowReceiver struct {
-	RowReceiver
-	ctx context.Context
-	acc *mon.BoundAccount
-}
-
-func (r *accountClearingRowReceiver) Push(
-	row sqlbase.EncDatumRow, meta *ProducerMetadata,
-) ConsumerStatus {
-	r.acc.Clear(r.ctx)
-	return r.RowReceiver.Push(row, meta)
-}
-
 // setupOutboundStream sets up an output stream; if the stream is local, the
 // RowChannel is looked up in the localStreams map; otherwise an outgoing
 // mailbox is created.
-func (f *Flow) setupOutboundStream(spec StreamEndpointSpec) (RowReceiver, error) {
+func (f *Flow) setupOutboundStream(spec distsqlpb.StreamEndpointSpec) (RowReceiver, error) {
 	sid := spec.StreamID
 	switch spec.Type {
-	case StreamEndpointSpec_SYNC_RESPONSE:
-		// Wrap the syncFlowConsumer in a row receiver that clears the row's memory
-		// account.
-		return &accountClearingRowReceiver{
-			acc:         f.EvalCtx.ActiveMemAcc,
-			ctx:         f.EvalCtx.Ctx(),
-			RowReceiver: f.syncFlowConsumer,
-		}, nil
+	case distsqlpb.StreamEndpointSpec_SYNC_RESPONSE:
+		return f.syncFlowConsumer, nil
 
-	case StreamEndpointSpec_REMOTE:
-		outbox := newOutbox(&f.FlowCtx, spec.TargetNodeID, spec.DeprecatedTargetAddr, f.id, sid)
+	case distsqlpb.StreamEndpointSpec_REMOTE:
+		outbox := newOutbox(&f.FlowCtx, spec.TargetNodeID, f.id, sid)
 		f.startables = append(f.startables, outbox)
 		return outbox, nil
 
-	case StreamEndpointSpec_LOCAL:
+	case distsqlpb.StreamEndpointSpec_LOCAL:
 		rowChan, found := f.localStreams[sid]
 		if !found {
 			return nil, errors.Errorf("unconnected inbound stream %d", sid)
@@ -318,7 +286,7 @@ func (f *Flow) setupOutboundStream(spec StreamEndpointSpec) (RowReceiver, error)
 // setupRouter initializes a router and the outbound streams.
 //
 // Pass-through routers are not supported; they should be handled separately.
-func (f *Flow) setupRouter(spec *OutputRouterSpec) (router, error) {
+func (f *Flow) setupRouter(spec *distsqlpb.OutputRouterSpec) (router, error) {
 	streams := make([]RowReceiver, len(spec.Streams))
 	for i := range spec.Streams {
 		var err error
@@ -341,33 +309,30 @@ func checkNumInOut(inputs []RowSource, outputs []RowReceiver, numIn, numOut int)
 }
 
 func (f *Flow) makeProcessor(
-	ctx context.Context, ps *ProcessorSpec, inputs []RowSource,
+	ctx context.Context, ps *distsqlpb.ProcessorSpec, inputs []RowSource,
 ) (Processor, error) {
 	if len(ps.Output) != 1 {
 		return nil, errors.Errorf("only single-output processors supported")
 	}
-	outputs := make([]RowReceiver, len(ps.Output))
-	for i := range ps.Output {
-		spec := &ps.Output[i]
-		if spec.Type == OutputRouterSpec_PASS_THROUGH {
-			// There is no entity that corresponds to a pass-through router - we just
-			// use its output stream directly.
-			if len(spec.Streams) != 1 {
-				return nil, errors.Errorf("expected one stream for passthrough router")
-			}
-			var err error
-			outputs[i], err = f.setupOutboundStream(spec.Streams[0])
-			if err != nil {
-				return nil, err
-			}
-			continue
+	var output RowReceiver
+	spec := &ps.Output[0]
+	if spec.Type == distsqlpb.OutputRouterSpec_PASS_THROUGH {
+		// There is no entity that corresponds to a pass-through router - we just
+		// use its output stream directly.
+		if len(spec.Streams) != 1 {
+			return nil, errors.Errorf("expected one stream for passthrough router")
 		}
-
+		var err error
+		output, err = f.setupOutboundStream(spec.Streams[0])
+		if err != nil {
+			return nil, err
+		}
+	} else {
 		r, err := f.setupRouter(spec)
 		if err != nil {
 			return nil, err
 		}
-		outputs[i] = r
+		output = r
 		f.startables = append(f.startables, r)
 	}
 
@@ -378,10 +343,9 @@ func (f *Flow) makeProcessor(
 	// upstreams, though, which means that copyingRowReceivers are only used on
 	// non-fused processors like the output routers.
 
-	for i := range outputs {
-		outputs[i] = &copyingRowReceiver{RowReceiver: outputs[i]}
-	}
+	output = &copyingRowReceiver{RowReceiver: output}
 
+	outputs := []RowReceiver{output}
 	proc, err := newProcessor(ctx, &f.FlowCtx, ps.ProcessorID, &ps.Core, &ps.Post, inputs, outputs, f.localProcessors)
 	if err != nil {
 		return nil, err
@@ -389,73 +353,74 @@ func (f *Flow) makeProcessor(
 
 	// Initialize any routers (the setupRouter case above) and outboxes.
 	types := proc.OutputTypes()
-	for _, o := range outputs {
-		rowRecv := o.(*copyingRowReceiver).RowReceiver
-		clearer, ok := rowRecv.(*accountClearingRowReceiver)
-		if ok {
-			rowRecv = clearer.RowReceiver
-		}
-		switch o := rowRecv.(type) {
-		case router:
-			o.init(ctx, &f.FlowCtx, types)
-		case *outbox:
-			o.init(types)
-		}
+	rowRecv := output.(*copyingRowReceiver).RowReceiver
+	switch o := rowRecv.(type) {
+	case router:
+		o.init(ctx, &f.FlowCtx, types)
+	case *outbox:
+		o.init(types)
 	}
 	return proc, nil
 }
 
-func (f *Flow) setup(ctx context.Context, spec *FlowSpec) error {
-	f.spec = spec
-
-	// First step: setup the input synchronizers for all processors.
-	inputSyncs := make([][]RowSource, len(spec.Processors))
-	for pIdx, ps := range spec.Processors {
+// setupInputSyncs populates a slice of input syncs, one for each Processor in
+// f.Spec, each containing one RowSource for each input to that Processor.
+func (f *Flow) setupInputSyncs(ctx context.Context) ([][]RowSource, error) {
+	inputSyncs := make([][]RowSource, len(f.spec.Processors))
+	for pIdx, ps := range f.spec.Processors {
 		for _, is := range ps.Input {
 			if len(is.Streams) == 0 {
-				return errors.Errorf("input sync with no streams")
+				return nil, errors.Errorf("input sync with no streams")
 			}
 			var sync RowSource
 			switch is.Type {
-			case InputSyncSpec_UNORDERED:
+			case distsqlpb.InputSyncSpec_UNORDERED:
 				mrc := &RowChannel{}
 				mrc.InitWithNumSenders(is.ColumnTypes, len(is.Streams))
 				for _, s := range is.Streams {
 					if err := f.setupInboundStream(ctx, s, mrc); err != nil {
-						return err
+						return nil, err
 					}
 				}
 				sync = mrc
-			case InputSyncSpec_ORDERED:
+			case distsqlpb.InputSyncSpec_ORDERED:
 				// Ordered synchronizer: create a RowChannel for each input.
 				streams := make([]RowSource, len(is.Streams))
 				for i, s := range is.Streams {
 					rowChan := &RowChannel{}
 					rowChan.InitWithNumSenders(is.ColumnTypes, 1)
 					if err := f.setupInboundStream(ctx, s, rowChan); err != nil {
-						return err
+						return nil, err
 					}
 					streams[i] = rowChan
 				}
 				var err error
-				sync, err = makeOrderedSync(convertToColumnOrdering(is.Ordering), f.EvalCtx, streams)
+				sync, err = makeOrderedSync(distsqlpb.ConvertToColumnOrdering(is.Ordering), f.EvalCtx, streams)
 				if err != nil {
-					return err
+					return nil, err
 				}
 
 			default:
-				return errors.Errorf("unsupported input sync type %s", is.Type)
+				return nil, errors.Errorf("unsupported input sync type %s", is.Type)
 			}
 			inputSyncs[pIdx] = append(inputSyncs[pIdx], sync)
 		}
 	}
+	return inputSyncs, nil
+}
 
-	f.processors = make([]Processor, 0, len(spec.Processors))
+// setupProcessors creates processors for each spec in f.spec, fusing processors
+// together when possible (when an upstream processor implements RowSource, only
+// has one output, and that output is a simple PASS_THROUGH output), and
+// populates f.processors with all created processors that weren't fused to and
+// thus need their own goroutine.
+func (f *Flow) setupProcessors(ctx context.Context, inputSyncs [][]RowSource) error {
+	f.processors = make([]Processor, 0, len(f.spec.Processors))
 
 	// Populate f.processors: see which processors need their own goroutine and
 	// which are fused with their consumer.
-	for i := range spec.Processors {
-		pspec := &spec.Processors[i]
+	for i := range f.spec.Processors {
+		pspec := &f.spec.Processors[i]
 		p, err := f.makeProcessor(ctx, pspec, inputSyncs[i])
 		if err != nil {
 			return err
@@ -475,7 +440,7 @@ func (f *Flow) setup(ctx context.Context, spec *FlowSpec) error {
 				return false
 			}
 			ospec := &pspec.Output[0]
-			if ospec.Type != OutputRouterSpec_PASS_THROUGH {
+			if ospec.Type != distsqlpb.OutputRouterSpec_PASS_THROUGH {
 				// The output is not pass-through and thus is being sent through a
 				// router.
 				return false
@@ -485,7 +450,7 @@ func (f *Flow) setup(ctx context.Context, spec *FlowSpec) error {
 				return false
 			}
 
-			for pIdx, ps := range spec.Processors {
+			for pIdx, ps := range f.spec.Processors {
 				if pIdx <= i {
 					// Skip processors which have already been created.
 					continue
@@ -494,7 +459,7 @@ func (f *Flow) setup(ctx context.Context, spec *FlowSpec) error {
 					// Look for "simple" inputs: an unordered input (which, by definition,
 					// doesn't require an ordered synchronizer), with a single input stream
 					// (which doesn't require a multiplexed RowChannel).
-					if in.Type != InputSyncSpec_UNORDERED {
+					if in.Type != distsqlpb.InputSyncSpec_UNORDERED {
 						continue
 					}
 					if len(in.Streams) != 1 {
@@ -517,9 +482,51 @@ func (f *Flow) setup(ctx context.Context, spec *FlowSpec) error {
 	return nil
 }
 
-// startInternal starts the flow. All processors apart from the last one are
-// started, each in their own goroutine. The caller must forward any returned
-// error to syncFlowConsumer if set.
+func (f *Flow) setup(ctx context.Context, spec *distsqlpb.FlowSpec) error {
+	f.spec = spec
+
+	// First step: setup the input synchronizers for all processors.
+	inputSyncs, err := f.setupInputSyncs(ctx)
+	if err != nil {
+		return err
+	}
+
+	if f.EvalCtx.SessionData.Vectorize != sessiondata.VectorizeOff {
+		err := f.setupVectorized(ctx)
+		if err == nil {
+			log.VEventf(ctx, 1, "vectorized flow.")
+			return nil
+		}
+		// Vectorization attempt failed with an error.
+		if f.EvalCtx.SessionData.Vectorize == sessiondata.VectorizeAlways {
+			// Only return the error if we are running a local planNode that is an
+			// exception to the rule that failures to set up a vectorized flow when
+			// experimental_vectorize=always should return an error.
+			var isException bool
+			if len(spec.Processors) == 1 &&
+				spec.Processors[0].Core.LocalPlanNode != nil {
+				rsidx := spec.Processors[0].Core.LocalPlanNode.RowSourceIdx
+				if rsidx != nil {
+					lp := f.localProcessors[*rsidx]
+					if z, ok := lp.(vectorizeAlwaysException); ok {
+						isException = z.IsException()
+					}
+				}
+			}
+			if !isException {
+				return err
+			}
+		}
+		log.VEventf(ctx, 1, "failed to vectorize: %s", err)
+	}
+
+	// Then, populate f.processors.
+	return f.setupProcessors(ctx, inputSyncs)
+}
+
+// startInternal starts the flow. All processors are started, each in their own
+// goroutine. The caller must forward any returned error to syncFlowConsumer if
+// set.
 func (f *Flow) startInternal(ctx context.Context, doneFn func()) error {
 	f.doneFn = doneFn
 	log.VEventf(
@@ -553,11 +560,14 @@ func (f *Flow) startInternal(ctx context.Context, doneFn func()) error {
 	for _, s := range f.startables {
 		s.start(ctx, &f.waitGroup, f.ctxCancel)
 	}
-	for i := 0; i < len(f.processors)-1; i++ {
+	for i := 0; i < len(f.processors); i++ {
 		f.waitGroup.Add(1)
-		go f.processors[i].Run(ctx, &f.waitGroup)
+		go func(i int) {
+			f.processors[i].Run(ctx)
+			f.waitGroup.Done()
+		}(i)
 	}
-	f.startedGoroutines = len(f.startables) > 0 || len(f.processors) > 1 || !f.isLocal()
+	f.startedGoroutines = len(f.startables) > 0 || len(f.processors) > 0 || !f.isLocal()
 	return nil
 }
 
@@ -566,13 +576,15 @@ func (f *Flow) isLocal() bool {
 	return len(f.inboundStreams) == 0
 }
 
-// StartAsync starts the flow. Each processor runs in their own goroutine.
+// Start starts the flow. Processors run asynchronously in their own goroutines.
+// Wait() needs to be called to wait for the flow to finish.
+// See Run() for a synchronous version.
 //
 // Generally if errors are encountered during the setup part, they're returned.
 // But if the flow is a synchronous one, then no error is returned; instead the
 // setup error is pushed to the syncFlowConsumer. In this case, a subsequent
 // call to f.Wait() will not block.
-func (f *Flow) StartAsync(ctx context.Context, doneFn func()) error {
+func (f *Flow) Start(ctx context.Context, doneFn func()) error {
 	if err := f.startInternal(ctx, doneFn); err != nil {
 		// For sync flows, the error goes to the consumer.
 		if f.syncFlowConsumer != nil {
@@ -582,18 +594,26 @@ func (f *Flow) StartAsync(ctx context.Context, doneFn func()) error {
 		}
 		return err
 	}
-	if len(f.processors) > 0 {
-		f.waitGroup.Add(1)
-		go f.processors[len(f.processors)-1].Run(ctx, &f.waitGroup)
-		f.startedGoroutines = true
-	}
 	return nil
 }
 
-// StartSync starts the flow just like StartAsync but the last processor is run
-// from the main goroutine. Wait() must still be called afterwards as other
-// goroutines might be spawned.
-func (f *Flow) StartSync(ctx context.Context, doneFn func()) error {
+// Run runs the flow to completion. The last processor is run in the current
+// goroutine; others may run in different goroutines depending on how the flow
+// was configured.
+// f.Wait() is called internally, so the call blocks until all the flow's
+// goroutines are done.
+// The caller needs to call f.Cleanup().
+func (f *Flow) Run(ctx context.Context, doneFn func()) error {
+	defer f.Wait()
+
+	// We'll take care of the last processor in particular.
+	var headProc Processor
+	if len(f.processors) == 0 {
+		return pgerror.NewAssertionErrorf("no processors in flow")
+	}
+	headProc = f.processors[len(f.processors)-1]
+	f.processors = f.processors[:len(f.processors)-1]
+
 	if err := f.startInternal(ctx, doneFn); err != nil {
 		// For sync flows, the error goes to the consumer.
 		if f.syncFlowConsumer != nil {
@@ -603,9 +623,7 @@ func (f *Flow) StartSync(ctx context.Context, doneFn func()) error {
 		}
 		return err
 	}
-	if len(f.processors) > 0 {
-		f.processors[len(f.processors)-1].Run(ctx, nil)
-	}
+	headProc.Run(ctx)
 	return nil
 }
 
@@ -631,15 +649,27 @@ func (f *Flow) Wait() {
 	}
 }
 
+// Releasable is an interface for objects than can be Released back into a
+// memory pool when finished.
+type Releasable interface {
+	// Release allows this object to be returned to a memory pool. Objects must
+	// not be used after Release is called.
+	Release()
+}
+
 // Cleanup should be called when the flow completes (after all processors and
 // mailboxes exited).
 func (f *Flow) Cleanup(ctx context.Context) {
 	if f.status == FlowFinished {
 		panic("flow cleanup called twice")
 	}
-	// This closes the account and monitor opened in ServerImpl.setupFlow.
-	f.EvalCtx.ActiveMemAcc.Close(ctx)
+	// This closes the monitor opened in ServerImpl.setupFlow.
 	f.EvalCtx.Stop(ctx)
+	for _, p := range f.processors {
+		if d, ok := p.(Releasable); ok {
+			d.Release()
+		}
+	}
 	if log.V(1) {
 		log.Infof(ctx, "cleaning up")
 	}
@@ -667,22 +697,17 @@ func (f *Flow) cancel() {
 		return
 	}
 	f.flowRegistry.Lock()
-	defer f.flowRegistry.Unlock()
+	timedOutReceivers := f.flowRegistry.cancelPendingStreamsLocked(f.id)
+	f.flowRegistry.Unlock()
 
-	entry := f.flowRegistry.flows[f.id]
-	for streamID, is := range entry.inboundStreams {
-		// Connected, non-finished inbound streams will get an error
-		// returned in ProcessInboundStream(). Non-connected streams
-		// are handled below.
-		if !is.connected && !is.finished {
-			is.canceled = true
+	for _, receiver := range timedOutReceivers {
+		go func(receiver RowReceiver) {
 			// Stream has yet to be started; send an error to its
 			// receiver and prevent it from being connected.
-			is.receiver.Push(
+			receiver.Push(
 				nil, /* row */
 				&ProducerMetadata{Err: sqlbase.QueryCanceledError})
-			is.receiver.ProducerDone()
-			f.flowRegistry.finishInboundStreamLocked(f.id, streamID)
-		}
+			receiver.ProducerDone()
+		}(receiver)
 	}
 }
