@@ -18,9 +18,8 @@ import (
 	"context"
 	"fmt"
 
-	"math/rand"
-
-	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
@@ -84,7 +83,7 @@ type hashJoiner struct {
 	// We read a portion of both streams, in the hope that one is small. One of
 	// the containers will contain the entire "stored" stream, the other just the
 	// start of the other stream.
-	rows [2]memRowContainer
+	rows [2]rowcontainer.MemRowContainer
 
 	// storedSide is set by the initial buffering phase and indicates which
 	// stream we store fully and build the hashRowContainer from.
@@ -95,7 +94,7 @@ type hashJoiner struct {
 	nullEquality bool
 
 	useTempStorage bool
-	storedRows     hashRowContainer
+	storedRows     rowcontainer.HashRowContainer
 
 	// Used by tests to force a storedSide.
 	forcedStoredSide *joinSide
@@ -106,7 +105,7 @@ type hashJoiner struct {
 		row sqlbase.EncDatumRow
 		// iter is an iterator over the bucket that matches row on the equality
 		// columns.
-		iter rowMarkerIterator
+		iter rowcontainer.RowMarkerIterator
 		// matched represents whether any row that matches row on equality columns
 		// has also passed the ON condition.
 		matched bool
@@ -114,17 +113,12 @@ type hashJoiner struct {
 
 	// emittingUnmatchedState is used when hjEmittingUnmatched.
 	emittingUnmatchedState struct {
-		iter rowIterator
+		iter rowcontainer.RowIterator
 	}
 
 	// testingKnobMemFailPoint specifies a state in which the hashJoiner will
 	// fail at a random point during this phase.
 	testingKnobMemFailPoint hashJoinerState
-	// testingKnobFailProbability is a value in the range [0, 1] that specifies
-	// a probability of failure at each possible failure point in a phase
-	// specified by testingKnobMemFailPoint. Note that it becomes less likely
-	// to hit a specific failure point as execution in the phase continues.
-	testingKnobFailProbability float64
 
 	// Context cancellation checker.
 	cancelChecker *sqlbase.CancelChecker
@@ -138,10 +132,10 @@ const hashJoinerProcName = "hash joiner"
 func newHashJoiner(
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *HashJoinerSpec,
+	spec *distsqlpb.HashJoinerSpec,
 	leftSource RowSource,
 	rightSource RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	output RowReceiver,
 ) (*hashJoiner, error) {
 	h := &hashJoiner{
@@ -209,11 +203,11 @@ func newHashJoiner(
 		h.finishTrace = h.outputStatsToTrace
 	}
 
-	h.rows[leftSide].initWithMon(
-		nil /* ordering */, h.leftSource.OutputTypes(), h.evalCtx, h.MemMonitor,
+	h.rows[leftSide].InitWithMon(
+		nil /* ordering */, h.leftSource.OutputTypes(), h.evalCtx, h.MemMonitor, 0, /* rowCapacity */
 	)
-	h.rows[rightSide].initWithMon(
-		nil /* ordering */, h.rightSource.OutputTypes(), h.evalCtx, h.MemMonitor,
+	h.rows[rightSide].InitWithMon(
+		nil /* ordering */, h.rightSource.OutputTypes(), h.evalCtx, h.MemMonitor, 0, /* rowCapacity */
 	)
 
 	if h.joinType == sqlbase.IntersectAllJoin || h.joinType == sqlbase.ExceptAllJoin {
@@ -266,11 +260,6 @@ func (h *hashJoiner) Next() (sqlbase.EncDatumRow, *ProducerMetadata) {
 	return nil, h.DrainHelper()
 }
 
-// ConsumerDone is part of the RowSource interface.
-func (h *hashJoiner) ConsumerDone() {
-	h.MoveToDraining(nil /* err */)
-}
-
 // ConsumerClosed is part of the RowSource interface.
 func (h *hashJoiner) ConsumerClosed() {
 	h.close()
@@ -346,14 +335,20 @@ func (h *hashJoiner) build() (hashJoinerState, sqlbase.EncDatumRow, *ProducerMet
 		if err := h.rows[side].AddRow(h.Ctx, row); err != nil {
 			// If this error is a memory limit error, move to hjConsumingStoredSide.
 			h.storedSide = side
-			if spilled, spillErr := h.maybeSpillToDisk(err); spilled {
-				addErr := h.storedRows.AddRow(h.Ctx, row)
-				if addErr == nil {
-					return hjConsumingStoredSide, nil, nil
+			if sqlbase.IsOutOfMemoryError(err) {
+				if !h.useTempStorage {
+					err = errors.Wrap(err, "error while attempting hashJoiner disk spill: temp storage disabled")
+				} else {
+					if err := h.initStoredRows(); err != nil {
+						h.MoveToDraining(err)
+						return hjStateUnknown, nil, h.DrainHelper()
+					}
+					addErr := h.storedRows.AddRow(h.Ctx, row)
+					if addErr == nil {
+						return hjConsumingStoredSide, nil, nil
+					}
+					err = errors.Wrap(err, addErr.Error())
 				}
-				err = errors.Wrap(err, addErr.Error())
-			} else if spillErr != nil {
-				err = errors.Wrap(err, spillErr.Error())
 			}
 			h.MoveToDraining(err)
 			return hjStateUnknown, nil, h.DrainHelper()
@@ -384,42 +379,30 @@ func (h *hashJoiner) consumeStoredSide() (hashJoinerState, sqlbase.EncDatumRow,
 		if row == nil {
 			// The stored side has been fully consumed, move on to hjReadingProbeSide.
 			// If storedRows is in-memory, pre-reserve the memory needed to mark.
-			if rc, ok := h.storedRows.(*hashMemRowContainer); ok {
-				err = h.maybeMakeMemErr("reserving mark memory")
-				if err == nil {
-					err = rc.reserveMarkMemoryMaybe(h.Ctx)
-				}
-				if err != nil {
-					if spilled, spillErr := h.maybeSpillToDisk(err); spilled {
-						return hjReadingProbeSide, nil, nil
-					} else if spillErr != nil {
-						err = errors.Wrap(err, spillErr.Error())
-					}
-					h.MoveToDraining(err)
-					return hjStateUnknown, nil, h.DrainHelper()
-				}
+			if rc, ok := h.storedRows.(*rowcontainer.HashMemRowContainer); ok {
+				// h.storedRows is hashMemRowContainer and not a disk backed one, so
+				// h.useTempStorage is false and we cannot spill to disk, so we simply
+				// will return an error if it occurs.
+				err = rc.ReserveMarkMemoryMaybe(h.Ctx)
+			} else if hdbrc, ok := h.storedRows.(*rowcontainer.HashDiskBackedRowContainer); ok {
+				err = hdbrc.ReserveMarkMemoryMaybe(h.Ctx)
+			} else {
+				panic("unexpected type of storedRows in hashJoiner")
+			}
+			if err != nil {
+				h.MoveToDraining(err)
+				return hjStateUnknown, nil, h.DrainHelper()
 			}
 			return hjReadingProbeSide, nil, nil
 		}
 
-		err = h.maybeMakeMemErr("consuming stored side")
-		if err == nil {
-			err = h.storedRows.AddRow(h.Ctx, row)
-		}
+		err = h.storedRows.AddRow(h.Ctx, row)
+		// Regardless of the underlying row container (disk backed or in-memory
+		// only), we cannot do anything about an error if it occurs.
 		if err != nil {
-			if spilled, spillErr := h.maybeSpillToDisk(err); spilled {
-				if err := h.storedRows.AddRow(h.Ctx, row); err != nil {
-					h.MoveToDraining(err)
-					return hjStateUnknown, nil, h.DrainHelper()
-				}
-				continue
-			} else if spillErr != nil {
-				err = errors.Wrap(err, spillErr.Error())
-			}
 			h.MoveToDraining(err)
 			return hjStateUnknown, nil, h.DrainHelper()
 		}
-
 	}
 }
 
@@ -609,10 +592,19 @@ func (h *hashJoiner) emitUnmatched() (hashJoinerState, sqlbase.EncDatumRow, *Pro
 
 func (h *hashJoiner) close() {
 	if h.InternalClose() {
-		h.rows[leftSide].Close(h.Ctx)
-		h.rows[rightSide].Close(h.Ctx)
+		// We need to close only memRowContainer of the probe side because the
+		// stored side container will be closed by closing h.storedRows.
+		if h.storedSide == rightSide {
+			h.rows[leftSide].Close(h.Ctx)
+		} else {
+			h.rows[rightSide].Close(h.Ctx)
+		}
 		if h.storedRows != nil {
 			h.storedRows.Close(h.Ctx)
+		} else {
+			// h.storedRows has not been initialized, so we need to close the stored
+			// side container explicitly.
+			h.rows[h.storedSide].Close(h.Ctx)
 		}
 		if h.probingRowState.iter != nil {
 			h.probingRowState.iter.Close()
@@ -650,6 +642,49 @@ func (h *hashJoiner) receiveNext(
 		} else if row == nil {
 			return nil, nil, false, nil
 		}
+		// We make the explicit check for whether or not the row contained a NULL value
+		// on an equality column. The reasoning here is because of the way we expect
+		// NULL equality checks to behave (i.e. NULL != NULL) and the fact that we
+		// use the encoding of any given row as key into our bucket. Thus if we
+		// encountered a NULL row when building the hashmap we have to store in
+		// order to use it for RIGHT OUTER joins but if we encounter another
+		// NULL row when going through the left stream (probing phase), matching
+		// this with the first NULL row would be incorrect.
+		//
+		// If we have have the following:
+		// CREATE TABLE t(x INT); INSERT INTO t(x) VALUES (NULL);
+		//    |  x   |
+		//     ------
+		//    | NULL |
+		//
+		// For the following query:
+		// SELECT * FROM t AS a FULL OUTER JOIN t AS b USING(x);
+		//
+		// We expect:
+		//    |  x   |
+		//     ------
+		//    | NULL |
+		//    | NULL |
+		//
+		// The following examples illustrates the behavior when joining on two
+		// or more columns, and only one of them contains NULL.
+		// If we have have the following:
+		// CREATE TABLE t(x INT, y INT);
+		// INSERT INTO t(x, y) VALUES (44,51), (NULL,52);
+		//    |  x   |  y   |
+		//     ------
+		//    |  44  |  51  |
+		//    | NULL |  52  |
+		//
+		// For the following query:
+		// SELECT * FROM t AS a FULL OUTER JOIN t AS b USING(x, y);
+		//
+		// We expect:
+		//    |  x   |  y   |
+		//     ------
+		//    |  44  |  51  |
+		//    | NULL |  52  |
+		//    | NULL |  52  |
 		hasNull := false
 		for _, c := range h.eqCols[side] {
 			if row[c].IsNull() {
@@ -686,110 +721,29 @@ func (h *hashJoiner) shouldEmitUnmatched(
 
 // initStoredRows initializes a hashRowContainer and sets h.storedRows.
 func (h *hashJoiner) initStoredRows() error {
-	storedMemRows := makeHashMemRowContainer(&h.rows[h.storedSide])
-	err := storedMemRows.Init(
+	if h.useTempStorage {
+		hrc := rowcontainer.MakeHashDiskBackedRowContainer(
+			&h.rows[h.storedSide],
+			h.evalCtx,
+			h.MemMonitor,
+			h.diskMonitor,
+			h.flowCtx.TempStorage,
+		)
+		h.storedRows = &hrc
+	} else {
+		hrc := rowcontainer.MakeHashMemRowContainer(&h.rows[h.storedSide])
+		h.storedRows = &hrc
+	}
+	return h.storedRows.Init(
 		h.Ctx,
 		shouldMark(h.storedSide, h.joinType),
-		h.rows[h.storedSide].types,
+		h.rows[h.storedSide].Types(),
 		h.eqCols[h.storedSide],
 		h.nullEquality,
 	)
-	if err == nil {
-		err = h.maybeMakeMemErr("initializing mem rows")
-		// Close the container on an artificial error.
-		if err != nil {
-			storedMemRows.Close(h.Ctx)
-		}
-	}
-	if err != nil {
-		if spilled, spillErr := h.maybeSpillToDisk(err); spilled {
-			return nil
-		} else if spillErr != nil {
-			return errors.Wrap(err, spillErr.Error())
-		}
-		return err
-	}
-
-	h.storedRows = &storedMemRows
-	return nil
 }
 
-// maybeSpillToDisk checks err and spills h.rows[h.storedSide] to disk if the
-// error is a memory error and h.storedRows is not a hashDiskRowContainer. On
-// a successful disk spill, maybeSpillToDisk Close()s whatever h.storedRows
-// previously pointed to and sets h.storedRows a hashDiskRowContainer.
-// Returns whether the hashJoiner spilled to disk and an error if one occurred
-// while doing so.
-// TODO(asubiotto): This should be behind an auto-fallback hashDiskRowContainer.
-func (h *hashJoiner) maybeSpillToDisk(err error) (bool, error) {
-	newDiskSpillErr := func(errString string) error {
-		return fmt.Errorf("error while attempting hashJoiner disk spill: %s", errString)
-	}
-	if !h.useTempStorage {
-		return false, newDiskSpillErr("temp storage disabled")
-	}
-	if pgErr, ok := pgerror.GetPGCause(err); !(ok && pgErr.Code == pgerror.CodeOutOfMemoryError) {
-		return false, nil
-	}
-	if _, ok := h.storedRows.(*hashDiskRowContainer); ok {
-		return false, newDiskSpillErr("already using disk")
-	}
-
-	storedDiskRows := makeHashDiskRowContainer(h.diskMonitor, h.flowCtx.TempStorage)
-	if err := storedDiskRows.Init(
-		h.Ctx,
-		shouldMark(h.storedSide, h.joinType),
-		h.rows[h.storedSide].types,
-		h.eqCols[h.storedSide],
-		h.nullEquality,
-	); err != nil {
-		return false, newDiskSpillErr(err.Error())
-	}
-
-	if h.storedRows != nil {
-		h.storedRows.Close(h.Ctx)
-	}
-	h.storedRows = &storedDiskRows
-
-	// Transfer rows from memory.
-	i := h.rows[h.storedSide].NewFinalIterator(h.Ctx)
-	defer i.Close()
-	for i.Rewind(); ; i.Next() {
-		if err := h.cancelChecker.Check(); err != nil {
-			return false, newDiskSpillErr(err.Error())
-		}
-		if ok, err := i.Valid(); err != nil {
-			return false, newDiskSpillErr(err.Error())
-		} else if !ok {
-			break
-		}
-		memRow, err := i.Row()
-		if err != nil {
-			return false, newDiskSpillErr(err.Error())
-		}
-		if err := storedDiskRows.AddRow(h.Ctx, memRow); err != nil {
-			return false, newDiskSpillErr(err.Error())
-		}
-	}
-	return true, nil
-}
-
-// maybeMakeMemErr is a utility function that returns a memory error with a
-// probability of h.testingKnobMemFailProbability if h.runningState matches the
-// h.testingKnobMemFailPoint.
-func (h *hashJoiner) maybeMakeMemErr(action string) error {
-	if h.testingKnobMemFailPoint == h.runningState &&
-		rand.Float64() < h.testingKnobFailProbability {
-		// Unset the testing knob fail point.
-		h.testingKnobMemFailPoint = hjStateUnknown
-		return pgerror.NewErrorf(
-			pgerror.CodeOutOfMemoryError, fmt.Sprintf("%s test induced error", action),
-		)
-	}
-	return nil
-}
-
-var _ DistSQLSpanStats = &HashJoinerStats{}
+var _ distsqlpb.DistSQLSpanStats = &HashJoinerStats{}
 
 const hashJoinerTagPrefix = "hashjoiner."
 
@@ -876,32 +830,3 @@ func shouldShortCircuit(storedSide joinSide, joinType sqlbase.JoinType) bool {
 		return false
 	}
 }
-
-// encodeColumnsOfRow returns the encoding for the grouping columns. This is
-// then used as our group key to determine which bucket to add to.
-// If the row contains any NULLs and encodeNull is false, hasNull is true and
-// no encoding is returned. If encodeNull is true, hasNull is never set.
-func encodeColumnsOfRow(
-	da *sqlbase.DatumAlloc,
-	appendTo []byte,
-	row sqlbase.EncDatumRow,
-	cols columns,
-	colTypes []sqlbase.ColumnType,
-	encodeNull bool,
-) (encoding []byte, hasNull bool, err error) {
-	for i, colIdx := range cols {
-		if row[colIdx].IsNull() && !encodeNull {
-			return nil, true, nil
-		}
-		// Note: we cannot compare VALUE encodings because they contain column IDs
-		// which can vary.
-		// TODO(radu): we should figure out what encoding is readily available and
-		// use that (though it needs to be consistent across all rows). We could add
-		// functionality to compare VALUE encodings ignoring the column ID.
-		appendTo, err = row[colIdx].Encode(&colTypes[i], da, sqlbase.DatumEncoding_ASCENDING_KEY, appendTo)
-		if err != nil {
-			return appendTo, false, err
-		}
-	}
-	return appendTo, false, nil
-}
