@@ -17,13 +17,14 @@ package distsqlrun
 import (
 	"context"
 	"fmt"
-	math "math"
+	"math"
 	"sort"
 	"strings"
 	"testing"
 
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
@@ -92,8 +93,8 @@ func TestHashJoiner(t *testing.T) {
 				if flowCtxSetup != nil {
 					flowCtxSetup(&flowCtx)
 				}
-				post := PostProcessSpec{Projection: true, OutputColumns: c.outCols}
-				spec := &HashJoinerSpec{
+				post := distsqlpb.PostProcessSpec{Projection: true, OutputColumns: c.outCols}
+				spec := &distsqlpb.HashJoinerSpec{
 					LeftEqColumns:  c.leftEqCols,
 					RightEqColumns: c.rightEqCols,
 					Type:           c.joinType,
@@ -111,7 +112,7 @@ func TestHashJoiner(t *testing.T) {
 				if i == 1 {
 					h.forcedStoredSide = &side
 				}
-				h.Run(context.Background(), nil /* wg */)
+				h.Run(context.Background())
 				side = otherSide(h.storedSide)
 
 				if !out.ProducerClosed() {
@@ -136,26 +137,6 @@ func TestHashJoiner(t *testing.T) {
 			})
 		}
 
-		// Run tests with a probability of the run failing with a memory error.
-		// These verify that the hashJoiner falls back to disk correctly in all
-		// cases.
-		for _, memFailPoint := range []hashJoinerState{hjBuilding, hjConsumingStoredSide} {
-			name := "Building"
-			if memFailPoint == hjConsumingStoredSide {
-				name = "ConsumingStoredSide"
-			}
-			for i := 0; i < 5; i++ {
-				t.Run(fmt.Sprintf("MemFailPoint=%s", name), func(t *testing.T) {
-					if err := testFunc(t, nil, func(h *hashJoiner) {
-						h.testingKnobMemFailPoint = memFailPoint
-						h.testingKnobFailProbability = 0.5
-					}); err != nil {
-						t.Fatal(err)
-					}
-				})
-			}
-		}
-
 		// Run test with a variety of memory limits.
 		for _, memLimit := range []int64{1, 256, 512, 1024, 2048} {
 			t.Run(fmt.Sprintf("MemLimit=%d", memLimit), func(t *testing.T) {
@@ -216,8 +197,8 @@ func TestHashJoinerError(t *testing.T) {
 				diskMonitor: &diskMonitor,
 			}
 
-			post := PostProcessSpec{Projection: true, OutputColumns: c.outCols}
-			spec := &HashJoinerSpec{
+			post := distsqlpb.PostProcessSpec{Projection: true, OutputColumns: c.outCols}
+			spec := &distsqlpb.HashJoinerSpec{
 				LeftEqColumns:  c.leftEqCols,
 				RightEqColumns: c.rightEqCols,
 				Type:           c.joinType,
@@ -229,7 +210,7 @@ func TestHashJoinerError(t *testing.T) {
 			}
 			outTypes := h.OutputTypes()
 			setup(h)
-			h.Run(context.Background(), nil /* wg */)
+			h.Run(context.Background())
 
 			if !out.ProducerClosed() {
 				return errors.New("output RowReceiver not closed")
@@ -294,7 +275,7 @@ func TestHashJoinerDrain(t *testing.T) {
 	for i := range v {
 		v[i] = sqlbase.DatumToEncDatum(columnTypeInt, tree.NewDInt(tree.DInt(i)))
 	}
-	spec := HashJoinerSpec{
+	spec := distsqlpb.HashJoinerSpec{
 		LeftEqColumns:  []uint32{0},
 		RightEqColumns: []uint32{0},
 		Type:           sqlbase.InnerJoin,
@@ -330,13 +311,13 @@ func TestHashJoinerDrain(t *testing.T) {
 		leftInputDrainNotification <- nil
 	}
 	leftInput := NewRowBuffer(
-		oneIntCol,
+		sqlbase.OneIntCol,
 		inputs[0],
 		RowBufferArgs{OnConsumerDone: leftInputConsumerDone},
 	)
-	rightInput := NewRowBuffer(oneIntCol, inputs[1], RowBufferArgs{})
+	rightInput := NewRowBuffer(sqlbase.OneIntCol, inputs[1], RowBufferArgs{})
 	out := NewRowBuffer(
-		oneIntCol,
+		sqlbase.OneIntCol,
 		nil, /* rows */
 		RowBufferArgs{AccumulateRowsWhileDraining: true},
 	)
@@ -354,7 +335,7 @@ func TestHashJoinerDrain(t *testing.T) {
 		EvalCtx:  &evalCtx,
 	}
 
-	post := PostProcessSpec{Projection: true, OutputColumns: outCols}
+	post := distsqlpb.PostProcessSpec{Projection: true, OutputColumns: outCols}
 	h, err := newHashJoiner(&flowCtx, 0 /* processorID */, &spec, leftInput, rightInput, &post, out)
 	if err != nil {
 		t.Fatal(err)
@@ -365,7 +346,7 @@ func TestHashJoinerDrain(t *testing.T) {
 	h.initialBufferSize = 0
 
 	out.ConsumerDone()
-	h.Run(context.Background(), nil /* wg */)
+	h.Run(context.Background())
 
 	if !out.ProducerClosed() {
 		t.Fatalf("output RowReceiver not closed")
@@ -382,7 +363,7 @@ func TestHashJoinerDrain(t *testing.T) {
 		t.Fatalf("left input not drained; still %d rows in it", len(leftInput.mu.records))
 	}
 
-	if err := checkExpectedRows(oneIntCol, expected, out); err != nil {
+	if err := checkExpectedRows(sqlbase.OneIntCol, expected, out); err != nil {
 		t.Fatal(err)
 	}
 }
@@ -398,7 +379,7 @@ func TestHashJoinerDrainAfterBuildPhaseError(t *testing.T) {
 	for i := range v {
 		v[i] = sqlbase.DatumToEncDatum(columnTypeInt, tree.NewDInt(tree.DInt(i)))
 	}
-	spec := HashJoinerSpec{
+	spec := distsqlpb.HashJoinerSpec{
 		LeftEqColumns:  []uint32{0},
 		RightEqColumns: []uint32{0},
 		Type:           sqlbase.InnerJoin,
@@ -452,12 +433,12 @@ func TestHashJoinerDrainAfterBuildPhaseError(t *testing.T) {
 		return nil, nil
 	}
 	leftInput := NewRowBuffer(
-		oneIntCol,
+		sqlbase.OneIntCol,
 		inputs[0],
 		RowBufferArgs{OnConsumerDone: leftInputConsumerDone},
 	)
 	rightInput := NewRowBuffer(
-		oneIntCol,
+		sqlbase.OneIntCol,
 		inputs[1],
 		RowBufferArgs{
 			OnConsumerDone: rightInputConsumerDone,
@@ -465,7 +446,7 @@ func TestHashJoinerDrainAfterBuildPhaseError(t *testing.T) {
 		},
 	)
 	out := NewRowBuffer(
-		oneIntCol,
+		sqlbase.OneIntCol,
 		nil, /* rows */
 		RowBufferArgs{},
 	)
@@ -481,7 +462,7 @@ func TestHashJoinerDrainAfterBuildPhaseError(t *testing.T) {
 	// infrastructure.
 	settingUseTempStorageJoins.Override(&st.SV, false)
 
-	post := PostProcessSpec{Projection: true, OutputColumns: outCols}
+	post := distsqlpb.PostProcessSpec{Projection: true, OutputColumns: outCols}
 	h, err := newHashJoiner(&flowCtx, 0 /* processorID */, &spec, leftInput, rightInput, &post, out)
 	if err != nil {
 		t.Fatal(err)
@@ -489,7 +470,7 @@ func TestHashJoinerDrainAfterBuildPhaseError(t *testing.T) {
 	// Disable initial buffering. We always store the right stream in this case.
 	h.initialBufferSize = 0
 
-	h.Run(context.Background(), nil /* wg */)
+	h.Run(context.Background())
 
 	if !out.ProducerClosed() {
 		t.Fatalf("output RowReceiver not closed")
@@ -547,13 +528,13 @@ func BenchmarkHashJoiner(b *testing.B) {
 	defer tempEngine.Close()
 	flowCtx.TempStorage = tempEngine
 
-	spec := &HashJoinerSpec{
+	spec := &distsqlpb.HashJoinerSpec{
 		LeftEqColumns:  []uint32{0},
 		RightEqColumns: []uint32{0},
 		Type:           sqlbase.InnerJoin,
 		// Implicit @1 = @2 constraint.
 	}
-	post := &PostProcessSpec{}
+	post := &distsqlpb.PostProcessSpec{}
 
 	const numCols = 1
 	for _, spill := range []bool{true, false} {
@@ -569,9 +550,9 @@ func BenchmarkHashJoiner(b *testing.B) {
 					continue
 				}
 				b.Run(fmt.Sprintf("rows=%d", numRows), func(b *testing.B) {
-					rows := makeIntRows(numRows, numCols)
-					leftInput := NewRepeatableRowSource(oneIntCol, rows)
-					rightInput := NewRepeatableRowSource(oneIntCol, rows)
+					rows := sqlbase.MakeIntRows(numRows, numCols)
+					leftInput := NewRepeatableRowSource(sqlbase.OneIntCol, rows)
+					rightInput := NewRepeatableRowSource(sqlbase.OneIntCol, rows)
 					b.SetBytes(int64(8 * numRows * numCols * 2))
 					b.ResetTimer()
 					for i := 0; i < b.N; i++ {
@@ -581,7 +562,7 @@ func BenchmarkHashJoiner(b *testing.B) {
 						if err != nil {
 							b.Fatal(err)
 						}
-						h.Run(context.Background(), nil /* wg */)
+						h.Run(context.Background())
 						leftInput.Reset()
 						rightInput.Reset()
 					}
