@@ -15,6 +15,7 @@
 package distsqlrun
 
 import (
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/pkg/errors"
@@ -24,7 +25,7 @@ type joinerTestCase struct {
 	leftEqCols  []uint32
 	rightEqCols []uint32
 	joinType    sqlbase.JoinType
-	onExpr      Expression
+	onExpr      distsqlpb.Expression
 	outCols     []uint32
 	leftTypes   []sqlbase.ColumnType
 	leftInput   sqlbase.EncDatumRows
@@ -49,7 +50,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.InnerJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -58,7 +59,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[4], v[5]},
 				{v[5], v[5]},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
@@ -76,12 +77,12 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.InnerJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1, 3},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[1]},
@@ -107,17 +108,17 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.InnerJoin,
-			onExpr:      Expression{Expr: "@4 >= 4"},
+			onExpr:      distsqlpb.Expression{Expr: "@4 >= 4"},
 			// Implicit AND @1 = @3 constraint.
 			outCols:   []uint32{0, 1, 3},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
 				{v[1], v[0]},
 				{v[1], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[1]},
@@ -151,7 +152,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftOuterJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -160,7 +161,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[4], v[5]},
 				{v[5], v[5]},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
@@ -181,13 +182,13 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.RightOuterJoin,
 			// Implicit @1 = @4 constraint.
 			outCols:   []uint32{3, 1, 2},
-			leftTypes: threeIntCols,
+			leftTypes: sqlbase.ThreeIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
 				{v[4], v[4], v[5]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -211,7 +212,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.FullOuterJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -219,7 +220,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[3], v[1]},
 				{v[4], v[5]},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
@@ -241,7 +242,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.InnerJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[2], v[4]},
@@ -249,7 +250,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[4], v[5]},
 				{v[5], v[5]},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
@@ -265,9 +266,9 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.LeftOuterJoin,
-			onExpr:      Expression{Expr: "@3 = 9"},
+			onExpr:      distsqlpb.Expression{Expr: "@3 = 9"},
 			outCols:     []uint32{0, 1},
-			leftTypes:   oneIntCol,
+			leftTypes:   sqlbase.OneIntCol,
 			leftInput: sqlbase.EncDatumRows{
 				{v[1]},
 				{v[2]},
@@ -276,7 +277,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[6]},
 				{v[7]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[2], v[8]},
 				{v[3], v[9]},
@@ -309,15 +310,15 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.RightOuterJoin,
-			onExpr:      Expression{Expr: "@2 > 1"},
+			onExpr:      distsqlpb.Expression{Expr: "@2 > 1"},
 			outCols:     []uint32{0, 1},
-			leftTypes:   oneIntCol,
+			leftTypes:   sqlbase.OneIntCol,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0]},
 				{v[1]},
 				{v[2]},
 			},
-			rightTypes: oneIntCol,
+			rightTypes: sqlbase.OneIntCol,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1]},
 				{v[2]},
@@ -334,15 +335,15 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.FullOuterJoin,
-			onExpr:      Expression{Expr: "@2 > 1"},
+			onExpr:      distsqlpb.Expression{Expr: "@2 > 1"},
 			outCols:     []uint32{0, 1},
-			leftTypes:   oneIntCol,
+			leftTypes:   sqlbase.OneIntCol,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0]},
 				{v[1]},
 				{v[2]},
 			},
-			rightTypes: oneIntCol,
+			rightTypes: sqlbase.OneIntCol,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1]},
 				{v[2]},
@@ -364,14 +365,14 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.InnerJoin,
 			// Implicit @1,@2 = @3,@4 constraint.
 			outCols:   []uint32{0, 1, 2, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], null},
 				{null, v[2]},
 				{null, null},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[0], v[4]},
 				{v[1], null, v[5]},
@@ -389,14 +390,14 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftOuterJoin,
 			// Implicit @1,@2 = @3,@4 constraint.
 			outCols:   []uint32{0, 1, 2, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], null},
 				{null, v[2]},
 				{null, null},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[0], v[4]},
 				{v[1], null, v[5]},
@@ -417,14 +418,14 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.RightOuterJoin,
 			// Implicit @1,@2 = @3,@4 constraint.
 			outCols:   []uint32{0, 1, 2, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], null},
 				{null, v[2]},
 				{null, null},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[0], v[4]},
 				{v[1], null, v[5]},
@@ -445,14 +446,14 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.FullOuterJoin,
 			// Implicit @1,@2 = @3,@4 constraint.
 			outCols:   []uint32{0, 1, 2, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], null},
 				{null, v[2]},
 				{null, null},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[0], v[4]},
 				{v[1], null, v[5]},
@@ -478,7 +479,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftSemiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[2], v[0]},
@@ -487,7 +488,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[3], v[4]},
 				{v[3], v[3]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
@@ -509,14 +510,14 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftSemiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[2], v[0]},
@@ -539,14 +540,14 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftSemiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{null, v[1]},
@@ -567,14 +568,14 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftSemiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{null, v[1]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{null, v[1]},
@@ -591,17 +592,17 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.LeftSemiJoin,
-			onExpr:      Expression{Expr: "@1 > 1"},
+			onExpr:      distsqlpb.Expression{Expr: "@1 > 1"},
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 				{v[2], v[2]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[4]},
@@ -619,17 +620,17 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.LeftSemiJoin,
-			onExpr:      Expression{Expr: "@4 > 4 and @2 + @4 = 8"},
+			onExpr:      distsqlpb.Expression{Expr: "@4 > 4 and @2 + @4 = 8"},
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 				{v[2], v[2]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[4]},
@@ -649,13 +650,13 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftAntiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[2], v[5]},
@@ -674,7 +675,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftAntiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[0]},
@@ -683,7 +684,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[2], v[1]},
 				{v[3], v[4]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[2], v[5]},
@@ -702,7 +703,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftAntiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[0]},
@@ -711,7 +712,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[2], v[1]},
 				{v[3], v[4]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{null, v[5]},
@@ -730,7 +731,7 @@ func joinerTestCases() []joinerTestCase {
 			joinType:    sqlbase.LeftAntiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[0]},
@@ -739,7 +740,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[2], v[1]},
 				{v[3], v[4]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{null, v[5]},
@@ -757,10 +758,10 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.LeftAntiJoin,
-			onExpr:      Expression{Expr: "(@2 + @4) % 2 = 0"},
+			onExpr:      distsqlpb.Expression{Expr: "(@2 + @4) % 2 = 0"},
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[1], v[2]},
@@ -768,7 +769,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[2], v[2]},
 				{v[2], v[3]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[2]},
 				{v[2], v[1]},
@@ -786,10 +787,10 @@ func joinerTestCases() []joinerTestCase {
 			leftEqCols:  []uint32{0},
 			rightEqCols: []uint32{0},
 			joinType:    sqlbase.LeftAntiJoin,
-			onExpr:      Expression{Expr: "(@2 + @4) % 2 = 0"},
+			onExpr:      distsqlpb.Expression{Expr: "(@2 + @4) % 2 = 0"},
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[1], v[2]},
@@ -797,7 +798,7 @@ func joinerTestCases() []joinerTestCase {
 				{v[2], v[2]},
 				{v[2], v[3]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[2]},
 				{v[2], v[1]},
@@ -824,7 +825,7 @@ type joinerErrorTestCase struct {
 	leftEqCols  []uint32
 	rightEqCols []uint32
 	joinType    sqlbase.JoinType
-	onExpr      Expression
+	onExpr      distsqlpb.Expression
 	outCols     []uint32
 	leftTypes   []sqlbase.ColumnType
 	leftInput   sqlbase.EncDatumRows
@@ -849,14 +850,14 @@ func joinerErrorTestCases() []joinerErrorTestCase {
 			joinType:    sqlbase.LeftSemiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1, 2},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 				{v[2], v[2]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[4]},
@@ -873,14 +874,14 @@ func joinerErrorTestCases() []joinerErrorTestCase {
 			joinType:    sqlbase.LeftAntiJoin,
 			// Implicit @1 = @3 constraint.
 			outCols:   []uint32{0, 1, 2},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[1]},
 				{v[2], v[1]},
 				{v[2], v[2]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[4]},
