@@ -21,6 +21,7 @@ import (
 	"testing"
 
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
@@ -28,7 +29,7 @@ import (
 )
 
 type mergeJoinerTestCase struct {
-	spec          MergeJoinerSpec
+	spec          distsqlpb.MergeJoinerSpec
 	outCols       []uint32
 	leftTypes     []sqlbase.ColumnType
 	leftInput     sqlbase.EncDatumRows
@@ -50,12 +51,12 @@ func TestMergeJoiner(t *testing.T) {
 
 	testCases := []mergeJoinerTestCase{
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
@@ -63,7 +64,7 @@ func TestMergeJoiner(t *testing.T) {
 				// Implicit @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -72,13 +73,13 @@ func TestMergeJoiner(t *testing.T) {
 				{v[4], v[5]},
 				{v[5], v[5]},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
 				{v[4], v[4], v[5]},
 			},
-			expectedTypes: threeIntCols,
+			expectedTypes: sqlbase.ThreeIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
@@ -86,12 +87,12 @@ func TestMergeJoiner(t *testing.T) {
 			},
 		},
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
@@ -99,12 +100,12 @@ func TestMergeJoiner(t *testing.T) {
 				// Implicit @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 1, 3},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[1]},
@@ -112,7 +113,7 @@ func TestMergeJoiner(t *testing.T) {
 				{v[0], v[5]},
 				{v[0], v[4]},
 			},
-			expectedTypes: threeIntCols,
+			expectedTypes: sqlbase.ThreeIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[0], v[0], v[4]},
 				{v[0], v[0], v[1]},
@@ -127,28 +128,28 @@ func TestMergeJoiner(t *testing.T) {
 			},
 		},
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type:   sqlbase.InnerJoin,
-				OnExpr: Expression{Expr: "@4 >= 4"},
+				OnExpr: distsqlpb.Expression{Expr: "@4 >= 4"},
 				// Implicit AND @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 1, 3},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
 				{v[1], v[0]},
 				{v[1], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[1]},
@@ -161,7 +162,7 @@ func TestMergeJoiner(t *testing.T) {
 				{v[1], v[5]},
 				{v[1], v[4]},
 			},
-			expectedTypes: threeIntCols,
+			expectedTypes: sqlbase.ThreeIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[0], v[0], v[4]},
 				{v[0], v[0], v[5]},
@@ -178,21 +179,21 @@ func TestMergeJoiner(t *testing.T) {
 			},
 		},
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type:   sqlbase.FullOuterJoin,
-				OnExpr: Expression{Expr: "@2 >= @4"},
+				OnExpr: distsqlpb.Expression{Expr: "@2 >= @4"},
 				// Implicit AND @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 1, 3},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[0]},
@@ -206,7 +207,7 @@ func TestMergeJoiner(t *testing.T) {
 
 				{v[6], v[0]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[5]},
 				{v[0], v[5]},
@@ -221,7 +222,7 @@ func TestMergeJoiner(t *testing.T) {
 
 				{v[5], v[0]},
 			},
-			expectedTypes: threeIntCols,
+			expectedTypes: sqlbase.ThreeIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[0], v[0], null},
 				{v[0], v[0], null},
@@ -243,12 +244,12 @@ func TestMergeJoiner(t *testing.T) {
 			},
 		},
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
@@ -256,7 +257,7 @@ func TestMergeJoiner(t *testing.T) {
 				// Implicit @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -265,13 +266,13 @@ func TestMergeJoiner(t *testing.T) {
 				{v[4], v[5]},
 				{v[5], v[5]},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
 				{v[4], v[4], v[5]},
 			},
-			expectedTypes: threeIntCols,
+			expectedTypes: sqlbase.ThreeIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[0], null, null},
 				{v[1], v[0], v[4]},
@@ -282,12 +283,12 @@ func TestMergeJoiner(t *testing.T) {
 			},
 		},
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
@@ -295,13 +296,13 @@ func TestMergeJoiner(t *testing.T) {
 				// Implicit @1 = @3 constraint.
 			},
 			outCols:   []uint32{3, 1, 2},
-			leftTypes: threeIntCols,
+			leftTypes: sqlbase.ThreeIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
 				{v[4], v[4], v[5]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -310,7 +311,7 @@ func TestMergeJoiner(t *testing.T) {
 				{v[4], v[5]},
 				{v[5], v[5]},
 			},
-			expectedTypes: threeIntCols,
+			expectedTypes: sqlbase.ThreeIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[0], null, null},
 				{v[1], v[0], v[4]},
@@ -321,12 +322,12 @@ func TestMergeJoiner(t *testing.T) {
 			},
 		},
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
@@ -334,7 +335,7 @@ func TestMergeJoiner(t *testing.T) {
 				// Implicit @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 3, 4},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[1], v[4]},
@@ -342,14 +343,14 @@ func TestMergeJoiner(t *testing.T) {
 				{v[3], v[1]},
 				{v[4], v[5]},
 			},
-			rightTypes: threeIntCols,
+			rightTypes: sqlbase.ThreeIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[1], v[0], v[4]},
 				{v[3], v[4], v[1]},
 				{v[4], v[4], v[5]},
 				{v[5], v[5], v[1]},
 			},
-			expectedTypes: threeIntCols,
+			expectedTypes: sqlbase.ThreeIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[0], null, null},
 				{v[1], v[0], v[4]},
@@ -360,13 +361,13 @@ func TestMergeJoiner(t *testing.T) {
 			},
 		},
 		{
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 						{ColIdx: 1, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 						{ColIdx: 1, Direction: encoding.Ascending},
@@ -374,21 +375,21 @@ func TestMergeJoiner(t *testing.T) {
 				Type: sqlbase.FullOuterJoin,
 			},
 			outCols:   []uint32{0, 1, 2, 3},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{null, v[4]},
 				{v[0], null},
 				{v[0], v[1]},
 				{v[2], v[4]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{null, v[4]},
 				{v[0], null},
 				{v[0], v[1]},
 				{v[2], v[4]},
 			},
-			expectedTypes: []sqlbase.ColumnType{intType, intType, intType, intType},
+			expectedTypes: []sqlbase.ColumnType{sqlbase.IntType, sqlbase.IntType, sqlbase.IntType, sqlbase.IntType},
 			expected: sqlbase.EncDatumRows{
 				{null, v[4], null, null},
 				{null, null, null, v[4]},
@@ -400,58 +401,58 @@ func TestMergeJoiner(t *testing.T) {
 		},
 		{
 			// Ensure that NULL = NULL is not matched.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type: sqlbase.InnerJoin,
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: oneIntCol,
+			leftTypes: sqlbase.OneIntCol,
 			leftInput: sqlbase.EncDatumRows{
 				{null},
 				{v[0]},
 			},
-			rightTypes: oneIntCol,
+			rightTypes: sqlbase.OneIntCol,
 			rightInput: sqlbase.EncDatumRows{
 				{null},
 				{v[1]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected:      sqlbase.EncDatumRows{},
 		},
 		{
 			// Ensure that semi joins doesn't output duplicates from
 			// the right side.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type: sqlbase.LeftSemiJoin,
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[1], v[2]},
 				{v[2], v[3]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[2], v[2]},
 				{v[2], v[2]},
 				{v[3], v[3]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[2], v[3]},
 			},
@@ -459,19 +460,19 @@ func TestMergeJoiner(t *testing.T) {
 		{
 			// Ensure that duplicate rows in the left are matched
 			// in the output in semi-joins.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type: sqlbase.LeftSemiJoin,
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[1], v[2]},
 				{v[1], v[2]},
@@ -479,12 +480,12 @@ func TestMergeJoiner(t *testing.T) {
 				{v[3], v[4]},
 				{v[3], v[5]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[2], v[2]},
 				{v[3], v[3]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[2], v[3]},
 				{v[3], v[4]},
@@ -493,51 +494,51 @@ func TestMergeJoiner(t *testing.T) {
 		},
 		{
 			// Ensure that NULL == NULL doesn't match in semi-join.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type: sqlbase.LeftSemiJoin,
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{null, v[2]},
 				{v[2], v[3]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{null, v[3]},
 				{v[2], v[4]},
 				{v[2], v[5]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[2], v[3]},
 			},
 		},
 		{
 			// Ensure that OnExprs are satisfied for semi-joins.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type:   sqlbase.LeftSemiJoin,
-				OnExpr: Expression{Expr: "@1 >= 4"},
+				OnExpr: distsqlpb.Expression{Expr: "@1 >= 4"},
 				// Implicit AND @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
@@ -548,7 +549,7 @@ func TestMergeJoiner(t *testing.T) {
 				{v[6], v[0]},
 				{v[6], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[1]},
@@ -561,7 +562,7 @@ func TestMergeJoiner(t *testing.T) {
 				{v[5], v[5]},
 				{v[5], v[4]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[5], v[0]},
 				{v[5], v[1]},
@@ -570,19 +571,19 @@ func TestMergeJoiner(t *testing.T) {
 		{
 			// Ensure that duplicate rows in the left are matched
 			// in the output in anti-joins.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type: sqlbase.LeftAntiJoin,
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[1], v[2]},
 				{v[1], v[3]},
@@ -590,12 +591,12 @@ func TestMergeJoiner(t *testing.T) {
 				{v[3], v[4]},
 				{v[3], v[5]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[2], v[2]},
 				{v[3], v[3]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[1], v[2]},
 				{v[1], v[3]},
@@ -603,51 +604,51 @@ func TestMergeJoiner(t *testing.T) {
 		},
 		{
 			// Ensure that NULL == NULL doesn't match in anti-join.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type: sqlbase.LeftAntiJoin,
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{null, v[2]},
 				{v[2], v[3]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{null, v[3]},
 				{v[2], v[4]},
 				{v[2], v[5]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected: sqlbase.EncDatumRows{
 				{null, v[2]},
 			},
 		},
 		{
 			// Ensure that OnExprs are satisfied for semi-joins.
-			spec: MergeJoinerSpec{
-				LeftOrdering: convertToSpecOrdering(
+			spec: distsqlpb.MergeJoinerSpec{
+				LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
-				RightOrdering: convertToSpecOrdering(
+				RightOrdering: distsqlpb.ConvertToSpecOrdering(
 					sqlbase.ColumnOrdering{
 						{ColIdx: 0, Direction: encoding.Ascending},
 					}),
 				Type:   sqlbase.LeftAntiJoin,
-				OnExpr: Expression{Expr: "@1 >= 4"},
+				OnExpr: distsqlpb.Expression{Expr: "@1 >= 4"},
 				// Implicit AND @1 = @3 constraint.
 			},
 			outCols:   []uint32{0, 1},
-			leftTypes: twoIntCols,
+			leftTypes: sqlbase.TwoIntCols,
 			leftInput: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
@@ -658,7 +659,7 @@ func TestMergeJoiner(t *testing.T) {
 				{v[6], v[0]},
 				{v[6], v[1]},
 			},
-			rightTypes: twoIntCols,
+			rightTypes: sqlbase.TwoIntCols,
 			rightInput: sqlbase.EncDatumRows{
 				{v[0], v[4]},
 				{v[0], v[1]},
@@ -671,7 +672,7 @@ func TestMergeJoiner(t *testing.T) {
 				{v[5], v[5]},
 				{v[5], v[4]},
 			},
-			expectedTypes: twoIntCols,
+			expectedTypes: sqlbase.TwoIntCols,
 			expected: sqlbase.EncDatumRows{
 				{v[0], v[0]},
 				{v[0], v[1]},
@@ -707,13 +708,13 @@ func TestMergeJoiner(t *testing.T) {
 				EvalCtx:  &evalCtx,
 			}
 
-			post := PostProcessSpec{Projection: true, OutputColumns: c.outCols}
+			post := distsqlpb.PostProcessSpec{Projection: true, OutputColumns: c.outCols}
 			m, err := newMergeJoiner(&flowCtx, 0 /* processorID */, &ms, leftInput, rightInput, &post, out)
 			if err != nil {
 				t.Fatal(err)
 			}
 
-			m.Run(context.Background(), nil /* wg */)
+			m.Run(context.Background())
 
 			if !out.ProducerClosed() {
 				t.Fatalf("output RowReceiver not closed")
@@ -747,12 +748,12 @@ func TestConsumerClosed(t *testing.T) {
 		v[i] = sqlbase.DatumToEncDatum(columnTypeInt, tree.NewDInt(tree.DInt(i)))
 	}
 
-	spec := MergeJoinerSpec{
-		LeftOrdering: convertToSpecOrdering(
+	spec := distsqlpb.MergeJoinerSpec{
+		LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 			sqlbase.ColumnOrdering{
 				{ColIdx: 0, Direction: encoding.Ascending},
 			}),
-		RightOrdering: convertToSpecOrdering(
+		RightOrdering: distsqlpb.ConvertToSpecOrdering(
 			sqlbase.ColumnOrdering{
 				{ColIdx: 0, Direction: encoding.Ascending},
 			}),
@@ -760,8 +761,8 @@ func TestConsumerClosed(t *testing.T) {
 		// Implicit @1 = @2 constraint.
 	}
 	outCols := []uint32{0}
-	leftTypes := oneIntCol
-	rightTypes := oneIntCol
+	leftTypes := sqlbase.OneIntCol
+	rightTypes := sqlbase.OneIntCol
 
 	testCases := []struct {
 		typ       sqlbase.JoinType
@@ -813,13 +814,13 @@ func TestConsumerClosed(t *testing.T) {
 				Settings: st,
 				EvalCtx:  &evalCtx,
 			}
-			post := PostProcessSpec{Projection: true, OutputColumns: outCols}
+			post := distsqlpb.PostProcessSpec{Projection: true, OutputColumns: outCols}
 			m, err := newMergeJoiner(&flowCtx, 0 /* processorID */, &spec, leftInput, rightInput, &post, out)
 			if err != nil {
 				t.Fatal(err)
 			}
 
-			m.Run(context.Background(), nil /* wg */)
+			m.Run(context.Background())
 
 			if !out.ProducerClosed() {
 				t.Fatalf("output RowReceiver not closed")
@@ -838,27 +839,27 @@ func BenchmarkMergeJoiner(b *testing.B) {
 		EvalCtx:  &evalCtx,
 	}
 
-	spec := &MergeJoinerSpec{
-		LeftOrdering: convertToSpecOrdering(
+	spec := &distsqlpb.MergeJoinerSpec{
+		LeftOrdering: distsqlpb.ConvertToSpecOrdering(
 			sqlbase.ColumnOrdering{
 				{ColIdx: 0, Direction: encoding.Ascending},
 			}),
-		RightOrdering: convertToSpecOrdering(
+		RightOrdering: distsqlpb.ConvertToSpecOrdering(
 			sqlbase.ColumnOrdering{
 				{ColIdx: 0, Direction: encoding.Ascending},
 			}),
 		Type: sqlbase.InnerJoin,
 		// Implicit @1 = @2 constraint.
 	}
-	post := &PostProcessSpec{}
+	post := &distsqlpb.PostProcessSpec{}
 	disposer := &RowDisposer{}
 
 	const numCols = 1
 	for _, inputSize := range []int{0, 1 << 2, 1 << 4, 1 << 8, 1 << 12, 1 << 16} {
 		b.Run(fmt.Sprintf("InputSize=%d", inputSize), func(b *testing.B) {
-			rows := makeIntRows(inputSize, numCols)
-			leftInput := NewRepeatableRowSource(oneIntCol, rows)
-			rightInput := NewRepeatableRowSource(oneIntCol, rows)
+			rows := sqlbase.MakeIntRows(inputSize, numCols)
+			leftInput := NewRepeatableRowSource(sqlbase.OneIntCol, rows)
+			rightInput := NewRepeatableRowSource(sqlbase.OneIntCol, rows)
 			b.SetBytes(int64(8 * inputSize * numCols * 2))
 			b.ResetTimer()
 			for i := 0; i < b.N; i++ {
@@ -866,7 +867,7 @@ func BenchmarkMergeJoiner(b *testing.B) {
 				if err != nil {
 					b.Fatal(err)
 				}
-				m.Run(context.Background(), nil /* wg */)
+				m.Run(context.Background())
 				leftInput.Reset()
 				rightInput.Reset()
 			}
@@ -876,8 +877,8 @@ func BenchmarkMergeJoiner(b *testing.B) {
 	for _, inputSize := range []int{0, 1 << 2, 1 << 4, 1 << 8, 1 << 12, 1 << 16} {
 		numRepeats := inputSize
 		b.Run(fmt.Sprintf("OneSideRepeatInputSize=%d", inputSize), func(b *testing.B) {
-			leftInput := NewRepeatableRowSource(oneIntCol, makeIntRows(inputSize, numCols))
-			rightInput := NewRepeatableRowSource(oneIntCol, makeRepeatedIntRows(numRepeats, inputSize, numCols))
+			leftInput := NewRepeatableRowSource(sqlbase.OneIntCol, sqlbase.MakeIntRows(inputSize, numCols))
+			rightInput := NewRepeatableRowSource(sqlbase.OneIntCol, sqlbase.MakeRepeatedIntRows(numRepeats, inputSize, numCols))
 			b.SetBytes(int64(8 * inputSize * numCols * 2))
 			b.ResetTimer()
 			for i := 0; i < b.N; i++ {
@@ -885,7 +886,7 @@ func BenchmarkMergeJoiner(b *testing.B) {
 				if err != nil {
 					b.Fatal(err)
 				}
-				m.Run(context.Background(), nil /* wg */)
+				m.Run(context.Background())
 				leftInput.Reset()
 				rightInput.Reset()
 			}
@@ -895,9 +896,9 @@ func BenchmarkMergeJoiner(b *testing.B) {
 	for _, inputSize := range []int{0, 1 << 2, 1 << 4, 1 << 8, 1 << 12, 1 << 16} {
 		numRepeats := int(math.Sqrt(float64(inputSize)))
 		b.Run(fmt.Sprintf("BothSidesRepeatInputSize=%d", inputSize), func(b *testing.B) {
-			row := makeRepeatedIntRows(100, numRepeats, numCols)
-			leftInput := NewRepeatableRowSource(oneIntCol, row)
-			rightInput := NewRepeatableRowSource(oneIntCol, row)
+			row := sqlbase.MakeRepeatedIntRows(100, numRepeats, numCols)
+			leftInput := NewRepeatableRowSource(sqlbase.OneIntCol, row)
+			rightInput := NewRepeatableRowSource(sqlbase.OneIntCol, row)
 			b.SetBytes(int64(8 * inputSize * numCols * 2))
 			b.ResetTimer()
 			for i := 0; i < b.N; i++ {
@@ -905,7 +906,7 @@ func BenchmarkMergeJoiner(b *testing.B) {
 				if err != nil {
 					b.Fatal(err)
 				}
-				m.Run(context.Background(), nil /* wg */)
+				m.Run(context.Background())
 				leftInput.Reset()
 				rightInput.Reset()
 			}
