@@ -22,17 +22,17 @@ import (
 	"sync"
 	"testing"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
 )
 
 const staticNodeID roachpb.NodeID = 3
@@ -66,10 +66,10 @@ func TestOutbox(t *testing.T) {
 		EvalCtx:    &evalCtx,
 		nodeDialer: nodedialer.New(newInsecureRPCContext(stopper), staticAddressResolver(addr)),
 	}
-	flowID := FlowID{uuid.MakeV4()}
-	streamID := StreamID(42)
-	outbox := newOutbox(&flowCtx, staticNodeID, "", flowID, streamID)
-	outbox.init(oneIntCol)
+	flowID := distsqlpb.FlowID{UUID: uuid.MakeV4()}
+	streamID := distsqlpb.StreamID(42)
+	outbox := newOutbox(&flowCtx, staticNodeID, flowID, streamID)
+	outbox.init(sqlbase.OneIntCol)
 	var outboxWG sync.WaitGroup
 	ctx, cancel := context.WithCancel(context.TODO())
 	defer cancel()
@@ -83,7 +83,7 @@ func TestOutbox(t *testing.T) {
 	go func() {
 		producerC <- func() error {
 			row := sqlbase.EncDatumRow{
-				sqlbase.DatumToEncDatum(intType, tree.NewDInt(tree.DInt(0))),
+				sqlbase.DatumToEncDatum(sqlbase.IntType, tree.NewDInt(tree.DInt(0))),
 			}
 			if consumerStatus := outbox.Push(row, nil /* meta */); consumerStatus != NeedMoreRows {
 				return errors.Errorf("expected status: %d, got: %d", NeedMoreRows, consumerStatus)
@@ -92,7 +92,7 @@ func TestOutbox(t *testing.T) {
 			// Send rows until the drain request is observed.
 			for {
 				row = sqlbase.EncDatumRow{
-					sqlbase.DatumToEncDatum(intType, tree.NewDInt(tree.DInt(-1))),
+					sqlbase.DatumToEncDatum(sqlbase.IntType, tree.NewDInt(tree.DInt(-1))),
 				}
 				consumerStatus := outbox.Push(row, nil /* meta */)
 				if consumerStatus == DrainRequested {
@@ -104,7 +104,7 @@ func TestOutbox(t *testing.T) {
 			}
 
 			// Now send another row that the outbox will discard.
-			row = sqlbase.EncDatumRow{sqlbase.DatumToEncDatum(intType, tree.NewDInt(tree.DInt(2)))}
+			row = sqlbase.EncDatumRow{sqlbase.DatumToEncDatum(sqlbase.IntType, tree.NewDInt(tree.DInt(2)))}
 			if consumerStatus := outbox.Push(row, nil /* meta */); consumerStatus != DrainRequested {
 				return errors.Errorf("expected status: %d, got: %d", NeedMoreRows, consumerStatus)
 			}
@@ -145,12 +145,12 @@ func TestOutbox(t *testing.T) {
 		// about the draining.
 		last := -1
 		for i := 0; i < len(rows); i++ {
-			if rows[i].String(oneIntCol) != "[-1]" {
+			if rows[i].String(sqlbase.OneIntCol) != "[-1]" {
 				last = i
 				continue
 			}
 			for j := i; j < len(rows); j++ {
-				if rows[j].String(oneIntCol) == "[-1]" {
+				if rows[j].String(sqlbase.OneIntCol) == "[-1]" {
 					continue
 				}
 				rows[i] = rows[j]
@@ -163,7 +163,7 @@ func TestOutbox(t *testing.T) {
 
 		// After we receive one row, we're going to ask the producer to drain.
 		if !drainSignalSent && len(rows) > 0 {
-			sig := ConsumerSignal{DrainRequest: &DrainRequest{}}
+			sig := distsqlpb.ConsumerSignal{DrainRequest: &distsqlpb.DrainRequest{}}
 			if err := serverStream.Send(&sig); err != nil {
 				t.Fatal(err)
 			}
@@ -179,11 +179,11 @@ func TestOutbox(t *testing.T) {
 	}
 	for i, m := range metas {
 		expectedStr := fmt.Sprintf("meta %d", i)
-		if m.Err.Error() != expectedStr {
+		if !testutils.IsError(m.Err, expectedStr) {
 			t.Fatalf("expected: %q, got: %q", expectedStr, m.Err.Error())
 		}
 	}
-	str := rows.String(oneIntCol)
+	str := rows.String(sqlbase.OneIntCol)
 	expected := "[[0]]"
 	if str != expected {
 		t.Errorf("invalid results: %s, expected %s'", str, expected)
@@ -217,14 +217,14 @@ func TestOutboxInitializesStreamBeforeReceivingAnyRows(t *testing.T) {
 		EvalCtx:    &evalCtx,
 		nodeDialer: nodedialer.New(newInsecureRPCContext(stopper), staticAddressResolver(addr)),
 	}
-	flowID := FlowID{uuid.MakeV4()}
-	streamID := StreamID(42)
-	outbox := newOutbox(&flowCtx, staticNodeID, "", flowID, streamID)
+	flowID := distsqlpb.FlowID{UUID: uuid.MakeV4()}
+	streamID := distsqlpb.StreamID(42)
+	outbox := newOutbox(&flowCtx, staticNodeID, flowID, streamID)
 
 	var outboxWG sync.WaitGroup
 	ctx, cancel := context.WithCancel(context.TODO())
 	defer cancel()
-	outbox.init(oneIntCol)
+	outbox.init(sqlbase.OneIntCol)
 	// Start the outbox. This should cause the stream to connect, even though
 	// we're not sending any rows.
 	outbox.start(ctx, &outboxWG, cancel)
@@ -286,8 +286,8 @@ func TestOutboxClosesWhenConsumerCloses(t *testing.T) {
 				EvalCtx:    &evalCtx,
 				nodeDialer: nodedialer.New(newInsecureRPCContext(stopper), staticAddressResolver(addr)),
 			}
-			flowID := FlowID{uuid.MakeV4()}
-			streamID := StreamID(42)
+			flowID := distsqlpb.FlowID{UUID: uuid.MakeV4()}
+			streamID := distsqlpb.StreamID(42)
 			var outbox *outbox
 			var wg sync.WaitGroup
 			var expectedErr error
@@ -295,8 +295,8 @@ func TestOutboxClosesWhenConsumerCloses(t *testing.T) {
 			ctx, cancel := context.WithCancel(context.TODO())
 			defer cancel()
 			if tc.outboxIsClient {
-				outbox = newOutbox(&flowCtx, staticNodeID, "", flowID, streamID)
-				outbox.init(oneIntCol)
+				outbox = newOutbox(&flowCtx, staticNodeID, flowID, streamID)
+				outbox.init(sqlbase.OneIntCol)
 				outbox.start(ctx, &wg, cancel)
 
 				// Wait for the outbox to connect the stream.
@@ -324,8 +324,8 @@ func TestOutboxClosesWhenConsumerCloses(t *testing.T) {
 				if err != nil {
 					t.Fatal(err)
 				}
-				client := NewDistSQLClient(conn)
-				var outStream DistSQL_RunSyncFlowClient
+				client := distsqlpb.NewDistSQLClient(conn)
+				var outStream distsqlpb.DistSQL_RunSyncFlowClient
 				ctx, cancel := context.WithCancel(context.Background())
 				defer cancel()
 				expectedErr = errors.Errorf("context canceled")
@@ -357,7 +357,7 @@ func TestOutboxClosesWhenConsumerCloses(t *testing.T) {
 				call := <-mockServer.runSyncFlowCalls
 				outbox = newOutboxSyncFlowStream(call.stream)
 				outbox.setFlowCtx(&FlowCtx{Settings: cluster.MakeTestingClusterSettings(), stopper: stopper})
-				outbox.init(oneIntCol)
+				outbox.init(sqlbase.OneIntCol)
 				// In a RunSyncFlow call, the outbox runs under the call's context.
 				outbox.start(call.stream.Context(), &wg, cancel)
 				// Wait for the consumer to receive the header message that the outbox
@@ -412,8 +412,8 @@ func TestOutboxCancelsFlowOnError(t *testing.T) {
 		EvalCtx:    &evalCtx,
 		nodeDialer: nodedialer.New(newInsecureRPCContext(stopper), staticAddressResolver(addr)),
 	}
-	flowID := FlowID{uuid.MakeV4()}
-	streamID := StreamID(42)
+	flowID := distsqlpb.FlowID{UUID: uuid.MakeV4()}
+	streamID := distsqlpb.StreamID(42)
 	var outbox *outbox
 	var wg sync.WaitGroup
 	ctx, cancel := context.WithCancel(context.TODO())
@@ -426,8 +426,8 @@ func TestOutboxCancelsFlowOnError(t *testing.T) {
 		ctxCanceled = true
 	}
 
-	outbox = newOutbox(&flowCtx, staticNodeID, "", flowID, streamID)
-	outbox.init(oneIntCol)
+	outbox = newOutbox(&flowCtx, staticNodeID, flowID, streamID)
+	outbox.init(sqlbase.OneIntCol)
 	outbox.start(ctx, &wg, mockCancel)
 
 	// Wait for the outbox to connect the stream.
@@ -444,60 +444,62 @@ func TestOutboxCancelsFlowOnError(t *testing.T) {
 	}
 }
 
-// Test that the legacy interface (rpcCtx and addr instead of
-// nodeDialer and nodeID) still works. Copied from
-// TestOutboxCancelsFlowOnError simply because it was the shortest
-// test.
-// TODO(bdarnell): Remove after 2.1.
-func TestOutboxCancelsFlowOnErrorLegacyInterface(t *testing.T) {
+// Test that the outbox unblocks its producers if it fails to connect during
+// startup.
+func TestOutboxUnblocksProducers(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
 	stopper := stop.NewStopper()
-	defer stopper.Stop(context.TODO())
-	mockServer, addr, err := startMockDistSQLServer(stopper)
-	if err != nil {
-		t.Fatal(err)
-	}
+	ctx := context.TODO()
+	defer stopper.Stop(ctx)
 
 	st := cluster.MakeTestingClusterSettings()
 	evalCtx := tree.MakeTestingEvalContext(st)
-	defer evalCtx.Stop(context.Background())
+	defer evalCtx.Stop(ctx)
 	flowCtx := FlowCtx{
 		Settings: st,
 		stopper:  stopper,
 		EvalCtx:  &evalCtx,
-		rpcCtx:   newInsecureRPCContext(stopper),
+		// a nil nodeDialer will always fail to connect.
+		nodeDialer: nil,
 	}
-	flowID := FlowID{uuid.MakeV4()}
-	streamID := StreamID(42)
+	flowID := distsqlpb.FlowID{UUID: uuid.MakeV4()}
+	streamID := distsqlpb.StreamID(42)
 	var outbox *outbox
 	var wg sync.WaitGroup
-	ctx, cancel := context.WithCancel(context.TODO())
+	ctx, cancel := context.WithCancel(ctx)
 	defer cancel()
 
-	// We could test this on ctx.cancel(), but this mock
-	// cancellation method is simpler.
-	ctxCanceled := false
-	mockCancel := func() {
-		ctxCanceled = true
+	outbox = newOutbox(&flowCtx, staticNodeID, flowID, streamID)
+	outbox.init(sqlbase.OneIntCol)
+
+	// Fill up the outbox.
+	for i := 0; i < outboxBufRows; i++ {
+		outbox.Push(nil, &ProducerMetadata{})
 	}
 
-	outbox = newOutbox(&flowCtx, 0, addr.String(), flowID, streamID)
-	outbox.init(oneIntCol)
-	outbox.start(ctx, &wg, mockCancel)
+	var blockedPusherWg sync.WaitGroup
+	blockedPusherWg.Add(1)
+	go func() {
+		// Push to the outbox one last time, which will block since the channel
+		// is full.
+		outbox.Push(nil, &ProducerMetadata{})
+		// We should become unblocked once outbox.start fails.
+		blockedPusherWg.Done()
+	}()
 
-	// Wait for the outbox to connect the stream.
-	streamNotification := <-mockServer.inboundStreams
-	if _, err := streamNotification.stream.Recv(); err != nil {
-		t.Fatal(err)
-	}
-
-	streamNotification.donec <- sqlbase.QueryCanceledError
+	// This outbox will fail to connect, because it has a nil nodeDialer.
+	outbox.start(ctx, &wg, cancel)
 
 	wg.Wait()
-	if !ctxCanceled {
-		t.Fatal("flow ctx was not canceled")
+	// Also, make sure that pushing to the outbox after its failed shows that
+	// it's been correctly ConsumerClosed.
+	status := outbox.RowChannel.Push(nil, &ProducerMetadata{})
+	if status != ConsumerClosed {
+		t.Fatalf("expected status=ConsumerClosed, got %s", status)
 	}
+
+	blockedPusherWg.Wait()
 }
 
 func BenchmarkOutbox(b *testing.B) {
@@ -514,11 +516,11 @@ func BenchmarkOutbox(b *testing.B) {
 	for _, numCols := range []int{1, 2, 4, 8} {
 		row := sqlbase.EncDatumRow{}
 		for i := 0; i < numCols; i++ {
-			row = append(row, sqlbase.DatumToEncDatum(intType, tree.NewDInt(tree.DInt(2))))
+			row = append(row, sqlbase.DatumToEncDatum(sqlbase.IntType, tree.NewDInt(tree.DInt(2))))
 		}
 		b.Run(fmt.Sprintf("numCols=%d", numCols), func(b *testing.B) {
-			flowID := FlowID{uuid.MakeV4()}
-			streamID := StreamID(42)
+			flowID := distsqlpb.FlowID{UUID: uuid.MakeV4()}
+			streamID := distsqlpb.StreamID(42)
 			evalCtx := tree.MakeTestingEvalContext(st)
 			defer evalCtx.Stop(context.Background())
 			flowCtx := FlowCtx{
@@ -527,8 +529,8 @@ func BenchmarkOutbox(b *testing.B) {
 				EvalCtx:    &evalCtx,
 				nodeDialer: nodedialer.New(newInsecureRPCContext(stopper), staticAddressResolver(addr)),
 			}
-			outbox := newOutbox(&flowCtx, staticNodeID, "", flowID, streamID)
-			outbox.init(makeIntCols(numCols))
+			outbox := newOutbox(&flowCtx, staticNodeID, flowID, streamID)
+			outbox.init(sqlbase.MakeIntCols(numCols))
 			var outboxWG sync.WaitGroup
 			ctx, cancel := context.WithCancel(context.TODO())
 			defer cancel()
