@@ -17,14 +17,13 @@ package distsqlrun
 import (
 	"context"
 	"math"
-	"sync"
 	"sync/atomic"
 
-	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util"
-	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
@@ -32,9 +31,6 @@ import (
 	"github.com/pkg/errors"
 )
 
-// processorIDTagKey is the key used for processor id tags in tracing spans.
-const processorIDTagKey = tracing.TagPrefix + "processorid"
-
 // Processor is a common interface implemented by all processors, used by the
 // higher-level flow orchestration code.
 type Processor interface {
@@ -43,8 +39,7 @@ type Processor interface {
 	OutputTypes() []sqlbase.ColumnType
 
 	// Run is the main loop of the processor.
-	// If wg is non-nil, wg.Done is called before exiting.
-	Run(_ context.Context, wg *sync.WaitGroup)
+	Run(context.Context)
 }
 
 // ProcOutputHelper is a helper type that performs filtering and projection on
@@ -59,11 +54,12 @@ type ProcOutputHelper struct {
 	rowAlloc sqlbase.EncDatumRowAlloc
 
 	filter *exprHelper
-	// renderExprs is set if we have a rendering. Only one of renderExprs and
-	// outputCols can be set.
+	// renderExprs has length > 0 if we have a rendering. Only one of renderExprs
+	// and outputCols can be set.
 	renderExprs []exprHelper
-	// outputCols is set if we have a projection. Only one of renderExprs and
-	// outputCols can be set.
+	// outputCols is non-nil if we have a projection. Only one of renderExprs and
+	// outputCols can be set. Note that 0-length projections are possible, in
+	// which case outputCols will be 0-length but non-nil.
 	outputCols []uint32
 
 	outputRow sqlbase.EncDatumRow
@@ -87,13 +83,24 @@ type ProcOutputHelper struct {
 	rowIdx uint64
 }
 
+// Reset resets this ProcOutputHelper, retaining allocated memory in its slices.
+func (h *ProcOutputHelper) Reset() {
+	*h = ProcOutputHelper{
+		renderExprs: h.renderExprs[:0],
+		outputTypes: h.outputTypes[:0],
+	}
+}
+
 // Init sets up a ProcOutputHelper. The types describe the internal schema of
 // the processor (as described for each processor core spec); they can be
 // omitted if there is no filtering expression.
 // Note that the types slice may be stored directly; the caller should not
 // modify it.
 func (h *ProcOutputHelper) Init(
-	post *PostProcessSpec, types []sqlbase.ColumnType, evalCtx *tree.EvalContext, output RowReceiver,
+	post *distsqlpb.PostProcessSpec,
+	types []sqlbase.ColumnType,
+	evalCtx *tree.EvalContext,
+	output RowReceiver,
 ) error {
 	if !post.Projection && len(post.OutputColumns) > 0 {
 		return errors.Errorf("post-processing has projection unset but output columns set: %s", post)
@@ -103,7 +110,7 @@ func (h *ProcOutputHelper) Init(
 	}
 	h.output = output
 	h.numInternalCols = len(types)
-	if post.Filter.Expr != "" {
+	if post.Filter != (distsqlpb.Expression{}) {
 		h.filter = &exprHelper{}
 		if err := h.filter.init(post.Filter, types, evalCtx); err != nil {
 			return err
@@ -120,14 +127,28 @@ func (h *ProcOutputHelper) Init(
 			// nil indicates no projection; use an empty slice.
 			h.outputCols = make([]uint32, 0)
 		}
-		h.outputTypes = make([]sqlbase.ColumnType, len(h.outputCols))
+		nOutputCols := len(h.outputCols)
+		if cap(h.outputTypes) >= nOutputCols {
+			h.outputTypes = h.outputTypes[:nOutputCols]
+		} else {
+			h.outputTypes = make([]sqlbase.ColumnType, nOutputCols)
+		}
 		for i, c := range h.outputCols {
 			h.outputTypes[i] = types[c]
 		}
-	} else if len(post.RenderExprs) > 0 {
-		h.renderExprs = make([]exprHelper, len(post.RenderExprs))
-		h.outputTypes = make([]sqlbase.ColumnType, len(post.RenderExprs))
+	} else if nRenders := len(post.RenderExprs); nRenders > 0 {
+		if cap(h.renderExprs) >= nRenders {
+			h.renderExprs = h.renderExprs[:nRenders]
+		} else {
+			h.renderExprs = make([]exprHelper, nRenders)
+		}
+		if cap(h.outputTypes) >= nRenders {
+			h.outputTypes = h.outputTypes[:nRenders]
+		} else {
+			h.outputTypes = make([]sqlbase.ColumnType, nRenders)
+		}
 		for i, expr := range post.RenderExprs {
+			h.renderExprs[i] = exprHelper{}
 			if err := h.renderExprs[i].init(expr, types, evalCtx); err != nil {
 				return err
 			}
@@ -138,9 +159,15 @@ func (h *ProcOutputHelper) Init(
 			h.outputTypes[i] = colTyp
 		}
 	} else {
-		h.outputTypes = types
+		// No rendering or projection.
+		if cap(h.outputTypes) >= len(types) {
+			h.outputTypes = h.outputTypes[:len(types)]
+		} else {
+			h.outputTypes = make([]sqlbase.ColumnType, len(types))
+		}
+		copy(h.outputTypes, types)
 	}
-	if h.outputCols != nil || h.renderExprs != nil {
+	if h.outputCols != nil || len(h.renderExprs) > 0 {
 		// We're rendering or projecting, so allocate an output row.
 		h.outputRow = h.rowAlloc.AllocRow(len(h.outputTypes))
 	}
@@ -158,7 +185,7 @@ func (h *ProcOutputHelper) Init(
 // neededColumns calculates the set of internal processor columns that are
 // actually used by the post-processing stage.
 func (h *ProcOutputHelper) neededColumns() (colIdxs util.FastIntSet) {
-	if h.outputCols == nil && h.renderExprs == nil {
+	if h.outputCols == nil && len(h.renderExprs) == 0 {
 		// No projection or rendering; all columns are needed.
 		colIdxs.AddRange(0, h.numInternalCols-1)
 		return colIdxs
@@ -177,12 +204,10 @@ func (h *ProcOutputHelper) neededColumns() (colIdxs util.FastIntSet) {
 		}
 
 		// See if render expressions require this column.
-		if h.renderExprs != nil {
-			for j := range h.renderExprs {
-				if h.renderExprs[j].vars.IndexedVarUsed(i) {
-					colIdxs.Add(i)
-					break
-				}
+		for j := range h.renderExprs {
+			if h.renderExprs[j].vars.IndexedVarUsed(i) {
+				colIdxs.Add(i)
+				break
 			}
 		}
 	}
@@ -230,16 +255,15 @@ func emitHelper(
 		}
 		// Bypass EmitRow() and send directly to output.output.
 		consumerStatus = output.output.Push(nil /* row */, meta)
+		if meta.Err != nil {
+			consumerStatus = ConsumerClosed
+		}
 	} else {
 		var err error
 		consumerStatus, err = output.EmitRow(ctx, row)
 		if err != nil {
 			output.output.Push(nil /* row */, &ProducerMetadata{Err: err})
-			for _, input := range inputs {
-				input.ConsumerClosed()
-			}
-			output.Close()
-			return false
+			consumerStatus = ConsumerClosed
 		}
 	}
 	switch consumerStatus {
@@ -345,7 +369,7 @@ func (h *ProcOutputHelper) ProcessRow(
 		return nil, true, nil
 	}
 
-	if h.renderExprs != nil {
+	if len(h.renderExprs) > 0 {
 		// Rendering.
 		for i := range h.renderExprs {
 			datum, err := h.renderExprs[i].eval(row)
@@ -539,6 +563,16 @@ type ProcessorBase struct {
 	inputsToDrain []RowSource
 }
 
+// Reset resets this ProcessorBase, retaining allocated memory in slices.
+func (pb *ProcessorBase) Reset() {
+	pb.out.Reset()
+	*pb = ProcessorBase{
+		out:           pb.out,
+		trailingMeta:  pb.trailingMeta[:0],
+		inputsToDrain: pb.inputsToDrain[:0],
+	}
+}
+
 // procState represents the standard states that a processor can be in. These
 // states are relevant when the processor is using the draining utilities in
 // ProcessorBase.
@@ -559,6 +593,29 @@ const (
 	// stateDraining is the state in which the processor is forwarding metadata
 	// from its input and otherwise ignoring all rows. Once the input is
 	// exhausted, the processor will transition to stateTrailingMeta.
+	//
+	// In stateDraining, processors are required to swallow
+	// ReadWithinUncertaintyIntervalErrors received from its sources. We're
+	// already draining, so we don't care about whatever data generated this
+	// uncertainty error. Besides generally seeming like a good idea, doing this
+	// allows us to offer a nice guarantee to SQL clients: a read-only query that
+	// produces at most one row, run as an implicit txn, never produces retriable
+	// errors, regardless of the size of the row being returned (in relation to
+	// the size of the result buffer on the connection). One would naively expect
+	// that to be true: either the error happens before any rows have been
+	// delivered to the client, in which case the auto-retries kick in, or, if a
+	// row has been delivered, then the query is done and so how can there be an
+	// error? What our naive friend is ignoring is that, if it weren't for this
+	// code, it'd be possible for a retriable error to sneak in after the query's
+	// limit has been satisfied but while processors are still draining. Note
+	// that uncertainty errors are not retried automatically by the leaf
+	// TxnCoordSenders (i.e. by refresh txn interceptor).
+	//
+	// Other categories of errors might be safe to ignore too; however we
+	// can't ignore all of them. Generally, we need to ensure that all the
+	// trailing metadata (e.g. TxnCoordMeta's) make it to the gateway for
+	// successful flows. If an error is telling us that some metadata might
+	// have been dropped, we can't ignore that.
 	stateDraining
 
 	// stateTrailingMeta is the state in which the processor is outputting final
@@ -643,6 +700,19 @@ func (pb *ProcessorBase) DrainHelper() *ProducerMetadata {
 			continue
 		}
 		if meta != nil {
+			// Swallow ReadWithinUncertaintyIntervalErrors. See comments on
+			// stateDraining.
+			if err := meta.Err; err != nil {
+				// We only look for UnhandledRetryableErrors. Local reads (which would
+				// be transformed by the Root TxnCoordSender into
+				// TransactionRetryWithProtoRefreshErrors) don't have any uncertainty.
+				if ure, ok := err.(*roachpb.UnhandledRetryableError); ok {
+					uncertain := ure.PErr.Detail.GetReadWithinUncertaintyInterval()
+					if uncertain != nil {
+						continue
+					}
+				}
+			}
 			return meta
 		}
 	}
@@ -723,15 +793,12 @@ func (pb *ProcessorBase) OutputTypes() []sqlbase.ColumnType {
 }
 
 // Run is part of the processor interface.
-func (pb *ProcessorBase) Run(ctx context.Context, wg *sync.WaitGroup) {
+func (pb *ProcessorBase) Run(ctx context.Context) {
 	if pb.out.output == nil {
 		panic("processor output not initialized for emitting rows")
 	}
 	ctx = pb.self.Start(ctx)
 	Run(ctx, pb.self, pb.out.output)
-	if wg != nil {
-		wg.Done()
-	}
 }
 
 // ProcStateOpts contains fields used by the ProcessorBase's family of functions
@@ -750,7 +817,7 @@ type ProcStateOpts struct {
 // Init initializes the ProcessorBase.
 func (pb *ProcessorBase) Init(
 	self RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	types []sqlbase.ColumnType,
 	flowCtx *FlowCtx,
 	processorID int32,
@@ -766,7 +833,7 @@ func (pb *ProcessorBase) Init(
 // InitWithEvalCtx initializes the ProcessorBase with a given EvalContext.
 func (pb *ProcessorBase) InitWithEvalCtx(
 	self RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	types []sqlbase.ColumnType,
 	flowCtx *FlowCtx,
 	evalCtx *tree.EvalContext,
@@ -843,6 +910,11 @@ func (pb *ProcessorBase) InternalClose() bool {
 	return closing
 }
 
+// ConsumerDone is part of the RowSource interface.
+func (pb *ProcessorBase) ConsumerDone() {
+	pb.MoveToDraining(nil /* err */)
+}
+
 // NewMonitor is a utility function used by processors to create a new
 // memory monitor with the given name and start it. The returned monitor must
 // be closed.
@@ -891,7 +963,7 @@ func (rb *rowSourceBase) consumerDone() {
 func (rb *rowSourceBase) consumerClosed(name string) {
 	status := ConsumerStatus(atomic.LoadUint32((*uint32)(&rb.consumerStatus)))
 	if status == ConsumerClosed {
-		log.Fatalf(context.Background(), "%s already closed", name)
+		log.ReportOrPanic(context.Background(), nil, "%s already closed", log.Safe(name))
 	}
 	atomic.StoreUint32((*uint32)(&rb.consumerStatus), uint32(ConsumerClosed))
 }
@@ -906,8 +978,8 @@ func newProcessor(
 	ctx context.Context,
 	flowCtx *FlowCtx,
 	processorID int32,
-	core *ProcessorCoreUnion,
-	post *PostProcessSpec,
+	core *distsqlpb.ProcessorCoreUnion,
+	post *distsqlpb.PostProcessSpec,
 	inputs []RowSource,
 	outputs []RowReceiver,
 	localProcessors []LocalProcessor,
@@ -977,6 +1049,14 @@ func newProcessor(
 			flowCtx, processorID, core.InterleavedReaderJoiner, post, outputs[0],
 		)
 	}
+	if core.ZigzagJoiner != nil {
+		if err := checkNumInOut(inputs, outputs, 0, 1); err != nil {
+			return nil, err
+		}
+		return newZigzagJoiner(
+			flowCtx, processorID, core.ZigzagJoiner, nil, post, outputs[0],
+		)
+	}
 	if core.HashJoiner != nil {
 		if err := checkNumInOut(inputs, outputs, 2, 1); err != nil {
 			return nil, err
@@ -988,9 +1068,9 @@ func newProcessor(
 			return nil, err
 		}
 		switch core.Backfiller.Type {
-		case BackfillerSpec_Index:
+		case distsqlpb.BackfillerSpec_Index:
 			return newIndexBackfiller(flowCtx, processorID, *core.Backfiller, post, outputs[0])
-		case BackfillerSpec_Column:
+		case distsqlpb.BackfillerSpec_Column:
 			return newColumnBackfiller(flowCtx, processorID, *core.Backfiller, post, outputs[0])
 		}
 	}
@@ -1107,254 +1187,41 @@ func newProcessor(
 type LocalProcessor interface {
 	RowSourcedProcessor
 	// InitWithOutput initializes this processor.
-	InitWithOutput(post *PostProcessSpec, output RowReceiver) error
+	InitWithOutput(post *distsqlpb.PostProcessSpec, output RowReceiver) error
 	// SetInput initializes this LocalProcessor with an input RowSource. Not all
 	// LocalProcessors need inputs, but this needs to be called if a
 	// LocalProcessor expects to get its data from another RowSource.
 	SetInput(ctx context.Context, input RowSource) error
 }
 
+// vectorizeAlwaysException is an object that returns whether or not execution
+// should continue if experimental_vectorize=always and an error occurred when
+// setting up the vectorized flow. Consider the case in which
+// experimental_vectorize=always. The user must be able to unset this session
+// variable without getting an error.
+type vectorizeAlwaysException interface {
+	// IsException returns whether this object should be an exception to the rule
+	// that an inability to run this node in a vectorized flow should produce an
+	// error.
+	// TODO(asubiotto): This is the cleanest way I can think of to not error out
+	// on SET statements when running with experimental_vectorize = always. If
+	// there is a better way, we should get rid of this interface.
+	IsException() bool
+}
+
 // NewReadImportDataProcessor is externally implemented and registered by
 // ccl/sqlccl/csv.go.
-var NewReadImportDataProcessor func(*FlowCtx, int32, ReadImportDataSpec, RowReceiver) (Processor, error)
+var NewReadImportDataProcessor func(*FlowCtx, int32, distsqlpb.ReadImportDataSpec, RowReceiver) (Processor, error)
 
 // NewSSTWriterProcessor is externally implemented and registered by
 // ccl/sqlccl/csv.go.
-var NewSSTWriterProcessor func(*FlowCtx, int32, SSTWriterSpec, RowSource, RowReceiver) (Processor, error)
+var NewSSTWriterProcessor func(*FlowCtx, int32, distsqlpb.SSTWriterSpec, RowSource, RowReceiver) (Processor, error)
 
 // NewCSVWriterProcessor is externally implemented.
-var NewCSVWriterProcessor func(*FlowCtx, int32, CSVWriterSpec, RowSource, RowReceiver) (Processor, error)
+var NewCSVWriterProcessor func(*FlowCtx, int32, distsqlpb.CSVWriterSpec, RowSource, RowReceiver) (Processor, error)
 
 // NewChangeAggregatorProcessor is externally implemented.
-var NewChangeAggregatorProcessor func(*FlowCtx, int32, ChangeAggregatorSpec, RowReceiver) (Processor, error)
+var NewChangeAggregatorProcessor func(*FlowCtx, int32, distsqlpb.ChangeAggregatorSpec, RowReceiver) (Processor, error)
 
 // NewChangeFrontierProcessor is externally implemented.
-var NewChangeFrontierProcessor func(*FlowCtx, int32, ChangeFrontierSpec, RowSource, RowReceiver) (Processor, error)
-
-// Equals returns true if two aggregation specifiers are identical (and thus
-// will always yield the same result).
-func (a AggregatorSpec_Aggregation) Equals(b AggregatorSpec_Aggregation) bool {
-	if a.Func != b.Func || a.Distinct != b.Distinct {
-		return false
-	}
-	if a.FilterColIdx == nil {
-		if b.FilterColIdx != nil {
-			return false
-		}
-	} else {
-		if b.FilterColIdx == nil || *a.FilterColIdx != *b.FilterColIdx {
-			return false
-		}
-	}
-	if len(a.ColIdx) != len(b.ColIdx) {
-		return false
-	}
-	for i, c := range a.ColIdx {
-		if c != b.ColIdx[i] {
-			return false
-		}
-	}
-	return true
-}
-
-func (spec *WindowerSpec_Frame_Mode) initFromAST(w tree.WindowFrameMode) {
-	switch w {
-	case tree.RANGE:
-		*spec = WindowerSpec_Frame_RANGE
-	case tree.ROWS:
-		*spec = WindowerSpec_Frame_ROWS
-	case tree.GROUPS:
-		*spec = WindowerSpec_Frame_GROUPS
-	default:
-		panic("unexpected WindowFrameMode")
-	}
-}
-
-func (spec *WindowerSpec_Frame_BoundType) initFromAST(bt tree.WindowFrameBoundType) {
-	switch bt {
-	case tree.UnboundedPreceding:
-		*spec = WindowerSpec_Frame_UNBOUNDED_PRECEDING
-	case tree.OffsetPreceding:
-		*spec = WindowerSpec_Frame_OFFSET_PRECEDING
-	case tree.CurrentRow:
-		*spec = WindowerSpec_Frame_CURRENT_ROW
-	case tree.OffsetFollowing:
-		*spec = WindowerSpec_Frame_OFFSET_FOLLOWING
-	case tree.UnboundedFollowing:
-		*spec = WindowerSpec_Frame_UNBOUNDED_FOLLOWING
-	default:
-		panic("unexpected WindowFrameBoundType")
-	}
-}
-
-// If offset exprs are present, we evaluate them and save the encoded results
-// in the spec.
-func (spec *WindowerSpec_Frame_Bounds) initFromAST(
-	b tree.WindowFrameBounds, m tree.WindowFrameMode, evalCtx *tree.EvalContext,
-) error {
-	if b.StartBound == nil {
-		return errors.Errorf("unexpected: Start Bound is nil")
-	}
-	spec.Start = WindowerSpec_Frame_Bound{}
-	spec.Start.BoundType.initFromAST(b.StartBound.BoundType)
-	if b.StartBound.HasOffset() {
-		typedStartOffset := b.StartBound.OffsetExpr.(tree.TypedExpr)
-		dStartOffset, err := typedStartOffset.Eval(evalCtx)
-		if err != nil {
-			return err
-		}
-		if dStartOffset == tree.DNull {
-			return pgerror.NewErrorf(pgerror.CodeNullValueNotAllowedError, "frame starting offset must not be null")
-		}
-		switch m {
-		case tree.ROWS:
-			startOffset := int(tree.MustBeDInt(dStartOffset))
-			if startOffset < 0 {
-				return pgerror.NewErrorf(pgerror.CodeInvalidWindowFrameOffsetError, "frame starting offset must not be negative")
-			}
-			spec.Start.IntOffset = uint32(startOffset)
-		case tree.RANGE:
-			if isNegative(evalCtx, dStartOffset) {
-				return pgerror.NewErrorf(pgerror.CodeInvalidWindowFrameOffsetError, "invalid preceding or following size in window function")
-			}
-			typ, err := sqlbase.DatumTypeToColumnType(dStartOffset.ResolvedType())
-			if err != nil {
-				return err
-			}
-			spec.Start.OffsetType = DatumInfo{Encoding: sqlbase.DatumEncoding_VALUE, Type: typ}
-			var buf []byte
-			var a sqlbase.DatumAlloc
-			datum := sqlbase.DatumToEncDatum(typ, dStartOffset)
-			buf, err = datum.Encode(&typ, &a, sqlbase.DatumEncoding_VALUE, buf)
-			if err != nil {
-				return err
-			}
-			spec.Start.TypedOffset = buf
-		case tree.GROUPS:
-			startOffset := int(tree.MustBeDInt(dStartOffset))
-			if startOffset < 0 {
-				return pgerror.NewErrorf(pgerror.CodeInvalidWindowFrameOffsetError, "frame starting offset must not be negative")
-			}
-			spec.Start.IntOffset = uint32(startOffset)
-		}
-	}
-
-	if b.EndBound != nil {
-		spec.End = &WindowerSpec_Frame_Bound{}
-		spec.End.BoundType.initFromAST(b.EndBound.BoundType)
-		if b.EndBound.HasOffset() {
-			typedEndOffset := b.EndBound.OffsetExpr.(tree.TypedExpr)
-			dEndOffset, err := typedEndOffset.Eval(evalCtx)
-			if err != nil {
-				return err
-			}
-			if dEndOffset == tree.DNull {
-				return pgerror.NewErrorf(pgerror.CodeNullValueNotAllowedError, "frame ending offset must not be null")
-			}
-			switch m {
-			case tree.ROWS:
-				endOffset := int(tree.MustBeDInt(dEndOffset))
-				if endOffset < 0 {
-					return pgerror.NewErrorf(pgerror.CodeInvalidWindowFrameOffsetError, "frame ending offset must not be negative")
-				}
-				spec.End.IntOffset = uint32(endOffset)
-			case tree.RANGE:
-				if isNegative(evalCtx, dEndOffset) {
-					return pgerror.NewErrorf(pgerror.CodeInvalidWindowFrameOffsetError, "invalid preceding or following size in window function")
-				}
-				typ, err := sqlbase.DatumTypeToColumnType(dEndOffset.ResolvedType())
-				if err != nil {
-					return err
-				}
-				spec.End.OffsetType = DatumInfo{Encoding: sqlbase.DatumEncoding_VALUE, Type: typ}
-				var buf []byte
-				var a sqlbase.DatumAlloc
-				datum := sqlbase.DatumToEncDatum(typ, dEndOffset)
-				buf, err = datum.Encode(&typ, &a, sqlbase.DatumEncoding_VALUE, buf)
-				if err != nil {
-					return err
-				}
-				spec.End.TypedOffset = buf
-			case tree.GROUPS:
-				endOffset := int(tree.MustBeDInt(dEndOffset))
-				if endOffset < 0 {
-					return pgerror.NewErrorf(pgerror.CodeInvalidWindowFrameOffsetError, "frame ending offset must not be negative")
-				}
-				spec.End.IntOffset = uint32(endOffset)
-			}
-		}
-	}
-
-	return nil
-}
-
-// isNegative returns whether offset is negative.
-func isNegative(evalCtx *tree.EvalContext, offset tree.Datum) bool {
-	switch o := offset.(type) {
-	case *tree.DInt:
-		return *o < 0
-	case *tree.DDecimal:
-		return o.Negative
-	case *tree.DFloat:
-		return *o < 0
-	case *tree.DInterval:
-		return o.Compare(evalCtx, &tree.DInterval{Duration: duration.Duration{}}) < 0
-	default:
-		panic("unexpected offset type")
-	}
-}
-
-// InitFromAST initializes the spec based on tree.WindowFrame. It will evaluate
-// offset expressions if present in the frame.
-func (spec *WindowerSpec_Frame) InitFromAST(f *tree.WindowFrame, evalCtx *tree.EvalContext) error {
-	spec.Mode.initFromAST(f.Mode)
-	return spec.Bounds.initFromAST(f.Bounds, f.Mode, evalCtx)
-}
-
-func (spec WindowerSpec_Frame_Mode) convertToAST() tree.WindowFrameMode {
-	switch spec {
-	case WindowerSpec_Frame_RANGE:
-		return tree.RANGE
-	case WindowerSpec_Frame_ROWS:
-		return tree.ROWS
-	case WindowerSpec_Frame_GROUPS:
-		return tree.GROUPS
-	default:
-		panic("unexpected WindowerSpec_Frame_Mode")
-	}
-}
-
-func (spec WindowerSpec_Frame_BoundType) convertToAST() tree.WindowFrameBoundType {
-	switch spec {
-	case WindowerSpec_Frame_UNBOUNDED_PRECEDING:
-		return tree.UnboundedPreceding
-	case WindowerSpec_Frame_OFFSET_PRECEDING:
-		return tree.OffsetPreceding
-	case WindowerSpec_Frame_CURRENT_ROW:
-		return tree.CurrentRow
-	case WindowerSpec_Frame_OFFSET_FOLLOWING:
-		return tree.OffsetFollowing
-	case WindowerSpec_Frame_UNBOUNDED_FOLLOWING:
-		return tree.UnboundedFollowing
-	default:
-		panic("unexpected WindowerSpec_Frame_BoundType")
-	}
-}
-
-// convertToAST produces tree.WindowFrameBounds based on
-// WindowerSpec_Frame_Bounds. Note that it might not be fully equivalent to
-// original - if offsetExprs were present in original tree.WindowFrameBounds,
-// they are not included.
-func (spec WindowerSpec_Frame_Bounds) convertToAST() tree.WindowFrameBounds {
-	bounds := tree.WindowFrameBounds{StartBound: &tree.WindowFrameBound{
-		BoundType: spec.Start.BoundType.convertToAST(),
-	}}
-	if spec.End != nil {
-		bounds.EndBound = &tree.WindowFrameBound{BoundType: spec.End.BoundType.convertToAST()}
-	}
-	return bounds
-}
-
-func (spec *WindowerSpec_Frame) convertToAST() *tree.WindowFrame {
-	return &tree.WindowFrame{Mode: spec.Mode.convertToAST(), Bounds: spec.Bounds.convertToAST()}
-}
+var NewChangeFrontierProcessor func(*FlowCtx, int32, distsqlpb.ChangeFrontierSpec, RowSource, RowReceiver) (Processor, error)
