@@ -27,14 +27,15 @@ import (
 	"sync"
 	"sync/atomic"
 
-	"github.com/pkg/errors"
-
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
 type router interface {
@@ -48,7 +49,7 @@ type router interface {
 //
 // Pass-through routers are not supported; the higher layer is expected to elide
 // them.
-func makeRouter(spec *OutputRouterSpec, streams []RowReceiver) (router, error) {
+func makeRouter(spec *distsqlpb.OutputRouterSpec, streams []RowReceiver) (router, error) {
 	if len(streams) == 0 {
 		return nil, errors.Errorf("no streams in router")
 	}
@@ -57,13 +58,13 @@ func makeRouter(spec *OutputRouterSpec, streams []RowReceiver) (router, error) {
 	rb.setupStreams(spec, streams)
 
 	switch spec.Type {
-	case OutputRouterSpec_BY_HASH:
+	case distsqlpb.OutputRouterSpec_BY_HASH:
 		return makeHashRouter(rb, spec.HashColumns)
 
-	case OutputRouterSpec_MIRROR:
+	case distsqlpb.OutputRouterSpec_MIRROR:
 		return makeMirrorRouter(rb)
 
-	case OutputRouterSpec_BY_RANGE:
+	case distsqlpb.OutputRouterSpec_BY_RANGE:
 		return makeRangeRouter(rb, spec.RangeRouterSpec)
 
 	default:
@@ -76,7 +77,7 @@ const routerRowBufSize = rowChannelBufSize
 // routerOutput is the data associated with one router consumer.
 type routerOutput struct {
 	stream   RowReceiver
-	streamID StreamID
+	streamID distsqlpb.StreamID
 	mu       struct {
 		syncutil.Mutex
 		// cond is signaled whenever the main router routine adds a metadata item, a
@@ -95,7 +96,7 @@ type routerOutput struct {
 		// The "level 2" rowContainer is used when we need to buffer more rows than
 		// rowBuf allows. The row container always contains rows "older" than those
 		// in rowBuf. The oldest rows are at the beginning of the row container.
-		rowContainer diskBackedRowContainer
+		rowContainer rowcontainer.DiskBackedRowContainer
 		producerDone bool
 	}
 	// TODO(radu): add padding of size sys.CacheLineSize to ensure there is no
@@ -210,7 +211,7 @@ func (rb *routerBase) aggStatus() ConsumerStatus {
 	return ConsumerStatus(atomic.LoadUint32(&rb.aggregatedStatus))
 }
 
-func (rb *routerBase) setupStreams(spec *OutputRouterSpec, streams []RowReceiver) {
+func (rb *routerBase) setupStreams(spec *distsqlpb.OutputRouterSpec, streams []RowReceiver) {
 	rb.numNonDrainingStreams = int32(len(streams))
 	n := len(streams)
 	if spec.DisableBuffering {
@@ -253,13 +254,14 @@ func (rb *routerBase) init(ctx context.Context, flowCtx *FlowCtx, types []sqlbas
 			diskMonitor = NewMonitor(ctx, diskMonitor, diskMonitorName)
 		}
 
-		rb.outputs[i].mu.rowContainer.init(
+		rb.outputs[i].mu.rowContainer.Init(
 			nil, /* ordering */
 			types,
 			evalCtx,
 			flowCtx.TempStorage,
 			memoryMonitor,
 			diskMonitor,
+			0, /* rowCapacity */
 		)
 		rb.outputs[i].memoryMonitor = memoryMonitor
 		rb.outputs[i].diskMonitor = diskMonitor
@@ -279,7 +281,7 @@ func (rb *routerBase) start(ctx context.Context, wg *sync.WaitGroup, ctxCancel c
 			var span opentracing.Span
 			if rb.statsCollectionEnabled {
 				ctx, span = processorSpan(ctx, "router output")
-				span.SetTag(streamIDTagKey, ro.streamID)
+				span.SetTag(distsqlpb.StreamIDTagKey, ro.streamID)
 			}
 
 			drain := false
@@ -381,6 +383,10 @@ func (rb *routerBase) ProducerDone() {
 	}
 }
 
+func (rb *routerBase) Types() []sqlbase.ColumnType {
+	return rb.types
+}
+
 // updateStreamState updates the status of one stream and, if this was the last
 // open stream, it also updates rb.aggregatedStatus.
 func (rb *routerBase) updateStreamState(streamStatus *ConsumerStatus, newState ConsumerStatus) {
@@ -461,8 +467,8 @@ type rangeRouter struct {
 	alloc sqlbase.DatumAlloc
 	// b is a temp storage location used during encoding
 	b         []byte
-	encodings []OutputRouterSpec_RangeRouterSpec_ColumnEncoding
-	spans     []OutputRouterSpec_RangeRouterSpec_Span
+	encodings []distsqlpb.OutputRouterSpec_RangeRouterSpec_ColumnEncoding
+	spans     []distsqlpb.OutputRouterSpec_RangeRouterSpec_Span
 	// defaultDest, if set, sends any row not matching a span to this stream. If
 	// not set and a non-matching row is encountered, an error is returned and
 	// the router is shut down.
@@ -594,7 +600,9 @@ func (hr *hashRouter) computeDestination(row sqlbase.EncDatumRow) (int, error) {
 	return int(crc32.Update(0, crc32Table, hr.buffer) % uint32(len(hr.outputs))), nil
 }
 
-func makeRangeRouter(rb routerBase, spec OutputRouterSpec_RangeRouterSpec) (*rangeRouter, error) {
+func makeRangeRouter(
+	rb routerBase, spec distsqlpb.OutputRouterSpec_RangeRouterSpec,
+) (*rangeRouter, error) {
 	if len(spec.Encodings) == 0 {
 		return nil, errors.New("missing encodings")
 	}
