@@ -17,12 +17,8 @@ package distsqlrun
 import (
 	"context"
 	"io"
-	"time"
-
-	"github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-
 	"sync"
+	"time"
 
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
@@ -33,23 +29,26 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
-	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlutil"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine"
+	"github.com/cockroachdb/cockroach/pkg/storage/diskmap"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
 	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
+	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/cockroachdb/cockroach/pkg/util/log/logtags"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
-// DistSQLVersion identifies DistSQL engine versions.
-type DistSQLVersion uint32
-
 // Version identifies the distsqlrun protocol version.
 //
 // This version is separate from the main CockroachDB version numbering; it is
@@ -76,11 +75,11 @@ type DistSQLVersion uint32
 //
 // ATTENTION: When updating these fields, add to version_history.txt explaining
 // what changed.
-const Version DistSQLVersion = 21
+const Version distsqlpb.DistSQLVersion = 22
 
 // MinAcceptedVersion is the oldest version that the server is
 // compatible with; see above.
-const MinAcceptedVersion DistSQLVersion = 21
+const MinAcceptedVersion distsqlpb.DistSQLVersion = 21
 
 // minFlowDrainWait is the minimum amount of time a draining server allows for
 // any incoming flows to be registered. It acts as a grace period in which the
@@ -113,7 +112,8 @@ var noteworthyMemoryUsageBytes = envutil.EnvOrDefaultInt64("COCKROACH_NOTEWORTHY
 type ServerConfig struct {
 	log.AmbientContext
 
-	Settings *cluster.Settings
+	Settings     *cluster.Settings
+	RuntimeStats RuntimeStats
 
 	// DB is a handle to the cluster.
 	DB *client.DB
@@ -137,7 +137,11 @@ type ServerConfig struct {
 
 	// TempStorage is used by some DistSQL processors to store rows when the
 	// working set is larger than can be stored in memory.
-	TempStorage engine.Engine
+	TempStorage diskmap.Factory
+
+	// BulkAdder is used by some processors to bulk-ingest data as SSTs.
+	BulkAdder storagebase.BulkAdderFactory
+
 	// DiskMonitor is used to monitor temporary storage disk usage. Actual disk
 	// space used will be a small multiple (~1.1) of this because of RocksDB
 	// space amplification.
@@ -165,9 +169,15 @@ type ServerConfig struct {
 	// SessionBoundInternalExecutorFactory is used to construct session-bound
 	// executors. The idea is that a higher-layer binds some of the arguments
 	// required, so that users of ServerConfig don't have to care about them.
-	SessionBoundInternalExecutorFactory func(
-		ctx context.Context, sessionData *sessiondata.SessionData,
-	) sqlutil.InternalExecutor
+	SessionBoundInternalExecutorFactory sqlutil.SessionBoundInternalExecutorFactory
+}
+
+// RuntimeStats is an interface through which the distsqlrun layer can get
+// information about runtime statistics.
+type RuntimeStats interface {
+	// GetCPUCombinedPercentNorm returns the recent user+system cpu usage,
+	// normalized to 0-1 by number of cores.
+	GetCPUCombinedPercentNorm() float64
 }
 
 // ServerImpl implements the server for the distributed SQL APIs.
@@ -179,7 +189,7 @@ type ServerImpl struct {
 	regexpCache   *tree.RegexpCache
 }
 
-var _ DistSQLServer = &ServerImpl{}
+var _ distsqlpb.DistSQLServer = &ServerImpl{}
 
 // NewServer instantiates a DistSQLServer.
 func NewServer(ctx context.Context, cfg ServerConfig) *ServerImpl {
@@ -208,7 +218,7 @@ func (ds *ServerImpl) Start() {
 	// for us.
 	if err := ds.ServerConfig.Gossip.AddInfoProto(
 		gossip.MakeDistSQLNodeVersionKey(ds.ServerConfig.NodeID.Get()),
-		&DistSQLVersionGossipInfo{
+		&distsqlpb.DistSQLVersionGossipInfo{
 			Version:            Version,
 			MinAcceptedVersion: MinAcceptedVersion,
 		},
@@ -258,7 +268,7 @@ func (ds *ServerImpl) Undrain(ctx context.Context) {
 func (ds *ServerImpl) setDraining(drain bool) error {
 	return ds.ServerConfig.Gossip.AddInfoProto(
 		gossip.MakeDistSQLDrainingKey(ds.ServerConfig.NodeID.Get()),
-		&DistSQLDrainingInfo{
+		&distsqlpb.DistSQLDrainingInfo{
 			Draining: drain,
 		},
 		0, // ttl - no expiration
@@ -267,17 +277,23 @@ func (ds *ServerImpl) setDraining(drain bool) error {
 
 // FlowVerIsCompatible checks a flow's version is compatible with this node's
 // DistSQL version.
-func FlowVerIsCompatible(flowVer, minAcceptedVersion, serverVersion DistSQLVersion) bool {
+func FlowVerIsCompatible(flowVer, minAcceptedVersion, serverVersion distsqlpb.DistSQLVersion) bool {
 	return flowVer >= minAcceptedVersion && flowVer <= serverVersion
 }
 
+// setupFlow creates a Flow.
+//
+// Args:
+// localState: Specifies if the flow runs entirely on this node and, if it does,
+//   specifies the txn and other attributes.
+//
 // Note: unless an error is returned, the returned context contains a span that
 // must be finished through Flow.Cleanup.
 func (ds *ServerImpl) setupFlow(
 	ctx context.Context,
 	parentSpan opentracing.Span,
 	parentMonitor *mon.BytesMonitor,
-	req *SetupFlowRequest,
+	req *distsqlpb.SetupFlowRequest,
 	syncFlowConsumer RowReceiver,
 	localState LocalState,
 ) (context.Context, *Flow, error) {
@@ -297,18 +313,23 @@ func (ds *ServerImpl) setupFlow(
 	const opName = "flow"
 	var sp opentracing.Span
 	if parentSpan == nil {
-		sp = ds.Tracer.StartSpan(opName, tracing.LogTagsFromCtx(ctx))
+		sp = ds.Tracer.(*tracing.Tracer).StartRootSpan(
+			opName, logtags.FromContext(ctx), tracing.NonRecordableSpan)
 	} else {
 		// We use FollowsFrom because the flow's span outlives the SetupFlow request.
+		// TODO(andrei): We should use something more efficient than StartSpan; we
+		// should use AmbientContext.AnnotateCtxWithSpan() but that interface
+		// doesn't currently support FollowsFrom relationships.
 		sp = ds.Tracer.StartSpan(
 			opName,
 			opentracing.FollowsFrom(parentSpan.Context()),
 			tracing.LogTagsFromCtx(ctx),
 		)
 	}
+	// sp will be Finish()ed by Flow.Cleanup().
 	ctx = opentracing.ContextWithSpan(ctx, sp)
 
-	// The monitor and account opened here are closed in Flow.Cleanup().
+	// The monitor opened here are closed in Flow.Cleanup().
 	monitor := mon.MakeMonitor(
 		"flow",
 		mon.MemoryResource,
@@ -319,18 +340,13 @@ func (ds *ServerImpl) setupFlow(
 		ds.Settings,
 	)
 	monitor.Start(ctx, parentMonitor, mon.BoundAccount{})
-	acc := monitor.MakeBoundAccount()
 
-	txn := localState.Txn
-	if txn := req.DeprecatedTxn; txn != nil {
-		if req.TxnCoordMeta != nil {
-			return nil, nil, errors.Errorf("provided both Txn and TxnCoordMeta")
-		}
-		meta := roachpb.MakeTxnCoordMeta(*txn)
-		req.TxnCoordMeta = &meta
-	}
-	if meta := req.TxnCoordMeta; meta != nil {
-		if !localState.IsLocal {
+	// Figure out what txn the flow needs to run in, if any.
+	// For local flows, the txn comes from localState.Txn. For non-local ones, we
+	// create a txn based on the request's TxnCoordMeta.
+	var txn *client.Txn
+	if !localState.IsLocal {
+		if meta := req.TxnCoordMeta; meta != nil {
 			if meta.Txn.Status != roachpb.PENDING {
 				return nil, nil, errors.Errorf("cannot create flow in non-PENDING txn: %s",
 					meta.Txn)
@@ -339,13 +355,14 @@ func (ds *ServerImpl) setupFlow(
 			// Txn to heartbeat the transaction.
 			txn = client.NewTxnWithCoordMeta(ctx, ds.FlowDB, req.Flow.Gateway, client.LeafTxn, *meta)
 		}
+	} else {
+		txn = localState.Txn
 	}
 
 	var evalCtx *tree.EvalContext
 	if localState.EvalContext != nil {
 		evalCtx = localState.EvalContext
 		evalCtx.Mon = &monitor
-		evalCtx.ActiveMemAcc = &acc
 		evalCtx.Txn = txn
 	} else {
 		location, err := timeutil.TimeZoneStringToLocation(req.EvalContext.Location)
@@ -356,11 +373,11 @@ func (ds *ServerImpl) setupFlow(
 
 		var be sessiondata.BytesEncodeFormat
 		switch req.EvalContext.BytesEncodeFormat {
-		case BytesEncodeFormat_HEX:
+		case distsqlpb.BytesEncodeFormat_HEX:
 			be = sessiondata.BytesEncodeHex
-		case BytesEncodeFormat_ESCAPE:
+		case distsqlpb.BytesEncodeFormat_ESCAPE:
 			be = sessiondata.BytesEncodeEscape
-		case BytesEncodeFormat_BASE64:
+		case distsqlpb.BytesEncodeFormat_BASE64:
 			be = sessiondata.BytesEncodeBase64
 		default:
 			return nil, nil, errors.Errorf("unknown byte encode format: %s",
@@ -378,28 +395,32 @@ func (ds *ServerImpl) setupFlow(
 				ExtraFloatDigits:  int(req.EvalContext.ExtraFloatDigits),
 			},
 		}
+		// Enable better compatibility with PostgreSQL date math.
+		if req.Version >= 22 {
+			sd.DurationAdditionMode = duration.AdditionModeCompatible
+		} else {
+			sd.DurationAdditionMode = duration.AdditionModeLegacy
+		}
 		ie := &lazyInternalExecutor{
 			newInternalExecutor: func() tree.SessionBoundInternalExecutor {
 				return ds.SessionBoundInternalExecutorFactory(ctx, sd)
 			},
 		}
 
-		evalPlanner := &dummyEvalPlanner{}
-		sequence := &dummySequenceOperators{}
 		evalCtx = &tree.EvalContext{
-			Settings:     ds.ServerConfig.Settings,
-			SessionData:  sd,
-			ClusterID:    ds.ServerConfig.ClusterID.Get(),
-			NodeID:       nodeID,
-			ReCache:      ds.regexpCache,
-			Mon:          &monitor,
-			ActiveMemAcc: &acc,
+			Settings:    ds.ServerConfig.Settings,
+			SessionData: sd,
+			ClusterID:   ds.ServerConfig.ClusterID.Get(),
+			NodeID:      nodeID,
+			ReCache:     ds.regexpCache,
+			Mon:         &monitor,
 			// TODO(andrei): This is wrong. Each processor should override Ctx with its
 			// own context.
 			Context:          ctx,
 			Txn:              txn,
-			Planner:          evalPlanner,
-			Sequence:         sequence,
+			Planner:          &sqlbase.DummyEvalPlanner{},
+			SessionAccessor:  &sqlbase.DummySessionAccessor{},
+			Sequence:         &sqlbase.DummySequenceOperators{},
 			InternalExecutor: ie,
 		}
 		evalCtx.SetStmtTimestamp(timeutil.Unix(0 /* sec */, req.EvalContext.StmtTimestampNanos))
@@ -417,22 +438,23 @@ func (ds *ServerImpl) setupFlow(
 	// TODO(radu): we should sanity check some of these fields.
 	flowCtx := FlowCtx{
 		Settings:       ds.Settings,
+		RuntimeStats:   ds.RuntimeStats,
 		AmbientContext: ds.AmbientContext,
 		stopper:        ds.Stopper,
 		id:             req.Flow.FlowID,
 		EvalCtx:        evalCtx,
-		rpcCtx:         ds.RPCContext,
 		nodeDialer:     ds.NodeDialer,
 		Gossip:         ds.Gossip,
 		txn:            txn,
 		ClientDB:       ds.DB,
 		executor:       ds.Executor,
-		LeaseManager:   ds.ServerConfig.LeaseManager,
+		LeaseManager:   ds.LeaseManager,
 		testingKnobs:   ds.TestingKnobs,
 		nodeID:         nodeID,
 		TempStorage:    ds.TempStorage,
+		BulkAdder:      ds.BulkAdder,
 		diskMonitor:    ds.DiskMonitor,
-		JobRegistry:    ds.ServerConfig.JobRegistry,
+		JobRegistry:    ds.JobRegistry,
 		traceKV:        req.TraceKV,
 		local:          localState.IsLocal,
 	}
@@ -456,7 +478,10 @@ func (ds *ServerImpl) setupFlow(
 // Note: the returned context contains a span that must be finished through
 // Flow.Cleanup.
 func (ds *ServerImpl) SetupSyncFlow(
-	ctx context.Context, parentMonitor *mon.BytesMonitor, req *SetupFlowRequest, output RowReceiver,
+	ctx context.Context,
+	parentMonitor *mon.BytesMonitor,
+	req *distsqlpb.SetupFlowRequest,
+	output RowReceiver,
 ) (context.Context, *Flow, error) {
 	return ds.setupFlow(ds.AnnotateCtx(ctx), opentracing.SpanFromContext(ctx), parentMonitor, req, output, LocalState{})
 }
@@ -464,13 +489,19 @@ func (ds *ServerImpl) SetupSyncFlow(
 // LocalState carries information that is required to set up a flow with wrapped
 // planNodes.
 type LocalState struct {
+	EvalContext *tree.EvalContext
+
 	// IsLocal is true if the flow is being run locally in the first place.
 	IsLocal bool
+
+	/////////////////////////////////////////////
+	// Fields below are empty if IsLocal == false
+	/////////////////////////////////////////////
+
 	// LocalProcs is an array of planNodeToRowSource processors. It's in order and
 	// will be indexed into by the RowSourceIdx field in LocalPlanNodeSpec.
-	LocalProcs  []LocalProcessor
-	EvalContext *tree.EvalContext
-	Txn         *client.Txn
+	LocalProcs []LocalProcessor
+	Txn        *client.Txn
 }
 
 // SetupLocalSyncFlow sets up a synchronous flow on the current (planning) node.
@@ -479,7 +510,7 @@ type LocalState struct {
 func (ds *ServerImpl) SetupLocalSyncFlow(
 	ctx context.Context,
 	parentMonitor *mon.BytesMonitor,
-	req *SetupFlowRequest,
+	req *distsqlpb.SetupFlowRequest,
 	output RowReceiver,
 	localState LocalState,
 ) (context.Context, *Flow, error) {
@@ -487,7 +518,7 @@ func (ds *ServerImpl) SetupLocalSyncFlow(
 }
 
 // RunSyncFlow is part of the DistSQLServer interface.
-func (ds *ServerImpl) RunSyncFlow(stream DistSQL_RunSyncFlowServer) error {
+func (ds *ServerImpl) RunSyncFlow(stream distsqlpb.DistSQL_RunSyncFlowServer) error {
 	// Set up the outgoing mailbox for the stream.
 	mbox := newOutboxSyncFlowStream(stream)
 
@@ -510,11 +541,10 @@ func (ds *ServerImpl) RunSyncFlow(stream DistSQL_RunSyncFlowServer) error {
 		defer ctxCancel()
 		f.startables = append(f.startables, mbox)
 		ds.Metrics.FlowStart()
-		if err := f.StartSync(ctx, func() {}); err != nil {
+		if err := f.Run(ctx, func() {}); err != nil {
 			log.Fatalf(ctx, "unexpected error from syncFlow.Start(): %s "+
 				"The error should have gone to the consumer.", err)
 		}
-		f.Wait()
 		f.Cleanup(ctx)
 		ds.Metrics.FlowStop()
 	}); err != nil {
@@ -525,8 +555,9 @@ func (ds *ServerImpl) RunSyncFlow(stream DistSQL_RunSyncFlowServer) error {
 
 // SetupFlow is part of the DistSQLServer interface.
 func (ds *ServerImpl) SetupFlow(
-	ctx context.Context, req *SetupFlowRequest,
-) (*SimpleResponse, error) {
+	ctx context.Context, req *distsqlpb.SetupFlowRequest,
+) (*distsqlpb.SimpleResponse, error) {
+	log.VEventf(ctx, 1, "received SetupFlow request from n%v for flow %v", req.Flow.Gateway, req.Flow.FlowID)
 	parentSpan := opentracing.SpanFromContext(ctx)
 
 	// Note: the passed context will be canceled when this RPC completes, so we
@@ -540,12 +571,14 @@ func (ds *ServerImpl) SetupFlow(
 		// We return flow deployment errors in the response so that they are
 		// packaged correctly over the wire. If we return them directly to this
 		// function, they become part of an rpc error.
-		return &SimpleResponse{Error: NewError(err)}, nil
+		return &distsqlpb.SimpleResponse{Error: distsqlpb.NewError(err)}, nil
 	}
-	return &SimpleResponse{}, nil
+	return &distsqlpb.SimpleResponse{}, nil
 }
 
-func (ds *ServerImpl) flowStreamInt(ctx context.Context, stream DistSQL_FlowStreamServer) error {
+func (ds *ServerImpl) flowStreamInt(
+	ctx context.Context, stream distsqlpb.DistSQL_FlowStreamServer,
+) error {
 	// Receive the first message.
 	msg, err := stream.Recv()
 	if err != nil {
@@ -574,7 +607,7 @@ func (ds *ServerImpl) flowStreamInt(ctx context.Context, stream DistSQL_FlowStre
 }
 
 // FlowStream is part of the DistSQLServer interface.
-func (ds *ServerImpl) FlowStream(stream DistSQL_FlowStreamServer) error {
+func (ds *ServerImpl) FlowStream(stream distsqlpb.DistSQL_FlowStreamServer) error {
 	ctx := ds.AnnotateCtx(stream.Context())
 	err := ds.flowStreamInt(ctx, stream)
 	if err != nil {
@@ -640,89 +673,6 @@ const (
 // ModuleTestingKnobs is part of the base.ModuleTestingKnobs interface.
 func (*TestingKnobs) ModuleTestingKnobs() {}
 
-var errEvalPlanner = errors.New("cannot backfill such evaluated expression")
-
-// Implements the tree.EvalPlanner interface by returning errors.
-type dummyEvalPlanner struct {
-}
-
-var _ tree.EvalPlanner = &dummyEvalPlanner{}
-
-// Implements the tree.EvalDatabase interface.
-func (ep *dummyEvalPlanner) ParseQualifiedTableName(
-	ctx context.Context, sql string,
-) (*tree.TableName, error) {
-	return nil, errEvalPlanner
-}
-
-// Implements the tree.EvalDatabase interface.
-func (ep *dummyEvalPlanner) LookupSchema(
-	ctx context.Context, dbName, scName string,
-) (bool, tree.SchemaMeta, error) {
-	return false, nil, errEvalPlanner
-}
-
-// Implements the tree.EvalDatabase interface.
-func (ep *dummyEvalPlanner) ResolveTableName(ctx context.Context, tn *tree.TableName) error {
-	return errEvalPlanner
-}
-
-// Implements the tree.EvalPlanner interface.
-func (ep *dummyEvalPlanner) ParseType(sql string) (coltypes.CastTargetType, error) {
-	return nil, errEvalPlanner
-}
-
-// Implements the tree.EvalPlanner interface.
-func (ep *dummyEvalPlanner) EvalSubquery(expr *tree.Subquery) (tree.Datum, error) {
-	return nil, errEvalPlanner
-}
-
-var errSequenceOperators = errors.New("cannot backfill such sequence operation")
-
-// Implements the tree.SequenceOperators interface by returning errors.
-type dummySequenceOperators struct {
-}
-
-// Implements the tree.EvalDatabase interface.
-func (so *dummySequenceOperators) ParseQualifiedTableName(
-	ctx context.Context, sql string,
-) (*tree.TableName, error) {
-	return nil, errSequenceOperators
-}
-
-// Implements the tree.EvalDatabase interface.
-func (so *dummySequenceOperators) ResolveTableName(ctx context.Context, tn *tree.TableName) error {
-	return errSequenceOperators
-}
-
-// Implements the tree.EvalDatabase interface.
-func (so *dummySequenceOperators) LookupSchema(
-	ctx context.Context, dbName, scName string,
-) (bool, tree.SchemaMeta, error) {
-	return false, nil, errSequenceOperators
-}
-
-// Implements the tree.SequenceOperators interface.
-func (so *dummySequenceOperators) IncrementSequence(
-	ctx context.Context, seqName *tree.TableName,
-) (int64, error) {
-	return 0, errSequenceOperators
-}
-
-// Implements the tree.SequenceOperators interface.
-func (so *dummySequenceOperators) GetLatestValueInSessionForSequence(
-	ctx context.Context, seqName *tree.TableName,
-) (int64, error) {
-	return 0, errSequenceOperators
-}
-
-// Implements the tree.SequenceOperators interface.
-func (so *dummySequenceOperators) SetSequenceValue(
-	ctx context.Context, seqName *tree.TableName, newVal int64, isCalled bool,
-) error {
-	return errSequenceOperators
-}
-
 // lazyInternalExecutor is a tree.SessionBoundInternalExecutor that initializes
 // itself only on the first call to QueryRow.
 type lazyInternalExecutor struct {
