@@ -16,14 +16,17 @@ package distsqlrun
 
 import (
 	"context"
-
 	"fmt"
+	"math"
 
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
 // sorter sorts the input rows according to the specified ordering.
@@ -34,8 +37,8 @@ type sorterBase struct {
 	ordering sqlbase.ColumnOrdering
 	matchLen uint32
 
-	rows sortableRowContainer
-	i    rowIterator
+	rows rowcontainer.SortableRowContainer
+	i    rowcontainer.RowIterator
 
 	// Only set if the ability to spill to disk is enabled.
 	diskMonitor *mon.BytesMonitor
@@ -46,7 +49,7 @@ func (s *sorterBase) init(
 	flowCtx *FlowCtx,
 	processorID int32,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	output RowReceiver,
 	ordering sqlbase.ColumnOrdering,
 	matchLen uint32,
@@ -86,19 +89,20 @@ func (s *sorterBase) init(
 
 	if useTempStorage {
 		s.diskMonitor = NewMonitor(ctx, flowCtx.diskMonitor, "sorter-disk")
-		rc := diskBackedRowContainer{}
-		rc.init(
+		rc := rowcontainer.DiskBackedRowContainer{}
+		rc.Init(
 			ordering,
 			input.OutputTypes(),
 			s.evalCtx,
 			flowCtx.TempStorage,
 			memMonitor,
 			s.diskMonitor,
+			0, /* rowCapacity */
 		)
 		s.rows = &rc
 	} else {
-		rc := memRowContainer{}
-		rc.initWithMon(ordering, input.OutputTypes(), s.evalCtx, memMonitor)
+		rc := rowcontainer.MemRowContainer{}
+		rc.InitWithMon(ordering, input.OutputTypes(), s.evalCtx, memMonitor, 0 /* rowCapacity */)
 		s.rows = &rc
 	}
 
@@ -114,7 +118,7 @@ func (s *sorterBase) init(
 func (s *sorterBase) Next() (sqlbase.EncDatumRow, *ProducerMetadata) {
 	for s.State == StateRunning {
 		if ok, err := s.i.Valid(); err != nil || !ok {
-			s.MoveToDraining(nil /* err */)
+			s.MoveToDraining(err)
 			break
 		}
 
@@ -138,7 +142,7 @@ func (s *sorterBase) close() {
 		if s.i != nil {
 			s.i.Close()
 		}
-		ctx := s.evalCtx.Ctx()
+		ctx := s.Ctx
 		s.rows.Close(ctx)
 		s.MemMonitor.Stop(ctx)
 		if s.diskMonitor != nil {
@@ -147,7 +151,7 @@ func (s *sorterBase) close() {
 	}
 }
 
-var _ DistSQLSpanStats = &SorterStats{}
+var _ distsqlpb.DistSQLSpanStats = &SorterStats{}
 
 const sorterTagPrefix = "sorter."
 
@@ -191,16 +195,21 @@ func newSorter(
 	ctx context.Context,
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *SorterSpec,
+	spec *distsqlpb.SorterSpec,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	output RowReceiver,
 ) (Processor, error) {
-	count := int64(0)
-	if post.Limit != 0 && post.Filter.Expr == "" {
+	count := uint64(0)
+	if post.Limit != 0 && post.Filter.Empty() {
 		// The sorter needs to produce Offset + Limit rows. The ProcOutputHelper
 		// will discard the first Offset ones.
-		count = int64(post.Limit) + int64(post.Offset)
+		// LIMIT and OFFSET should each never be greater than math.MaxInt64, the
+		// parser ensures this.
+		if post.Limit > math.MaxInt64 || post.Offset > math.MaxInt64 {
+			return nil, errors.Errorf("error creating sorter: limit %d offset %d too large", post.Limit, post.Offset)
+		}
+		count = post.Limit + post.Offset
 	}
 
 	// Choose the optimal processor.
@@ -246,15 +255,15 @@ func newSortAllProcessor(
 	ctx context.Context,
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *SorterSpec,
+	spec *distsqlpb.SorterSpec,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	out RowReceiver,
 ) (Processor, error) {
 	proc := &sortAllProcessor{}
 	if err := proc.sorterBase.init(
 		proc, flowCtx, processorID, input, post, out,
-		convertToColumnOrdering(spec.OutputOrdering),
+		distsqlpb.ConvertToColumnOrdering(spec.OutputOrdering),
 		spec.OrderingMatchLen,
 		ProcStateOpts{
 			InputsToDrain: []RowSource{input},
@@ -346,7 +355,7 @@ func (s *sortAllProcessor) ConsumerClosed() {
 // sorted in linearithmic time.
 type sortTopKProcessor struct {
 	sorterBase
-	k int64
+	k uint64
 }
 
 var _ Processor = &sortTopKProcessor{}
@@ -354,16 +363,21 @@ var _ RowSource = &sortTopKProcessor{}
 
 const sortTopKProcName = "sortTopK"
 
+var errSortTopKZeroK = errors.New("invalid value 0 for k")
+
 func newSortTopKProcessor(
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *SorterSpec,
+	spec *distsqlpb.SorterSpec,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	out RowReceiver,
-	k int64,
+	k uint64,
 ) (Processor, error) {
-	ordering := convertToColumnOrdering(spec.OutputOrdering)
+	if k == 0 {
+		return nil, errors.Wrap(errSortTopKZeroK, "error creating top k sorter")
+	}
+	ordering := distsqlpb.ConvertToColumnOrdering(spec.OutputOrdering)
 	proc := &sortTopKProcessor{k: k}
 	if err := proc.sorterBase.init(
 		proc, flowCtx, processorID, input, post, out,
@@ -404,7 +418,7 @@ func (s *sortTopKProcessor) Start(ctx context.Context) context.Context {
 			break
 		}
 
-		if int64(s.rows.Len()) < s.k {
+		if uint64(s.rows.Len()) < s.k {
 			// Accumulate up to k values.
 			if err := s.rows.AddRow(ctx, row); err != nil {
 				s.MoveToDraining(err)
@@ -461,12 +475,12 @@ const sortChunksProcName = "sortChunks"
 func newSortChunksProcessor(
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *SorterSpec,
+	spec *distsqlpb.SorterSpec,
 	input RowSource,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	out RowReceiver,
 ) (Processor, error) {
-	ordering := convertToColumnOrdering(spec.OutputOrdering)
+	ordering := distsqlpb.ConvertToColumnOrdering(spec.OutputOrdering)
 
 	proc := &sortChunksProcessor{}
 	if err := proc.sorterBase.init(
@@ -508,7 +522,7 @@ func (s *sortChunksProcessor) chunkCompleted(
 // if a metadata record was encountered). The caller is expected to drain when
 // this returns false.
 func (s *sortChunksProcessor) fill() (bool, error) {
-	ctx := s.evalCtx.Ctx()
+	ctx := s.Ctx
 
 	var meta *ProducerMetadata
 
@@ -578,6 +592,7 @@ func (s *sortChunksProcessor) Start(ctx context.Context) context.Context {
 
 // Next is part of the RowSource interface.
 func (s *sortChunksProcessor) Next() (sqlbase.EncDatumRow, *ProducerMetadata) {
+	ctx := s.Ctx
 	for s.State == StateRunning {
 		ok, err := s.i.Valid()
 		if err != nil {
@@ -586,7 +601,6 @@ func (s *sortChunksProcessor) Next() (sqlbase.EncDatumRow, *ProducerMetadata) {
 		}
 		// If we don't have an active chunk, clear and refill it.
 		if !ok {
-			ctx := s.evalCtx.Ctx()
 			if err := s.rows.UnsafeReset(ctx); err != nil {
 				s.MoveToDraining(err)
 				break
@@ -620,11 +634,6 @@ func (s *sortChunksProcessor) Next() (sqlbase.EncDatumRow, *ProducerMetadata) {
 	return nil, s.DrainHelper()
 }
 
-// ConsumerDone is part of the RowSource interface.
-func (s *sortChunksProcessor) ConsumerDone() {
-	s.MoveToDraining(nil /* err */)
-}
-
 // ConsumerClosed is part of the RowSource interface.
 func (s *sortChunksProcessor) ConsumerClosed() {
 	// The consumer is done, Next() will not be called again.
