@@ -17,16 +17,17 @@ package distsqlrun
 import (
 	"context"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
 	"github.com/cockroachdb/cockroach/pkg/sql/scrub"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/pkg/errors"
 )
 
 // Consider the schema:
@@ -263,9 +264,9 @@ const zigzagJoinerProcName = "zigzagJoiner"
 func newZigzagJoiner(
 	flowCtx *FlowCtx,
 	processorID int32,
-	spec *ZigzagJoinerSpec,
+	spec *distsqlpb.ZigzagJoinerSpec,
 	fixedValues []sqlbase.EncDatumRow,
-	post *PostProcessSpec,
+	post *distsqlpb.PostProcessSpec,
 	output RowReceiver,
 ) (*zigzagJoiner, error) {
 	z := &zigzagJoiner{}
@@ -302,8 +303,16 @@ func newZigzagJoiner(
 
 	colOffset := 0
 	for i := 0; i < z.numTables; i++ {
-		if i < len(fixedValues) {
+		if fixedValues != nil && i < len(fixedValues) {
+			// Useful for testing. In cases where we plan a zigzagJoin in
+			// the planner, we specify fixed values as ValuesCoreSpecs in
+			// the spec itself.
 			z.infos[i].fixedValues = fixedValues[i]
+		} else if i < len(spec.FixedValues) {
+			z.infos[i].fixedValues, err = valuesSpecToEncDatum(spec.FixedValues[i])
+			if err != nil {
+				return nil, err
+			}
 		}
 		if err := z.setupInfo(spec, i, colOffset); err != nil {
 			return nil, err
@@ -314,6 +323,23 @@ func newZigzagJoiner(
 	return z, nil
 }
 
+// Helper function to convert a values spec containing one tuple into EncDatums for
+// each cell. Note that this function assumes that there is only one tuple in the
+// ValuesSpec (i.e. the way fixed values are encoded in the ZigzagJoinSpec).
+func valuesSpecToEncDatum(
+	valuesSpec *distsqlpb.ValuesCoreSpec,
+) (res []sqlbase.EncDatum, err error) {
+	res = make([]sqlbase.EncDatum, len(valuesSpec.Columns))
+	rem := valuesSpec.RawBytes[0]
+	for i, colInfo := range valuesSpec.Columns {
+		res[i], rem, err = sqlbase.EncDatumFromBuffer(&colInfo.Type, colInfo.Encoding, rem)
+		if err != nil {
+			return nil, err
+		}
+	}
+	return res, nil
+}
+
 // Start is part of the RowSource interface.
 func (z *zigzagJoiner) Start(ctx context.Context) context.Context {
 	ctx = z.StartInternal(ctx, zigzagJoinerProcName)
@@ -326,7 +352,7 @@ func (z *zigzagJoiner) Start(ctx context.Context) context.Context {
 // zigzagJoinerInfo contains all the information that needs to be
 // stored for each side of the join.
 type zigzagJoinerInfo struct {
-	fetcher    sqlbase.RowFetcher
+	fetcher    row.Fetcher
 	alloc      *sqlbase.DatumAlloc
 	table      *sqlbase.TableDescriptor
 	index      *sqlbase.IndexDescriptor
@@ -337,7 +363,8 @@ type zigzagJoinerInfo struct {
 	// the cartesian product of the containers will be emitted.
 	container sqlbase.EncDatumRowContainer
 
-	eqColumnIDs columns
+	// eqColumns is the ordinal positions of the equality columns.
+	eqColumns columns
 
 	// Prefix of the index key that has fixed values.
 	fixedValues sqlbase.EncDatumRow
@@ -357,12 +384,13 @@ type zigzagJoinerInfo struct {
 // colOffset is specified to determine the appropriate range of output columns
 // to process. It is the number of columns in the tables of all previous sides
 // of the join.
-func (z *zigzagJoiner) setupInfo(spec *ZigzagJoinerSpec, side int, colOffset int) error {
+func (z *zigzagJoiner) setupInfo(spec *distsqlpb.ZigzagJoinerSpec, side int, colOffset int) error {
 	z.side = side
 	info := z.infos[side]
 
+	info.alloc = &sqlbase.DatumAlloc{}
 	info.table = &spec.Tables[side]
-	info.eqColumnIDs = spec.EqColumns[side].Columns
+	info.eqColumns = spec.EqColumns[side].Columns
 	indexID := spec.IndexIds[side]
 	if indexID == 0 {
 		info.index = &info.table.PrimaryIndex
@@ -373,11 +401,9 @@ func (z *zigzagJoiner) setupInfo(spec *ZigzagJoinerSpec, side int, colOffset int
 	var columnIDs []sqlbase.ColumnID
 	columnIDs, info.indexDirs = info.index.FullColumnIDs()
 	info.indexTypes = make([]sqlbase.ColumnType, len(columnIDs))
-	indexCols := make([]uint32, len(columnIDs))
 	columnTypes := info.table.ColumnTypes()
 	colIdxMap := info.table.ColumnIdxMap()
 	for i, columnID := range columnIDs {
-		indexCols[i] = uint32(columnID)
 		info.indexTypes[i] = columnTypes[colIdxMap[columnID]]
 	}
 
@@ -391,18 +417,18 @@ func (z *zigzagJoiner) setupInfo(spec *ZigzagJoinerSpec, side int, colOffset int
 
 	// Add the fixed columns.
 	for i := 0; i < len(info.fixedValues); i++ {
-		neededCols.Add(int(indexCols[i]) - 1)
+		neededCols.Add(colIdxMap[columnIDs[i]])
 	}
 
 	// Add the equality columns.
-	for _, col := range info.eqColumnIDs {
+	for _, col := range info.eqColumns {
 		neededCols.Add(int(col))
 	}
 
 	// Setup the RowContainers.
 	info.container.Reset()
 
-	// Setup the RowFetcher.
+	// Setup the Fetcher.
 	_, _, err := initRowFetcher(
 		&(info.fetcher),
 		info.table,
@@ -412,7 +438,7 @@ func (z *zigzagJoiner) setupInfo(spec *ZigzagJoinerSpec, side int, colOffset int
 		neededCols,
 		false, /* check */
 		info.alloc,
-		ScanVisibility_PUBLIC,
+		distsqlpb.ScanVisibility_PUBLIC,
 	)
 	if err != nil {
 		return err
@@ -475,7 +501,7 @@ func (z *zigzagJoiner) fetchRowFromSide(
 	// Keep fetching until a row is found that does not have null in an equality
 	// column.
 	hasNull := func(row sqlbase.EncDatumRow) bool {
-		for _, c := range z.infos[side].eqColumnIDs {
+		for _, c := range z.infos[side].eqColumns {
 			if row[c].IsNull() {
 				return true
 			}
@@ -497,12 +523,68 @@ func (z *zigzagJoiner) fetchRowFromSide(
 // Return the datums from the equality columns from a given non-empty row
 // from the specified side.
 func (z *zigzagJoiner) extractEqDatums(row sqlbase.EncDatumRow, side int) sqlbase.EncDatumRow {
-	eqColIDs := z.infos[side].eqColumnIDs
-	eqCols := make(sqlbase.EncDatumRow, len(eqColIDs))
-	for i, id := range eqColIDs {
-		eqCols[i] = row[id]
+	eqCols := z.infos[side].eqColumns
+	eqDatums := make(sqlbase.EncDatumRow, len(eqCols))
+	for i, col := range eqCols {
+		eqDatums[i] = row[col]
 	}
-	return eqCols
+	return eqDatums
+}
+
+// Generates a Key for an inverted index from the passed datums and side
+// info. Used by produceKeyFromBaseRow.
+func (z *zigzagJoiner) produceInvertedIndexKey(
+	info *zigzagJoinerInfo, datums sqlbase.EncDatumRow,
+) (roachpb.Span, error) {
+	// For inverted indexes, the JSON field (first column in the index) is
+	// encoded a little differently. We need to explicitly call
+	// EncodeInvertedIndexKeys to generate the prefix. The rest of the
+	// index key containing the remaining neededDatums can be generated
+	// and appended using EncodeColumns.
+	colMap := make(map[sqlbase.ColumnID]int)
+	decodedDatums := make([]tree.Datum, len(datums))
+
+	// Ensure all EncDatums have been decoded.
+	for i, encDatum := range datums {
+		err := encDatum.EnsureDecoded(&info.indexTypes[i], info.alloc)
+		if err != nil {
+			return roachpb.Span{}, err
+		}
+
+		decodedDatums[i] = encDatum.Datum
+		if i < len(info.index.ColumnIDs) {
+			colMap[info.index.ColumnIDs[i]] = i
+		} else {
+			// This column's value will be encoded in the second part (i.e.
+			// EncodeColumns).
+			colMap[info.index.ExtraColumnIDs[i-len(info.index.ColumnIDs)]] = i
+		}
+	}
+
+	keys, err := sqlbase.EncodeInvertedIndexKeys(
+		info.table,
+		info.index,
+		colMap,
+		decodedDatums,
+		info.prefix,
+	)
+	if err != nil {
+		return roachpb.Span{}, err
+	}
+	if len(keys) != 1 {
+		return roachpb.Span{}, errors.Errorf("%d fixed values passed in for inverted index", len(keys))
+	}
+
+	// Append remaining (non-JSON) datums to the key.
+	keyBytes, _, err := sqlbase.EncodeColumns(
+		info.index.ExtraColumnIDs[:len(datums)-1],
+		info.indexDirs[1:],
+		colMap,
+		decodedDatums,
+		keys[0],
+	)
+	key := roachpb.Key(keyBytes)
+	return roachpb.Span{Key: key, EndKey: key.PrefixEnd()}, err
 }
 
 // Generates a Key, corresponding to the current `z.baseRow` in
@@ -517,6 +599,10 @@ func (z *zigzagJoiner) produceSpanFromBaseRow() (roachpb.Span, error) {
 
 	// Construct correct row by concatenating right fixed datums with
 	// primary key extracted from `row`.
+	if info.index.Type == sqlbase.IndexDescriptor_INVERTED {
+		return z.produceInvertedIndexKey(info, neededDatums)
+	}
+
 	return sqlbase.MakeSpanFromEncDatums(
 		info.prefix,
 		neededDatums,
@@ -530,28 +616,29 @@ func (z *zigzagJoiner) produceSpanFromBaseRow() (roachpb.Span, error) {
 
 // Returns the column types of the equality columns.
 func (zi *zigzagJoinerInfo) eqColTypes() []sqlbase.ColumnType {
-	eqColIDs := zi.eqColumnIDs
-	eqColTypes := make([]sqlbase.ColumnType, 0, len(eqColIDs))
-	for _, id := range eqColIDs {
-		eqColTypes = append(eqColTypes, zi.table.ColumnTypes()[id])
+	eqColTypes := make([]sqlbase.ColumnType, len(zi.eqColumns))
+	colTypes := zi.table.ColumnTypes()
+	for i := range eqColTypes {
+		eqColTypes[i] = colTypes[zi.eqColumns[i]]
 	}
 	return eqColTypes
 }
 
 // Returns the ordering of the equality columns.
 func (zi *zigzagJoinerInfo) eqOrdering() (sqlbase.ColumnOrdering, error) {
-	ordering := make(sqlbase.ColumnOrdering, len(zi.eqColumnIDs))
-	for i, colID := range zi.eqColumnIDs {
+	ordering := make(sqlbase.ColumnOrdering, len(zi.eqColumns))
+	for i := range zi.eqColumns {
+		colID := zi.table.Columns[zi.eqColumns[i]].ID
 		// Search the index columns, then the primary keys to find an ordering for
 		// the current column, 'colID'.
 		var direction encoding.Direction
 		var err error
-		if idx := findColumnID(zi.index.ColumnIDs, sqlbase.ColumnID(colID+1)); idx != -1 {
+		if idx := findColumnID(zi.index.ColumnIDs, colID); idx != -1 {
 			direction, err = zi.index.ColumnDirections[idx].ToEncodingDirection()
 			if err != nil {
 				return nil, err
 			}
-		} else if idx := findColumnID(zi.table.PrimaryIndex.ColumnIDs, sqlbase.ColumnID(colID+1)); idx != -1 {
+		} else if idx := findColumnID(zi.table.PrimaryIndex.ColumnIDs, colID); idx != -1 {
 			direction, err = zi.table.PrimaryIndex.ColumnDirections[idx].ToEncodingDirection()
 			if err != nil {
 				return nil, err
@@ -854,10 +941,6 @@ func (z *zigzagJoiner) Next() (sqlbase.EncDatumRow, *ProducerMetadata) {
 	return nil, z.DrainHelper()
 }
 
-// ConsumerDone is part of the RowSource interface.
-func (z *zigzagJoiner) ConsumerDone() {
-}
-
 // ConsumerClosed is part of the RowSource interface.
 func (z *zigzagJoiner) ConsumerClosed() {
 	// The consumer is done, Next() will not be called again.
