@@ -21,16 +21,18 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/sql/distsqlpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
+	"github.com/stretchr/testify/require"
 )
 
 type zigzagJoinerTestCase struct {
 	desc          string
-	spec          ZigzagJoinerSpec
+	spec          distsqlpb.ZigzagJoinerSpec
 	outCols       []uint32
 	fixedValues   []sqlbase.EncDatumRow
 	expectedTypes []sqlbase.ColumnType
@@ -40,7 +42,7 @@ type zigzagJoinerTestCase struct {
 func intCols(numCols int) []sqlbase.ColumnType {
 	cols := make([]sqlbase.ColumnType, numCols)
 	for i := range cols {
-		cols[i] = intType
+		cols[i] = sqlbase.IntType
 	}
 	return cols
 }
@@ -89,12 +91,16 @@ func TestZigzagJoiner(t *testing.T) {
 	}
 
 	sqlutils.CreateTableDebug(t, sqlDB, "empty",
-		"a INT, b INT, c INT, d INT, PRIMARY KEY (a,b), INDEX c (c), INDEX d (d)",
+		"a INT, b INT, x INT, c INT, d INT, PRIMARY KEY (a,b), INDEX c (c), INDEX d (d)",
 		0,
 		sqlutils.ToRowFn(aFn, bFn, cFn, dFn),
 		true, /* shouldPrint */
 	)
 
+	// Drop a column to test https://github.com/cockroachdb/cockroach/issues/37196
+	_, err := sqlDB.Exec("ALTER TABLE test.empty DROP COLUMN x")
+	require.NoError(t, err)
+
 	sqlutils.CreateTableDebug(t, sqlDB, "single",
 		"a INT, b INT, c INT, d INT, PRIMARY KEY (a,b), INDEX c (c), INDEX d (d)",
 		1,
@@ -187,9 +193,9 @@ func TestZigzagJoiner(t *testing.T) {
 	testCases := []zigzagJoinerTestCase{
 		{
 			desc: "join on an empty table with itself on its primary key",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*empty, *empty},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -200,9 +206,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join an empty table on the left with a populated table on its primary key",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*empty, *highRangeDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -213,9 +219,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a populated table on the left with an empty table on its primary key",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*highRangeDesc, *empty},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -226,9 +232,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a table with a single row with itself on its primary key",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*single, *single},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -239,9 +245,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a table with a few rows with itself on its primary key",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*smallDesc, *smallDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -252,9 +258,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a populated table that has a match in the last row with itself",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*medDesc, *medDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -265,9 +271,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "(a) is free, and outputs cartesian product",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*medDesc, *medDesc},
-				EqColumns: []Columns{{[]uint32{0}}, {[]uint32{0}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0}}, {Columns: []uint32{0}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -279,9 +285,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "set the fixed columns to be a part of the primary key",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*medDesc, *medDesc},
-				EqColumns: []Columns{{[]uint32{1}}, {[]uint32{1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{1}}, {Columns: []uint32{1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -292,9 +298,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join should work when there is a block of matches",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*highRangeDesc, *highRangeDesc},
-				EqColumns: []Columns{{[]uint32{0}}, {[]uint32{0}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0}}, {Columns: []uint32{0}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -306,9 +312,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join two different tables where first one is larger",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*medDesc, *smallDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -319,9 +325,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join two different tables where second is larger",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*smallDesc, *medDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c) */, 2 /* (d) */},
 			},
@@ -332,9 +338,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join on an index containing primary key columns explicitly",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*overlappingDesc, *overlappingDesc},
-				EqColumns: []Columns{{[]uint32{1}}, {[]uint32{1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{1}}, {Columns: []uint32{1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (a, c) */, 2 /* (d) */},
 			},
@@ -345,9 +351,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join two tables with different schemas",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*smallDesc, *t2Desc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{0 /* (a, b) */, 0 /* (a, b) */},
 			},
@@ -357,9 +363,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join two tables with different schemas flipped",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*t2Desc, *smallDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{0 /* (a, b) */, 0 /* (a, b) */},
 			},
@@ -369,9 +375,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join on a populated table with no fixed columns",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*smallDesc, *smallDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{0 /* (a, b) */, 0 /* (a, b) */},
 			},
@@ -381,9 +387,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join tables with different schemas with no locked columns",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*smallDesc, *t2Desc},
-				EqColumns: []Columns{{[]uint32{1}}, {[]uint32{1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{1}}, {Columns: []uint32{1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{0 /* (a, b) */, 0 /* (a, b) */},
 			},
@@ -394,9 +400,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a composite index with itself",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*compDesc, *compDesc},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c, a, b) */, 2 /* (d) */},
 			},
@@ -407,9 +413,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a composite index with the primary key reversed with itself",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*revCompDesc, *revCompDesc},
-				EqColumns: []Columns{{[]uint32{0}}, {[]uint32{0}}}, // join on a
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0}}, {Columns: []uint32{0}}}, // join on a
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c, b, a) */, 2 /* (d) */},
 			},
@@ -420,12 +426,12 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a composite index with the primary key reversed with itself with onExpr on value on one side",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*revCompDesc, *revCompDesc},
-				EqColumns: []Columns{{[]uint32{0}}, {[]uint32{0}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0}}, {Columns: []uint32{0}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c, b, a) */, 2 /* (d) */},
-				OnExpr:    Expression{Expr: "@1 > 1"},
+				OnExpr:    distsqlpb.Expression{Expr: "@1 > 1"},
 			},
 			fixedValues:   []sqlbase.EncDatumRow{{encInt(3), encInt(1)}, {encInt(7)}},
 			outCols:       []uint32{0, 1, 2, 7},
@@ -434,12 +440,12 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a composite index with the primary key reversed with itself and with onExpr comparing both sides",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*revCompDesc, *revCompDesc},
-				EqColumns: []Columns{{[]uint32{0}}, {[]uint32{0}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0}}, {Columns: []uint32{0}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (c, b, a) */, 2 /* (d) */},
-				OnExpr:    Expression{Expr: "@8 < 2 * @1"},
+				OnExpr:    distsqlpb.Expression{Expr: "@8 < 2 * @1"},
 			},
 			fixedValues:   []sqlbase.EncDatumRow{{encInt(3), encInt(1)}, {encInt(7)}},
 			outCols:       []uint32{0, 1, 2, 7},
@@ -448,9 +454,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "join a composite index that doesn't contain the full primary key with itself",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*compUnqDesc, *compUnqDesc},
-				EqColumns: []Columns{{[]uint32{1}}, {[]uint32{1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{1}}, {Columns: []uint32{1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{2 /* (c, b) */, 3 /* (d) */},
 			},
@@ -461,9 +467,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "test when equality columns may be null",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*nullableDesc, *nullableDesc},
-				EqColumns: []Columns{{[]uint32{2}}, {[]uint32{3}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{2}}, {Columns: []uint32{3}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{1 /* (e) */, 2 /* (d) */},
 			},
@@ -476,9 +482,9 @@ func TestZigzagJoiner(t *testing.T) {
 		},
 		{
 			desc: "test joining with primary key",
-			spec: ZigzagJoinerSpec{
+			spec: distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*medDesc, *medDesc},
-				EqColumns: []Columns{{[]uint32{0}}, {[]uint32{3}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0}}, {Columns: []uint32{3}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{0 /* primary (a, b) */, 2 /* (d) */},
 			},
@@ -502,13 +508,13 @@ func TestZigzagJoiner(t *testing.T) {
 			}
 
 			out := &RowBuffer{}
-			post := PostProcessSpec{Projection: true, OutputColumns: c.outCols}
+			post := distsqlpb.PostProcessSpec{Projection: true, OutputColumns: c.outCols}
 			z, err := newZigzagJoiner(&flowCtx, 0 /* processorID */, &c.spec, c.fixedValues, &post, out)
 			if err != nil {
 				t.Fatal(err)
 			}
 
-			z.Run(ctx, nil /* wg */)
+			z.Run(ctx)
 
 			if !out.ProducerClosed() {
 				t.Fatalf("output RowReceiver not closed")
@@ -566,7 +572,7 @@ func TestZigzagJoinerDrain(t *testing.T) {
 	}
 
 	encRow := make(sqlbase.EncDatumRow, 1)
-	encRow[0] = sqlbase.DatumToEncDatum(intType, tree.NewDInt(1))
+	encRow[0] = sqlbase.DatumToEncDatum(sqlbase.IntType, tree.NewDInt(1))
 
 	// ConsumerClosed verifies that when a joinReader's consumer is closed, the
 	// joinReader finishes gracefully.
@@ -576,20 +582,20 @@ func TestZigzagJoinerDrain(t *testing.T) {
 		zz, err := newZigzagJoiner(
 			&flowCtx,
 			0, /* processorID */
-			&ZigzagJoinerSpec{
+			&distsqlpb.ZigzagJoinerSpec{
 				Tables:    []sqlbase.TableDescriptor{*td, *td},
-				EqColumns: []Columns{{[]uint32{0, 1}}, {[]uint32{0, 1}}},
+				EqColumns: []distsqlpb.Columns{{Columns: []uint32{0, 1}}, {Columns: []uint32{0, 1}}},
 				Type:      sqlbase.InnerJoin,
 				IndexIds:  []uint32{0, 1},
 			},
 			[]sqlbase.EncDatumRow{{encThree}, {encSeven}},
-			&PostProcessSpec{Projection: true, OutputColumns: []uint32{0, 1}},
+			&distsqlpb.PostProcessSpec{Projection: true, OutputColumns: []uint32{0, 1}},
 			out,
 		)
 		if err != nil {
 			t.Fatal(err)
 		}
-		zz.Run(ctx, nil /* wg */)
+		zz.Run(ctx)
 	})
 
 	//TODO(pbardea): When RowSource inputs are added, ensure that meta is
