@@ -27,9 +27,6 @@ import (
 	"sync"
 	"time"
 
-	opentracing "github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
@@ -37,21 +34,25 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/jobs"
 	"github.com/cockroachdb/cockroach/pkg/kv"
+	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/rpc"
 	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
-	"github.com/cockroachdb/cockroach/pkg/server/status"
+	"github.com/cockroachdb/cockroach/pkg/server/status/statuspb"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql/distsqlrun"
+	"github.com/cockroachdb/cockroach/pkg/sql/opt"
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/querycache"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqlutil"
 	"github.com/cockroachdb/cockroach/pkg/sql/stats"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
+	"github.com/cockroachdb/cockroach/pkg/util/bitarray"
 	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
@@ -60,6 +61,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
 // ClusterOrganization is the organization name.
@@ -76,10 +79,22 @@ var ClusterSecret = func() *settings.StringSetting {
 		"cluster specific secret",
 		"",
 	)
-	s.Hide()
+	s.SetConfidential()
 	return s
 }()
 
+// defaultIntSize controls how a "naked" INT type will be parsed.
+// TODO(bob): Change this to 4 in v2.3; https://github.com/cockroachdb/cockroach/issues/32534
+// TODO(bob): Remove or n-op this in v2.4: https://github.com/cockroachdb/cockroach/issues/32844
+var defaultIntSize = settings.RegisterValidatedIntSetting(
+	"sql.defaults.default_int_size",
+	"the size, in bytes, of an INT type", 8, func(i int64) error {
+		if i != 4 && i != 8 {
+			return errors.New("only 4 or 8 are valid values")
+		}
+		return nil
+	})
+
 // traceTxnThreshold can be used to log SQL transactions that take
 // longer than duration to complete. For example, traceTxnThreshold=1s
 // will log the trace for any transaction that takes 1s or longer. To
@@ -113,17 +128,47 @@ var OptimizerClusterMode = settings.RegisterEnumSetting(
 	},
 )
 
+// ReorderJoinsLimitClusterSettingName is the name of the cluster setting for
+// the maximum number of joins to reorder.
+const ReorderJoinsLimitClusterSettingName = "sql.defaults.reorder_joins_limit"
+
+// ReorderJoinsLimitClusterValue controls the cluster default for the maximum
+// number of joins reordered.
+var ReorderJoinsLimitClusterValue = settings.RegisterValidatedIntSetting(
+	ReorderJoinsLimitClusterSettingName,
+	"default number of joins to reorder",
+	opt.DefaultJoinOrderLimit,
+	func(v int64) error {
+		if v < 0 {
+			return pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError,
+				"cannot set sql.defaults.reorder_joins_limit to a negative value: %d", v)
+		}
+		return nil
+	},
+)
+
+// VectorizeClusterMode controls the cluster default for when automatic
+// vectorization is enabled.
+var VectorizeClusterMode = settings.RegisterEnumSetting(
+	"sql.defaults.experimental_vectorize",
+	"default experimental_vectorize mode",
+	"off",
+	map[int64]string{
+		int64(sessiondata.VectorizeOff):    "off",
+		int64(sessiondata.VectorizeOn):     "on",
+		int64(sessiondata.VectorizeAlways): "always",
+	},
+)
+
 // DistSQLClusterExecMode controls the cluster default for when DistSQL is used.
 var DistSQLClusterExecMode = settings.RegisterEnumSetting(
 	"sql.defaults.distsql",
 	"default distributed SQL execution mode",
 	"auto",
 	map[int64]string{
-		int64(sessiondata.DistSQLOff):       "off",
-		int64(sessiondata.DistSQLAuto):      "auto",
-		int64(sessiondata.DistSQLOn):        "on",
-		int64(sessiondata.DistSQL2Dot0Auto): "2.0-auto",
-		int64(sessiondata.DistSQL2Dot0Off):  "2.0-off",
+		int64(sessiondata.DistSQLOff):  "off",
+		int64(sessiondata.DistSQLAuto): "auto",
+		int64(sessiondata.DistSQLOn):   "on",
 	},
 )
 
@@ -162,7 +207,7 @@ var (
 	}
 	MetaTxnAbort = metric.Metadata{
 		Name:        "sql.txn.abort.count",
-		Help:        "Number of SQL transaction ABORT statements",
+		Help:        "Number of SQL transaction abort errors",
 		Measurement: "SQL Statements",
 		Unit:        metric.Unit_COUNT,
 	}
@@ -202,6 +247,18 @@ var (
 		Measurement: "SQL Statements",
 		Unit:        metric.Unit_COUNT,
 	}
+	MetaSQLOptPlanCacheHits = metric.Metadata{
+		Name:        "sql.optimizer.plan_cache.hits",
+		Help:        "Number of non-prepared statements for which a cached plan was used",
+		Measurement: "SQL Statements",
+		Unit:        metric.Unit_COUNT,
+	}
+	MetaSQLOptPlanCacheMisses = metric.Metadata{
+		Name:        "sql.optimizer.plan_cache.misses",
+		Help:        "Number of non-prepared statements for which a cached plan was not used",
+		Measurement: "SQL Statements",
+		Unit:        metric.Unit_COUNT,
+	}
 	MetaDistSQLSelect = metric.Metadata{
 		Name:        "sql.distsql.select.count",
 		Help:        "Number of DistSQL SELECT statements",
@@ -238,6 +295,30 @@ var (
 		Measurement: "SQL Statements",
 		Unit:        metric.Unit_COUNT,
 	}
+	MetaSavepoint = metric.Metadata{
+		Name:        "sql.savepoint.count",
+		Help:        "Number of SQL SAVEPOINT statements",
+		Measurement: "SQL Statements",
+		Unit:        metric.Unit_COUNT,
+	}
+	MetaRestartSavepoint = metric.Metadata{
+		Name:        "sql.restart_savepoint.count",
+		Help:        "Number of `SAVEPOINT cockroach_restart` statements",
+		Measurement: "SQL Statements",
+		Unit:        metric.Unit_COUNT,
+	}
+	MetaReleaseRestartSavepoint = metric.Metadata{
+		Name:        "sql.restart_savepoint.release.count",
+		Help:        "Number of `RELEASE SAVEPOINT cockroach_restart` statements",
+		Measurement: "SQL Statements",
+		Unit:        metric.Unit_COUNT,
+	}
+	MetaRollbackToRestartSavepoint = metric.Metadata{
+		Name:        "sql.restart_savepoint.rollback.count",
+		Help:        "Number of `ROLLBACK TO SAVEPOINT cockroach_restart` statements",
+		Measurement: "SQL Statements",
+		Unit:        metric.Unit_COUNT,
+	}
 	MetaDdl = metric.Metadata{
 		Name:        "sql.ddl.count",
 		Help:        "Number of SQL DDL statements",
@@ -264,6 +345,15 @@ var (
 	}
 )
 
+func getMetricMeta(meta metric.Metadata, internal bool) metric.Metadata {
+	if internal {
+		meta.Name += ".internal"
+		meta.Help += " (internal queries)"
+		meta.Measurement = "SQL Internal Statements"
+	}
+	return meta
+}
+
 // NodeInfo contains metadata about the executing node and cluster.
 type NodeInfo struct {
 	ClusterID func() uuid.UUID
@@ -272,6 +362,12 @@ type NodeInfo struct {
 	PGURL     func(*url.Userinfo) (*url.URL, error)
 }
 
+// nodeStatusGenerator is a limited portion of the status.MetricsRecorder
+// struct, to avoid having to import all of status in sql.
+type nodeStatusGenerator interface {
+	GenerateNodeStatus(ctx context.Context) *statuspb.NodeStatus
+}
+
 // An ExecutorConfig encompasses the auxiliary objects and configuration
 // required to create an executor.
 // All fields holding a pointer or an interface are required to create
@@ -279,6 +375,7 @@ type NodeInfo struct {
 type ExecutorConfig struct {
 	Settings *cluster.Settings
 	NodeInfo
+	Locality         roachpb.Locality
 	AmbientCtx       log.AmbientContext
 	DB               *client.DB
 	Gossip           *gossip.Gossip
@@ -288,17 +385,20 @@ type ExecutorConfig struct {
 	Clock            *hlc.Clock
 	DistSQLSrv       *distsqlrun.ServerImpl
 	StatusServer     serverpb.StatusServer
-	MetricsRecorder  *status.MetricsRecorder
+	MetricsRecorder  nodeStatusGenerator
 	SessionRegistry  *SessionRegistry
 	JobRegistry      *jobs.Registry
 	VirtualSchemas   *VirtualSchemaHolder
 	DistSQLPlanner   *DistSQLPlanner
 	TableStatsCache  *stats.TableStatisticsCache
+	StatsRefresher   *stats.Refresher
 	ExecLogger       *log.SecondaryLogger
 	AuditLogger      *log.SecondaryLogger
 	InternalExecutor *InternalExecutor
+	QueryCache       *querycache.C
 
-	TestingKnobs              *ExecutorTestingKnobs
+	TestingKnobs              ExecutorTestingKnobs
+	PGWireTestingKnobs        *PGWireTestingKnobs
 	SchemaChangerTestingKnobs *SchemaChangerTestingKnobs
 	DistSQLRunTestingKnobs    *distsqlrun.TestingKnobs
 	EvalContextTestingKnobs   tree.EvalContextTestingKnobs
@@ -308,11 +408,6 @@ type ExecutorConfig struct {
 	// Caches updated by DistSQL.
 	RangeDescriptorCache *kv.RangeDescriptorCache
 	LeaseHolderCache     *kv.LeaseHolderCache
-
-	// ConnResultsBufferBytes is the size of the buffer in which each connection
-	// accumulates results set. Results are flushed to the network when this
-	// buffer overflows.
-	ConnResultsBufferBytes int
 }
 
 // Organization returns the value of cluster.organization.
@@ -372,6 +467,23 @@ type ExecutorTestingKnobs struct {
 	BeforeAutoCommit func(ctx context.Context, stmt string) error
 }
 
+// PGWireTestingKnobs contains knobs for the pgwire module.
+type PGWireTestingKnobs struct {
+	// CatchPanics causes the pgwire.conn to recover from panics in its execution
+	// thread and return them as errors to the client, closing the connection
+	// afterward.
+	CatchPanics bool
+
+	// AuthHook is used to override the normal authentication handling on new
+	// connections.
+	AuthHook func(context.Context) error
+}
+
+var _ base.ModuleTestingKnobs = &PGWireTestingKnobs{}
+
+// ModuleTestingKnobs implements the base.ModuleTestingKnobs interface.
+func (*PGWireTestingKnobs) ModuleTestingKnobs() {}
+
 // databaseCacheHolder is a thread-safe container for a *databaseCache.
 // It also allows clients to block until the cache is updated to a desired
 // state.
@@ -422,7 +534,7 @@ var _ dbCacheSubscriber = &databaseCacheHolder{}
 
 // updateSystemConfig is called whenever a new system config gossip entry is
 // received.
-func (dc *databaseCacheHolder) updateSystemConfig(cfg config.SystemConfig) {
+func (dc *databaseCacheHolder) updateSystemConfig(cfg *config.SystemConfig) {
 	dc.mu.Lock()
 	dc.mu.c = newDatabaseCache(cfg)
 	dc.mu.cv.Broadcast()
@@ -435,11 +547,6 @@ func (dc *databaseCacheHolder) updateSystemConfig(cfg config.SystemConfig) {
 func forEachRow(params runParams, p planNode, f func(tree.Datums) error) error {
 	next, err := p.Next(params)
 	for ; next; next, err = p.Next(params) {
-		// If we're tracking memory, clear the previous row's memory account.
-		if params.extendedEvalCtx.ActiveMemAcc != nil {
-			params.extendedEvalCtx.ActiveMemAcc.Clear(params.ctx)
-		}
-
 		if err := f(p.Values()); err != nil {
 			return err
 		}
@@ -467,29 +574,13 @@ func countRowsAffected(params runParams, p planNode) (int, error) {
 	return count, err
 }
 
-// shouldUseDistSQL returns true if the combination of mode and distribution
-// requirement given by shouldDistributeGivenRecAndMode requires that the
-// distSQL engine should be used.
-// This is always true unless the mode is set to 2.0-auto or 2.0-off.
-// 2.0-auto causes the distribution recommendation to control whether distsql is
-// used at all, and 2.0-off causes distsql to never be used regardless of the
-// recommendation.
-func shouldUseDistSQL(distributePlan bool, mode sessiondata.DistSQLExecMode) bool {
-	if mode == sessiondata.DistSQL2Dot0Off {
-		return false
-	} else if mode == sessiondata.DistSQL2Dot0Auto {
-		return distributePlan
-	}
-	return true
-}
-
 func shouldDistributeGivenRecAndMode(
 	rec distRecommendation, mode sessiondata.DistSQLExecMode,
 ) bool {
 	switch mode {
-	case sessiondata.DistSQLOff, sessiondata.DistSQL2Dot0Off:
+	case sessiondata.DistSQLOff:
 		return false
-	case sessiondata.DistSQLAuto, sessiondata.DistSQL2Dot0Auto:
+	case sessiondata.DistSQLAuto:
 		return rec == shouldDistribute
 	case sessiondata.DistSQLOn, sessiondata.DistSQLAlways:
 		return rec != cannotDistribute
@@ -545,7 +636,9 @@ func golangFillQueryArguments(args ...interface{}) tree.Datums {
 		case time.Time:
 			d = tree.MakeDTimestamp(t, time.Microsecond)
 		case time.Duration:
-			d = &tree.DInterval{Duration: duration.Duration{Nanos: t.Nanoseconds()}}
+			d = &tree.DInterval{Duration: duration.MakeDuration(t.Nanoseconds(), 0, 0)}
+		case bitarray.BitArray:
+			d = &tree.DBitArray{BitArray: t}
 		case *apd.Decimal:
 			dd := &tree.DDecimal{}
 			dd.Set(t)
@@ -584,10 +677,13 @@ func golangFillQueryArguments(args ...interface{}) tree.Datums {
 	return res
 }
 
+// checkResultType verifies that a table result can be returned to the
+// client.
 func checkResultType(typ types.T) error {
 	// Compare all types that can rely on == equality.
 	switch types.UnwrapType(typ) {
 	case types.Unknown:
+	case types.BitArray:
 	case types.Bool:
 	case types.Int:
 	case types.Float:
@@ -615,7 +711,13 @@ func checkResultType(typ types.T) error {
 		switch {
 		case istype(types.FamArray):
 			if istype(types.UnwrapType(typ).(types.TArray).Typ) {
-				return pgerror.Unimplemented("nested arrays", "arrays cannot have arrays as element type")
+				// Technically we could probably return arrays of arrays to a
+				// client (the encoding exists) but we don't want to give
+				// mixed signals -- that nested arrays appear to be supported
+				// in this case, and not in other cases (eg. CREATE). So we
+				// reject them in every case instead.
+				return pgerror.UnimplementedWithIssueDetailError(32552,
+					"result", "arrays cannot have arrays as element type")
 			}
 		case istype(types.FamCollatedString):
 		case istype(types.FamTuple):
@@ -630,10 +732,16 @@ func checkResultType(typ types.T) error {
 
 // EvalAsOfTimestamp evaluates and returns the timestamp from an AS OF SYSTEM
 // TIME clause.
-func (p *planner) EvalAsOfTimestamp(
-	asOf tree.AsOfClause, max hlc.Timestamp,
-) (hlc.Timestamp, error) {
-	return tree.EvalAsOfTimestamp(asOf, max, &p.semaCtx, p.EvalContext())
+func (p *planner) EvalAsOfTimestamp(asOf tree.AsOfClause) (_ hlc.Timestamp, err error) {
+	ts, err := tree.EvalAsOfTimestamp(asOf, &p.semaCtx, p.EvalContext())
+	if err != nil {
+		return hlc.Timestamp{}, err
+	}
+	if now := p.execCfg.Clock.Now(); now.Less(ts) {
+		return hlc.Timestamp{}, errors.Errorf(
+			"AS OF SYSTEM TIME: cannot specify timestamp in the future (%s > %s)", ts, now)
+	}
+	return ts, nil
 }
 
 // ParseHLC parses a string representation of an `hlc.Timestamp`.
@@ -649,11 +757,8 @@ func ParseHLC(s string) (hlc.Timestamp, error) {
 // that requires the transaction to be started already. If the returned
 // timestamp is not nil, it is the timestamp to which a transaction
 // should be set. The statements that will be checked are Select,
-// ShowTrace (of a Select statement), and Scrub.
-//
-// max is a lower bound on what the transaction's timestamp will be.
-// Used to check that the user didn't specify a timestamp in the future.
-func (p *planner) isAsOf(stmt tree.Statement, max hlc.Timestamp) (*hlc.Timestamp, error) {
+// ShowTrace (of a Select statement), Scrub, Export, and CreateStats.
+func (p *planner) isAsOf(stmt tree.Statement) (*hlc.Timestamp, error) {
 	var asOf tree.AsOfClause
 	switch s := stmt.(type) {
 	case *tree.Select:
@@ -679,12 +784,16 @@ func (p *planner) isAsOf(stmt tree.Statement, max hlc.Timestamp) (*hlc.Timestamp
 		}
 		asOf = s.AsOf
 	case *tree.Export:
-		return p.isAsOf(s.Query, max)
+		return p.isAsOf(s.Query)
+	case *tree.CreateStats:
+		if s.Options.AsOf.Expr == nil {
+			return nil, nil
+		}
+		asOf = s.Options.AsOf
 	default:
 		return nil, nil
 	}
-
-	ts, err := p.EvalAsOfTimestamp(asOf, max)
+	ts, err := p.EvalAsOfTimestamp(asOf)
 	return &ts, err
 }
 
@@ -754,7 +863,8 @@ type SessionArgs struct {
 	SessionDefaults SessionDefaults
 	// RemoteAddr is the client's address. This is nil iff this is an internal
 	// client.
-	RemoteAddr net.Addr
+	RemoteAddr            net.Addr
+	ConnResultsBufferSize int64
 }
 
 // isDefined returns true iff the SessionArgs is well-defined.
@@ -893,7 +1003,10 @@ func (scc *schemaChangerCollection) reset() {
 //
 // The list of closures is cleared after (attempting) execution.
 func (scc *schemaChangerCollection) execSchemaChanges(
-	ctx context.Context, cfg *ExecutorConfig, tracing *SessionTracing,
+	ctx context.Context,
+	cfg *ExecutorConfig,
+	tracing *SessionTracing,
+	ieFactory sqlutil.SessionBoundInternalExecutorFactory,
 ) error {
 	if len(scc.schemaChangers) == 0 {
 		return nil
@@ -909,8 +1022,9 @@ func (scc *schemaChangerCollection) execSchemaChanges(
 		sc.testingKnobs = cfg.SchemaChangerTestingKnobs
 		sc.distSQLPlanner = cfg.DistSQLPlanner
 		sc.settings = cfg.Settings
+		sc.ieFactory = ieFactory
 		for r := retry.Start(base.DefaultRetryOptions()); r.Next(); {
-			evalCtx := createSchemaChangeEvalCtx(cfg.Clock.Now(), tracing)
+			evalCtx := createSchemaChangeEvalCtx(ctx, cfg.Clock.Now(), tracing, ieFactory)
 			if err := sc.exec(ctx, true /* inSession */, &evalCtx); err != nil {
 				if onError := cfg.SchemaChangerTestingKnobs.OnError; onError != nil {
 					onError(err)
@@ -946,7 +1060,7 @@ func (scc *schemaChangerCollection) execSchemaChanges(
 	return firstError
 }
 
-const panicLogOutputCutoffChars = 500
+const panicLogOutputCutoffChars = 10000
 
 func anonymizeStmtAndConstants(stmt tree.Statement) string {
 	return tree.AsStringWithFlags(stmt, tree.FmtAnonymize|tree.FmtHideConstants)
@@ -960,8 +1074,8 @@ func AnonymizeStatementsForReporting(action, sqlStmts string, r interface{}) err
 	{
 		stmts, err := parser.Parse(sqlStmts)
 		if err == nil {
-			for _, stmt := range stmts {
-				anonymized = append(anonymized, anonymizeStmtAndConstants(stmt))
+			for i := range stmts {
+				anonymized = append(anonymized, anonymizeStmtAndConstants(stmts[i].AST))
 			}
 		}
 	}
@@ -1107,6 +1221,10 @@ func (st *SessionTracing) StartTracing(
 	opName := "session recording"
 	var sp opentracing.Span
 	connCtx := st.ex.ctxHolder.connCtx
+
+	// TODO(andrei): use tracing.EnsureChildSpan() or something more efficient
+	// than StartSpan(). The problem is that the current interface doesn't allow
+	// the Recordable option to be passed.
 	if parentSp := opentracing.SpanFromContext(connCtx); parentSp != nil {
 		// Create a child span while recording.
 		sp = parentSp.Tracer().StartSpan(
@@ -1206,14 +1324,14 @@ func (st *SessionTracing) TracePlanCheckEnd(ctx context.Context, err error, dist
 	if err != nil {
 		log.VEventfDepth(ctx, 2, 1, "distributability check error: %v", err)
 	} else {
-		log.VEventfDepth(ctx, 2, 1, "distributable plan: %v", dist)
+		log.VEventfDepth(ctx, 2, 1, "will distribute plan: %v", dist)
 	}
 }
 
 // TraceExecStart conditionally emits a trace message at the moment
 // plan execution starts.
 func (st *SessionTracing) TraceExecStart(ctx context.Context, engine string) {
-	log.VEventfDepth(ctx, 2, 1, "execution starts: %s", engine)
+	log.VEventfDepth(ctx, 2, 1, "execution starts: %s engine", engine)
 }
 
 // TraceExecConsume creates a context for TraceExecRowsResult below.
@@ -1367,9 +1485,7 @@ func generateSessionTraceVTable(spans []tracing.RecordedSpan) ([]traceRow, error
 			opMap[spanIdx] = tree.NewDString(lrr.span.Operation)
 			if lrr.span.Duration != 0 {
 				durMap[spanIdx] = &tree.DInterval{
-					Duration: duration.Duration{
-						Nanos: lrr.span.Duration.Nanoseconds(),
-					},
+					Duration: duration.MakeDuration(lrr.span.Duration.Nanoseconds(), 0, 0),
 				}
 			}
 		}
@@ -1423,7 +1539,7 @@ func generateSessionTraceVTable(spans []tracing.RecordedSpan) ([]traceRow, error
 
 		ts := res[i][traceTimestampCol].(*tree.DTimestampTZ)
 		res[i][traceAgeCol] = &tree.DInterval{
-			Duration: duration.Duration{Nanos: ts.Sub(minTimestamp).Nanoseconds()},
+			Duration: duration.MakeDuration(ts.Sub(minTimestamp).Nanoseconds(), 0, 0),
 		}
 	}
 
@@ -1553,8 +1669,8 @@ type sessionDataMutator struct {
 	data     *sessiondata.SessionData
 	defaults SessionDefaults
 	settings *cluster.Settings
-	// curTxnReadOnly is a value to be mutated through SET transaction_read_only = ...
-	curTxnReadOnly *bool
+	// setCurTxnReadOnly is called when we execute SET transaction_read_only = ...
+	setCurTxnReadOnly func(val bool)
 	// applicationNamedChanged, if set, is called when the "application name"
 	// variable is updated.
 	applicationNameChanged func(newName string)
@@ -1580,8 +1696,8 @@ func (m *sessionDataMutator) SetDatabase(dbName string) {
 	m.data.Database = dbName
 }
 
-func (m *sessionDataMutator) SetDefaultIsolationLevel(iso enginepb.IsolationType) {
-	m.data.DefaultIsolationLevel = iso
+func (m *sessionDataMutator) SetDefaultIntSize(size int) {
+	m.data.DefaultIntSize = size
 }
 
 func (m *sessionDataMutator) SetDefaultReadOnly(val bool) {
@@ -1592,8 +1708,8 @@ func (m *sessionDataMutator) SetDistSQLMode(val sessiondata.DistSQLExecMode) {
 	m.data.DistSQLMode = val
 }
 
-func (m *sessionDataMutator) SetLookupJoinEnabled(val bool) {
-	m.data.LookupJoinEnabled = val
+func (m *sessionDataMutator) SetForceSavepointRestart(val bool) {
+	m.data.ForceSavepointRestart = val
 }
 
 func (m *sessionDataMutator) SetForceSplitAt(val bool) {
@@ -1604,6 +1720,14 @@ func (m *sessionDataMutator) SetZigzagJoinEnabled(val bool) {
 	m.data.ZigzagJoinEnabled = val
 }
 
+func (m *sessionDataMutator) SetReorderJoinsLimit(val int) {
+	m.data.ReorderJoinsLimit = val
+}
+
+func (m *sessionDataMutator) SetVectorize(val sessiondata.VectorizeExecMode) {
+	m.data.Vectorize = val
+}
+
 func (m *sessionDataMutator) SetOptimizerMode(val sessiondata.OptimizerMode) {
 	m.data.OptimizerMode = val
 }
@@ -1625,13 +1749,17 @@ func (m *sessionDataMutator) SetLocation(loc *time.Location) {
 }
 
 func (m *sessionDataMutator) SetReadOnly(val bool) {
-	*m.curTxnReadOnly = val
+	m.setCurTxnReadOnly(val)
 }
 
 func (m *sessionDataMutator) SetStmtTimeout(timeout time.Duration) {
 	m.data.StmtTimeout = timeout
 }
 
+func (m *sessionDataMutator) SetAllowPrepareAsOptPlan(val bool) {
+	m.data.AllowPrepareAsOptPlan = val
+}
+
 // RecordLatestSequenceValue records that value to which the session incremented
 // a sequence.
 func (m *sessionDataMutator) RecordLatestSequenceVal(seqID uint32, val int64) {
@@ -1672,8 +1800,11 @@ func (s *sqlStatsCollectorImpl) PhaseTimes() *phaseTimes {
 }
 
 // RecordStatement is part of the sqlStatsCollector interface.
+//
+// samplePlanDescription can be nil, as these are only sampled periodically per unique fingerprint.
 func (s *sqlStatsCollectorImpl) RecordStatement(
-	stmt Statement,
+	stmt *Statement,
+	samplePlanDescription *roachpb.ExplainTreePlanNode,
 	distSQLUsed bool,
 	optUsed bool,
 	automaticRetryCount int,
@@ -1682,7 +1813,7 @@ func (s *sqlStatsCollectorImpl) RecordStatement(
 	parseLat, planLat, runLat, svcLat, ovhLat float64,
 ) {
 	s.appStats.recordStatement(
-		stmt, distSQLUsed, optUsed, automaticRetryCount, numRows, err,
+		stmt, samplePlanDescription, distSQLUsed, optUsed, automaticRetryCount, numRows, err,
 		parseLat, planLat, runLat, svcLat, ovhLat)
 }
 
