@@ -121,8 +121,25 @@ func doExpandPlan(
 		n.source, err = doExpandPlan(ctx, p, noParams, n.source)
 
 	case *deleteNode:
+		// If the source of the delete is a scan node (optionally with a render on
+		// top), mark it as such. Note that this parallels the logic in
+		// canDeleteFast.
+		maybeScan := n.source
+		if sel, ok := maybeScan.(*renderNode); ok {
+			maybeScan = sel.source.plan
+		}
+		scan, ok := maybeScan.(*scanNode)
+		if ok {
+			scan.isDeleteSource = true
+		}
+
 		n.source, err = doExpandPlan(ctx, p, noParams, n.source)
 
+		if fastDelete, ok := maybeCreateDeleteFastNode(
+			ctx, n.source, n.run.td.tableDesc(), n.run.fastPathInterleaved, n.run.rowsNeeded); ok {
+			plan = fastDelete
+		}
+
 	case *rowCountNode:
 		var newPlan planNode
 		newPlan, err = doExpandPlan(ctx, p, noParams, n.source)
@@ -271,6 +288,9 @@ func doExpandPlan(
 	case *windowNode:
 		n.plan, err = doExpandPlan(ctx, p, noParams, n.plan)
 
+	case *max1RowNode:
+		n.plan, err = doExpandPlan(ctx, p, noParams, n.plan)
+
 	case *sortNode:
 		if !n.ordering.IsPrefixOf(params.desiredOrdering) {
 			params.desiredOrdering = n.ordering
@@ -330,10 +350,14 @@ func doExpandPlan(
 		n.source, err = doExpandPlan(ctx, p, noParams, n.source)
 
 	case *valuesNode:
+	case *virtualTableNode:
 	case *alterIndexNode:
 	case *alterTableNode:
 	case *alterSequenceNode:
 	case *alterUserSetPasswordNode:
+	case *commentOnColumnNode:
+	case *commentOnDatabaseNode:
+	case *commentOnTableNode:
 	case *renameColumnNode:
 	case *renameDatabaseNode:
 	case *renameIndexNode:
@@ -366,7 +390,6 @@ func doExpandPlan(
 	case *setClusterSettingNode:
 	case *setZoneConfigNode:
 	case *showZoneConfigNode:
-	case *showRangesNode:
 	case *showFingerprintsNode:
 	case *showTraceNode:
 	case *scatterNode:
@@ -665,6 +688,8 @@ func (p *planner) simplifyOrderings(plan planNode, usefulOrdering sqlbase.Column
 	case *deleteNode:
 		n.source = p.simplifyOrderings(n.source, nil)
 
+	case *deleteRangeNode:
+
 	case *rowCountNode:
 		n.source = p.simplifyOrderings(n.source, nil).(batchedPlanNode)
 
@@ -739,6 +764,9 @@ func (p *planner) simplifyOrderings(plan planNode, usefulOrdering sqlbase.Column
 	case *limitNode:
 		n.plan = p.simplifyOrderings(n.plan, usefulOrdering)
 
+	case *max1RowNode:
+		n.plan = p.simplifyOrderings(n.plan, usefulOrdering)
+
 	case *spoolNode:
 		n.source = p.simplifyOrderings(n.source, usefulOrdering)
 
@@ -836,10 +864,14 @@ func (p *planner) simplifyOrderings(plan planNode, usefulOrdering sqlbase.Column
 		n.rows = p.simplifyOrderings(n.rows, nil)
 
 	case *valuesNode:
+	case *virtualTableNode:
 	case *alterIndexNode:
 	case *alterTableNode:
 	case *alterSequenceNode:
 	case *alterUserSetPasswordNode:
+	case *commentOnColumnNode:
+	case *commentOnDatabaseNode:
+	case *commentOnTableNode:
 	case *renameColumnNode:
 	case *renameDatabaseNode:
 	case *renameIndexNode:
@@ -866,7 +898,6 @@ func (p *planner) simplifyOrderings(plan planNode, usefulOrdering sqlbase.Column
 	case *setClusterSettingNode:
 	case *setZoneConfigNode:
 	case *showZoneConfigNode:
-	case *showRangesNode:
 	case *showFingerprintsNode:
 	case *showTraceNode:
 	case *scatterNode:
