@@ -19,14 +19,15 @@ import (
 	"fmt"
 	"sync"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/privilege"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 var insertNodePool = sync.Pool{
@@ -111,20 +112,27 @@ func (p *planner) Insert(
 		}
 	}
 
-	// Determine what are the foreign key tables that are involved in the update.
-	var fkCheckType sqlbase.FKCheck
-	if n.OnConflict == nil || n.OnConflict.DoNothing {
-		fkCheckType = sqlbase.CheckInserts
-	} else {
-		fkCheckType = sqlbase.CheckUpdates
+	// Set up any check constraints.
+	checkHelper, err := sqlbase.NewEvalCheckHelper(ctx, p.analyzeExpr, desc)
+	if err != nil {
+		return nil, err
 	}
-	fkTables, err := sqlbase.TablesNeededForFKs(
+
+	// Determine what are the foreign key tables that are involved in the update.
+	var fkCheckType row.FKCheckType
+	if n.OnConflict == nil || n.OnConflict.DoNothing {
+		fkCheckType = row.CheckInserts
+	} else {
+		fkCheckType = row.CheckUpdates
+	}
+	fkTables, err := row.MakeFkMetadata(
 		ctx,
-		*desc,
+		desc,
 		fkCheckType,
-		p.lookupFKTable,
+		p.LookupTableByID,
 		p.CheckPrivilege,
 		p.analyzeExpr,
+		checkHelper,
 	)
 	if err != nil {
 		return nil, err
@@ -140,17 +148,7 @@ func (p *planner) Insert(
 		// not care: the backfill is also inserting defaults, and we do
 		// not provide guarantees about the evaluation order of default
 		// expressions.
-		insertCols = desc.Columns
-		// Add mutation columns.
-		if len(desc.Mutations) > 0 {
-			insertCols = make([]sqlbase.ColumnDescriptor, 0, len(desc.Columns)+len(desc.Mutations))
-			insertCols = append(insertCols, desc.Columns...)
-			for i := range desc.Mutations {
-				if c := desc.Mutations[i].GetColumn(); c != nil && desc.Mutations[i].State == sqlbase.DescriptorMutation_DELETE_AND_WRITE_ONLY {
-					insertCols = append(insertCols, *c)
-				}
-			}
-		}
+		insertCols = desc.WritableColumns()
 	} else {
 		var err error
 		if insertCols, err = p.processColumns(desc, n.Columns,
@@ -230,7 +228,7 @@ func (p *planner) Insert(
 					// TODO(justin): this is too restrictive. It should
 					// be possible to allow INSERT INTO (x) VALUES (DEFAULT)
 					// if x is a computed column. See #22434.
-					return nil, sqlbase.CannotWriteToComputedColError(&insertCols[maxInsertIdx])
+					return nil, sqlbase.CannotWriteToComputedColError(insertCols[maxInsertIdx].Name)
 				}
 				arityChecked = true
 			}
@@ -260,7 +258,7 @@ func (p *planner) Insert(
 			return nil, err
 		}
 		if numExprs > maxInsertIdx {
-			return nil, sqlbase.CannotWriteToComputedColError(&insertCols[maxInsertIdx])
+			return nil, sqlbase.CannotWriteToComputedColError(insertCols[maxInsertIdx].Name)
 		}
 	}
 
@@ -275,8 +273,8 @@ func (p *planner) Insert(
 	}
 
 	// Create the table insert, which does the bulk of the work.
-	ri, err := sqlbase.MakeRowInserter(p.txn, desc, fkTables, insertCols,
-		sqlbase.CheckFKs, &p.alloc)
+	ri, err := row.MakeInserter(p.txn, desc, fkTables, insertCols,
+		row.CheckFKs, &p.alloc)
 	if err != nil {
 		return nil, err
 	}
@@ -313,7 +311,7 @@ func (p *planner) Insert(
 			columns: columns,
 			run: insertRun{
 				ti:           tableInserter{ri: ri},
-				checkHelper:  fkTables[desc.ID].CheckHelper,
+				checkHelper:  checkHelper,
 				rowsNeeded:   rowsNeeded,
 				computedCols: computedCols,
 				computeExprs: computeExprs,
@@ -368,7 +366,7 @@ type insertRun struct {
 	done bool
 
 	// rows contains the accumulated result rows if rowsNeeded is set.
-	rows *sqlbase.RowContainer
+	rows *rowcontainer.RowContainer
 
 	// resultRowBuffer is used to prepare a result row for accumulation
 	// into the row container above, when rowsNeeded is set.
@@ -381,10 +379,6 @@ type insertRun struct {
 	// index is not public.
 	rowIdxToRetIdx []int
 
-	// autoCommit indicates whether the last KV batch processed by
-	// this update will also commit the KV txn.
-	autoCommit autoCommitOpt
-
 	// traceKV caches the current KV tracing flag.
 	traceKV bool
 }
@@ -404,7 +398,7 @@ func (n *insertNode) startExec(params runParams) error {
 	n.run.traceKV = params.p.ExtendedEvalContext().Tracing.KVTracingEnabled()
 
 	if n.run.rowsNeeded {
-		n.run.rows = sqlbase.NewRowContainer(
+		n.run.rows = rowcontainer.NewRowContainer(
 			params.EvalContext().Mon.MakeBoundAccount(),
 			sqlbase.ColTypeInfoFromResCols(n.columns), 0)
 
@@ -511,13 +505,16 @@ func (n *insertNode) BatchedNext(params runParams) (bool, error) {
 	}
 
 	if lastBatch {
-		if _, err := n.run.ti.finalize(params.ctx, n.run.autoCommit, n.run.traceKV); err != nil {
+		if _, err := n.run.ti.finalize(params.ctx, n.run.traceKV); err != nil {
 			return false, err
 		}
 		// Remember we're done for the next call to BatchedNext().
 		n.run.done = true
 	}
 
+	// Possibly initiate a run of CREATE STATISTICS.
+	params.ExecCfg().StatsRefresher.NotifyMutation(n.run.ti.tableDesc().ID, n.run.rowCount)
+
 	return n.run.rowCount > 0, nil
 }
 
@@ -532,7 +529,7 @@ func (n *insertNode) processSourceRow(params runParams, sourceVals tree.Datums)
 		n.run.computeExprs,
 		n.run.insertCols,
 		n.run.computedCols,
-		*params.EvalContext(),
+		*params.EvalContext().Copy(),
 		n.run.ti.tableDesc(),
 		sourceVals,
 		&n.run.iVarContainerForComputedCols,
@@ -541,19 +538,29 @@ func (n *insertNode) processSourceRow(params runParams, sourceVals tree.Datums)
 		return err
 	}
 
-	// Run the CHECK constraints, if any.
-	if len(n.run.checkHelper.Exprs) > 0 {
-		if err := n.run.checkHelper.LoadRow(n.run.ti.ri.InsertColIDtoRowIndex, rowVals, false); err != nil {
-			return err
-		}
-		if err := n.run.checkHelper.Check(params.EvalContext()); err != nil {
-			return err
+	// Run the CHECK constraints, if any. CheckHelper will either evaluate the
+	// constraints itself, or else inspect boolean columns from the input that
+	// contain the results of evaluation.
+	if n.run.checkHelper != nil {
+		if n.run.checkHelper.NeedsEval() {
+			if err := n.run.checkHelper.LoadEvalRow(
+				n.run.ti.ri.InsertColIDtoRowIndex, rowVals, false); err != nil {
+				return err
+			}
+			if err := n.run.checkHelper.CheckEval(params.EvalContext()); err != nil {
+				return err
+			}
+		} else {
+			checkVals := rowVals[len(n.run.insertCols):]
+			if err := n.run.checkHelper.CheckInput(checkVals); err != nil {
+				return err
+			}
+			rowVals = rowVals[:len(n.run.insertCols)]
 		}
 	}
 
 	// Queue the insert in the KV batch.
-	_, err = n.run.ti.row(params.ctx, rowVals, n.run.traceKV)
-	if err != nil {
+	if err = n.run.ti.row(params.ctx, rowVals, n.run.traceKV); err != nil {
 		return err
 	}
 
@@ -593,7 +600,7 @@ func (n *insertNode) Close(ctx context.Context) {
 
 // enableAutoCommit is part of the autoCommitNode interface.
 func (n *insertNode) enableAutoCommit() {
-	n.run.autoCommit = autoCommitEnabled
+	n.run.ti.enableAutoCommit()
 }
 
 // GenerateInsertRow prepares a row tuple for insertion. It fills in default
@@ -618,7 +625,7 @@ func GenerateInsertRow(
 	insertCols []sqlbase.ColumnDescriptor,
 	computedCols []sqlbase.ColumnDescriptor,
 	evalCtx tree.EvalContext,
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.ImmutableTableDescriptor,
 	rowVals tree.Datums,
 	rowContainerForComputedVals *sqlbase.RowIndexedVarContainer,
 ) (tree.Datums, error) {
@@ -665,33 +672,39 @@ func GenerateInsertRow(
 		evalCtx.PopIVarContainer()
 	}
 
+	// Verify the column constraints.
+	//
+	// We would really like to use enforceLocalColumnConstraints() here,
+	// but this is not possible because of some brain damage in the
+	// Insert() constructor, which causes insertCols to contain
+	// duplicate columns descriptors: computed columns are listed twice,
+	// one will receive a NULL value and one will receive a comptued
+	// value during execution. It "works out in the end" because the
+	// latter (non-NULL) value overwrites the earlier, but
+	// enforceLocalColumnConstraints() does not know how to reason about
+	// this.
+	//
+	// In the end it does not matter much, this code is going away in
+	// favor of the (simpler, correct) code in the CBO.
+
 	// Check to see if NULL is being inserted into any non-nullable column.
-	checkNullViolation := func(col *sqlbase.ColumnDescriptor) error {
-		if i, ok := rowContainerForComputedVals.Mapping[col.ID]; !col.Nullable && (!ok || rowVals[i] == tree.DNull) {
-			return sqlbase.NewNonNullViolationError(col.Name)
-		}
-		return nil
-	}
-	for _, col := range tableDesc.Columns {
-		if err := checkNullViolation(&col); err != nil {
-			return nil, err
-		}
-	}
-	for _, m := range tableDesc.Mutations {
-		if c := m.GetColumn(); c != nil && m.State == sqlbase.DescriptorMutation_DELETE_AND_WRITE_ONLY {
-			if err := checkNullViolation(c); err != nil {
-				return nil, err
+	for _, col := range tableDesc.WritableColumns() {
+		if !col.Nullable {
+			if i, ok := rowContainerForComputedVals.Mapping[col.ID]; !ok || rowVals[i] == tree.DNull {
+				return nil, sqlbase.NewNonNullViolationError(col.Name)
 			}
 		}
 	}
 
 	// Ensure that the values honor the specified column widths.
-	for i := range rowVals {
-		if err := sqlbase.CheckValueWidth(
-			insertCols[i].Type, rowVals[i], &insertCols[i].Name); err != nil {
+	for i := 0; i < len(insertCols); i++ {
+		outVal, err := sqlbase.LimitValueWidth(insertCols[i].Type, rowVals[i], &insertCols[i].Name)
+		if err != nil {
 			return nil, err
 		}
+		rowVals[i] = outVal
 	}
+
 	return rowVals, nil
 }
 
@@ -702,7 +715,9 @@ func GenerateInsertRow(
 // are returned. If allowMutations is set, even columns undergoing
 // mutations are added.
 func (p *planner) processColumns(
-	tableDesc *sqlbase.TableDescriptor, nameList tree.NameList, ensureColumns, allowMutations bool,
+	tableDesc *sqlbase.ImmutableTableDescriptor,
+	nameList tree.NameList,
+	ensureColumns, allowMutations bool,
 ) ([]sqlbase.ColumnDescriptor, error) {
 	if len(nameList) == 0 {
 		if ensureColumns {
@@ -730,7 +745,8 @@ func (p *planner) processColumns(
 		}
 
 		if _, ok := colIDSet[col.ID]; ok {
-			return nil, fmt.Errorf("multiple assignments to the same column %q", &nameList[i])
+			return nil, pgerror.NewErrorf(pgerror.CodeSyntaxError,
+				"multiple assignments to the same column %q", &nameList[i])
 		}
 		colIDSet[col.ID] = struct{}{}
 		cols[i] = col
