@@ -17,7 +17,6 @@ package sql
 import (
 	"context"
 	"math"
-	"strconv"
 	"sync"
 
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
@@ -27,6 +26,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgwirebase"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
+	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlutil"
@@ -74,15 +74,18 @@ type internalExecutorImpl struct {
 
 	// sessionData, if not nil, represents the session variables used by
 	// statements executed on this internalExecutor. This field supports the
-	// implementation of SessionBoundInternalExecutor. Note that different
-	// executors never share sessionData; instead they make copies. Updating the
-	// sessionData through an internalExecutor doesn't affect the real client
-	// session from whence it came (and thus updating this data is probably not
-	// what you want).
-	// NOTE: We could consider allowing an internalExecutor to share/update the
-	// parent session state if need be, but then we'd need to share a
-	// sessionDataMutator.
+	// implementation of SessionBoundInternalExecutor. Note that a session bound
+	// internal executor cannot modify session data.
 	sessionData *sessiondata.SessionData
+
+	// The internal executor uses its own TableCollection. A TableCollection
+	// is a schema cache for each transaction and contains data like the schema
+	// modified by a transaction. Occasionally an internal executor is called
+	// within the context of a transaction that has modified the schema, the
+	// internal executor should see the modified schema. This interface allows
+	// the internal executor to modify its TableCollection to match the
+	// TableCollection of the parent executor.
+	tcModifier tableCollectionModifier
 }
 
 // MakeInternalExecutor creates an InternalExecutor.
@@ -107,14 +110,14 @@ func MakeInternalExecutor(
 	}
 }
 
-// MakeSessionBoundInternalExecutor creates a SessionBoundInternalExecutor.
-func MakeSessionBoundInternalExecutor(
+// NewSessionBoundInternalExecutor creates a SessionBoundInternalExecutor.
+func NewSessionBoundInternalExecutor(
 	ctx context.Context,
 	sessionData *sessiondata.SessionData,
 	s *Server,
 	memMetrics MemoryMetrics,
 	settings *cluster.Settings,
-) SessionBoundInternalExecutor {
+) *SessionBoundInternalExecutor {
 	monitor := mon.MakeUnlimitedMonitor(
 		ctx,
 		"internal SQL executor",
@@ -124,15 +127,12 @@ func MakeSessionBoundInternalExecutor(
 		math.MaxInt64, /* noteworthy */
 		settings,
 	)
-	dataCopy := sessionData.Copy()
-	return SessionBoundInternalExecutor{
+	return &SessionBoundInternalExecutor{
 		impl: internalExecutorImpl{
-			s:          s,
-			mon:        &monitor,
-			memMetrics: memMetrics,
-			// The internal executor gets a copy of the session data, so it can't
-			// update the parent session.
-			sessionData: &dataCopy,
+			s:           s,
+			mon:         &monitor,
+			memMetrics:  memMetrics,
+			sessionData: sessionData,
 		},
 	}
 }
@@ -159,17 +159,19 @@ func (ie *internalExecutorImpl) initConnEx(
 		lastDelivered: -1,
 	}
 
-	var sp sessionParams
+	var sd *sessiondata.SessionData
+	var sdMut *sessionDataMutator
 	if sargs.isDefined() {
 		if ie.sessionData != nil {
 			log.Fatal(ctx, "sargs used on a session bound executor")
 		}
-		sp.args = sargs
+		sd, sdMut = ie.s.newSessionDataAndMutator(sargs)
 	} else {
 		if ie.sessionData == nil {
 			log.Fatal(ctx, "initConnEx called with no sargs, and the executor is also not session bound")
 		}
-		sp.data = ie.sessionData
+		sd = ie.sessionData
+		// sdMut stays nil for session bound executors.
 	}
 
 	stmtBuf := NewStmtBuf()
@@ -178,39 +180,32 @@ func (ie *internalExecutorImpl) initConnEx(
 	if txn == nil {
 		ex, err = ie.s.newConnExecutor(
 			ctx,
-			sp,
+			sd, sdMut,
 			stmtBuf,
 			clientComm,
-			ie.memMetrics)
+			ie.memMetrics,
+			&ie.s.InternalMetrics)
 	} else {
 		ex, err = ie.s.newConnExecutorWithTxn(
 			ctx,
-			sp,
+			sd, sdMut,
 			stmtBuf,
 			clientComm,
 			ie.mon,
 			ie.memMetrics,
-			txn)
+			&ie.s.InternalMetrics,
+			txn,
+			ie.tcModifier)
 	}
 	if err != nil {
 		return nil, nil, err
 	}
-	ex.stmtCounterDisabled = true
 
 	var wg sync.WaitGroup
 	wg.Add(1)
 	go func() {
-		// Once ex.run() returns, we call ex.resetExtraTxnState() to release the
-		// leases that might have been acquired by the connEx. If there is no
-		// higher-level txn, then the leases were already released when the implicit
-		// txn committed. But if there is a higher-level txn, they don't get
-		// released.
-		defer func() {
-			if err := ex.resetExtraTxnState(ctx, ex.server.dbCache); err != nil {
-				log.Warningf(ctx, "error while cleaning up connExecutor: %s", err)
-			}
-		}()
 		if err := ex.run(ctx, ie.mon, mon.BoundAccount{} /*reserved*/, nil /* cancel */); err != nil {
+			ex.recordError(err)
 			errCallback(err)
 		}
 		closeMode := normalClose
@@ -229,6 +224,19 @@ func (ie *internalExecutorImpl) initConnEx(
 // If txn is not nil, the statement will be executed in the respective txn.
 func (ie *InternalExecutor) Query(
 	ctx context.Context, opName string, txn *client.Txn, stmt string, qargs ...interface{},
+) ([]tree.Datums, error) {
+	datums, _, err := ie.queryInternal(
+		ctx, opName, txn,
+		internalExecRootSession,
+		SessionArgs{},
+		stmt, qargs...)
+	return datums, err
+}
+
+// QueryWithCols is like Query, but it also returns the computed ResultColumns
+// of the input query.
+func (ie *InternalExecutor) QueryWithCols(
+	ctx context.Context, opName string, txn *client.Txn, stmt string, qargs ...interface{},
 ) ([]tree.Datums, sqlbase.ResultColumns, error) {
 	return ie.queryInternal(
 		ctx, opName, txn,
@@ -272,7 +280,7 @@ func (ie *InternalExecutor) QueryWithUser(
 func (ie *InternalExecutor) QueryRow(
 	ctx context.Context, opName string, txn *client.Txn, stmt string, qargs ...interface{},
 ) (tree.Datums, error) {
-	rows, _, err := ie.Query(ctx, opName, txn, stmt, qargs...)
+	rows, err := ie.Query(ctx, opName, txn, stmt, qargs...)
 	if err != nil {
 		return nil, err
 	}
@@ -328,6 +336,19 @@ func (ie *InternalExecutor) ExecWithUser(
 // If txn is not nil, the statement will be executed in the respective txn.
 func (ie *SessionBoundInternalExecutor) Query(
 	ctx context.Context, opName string, txn *client.Txn, stmt string, qargs ...interface{},
+) ([]tree.Datums, error) {
+	rows, _, err := ie.impl.queryInternal(
+		ctx, opName, txn,
+		internalExecInheritSession,
+		SessionArgs{},
+		stmt, qargs...)
+	return rows, err
+}
+
+// QueryWithCols is like Query, but it also returns the computed ResultColumns
+// of the input query.
+func (ie *SessionBoundInternalExecutor) QueryWithCols(
+	ctx context.Context, opName string, txn *client.Txn, stmt string, qargs ...interface{},
 ) ([]tree.Datums, sqlbase.ResultColumns, error) {
 	return ie.impl.queryInternal(
 		ctx, opName, txn,
@@ -441,7 +462,7 @@ func (ie *internalExecutorImpl) execInternal(
 	case internalExecRootSession:
 		sargs.SessionDefaults = map[string]string{
 			"database":         "system",
-			"application_name": InternalAppNamePrefix + "internal-" + opName,
+			"application_name": InternalAppNamePrefix + "-" + opName,
 		}
 	}
 
@@ -457,12 +478,12 @@ func (ie *internalExecutorImpl) execInternal(
 		}
 	}()
 
-	ctx, finishSp := tracing.EnsureChildSpan(ctx, ie.s.cfg.AmbientCtx.Tracer, opName)
-	defer finishSp()
+	ctx, sp := tracing.EnsureChildSpan(ctx, ie.s.cfg.AmbientCtx.Tracer, opName)
+	defer sp.Finish()
 
 	timeReceived := timeutil.Now()
 	parseStart := timeReceived
-	s, err := parser.ParseOne(stmt)
+	parsed, err := parser.ParseOne(stmt)
 	if err != nil {
 		return result{}, err
 	}
@@ -504,17 +525,17 @@ func (ie *internalExecutorImpl) execInternal(
 	// Transforms the args to datums. The datum types will be passed as type hints
 	// to the PrepareStmt command.
 	datums := golangFillQueryArguments(qargs...)
-	types := make(tree.PlaceholderTypes, len(datums))
+	typeHints := make(tree.PlaceholderTypes, len(datums))
 	for i, d := range datums {
 		// Arg numbers start from 1.
-		types[strconv.Itoa(i+1)] = d.ResolvedType()
+		typeHints[types.PlaceholderIdx(i)] = d.ResolvedType()
 	}
 	if len(qargs) == 0 {
 		resPos = 0
 		if err := stmtBuf.Push(
 			ctx,
 			ExecStmt{
-				Stmt:         s,
+				Statement:    parsed,
 				TimeReceived: timeReceived,
 				ParseStart:   parseStart,
 				ParseEnd:     parseEnd,
@@ -526,10 +547,10 @@ func (ie *internalExecutorImpl) execInternal(
 		if err := stmtBuf.Push(
 			ctx,
 			PrepareStmt{
-				Stmt:       s,
+				Statement:  parsed,
 				ParseStart: parseStart,
 				ParseEnd:   parseEnd,
-				TypeHints:  types,
+				TypeHints:  typeHints,
 			},
 		); err != nil {
 			return result{}, err
@@ -637,7 +658,7 @@ func (icc *internalClientComm) Flush(pos CmdPos) error {
 
 // CreateDescribeResult is part of the ClientComm interface.
 func (icc *internalClientComm) CreateDescribeResult(pos CmdPos) DescribeResult {
-	panic("unimplemented")
+	return icc.createRes(pos, nil /* onClose */)
 }
 
 // CreateDeleteResult is part of the ClientComm interface.
