@@ -25,8 +25,6 @@ import (
 	"sync/atomic"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
@@ -37,7 +35,6 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
@@ -46,6 +43,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil/singleflight"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 var errRenewLease = errors.New("renew lease on id")
@@ -65,7 +63,7 @@ type storedTableLease struct {
 type tableVersionState struct {
 	// This descriptor is immutable and can be shared by many goroutines.
 	// Care must be taken to not modify it.
-	sqlbase.TableDescriptor
+	sqlbase.ImmutableTableDescriptor
 
 	// The expiration time for the table version. A transaction with
 	// timestamp T can use this table descriptor version iff
@@ -108,6 +106,14 @@ func (s *tableVersionState) hasExpired(timestamp hlc.Timestamp) bool {
 	return !timestamp.Less(s.expiration)
 }
 
+// hasValidExpiration checks that this table have a larger expiration than
+// the existing one it is replacing. This can be used to check the
+// monotonicity of the expiration times on a table at a particular version.
+// The version is not explicitly checked here.
+func (s *tableVersionState) hasValidExpiration(existing *tableVersionState) bool {
+	return existing.expiration.Less(s.expiration)
+}
+
 func (s *tableVersionState) incRefcount() {
 	s.mu.Lock()
 	s.incRefcountLocked()
@@ -165,11 +171,20 @@ func (s LeaseStore) jitteredLeaseDuration() time.Duration {
 // acquire a lease on the most recent version of a table descriptor.
 // If the lease cannot be obtained because the descriptor is in the process of
 // being dropped, the error will be errTableDropped.
-func (s LeaseStore) acquire(ctx context.Context, tableID sqlbase.ID) (*tableVersionState, error) {
+// The expiration time set for the lease > minExpiration.
+func (s LeaseStore) acquire(
+	ctx context.Context, minExpiration hlc.Timestamp, tableID sqlbase.ID,
+) (*tableVersionState, error) {
 	var table *tableVersionState
 	err := s.execCfg.DB.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
 		expiration := txn.OrigTimestamp()
 		expiration.WallTime += int64(s.jitteredLeaseDuration())
+		if !minExpiration.Less(expiration) {
+			// In the rare circumstances where expiration <= minExpiration
+			// use an expiration based on the minExpiration to guarantee
+			// a monotonically increasing expiration.
+			expiration = minExpiration.Add(int64(time.Millisecond), 0)
+		}
 
 		tableDesc, err := sqlbase.GetTableDescFromID(ctx, txn, tableID)
 		if err != nil {
@@ -187,8 +202,8 @@ func (s LeaseStore) acquire(ctx context.Context, tableID sqlbase.ID) (*tableVers
 			expiration: storedLeaseExpiration(expiration),
 		}
 		table = &tableVersionState{
-			TableDescriptor: *tableDesc,
-			expiration:      expiration,
+			ImmutableTableDescriptor: *sqlbase.NewImmutableTableDescriptor(*tableDesc),
+			expiration:               expiration,
 		}
 		table.mu.lease = storedLease
 
@@ -210,7 +225,7 @@ func (s LeaseStore) acquire(ctx context.Context, tableID sqlbase.ID) (*tableVers
 		// read from the database for the special descriptor of a system table
 		// (#23937).
 		insertLease := fmt.Sprintf(
-			`INSERT INTO system.lease ("descID", version, "nodeID", expiration) VALUES (%d, %d, %d, %s)`,
+			`INSERT INTO system.public.lease ("descID", version, "nodeID", expiration) VALUES (%d, %d, %d, %s)`,
 			storedLease.id, storedLease.version, nodeID, &storedLease.expiration,
 		)
 		count, err := s.execCfg.InternalExecutor.Exec(ctx, "lease-insert", txn, insertLease)
@@ -244,7 +259,7 @@ func (s LeaseStore) release(ctx context.Context, stopper *stop.Stopper, lease *s
 		if nodeID == 0 {
 			panic("zero nodeID")
 		}
-		const deleteLease = `DELETE FROM system.lease ` +
+		const deleteLease = `DELETE FROM system.public.lease ` +
 			`WHERE ("descID", version, "nodeID", expiration) = ($1, $2, $3, $4)`
 		count, err := s.execCfg.InternalExecutor.Exec(
 			ctx,
@@ -298,13 +313,7 @@ func (s LeaseStore) WaitForOneVersion(
 		// Check to see if there are any leases that still exist on the previous
 		// version of the descriptor.
 		now := s.execCfg.Clock.Now()
-		tables := []IDVersion{
-			{
-				name:    tableDesc.Name,
-				id:      tableDesc.ID,
-				version: tableDesc.Version - 1,
-			},
-		}
+		tables := []IDVersion{NewIDVersionPrev(tableDesc)}
 		count, err := CountLeases(ctx, s.execCfg.InternalExecutor, tables, now)
 		if err != nil {
 			return 0, err
@@ -317,16 +326,14 @@ func (s LeaseStore) WaitForOneVersion(
 	return tableDesc.Version, nil
 }
 
-func incrementVersion(
-	ctx context.Context, tableDesc *sqlbase.TableDescriptor, txn *client.Txn,
+func maybeIncrementVersion(
+	ctx context.Context, desc *sqlbase.MutableTableDescriptor, txn *client.Txn,
 ) error {
-	// Use SERIALIZABLE transactions so that the ModificationTime on the
-	// descriptor is the commit timestamp.
-	if txn.Isolation() != enginepb.SERIALIZABLE {
-		return pgerror.NewErrorf(pgerror.CodeInvalidTransactionStateError,
-			"transaction involving a schemas change needs to be SERIALIZABLE")
+	// Already incremented, no-op.
+	if desc.Version == desc.ClusterVersion.Version+1 {
+		return nil
 	}
-	tableDesc.Version++
+	desc.Version++
 	// We need to set ModificationTime to the transaction's commit
 	// timestamp. Using CommitTimestamp() guarantees that the
 	// transaction will commit at the CommitTimestamp().
@@ -341,9 +348,9 @@ func incrementVersion(
 	// so updating this policy will also need to consider not doing
 	// that.
 	modTime := txn.CommitTimestamp()
-	tableDesc.ModificationTime = modTime
+	desc.ModificationTime = modTime
 	log.Infof(ctx, "publish: descID=%d (%s) version=%d mtime=%s",
-		tableDesc.ID, tableDesc.Name, tableDesc.Version, modTime.GoTime())
+		desc.ID, desc.Name, desc.Version, modTime.GoTime())
 	return nil
 }
 
@@ -362,9 +369,9 @@ var errDidntUpdateDescriptor = errors.New("didn't update the table descriptor")
 func (s LeaseStore) Publish(
 	ctx context.Context,
 	tableID sqlbase.ID,
-	update func(*sqlbase.TableDescriptor) error,
+	update func(*sqlbase.MutableTableDescriptor) error,
 	logEvent func(*client.Txn) error,
-) (*sqlbase.Descriptor, error) {
+) (*sqlbase.ImmutableTableDescriptor, error) {
 	errLeaseVersionChanged := errors.New("lease version changed")
 	// Retry while getting errLeaseVersionChanged.
 	for r := retry.Start(base.DefaultRetryOptions()); r.Next(); {
@@ -375,21 +382,18 @@ func (s LeaseStore) Publish(
 			return nil, err
 		}
 
-		desc := &sqlbase.Descriptor{}
+		var tableDesc *sqlbase.MutableTableDescriptor
 		// There should be only one version of the descriptor, but it's
 		// a race now to update to the next version.
 		err = s.execCfg.DB.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
-			descKey := sqlbase.MakeDescMetadataKey(tableID)
-
 			// Re-read the current version of the table descriptor, this time
 			// transactionally.
-			if err := txn.GetProto(ctx, descKey, desc); err != nil {
+			var err error
+			tableDesc, err = sqlbase.GetMutableTableDescFromID(ctx, txn, tableID)
+			if err != nil {
 				return err
 			}
-			tableDesc := desc.GetTable()
-			if tableDesc == nil {
-				return errors.Errorf("ID %d is not a table", tableID)
-			}
+
 			if expectedVersion != tableDesc.Version {
 				// The version changed out from under us. Someone else must be
 				// performing a schema change operation.
@@ -409,7 +413,7 @@ func (s LeaseStore) Publish(
 					tableDesc.Version, version)
 			}
 
-			if err := incrementVersion(ctx, tableDesc, txn); err != nil {
+			if err := maybeIncrementVersion(ctx, tableDesc, txn); err != nil {
 				return err
 			}
 			if err := tableDesc.ValidateTable(s.execCfg.Settings); err != nil {
@@ -421,7 +425,7 @@ func (s LeaseStore) Publish(
 				return err
 			}
 			b := txn.NewBatch()
-			b.Put(descKey, desc)
+			b.Put(sqlbase.MakeDescMetadataKey(tableID), sqlbase.WrapDescriptor(tableDesc))
 			if logEvent != nil {
 				// If an event log is required for this update, ensure that the
 				// descriptor change occurs first in the transaction. This is
@@ -443,7 +447,7 @@ func (s LeaseStore) Publish(
 
 		switch err {
 		case nil, errDidntUpdateDescriptor:
-			return desc, nil
+			return sqlbase.NewImmutableTableDescriptor(tableDesc.TableDescriptor), nil
 		case errLeaseVersionChanged:
 			// will loop around to retry
 		default:
@@ -463,9 +467,10 @@ type IDVersion struct {
 	version sqlbase.DescriptorVersion
 }
 
-// NewIDVersion returns an initialized IDVersion.
-func NewIDVersion(name string, id sqlbase.ID, version sqlbase.DescriptorVersion) IDVersion {
-	return IDVersion{name: name, id: id, version: version}
+// NewIDVersionPrev returns an initialized IDVersion with the
+// previous version of the descriptor.
+func NewIDVersionPrev(desc *sqlbase.TableDescriptor) IDVersion {
+	return IDVersion{name: desc.Name, id: desc.ID, version: desc.Version - 1}
 }
 
 // CountLeases returns the number of unexpired leases for a number of tables
@@ -481,7 +486,7 @@ func CountLeases(
 		)
 	}
 
-	stmt := fmt.Sprintf(`SELECT count(1) FROM system.lease AS OF SYSTEM TIME %s WHERE `,
+	stmt := fmt.Sprintf(`SELECT count(1) FROM system.public.lease AS OF SYSTEM TIME %s WHERE `,
 		at.AsOfSystemTime()) +
 		strings.Join(whereClauses, " OR ")
 	values, err := executor.QueryRow(
@@ -523,8 +528,8 @@ func (s LeaseStore) getForExpiration(
 		}
 		// Create a tableVersionState with the table and without a lease.
 		table = &tableVersionState{
-			TableDescriptor: *tableDesc,
-			expiration:      expiration,
+			ImmutableTableDescriptor: *sqlbase.NewImmutableTableDescriptor(*tableDesc),
+			expiration:               expiration,
 		}
 		return nil
 	})
@@ -869,19 +874,33 @@ func (m *LeaseManager) AcquireFreshestFromStore(ctx context.Context, tableID sql
 // upsertLocked inserts a lease for a particular table version.
 // If an existing lease exists for the table version it replaces
 // it and returns it.
-func (t *tableState) upsertLocked(ctx context.Context, table *tableVersionState) *storedTableLease {
+func (t *tableState) upsertLocked(
+	ctx context.Context, table *tableVersionState,
+) (*storedTableLease, error) {
 	s := t.mu.active.find(table.Version)
 	if s == nil {
 		if t.mu.active.findNewest() != nil {
 			log.Infof(ctx, "new lease: %s", table)
 		}
 		t.mu.active.insert(table)
-		return nil
+		return nil, nil
+	}
+
+	// The table is replacing an existing one at the same version.
+	if !table.hasValidExpiration(s) {
+		// This is a violation of an invariant and can actually not
+		// happen. We return an error here to aid in further investigations.
+		return nil, errors.Errorf("lease expiration monotonicity violation, (%s) vs (%s)", s, table)
 	}
 
 	s.mu.Lock()
 	table.mu.Lock()
-	// subsume the refcount of the older lease.
+	// subsume the refcount of the older lease. This is permitted because
+	// the new lease has a greater expiration than the older lease and
+	// any transaction using the older lease can safely use a deadline set
+	// to the older lease's expiration even though the older lease is
+	// released! This is because the new lease is valid at the same table
+	// version at a greater expiration.
 	table.mu.refcount += s.mu.refcount
 	s.mu.refcount = 0
 	l := s.mu.lease
@@ -893,7 +912,7 @@ func (t *tableState) upsertLocked(ctx context.Context, table *tableVersionState)
 	s.mu.Unlock()
 	t.mu.active.remove(s)
 	t.mu.active.insert(table)
-	return l
+	return l, nil
 }
 
 // removeInactiveVersions removes inactive versions in t.mu.active.data with refcount 0.
@@ -920,26 +939,30 @@ func (t *tableState) removeInactiveVersions() []*storedTableLease {
 
 // If the lease cannot be obtained because the descriptor is in the process of
 // being dropped, the error will be errTableDropped.
-// minExpirationTime, if not set to the zero value, will be used as a lower
-// bound on the expiration of the new table. This can be used to eliminate the
-// jitter in the expiration time, and guarantee that we get a lease that will be
-// inserted at the end of the lease set (i.e. it will be returned by
-// findNewest() from now on). The boolean returned is true if this call was actually
-// responsible for the lease acquisition.
+// The boolean returned is true if this call was actually responsible for the
+// lease acquisition.
 func acquireNodeLease(ctx context.Context, m *LeaseManager, id sqlbase.ID) (bool, error) {
 	var toRelease *storedTableLease
 	resultChan, didAcquire := m.group.DoChan(fmt.Sprintf("acquire%d", id), func() (interface{}, error) {
 		if m.isDraining() {
 			return nil, errors.New("cannot acquire lease when draining")
 		}
-		table, err := m.LeaseStore.acquire(ctx, id)
+		newest := m.findNewest(id)
+		var minExpiration hlc.Timestamp
+		if newest != nil {
+			minExpiration = newest.expiration
+		}
+		table, err := m.LeaseStore.acquire(ctx, minExpiration, id)
 		if err != nil {
 			return nil, err
 		}
 		t := m.findTableState(id, false /* create */)
 		t.mu.Lock()
 		defer t.mu.Unlock()
-		toRelease = t.upsertLocked(ctx, table)
+		toRelease, err = t.upsertLocked(ctx, table)
+		if err != nil {
+			return nil, err
+		}
 		m.tableNames.insert(table)
 		return leaseToken(table), nil
 	})
@@ -956,7 +979,7 @@ func acquireNodeLease(ctx context.Context, m *LeaseManager, id sqlbase.ID) (bool
 // release returns a tableVersionState that needs to be released from
 // the store.
 func (t *tableState) release(
-	table *sqlbase.TableDescriptor, removeOnceDereferenced bool,
+	table *sqlbase.ImmutableTableDescriptor, removeOnceDereferenced bool,
 ) (*storedTableLease, error) {
 	t.mu.Lock()
 	defer t.mu.Unlock()
@@ -1076,7 +1099,7 @@ func purgeOldVersions(
 	if dropped := err == errTableDropped; dropped || err == nil {
 		removeInactives(dropped)
 		if table != nil {
-			s, err := t.release(&table.TableDescriptor, m.removeOnceDereferenced())
+			s, err := t.release(&table.ImmutableTableDescriptor, m.removeOnceDereferenced())
 			if err != nil {
 				return err
 			}
@@ -1169,9 +1192,9 @@ type LeaseManagerTestingKnobs struct {
 	// refreshed. Careful when using this to block for too long - you can block
 	// all the gossip users in the system. If it returns an error the gossip
 	// update is ignored.
-	GossipUpdateEvent func(config.SystemConfig) error
+	GossipUpdateEvent func(*config.SystemConfig) error
 	// A callback called after the leases are refreshed as a result of a gossip update.
-	TestingLeasesRefreshedEvent func(config.SystemConfig)
+	TestingLeasesRefreshedEvent func(*config.SystemConfig)
 	// To disable the deletion of orphaned leases at server startup.
 	DisableDeleteOrphanedLeases bool
 	LeaseStoreTestingKnobs      LeaseStoreTestingKnobs
@@ -1223,7 +1246,7 @@ func (c *tableNameCache) get(
 
 	defer table.mu.Unlock()
 
-	if !nameMatchesTable(&table.TableDescriptor, dbID, tableName) {
+	if !nameMatchesTable(&table.ImmutableTableDescriptor, dbID, tableName) {
 		panic(fmt.Sprintf("Out of sync entry in the name cache. "+
 			"Cache entry: %d.%q -> %d. Lease: %d.%q.",
 			dbID, tableName, table.ID, table.ParentID, table.Name))
@@ -1251,7 +1274,7 @@ func (c *tableNameCache) insert(table *tableVersionState) {
 	// If we already have a lease in the cache for this name, see if this one is
 	// better (higher version or later expiration).
 	if table.Version > existing.Version ||
-		(table.Version == existing.Version && (existing.expiration.Less(table.expiration))) {
+		(table.Version == existing.Version && table.hasValidExpiration(existing)) {
 		// Overwrite the old table. The new one is better. From now on, we want
 		// clients to use the new one.
 		c.tables[key] = table
@@ -1354,7 +1377,9 @@ func (m *LeaseManager) SetExecCfg(execCfg *ExecutorConfig) {
 	m.execCfg = execCfg
 }
 
-func nameMatchesTable(table *sqlbase.TableDescriptor, dbID sqlbase.ID, tableName string) bool {
+func nameMatchesTable(
+	table *sqlbase.ImmutableTableDescriptor, dbID sqlbase.ID, tableName string,
+) bool {
 	return table.ParentID == dbID && table.Name == tableName
 }
 
@@ -1373,9 +1398,20 @@ func (m *LeaseManager) findNewest(tableID sqlbase.ID) *tableVersionState {
 // the returned descriptor. Renewal of a lease may begin in the
 // background. Renewal is done in order to prevent blocking on future
 // acquisitions.
+//
+// Known limitation: AcquireByName() calls Acquire() and therefore suffers
+// from the same limitation as Acquire (See Acquire). AcquireByName() is
+// unable to function correctly on a timestamp less than the timestamp
+// of a transaction with a DROP/TRUNCATE on a table. The limitation in
+// the face of a DROP follows directly from the limitation on Acquire().
+// A TRUNCATE is implemented by changing the name -> id mapping for a table
+// and by dropping the descriptor with the old id. While AcquireByName
+// can use the timestamp and get the correct name->id  mapping at a
+// timestamp, it uses Acquire() to get a descriptor with the corresponding
+// id and fails because the id has been dropped by the TRUNCATE.
 func (m *LeaseManager) AcquireByName(
 	ctx context.Context, timestamp hlc.Timestamp, dbID sqlbase.ID, tableName string,
-) (*sqlbase.TableDescriptor, hlc.Timestamp, error) {
+) (*sqlbase.ImmutableTableDescriptor, hlc.Timestamp, error) {
 	// Check if we have cached an ID for this name.
 	tableVersion := m.tableNames.get(dbID, tableName, timestamp)
 	if tableVersion != nil {
@@ -1390,9 +1426,9 @@ func (m *LeaseManager) AcquireByName(
 					}
 				}
 			}
-			return &tableVersion.TableDescriptor, tableVersion.expiration, nil
+			return &tableVersion.ImmutableTableDescriptor, tableVersion.expiration, nil
 		}
-		if err := m.Release(&tableVersion.TableDescriptor); err != nil {
+		if err := m.Release(&tableVersion.ImmutableTableDescriptor); err != nil {
 			return nil, hlc.Timestamp{}, err
 		}
 		// Return a valid table descriptor for the timestamp.
@@ -1506,9 +1542,15 @@ func (m *LeaseManager) resolveName(
 // A transaction using this descriptor must ensure that its
 // commit-timestamp < expiration-time. Care must be taken to not modify
 // the returned descriptor.
+//
+// Known limitation: Acquire() can return an error after the table with
+// the tableID has been dropped. This is true even when using a timestamp
+// less than the timestamp of the DROP command. This is because Acquire
+// can only return an older version of a descriptor if the latest version
+// can be leased; as it stands a dropped table cannot be leased.
 func (m *LeaseManager) Acquire(
 	ctx context.Context, timestamp hlc.Timestamp, tableID sqlbase.ID,
-) (*sqlbase.TableDescriptor, hlc.Timestamp, error) {
+) (*sqlbase.ImmutableTableDescriptor, hlc.Timestamp, error) {
 	for {
 		t := m.findTableState(tableID, true /*create*/)
 		table, latest, err := t.findForTimestamp(ctx, timestamp)
@@ -1522,7 +1564,7 @@ func (m *LeaseManager) Acquire(
 					}
 				}
 			}
-			return &table.TableDescriptor, table.expiration, nil
+			return &table.ImmutableTableDescriptor, table.expiration, nil
 		}
 		switch err {
 		case errRenewLease:
@@ -1550,7 +1592,7 @@ func (m *LeaseManager) Acquire(
 }
 
 // Release releases a previously acquired table.
-func (m *LeaseManager) Release(desc *sqlbase.TableDescriptor) error {
+func (m *LeaseManager) Release(desc *sqlbase.ImmutableTableDescriptor) error {
 	t := m.findTableState(desc.ID, false /* create */)
 	if t == nil {
 		return errors.Errorf("table %d not found", desc.ID)
@@ -1626,7 +1668,7 @@ func (m *LeaseManager) RefreshLeases(s *stop.Stopper, db *client.DB, g *gossip.G
 		for {
 			select {
 			case <-gossipUpdateC:
-				cfg, _ := g.GetSystemConfig()
+				cfg := g.GetSystemConfig()
 				if m.testingKnobs.GossipUpdateEvent != nil {
 					if err := m.testingKnobs.GossipUpdateEvent(cfg); err != nil {
 						break
@@ -1775,7 +1817,7 @@ func (m *LeaseManager) DeleteOrphanedLeases(timeThreshold int64) {
 
 		// Read orphaned leases.
 		sqlQuery := fmt.Sprintf(`
-SELECT "descID", version, expiration FROM system.lease AS OF SYSTEM TIME %d WHERE "nodeID" = %d
+SELECT "descID", version, expiration FROM system.public.lease AS OF SYSTEM TIME %d WHERE "nodeID" = %d
 `, timeThreshold, nodeID)
 		var rows []tree.Datums
 		retryOptions := base.DefaultRetryOptions()
@@ -1783,7 +1825,7 @@ SELECT "descID", version, expiration FROM system.lease AS OF SYSTEM TIME %d WHER
 		// The retry is required because of errors caused by node restarts. Retry 30 times.
 		if err := retry.WithMaxAttempts(ctx, retryOptions, 30, func() error {
 			var err error
-			rows, _, err = m.LeaseStore.execCfg.InternalExecutor.Query(
+			rows, err = m.LeaseStore.execCfg.InternalExecutor.Query(
 				ctx, "read orphaned table leases", nil /*txn*/, sqlQuery)
 			return err
 		}); err != nil {
