@@ -15,13 +15,14 @@
 package norm
 
 import (
-	"fmt"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/opt"
+	"github.com/cockroachdb/cockroach/pkg/sql/opt/cat"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/memo"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/props"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util"
+	"github.com/cockroachdb/cockroach/pkg/util/log"
 )
 
 // ----------------------------------------------------------------------
@@ -31,42 +32,47 @@ import (
 //
 // ----------------------------------------------------------------------
 
+// EmptyJoinPrivate returns an unset JoinPrivate.
+func (c *CustomFuncs) EmptyJoinPrivate() *memo.JoinPrivate {
+	return memo.EmptyJoinPrivate
+}
+
 // ConstructNonLeftJoin maps a left join to an inner join and a full join to a
 // right join when it can be proved that the right side of the join always
 // produces at least one row for every row on the left.
 func (c *CustomFuncs) ConstructNonLeftJoin(
-	joinOp opt.Operator, left, right, on memo.GroupID,
-) memo.GroupID {
+	joinOp opt.Operator, left, right memo.RelExpr, on memo.FiltersExpr, private *memo.JoinPrivate,
+) memo.RelExpr {
 	switch joinOp {
 	case opt.LeftJoinOp:
-		return c.f.ConstructInnerJoin(left, right, on)
+		return c.f.ConstructInnerJoin(left, right, on, private)
 	case opt.LeftJoinApplyOp:
-		return c.f.ConstructInnerJoinApply(left, right, on)
+		return c.f.ConstructInnerJoinApply(left, right, on, private)
 	case opt.FullJoinOp:
-		return c.f.ConstructRightJoin(left, right, on)
+		return c.f.ConstructRightJoin(left, right, on, private)
 	case opt.FullJoinApplyOp:
-		return c.f.ConstructRightJoinApply(left, right, on)
+		return c.f.ConstructRightJoinApply(left, right, on, private)
 	}
-	panic(fmt.Sprintf("unexpected join operator: %v", joinOp))
+	panic(pgerror.NewAssertionErrorf("unexpected join operator: %v", log.Safe(joinOp)))
 }
 
 // ConstructNonRightJoin maps a right join to an inner join and a full join to a
 // left join when it can be proved that the left side of the join always
 // produces at least one row for every row on the right.
 func (c *CustomFuncs) ConstructNonRightJoin(
-	joinOp opt.Operator, left, right, on memo.GroupID,
-) memo.GroupID {
+	joinOp opt.Operator, left, right memo.RelExpr, on memo.FiltersExpr, private *memo.JoinPrivate,
+) memo.RelExpr {
 	switch joinOp {
 	case opt.RightJoinOp:
-		return c.f.ConstructInnerJoin(left, right, on)
+		return c.f.ConstructInnerJoin(left, right, on, private)
 	case opt.RightJoinApplyOp:
-		return c.f.ConstructInnerJoinApply(left, right, on)
+		return c.f.ConstructInnerJoinApply(left, right, on, private)
 	case opt.FullJoinOp:
-		return c.f.ConstructLeftJoin(left, right, on)
+		return c.f.ConstructLeftJoin(left, right, on, private)
 	case opt.FullJoinApplyOp:
-		return c.f.ConstructLeftJoinApply(left, right, on)
+		return c.f.ConstructLeftJoinApply(left, right, on, private)
 	}
-	panic(fmt.Sprintf("unexpected join operator: %v", joinOp))
+	panic(pgerror.NewAssertionErrorf("unexpected join operator: %v", log.Safe(joinOp)))
 }
 
 // SimplifyNotNullEquality simplifies an expression of the following form:
@@ -77,8 +83,8 @@ func (c *CustomFuncs) ConstructNonRightJoin(
 // The testOp argument must be IsOp or IsNotOp, and the constOp argument must be
 // TrueOp, FalseOp, or NullOp.
 func (c *CustomFuncs) SimplifyNotNullEquality(
-	eq memo.GroupID, testOp, constOp opt.Operator,
-) memo.GroupID {
+	eq opt.ScalarExpr, testOp, constOp opt.Operator,
+) opt.ScalarExpr {
 	switch testOp {
 	case opt.IsOp:
 		switch constOp {
@@ -100,7 +106,7 @@ func (c *CustomFuncs) SimplifyNotNullEquality(
 			return c.f.ConstructTrue()
 		}
 	}
-	panic(fmt.Sprintf("invalid ops: %v, %v", testOp, constOp))
+	panic(pgerror.NewAssertionErrorf("invalid ops: %v, %v", testOp, constOp))
 }
 
 // CanMap returns true if it is possible to map a boolean expression src, which
@@ -124,22 +130,23 @@ func (c *CustomFuncs) SimplifyNotNullEquality(
 // dst is (Scan a).
 //
 // If src has a correlated subquery, CanMap returns false.
-func (c *CustomFuncs) CanMap(filters, src, dst memo.GroupID) bool {
+func (c *CustomFuncs) CanMap(
+	filters memo.FiltersExpr, src *memo.FiltersItem, dst memo.RelExpr,
+) bool {
 	// Fast path if src is already bound by dst.
 	if c.IsBoundBy(src, c.OutputCols(dst)) {
 		return true
 	}
 
-	if c.HasHoistableSubquery(src) {
+	scalarProps := src.ScalarProps(c.mem)
+	if scalarProps.HasCorrelatedSubquery {
 		return false
 	}
 
-	fd := c.LookupLogical(filters).Scalar.FuncDeps
-
 	// For CanMap to be true, each column in src must map to at least one column
 	// in dst.
-	for i, ok := c.OuterCols(src).Next(0); ok; i, ok = c.OuterCols(src).Next(i + 1) {
-		eqCols := c.GetEquivColsWithEquivType(opt.ColumnID(i), &fd)
+	for i, ok := scalarProps.OuterCols.Next(0); ok; i, ok = scalarProps.OuterCols.Next(i + 1) {
+		eqCols := c.GetEquivColsWithEquivType(opt.ColumnID(i), filters)
 		if !eqCols.Intersects(c.OutputCols(dst)) {
 			return false
 		}
@@ -167,61 +174,57 @@ func (c *CustomFuncs) CanMap(filters, src, dst memo.GroupID) bool {
 // If Map is called with src as a.x + b.y = 5 and dst as (Scan b), it returns
 // b.x + b.y = 5. Map should not be called with the equality predicate
 // a.x = b.x, because it would just return the tautology b.x = b.x.
-func (c *CustomFuncs) Map(filters, src, dst memo.GroupID) memo.GroupID {
+func (c *CustomFuncs) Map(
+	filters memo.FiltersExpr, src *memo.FiltersItem, dst memo.RelExpr,
+) opt.ScalarExpr {
 	// Fast path if src is already bound by dst.
 	if c.IsBoundBy(src, c.OutputCols(dst)) {
-		return src
+		return src.Condition
 	}
 
-	fd := c.LookupLogical(filters).Scalar.FuncDeps
-
 	// Map each column in src to one column in dst. We choose an arbitrary column
 	// (the one with the smallest ColumnID) if there are multiple choices.
 	var colMap util.FastIntMap
-	c.OuterCols(src).ForEach(func(srcCol int) {
-		eqCols := c.GetEquivColsWithEquivType(opt.ColumnID(srcCol), &fd)
+	outerCols := src.ScalarProps(c.mem).OuterCols
+	for srcCol, ok := outerCols.Next(0); ok; srcCol, ok = outerCols.Next(srcCol + 1) {
+		eqCols := c.GetEquivColsWithEquivType(opt.ColumnID(srcCol), filters)
 		eqCols.IntersectionWith(c.OutputCols(dst))
 		if eqCols.Contains(srcCol) {
 			colMap.Set(srcCol, srcCol)
 		} else {
 			dstCol, ok := eqCols.Next(0)
 			if !ok {
-				panic(fmt.Errorf(
+				panic(pgerror.NewAssertionErrorf(
 					"Map called on src that cannot be mapped to dst. src:\n%s\ndst:\n%s",
-					memo.MakeNormExprView(c.mem, src).FormatString(memo.ExprFmtHideScalars),
-					memo.MakeNormExprView(c.mem, dst).FormatString(memo.ExprFmtHideScalars),
+					src, dst,
 				))
 			}
 			colMap.Set(srcCol, dstCol)
 		}
-	})
+	}
 
 	// Recursively walk the scalar sub-tree looking for references to columns
 	// that need to be replaced.
-	var replace memo.ReplaceChildFunc
-	replace = func(child memo.GroupID) memo.GroupID {
-		expr := c.mem.NormExpr(child)
-
-		switch expr.Operator() {
-		case opt.VariableOp:
-			varColID := c.ExtractColID(expr.AsVariable().Col())
-			outCol, _ := colMap.Get(int(varColID))
-			if int(varColID) == outCol {
+	var replace ReplaceFunc
+	replace = func(nd opt.Expr) opt.Expr {
+		switch t := nd.(type) {
+		case *memo.VariableExpr:
+			outCol, _ := colMap.Get(int(t.Col))
+			if int(t.Col) == outCol {
 				// Avoid constructing a new variable if possible.
-				return child
+				return nd
 			}
-			return c.f.ConstructVariable(c.f.InternColumnID(opt.ColumnID(outCol)))
+			return c.f.ConstructVariable(opt.ColumnID(outCol))
 
-		case opt.SubqueryOp, opt.ExistsOp, opt.AnyOp:
+		case *memo.SubqueryExpr, *memo.ExistsExpr, *memo.AnyExpr:
 			// There are no correlated subqueries, so we don't need to recurse here.
-			return child
+			return nd
 		}
 
-		ev := memo.MakeNormExprView(c.mem, child)
-		return ev.Replace(c.f.evalCtx, replace).Group()
+		return c.f.Replace(nd, replace)
 	}
 
-	return replace(src)
+	return replace(src.Condition).(opt.ScalarExpr)
 }
 
 // GetEquivColsWithEquivType uses the given FuncDepSet to find columns that are
@@ -251,9 +254,11 @@ func (c *CustomFuncs) Map(filters, src, dst memo.GroupID) memo.GroupID {
 //
 // TODO(rytaft): In the future, we may want to allow the mapping if the
 // filter involves a comparison operator, such as x < 5.
-func (c *CustomFuncs) GetEquivColsWithEquivType(col opt.ColumnID, fd *props.FuncDepSet) opt.ColSet {
+func (c *CustomFuncs) GetEquivColsWithEquivType(
+	col opt.ColumnID, filters memo.FiltersExpr,
+) opt.ColSet {
 	var res opt.ColSet
-	colType := c.f.Metadata().ColumnType(col)
+	colType := c.f.Metadata().ColumnMeta(col).Type
 
 	// Don't bother looking for equivalent columns if colType has a composite
 	// key encoding.
@@ -262,10 +267,15 @@ func (c *CustomFuncs) GetEquivColsWithEquivType(col opt.ColumnID, fd *props.Func
 		return res
 	}
 
-	eqCols := fd.ComputeEquivClosure(util.MakeFastIntSet(int(col)))
+	// Compute all equivalent columns.
+	eqCols := util.MakeFastIntSet(int(col))
+	for i := range filters {
+		eqCols = filters[i].ScalarProps(c.mem).FuncDeps.ComputeEquivClosure(eqCols)
+	}
+
 	eqCols.ForEach(func(i int) {
-		eqColType := c.f.Metadata().ColumnType(opt.ColumnID(i))
 		// Only include columns that have the same type as col.
+		eqColType := c.f.Metadata().ColumnMeta(opt.ColumnID(i)).Type
 		if colType.Equivalent(eqColType) {
 			res.Add(i)
 		}
@@ -274,6 +284,36 @@ func (c *CustomFuncs) GetEquivColsWithEquivType(col opt.ColumnID, fd *props.Func
 	return res
 }
 
+// eqConditionsToColMap returns a map of left columns to right columns
+// that are being equated in the specified conditions. leftCols is used
+// to identify which column is a left column.
+func (c *CustomFuncs) eqConditionsToColMap(
+	filters memo.FiltersExpr, leftCols opt.ColSet,
+) map[opt.ColumnID]opt.ColumnID {
+	eqColMap := make(map[opt.ColumnID]opt.ColumnID)
+
+	for i := range filters {
+		eq, _ := filters[i].Condition.(*memo.EqExpr)
+		if eq == nil {
+			continue
+		}
+
+		leftVarExpr, _ := eq.Left.(*memo.VariableExpr)
+		rightVarExpr, _ := eq.Right.(*memo.VariableExpr)
+		if leftVarExpr == nil || rightVarExpr == nil {
+			continue
+		}
+
+		if leftCols.Contains(int(leftVarExpr.Col)) {
+			eqColMap[leftVarExpr.Col] = rightVarExpr.Col
+		} else {
+			eqColMap[rightVarExpr.Col] = leftVarExpr.Col
+		}
+	}
+
+	return eqColMap
+}
+
 // JoinFiltersMatchAllLeftRows returns true when each row in the given join's
 // left input matches at least one row from the right input, according to the
 // join filters. This is true when the following conditions are satisfied:
@@ -294,7 +334,9 @@ func (c *CustomFuncs) GetEquivColsWithEquivType(col opt.ColumnID, fd *props.Func
 //      a foreign key on the left equality table, and the right equality columns
 //      to the corresponding referenced columns in the right equality table.
 //
-func (c *CustomFuncs) JoinFiltersMatchAllLeftRows(left, right, filters memo.GroupID) bool {
+func (c *CustomFuncs) JoinFiltersMatchAllLeftRows(
+	left, right memo.RelExpr, filters memo.FiltersExpr,
+) bool {
 	unfilteredCols := c.deriveUnfilteredCols(right)
 	if unfilteredCols.Empty() {
 		// Condition #3: right input has no columns which contain values from
@@ -302,93 +344,181 @@ func (c *CustomFuncs) JoinFiltersMatchAllLeftRows(left, right, filters memo.Grou
 		return false
 	}
 
-	filtersExpr := c.mem.NormExpr(filters).AsFilters()
-	if filtersExpr == nil {
-		return false
-	}
-
-	leftCols := c.LookupLogical(left).Relational.NotNullCols
-	rightCols := c.LookupLogical(right).Relational.NotNullCols
+	leftColIDs := left.Relational().NotNullCols
+	rightColIDs := right.Relational().NotNullCols
 
 	md := c.f.Metadata()
 
-	var leftTab, rightTab opt.Table
-	for _, condition := range c.mem.LookupList(filtersExpr.Conditions()) {
-		eqExpr := c.mem.NormExpr(condition).AsEq()
-		if eqExpr == nil {
+	var leftTab, rightTab opt.TableID
+
+	// Any left columns that don't match conditions 1-4 end up in this set.
+	var remainingLeftColIDs opt.ColSet
+
+	for i := range filters {
+		eq, _ := filters[i].Condition.(*memo.EqExpr)
+		if eq == nil {
 			// Condition #1: conjunct is not an equality comparison.
 			return false
 		}
 
-		leftVarExpr := c.mem.NormExpr(eqExpr.Left()).AsVariable()
-		rightVarExpr := c.mem.NormExpr(eqExpr.Right()).AsVariable()
-		if leftVarExpr == nil || rightVarExpr == nil {
+		leftVar, _ := eq.Left.(*memo.VariableExpr)
+		rightVar, _ := eq.Right.(*memo.VariableExpr)
+		if leftVar == nil || rightVar == nil {
 			// Condition #1: conjunct does not compare two columns.
 			return false
 		}
 
-		leftCol := c.ExtractColID(leftVarExpr.Col())
-		rightCol := c.ExtractColID(rightVarExpr.Col())
+		leftColID := leftVar.Col
+		rightColID := rightVar.Col
 
-		// Normalize leftCol to come from leftCols.
-		if !leftCols.Contains(int(leftCol)) {
-			leftCol, rightCol = rightCol, leftCol
+		// Normalize leftColID to come from leftColIDs.
+		if !leftColIDs.Contains(int(leftColID)) {
+			leftColID, rightColID = rightColID, leftColID
 		}
-		if !leftCols.Contains(int(leftCol)) || !rightCols.Contains(int(rightCol)) {
+		if !leftColIDs.Contains(int(leftColID)) || !rightColIDs.Contains(int(rightColID)) {
 			// Condition #1: columns don't come from both sides of join, or
 			// columns are nullable.
 			return false
 		}
 
-		if !unfilteredCols.Contains(int(rightCol)) {
+		if !unfilteredCols.Contains(int(rightColID)) {
 			// Condition #3: right column doesn't contain values from every row.
 			return false
 		}
 
-		if leftTab == nil {
-			leftTabID := md.ColumnTableID(leftCol)
-			rightTabID := md.ColumnTableID(rightCol)
-			if leftTabID == 0 || rightTabID == 0 {
+		if leftTab == 0 {
+			leftTab = md.ColumnMeta(leftColID).Table
+			rightTab = md.ColumnMeta(rightColID).Table
+			if leftTab == 0 || rightTab == 0 {
 				// Condition #2: Columns don't come from base tables.
 				return false
 			}
-			leftTab = md.Table(leftTabID)
-			rightTab = md.Table(rightTabID)
-		} else if md.Table(md.ColumnTableID(leftCol)) != leftTab {
+		} else if md.ColumnMeta(leftColID).Table != leftTab {
 			// Condition #2: All left columns don't come from same table.
 			return false
-		} else if md.Table(md.ColumnTableID(rightCol)) != rightTab {
+		} else if md.ColumnMeta(rightColID).Table != rightTab {
 			// Condition #2: All right columns don't come from same table.
 			return false
 		}
 
-		if leftTab == rightTab {
+		if md.TableMeta(leftTab).Table == md.TableMeta(rightTab).Table {
 			// Check self-join case.
-			if md.ColumnOrdinal(leftCol) != md.ColumnOrdinal(rightCol) {
+			leftColOrd := leftTab.ColumnOrdinal(leftColID)
+			rightColOrd := rightTab.ColumnOrdinal(rightColID)
+			if leftColOrd != rightColOrd {
 				// Condition #4: Left and right column ordinals do not match.
 				return false
 			}
 		} else {
-			// TODO(andyk): Check foreign key case.
-			return false
+			// Column could be a potential foreign key match so save it.
+			remainingLeftColIDs.Add(int(leftColID))
 		}
 	}
 
-	return true
+	if remainingLeftColIDs.Empty() {
+		return true
+	}
+
+	var leftRightColMap map[opt.ColumnID]opt.ColumnID
+	// Condition #5: All remaining left columns correspond to a validated foreign
+	// key relation.
+	leftTabMeta := md.TableMeta(leftTab)
+	rightTabMeta := md.TableMeta(rightTab)
+	for i, cnt := 0, leftTabMeta.Table.IndexCount(); i < cnt; i++ {
+		index := leftTabMeta.Table.Index(i)
+		fkRef, ok := index.ForeignKey()
+
+		if !ok || !fkRef.Validated {
+			// No validated foreign key reference on this index.
+			continue
+		}
+
+		fkTable := md.TableByStableID(fkRef.TableID)
+		fkPrefix := int(fkRef.PrefixLen)
+		if fkPrefix <= 0 {
+			panic(pgerror.NewAssertionErrorf("fkPrefix should always be positive"))
+		}
+		if fkTable == nil || fkTable.ID() != rightTabMeta.Table.ID() {
+			continue
+		}
+
+		// Find the index corresponding to fkRef.IndexID - the index
+		// on the right table that forms the destination end of
+		// the fk relation.
+		var fkIndex cat.Index
+		found := false
+		for j, cnt2 := 0, fkTable.IndexCount(); j < cnt2; j++ {
+			if fkTable.Index(j).ID() == fkRef.IndexID {
+				found = true
+				fkIndex = fkTable.Index(j)
+				break
+			}
+		}
+		if !found {
+			panic(pgerror.NewAssertionErrorf("Foreign key referenced index not found in table"))
+		}
+
+		var leftIndexCols opt.ColSet
+		for j := 0; j < fkPrefix; j++ {
+			ord := index.Column(j).Ordinal
+			leftIndexCols.Add(int(leftTab.ColumnID(ord)))
+		}
+
+		if !remainingLeftColIDs.SubsetOf(leftIndexCols) {
+			continue
+		}
+
+		// Build a mapping of left to right columns as specified
+		// in the filter conditions - this is used to detect
+		// whether the filter conditions follow the foreign key
+		// constraint exactly.
+		if leftRightColMap == nil {
+			leftRightColMap = c.eqConditionsToColMap(filters, leftColIDs)
+		}
+
+		// Loop through all columns in fk index that also exist in LHS of match condition,
+		// and ensure that they correspond to the correct RHS column according to the
+		// foreign key relation. In other words, each LHS column's index ordinal
+		// in the foreign key index matches that of the RHS column (in the index being
+		// referenced) that it's being equated to.
+		fkMatch := true
+		for j := 0; j < fkPrefix; j++ {
+			indexLeftCol := leftTab.ColumnID(index.Column(j).Ordinal)
+
+			// Not every fk column needs to be in the equality conditions.
+			if !remainingLeftColIDs.Contains(int(indexLeftCol)) {
+				continue
+			}
+
+			indexRightCol := rightTab.ColumnID(fkIndex.Column(j).Ordinal)
+
+			if rightCol, ok := leftRightColMap[indexLeftCol]; !ok || rightCol != indexRightCol {
+				fkMatch = false
+				break
+			}
+		}
+
+		// Condition #5 satisfied.
+		if fkMatch {
+			return true
+		}
+	}
+
+	return false
 }
 
-// deriveUnfilteredCols returns the subset of the given group's output columns
-// that have values for every row in their owner table. In other words, columns
-// from tables that have had none of their rows filtered (but it's OK if rows
-// have been duplicated).
+// deriveUnfilteredCols returns the subset of the given input expression's
+// output columns that have values for every row in their owner table. In other
+// words, columns from tables that have had none of their rows filtered (but
+// it's OK if rows have been duplicated).
 //
 // deriveUnfilteredCols recursively derives the property, and populates the
 // props.Relational.Rule.UnfilteredCols field as it goes to make future calls
 // faster.
-func (c *CustomFuncs) deriveUnfilteredCols(group memo.GroupID) opt.ColSet {
+func (c *CustomFuncs) deriveUnfilteredCols(in memo.RelExpr) opt.ColSet {
 	// If the UnfilteredCols property has already been derived, return it
 	// immediately.
-	relational := c.LookupLogical(group).Relational
+	relational := in.Relational()
 	if relational.IsAvailable(props.UnfilteredCols) {
 		return relational.Rule.UnfilteredCols
 	}
@@ -396,27 +526,25 @@ func (c *CustomFuncs) deriveUnfilteredCols(group memo.GroupID) opt.ColSet {
 
 	// Derive the UnfilteredCols property now.
 	// TODO(andyk): Could add other cases, such as outer joins and union.
-	expr := c.mem.NormExpr(group)
-	switch expr.Operator() {
-	case opt.ScanOp:
+	switch t := in.(type) {
+	case *memo.ScanExpr:
 		// All un-limited, unconstrained output columns are unfiltered columns.
-		def := expr.Private(c.mem).(*memo.ScanOpDef)
-		if def.HardLimit == 0 && def.Constraint == nil {
+		if t.HardLimit == 0 && t.Constraint == nil {
 			relational.Rule.UnfilteredCols = relational.OutputCols
 		}
 
-	case opt.ProjectOp:
+	case *memo.ProjectExpr:
 		// Project never filters rows, so it passes through unfiltered columns.
-		unfilteredCols := c.deriveUnfilteredCols(expr.ChildGroup(c.mem, 0))
+		unfilteredCols := c.deriveUnfilteredCols(t.Input)
 		relational.Rule.UnfilteredCols = unfilteredCols.Intersection(relational.OutputCols)
 
-	case opt.InnerJoinOp, opt.InnerJoinApplyOp:
-		left := expr.ChildGroup(c.mem, 0)
-		right := expr.ChildGroup(c.mem, 1)
-		on := expr.ChildGroup(c.mem, 2)
+	case *memo.InnerJoinExpr, *memo.InnerJoinApplyExpr:
+		left := t.Child(0).(memo.RelExpr)
+		right := t.Child(1).(memo.RelExpr)
+		on := *t.Child(2).(*memo.FiltersExpr)
 
 		// Cross join always preserves left/right rows.
-		isCrossJoin := c.mem.NormOp(on) == opt.TrueOp
+		isCrossJoin := on.IsTrue()
 
 		// Inner joins may preserve left/right rows, according to
 		// JoinFiltersMatchAllLeftRows conditions.
@@ -441,74 +569,83 @@ func (c *CustomFuncs) deriveUnfilteredCols(group memo.GroupID) opt.ColSet {
 // Such an equality can be converted to a column equality by pushing down
 // expressions as projections.
 func (c *CustomFuncs) CanExtractJoinEquality(
-	a, b memo.GroupID, leftCols, rightCols opt.ColSet,
+	a, b opt.ScalarExpr, leftCols, rightCols opt.ColSet,
 ) bool {
-	if (c.IsBoundBy(a, leftCols) && c.IsBoundBy(b, rightCols)) ||
-		(c.IsBoundBy(a, rightCols) && c.IsBoundBy(b, leftCols)) {
+	// Disallow simple equality between variables.
+	if a.Op() == opt.VariableOp && b.Op() == opt.VariableOp {
+		return false
+	}
+
+	// Recursively compute properties for left and right sides.
+	var leftProps, rightProps props.Shared
+	memo.BuildSharedProps(c.mem, a, &leftProps)
+	memo.BuildSharedProps(c.mem, b, &rightProps)
+
+	// Disallow cases when one side has a correlated subquery.
+	// TODO(radu): investigate relaxing this.
+	if leftProps.HasCorrelatedSubquery || rightProps.HasCorrelatedSubquery {
+		return false
+	}
+
+	if (leftProps.OuterCols.SubsetOf(leftCols) && rightProps.OuterCols.SubsetOf(rightCols)) ||
+		(leftProps.OuterCols.SubsetOf(rightCols) && rightProps.OuterCols.SubsetOf(leftCols)) {
 		// The equality is of the form:
 		//   expression(leftCols) = expression(rightCols)
-
-		// Disallow cases when one side has a correlated subquery.
-		// TODO(radu): investigate relaxing this.
-		if c.HasCorrelatedSubquery(a) || c.HasCorrelatedSubquery(b) {
-			return false
-		}
-		aExpr := c.mem.NormExpr(a)
-		bExpr := c.mem.NormExpr(b)
-
-		// Disallow cases where one side is constant.
-		if aExpr.IsConstValue() || bExpr.IsConstValue() {
-			return false
-		}
-		// Disallow simple equality between variables.
-		if aExpr.Operator() == opt.VariableOp && bExpr.Operator() == opt.VariableOp {
-			return false
-		}
 		return true
 	}
 	return false
 }
 
-// ExtractJoinEqualities converts all join conditions that can satisfy
-// CanExtractJoinEquality to equality on input columns, by pushing down more
-// complicated expressions as projections. See the ExtractJoinEqualities rule.
-func (c *CustomFuncs) ExtractJoinEqualities(
-	joinOp opt.Operator, left, right memo.GroupID, filters memo.ListID,
-) memo.GroupID {
+// ExtractJoinEquality takes an equality FiltersItem that was identified via a
+// call to CanExtractJoinEquality, and converts it to an equality on "bare"
+// variables, by pushing down more complicated expressions as projections. See
+// the ExtractJoinEqualities rule.
+func (c *CustomFuncs) ExtractJoinEquality(
+	joinOp opt.Operator,
+	left, right memo.RelExpr,
+	filters memo.FiltersExpr,
+	item *memo.FiltersItem,
+	private *memo.JoinPrivate,
+) memo.RelExpr {
 	leftCols := c.OutputCols(left)
 	rightCols := c.OutputCols(right)
 
-	var leftProj, rightProj projectBuilder
-	leftProj.init(c)
-	rightProj.init(c)
+	eq := item.Condition.(*memo.EqExpr)
+	a, b := eq.Left, eq.Right
 
-	newFilters := MakeListBuilder(c)
-	conditions := c.f.mem.LookupList(filters)
-	for i := range conditions {
-		expr := c.f.mem.NormExpr(conditions[i]).AsEq()
-		if expr == nil || !c.CanExtractJoinEquality(expr.Left(), expr.Right(), leftCols, rightCols) {
-			newFilters.AddItem(conditions[i])
+	var eqLeftProps props.Shared
+	memo.BuildSharedProps(c.mem, eq.Left, &eqLeftProps)
+	if eqLeftProps.OuterCols.SubsetOf(rightCols) {
+		a, b = b, a
+	}
+
+	var leftProj, rightProj projectBuilder
+	leftProj.init(c.f)
+	rightProj.init(c.f)
+
+	newFilters := make(memo.FiltersExpr, len(filters))
+	for i := range filters {
+		if &filters[i] != item {
+			newFilters[i] = filters[i]
 			continue
 		}
-		a, b := expr.Left(), expr.Right()
-		if !(c.IsBoundBy(a, leftCols) && c.IsBoundBy(b, rightCols)) {
-			a, b = b, a
+
+		newFilters[i] = memo.FiltersItem{
+			Condition: c.f.ConstructEq(leftProj.add(a), rightProj.add(b)),
 		}
-		newFilters.AddItem(c.f.ConstructEq(
-			leftProj.add(a),
-			rightProj.add(b),
-		))
 	}
 	if leftProj.empty() && rightProj.empty() {
-		panic("no equalities to extract")
+		panic(pgerror.NewAssertionErrorf("no equalities to extract"))
 	}
 
-	join := c.f.DynamicConstruct(joinOp, memo.DynamicOperands{
-		memo.DynamicID(leftProj.buildProject(left)),
-		memo.DynamicID(rightProj.buildProject(right)),
-		memo.DynamicID(c.f.ConstructFilters(newFilters.BuildList())),
-	})
+	join := c.f.ConstructJoin(
+		joinOp,
+		leftProj.buildProject(left, leftCols),
+		rightProj.buildProject(right, rightCols),
+		newFilters,
+		private,
+	)
 
 	// Project away the synthesized columns.
-	return c.f.ConstructSimpleProject(join, leftCols.Union(rightCols))
+	return c.f.ConstructProject(join, memo.EmptyProjectionsExpr, leftCols.Union(rightCols))
 }
