@@ -40,6 +40,8 @@ package optbuilder
 //   post-projection: 1 + col3
 
 import (
+	"fmt"
+
 	"github.com/cockroachdb/cockroach/pkg/sql/opt"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/memo"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
@@ -57,10 +59,6 @@ type groupby struct {
 	// about aggregate functions encountered.
 	aggs []aggregateInfo
 
-	// groupings contains all group by expressions that were extracted from the
-	// query and which will become columns in this scope.
-	groupings []memo.GroupID
-
 	// groupStrs contains a string representation of each GROUP BY expression
 	// using symbolic notation. These strings are used to determine if SELECT
 	// and HAVING expressions contain sub-expressions matching a GROUP BY
@@ -87,9 +85,10 @@ type groupby struct {
 type aggregateInfo struct {
 	*tree.FuncExpr
 
-	def      memo.FuncOpDef
+	def      memo.FunctionPrivate
 	distinct bool
-	args     []memo.GroupID
+	args     memo.ScalarListExpr
+	filter   opt.ScalarExpr
 
 	// col is the output column of the aggregation.
 	col *scopeColumn
@@ -115,7 +114,7 @@ func (a *aggregateInfo) TypeCheck(ctx *tree.SemaContext, desired types.T) (tree.
 
 // Eval is part of the tree.TypedExpr interface.
 func (a *aggregateInfo) Eval(_ *tree.EvalContext) (tree.Datum, error) {
-	panic("aggregateInfo must be replaced before evaluation")
+	panic(pgerror.NewAssertionErrorf("aggregateInfo must be replaced before evaluation"))
 }
 
 var _ tree.Expr = &aggregateInfo{}
@@ -134,48 +133,39 @@ func (b *Builder) needsAggregation(sel *tree.SelectClause, scope *scope) bool {
 }
 
 func (b *Builder) constructGroupBy(
-	input memo.GroupID, groupingColSet opt.ColSet, aggCols []scopeColumn, ordering opt.Ordering,
-) memo.GroupID {
-	colList := make(opt.ColList, 0, len(aggCols))
-	groupList := make([]memo.GroupID, 0, len(aggCols))
+	input memo.RelExpr, groupingColSet opt.ColSet, aggCols []scopeColumn, ordering opt.Ordering,
+) memo.RelExpr {
+	aggs := make(memo.AggregationsExpr, 0, len(aggCols))
 
 	// Deduplicate the columns; we don't need to produce the same aggregation
 	// multiple times.
 	colSet := opt.ColSet{}
 	for i := range aggCols {
-		if id, group := aggCols[i].id, aggCols[i].group; !colSet.Contains(int(id)) {
-			if group == 0 {
+		if id, scalar := aggCols[i].id, aggCols[i].scalar; !colSet.Contains(int(id)) {
+			if scalar == nil {
 				// A "pass through" column (i.e. a VariableOp) is not legal as an
 				// aggregation.
-				panic("variable as aggregation")
+				panic(pgerror.NewAssertionErrorf("variable as aggregation"))
 			}
-			colList = append(colList, id)
-			groupList = append(groupList, group)
+			aggs = append(aggs, memo.AggregationsItem{
+				Agg:        scalar,
+				ColPrivate: memo.ColPrivate{Col: id},
+			})
 			colSet.Add(int(id))
 		}
 	}
 
-	aggs := b.factory.ConstructAggregations(
-		b.factory.InternList(groupList),
-		b.factory.InternColList(colList),
-	)
-	def := memo.GroupByDef{GroupingCols: groupingColSet}
+	private := memo.GroupingPrivate{GroupingCols: groupingColSet}
+
 	// The ordering of the GROUP BY is inherited from the input. This ordering is
 	// only useful for intra-group ordering (for order-sensitive aggregations like
 	// ARRAY_AGG). So we add the grouping columns as optional columns.
-	def.Ordering.FromOrderingWithOptCols(ordering, groupingColSet)
-
-	// TODO(radu): the execbuilder doesn't correctly handle orderings with
-	// optional columns that are not constant (#31882). The fix for that is
-	// involved; for now we don't set the grouping columns as optional. The
-	// exploration rule that looks at interesting orderings negates (or at least
-	// alleviates) the impact of restricting the Ordering in this way.
-	def.Ordering.Optional = opt.ColSet{}
+	private.Ordering.FromOrderingWithOptCols(ordering, groupingColSet)
 
 	if groupingColSet.Empty() {
-		return b.factory.ConstructScalarGroupBy(input, aggs, b.factory.InternGroupByDef(&def))
+		return b.factory.ConstructScalarGroupBy(input, aggs, &private)
 	}
-	return b.factory.ConstructGroupBy(input, aggs, b.factory.InternGroupByDef(&def))
+	return b.factory.ConstructGroupBy(input, aggs, &private)
 }
 
 // buildAggregation builds the pre-projection and the aggregation operators.
@@ -224,13 +214,13 @@ func (b *Builder) buildAggregation(
 
 	// The "from" columns are visible to any grouping expressions.
 	b.buildGroupingList(sel.GroupBy, sel.Exprs, fromScope, aggInScope)
-	groupingsLen := len(fromScope.groupby.groupings)
+	groupingsLen := len(fromScope.groupby.groupStrs)
 
 	// Copy the grouping columns to the aggOutScope.
 	groupingCols := aggInScope.getGroupingCols(groupingsLen)
 	aggOutScope.appendColumns(groupingCols)
 
-	var having memo.GroupID
+	var having opt.ScalarExpr
 	if sel.Having != nil {
 		// Any "grouping" columns are visible to both the "having" and "projection"
 		// expressions.
@@ -241,7 +231,7 @@ func (b *Builder) buildAggregation(
 	b.buildOrderBy(fromScope, projectionsScope, orderByScope)
 	b.buildDistinctOnArgs(fromScope, projectionsScope, distinctOnScope)
 	if len(fromScope.srfs) > 0 {
-		fromScope.group = b.constructProjectSet(fromScope.group, fromScope.srfs)
+		fromScope.expr = b.constructProjectSet(fromScope.expr, fromScope.srfs)
 	}
 
 	aggInfos := aggOutScope.groupby.aggs
@@ -256,22 +246,38 @@ func (b *Builder) buildAggregation(
 		fromCols = fromScope.colSet()
 	}
 	for i, agg := range aggInfos {
-		var operands memo.DynamicOperands
-		for j := range agg.args {
+		args := make([]opt.ScalarExpr, 0, 2)
+		if len(agg.args) > 0 {
 			colID := argCols[0].id
 			argCols = argCols[1:]
-			arg := b.factory.ConstructVariable(b.factory.InternColumnID(colID))
+			args = append(args, b.factory.ConstructVariable(colID))
 			if agg.distinct {
 				// Wrap the argument with AggDistinct.
-				arg = b.factory.ConstructAggDistinct(arg)
+				args[0] = b.factory.ConstructAggDistinct(args[0].(opt.ScalarExpr))
+			}
+
+			// Append any constant arguments without further processing.
+			constArgs := agg.args[1:]
+			args = append(args, constArgs...)
+			argCols = argCols[len(constArgs):]
+
+			// If the aggregate had a filter, there's an extra column in argCols
+			// corresponding to the filter.
+			// TODO(justin): add a norm rule to push these filters below group bys where appropriate.
+			if agg.filter != nil {
+				colID := argCols[0].id
+				argCols = argCols[1:]
+				// Wrap the argument with AggFilter.
+				args[0] = b.factory.ConstructAggFilter(args[0], b.factory.ConstructVariable(colID))
 			}
-			operands[j] = memo.DynamicID(arg)
 		}
-		aggOp := aggOpLookup[agg.def.Name]
-		if opt.AggregateIsOrderingSensitive(aggOp) {
+
+		aggCols[i].scalar = b.constructAggregate(agg.def.Name, args).(opt.ScalarExpr)
+
+		if opt.AggregateIsOrderingSensitive(aggCols[i].scalar.Op()) {
 			haveOrderingSensitiveAgg = true
 		}
-		aggCols[i].group = b.factory.DynamicConstruct(aggOp, operands)
+
 		if b.subquery != nil {
 			// Update the subquery with any outer columns from the aggregate
 			// arguments. The outer columns were not added in finishBuildScalarRef
@@ -297,18 +303,18 @@ func (b *Builder) buildAggregation(
 		groupingColSet.Add(int(groupingCols[i].id))
 	}
 
-	aggOutScope.group = b.constructGroupBy(
-		aggInScope.group,
+	aggOutScope.expr = b.constructGroupBy(
+		aggInScope.expr.(memo.RelExpr),
 		groupingColSet,
 		aggCols,
 		aggInScope.ordering,
 	)
 
 	// Wrap with having filter if it exists.
-	if having != 0 {
-		// Wrap the filter in a FiltersOp.
-		having = b.factory.ConstructFilters(b.factory.InternList([]memo.GroupID{having}))
-		aggOutScope.group = b.factory.ConstructSelect(aggOutScope.group, having)
+	if having != nil {
+		input := aggOutScope.expr.(memo.RelExpr)
+		filters := memo.FiltersExpr{{Condition: having}}
+		aggOutScope.expr = b.factory.ConstructSelect(input, filters)
 	}
 	return aggOutScope
 }
@@ -335,7 +341,7 @@ func (b *Builder) analyzeHaving(having *tree.Where, inScope *scope) tree.TypedEx
 //
 // The return value corresponds to the top-level memo group ID for this
 // HAVING clause.
-func (b *Builder) buildHaving(having tree.TypedExpr, inScope *scope) memo.GroupID {
+func (b *Builder) buildHaving(having tree.TypedExpr, inScope *scope) opt.ScalarExpr {
 	return b.buildScalar(having, inScope, nil, nil, nil)
 }
 
@@ -352,7 +358,6 @@ func (b *Builder) buildHaving(having tree.TypedExpr, inScope *scope) memo.GroupI
 func (b *Builder) buildGroupingList(
 	groupBy tree.GroupBy, selects tree.SelectExprs, inScope *scope, outScope *scope,
 ) {
-	inScope.groupby.groupings = make([]memo.GroupID, 0, len(groupBy))
 	inScope.groupby.groupStrs = make(groupByStrSet, len(groupBy))
 	if outScope.cols == nil {
 		outScope.cols = make([]scopeColumn, 0, len(groupBy))
@@ -382,10 +387,10 @@ func (b *Builder) buildGrouping(
 	// Check whether the GROUP BY clause refers to a column in the SELECT list
 	// by index, e.g. `SELECT a, SUM(b) FROM y GROUP BY 1`.
 	col := colIndex(len(selects), groupBy, "GROUP BY")
-	label := ""
+	alias := ""
 	if col != -1 {
 		groupBy = selects[col].Expr
-		label = string(selects[col].As)
+		alias = string(selects[col].As)
 	}
 
 	// We need to save and restore the previous value of the field in semaCtx
@@ -402,16 +407,38 @@ func (b *Builder) buildGrouping(
 
 	// Finally, build each of the GROUP BY columns.
 	for _, e := range exprs {
+		// If a grouping column has already been added, don't add it again.
+		// GROUP BY a, a is semantically equivalent to GROUP BY a.
+		exprStr := symbolicExprStr(e)
+		if _, ok := inScope.groupby.groupStrs[exprStr]; ok {
+			continue
+		}
+
 		// Save a representation of the GROUP BY expression for validation of the
 		// SELECT and HAVING expressions. This enables queries such as:
 		//   SELECT x+y FROM t GROUP BY x+y
-		col := b.addColumn(outScope, label, e.ResolvedType(), e)
+		col := b.addColumn(outScope, alias, e)
 		b.buildScalar(e, inScope, outScope, col, nil)
-		inScope.groupby.groupStrs[symbolicExprStr(e)] = col
-		inScope.groupby.groupings = append(inScope.groupby.groupings, col.group)
+		inScope.groupby.groupStrs[exprStr] = col
 	}
 }
 
+// buildAggArg builds a scalar expression which is used as an input in some form
+// to an aggregate expression. The scopeColumn for the built expression will
+// be added to tempScope.
+func (b *Builder) buildAggArg(
+	e tree.TypedExpr, info *aggregateInfo, tempScope, inScope *scope,
+) opt.ScalarExpr {
+	// This synthesizes a new tempScope column, unless the argument is a
+	// simple VariableOp.
+	col := b.addColumn(tempScope, "" /* alias */, e)
+	b.buildScalar(e, inScope, tempScope, col, &info.colRefs)
+	if col.scalar != nil {
+		return col.scalar
+	}
+	return b.factory.ConstructVariable(col.id)
+}
+
 // buildAggregateFunction is called when we are building a function which is an
 // aggregate. Any non-trivial parameters (i.e. not column reference) to the
 // aggregate function are extracted and added to aggInScope. The aggregate
@@ -426,24 +453,16 @@ func (b *Builder) buildGrouping(
 // the function definition, fully built arguments, and the aggregate output
 // column.
 func (b *Builder) buildAggregateFunction(
-	f *tree.FuncExpr, funcDef memo.FuncOpDef, inScope *scope,
+	f *tree.FuncExpr, def *memo.FunctionPrivate, inScope *scope,
 ) *aggregateInfo {
-	if len(f.Exprs) > 1 {
-		// TODO: #10495
-		panic(builderError{pgerror.UnimplementedWithIssueError(
-			10495,
-			"aggregate functions with multiple arguments are not supported yet"),
-		})
-	}
-
 	tempScope := inScope.startAggFunc()
 	tempScopeColsBefore := len(tempScope.cols)
 
 	info := aggregateInfo{
 		FuncExpr: f,
-		def:      funcDef,
+		def:      *def,
 		distinct: (f.Type == tree.DistinctFuncType),
-		args:     make([]memo.GroupID, len(f.Exprs)),
+		args:     make(memo.ScalarListExpr, len(f.Exprs)),
 	}
 
 	// Temporarily set b.subquery to nil so we don't add outer columns to the
@@ -453,15 +472,13 @@ func (b *Builder) buildAggregateFunction(
 	defer func() { b.subquery = subq }()
 
 	for i, pexpr := range f.Exprs {
-		// This synthesizes a new tempScope column, unless the argument is a
-		// simple VariableOp.
-		texpr := pexpr.(tree.TypedExpr)
-		col := b.addColumn(tempScope, "" /* label */, texpr.ResolvedType(), texpr)
-		b.buildScalar(texpr, inScope, tempScope, col, &info.colRefs)
-		info.args[i] = col.group
-		if info.args[i] == 0 {
-			info.args[i] = b.factory.ConstructVariable(b.factory.InternColumnID(col.id))
-		}
+		info.args[i] = b.buildAggArg(pexpr.(tree.TypedExpr), &info, tempScope, inScope)
+	}
+
+	// If we have a filter, add it to tempScope after all the arguments. We'll
+	// later extract the column that gets added here in buildAggregation.
+	if f.Filter != nil {
+		info.filter = b.buildAggArg(f.Filter.(tree.TypedExpr), &info, tempScope, inScope)
 	}
 
 	// Find the appropriate aggregation scopes for this aggregate now that we
@@ -476,7 +493,7 @@ func (b *Builder) buildAggregateFunction(
 	if info.col == nil {
 		// Use 0 as the group for now; it will be filled in later by the
 		// buildAggregation method.
-		info.col = b.synthesizeColumn(aggOutScope, funcDef.Name, f.ResolvedType(), f, 0 /* group */)
+		info.col = b.synthesizeColumn(aggOutScope, def.Name, f.ResolvedType(), f, nil /* scalar */)
 
 		// Move the columns for the aggregate input expressions to the correct scope.
 		if aggInScope != tempScope {
@@ -496,6 +513,53 @@ func (b *Builder) buildAggregateFunction(
 	return &info
 }
 
+func (b *Builder) constructAggregate(name string, args []opt.ScalarExpr) opt.ScalarExpr {
+	switch name {
+	case "array_agg":
+		return b.factory.ConstructArrayAgg(args[0])
+	case "avg":
+		return b.factory.ConstructAvg(args[0])
+	case "bool_and":
+		return b.factory.ConstructBoolAnd(args[0])
+	case "bool_or":
+		return b.factory.ConstructBoolOr(args[0])
+	case "concat_agg":
+		return b.factory.ConstructConcatAgg(args[0])
+	case "count":
+		return b.factory.ConstructCount(args[0])
+	case "count_rows":
+		return b.factory.ConstructCountRows()
+	case "max":
+		return b.factory.ConstructMax(args[0])
+	case "min":
+		return b.factory.ConstructMin(args[0])
+	case "sum_int":
+		return b.factory.ConstructSumInt(args[0])
+	case "sum":
+		return b.factory.ConstructSum(args[0])
+	case "sqrdiff":
+		return b.factory.ConstructSqrDiff(args[0])
+	case "variance":
+		return b.factory.ConstructVariance(args[0])
+	case "stddev":
+		return b.factory.ConstructStdDev(args[0])
+	case "xor_agg":
+		return b.factory.ConstructXorAgg(args[0])
+	case "json_agg":
+		return b.factory.ConstructJsonAgg(args[0])
+	case "jsonb_agg":
+		return b.factory.ConstructJsonbAgg(args[0])
+	case "string_agg":
+		if !memo.CanExtractConstDatum(args[1]) {
+			panic(builderError{
+				fmt.Errorf("unimplemented: aggregate functions with multiple non-constant expressions are not supported"),
+			})
+		}
+		return b.factory.ConstructStringAgg(args[0], args[1])
+	}
+	panic(fmt.Sprintf("unhandled aggregate: %s", name))
+}
+
 func isAggregate(def *tree.FunctionDefinition) bool {
 	return def.Class == tree.AggregateClass
 }
@@ -504,26 +568,6 @@ func isGenerator(def *tree.FunctionDefinition) bool {
 	return def.Class == tree.GeneratorClass
 }
 
-var aggOpLookup = map[string]opt.Operator{
-	"array_agg":  opt.ArrayAggOp,
-	"avg":        opt.AvgOp,
-	"bool_and":   opt.BoolAndOp,
-	"bool_or":    opt.BoolOrOp,
-	"concat_agg": opt.ConcatAggOp,
-	"count":      opt.CountOp,
-	"count_rows": opt.CountRowsOp,
-	"max":        opt.MaxOp,
-	"min":        opt.MinOp,
-	"sum_int":    opt.SumIntOp,
-	"sum":        opt.SumOp,
-	"sqrdiff":    opt.SqrDiffOp,
-	"variance":   opt.VarianceOp,
-	"stddev":     opt.StdDevOp,
-	"xor_agg":    opt.XorAggOp,
-	"json_agg":   opt.JsonAggOp,
-	"jsonb_agg":  opt.JsonbAggOp,
-}
-
 func newGroupingError(name *tree.Name) error {
 	return pgerror.NewErrorf(pgerror.CodeGroupingError,
 		"column \"%s\" must appear in the GROUP BY clause or be used in an aggregate function",
