@@ -18,87 +18,26 @@ import (
 	"context"
 	"fmt"
 
+	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/memo"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/norm"
-	"github.com/cockroachdb/cockroach/pkg/sql/opt/props"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqltelemetry"
 )
 
-type unaryFactoryFunc func(f *norm.Factory, input memo.GroupID) memo.GroupID
-type binaryFactoryFunc func(f *norm.Factory, left, right memo.GroupID) memo.GroupID
-
 func checkArrayElementType(t types.T) error {
-	if !types.IsValidArrayElementType(t) {
-		return pgerror.NewErrorf(pgerror.CodeFeatureNotSupportedError,
+	if ok, issueNum := types.IsValidArrayElementType(t); !ok {
+		return pgerror.UnimplementedWithIssueDetailErrorf(issueNum, t.String(),
 			"arrays of %s not allowed", t)
 	}
 	return nil
 }
 
-// Map from tree.ComparisonOperator to Factory constructor function.
-var comparisonOpMap = [tree.NumComparisonOperators]binaryFactoryFunc{
-	tree.EQ:                (*norm.Factory).ConstructEq,
-	tree.LT:                (*norm.Factory).ConstructLt,
-	tree.GT:                (*norm.Factory).ConstructGt,
-	tree.LE:                (*norm.Factory).ConstructLe,
-	tree.GE:                (*norm.Factory).ConstructGe,
-	tree.NE:                (*norm.Factory).ConstructNe,
-	tree.In:                (*norm.Factory).ConstructIn,
-	tree.NotIn:             (*norm.Factory).ConstructNotIn,
-	tree.Like:              (*norm.Factory).ConstructLike,
-	tree.NotLike:           (*norm.Factory).ConstructNotLike,
-	tree.ILike:             (*norm.Factory).ConstructILike,
-	tree.NotILike:          (*norm.Factory).ConstructNotILike,
-	tree.SimilarTo:         (*norm.Factory).ConstructSimilarTo,
-	tree.NotSimilarTo:      (*norm.Factory).ConstructNotSimilarTo,
-	tree.RegMatch:          (*norm.Factory).ConstructRegMatch,
-	tree.NotRegMatch:       (*norm.Factory).ConstructNotRegMatch,
-	tree.RegIMatch:         (*norm.Factory).ConstructRegIMatch,
-	tree.NotRegIMatch:      (*norm.Factory).ConstructNotRegIMatch,
-	tree.IsDistinctFrom:    (*norm.Factory).ConstructIsNot,
-	tree.IsNotDistinctFrom: (*norm.Factory).ConstructIs,
-	tree.Contains:          (*norm.Factory).ConstructContains,
-	tree.ContainedBy: func(f *norm.Factory, left, right memo.GroupID) memo.GroupID {
-		// This is just syntatic sugar that reverses the operands.
-		return f.ConstructContains(right, left)
-	},
-	tree.JSONExists:     (*norm.Factory).ConstructJsonExists,
-	tree.JSONAllExists:  (*norm.Factory).ConstructJsonAllExists,
-	tree.JSONSomeExists: (*norm.Factory).ConstructJsonSomeExists,
-}
-
-// Map from tree.BinaryOperator to Factory constructor function.
-var binaryOpMap = [tree.NumBinaryOperators]binaryFactoryFunc{
-	tree.Bitand:            (*norm.Factory).ConstructBitand,
-	tree.Bitor:             (*norm.Factory).ConstructBitor,
-	tree.Bitxor:            (*norm.Factory).ConstructBitxor,
-	tree.Plus:              (*norm.Factory).ConstructPlus,
-	tree.Minus:             (*norm.Factory).ConstructMinus,
-	tree.Mult:              (*norm.Factory).ConstructMult,
-	tree.Div:               (*norm.Factory).ConstructDiv,
-	tree.FloorDiv:          (*norm.Factory).ConstructFloorDiv,
-	tree.Mod:               (*norm.Factory).ConstructMod,
-	tree.Pow:               (*norm.Factory).ConstructPow,
-	tree.Concat:            (*norm.Factory).ConstructConcat,
-	tree.LShift:            (*norm.Factory).ConstructLShift,
-	tree.RShift:            (*norm.Factory).ConstructRShift,
-	tree.JSONFetchText:     (*norm.Factory).ConstructFetchText,
-	tree.JSONFetchVal:      (*norm.Factory).ConstructFetchVal,
-	tree.JSONFetchValPath:  (*norm.Factory).ConstructFetchValPath,
-	tree.JSONFetchTextPath: (*norm.Factory).ConstructFetchTextPath,
-}
-
-// Map from tree.UnaryOperator to Factory constructor function.
-var unaryOpMap = [tree.NumUnaryOperators]unaryFactoryFunc{
-	tree.UnaryMinus:      (*norm.Factory).ConstructUnaryMinus,
-	tree.UnaryComplement: (*norm.Factory).ConstructUnaryComplement,
-}
-
 // buildScalar builds a set of memo groups that represent the given scalar
 // expression. If outScope is not nil, then this is a projection context, and
 // the resulting memo group will be projected as the output column outCol.
@@ -113,7 +52,7 @@ var unaryOpMap = [tree.NumUnaryOperators]unaryFactoryFunc{
 // values.
 func (b *Builder) buildScalar(
 	scalar tree.TypedExpr, inScope, outScope *scope, outCol *scopeColumn, colRefs *opt.ColSet,
-) (out memo.GroupID) {
+) (out opt.ScalarExpr) {
 	// If we are in a grouping context and this expression corresponds to a
 	// GROUP BY expression, return a reference to the GROUP BY column.
 	// Note that GROUP BY columns cannot be reused inside an aggregate input
@@ -153,11 +92,10 @@ func (b *Builder) buildScalar(
 	case *tree.AndExpr:
 		left := b.buildScalar(t.TypedLeft(), inScope, nil, nil, colRefs)
 		right := b.buildScalar(t.TypedRight(), inScope, nil, nil, colRefs)
-		conditions := b.factory.InternList([]memo.GroupID{left, right})
-		out = b.factory.ConstructAnd(conditions)
+		out = b.factory.ConstructAnd(left, right)
 
 	case *tree.Array:
-		els := make([]memo.GroupID, len(t.Exprs))
+		els := make(memo.ScalarListExpr, len(t.Exprs))
 		arrayType := t.ResolvedType()
 		elementType := arrayType.(types.TArray).Typ
 		if err := checkArrayElementType(elementType); err != nil {
@@ -167,82 +105,81 @@ func (b *Builder) buildScalar(
 			texpr := t.Exprs[i].(tree.TypedExpr)
 			els[i] = b.buildScalar(texpr, inScope, nil, nil, colRefs)
 		}
-		elements := b.factory.InternList(els)
-		out = b.factory.ConstructArray(elements, b.factory.InternType(arrayType))
+		out = b.factory.ConstructArray(els, arrayType)
+
+	case *tree.CollateExpr:
+		in := b.buildScalar(t.Expr.(tree.TypedExpr), inScope, nil, nil, colRefs)
+		out = b.factory.ConstructCollate(in, t.Locale)
 
 	case *tree.ArrayFlatten:
 		if b.AllowUnsupportedExpr {
-			out = b.factory.ConstructUnsupportedExpr(b.factory.InternTypedExpr(scalar))
+			out = b.factory.ConstructUnsupportedExpr(t)
 			break
 		}
 
-		// We build
-		//
-		//  ARRAY(<subquery>)
-		//
-		// as
-		//
-		//   COALESCE(
-		//     (SELECT array_agg(x) FROM (<subquery>)),
-		//     ARRAY[]
-		//   )
-		//
-		// The COALESCE is needed because ARRAY(<empty subquery>) needs to return
-		// an empty array, while ARRAY_AGG with no inputs returns NULL.
-
 		s := t.Subquery.(*subquery)
-		aggInputColID := s.cols[0].id
 
-		elemType := b.factory.Metadata().ColumnType(aggInputColID)
-		if err := checkArrayElementType(elemType); err != nil {
+		inCol := s.cols[0].id
+
+		// This looks kind of arbitrary and strange, because it is:
+		// We cannot array_agg over some types, but we can only decorrelate via array_agg.
+		// Thus, we reject a query that is correlated and over a type that we can't array_agg.
+		typ := b.factory.Metadata().ColumnMeta(inCol).Type
+		if !s.outerCols.Empty() && !memo.AggregateOverloadExists(opt.ArrayAggOp, typ) {
+			panic(unimplementedWithIssueDetailf(35710, "", "can't execute a correlated ARRAY(...) over %s", typ))
+		}
+
+		if err := checkArrayElementType(typ); err != nil {
 			panic(builderError{err})
 		}
 
-		switch elemType.(type) {
-		case types.TTuple, types.TArray:
-			// We build this into ARRAY_AGG which doesn't support non-scalar types.
-			panic(unimplementedf("can't build ARRAY(%s)", elemType))
-		}
-
-		aggColID := b.factory.Metadata().AddColumn(
-			"array_agg",
-			types.TArray{Typ: elemType},
-		)
-
-		typID := b.factory.InternType(types.TArray{Typ: elemType})
-
-		var oc props.OrderingChoice
-		oc.FromOrdering(s.ordering)
-
-		out = b.factory.ConstructCoalesce(b.factory.InternList([]memo.GroupID{
-			b.factory.ConstructSubquery(
-				// A ScalarGroupBy always returns exactly one row, so there's no need
-				// for a Max1Row here.
-				b.factory.ConstructScalarGroupBy(
-					s.group,
-					b.factory.ConstructAggregations(
-						b.factory.InternList([]memo.GroupID{
-							b.factory.ConstructArrayAgg(
-								b.factory.ConstructVariable(
-									b.factory.InternColumnID(aggInputColID),
-								),
-							),
-						}),
-						b.factory.InternColList(opt.ColList{aggColID}),
-					),
-					b.factory.InternGroupByDef(&memo.GroupByDef{
-						Ordering: oc,
-					}),
-				),
-				b.factory.InternSubqueryDef(&memo.SubqueryDef{OriginalExpr: s.Subquery}),
-			),
-			b.factory.ConstructArray(memo.EmptyList, typID),
-		}))
-
 		// Perform correctness checks on the outer cols, update colRefs and
 		// b.subquery.outerCols.
 		b.checkSubqueryOuterCols(s.outerCols, inGroupingContext, inScope, colRefs)
 
+		subqueryPrivate := memo.SubqueryPrivate{
+			OriginalExpr: s.Subquery,
+			Ordering:     s.ordering,
+			RequestedCol: inCol,
+		}
+		out = b.factory.ConstructArrayFlatten(s.node, &subqueryPrivate)
+
+	case *tree.IndirectionExpr:
+		expr := b.buildScalar(t.Expr.(tree.TypedExpr), inScope, nil, nil, colRefs)
+
+		if len(t.Indirection) != 1 {
+			panic(unimplementedWithIssueDetailf(32552, "ind", "multidimensional indexing is not supported"))
+		}
+
+		subscript := t.Indirection[0]
+		if subscript.Slice {
+			panic(unimplementedWithIssueDetailf(32551, "", "array slicing is not supported"))
+		}
+
+		out = b.factory.ConstructIndirection(
+			expr,
+			b.buildScalar(subscript.Begin.(tree.TypedExpr), inScope, nil, nil, colRefs),
+		)
+
+	case *tree.IfErrExpr:
+		cond := b.buildScalar(t.Cond.(tree.TypedExpr), inScope, nil, nil, colRefs)
+
+		orElse := memo.EmptyScalarListExpr
+		if t.Else != nil {
+			orElse = memo.ScalarListExpr{
+				b.buildScalar(t.Else.(tree.TypedExpr), inScope, nil, nil, colRefs),
+			}
+		}
+
+		errCode := memo.EmptyScalarListExpr
+		if t.ErrCode != nil {
+			errCode = memo.ScalarListExpr{
+				b.buildScalar(t.ErrCode.(tree.TypedExpr), inScope, nil, nil, colRefs),
+			}
+		}
+
+		out = b.factory.ConstructIfErr(cond, orElse, errCode)
+
 	case *tree.BinaryExpr:
 		// It's possible for an overload to be selected that expects different
 		// types than the TypedExpr arguments return:
@@ -255,24 +192,24 @@ func (b *Builder) buildScalar(
 		// select the right overload. The solution is to wrap any mismatched
 		// arguments with a CastExpr that preserves the static type.
 
-		fn := binaryOpMap[t.Operator]
 		left, _ := tree.ReType(t.TypedLeft(), t.ResolvedBinOp().LeftType)
 		right, _ := tree.ReType(t.TypedRight(), t.ResolvedBinOp().RightType)
-		out = fn(b.factory,
+		out = b.constructBinary(t.Operator,
 			b.buildScalar(left, inScope, nil, nil, colRefs),
 			b.buildScalar(right, inScope, nil, nil, colRefs),
+			t.ResolvedType(),
 		)
 
 	case *tree.CaseExpr:
-		var input memo.GroupID
+		var input opt.ScalarExpr
 		if t.Expr != nil {
 			texpr := t.Expr.(tree.TypedExpr)
 			input = b.buildScalar(texpr, inScope, nil, nil, colRefs)
 		} else {
-			input = b.factory.ConstructTrue()
+			input = memo.TrueSingleton
 		}
 
-		whens := make([]memo.GroupID, 0, len(t.Whens)+1)
+		whens := make(memo.ScalarListExpr, 0, len(t.Whens)+1)
 		for i := range t.Whens {
 			texpr := t.Whens[i].Cond.(tree.TypedExpr)
 			cond := b.buildScalar(texpr, inScope, nil, nil, colRefs)
@@ -281,101 +218,85 @@ func (b *Builder) buildScalar(
 			whens = append(whens, b.factory.ConstructWhen(cond, val))
 		}
 		// Add the ELSE expression to the end of whens as a raw scalar expression.
+		var orElse opt.ScalarExpr
 		if t.Else != nil {
 			texpr := t.Else.(tree.TypedExpr)
-			elseExpr := b.buildScalar(texpr, inScope, nil, nil, colRefs)
-			whens = append(whens, elseExpr)
+			orElse = b.buildScalar(texpr, inScope, nil, nil, colRefs)
 		} else {
-			whens = append(whens, b.factory.ConstructConstVal(tree.DNull))
+			orElse = memo.NullSingleton
 		}
-		out = b.factory.ConstructCase(input, b.factory.InternList(whens))
-
-	case *tree.IfExpr:
-		cond := b.buildScalar(t.Cond.(tree.TypedExpr), inScope, nil, nil, colRefs)
-		tru := b.buildScalar(t.True.(tree.TypedExpr), inScope, nil, nil, colRefs)
-		els := b.buildScalar(t.Else.(tree.TypedExpr), inScope, nil, nil, colRefs)
-		whens := []memo.GroupID{
-			b.factory.ConstructWhen(b.factory.ConstructTrue(), tru),
-			els,
-		}
-		out = b.factory.ConstructCase(cond, b.factory.InternList(whens))
-
-	case *tree.NullIfExpr:
-		e1 := b.buildScalar(t.Expr1.(tree.TypedExpr), inScope, nil, nil, colRefs)
-		e2 := b.buildScalar(t.Expr2.(tree.TypedExpr), inScope, nil, nil, colRefs)
-		whens := []memo.GroupID{
-			b.factory.ConstructWhen(e2, b.factory.ConstructConstVal(tree.DNull)),
-			e1,
-		}
-		out = b.factory.ConstructCase(e1, b.factory.InternList(whens))
+		out = b.factory.ConstructCase(input, whens, orElse)
 
 	case *tree.CastExpr:
 		texpr := t.Expr.(tree.TypedExpr)
 		arg := b.buildScalar(texpr, inScope, nil, nil, colRefs)
-		out = b.factory.ConstructCast(arg, b.factory.InternColType(t.Type.(coltypes.T)))
+		out = b.factory.ConstructCast(arg, t.Type.(coltypes.T))
 
 	case *tree.CoalesceExpr:
-		args := make([]memo.GroupID, len(t.Exprs))
+		args := make(memo.ScalarListExpr, len(t.Exprs))
 		for i := range args {
 			args[i] = b.buildScalar(t.TypedExprAt(i), inScope, nil, nil, colRefs)
 		}
-		out = b.factory.ConstructCoalesce(b.factory.InternList(args))
+		out = b.factory.ConstructCoalesce(args)
 
 	case *tree.ColumnAccessExpr:
 		input := b.buildScalar(t.Expr.(tree.TypedExpr), inScope, nil, nil, colRefs)
-		out = b.factory.ConstructColumnAccess(
-			input, b.factory.InternTupleOrdinal(memo.TupleOrdinal(t.ColIndex)),
-		)
+		out = b.factory.ConstructColumnAccess(input, memo.TupleOrdinal(t.ColIndex))
 
 	case *tree.ComparisonExpr:
-		if sub, ok := t.Right.(*subquery); ok && sub.wrapInTuple {
+		if sub, ok := t.Right.(*subquery); ok && sub.isMultiRow() {
 			out, _ = b.buildMultiRowSubquery(t, inScope, colRefs)
 			// Perform correctness checks on the outer cols, update colRefs and
 			// b.subquery.outerCols.
 			b.checkSubqueryOuterCols(sub.outerCols, inGroupingContext, inScope, colRefs)
 		} else if b.hasSubOperator(t) {
-			// Cases where the RHS is a subquery and not a scalar (of which only an
-			// array or tuple is legal) were handled above.
+			// Cases where the RHS is a multi-row subquery were handled above, so this
+			// only handles explicit tuples and arrays.
 			out = b.buildAnyScalar(t, inScope, colRefs)
 		} else {
 			left := b.buildScalar(t.TypedLeft(), inScope, nil, nil, colRefs)
 			right := b.buildScalar(t.TypedRight(), inScope, nil, nil, colRefs)
-
-			fn := comparisonOpMap[t.Operator]
-
-			if fn == nil {
-				panic(builderError{fmt.Errorf("unknown comparison operator: %s", t.Operator)})
-			}
-
-			// Most comparison ops map directly to a factory method.
-			out = fn(b.factory, left, right)
+			out = b.constructComparison(t.Operator, left, right)
 		}
 
 	case *tree.DTuple:
-		list := make([]memo.GroupID, len(t.D))
+		els := make(memo.ScalarListExpr, len(t.D))
 		for i := range t.D {
-			list[i] = b.buildScalar(t.D[i], inScope, nil, nil, colRefs)
+			els[i] = b.buildScalar(t.D[i], inScope, nil, nil, colRefs)
 		}
-		out = b.factory.ConstructTuple(b.factory.InternList(list), b.factory.InternType(t.ResolvedType()))
+		out = b.factory.ConstructTuple(els, t.ResolvedType())
 
 	case *tree.FuncExpr:
 		return b.buildFunction(t, inScope, outScope, outCol, colRefs)
 
+	case *tree.IfExpr:
+		input := b.buildScalar(t.Cond.(tree.TypedExpr), inScope, nil, nil, colRefs)
+		ifTrue := b.buildScalar(t.True.(tree.TypedExpr), inScope, nil, nil, colRefs)
+		whens := memo.ScalarListExpr{b.factory.ConstructWhen(memo.TrueSingleton, ifTrue)}
+		orElse := b.buildScalar(t.Else.(tree.TypedExpr), inScope, nil, nil, colRefs)
+		out = b.factory.ConstructCase(input, whens, orElse)
+
 	case *tree.IndexedVar:
 		if t.Idx < 0 || t.Idx >= len(inScope.cols) {
-			panic(builderError{pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
-				"invalid column ordinal: @%d", t.Idx+1)})
+			panic(pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
+				"invalid column ordinal: @%d", t.Idx+1))
 		}
-		out = b.factory.ConstructVariable(b.factory.InternColumnID(inScope.cols[t.Idx].id))
+		out = b.factory.ConstructVariable(inScope.cols[t.Idx].id)
 
 	case *tree.NotExpr:
-		out = b.factory.ConstructNot(b.buildScalar(t.TypedInnerExpr(), inScope, nil, nil, colRefs))
+		input := b.buildScalar(t.TypedInnerExpr(), inScope, nil, nil, colRefs)
+		out = b.factory.ConstructNot(input)
+
+	case *tree.NullIfExpr:
+		input := b.buildScalar(t.Expr1.(tree.TypedExpr), inScope, nil, nil, colRefs)
+		cond := b.buildScalar(t.Expr2.(tree.TypedExpr), inScope, nil, nil, colRefs)
+		whens := memo.ScalarListExpr{b.factory.ConstructWhen(cond, memo.NullSingleton)}
+		out = b.factory.ConstructCase(input, whens, input)
 
 	case *tree.OrExpr:
 		left := b.buildScalar(t.TypedLeft(), inScope, nil, nil, colRefs)
 		right := b.buildScalar(t.TypedRight(), inScope, nil, nil, colRefs)
-		conditions := b.factory.InternList([]memo.GroupID{left, right})
-		out = b.factory.ConstructOr(conditions)
+		out = b.factory.ConstructOr(left, right)
 
 	case *tree.ParenExpr:
 		// Treat ParenExpr as if it wasn't present.
@@ -383,14 +304,15 @@ func (b *Builder) buildScalar(
 
 	case *tree.Placeholder:
 		if !b.KeepPlaceholders && b.evalCtx.HasPlaceholders() {
+			b.HadPlaceholders = true
 			// Replace placeholders with their value.
 			d, err := t.Eval(b.evalCtx)
 			if err != nil {
 				panic(builderError{err})
 			}
-			out = b.factory.ConstructConstVal(d)
+			out = b.factory.ConstructConstVal(d, t.ResolvedType())
 		} else {
-			out = b.factory.ConstructPlaceholder(b.factory.InternTypedExpr(t))
+			out = b.factory.ConstructPlaceholder(t)
 		}
 
 	case *tree.RangeCond:
@@ -409,11 +331,11 @@ func (b *Builder) buildScalar(
 			}
 			return b.finishBuildScalarRef(&t.cols[0], inScope, outScope, outCol, colRefs)
 		}
-		list := make([]memo.GroupID, len(t.cols))
+		els := make(memo.ScalarListExpr, len(t.cols))
 		for i := range t.cols {
-			list[i] = b.buildScalar(&t.cols[i], inScope, nil, nil, colRefs)
+			els[i] = b.buildScalar(&t.cols[i], inScope, nil, nil, colRefs)
 		}
-		out = b.factory.ConstructTuple(b.factory.InternList(list), b.factory.InternType(t.ResolvedType()))
+		out = b.factory.ConstructTuple(els, t.ResolvedType())
 
 	case *subquery:
 		out, _ = b.buildSingleRowSubquery(t, inScope)
@@ -422,27 +344,47 @@ func (b *Builder) buildScalar(
 		b.checkSubqueryOuterCols(t.outerCols, inGroupingContext, inScope, colRefs)
 
 	case *tree.Tuple:
-		list := make([]memo.GroupID, len(t.Exprs))
+		els := make(memo.ScalarListExpr, len(t.Exprs))
 		for i := range t.Exprs {
-			list[i] = b.buildScalar(t.Exprs[i].(tree.TypedExpr), inScope, nil, nil, colRefs)
+			els[i] = b.buildScalar(t.Exprs[i].(tree.TypedExpr), inScope, nil, nil, colRefs)
 		}
-		out = b.factory.ConstructTuple(b.factory.InternList(list), b.factory.InternType(t.ResolvedType()))
+		out = b.factory.ConstructTuple(els, t.ResolvedType())
 
 	case *tree.UnaryExpr:
 		out = b.buildScalar(t.TypedInnerExpr(), inScope, nil, nil, colRefs)
-		out = unaryOpMap[t.Operator](b.factory, out)
+		out = b.constructUnary(t.Operator, out, t.ResolvedType())
+
+	case *tree.IsOfTypeExpr:
+		// IsOfTypeExpr is a little strange because its value can be determined
+		// statically just from the type of the expression.
+		actualType := t.Expr.(tree.TypedExpr).ResolvedType()
+
+		found := false
+		for _, typ := range t.Types {
+			wantTyp := coltypes.CastTargetToDatumType(typ)
+			if actualType.Equivalent(wantTyp) {
+				found = true
+				break
+			}
+		}
+
+		if found != t.Not {
+			out = b.factory.ConstructTrue()
+		} else {
+			out = b.factory.ConstructFalse()
+		}
 
 	// NB: this is the exception to the sorting of the case statements. The
 	// tree.Datum case needs to occur after *tree.Placeholder which implements
 	// Datum.
 	case tree.Datum:
-		out = b.factory.ConstructConstVal(t)
+		out = b.factory.ConstructConstVal(t, t.ResolvedType())
 
 	default:
 		if b.AllowUnsupportedExpr {
-			out = b.factory.ConstructUnsupportedExpr(b.factory.InternTypedExpr(scalar))
+			out = b.factory.ConstructUnsupportedExpr(scalar)
 		} else {
-			panic(unimplementedf("not yet implemented: scalar expression: %T", scalar))
+			panic(unimplementedWithIssueDetailf(34848, fmt.Sprintf("%T", scalar), "not yet implemented: scalar expression: %T", scalar))
 		}
 	}
 
@@ -455,7 +397,7 @@ func (b *Builder) hasSubOperator(t *tree.ComparisonExpr) bool {
 
 func (b *Builder) buildAnyScalar(
 	t *tree.ComparisonExpr, inScope *scope, colRefs *opt.ColSet,
-) memo.GroupID {
+) opt.ScalarExpr {
 	left := b.buildScalar(t.TypedLeft(), inScope, nil, nil, colRefs)
 	right := b.buildScalar(t.TypedRight(), inScope, nil, nil, colRefs)
 
@@ -465,12 +407,7 @@ func (b *Builder) buildAnyScalar(
 		subop = opt.NegateOpMap[subop]
 	}
 
-	out := b.factory.ConstructAnyScalar(
-		left,
-		right,
-		b.factory.InternOperator(subop),
-	)
-
+	out := b.factory.ConstructAnyScalar(left, right, subop)
 	if t.Operator == tree.All {
 		out = b.factory.ConstructNot(out)
 	}
@@ -481,8 +418,6 @@ func (b *Builder) buildAnyScalar(
 // expression.
 //
 // f        The given function expression.
-// label    If a new column is synthesized, it will be labeled with this
-//          string.
 // outCol   The output column of the function being built.
 // colRefs  The set of columns referenced so far by the scalar expression
 //          being built. If not nil, it is updated with any columns seen in
@@ -492,12 +427,12 @@ func (b *Builder) buildAnyScalar(
 // return values.
 func (b *Builder) buildFunction(
 	f *tree.FuncExpr, inScope, outScope *scope, outCol *scopeColumn, colRefs *opt.ColSet,
-) (out memo.GroupID) {
+) (out opt.ScalarExpr) {
 	if f.WindowDef != nil {
 		if inScope.groupby.inAgg {
 			panic(builderError{sqlbase.NewWindowInAggError()})
 		}
-		panic(unimplementedf("window functions are not supported"))
+		panic(unimplementedWithIssueDetailf(34251, "", "window functions are not supported"))
 	}
 
 	def, err := f.Func.Resolve(b.semaCtx.SearchPath)
@@ -505,26 +440,22 @@ func (b *Builder) buildFunction(
 		panic(builderError{err})
 	}
 
-	funcDef := memo.FuncOpDef{
-		Name:       def.Name,
-		Type:       f.ResolvedType(),
-		Properties: &def.FunctionProperties,
-		Overload:   f.ResolvedOverload(),
-	}
-
 	if isAggregate(def) {
-		panic("aggregate function should have been replaced")
+		panic(pgerror.NewAssertionErrorf("aggregate function should have been replaced"))
 	}
 
-	argList := make([]memo.GroupID, len(f.Exprs))
+	args := make(memo.ScalarListExpr, len(f.Exprs))
 	for i, pexpr := range f.Exprs {
-		argList[i] = b.buildScalar(pexpr.(tree.TypedExpr), inScope, nil, nil, colRefs)
+		args[i] = b.buildScalar(pexpr.(tree.TypedExpr), inScope, nil, nil, colRefs)
 	}
 
 	// Construct a private FuncOpDef that refers to a resolved function overload.
-	out = b.factory.ConstructFunction(
-		b.factory.InternList(argList), b.factory.InternFuncOpDef(&funcDef),
-	)
+	out = b.factory.ConstructFunction(args, &memo.FunctionPrivate{
+		Name:       def.Name,
+		Typ:        f.ResolvedType(),
+		Properties: &def.FunctionProperties,
+		Overload:   f.ResolvedOverload(),
+	})
 
 	if isGenerator(def) {
 		columns := len(def.ReturnLabels)
@@ -544,28 +475,22 @@ func (b *Builder) buildFunction(
 // push down the negation).
 // TODO(radu): this doesn't work when the expressions have side-effects.
 func (b *Builder) buildRangeCond(
-	not bool, symmetric bool, input, from, to memo.GroupID,
-) memo.GroupID {
+	not bool, symmetric bool, input, from, to opt.ScalarExpr,
+) opt.ScalarExpr {
 	// Build "input >= from AND input <= to".
 	out := b.factory.ConstructAnd(
-		b.factory.InternList([]memo.GroupID{
-			b.factory.ConstructGe(input, from),
-			b.factory.ConstructLe(input, to),
-		}),
+		b.factory.ConstructGe(input, from),
+		b.factory.ConstructLe(input, to),
 	)
 
 	if symmetric {
 		// Build "(input >= from AND input <= to) OR (input >= to AND input <= from)".
 		lhs := out
 		rhs := b.factory.ConstructAnd(
-			b.factory.InternList([]memo.GroupID{
-				b.factory.ConstructGe(input, to),
-				b.factory.ConstructLe(input, from),
-			}),
-		)
-		out = b.factory.ConstructOr(
-			b.factory.InternList([]memo.GroupID{lhs, rhs}),
+			b.factory.ConstructGe(input, to),
+			b.factory.ConstructLe(input, from),
 		)
+		out = b.factory.ConstructOr(lhs, rhs)
 	}
 
 	if not {
@@ -590,6 +515,18 @@ func (b *Builder) checkSubqueryOuterCols(
 		return
 	}
 
+	if !b.IsCorrelated {
+		// Remember whether the query was correlated for the heuristic planner,
+		// to enhance error messages.
+		// TODO(knz): this can go away when the HP disappears.
+		b.IsCorrelated = true
+
+		// Register the use of correlation to telemetry.
+		// Note: we don't blindly increment the counter every time this
+		// method is called, to avoid double counting the same query.
+		telemetry.Inc(sqltelemetry.CorrelatedSubqueryUseCounter)
+	}
+
 	var inScopeCols opt.ColSet
 	if b.subquery != nil || inGroupingContext {
 		// Only calculate the set of inScope columns if it will be used below.
@@ -623,14 +560,127 @@ func (b *Builder) checkSubqueryOuterCols(
 			subqueryOuterCols.DifferenceWith(inScope.groupby.aggOutScope.colSet())
 			colID, _ := subqueryOuterCols.Next(0)
 			col := inScope.getColumn(opt.ColumnID(colID))
-			panic(builderError{pgerror.NewErrorf(
+			panic(pgerror.NewErrorf(
 				pgerror.CodeGroupingError,
 				"subquery uses ungrouped column \"%s\" from outer query",
-				tree.ErrString(&col.name))})
+				tree.ErrString(&col.name)))
 		}
 	}
 }
 
+func (b *Builder) constructComparison(
+	cmp tree.ComparisonOperator, left, right opt.ScalarExpr,
+) opt.ScalarExpr {
+	switch cmp {
+	case tree.EQ:
+		return b.factory.ConstructEq(left, right)
+	case tree.LT:
+		return b.factory.ConstructLt(left, right)
+	case tree.GT:
+		return b.factory.ConstructGt(left, right)
+	case tree.LE:
+		return b.factory.ConstructLe(left, right)
+	case tree.GE:
+		return b.factory.ConstructGe(left, right)
+	case tree.NE:
+		return b.factory.ConstructNe(left, right)
+	case tree.In:
+		return b.factory.ConstructIn(left, right)
+	case tree.NotIn:
+		return b.factory.ConstructNotIn(left, right)
+	case tree.Like:
+		return b.factory.ConstructLike(left, right)
+	case tree.NotLike:
+		return b.factory.ConstructNotLike(left, right)
+	case tree.ILike:
+		return b.factory.ConstructILike(left, right)
+	case tree.NotILike:
+		return b.factory.ConstructNotILike(left, right)
+	case tree.SimilarTo:
+		return b.factory.ConstructSimilarTo(left, right)
+	case tree.NotSimilarTo:
+		return b.factory.ConstructNotSimilarTo(left, right)
+	case tree.RegMatch:
+		return b.factory.ConstructRegMatch(left, right)
+	case tree.NotRegMatch:
+		return b.factory.ConstructNotRegMatch(left, right)
+	case tree.RegIMatch:
+		return b.factory.ConstructRegIMatch(left, right)
+	case tree.NotRegIMatch:
+		return b.factory.ConstructNotRegIMatch(left, right)
+	case tree.IsDistinctFrom:
+		return b.factory.ConstructIsNot(left, right)
+	case tree.IsNotDistinctFrom:
+		return b.factory.ConstructIs(left, right)
+	case tree.Contains:
+		return b.factory.ConstructContains(left, right)
+	case tree.ContainedBy:
+		// This is just syntatic sugar that reverses the operands.
+		return b.factory.ConstructContains(right, left)
+	case tree.JSONExists:
+		return b.factory.ConstructJsonExists(left, right)
+	case tree.JSONAllExists:
+		return b.factory.ConstructJsonAllExists(left, right)
+	case tree.JSONSomeExists:
+		return b.factory.ConstructJsonSomeExists(left, right)
+	}
+	panic(fmt.Sprintf("unhandled comparison operator: %s", cmp))
+}
+
+func (b *Builder) constructBinary(
+	bin tree.BinaryOperator, left, right opt.ScalarExpr, typ types.T,
+) opt.ScalarExpr {
+	switch bin {
+	case tree.Bitand:
+		return b.factory.ConstructBitand(left, right)
+	case tree.Bitor:
+		return b.factory.ConstructBitor(left, right)
+	case tree.Bitxor:
+		return b.factory.ConstructBitxor(left, right)
+	case tree.Plus:
+		return b.factory.ConstructPlus(left, right)
+	case tree.Minus:
+		return b.factory.ConstructMinus(left, right)
+	case tree.Mult:
+		return b.factory.ConstructMult(left, right)
+	case tree.Div:
+		return b.factory.ConstructDiv(left, right)
+	case tree.FloorDiv:
+		return b.factory.ConstructFloorDiv(left, right)
+	case tree.Mod:
+		return b.factory.ConstructMod(left, right)
+	case tree.Pow:
+		return b.factory.ConstructPow(left, right)
+	case tree.Concat:
+		return b.factory.ConstructConcat(left, right)
+	case tree.LShift:
+		return b.factory.ConstructLShift(left, right)
+	case tree.RShift:
+		return b.factory.ConstructRShift(left, right)
+	case tree.JSONFetchText:
+		return b.factory.ConstructFetchText(left, right)
+	case tree.JSONFetchVal:
+		return b.factory.ConstructFetchVal(left, right)
+	case tree.JSONFetchValPath:
+		return b.factory.ConstructFetchValPath(left, right)
+	case tree.JSONFetchTextPath:
+		return b.factory.ConstructFetchTextPath(left, right)
+	}
+	panic(fmt.Sprintf("unhandled binary operator: %s", bin))
+}
+
+func (b *Builder) constructUnary(
+	un tree.UnaryOperator, input opt.ScalarExpr, typ types.T,
+) opt.ScalarExpr {
+	switch un {
+	case tree.UnaryMinus:
+		return b.factory.ConstructUnaryMinus(input)
+	case tree.UnaryComplement:
+		return b.factory.ConstructUnaryComplement(input)
+	}
+	panic(fmt.Sprintf("unhandled unary operator: %s", un))
+}
+
 // ScalarBuilder is a specialized variant of Builder that can be used to create
 // a scalar from a TypedExpr. This is used to build scalar expressions for
 // testing. It is also used temporarily to interface with the old planning code.
@@ -662,10 +712,10 @@ func NewScalar(
 	// Put all the columns in the current scope.
 	sb.scope.cols = make([]scopeColumn, 0, md.NumColumns())
 	for colID := opt.ColumnID(1); int(colID) <= md.NumColumns(); colID++ {
-		name := tree.Name(md.ColumnLabel(colID))
+		colMeta := md.ColumnMeta(colID)
 		sb.scope.cols = append(sb.scope.cols, scopeColumn{
-			name: name,
-			typ:  md.ColumnType(colID),
+			name: tree.Name(colMeta.Alias),
+			typ:  colMeta.Type,
 			id:   colID,
 		})
 	}
@@ -690,7 +740,7 @@ func (sb *ScalarBuilder) Build(expr tree.TypedExpr) (err error) {
 		}
 	}()
 
-	group := sb.buildScalar(expr, &sb.scope, nil, nil, nil)
-	sb.factory.Memo().SetRoot(group, memo.MinPhysPropsID)
+	scalar := sb.buildScalar(expr, &sb.scope, nil, nil, nil)
+	sb.factory.Memo().SetScalarRoot(scalar)
 	return nil
 }
