@@ -20,15 +20,19 @@ import (
 	"fmt"
 	"strings"
 
+	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/memo"
-	"github.com/cockroachdb/cockroach/pkg/sql/opt/props"
+	"github.com/cockroachdb/cockroach/pkg/sql/opt/props/physical"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 )
 
+// scopeOrdinal identifies an ordinal position with a list of scope columns.
+type scopeOrdinal int
+
 // scope is used for the build process and maintains the variables that have
 // been bound within the current scope as columnProps. Variables bound in the
 // parent scope are also visible in this scope.
@@ -52,8 +56,8 @@ type scope struct {
 	// which don't appear in cols.
 	extraCols []scopeColumn
 
-	// group is the memo.GroupID of the relational operator built with this scope.
-	group memo.GroupID
+	// expr is the SQL node built with this scope.
+	expr memo.RelExpr
 
 	// Desired number of columns for subqueries found during name resolution and
 	// type checking. This only applies to the top-level subqueries that are
@@ -64,6 +68,10 @@ type scope struct {
 	// the replaceSRF() function for more details.
 	replaceSRFs bool
 
+	// singleSRFColumn is true if this scope has a single column that comes from
+	// an SRF. The flag is used to allow renaming the column to the table alias.
+	singleSRFColumn bool
+
 	// srfs contains all the SRFs that were replaced in this scope. It will be
 	// used by the Builder to convert the input from the FROM clause to a lateral
 	// cross join between the input and a Zip of all the srfs in this slice.
@@ -81,9 +89,9 @@ type scope struct {
 
 // cteSource represents a CTE in the given query.
 type cteSource struct {
-	name  tree.AliasClause
-	cols  []scopeColumn
-	group memo.GroupID
+	name tree.AliasClause
+	cols []scopeColumn
+	expr memo.RelExpr
 
 	// used tracks if this CTE has been referenced.  We are currently limited
 	// to only having a single reference to a given CTE, so if this is set then
@@ -134,29 +142,38 @@ func (s *scope) replace() *scope {
 }
 
 // appendColumnsFromScope adds newly bound variables to this scope.
-// The groups in the new columns are reset to 0.
+// The expressions in the new columns are reset to nil.
 func (s *scope) appendColumnsFromScope(src *scope) {
 	l := len(s.cols)
 	s.cols = append(s.cols, src.cols...)
-	// We want to reset the groups, as these become pass-through columns in the
-	// new scope.
+	// We want to reset the expressions, as these become pass-through columns in
+	// the new scope.
 	for i := l; i < len(s.cols); i++ {
-		s.cols[i].group = 0
+		s.cols[i].scalar = nil
 	}
 }
 
 // appendColumns adds newly bound variables to this scope.
-// The groups in the new columns are reset to 0.
+// The expressions in the new columns are reset to nil.
 func (s *scope) appendColumns(cols []scopeColumn) {
 	l := len(s.cols)
 	s.cols = append(s.cols, cols...)
-	// We want to reset the groups, as these become pass-through columns in the
-	// new scope.
+	// We want to reset the expressions, as these become pass-through columns in
+	// the new scope.
 	for i := l; i < len(s.cols); i++ {
-		s.cols[i].group = 0
+		s.cols[i].scalar = nil
 	}
 }
 
+// appendColumn adds a newly bound variable to this scope.
+// The expression in the new column is reset to nil.
+func (s *scope) appendColumn(col *scopeColumn) {
+	s.cols = append(s.cols, *col)
+	// We want to reset the expression, as this becomes a pass-through column in
+	// the new scope.
+	s.cols[len(s.cols)-1].scalar = nil
+}
+
 // addExtraColumns adds the given columns as extra columns, ignoring any
 // duplicate columns that are already in the scope.
 func (s *scope) addExtraColumns(cols []scopeColumn) {
@@ -189,7 +206,7 @@ func (s *scope) copyOrdering(src *scope) {
 			col := *src.getColumn(ordCol.ID())
 			// We want to reset the group, as this becomes a pass-through column in
 			// the new scope.
-			col.group = 0
+			col.scalar = nil
 			s.extraCols = append(s.extraCols, col)
 		}
 	}
@@ -212,24 +229,24 @@ func (s *scope) getColumn(col opt.ColumnID) *scopeColumn {
 }
 
 // makeOrderingChoice returns an OrderingChoice that corresponds to s.ordering.
-func (s *scope) makeOrderingChoice() props.OrderingChoice {
-	var oc props.OrderingChoice
+func (s *scope) makeOrderingChoice() physical.OrderingChoice {
+	var oc physical.OrderingChoice
 	oc.FromOrdering(s.ordering)
 	return oc
 }
 
 // makePhysicalProps constructs physical properties using the columns in the
 // scope for presentation and s.ordering for required ordering.
-func (s *scope) makePhysicalProps() *props.Physical {
-	p := &props.Physical{}
+func (s *scope) makePhysicalProps() *physical.Required {
+	p := &physical.Required{}
 
 	if len(s.cols) > 0 {
-		p.Presentation = make(props.Presentation, 0, len(s.cols))
+		p.Presentation = make(physical.Presentation, 0, len(s.cols))
 		for i := range s.cols {
 			col := &s.cols[i]
 			if !col.hidden {
-				p.Presentation = append(p.Presentation, opt.LabeledColumn{
-					Label: string(col.name),
+				p.Presentation = append(p.Presentation, opt.AliasedColumn{
+					Alias: string(col.name),
 					ID:    col.id,
 				})
 			}
@@ -282,15 +299,16 @@ func (s *scope) resolveCTE(name *tree.TableName) *cteSource {
 //
 // The desired type is a suggestion, but resolveType does not throw an error if
 // the resolved type turns out to be different from desired (in contrast to
-// resolveAndRequireType, which panics with a builderError).
+// resolveAndRequireType, which panics with a builderError). If the result
+// type is types.Unknown, then resolveType will wrap the expression in a type
+// cast in order to produce the desired type.
 func (s *scope) resolveType(expr tree.Expr, desired types.T) tree.TypedExpr {
 	expr = s.walkExprTree(expr)
 	texpr, err := tree.TypeCheck(expr, s.builder.semaCtx, desired)
 	if err != nil {
 		panic(builderError{err})
 	}
-
-	return texpr
+	return s.ensureNullType(texpr, desired)
 }
 
 // resolveAndRequireType converts the given expr to a tree.TypedExpr. As part
@@ -300,14 +318,34 @@ func (s *scope) resolveType(expr tree.Expr, desired types.T) tree.TypedExpr {
 //
 // If the resolved type does not match the desired type, resolveAndRequireType
 // panics with a builderError (in contrast to resolveType, which returns the
-// typed expression with no error).
+// typed expression with no error). If the result type is types.Unknown, then
+// resolveType will wrap the expression in a type cast in order to produce the
+// desired type.
 func (s *scope) resolveAndRequireType(expr tree.Expr, desired types.T) tree.TypedExpr {
 	expr = s.walkExprTree(expr)
 	texpr, err := tree.TypeCheckAndRequire(expr, s.builder.semaCtx, desired, s.context)
 	if err != nil {
 		panic(builderError{err})
 	}
+	return s.ensureNullType(texpr, desired)
+}
 
+// ensureNullType tests the type of the given expression. If types.Unknown, then
+// ensureNullType wraps the expression in a CAST to the desired type (assuming
+// it is not types.Any). types.Unknown is a special type used for null values,
+// and can be cast to any other type.
+func (s *scope) ensureNullType(texpr tree.TypedExpr, desired types.T) tree.TypedExpr {
+	if desired != types.Any && texpr.ResolvedType() == types.Unknown {
+		// Should always be able to convert null value to any other type.
+		colType, err := coltypes.DatumTypeToColumnType(desired)
+		if err != nil {
+			panic(err)
+		}
+		texpr, err = tree.NewTypedCastExpr(texpr, colType)
+		if err != nil {
+			panic(err)
+		}
+	}
 	return texpr
 }
 
@@ -397,7 +435,6 @@ func (s *scope) findExistingCol(expr tree.TypedExpr) *scopeColumn {
 			return col
 		}
 	}
-
 	return nil
 }
 
@@ -411,7 +448,8 @@ func (s *scope) getAggregateCols() []scopeColumn {
 
 // getAggregateArgCols returns the columns in this scope corresponding
 // to arguments to aggregate functions. This call is only valid on an
-// aggInScope.
+// aggInScope. If the aggregate has a filter, the column corresponding
+// to its input will immediately follow its inputs.
 func (s *scope) getAggregateArgCols(groupingsLen int) []scopeColumn {
 	// Aggregate args are always clustered at the beginning of the column list.
 	return s.cols[:len(s.cols)-groupingsLen]
@@ -438,20 +476,22 @@ func (s *scope) findAggregate(agg aggregateInfo) *scopeColumn {
 	}
 
 	for i, a := range s.groupby.aggs {
-		// Find an existing aggregate that has the same function and the same
-		// arguments.
-		if a.def == agg.def && a.distinct == agg.distinct && len(a.args) == len(agg.args) {
-			match := true
-			for j, arg := range a.args {
-				if arg != agg.args[j] {
-					match = false
-					break
+		// Find an existing aggregate that uses the same function overload.
+		if a.def.Overload == agg.def.Overload && a.distinct == agg.distinct && a.filter == agg.filter {
+			// Now check that the arguments are identical.
+			if len(a.args) == len(agg.args) {
+				match := true
+				for j, arg := range a.args {
+					if arg != agg.args[j] {
+						match = false
+						break
+					}
+				}
+				if match {
+					// Aggregate already exists, so return information about the
+					// existing column that computes it.
+					return &s.getAggregateCols()[i]
 				}
-			}
-			if match {
-				// Aggregate already exists, so return information about the
-				// existing column that computes it.
-				return &s.getAggregateCols()[i]
 			}
 		}
 	}
@@ -495,7 +535,7 @@ func (s *scope) startAggFunc() *scope {
 // are only used in a groupings scope.
 func (s *scope) endAggFunc(cols opt.ColSet) (aggInScope, aggOutScope *scope) {
 	if !s.groupby.inAgg {
-		panic(fmt.Errorf("mismatched calls to start/end aggFunc"))
+		panic(pgerror.NewAssertionErrorf("mismatched calls to start/end aggFunc"))
 	}
 	s.groupby.inAgg = false
 
@@ -511,7 +551,7 @@ func (s *scope) endAggFunc(cols opt.ColSet) (aggInScope, aggOutScope *scope) {
 		}
 	}
 
-	panic(fmt.Errorf("aggregate function is not allowed in this context"))
+	panic(pgerror.NewAssertionErrorf("aggregate function is not allowed in this context"))
 }
 
 // startBuildingGroupingCols is called when the builder starts building the
@@ -532,7 +572,7 @@ func (s *scope) startBuildingGroupingCols() {
 // to ensure that a grouping error is not called prematurely.
 func (s *scope) endBuildingGroupingCols() {
 	if !s.groupby.buildingGroupingCols {
-		panic(fmt.Errorf("mismatched calls to start/end groupings"))
+		panic(pgerror.NewAssertionErrorf("mismatched calls to start/end groupings"))
 	}
 	s.groupby.buildingGroupingCols = false
 }
@@ -571,30 +611,39 @@ func (s *scope) FindSourceProvidingColumn(
 	// due to ambiguity. If no columns match in the current scope, we
 	// search the parent scope. If the column is not found in any of the
 	// ancestor scopes, we return an error.
+	reportBackfillError := false
 	for ; s != nil; s, allowHidden = s.parent, false {
 		for i := range s.cols {
 			col := &s.cols[i]
-			// TODO(rytaft): Do not return a match if this column is being
-			// backfilled, or the column expression being resolved is not from
-			// a selector column expression from an UPDATE/DELETE.
-			if col.name == colName {
-				if col.table.TableName == "" && !col.hidden {
-					if candidateFromAnonSource != nil {
-						moreThanOneCandidateFromAnonSource = true
-						break
-					}
-					candidateFromAnonSource = col
-				} else if !col.hidden {
-					if candidateWithPrefix != nil {
-						moreThanOneCandidateWithPrefix = true
-					}
-					candidateWithPrefix = col
-				} else if allowHidden {
-					if hiddenCandidate != nil {
-						moreThanOneHiddenCandidate = true
-					}
-					hiddenCandidate = col
+			if col.name != colName {
+				continue
+			}
+
+			// If the matching column is a mutation column, then act as if it's not
+			// present so that matches in higher scopes can be found. However, if
+			// no match is found in higher scopes, report a backfill error rather
+			// than a "not found" error.
+			if col.mutation {
+				reportBackfillError = true
+				continue
+			}
+
+			if col.table.TableName == "" && !col.hidden {
+				if candidateFromAnonSource != nil {
+					moreThanOneCandidateFromAnonSource = true
+					break
 				}
+				candidateFromAnonSource = col
+			} else if !col.hidden {
+				if candidateWithPrefix != nil {
+					moreThanOneCandidateWithPrefix = true
+				}
+				candidateWithPrefix = col
+			} else if allowHidden {
+				if hiddenCandidate != nil {
+					moreThanOneHiddenCandidate = true
+				}
+				hiddenCandidate = col
 			}
 		}
 
@@ -631,6 +680,9 @@ func (s *scope) FindSourceProvidingColumn(
 	// not cause colName to be allocated on the heap in the happy (no error) path
 	// above.
 	tmpName := colName
+	if reportBackfillError {
+		return nil, nil, -1, makeBackfillError(tmpName)
+	}
 	return nil, nil, -1, sqlbase.NewUndefinedColumnError(tree.ErrString(&tmpName))
 }
 
@@ -771,7 +823,7 @@ func (s *scope) VisitPre(expr tree.Expr) (recurse bool, newExpr tree.Expr) {
 
 	case *tree.FuncExpr:
 		if t.WindowDef != nil {
-			panic(unimplementedf("window functions are not supported"))
+			panic(unimplementedWithIssueDetailf(34251, "", "window functions are not supported"))
 		}
 
 		def, err := t.Func.Resolve(s.builder.semaCtx.SearchPath)
@@ -798,7 +850,7 @@ func (s *scope) VisitPre(expr tree.Expr) (recurse bool, newExpr tree.Expr) {
 		if sub, ok := t.Subquery.(*tree.Subquery); ok {
 			// Copy the ArrayFlatten expression so that the tree isn't mutated.
 			copy := *t
-			copy.Subquery = s.replaceSubquery(sub, false /* wrapInTuple */, 1 /* desiredColumns */)
+			copy.Subquery = s.replaceSubquery(sub, false /* wrapInTuple */, 1 /* desiredColumns */, extraColsAllowed)
 			expr = &copy
 		}
 
@@ -813,7 +865,7 @@ func (s *scope) VisitPre(expr tree.Expr) (recurse bool, newExpr tree.Expr) {
 			if sub, ok := t.Right.(*tree.Subquery); ok {
 				// Copy the Comparison expression so that the tree isn't mutated.
 				copy := *t
-				copy.Right = s.replaceSubquery(sub, true /* wrapInTuple */, -1 /* desiredColumns */)
+				copy.Right = s.replaceSubquery(sub, true /* wrapInTuple */, -1 /* desiredColumns */, noExtraColsAllowed)
 				expr = &copy
 			}
 		}
@@ -825,9 +877,9 @@ func (s *scope) VisitPre(expr tree.Expr) (recurse bool, newExpr tree.Expr) {
 		}
 
 		if t.Exists {
-			expr = s.replaceSubquery(t, true /* wrapInTuple */, -1 /* desiredColumns */)
+			expr = s.replaceSubquery(t, true /* wrapInTuple */, -1 /* desiredColumns */, noExtraColsAllowed)
 		} else {
-			expr = s.replaceSubquery(t, false /* wrapInTuple */, s.columns /* desiredColumns */)
+			expr = s.replaceSubquery(t, false /* wrapInTuple */, s.columns /* desiredColumns */, noExtraColsAllowed)
 		}
 	}
 
@@ -866,13 +918,13 @@ func (s *scope) replaceSRF(f *tree.FuncExpr, def *tree.FunctionDefinition) *srf
 	srfScope := s.push()
 	var outCol *scopeColumn
 	if len(def.ReturnLabels) == 1 {
-		outCol = s.builder.addColumn(srfScope, def.Name, typedFunc.ResolvedType(), typedFunc)
+		outCol = s.builder.addColumn(srfScope, def.Name, typedFunc)
 	}
 	out := s.builder.buildFunction(typedFunc.(*tree.FuncExpr), s, srfScope, outCol, nil)
 	srf := &srf{
 		FuncExpr: typedFunc.(*tree.FuncExpr),
 		cols:     srfScope.cols,
-		group:    out,
+		fn:       out,
 	}
 	s.srfs = append(s.srfs, srf)
 
@@ -894,10 +946,6 @@ func (s *scope) replaceSRF(f *tree.FuncExpr, def *tree.FunctionDefinition) *srf
 // aggregate references no variables). The aggOutScope.groupby.aggs slice is
 // used later by the Builder to build aggregations in the aggregation scope.
 func (s *scope) replaceAggregate(f *tree.FuncExpr, def *tree.FunctionDefinition) tree.Expr {
-	if f.Filter != nil {
-		panic(unimplementedf("aggregates with FILTER are not supported yet"))
-	}
-
 	f, def = s.replaceCount(f, def)
 
 	// We need to save and restore the previous value of the field in
@@ -909,6 +957,23 @@ func (s *scope) replaceAggregate(f *tree.FuncExpr, def *tree.FunctionDefinition)
 		tree.RejectNestedAggregates|tree.RejectWindowApplications)
 
 	expr := f.Walk(s)
+
+	// We need to do this check here to ensure that we check the usage of special
+	// functions with the right error message.
+	if f.Filter != nil {
+		func() {
+			oldProps := s.builder.semaCtx.Properties
+			defer func() { s.builder.semaCtx.Properties.Restore(oldProps) }()
+
+			s.builder.semaCtx.Properties.Require("FILTER", tree.RejectSpecial)
+			_, err := tree.TypeCheck(expr.(*tree.FuncExpr).Filter, s.builder.semaCtx, types.Any)
+			if err != nil {
+				panic(builderError{err})
+			}
+
+		}()
+	}
+
 	typedFunc, err := tree.TypeCheck(expr, s.builder.semaCtx, types.Any)
 	if err != nil {
 		panic(builderError{err})
@@ -919,14 +984,13 @@ func (s *scope) replaceAggregate(f *tree.FuncExpr, def *tree.FunctionDefinition)
 
 	f = typedFunc.(*tree.FuncExpr)
 
-	funcDef := memo.FuncOpDef{
+	private := memo.FunctionPrivate{
 		Name:       def.Name,
-		Type:       f.ResolvedType(),
 		Properties: &def.FunctionProperties,
 		Overload:   f.ResolvedOverload(),
 	}
 
-	return s.builder.buildAggregateFunction(f, funcDef, s)
+	return s.builder.buildAggregateFunction(f, &private, s)
 }
 
 // replaceCount replaces count(*) with count_rows().
@@ -948,11 +1012,31 @@ func (s *scope) replaceCount(
 
 	if strings.EqualFold(def.Name, "count") && f.Type == 0 {
 		if _, ok := vn.(tree.UnqualifiedStar); ok {
-			// Special case handling for COUNT(*). This is a special construct to
-			// count the number of rows; in this case * does NOT refer to a set of
-			// columns. A * is invalid elsewhere (and will be caught by TypeCheck()).
-			// Replace the function with COUNT_ROWS (which doesn't take any
-			// arguments).
+			if f.Filter != nil {
+				// If we have a COUNT(*) with a FILTER, we need to synthesize an input
+				// for the aggregation to be over, because otherwise we have no input
+				// to hang the AggFilter off of.
+				// Thus, we convert
+				//   COUNT(*) FILTER (WHERE foo)
+				// to
+				//   COUNT(true) FILTER (WHERE foo).
+				cpy := *f
+				e := &cpy
+				e.Exprs = tree.Exprs{tree.DBoolTrue}
+
+				newDef, err := e.Func.Resolve(s.builder.semaCtx.SearchPath)
+				if err != nil {
+					panic(builderError{err})
+				}
+
+				return e, newDef
+			}
+
+			// Special case handling for COUNT(*) with no FILTER. This is a special
+			// construct to count the number of rows; in this case * does NOT refer
+			// to a set of columns. A * is invalid elsewhere (and will be caught by
+			// TypeCheck()).  Replace the function with COUNT_ROWS (which doesn't
+			// take any arguments).
 			e := &tree.FuncExpr{
 				Func: tree.ResolvableFunctionReference{
 					FunctionReference: &tree.UnresolvedName{
@@ -981,6 +1065,11 @@ func (s *scope) replaceCount(
 	return f, def
 }
 
+const (
+	extraColsAllowed   = true
+	noExtraColsAllowed = false
+)
+
 // Replace a raw subquery node with a typed subquery. wrapInTuple specifies
 // whether the return type of the subquery should be wrapped in a tuple.
 // wrapInTuple is true for subqueries that may return multiple rows in
@@ -989,8 +1078,12 @@ func (s *scope) replaceCount(
 // subquery. Specifying -1 for desiredColumns allows the subquery to return any
 // number of columns and is used when the normal type checking machinery will
 // verify that the correct number of columns is returned.
+// If extraColsAllowed is true, extra columns built from the subquery (such as
+// columns for which orderings have been requested) will not be stripped away.
+// It is the duty of the caller to ensure that those columns are eventually
+// dealt with.
 func (s *scope) replaceSubquery(
-	sub *tree.Subquery, wrapInTuple bool, desiredColumns int,
+	sub *tree.Subquery, wrapInTuple bool, desiredColumns int, extraColsAllowed bool,
 ) *subquery {
 	if s.replaceSRFs {
 		// We need to save and restore the previous value of the replaceSRFs field in
@@ -1011,6 +1104,7 @@ func (s *scope) replaceSubquery(
 	s.builder.subquery = &subq
 
 	outScope := s.builder.buildStmt(sub.Select, s)
+	ord := outScope.ordering
 
 	// Treat the subquery result as an anonymous data source (i.e. column names
 	// are not qualified). Remove hidden columns, as they are not accessible
@@ -1022,25 +1116,25 @@ func (s *scope) replaceSubquery(
 		n := len(outScope.cols)
 		switch desiredColumns {
 		case 1:
-			panic(builderError{pgerror.NewErrorf(pgerror.CodeSyntaxError,
-				"subquery must return only one column, found %d", n)})
+			panic(pgerror.NewErrorf(pgerror.CodeSyntaxError,
+				"subquery must return only one column, found %d", n))
 		default:
-			panic(builderError{pgerror.NewErrorf(pgerror.CodeSyntaxError,
-				"subquery must return %d columns, found %d", desiredColumns, n)})
+			panic(pgerror.NewErrorf(pgerror.CodeSyntaxError,
+				"subquery must return %d columns, found %d", desiredColumns, n))
 		}
 	}
 
-	if len(outScope.extraCols) > 0 {
+	if len(outScope.extraCols) > 0 && !extraColsAllowed {
 		// We need to add a projection to remove the extra columns.
 		projScope := outScope.push()
 		projScope.appendColumnsFromScope(outScope)
-		projScope.group = s.builder.constructProject(outScope.group, projScope.cols)
+		projScope.expr = s.builder.constructProject(outScope.expr.(memo.RelExpr), projScope.cols)
 		outScope = projScope
 	}
 
 	subq.cols = outScope.cols
-	subq.group = outScope.group
-	subq.ordering = outScope.ordering
+	subq.node = outScope.expr.(memo.RelExpr)
+	subq.ordering = ord
 	return &subq
 }
 
@@ -1056,25 +1150,25 @@ var _ tree.IndexedVarContainer = &scope{}
 
 // IndexedVarEval is part of the IndexedVarContainer interface.
 func (s *scope) IndexedVarEval(idx int, ctx *tree.EvalContext) (tree.Datum, error) {
-	panic("unimplemented: scope.IndexedVarEval")
+	panic(pgerror.NewAssertionErrorf("unimplemented: scope.IndexedVarEval"))
 }
 
 // IndexedVarResolvedType is part of the IndexedVarContainer interface.
 func (s *scope) IndexedVarResolvedType(idx int) types.T {
 	if idx >= len(s.cols) {
 		if len(s.cols) == 0 {
-			panic(builderError{pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
-				"column reference @%d not allowed in this context", idx+1)})
+			panic(pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
+				"column reference @%d not allowed in this context", idx+1))
 		}
-		panic(builderError{pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
-			"invalid column ordinal: @%d", idx+1)})
+		panic(pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
+			"invalid column ordinal: @%d", idx+1))
 	}
 	return s.cols[idx].typ
 }
 
 // IndexedVarNodeFormatter is part of the IndexedVarContainer interface.
 func (s *scope) IndexedVarNodeFormatter(idx int) tree.NodeFormatter {
-	panic("unimplemented: scope.IndexedVarNodeFormatter")
+	panic(pgerror.NewAssertionErrorf("unimplemented: scope.IndexedVarNodeFormatter"))
 }
 
 // newAmbiguousColumnError returns an error with a helpful error message to be
