@@ -17,16 +17,24 @@ package optbuilder
 import (
 	"fmt"
 
+	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt"
+	"github.com/cockroachdb/cockroach/pkg/sql/opt/cat"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/memo"
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/privilege"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
-	"github.com/cockroachdb/cockroach/pkg/util/hlc"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqltelemetry"
 	"github.com/pkg/errors"
 )
 
+const (
+	excludeMutations = false
+	includeMutations = true
+)
+
 // buildDataSource builds a set of memo groups that represent the given table
 // expression. For example, if the tree.TableExpr consists of a single table,
 // the resulting set of memo groups will consist of a single group with a
@@ -43,6 +51,7 @@ func (b *Builder) buildDataSource(
 	switch source := texpr.(type) {
 	case *tree.AliasedTableExpr:
 		if source.IndexFlags != nil {
+			telemetry.Inc(sqltelemetry.IndexHintUseCounter)
 			indexFlags = source.IndexFlags
 		}
 
@@ -60,16 +69,13 @@ func (b *Builder) buildDataSource(
 	case *tree.JoinTableExpr:
 		return b.buildJoin(source, inScope)
 
-	case *tree.NormalizableTableName:
-		tn, err := source.Normalize()
-		if err != nil {
-			panic(builderError{err})
-		}
+	case *tree.TableName:
+		tn := source
 
 		// CTEs take precedence over other data sources.
 		if cte := inScope.resolveCTE(tn); cte != nil {
 			if cte.used {
-				panic(builderError{fmt.Errorf("unsupported multiple use of CTE clause %q", tn)})
+				panic(unimplementedWithIssueDetailf(21084, "", "unsupported multiple use of CTE clause %q", tn))
 			}
 			cte.used = true
 
@@ -77,19 +83,22 @@ func (b *Builder) buildDataSource(
 
 			// TODO(justin): once we support mutations here, we will want to include a
 			// spool operation.
-			outScope.group = cte.group
+			outScope.expr = cte.expr
 			outScope.cols = cte.cols
 			return outScope
 		}
 
-		ds := b.resolveDataSource(tn)
+		ds, resName := b.resolveDataSource(tn, privilege.SELECT)
 		switch t := ds.(type) {
-		case opt.Table:
-			return b.buildScan(t, tn, nil /* ordinals */, indexFlags, inScope)
-		case opt.View:
+		case cat.Table:
+			tabID := b.factory.Metadata().AddTableWithAlias(t, &resName)
+			return b.buildScan(tabID, nil /* ordinals */, indexFlags, excludeMutations, inScope)
+		case cat.View:
 			return b.buildView(t, inScope)
+		case cat.Sequence:
+			return b.buildSequenceSelect(t, inScope)
 		default:
-			panic(unimplementedf("sequences are not supported"))
+			panic(pgerror.NewAssertionErrorf("unknown DataSource type %T", ds))
 		}
 
 	case *tree.ParenTableExpr:
@@ -111,29 +120,33 @@ func (b *Builder) buildDataSource(
 
 	case *tree.StatementSource:
 		outScope = b.buildStmt(source.Statement, inScope)
+		if len(outScope.cols) == 0 {
+			panic(pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
+				"statement source \"%v\" does not return any columns", source.Statement))
+		}
 		return outScope
 
 	case *tree.TableRef:
-		ds := b.resolveDataSourceRef(source)
+		ds := b.resolveDataSourceRef(source, privilege.SELECT)
 		switch t := ds.(type) {
-		case opt.Table:
+		case cat.Table:
 			outScope = b.buildScanFromTableRef(t, source, indexFlags, inScope)
 		default:
-			panic(unimplementedf("view and sequence numeric refs are not supported"))
+			panic(unimplementedWithIssueDetailf(35708, fmt.Sprintf("%T", t), "view and sequence numeric refs are not supported"))
 		}
 		b.renameSource(source.As, outScope)
 		return outScope
 
 	default:
-		panic(builderError{fmt.Errorf("unknown table expr: %T", texpr)})
+		panic(pgerror.NewAssertionErrorf("unknown table expr: %T", texpr))
 	}
 }
 
 // buildView parses the view query text and builds it as a Select expression.
-func (b *Builder) buildView(view opt.View, inScope *scope) (outScope *scope) {
+func (b *Builder) buildView(view cat.View, inScope *scope) (outScope *scope) {
 	// Cache the AST so that multiple references won't need to reparse.
 	if b.views == nil {
-		b.views = make(map[opt.View]*tree.Select)
+		b.views = make(map[cat.View]*tree.Select)
 	}
 
 	// Check whether view has already been parsed, and if not, parse now.
@@ -145,12 +158,15 @@ func (b *Builder) buildView(view opt.View, inScope *scope) (outScope *scope) {
 			panic(builderError{wrapped})
 		}
 
-		sel, ok = stmt.(*tree.Select)
+		sel, ok = stmt.AST.(*tree.Select)
 		if !ok {
-			panic("expected SELECT statement")
+			panic(pgerror.NewAssertionErrorf("expected SELECT statement"))
 		}
 
 		b.views[view] = sel
+
+		// Keep track of referenced views for EXPLAIN (opt, env).
+		b.factory.Metadata().AddView(view)
 	}
 
 	// When building the view, we don't want to check for the SELECT privilege
@@ -163,7 +179,7 @@ func (b *Builder) buildView(view opt.View, inScope *scope) (outScope *scope) {
 		defer func() { b.skipSelectPrivilegeChecks = false }()
 	}
 
-	outScope = b.buildSelect(sel, &scope{builder: b})
+	outScope = b.buildSelect(sel, nil /* desiredTypes */, &scope{builder: b})
 
 	// Update data source name to be the name of the view. And if view columns
 	// are specified, then update names of output columns.
@@ -180,46 +196,54 @@ func (b *Builder) buildView(view opt.View, inScope *scope) (outScope *scope) {
 
 // renameSource applies an AS clause to the columns in scope.
 func (b *Builder) renameSource(as tree.AliasClause, scope *scope) {
-	var tableAlias tree.TableName
-	colAlias := as.Cols
-
 	if as.Alias != "" {
+		colAlias := as.Cols
+
 		// Special case for Postgres compatibility: if a data source does not
 		// currently have a name, and it is a set-generating function or a scalar
 		// function with just one column, and the AS clause doesn't specify column
 		// names, then use the specified table name both as the column name and
 		// table name.
 		noColNameSpecified := len(colAlias) == 0
-		if scope.isAnonymousTable() && noColNameSpecified {
-			// SRFs and scalar functions used as a data source are always wrapped in
-			// a Zip operation.
-			ev := memo.MakeNormExprView(b.factory.Memo(), scope.group)
-			if ev.Operator() == opt.ZipOp && ev.Logical().Relational.OutputCols.Len() == 1 {
-				colAlias = tree.NameList{as.Alias}
-			}
+		if scope.isAnonymousTable() && noColNameSpecified && scope.singleSRFColumn {
+			colAlias = tree.NameList{as.Alias}
 		}
 
 		// If an alias was specified, use that to qualify the column names.
-		tableAlias = tree.MakeUnqualifiedTableName(as.Alias)
+		tableAlias := tree.MakeUnqualifiedTableName(as.Alias)
 		scope.setTableAlias(as.Alias)
-	}
 
-	if len(colAlias) > 0 {
-		// The column aliases can only refer to explicit columns.
-		for colIdx, aliasIdx := 0, 0; aliasIdx < len(colAlias); colIdx++ {
-			if colIdx >= len(scope.cols) {
-				srcName := tree.ErrString(&tableAlias)
-				panic(builderError{pgerror.NewErrorf(
-					pgerror.CodeInvalidColumnReferenceError,
-					"source %q has %d columns available but %d columns specified",
-					srcName, aliasIdx, len(colAlias),
-				)})
+		// If input expression is a ScanExpr, then override metadata aliases for
+		// pretty-printing.
+		scan, isScan := scope.expr.(*memo.ScanExpr)
+		if isScan {
+			tabMeta := b.factory.Metadata().TableMeta(scan.ScanPrivate.Table)
+			tabMeta.Alias = tree.MakeUnqualifiedTableName(as.Alias)
+		}
+
+		if len(colAlias) > 0 {
+			// The column aliases can only refer to explicit columns.
+			for colIdx, aliasIdx := 0, 0; aliasIdx < len(colAlias); colIdx++ {
+				if colIdx >= len(scope.cols) {
+					srcName := tree.ErrString(&tableAlias)
+					panic(pgerror.NewErrorf(
+						pgerror.CodeInvalidColumnReferenceError,
+						"source %q has %d columns available but %d columns specified",
+						srcName, aliasIdx, len(colAlias),
+					))
+				}
+				col := &scope.cols[colIdx]
+				if col.hidden {
+					continue
+				}
+				col.name = colAlias[aliasIdx]
+				if isScan {
+					// Override column metadata alias.
+					colMeta := b.factory.Metadata().ColumnMeta(col.id)
+					colMeta.Alias = string(colAlias[aliasIdx])
+				}
+				aliasIdx++
 			}
-			if scope.cols[colIdx].hidden {
-				continue
-			}
-			scope.cols[colIdx].name = colAlias[aliasIdx]
-			aliasIdx++
 		}
 	}
 }
@@ -232,11 +256,11 @@ func (b *Builder) renameSource(as tree.AliasClause, scope *scope) {
 // Note, the query SELECT * FROM [53() as t] is unsupported. Column lists must
 // be non-empty
 func (b *Builder) buildScanFromTableRef(
-	tab opt.Table, ref *tree.TableRef, indexFlags *tree.IndexFlags, inScope *scope,
+	tab cat.Table, ref *tree.TableRef, indexFlags *tree.IndexFlags, inScope *scope,
 ) (outScope *scope) {
 	if ref.Columns != nil && len(ref.Columns) == 0 {
-		panic(builderError{pgerror.NewErrorf(pgerror.CodeSyntaxError,
-			"an explicit list of column IDs must include at least one column")})
+		panic(pgerror.NewErrorf(pgerror.CodeSyntaxError,
+			"an explicit list of column IDs must include at least one column"))
 	}
 
 	// See tree.TableRef: "Note that a nil [Columns] array means 'unspecified'
@@ -248,37 +272,62 @@ func (b *Builder) buildScanFromTableRef(
 	if ref.Columns != nil {
 		ordinals = make([]int, len(ref.Columns))
 		for i, c := range ref.Columns {
-			ord, err := tab.LookupColumnOrdinal(uint32(c))
-			if err != nil {
-				panic(builderError{err})
+			ord := 0
+			cnt := tab.ColumnCount()
+			for ord < cnt {
+				if tab.Column(ord).ColID() == cat.StableID(c) {
+					break
+				}
+				ord++
+			}
+			if ord >= cnt {
+				panic(pgerror.NewErrorf(pgerror.CodeUndefinedColumnError,
+					"column [%d] does not exist", c))
 			}
 			ordinals[i] = ord
 		}
 	}
-	return b.buildScan(tab, tab.Name(), ordinals, indexFlags, inScope)
+
+	tabID := b.factory.Metadata().AddTable(tab)
+	return b.buildScan(tabID, ordinals, indexFlags, excludeMutations, inScope)
 }
 
 // buildScan builds a memo group for a ScanOp or VirtualScanOp expression on the
-// given table with the given table name. If the ordinals slice is not nil, then
-// only columns with ordinals in that list are projected by the scan. Otherwise,
-// all columns from the table are projected.
+// given table with the given table name. Note that the table name is passed
+// separately in order to preserve knowledge of whether the catalog and schema
+// names were explicitly specified.
 //
-// See Builder.buildStmt for a description of the remaining input and
-// return values.
+// If the ordinals slice is not nil, then only columns with ordinals in that
+// list are projected by the scan. Otherwise, all columns from the table are
+// projected.
+//
+// See Builder.buildStmt for a description of the remaining input and return
+// values.
 func (b *Builder) buildScan(
-	tab opt.Table, tn *tree.TableName, ordinals []int, indexFlags *tree.IndexFlags, inScope *scope,
+	tabID opt.TableID,
+	ordinals []int,
+	indexFlags *tree.IndexFlags,
+	scanMutationCols bool,
+	inScope *scope,
 ) (outScope *scope) {
 	md := b.factory.Metadata()
-	tabID := md.AddTable(tab)
+	tabMeta := md.TableMeta(tabID)
+	tab := tabMeta.Table
 
 	colCount := len(ordinals)
 	if colCount == 0 {
-		colCount = tab.ColumnCount()
+		// If scanning mutation columns, then include writable and deletable
+		// columns in the output, in addition to public columns.
+		if scanMutationCols {
+			colCount = tab.DeletableColumnCount()
+		} else {
+			colCount = tab.ColumnCount()
+		}
 	}
 
 	var tabColIDs opt.ColSet
 	outScope = inScope.push()
-	outScope.cols = make([]scopeColumn, colCount)
+	outScope.cols = make([]scopeColumn, 0, colCount)
 	for i := 0; i < colCount; i++ {
 		ord := i
 		if ordinals != nil {
@@ -289,31 +338,34 @@ func (b *Builder) buildScan(
 		colID := tabID.ColumnID(ord)
 		tabColIDs.Add(int(colID))
 		name := col.ColName()
-		outScope.cols[i] = scopeColumn{
-			id:     colID,
-			name:   name,
-			table:  *tn,
-			typ:    col.DatumType(),
-			hidden: col.IsHidden(),
-		}
+		isMutation := cat.IsMutationColumn(tab, ord)
+		outScope.cols = append(outScope.cols, scopeColumn{
+			id:       colID,
+			name:     name,
+			table:    tabMeta.Alias,
+			typ:      col.DatumType(),
+			hidden:   col.IsHidden() || isMutation,
+			mutation: isMutation,
+		})
 	}
 
 	if tab.IsVirtualTable() {
 		if indexFlags != nil {
-			panic(builderError{errors.Errorf("index flags not allowed with virtual tables")})
+			panic(pgerror.NewErrorf(pgerror.CodeSyntaxError,
+				"index flags not allowed with virtual tables"))
 		}
-		def := memo.VirtualScanOpDef{Table: tabID, Cols: tabColIDs}
-		outScope.group = b.factory.ConstructVirtualScan(b.factory.InternVirtualScanOpDef(&def))
+		private := memo.VirtualScanPrivate{Table: tabID, Cols: tabColIDs}
+		outScope.expr = b.factory.ConstructVirtualScan(&private)
 	} else {
-		def := memo.ScanOpDef{Table: tabID, Cols: tabColIDs}
+		private := memo.ScanPrivate{Table: tabID, Cols: tabColIDs}
 
 		if indexFlags != nil {
-			def.Flags.NoIndexJoin = indexFlags.NoIndexJoin
+			private.Flags.NoIndexJoin = indexFlags.NoIndexJoin
 			if indexFlags.Index != "" || indexFlags.IndexID != 0 {
 				idx := -1
 				for i := 0; i < tab.IndexCount(); i++ {
-					if tab.Index(i).IdxName() == string(indexFlags.Index) ||
-						tab.Index(i).InternalID() == uint64(indexFlags.IndexID) {
+					if tab.Index(i).Name() == tree.Name(indexFlags.Index) ||
+						tab.Index(i).ID() == cat.StableID(indexFlags.IndexID) {
 						idx = i
 						break
 					}
@@ -327,17 +379,47 @@ func (b *Builder) buildScan(
 					}
 					panic(builderError{err})
 				}
-				def.Flags.ForceIndex = true
-				def.Flags.Index = idx
-				def.Flags.Direction = indexFlags.Direction
+				private.Flags.ForceIndex = true
+				private.Flags.Index = idx
+				private.Flags.Direction = indexFlags.Direction
 			}
 		}
 
-		outScope.group = b.factory.ConstructScan(b.factory.InternScanOpDef(&def))
+		outScope.expr = b.factory.ConstructScan(&private)
 	}
 	return outScope
 }
 
+func (b *Builder) buildSequenceSelect(seq cat.Sequence, inScope *scope) (outScope *scope) {
+	tn := seq.SequenceName()
+	md := b.factory.Metadata()
+	outScope = inScope.push()
+
+	cols := opt.ColList{
+		md.AddColumn("last_value", types.Int),
+		md.AddColumn("log_cnt", types.Int),
+		md.AddColumn("is_called", types.Bool),
+	}
+
+	outScope.cols = make([]scopeColumn, 3)
+	for i, c := range cols {
+		col := md.ColumnMeta(c)
+		outScope.cols[i] = scopeColumn{
+			id:    c,
+			name:  tree.Name(col.Alias),
+			table: *tn,
+			typ:   col.Type,
+		}
+	}
+
+	private := memo.SequenceSelectPrivate{
+		Sequence: md.AddSequence(seq),
+		Cols:     cols,
+	}
+	outScope.expr = b.factory.ConstructSequenceSelect(&private)
+	return outScope
+}
+
 // buildWithOrdinality builds a group which appends an increasing integer column to
 // the output. colName optionally denotes the name this column is given, or can
 // be blank for none.
@@ -345,18 +427,16 @@ func (b *Builder) buildScan(
 // See Builder.buildStmt for a description of the remaining input and
 // return values.
 func (b *Builder) buildWithOrdinality(colName string, inScope *scope) (outScope *scope) {
-	col := b.synthesizeColumn(inScope, colName, types.Int, nil, 0)
+	col := b.synthesizeColumn(inScope, colName, types.Int, nil, nil /* scalar */)
 
 	// See https://www.cockroachlabs.com/docs/stable/query-order.html#order-preservation
 	// for the semantics around WITH ORDINALITY and ordering.
 
-	inScope.group = b.factory.ConstructRowNumber(
-		inScope.group,
-		b.factory.InternRowNumberDef(&memo.RowNumberDef{
-			Ordering: inScope.makeOrderingChoice(),
-			ColID:    col.id,
-		}),
-	)
+	input := inScope.expr.(memo.RelExpr)
+	inScope.expr = b.factory.ConstructRowNumber(input, &memo.RowNumberPrivate{
+		Ordering: inScope.makeOrderingChoice(),
+		ColID:    col.id,
+	})
 
 	return inScope
 }
@@ -395,22 +475,70 @@ func (b *Builder) buildCTE(ctes []*tree.CTE, inScope *scope) (outScope *scope) {
 				cols[j].table = tableName
 			}
 		}
+
+		if len(cols) == 0 {
+			panic(pgerror.NewErrorf(pgerror.CodeFeatureNotSupportedError,
+				"WITH clause %q does not have a RETURNING clause", tree.ErrString(&name)))
+		}
+
 		outScope.ctes[ctes[i].Name.Alias.String()] = &cteSource{
-			name:  ctes[i].Name,
-			cols:  cols,
-			group: cteScope.group,
+			name: ctes[i].Name,
+			cols: cols,
+			expr: cteScope.expr,
 		}
 	}
 
+	telemetry.Inc(sqltelemetry.CteUseCounter)
+
 	return outScope
 }
 
-// buildSelect builds a set of memo groups that represent the given select
+// checkCTEUsage ensures that a CTE that contains a mutation (like INSERT) is
+// used at least once by the query. Otherwise, it might not be executed.
+func (b *Builder) checkCTEUsage(inScope *scope) {
+	for alias, source := range inScope.ctes {
+		if !source.used && source.expr.Relational().CanMutate {
+			panic(pgerror.UnimplementedWithIssueErrorf(24307,
+				"common table expression %q with side effects was not used in query", alias))
+		}
+	}
+}
+
+// buildSelectStmt builds a set of memo groups that represent the given select
 // statement.
 //
 // See Builder.buildStmt for a description of the remaining input and
 // return values.
-func (b *Builder) buildSelect(stmt *tree.Select, inScope *scope) (outScope *scope) {
+func (b *Builder) buildSelectStmt(
+	stmt tree.SelectStatement, desiredTypes []types.T, inScope *scope,
+) (outScope *scope) {
+	// NB: The case statements are sorted lexicographically.
+	switch stmt := stmt.(type) {
+	case *tree.ParenSelect:
+		return b.buildSelect(stmt.Select, desiredTypes, inScope)
+
+	case *tree.SelectClause:
+		return b.buildSelectClause(stmt, nil /* orderBy */, desiredTypes, inScope)
+
+	case *tree.UnionClause:
+		return b.buildUnion(stmt, desiredTypes, inScope)
+
+	case *tree.ValuesClause:
+		return b.buildValuesClause(stmt, desiredTypes, inScope)
+
+	default:
+		panic(pgerror.NewAssertionErrorf("unknown select statement type: %T", stmt))
+	}
+}
+
+// buildSelect builds a set of memo groups that represent the given select
+// expression.
+//
+// See Builder.buildStmt for a description of the remaining input and
+// return values.
+func (b *Builder) buildSelect(
+	stmt *tree.Select, desiredTypes []types.T, inScope *scope,
+) (outScope *scope) {
 	wrapped := stmt.Select
 	orderBy := stmt.OrderBy
 	limit := stmt.Limit
@@ -423,25 +551,24 @@ func (b *Builder) buildSelect(stmt *tree.Select, inScope *scope) (outScope *scop
 				// (WITH ... (WITH ...))
 				// Currently we are unable to nest the scopes inside ParenSelect so we
 				// must refuse the syntax so that the query does not get invalid results.
-				panic(builderError{pgerror.UnimplementedWithIssueError(24303,
-					"multiple WITH clauses in parentheses")})
+				panic(pgerror.UnimplementedWithIssueError(24303, "multiple WITH clauses in parentheses"))
 			}
 			with = s.Select.With
 		}
 		wrapped = stmt.Select
 		if stmt.OrderBy != nil {
 			if orderBy != nil {
-				panic(builderError{pgerror.NewErrorf(
+				panic(pgerror.NewErrorf(
 					pgerror.CodeSyntaxError, "multiple ORDER BY clauses not allowed",
-				)})
+				))
 			}
 			orderBy = stmt.OrderBy
 		}
 		if stmt.Limit != nil {
 			if limit != nil {
-				panic(builderError{pgerror.NewErrorf(
+				panic(pgerror.NewErrorf(
 					pgerror.CodeSyntaxError, "multiple LIMIT clauses not allowed",
-				)})
+				))
 			}
 			limit = stmt.Limit
 		}
@@ -449,21 +576,23 @@ func (b *Builder) buildSelect(stmt *tree.Select, inScope *scope) (outScope *scop
 
 	if with != nil {
 		inScope = b.buildCTE(with.CTEList, inScope)
+		defer b.checkCTEUsage(inScope)
 	}
 
 	// NB: The case statements are sorted lexicographically.
 	switch t := stmt.Select.(type) {
 	case *tree.SelectClause:
-		outScope = b.buildSelectClause(t, orderBy, inScope)
+		outScope = b.buildSelectClause(t, orderBy, desiredTypes, inScope)
 
 	case *tree.UnionClause:
-		outScope = b.buildUnion(t, inScope)
+		outScope = b.buildUnion(t, desiredTypes, inScope)
 
 	case *tree.ValuesClause:
-		outScope = b.buildValuesClause(t, inScope)
+		outScope = b.buildValuesClause(t, desiredTypes, inScope)
 
 	default:
-		panic(fmt.Errorf("unknown select statement: %T", stmt.Select))
+		panic(pgerror.NewErrorf(pgerror.CodeFeatureNotSupportedError,
+			"unknown select statement: %T", stmt.Select))
 	}
 
 	if outScope.ordering.Empty() && orderBy != nil {
@@ -471,7 +600,7 @@ func (b *Builder) buildSelect(stmt *tree.Select, inScope *scope) (outScope *scop
 		projectionsScope.cols = make([]scopeColumn, 0, len(outScope.cols))
 		for i := range outScope.cols {
 			expr := &outScope.cols[i]
-			col := b.addColumn(projectionsScope, "" /* label */, expr.ResolvedType(), expr)
+			col := b.addColumn(projectionsScope, "" /* alias */, expr)
 			b.buildScalar(expr, outScope, projectionsScope, col, nil)
 		}
 		orderByScope := b.analyzeOrderBy(orderBy, outScope, projectionsScope)
@@ -497,16 +626,18 @@ func (b *Builder) buildSelect(stmt *tree.Select, inScope *scope) (outScope *scop
 // See Builder.buildStmt for a description of the remaining input and
 // return values.
 func (b *Builder) buildSelectClause(
-	sel *tree.SelectClause, orderBy tree.OrderBy, inScope *scope,
+	sel *tree.SelectClause, orderBy tree.OrderBy, desiredTypes []types.T, inScope *scope,
 ) (outScope *scope) {
-	fromScope := b.buildFrom(sel.From, sel.Where, inScope)
+	fromScope := b.buildFrom(sel.From, inScope)
+	b.buildWhere(sel.Where, fromScope)
+
 	projectionsScope := fromScope.replace()
 
 	// This is where the magic happens. When this call reaches an aggregate
 	// function that refers to variables in fromScope or an ancestor scope,
 	// buildAggregateFunction is called which adds columns to the appropriate
 	// aggInScope and aggOutScope.
-	b.analyzeProjectionList(sel.Exprs, fromScope, projectionsScope)
+	b.analyzeProjectionList(sel.Exprs, desiredTypes, fromScope, projectionsScope)
 
 	// Any aggregates in the HAVING, ORDER BY and DISTINCT ON clauses (if they
 	// exist) will be added here.
@@ -523,7 +654,7 @@ func (b *Builder) buildSelectClause(
 		b.buildOrderBy(fromScope, projectionsScope, orderByScope)
 		b.buildDistinctOnArgs(fromScope, projectionsScope, distinctOnScope)
 		if len(fromScope.srfs) > 0 {
-			fromScope.group = b.constructProjectSet(fromScope.group, fromScope.srfs)
+			fromScope.expr = b.constructProjectSet(fromScope.expr, fromScope.srfs)
 		}
 		outScope = fromScope
 	}
@@ -534,7 +665,7 @@ func (b *Builder) buildSelectClause(
 
 	if sel.Distinct {
 		if projectionsScope.distinctOnCols.Empty() {
-			outScope.group = b.constructDistinct(outScope)
+			outScope.expr = b.constructDistinct(outScope)
 		} else {
 			outScope = b.buildDistinctOn(projectionsScope.distinctOnCols, outScope)
 		}
@@ -542,12 +673,11 @@ func (b *Builder) buildSelectClause(
 	return outScope
 }
 
-// buildFrom builds a set of memo groups that represent the given FROM statement
-// and WHERE clause.
+// buildFrom builds a set of memo groups that represent the given FROM clause.
 //
-// See Builder.buildStmt for a description of the remaining input and
-// return values.
-func (b *Builder) buildFrom(from *tree.From, where *tree.Where, inScope *scope) (outScope *scope) {
+// See Builder.buildStmt for a description of the remaining input and return
+// values.
+func (b *Builder) buildFrom(from *tree.From, inScope *scope) (outScope *scope) {
 	// The root AS OF clause is recognized and handled by the executor. The only
 	// thing that must be done at this point is to ensure that if any timestamps
 	// are specified, the root SELECT was an AS OF SYSTEM TIME and that the time
@@ -559,36 +689,44 @@ func (b *Builder) buildFrom(from *tree.From, where *tree.Where, inScope *scope)
 	if len(from.Tables) > 0 {
 		outScope = b.buildFromTables(from.Tables, inScope)
 	} else {
-		rows := []memo.GroupID{b.factory.ConstructTuple(
-			b.factory.InternList(nil), b.factory.InternType(memo.EmptyTupleType),
-		)}
 		outScope = inScope.push()
-		outScope.group = b.factory.ConstructValues(
-			b.factory.InternList(rows),
-			b.factory.InternColList(opt.ColList{}),
-		)
-	}
-
-	if where != nil {
-		// We need to save and restore the previous value of the field in
-		// semaCtx in case we are recursively called within a subquery
-		// context.
-		defer b.semaCtx.Properties.Restore(b.semaCtx.Properties)
-		b.semaCtx.Properties.Require("WHERE", tree.RejectSpecial)
-		outScope.context = "WHERE"
-
-		// All "from" columns are visible to the filter expression.
-		texpr := outScope.resolveAndRequireType(where.Expr, types.Bool)
-
-		filter := b.buildScalar(texpr, outScope, nil, nil, nil)
-		// Wrap the filter in a FiltersOp.
-		filter = b.factory.ConstructFilters(b.factory.InternList([]memo.GroupID{filter}))
-		outScope.group = b.factory.ConstructSelect(outScope.group, filter)
+		outScope.expr = b.factory.ConstructValues(memo.ScalarListWithEmptyTuple, &memo.ValuesPrivate{
+			Cols: opt.ColList{},
+			ID:   b.factory.Metadata().NextValuesID(),
+		})
 	}
 
 	return outScope
 }
 
+// buildWhere builds a set of memo groups that represent the given WHERE clause.
+//
+// See Builder.buildStmt for a description of the remaining input and return
+// values.
+func (b *Builder) buildWhere(where *tree.Where, inScope *scope) {
+	if where == nil {
+		return
+	}
+
+	// We need to save and restore the previous value of the field in
+	// semaCtx in case we are recursively called within a subquery
+	// context.
+	defer b.semaCtx.Properties.Restore(b.semaCtx.Properties)
+	b.semaCtx.Properties.Require("WHERE", tree.RejectSpecial)
+	inScope.context = "WHERE"
+
+	// All "from" columns are visible to the filter expression.
+	texpr := inScope.resolveAndRequireType(where.Expr, types.Bool)
+
+	filter := b.buildScalar(texpr, inScope, nil, nil, nil)
+
+	// Wrap the filter in a FiltersOp.
+	inScope.expr = b.factory.ConstructSelect(
+		inScope.expr.(memo.RelExpr),
+		memo.FiltersExpr{{Condition: filter}},
+	)
+}
+
 // buildFromTables recursively builds a series of InnerJoin expressions that
 // join together the given FROM tables. The tables are joined in the reverse
 // order that they appear in the list, with the innermost join involving the
@@ -616,25 +754,28 @@ func (b *Builder) buildFromTables(tables tree.TableExprs, inScope *scope) (outSc
 	b.validateJoinTableNames(outScope, tableScope)
 
 	outScope.appendColumnsFromScope(tableScope)
-	outScope.group = b.factory.ConstructInnerJoin(
-		outScope.group, tableScope.group, b.factory.ConstructTrue(),
-	)
+
+	left := outScope.expr.(memo.RelExpr)
+	right := tableScope.expr.(memo.RelExpr)
+	outScope.expr = b.factory.ConstructInnerJoin(left, right, memo.TrueFilter, memo.EmptyJoinPrivate)
 	return outScope
 }
 
 // validateAsOf ensures that any AS OF SYSTEM TIME timestamp is consistent with
 // that of the root statement.
 func (b *Builder) validateAsOf(asOf tree.AsOfClause) {
-	ts, err := tree.EvalAsOfTimestamp(asOf, hlc.MaxTimestamp, b.semaCtx, b.evalCtx)
+	ts, err := tree.EvalAsOfTimestamp(asOf, b.semaCtx, b.evalCtx)
 	if err != nil {
 		panic(builderError{err})
 	}
 
 	if b.semaCtx.AsOfTimestamp == nil {
-		panic(builderError{errors.Errorf("AS OF SYSTEM TIME must be provided on a top-level statement")})
+		panic(pgerror.NewErrorf(pgerror.CodeSyntaxError,
+			"AS OF SYSTEM TIME must be provided on a top-level statement"))
 	}
 
 	if *b.semaCtx.AsOfTimestamp != ts {
-		panic(builderError{errors.Errorf("cannot specify AS OF SYSTEM TIME with different timestamps")})
+		panic(unimplementedWithIssueDetailf(35712, "",
+			"cannot specify AS OF SYSTEM TIME with different timestamps"))
 	}
 }
