@@ -19,7 +19,9 @@ import (
 	"strings"
 
 	"github.com/cockroachdb/cockroach/pkg/sql/opt"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/util"
+	"github.com/cockroachdb/cockroach/pkg/util/log"
 )
 
 // FuncDepSet is a set of functional dependencies (FDs) that encode useful
@@ -615,7 +617,7 @@ func (f *FuncDepSet) ComputeEquivClosure(cols opt.ColSet) opt.ColSet {
 // the new key is adopted in its place.
 func (f *FuncDepSet) AddStrictKey(keyCols, allCols opt.ColSet) {
 	if !keyCols.SubsetOf(allCols) {
-		panic("allCols does not include keyCols")
+		panic(pgerror.NewAssertionErrorf("allCols does not include keyCols"))
 	}
 
 	// Ensure we have candidate key (i.e. has no columns that are functionally
@@ -637,7 +639,7 @@ func (f *FuncDepSet) AddStrictKey(keyCols, allCols opt.ColSet) {
 //
 func (f *FuncDepSet) AddLaxKey(keyCols, allCols opt.ColSet) {
 	if !keyCols.SubsetOf(allCols) {
-		panic("allCols does not include keyCols")
+		panic(pgerror.NewAssertionErrorf("allCols does not include keyCols"))
 	}
 
 	// Ensure we have candidate key (i.e. has no columns that are functionally
@@ -812,7 +814,7 @@ func (f *FuncDepSet) AddConstants(cols opt.ColSet) {
 //
 func (f *FuncDepSet) AddSynthesizedCol(from opt.ColSet, col opt.ColumnID) {
 	if from.Contains(int(col)) {
-		panic("synthesized column cannot depend upon itself")
+		panic(pgerror.NewAssertionErrorf("synthesized column cannot depend upon itself"))
 	}
 
 	var colSet opt.ColSet
@@ -977,6 +979,17 @@ func (f *FuncDepSet) AddFrom(fdset *FuncDepSet) {
 	}
 }
 
+// AddEquivFrom is similar to AddFrom, except that it only adds equivalence
+// dependencies from the given set to this set.
+func (f *FuncDepSet) AddEquivFrom(fdset *FuncDepSet) {
+	for i := range fdset.deps {
+		fd := &fdset.deps[i]
+		if fd.equiv {
+			f.addDependency(fd.from, fd.to, fd.strict, fd.equiv)
+		}
+	}
+}
+
 // MakeProduct modifies the FD set to reflect the impact of a cartesian product
 // operation between this set and the given set. The result is a union of the
 // FDs from each set, as well as a union of their keys. The two FD sets are
@@ -1230,45 +1243,45 @@ func (f *FuncDepSet) Verify() {
 		fd := &f.deps[i]
 
 		if fd.from.Intersects(fd.to) {
-			panic(fmt.Sprintf("expected FD determinant and dependants to be disjoint: %s (%d)", f, i))
+			panic(pgerror.NewAssertionErrorf("expected FD determinant and dependants to be disjoint: %s (%d)", log.Safe(f), log.Safe(i)))
 		}
 
 		if fd.strict && fd.from.Empty() {
 			if i != 0 {
-				panic(fmt.Sprintf("expected strict constant FD to be first FD in set: %s (%d)", f, i))
+				panic(pgerror.NewAssertionErrorf("expected strict constant FD to be first FD in set: %s (%d)", log.Safe(f), log.Safe(i)))
 			}
 		}
 
 		if fd.equiv {
 			if !fd.strict {
-				panic(fmt.Sprintf("unexpected lax equivalency: %s (%d)", f, i))
+				panic(pgerror.NewAssertionErrorf("unexpected lax equivalency: %s (%d)", f, i))
 			}
 
 			if fd.from.Len() != 1 {
-				panic(fmt.Sprintf("expected equivalence determinant to be single col: %s (%d)", f, i))
+				panic(pgerror.NewAssertionErrorf("expected equivalence determinant to be single col: %s (%d)", log.Safe(f), log.Safe(i)))
 			}
 
 			if !f.ComputeEquivClosure(fd.from).Equals(fd.from.Union(fd.to)) {
-				panic(fmt.Sprintf("expected equivalence dependants to be its closure: %s (%d)", f, i))
+				panic(pgerror.NewAssertionErrorf("expected equivalence dependants to be its closure: %s (%d)", log.Safe(f), log.Safe(i)))
 			}
 		}
 	}
 
 	if f.hasKey != noKey {
 		if reduced := f.ReduceCols(f.key); !reduced.Equals(f.key) {
-			panic(fmt.Sprintf("expected FD to have candidate key: %s", f))
+			panic(pgerror.NewAssertionErrorf("expected FD to have candidate key: %s", f))
 		}
 
 		if f.hasKey == strictKey {
 			allCols := f.ColSet()
 			allCols.UnionWith(f.key)
 			if !f.ComputeClosure(f.key).Equals(allCols) {
-				panic(fmt.Sprintf("expected closure of FD key to include all known cols: %s", f))
+				panic(pgerror.NewAssertionErrorf("expected closure of FD key to include all known cols: %s", log.Safe(f)))
 			}
 		}
 	} else {
 		if !f.key.Empty() {
-			panic(fmt.Sprintf("expected empty key columns since no key: %s", f))
+			panic(pgerror.NewAssertionErrorf("expected empty key columns since no key: %s", f))
 		}
 	}
 }
@@ -1307,7 +1320,7 @@ func (f FuncDepSet) formatFDs(b *strings.Builder) {
 		}
 		if fd.equiv {
 			if !fd.strict {
-				panic("lax equivalent columns are not supported")
+				panic(pgerror.NewAssertionErrorf("lax equivalent columns are not supported"))
 			}
 			fmt.Fprintf(b, "%s==%s", fd.from, fd.to)
 		} else {
