@@ -15,30 +15,12 @@
 package props
 
 import (
-	"fmt"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/opt"
 	"github.com/cockroachdb/cockroach/pkg/sql/opt/constraint"
-	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/util/log"
 )
 
-// Logical properties describe the content and characteristics of data returned
-// by all expression variants within a memo group. While each expression in the
-// group may return rows or columns in a different order, or compute the result
-// using different algorithms, the same set of data is returned and can then be
-// transformed into whatever layout or presentation format that is desired,
-// according to the required physical properties.
-type Logical struct {
-	// Relational contains the set of properties that describe relational
-	// operators, like select, join, and project. It is nil for scalar
-	// operators.
-	Relational *Relational
-
-	// Scalar contains the set of properties that describe scalar operators,
-	// like And, Plus, and Const. It is nil for relational operators.
-	Scalar *Scalar
-}
-
 // AvailableRuleProps is a bit set that indicates when lazily-populated Rule
 // properties are initialized and ready for use.
 type AvailableRuleProps int
@@ -62,46 +44,153 @@ const (
 	// HasHoistableSubquery is set when the Scalar.Rule.HasHoistableSubquery
 	// is populated.
 	HasHoistableSubquery
+
+	// JoinSize is set when the Scalar.Rule.JoinSize field is populated.
+	JoinSize
 )
 
-// Relational properties are the subset of logical properties that are computed
-// for relational expressions that return rows and columns rather than scalar
-// values.
-type Relational struct {
-	// OutputCols is the set of columns that can be projected by the
-	// expression. Ordering, naming, and duplication of columns is not
-	// representable by this property; those are physical properties.
-	OutputCols opt.ColSet
-
-	// NotNullCols is the subset of output columns which cannot be NULL.
-	// The NULL-ability of columns flows from the inputs and can also be
-	// derived from filters that are NULL-intolerant.
-	NotNullCols opt.ColSet
-
+// Shared are properties that are shared by both relational and scalar
+// expressions.
+type Shared struct {
 	// OuterCols is the set of columns that are referenced by variables within
-	// this relational sub-expression, but are not bound within the scope of
-	// the expression. For example:
+	// this sub-expression, but are not bound within the scope of the expression.
+	// For example:
 	//
 	//   SELECT *
 	//   FROM a
 	//   WHERE EXISTS(SELECT * FROM b WHERE b.x = a.x AND b.y = 5)
 	//
-	// For the inner SELECT expression, a.x is an outer column, meaning that it
-	// is defined "outside" the SELECT expression (hence the name "outer"). The
-	// SELECT expression binds the b.x and b.y references, so they are not
-	// part of the outer column set. The outer SELECT binds the a.x column, and
-	// so its outer column set is empty.
+	// For the EXISTS expression, a.x is an outer column, meaning that it is
+	// defined "outside" the EXISTS expression (hence the name "outer"). The
+	// SELECT expression binds the b.x and b.y references, so they are not part
+	// of the outer column set. The outer SELECT binds the a.x column, and so
+	// its outer column set is empty.
+	//
+	// Note that what constitutes an "outer column" is dependent on an
+	// expression's location in the query. For example, while the b.x and b.y
+	// columns are not outer columns on the EXISTS expression, they *are* outer
+	// columns on the inner WHERE condition.
 	OuterCols opt.ColSet
 
-	// CanHaveSideEffects is true if the subtree rooted at this expression might
-	// trigger a run-time error, might modify outside state, or might not always
-	// return the same output given the same input. For more details, see the
-	// comment for Logical.CanHaveSideEffects.
+	// HasSubquery is true if the subtree rooted at this node contains a subquery.
+	// The subquery can be a Subquery, Exists, Any, or ArrayFlatten expression.
+	// Subqueries are the only place where a relational node can be nested within a
+	// scalar expression.
+	HasSubquery bool
+
+	// HasCorrelatedSubquery is true if the scalar expression tree contains a
+	// subquery having one or more outer columns. The subquery can be a Subquery,
+	// Exists, or Any operator. These operators usually need to be hoisted out of
+	// scalar expression trees and turned into top-level apply joins. This
+	// property makes detection fast and easy so that the hoister doesn't waste
+	// time searching subtrees that don't contain subqueries.
+	HasCorrelatedSubquery bool
+
+	// CanHaveSideEffects is true if the expression modifies state outside its
+	// own scope, or if depends upon state that may change across evaluations. An
+	// expression can have side effects if it can do any of the following:
+	//
+	//   1. Trigger a run-time error
+	//        10 / col                          -- division by zero error possible
+	//        crdb_internal.force_error('', '') -- triggers run-time error
+	//
+	//   2. Modify outside session or database state
+	//        nextval(seq)               -- modifies database sequence value
+	//        SELECT * FROM [INSERT ...] -- inserts rows into database
+	//
+	//   3. Return different results when repeatedly called with same input
+	//        ORDER BY random()      -- random can return different values
+	//        ts < clock_timestamp() -- clock_timestamp can return different vals
+	//
+	// The optimizer makes *only* the following side-effect related guarantees:
+	//
+	//   1. CASE/IF branches are only evaluated if the branch condition is true.
+	//      Therefore, the following is guaranteed to never raise a divide by
+	//      zero error, regardless of how cleverly the optimizer rewrites the
+	//      expression:
+	//
+	//        CASE WHEN divisor<>0 THEN dividend / divisor ELSE NULL END
+	//
+	//      While this example is trivial, a more complex example might have
+	//      correlated subqueries that cannot be hoisted outside the CASE
+	//      expression in the usual way, since that would trigger premature
+	//      evaluation.
+	//
+	//   2. Expressions with side effects are never treated as constant
+	//      expressions, even though they do not depend on other columns in the
+	//      query:
+	//
+	//        SELECT * FROM xy ORDER BY random()
+	//
+	//      If the random() expression were treated as a constant, then the ORDER
+	//      BY could be dropped by the optimizer, since ordering by a constant is
+	//      a no-op. Instead, the optimizer treats it like it would an expression
+	//      that depends upon a column.
+	//
+	//   3. A common table expression (CTE) with side effects will only be
+	//      evaluated one time. This will typically prevent inlining of the CTE
+	//      into the query body. For example:
+	//
+	//        WITH a AS (INSERT ... RETURNING ...) SELECT * FROM a, a
+	//
+	//      Although the "a" CTE is referenced twice, it must be evaluated only
+	//      one time (and its results cached to satisfy the second reference).
+	//
+	// As long as the optimizer provides these guarantees, it is free to rewrite,
+	// reorder, duplicate, and eliminate as if no side effects were present. As an
+	// example, the optimizer is free to eliminate the unused "nextval" column in
+	// this query:
+	//
+	//   SELECT x FROM (SELECT nextval(seq), x FROM xy)
+	//   =>
+	//   SELECT x FROM xy
+	//
+	// It's also allowed to duplicate side-effecting expressions during predicate
+	// pushdown:
+	//
+	//   SELECT * FROM xy INNER JOIN xz ON xy.x=xz.x WHERE xy.x=random()
+	//   =>
+	//   SELECT *
+	//   FROM (SELECT * FROM xy WHERE xy.x=random())
+	//   INNER JOIN (SELECT * FROM xz WHERE xz.x=random())
+	//   ON xy.x=xz.x
+	//
 	CanHaveSideEffects bool
 
+	// CanMutate is true if the subtree rooted at this expression contains at
+	// least one operator that modifies schema (like CreateTable) or writes or
+	// deletes rows (like Insert).
+	CanMutate bool
+
 	// HasPlaceholder is true if the subtree rooted at this expression contains
 	// at least one Placeholder operator.
 	HasPlaceholder bool
+}
+
+// Relational properties describe the content and characteristics of relational
+// data returned by all expression variants within a memo group. While each
+// expression in the group may return rows or columns in a different order, or
+// compute the result using different algorithms, the same set of data is
+// returned and can then be  transformed into whatever layout or presentation
+// format that is desired, according to the required physical properties.
+type Relational struct {
+	Shared
+
+	// OutputCols is the set of columns that can be projected by the expression.
+	// Ordering, naming, and duplication of columns is not representable by this
+	// property; those are physical properties.
+	OutputCols opt.ColSet
+
+	// NotNullCols is the subset of output columns which cannot be NULL. The
+	// nullability of columns flows from the inputs and can also be derived from
+	// filters that reject nulls.
+	NotNullCols opt.ColSet
+
+	// Cardinality is the number of rows that can be returned from this relational
+	// expression. The number of rows will always be between the inclusive Min and
+	// Max bounds. If Max=math.MaxUint32, then there is no limit to the number of
+	// rows returned by the expression.
+	Cardinality Cardinality
 
 	// FuncDepSet is a set of functional dependencies (FDs) that encode useful
 	// relationships between columns in a base or derived relation. Given two sets
@@ -128,12 +217,6 @@ type Relational struct {
 	// For more details, see the header comment for FuncDepSet.
 	FuncDeps FuncDepSet
 
-	// Cardinality is the number of rows that can be returned from this relational
-	// expression. The number of rows will always be between the inclusive Min and
-	// Max bounds. If Max=math.MaxUint32, then there is no limit to the number of
-	// rows returned by the expression.
-	Cardinality Cardinality
-
 	// Stats is the set of statistics that apply to this relational expression.
 	// See statistics.go and memo/statistics_builder.go for more details.
 	Stats Statistics
@@ -235,64 +318,37 @@ type Relational struct {
 		// and SimplifyRightJoinWithFilters rules. It is only valid once the
 		// Rule.Available.UnfilteredCols bit has been set.
 		UnfilteredCols opt.ColSet
+
+		// JoinSize is the number of relations being *inner* joined underneath
+		// this node. It is used to only reorder joins via AssociateJoin up to
+		// a certain limit.
+		JoinSize int
 	}
 }
 
-// Scalar properties are the subset of logical properties that are computed for
-// scalar expressions that return primitive-valued types.
+// Scalar properties are logical properties that are computed for scalar
+// expressions that return primitive-valued types. Scalar properties are
+// lazily populated on request.
 type Scalar struct {
-	// Type is the data type of the scalar expression (int, string, etc).
-	Type types.T
+	Shared
 
-	// OuterCols is the set of columns that are referenced by variables within
-	// this scalar sub-expression, but are not bound within the scope of the
-	// expression. For example:
-	//
-	//   SELECT *
-	//   FROM a
-	//   WHERE EXISTS(SELECT * FROM b WHERE b.x = a.x AND b.y = 5)
-	//
-	// For the EXISTS expression, only a.x is an outer column, meaning that
-	// only it is defined "outside" the EXISTS expression (hence the name
-	// "outer"). Note that what constitutes an "outer column" is dependent on
-	// an expression's location in the query. For example, while the b.x and
-	// b.y columns are not outer columns on the EXISTS expression, they *are*
-	// outer columns on the inner WHERE condition.
-	OuterCols opt.ColSet
-
-	// CanHaveSideEffects is true if the subtree rooted at this expression might
-	// trigger a run-time error, might modify outside state, or might not always
-	// return the same output given the same input. For more details, see the
-	// comment for Logical.CanHaveSideEffects.
-	CanHaveSideEffects bool
-
-	// HasPlaceholder is true if the subtree rooted at this expression contains
-	// at least one Placeholder operator.
-	HasPlaceholder bool
-
-	// HasCorrelatedSubquery is true if the scalar expression tree contains a
-	// subquery having one or more outer columns. The subquery can be a Subquery,
-	// Exists, or Any operator. These operators need to be hoisted out of scalar
-	// expression trees and turned into top-level apply joins. This property makes
-	// detection fast and easy so that the hoister doesn't waste time searching
-	// subtrees that don't contain subqueries.
-	HasCorrelatedSubquery bool
+	// Populated is set to true once the scalar properties have been set, usually
+	// triggered by a call to the ScalarPropsExpr.ScalarProps interface.
+	Populated bool
 
 	// Constraints is the set of constraints deduced from a boolean expression.
 	// For the expression to be true, all constraints in the set must be
 	// satisfied.
-	// This field is populated lazily, as necessary.
 	Constraints *constraint.Set
 
 	// TightConstraints is true if the expression is exactly equivalent to the
 	// constraints. If it is false, the constraints are weaker than the
 	// expression.
-	// This field is populated lazily, as necessary.
 	TightConstraints bool
 
 	// FuncDeps is a set of functional dependencies (FDs) inferred from a
 	// boolean expression. This field is only populated for Filters expressions.
-	// FDs that can be inferred from Filters expressions include:
+	//
 	//  - Constant column FDs such as ()-->(1,2) from conjuncts such as
 	//    x = 5 AND y = 10.
 	//  - Equivalent column FDs such as (1)==(2), (2)==(1) from conjuncts such
@@ -363,98 +419,18 @@ func (s *Scalar) SetAvailable(p AvailableRuleProps) {
 	s.Rule.Available |= p
 }
 
-// OuterCols is a helper method that returns either the relational or scalar
-// OuterCols field, depending on the operator's type.
-func (p *Logical) OuterCols() opt.ColSet {
-	if p.Scalar != nil {
-		return p.Scalar.OuterCols
+// Verify runs consistency checks against the shared properties, in order to
+// ensure that they conform to several invariants:
+//
+//   1. If HasCorrelatedSubquery is true, then HasSubquery must be true as well.
+//
+func (s *Shared) Verify() {
+	if s.HasCorrelatedSubquery && !s.HasSubquery {
+		panic(pgerror.NewAssertionErrorf("HasSubquery cannot be false if HasCorrelatedSubquery is true"))
 	}
-	return p.Relational.OuterCols
 }
 
-// CanHaveSideEffects is true if the expression modifies state outside its own
-// scope, or if depends upon state that may change across evaluations. An
-// expression can have side effects if it can do any of the following:
-//
-//   1. Trigger a run-time error
-//        10 / col                           -- division by zero error possible
-//        crdb_internal.force_error('', '')  -- triggers run-time error
-//
-//   2. Modify outside session or database state
-//        nextval(seq)                -- modifies database sequence value
-//        SELECT * FROM [INSERT ...]  -- inserts rows into database
-//
-//   3. Return different results when repeatedly called with same input
-//        ORDER BY random()       -- random can return different values
-//        ts < clock_timestamp()  -- clock_timestamp can return different values
-//
-// The optimizer makes *only* the following side-effect related guarantees:
-//
-//   1. CASE/IF branches are only evaluated if the branch condition is true.
-//      Therefore, the following is guaranteed to never raise a divide by zero
-//      error, regardless of how cleverly the optimizer rewrites the expression:
-//
-//        CASE WHEN divisor<>0 THEN dividend / divisor ELSE NULL END
-//
-//      While this example is trivial, a more complex example might have
-//      correlated subqueries that cannot be hoisted outside the CASE expression
-//      in the usual way, since that would trigger premature evaluation.
-//
-//   2. Expressions with side effects are never treated as constant expressions,
-//      even though they do not depend on other columns in the query:
-//
-//        SELECT * FROM xy ORDER BY random()
-//
-//      If the random() expression were treated as a constant, then the ORDER BY
-//      could be dropped by the optimizer, since ordering by a constant is a
-//      no-op. Instead, the optimizer treats it like it would an expression that
-//      depends upon a column.
-//
-//   3. A common table expression (CTE) with side effects will only be evaluated
-//      one time. This will typically prevent inlining of the CTE into the query
-//      body. For example:
-//
-//        WITH a AS (INSERT ... RETURNING ...) SELECT * FROM a, a
-//
-//      Although the "a" CTE is referenced twice, it must be evaluated only one
-//      time (and its results cached to satisfy the second reference).
-//
-// As long as the optimizer provides these guarantees, it is free to rewrite,
-// reorder, duplicate, and eliminate as if no side effects were present. As an
-// example, the optimizer is free to eliminate the unused "nextval" column in
-// this query:
-//
-//   SELECT x FROM (SELECT nextval(seq), x FROM xy)
-//   =>
-//   SELECT x FROM xy
-//
-// It's also allowed to duplicate side-effecting expressions during predicate
-// pushdown:
-//
-//   SELECT * FROM xy INNER JOIN xz ON xy.x=xz.x WHERE xy.x=random()
-//   =>
-//   SELECT *
-//   FROM (SELECT * FROM xy WHERE xy.x=random())
-//   INNER JOIN (SELECT * FROM xz WHERE xz.x=random())
-//   ON xy.x=xz.x
-//
-func (p *Logical) CanHaveSideEffects() bool {
-	if p.Scalar != nil {
-		return p.Scalar.CanHaveSideEffects
-	}
-	return p.Relational.CanHaveSideEffects
-}
-
-// HasPlaceholder is true if the subtree rooted at this expression contains
-// at least one Placeholder operator.
-func (p *Logical) HasPlaceholder() bool {
-	if p.Scalar != nil {
-		return p.Scalar.HasPlaceholder
-	}
-	return p.Relational.HasPlaceholder
-}
-
-// Verify runs consistency checks against the logical properties, in order to
+// Verify runs consistency checks against the relational properties, in order to
 // ensure that they conform to several invariants:
 //
 //   1. Functional dependencies are internally consistent.
@@ -463,32 +439,22 @@ func (p *Logical) HasPlaceholder() bool {
 //   4. If functional dependencies indicate that the relation can have at most
 //      one row, then the cardinality reflects that as well.
 //
-func (p *Logical) Verify() {
-	scalar := p.Scalar
-	if scalar != nil {
-		scalar.FuncDeps.Verify()
+func (r *Relational) Verify() {
+	r.Shared.Verify()
+	r.FuncDeps.Verify()
 
-		if p.Relational != nil {
-			panic("relational and scalar properties cannot both be set")
-		}
-		return
+	if !r.NotNullCols.SubsetOf(r.OutputCols) {
+		panic(pgerror.NewAssertionErrorf("not null cols %s not a subset of output cols %s",
+			log.Safe(r.NotNullCols), log.Safe(r.OutputCols)))
 	}
-
-	relational := p.Relational
-	relational.FuncDeps.Verify()
-
-	if !relational.NotNullCols.SubsetOf(relational.OutputCols) {
-		panic(fmt.Sprintf("not null cols %s not a subset of output cols %s",
-			relational.NotNullCols, relational.OutputCols))
+	if r.OuterCols.Intersects(r.OutputCols) {
+		panic(pgerror.NewAssertionErrorf("outer cols %s intersect output cols %s",
+			log.Safe(r.OuterCols), log.Safe(r.OutputCols)))
 	}
-	if relational.OuterCols.Intersects(relational.OutputCols) {
-		panic(fmt.Sprintf("outer cols %s intersect output cols %s",
-			relational.OuterCols, relational.OutputCols))
-	}
-	if relational.FuncDeps.HasMax1Row() {
-		if relational.Cardinality.Max > 1 {
-			panic(fmt.Sprintf(
-				"max cardinality must be <= 1 if FDs have max 1 row: %s", relational.Cardinality))
+	if r.FuncDeps.HasMax1Row() {
+		if r.Cardinality.Max > 1 {
+			panic(pgerror.NewAssertionErrorf(
+				"max cardinality must be <= 1 if FDs have max 1 row: %s", r.Cardinality))
 		}
 	}
 }
@@ -496,35 +462,38 @@ func (p *Logical) Verify() {
 // VerifyAgainst checks that the two properties don't contradict each other.
 // Used for testing (e.g. to cross-check derived properties from expressions in
 // the same group).
-func (p *Logical) VerifyAgainst(other *Logical) {
-	if r1, r2 := p.Relational, other.Relational; r1 != nil || r2 != nil {
-		if !r1.OutputCols.Equals(r2.OutputCols) {
-			panic(fmt.Sprintf("output cols mismatch: %s vs %s", r1.OutputCols, r2.OutputCols))
-		}
-
-		// NotNullCols, FuncDeps are best effort, so they might differ.
-
-		if r1.Cardinality.Max < r2.Cardinality.Min ||
-			r1.Cardinality.Min > r2.Cardinality.Max {
-			panic(fmt.Sprintf("cardinality mismatch: %s vs %s", r1.Cardinality, r2.Cardinality))
-		}
-
-		// TODO(radu): these checks might be overzealous - conceivably a
-		// subexpression with outer columns/side-effects/placeholders could be
-		// elided.
-		if !r1.OuterCols.Equals(r2.OuterCols) {
-			panic(fmt.Sprintf("outer cols mismatch: %s vs %s", r1.OuterCols, r2.OuterCols))
-		}
-		if r1.CanHaveSideEffects != r2.CanHaveSideEffects {
-			panic(fmt.Sprintf("can-have-side-effects mismatch"))
-		}
-		if r1.HasPlaceholder != r2.HasPlaceholder {
-			panic(fmt.Sprintf("has-placeholder mismatch"))
-		}
+func (r *Relational) VerifyAgainst(other *Relational) {
+	if !r.OutputCols.Equals(other.OutputCols) {
+		panic(pgerror.NewAssertionErrorf("output cols mismatch: %s vs %s", log.Safe(r.OutputCols), log.Safe(other.OutputCols)))
 	}
-	if s1, s2 := p.Scalar, other.Scalar; s1 != nil || s2 != nil {
-		// TODO(radu): implement this if necessary. Currently we won't ever have
-		// multiple expressions in a scalar group.
-		panic("unimplemented")
+
+	// NotNullCols, FuncDeps are best effort, so they might differ.
+
+	if r.Cardinality.Max < other.Cardinality.Min ||
+		r.Cardinality.Min > other.Cardinality.Max {
+		panic(pgerror.NewAssertionErrorf("cardinality mismatch: %s vs %s", log.Safe(r.Cardinality), log.Safe(other.Cardinality)))
+	}
+
+	// TODO(radu): these checks might be overzealous - conceivably a
+	// subexpression with outer columns/side-effects/placeholders could be
+	// elided.
+	if !r.OuterCols.Equals(other.OuterCols) {
+		panic(pgerror.NewAssertionErrorf("outer cols mismatch: %s vs %s", log.Safe(r.OuterCols), log.Safe(other.OuterCols)))
+	}
+	if r.CanHaveSideEffects != other.CanHaveSideEffects {
+		panic(pgerror.NewAssertionErrorf("can-have-side-effects mismatch"))
+	}
+	if r.HasPlaceholder != other.HasPlaceholder {
+		panic(pgerror.NewAssertionErrorf("has-placeholder mismatch"))
 	}
 }
+
+// Verify runs consistency checks against the relational properties, in order to
+// ensure that they conform to several invariants:
+//
+//   1. Functional dependencies are internally consistent.
+//
+func (s *Scalar) Verify() {
+	s.Shared.Verify()
+	s.FuncDeps.Verify()
+}
