@@ -14,33 +14,22 @@
 
 package testcat
 
-import (
-	"bytes"
-	"fmt"
-
-	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
-)
+import "github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 
 // CreateView creates a test view from a parsed DDL statement and adds it to the
 // catalog.
 func (tc *Catalog) CreateView(stmt *tree.CreateView) *View {
-	tn, err := stmt.Name.Normalize()
-	if err != nil {
-		panic(fmt.Errorf("%s", err))
-	}
-
 	// Update the view name to include catalog and schema if not provided.
-	tc.qualifyTableName(tn)
+	tc.qualifyTableName(&stmt.Name)
 
-	var buf bytes.Buffer
-	fmtCtx := tree.MakeFmtCtx(&buf, tree.FmtParsable)
-	stmt.AsSource.Format(&fmtCtx)
+	fmtCtx := tree.NewFmtCtx(tree.FmtParsable)
+	stmt.AsSource.Format(fmtCtx)
 
 	view := &View{
-		ViewFingerprint: tc.nextFingerprint(),
-		ViewName:        *tn,
-		QueryText:       buf.String(),
-		ColumnNames:     stmt.ColumnNames,
+		ViewID:      tc.nextStableID(),
+		ViewName:    stmt.Name,
+		QueryText:   fmtCtx.CloseAndGetString(),
+		ColumnNames: stmt.ColumnNames,
 	}
 
 	// Add the new view to the catalog.
