@@ -14,17 +14,12 @@
 
 package testcat
 
-import (
-	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
-)
+import "github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 
 // DropTable is a partial implementation of the DROP TABLE statement.
 func (tc *Catalog) DropTable(stmt *tree.DropTable) {
-	for _, ntn := range stmt.Names {
-		tn, err := ntn.Normalize()
-		if err != nil {
-			panic(err)
-		}
+	for i := range stmt.Names {
+		tn := &stmt.Names[i]
 
 		// Update the table name to include catalog and schema if not provided.
 		tc.qualifyTableName(tn)
