@@ -244,6 +244,11 @@ func (p *planner) propagateFilters(
 			return plan, extraFilter, err
 		}
 
+	case *max1RowNode:
+		if n.plan, err = p.triggerFilterPropagation(ctx, n.plan); err != nil {
+			return plan, extraFilter, err
+		}
+
 	case *windowNode:
 		if n.plan, err = p.triggerFilterPropagation(ctx, n.plan); err != nil {
 			return plan, extraFilter, err
@@ -366,12 +371,16 @@ func (p *planner) propagateFilters(
 	case *renameTableNode:
 	case *scrubNode:
 	case *truncateNode:
+	case *commentOnColumnNode:
+	case *commentOnDatabaseNode:
+	case *commentOnTableNode:
 	case *createDatabaseNode:
 	case *createIndexNode:
 	case *CreateUserNode:
 	case *createViewNode:
 	case *createSequenceNode:
 	case *createStatsNode:
+	case *deleteRangeNode:
 	case *dropDatabaseNode:
 	case *dropIndexNode:
 	case *dropTableNode:
@@ -380,12 +389,12 @@ func (p *planner) propagateFilters(
 	case *DropUserNode:
 	case *hookFnNode:
 	case *valuesNode:
+	case *virtualTableNode:
 	case *sequenceSelectNode:
 	case *setVarNode:
 	case *setClusterSettingNode:
 	case *setZoneConfigNode:
 	case *showZoneConfigNode:
-	case *showRangesNode:
 	case *showFingerprintsNode:
 	case *showTraceNode:
 	case *scatterNode:
