@@ -20,14 +20,14 @@ import (
 	"reflect"
 	"testing"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/security"
+	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
+	"github.com/pkg/errors"
 )
 
 func newPlanNode() planNode {
@@ -76,7 +76,7 @@ func (pq *ParallelizeQueue) MustAdd(t *testing.T, plan planNode, exec func() err
 		extendedEvalCtx: &p.extendedEvalCtx,
 	}
 	params.p.curPlan.plan = plan
-	if err := pq.Add(params, exec); err != nil {
+	if err := pq.Add(params, exec, p.curPlan.close /* cleanup */); err != nil {
 		t.Fatalf("ParallelizeQueue.Add failed: %v", err)
 	}
 }
@@ -325,15 +325,12 @@ func planQuery(t *testing.T, s serverutils.TestServerInterface, sql string) (*pl
 	// planner with the session so that its copy of the session data gets updated.
 	p.extendedEvalCtx.SessionData.Database = "test"
 
-	stmts, err := p.parser.Parse(sql)
+	stmt, err := parser.ParseOne(sql)
 	if err != nil {
 		t.Fatal(err)
 	}
-	if len(stmts) != 1 {
-		t.Fatalf("expected to parse 1 statement, got: %d", len(stmts))
-	}
-	stmt := stmts[0]
-	if err := p.makePlan(context.TODO(), Statement{AST: stmt}); err != nil {
+	p.stmt = &Statement{Statement: stmt}
+	if err := p.makePlan(context.TODO()); err != nil {
 		t.Fatal(err)
 	}
 	return p, func() {
