@@ -15,16 +15,14 @@
 package parser
 
 import (
-	"bytes"
 	"fmt"
 	"go/constant"
 	"go/token"
 	"strconv"
-	"strings"
 	"unicode/utf8"
+	"unsafe"
 
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
-	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 )
 
@@ -35,206 +33,68 @@ const errInvalidHexNumeric = "invalid hexadecimal numeric literal"
 const singleQuote = '\''
 const identQuote = '"'
 
-// Scanner lexes SQL statements.
-type Scanner struct {
-	in        string
-	pos       int
-	tokBuf    sqlSymType
-	lastTok   sqlSymType
-	nextTok   *sqlSymType
-	lastError *scanErr
-
-	// stmts contains the list of statements at the end of parsing.
-	stmts []tree.Statement
-
-	initialized bool
+// scanner lexes SQL statements.
+type scanner struct {
+	in            string
+	pos           int
+	bytesPrealloc []byte
 }
 
-// scanErr holds error state for a Scanner.
-type scanErr struct {
-	msg                  string
-	hint                 string
-	detail               string
-	unimplementedFeature string
-}
-
-// MakeScanner makes a Scanner from str.
-func MakeScanner(str string) Scanner {
-	var s Scanner
+func makeScanner(str string) scanner {
+	var s scanner
 	s.init(str)
 	return s
 }
 
-func (s *Scanner) init(str string) {
-	if s.initialized {
-		panic("scanner already initialized; a scanner cannot be reused.")
-	}
-	s.initialized = true
+func (s *scanner) init(str string) {
 	s.in = str
+	s.pos = 0
+	// Preallocate some buffer space for identifiers etc.
+	s.bytesPrealloc = make([]byte, len(str))
 }
 
-// Tokens calls f on all tokens of the input until an EOF is encountered.
-func (s *Scanner) Tokens(f func(token int)) {
-	for {
-		t := s.Lex(&s.tokBuf)
-		if t == 0 {
-			return
-		}
-		f(t)
+// cleanup is used to avoid holding on to memory unnecessarily (for the cases
+// where we reuse a scanner).
+func (s *scanner) cleanup() {
+	s.bytesPrealloc = nil
+}
+
+func (s *scanner) allocBytes(length int) []byte {
+	if len(s.bytesPrealloc) >= length {
+		res := s.bytesPrealloc[:length:length]
+		s.bytesPrealloc = s.bytesPrealloc[length:]
+		return res
+	}
+	return make([]byte, length)
+}
+
+// buffer returns an empty []byte buffer that can be appended to. Any unused
+// portion can be returned later using returnBuffer.
+func (s *scanner) buffer() []byte {
+	buf := s.bytesPrealloc[:0]
+	s.bytesPrealloc = nil
+	return buf
+}
+
+// returnBuffer returns the unused portion of buf to the scanner, to be used for
+// future allocBytes() or buffer() calls. The caller must not use buf again.
+func (s *scanner) returnBuffer(buf []byte) {
+	if len(buf) < cap(buf) {
+		s.bytesPrealloc = buf[len(buf):]
 	}
 }
 
-// Until returns the position of token or 0 if it is not found.
-func (s *Scanner) Until(token int) int {
-	var t int
-	for {
-		t = s.Lex(&s.tokBuf)
-		switch t {
-		case 0:
-			return 0
-		case token:
-			return s.pos
-		}
-	}
+// finishString casts the given buffer to a string and returns the unused
+// portion of the buffer. The caller must not use buf again.
+func (s *scanner) finishString(buf []byte) string {
+	str := *(*string)(unsafe.Pointer(&buf))
+	s.returnBuffer(buf)
+	return str
 }
 
-// Lex lexes a token from input.
-func (s *Scanner) Lex(lval *sqlSymType) int {
-	// The core lexing takes place in scan(). Here we do a small bit of post
-	// processing of the lexical tokens so that the grammar only requires
-	// one-token lookahead despite SQL requiring multi-token lookahead in some
-	// cases. These special cases are handled below and the returned tokens are
-	// adjusted to reflect the lookahead (LA) that occurred.
-
-	if s.nextTok != nil {
-		*lval = *s.nextTok
-		s.nextTok = nil
-	} else {
-		s.scan(lval)
-	}
-
-	switch lval.id {
-	case NOT, WITH, AS:
-	default:
-		s.lastTok = *lval
-		return lval.id
-	}
-
-	s.nextTok = &s.tokBuf
-	s.scan(s.nextTok)
-
-	// If you update these cases, update lookaheadKeywords below.
-	switch lval.id {
-	case AS:
-		switch s.nextTok.id {
-		case OF:
-			lval.id = AS_LA
-		}
-	case NOT:
-		switch s.nextTok.id {
-		case BETWEEN, IN, LIKE, ILIKE, SIMILAR:
-			lval.id = NOT_LA
-		}
-
-	case WITH:
-		switch s.nextTok.id {
-		case TIME, ORDINALITY:
-			lval.id = WITH_LA
-		}
-	}
-
-	s.lastTok = *lval
-	return lval.id
-}
-
-func (s *Scanner) initLastErr() {
-	if s.lastError == nil {
-		s.lastError = new(scanErr)
-	}
-}
-
-// Unimplemented wraps Error, setting lastUnimplementedError.
-func (s *Scanner) Unimplemented(feature string) {
-	s.Error("unimplemented")
-	s.lastError.unimplementedFeature = feature
-}
-
-// UnimplementedWithIssue wraps Error, setting lastUnimplementedError.
-func (s *Scanner) UnimplementedWithIssue(issue int) {
-	s.Error("unimplemented")
-	s.lastError.unimplementedFeature = fmt.Sprintf("#%d", issue)
-	s.lastError.hint = fmt.Sprintf("See: https://github.com/cockroachdb/cockroach/issues/%d", issue)
-}
-
-// UnimplementedWithIssueDetail wraps Error, setting lastUnimplementedError.
-func (s *Scanner) UnimplementedWithIssueDetail(issue int, detail string) {
-	s.Error("unimplemented")
-	s.lastError.unimplementedFeature = fmt.Sprintf("#%d.%s", issue, detail)
-	s.lastError.hint = fmt.Sprintf("See: https://github.com/cockroachdb/cockroach/issues/%d", issue)
-}
-
-func (s *Scanner) Error(e string) {
-	s.initLastErr()
-	if s.lastTok.id == ERROR {
-		// This is a tokenizer (lexical) error: just emit the invalid
-		// input as error.
-		s.lastError.msg = s.lastTok.str
-	} else {
-		// This is a contextual error. Print the provided error message
-		// and the error context.
-		s.lastError.msg = fmt.Sprintf("%s at or near \"%s\"", e, s.lastTok.str)
-	}
-
-	// Find the end of the line containing the last token.
-	i := strings.IndexByte(s.in[s.lastTok.pos:], '\n')
-	if i == -1 {
-		i = len(s.in)
-	} else {
-		i += s.lastTok.pos
-	}
-	// Find the beginning of the line containing the last token. Note that
-	// LastIndexByte returns -1 if '\n' could not be found.
-	j := strings.LastIndexByte(s.in[:s.lastTok.pos], '\n') + 1
-	// Output everything up to and including the line containing the last token.
-	var buf bytes.Buffer
-	fmt.Fprintf(&buf, "source SQL:\n%s\n", s.in[:i])
-	// Output a caret indicating where the last token starts.
-	fmt.Fprintf(&buf, "%s^", strings.Repeat(" ", s.lastTok.pos-j))
-	s.lastError.detail = buf.String()
-	s.lastError.unimplementedFeature = ""
-	s.lastError.hint = ""
-}
-
-// SetHelp marks the "last error" field in the Scanner to become a
-// help text. This method is invoked in the error action of the
-// parser, so the help text is only produced if the last token
-// encountered was HELPTOKEN -- other cases are just syntax errors,
-// and in that case we do not want the help text to overwrite the
-// lastError field, which was set earlier to contain details about the
-// syntax error.
-func (s *Scanner) SetHelp(msg HelpMessage) {
-	if s.lastTok.id == HELPTOKEN {
-		s.populateHelpMsg(msg.String())
-	} else {
-		s.initLastErr()
-		if msg.Command != "" {
-			s.lastError.hint = `try \h ` + msg.Command
-		} else {
-			s.lastError.hint = `try \hf ` + msg.Function
-		}
-	}
-}
-
-func (s *Scanner) populateHelpMsg(msg string) {
-	s.initLastErr()
-	s.lastError.unimplementedFeature = ""
-	s.lastError.msg = "help token in input"
-	s.lastError.hint = msg
-}
-
-func (s *Scanner) scan(lval *sqlSymType) {
+func (s *scanner) scan(lval *sqlSymType) {
 	lval.id = 0
-	lval.pos = s.pos
+	lval.pos = int32(s.pos)
 	lval.str = "EOF"
 
 	if _, ok := s.skipWhitespace(lval, true); !ok {
@@ -243,12 +103,12 @@ func (s *Scanner) scan(lval *sqlSymType) {
 
 	ch := s.next()
 	if ch == eof {
-		lval.pos = s.pos
+		lval.pos = int32(s.pos)
 		return
 	}
 
-	lval.id = ch
-	lval.pos = s.pos - 1
+	lval.id = int32(ch)
+	lval.pos = int32(s.pos - 1)
 	lval.str = s.in[lval.pos:s.pos]
 
 	switch ch {
@@ -274,10 +134,10 @@ func (s *Scanner) scan(lval *sqlSymType) {
 		}
 		return
 
-	case 'b', 'B':
+	case 'b':
 		// Bytes?
 		if s.peek() == singleQuote {
-			// [bB]'[^']'
+			// b'[^']'
 			s.pos++
 			if s.scanString(lval, singleQuote, true /* allowEscapes */, false /* requireUTF8 */) {
 				lval.id = BCONST
@@ -304,6 +164,17 @@ func (s *Scanner) scan(lval *sqlSymType) {
 		s.scanIdent(lval)
 		return
 
+	case 'B':
+		// Bit array literal?
+		if s.peek() == singleQuote {
+			// B'[01]*'
+			s.pos++
+			s.scanBitString(lval, singleQuote)
+			return
+		}
+		s.scanIdent(lval)
+		return
+
 	case 'x', 'X':
 		// Hex literal?
 		if s.peek() == singleQuote {
@@ -518,14 +389,14 @@ func (s *Scanner) scan(lval *sqlSymType) {
 	// lval for above.
 }
 
-func (s *Scanner) peek() int {
+func (s *scanner) peek() int {
 	if s.pos >= len(s.in) {
 		return eof
 	}
 	return int(s.in[s.pos])
 }
 
-func (s *Scanner) peekN(n int) int {
+func (s *scanner) peekN(n int) int {
 	pos := s.pos + n
 	if pos >= len(s.in) {
 		return eof
@@ -533,7 +404,7 @@ func (s *Scanner) peekN(n int) int {
 	return int(s.in[pos])
 }
 
-func (s *Scanner) next() int {
+func (s *scanner) next() int {
 	ch := s.peek()
 	if ch != eof {
 		s.pos++
@@ -541,7 +412,7 @@ func (s *Scanner) next() int {
 	return ch
 }
 
-func (s *Scanner) skipWhitespace(lval *sqlSymType, allowComments bool) (newline, ok bool) {
+func (s *scanner) skipWhitespace(lval *sqlSymType, allowComments bool) (newline, ok bool) {
 	newline = false
 	for {
 		ch := s.peek()
@@ -566,7 +437,7 @@ func (s *Scanner) skipWhitespace(lval *sqlSymType, allowComments bool) (newline,
 	return newline, true
 }
 
-func (s *Scanner) scanComment(lval *sqlSymType) (present, ok bool) {
+func (s *scanner) scanComment(lval *sqlSymType) (present, ok bool) {
 	start := s.pos
 	ch := s.peek()
 
@@ -599,7 +470,7 @@ func (s *Scanner) scanComment(lval *sqlSymType) (present, ok bool) {
 
 			case eof:
 				lval.id = ERROR
-				lval.pos = start
+				lval.pos = int32(start)
 				lval.str = "unterminated comment"
 				return false, false
 			}
@@ -623,7 +494,7 @@ func (s *Scanner) scanComment(lval *sqlSymType) (present, ok bool) {
 	return false, true
 }
 
-func (s *Scanner) scanIdent(lval *sqlSymType) {
+func (s *scanner) scanIdent(lval *sqlSymType) {
 	s.pos--
 	start := s.pos
 	isASCII := true
@@ -659,26 +530,24 @@ func (s *Scanner) scanIdent(lval *sqlSymType) {
 	} else if isASCII {
 		// We know that the identifier we've seen so far is ASCII, so we don't need
 		// to unicode normalize. Instead, just lowercase as normal.
-		b := make([]byte, s.pos-start)
+		b := s.allocBytes(s.pos - start)
+		_ = b[s.pos-start-1] // For bounds check elimination.
 		for i, c := range s.in[start:s.pos] {
 			if c >= 'A' && c <= 'Z' {
 				c += 'a' - 'A'
 			}
 			b[i] = byte(c)
 		}
-		lval.str = string(b)
+		lval.str = *(*string)(unsafe.Pointer(&b))
 	} else {
 		// The string has unicode in it. No choice but to run Normalize.
 		lval.str = lex.NormalizeName(s.in[start:s.pos])
 	}
-	if id, ok := lex.Keywords[lval.str]; ok {
-		lval.id = id.Tok
-	} else {
-		lval.id = IDENT
-	}
+
+	lval.id = lex.GetKeywordID(lval.str)
 }
 
-func (s *Scanner) scanNumber(lval *sqlSymType, ch int) {
+func (s *scanner) scanNumber(lval *sqlSymType, ch int) {
 	start := s.pos - 1
 	isHex := false
 	hasDecimal := ch == '.'
@@ -776,29 +645,27 @@ func (s *Scanner) scanNumber(lval *sqlSymType, ch int) {
 	}
 }
 
-func (s *Scanner) scanPlaceholder(lval *sqlSymType) {
+func (s *scanner) scanPlaceholder(lval *sqlSymType) {
 	start := s.pos
 	for lex.IsDigit(s.peek()) {
 		s.pos++
 	}
 	lval.str = s.in[start:s.pos]
 
-	uval, err := strconv.ParseUint(lval.str, 10, 64)
-	if err == nil && uval > 1<<63 {
-		err = pgerror.NewErrorf(pgerror.CodeNumericValueOutOfRangeError, "integer value out of range: %d", uval)
-	}
+	placeholder, err := tree.NewPlaceholder(lval.str)
 	if err != nil {
 		lval.id = ERROR
 		lval.str = err.Error()
 		return
 	}
-
 	lval.id = PLACEHOLDER
+	lval.union.val = placeholder
 }
 
 // scanHexString scans the content inside x'....'.
-func (s *Scanner) scanHexString(lval *sqlSymType, ch int) bool {
-	var buf []byte
+func (s *scanner) scanHexString(lval *sqlSymType, ch int) bool {
+	buf := s.buffer()
+
 	var curbyte byte
 	bytep := 0
 	const errInvalidBytesLiteral = "invalid hexadecimal bytes literal"
@@ -848,15 +715,51 @@ outer:
 	}
 
 	lval.id = BCONST
-	lval.str = string(buf)
+	lval.str = s.finishString(buf)
+	return true
+}
+
+// scanBitString scans the content inside B'....'.
+func (s *scanner) scanBitString(lval *sqlSymType, ch int) bool {
+	buf := s.buffer()
+outer:
+	for {
+		b := s.next()
+		switch b {
+		case ch:
+			newline, ok := s.skipWhitespace(lval, false)
+			if !ok {
+				return false
+			}
+			// SQL allows joining adjacent strings separated by whitespace
+			// as long as that whitespace contains at least one
+			// newline. Kind of strange to require the newline, but that
+			// is the standard.
+			if s.peek() == ch && newline {
+				s.pos++
+				continue
+			}
+			break outer
+
+		case '0', '1':
+			buf = append(buf, byte(b))
+		default:
+			lval.id = ERROR
+			lval.str = fmt.Sprintf(`"%c" is not a valid binary digit`, rune(b))
+			return false
+		}
+	}
+
+	lval.id = BITCONST
+	lval.str = s.finishString(buf)
 	return true
 }
 
 // scanString scans the content inside '...'. This is used for simple
 // string literals '...' but also e'....' and b'...'. For x'...', see
 // scanHexString().
-func (s *Scanner) scanString(lval *sqlSymType, ch int, allowEscapes, requireUTF8 bool) bool {
-	var buf []byte
+func (s *scanner) scanString(lval *sqlSymType, ch int, allowEscapes, requireUTF8 bool) bool {
+	buf := s.buffer()
 	var runeTmp [utf8.UTFMax]byte
 	start := s.pos
 
@@ -947,6 +850,66 @@ outer:
 		return false
 	}
 
-	lval.str = string(buf)
+	lval.str = s.finishString(buf)
 	return true
 }
+
+// SplitFirstStatement returns the length of the prefix of the string up to and
+// including the first semicolon that separates statements. If there is no
+// semicolon, returns ok=false.
+func SplitFirstStatement(sql string) (pos int, ok bool) {
+	s := makeScanner(sql)
+	var lval sqlSymType
+	for {
+		s.scan(&lval)
+		switch lval.id {
+		case 0, ERROR:
+			return 0, false
+		case ';':
+			return s.pos, true
+		}
+	}
+}
+
+// Tokens decomposes the input into lexical tokens.
+func Tokens(sql string) (tokens []TokenString, ok bool) {
+	s := makeScanner(sql)
+	for {
+		var lval sqlSymType
+		s.scan(&lval)
+		if lval.id == ERROR {
+			return nil, false
+		}
+		if lval.id == 0 {
+			break
+		}
+		tokens = append(tokens, TokenString{TokenID: lval.id, Str: lval.str})
+	}
+	return tokens, true
+}
+
+// TokenString is the unit value returned by Tokens.
+type TokenString struct {
+	TokenID int32
+	Str     string
+}
+
+// LastLexicalToken returns the last lexical token. If the string has no lexical
+// tokens, returns 0 and ok=false.
+func LastLexicalToken(sql string) (lastTok int, ok bool) {
+	s := makeScanner(sql)
+	var lval sqlSymType
+	for {
+		last := lval.id
+		s.scan(&lval)
+		if lval.id == 0 {
+			return int(last), last != 0
+		}
+	}
+}
+
+// EndsInSemicolon returns true if the last lexical token is a semicolon.
+func EndsInSemicolon(sql string) bool {
+	lastTok, ok := LastLexicalToken(sql)
+	return ok && lastTok == ';'
+}
