@@ -17,20 +17,20 @@ package pgwire
 import (
 	"bytes"
 	"context"
+	gosql "database/sql"
+	"fmt"
 	"io"
 	"io/ioutil"
 	"net"
+	"net/url"
 	"strconv"
 	"strings"
 	"sync"
 	"testing"
 	"time"
 
-	"github.com/jackc/pgx"
-	"github.com/pkg/errors"
-	"golang.org/x/sync/errgroup"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
+	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/sql"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgwirebase"
@@ -47,6 +47,11 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
+	"github.com/jackc/pgx"
+	"github.com/jackc/pgx/pgproto3"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/require"
+	"golang.org/x/sync/errgroup"
 )
 
 // Test the conn struct: check that it marshalls the correct commands to the
@@ -100,13 +105,15 @@ func TestConn(t *testing.T) {
 	// buffer.
 	serveCtx, stopServe := context.WithCancel(ctx)
 	g.Go(func() error {
-		return conn.serveImpl(
+		conn.serveImpl(
 			serveCtx,
 			func() bool { return false }, /* draining */
 			// sqlServer - nil means don't create a command processor and a write side of the conn
 			nil,
 			mon.BoundAccount{}, /* reserved */
+			authOptions{skipAuth: true},
 			s.Stopper())
+		return nil
 	})
 	defer stopServe()
 
@@ -202,7 +209,7 @@ func processPgxStartup(ctx context.Context, s serverutils.TestServerInterface, c
 			log.Infof(ctx, "stop wait at: %v", cmd)
 			return nil
 		}
-		query := exec.Stmt.String()
+		query := exec.AST.String()
 		if !strings.HasPrefix(query, "SELECT t.oid") {
 			log.Infof(ctx, "stop wait at query: %s", query)
 			return nil
@@ -220,7 +227,7 @@ func processPgxStartup(ctx context.Context, s serverutils.TestServerInterface, c
 func execQuery(
 	ctx context.Context, query string, s serverutils.TestServerInterface, c *conn,
 ) error {
-	rows, cols, err := s.InternalExecutor().(sqlutil.InternalExecutor).Query(
+	rows, cols, err := s.InternalExecutor().(sqlutil.InternalExecutor).QueryWithCols(
 		ctx, "test", nil /* txn */, query,
 	)
 	if err != nil {
@@ -329,7 +336,7 @@ func waitForClientConn(ln net.Listener) (*conn, error) {
 	}
 
 	metrics := makeServerMetrics(sql.MemoryMetrics{} /* sqlMemMetrics */, metric.TestSampleInterval)
-	pgwireConn := newConn(conn, sql.SessionArgs{}, &metrics, &sql.ExecutorConfig{})
+	pgwireConn := newConn(conn, sql.SessionArgs{ConnResultsBufferSize: 16 << 10}, &metrics, nil)
 	return pgwireConn, nil
 }
 
@@ -389,8 +396,8 @@ func expectExecStmt(
 		t.Fatalf("%s: expected command ExecStmt, got: %T (%+v)", testutils.Caller(1), cmd, cmd)
 	}
 
-	if es.Stmt.String() != expSQL {
-		t.Fatalf("%s: expected %s, got %s", testutils.Caller(1), expSQL, es.Stmt.String())
+	if es.AST.String() != expSQL {
+		t.Fatalf("%s: expected %s, got %s", testutils.Caller(1), expSQL, es.AST.String())
 	}
 
 	if es.ParseStart == (time.Time{}) {
@@ -428,8 +435,8 @@ func expectPrepareStmt(
 		t.Fatalf("%s: expected name %s, got %s", testutils.Caller(1), expName, pr.Name)
 	}
 
-	if pr.Stmt.String() != expSQL {
-		t.Fatalf("%s: expected %s, got %s", testutils.Caller(1), expSQL, pr.Stmt.String())
+	if pr.AST.String() != expSQL {
+		t.Fatalf("%s: expected %s, got %s", testutils.Caller(1), expSQL, pr.AST.String())
 	}
 
 	if err := finishQuery(prepare, c); err != nil {
@@ -648,38 +655,38 @@ var _ pgx.Logger = pgxTestLogger{}
 // network errors.
 func TestConnClose(t *testing.T) {
 	defer leaktest.AfterTest(t)()
-	s, db, _ := serverutils.StartServer(t, base.TestServerArgs{
-		// Make the connections' results buffers really small so that it overflows
-		// when we produce a few results.
-		ConnResultsBufferBytes: 10,
-		// Andrei is too lazy to figure out the incantation for telling pgx about
-		// our test certs.
-		Insecure: true,
-	})
+	s, db, _ := serverutils.StartServer(t, base.TestServerArgs{})
 	ctx := context.TODO()
 	defer s.Stopper().Stop(ctx)
 
-	r := sqlutils.MakeSQLRunner(db)
-	r.Exec(t, "CREATE DATABASE test")
-	r.Exec(t, "CREATE TABLE test.test AS SELECT * FROM generate_series(1,100)")
-
-	host, ports, err := net.SplitHostPort(s.ServingAddr())
+	// Disable results buffering.
+	if _, err := db.Exec(
+		`SET CLUSTER SETTING sql.defaults.results_buffer.size = '0'`,
+	); err != nil {
+		t.Fatal(err)
+	}
+	pgURL, cleanupFunc := sqlutils.PGUrl(
+		t, s.ServingAddr(), "testConnClose" /* prefix */, url.User(security.RootUser),
+	)
+	defer cleanupFunc()
+	noBufferDB, err := gosql.Open("postgres", pgURL.String())
 	if err != nil {
 		t.Fatal(err)
 	}
-	port, err := strconv.Atoi(ports)
+	defer noBufferDB.Close()
+
+	r := sqlutils.MakeSQLRunner(noBufferDB)
+	r.Exec(t, "CREATE DATABASE test")
+	r.Exec(t, "CREATE TABLE test.test AS SELECT * FROM generate_series(1,100)")
+
+	pgxConfig, err := pgx.ParseConnectionString(pgURL.String())
 	if err != nil {
 		t.Fatal(err)
 	}
 	// We test both with and without DistSQL, as the way that network errors are
 	// observed depends on the engine.
 	testutils.RunTrueAndFalse(t, "useDistSQL", func(t *testing.T, useDistSQL bool) {
-		conn, err := pgx.Connect(pgx.ConnConfig{
-			Host:     host,
-			Port:     uint16(port),
-			User:     "root",
-			Database: "system",
-		})
+		conn, err := pgx.Connect(pgxConfig)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -764,14 +771,20 @@ func TestMaliciousInputs(t *testing.T) {
 			sqlMetrics := sql.MakeMemMetrics("test" /* endpoint */, time.Second /* histogramWindow */)
 			metrics := makeServerMetrics(sqlMetrics, time.Second /* histogramWindow */)
 
-			conn := newConn(r, sql.SessionArgs{}, &metrics, nil /* execCfg */)
+			conn := newConn(
+				// ConnResultsBufferBytes - really small so that it overflows
+				// when we produce a few results.
+				r, sql.SessionArgs{ConnResultsBufferSize: 10}, &metrics,
+				nil,
+			)
 			// Ignore the error from serveImpl. There might be one when the client
 			// sends malformed input.
-			_ /* err */ = conn.serveImpl(
+			conn.serveImpl(
 				ctx,
 				func() bool { return false }, /* draining */
-				nil,                /* sqlServer */
-				mon.BoundAccount{}, /* reserved */
+				nil,                          /* sqlServer */
+				mon.BoundAccount{},           /* reserved */
+				authOptions{skipAuth: true},
 				stopper,
 			)
 			if err := <-errChan; err != nil {
@@ -810,7 +823,7 @@ func TestReadTimeoutConnExits(t *testing.T) {
 			}
 			defer c.Close()
 
-			readTimeoutConn := newReadTimeoutConn(c, ctx.Err)
+			readTimeoutConn := newReadTimeoutConn(c, func(_ *readTimeoutConn) error { return ctx.Err() })
 			// Assert that reads are performed normally.
 			readBytes := make([]byte, len(expectedRead))
 			if _, err := readTimeoutConn.Read(readBytes); err != nil {
@@ -847,3 +860,122 @@ func TestReadTimeoutConnExits(t *testing.T) {
 		t.Fatalf("unexpected error: %v", err)
 	}
 }
+
+func TestConnResultsBufferSize(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	s, db, _ := serverutils.StartServer(t, base.TestServerArgs{})
+	defer s.Stopper().Stop(context.Background())
+
+	// Check that SHOW results_buffer_size correctly exposes the value when it
+	// inherits the default.
+	{
+		var size string
+		require.NoError(t, db.QueryRow(`SHOW results_buffer_size`).Scan(&size))
+		require.Equal(t, `16384`, size)
+	}
+
+	pgURL, cleanup := sqlutils.PGUrl(t, s.ServingAddr(), t.Name(), url.User(security.RootUser))
+	defer cleanup()
+	q := pgURL.Query()
+
+	q.Add(`results_buffer_size`, `foo`)
+	pgURL.RawQuery = q.Encode()
+	{
+		errDB, err := gosql.Open("postgres", pgURL.String())
+		require.NoError(t, err)
+		defer errDB.Close()
+		_, err = errDB.Exec(`SELECT 1`)
+		require.EqualError(t, err,
+			`pq: error parsing results_buffer_size option value 'foo' as bytes`)
+	}
+
+	q.Del(`results_buffer_size`)
+	q.Add(`results_buffer_size`, `-1`)
+	pgURL.RawQuery = q.Encode()
+	{
+		errDB, err := gosql.Open("postgres", pgURL.String())
+		require.NoError(t, err)
+		defer errDB.Close()
+		_, err = errDB.Exec(`SELECT 1`)
+		require.EqualError(t, err, `pq: results_buffer_size option value '-1' cannot be negative`)
+	}
+
+	// Set the results_buffer_size to a very small value, eliminating buffering.
+	q.Del(`results_buffer_size`)
+	q.Add(`results_buffer_size`, `2`)
+	pgURL.RawQuery = q.Encode()
+
+	noBufferDB, err := gosql.Open("postgres", pgURL.String())
+	require.NoError(t, err)
+	defer noBufferDB.Close()
+
+	var size string
+	require.NoError(t, noBufferDB.QueryRow(`SHOW results_buffer_size`).Scan(&size))
+	require.Equal(t, `2`, size)
+
+	// Run a query that immediately returns one result and then pauses for a
+	// long time while computing the second.
+	rows, err := noBufferDB.Query(
+		`SELECT a, if(a = 1, pg_sleep(99999), false) from (VALUES (0), (1)) AS foo (a)`)
+	require.NoError(t, err)
+
+	// Verify that the first result has been flushed.
+	require.True(t, rows.Next())
+	var a int
+	var b bool
+	require.NoError(t, rows.Scan(&a, &b))
+	require.Equal(t, 0, a)
+	require.False(t, b)
+}
+
+// Test that closing a connection while authentication was ongoing cancels the
+// auhentication process. In other words, this checks that the server is reading
+// from the connection while authentication is ongoing and so it reacts to the
+// connection closing.
+func TestConnCloseCancelsAuth(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	authBlocked := make(chan struct{})
+	s, _, _ := serverutils.StartServer(t,
+		base.TestServerArgs{
+			Insecure: true,
+			Knobs: base.TestingKnobs{
+				PGWireTestingKnobs: &sql.PGWireTestingKnobs{
+					AuthHook: func(ctx context.Context) error {
+						// Notify the test.
+						authBlocked <- struct{}{}
+						// Wait for context cancelation.
+						<-ctx.Done()
+						// Notify the test.
+						close(authBlocked)
+						return fmt.Errorf("test auth canceled")
+					},
+				},
+			},
+		})
+	ctx := context.Background()
+	defer s.Stopper().Stop(ctx)
+
+	// We're going to open a client connection and do the minimum so that the
+	// server gets to the authentication phase, where it will block.
+	conn, err := net.Dial("tcp", s.ServingAddr())
+	if err != nil {
+		t.Fatal(err)
+	}
+	fe, err := pgproto3.NewFrontend(conn, conn)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if err := fe.Send(&pgproto3.StartupMessage{ProtocolVersion: version30}); err != nil {
+		t.Fatal(err)
+	}
+
+	// Wait for server to block the auth.
+	<-authBlocked
+	// Close the connection. This is supposed to unblock the auth by canceling its
+	// ctx.
+	if err := conn.Close(); err != nil {
+		t.Fatal(err)
+	}
+	// Check that the auth process indeed noticed the cancelation.
+	<-authBlocked
+}
