@@ -15,7 +15,6 @@
 package pgwire_test
 
 import (
-	"bytes"
 	"context"
 	gosql "database/sql"
 	"database/sql/driver"
@@ -34,11 +33,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/jackc/pgx"
-	"github.com/jackc/pgx/pgproto3"
-	"github.com/lib/pq"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/security/securitytest"
@@ -52,15 +46,14 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/jackc/pgx"
+	"github.com/jackc/pgx/pgproto3"
+	"github.com/lib/pq"
+	"github.com/pkg/errors"
 )
 
-// https://golang.org/cl/38533 and https://golang.org/cl/91115 changed the
-// validation message.
 func wrongArgCountString(want, got int) string {
-	if strings.HasPrefix(runtime.Version(), "go1.10") {
-		return fmt.Sprintf("sql: expected %d arguments, got %d", want, got)
-	}
-	return fmt.Sprintf("sql: statement expects %d inputs; got %d", want, got)
+	return fmt.Sprintf("sql: expected %d arguments, got %d", want, got)
 }
 
 func trivialQuery(pgURL url.URL) error {
@@ -267,7 +260,7 @@ func TestPGWireDrainClient(t *testing.T) {
 
 	// Ensure server is in draining mode and rejects new connections.
 	testutils.SucceedsSoon(t, func() error {
-		if err := trivialQuery(pgBaseURL); !testutils.IsError(err, pgwire.ErrDraining) {
+		if err := trivialQuery(pgBaseURL); !testutils.IsError(err, pgwire.ErrDrainingNewConn) {
 			return errors.Errorf("unexpected error: %v", err)
 		}
 		return nil
@@ -476,10 +469,10 @@ func TestPGPrepareFail(t *testing.T) {
 		"SELECT $1 + $1":                            "pq: could not determine data type of placeholder $1",
 		"SELECT CASE WHEN TRUE THEN $1 END":         "pq: could not determine data type of placeholder $1",
 		"SELECT CASE WHEN TRUE THEN $1 ELSE $2 END": "pq: could not determine data type of placeholder $1",
-		"SELECT $1 > 0 AND NOT $1":                  "pq: placeholder 1 already has type int, cannot assign bool",
+		"SELECT $1 > 0 AND NOT $1":                  "pq: placeholder $1 already has type int, cannot assign bool",
 		"CREATE TABLE $1 (id INT)":                  "pq: syntax error at or near \"1\"",
 		"UPDATE d.t SET s = i + $1":                 "pq: unsupported binary operator: <int> + <placeholder{1}> (desired <string>)",
-		"SELECT $0 > 0":                             "pq: invalid placeholder name: $0",
+		"SELECT $0 > 0":                             "pq: lexical error: placeholder index must be between 1 and 65536",
 		"SELECT $2 > 0":                             "pq: could not determine data type of placeholder $1",
 		"SELECT 3 + CASE (4) WHEN 4 THEN $1 END":    "pq: could not determine data type of placeholder $1",
 		"SELECT ($1 + $1) + current_date()":         "pq: could not determine data type of placeholder $1",
@@ -704,6 +697,7 @@ func TestPGPreparedQuery(t *testing.T) {
 		}},
 		{"SELECT $1[2] LIKE 'b'", []preparedQueryTest{
 			baseTest.SetArgs(pq.Array([]string{"a", "b", "c"})).Results(true),
+			baseTest.SetArgs(pq.Array([]gosql.NullString{{String: "a", Valid: true}, {Valid: false}, {String: "c", Valid: true}})).Results(gosql.NullBool{Valid: false}),
 		}},
 		{"SET application_name = $1", []preparedQueryTest{
 			baseTest.SetArgs("hello world"),
@@ -719,7 +713,7 @@ func TestPGPreparedQuery(t *testing.T) {
 		}},
 		{"SHOW COLUMNS FROM system.users", []preparedQueryTest{
 			baseTest.
-				Results("username", "STRING", false, gosql.NullBool{}, "", "{\"primary\"}", false).
+				Results("username", "STRING", false, gosql.NullBool{}, "", "{primary}", false).
 				Results("hashedPassword", "BYTES", true, gosql.NullBool{}, "", "{}", false).
 				Results("isRole", "BOOL", false, false, "", "{}", false),
 		}},
@@ -742,7 +736,7 @@ func TestPGPreparedQuery(t *testing.T) {
 			baseTest.Results("users", "primary", false, 1, "username", "ASC", false, false),
 		}},
 		{"SHOW TABLES FROM system", []preparedQueryTest{
-			baseTest.Results("descriptor").Others(13),
+			baseTest.Results("comments").Others(14),
 		}},
 		{"SHOW SCHEMAS FROM system", []preparedQueryTest{
 			baseTest.Results("crdb_internal").Others(3),
@@ -912,7 +906,7 @@ func TestPGPreparedQuery(t *testing.T) {
 		}},
 		// #14238
 		{"EXPLAIN SELECT 1", []preparedQueryTest{
-			baseTest.SetArgs().Results("render", "", "").Results(" └── emptyrow", "", ""),
+			baseTest.SetArgs().Results("values", "", "").Results("", "size", "1 column, 1 row"),
 		}},
 		// #14245
 		{"SELECT 1::oid = $1", []preparedQueryTest{
@@ -941,17 +935,27 @@ func TestPGPreparedQuery(t *testing.T) {
 			baseTest.SetArgs("{1.000}").Results("{1.000}"),
 		}},
 		{"SELECT $1:::STRING[]", []preparedQueryTest{
-			baseTest.SetArgs(`{aaa}`).Results(`{"aaa"}`),
-			baseTest.SetArgs(`{"aaa"}`).Results(`{"aaa"}`),
-			baseTest.SetArgs(`{aaa,bbb,ccc}`).Results(`{"aaa","bbb","ccc"}`),
+			baseTest.SetArgs(`{aaa}`).Results(`{aaa}`),
+			baseTest.SetArgs(`{"aaa"}`).Results(`{aaa}`),
+			baseTest.SetArgs(`{aaa,bbb,ccc}`).Results(`{aaa,bbb,ccc}`),
 		}},
 		{"SELECT $1:::JSON", []preparedQueryTest{
 			baseTest.SetArgs(`true`).Results(`true`),
 			baseTest.SetArgs(`"hello"`).Results(`"hello"`),
 		}},
+		{"SELECT $1:::BIT(4)", []preparedQueryTest{
+			baseTest.SetArgs(`1101`).Results(`1101`),
+		}},
+		{"SELECT $1:::VARBIT", []preparedQueryTest{
+			baseTest.SetArgs(`1101`).Results(`1101`),
+			baseTest.SetArgs(`1101001`).Results(`1101001`),
+		}},
 		{"SELECT $1::INT[]", []preparedQueryTest{
 			baseTest.SetArgs(pq.Array([]int64{10})).Results(pq.Array([]int64{10})),
 		}},
+		{"INSERT INTO d.arr VALUES($1, $2)", []preparedQueryTest{
+			baseTest.SetArgs(pq.Array([]int64{}), pq.Array([]string{})),
+		}},
 		{"EXPERIMENTAL SCRUB TABLE system.locations", []preparedQueryTest{
 			baseTest.SetArgs(),
 		}},
@@ -1080,6 +1084,7 @@ CREATE TABLE d.ts (a TIMESTAMP, b DATE);
 CREATE TABLE d.two (a INT, b INT);
 CREATE TABLE d.intStr (a INT, s STRING);
 CREATE TABLE d.str (s STRING, b BYTES);
+CREATE TABLE d.arr (a INT[], b TEXT[]);
 CREATE TABLE d.emptynorows (); -- zero columns, zero rows
 CREATE TABLE d.emptyrows (x INT);
 INSERT INTO d.emptyrows VALUES (1),(2),(3);
@@ -1145,6 +1150,45 @@ func (p preparedExecTest) RowsAffectedErr(err string) preparedExecTest {
 	return p
 }
 
+// Verify that bound dates are evaluated using session timezone.
+func TestPGPrepareDate(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	s, db, _ := serverutils.StartServer(t, base.TestServerArgs{})
+	defer s.Stopper().Stop(context.TODO())
+
+	if _, err := db.Exec("CREATE TABLE test (t TIMESTAMPTZ)"); err != nil {
+		t.Fatal(err)
+	}
+
+	if _, err := db.Exec("SET TIME ZONE +08"); err != nil {
+		t.Fatal(err)
+	}
+
+	stmt, err := db.Prepare("INSERT INTO test VALUES ($1)")
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	if _, err := stmt.Exec("2018-01-01 12:34:56"); err != nil {
+		t.Fatal(err)
+	}
+
+	// Reset to UTC for the query.
+	if _, err := db.Exec("SET TIME ZONE UTC"); err != nil {
+		t.Fatal(err)
+	}
+
+	var ts time.Time
+	if err := db.QueryRow("SELECT t FROM test").Scan(&ts); err != nil {
+		t.Fatal(err)
+	}
+
+	exp := time.Date(2018, 1, 1, 4, 34, 56, 0, time.UTC)
+	if !exp.Equal(ts) {
+		t.Fatalf("expected %s, got %s", exp, ts)
+	}
+}
+
 func TestPGPreparedExec(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	var baseTest preparedExecTest
@@ -1853,8 +1897,8 @@ func TestPGWireAuth(t *testing.T) {
 				Host:     net.JoinHostPort(host, port),
 				RawQuery: "sslmode=require",
 			}
-			if err := trivialQuery(unicodeUserPgURL); !testutils.IsError(err, "pq: password authentication failed for user") {
-				t.Fatalf("unexpected error: %v", err)
+			if err := trivialQuery(unicodeUserPgURL); !testutils.IsError(err, "password authentication failed for user") {
+				t.Fatalf("expected \"password authentication failed for user\", got: %v", err)
 			}
 
 			// Supply correct password.
@@ -1887,12 +1931,174 @@ func TestPGWireAuth(t *testing.T) {
 		// Even though the correct password is supplied (empty string), this
 		// should fail because we do not support password authentication for
 		// users with empty passwords.
-		if err := trivialQuery(testUserPgURL); !testutils.IsError(err, "pq: password authentication failed for user") {
-			t.Fatalf("unexpected error: %v", err)
+		if err := trivialQuery(testUserPgURL); !testutils.IsError(err, "password authentication failed for user") {
+			t.Fatalf("expected \"password authentication failed for user\", got: %v", err)
 		}
 	})
 }
 
+func TestHBA(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	s, conn, _ := serverutils.StartServer(t, base.TestServerArgs{})
+	defer s.Stopper().Stop(context.TODO())
+	db := sqlutils.MakeSQLRunner(conn)
+
+	const (
+		user = "passworduser"
+		pass = "pass"
+	)
+
+	db.Exec(t, fmt.Sprintf(`CREATE USER %s`, server.TestUser))
+	db.Exec(t, fmt.Sprintf(`CREATE USER %s WITH PASSWORD '%s'`, user, pass))
+
+	tests := []struct {
+		// The hba.conf file/setting.
+		conf string
+		// Error message when setting the config.
+		confErr string
+		// Error message of password login.
+		passErr string
+		// Error message of cert login.
+		certErr string
+	}{
+		{
+			conf:    `bad`,
+			confErr: "entry 1 invalid",
+		},
+		{
+			conf:    `#empty`,
+			confErr: "no entries",
+		},
+		{
+			conf:    `host all all 1.1.1/0 cert`,
+			confErr: "invalid CIDR address",
+		},
+		{
+			conf:    `host all all 0.0.0.0/0 invalid`,
+			confErr: "unknown auth method",
+		},
+		{
+			conf:    `host db all 0.0.0.0/0 cert`,
+			confErr: "database must be specified as all",
+		},
+		{
+			// quoted all isn't ok since it strips the special meaning
+			conf:    `host "all" all 0.0.0.0/0 cert`,
+			confErr: "database must be specified as all",
+		},
+		{
+			// only the all hostname is supported
+			conf:    `host all all hostname cert`,
+			confErr: "host addresses not supported",
+		},
+		{
+			// valid for both specified users
+			conf: `
+				host all testuser 0.0.0.0/0 cert
+				host all passworduser 0.0.0.0/0 password
+			`,
+		},
+		{
+			// valid for both specified users
+			conf: `
+				host all testuser,passworduser all cert-password
+			`,
+		},
+		{
+			// the "all" user means password never is checked
+			conf: `
+				host all testuser,all 0.0.0.0/0 cert
+				host all passworduser 0.0.0.0/0 password
+			`,
+			passErr: "no TLS peer certificates",
+		},
+		{
+			// but double quoting removes special meaning
+			conf: `
+				host all testuser,"all" 0.0.0.0/0 cert
+				host all passworduser 0.0.0.0/0 password
+			`,
+		},
+		{
+			// disallow passwords
+			conf:    "host all all 0.0.0.0/0 cert",
+			passErr: "no TLS peer certificates",
+		},
+		{
+			// disallow certs
+			conf:    "host all all 0.0.0.0/0 password",
+			certErr: "password authentication failed for user testuser",
+		},
+		{
+			// invalid user name
+			conf:    "host all invalid 0.0.0.0/0 cert",
+			certErr: "no .* entry",
+			passErr: "no .* entry",
+		},
+		{
+			// invalid IP
+			conf:    "host all all 0.0.0.0/32 cert",
+			certErr: "no .* entry",
+			passErr: "no .* entry",
+		},
+		{
+			// allow certs from 127.*.*.*
+			conf:    "host all all 127.0.0.0/8 cert",
+			passErr: "no TLS peer certificates",
+		},
+		{
+			// allow certs from 128.*.*.*, but connect from 127.0.0.0
+			conf:    "host all all 128.0.0.0/8 cert",
+			certErr: "no .* entry",
+			passErr: "no .* entry",
+		},
+	}
+	for i, tc := range tests {
+		t.Run(fmt.Sprint(i), func(t *testing.T) {
+			t.Log(tc.conf)
+			db.ExpectErr(t, tc.confErr, `SET CLUSTER SETTING server.host_based_authentication.configuration = $1`, tc.conf)
+
+			t.Run("root", func(t *testing.T) {
+				// Authenticate as root with certificate and expect success.
+				rootPgURL, cleanupFn := sqlutils.PGUrl(
+					t, s.ServingAddr(), t.Name(), url.User(security.RootUser))
+				defer cleanupFn()
+				if err := trivialQuery(rootPgURL); err != nil {
+					t.Fatalf("could not auth as root: %v", err)
+				}
+			})
+
+			t.Run("cert", func(t *testing.T) {
+				testUserPgURL, cleanupFn := sqlutils.PGUrl(
+					t, s.ServingAddr(), t.Name(), url.User(server.TestUser))
+				defer cleanupFn()
+				err := trivialQuery(testUserPgURL)
+				if !testutils.IsError(err, tc.certErr) {
+					t.Errorf("expected err %v, got %v", tc.certErr, err)
+				}
+			})
+
+			t.Run("password", func(t *testing.T) {
+				host, port, err := net.SplitHostPort(s.ServingAddr())
+				if err != nil {
+					t.Fatal(err)
+				}
+				testURL := url.URL{
+					Scheme:   "postgres",
+					User:     url.UserPassword(user, pass),
+					Host:     net.JoinHostPort(host, port),
+					RawQuery: "sslmode=require",
+				}
+				err = trivialQuery(testURL)
+				if !testutils.IsError(err, tc.passErr) {
+					t.Errorf("expected err %v, got %v", tc.passErr, err)
+				}
+			})
+		})
+	}
+}
+
 func TestPGWireResultChange(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	s, _, _ := serverutils.StartServer(t, base.TestServerArgs{})
@@ -1956,39 +2162,6 @@ func TestPGWireResultChange(t *testing.T) {
 	}
 }
 
-func TestPGWireTooManyArguments(t *testing.T) {
-	defer leaktest.AfterTest(t)()
-	s, _, _ := serverutils.StartServer(t, base.TestServerArgs{})
-	defer s.Stopper().Stop(context.TODO())
-
-	pgURL, cleanupFn := sqlutils.PGUrl(t, s.ServingAddr(), t.Name(), url.User(security.RootUser))
-	defer cleanupFn()
-
-	db, err := gosql.Open("postgres", pgURL.String())
-	if err != nil {
-		t.Fatal(err)
-	}
-	defer db.Close()
-
-	var b bytes.Buffer
-	b.WriteString("SELECT")
-
-	args := make([]interface{}, 1<<16)
-	for i := 1; i <= 1<<16; i++ {
-		comma := ","
-		if i == 1 {
-			comma = ""
-		}
-		b.WriteString(fmt.Sprintf("%s $%d::int", comma, i))
-		args[i-1] = i
-	}
-	b.WriteString(";")
-	query := b.String()
-	if _, err := db.Prepare(query); !testutils.IsError(err, "more than 65535 arguments") {
-		t.Fatalf("unexpected error: %v", err)
-	}
-}
-
 func TestSessionParameters(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
