@@ -25,18 +25,18 @@ import (
 	"unicode/utf8"
 	"unsafe"
 
-	"github.com/lib/pq"
-	"github.com/lib/pq/oid"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
+	"github.com/cockroachdb/cockroach/pkg/util/bitarray"
 	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/ipaddr"
 	"github.com/cockroachdb/cockroach/pkg/util/timeofday"
 	"github.com/cockroachdb/cockroach/pkg/util/uint128"
+	"github.com/jackc/pgx/pgtype"
+	"github.com/lib/pq/oid"
+	"github.com/pkg/errors"
 )
 
 const secondsInDay = 24 * 60 * 60
@@ -189,9 +189,31 @@ func NewProtocolViolationErrorf(format string, args ...interface{}) error {
 	return pgerror.NewErrorf(pgerror.CodeProtocolViolationError, format, args...)
 }
 
+// validateArrayDimensions takes the number of dimensions and elements and
+// returns an error if we don't support that combination.
+func validateArrayDimensions(nDimensions int, nElements int) error {
+	switch nDimensions {
+	case 1:
+		break
+	case 0:
+		// 0-dimensional array means 0-length array: validate that.
+		if nElements == 0 {
+			break
+		}
+		fallthrough
+	default:
+		return pgerror.UnimplementedWithIssueErrorf(32552,
+			"%d-dimension arrays not supported; only 1-dimension", nDimensions)
+	}
+	return nil
+}
+
 // DecodeOidDatum decodes bytes with specified Oid and format code into
-// a datum.
-func DecodeOidDatum(id oid.Oid, code FormatCode, b []byte) (tree.Datum, error) {
+// a datum. If the ParseTimeContext is nil, reasonable defaults
+// will be applied.
+func DecodeOidDatum(
+	ctx tree.ParseTimeContext, id oid.Oid, code FormatCode, b []byte,
+) (tree.Datum, error) {
 	switch code {
 	case FormatText:
 		switch id {
@@ -201,6 +223,12 @@ func DecodeOidDatum(id oid.Oid, code FormatCode, b []byte) (tree.Datum, error) {
 				return nil, err
 			}
 			return tree.MakeDBool(tree.DBool(t)), nil
+		case oid.T_bit, oid.T_varbit:
+			t, err := tree.ParseDBitArray(string(b))
+			if err != nil {
+				return nil, err
+			}
+			return t, nil
 		case oid.T_int2, oid.T_int4, oid.T_int8:
 			i, err := strconv.ParseInt(string(b), 10, 64)
 			if err != nil {
@@ -232,30 +260,25 @@ func DecodeOidDatum(id oid.Oid, code FormatCode, b []byte) (tree.Datum, error) {
 			}
 			return tree.NewDBytes(tree.DBytes(res)), nil
 		case oid.T_timestamp:
-			d, err := tree.ParseDTimestamp(string(b), time.Microsecond)
+			d, err := tree.ParseDTimestamp(ctx, string(b), time.Microsecond)
 			if err != nil {
 				return nil, errors.Errorf("could not parse string %q as timestamp", b)
 			}
 			return d, nil
 		case oid.T_timestamptz:
-			d, err := tree.ParseDTimestampTZ(string(b), time.UTC, time.Microsecond)
+			d, err := tree.ParseDTimestampTZ(ctx, string(b), time.Microsecond)
 			if err != nil {
 				return nil, errors.Errorf("could not parse string %q as timestamptz", b)
 			}
 			return d, nil
 		case oid.T_date:
-			ts, err := tree.ParseDTimestamp(string(b), time.Microsecond)
+			d, err := tree.ParseDDate(ctx, string(b))
 			if err != nil {
-				res, err := tree.ParseDDate(string(b), time.UTC)
-				if err != nil {
-					return nil, errors.Errorf("could not parse string %q as date", b)
-				}
-				return res, nil
+				return nil, errors.Errorf("could not parse string %q as date", b)
 			}
-			daysSinceEpoch := ts.Unix() / secondsInDay
-			return tree.NewDDate(tree.DDate(daysSinceEpoch)), nil
+			return d, nil
 		case oid.T_time:
-			d, err := tree.ParseDTime(string(b))
+			d, err := tree.ParseDTime(nil, string(b))
 			if err != nil {
 				return nil, errors.Errorf("could not parse string %q as time", b)
 			}
@@ -280,32 +303,55 @@ func DecodeOidDatum(id oid.Oid, code FormatCode, b []byte) (tree.Datum, error) {
 			}
 			return d, nil
 		case oid.T__int2, oid.T__int4, oid.T__int8:
-			var arr pq.Int64Array
-			if err := (&arr).Scan(b); err != nil {
+			var arr pgtype.Int8Array
+			if err := arr.DecodeText(nil, b); err != nil {
+				return nil, errors.Wrapf(err, "could not parse string %q as int array", b)
+			}
+			if arr.Status != pgtype.Present {
+				return tree.DNull, nil
+			}
+			if err := validateArrayDimensions(len(arr.Dimensions), len(arr.Elements)); err != nil {
 				return nil, err
 			}
 			out := tree.NewDArray(types.Int)
-			for _, v := range arr {
-				if err := out.Append(tree.NewDInt(tree.DInt(v))); err != nil {
+			var d tree.Datum
+			for _, v := range arr.Elements {
+				if v.Status != pgtype.Present {
+					d = tree.DNull
+				} else {
+					d = tree.NewDInt(tree.DInt(v.Int))
+				}
+				if err := out.Append(d); err != nil {
 					return nil, err
 				}
 			}
 			return out, nil
 		case oid.T__text, oid.T__name:
-			var arr pq.StringArray
-			if err := (&arr).Scan(b); err != nil {
+			var arr pgtype.TextArray
+			if err := arr.DecodeText(nil, b); err != nil {
+				return nil, errors.Wrapf(err, "could not parse string %q as text array", b)
+			}
+			if arr.Status != pgtype.Present {
+				return tree.DNull, nil
+			}
+			if err := validateArrayDimensions(len(arr.Dimensions), len(arr.Elements)); err != nil {
 				return nil, err
 			}
 			out := tree.NewDArray(types.String)
 			if id == oid.T__name {
 				out.ParamTyp = types.Name
 			}
-			for _, v := range arr {
-				var s tree.Datum = tree.NewDString(v)
-				if id == oid.T__name {
-					s = tree.NewDNameFromDString(s.(*tree.DString))
+			var d tree.Datum
+			for _, v := range arr.Elements {
+				if v.Status != pgtype.Present {
+					d = tree.DNull
+				} else {
+					d = tree.NewDString(v.String)
+					if id == oid.T__name {
+						d = tree.NewDNameFromDString(d.(*tree.DString))
+					}
 				}
-				if err := out.Append(s); err != nil {
+				if err := out.Append(d); err != nil {
 					return nil, err
 				}
 			}
@@ -434,7 +480,7 @@ func DecodeOidDatum(id oid.Oid, code FormatCode, b []byte) (tree.Datum, error) {
 				if _, ok := alloc.dd.Coeff.SetString(decString, 10); !ok {
 					return nil, errors.Errorf("could not parse string %q as decimal", decString)
 				}
-				alloc.dd.SetExponent(-int32(Dscale))
+				alloc.dd.Exponent = -int32(Dscale)
 			}
 
 			switch alloc.pgNum.Sign {
@@ -483,11 +529,7 @@ func DecodeOidDatum(id oid.Oid, code FormatCode, b []byte) (tree.Datum, error) {
 			days := int32(binary.BigEndian.Uint32(b[8:]))
 			months := int32(binary.BigEndian.Uint32(b[12:]))
 
-			duration := duration.Duration{
-				Nanos:  nanos,
-				Days:   int64(days),
-				Months: int64(months),
-			}
+			duration := duration.MakeDuration(nanos, int64(days), int64(months))
 			return &tree.DInterval{Duration: duration}, nil
 		case oid.T_uuid:
 			u, err := tree.ParseDUuidFromBytes(b)
@@ -508,9 +550,41 @@ func DecodeOidDatum(id oid.Oid, code FormatCode, b []byte) (tree.Datum, error) {
 				return nil, err
 			}
 			return tree.ParseDJSON(string(b))
+		case oid.T_varbit, oid.T_bit:
+			if len(b) < 4 {
+				return nil, errors.Errorf("missing varbit bitlen prefix")
+			}
+			bitlen := binary.BigEndian.Uint32(b)
+			b = b[4:]
+			lastBitsUsed := uint64(bitlen % 64)
+			if bitlen != 0 && lastBitsUsed == 0 {
+				lastBitsUsed = 64
+			}
+			if len(b)*8 < int(bitlen) {
+				return nil, errors.Errorf("unexpected varbit bitlen %d (b: %d)", bitlen, len(b))
+			}
+			words := make([]uint64, (len(b)+7)/8)
+			// We need two loops here. The first loop does full 8-byte decoding. The
+			// last word is not guaranteed to be a full 8 bytes, and so the second loop
+			// does manual per-byte decoding.
+			for i := 0; i < len(words)-1; i++ {
+				words[i] = binary.BigEndian.Uint64(b)
+				b = b[8:]
+			}
+			if len(words) > 0 {
+				var w uint64
+				i := uint(0)
+				for ; i < uint(lastBitsUsed); i += 8 {
+					w = (w << 8) | uint64(b[0])
+					b = b[1:]
+				}
+				words[len(words)-1] = w << (64 - i)
+			}
+			ba, err := bitarray.FromEncodingParts(words, lastBitsUsed)
+			return &tree.DBitArray{BitArray: ba}, err
 		default:
 			if _, ok := types.ArrayOids[id]; ok {
-				return decodeBinaryArray(b, code)
+				return decodeBinaryArray(ctx, b, code)
 			}
 		}
 	default:
@@ -615,7 +689,7 @@ func pgBinaryToIPAddr(b []byte) (ipaddr.IPAddr, error) {
 	}, nil
 }
 
-func decodeBinaryArray(b []byte, code FormatCode) (tree.Datum, error) {
+func decodeBinaryArray(ctx tree.ParseTimeContext, b []byte, code FormatCode) (tree.Datum, error) {
 	hdr := struct {
 		Ndims int32
 		// Nullflag
@@ -632,9 +706,8 @@ func decodeBinaryArray(b []byte, code FormatCode) (tree.Datum, error) {
 	if err := binary.Read(r, binary.BigEndian, &hdr); err != nil {
 		return nil, err
 	}
-	// Only 1-dimensional arrays are supported for now.
-	if hdr.Ndims != 1 {
-		return nil, errors.Errorf("unsupported number of array dimensions: %d", hdr.Ndims)
+	if err := validateArrayDimensions(int(hdr.Ndims), int(hdr.DimSize)); err != nil {
+		return nil, err
 	}
 
 	elemOid := oid.Oid(hdr.ElemOid)
@@ -651,7 +724,7 @@ func decodeBinaryArray(b []byte, code FormatCode) (tree.Datum, error) {
 			continue
 		}
 		buf := r.Next(int(vlen))
-		elem, err := DecodeOidDatum(elemOid, code, buf)
+		elem, err := DecodeOidDatum(ctx, elemOid, code, buf)
 		if err != nil {
 			return nil, err
 		}
