@@ -23,23 +23,46 @@ import (
 	"strings"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
+	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/hba"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgwirebase"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqltelemetry"
 	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
+	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/pkg/errors"
+)
+
+// ATTENTION: After changing this value in a unit test, you probably want to
+// open a new connection pool since the connections in the existing one are not
+// affected.
+//
+// The "results_buffer_size" connection parameter can be used to override this
+// default for an individual connection.
+var connResultsBufferSize = settings.RegisterByteSizeSetting(
+	"sql.defaults.results_buffer.size",
+	"default size of the buffer that accumulates results for a statement or a batch "+
+		"of statements before they are sent to the client. This can be overridden on "+
+		"an individual connection with the 'results_buffer_size' parameter. Note that auto-retries "+
+		"generally only happen while no results have been delivered to the client, so "+
+		"reducing this size can increase the number of retriable errors a client "+
+		"receives. On the other hand, increasing the buffer size can increase the "+
+		"delay until the client receives the first result row. "+
+		"Updating the setting only affects new connections. "+
+		"Setting to 0 disables any buffering.",
+	16<<10, // 16 KiB
 )
 
 const (
@@ -47,9 +70,12 @@ const (
 	// secure server in cleartext.
 	ErrSSLRequired = "node is running secure mode, SSL connection required"
 
-	// ErrDraining is returned when a client attempts to connect to a server
+	// ErrDrainingNewConn is returned when a client attempts to connect to a server
 	// which is not accepting client connections.
-	ErrDraining = "server is not accepting clients"
+	ErrDrainingNewConn = "server is not accepting clients"
+	// ErrDrainingExistingConn is returned when a connection is shut down because
+	// the server is draining.
+	ErrDrainingExistingConn = "server is shutting down"
 )
 
 // Fully-qualified names for metrics.
@@ -120,6 +146,11 @@ type Server struct {
 		draining      bool
 	}
 
+	auth struct {
+		syncutil.RWMutex
+		conf *hba.Conf
+	}
+
 	sqlMemoryPool mon.BytesMonitor
 	connMonitor   mon.BytesMonitor
 
@@ -194,6 +225,30 @@ func MakeServer(
 	server.mu.connCancelMap = make(cancelChanMap)
 	server.mu.Unlock()
 
+	connAuthConf.SetOnChange(&st.SV, func() {
+		val := connAuthConf.Get(&st.SV)
+		server.auth.Lock()
+		defer server.auth.Unlock()
+		if val == "" {
+			server.auth.conf = nil
+			return
+		}
+		conf, err := hba.Parse(val)
+		if err != nil {
+			log.Warningf(ambientCtx.AnnotateCtx(context.Background()), "invalid %s: %v", serverHBAConfSetting, err)
+			conf = nil
+		}
+		// Usernames are normalized during session init. Normalize the HBA usernames
+		// in the same way.
+		for _, entry := range conf.Entries {
+			for iu := range entry.User {
+				user := &entry.User[iu]
+				user.Value = tree.Name(user.Value).Normalize()
+			}
+		}
+		server.auth.conf = conf
+	})
+
 	return server
 }
 
@@ -225,19 +280,15 @@ func (s *Server) IsDraining() bool {
 	return s.mu.draining
 }
 
-// Metrics returns the metrics struct.
-func (s *Server) Metrics() *ServerMetrics {
-	return &s.metrics
-}
-
-// StatementCounters returns the Server's StatementCounters.
-func (s *Server) StatementCounters() *sql.StatementCounters {
-	return &s.SQLServer.StatementCounters
-}
-
-// EngineMetrics returns the Server's EngineMetrics.
-func (s *Server) EngineMetrics() *sql.EngineMetrics {
-	return &s.SQLServer.EngineMetrics
+// Metrics returns the set of metrics structs.
+func (s *Server) Metrics() (res []interface{}) {
+	return []interface{}{
+		&s.metrics,
+		&s.SQLServer.Metrics.StatementCounters,
+		&s.SQLServer.Metrics.EngineMetrics,
+		&s.SQLServer.InternalMetrics.StatementCounters,
+		&s.SQLServer.InternalMetrics.EngineMetrics,
+	}
 }
 
 // Drain prevents new connections from being served and waits for drainWait for
@@ -351,6 +402,8 @@ func (s *Server) drainImpl(drainWait time.Duration, cancelWait time.Duration) er
 
 // ServeConn serves a single connection, driving the handshake process and
 // delegating to the appropriate connection type.
+//
+// An error is returned if the initial handshake of the connection fails.
 func (s *Server) ServeConn(ctx context.Context, conn net.Conn) error {
 	s.mu.Lock()
 	draining := s.mu.draining
@@ -424,14 +477,14 @@ func (s *Server) ServeConn(ctx context.Context, conn net.Conn) error {
 
 	sendErr := func(err error) error {
 		msgBuilder := newWriteBuffer(s.metrics.BytesOutCount)
-		_ /* err */ = writeErr(err, msgBuilder, conn)
+		_ /* err */ = writeErr(ctx, &s.execCfg.Settings.SV, err, msgBuilder, conn)
 		_ = conn.Close()
 		return err
 	}
 
 	if version != version30 {
 		if version == versionCancel {
-			telemetry.Count("pgwire.unimplemented.cancel_request")
+			telemetry.Inc(sqltelemetry.CancelRequestCounter)
 			_ = conn.Close()
 			return nil
 		}
@@ -441,7 +494,7 @@ func (s *Server) ServeConn(ctx context.Context, conn net.Conn) error {
 		return sendErr(pgerror.NewError(pgerror.CodeProtocolViolationError, ErrSSLRequired))
 	}
 	if draining {
-		return sendErr(newAdminShutdownErr(errors.New(ErrDraining)))
+		return sendErr(newAdminShutdownErr(ErrDrainingNewConn))
 	}
 
 	var sArgs sql.SessionArgs
@@ -449,6 +502,9 @@ func (s *Server) ServeConn(ctx context.Context, conn net.Conn) error {
 		return sendErr(err)
 	}
 	sArgs.User = tree.Name(sArgs.User).Normalize()
+	if sArgs.ConnResultsBufferSize == connResultsBufferSizeUnsetSentinel {
+		sArgs.ConnResultsBufferSize = connResultsBufferSize.Get(&s.execCfg.Settings.SV)
+	}
 
 	// Reserve some memory for this connection using the server's monitor. This
 	// reduces pressure on the shared pool because the server monitor allocates in
@@ -459,12 +515,38 @@ func (s *Server) ServeConn(ctx context.Context, conn net.Conn) error {
 		return errors.Errorf("unable to pre-allocate %d bytes for this connection: %v",
 			baseSQLMemoryBudget, err)
 	}
-	return serveConn(ctx, conn, sArgs, &s.metrics, reserved, s.SQLServer,
-		s.IsDraining, s.execCfg, s.stopper, s.cfg.Insecure)
+
+	s.auth.RLock()
+	auth := s.auth.conf
+	s.auth.RUnlock()
+
+	var authHook func(context.Context) error
+	if k := s.execCfg.PGWireTestingKnobs; k != nil {
+		authHook = k.AuthHook
+	}
+
+	serveConn(
+		ctx, conn, sArgs,
+		&s.metrics, reserved, s.SQLServer,
+		s.IsDraining,
+		authOptions{
+			insecure: s.cfg.Insecure,
+			ie:       s.execCfg.InternalExecutor,
+			auth:     auth,
+			authHook: authHook,
+		},
+		s.stopper)
+	return nil
 }
 
+// -1 for the sentinel in case someone wants to set it to 0.
+const connResultsBufferSizeUnsetSentinel = -1
+
 func parseOptions(ctx context.Context, data []byte) (sql.SessionArgs, error) {
-	args := sql.SessionArgs{SessionDefaults: make(map[string]string)}
+	args := sql.SessionArgs{
+		SessionDefaults:       make(map[string]string),
+		ConnResultsBufferSize: connResultsBufferSizeUnsetSentinel,
+	}
 	buf := pgwirebase.ReadBuffer{Msg: data}
 	for {
 		key, err := buf.GetString()
@@ -484,12 +566,25 @@ func parseOptions(ctx context.Context, data []byte) (sql.SessionArgs, error) {
 		switch key {
 		case "user":
 			args.User = value
+		case "results_buffer_size":
+			if args.ConnResultsBufferSize, err = humanizeutil.ParseBytes(value); err != nil {
+				return sql.SessionArgs{}, pgerror.NewErrorf(pgerror.CodeProtocolViolationError,
+					"error parsing results_buffer_size option value '%s' as bytes", value)
+			}
+			if args.ConnResultsBufferSize < 0 {
+				return sql.SessionArgs{}, pgerror.NewErrorf(pgerror.CodeProtocolViolationError,
+					"results_buffer_size option value '%s' cannot be negative", value)
+			}
 		default:
 			exists, configurable := sql.IsSessionVariableConfigurable(key)
 			if exists && configurable {
 				args.SessionDefaults[key] = value
 			} else {
 				if !exists {
+					if _, ok := sql.UnsupportedVars[key]; ok {
+						counter := sqltelemetry.UnimplementedClientStatusParameterCounter(key)
+						telemetry.Inc(counter)
+					}
 					log.Warningf(ctx, "unknown configuration parameter: %q", key)
 				} else {
 					return sql.SessionArgs{}, pgerror.NewErrorf(pgerror.CodeCantChangeRuntimeParamError,
@@ -508,6 +603,6 @@ func parseOptions(ctx context.Context, data []byte) (sql.SessionArgs, error) {
 	return args, nil
 }
 
-func newAdminShutdownErr(err error) error {
-	return pgerror.NewErrorf(pgerror.CodeAdminShutdownError, err.Error())
+func newAdminShutdownErr(msg string) error {
+	return pgerror.NewErrorf(pgerror.CodeAdminShutdownError, msg)
 }
