@@ -256,12 +256,14 @@ func (pp *physicalProps) reduce(order sqlbase.ColumnOrdering) sqlbase.ColumnOrde
 //   a=b=c; d=e=f; g=CONST; h=CONST; b+,d-
 func (pp *physicalProps) Format(buf *bytes.Buffer, columns sqlbase.ResultColumns) {
 	pp.check()
-	fmtCtx := tree.MakeFmtCtx(buf, tree.FmtSimple)
 	printCol := func(buf *bytes.Buffer, columns sqlbase.ResultColumns, colIdx int) {
 		if columns == nil || colIdx >= len(columns) {
 			fmt.Fprintf(buf, "@%d", colIdx+1)
 		} else {
+			fmtCtx := tree.NewFmtCtx(tree.FmtSimple)
 			fmtCtx.FormatNameP(&columns[colIdx].Name)
+			_, _ = fmtCtx.WriteTo(buf)
+			fmtCtx.Close()
 		}
 	}
 
