@@ -16,14 +16,8 @@ package sql
 
 import (
 	"context"
-	"fmt"
-
-	"github.com/pkg/errors"
 
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
-	"github.com/cockroachdb/cockroach/pkg/sql/opt/exec/execbuilder"
-	"github.com/cockroachdb/cockroach/pkg/sql/opt/memo"
-	"github.com/cockroachdb/cockroach/pkg/sql/opt/optbuilder"
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
@@ -32,6 +26,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 type planMaker interface {
@@ -58,7 +53,7 @@ type planMaker interface {
 	// iterating using curPlan.plan.Next() and curPlan.plan.Values() in
 	// order to retrieve matching rows. Finally, the plan must be closed
 	// with curPlan.close().
-	makePlan(ctx context.Context, stmt Statement) error
+	makePlan(ctx context.Context) error
 
 	// prepare does the same checks as makePlan but skips building some
 	// data structures necessary for execution, based on the assumption
@@ -98,6 +93,11 @@ func (r *runParams) SessionData() *sessiondata.SessionData {
 	return r.extendedEvalCtx.SessionData
 }
 
+// ExecCfg gives convenient access to the runParam's ExecutorConfig.
+func (r *runParams) ExecCfg() *ExecutorConfig {
+	return r.extendedEvalCtx.ExecCfg
+}
+
 // planNode defines the interface for executing a query or portion of a query.
 //
 // The following methods apply to planNodes and contain special cases
@@ -116,10 +116,11 @@ func (r *runParams) SessionData() *sessiondata.SessionData {
 // - planColumns()                 (plan_columns.go)
 //
 // Also, there are optional interfaces that new nodes may want to implement:
-// - execStartable
 // - autoCommitNode
 //
 type planNode interface {
+	startExec(params runParams) error
+
 	// Next performs one unit of work, returning false if an error is
 	// encountered or if there is no more work to do. For statements
 	// that return a result set, the Values() method will return one row
@@ -141,6 +142,10 @@ type planNode interface {
 	// This method should be called if the node has been used in any way (any
 	// methods on it have been called) after it was constructed. Note that this
 	// doesn't imply that startPlan() has been necessarily called.
+	//
+	// This method must not be called during execution - the planNode
+	// tree must remain "live" and readable via walk() even after
+	// execution completes.
 	Close(ctx context.Context)
 }
 
@@ -161,6 +166,8 @@ type planNodeFastPath interface {
 var _ planNode = &alterIndexNode{}
 var _ planNode = &alterSequenceNode{}
 var _ planNode = &alterTableNode{}
+var _ planNode = &cancelQueriesNode{}
+var _ planNode = &cancelSessionsNode{}
 var _ planNode = &createDatabaseNode{}
 var _ planNode = &createIndexNode{}
 var _ planNode = &createSequenceNode{}
@@ -170,6 +177,7 @@ var _ planNode = &CreateUserNode{}
 var _ planNode = &createViewNode{}
 var _ planNode = &delayedNode{}
 var _ planNode = &deleteNode{}
+var _ planNode = &deleteRangeNode{}
 var _ planNode = &distinctNode{}
 var _ planNode = &dropDatabaseNode{}
 var _ planNode = &dropIndexNode{}
@@ -186,6 +194,7 @@ var _ planNode = &indexJoinNode{}
 var _ planNode = &insertNode{}
 var _ planNode = &joinNode{}
 var _ planNode = &limitNode{}
+var _ planNode = &max1RowNode{}
 var _ planNode = &ordinalityNode{}
 var _ planNode = &projectSetNode{}
 var _ planNode = &relocateNode{}
@@ -198,8 +207,8 @@ var _ planNode = &rowCountNode{}
 var _ planNode = &scanNode{}
 var _ planNode = &scatterNode{}
 var _ planNode = &serializeNode{}
+var _ planNode = &sequenceSelectNode{}
 var _ planNode = &showFingerprintsNode{}
-var _ planNode = &showRangesNode{}
 var _ planNode = &showTraceNode{}
 var _ planNode = &sortNode{}
 var _ planNode = &splitNode{}
@@ -209,6 +218,7 @@ var _ planNode = &unionNode{}
 var _ planNode = &updateNode{}
 var _ planNode = &upsertNode{}
 var _ planNode = &valuesNode{}
+var _ planNode = &virtualTableNode{}
 var _ planNode = &windowNode{}
 var _ planNode = &zeroNode{}
 
@@ -216,7 +226,7 @@ var _ planNodeFastPath = &CreateUserNode{}
 var _ planNodeFastPath = &DropUserNode{}
 var _ planNodeFastPath = &alterUserSetPasswordNode{}
 var _ planNodeFastPath = &createTableNode{}
-var _ planNodeFastPath = &deleteNode{}
+var _ planNodeFastPath = &deleteRangeNode{}
 var _ planNodeFastPath = &rowCountNode{}
 var _ planNodeFastPath = &serializeNode{}
 var _ planNodeFastPath = &setZoneConfigNode{}
@@ -276,16 +286,30 @@ type planTop struct {
 	// #10028 is addressed.
 	hasStar bool
 
-	// isCorrelated collects whether the query was found to be correlated.
-	// Used to produce better error messages.
-	isCorrelated bool
-
 	// subqueryPlans contains all the sub-query plans.
 	subqueryPlans []subquery
 
 	// auditEvents becomes non-nil if any of the descriptors used by
 	// current statement is causing an auditing event. See exec_log.go.
 	auditEvents []auditEvent
+
+	// flags is populated during planning and execution.
+	flags planFlags
+
+	// execErr retains the last execution error, if any.
+	execErr error
+
+	// maybeSavePlan, if defined, is called during close() to
+	// conditionally save the logical plan to savedPlanForStats.
+	maybeSavePlan func(context.Context) *roachpb.ExplainTreePlanNode
+
+	// savedPlanForStats is conditionally populated at the end of
+	// statement execution, for registration in statement statistics.
+	savedPlanForStats *roachpb.ExplainTreePlanNode
+
+	// avoidBuffering, when set, causes the execution to avoid buffering
+	// results.
+	avoidBuffering bool
 }
 
 // makePlan implements the Planner interface. It populates the
@@ -294,13 +318,14 @@ type planTop struct {
 // The caller is responsible for populating the placeholders
 // beforehand (currently in semaCtx.Placeholders).
 //
-// After makePlan(), the caller should be careful to also call
-// p.curPlan.Close().
-func (p *planner) makePlan(ctx context.Context, stmt Statement) error {
+// If no error is returned, the caller must call p.curPlan.Close() once the plan
+// is no longer needed.
+func (p *planner) makePlan(ctx context.Context) error {
 	// Reinitialize.
+	stmt := p.stmt
 	p.curPlan = planTop{AST: stmt.AST}
 
-	log.VEvent(ctx, 1, "heuristic planner starts")
+	log.VEvent(ctx, 2, "heuristic planner starts")
 
 	var err error
 	p.curPlan.plan, err = p.newPlan(ctx, stmt.AST, nil /*desiredTypes*/)
@@ -329,7 +354,7 @@ func (p *planner) makePlan(ctx context.Context, stmt Statement) error {
 		return err
 	}
 
-	log.VEvent(ctx, 1, "heuristic planner optimizes plan")
+	log.VEvent(ctx, 2, "heuristic planner optimizes plan")
 
 	needed := allColumns(p.curPlan.plan)
 	p.curPlan.plan, err = p.optimizePlan(ctx, p.curPlan.plan, needed)
@@ -338,7 +363,7 @@ func (p *planner) makePlan(ctx context.Context, stmt Statement) error {
 		return err
 	}
 
-	log.VEvent(ctx, 1, "heuristic planner optimizes subqueries")
+	log.VEvent(ctx, 2, "heuristic planner optimizes subqueries")
 
 	// Now do the same work for all sub-queries.
 	for i := range p.curPlan.subqueryPlans {
@@ -356,147 +381,7 @@ func (p *planner) makePlan(ctx context.Context, stmt Statement) error {
 	return nil
 }
 
-// makeOptimizerPlan is an alternative to makePlan which uses the cost-based
-// optimizer.
-func (p *planner) makeOptimizerPlan(ctx context.Context, stmt Statement) error {
-	// Ensure that p.curPlan is populated in case an error occurs early,
-	// so that maybeLogStatement in the error case does not find an empty AST.
-	p.curPlan = planTop{AST: stmt.AST}
-
-	// Start with fast check to see if top-level statement is supported.
-	switch stmt.AST.(type) {
-	case *tree.ParenSelect, *tree.Select, *tree.SelectClause,
-		*tree.UnionClause, *tree.ValuesClause, *tree.Explain:
-
-	default:
-		return pgerror.Unimplemented("statement", fmt.Sprintf("unsupported statement: %T", stmt.AST))
-	}
-
-	var catalog optCatalog
-	catalog.init(p.execCfg.TableStatsCache, p)
-
-	p.optimizer.Init(p.EvalContext())
-	f := p.optimizer.Factory()
-
-	// If the statement includes a PreparedStatement, then separate planning into
-	// two distinct phases:
-	//
-	//   PREPARE - Build the Memo (optbuild) and apply normalization rules to it.
-	//             If the query contains placeholders, values are not assigned
-	//             during this phase, as that only happens during the EXECUTE
-	//             phase. If the query does not contain placeholders, then also
-	//             apply exploration rules to the Memo so that there's even less
-	//             to do during the EXECUTE phase.
-	//
-	//   EXECUTE - Before the query can be executed, first any placeholders must
-	//             be assigned values. This can trigger additional normalization
-	//             rules, such as with this example:
-	//
-	//               SELECT * FROM abc WHERE b = $1 - 5
-	//
-	//             Without folding the Sub expression, any index on the "b" column
-	//             won't be found. This also means that after placeholders are
-	//             assigned, exploration rules must be applied (vs. applying them
-	//             during PREPARE when there are no placeholders).
-	//
-	//             Whether there were placeholders or not, after exploration the
-	//             plan tree must be built (execbuild). This tree is set as the
-	//             planner.curPlan, and the EXECUTE phase of planning is complete.
-	//
-	var prepMemo *memo.Memo
-	inPreparePhase := p.EvalContext().PrepareOnly
-	if stmt.Prepared != nil {
-		// Don't use memo if it was never prepared, typically because of fallback
-		// to the heuristic planner.
-		if inPreparePhase || stmt.Prepared.Memo.RootGroup() != 0 {
-			prepMemo = &stmt.Prepared.Memo
-		}
-	}
-
-	// If this is the prepare phase, or if a prepared memo:
-	//   1. doesn't yet exist, or
-	//   2. it's been invalidated by schema or other changes
-	//
-	// Then entirely rebuild the memo from the AST.
-	if inPreparePhase || prepMemo == nil || prepMemo.IsStale(ctx, p.EvalContext(), &catalog) {
-		bld := optbuilder.New(ctx, &p.semaCtx, p.EvalContext(), &catalog, f, stmt.AST)
-		bld.KeepPlaceholders = prepMemo != nil
-		err := bld.Build()
-		if err != nil {
-			// isCorrelated is used in the fallback case to create a better error.
-			p.curPlan.isCorrelated = bld.IsCorrelated
-			return err
-		}
-
-		if prepMemo != nil {
-			// If the memo doesn't have placeholders, then fully optimize it, since
-			// it can be reused without further changes to build the execution tree.
-			if !f.Memo().HasPlaceholders() {
-				p.optimizer.Optimize()
-			}
-
-			// Update the prepared memo.
-			prepMemo.InitFrom(f.Memo())
-		}
-	}
-
-	// If in the PREPARE phase, construct a dummy plan that has correct output
-	// columns. Only output columns and placeholder types are needed.
-	if inPreparePhase {
-		mem := f.Memo()
-		md := mem.Metadata()
-		physical := mem.LookupPhysicalProps(mem.RootProps())
-		resultCols := make(sqlbase.ResultColumns, len(physical.Presentation))
-		for i, col := range physical.Presentation {
-			resultCols[i].Name = col.Label
-			resultCols[i].Typ = md.ColumnType(col.ID)
-		}
-		p.curPlan.plan = &zeroNode{columns: resultCols}
-		return nil
-	}
-
-	// This is the EXECUTE phase, so finish optimization by assigning any
-	// remaining placeholders and applying exploration rules.
-	var ev memo.ExprView
-	if prepMemo == nil {
-		ev = p.optimizer.Optimize()
-	} else {
-		if prepMemo.HasPlaceholders() {
-			// Assign placeholders in the prepared memo.
-			f.Memo().InitFrom(prepMemo)
-			err := f.AssignPlaceholders()
-			if err != nil {
-				return err
-			}
-			ev = p.optimizer.Optimize()
-		} else {
-			ev = prepMemo.Root()
-		}
-	}
-
-	// Build the plan tree and store it in planner.curPlan.
-	execFactory := makeExecFactory(p)
-	plan, err := execbuilder.New(&execFactory, ev, p.EvalContext()).Build()
-	if err != nil {
-		return err
-	}
-
-	p.curPlan = *plan.(*planTop)
-	// Since the assignment above just cleared the AST, we need to set it again.
-	p.curPlan.AST = stmt.AST
-
-	cols := planColumns(p.curPlan.plan)
-	if stmt.ExpectedTypes != nil {
-		if !stmt.ExpectedTypes.TypesEqual(cols) {
-			return pgerror.NewError(pgerror.CodeFeatureNotSupportedError,
-				"cached plan must not change result type")
-		}
-	}
-
-	return nil
-}
-
-// hideHiddenColumn ensures that if the plan is returning some hidden
+// hideHiddenColumns ensures that if the plan is returning some hidden
 // column(s), it is wrapped into a renderNode which only renders the
 // visible columns.
 func (p *planner) hideHiddenColumns(
@@ -528,6 +413,9 @@ func (p *planner) hideHiddenColumns(
 // close ensures that the plan's resources have been deallocated.
 func (p *planTop) close(ctx context.Context) {
 	if p.plan != nil {
+		if p.maybeSavePlan != nil && p.flags.IsSet(planFlagExecDone) {
+			p.savedPlanForStats = p.maybeSavePlan(ctx)
+		}
 		p.plan.Close(ctx)
 		p.plan = nil
 	}
@@ -542,14 +430,6 @@ func (p *planTop) close(ctx context.Context) {
 	}
 }
 
-// start starts the plan.
-func (p *planTop) start(params runParams) error {
-	if err := p.evalSubqueries(params); err != nil {
-		return err
-	}
-	return startPlan(params, p.plan)
-}
-
 // columns retrieves the plan's columns.
 func (p *planTop) columns() sqlbase.ResultColumns {
 	return planColumns(p.plan)
@@ -571,26 +451,6 @@ func (p *planTop) collectSpans(params runParams) (readSpans, writeSpans roachpb.
 	return readSpans, writeSpans, nil
 }
 
-// startPlan starts the given plan and all its sub-query nodes.
-func startPlan(params runParams, plan planNode) error {
-	// Now start execution.
-	if err := startExec(params, plan); err != nil {
-		return err
-	}
-
-	// Finally, trigger limit propagation through the plan.  The actual
-	// LIMIT values will have been evaluated by startExec().
-	params.p.setUnlimited(plan)
-
-	return nil
-}
-
-// execStartable is implemented by planNodes that have an initial
-// execution step.
-type execStartable interface {
-	startExec(params runParams) error
-}
-
 // autoCommitNode is implemented by planNodes that might be able to commit the
 // KV txn in which they operate. Some nodes might want to do this to take
 // advantage of the 1PC optimization in case they're running as an implicit
@@ -614,9 +474,8 @@ var _ autoCommitNode = &insertNode{}
 var _ autoCommitNode = &updateNode{}
 var _ autoCommitNode = &upsertNode{}
 
-// startExec calls startExec() on each planNode that supports
-// execStartable using a depth-first, post-order traversal.
-// The subqueries, if any, are also started.
+// startExec calls startExec() on each planNode using a depth-first, post-order
+// traversal.  The subqueries, if any, are also started.
 //
 // Reminder: walkPlan() ensures that subqueries and sub-plans are
 // started before startExec() is called.
@@ -630,16 +489,11 @@ func startExec(params runParams, plan planNode) error {
 			case *showTraceNode:
 				// showTrace needs to override the params struct, and does so in its startExec() method.
 				return false, nil
-			case *createStatsNode:
-				return false, errors.Errorf("statistics can only be created via DistSQL")
 			}
 			return true, nil
 		},
 		leaveNode: func(_ string, n planNode) error {
-			if s, ok := n.(execStartable); ok {
-				return s.startExec(params)
-			}
-			return nil
+			return n.startExec(params)
 		},
 	}
 	return walkPlan(params.ctx, plan, o)
@@ -652,9 +506,12 @@ func (p *planner) maybePlanHook(ctx context.Context, stmt tree.Statement) (planN
 	// upcoming IR work will provide unique numeric type tags, which will
 	// elegantly solve this.
 	for _, planHook := range planHooks {
-		if fn, header, subplans, err := planHook(ctx, stmt, p); err != nil {
+		if fn, header, subplans, avoidBuffering, err := planHook(ctx, stmt, p); err != nil {
 			return nil, err
 		} else if fn != nil {
+			if avoidBuffering {
+				p.curPlan.avoidBuffering = true
+			}
 			return &hookFnNode{f: fn, header: header, subplans: subplans}, nil
 		}
 	}
@@ -681,14 +538,14 @@ func (p *planner) delegateQuery(
 	initialCheck func(ctx context.Context) error,
 	desiredTypes []types.T,
 ) (planNode, error) {
-	// log.VEventf(ctx, 2, "delegated query: %q", sql)
+	log.VEventf(ctx, 2, "delegated query: %q", sql)
 
 	// Prepare the sub-plan.
 	stmt, err := parser.ParseOne(sql)
 	if err != nil {
 		return nil, err
 	}
-	plan, err := p.newPlan(ctx, stmt, desiredTypes)
+	plan, err := p.newPlan(ctx, stmt.AST, desiredTypes)
 	if err != nil {
 		return nil, err
 	}
@@ -736,7 +593,7 @@ func (p *planner) newPlan(
 	// schema-modifying statements that have no effect, such as
 	// `BEGIN; INSERT INTO ...; CREATE TABLE IF NOT EXISTS ...; COMMIT;`
 	// where the table already exists. This will generate some false
-	// refreshes, but that's expected to be quite rare in practice.
+	// schema cache refreshes, but that's expected to be quite rare in practice.
 	canModifySchema := tree.CanModifySchema(stmt)
 	if canModifySchema {
 		if err := p.txn.SetSystemConfigTrigger(); err != nil {
@@ -769,6 +626,12 @@ func (p *planner) newPlan(
 		return p.CancelQueries(ctx, n)
 	case *tree.CancelSessions:
 		return p.CancelSessions(ctx, n)
+	case *tree.CommentOnColumn:
+		return p.CommentOnColumn(ctx, n)
+	case *tree.CommentOnDatabase:
+		return p.CommentOnDatabase(ctx, n)
+	case *tree.CommentOnTable:
+		return p.CommentOnTable(ctx, n)
 	case *tree.ControlJobs:
 		return p.ControlJobs(ctx, n)
 	case *tree.Scrub:
@@ -878,6 +741,8 @@ func (p *planner) newPlan(
 		return p.ShowTables(ctx, n)
 	case *tree.ShowSchemas:
 		return p.ShowSchemas(ctx, n)
+	case *tree.ShowSequences:
+		return p.ShowSequences(ctx, n)
 	case *tree.ShowTraceForSession:
 		return p.ShowTrace(ctx, n)
 	case *tree.ShowTransactionStatus:
@@ -990,6 +855,10 @@ func (p *planner) doPrepare(ctx context.Context, stmt tree.Statement) (planNode,
 		return p.ShowJobs(ctx, n)
 	case *tree.ShowRoleGrants:
 		return p.ShowRoleGrants(ctx, n)
+	case *tree.ShowHistogram:
+		return p.ShowHistogram(ctx, n)
+	case *tree.ShowRanges:
+		return p.ShowRanges(ctx, n)
 	case *tree.ShowRoles:
 		return p.ShowRoles(ctx, n)
 	case *tree.ShowSessions:
@@ -998,14 +867,18 @@ func (p *planner) doPrepare(ctx context.Context, stmt tree.Statement) (planNode,
 		return p.ShowTables(ctx, n)
 	case *tree.ShowSchemas:
 		return p.ShowSchemas(ctx, n)
+	case *tree.ShowSequences:
+		return p.ShowSequences(ctx, n)
+	case *tree.ShowTableStats:
+		return p.ShowTableStats(ctx, n)
 	case *tree.ShowTraceForSession:
 		return p.ShowTrace(ctx, n)
 	case *tree.ShowUsers:
 		return p.ShowUsers(ctx, n)
 	case *tree.ShowTransactionStatus:
 		return p.ShowTransactionStatus(ctx)
-	case *tree.ShowRanges:
-		return p.ShowRanges(ctx, n)
+	case *tree.ShowZoneConfig:
+		return p.ShowZoneConfig(ctx, n)
 	case *tree.Split:
 		return p.Split(ctx, n)
 	case *tree.Truncate:
@@ -1033,3 +906,47 @@ func (p *planner) maybeSetSystemConfig(id sqlbase.ID) error {
 	// Mark transaction as operating on the system DB.
 	return p.txn.SetSystemConfigTrigger()
 }
+
+// planFlags is used throughout the planning code to keep track of various
+// events or decisions along the way.
+type planFlags uint32
+
+const (
+	// planFlagOptUsed is set if the optimizer was used to create the plan.
+	planFlagOptUsed planFlags = (1 << iota)
+
+	// planFlagIsCorrelated is set if the plan contained a correlated subquery.
+	// This is used to enhance the error fallback and for telemetry.
+	planFlagOptIsCorrelated
+
+	// planFlagOptFallback is set if the optimizer was enabled but did not support the
+	// statement.
+	planFlagOptFallback
+
+	// planFlagOptCacheHit is set if a plan from the query plan cache was used (and
+	// re-optimized).
+	planFlagOptCacheHit
+
+	// planFlagOptCacheMiss is set if we looked for a plan in the query plan cache but
+	// did not find one.
+	planFlagOptCacheMiss
+
+	// planFlagDistributed is set if the plan is for the DistSQL engine, in
+	// distributed mode.
+	planFlagDistributed
+
+	// planFlagDistSQLLocal is set if the plan is for the DistSQL engine,
+	// but in local mode.
+	planFlagDistSQLLocal
+
+	// planFlagExecDone marks that execution has been completed.
+	planFlagExecDone
+)
+
+func (pf planFlags) IsSet(flag planFlags) bool {
+	return (pf & flag) != 0
+}
+
+func (pf *planFlags) Set(flag planFlags) {
+	*pf |= flag
+}
