@@ -21,11 +21,13 @@ import (
 
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
-	"github.com/cockroachdb/cockroach/pkg/security"
+	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
-	"github.com/cockroachdb/cockroach/pkg/sql/opt/xform"
 	"github.com/cockroachdb/cockroach/pkg/sql/parser"
+	"github.com/cockroachdb/cockroach/pkg/sql/querycache"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/transform"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
@@ -74,6 +76,13 @@ type extendedEvalContext struct {
 	schemaAccessors *schemaInterface
 }
 
+// copy returns a deep copy of ctx.
+func (ctx *extendedEvalContext) copy() *extendedEvalContext {
+	cpy := *ctx
+	cpy.EvalContext = *ctx.EvalContext.Copy()
+	return &cpy
+}
+
 // schemaInterface provides access to the database and table descriptors.
 // See schema_accessors.go.
 type schemaInterface struct {
@@ -113,16 +122,11 @@ type planner struct {
 
 	// avoidCachedDescriptors, when true, instructs all code that
 	// accesses table/view descriptors to force reading the descriptors
-	// within the transaction. This is necessary to:
-	// - ensure that queries ran with AS OF SYSTEM TIME get the right
-	//   version of descriptors.
-	// - queries that create/update descriptors read all their dependencies
-	//   in the same txn that they write new descriptors or update their
-	//   dependencies, so that update/creation appears transactional
-	//   to the rest of the cluster.
-	// Code that sets this to true should probably also check that
-	// the txn isolation level is SERIALIZABLE, and reject any update
-	// if it is SNAPSHOT.
+	// within the transaction. This is necessary to read descriptors
+	// from the store for:
+	// 1. Descriptors that are part of a schema change but are not
+	// modified by the schema change. (reading a table in CREATE VIEW)
+	// 2. Disable the use of the table cache in tests.
 	avoidCachedDescriptors bool
 
 	// If set, the planner should skip checking for the SELECT privilege when
@@ -138,6 +142,12 @@ type planner struct {
 	// want to do 1PC transactions have to implement the autoCommitNode interface.
 	autoCommit bool
 
+	// discardRows is set if we want to discard any results rather than sending
+	// them back to the client. Used for testing/benchmarking. Note that the
+	// resulting schema or the plan are not affected.
+	// See EXECUTE .. DISCARD ROWS.
+	discardRows bool
+
 	// cancelChecker is used by planNodes to check for cancellation of the associated
 	// query.
 	cancelChecker *sqlbase.CancelChecker
@@ -151,7 +161,6 @@ type planner struct {
 	curPlan planTop
 
 	// Avoid allocations by embedding commonly used objects and visitors.
-	parser                parser.Parser
 	txCtx                 transform.ExprTransformContext
 	subqueryVisitor       subqueryVisitor
 	nameResolutionVisitor sqlbase.NameResolutionVisitor
@@ -163,9 +172,11 @@ type planner struct {
 	// be pool allocated.
 	alloc sqlbase.DatumAlloc
 
-	// optimizer caches an instance of the cost-based optimizer that can be reused
-	// to plan queries (reused in order to reduce allocations).
-	optimizer xform.Optimizer
+	// optPlanningCtx stores the optimizer planning context, which contains
+	// data structures that can be reused between queries (for efficiency).
+	optPlanningCtx optPlanningCtx
+
+	queryCacheSession querycache.Session
 }
 
 // noteworthyInternalMemoryUsageBytes is the minimum size tracked by each
@@ -212,17 +223,16 @@ func newInternalPlanner(
 	}
 	tables := &TableCollection{
 		leaseMgr:      execCfg.LeaseManager,
-		databaseCache: newDatabaseCache(config.SystemConfig{}),
+		databaseCache: newDatabaseCache(config.NewSystemConfig()),
 	}
-	txnReadOnly := new(bool)
 	dataMutator := &sessionDataMutator{
 		data: sd,
 		defaults: SessionDefaults(map[string]string{
 			"application_name": "crdb-internal",
 			"database":         "system",
 		}),
-		settings:       execCfg.Settings,
-		curTxnReadOnly: txnReadOnly,
+		settings:          execCfg.Settings,
+		setCurTxnReadOnly: func(bool) {},
 	}
 
 	var ts time.Time
@@ -240,7 +250,7 @@ func newInternalPlanner(
 	p.stmt = nil
 	p.cancelChecker = sqlbase.NewCancelChecker(ctx)
 
-	p.semaCtx = tree.MakeSemaContext(sd.User == security.RootUser /* privileged */)
+	p.semaCtx = tree.MakeSemaContext()
 	p.semaCtx.Location = &sd.DataConversion.Location
 	p.semaCtx.SearchPath = sd.SearchPath
 
@@ -254,9 +264,11 @@ func newInternalPlanner(
 		ctx, sd, dataMutator, tables, txn, ts, ts, execCfg, &plannerMon,
 	)
 	p.extendedEvalCtx.Planner = p
+	p.extendedEvalCtx.SessionAccessor = p
 	p.extendedEvalCtx.Sequence = p
 	p.extendedEvalCtx.ClusterID = execCfg.ClusterID()
 	p.extendedEvalCtx.NodeID = execCfg.NodeID.Get()
+	p.extendedEvalCtx.Locality = execCfg.Locality
 
 	p.sessionDataMutator = dataMutator
 	p.autoCommit = false
@@ -266,13 +278,11 @@ func newInternalPlanner(
 	p.extendedEvalCtx.Placeholders = &p.semaCtx.Placeholders
 	p.extendedEvalCtx.Tables = tables
 
-	acc := plannerMon.MakeBoundAccount()
-	p.extendedEvalCtx.ActiveMemAcc = &acc
+	p.queryCacheSession.Init()
 
 	return p, func() {
 		// Note that we capture ctx here. This is only valid as long as we create
 		// the context as explained at the top of the method.
-		acc.Close(ctx)
 		plannerMon.Stop(ctx)
 	}
 }
@@ -301,7 +311,7 @@ func internalExtendedEvalCtx(
 		EvalContext: tree.EvalContext{
 			Txn:           txn,
 			SessionData:   sd,
-			TxnReadOnly:   *dataMutator.curTxnReadOnly,
+			TxnReadOnly:   false,
 			TxnImplicit:   true,
 			Settings:      execCfg.Settings,
 			Context:       ctx,
@@ -329,10 +339,15 @@ func (p *planner) LogicalSchemaAccessor() SchemaAccessor {
 	return p.extendedEvalCtx.schemaAccessors.logical
 }
 
+// Note: if the context will be modified, use ExtendedEvalContextCopy instead.
 func (p *planner) ExtendedEvalContext() *extendedEvalContext {
 	return &p.extendedEvalCtx
 }
 
+func (p *planner) ExtendedEvalContextCopy() *extendedEvalContext {
+	return p.extendedEvalCtx.copy()
+}
+
 func (p *planner) CurrentDatabase() string {
 	return p.SessionData().Database
 }
@@ -391,19 +406,19 @@ func (p *planner) ResolveTableName(ctx context.Context, tn *tree.TableName) erro
 	return err
 }
 
-func (p *planner) lookupFKTable(
-	ctx context.Context, tableID sqlbase.ID,
-) (sqlbase.TableLookup, error) {
-	flags := ObjectLookupFlags{
-		CommonLookupFlags{txn: p.txn, avoidCached: p.avoidCachedDescriptors}}
-	table, err := p.Tables().getTableVersionByID(ctx, tableID, flags)
+// LookupTableByID looks up a table, by the given descriptor ID. Based on the
+// CommonLookupFlags, it could use or skip the TableCollection cache. See
+// TableCollection.getTableVersionByID for how it's used.
+func (p *planner) LookupTableByID(ctx context.Context, tableID sqlbase.ID) (row.TableEntry, error) {
+	flags := ObjectLookupFlags{CommonLookupFlags: CommonLookupFlags{avoidCached: p.avoidCachedDescriptors}}
+	table, err := p.Tables().getTableVersionByID(ctx, p.txn, tableID, flags)
 	if err != nil {
 		if err == errTableAdding {
-			return sqlbase.TableLookup{IsAdding: true}, nil
+			return row.TableEntry{IsAdding: true}, nil
 		}
-		return sqlbase.TableLookup{}, err
+		return row.TableEntry{}, err
 	}
-	return sqlbase.TableLookup{Table: table}, nil
+	return row.TableEntry{Desc: table}, nil
 }
 
 // TypeAsString enforces (not hints) that the given expression typechecks as a
@@ -533,36 +548,67 @@ func (p *planner) prepareForDistSQLSupportCheck() {
 	p.setUnlimited(p.curPlan.plan)
 }
 
-// optionallyUseOptimizer will attempt to make an optimizer plan based on the
-// optimizerMode setting. If it is run, it will return true. If it returns false
-// and no error is returned, it is safe to fallback to a non-optimizer plan.
-func (p *planner) optionallyUseOptimizer(
-	ctx context.Context, sd sessiondata.SessionData, stmt Statement,
-) (bool, error) {
-	if sd.OptimizerMode == sessiondata.OptimizerOff {
-		log.VEvent(ctx, 1, "optimizer disabled")
-		return false, nil
+// runWithDistSQL runs a planNode tree synchronously via DistSQL, returning the
+// results in a RowContainer. There's no streaming on this, so use sparingly.
+// In general, you should always prefer to use the internal executor if you can.
+func (p *planner) runWithDistSQL(
+	ctx context.Context, plan planNode,
+) (*rowcontainer.RowContainer, error) {
+	params := runParams{
+		ctx:             ctx,
+		extendedEvalCtx: &p.extendedEvalCtx,
+		p:               p,
 	}
+	// Create the DistSQL plan for the input.
+	planCtx := params.extendedEvalCtx.DistSQLPlanner.NewPlanningCtx(ctx, params.extendedEvalCtx, params.p.txn)
+	log.VEvent(ctx, 1, "creating DistSQL plan")
+	physPlan, err := planCtx.ExtendedEvalCtx.DistSQLPlanner.createPlanForNode(planCtx, plan)
+	if err != nil {
+		return nil, err
+	}
+	planCtx.ExtendedEvalCtx.DistSQLPlanner.FinalizePlan(planCtx, &physPlan)
+	columns := planColumns(plan)
 
-	log.VEvent(ctx, 1, "generating optimizer plan")
+	// Initialize a row container for the DistSQL execution engine to write into.
+	// The caller of this method will call Close on the returned RowContainer,
+	// which will close this account.
+	acc := planCtx.EvalContext().Mon.MakeBoundAccount()
+	ci := sqlbase.ColTypeInfoFromResCols(columns)
+	rows := rowcontainer.NewRowContainer(acc, ci, 0 /* rowCapacity */)
+	rowResultWriter := NewRowResultWriter(rows)
+	recv := MakeDistSQLReceiver(
+		ctx,
+		rowResultWriter,
+		tree.Rows,
+		p.ExecCfg().RangeDescriptorCache,
+		p.ExecCfg().LeaseHolderCache,
+		p.txn,
+		func(ts hlc.Timestamp) {
+			_ = p.ExecCfg().Clock.Update(ts)
+		},
+		p.extendedEvalCtx.Tracing,
+	)
+	defer recv.Release()
 
-	err := p.makeOptimizerPlan(ctx, stmt)
-	if err == nil {
-		log.VEvent(ctx, 1, "optimizer plan succeeded")
-		return true, nil
+	// Copy the evalCtx, as dsp.Run() might change it.
+	evalCtxCopy := p.extendedEvalCtx
+	// Run the plan, writing to the row container we initialized earlier.
+	p.extendedEvalCtx.DistSQLPlanner.Run(
+		planCtx, p.txn, &physPlan, recv, &evalCtxCopy, nil /* finishedSetupFn */)
+	if rowResultWriter.Err() != nil {
+		rows.Close(ctx)
+		return nil, rowResultWriter.Err()
 	}
-	log.VEventf(ctx, 1, "optimizer plan failed: %v", err)
-	if canFallbackFromOpt(err, sd.OptimizerMode, stmt) {
-		log.VEvent(ctx, 1, "optimizer falls back on heuristic planner")
-		return false, nil
-	}
-	return false, err
+	return rows, nil
 }
 
 // txnModesSetter is an interface used by SQL execution to influence the current
 // transaction.
 type txnModesSetter interface {
-	setTransactionModes(modes tree.TransactionModes) error
+	// setTransactionModes updates some characteristics of the current
+	// transaction.
+	// asOfTs, if not empty, is the evaluation of modes.AsOf.
+	setTransactionModes(modes tree.TransactionModes, asOfTs hlc.Timestamp) error
 }
 
 // sqlStatsCollector is the interface used by SQL execution, through the
@@ -574,8 +620,11 @@ type sqlStatsCollector interface {
 	PhaseTimes() *phaseTimes
 
 	// RecordStatement record stats for one statement.
+	//
+	// samplePlanDescription can be nil, as these are only sampled periodically per unique fingerprint.
 	RecordStatement(
-		stmt Statement,
+		stmt *Statement,
+		samplePlanDescription *roachpb.ExplainTreePlanNode,
 		distSQLUsed bool,
 		optUsed bool,
 		automaticRetryCount int,
