@@ -80,14 +80,12 @@ type renderNode struct {
 	// modified by index selection.
 	props physicalProps
 
-	run renderRun
-
 	// This struct must be allocated on the heap and its location stay
 	// stable after construction because it implements
 	// IndexedVarContainer and the IndexedVar objects in sub-expressions
 	// will link to it by reference after checkRenderStar / analyzeExpr.
 	// Enforce this using NoCopy.
-	noCopy util.NoCopy
+	_ util.NoCopy
 }
 
 // Select selects rows from a SELECT/UNION/VALUES, ordering and/or limiting them.
@@ -344,7 +342,7 @@ func (p *planner) SelectClause(
 
 // IndexedVarEval implements the tree.IndexedVarContainer interface.
 func (r *renderNode) IndexedVarEval(idx int, ctx *tree.EvalContext) (tree.Datum, error) {
-	return r.run.curSourceRow[idx].Eval(ctx)
+	panic("renderNode can't be run in local mode")
 }
 
 // IndexedVarResolvedType implements the tree.IndexedVarContainer interface.
@@ -357,29 +355,18 @@ func (r *renderNode) IndexedVarNodeFormatter(idx int) tree.NodeFormatter {
 	return r.sourceInfo[0].NodeFormatter(idx)
 }
 
-// renderRun contains the run-time state of renderNode during local execution.
-type renderRun struct {
-	// The current source row, with one value per source column.
-	// populated by Next(), used by renderRow().
-	curSourceRow tree.Datums
-
-	// The rendered row, with one value for each render expression.
-	// populated by Next().
-	row tree.Datums
+func (r *renderNode) startExec(runParams) error {
+	panic("renderNode can't be run in local mode")
 }
 
 func (r *renderNode) Next(params runParams) (bool, error) {
-	if next, err := r.source.plan.Next(params); !next {
-		return false, err
-	}
+	panic("renderNode can't be run in local mode")
+}
 
-	r.run.curSourceRow = r.source.plan.Values()
-
-	err := r.renderRow(params.EvalContext())
-	return err == nil, err
+func (r *renderNode) Values() tree.Datums {
+	panic("renderNode can't be run in local mode")
 }
 
-func (r *renderNode) Values() tree.Datums       { return r.run.row }
 func (r *renderNode) Close(ctx context.Context) { r.source.plan.Close(ctx) }
 
 // initFrom initializes the table node, given the parsed select expression
@@ -490,20 +477,22 @@ func (p *planner) getTimestamp(asOf tree.AsOfClause) (hlc.Timestamp, bool, error
 		// would not be set globally for the entire txn.
 		if p.semaCtx.AsOfTimestamp == nil {
 			return hlc.MaxTimestamp, false,
-				fmt.Errorf("AS OF SYSTEM TIME must be provided on a top-level statement")
+				pgerror.NewErrorf(pgerror.CodeSyntaxError,
+					"AS OF SYSTEM TIME must be provided on a top-level statement")
 		}
 
 		// The Executor found an AS OF SYSTEM TIME clause at the top
 		// level. We accept AS OF SYSTEM TIME in multiple places (e.g. in
 		// subqueries or view queries) but they must all point to the same
 		// timestamp.
-		ts, err := p.EvalAsOfTimestamp(asOf, hlc.MaxTimestamp)
+		ts, err := p.EvalAsOfTimestamp(asOf)
 		if err != nil {
 			return hlc.MaxTimestamp, false, err
 		}
 		if ts != *p.semaCtx.AsOfTimestamp {
 			return hlc.MaxTimestamp, false,
-				fmt.Errorf("cannot specify AS OF SYSTEM TIME with different timestamps")
+				pgerror.UnimplementedWithIssueError(35712,
+					"cannot specify AS OF SYSTEM TIME with different timestamps")
 		}
 		return ts, true, nil
 	}
@@ -562,22 +551,6 @@ func (r *renderNode) resetRenderColumns(exprs []tree.TypedExpr, cols sqlbase.Res
 	r.columns = cols
 }
 
-// renderRow renders the row by evaluating the render expressions.
-func (r *renderNode) renderRow(evalCtx *tree.EvalContext) error {
-	if r.run.row == nil {
-		r.run.row = make([]tree.Datum, len(r.render))
-	}
-	evalCtx.IVarContainer = r
-	for i, e := range r.render {
-		var err error
-		r.run.row[i], err = e.Eval(evalCtx)
-		if err != nil {
-			return err
-		}
-	}
-	return nil
-}
-
 // computePhysicalPropsForRender computes ordering information for the
 // render node, given ordering information for the "from" node.
 //
@@ -648,7 +621,7 @@ func (r *renderNode) colIdxByRenderAlias(
 			return 0, err
 		}
 
-		if c, ok := v.(*tree.ColumnItem); ok && c.TableName.Parts[0] == "" {
+		if c, ok := v.(*tree.ColumnItem); ok && c.TableName == nil {
 			// Look for an output column that matches the name. This
 			// handles cases like:
 			//
