@@ -21,6 +21,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/privilege"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
@@ -41,8 +42,9 @@ type SchemaResolver interface {
 	LogicalSchemaAccessor() SchemaAccessor
 	CurrentDatabase() string
 	CurrentSearchPath() sessiondata.SearchPath
-	CommonLookupFlags(ctx context.Context, required bool) CommonLookupFlags
-	ObjectLookupFlags(ctx context.Context, required bool) ObjectLookupFlags
+	CommonLookupFlags(required bool) CommonLookupFlags
+	ObjectLookupFlags(required bool, requireMutable bool) ObjectLookupFlags
+	LookupTableByID(ctx context.Context, id sqlbase.ID) (row.TableEntry, error)
 }
 
 var _ SchemaResolver = &planner{}
@@ -61,8 +63,8 @@ func (p *planner) ResolveUncachedDatabaseByName(
 	ctx context.Context, dbName string, required bool,
 ) (res *UncachedDatabaseDescriptor, err error) {
 	p.runWithOptions(resolveFlags{skipCache: true}, func() {
-		res, err = p.LogicalSchemaAccessor().GetDatabaseDesc(dbName,
-			p.CommonLookupFlags(ctx, required))
+		res, err = p.LogicalSchemaAccessor().GetDatabaseDesc(ctx, p.txn, dbName,
+			p.CommonLookupFlags(required))
 	})
 	return res, err
 }
@@ -70,14 +72,15 @@ func (p *planner) ResolveUncachedDatabaseByName(
 // GetObjectNames retrieves the names of all objects in the target database/schema.
 func GetObjectNames(
 	ctx context.Context,
+	txn *client.Txn,
 	sc SchemaResolver,
 	dbDesc *DatabaseDescriptor,
 	scName string,
 	explicitPrefix bool,
 ) (res TableNames, err error) {
-	return sc.LogicalSchemaAccessor().GetObjectNames(dbDesc, scName,
+	return sc.LogicalSchemaAccessor().GetObjectNames(ctx, txn, dbDesc, scName,
 		DatabaseListFlags{
-			CommonLookupFlags: sc.CommonLookupFlags(ctx, true /*required*/),
+			CommonLookupFlags: sc.CommonLookupFlags(true /*required*/),
 			explicitPrefix:    explicitPrefix,
 		})
 }
@@ -91,8 +94,40 @@ func GetObjectNames(
 // if no object is found.
 func ResolveExistingObject(
 	ctx context.Context, sc SchemaResolver, tn *ObjectName, required bool, requiredType requiredType,
-) (res *ObjectDescriptor, err error) {
-	found, descI, err := tn.ResolveExisting(ctx, sc, sc.CurrentDatabase(), sc.CurrentSearchPath())
+) (res *ImmutableTableDescriptor, err error) {
+	desc, err := resolveExistingObjectImpl(ctx, sc, tn, required, false /* requiredMutable */, requiredType)
+	if err != nil || desc == nil {
+		return nil, err
+	}
+	return desc.(*ImmutableTableDescriptor), nil
+}
+
+// ResolveMutableExistingObject looks up an existing mutable object.
+// If required is true, an error is returned if the object does not exist.
+// Optionally, if a desired descriptor type is specified, that type is checked.
+//
+// The object name is modified in-place with the result of the name
+// resolution, if successful. It is not modified in case of error or
+// if no object is found.
+func ResolveMutableExistingObject(
+	ctx context.Context, sc SchemaResolver, tn *ObjectName, required bool, requiredType requiredType,
+) (res *MutableTableDescriptor, err error) {
+	desc, err := resolveExistingObjectImpl(ctx, sc, tn, required, true /* requiredMutable */, requiredType)
+	if err != nil || desc == nil {
+		return nil, err
+	}
+	return desc.(*MutableTableDescriptor), nil
+}
+
+func resolveExistingObjectImpl(
+	ctx context.Context,
+	sc SchemaResolver,
+	tn *ObjectName,
+	required bool,
+	requiredMutable bool,
+	requiredType requiredType,
+) (res tree.NameResolutionResult, err error) {
+	found, descI, err := tn.ResolveExisting(ctx, sc, requiredMutable, sc.CurrentDatabase(), sc.CurrentSearchPath())
 	if err != nil {
 		return nil, err
 	}
@@ -102,22 +137,28 @@ func ResolveExistingObject(
 		}
 		return nil, nil
 	}
-	desc := descI.(*ObjectDescriptor)
+	obj := descI.(ObjectDescriptor)
+
 	goodType := true
 	switch requiredType {
 	case requireTableDesc:
-		goodType = desc.IsTable()
+		goodType = obj.TableDesc().IsTable()
 	case requireViewDesc:
-		goodType = desc.IsView()
+		goodType = obj.TableDesc().IsView()
 	case requireTableOrViewDesc:
-		goodType = desc.IsTable() || desc.IsView()
+		goodType = obj.TableDesc().IsTable() || obj.TableDesc().IsView()
 	case requireSequenceDesc:
-		goodType = desc.IsSequence()
+		goodType = obj.TableDesc().IsSequence()
 	}
 	if !goodType {
 		return nil, sqlbase.NewWrongObjectTypeError(tn, requiredTypeNames[requiredType])
 	}
-	return desc, nil
+
+	if requiredMutable {
+		return descI.(*MutableTableDescriptor), nil
+	}
+
+	return descI.(*ImmutableTableDescriptor), nil
 }
 
 // runWithOptions sets the provided resolution flags for the
@@ -147,15 +188,12 @@ type resolveFlags struct {
 func (p *planner) ResolveMutableTableDescriptor(
 	ctx context.Context, tn *ObjectName, required bool, requiredType requiredType,
 ) (table *MutableTableDescriptor, err error) {
-	p.runWithOptions(resolveFlags{skipCache: true}, func() {
-		table, err = ResolveExistingObject(ctx, p, tn, required, requiredType)
-	})
-	return table, err
+	return ResolveMutableExistingObject(ctx, p, tn, required, requiredType)
 }
 
 func (p *planner) ResolveUncachedTableDescriptor(
 	ctx context.Context, tn *ObjectName, required bool, requiredType requiredType,
-) (table *UncachedTableDescriptor, err error) {
+) (table *ImmutableTableDescriptor, err error) {
 	p.runWithOptions(resolveFlags{skipCache: true}, func() {
 		table, err = ResolveExistingObject(ctx, p, tn, required, requiredType)
 	})
@@ -220,7 +258,7 @@ func (p *planner) LookupSchema(
 	ctx context.Context, dbName, scName string,
 ) (found bool, scMeta tree.SchemaMeta, err error) {
 	sc := p.LogicalSchemaAccessor()
-	dbDesc, err := sc.GetDatabaseDesc(dbName, p.CommonLookupFlags(ctx, false /*required*/))
+	dbDesc, err := sc.GetDatabaseDesc(ctx, p.txn, dbName, p.CommonLookupFlags(false /*required*/))
 	if err != nil || dbDesc == nil {
 		return false, nil, err
 	}
@@ -229,27 +267,25 @@ func (p *planner) LookupSchema(
 
 // LookupObject implements the tree.TableNameExistingResolver interface.
 func (p *planner) LookupObject(
-	ctx context.Context, dbName, scName, tbName string,
+	ctx context.Context, requireMutable bool, dbName, scName, tbName string,
 ) (found bool, objMeta tree.NameResolutionResult, err error) {
 	sc := p.LogicalSchemaAccessor()
 	p.tableName = tree.MakeTableNameWithSchema(tree.Name(dbName), tree.Name(scName), tree.Name(tbName))
-	objDesc, _, err := sc.GetObjectDesc(&p.tableName, p.ObjectLookupFlags(ctx, false /*required*/))
+	objDesc, err := sc.GetObjectDesc(ctx, p.txn, &p.tableName, p.ObjectLookupFlags(false /*required*/, requireMutable))
 	return objDesc != nil, objDesc, err
 }
 
-func (p *planner) CommonLookupFlags(ctx context.Context, required bool) CommonLookupFlags {
+func (p *planner) CommonLookupFlags(required bool) CommonLookupFlags {
 	return CommonLookupFlags{
-		ctx:            ctx,
-		txn:            p.txn,
-		required:       required,
-		avoidCached:    p.avoidCachedDescriptors,
-		fixedTimestamp: p.semaCtx.AsOfTimestamp != nil,
+		required:    required,
+		avoidCached: p.avoidCachedDescriptors,
 	}
 }
 
-func (p *planner) ObjectLookupFlags(ctx context.Context, required bool) ObjectLookupFlags {
+func (p *planner) ObjectLookupFlags(required, requireMutable bool) ObjectLookupFlags {
 	return ObjectLookupFlags{
-		CommonLookupFlags: p.CommonLookupFlags(ctx, required),
+		CommonLookupFlags: p.CommonLookupFlags(required),
+		requireMutable:    requireMutable,
 	}
 }
 
@@ -284,13 +320,12 @@ func getDescriptorsFromTargetList(
 		if err != nil {
 			return nil, err
 		}
-		tableNames, err := expandTableGlob(ctx, p, tableGlob)
+		tableNames, err := expandTableGlob(ctx, p.txn, p, tableGlob)
 		if err != nil {
 			return nil, err
 		}
 		for i := range tableNames {
-			descriptor, _, err := p.LogicalSchemaAccessor().GetObjectDesc(&tableNames[i],
-				p.ObjectLookupFlags(ctx, true /*required*/))
+			descriptor, err := ResolveMutableExistingObject(ctx, p, &tableNames[i], true, anyDescType)
 			if err != nil {
 				return nil, err
 			}
@@ -326,28 +361,29 @@ func (p *planner) getQualifiedTableName(
 // error will be returned, otherwise the TableName and descriptor
 // returned will be nil.
 func findTableContainingIndex(
-	sc SchemaAccessor,
+	ctx context.Context,
+	txn *client.Txn,
+	sc SchemaResolver,
 	dbName, scName string,
 	idxName tree.UnrestrictedName,
 	lookupFlags CommonLookupFlags,
 ) (result *tree.TableName, desc *MutableTableDescriptor, err error) {
-	dbDesc, err := sc.GetDatabaseDesc(dbName, lookupFlags)
+	sa := sc.LogicalSchemaAccessor()
+	dbDesc, err := sa.GetDatabaseDesc(ctx, txn, dbName, lookupFlags)
 	if dbDesc == nil || err != nil {
 		return nil, nil, err
 	}
 
-	tns, err := sc.GetObjectNames(dbDesc, scName,
+	tns, err := sa.GetObjectNames(ctx, txn, dbDesc, scName,
 		DatabaseListFlags{CommonLookupFlags: lookupFlags, explicitPrefix: true})
 	if err != nil {
 		return nil, nil, err
 	}
 
 	result = nil
-	tblLookupFlags := ObjectLookupFlags{CommonLookupFlags: lookupFlags}
-	tblLookupFlags.required = false
 	for i := range tns {
 		tn := &tns[i]
-		tableDesc, _, err := sc.GetObjectDesc(tn, tblLookupFlags)
+		tableDesc, err := ResolveMutableExistingObject(ctx, sc, tn, true, anyDescType)
 		if err != nil {
 			return nil, nil, err
 		}
@@ -386,69 +422,61 @@ func findTableContainingIndex(
 // on the lookup path. This can be used in the caller to avoid a 2nd
 // lookup.
 func expandMutableIndexName(
-	ctx context.Context, p *planner, index *tree.TableNameWithIndex, requireTable bool,
+	ctx context.Context, p *planner, index *tree.TableIndexName, requireTable bool,
 ) (tn *tree.TableName, desc *MutableTableDescriptor, err error) {
 	p.runWithOptions(resolveFlags{skipCache: true}, func() {
-		tn, desc, err = expandIndexName(ctx, p, index, requireTable)
+		tn, desc, err = expandIndexName(ctx, p.txn, p, index, requireTable)
 	})
 	return tn, desc, err
 }
 
 func expandIndexName(
-	ctx context.Context, sc SchemaResolver, index *tree.TableNameWithIndex, requireTable bool,
+	ctx context.Context,
+	txn *client.Txn,
+	sc SchemaResolver,
+	index *tree.TableIndexName,
+	requireTable bool,
 ) (tn *tree.TableName, desc *MutableTableDescriptor, err error) {
-	tn, err = index.Table.Normalize()
+	tn = &index.Table
+	if tn.Table() != "" {
+		// The index and its table prefix must exist already. Resolve the table.
+		desc, err = ResolveMutableExistingObject(ctx, sc, tn, requireTable, requireTableDesc)
+		if err != nil {
+			return nil, nil, err
+		}
+		return tn, desc, nil
+	}
+
+	// On the first call to expandMutableIndexName(), index.Table.Table() is empty.
+	// Once the table name is resolved for the index below, index.Table
+	// references the table name.
+
+	// Look up the table prefix.
+	found, _, err := tn.TableNamePrefix.Resolve(ctx, sc, sc.CurrentDatabase(), sc.CurrentSearchPath())
+	if err != nil {
+		return nil, nil, err
+	}
+	if !found {
+		if requireTable {
+			return nil, nil, pgerror.NewErrorf(pgerror.CodeUndefinedObjectError,
+				"schema or database was not found while searching index: %q",
+				tree.ErrString(&index.Index)).SetHintf(
+				"check the current database and search_path are valid")
+		}
+		return nil, nil, nil
+	}
+
+	lookupFlags := sc.CommonLookupFlags(requireTable)
+	var foundTn *tree.TableName
+	foundTn, desc, err = findTableContainingIndex(ctx, txn, sc, tn.Catalog(), tn.Schema(), index.Index, lookupFlags)
 	if err != nil {
 		return nil, nil, err
 	}
 
-	if !index.SearchTable {
-		// The index and its table prefix must exist already. Resolve the table.
-		desc, err = ResolveExistingObject(ctx, sc, tn, requireTable, requireTableDesc)
-		if err != nil {
-			return nil, nil, err
-		}
-	} else {
-		// On the first call to expandMutableIndexName(), index.SearchTable is
-		// true, index.Index is empty and tn.Table() is the index
-		// name. Once the table name is resolved for the index below,
-		// index.Table references a new table name (not the index), so a
-		// subsequent call to expandMutableIndexName() will generate tn using the
-		// new value of index.Table, which is a table name.
-
-		// Just an assertion: if we got there, there cannot be a value in index.Index yet.
-		if index.Index != "" {
-			return nil, nil, pgerror.NewAssertionErrorf("programmer error: not-searched index name found already qualified: %s@%s", tn, index.Index)
-		}
-
-		index.Index = tree.UnrestrictedName(tn.TableName)
-
-		// Look up the table prefix.
-		found, _, err := tn.TableNamePrefix.Resolve(ctx, sc, sc.CurrentDatabase(), sc.CurrentSearchPath())
-		if err != nil {
-			return nil, nil, err
-		}
-		if !found {
-			if requireTable {
-				return nil, nil, pgerror.NewErrorf(pgerror.CodeUndefinedObjectError,
-					"database or schema was not found while searching for index %q",
-					tree.ErrString(&index.Index)).SetHintf(
-					"verify that the current database and search_path are valid")
-			}
-			return nil, nil, nil
-		}
-
-		lookupFlags := sc.CommonLookupFlags(ctx, requireTable)
-		var foundTn *tree.TableName
-		foundTn, desc, err = findTableContainingIndex(
-			sc.LogicalSchemaAccessor(), tn.Catalog(), tn.Schema(), index.Index, lookupFlags)
-		if err != nil {
-			return nil, nil, err
-		} else if foundTn != nil {
-			// Memoize the table name that was found. tn is a reference to the table name
-			// stored in index.Table.
-			*tn = *foundTn
-		}
+	if foundTn != nil {
+		// Memoize the table name that was found. tn is a reference to the table name
+		// stored in index.Table.
+		*tn = *foundTn
 	}
 	return tn, desc, nil
 }
@@ -460,20 +488,17 @@ func expandIndexName(
 // It can return indexes that are being rolled out.
 func (p *planner) getTableAndIndex(
 	ctx context.Context,
-	table *tree.NormalizableTableName,
-	tableWithIndex *tree.TableNameWithIndex,
+	table *tree.TableName,
+	tableWithIndex *tree.TableIndexName,
 	privilege privilege.Kind,
-) (*sqlbase.TableDescriptor, *sqlbase.IndexDescriptor, error) {
-	var tn *tree.TableName
+) (*MutableTableDescriptor, *sqlbase.IndexDescriptor, error) {
 	var tableDesc *MutableTableDescriptor
 	var err error
 	if tableWithIndex == nil {
 		// Variant: ALTER TABLE
-		tn, err = table.Normalize()
-		if err != nil {
-			return nil, nil, err
-		}
-		tableDesc, err = p.ResolveMutableTableDescriptor(ctx, tn, true /*required*/, requireTableDesc)
+		tableDesc, err = p.ResolveMutableTableDescriptor(
+			ctx, table, true /*required*/, requireTableDesc,
+		)
 	} else {
 		// Variant: ALTER INDEX
 		_, tableDesc, err = expandMutableIndexName(ctx, p, tableWithIndex, true /* requireTable */)
@@ -487,9 +512,9 @@ func (p *planner) getTableAndIndex(
 	}
 
 	// Determine which index to use.
-	var index sqlbase.IndexDescriptor
+	var index *sqlbase.IndexDescriptor
 	if tableWithIndex == nil {
-		index = tableDesc.PrimaryIndex
+		index = &tableDesc.PrimaryIndex
 	} else {
 		idx, dropped, err := tableDesc.FindIndexByName(string(tableWithIndex.Index))
 		if err != nil {
@@ -500,13 +525,13 @@ func (p *planner) getTableAndIndex(
 		}
 		index = idx
 	}
-	return tableDesc, &index, nil
+	return tableDesc, index, nil
 }
 
 // expandTableGlob expands pattern into a list of tables represented
 // as a tree.TableNames.
 func expandTableGlob(
-	ctx context.Context, sc SchemaResolver, pattern tree.TablePattern,
+	ctx context.Context, txn *client.Txn, sc SchemaResolver, pattern tree.TablePattern,
 ) (tree.TableNames, error) {
 	if t, ok := pattern.(*tree.TableName); ok {
 		_, err := ResolveExistingObject(ctx, sc, t, true /*required*/, anyDescType)
@@ -526,7 +551,7 @@ func expandTableGlob(
 		return nil, sqlbase.NewInvalidWildcardError(tree.ErrString(glob))
 	}
 
-	return GetObjectNames(ctx, sc, descI.(*DatabaseDescriptor), glob.Schema(), glob.ExplicitSchema)
+	return GetObjectNames(ctx, txn, sc, descI.(*DatabaseDescriptor), glob.Schema(), glob.ExplicitSchema)
 }
 
 // fkSelfResolver is a SchemaResolver that inserts itself between a
@@ -544,14 +569,18 @@ var _ SchemaResolver = &fkSelfResolver{}
 
 // LookupObject implements the tree.TableNameExistingResolver interface.
 func (r *fkSelfResolver) LookupObject(
-	ctx context.Context, dbName, scName, tbName string,
+	ctx context.Context, requireMutable bool, dbName, scName, tbName string,
 ) (found bool, objMeta tree.NameResolutionResult, err error) {
 	if dbName == r.newTableName.Catalog() &&
 		scName == r.newTableName.Schema() &&
 		tbName == r.newTableName.Table() {
-		return true, r.newTableDesc, nil
+		table := r.newTableDesc
+		if requireMutable {
+			return true, sqlbase.NewMutableExistingTableDescriptor(*table), nil
+		}
+		return true, sqlbase.NewImmutableTableDescriptor(*table), nil
 	}
-	return r.SchemaResolver.LookupObject(ctx, dbName, scName, tbName)
+	return r.SchemaResolver.LookupObject(ctx, requireMutable, dbName, scName, tbName)
 }
 
 // internalLookupCtx can be used in contexts where all descriptors
