@@ -22,8 +22,6 @@ import (
 	"strings"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
@@ -32,12 +30,14 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/kv"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqlutil"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/grpcutil"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
@@ -47,6 +47,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 var schemaChangeLeaseDuration = settings.RegisterNonNegativeDurationSetting(
@@ -58,7 +59,7 @@ var schemaChangeLeaseDuration = settings.RegisterNonNegativeDurationSetting(
 var schemaChangeLeaseRenewFraction = settings.RegisterFloatSetting(
 	"schemachanger.lease.renew_fraction",
 	"the fraction of schemachanger.lease_duration remaining to trigger a renew of the lease",
-	0.4,
+	0.5,
 )
 
 // This is a delay [0.9 * asyncSchemaChangeDelay, 1.1 * asyncSchemaChangeDelay)
@@ -78,6 +79,12 @@ var schemaChangeLeaseRenewFraction = settings.RegisterFloatSetting(
 // attempting to become the job coordinator.
 const asyncSchemaChangeDelay = 1 * time.Minute
 
+type droppedIndex struct {
+	indexID  sqlbase.IndexID
+	dropTime int64
+	deadline int64
+}
+
 // SchemaChanger is used to change the schema on a table.
 type SchemaChanger struct {
 	tableID    sqlbase.ID
@@ -92,6 +99,8 @@ type SchemaChanger struct {
 	// table.DropTime.
 	dropTime int64
 
+	dropIndexTimes []droppedIndex
+
 	testingKnobs   *SchemaChangerTestingKnobs
 	distSQLPlanner *DistSQLPlanner
 	jobRegistry    *jobs.Registry
@@ -107,6 +116,7 @@ type SchemaChanger struct {
 	clock                *hlc.Clock
 	settings             *cluster.Settings
 	execCfg              *ExecutorConfig
+	ieFactory            sqlutil.SessionBoundInternalExecutorFactory
 }
 
 // NewSchemaChangerForTesting only for tests.
@@ -158,7 +168,7 @@ func isPermanentSchemaChangeError(err error) bool {
 	// Ignore error thrown because of a read at a very old timestamp.
 	// The Backfill will grab a new timestamp to read at for the rest
 	// of the backfill.
-	if strings.Contains(err.Error(), "must be after GC threshold") {
+	if strings.Contains(err.Error(), "must be after replica GC threshold") {
 		return false
 	}
 
@@ -182,10 +192,9 @@ func isPermanentSchemaChangeError(err error) bool {
 			return false
 
 		case pgerror.CodeInternalError:
-			if err.Message == context.DeadlineExceeded.Error() {
+			if strings.Contains(err.Message, context.DeadlineExceeded.Error()) {
 				return false
 			}
-
 		}
 	}
 
@@ -326,15 +335,19 @@ func (sc *SchemaChanger) ExtendLease(
 	return nil
 }
 
+func (sc *SchemaChanger) canClearRangeForDrop(index *sqlbase.IndexDescriptor) bool {
+	return sc.execCfg.Settings.Version.IsActive(cluster.VersionClearRange) && !index.IsInterleaved()
+}
+
 // DropTableDesc removes a descriptor from the KV database.
-func DropTableDesc(
-	ctx context.Context, tableDesc *sqlbase.TableDescriptor, db *client.DB, traceKV bool,
+func (sc *SchemaChanger) DropTableDesc(
+	ctx context.Context, tableDesc *sqlbase.TableDescriptor, traceKV bool,
 ) error {
 	descKey := sqlbase.MakeDescMetadataKey(tableDesc.ID)
 	zoneKeyPrefix := config.MakeZoneKeyPrefix(uint32(tableDesc.ID))
 
 	// Finished deleting all the table data, now delete the table meta data.
-	return db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
+	return sc.db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
 		// Delete table descriptor
 		b := &client.Batch{}
 		if traceKV {
@@ -348,6 +361,31 @@ func DropTableDesc(
 		if err := txn.SetSystemConfigTrigger(); err != nil {
 			return err
 		}
+
+		if tableDesc.GetDropJobID() != 0 {
+			if err := sc.updateDropTableJob(
+				ctx,
+				txn,
+				tableDesc.GetDropJobID(),
+				tableDesc.ID,
+				jobspb.Status_DONE,
+				func(ctx context.Context, txn *client.Txn, job *jobs.Job) error {
+					// Delete the zone config entry for the dropped database associated
+					// with the job, if it exists.
+					details := job.Details().(jobspb.SchemaChangeDetails)
+					if details.DroppedDatabaseID == sqlbase.InvalidID {
+						return nil
+					}
+					dbZoneKeyPrefix := config.MakeZoneKeyPrefix(uint32(details.DroppedDatabaseID))
+					if traceKV {
+						log.VEventf(ctx, 2, "DelRange %s", zoneKeyPrefix)
+					}
+					b.DelRange(dbZoneKeyPrefix, dbZoneKeyPrefix.PrefixEnd(), false /* returnKeys */)
+					return nil
+				}); err != nil {
+				return errors.Wrapf(err, "failed to update job %d", tableDesc.GetDropJobID())
+			}
+		}
 		return txn.Run(ctx, b)
 	})
 }
@@ -433,60 +471,77 @@ func (sc *SchemaChanger) truncateTable(
 	return nil
 }
 
-// maybe Add/Drop a table depending on the state of a table descriptor.
-// This method returns true if the table is deleted.
-func (sc *SchemaChanger) maybeAddDrop(
-	ctx context.Context,
-	inSession bool,
-	lease *sqlbase.TableDescriptor_SchemaChangeLease,
-	table *sqlbase.TableDescriptor,
-	evalCtx *extendedEvalContext,
-) (bool, error) {
-	if table.Dropped() {
-		if err := sc.ExtendLease(ctx, lease); err != nil {
-			return false, err
-		}
-
-		if inSession {
-			return false, nil
-		}
-
-		// This can happen if a change other than the drop originally
-		// scheduled the changer for this table. If that's the case,
-		// we still need to wait for the deadline to expire.
-		if table.DropTime != 0 {
-			var timeRemaining time.Duration
-			if err := sc.db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
-				timeRemaining = 0
-				_, zoneCfg, _, err := GetZoneConfigInTxn(
-					ctx, txn, uint32(table.ID), &sqlbase.IndexDescriptor{}, "", false,
-				)
-				if err != nil {
-					return err
-				}
-				deadline := table.DropTime + int64(zoneCfg.GC.TTLSeconds)*time.Second.Nanoseconds()
-				timeRemaining = timeutil.Since(timeutil.Unix(0, deadline))
-				return nil
-			}); err != nil {
-				return false, err
-			}
-			if timeRemaining < 0 {
-				return false, errNotHitGCTTLDeadline
-			}
-		}
-		// Do all the hard work of deleting the table data and the table ID.
-		if err := sc.truncateTable(ctx, lease, table, evalCtx); err != nil {
-			return false, err
-		}
-
-		return true, DropTableDesc(ctx, table, sc.db, false /* traceKV */)
+// maybe Drop a table. Return nil if successfully dropped.
+func (sc *SchemaChanger) maybeDropTable(
+	ctx context.Context, inSession bool, table *sqlbase.TableDescriptor, evalCtx *extendedEvalContext,
+) error {
+	if !table.Dropped() || inSession {
+		return nil
 	}
 
+	// This can happen if a change other than the drop originally
+	// scheduled the changer for this table. If that's the case,
+	// we still need to wait for the deadline to expire.
+	if table.DropTime != 0 {
+		var timeRemaining time.Duration
+		if err := sc.db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
+			timeRemaining = 0
+			_, zoneCfg, _, err := GetZoneConfigInTxn(ctx, txn, uint32(table.ID),
+				&sqlbase.IndexDescriptor{}, "", false /* getInheritedDefault */)
+			if err != nil {
+				return err
+			}
+			deadline := table.DropTime + int64(zoneCfg.GC.TTLSeconds)*time.Second.Nanoseconds()
+			timeRemaining = timeutil.Since(timeutil.Unix(0, deadline))
+			return nil
+		}); err != nil {
+			return err
+		}
+		if timeRemaining < 0 {
+			return errNotHitGCTTLDeadline
+		}
+	}
+
+	// Acquire lease.
+	lease, err := sc.AcquireLease(ctx)
+	if err != nil {
+		return err
+	}
+	needRelease := true
+	// Always try to release lease.
+	defer func() {
+		// If the schema changer deleted the descriptor, there's no longer a lease to be
+		// released.
+		if !needRelease {
+			return
+		}
+		if err := sc.ReleaseLease(ctx, lease); err != nil {
+			log.Warning(ctx, err)
+		}
+	}()
+
+	// Do all the hard work of deleting the table data and the table ID.
+	if err := sc.truncateTable(ctx, &lease, table, evalCtx); err != nil {
+		return err
+	}
+
+	if err := sc.DropTableDesc(ctx, table, false /* traceKV */); err != nil {
+		return err
+	}
+	// The descriptor was deleted.
+	needRelease = false
+	return nil
+}
+
+// maybe make a table PUBLIC if it's in the ADD state.
+func (sc *SchemaChanger) maybeMakeAddTablePublic(
+	ctx context.Context, table *sqlbase.TableDescriptor,
+) error {
 	if table.Adding() {
 		for _, idx := range table.AllNonDropIndexes() {
 			if idx.ForeignKey.IsSet() {
 				if err := sc.waitToUpdateLeases(ctx, idx.ForeignKey.Table); err != nil {
-					return false, err
+					return err
 				}
 			}
 		}
@@ -494,50 +549,181 @@ func (sc *SchemaChanger) maybeAddDrop(
 		if _, err := sc.leaseMgr.Publish(
 			ctx,
 			table.ID,
-			func(tbl *sqlbase.TableDescriptor) error {
+			func(tbl *sqlbase.MutableTableDescriptor) error {
+				if !tbl.Adding() {
+					return errDidntUpdateDescriptor
+				}
 				tbl.State = sqlbase.TableDescriptor_PUBLIC
 				return nil
 			},
 			func(txn *client.Txn) error { return nil },
 		); err != nil {
-			return false, err
+			return err
 		}
 	}
 
-	return false, nil
+	return nil
 }
 
-// Drain old names from the cluster.
-func (sc *SchemaChanger) drainNames(
-	ctx context.Context, lease *sqlbase.TableDescriptor_SchemaChangeLease,
+func (sc *SchemaChanger) maybeGCMutations(
+	ctx context.Context, inSession bool, table *sqlbase.TableDescriptor,
 ) error {
-	if err := sc.ExtendLease(ctx, lease); err != nil {
+	if inSession || len(table.GCMutations) == 0 || len(sc.dropIndexTimes) == 0 {
+		return nil
+	}
+
+	// Don't perform GC work if there are non-GC mutations waiting.
+	if len(table.Mutations) > 0 {
+		return nil
+	}
+
+	// Find dropped index with earliest GC deadline.
+	dropped := sc.dropIndexTimes[0]
+	for i := 1; i < len(sc.dropIndexTimes); i++ {
+		if other := sc.dropIndexTimes[i]; other.deadline < dropped.deadline {
+			dropped = other
+		}
+	}
+
+	var mutation sqlbase.TableDescriptor_GCDescriptorMutation
+	found := false
+	for _, gcm := range table.GCMutations {
+		if gcm.IndexID == sc.dropIndexTimes[0].indexID {
+			found = true
+			mutation = gcm
+			break
+		}
+	}
+	if !found {
+		return errors.Errorf("no GC mutation for index %d", sc.dropIndexTimes[0].indexID)
+	}
+
+	// Check if the deadline for GC'd dropped index expired because
+	// a change other than the drop could have scheduled the changer
+	// for this table.
+	timeRemaining := timeutil.Since(timeutil.Unix(0, dropped.deadline))
+	if timeRemaining < 0 {
+		// Return nil to allow other any mutations to make progress.
+		return nil
+	}
+
+	// Acquire lease.
+	lease, err := sc.AcquireLease(ctx)
+	if err != nil {
+		return err
+	}
+	// Always try to release lease.
+	defer func() {
+		if err := sc.ReleaseLease(ctx, lease); err != nil {
+			log.Warning(ctx, err)
+		}
+	}()
+
+	if err := sc.truncateIndexes(ctx, &lease, table.Version, []sqlbase.IndexDescriptor{{ID: mutation.IndexID}}); err != nil {
 		return err
 	}
 
+	_, err = sc.leaseMgr.Publish(
+		ctx,
+		table.ID,
+		func(tbl *sqlbase.MutableTableDescriptor) error {
+			found := false
+			for i := 0; i < len(tbl.GCMutations); i++ {
+				if other := tbl.GCMutations[i]; other.IndexID == mutation.IndexID {
+					tbl.GCMutations = append(tbl.GCMutations[:i], tbl.GCMutations[i+1:]...)
+					found = true
+					break
+				}
+			}
+
+			if !found {
+				return errDidntUpdateDescriptor
+			}
+
+			return nil
+		},
+		func(txn *client.Txn) error {
+			job, err := sc.jobRegistry.LoadJobWithTxn(ctx, mutation.JobID, txn)
+			if err != nil {
+				return err
+			}
+			return job.WithTxn(txn).Succeeded(ctx, jobs.NoopFn)
+		},
+	)
+
+	return err
+}
+
+func (sc *SchemaChanger) updateDropTableJob(
+	ctx context.Context,
+	txn *client.Txn,
+	jobID int64,
+	tableID sqlbase.ID,
+	status jobspb.Status,
+	onSuccess func(context.Context, *client.Txn, *jobs.Job) error,
+) error {
+	job, err := sc.jobRegistry.LoadJobWithTxn(ctx, jobID, txn)
+	if err != nil {
+		return err
+	}
+
+	schemaDetails, ok := job.Details().(jobspb.SchemaChangeDetails)
+	if !ok {
+		return errors.Errorf("unexpected details for job %d", *job.ID())
+	}
+
+	lowestStatus := jobspb.Status_DONE
+	for i := range schemaDetails.DroppedTables {
+		if tableID == schemaDetails.DroppedTables[i].ID {
+			schemaDetails.DroppedTables[i].Status = status
+		}
+
+		if lowestStatus > schemaDetails.DroppedTables[i].Status {
+			lowestStatus = schemaDetails.DroppedTables[i].Status
+		}
+	}
+
+	var runningStatus jobs.RunningStatus
+	switch lowestStatus {
+	case jobspb.Status_DRAINING_NAMES:
+		runningStatus = jobs.RunningStatusDrainingNames
+	case jobspb.Status_WAIT_FOR_GC_INTERVAL:
+		runningStatus = jobs.RunningStatusWaitingGC
+	case jobspb.Status_ROCKSDB_COMPACTION:
+		runningStatus = jobs.RunningStatusCompaction
+	case jobspb.Status_DONE:
+		return job.WithTxn(txn).Succeeded(ctx, onSuccess)
+	default:
+		return errors.Errorf("unexpected dropped table status %d", lowestStatus)
+	}
+
+	if err := job.WithTxn(txn).SetDetails(ctx, schemaDetails); err != nil {
+		return err
+	}
+
+	return job.WithTxn(txn).RunningStatus(ctx, func(ctx context.Context, _ jobspb.Details) (jobs.RunningStatus, error) {
+		return runningStatus, nil
+	})
+}
+
+// Drain old names from the cluster.
+func (sc *SchemaChanger) drainNames(ctx context.Context) error {
 	// Publish a new version with all the names drained after everyone
 	// has seen the version with the new name. All the draining names
 	// can be reused henceforth.
 	var namesToReclaim []sqlbase.TableDescriptor_NameInfo
+	var dropJobID int64
 	_, err := sc.leaseMgr.Publish(
 		ctx,
 		sc.tableID,
-		func(desc *sqlbase.TableDescriptor) error {
+		func(desc *sqlbase.MutableTableDescriptor) error {
 			if sc.testingKnobs.OldNamesDrainedNotification != nil {
 				sc.testingKnobs.OldNamesDrainedNotification()
 			}
-			// Check that another schema change didn't run during the
-			// drain phase. This ensures that we don't reclaim old names
-			// here without explicitly going through a drain phase for the
-			// names. On seeing a need to increment the version return an
-			// error, so that sc.exec() is reexecuted and increments the
-			// version before coming back here to correctly drain the names.
-			if desc.UpVersion {
-				return errSchemaChangeDuringDrain
-			}
 			// Free up the old name(s) for reuse.
 			namesToReclaim = desc.DrainingNames
 			desc.DrainingNames = nil
+			dropJobID = desc.GetDropJobID()
 			return nil
 		},
 		// Reclaim all the old names.
@@ -547,6 +733,13 @@ func (sc *SchemaChanger) drainNames(
 				tbKey := tableKey{drain.ParentID, drain.Name}.Key()
 				b.Del(tbKey)
 			}
+
+			if dropJobID != 0 {
+				if err := sc.updateDropTableJob(ctx, txn, dropJobID, sc.tableID, jobspb.Status_WAIT_FOR_GC_INTERVAL, jobs.NoopFn); err != nil {
+					return err
+				}
+			}
+
 			return txn.Run(ctx, b)
 		},
 	)
@@ -567,25 +760,8 @@ func (sc *SchemaChanger) exec(
 		log.Infof(ctx, "exec pending schema change; table: %d, mutation: %d",
 			sc.tableID, sc.mutationID)
 	}
-	// Acquire lease.
-	lease, err := sc.AcquireLease(ctx)
-	if err != nil {
-		return err
-	}
-	needRelease := true
-	// Always try to release lease.
-	defer func() {
-		// If the schema changer deleted the descriptor, there's no longer a lease to be
-		// released.
-		if !needRelease {
-			return
-		}
-		if err := sc.ReleaseLease(ctx, lease); err != nil {
-			log.Warning(ctx, err)
-		}
-	}()
 
-	notFirst, err := sc.notFirstInLine(ctx)
+	tableDesc, notFirst, err := sc.notFirstInLine(ctx)
 	if err != nil {
 		return err
 	}
@@ -593,40 +769,57 @@ func (sc *SchemaChanger) exec(
 		return errSchemaChangeNotFirstInLine
 	}
 
-	// Increment the version and unset tableDescriptor.UpVersion.
-	desc, err := sc.MaybeIncrementVersion(ctx)
-	if err != nil {
-		return err
-	}
-
-	tableDesc := desc.GetTable()
 	if tableDesc.HasDrainingNames() {
-		if err := sc.drainNames(ctx, &lease); err != nil {
+		if err := sc.drainNames(ctx); err != nil {
 			return err
 		}
 	}
 
-	if drop, err := sc.maybeAddDrop(ctx, inSession, &lease, tableDesc, evalCtx); err != nil {
+	// Delete dropped table data if possible.
+	if err := sc.maybeDropTable(ctx, inSession, tableDesc, evalCtx); err != nil {
+		return err
+	}
+
+	if err := sc.maybeMakeAddTablePublic(ctx, tableDesc); err != nil {
+		return err
+	}
+
+	if err := sc.maybeGCMutations(ctx, inSession, tableDesc); err != nil {
 		return err
-	} else if drop {
-		needRelease = false
-		return nil
 	}
 
 	// Wait for the schema change to propagate to all nodes after this function
 	// returns, so that the new schema is live everywhere. This is not needed for
 	// correctness but is done to make the UI experience/tests predictable.
-	defer func() {
+	waitToUpdateLeases := func(refreshStats bool) {
 		if err := sc.waitToUpdateLeases(ctx, sc.tableID); err != nil {
 			log.Warning(ctx, err)
 		}
-	}()
+		// We wait to trigger a stats refresh until we know the leases have been
+		// updated.
+		if refreshStats {
+			sc.refreshStats()
+		}
+	}
 
 	if sc.mutationID == sqlbase.InvalidMutationID {
 		// Nothing more to do.
+		waitToUpdateLeases(false /* refreshStats */)
 		return nil
 	}
 
+	// Acquire lease.
+	lease, err := sc.AcquireLease(ctx)
+	if err != nil {
+		return err
+	}
+	// Always try to release lease.
+	defer func() {
+		if err := sc.ReleaseLease(ctx, lease); err != nil {
+			log.Warning(ctx, err)
+		}
+	}()
+
 	// Find our job.
 	foundJobID := false
 	for _, g := range tableDesc.MutationJobs {
@@ -640,6 +833,7 @@ func (sc *SchemaChanger) exec(
 			break
 		}
 	}
+
 	if !foundJobID {
 		// No job means we've already run and completed this schema change
 		// successfully, so we can just exit.
@@ -652,12 +846,11 @@ func (sc *SchemaChanger) exec(
 		}
 	}
 
-	// Another transaction might set the up_version bit again,
-	// but we're no longer responsible for taking care of that.
-
 	// Run through mutation state machine and backfill.
 	err = sc.runStateMachineAndBackfill(ctx, &lease, evalCtx)
 
+	defer waitToUpdateLeases(err == nil /* refreshStats */)
+
 	// Purge the mutations if the application of the mutations failed due to
 	// a permanent error. All other errors are transient errors that are
 	// resolved by retrying the backfill.
@@ -697,27 +890,11 @@ func (sc *SchemaChanger) rollbackSchemaChange(
 	return nil
 }
 
-// MaybeIncrementVersion increments the version if needed.
-// If the version is to be incremented, it also assures that all nodes are on
-// the current (pre-increment) version of the descriptor.
-// Returns the (potentially updated) descriptor.
-func (sc *SchemaChanger) MaybeIncrementVersion(ctx context.Context) (*sqlbase.Descriptor, error) {
-	return sc.leaseMgr.Publish(ctx, sc.tableID, func(desc *sqlbase.TableDescriptor) error {
-		if !desc.UpVersion {
-			// Return error so that Publish() doesn't increment the version.
-			return errDidntUpdateDescriptor
-		}
-		desc.UpVersion = false
-		// Publish() will increment the version.
-		return nil
-	}, nil)
-}
-
 // RunStateMachineBeforeBackfill moves the state machine forward
 // and wait to ensure that all nodes are seeing the latest version
 // of the table.
 func (sc *SchemaChanger) RunStateMachineBeforeBackfill(ctx context.Context) error {
-	if _, err := sc.leaseMgr.Publish(ctx, sc.tableID, func(desc *sqlbase.TableDescriptor) error {
+	if _, err := sc.leaseMgr.Publish(ctx, sc.tableID, func(desc *sqlbase.MutableTableDescriptor) error {
 		var modified bool
 		// Apply mutations belonging to the same version.
 		for i, mutation := range desc.Mutations {
@@ -789,10 +966,16 @@ func (sc *SchemaChanger) waitToUpdateLeases(ctx context.Context, tableID sqlbase
 // done finalizes the mutations (adds new cols/indexes to the table).
 // It ensures that all nodes are on the current (pre-update) version of the
 // schema.
-// Returns the updated of the descriptor.
-func (sc *SchemaChanger) done(ctx context.Context) (*sqlbase.Descriptor, error) {
+// Returns the updated descriptor.
+func (sc *SchemaChanger) done(ctx context.Context) (*sqlbase.ImmutableTableDescriptor, error) {
 	isRollback := false
-	return sc.leaseMgr.Publish(ctx, sc.tableID, func(desc *sqlbase.TableDescriptor) error {
+	jobSucceeded := true
+	now := timeutil.Now().UnixNano()
+	return sc.leaseMgr.Publish(ctx, sc.tableID, func(desc *sqlbase.MutableTableDescriptor) error {
+		// Reset vars here because update function can be called multiple times in a retry.
+		isRollback = false
+		jobSucceeded = true
+
 		i := 0
 		for _, mutation := range desc.Mutations {
 			if mutation.MutationID != sc.mutationID {
@@ -801,6 +984,19 @@ func (sc *SchemaChanger) done(ctx context.Context) (*sqlbase.Descriptor, error)
 				break
 			}
 			isRollback = mutation.Rollback
+			if indexDesc := mutation.GetIndex(); mutation.Direction == sqlbase.DescriptorMutation_DROP &&
+				indexDesc != nil {
+				if sc.canClearRangeForDrop(indexDesc) {
+					jobSucceeded = false
+					desc.GCMutations = append(
+						desc.GCMutations,
+						sqlbase.TableDescriptor_GCDescriptorMutation{
+							IndexID:  indexDesc.ID,
+							DropTime: now,
+							JobID:    *sc.job.ID(),
+						})
+				}
+			}
 			if err := desc.MakeMutationComplete(mutation); err != nil {
 				return err
 			}
@@ -823,8 +1019,16 @@ func (sc *SchemaChanger) done(ctx context.Context) (*sqlbase.Descriptor, error)
 		}
 		return nil
 	}, func(txn *client.Txn) error {
-		if err := sc.job.WithTxn(txn).Succeeded(ctx, jobs.NoopFn); err != nil {
-			return errors.Wrapf(err, "failed to mark job %d as as successful", *sc.job.ID())
+		if jobSucceeded {
+			if err := sc.job.WithTxn(txn).Succeeded(ctx, jobs.NoopFn); err != nil {
+				return errors.Wrapf(err, "failed to mark job %d as successful", *sc.job.ID())
+			}
+		} else {
+			if err := sc.job.WithTxn(txn).RunningStatus(ctx, func(ctx context.Context, details jobspb.Details) (jobs.RunningStatus, error) {
+				return jobs.RunningStatusWaitingGC, nil
+			}); err != nil {
+				return errors.Wrapf(err, "failed to update running status of job %d", *sc.job.ID())
+			}
 		}
 
 		schemaChangeEventType := EventLogFinishSchemaChange
@@ -851,11 +1055,15 @@ func (sc *SchemaChanger) done(ctx context.Context) (*sqlbase.Descriptor, error)
 
 // notFirstInLine returns true whenever the schema change has been queued
 // up for execution after another schema change.
-func (sc *SchemaChanger) notFirstInLine(ctx context.Context) (bool, error) {
+func (sc *SchemaChanger) notFirstInLine(
+	ctx context.Context,
+) (*sqlbase.TableDescriptor, bool, error) {
 	var notFirst bool
+	var desc *sqlbase.TableDescriptor
 	err := sc.db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
 		notFirst = false
-		desc, err := sqlbase.GetTableDescFromID(ctx, txn, sc.tableID)
+		var err error
+		desc, err = sqlbase.GetTableDescFromID(ctx, txn, sc.tableID)
 		if err != nil {
 			return err
 		}
@@ -867,7 +1075,7 @@ func (sc *SchemaChanger) notFirstInLine(ctx context.Context) (bool, error) {
 		}
 		return nil
 	})
-	return notFirst, err
+	return desc, notFirst, err
 }
 
 // runStateMachineAndBackfill runs the schema change state machine followed by
@@ -895,6 +1103,13 @@ func (sc *SchemaChanger) runStateMachineAndBackfill(
 	return err
 }
 
+func (sc *SchemaChanger) refreshStats() {
+	// Initiate an asynchronous run of CREATE STATISTICS. We use a large number
+	// for rowsAffected because we want to make sure that stats always get
+	// created/refreshed here.
+	sc.execCfg.StatsRefresher.NotifyMutation(sc.tableID, math.MaxInt32 /* rowsAffected */)
+}
+
 // reverseMutations reverses the direction of all the mutations with the
 // mutationID. This is called after hitting an irrecoverable error while
 // applying a schema change. If a column being added is reversed and droped,
@@ -907,7 +1122,7 @@ func (sc *SchemaChanger) reverseMutations(ctx context.Context, causingError erro
 	// where the indexes refer columns. Whenever a column schema change
 	// is reversed, any index mutation referencing it is also reversed.
 	var droppedMutations map[sqlbase.MutationID]struct{}
-	_, err := sc.leaseMgr.Publish(ctx, sc.tableID, func(desc *sqlbase.TableDescriptor) error {
+	_, err := sc.leaseMgr.Publish(ctx, sc.tableID, func(desc *sqlbase.MutableTableDescriptor) error {
 		// Keep track of the column mutations being reversed so that indexes
 		// referencing them can be dropped.
 		columns := make(map[string]struct{})
@@ -930,7 +1145,17 @@ func (sc *SchemaChanger) reverseMutations(ctx context.Context, causingError erro
 			}
 
 			log.Warningf(ctx, "reverse schema change mutation: %+v", mutation)
-			desc.Mutations[i], columns = reverseMutation(mutation, false /*notStarted*/, columns)
+			desc.Mutations[i], columns = sc.reverseMutation(mutation, false /*notStarted*/, columns)
+
+			// If the mutation is for validating a constraint that is being added,
+			// drop the constraint because validation has failed
+			if constraint := mutation.GetConstraint(); constraint != nil &&
+				mutation.Direction == sqlbase.DescriptorMutation_ADD {
+				log.Warningf(ctx, "dropping constraint %+v", constraint)
+				if err := sc.maybeDropValidatingConstraint(ctx, desc, constraint); err != nil {
+					return err
+				}
+			}
 
 			desc.Mutations[i].Rollback = true
 		}
@@ -1080,12 +1305,36 @@ func (sc *SchemaChanger) createRollbackJob(
 	return nil, fmt.Errorf("no job found for table %d mutation %d", sc.tableID, sc.mutationID)
 }
 
+func (sc *SchemaChanger) maybeDropValidatingConstraint(
+	ctx context.Context, desc *MutableTableDescriptor, constraint *sqlbase.ConstraintToUpdate,
+) error {
+	switch constraint.ConstraintType {
+	case sqlbase.ConstraintToUpdate_CHECK:
+		for j, c := range desc.Checks {
+			if c.Name == constraint.Name {
+				desc.Checks = append(desc.Checks[:j], desc.Checks[j+1:]...)
+				return nil
+			}
+		}
+		if log.V(2) {
+			log.Infof(
+				ctx,
+				"attempted to drop constraint %s, but it hadn't been added to the table descriptor yet",
+				constraint.Name,
+			)
+		}
+	default:
+		return errors.Errorf("unsupported constraint type: %d", constraint.ConstraintType)
+	}
+	return nil
+}
+
 // deleteIndexMutationsWithReversedColumns deletes mutations with a
 // different mutationID than the schema changer and with an index that
 // references one of the reversed columns. Execute this as a breadth
 // first search graph traversal.
 func (sc *SchemaChanger) deleteIndexMutationsWithReversedColumns(
-	ctx context.Context, desc *sqlbase.TableDescriptor, columns map[string]struct{},
+	ctx context.Context, desc *sqlbase.MutableTableDescriptor, columns map[string]struct{},
 ) (map[sqlbase.MutationID]struct{}, error) {
 	dropMutations := make(map[sqlbase.MutationID]struct{})
 	// Run breadth first search traversal that reverses mutations
@@ -1125,7 +1374,7 @@ func (sc *SchemaChanger) deleteIndexMutationsWithReversedColumns(
 				// Reverse mutation. Update columns to reflect additional
 				// columns that have been purged. This mutation doesn't need
 				// a rollback because it was not started.
-				mutation, columns = reverseMutation(mutation, true /*notStarted*/, columns)
+				mutation, columns = sc.reverseMutation(mutation, true /*notStarted*/, columns)
 				// Mark as complete because this mutation needs no backfill.
 				if err := desc.MakeMutationComplete(mutation); err != nil {
 					return nil, err
@@ -1143,7 +1392,7 @@ func (sc *SchemaChanger) deleteIndexMutationsWithReversedColumns(
 // Reverse a mutation. Returns the updated mutation and updated columns.
 // notStarted is set to true only if the schema change state machine
 // was not started for the mutation.
-func reverseMutation(
+func (sc *SchemaChanger) reverseMutation(
 	mutation sqlbase.DescriptorMutation, notStarted bool, columns map[string]struct{},
 ) (sqlbase.DescriptorMutation, map[string]struct{}) {
 	switch mutation.Direction {
@@ -1253,9 +1502,11 @@ type SchemaChangeManager struct {
 	testingKnobs *SchemaChangerTestingKnobs
 	// Create a schema changer for every outstanding schema change seen.
 	schemaChangers map[sqlbase.ID]SchemaChanger
-	// Create a schema changer for every dropped table that needs to be GC-ed.
+	// Create a schema changer for every table that is dropped or has
+	// dropped indexes that needs to be GC-ed.
 	forGC          map[sqlbase.ID]SchemaChanger
 	distSQLPlanner *DistSQLPlanner
+	ieFactory      sqlutil.SessionBoundInternalExecutorFactory
 }
 
 // NewSchemaChangeManager returns a new SchemaChangeManager.
@@ -1266,6 +1517,7 @@ func NewSchemaChangeManager(
 	db client.DB,
 	nodeDesc roachpb.NodeDescriptor,
 	dsp *DistSQLPlanner,
+	ieFactory sqlutil.SessionBoundInternalExecutorFactory,
 ) *SchemaChangeManager {
 	return &SchemaChangeManager{
 		ambientCtx:     ambientCtx,
@@ -1274,6 +1526,7 @@ func NewSchemaChangeManager(
 		schemaChangers: make(map[sqlbase.ID]SchemaChanger),
 		forGC:          make(map[sqlbase.ID]SchemaChanger),
 		distSQLPlanner: dsp,
+		ieFactory:      ieFactory,
 	}
 }
 
@@ -1312,11 +1565,12 @@ func (s *SchemaChangeManager) Start(stopper *stop.Stopper) {
 		if s.testingKnobs.AsyncExecQuickly {
 			delay = 20 * time.Millisecond
 		}
+		defTTL := config.DefaultZoneConfig().GC.TTLSeconds
 
 		execOneSchemaChange := func(schemaChangers map[sqlbase.ID]SchemaChanger) {
 			for tableID, sc := range schemaChangers {
 				if timeutil.Since(sc.execAfter) > 0 {
-					evalCtx := createSchemaChangeEvalCtx(s.execCfg.Clock.Now(), &SessionTracing{})
+					evalCtx := createSchemaChangeEvalCtx(ctx, s.execCfg.Clock.Now(), &SessionTracing{}, s.ieFactory)
 
 					execCtx, cleanup := tracing.EnsureContext(ctx, s.ambientCtx.Tracer, "schema change [async]")
 					err := sc.exec(execCtx, false /* inSession */, &evalCtx)
@@ -1351,7 +1605,7 @@ func (s *SchemaChangeManager) Start(stopper *stop.Stopper) {
 		for {
 			select {
 			case <-gossipUpdateC:
-				cfg, _ := s.execCfg.Gossip.GetSystemConfig()
+				cfg := s.execCfg.Gossip.GetSystemConfig()
 				// Read all tables and their versions
 				if log.V(2) {
 					log.Info(ctx, "received a new config")
@@ -1369,24 +1623,60 @@ func (s *SchemaChangeManager) Start(stopper *stop.Stopper) {
 					// GC TTL for a table has indeed changed it is modified
 					// and enqueued with the new TTL timeout.
 					for id, sc := range s.forGC {
+						zoneCfg, placeholder, _, err := ZoneConfigHook(cfg, uint32(id))
+						if err != nil {
+							log.Errorf(ctx, "zone config for desc: %d, err = %+v", id, err)
+							return
+						}
+						if zoneCfg == nil {
+							// Do nothing, use the old zone config's TTL.
+							continue
+						}
+						if placeholder == nil {
+							placeholder = zoneCfg
+						}
+
+						newExecTime := sc.execAfter
 						if sc.dropTime > 0 {
-							zoneCfg, _, err := ZoneConfigHook(cfg, uint32(id), nil)
-							if err != nil {
-								log.Errorf(ctx, "no zone config for desc: %d", id)
-								return
-							}
 							deadline := sc.dropTime +
 								int64(zoneCfg.GC.TTLSeconds)*time.Second.Nanoseconds() +
 								int64(delay)
-							if ea := timeutil.Unix(0, deadline); ea != sc.execAfter {
-								resetTimer = true
-								sc.execAfter = ea
-								// Safe to modify map inplace while iterating over it.
-								s.forGC[id] = sc
-								if log.V(2) {
-									log.Infof(ctx,
-										"re-queue up pending drop table GC; table: %d", id)
+							newExecTime = timeutil.Unix(0, deadline)
+						}
+						if len(sc.dropIndexTimes) > 0 {
+							var earliestIndexExec time.Time
+
+							for i := 0; i < len(sc.dropIndexTimes); i++ {
+								droppedIdx := &sc.dropIndexTimes[i]
+
+								ttlSeconds := zoneCfg.GC.TTLSeconds
+								if subzone := placeholder.GetSubzone(
+									uint32(droppedIdx.indexID), ""); subzone != nil && subzone.Config.GC != nil {
+									ttlSeconds = subzone.Config.GC.TTLSeconds
 								}
+
+								deadline := droppedIdx.dropTime +
+									int64(ttlSeconds)*time.Second.Nanoseconds() +
+									int64(delay)
+								execTime := timeutil.Unix(0, deadline)
+								droppedIdx.deadline = deadline
+
+								if earliestIndexExec.IsZero() || execTime.Before(earliestIndexExec) {
+									earliestIndexExec = execTime
+								}
+							}
+							if newExecTime.After(earliestIndexExec) {
+								newExecTime = earliestIndexExec
+							}
+						}
+						if newExecTime != sc.execAfter {
+							resetTimer = true
+							sc.execAfter = newExecTime
+							// Safe to modify map inplace while iterating over it.
+							s.forGC[id] = sc
+							if log.V(2) {
+								log.Infof(ctx,
+									"re-queue up pending drop table GC; table: %d", id)
 							}
 						}
 					}
@@ -1404,6 +1694,7 @@ func (s *SchemaChangeManager) Start(stopper *stop.Stopper) {
 					rangeDescriptorCache: s.execCfg.RangeDescriptorCache,
 					clock:                s.execCfg.Clock,
 					settings:             s.execCfg.Settings,
+					ieFactory:            s.ieFactory,
 				}
 
 				execAfter := timeutil.Now().Add(delay)
@@ -1431,14 +1722,42 @@ func (s *SchemaChangeManager) Start(stopper *stop.Stopper) {
 						schemaChanger.execAfter = execAfter
 						schemaChanger.dropTime = 0
 
+						var minDeadline int64
+						if len(table.GCMutations) > 0 {
+							zoneCfg, placeholder, _, err := ZoneConfigHook(cfg, uint32(table.ID))
+							if err != nil {
+								log.Errorf(ctx, "zone config for desc: %d, err = %+v", table.ID, err)
+								return
+							}
+
+							if placeholder == nil {
+								placeholder = zoneCfg
+							}
+
+							for _, m := range table.GCMutations {
+								// Initialize TTL without a zone config in case it's not present.
+								ttlSeconds := defTTL
+								if zoneCfg != nil {
+									ttlSeconds = zoneCfg.GC.TTLSeconds
+									if subzone := placeholder.GetSubzone(
+										uint32(m.IndexID), ""); subzone != nil && subzone.Config.GC != nil {
+										ttlSeconds = subzone.Config.GC.TTLSeconds
+									}
+								}
+								deadline := m.DropTime + int64(delay) + int64(ttlSeconds)*time.Second.Nanoseconds()
+
+								dropped := droppedIndex{m.IndexID, m.DropTime, deadline}
+								if minDeadline == 0 || deadline < minDeadline {
+									minDeadline = deadline
+								}
+								schemaChanger.dropIndexTimes = append(schemaChanger.dropIndexTimes, dropped)
+							}
+						}
+
 						// Keep track of outstanding schema changes.
-						// If all schema change commands always set UpVersion, why
-						// check for the presence of mutations?
-						// A schema change execution might fail soon after
-						// unsetting UpVersion, and we still want to process
-						// outstanding mutations.
-						if table.UpVersion || table.Adding() ||
-							table.HasDrainingNames() || len(table.Mutations) > 0 {
+						pendingChanges := table.Adding() ||
+							table.HasDrainingNames() || len(table.Mutations) > 0
+						if pendingChanges {
 							if log.V(2) {
 								log.Infof(ctx, "%s: queue up pending schema change; table: %d, version: %d",
 									kv.Key, table.ID, table.Version)
@@ -1458,21 +1777,33 @@ func (s *SchemaChangeManager) Start(stopper *stop.Stopper) {
 
 							if table.DropTime > 0 {
 								schemaChanger.dropTime = table.DropTime
-								zoneCfg, _, err := ZoneConfigHook(cfg, uint32(table.ID), nil)
+								zoneCfg, _, _, err := ZoneConfigHook(cfg, uint32(table.ID))
 								if err != nil {
-									log.Errorf(ctx, "no zone config for desc: %d", table.ID)
+									log.Errorf(ctx, "zone config for desc: %d, err: %+v", table.ID, err)
 									return
 								}
-								deadline := table.DropTime +
-									int64(zoneCfg.GC.TTLSeconds)*time.Second.Nanoseconds() +
-									int64(delay)
-								schemaChanger.execAfter = timeutil.Unix(0, deadline)
-							}
 
+								// Initialize deadline without a zone config in case it's not present.
+								deadline := table.DropTime + int64(delay)
+								if zoneCfg != nil {
+									deadline += int64(zoneCfg.GC.TTLSeconds) * time.Second.Nanoseconds()
+								} else {
+									deadline += int64(defTTL) * time.Second.Nanoseconds()
+								}
+								if minDeadline == 0 || deadline < minDeadline {
+									minDeadline = deadline
+								}
+							}
+						}
+
+						if minDeadline != 0 {
+							schemaChanger.execAfter = timeutil.Unix(0, minDeadline)
+						}
+
+						if table.Dropped() || (!pendingChanges && len(table.GCMutations) > 0) {
 							s.forGC[table.ID] = schemaChanger
 							// Remove from schema change map if present because
-							// this table has been dropped and is only waiting
-							// to be GC-ed.
+							// this table has been dropped or only has other GC waiting mutations.
 							delete(s.schemaChangers, table.ID)
 						}
 
@@ -1519,27 +1850,43 @@ func (s *SchemaChangeManager) Start(stopper *stop.Stopper) {
 //
 // TODO(andrei): This EvalContext() will be broken for backfills trying to use
 // functions marked with distsqlBlacklist.
-func createSchemaChangeEvalCtx(ts hlc.Timestamp, tracing *SessionTracing) extendedEvalContext {
+func createSchemaChangeEvalCtx(
+	ctx context.Context,
+	ts hlc.Timestamp,
+	tracing *SessionTracing,
+	ieFactory sqlutil.SessionBoundInternalExecutorFactory,
+) extendedEvalContext {
 	dummyLocation := time.UTC
+
+	sd := &sessiondata.SessionData{
+		SearchPath: sqlbase.DefaultSearchPath,
+		// The database is not supposed to be needed in schema changes, as there
+		// shouldn't be unqualified identifiers in backfills, and the pure functions
+		// that need it should have already been evaluated.
+		//
+		// TODO(andrei): find a way to assert that this field is indeed not used.
+		// And in fact it is used by `current_schemas()`, which, although is a pure
+		// function, takes arguments which might be impure (so it can't always be
+		// pre-evaluated).
+		Database:      "",
+		SequenceState: sessiondata.NewSequenceState(),
+		DataConversion: sessiondata.DataConversionConfig{
+			Location: dummyLocation,
+		},
+		User: security.NodeUser,
+	}
+
 	evalCtx := extendedEvalContext{
 		Tracing: tracing,
 		EvalContext: tree.EvalContext{
-			SessionData: &sessiondata.SessionData{
-				SearchPath: sqlbase.DefaultSearchPath,
-				// The database is not supposed to be needed in schema changes, as there
-				// shouldn't be unqualified identifiers in backfills, and the pure functions
-				// that need it should have already been evaluated.
-				//
-				// TODO(andrei): find a way to assert that this field is indeed not used.
-				// And in fact it is used by `current_schemas()`, which, although is a pure
-				// function, takes arguments which might be impure (so it can't always be
-				// pre-evaluated).
-				Database:      "",
-				SequenceState: sessiondata.NewSequenceState(),
-				DataConversion: sessiondata.DataConversionConfig{
-					Location: dummyLocation,
-				},
-			},
+			SessionData:      sd,
+			InternalExecutor: ieFactory(ctx, sd),
+			// TODO(andrei): This is wrong (just like on the main code path on
+			// setupFlow). Each processor should override Ctx with its own context.
+			Context:         ctx,
+			Sequence:        &sqlbase.DummySequenceOperators{},
+			Planner:         &sqlbase.DummyEvalPlanner{},
+			SessionAccessor: &sqlbase.DummySessionAccessor{},
 		},
 	}
 	// The backfill is going to use the current timestamp for the various
