@@ -34,6 +34,10 @@ import (
 func initAggregateBuiltins() {
 	// Add all aggregates to the Builtins map after a few sanity checks.
 	for k, v := range aggregates {
+		if _, exists := builtins[k]; exists {
+			panic("duplicate builtin: " + k)
+		}
+
 		if !v.props.Impure {
 			panic(fmt.Sprintf("%s: aggregate functions should all be impure, found %v", k, v))
 		}
@@ -1018,12 +1022,12 @@ func (a *intSumAggregate) Add(ctx context.Context, datum tree.Datum, _ ...tree.D
 				// And overflow was detected; go to large integers, but keep the
 				// sum computed so far.
 				a.large = true
-				a.decSum.SetCoefficient(a.intSum)
+				a.decSum.SetFinite(a.intSum, 0)
 			}
 		}
 
 		if a.large {
-			a.tmpDec.SetCoefficient(t)
+			a.tmpDec.SetFinite(t, 0)
 			_, err := tree.ExactCtx.Add(&a.decSum, &a.decSum, &a.tmpDec)
 			if err != nil {
 				return err
@@ -1046,7 +1050,7 @@ func (a *intSumAggregate) Result() (tree.Datum, error) {
 	if a.large {
 		dd.Set(&a.decSum)
 	} else {
-		dd.SetCoefficient(a.intSum)
+		dd.SetFinite(a.intSum, 0)
 	}
 	return dd, nil
 }
@@ -1221,7 +1225,7 @@ func (a *intSqrDiffAggregate) Add(ctx context.Context, datum tree.Datum, _ ...tr
 		return nil
 	}
 
-	a.tmpDec.SetCoefficient(int64(tree.MustBeDInt(datum)))
+	a.tmpDec.SetFinite(int64(tree.MustBeDInt(datum)), 0)
 	return a.agg.Add(ctx, &a.tmpDec)
 }
 
