@@ -19,17 +19,16 @@ import (
 	"strings"
 	"time"
 
-	"github.com/lib/pq/oid"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/privilege"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/ipaddr"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/lib/pq/oid"
+	"github.com/pkg/errors"
 )
 
 // This file contains builtin functions that we implement primarily for
@@ -57,7 +56,7 @@ func makeNotUsableFalseBuiltin() builtinDefinition {
 
 // typeBuiltinsHaveUnderscore is a map to keep track of which types have i/o
 // builtins with underscores in between their type name and the i/o builtin
-// name, like date_in vs int8int. There seems to be no other way to
+// name, like date_in vs int8in. There seems to be no other way to
 // programmatically determine whether or not this underscore is present, hence
 // the existence of this map.
 var typeBuiltinsHaveUnderscore = map[oid.Oid]struct{}{
@@ -69,6 +68,8 @@ var typeBuiltinsHaveUnderscore = map[oid.Oid]struct{}{
 	types.Interval.Oid():    {},
 	types.JSON.Oid():        {},
 	types.UUID.Oid():        {},
+	oid.T_varbit:            {},
+	oid.T_bit:               {},
 	types.Timestamp.Oid():   {},
 	types.TimestampTZ.Oid(): {},
 	types.FamTuple.Oid():    {},
@@ -88,6 +89,9 @@ func PGIOBuiltinPrefix(typ types.T) string {
 // initPGBuiltins adds all of the postgres builtins to the Builtins map.
 func initPGBuiltins() {
 	for k, v := range pgBuiltins {
+		if _, exists := builtins[k]; exists {
+			panic("duplicate builtin: " + k)
+		}
 		v.props.Category = categoryCompatibility
 		builtins[k] = v
 	}
@@ -711,8 +715,24 @@ var pgBuiltins = map[string]builtinDefinition{
 		tree.Overload{
 			Types:      tree.ArgTypes{{"table_oid", types.Oid}, {"column_number", types.Int}},
 			ReturnType: tree.FixedReturnType(types.String),
-			Fn: func(_ *tree.EvalContext, _ tree.Datums) (tree.Datum, error) {
-				return tree.DNull, nil
+			Fn: func(ctx *tree.EvalContext, args tree.Datums) (tree.Datum, error) {
+				if *args[1].(*tree.DInt) == 0 {
+					// column ID 0 never exists, and we don't want the query
+					// below to pick up the table comment by accident.
+					return tree.DNull, nil
+				}
+				r, err := ctx.InternalExecutor.QueryRow(
+					ctx.Ctx(), "pg_get_coldesc",
+					ctx.Txn,
+					"SELECT description FROM pg_catalog.pg_description WHERE objoid=$1 AND objsubid=$2 LIMIT 1",
+					args[0], args[1])
+				if err != nil {
+					return nil, err
+				}
+				if len(r) == 0 {
+					return tree.DNull, nil
+				}
+				return r[0], nil
 			},
 			Info: notUsableInfo,
 		},
@@ -722,16 +742,19 @@ var pgBuiltins = map[string]builtinDefinition{
 		tree.Overload{
 			Types:      tree.ArgTypes{{"object_oid", types.Oid}},
 			ReturnType: tree.FixedReturnType(types.String),
-			Fn: func(_ *tree.EvalContext, _ tree.Datums) (tree.Datum, error) {
-				return tree.DNull, nil
+			Fn: func(ctx *tree.EvalContext, args tree.Datums) (tree.Datum, error) {
+				return getPgObjDesc(ctx, "", int(args[0].(*tree.DOid).DInt))
 			},
 			Info: notUsableInfo,
 		},
 		tree.Overload{
 			Types:      tree.ArgTypes{{"object_oid", types.Oid}, {"catalog_name", types.String}},
 			ReturnType: tree.FixedReturnType(types.String),
-			Fn: func(_ *tree.EvalContext, _ tree.Datums) (tree.Datum, error) {
-				return tree.DNull, nil
+			Fn: func(ctx *tree.EvalContext, args tree.Datums) (tree.Datum, error) {
+				return getPgObjDesc(ctx,
+					string(tree.MustBeDString(args[1])),
+					int(args[0].(*tree.DOid).DInt),
+				)
 			},
 			Info: notUsableInfo,
 		},
@@ -752,8 +775,34 @@ var pgBuiltins = map[string]builtinDefinition{
 		tree.Overload{
 			Types:      tree.ArgTypes{{"object_oid", types.Oid}, {"catalog_name", types.String}},
 			ReturnType: tree.FixedReturnType(types.String),
-			Fn: func(_ *tree.EvalContext, _ tree.Datums) (tree.Datum, error) {
-				return tree.DNull, nil
+			Fn: func(ctx *tree.EvalContext, args tree.Datums) (tree.Datum, error) {
+				catalogName := string(tree.MustBeDString(args[1]))
+				objOid := int(args[0].(*tree.DOid).DInt)
+
+				classOid, ok := getCatalogOidForComments(catalogName)
+				if !ok {
+					// No such catalog - return null, matching pg.
+					return tree.DNull, nil
+				}
+
+				r, err := ctx.InternalExecutor.QueryRow(
+					ctx.Ctx(), "pg_get_shobjdesc", ctx.Txn,
+					fmt.Sprintf(`
+SELECT description
+  FROM pg_catalog.pg_shdescription
+ WHERE objoid = %[1]d
+   AND classoid = %[2]d
+ LIMIT 1`,
+						objOid,
+						classOid,
+					))
+				if err != nil {
+					return nil, err
+				}
+				if len(r) == 0 {
+					return tree.DNull, nil
+				}
+				return r[0], nil
 			},
 			Info: notUsableInfo,
 		},
@@ -807,7 +856,7 @@ var pgBuiltins = map[string]builtinDefinition{
 					ctx.Ctx(), "pg_table_is_visible",
 					ctx.Txn,
 					"SELECT nspname FROM pg_catalog.pg_class c JOIN pg_catalog.pg_namespace n ON c.relnamespace=n.oid "+
-						"WHERE c.oid=$1 AND nspname=ANY(current_schemas(true));", oid)
+						"WHERE c.oid=$1 AND nspname=ANY(current_schemas(true))", oid)
 				if err != nil {
 					return nil, err
 				}
@@ -1443,6 +1492,52 @@ var pgBuiltins = map[string]builtinDefinition{
 		},
 	),
 
+	// See https://www.postgresql.org/docs/10/functions-admin.html#FUNCTIONS-ADMIN-SET
+	"current_setting": makeBuiltin(
+		tree.FunctionProperties{
+			Category:         categorySystemInfo,
+			DistsqlBlacklist: true,
+		},
+		tree.Overload{
+			Types:      tree.ArgTypes{{"setting_name", types.String}},
+			ReturnType: tree.FixedReturnType(types.String),
+			Fn: func(ctx *tree.EvalContext, args tree.Datums) (tree.Datum, error) {
+				return getSessionVar(ctx, string(tree.MustBeDString(args[0])), false /* missingOk */)
+			},
+			Info: notUsableInfo,
+		},
+		tree.Overload{
+			Types:      tree.ArgTypes{{"setting_name", types.String}, {"missing_ok", types.Bool}},
+			ReturnType: tree.FixedReturnType(types.String),
+			Fn: func(ctx *tree.EvalContext, args tree.Datums) (tree.Datum, error) {
+				return getSessionVar(ctx, string(tree.MustBeDString(args[0])), bool(tree.MustBeDBool(args[1])))
+			},
+			Info: notUsableInfo,
+		},
+	),
+
+	// See https://www.postgresql.org/docs/10/functions-admin.html#FUNCTIONS-ADMIN-SET
+	"set_config": makeBuiltin(
+		tree.FunctionProperties{
+			Category:         categorySystemInfo,
+			DistsqlBlacklist: true,
+			Impure:           true,
+		},
+		tree.Overload{
+			Types:      tree.ArgTypes{{"setting_name", types.String}, {"new_value", types.String}, {"is_local", types.Bool}},
+			ReturnType: tree.FixedReturnType(types.String),
+			Fn: func(ctx *tree.EvalContext, args tree.Datums) (tree.Datum, error) {
+				varName := string(tree.MustBeDString(args[0]))
+				err := setSessionVar(ctx, varName, string(tree.MustBeDString(args[1])), bool(tree.MustBeDBool(args[2])))
+				if err != nil {
+					return nil, err
+				}
+				return getSessionVar(ctx, varName, false /* missingOk */)
+			},
+			Info: notUsableInfo,
+		},
+	),
+
 	// inet_{client,server}_{addr,port} return either an INet address or integer
 	// port that corresponds to either the client or server side of the current
 	// session's connection.
@@ -1498,3 +1593,79 @@ var pgBuiltins = map[string]builtinDefinition{
 		},
 	),
 }
+
+func getSessionVar(ctx *tree.EvalContext, settingName string, missingOk bool) (tree.Datum, error) {
+	if ctx.SessionAccessor == nil {
+		return nil, pgerror.NewAssertionErrorf("session accessor not set")
+	}
+	ok, s, err := ctx.SessionAccessor.GetSessionVar(ctx.Context, settingName, missingOk)
+	if err != nil {
+		return nil, err
+	}
+	if !ok {
+		return tree.DNull, nil
+	}
+	return tree.NewDString(s), nil
+}
+
+func setSessionVar(ctx *tree.EvalContext, settingName, newVal string, isLocal bool) error {
+	if ctx.SessionAccessor == nil {
+		return pgerror.NewAssertionErrorf("session accessor not set")
+	}
+	if isLocal {
+		return pgerror.UnimplementedWithIssueErrorf(32562, "transaction-scoped settings are not supported")
+	}
+	return ctx.SessionAccessor.SetSessionVar(ctx.Context, settingName, newVal)
+}
+
+// getCatalogOidForComments returns the "catalog table oid" (the oid of a
+// catalog table like pg_database, in the pg_class table) for an input catalog
+// name (like pg_class or pg_database). It returns false if there is no such
+// catalog table.
+func getCatalogOidForComments(catalogName string) (id int, ok bool) {
+	switch catalogName {
+	case "pg_class":
+		return sqlbase.PgCatalogClassTableID, true
+	case "pg_database":
+		return sqlbase.PgCatalogDatabaseTableID, true
+	default:
+		// We currently only support comments on pg_class objects
+		// (columns, tables) in this context.
+		// see a different name, matching pg.
+		return 0, false
+	}
+}
+
+// getPgObjDesc queries pg_description for object comments. catalog_name, if not
+// empty, provides a constraint on which "system catalog" the comment is in.
+// System catalogs are things like pg_class, pg_type, pg_database, and so on.
+func getPgObjDesc(ctx *tree.EvalContext, catalogName string, oid int) (tree.Datum, error) {
+	classOidFilter := ""
+	if catalogName != "" {
+		classOid, ok := getCatalogOidForComments(catalogName)
+		if !ok {
+			// Return NULL for no comment if we can't find the catalog, matching pg.
+			return tree.DNull, nil
+		}
+		classOidFilter = fmt.Sprintf("AND classoid = %d", classOid)
+	}
+	r, err := ctx.InternalExecutor.QueryRow(
+		ctx.Ctx(), "pg_get_objdesc", ctx.Txn,
+		fmt.Sprintf(`
+SELECT description
+  FROM pg_catalog.pg_description
+ WHERE objoid = %[1]d
+   AND objsubid = 0
+   %[2]s
+ LIMIT 1`,
+			oid,
+			classOidFilter,
+		))
+	if err != nil {
+		return nil, err
+	}
+	if len(r) == 0 {
+		return tree.DNull, nil
+	}
+	return r[0], nil
+}
