@@ -26,6 +26,10 @@ import (
 func initWindowBuiltins() {
 	// Add all windows to the Builtins map after a few sanity checks.
 	for k, v := range windows {
+		if _, exists := builtins[k]; exists {
+			panic("duplicate builtin: " + k)
+		}
+
 		if !v.props.Impure {
 			panic(fmt.Sprintf("%s: window functions should all be impure, found %v", k, v))
 		}
@@ -199,10 +203,23 @@ func (w *aggregateWindowFunc) Compute(
 	// Accumulate all values in the peer group at the same time, as these
 	// must return the same value.
 	for i := 0; i < wfr.PeerHelper.GetRowCount(wfr.CurRowPeerGroupNum); i++ {
-		if wfr.FilterColIdx != noFilterIdx && wfr.Rows.GetRow(wfr.RowIdx+i).GetDatum(wfr.FilterColIdx) != tree.DBoolTrue {
-			continue
+		if wfr.FilterColIdx != noFilterIdx {
+			row, err := wfr.Rows.GetRow(ctx, wfr.RowIdx+i)
+			if err != nil {
+				return nil, err
+			}
+			datum, err := row.GetDatum(wfr.FilterColIdx)
+			if err != nil {
+				return nil, err
+			}
+			if datum != tree.DBoolTrue {
+				continue
+			}
+		}
+		args, err := wfr.ArgsWithRowOffset(ctx, i)
+		if err != nil {
+			return nil, err
 		}
-		args := wfr.ArgsWithRowOffset(i)
 		var value tree.Datum
 		var others tree.Datums
 		// COUNT_ROWS takes no arguments.
@@ -272,11 +289,32 @@ func (w *framableAggregateWindowFunc) Compute(
 	*w.agg = aggregateWindowFunc{w.aggConstructor(evalCtx, nil /* arguments */), tree.DNull}
 
 	// Accumulate all values in the window frame.
-	for i := wfr.FrameStartIdx(evalCtx); i < wfr.FrameEndIdx(evalCtx); i++ {
-		if wfr.FilterColIdx != noFilterIdx && wfr.Rows.GetRow(i).GetDatum(wfr.FilterColIdx) != tree.DBoolTrue {
-			continue
+	frameStartIdx, err := wfr.FrameStartIdx(ctx, evalCtx)
+	if err != nil {
+		return nil, err
+	}
+	frameEndIdx, err := wfr.FrameEndIdx(ctx, evalCtx)
+	if err != nil {
+		return nil, err
+	}
+	for i := frameStartIdx; i < frameEndIdx; i++ {
+		if wfr.FilterColIdx != noFilterIdx {
+			row, err := wfr.Rows.GetRow(ctx, i)
+			if err != nil {
+				return nil, err
+			}
+			datum, err := row.GetDatum(wfr.FilterColIdx)
+			if err != nil {
+				return nil, err
+			}
+			if datum != tree.DBoolTrue {
+				continue
+			}
+		}
+		args, err := wfr.ArgsByRowIdx(ctx, i)
+		if err != nil {
+			return nil, err
 		}
-		args := wfr.ArgsByRowIdx(i)
 		var value tree.Datum
 		var others tree.Datums
 		// COUNT_ROWS takes no arguments.
@@ -428,13 +466,17 @@ var errInvalidArgumentForNtile = pgerror.NewErrorf(
 	pgerror.CodeInvalidParameterValueError, "argument of ntile() must be greater than zero")
 
 func (w *ntileWindow) Compute(
-	_ context.Context, _ *tree.EvalContext, wfr *tree.WindowFrameRun,
+	ctx context.Context, _ *tree.EvalContext, wfr *tree.WindowFrameRun,
 ) (tree.Datum, error) {
 	if w.ntile == nil {
 		// If this is the first call to ntileWindow.Compute, set up the buckets.
 		total := wfr.PartitionSize()
 
-		arg := wfr.Args()[0]
+		args, err := wfr.Args(ctx)
+		if err != nil {
+			return nil, err
+		}
+		arg := args[0]
 		if arg == tree.DNull {
 			// per spec: If argument is the null value, then the result is the null value.
 			return tree.DNull, nil
@@ -498,11 +540,15 @@ func makeLeadLagWindowConstructor(
 }
 
 func (w *leadLagWindow) Compute(
-	_ context.Context, _ *tree.EvalContext, wfr *tree.WindowFrameRun,
+	ctx context.Context, _ *tree.EvalContext, wfr *tree.WindowFrameRun,
 ) (tree.Datum, error) {
 	offset := 1
 	if w.withOffset {
-		offsetArg := wfr.Args()[1]
+		args, err := wfr.Args(ctx)
+		if err != nil {
+			return nil, err
+		}
+		offsetArg := args[1]
 		if offsetArg == tree.DNull {
 			return tree.DNull, nil
 		}
@@ -516,12 +562,20 @@ func (w *leadLagWindow) Compute(
 		// Target row is out of the partition; supply default value if provided,
 		// otherwise return NULL.
 		if w.withDefault {
-			return wfr.Args()[2], nil
+			args, err := wfr.Args(ctx)
+			if err != nil {
+				return nil, err
+			}
+			return args[2], nil
 		}
 		return tree.DNull, nil
 	}
 
-	return wfr.ArgsWithRowOffset(offset)[0], nil
+	args, err := wfr.ArgsWithRowOffset(ctx, offset)
+	if err != nil {
+		return nil, err
+	}
+	return args[0], nil
 }
 
 func (w *leadLagWindow) Close(context.Context, *tree.EvalContext) {}
@@ -534,9 +588,17 @@ func newFirstValueWindow([]types.T, *tree.EvalContext) tree.WindowFunc {
 }
 
 func (firstValueWindow) Compute(
-	_ context.Context, evalCtx *tree.EvalContext, wfr *tree.WindowFrameRun,
+	ctx context.Context, evalCtx *tree.EvalContext, wfr *tree.WindowFrameRun,
 ) (tree.Datum, error) {
-	return wfr.Rows.GetRow(wfr.FrameStartIdx(evalCtx)).GetDatum(wfr.ArgIdxStart), nil
+	frameStartIdx, err := wfr.FrameStartIdx(ctx, evalCtx)
+	if err != nil {
+		return nil, err
+	}
+	row, err := wfr.Rows.GetRow(ctx, frameStartIdx)
+	if err != nil {
+		return nil, err
+	}
+	return row.GetDatum(wfr.ArgIdxStart)
 }
 
 func (firstValueWindow) Close(context.Context, *tree.EvalContext) {}
@@ -549,9 +611,17 @@ func newLastValueWindow([]types.T, *tree.EvalContext) tree.WindowFunc {
 }
 
 func (lastValueWindow) Compute(
-	_ context.Context, evalCtx *tree.EvalContext, wfr *tree.WindowFrameRun,
+	ctx context.Context, evalCtx *tree.EvalContext, wfr *tree.WindowFrameRun,
 ) (tree.Datum, error) {
-	return wfr.Rows.GetRow(wfr.FrameEndIdx(evalCtx) - 1).GetDatum(wfr.ArgIdxStart), nil
+	frameEndIdx, err := wfr.FrameEndIdx(ctx, evalCtx)
+	if err != nil {
+		return nil, err
+	}
+	row, err := wfr.Rows.GetRow(ctx, frameEndIdx-1)
+	if err != nil {
+		return nil, err
+	}
+	return row.GetDatum(wfr.ArgIdxStart)
 }
 
 func (lastValueWindow) Close(context.Context, *tree.EvalContext) {}
@@ -568,9 +638,13 @@ var errInvalidArgumentForNthValue = pgerror.NewErrorf(
 	pgerror.CodeInvalidParameterValueError, "argument of nth_value() must be greater than zero")
 
 func (nthValueWindow) Compute(
-	_ context.Context, evalCtx *tree.EvalContext, wfr *tree.WindowFrameRun,
+	ctx context.Context, evalCtx *tree.EvalContext, wfr *tree.WindowFrameRun,
 ) (tree.Datum, error) {
-	arg := wfr.Args()[1]
+	args, err := wfr.Args(ctx)
+	if err != nil {
+		return nil, err
+	}
+	arg := args[1]
 	if arg == tree.DNull {
 		return tree.DNull, nil
 	}
@@ -580,10 +654,23 @@ func (nthValueWindow) Compute(
 		return nil, errInvalidArgumentForNthValue
 	}
 
-	if nth > wfr.FrameSize(evalCtx) {
+	frameSize, err := wfr.FrameSize(ctx, evalCtx)
+	if err != nil {
+		return nil, err
+	}
+	if nth > frameSize {
 		return tree.DNull, nil
 	}
-	return wfr.Rows.GetRow(wfr.FrameStartIdx(evalCtx) + nth - 1).GetDatum(wfr.ArgIdxStart), nil
+
+	frameStartIdx, err := wfr.FrameStartIdx(ctx, evalCtx)
+	if err != nil {
+		return nil, err
+	}
+	row, err := wfr.Rows.GetRow(ctx, frameStartIdx+nth-1)
+	if err != nil {
+		return nil, err
+	}
+	return row.GetDatum(wfr.ArgIdxStart)
 }
 
 func (nthValueWindow) Close(context.Context, *tree.EvalContext) {}
