@@ -21,12 +21,11 @@ import (
 	"math"
 	"strings"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
+	"github.com/pkg/errors"
 )
 
 // Constant is an constant literal expression which may be resolved to more than one type.
@@ -396,7 +395,7 @@ func (expr *StrVal) RawString() string {
 
 // Format implements the NodeFormatter interface.
 func (expr *StrVal) Format(ctx *FmtCtx) {
-	buf, f := ctx.Buffer, ctx.flags
+	buf, f := &ctx.Buffer, ctx.flags
 	if expr.scannedAsBytes {
 		lex.EncodeSQLBytes(buf, expr.s)
 	} else {
@@ -421,6 +420,7 @@ var (
 		types.UUID,
 		types.INet,
 		types.JSON,
+		types.BitArray,
 	}
 	// StrValAvailBytes is the set of types convertible to byte array.
 	StrValAvailBytes = []types.T{types.Bytes, types.UUID, types.String}
