@@ -29,10 +29,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
-
-	"golang.org/x/text/language"
-
 	"github.com/pkg/errors"
+	"golang.org/x/text/language"
 )
 
 // CreateDatabase represents a CREATE DATABASE statement.
@@ -54,19 +52,19 @@ func (node *CreateDatabase) Format(ctx *FmtCtx) {
 	ctx.FormatNode(&node.Name)
 	if node.Template != "" {
 		ctx.WriteString(" TEMPLATE = ")
-		lex.EncodeSQLStringWithFlags(ctx.Buffer, node.Template, ctx.flags.EncodeFlags())
+		lex.EncodeSQLStringWithFlags(&ctx.Buffer, node.Template, ctx.flags.EncodeFlags())
 	}
 	if node.Encoding != "" {
 		ctx.WriteString(" ENCODING = ")
-		lex.EncodeSQLStringWithFlags(ctx.Buffer, node.Encoding, ctx.flags.EncodeFlags())
+		lex.EncodeSQLStringWithFlags(&ctx.Buffer, node.Encoding, ctx.flags.EncodeFlags())
 	}
 	if node.Collate != "" {
 		ctx.WriteString(" LC_COLLATE = ")
-		lex.EncodeSQLStringWithFlags(ctx.Buffer, node.Collate, ctx.flags.EncodeFlags())
+		lex.EncodeSQLStringWithFlags(&ctx.Buffer, node.Collate, ctx.flags.EncodeFlags())
 	}
 	if node.CType != "" {
 		ctx.WriteString(" LC_CTYPE = ")
-		lex.EncodeSQLStringWithFlags(ctx.Buffer, node.CType, ctx.flags.EncodeFlags())
+		lex.EncodeSQLStringWithFlags(&ctx.Buffer, node.CType, ctx.flags.EncodeFlags())
 	}
 }
 
@@ -102,7 +100,7 @@ func (l *IndexElemList) Format(ctx *FmtCtx) {
 // CreateIndex represents a CREATE INDEX statement.
 type CreateIndex struct {
 	Name        Name
-	Table       NormalizableTableName
+	Table       TableName
 	Unique      bool
 	Inverted    bool
 	IfNotExists bool
@@ -207,10 +205,11 @@ type ColumnTableDef struct {
 	}
 	CheckExprs []ColumnTableDefCheckExpr
 	References struct {
-		Table          NormalizableTableName
+		Table          *TableName
 		Col            Name
 		ConstraintName Name
 		Actions        ReferenceActions
+		Match          CompositeKeyMatchMethod
 	}
 	Computed struct {
 		Computed bool
@@ -312,10 +311,11 @@ func NewColumnTableDef(
 				return nil, pgerror.NewErrorf(pgerror.CodeInvalidTableDefinitionError,
 					"multiple foreign key constraints specified for column %q", name)
 			}
-			d.References.Table = t.Table
+			d.References.Table = &t.Table
 			d.References.Col = t.Col
 			d.References.ConstraintName = c.Name
 			d.References.Actions = t.Actions
+			d.References.Match = t.Match
 		case *ColumnComputedDef:
 			d.Computed.Computed = true
 			d.Computed.Expr = t.Expr
@@ -346,7 +346,7 @@ func (node *ColumnTableDef) HasDefaultExpr() bool {
 
 // HasFKConstraint returns if the ColumnTableDef has a foreign key constraint.
 func (node *ColumnTableDef) HasFKConstraint() bool {
-	return node.References.Table.TableNameReference != nil
+	return node.References.Table != nil
 }
 
 // IsComputed returns if the ColumnTableDef is a computed column.
@@ -363,7 +363,7 @@ func (node *ColumnTableDef) HasColumnFamily() bool {
 func (node *ColumnTableDef) Format(ctx *FmtCtx) {
 	ctx.FormatNode(&node.Name)
 	ctx.WriteByte(' ')
-	node.Type.Format(ctx.Buffer, ctx.flags.EncodeFlags())
+	node.Type.Format(&ctx.Buffer, ctx.flags.EncodeFlags())
 	if node.Nullable.Nullability != SilentNull && node.Nullable.ConstraintName != "" {
 		ctx.WriteString(" CONSTRAINT ")
 		ctx.FormatNode(&node.Nullable.ConstraintName)
@@ -408,12 +408,16 @@ func (node *ColumnTableDef) Format(ctx *FmtCtx) {
 			ctx.FormatNode(&node.References.ConstraintName)
 		}
 		ctx.WriteString(" REFERENCES ")
-		ctx.FormatNode(&node.References.Table)
+		ctx.FormatNode(node.References.Table)
 		if node.References.Col != "" {
 			ctx.WriteString(" (")
 			ctx.FormatNode(&node.References.Col)
 			ctx.WriteByte(')')
 		}
+		if node.References.Match != MatchSimple {
+			ctx.WriteByte(' ')
+			ctx.WriteString(node.References.Match.String())
+		}
 		ctx.FormatNode(&node.References.Actions)
 	}
 	if node.IsComputed() {
@@ -488,9 +492,10 @@ type ColumnCheckConstraint struct {
 
 // ColumnFKConstraint represents a FK-constaint on a column.
 type ColumnFKConstraint struct {
-	Table   NormalizableTableName
+	Table   TableName
 	Col     Name // empty-string means use PK
 	Actions ReferenceActions
+	Match   CompositeKeyMatchMethod
 }
 
 // ColumnComputedDef represents the description of a computed column.
@@ -637,13 +642,37 @@ func (node *ReferenceActions) Format(ctx *FmtCtx) {
 	}
 }
 
+// CompositeKeyMatchMethod is the algorithm use when matching composite keys.
+// See https://github.com/cockroachdb/cockroach/issues/20305 or
+// https://www.postgresql.org/docs/11/sql-createtable.html for details on the
+// different composite foreign key matching methods.
+type CompositeKeyMatchMethod int
+
+// The values for CompositeKeyMatchMethod.
+const (
+	MatchSimple CompositeKeyMatchMethod = iota
+	MatchFull
+	MatchPartial // Note: PARTIAL not actually supported at this point.
+)
+
+var compositeKeyMatchMethodName = [...]string{
+	MatchSimple:  "MATCH SIMPLE",
+	MatchFull:    "MATCH FULL",
+	MatchPartial: "MATCH PARTIAL",
+}
+
+func (c CompositeKeyMatchMethod) String() string {
+	return compositeKeyMatchMethodName[c]
+}
+
 // ForeignKeyConstraintTableDef represents a FOREIGN KEY constraint in the AST.
 type ForeignKeyConstraintTableDef struct {
 	Name     Name
-	Table    NormalizableTableName
+	Table    TableName
 	FromCols NameList
 	ToCols   NameList
 	Actions  ReferenceActions
+	Match    CompositeKeyMatchMethod
 }
 
 // Format implements the NodeFormatter interface.
@@ -665,6 +694,11 @@ func (node *ForeignKeyConstraintTableDef) Format(ctx *FmtCtx) {
 		ctx.WriteByte(')')
 	}
 
+	if node.Match != MatchSimple {
+		ctx.WriteByte(' ')
+		ctx.WriteString(node.Match.String())
+	}
+
 	ctx.FormatNode(&node.Actions)
 }
 
@@ -730,7 +764,7 @@ func (node *FamilyTableDef) Format(ctx *FmtCtx) {
 // InterleaveDef represents an interleave definition within a CREATE TABLE
 // or CREATE INDEX statement.
 type InterleaveDef struct {
-	Parent       *NormalizableTableName
+	Parent       TableName
 	Fields       NameList
 	DropBehavior DropBehavior
 }
@@ -738,7 +772,7 @@ type InterleaveDef struct {
 // Format implements the NodeFormatter interface.
 func (node *InterleaveDef) Format(ctx *FmtCtx) {
 	ctx.WriteString(" INTERLEAVE IN PARENT ")
-	ctx.FormatNode(node.Parent)
+	ctx.FormatNode(&node.Parent)
 	ctx.WriteString(" (")
 	for i := range node.Fields {
 		if i > 0 {
@@ -844,7 +878,7 @@ func (node *RangePartition) Format(ctx *FmtCtx) {
 // CreateTable represents a CREATE TABLE statement.
 type CreateTable struct {
 	IfNotExists   bool
-	Table         NormalizableTableName
+	Table         TableName
 	Interleave    *InterleaveDef
 	PartitionBy   *PartitionBy
 	Defs          TableDefs
@@ -865,6 +899,12 @@ func (node *CreateTable) Format(ctx *FmtCtx) {
 		ctx.WriteString("IF NOT EXISTS ")
 	}
 	ctx.FormatNode(&node.Table)
+	node.FormatBody(ctx)
+}
+
+// FormatBody formats the "body" of the create table definition - everything
+// but the CREATE TABLE tableName part.
+func (node *CreateTable) FormatBody(ctx *FmtCtx) {
 	if node.As() {
 		if len(node.AsColumnNames) > 0 {
 			ctx.WriteString(" (")
@@ -930,13 +970,14 @@ func (node *CreateTable) HoistConstraints() {
 					targetCol = append(targetCol, col.References.Col)
 				}
 				node.Defs = append(node.Defs, &ForeignKeyConstraintTableDef{
-					Table:    col.References.Table,
+					Table:    *col.References.Table,
 					FromCols: NameList{col.Name},
 					ToCols:   targetCol,
 					Name:     col.References.ConstraintName,
 					Actions:  col.References.Actions,
+					Match:    col.References.Match,
 				})
-				col.References.Table = NormalizableTableName{}
+				col.References.Table = nil
 			}
 		}
 	}
@@ -945,7 +986,7 @@ func (node *CreateTable) HoistConstraints() {
 // CreateSequence represents a CREATE SEQUENCE statement.
 type CreateSequence struct {
 	IfNotExists bool
-	Name        NormalizableTableName
+	Name        TableName
 	Options     SequenceOptions
 }
 
@@ -1100,7 +1141,7 @@ func (node *CreateRole) Format(ctx *FmtCtx) {
 
 // CreateView represents a CREATE VIEW statement.
 type CreateView struct {
-	Name        NormalizableTableName
+	Name        TableName
 	ColumnNames NameList
 	AsSource    *Select
 }
@@ -1125,7 +1166,8 @@ func (node *CreateView) Format(ctx *FmtCtx) {
 type CreateStats struct {
 	Name        Name
 	ColumnNames NameList
-	Table       NormalizableTableName
+	Table       TableExpr
+	Options     CreateStatsOptions
 }
 
 // Format implements the NodeFormatter interface.
@@ -1133,9 +1175,63 @@ func (node *CreateStats) Format(ctx *FmtCtx) {
 	ctx.WriteString("CREATE STATISTICS ")
 	ctx.FormatNode(&node.Name)
 
-	ctx.WriteString(" ON ")
-	ctx.FormatNode(&node.ColumnNames)
+	if len(node.ColumnNames) > 0 {
+		ctx.WriteString(" ON ")
+		ctx.FormatNode(&node.ColumnNames)
+	}
 
 	ctx.WriteString(" FROM ")
-	ctx.FormatNode(&node.Table)
+	ctx.FormatNode(node.Table)
+
+	if !node.Options.Empty() {
+		ctx.WriteString(" WITH OPTIONS ")
+		ctx.FormatNode(&node.Options)
+	}
+}
+
+// CreateStatsOptions contains options for CREATE STATISTICS.
+type CreateStatsOptions struct {
+	// Throttling enables throttling and indicates the fraction of time we are
+	// idling (between 0 and 1).
+	Throttling float64
+
+	// AsOf performs a historical read at the given timestamp.
+	AsOf AsOfClause
+}
+
+// Empty returns true if no options were provided.
+func (o *CreateStatsOptions) Empty() bool {
+	return o.Throttling == 0 && o.AsOf.Expr == nil
+}
+
+// Format implements the NodeFormatter interface.
+func (o *CreateStatsOptions) Format(ctx *FmtCtx) {
+	sep := ""
+	if o.Throttling != 0 {
+		fmt.Fprintf(ctx, "THROTTLING %g", o.Throttling)
+		sep = " "
+	}
+	if o.AsOf.Expr != nil {
+		ctx.WriteString(sep)
+		ctx.FormatNode(&o.AsOf)
+		sep = " "
+	}
+}
+
+// CombineWith combines two options, erroring out if the two options contain
+// incompatible settings.
+func (o *CreateStatsOptions) CombineWith(other *CreateStatsOptions) error {
+	if other.Throttling != 0 {
+		if o.Throttling != 0 {
+			return errors.New("THROTTLING specified multiple times")
+		}
+		o.Throttling = other.Throttling
+	}
+	if other.AsOf.Expr != nil {
+		if o.AsOf.Expr != nil {
+			return errors.New("AS OF specified multiple times")
+		}
+		o.AsOf = other.AsOf
+	}
+	return nil
 }
