@@ -20,7 +20,6 @@ import (
 	"math"
 	"math/big"
 	"net"
-	"regexp"
 	"sort"
 	"strconv"
 	"strings"
@@ -28,25 +27,26 @@ import (
 	"unicode"
 	"unsafe"
 
-	"github.com/lib/pq/oid"
-	"github.com/pkg/errors"
-	"golang.org/x/text/collate"
-	"golang.org/x/text/language"
-
 	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
+	"github.com/cockroachdb/cockroach/pkg/util/bitarray"
 	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/ipaddr"
 	"github.com/cockroachdb/cockroach/pkg/util/json"
 	"github.com/cockroachdb/cockroach/pkg/util/stringencoding"
 	"github.com/cockroachdb/cockroach/pkg/util/timeofday"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/cockroachdb/cockroach/pkg/util/timeutil/pgdate"
 	"github.com/cockroachdb/cockroach/pkg/util/uint128"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/lib/pq/oid"
+	"github.com/pkg/errors"
+	"golang.org/x/text/collate"
+	"golang.org/x/text/language"
 )
 
 var (
@@ -142,13 +142,6 @@ type Datums []Datum
 // Len returns the number of Datum values.
 func (d Datums) Len() int { return len(d) }
 
-// Reverse reverses the order of the Datum values.
-func (d Datums) Reverse() {
-	for i, j := 0, d.Len()-1; i < j; i, j = i+1, j-1 {
-		d[i], d[j] = d[j], d[i]
-	}
-}
-
 // Format implements the NodeFormatter interface.
 func (d *Datums) Format(ctx *FmtCtx) {
 	ctx.WriteByte('(')
@@ -364,10 +357,15 @@ func (d *DBool) Compare(ctx *EvalContext, other Datum) int {
 	if !ok {
 		panic(makeUnsupportedComparisonMessage(d, other))
 	}
-	if !*d && *v {
+	return CompareBools(bool(*d), bool(*v))
+}
+
+// CompareBools compares the input bools according to the SQL comparison rules.
+func CompareBools(d, v bool) int {
+	if !d && v {
 		return -1
 	}
-	if *d && !*v {
+	if d && !v {
 		return 1
 	}
 	return 0
@@ -424,6 +422,160 @@ func (d *DBool) Size() uintptr {
 	return unsafe.Sizeof(*d)
 }
 
+// DBitArray is the BIT/VARBIT Datum.
+type DBitArray struct {
+	bitarray.BitArray
+}
+
+// ParseDBitArray parses a string representation of binary digits.
+func ParseDBitArray(s string) (*DBitArray, error) {
+	var a DBitArray
+	var err error
+	a.BitArray, err = bitarray.Parse(s)
+	if err != nil {
+		return nil, err
+	}
+	return &a, nil
+}
+
+// NewDBitArray returns a DBitArray.
+func NewDBitArray(bitLen uint) *DBitArray {
+	a := MakeDBitArray(bitLen)
+	return &a
+}
+
+// MakeDBitArray returns a DBitArray.
+func MakeDBitArray(bitLen uint) DBitArray {
+	return DBitArray{BitArray: bitarray.MakeZeroBitArray(bitLen)}
+}
+
+// MustBeDBitArray attempts to retrieve a DBitArray from an Expr, panicking if the
+// assertion fails.
+func MustBeDBitArray(e Expr) *DBitArray {
+	b, ok := AsDBitArray(e)
+	if !ok {
+		panic(pgerror.NewAssertionErrorf("expected *DBitArray, found %T", e))
+	}
+	return b
+}
+
+// AsDBitArray attempts to retrieve a *DBitArray from an Expr, returning a *DBitArray and
+// a flag signifying whether the assertion was successful. The function should
+// be used instead of direct type assertions.
+func AsDBitArray(e Expr) (*DBitArray, bool) {
+	switch t := e.(type) {
+	case *DBitArray:
+		return t, true
+	}
+	return nil, false
+}
+
+var errCannotCastNegativeIntToBitArray = pgerror.NewErrorf(pgerror.CodeCannotCoerceError,
+	"cannot cast negative integer to bit varying with unbounded width")
+
+// NewDBitArrayFromInt creates a bit array from the specified integer
+// at the specified width.
+// If the width is zero, only positive integers can be converted.
+// If the width is nonzero, the value is truncated to that width.
+// Negative values are encoded using two's complement.
+func NewDBitArrayFromInt(i int64, width uint) (*DBitArray, error) {
+	if width == 0 && i < 0 {
+		return nil, errCannotCastNegativeIntToBitArray
+	}
+	return &DBitArray{
+		BitArray: bitarray.MakeBitArrayFromInt64(width, i, 64),
+	}, nil
+}
+
+// AsDInt computes the integer value of the given bit array.
+// The value is assumed to be encoded using two's complement.
+// The result is truncated to the given integer number of bits,
+// if specified.
+// The given width must be 64 or smaller. The results are undefined
+// if n is greater than 64.
+func (d *DBitArray) AsDInt(n uint) *DInt {
+	if n == 0 {
+		n = 64
+	}
+	return NewDInt(DInt(d.BitArray.AsInt64(n)))
+}
+
+// ResolvedType implements the TypedExpr interface.
+func (*DBitArray) ResolvedType() types.T {
+	return types.BitArray
+}
+
+// Compare implements the Datum interface.
+func (d *DBitArray) Compare(ctx *EvalContext, other Datum) int {
+	if other == DNull {
+		// NULL is less than any non-NULL value.
+		return 1
+	}
+	v, ok := UnwrapDatum(ctx, other).(*DBitArray)
+	if !ok {
+		panic(makeUnsupportedComparisonMessage(d, other))
+	}
+	return bitarray.Compare(d.BitArray, v.BitArray)
+}
+
+// Prev implements the Datum interface.
+func (d *DBitArray) Prev(_ *EvalContext) (Datum, bool) {
+	return nil, false
+}
+
+// Next implements the Datum interface.
+func (d *DBitArray) Next(_ *EvalContext) (Datum, bool) {
+	a := bitarray.Next(d.BitArray)
+	return &DBitArray{BitArray: a}, true
+}
+
+// IsMax implements the Datum interface.
+func (d *DBitArray) IsMax(_ *EvalContext) bool {
+	return false
+}
+
+// IsMin implements the Datum interface.
+func (d *DBitArray) IsMin(_ *EvalContext) bool {
+	return d.BitArray.IsEmpty()
+}
+
+var bitArrayZero = NewDBitArray(0)
+
+// Min implements the Datum interface.
+func (d *DBitArray) Min(_ *EvalContext) (Datum, bool) {
+	return bitArrayZero, true
+}
+
+// Max implements the Datum interface.
+func (d *DBitArray) Max(_ *EvalContext) (Datum, bool) {
+	return nil, false
+}
+
+// AmbiguousFormat implements the Datum interface.
+func (*DBitArray) AmbiguousFormat() bool { return false }
+
+// Format implements the NodeFormatter interface.
+func (d *DBitArray) Format(ctx *FmtCtx) {
+	f := ctx.flags
+	if f.HasFlags(fmtPgwireFormat) {
+		d.BitArray.Format(&ctx.Buffer)
+	} else {
+		withQuotes := !f.HasFlags(FmtFlags(lex.EncBareStrings))
+		if withQuotes {
+			ctx.WriteString("B'")
+		}
+		d.BitArray.Format(&ctx.Buffer)
+		if withQuotes {
+			ctx.WriteByte('\'')
+		}
+	}
+}
+
+// Size implements the Datum interface.
+func (d *DBitArray) Size() uintptr {
+	return d.BitArray.Sizeof()
+}
+
 // DInt is the int Datum.
 type DInt int64
 
@@ -756,7 +908,7 @@ func (d *DDecimal) Compare(ctx *EvalContext, other Datum) int {
 	case *DDecimal:
 		v = &t.Decimal
 	case *DInt:
-		v.SetCoefficient(int64(*t)).SetExponent(0)
+		v.SetFinite(int64(*t), 0)
 	case *DFloat:
 		if _, err := v.SetFloat64(float64(*t)); err != nil {
 			panic(err)
@@ -764,6 +916,12 @@ func (d *DDecimal) Compare(ctx *EvalContext, other Datum) int {
 	default:
 		panic(makeUnsupportedComparisonMessage(d, other))
 	}
+	return CompareDecimals(&d.Decimal, v)
+}
+
+// CompareDecimals compares 2 apd.Decimals according to the SQL comparison
+// rules, making sure that NaNs sort first.
+func CompareDecimals(d *apd.Decimal, v *apd.Decimal) int {
 	// NaNs sort first in SQL.
 	if dn, vn := d.Form == apd.NaN, v.Form == apd.NaN; dn && !vn {
 		return -1
@@ -957,8 +1115,8 @@ func (*DString) AmbiguousFormat() bool { return true }
 
 // Format implements the NodeFormatter interface.
 func (d *DString) Format(ctx *FmtCtx) {
-	buf, f := ctx.Buffer, ctx.flags
-	if f.HasFlags(fmtUnicodeStrings) {
+	buf, f := &ctx.Buffer, ctx.flags
+	if f.HasFlags(fmtRawStrings) {
 		buf.WriteString(string(*d))
 	} else {
 		lex.EncodeSQLStringWithFlags(buf, string(*d), f.EncodeFlags())
@@ -1026,9 +1184,9 @@ func (*DCollatedString) AmbiguousFormat() bool { return false }
 
 // Format implements the NodeFormatter interface.
 func (d *DCollatedString) Format(ctx *FmtCtx) {
-	lex.EncodeSQLString(ctx.Buffer, d.Contents)
+	lex.EncodeSQLString(&ctx.Buffer, d.Contents)
 	ctx.WriteString(" COLLATE ")
-	lex.EncodeUnrestrictedSQLIdent(ctx.Buffer, d.Locale, lex.EncNoFlags)
+	lex.EncodeUnrestrictedSQLIdent(&ctx.Buffer, d.Locale, lex.EncNoFlags)
 }
 
 // ResolvedType implements the TypedExpr interface.
@@ -1267,8 +1425,8 @@ func (d *DUuid) IsMin(_ *EvalContext) bool {
 }
 
 var dMinUUID = NewDUuid(DUuid{uuid.UUID{}})
-var dMaxUUID = NewDUuid(DUuid{uuid.UUID{UUID: [16]byte{0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
-	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}}})
+var dMaxUUID = NewDUuid(DUuid{uuid.UUID{0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
+	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}})
 
 // Min implements the Datum interface.
 func (*DUuid) Min(_ *EvalContext) (Datum, bool) {
@@ -1478,11 +1636,60 @@ func NewDDateFromTime(t time.Time, loc *time.Location) *DDate {
 	return NewDDate(DDate(secs / SecondsInDay))
 }
 
+// ParseTimeContext provides the information necessary for
+// parsing dates, times, and timestamps. A nil value is generally
+// acceptable and will result in reasonable defaults being applied.
+type ParseTimeContext interface {
+	duration.Context
+	// GetRelativeParseTime returns the transaction time in the session's
+	// timezone (i.e. now()). This is used to calculate relative dates,
+	// like "tomorrow", and also provides a default time.Location for
+	// parsed times.
+	GetRelativeParseTime() time.Time
+}
+
+var _ ParseTimeContext = &EvalContext{}
+var _ ParseTimeContext = &SemaContext{}
+var _ ParseTimeContext = &simpleParseTimeContext{}
+
+// NewParseTimeContext constructs a ParseTimeContext that returns
+// the given values.
+func NewParseTimeContext(mode duration.AdditionMode, relativeParseTime time.Time) ParseTimeContext {
+	return &simpleParseTimeContext{
+		AdditionMode:      mode,
+		RelativeParseTime: relativeParseTime,
+	}
+}
+
+type simpleParseTimeContext struct {
+	AdditionMode      duration.AdditionMode
+	RelativeParseTime time.Time
+}
+
+// GetAdditionMode implements ParseTimeContext.
+func (ctx simpleParseTimeContext) GetAdditionMode() duration.AdditionMode {
+	return ctx.AdditionMode
+}
+
+// GetRelativeParseTime implements ParseTimeContext.
+func (ctx simpleParseTimeContext) GetRelativeParseTime() time.Time {
+	return ctx.RelativeParseTime
+}
+
+// relativeParseTime chooses a reasonable "now" value for
+// performing date parsing.
+func relativeParseTime(ctx ParseTimeContext) time.Time {
+	if ctx == nil {
+		return timeutil.Now()
+	}
+	return ctx.GetRelativeParseTime()
+}
+
 // ParseDDate parses and returns the *DDate Datum value represented by the provided
 // string in the provided location, or an error if parsing is unsuccessful.
-func ParseDDate(s string, loc *time.Location) (*DDate, error) {
-	// No need to ParseInLocation here because we're only parsing dates.
-	t, err := parseTimestampInLocation(s, time.UTC, types.Date)
+func ParseDDate(ctx ParseTimeContext, s string) (*DDate, error) {
+	now := relativeParseTime(ctx)
+	t, err := pgdate.ParseDate(now, 0 /* mode */, s)
 	if err != nil {
 		return nil, err
 	}
@@ -1583,8 +1790,9 @@ func MakeDTime(t timeofday.TimeOfDay) *DTime {
 
 // ParseDTime parses and returns the *DTime Datum value represented by the
 // provided string, or an error if parsing is unsuccessful.
-func ParseDTime(s string) (*DTime, error) {
-	t, err := parseTimestampInLocation("1970-01-01 "+s, time.UTC, types.Time)
+func ParseDTime(ctx ParseTimeContext, s string) (*DTime, error) {
+	now := relativeParseTime(ctx)
+	t, err := pgdate.ParseTime(now, 0 /* mode */, s)
 	if err != nil {
 		// Build our own error message to avoid exposing the dummy date.
 		return nil, makeParseError(s, types.Time, nil)
@@ -1674,129 +1882,23 @@ func MakeDTimestamp(t time.Time, precision time.Duration) *DTimestamp {
 
 // time.Time formats.
 const (
-	dateFormat                = "2006-01-02"
-	dateFormatWithOffset      = dateFormat + " -070000"
-	dateFormatNoPad           = "2006-1-2"
-	dateFormatNoPadWithOffset = dateFormatNoPad + " -070000"
-
-	timestampFormat                      = dateFormatNoPad + " 15:04:05"
-	timestampWithOffsetZoneFormat        = timestampFormat + "-07"
-	timestampWithOffsetMinutesZoneFormat = timestampWithOffsetZoneFormat + ":00"
-	timestampWithOffsetSecondsZoneFormat = timestampWithOffsetMinutesZoneFormat + ":00"
-	timestampWithNamedZoneFormat         = timestampFormat + " MST"
-	timestampRFC3339WithoutZoneFormat    = dateFormat + "T15:04:05"
-	timestampSequelizeFormat             = timestampFormat + ".000 -07:00"
-
-	timestampJdbcFormat = timestampFormat + ".999999 -070000"
-	timestampNodeFormat = timestampFormat + ".999999-07:00"
-
-	// See https://github.com/lib/pq/blob/8df6253/encode.go#L480.
-	timestampPgwireFormat = "2006-01-02 15:04:05.999999999Z07:00"
+	dateFormat = "2006-01-02"
 
 	// TimestampOutputFormat is used to output all timestamps.
 	TimestampOutputFormat = "2006-01-02 15:04:05.999999-07:00"
 )
 
-var timeFormats = []string{
-	dateFormat,
-	dateFormatWithOffset,
-	dateFormatNoPad,
-	dateFormatNoPadWithOffset,
-	time.RFC3339Nano,
-	timestampPgwireFormat,
-	timestampWithOffsetZoneFormat,
-	timestampWithOffsetMinutesZoneFormat,
-	timestampWithOffsetSecondsZoneFormat,
-	timestampFormat,
-	timestampWithNamedZoneFormat,
-	timestampRFC3339WithoutZoneFormat,
-	timestampSequelizeFormat,
-	timestampNodeFormat,
-	timestampJdbcFormat,
-}
-
-var (
-	tzMatch        = regexp.MustCompile(` [+-]`)
-	loneZeroRMatch = regexp.MustCompile(`:(\d(?:[^\d]|$))`)
-)
-
-func parseTimestampInLocation(s string, loc *time.Location, typ types.T) (time.Time, error) {
-	origS := s
-	l := len(s)
-	if loneZeroRMatch.MatchString(s) {
-		// HACK: go doesn't handle offsets that are not zero-padded from psql/jdbc.
-		// Thus, if we see `2015-10-05 3:0:5 +0:0:0` we need to change it to
-		// `... 3:00:50 +00:00:00`.
-		s = loneZeroRMatch.ReplaceAllString(s, ":0${1}")
-		// This must be run twice, since ReplaceAllString doesn't touch overlapping
-		// matches and thus wouldn't fix a string of the form 3:3:3.
-		s = loneZeroRMatch.ReplaceAllString(s, ":0${1}")
-	}
-
-	if loc := tzMatch.FindStringIndex(s); loc != nil && l > loc[1] {
-		// Remove `:` characters from timezone specifier and pad to 6 digits. A
-		// leading 0 will be added if there are an odd number of digits in the
-		// specifier, since this is short-hand for an offset with number of hours
-		// equal to the leading digit.
-		// This converts all timezone specifiers to the stdNumSecondsTz format in
-		// time/format.go: `-070000`.
-		tzPos := loc[1]
-		tzSpec := strings.Replace(s[tzPos:], ":", "", -1)
-		if len(tzSpec)%2 == 1 {
-			tzSpec = "0" + tzSpec
-		}
-		if len(tzSpec) < 6 {
-			tzSpec += strings.Repeat("0", 6-len(tzSpec))
-		}
-		s = s[:tzPos] + tzSpec
-	}
-
-	for _, format := range timeFormats {
-		if t, err := time.ParseInLocation(format, s, loc); err == nil {
-			if err := checkForMissingZone(t, loc); err != nil {
-				return time.Time{}, makeParseError(origS, typ, err)
-			}
-			return t, nil
-		}
-	}
-	return time.Time{}, makeParseError(origS, typ, nil)
-}
-
-// Unfortunately Go is very strict when parsing abbreviated zone names -- with
-// the exception of 'UTC' and 'GMT', it only supports abbreviations that are
-// defined in the local in which it is parsing. Worse, it doesn't return any
-// sort of error for unresolved zones, but rather simply pretends they have a
-// zero offset. This means changing the session zone such that an abbreviation
-// like 'CET' stops being resolved *silently* changes the offsets of parsed
-// strings with 'CET' offsets to zero.
-// We attempt to detect when this has happened and return an error instead.
-//
-// Postgres does its own parsing and just maintains a list of zone abbreviations
-// that are always recognized, regardless of the session location. If this check
-// ends up catching too many users, we may need to do the same.
-func checkForMissingZone(t time.Time, parseLoc *time.Location) error {
-	if z, off := t.Zone(); off == 0 && t.Location() != parseLoc && z != "UTC" && !strings.HasPrefix(z, "GMT") {
-		return pgerror.NewErrorf(pgerror.CodeInvalidDatetimeFormatError, "unknown zone %q", z)
-	}
-	return nil
-}
-
 // ParseDTimestamp parses and returns the *DTimestamp Datum value represented by
 // the provided string in UTC, or an error if parsing is unsuccessful.
-func ParseDTimestamp(s string, precision time.Duration) (*DTimestamp, error) {
-	// `ParseInLocation` uses the location provided both for resolving an explicit
-	// abbreviated zone as well as for the default zone if not specified
-	// explicitly. For non-'WITH TIME ZONE' strings (which this is used to parse),
-	// we do not want to add a non-UTC zone if one is not explicitly stated, so we
-	// use time.UTC rather than the session location. Unfortunately this also means
-	// we do not use the session zone for resolving abbreviations.
-	t, err := parseTimestampInLocation(s, time.UTC, types.Timestamp)
+func ParseDTimestamp(ctx ParseTimeContext, s string, precision time.Duration) (*DTimestamp, error) {
+	now := relativeParseTime(ctx)
+	t, err := pgdate.ParseTimestamp(now, 0 /* mode */, s)
 	if err != nil {
 		return nil, err
 	}
 	// Truncate the timezone. DTimestamp doesn't carry its timezone around.
 	_, offset := t.Zone()
-	t = duration.Add(t, duration.FromInt64(int64(offset))).UTC()
+	t = duration.Add(ctx, t, duration.FromInt64(int64(offset))).UTC()
 	return MakeDTimestamp(t, precision), nil
 }
 
@@ -1945,9 +2047,10 @@ func MakeDTimestampTZFromDate(loc *time.Location, d *DDate) *DTimestampTZ {
 // ParseDTimestampTZ parses and returns the *DTimestampTZ Datum value represented by
 // the provided string in the provided location, or an error if parsing is unsuccessful.
 func ParseDTimestampTZ(
-	s string, loc *time.Location, precision time.Duration,
+	ctx ParseTimeContext, s string, precision time.Duration,
 ) (*DTimestampTZ, error) {
-	t, err := parseTimestampInLocation(s, loc, types.TimestampTZ)
+	now := relativeParseTime(ctx)
+	t, err := pgdate.ParseTimestamp(now, 0 /* mode */, s)
 	if err != nil {
 		return nil, err
 	}
@@ -2030,7 +2133,7 @@ func (d *DTimestampTZ) Size() uintptr {
 //             '2012-01-01 12:00:00'.
 func (d *DTimestampTZ) stripTimeZone(ctx *EvalContext) *DTimestamp {
 	_, locOffset := d.Time.In(ctx.GetLocation()).Zone()
-	newTime := duration.Add(d.Time.UTC(), duration.FromInt64(int64(locOffset)))
+	newTime := duration.Add(ctx, d.Time.UTC(), duration.FromInt64(int64(locOffset)))
 	return MakeDTimestamp(newTime, time.Microsecond)
 }
 
@@ -2063,7 +2166,7 @@ const (
 // ParseDInterval parses and returns the *DInterval Datum value represented by the provided
 // string, or an error if parsing is unsuccessful.
 func ParseDInterval(s string) (*DInterval, error) {
-	return parseDInterval(s, Second)
+	return ParseDIntervalWithField(s, Second)
 }
 
 // truncateDInterval truncates the input DInterval downward to the nearest
@@ -2073,16 +2176,16 @@ func truncateDInterval(d *DInterval, field DurationField) {
 	case Year:
 		d.Duration.Months = d.Duration.Months - d.Duration.Months%12
 		d.Duration.Days = 0
-		d.Duration.Nanos = 0
+		d.Duration.SetNanos(0)
 	case Month:
 		d.Duration.Days = 0
-		d.Duration.Nanos = 0
+		d.Duration.SetNanos(0)
 	case Day:
-		d.Duration.Nanos = 0
+		d.Duration.SetNanos(0)
 	case Hour:
-		d.Duration.Nanos = d.Duration.Nanos - d.Duration.Nanos%time.Hour.Nanoseconds()
+		d.Duration.SetNanos(d.Duration.Nanos() - d.Duration.Nanos()%time.Hour.Nanoseconds())
 	case Minute:
-		d.Duration.Nanos = d.Duration.Nanos - d.Duration.Nanos%time.Minute.Nanoseconds()
+		d.Duration.SetNanos(d.Duration.Nanos() - d.Duration.Nanos()%time.Minute.Nanoseconds())
 	case Second:
 		// Postgres doesn't truncate to whole seconds.
 	}
@@ -2134,13 +2237,13 @@ func parseDInterval(s string, field DurationField) (*DInterval, error) {
 		case Day:
 			ret.Days = int64(f)
 		case Hour:
-			ret.Nanos = time.Hour.Nanoseconds() * int64(f)
+			ret.SetNanos(time.Hour.Nanoseconds() * int64(f))
 		case Minute:
-			ret.Nanos = time.Minute.Nanoseconds() * int64(f)
+			ret.SetNanos(time.Minute.Nanoseconds() * int64(f))
 		case Second:
-			ret.Nanos = int64(float64(time.Second.Nanoseconds()) * f)
+			ret.SetNanos(int64(float64(time.Second.Nanoseconds()) * f))
 		case Millisecond:
-			ret.Nanos = int64(float64(time.Millisecond.Nanoseconds()) * f)
+			ret.SetNanos(int64(float64(time.Millisecond.Nanoseconds()) * f))
 		default:
 			panic(fmt.Sprintf("unhandled DurationField constant %d", field))
 		}
@@ -2194,26 +2297,18 @@ func (d *DInterval) Next(_ *EvalContext) (Datum, bool) {
 
 // IsMax implements the Datum interface.
 func (d *DInterval) IsMax(_ *EvalContext) bool {
-	return d.Months == math.MaxInt64 && d.Days == math.MaxInt64 && d.Nanos == math.MaxInt64
+	return d.Duration == dMaxInterval.Duration
 }
 
 // IsMin implements the Datum interface.
 func (d *DInterval) IsMin(_ *EvalContext) bool {
-	return d.Months == math.MinInt64 && d.Days == math.MinInt64 && d.Nanos == math.MinInt64
+	return d.Duration == dMinInterval.Duration
 }
 
-var dMaxInterval = &DInterval{
-	duration.Duration{
-		Months: math.MaxInt64,
-		Days:   math.MaxInt64,
-		Nanos:  math.MaxInt64,
-	}}
-var dMinInterval = &DInterval{
-	duration.Duration{
-		Months: math.MinInt64,
-		Days:   math.MinInt64,
-		Nanos:  math.MinInt64,
-	}}
+var (
+	dMaxInterval = &DInterval{duration.MakeDuration(math.MaxInt64, math.MaxInt64, math.MaxInt64)}
+	dMinInterval = &DInterval{duration.MakeDuration(math.MinInt64, math.MinInt64, math.MinInt64)}
+)
 
 // Max implements the Datum interface.
 func (d *DInterval) Max(_ *EvalContext) (Datum, bool) {
@@ -2240,7 +2335,7 @@ func (d *DInterval) Format(ctx *FmtCtx) {
 	if !bareStrings {
 		ctx.WriteByte('\'')
 	}
-	d.Duration.Format(ctx.Buffer)
+	d.Duration.Format(&ctx.Buffer)
 	if !bareStrings {
 		ctx.WriteByte('\'')
 	}
@@ -2345,7 +2440,16 @@ func AsJSON(d Datum) (json.JSON, error) {
 			builder.Add(fmt.Sprintf("f%d", i+1), j)
 		}
 		return builder.Build(), nil
-	case *DTimestamp, *DTimestampTZ, *DDate, *DUuid, *DOid, *DInterval, *DBytes, *DIPAddr, *DTime:
+	case *DTimestampTZ:
+		// Our normal timestamp-formatting code uses a variation on RFC 3339,
+		// without the T separator. This causes some compatibility problems
+		// with certain JSON consumers, so we'll use an alternate formatting
+		// path here to maintain consistency with PostgreSQL.
+		return json.FromString(t.Time.Format(time.RFC3339Nano)), nil
+	case *DTimestamp:
+		// This is RFC3339Nano, but without the TZ fields.
+		return json.FromString(t.UTC().Format("2006-01-02T15:04:05.999999999")), nil
+	case *DDate, *DUuid, *DOid, *DInterval, *DBytes, *DIPAddr, *DTime, *DBitArray:
 		return json.FromString(AsStringWithFlags(t, FmtBareStrings)), nil
 	default:
 		if d == DNull {
@@ -2419,11 +2523,11 @@ func (d *DJSON) Format(ctx *FmtCtx) {
 	// TODO(justin): ideally the JSON string encoder should know it needs to
 	// escape things to be inside SQL strings in order to avoid this allocation.
 	s := d.JSON.String()
-	if ctx.flags.HasFlags(fmtUnicodeStrings) {
-		ctx.Buffer.WriteString(s)
-		return
+	if ctx.flags.HasFlags(fmtRawStrings) {
+		ctx.WriteString(s)
+	} else {
+		lex.EncodeSQLStringWithFlags(&ctx.Buffer, s, ctx.flags.EncodeFlags())
 	}
-	lex.EncodeSQLStringWithFlags(ctx.Buffer, s, ctx.flags.EncodeFlags())
 }
 
 // Size implements the Datum interface.
@@ -2655,7 +2759,7 @@ func (d *DTuple) Format(ctx *FmtCtx) {
 			// case, if necessary, so just skip the annotation and continue.
 			if err == nil {
 				ctx.WriteString("::")
-				coltype.Format(ctx.Buffer, ctx.flags.EncodeFlags())
+				coltype.Format(&ctx.Buffer, ctx.flags.EncodeFlags())
 			}
 		}
 		comma = ", "
@@ -3113,12 +3217,12 @@ func (d *DOid) Format(ctx *FmtCtx) {
 		ctx.WriteByte('(')
 		d.DInt.Format(ctx)
 		ctx.WriteByte(',')
-		lex.EncodeSQLStringWithFlags(ctx.Buffer, d.name, lex.EncNoFlags)
+		lex.EncodeSQLStringWithFlags(&ctx.Buffer, d.name, lex.EncNoFlags)
 		ctx.WriteByte(')')
 	} else {
 		// This is used to print the name of pseudo-procedures in e.g.
 		// pg_catalog.pg_type.typinput
-		lex.EncodeSQLStringWithFlags(ctx.Buffer, d.name, lex.EncBareStrings)
+		lex.EncodeSQLStringWithFlags(&ctx.Buffer, d.name, lex.EncBareStrings)
 	}
 }
 
@@ -3301,7 +3405,7 @@ func (d *Placeholder) AmbiguousFormat() bool {
 }
 
 func (d *Placeholder) mustGetValue(ctx *EvalContext) Datum {
-	e, ok := ctx.Placeholders.Value(d.Name)
+	e, ok := ctx.Placeholders.Value(d.Idx)
 	if !ok {
 		panic("fail")
 	}
@@ -3434,6 +3538,7 @@ var baseDatumTypeSizes = map[types.T]struct {
 }{
 	types.Unknown:     {unsafe.Sizeof(dNull{}), fixedSize},
 	types.Bool:        {unsafe.Sizeof(DBool(false)), fixedSize},
+	types.BitArray:    {unsafe.Sizeof(DBitArray{}), variableSize},
 	types.Int:         {unsafe.Sizeof(DInt(0)), fixedSize},
 	types.Float:       {unsafe.Sizeof(DFloat(0.0)), fixedSize},
 	types.Decimal:     {unsafe.Sizeof(DDecimal{}), variableSize},
