@@ -110,20 +110,20 @@ func TestDatumOrdering(t *testing.T) {
 
 		// Intervals
 		{`'1 day':::interval`, noPrev, noNext,
-			`'-768614336404564650y-8mon-9223372036854775808d-2562047h-47m-16s-854ms-775µs-808ns'`,
-			`'768614336404564650y7mon9223372036854775807d2562047h47m16s854ms775µs807ns'`},
+			`'-768614336404564650 years -8 mons -9223372036854775808 days -2562047:47:16.854775'`,
+			`'768614336404564650 years 7 mons 9223372036854775807 days 2562047:47:16.854775'`},
 		// Max interval: we use Postgres syntax, because Go doesn't accept
 		// months/days and ISO8601 doesn't accept nanoseconds.
 		{`'9223372036854775807 months 9223372036854775807 days ` +
-			`2562047 hours 47 minutes 16 seconds 854775807 nanoseconds':::interval`,
+			`2562047 hours 47 minutes 16 seconds 854775 us':::interval`,
 			noPrev, valIsMax,
-			`'-768614336404564650y-8mon-9223372036854775808d-2562047h-47m-16s-854ms-775µs-808ns'`,
-			`'768614336404564650y7mon9223372036854775807d2562047h47m16s854ms775µs807ns'`},
+			`'-768614336404564650 years -8 mons -9223372036854775808 days -2562047:47:16.854775'`,
+			`'768614336404564650 years 7 mons 9223372036854775807 days 2562047:47:16.854775'`},
 		{`'-9223372036854775808 months -9223372036854775808 days ` +
-			`-2562047 h -47 m -16 s -854775808 ns':::interval`,
+			`-2562047 h -47 m -16 s -854775 us':::interval`,
 			valIsMin, noNext,
-			`'-768614336404564650y-8mon-9223372036854775808d-2562047h-47m-16s-854ms-775µs-808ns'`,
-			`'768614336404564650y7mon9223372036854775807d2562047h47m16s854ms775µs807ns'`},
+			`'-768614336404564650 years -8 mons -9223372036854775808 days -2562047:47:16.854775'`,
+			`'768614336404564650 years 7 mons 9223372036854775807 days 2562047:47:16.854775'`},
 
 		// UUIDs
 		{`'ffffffff-ffff-ffff-ffff-ffffffffffff'::uuid`, `'ffffffff-ffff-ffff-ffff-fffffffffffe'`, valIsMax,
@@ -396,12 +396,12 @@ func TestParseDDate(t *testing.T) {
 		{"2017-3-3", "2017-03-03"},
 	}
 	for _, td := range testData {
-		actual, err := tree.ParseDDate(td.str, time.UTC)
+		actual, err := tree.ParseDDate(nil, td.str)
 		if err != nil {
 			t.Errorf("unexpected error while parsing DATE %s: %s", td.str, err)
 			continue
 		}
-		expected, err := tree.ParseDDate(td.expected, time.UTC)
+		expected, err := tree.ParseDDate(nil, td.expected)
 		if err != nil {
 			t.Errorf("unexpected error while parsing expected value DATE %s: %s", td.expected, err)
 			continue
@@ -488,7 +488,7 @@ func TestParseDTime(t *testing.T) {
 		{"4:5:6", timeofday.New(4, 5, 6, 0)},
 	}
 	for _, td := range testData {
-		actual, err := tree.ParseDTime(td.str)
+		actual, err := tree.ParseDTime(nil, td.str)
 		if err != nil {
 			t.Errorf("unexpected error while parsing TIME %s: %s", td.str, err)
 			continue
@@ -504,11 +504,10 @@ func TestParseDTimeError(t *testing.T) {
 		"",
 		"foo",
 		"01",
-		"2001-02-03 04:05:06",
 		"24:00:00",
 	}
 	for _, s := range testData {
-		actual, _ := tree.ParseDTime(s)
+		actual, _ := tree.ParseDTime(nil, s)
 		if actual != nil {
 			t.Errorf("TIME %s: got %s, expected error", s, actual)
 		}
@@ -555,7 +554,7 @@ func TestParseDTimestamp(t *testing.T) {
 			time.FixedZone("", 0))},
 	}
 	for _, td := range testData {
-		actual, err := tree.ParseDTimestamp(td.str, time.Nanosecond)
+		actual, err := tree.ParseDTimestamp(nil, td.str, time.Nanosecond)
 		if err != nil {
 			t.Errorf("unexpected error while parsing TIMESTAMP %s: %s", td.str, err)
 			continue
