@@ -17,9 +17,12 @@ package tree
 import (
 	"bytes"
 	"fmt"
+	"strconv"
 
+	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/util"
 )
@@ -335,6 +338,8 @@ const (
 	NumComparisonOperators
 )
 
+var _ = NumComparisonOperators
+
 var comparisonOpName = [...]string{
 	EQ:                "=",
 	LT:                "<",
@@ -437,6 +442,51 @@ func NewTypedComparisonExprWithSubOp(
 	return node
 }
 
+// NewTypedIndirectionExpr returns a new IndirectionExpr that is verified to be well-typed.
+func NewTypedIndirectionExpr(expr, index TypedExpr) *IndirectionExpr {
+	node := &IndirectionExpr{
+		Expr:        expr,
+		Indirection: ArraySubscripts{&ArraySubscript{Begin: index}},
+	}
+	node.typ = types.UnwrapType(expr.(TypedExpr).ResolvedType()).(types.TArray).Typ
+	return node
+}
+
+// NewTypedCollateExpr returns a new CollateExpr that is verified to be well-typed.
+func NewTypedCollateExpr(expr TypedExpr, locale string) *CollateExpr {
+	node := &CollateExpr{
+		Expr:   expr,
+		Locale: locale,
+	}
+	node.typ = types.TCollatedString{Locale: locale}
+	return node
+}
+
+// NewTypedArrayFlattenExpr returns a new ArrayFlattenExpr that is verified to be well-typed.
+func NewTypedArrayFlattenExpr(input Expr) *ArrayFlatten {
+	inputTyp := input.(TypedExpr).ResolvedType()
+	node := &ArrayFlatten{
+		Subquery: input,
+	}
+	node.typ = types.TArray{Typ: inputTyp}
+	return node
+}
+
+// NewTypedIfErrExpr returns a new IfErrExpr that is verified to be well-typed.
+func NewTypedIfErrExpr(cond, orElse, errCode TypedExpr) *IfErrExpr {
+	node := &IfErrExpr{
+		Cond:    cond,
+		Else:    orElse,
+		ErrCode: errCode,
+	}
+	if orElse == nil {
+		node.typ = types.Bool
+	} else {
+		node.typ = cond.ResolvedType()
+	}
+	return node
+}
+
 func (node *ComparisonExpr) memoizeFn() {
 	fOp, fLeft, fRight, _, _ := foldComparisonExpr(node.Operator, node.Left, node.Right)
 	leftRet, rightRet := fLeft.(TypedExpr).ResolvedType(), fRight.(TypedExpr).ResolvedType()
@@ -545,7 +595,7 @@ func (node *IsOfTypeExpr) Format(ctx *FmtCtx) {
 		if i > 0 {
 			ctx.WriteString(", ")
 		}
-		t.Format(ctx.Buffer, ctx.flags.EncodeFlags())
+		t.Format(&ctx.Buffer, ctx.flags.EncodeFlags())
 	}
 	ctx.WriteByte(')')
 }
@@ -687,32 +737,44 @@ func (node DefaultVal) Format(ctx *FmtCtx) {
 // ResolvedType implements the TypedExpr interface.
 func (DefaultVal) ResolvedType() types.T { return nil }
 
-// MaxVal represents the MAXVALUE expression.
-type MaxVal struct{}
+// PartitionMaxVal represents the MAXVALUE expression.
+type PartitionMaxVal struct{}
 
 // Format implements the NodeFormatter interface.
-func (node MaxVal) Format(ctx *FmtCtx) {
+func (node PartitionMaxVal) Format(ctx *FmtCtx) {
 	ctx.WriteString("MAXVALUE")
 }
 
-// MinVal represents the MINVALUE expression.
-type MinVal struct{}
+// PartitionMinVal represents the MINVALUE expression.
+type PartitionMinVal struct{}
 
 // Format implements the NodeFormatter interface.
-func (node MinVal) Format(ctx *FmtCtx) {
+func (node PartitionMinVal) Format(ctx *FmtCtx) {
 	ctx.WriteString("MINVALUE")
 }
 
 // Placeholder represents a named placeholder.
 type Placeholder struct {
-	Name string
+	Idx types.PlaceholderIdx
 
 	typeAnnotation
 }
 
 // NewPlaceholder allocates a Placeholder.
-func NewPlaceholder(name string) *Placeholder {
-	return &Placeholder{Name: name}
+func NewPlaceholder(name string) (*Placeholder, error) {
+	uval, err := strconv.ParseUint(name, 10, 64)
+	if err != nil {
+		return nil, err
+	}
+	// The string is the number that follows $ which is a 1-based index ($1, $2,
+	// etc), while PlaceholderIdx is 0-based.
+	if uval == 0 || uval > types.MaxPlaceholderIdx+1 {
+		return nil, pgerror.NewErrorf(
+			pgerror.CodeNumericValueOutOfRangeError,
+			"placeholder index must be between 1 and %d", types.MaxPlaceholderIdx+1,
+		)
+	}
+	return &Placeholder{Idx: types.PlaceholderIdx(uval - 1)}, nil
 }
 
 // Format implements the NodeFormatter interface.
@@ -721,14 +783,13 @@ func (node *Placeholder) Format(ctx *FmtCtx) {
 		ctx.placeholderFormat(ctx, node)
 		return
 	}
-	ctx.WriteByte('$')
-	ctx.WriteString(node.Name)
+	ctx.Printf("$%d", node.Idx+1)
 }
 
 // ResolvedType implements the TypedExpr interface.
 func (node *Placeholder) ResolvedType() types.T {
 	if node.typ == nil {
-		node.typ = &types.TPlaceholder{Name: node.Name}
+		node.typ = &types.TPlaceholder{Idx: node.Idx}
 	}
 	return node.typ
 }
@@ -846,7 +907,7 @@ func (node *ArrayFlatten) Format(ctx *FmtCtx) {
 			if len(t.D) == 0 {
 				if colTyp, err := coltypes.DatumTypeToColumnType(node.typ); err == nil {
 					ctx.WriteString(":::")
-					colTyp.Format(ctx.Buffer, ctx.flags.EncodeFlags())
+					colTyp.Format(&ctx.Buffer, ctx.flags.EncodeFlags())
 				}
 			}
 		}
@@ -909,19 +970,18 @@ func (node *Subquery) Format(ctx *FmtCtx) {
 	} else {
 		// Ensure that type printing is disabled during the recursion, as
 		// the type annotations are not available in subqueries.
-		noTypesCtx := *ctx
-		noTypesCtx.flags &= ^FmtShowTypes
-
-		if node.Exists {
-			noTypesCtx.WriteString("EXISTS ")
-		}
-		if node.Select == nil {
-			// If the subquery is generated by the optimizer, we
-			// don't have an actual statement.
-			noTypesCtx.WriteString("<unknown>")
-		} else {
-			noTypesCtx.FormatNode(node.Select)
-		}
+		ctx.WithFlags(ctx.flags & ^FmtShowTypes, func() {
+			if node.Exists {
+				ctx.WriteString("EXISTS ")
+			}
+			if node.Select == nil {
+				// If the subquery is generated by the optimizer, we
+				// don't have an actual statement.
+				ctx.WriteString("<unknown>")
+			} else {
+				ctx.FormatNode(node.Select)
+			}
+		})
 	}
 }
 
@@ -953,6 +1013,8 @@ const (
 	NumBinaryOperators
 )
 
+var _ = NumBinaryOperators
+
 var binaryOpName = [...]string{
 	Bitand:            "&",
 	Bitor:             "|",
@@ -1091,6 +1153,8 @@ const (
 	NumUnaryOperators
 )
 
+var _ = NumUnaryOperators
+
 var unaryOpName = [...]string{
 	UnaryMinus:      "-",
 	UnaryComplement: "~",
@@ -1209,18 +1273,6 @@ func (node *FuncExpr) GetAggregateConstructor() func(*EvalContext, Datums) Aggre
 	}
 }
 
-// GetWindowConstructor returns a window function constructor if the
-// FuncExpr is a built-in window function.
-func (node *FuncExpr) GetWindowConstructor() func(*EvalContext) WindowFunc {
-	if node.fn == nil || node.fn.WindowFunc == nil {
-		return nil
-	}
-	return func(evalCtx *EvalContext) WindowFunc {
-		types := typesOfExprs(node.Exprs)
-		return node.fn.WindowFunc(types, evalCtx)
-	}
-}
-
 func typesOfExprs(exprs Exprs) []types.T {
 	types := make([]types.T, len(exprs))
 	for i, expr := range exprs {
@@ -1274,8 +1326,9 @@ func (node *FuncExpr) Format(ctx *FmtCtx) {
 
 	// We need to remove name anonymization for the function name in
 	// particular. Do this by overriding the flags.
-	subCtx := ctx.CopyWithFlags(ctx.flags & ^FmtAnonymize)
-	subCtx.FormatNode(&node.Func)
+	ctx.WithFlags(ctx.flags&^FmtAnonymize, func() {
+		ctx.FormatNode(&node.Func)
+	})
 
 	ctx.WriteByte('(')
 	ctx.WriteString(typ)
@@ -1380,7 +1433,7 @@ type CastExpr struct {
 
 // Format implements the NodeFormatter interface.
 func (node *CastExpr) Format(ctx *FmtCtx) {
-	buf := ctx.Buffer
+	buf := &ctx.Buffer
 	switch node.SyntaxMode {
 	case CastPrepend:
 		// This is a special case for things like INTERVAL '1s'. These only work
@@ -1410,7 +1463,7 @@ func (node *CastExpr) Format(ctx *FmtCtx) {
 		ctx.WriteByte(')')
 		if isCollatedString {
 			ctx.WriteString(" COLLATE ")
-			lex.EncodeUnrestrictedSQLIdent(ctx.Buffer, t.Locale, lex.EncNoFlags)
+			lex.EncodeUnrestrictedSQLIdent(&ctx.Buffer, t.Locale, lex.EncNoFlags)
 		}
 	}
 }
@@ -1426,33 +1479,42 @@ func (node *CastExpr) castType() types.T {
 	return coltypes.CastTargetToDatumType(node.Type)
 }
 
+type castInfo struct {
+	fromT   types.T
+	counter telemetry.Counter
+}
+
 var (
-	boolCastTypes = []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString}
-	intCastTypes  = []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
-		types.Timestamp, types.TimestampTZ, types.Date, types.Interval, types.Oid}
-	floatCastTypes = []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
-		types.Timestamp, types.TimestampTZ, types.Date, types.Interval}
-	decimalCastTypes = []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
-		types.Timestamp, types.TimestampTZ, types.Date, types.Interval}
-	stringCastTypes = []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
+	bitArrayCastTypes = annotateCast(types.BitArray, []types.T{types.Unknown, types.BitArray, types.Int, types.String, types.FamCollatedString})
+	boolCastTypes     = annotateCast(types.Bool, []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString})
+	intCastTypes      = annotateCast(types.Int, []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
+		types.Timestamp, types.TimestampTZ, types.Date, types.Interval, types.Oid, types.BitArray})
+	floatCastTypes = annotateCast(types.Float, []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
+		types.Timestamp, types.TimestampTZ, types.Date, types.Interval})
+	decimalCastTypes = annotateCast(types.Decimal, []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
+		types.Timestamp, types.TimestampTZ, types.Date, types.Interval})
+	stringCastTypes = annotateCast(types.String, []types.T{types.Unknown, types.Bool, types.Int, types.Float, types.Decimal, types.String, types.FamCollatedString,
+		types.BitArray,
 		types.FamArray, types.FamTuple,
-		types.Bytes, types.Timestamp, types.TimestampTZ, types.Interval, types.UUID, types.Date, types.Time, types.Oid, types.INet, types.JSON}
-	bytesCastTypes = []types.T{types.Unknown, types.String, types.FamCollatedString, types.Bytes, types.UUID}
-	dateCastTypes  = []types.T{types.Unknown, types.String, types.FamCollatedString, types.Date, types.Timestamp, types.TimestampTZ, types.Int}
-	timeCastTypes  = []types.T{types.Unknown, types.String, types.FamCollatedString, types.Time,
-		types.Timestamp, types.TimestampTZ, types.Interval}
-	timestampCastTypes = []types.T{types.Unknown, types.String, types.FamCollatedString, types.Date, types.Timestamp, types.TimestampTZ, types.Int}
-	intervalCastTypes  = []types.T{types.Unknown, types.String, types.FamCollatedString, types.Int, types.Time, types.Interval, types.Float, types.Decimal}
-	oidCastTypes       = []types.T{types.Unknown, types.String, types.FamCollatedString, types.Int, types.Oid}
-	uuidCastTypes      = []types.T{types.Unknown, types.String, types.FamCollatedString, types.Bytes, types.UUID}
-	inetCastTypes      = []types.T{types.Unknown, types.String, types.FamCollatedString, types.INet}
-	arrayCastTypes     = []types.T{types.Unknown, types.String}
-	jsonCastTypes      = []types.T{types.Unknown, types.String, types.JSON}
+		types.Bytes, types.Timestamp, types.TimestampTZ, types.Interval, types.UUID, types.Date, types.Time, types.Oid, types.INet, types.JSON})
+	bytesCastTypes = annotateCast(types.Bytes, []types.T{types.Unknown, types.String, types.FamCollatedString, types.Bytes, types.UUID})
+	dateCastTypes  = annotateCast(types.Date, []types.T{types.Unknown, types.String, types.FamCollatedString, types.Date, types.Timestamp, types.TimestampTZ, types.Int})
+	timeCastTypes  = annotateCast(types.Time, []types.T{types.Unknown, types.String, types.FamCollatedString, types.Time,
+		types.Timestamp, types.TimestampTZ, types.Interval})
+	timestampCastTypes = annotateCast(types.Timestamp, []types.T{types.Unknown, types.String, types.FamCollatedString, types.Date, types.Timestamp, types.TimestampTZ, types.Int})
+	intervalCastTypes  = annotateCast(types.Interval, []types.T{types.Unknown, types.String, types.FamCollatedString, types.Int, types.Time, types.Interval, types.Float, types.Decimal})
+	oidCastTypes       = annotateCast(types.Oid, []types.T{types.Unknown, types.String, types.FamCollatedString, types.Int, types.Oid})
+	uuidCastTypes      = annotateCast(types.UUID, []types.T{types.Unknown, types.String, types.FamCollatedString, types.Bytes, types.UUID})
+	inetCastTypes      = annotateCast(types.INet, []types.T{types.Unknown, types.String, types.FamCollatedString, types.INet})
+	arrayCastTypes     = annotateCast(types.FamArray, []types.T{types.Unknown, types.String})
+	jsonCastTypes      = annotateCast(types.JSON, []types.T{types.Unknown, types.String, types.JSON})
 )
 
 // validCastTypes returns a set of types that can be cast into the provided type.
-func validCastTypes(t types.T) []types.T {
+func validCastTypes(t types.T) []castInfo {
 	switch types.UnwrapType(t) {
+	case types.BitArray:
+		return bitArrayCastTypes
 	case types.Bool:
 		return boolCastTypes
 	case types.Int:
@@ -1487,7 +1549,7 @@ func validCastTypes(t types.T) []types.T {
 		if t.FamilyEqual(types.FamCollatedString) {
 			return stringCastTypes
 		} else if t.FamilyEqual(types.FamArray) {
-			ret := make([]types.T, len(arrayCastTypes))
+			ret := make([]castInfo, len(arrayCastTypes))
 			copy(ret, arrayCastTypes)
 			return ret
 		}
@@ -1537,7 +1599,7 @@ type AnnotateTypeExpr struct {
 
 // Format implements the NodeFormatter interface.
 func (node *AnnotateTypeExpr) Format(ctx *FmtCtx) {
-	buf := ctx.Buffer
+	buf := &ctx.Buffer
 	switch node.SyntaxMode {
 	case AnnotateShort:
 		exprFmtWithParen(ctx, node.Expr)
@@ -1574,7 +1636,7 @@ type CollateExpr struct {
 func (node *CollateExpr) Format(ctx *FmtCtx) {
 	exprFmtWithParen(ctx, node.Expr)
 	ctx.WriteString(" COLLATE ")
-	lex.EncodeUnrestrictedSQLIdent(ctx.Buffer, node.Locale, lex.EncNoFlags)
+	lex.EncodeUnrestrictedSQLIdent(&ctx.Buffer, node.Locale, lex.EncNoFlags)
 }
 
 // TupleStar represents (E).* expressions.
@@ -1637,6 +1699,7 @@ func (node *ColumnAccessExpr) String() string { return AsString(node) }
 func (node *CollateExpr) String() string      { return AsString(node) }
 func (node *ComparisonExpr) String() string   { return AsString(node) }
 func (node *Datums) String() string           { return AsString(node) }
+func (node *DBitArray) String() string        { return AsString(node) }
 func (node *DBool) String() string            { return AsString(node) }
 func (node *DBytes) String() string           { return AsString(node) }
 func (node *DDate) String() string            { return AsString(node) }
@@ -1679,8 +1742,8 @@ func (node *TupleStar) String() string        { return AsString(node) }
 func (node *AnnotateTypeExpr) String() string { return AsString(node) }
 func (node *UnaryExpr) String() string        { return AsString(node) }
 func (node DefaultVal) String() string        { return AsString(node) }
-func (node MaxVal) String() string            { return AsString(node) }
-func (node MinVal) String() string            { return AsString(node) }
+func (node PartitionMaxVal) String() string   { return AsString(node) }
+func (node PartitionMinVal) String() string   { return AsString(node) }
 func (node *Placeholder) String() string      { return AsString(node) }
 func (node dNull) String() string             { return AsString(node) }
 func (list *NameList) String() string         { return AsString(list) }
