@@ -33,32 +33,6 @@ import (
 //   or pattern to something looked up from the database.
 //
 
-// NormalizeTableName transforms an UnresolvedName into a TableName.
-// This does not perform name resolution.
-func NormalizeTableName(n *UnresolvedName) (res TableName, err error) {
-	if n.NumParts < 1 || n.NumParts > 3 || n.Star {
-		// The Star part of the condition is really an assertion. The
-		// parser should not have let this star propagate to a point where
-		// this method is called.
-		return res, newInvTableNameError(n)
-	}
-
-	// Check that all the parts specified are not empty.
-	// It's OK if the catalog name is empty.
-	// We allow this in e.g. `select * from "".crdb_internal.tables`.
-	lastCheck := n.NumParts
-	if lastCheck > 2 {
-		lastCheck = 2
-	}
-	for i := 0; i < lastCheck; i++ {
-		if len(n.Parts[i]) == 0 {
-			return res, newInvTableNameError(n)
-		}
-	}
-
-	return makeTableNameFromUnresolvedName(n), nil
-}
-
 // classifyTablePattern distinguishes between a TableName (last name
 // part is a table name) and an AllTablesSelector.
 // Used e.g. for GRANT.
@@ -119,12 +93,16 @@ func classifyColumnItem(n *UnresolvedName) (VarName, error) {
 	}
 
 	// Construct the result.
-	tn := UnresolvedName{
-		NumParts: n.NumParts - 1,
-		Parts:    NameParts{n.Parts[1], n.Parts[2], n.Parts[3]},
+	var tn *UnresolvedObjectName
+	if n.NumParts > 1 {
+		var err error
+		tn, err = NewUnresolvedObjectName(n.NumParts-1, [3]string{n.Parts[1], n.Parts[2], n.Parts[3]})
+		if err != nil {
+			return nil, err
+		}
 	}
 	if n.Star {
-		return &AllColumnsSelector{tn}, nil
+		return &AllColumnsSelector{TableName: tn}, nil
 	}
 	return &ColumnItem{TableName: tn, ColumnName: Name(n.Parts[0])}, nil
 }
@@ -192,7 +170,7 @@ type ColumnResolutionResult interface {
 func (a *AllColumnsSelector) Resolve(
 	ctx context.Context, r ColumnItemResolver,
 ) (srcName *TableName, srcMeta ColumnSourceMeta, err error) {
-	prefix := makeTableNameFromUnresolvedName(&a.TableName)
+	prefix := a.TableName.ToTableName()
 
 	// Is there a data source with this prefix?
 	var res NumResolutionResults
@@ -223,7 +201,7 @@ func (c *ColumnItem) Resolve(
 	ctx context.Context, r ColumnItemResolver,
 ) (ColumnResolutionResult, error) {
 	colName := c.ColumnName
-	if c.TableName.NumParts == 0 {
+	if c.TableName == nil {
 		// Naked column name: simple case.
 		srcName, srcMeta, cHint, err := r.FindSourceProvidingColumn(ctx, colName)
 		if err != nil {
@@ -233,7 +211,7 @@ func (c *ColumnItem) Resolve(
 	}
 
 	// There is a prefix. We need to search for it.
-	prefix := makeTableNameFromUnresolvedName(&c.TableName)
+	prefix := c.TableName.ToTableName()
 
 	// Is there a data source with this prefix?
 	res, srcName, srcMeta, err := r.FindSourceMatchingName(ctx, prefix)
@@ -253,7 +231,7 @@ func (c *ColumnItem) Resolve(
 		}
 	}
 	if res == NoResults {
-		return nil, newSourceNotFoundError("no data source matches prefix: %s", &c.TableName)
+		return nil, newSourceNotFoundError("no data source matches prefix: %s", c.TableName)
 	}
 	return r.Resolve(ctx, srcName, srcMeta, -1, colName)
 }
@@ -271,9 +249,13 @@ type SchemaMeta interface {
 }
 
 // TableNameExistingResolver is the helper interface to resolve table
-// names when the object is expected to exist already.
+// names when the object is expected to exist already. The boolean passed
+// is used to specify if a MutableTableDescriptor is to be returned in the
+// result.
 type TableNameExistingResolver interface {
-	LookupObject(ctx context.Context, dbName, scName, obName string) (found bool, objMeta NameResolutionResult, err error)
+	LookupObject(ctx context.Context, requireMutation bool, dbName, scName, obName string) (
+		found bool, objMeta NameResolutionResult, err error,
+	)
 }
 
 // NameResolutionResult is an opaque reference returned by LookupObject().
@@ -285,12 +267,16 @@ type NameResolutionResult interface {
 // ResolveExisting performs name resolution for a table name when
 // the target object is expected to exist already.
 func (t *TableName) ResolveExisting(
-	ctx context.Context, r TableNameExistingResolver, curDb string, searchPath sessiondata.SearchPath,
+	ctx context.Context,
+	r TableNameExistingResolver,
+	requireMutable bool,
+	curDb string,
+	searchPath sessiondata.SearchPath,
 ) (bool, NameResolutionResult, error) {
 	if t.ExplicitSchema {
 		if t.ExplicitCatalog {
 			// Already 3 parts: nothing to search. Delegate to the resolver.
-			return r.LookupObject(ctx, t.Catalog(), t.Schema(), t.Table())
+			return r.LookupObject(ctx, requireMutable, t.Catalog(), t.Schema(), t.Table())
 		}
 		// Two parts: D.T.
 		// Try to use the current database, and be satisfied if it's sufficient to find the object.
@@ -300,14 +286,14 @@ func (t *TableName) ResolveExisting(
 		// database is not set. For example, `select * from
 		// pg_catalog.pg_tables` is meant to show all tables across all
 		// databases when there is no current database set.
-		if found, objMeta, err := r.LookupObject(ctx, curDb, t.Schema(), t.Table()); found || err != nil {
+		if found, objMeta, err := r.LookupObject(ctx, requireMutable, curDb, t.Schema(), t.Table()); found || err != nil {
 			if err == nil {
 				t.CatalogName = Name(curDb)
 			}
 			return found, objMeta, err
 		}
 		// No luck so far. Compatibility with CockroachDB v1.1: try D.public.T instead.
-		if found, objMeta, err := r.LookupObject(ctx, t.Schema(), PublicSchema, t.Table()); found || err != nil {
+		if found, objMeta, err := r.LookupObject(ctx, requireMutable, t.Schema(), PublicSchema, t.Table()); found || err != nil {
 			if err == nil {
 				t.CatalogName = t.SchemaName
 				t.SchemaName = PublicSchemaName
@@ -322,7 +308,7 @@ func (t *TableName) ResolveExisting(
 	// This is a naked table name. Use the search path.
 	iter := searchPath.Iter()
 	for next, ok := iter.Next(); ok; next, ok = iter.Next() {
-		if found, objMeta, err := r.LookupObject(ctx, curDb, next, t.Table()); found || err != nil {
+		if found, objMeta, err := r.LookupObject(ctx, requireMutable, curDb, next, t.Table()); found || err != nil {
 			if err == nil {
 				t.CatalogName = Name(curDb)
 				t.SchemaName = Name(next)
@@ -503,7 +489,7 @@ func newInvColRef(fmt string, n *UnresolvedName) error {
 	return pgerror.NewErrorWithDepthf(1, pgerror.CodeInvalidColumnReferenceError, fmt, n)
 }
 
-func newInvTableNameError(n *UnresolvedName) error {
+func newInvTableNameError(n fmt.Stringer) error {
 	return pgerror.NewErrorWithDepthf(1, pgerror.CodeInvalidNameError,
 		"invalid table name: %s", n)
 }
