@@ -27,65 +27,6 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
 )
 
-func TestNormalizeTableName(t *testing.T) {
-	testCases := []struct {
-		in, out  string
-		expanded string
-		err      string
-	}{
-		{`a`, `a`, `""."".a`, ``},
-		{`a.b`, `a.b`, `"".a.b`, ``},
-		{`a.b.c`, `a.b.c`, `a.b.c`, ``},
-		{`a.b.c.d`, ``, ``, `syntax error at or near "\."`},
-		{`a.""`, ``, ``, `invalid table name: a\.""`},
-		{`a.b.""`, ``, ``, `invalid table name: a\.b\.""`},
-		{`a.b.c.""`, ``, ``, `syntax error at or near "\."`},
-		{`a."".c`, ``, ``, `invalid table name: a\.""\.c`},
-
-		// CockroachDB extension: empty catalog name.
-		{`"".b.c`, `"".b.c`, `"".b.c`, ``},
-
-		// Check keywords: disallowed in first position, ok afterwards.
-		{`user.x.y`, ``, ``, `syntax error`},
-		{`"user".x.y`, `"user".x.y`, `"user".x.y`, ``},
-		{`x.user.y`, `x."user".y`, `x."user".y`, ``},
-		{`x.user`, `x."user"`, `"".x."user"`, ``},
-
-		{`foo@bar`, ``, ``, `syntax error at or near "@"`},
-		{`test.*`, ``, ``, `syntax error at or near "\*"`},
-	}
-
-	for _, tc := range testCases {
-		t.Run(tc.in, func(t *testing.T) {
-			tn, err := func() (*tree.TableName, error) {
-				stmt, err := parser.ParseOne(fmt.Sprintf("ALTER TABLE %s RENAME TO x", tc.in))
-				if err != nil {
-					return nil, err
-				}
-				tn, err := stmt.(*tree.RenameTable).Name.Normalize()
-				if err != nil {
-					return nil, err
-				}
-				return tn, nil
-			}()
-			if !testutils.IsError(err, tc.err) {
-				t.Fatalf("%s: expected %s, but found %v", tc.in, tc.err, err)
-			}
-			if tc.err != "" {
-				return
-			}
-			if out := tn.String(); tc.out != out {
-				t.Fatalf("%s: expected %s, but found %s", tc.in, tc.out, out)
-			}
-			tn.ExplicitSchema = true
-			tn.ExplicitCatalog = true
-			if out := tn.String(); tc.expanded != out {
-				t.Fatalf("%s: expected full %s, but found %s", tc.in, tc.expanded, out)
-			}
-		})
-	}
-}
-
 func TestClassifyTablePattern(t *testing.T) {
 	testCases := []struct {
 		in, out  string
@@ -118,7 +59,7 @@ func TestClassifyTablePattern(t *testing.T) {
 		{`*.b`, ``, ``, `syntax error at or near "\."`},
 		{`"".*`, ``, ``, `invalid table name: "".\*`},
 		{`a."".*`, ``, ``, `invalid table name: a\.""\.\*`},
-		{`a.b."".*`, ``, ``, `syntax error at or near "\."`},
+		{`a.b."".*`, ``, ``, `invalid table name: a.b.""`},
 		// CockroachDB extension: empty catalog name.
 		{`"".b.*`, `"".b.*`, `"".b.*`, ``},
 
@@ -137,7 +78,7 @@ func TestClassifyTablePattern(t *testing.T) {
 				if err != nil {
 					return nil, err
 				}
-				tp, err := stmt.(*tree.Grant).Targets.Tables[0].NormalizeTablePattern()
+				tp, err := stmt.AST.(*tree.Grant).Targets.Tables[0].NormalizeTablePattern()
 				if err != nil {
 					return nil, err
 				}
@@ -233,7 +174,7 @@ func TestClassifyColumnName(t *testing.T) {
 				if err != nil {
 					return nil, err
 				}
-				v := stmt.(*tree.Select).Select.(*tree.SelectClause).Exprs[0].Expr.(tree.VarName)
+				v := stmt.AST.(*tree.Select).Select.(*tree.SelectClause).Exprs[0].Expr.(tree.VarName)
 				return v.NormalizeVarName()
 			}()
 			if !testutils.IsError(err, tc.err) {
@@ -507,7 +448,7 @@ func (fakeResResult) NameResolutionResult() {}
 
 // LookupObject implements the TableNameResolver interface.
 func (f *fakeMetadata) LookupObject(
-	_ context.Context, dbName, scName, tbName string,
+	_ context.Context, requireMutable bool, dbName, scName, tbName string,
 ) (found bool, obMeta tree.NameResolutionResult, err error) {
 	defer func() {
 		f.t.Logf("LookupObject(%s, %s, %s) -> found %v meta %v err %v",
@@ -761,7 +702,7 @@ func TestResolveTablePatternOrName(t *testing.T) {
 				if err != nil {
 					return nil, "", err
 				}
-				tp, err := stmt.(*tree.Grant).Targets.Tables[0].NormalizeTablePattern()
+				tp, err := stmt.AST.(*tree.Grant).Targets.Tables[0].NormalizeTablePattern()
 				if err != nil {
 					return nil, "", err
 				}
@@ -778,7 +719,7 @@ func TestResolveTablePatternOrName(t *testing.T) {
 					ctPrefix = tpv.Catalog()
 				case *tree.TableName:
 					if tc.expected {
-						found, obMeta, err = tpv.ResolveExisting(ctx, fakeResolver, tc.curDb, tc.searchPath)
+						found, obMeta, err = tpv.ResolveExisting(ctx, fakeResolver, false /*requireMutable*/, tc.curDb, tc.searchPath)
 					} else {
 						found, scMeta, err = tpv.ResolveTarget(ctx, fakeResolver, tc.curDb, tc.searchPath)
 					}
