@@ -23,9 +23,7 @@
 
 package tree
 
-import (
-	"fmt"
-)
+import "fmt"
 
 // Instructions for creating new types: If a type needs to satisfy an
 // interface, declare that function along with that interface. This
@@ -117,6 +115,26 @@ func CanWriteData(stmt Statement) bool {
 	return false
 }
 
+// IsStmtParallelized determines if a given statement's execution should be
+// parallelized. This means that its results should be mocked out, and that
+// it should be run asynchronously and in parallel with other statements that
+// are independent.
+func IsStmtParallelized(stmt Statement) bool {
+	parallelizedRetClause := func(ret ReturningClause) bool {
+		_, ok := ret.(*ReturningNothing)
+		return ok
+	}
+	switch s := stmt.(type) {
+	case *Delete:
+		return parallelizedRetClause(s.Returning)
+	case *Insert:
+		return parallelizedRetClause(s.Returning)
+	case *Update:
+		return parallelizedRetClause(s.Returning)
+	}
+	return false
+}
+
 // HiddenFromShowQueries is a pseudo-interface to be implemented
 // by statements that should not show up in SHOW QUERIES (and are hence
 // not cancellable using CANCEL QUERIES either). Usually implemented by
@@ -135,19 +153,6 @@ type IndependentFromParallelizedPriors interface {
 	independentFromParallelizedPriors()
 }
 
-// StatementList is a list of statements.
-type StatementList []Statement
-
-// Format implements the NodeFormatter interface.
-func (l *StatementList) Format(ctx *FmtCtx) {
-	for i, s := range *l {
-		if i > 0 {
-			ctx.WriteString("; ")
-		}
-		ctx.FormatNode(s)
-	}
-}
-
 // ObserverStatement is a marker interface for statements which are allowed to
 // run regardless of the current transaction state: statements other than
 // rollback are generally rejected if the session is in a failed transaction
@@ -233,6 +238,30 @@ func (*CancelSessions) StatementTag() string { return "CANCEL SESSIONS" }
 
 func (*CancelSessions) independentFromParallelizedPriors() {}
 
+// StatementType implements the Statement interface.
+func (*CannedOptPlan) StatementType() StatementType { return Rows }
+
+// StatementTag returns a short string identifying the type of statement.
+func (*CannedOptPlan) StatementTag() string { return "PREPARE AS OPT PLAN" }
+
+// StatementType implements the Statement interface.
+func (*CommentOnColumn) StatementType() StatementType { return DDL }
+
+// StatementTag returns a short string identifying the type of statement.
+func (*CommentOnColumn) StatementTag() string { return "COMMENT ON COLUMN" }
+
+// StatementType implements the Statement interface.
+func (*CommentOnDatabase) StatementType() StatementType { return DDL }
+
+// StatementTag returns a short string identifying the type of statement.
+func (*CommentOnDatabase) StatementTag() string { return "COMMENT ON DATABASE" }
+
+// StatementType implements the Statement interface.
+func (*CommentOnTable) StatementType() StatementType { return DDL }
+
+// StatementTag returns a short string identifying the type of statement.
+func (*CommentOnTable) StatementTag() string { return "COMMENT ON TABLE" }
+
 // StatementType implements the Statement interface.
 func (*CommitTransaction) StatementType() StatementType { return Ack }
 
@@ -249,7 +278,12 @@ func (*CopyFrom) StatementTag() string { return "COPY" }
 func (*CreateChangefeed) StatementType() StatementType { return Rows }
 
 // StatementTag returns a short string identifying the type of statement.
-func (*CreateChangefeed) StatementTag() string { return "CREATE CHANGEFEED" }
+func (n *CreateChangefeed) StatementTag() string {
+	if n.SinkURI == nil {
+		return "EXPERIMENTAL CHANGEFEED"
+	}
+	return "CREATE CHANGEFEED"
+}
 
 // StatementType implements the Statement interface.
 func (*CreateDatabase) StatementType() StatementType { return DDL }
@@ -753,6 +787,12 @@ func (*ShowSchemas) StatementType() StatementType { return Rows }
 // StatementTag returns a short string identifying the type of statement.
 func (*ShowSchemas) StatementTag() string { return "SHOW SCHEMAS" }
 
+// StatementType implements the Statement interface.
+func (*ShowSequences) StatementType() StatementType { return Rows }
+
+// StatementTag returns a short string identifying the type of statement.
+func (*ShowSequences) StatementTag() string { return "SHOW SCHEMAS" }
+
 // StatementType implements the Statement interface.
 func (*Split) StatementType() StatementType { return Rows }
 
@@ -804,6 +844,10 @@ func (n *BeginTransaction) String() string          { return AsString(n) }
 func (n *ControlJobs) String() string               { return AsString(n) }
 func (n *CancelQueries) String() string             { return AsString(n) }
 func (n *CancelSessions) String() string            { return AsString(n) }
+func (n *CannedOptPlan) String() string             { return AsString(n) }
+func (n *CommentOnColumn) String() string           { return AsString(n) }
+func (n *CommentOnDatabase) String() string         { return AsString(n) }
+func (n *CommentOnTable) String() string            { return AsString(n) }
 func (n *CommitTransaction) String() string         { return AsString(n) }
 func (n *CopyFrom) String() string                  { return AsString(n) }
 func (n *CreateChangefeed) String() string          { return AsString(n) }
@@ -870,6 +914,7 @@ func (n *ShowRanges) String() string                { return AsString(n) }
 func (n *ShowRoleGrants) String() string            { return AsString(n) }
 func (n *ShowRoles) String() string                 { return AsString(n) }
 func (n *ShowSchemas) String() string               { return AsString(n) }
+func (n *ShowSequences) String() string             { return AsString(n) }
 func (n *ShowSessions) String() string              { return AsString(n) }
 func (n *ShowSyntax) String() string                { return AsString(n) }
 func (n *ShowTableStats) String() string            { return AsString(n) }
@@ -881,7 +926,6 @@ func (n *ShowVar) String() string                   { return AsString(n) }
 func (n *ShowZoneConfig) String() string            { return AsString(n) }
 func (n *ShowFingerprints) String() string          { return AsString(n) }
 func (n *Split) String() string                     { return AsString(n) }
-func (l *StatementList) String() string             { return AsString(l) }
 func (n *Truncate) String() string                  { return AsString(n) }
 func (n *UnionClause) String() string               { return AsString(n) }
 func (n *Update) String() string                    { return AsString(n) }
