@@ -19,15 +19,18 @@ import (
 	"strings"
 	"time"
 
-	"github.com/pkg/errors"
-	"golang.org/x/text/language"
-
+	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/lex"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqltelemetry"
+	"github.com/cockroachdb/cockroach/pkg/util/duration"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
+	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/pkg/errors"
+	"golang.org/x/text/language"
 )
 
 // SemaContext defines the context in which to perform semantic analysis on an
@@ -47,12 +50,6 @@ type SemaContext struct {
 	// already.
 	SearchPath sessiondata.SearchPath
 
-	// privileged, if true, enables "unsafe" builtins, e.g. those
-	// from the crdb_internal namespace. Must be set only for
-	// the root user.
-	// TODO(knz): this attribute can be moved to EvalContext pending #15363.
-	privileged bool
-
 	// AsOfTimestamp denotes the explicit AS OF SYSTEM TIME timestamp for the
 	// query, if any. If the query is not an AS OF SYSTEM TIME query,
 	// AsOfTimestamp is nil.
@@ -186,11 +183,10 @@ func (sp *ScalarProperties) Clear() {
 // MakeSemaContext initializes a simple SemaContext suitable
 // for "lightweight" type checking such as the one performed for default
 // expressions.
-func MakeSemaContext(privileged bool) SemaContext {
-	return SemaContext{
-		Placeholders: MakePlaceholderInfo(),
-		privileged:   privileged,
-	}
+// Note: if queries with placeholders are going to be used,
+// SemaContext.Placeholders.Init must be called separately.
+func MakeSemaContext() SemaContext {
+	return SemaContext{}
 }
 
 // isUnresolvedPlaceholder provides a nil-safe method to determine whether expr is an
@@ -210,12 +206,22 @@ func (sc *SemaContext) GetLocation() *time.Location {
 	return *sc.Location
 }
 
+// GetAdditionMode implements ParseTimeContext.
+func (sc *SemaContext) GetAdditionMode() duration.AdditionMode {
+	return duration.AdditionModeCompatible
+}
+
+// GetRelativeParseTime implements ParseTimeContext.
+func (sc *SemaContext) GetRelativeParseTime() time.Time {
+	return timeutil.Now().In(sc.GetLocation())
+}
+
 type placeholderTypeAmbiguityError struct {
-	v *Placeholder
+	idx types.PlaceholderIdx
 }
 
 func (err placeholderTypeAmbiguityError) Error() string {
-	return fmt.Sprintf("could not determine data type of placeholder %s", err.v)
+	return fmt.Sprintf("could not determine data type of placeholder %s", err.idx)
 }
 
 type unexpectedTypeError struct {
@@ -338,6 +344,12 @@ func (expr *BinaryExpr) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr,
 	}
 
 	binOp := fns[0].(*BinOp)
+
+	// Register operator usage in telemetry.
+	if binOp.counter != nil {
+		telemetry.Inc(binOp.counter)
+	}
+
 	expr.Left, expr.Right = leftTyped, rightTyped
 	expr.fn = binOp
 	expr.typ = binOp.returnType()(typedSubExprs)
@@ -396,18 +408,22 @@ func (expr *CaseExpr) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr, e
 	return expr, nil
 }
 
-func isCastDeepValid(castFrom, castTo types.T) bool {
+func isCastDeepValid(castFrom, castTo types.T) (bool, telemetry.Counter) {
 	castFrom = types.UnwrapType(castFrom)
 	castTo = types.UnwrapType(castTo)
 	if castTo.FamilyEqual(types.FamArray) && castFrom.FamilyEqual(types.FamArray) {
-		return isCastDeepValid(castFrom.(types.TArray).Typ, castTo.(types.TArray).Typ)
+		ok, c := isCastDeepValid(castFrom.(types.TArray).Typ, castTo.(types.TArray).Typ)
+		if ok {
+			telemetry.Inc(sqltelemetry.ArrayCastCounter)
+		}
+		return ok, c
 	}
 	for _, t := range validCastTypes(castTo) {
-		if castFrom.FamilyEqual(t) {
-			return true
+		if castFrom.FamilyEqual(t.fromT) {
+			return true, t.counter
 		}
 	}
-	return false
+	return false, nil
 }
 
 // TypeCheck implements the Expr interface.
@@ -449,7 +465,8 @@ func (expr *CastExpr) TypeCheck(ctx *SemaContext, _ types.T) (TypedExpr, error)
 
 	castFrom := typedSubExpr.ResolvedType()
 
-	if isCastDeepValid(castFrom, returnType) {
+	if ok, c := isCastDeepValid(castFrom, returnType); ok {
+		telemetry.Inc(c)
 		expr.Expr = typedSubExpr
 		expr.typ = returnType
 		return expr, nil
@@ -465,7 +482,7 @@ func (expr *IndirectionExpr) TypeCheck(ctx *SemaContext, desired types.T) (Typed
 			return nil, pgerror.UnimplementedWithIssueErrorf(32551, "ARRAY slicing in %s", expr)
 		}
 		if i > 0 {
-			return nil, pgerror.UnimplementedWithIssueErrorf(32552, "multidimensional ARRAY %s", expr)
+			return nil, pgerror.UnimplementedWithIssueDetailErrorf(32552, "ind", "multidimensional indexing: %s", expr)
 		}
 
 		beginExpr, err := typeCheckAndRequire(ctx, t.Begin, types.Int, "ARRAY subscript")
@@ -486,6 +503,8 @@ func (expr *IndirectionExpr) TypeCheck(ctx *SemaContext, desired types.T) (Typed
 	}
 	expr.Expr = subExpr
 	expr.typ = arrType.Typ
+
+	telemetry.Inc(sqltelemetry.ArraySubscriptCounter)
 	return expr, nil
 }
 
@@ -581,7 +600,7 @@ func (expr *ColumnAccessExpr) TypeCheck(ctx *SemaContext, desired types.T) (Type
 	if expr.ColIndex < 0 {
 		return nil, pgerror.NewErrorf(pgerror.CodeDatatypeMismatchError,
 			"could not identify column %q in %s",
-			ErrNameString(&expr.ColName), resolvedType,
+			ErrNameStringP(&expr.ColName), resolvedType,
 		)
 	}
 
@@ -639,6 +658,11 @@ func (expr *ComparisonExpr) TypeCheck(ctx *SemaContext, desired types.T) (TypedE
 		return DNull, nil
 	}
 
+	// Register operator usage in telemetry.
+	if fn.counter != nil {
+		telemetry.Inc(fn.counter)
+	}
+
 	expr.Left, expr.Right = leftTyped, rightTyped
 	expr.fn = fn
 	expr.typ = types.Bool
@@ -652,7 +676,6 @@ var (
 	errInvalidMaxUsage      = pgerror.NewError(pgerror.CodeSyntaxError, "MAXVALUE can only appear within a range partition expression")
 	errInvalidMinUsage      = pgerror.NewError(pgerror.CodeSyntaxError, "MINVALUE can only appear within a range partition expression")
 	errPrivateFunction      = pgerror.NewError(pgerror.CodeFeatureNotSupportedError, "function reserved for internal use")
-	errInsufficientPriv     = pgerror.NewError(pgerror.CodeInsufficientPrivilegeError, "insufficient privilege")
 )
 
 // NewAggInAggError creates an error for the case when an aggregate function is
@@ -700,12 +723,6 @@ func (sc *SemaContext) checkFunctionUsage(expr *FuncExpr, def *FunctionDefinitio
 	if def.Private {
 		return errors.Wrapf(errPrivateFunction, "%s()", def.Name)
 	}
-	// Check that the built-in is allowed for the current user.
-	// TODO(knz): this check can be moved to evaluation time pending #15363.
-	if def.Privileged && (sc == nil || !sc.privileged) {
-		return errors.Wrapf(errInsufficientPriv, "%s()", def.Name)
-	}
-
 	if sc == nil {
 		// We can't check anything further. Give up.
 		return nil
@@ -893,6 +910,9 @@ func (expr *FuncExpr) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr, e
 			strings.Join(typeNames, ", "),
 		)
 	}
+	if overloadImpl.counter != nil {
+		telemetry.Inc(overloadImpl.counter)
+	}
 	return expr, nil
 }
 
@@ -982,6 +1002,8 @@ func (expr *IfErrExpr) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr,
 	expr.Else = typedElse
 	expr.ErrCode = typedErrCode
 	expr.typ = retType
+
+	telemetry.Inc(sqltelemetry.IfErrCounter)
 	return expr, nil
 }
 
@@ -1158,6 +1180,12 @@ func (expr *UnaryExpr) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr,
 	}
 
 	unaryOp := fns[0].(*UnaryOp)
+
+	// Register operator usage in telemetry.
+	if unaryOp.counter != nil {
+		telemetry.Inc(unaryOp.counter)
+	}
+
 	expr.Expr = exprTyped
 	expr.fn = unaryOp
 	expr.typ = unaryOp.returnType()(typedSubExprs)
@@ -1170,12 +1198,12 @@ func (expr DefaultVal) TypeCheck(_ *SemaContext, desired types.T) (TypedExpr, er
 }
 
 // TypeCheck implements the Expr interface.
-func (expr MinVal) TypeCheck(_ *SemaContext, desired types.T) (TypedExpr, error) {
+func (expr PartitionMinVal) TypeCheck(_ *SemaContext, desired types.T) (TypedExpr, error) {
 	return nil, errInvalidMinUsage
 }
 
 // TypeCheck implements the Expr interface.
-func (expr MaxVal) TypeCheck(_ *SemaContext, desired types.T) (TypedExpr, error) {
+func (expr PartitionMaxVal) TypeCheck(_ *SemaContext, desired types.T) (TypedExpr, error) {
 	return nil, errInvalidMaxUsage
 }
 
@@ -1219,7 +1247,7 @@ func (expr *Tuple) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr, erro
 			for j := 0; j < i; j++ {
 				if expr.Labels[i] == expr.Labels[j] {
 					return nil, pgerror.NewErrorf(pgerror.CodeSyntaxError,
-						"found duplicate tuple label: %q", ErrNameString(&expr.Labels[i]),
+						"found duplicate tuple label: %q", ErrNameStringP(&expr.Labels[i]),
 					)
 				}
 			}
@@ -1260,6 +1288,8 @@ func (expr *Array) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr, erro
 	for i := range typedSubExprs {
 		expr.Exprs[i] = typedSubExprs[i]
 	}
+
+	telemetry.Inc(sqltelemetry.ArrayConstructorCounter)
 	return expr, nil
 }
 
@@ -1276,6 +1306,8 @@ func (expr *ArrayFlatten) TypeCheck(ctx *SemaContext, desired types.T) (TypedExp
 	}
 	expr.Subquery = subqueryTyped
 	expr.typ = types.TArray{Typ: subqueryTyped.ResolvedType()}
+
+	telemetry.Inc(sqltelemetry.ArrayFlattenCounter)
 	return expr, nil
 }
 
@@ -1285,7 +1317,7 @@ func (expr *Placeholder) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr
 	// when there are no available values for the placeholders yet, because
 	// during Execute all placeholders are replaced from the AST before type
 	// checking.
-	if typ, ok := ctx.Placeholders.Type(expr.Name, true); ok {
+	if typ, ok := ctx.Placeholders.Type(expr.Idx); ok {
 		if !desired.Equivalent(typ) {
 			// This indicates there's a conflict between what the type system thinks
 			// the type for this position should be, and the actual type of the
@@ -1301,22 +1333,26 @@ func (expr *Placeholder) TypeCheck(ctx *SemaContext, desired types.T) (TypedExpr
 		// We call SetType regardless of the above condition to inform the
 		// placeholder struct that this placeholder is locked to its type and cannot
 		// be overridden again.
-		if err := ctx.Placeholders.SetType(expr.Name, typ); err != nil {
+		if err := ctx.Placeholders.SetType(expr.Idx, typ); err != nil {
 			return nil, err
 		}
 		expr.typ = typ
 		return expr, nil
 	}
 	if desired.IsAmbiguous() {
-		return nil, placeholderTypeAmbiguityError{expr}
+		return nil, placeholderTypeAmbiguityError{expr.Idx}
 	}
-	if err := ctx.Placeholders.SetType(expr.Name, desired); err != nil {
+	if err := ctx.Placeholders.SetType(expr.Idx, desired); err != nil {
 		return nil, err
 	}
 	expr.typ = desired
 	return expr, nil
 }
 
+// TypeCheck implements the Expr interface. It is implemented as an idempotent
+// identity function for Datum.
+func (d *DBitArray) TypeCheck(_ *SemaContext, _ types.T) (TypedExpr, error) { return d, nil }
+
 // TypeCheck implements the Expr interface. It is implemented as an idempotent
 // identity function for Datum.
 func (d *DBool) TypeCheck(_ *SemaContext, _ types.T) (TypedExpr, error) { return d, nil }
@@ -1770,7 +1806,8 @@ func TypeCheckSameTypedExprs(
 			case len(constIdxs) > 0:
 				return typeCheckConstsAndPlaceholdersWithDesired(s, desired)
 			case len(placeholderIdxs) > 0:
-				return nil, nil, placeholderTypeAmbiguityError{s.exprs[placeholderIdxs[0]].(*Placeholder)}
+				p := s.exprs[placeholderIdxs[0]].(*Placeholder)
+				return nil, nil, placeholderTypeAmbiguityError{p.Idx}
 			default:
 				return typedExprs, types.Unknown, nil
 			}
@@ -2029,64 +2066,129 @@ func checkTupleHasLength(t *Tuple, expectedLen int) error {
 }
 
 type placeholderAnnotationVisitor struct {
-	placeholders map[string]annotationState
+	types PlaceholderTypes
+	state []annotationState
+	err   error
+	// errIdx stores the placeholder to which err applies. Used to select the
+	// error for the smallest index.
+	errIdx types.PlaceholderIdx
 }
 
-// annotationState holds the state of an unresolved type annotation for a given placeholder.
-type annotationState struct {
-	sawAssertion   bool // marks if the placeholder has been subject to at least one type assetion
-	shouldAnnotate bool // marks if the placeholder should be annotated with the type typ
-	typ            types.T
+// annotationState keeps extra information relating to the type of a
+// placeholder.
+type annotationState uint8
+
+const (
+	noType annotationState = iota
+
+	// typeFromHint indicates that the type for the placeholder is set from a
+	// provided type hint (from pgwire, or PREPARE AS arguments).
+	typeFromHint
+
+	// typeFromAnnotation indicates that the type for this placeholder is set from
+	// a type annotation.
+	typeFromAnnotation
+
+	// typeFromCast indicates that the type for this placeholder is set from
+	// a cast. This type can be replaced if an annotation is found, or discarded
+	// if conflicting casts are found.
+	typeFromCast
+
+	// conflictingCasts indicates that we haven't found an annotation for the
+	// placeholder, and we cannot determine the type from casts: either we found
+	// conflicting casts, or we found an appearance of the placeholder without a
+	// cast.
+	conflictingCasts
+)
+
+func (v *placeholderAnnotationVisitor) setErr(idx types.PlaceholderIdx, err error) {
+	if v.err == nil || v.errIdx >= idx {
+		v.err = err
+		v.errIdx = idx
+	}
 }
 
 func (v *placeholderAnnotationVisitor) VisitPre(expr Expr) (recurse bool, newExpr Expr) {
 	switch t := expr.(type) {
 	case *AnnotateTypeExpr:
 		if arg, ok := t.Expr.(*Placeholder); ok {
-			assertType := t.annotationType()
-			if state, ok := v.placeholders[arg.Name]; ok && state.sawAssertion {
-				if state.shouldAnnotate && !assertType.Equivalent(state.typ) {
-					state.shouldAnnotate = false
-					v.placeholders[arg.Name] = state
+			annotationType := t.annotationType()
+			switch v.state[arg.Idx] {
+			case noType, typeFromCast, conflictingCasts:
+				// An annotation overrides casts.
+				v.types[arg.Idx] = annotationType
+				v.state[arg.Idx] = typeFromAnnotation
+
+			case typeFromAnnotation:
+				// Verify that the annotations are consistent.
+				if !annotationType.Equivalent(v.types[arg.Idx]) {
+					v.setErr(arg.Idx, pgerror.NewErrorf(
+						pgerror.CodeDatatypeMismatchError,
+						"multiple conflicting type annotations around %s",
+						arg.Idx,
+					))
 				}
-			} else {
-				// Ignore any previous casts now that we see an annotation.
-				v.placeholders[arg.Name] = annotationState{
-					sawAssertion:   true,
-					shouldAnnotate: true,
-					typ:            assertType,
+
+			case typeFromHint:
+				// Verify that the annotation is consistent with the type hint.
+				if prevType := v.types[arg.Idx]; !annotationType.Equivalent(prevType) {
+					v.setErr(arg.Idx, pgerror.NewErrorf(
+						pgerror.CodeDatatypeMismatchError,
+						"type annotation around %s conflicts with specified type %s",
+						arg.Idx, v.types[arg.Idx],
+					))
 				}
+
+			default:
+				panic("unhandled state")
 			}
 			return false, expr
 		}
+
 	case *CastExpr:
 		if arg, ok := t.Expr.(*Placeholder); ok {
 			castType := t.castType()
-			if state, ok := v.placeholders[arg.Name]; ok {
-				// Ignore casts once an assertion has been seen.
-				if state.sawAssertion {
-					return false, expr
+			switch v.state[arg.Idx] {
+			case noType:
+				v.types[arg.Idx] = castType
+				v.state[arg.Idx] = typeFromCast
+
+			case typeFromCast:
+				// Verify that the casts are consistent.
+				if !castType.Equivalent(v.types[arg.Idx]) {
+					v.state[arg.Idx] = conflictingCasts
+					v.types[arg.Idx] = nil
 				}
 
-				if state.shouldAnnotate && !castType.Equivalent(state.typ) {
-					state.shouldAnnotate = false
-					v.placeholders[arg.Name] = state
-				}
-			} else {
-				v.placeholders[arg.Name] = annotationState{
-					shouldAnnotate: true,
-					typ:            castType,
-				}
+			case typeFromHint, typeFromAnnotation:
+				// A cast never overrides a hinted or annotated type.
+
+			case conflictingCasts:
+				// We already saw inconsistent casts, or a "bare" placeholder; ignore
+				// this cast.
+
+			default:
+				panic("unhandled state")
 			}
 			return false, expr
 		}
+
 	case *Placeholder:
-		if state, ok := v.placeholders[t.Name]; !(ok && state.sawAssertion) {
-			// Ignore non-annotated placeholders once an assertion has been seen.
-			state.shouldAnnotate = false
-			v.placeholders[t.Name] = state
+		switch v.state[t.Idx] {
+		case noType, typeFromCast:
+			// A "bare" placeholder prevents type determination from casts.
+			v.state[t.Idx] = conflictingCasts
+			v.types[t.Idx] = nil
+
+		case typeFromHint, typeFromAnnotation:
+			// We are not relying on casts to determine the type, nothing to do.
+
+		case conflictingCasts:
+			// We already decided not to use casts, nothing to do.
+
+		default:
+			panic("unhandled state")
 		}
-		return false, expr
 	}
 	return true, expr
 }
@@ -2095,48 +2197,42 @@ func (*placeholderAnnotationVisitor) VisitPost(expr Expr) Expr { return expr }
 
 // ProcessPlaceholderAnnotations performs an order-independent global traversal of the
 // provided Statement, annotating all placeholders with a type in either of the following
-// situations.
-// - the placeholder is the subject of an explicit type annotation in at least one
-//   of its occurrences. If it is subject to multiple explicit type annotations where
-//   the types are not all in agreement, or if the placeholder already has an inferred
-//   type in the placeholder map which conflicts with the explicit type annotation
-//   type, an error will be thrown.
-// - the placeholder is the subject to an implicit type annotation, meaning that it
-//   is not subject to an explicit type annotation, and that in all occurrences of the
-//   placeholder, it is subject to a cast to the same type. If it is subject to casts
-//   of multiple types, no error will be thrown, but the placeholder type will not be
-//   inferred. If a type has already been assigned for the placeholder in the placeholder
-//   map, no error will be thrown, and the placeholder will keep it's previously
-//   inferred type.
+// situations:
+//
+//  - the placeholder is the subject of an explicit type annotation in at least one
+//    of its occurrences. If it is subject to multiple explicit type annotations
+//    where the types are not all in agreement, or if the placeholder already has
+//    a type hint in the placeholder map which conflicts with the explicit type
+//    annotation type, an error will be thrown.
+//
+//  - the placeholder is the subject to a cast of the same type in all
+//    occurrences of the placeholder. If the placeholder is subject to casts of
+//    multiple types, or if it has occurrences without a cast, no error will be
+//    thrown but the type will not be inferred. If the placeholder already has a
+//    type hint, that type will be kept regardless of any casts.
+//
+// See docs/RFCS/20160203_typing.md for more details on placeholder typing (in
+// particular section "First pass: placeholder annotations").
+//
+// The typeHints slice contains the client-provided hints and is populated with
+// any newly assigned types. It is assumed to be pre-sized to the number of
+// placeholders in the statement and is populated accordingly.
 //
 // TODO(nvanbenschoten): Can this visitor and map be preallocated (like normalizeVisitor)?
-func (p PlaceholderTypes) ProcessPlaceholderAnnotations(stmt Statement) error {
-	v := placeholderAnnotationVisitor{make(map[string]annotationState)}
+func ProcessPlaceholderAnnotations(stmt Statement, typeHints PlaceholderTypes) error {
+	v := placeholderAnnotationVisitor{
+		types: typeHints,
+		state: make([]annotationState, len(typeHints)),
+	}
 
-	// We treat existing inferred types in the MapPlaceholderTypes as initial assertions.
-	for placeholder, typ := range p {
-		v.placeholders[placeholder] = annotationState{
-			sawAssertion:   true,
-			shouldAnnotate: true,
-			typ:            typ,
+	for placeholder := range typeHints {
+		if typeHints[placeholder] != nil {
+			v.state[placeholder] = typeFromHint
 		}
 	}
 
 	walkStmt(&v, stmt)
-	for placeholder, state := range v.placeholders {
-		if state.shouldAnnotate {
-			p[placeholder] = state.typ
-		} else if state.sawAssertion {
-			// If we should not annotate the type but we did see a type assertion,
-			// there were conflicting type assertions.
-			if prevType, ok := p[placeholder]; ok {
-				return pgerror.NewErrorf(pgerror.CodeDatatypeMismatchError, "found type annotation around %s that conflicts with previously "+
-					"inferred type %s", placeholder, prevType)
-			}
-			return pgerror.NewErrorf(pgerror.CodeDatatypeMismatchError, "found multiple conflicting type annotations around %s", placeholder)
-		}
-	}
-	return nil
+	return v.err
 }
 
 // StripMemoizedFuncs strips memoized function references from expression trees.
