@@ -16,6 +16,7 @@ package tree_test
 
 import (
 	"context"
+	"fmt"
 	"go/constant"
 	"go/token"
 	"reflect"
@@ -36,7 +37,10 @@ func BenchmarkTypeCheck(b *testing.B) {
 	if err != nil {
 		b.Fatalf("%s: %v", expr, err)
 	}
-	ctx := tree.MakeSemaContext(false)
+	ctx := tree.MakeSemaContext()
+	if err := ctx.Placeholders.Init(1 /* numPlaceholders */, nil /* typeHints */); err != nil {
+		b.Fatal(err)
+	}
 	for i := 0; i < b.N; i++ {
 		_, err := tree.TypeCheck(expr, &ctx, types.Int)
 		if err != nil {
@@ -61,7 +65,7 @@ func TestTypeCheckNormalize(t *testing.T) {
 			if err != nil {
 				t.Fatal(err)
 			}
-			ctx := tree.MakeSemaContext(false)
+			ctx := tree.MakeSemaContext()
 			typeChecked, err := tree.TypeCheck(expr, &ctx, types.Any)
 			if err != nil {
 				t.Fatal(err)
@@ -80,11 +84,12 @@ func TestTypeCheckNormalize(t *testing.T) {
 }
 
 var (
-	mapPTypesInt               = tree.PlaceholderTypes{"a": types.Int}
-	mapPTypesDecimal           = tree.PlaceholderTypes{"a": types.Decimal}
-	mapPTypesIntAndInt         = tree.PlaceholderTypes{"a": types.Int, "b": types.Int}
-	mapPTypesIntAndDecimal     = tree.PlaceholderTypes{"a": types.Int, "b": types.Decimal}
-	mapPTypesDecimalAndDecimal = tree.PlaceholderTypes{"a": types.Decimal, "b": types.Decimal}
+	ptypesNone              = tree.PlaceholderTypes{nil, nil}
+	ptypesInt               = tree.PlaceholderTypes{types.Int, nil}
+	ptypesDecimal           = tree.PlaceholderTypes{types.Decimal, nil}
+	ptypesIntAndInt         = tree.PlaceholderTypes{types.Int, types.Int}
+	ptypesIntAndDecimal     = tree.PlaceholderTypes{types.Int, types.Decimal}
+	ptypesDecimalAndDecimal = tree.PlaceholderTypes{types.Decimal, types.Decimal}
 )
 
 // copyableExpr can provide each test permutation with a deep copy of the expression tree.
@@ -129,9 +134,9 @@ func ddecimal(f float64) copyableExpr {
 		return dd
 	}
 }
-func placeholder(name string) copyableExpr {
+func placeholder(id types.PlaceholderIdx) copyableExpr {
 	return func() tree.Expr {
-		return tree.NewPlaceholder(name)
+		return newPlaceholder(id)
 	}
 }
 func tuple(exprs ...copyableExpr) copyableExpr {
@@ -155,7 +160,7 @@ func forEachPerm(exprs []copyableExpr, i int, fn func([]copyableExpr)) {
 }
 
 func clonePlaceholderTypes(args tree.PlaceholderTypes) tree.PlaceholderTypes {
-	clone := make(tree.PlaceholderTypes)
+	clone := make(tree.PlaceholderTypes, len(args))
 	for k, v := range args {
 		clone[k] = v
 	}
@@ -173,11 +178,13 @@ type sameTypedExprsTestCase struct {
 
 func attemptTypeCheckSameTypedExprs(t *testing.T, idx int, test sameTypedExprsTestCase) {
 	if test.expectedPTypes == nil {
-		test.expectedPTypes = make(tree.PlaceholderTypes)
+		test.expectedPTypes = tree.PlaceholderTypes{}
 	}
 	forEachPerm(test.exprs, 0, func(exprs []copyableExpr) {
-		ctx := tree.MakeSemaContext(false)
-		ctx.Placeholders.SetTypeHints(clonePlaceholderTypes(test.ptypes))
+		ctx := tree.MakeSemaContext()
+		if err := ctx.Placeholders.Init(len(test.ptypes), clonePlaceholderTypes(test.ptypes)); err != nil {
+			t.Fatal(err)
+		}
 		desired := types.Any
 		if test.desired != nil {
 			desired = test.desired
@@ -215,16 +222,16 @@ func TestTypeCheckSameTypedExprs(t *testing.T) {
 		{nil, nil, exprs(ddecimal(1), decConst("1.1")), types.Decimal, nil},
 		{nil, nil, exprs(ddecimal(1), ddecimal(1)), types.Decimal, nil},
 		// Mixing resolved placeholders with constants and resolved exprs.
-		{mapPTypesDecimal, nil, exprs(ddecimal(1), placeholder("a")), types.Decimal, mapPTypesDecimal},
-		{mapPTypesDecimal, nil, exprs(intConst("1"), placeholder("a")), types.Decimal, mapPTypesDecimal},
-		{mapPTypesDecimal, nil, exprs(decConst("1.1"), placeholder("a")), types.Decimal, mapPTypesDecimal},
-		{mapPTypesInt, nil, exprs(intConst("1"), placeholder("a")), types.Int, mapPTypesInt},
-		{mapPTypesInt, nil, exprs(decConst("1.0"), placeholder("a")), types.Int, mapPTypesInt},
-		{mapPTypesDecimalAndDecimal, nil, exprs(placeholder("b"), placeholder("a")), types.Decimal, mapPTypesDecimalAndDecimal},
+		{ptypesDecimal, nil, exprs(ddecimal(1), placeholder(0)), types.Decimal, ptypesDecimal},
+		{ptypesDecimal, nil, exprs(intConst("1"), placeholder(0)), types.Decimal, ptypesDecimal},
+		{ptypesDecimal, nil, exprs(decConst("1.1"), placeholder(0)), types.Decimal, ptypesDecimal},
+		{ptypesInt, nil, exprs(intConst("1"), placeholder(0)), types.Int, ptypesInt},
+		{ptypesInt, nil, exprs(decConst("1.0"), placeholder(0)), types.Int, ptypesInt},
+		{ptypesDecimalAndDecimal, nil, exprs(placeholder(1), placeholder(0)), types.Decimal, ptypesDecimalAndDecimal},
 		// Mixing unresolved placeholders with constants and resolved exprs.
-		{nil, nil, exprs(ddecimal(1), placeholder("a")), types.Decimal, mapPTypesDecimal},
-		{nil, nil, exprs(intConst("1"), placeholder("a")), types.Int, mapPTypesInt},
-		{nil, nil, exprs(decConst("1.1"), placeholder("a")), types.Decimal, mapPTypesDecimal},
+		{ptypesNone, nil, exprs(ddecimal(1), placeholder(0)), types.Decimal, ptypesDecimal},
+		{ptypesNone, nil, exprs(intConst("1"), placeholder(0)), types.Int, ptypesInt},
+		{ptypesNone, nil, exprs(decConst("1.1"), placeholder(0)), types.Decimal, ptypesDecimal},
 		// Verify dealing with Null.
 		{nil, nil, exprs(dnull), types.Unknown, nil},
 		{nil, nil, exprs(dnull, dnull), types.Unknown, nil},
@@ -248,11 +255,13 @@ func TestTypeCheckSameTypedExprs(t *testing.T) {
 		{nil, types.Int, exprs(intConst("1"), decConst("1.1")), types.Decimal, nil},
 		{nil, types.Decimal, exprs(intConst("1"), decConst("1.1")), types.Decimal, nil},
 		// Verify desired type when possible with unresolved placeholders.
-		{nil, types.Decimal, exprs(placeholder("a")), types.Decimal, mapPTypesDecimal},
-		{nil, types.Decimal, exprs(intConst("1"), placeholder("a")), types.Decimal, mapPTypesDecimal},
-		{nil, types.Decimal, exprs(decConst("1.1"), placeholder("a")), types.Decimal, mapPTypesDecimal},
+		{ptypesNone, types.Decimal, exprs(placeholder(0)), types.Decimal, ptypesDecimal},
+		{ptypesNone, types.Decimal, exprs(intConst("1"), placeholder(0)), types.Decimal, ptypesDecimal},
+		{ptypesNone, types.Decimal, exprs(decConst("1.1"), placeholder(0)), types.Decimal, ptypesDecimal},
 	} {
-		attemptTypeCheckSameTypedExprs(t, i, d)
+		t.Run(fmt.Sprintf("%d", i), func(t *testing.T) {
+			attemptTypeCheckSameTypedExprs(t, i, d)
+		})
 	}
 }
 
@@ -271,19 +280,19 @@ func TestTypeCheckSameTypedTupleExprs(t *testing.T) {
 		{nil, nil, exprs(tuple(dint(1), decConst("1.1")), tuple(intConst("1"), ddecimal(1))), ttuple(types.Int, types.Decimal), nil},
 		{nil, nil, exprs(tuple(dint(1), decConst("1.0")), tuple(intConst("1"), dint(1))), ttuple(types.Int, types.Int), nil},
 		// Mixing resolved placeholders with constants and resolved exprs.
-		{mapPTypesDecimal, nil, exprs(tuple(ddecimal(1), intConst("1")), tuple(placeholder("a"), placeholder("a"))), ttuple(types.Decimal, types.Decimal), mapPTypesDecimal},
-		{mapPTypesDecimalAndDecimal, nil, exprs(tuple(placeholder("b"), intConst("1")), tuple(placeholder("a"), placeholder("a"))), ttuple(types.Decimal, types.Decimal), mapPTypesDecimalAndDecimal},
-		{mapPTypesIntAndDecimal, nil, exprs(tuple(intConst("1"), intConst("1")), tuple(placeholder("a"), placeholder("b"))), ttuple(types.Int, types.Decimal), mapPTypesIntAndDecimal},
+		{ptypesDecimal, nil, exprs(tuple(ddecimal(1), intConst("1")), tuple(placeholder(0), placeholder(0))), ttuple(types.Decimal, types.Decimal), ptypesDecimal},
+		{ptypesDecimalAndDecimal, nil, exprs(tuple(placeholder(1), intConst("1")), tuple(placeholder(0), placeholder(0))), ttuple(types.Decimal, types.Decimal), ptypesDecimalAndDecimal},
+		{ptypesIntAndDecimal, nil, exprs(tuple(intConst("1"), intConst("1")), tuple(placeholder(0), placeholder(1))), ttuple(types.Int, types.Decimal), ptypesIntAndDecimal},
 		// Mixing unresolved placeholders with constants and resolved exprs.
-		{nil, nil, exprs(tuple(ddecimal(1), intConst("1")), tuple(placeholder("a"), placeholder("a"))), ttuple(types.Decimal, types.Decimal), mapPTypesDecimal},
-		{nil, nil, exprs(tuple(intConst("1"), intConst("1")), tuple(placeholder("a"), placeholder("b"))), ttuple(types.Int, types.Int), mapPTypesIntAndInt},
+		{ptypesNone, nil, exprs(tuple(ddecimal(1), intConst("1")), tuple(placeholder(0), placeholder(0))), ttuple(types.Decimal, types.Decimal), ptypesDecimal},
+		{ptypesNone, nil, exprs(tuple(intConst("1"), intConst("1")), tuple(placeholder(0), placeholder(1))), ttuple(types.Int, types.Int), ptypesIntAndInt},
 		// Verify dealing with Null.
 		{nil, nil, exprs(tuple(intConst("1"), dnull), tuple(dnull, decConst("1"))), ttuple(types.Int, types.Decimal), nil},
 		{nil, nil, exprs(tuple(dint(1), dnull), tuple(dnull, ddecimal(1))), ttuple(types.Int, types.Decimal), nil},
 		// Verify desired type when possible.
 		{nil, ttuple(types.Int, types.Decimal), exprs(tuple(intConst("1"), intConst("1")), tuple(intConst("1"), intConst("1"))), ttuple(types.Int, types.Decimal), nil},
 		// Verify desired type when possible with unresolved constants.
-		{nil, ttuple(types.Int, types.Decimal), exprs(tuple(placeholder("a"), intConst("1")), tuple(intConst("1"), placeholder("b"))), ttuple(types.Int, types.Decimal), mapPTypesIntAndDecimal},
+		{ptypesNone, ttuple(types.Int, types.Decimal), exprs(tuple(placeholder(0), intConst("1")), tuple(intConst("1"), placeholder(1))), ttuple(types.Int, types.Decimal), ptypesIntAndDecimal},
 	} {
 		attemptTypeCheckSameTypedExprs(t, i, d)
 	}
@@ -306,17 +315,20 @@ func TestTypeCheckSameTypedExprsError(t *testing.T) {
 		// Single type mismatches.
 		{nil, nil, exprs(dint(1), decConst("1.1")), decimalIntMismatchErr},
 		{nil, nil, exprs(dint(1), ddecimal(1)), decimalIntMismatchErr},
-		{mapPTypesInt, nil, exprs(decConst("1.1"), placeholder("a")), decimalIntMismatchErr},
+		{ptypesInt, nil, exprs(decConst("1.1"), placeholder(0)), decimalIntMismatchErr},
 		// Tuple type mismatches.
 		{nil, nil, exprs(tuple(dint(1)), tuple(ddecimal(1))), tupleFloatIntMismatchErr},
 		{nil, nil, exprs(tuple(dint(1)), dint(1), dint(1)), tupleIntMismatchErr},
 		{nil, nil, exprs(tuple(dint(1)), tuple(dint(1), dint(1))), tupleLenErr},
 		// Placeholder ambiguity.
-		{nil, nil, exprs(placeholder("b"), placeholder("a")), placeholderErr},
+		{ptypesNone, nil, exprs(placeholder(1), placeholder(0)), placeholderErr},
 	}
 	for i, d := range testData {
-		ctx := tree.MakeSemaContext(false)
-		ctx.Placeholders.SetTypeHints(d.ptypes)
+		ctx := tree.MakeSemaContext()
+		if err := ctx.Placeholders.Init(len(d.ptypes), d.ptypes); err != nil {
+			t.Error(err)
+			continue
+		}
 		desired := types.Any
 		if d.desired != nil {
 			desired = d.desired
@@ -337,7 +349,7 @@ func annot(p *tree.Placeholder, typ coltypes.T) tree.Expr {
 }
 
 func TestProcessPlaceholderAnnotations(t *testing.T) {
-	intType := coltypes.Int
+	intType := coltypes.Int8
 	boolType := coltypes.Bool
 
 	testData := []struct {
@@ -345,171 +357,176 @@ func TestProcessPlaceholderAnnotations(t *testing.T) {
 		stmtExprs []tree.Expr
 		desired   tree.PlaceholderTypes
 	}{
-		{
-			tree.PlaceholderTypes{},
-			[]tree.Expr{tree.NewPlaceholder("a")},
-			tree.PlaceholderTypes{},
+		{ // 0
+			tree.PlaceholderTypes{nil},
+			[]tree.Expr{newPlaceholder(0)},
+			tree.PlaceholderTypes{nil},
 		},
-		{
-			tree.PlaceholderTypes{},
-			[]tree.Expr{tree.NewPlaceholder("a"), tree.NewPlaceholder("b")},
-			tree.PlaceholderTypes{},
+		{ // 1
+			tree.PlaceholderTypes{nil, nil},
+			[]tree.Expr{newPlaceholder(0), newPlaceholder(1)},
+			tree.PlaceholderTypes{nil, nil},
 		},
-		{
-			tree.PlaceholderTypes{"b": types.Bool},
-			[]tree.Expr{tree.NewPlaceholder("a"), tree.NewPlaceholder("b")},
-			tree.PlaceholderTypes{"b": types.Bool},
+		{ // 2
+			tree.PlaceholderTypes{nil, nil, types.Bool},
+			[]tree.Expr{newPlaceholder(0), newPlaceholder(1)},
+			tree.PlaceholderTypes{nil, nil, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{"c": types.Float},
-			[]tree.Expr{tree.NewPlaceholder("a"), tree.NewPlaceholder("b")},
-			tree.PlaceholderTypes{"c": types.Float},
+		{ // 3
+			tree.PlaceholderTypes{nil, nil, types.Float},
+			[]tree.Expr{newPlaceholder(0), newPlaceholder(1)},
+			tree.PlaceholderTypes{nil, nil, types.Float},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 4
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("a"), boolType),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(0), boolType),
 			},
-			tree.PlaceholderTypes{},
+			tree.PlaceholderTypes{nil, nil},
 		},
-		{
-			tree.PlaceholderTypes{"a": types.Float},
+		{ // 5
+			tree.PlaceholderTypes{types.Float},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("a"), boolType),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(0), boolType),
 			},
-			tree.PlaceholderTypes{"a": types.Float},
+			tree.PlaceholderTypes{types.Float},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 6
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), boolType),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), boolType),
 			},
-			tree.PlaceholderTypes{"a": types.Int, "b": types.Bool},
+			tree.PlaceholderTypes{types.Int, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 7
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("b"), boolType),
+				annot(newPlaceholder(0), intType),
+				annot(newPlaceholder(1), boolType),
 			},
-			tree.PlaceholderTypes{"a": types.Int, "b": types.Bool},
+			tree.PlaceholderTypes{types.Int, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{"b": types.Bool},
+		{ // 8
+			tree.PlaceholderTypes{nil, types.Bool},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), intType),
+				annot(newPlaceholder(0), intType),
 			},
-			tree.PlaceholderTypes{"a": types.Int, "b": types.Bool},
+			tree.PlaceholderTypes{types.Int, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 9
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), boolType),
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), intType),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), boolType),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), intType),
 			},
-			tree.PlaceholderTypes{"a": types.Int},
+			tree.PlaceholderTypes{types.Int, nil},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 10
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("b"), boolType),
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), intType),
+				cast(newPlaceholder(0), intType),
+				annot(newPlaceholder(1), boolType),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), intType),
 			},
-			tree.PlaceholderTypes{"a": types.Int, "b": types.Bool},
+			tree.PlaceholderTypes{types.Int, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 11
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), boolType),
-				tree.NewPlaceholder("a"),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), boolType),
+				newPlaceholder(0),
 			},
-			tree.PlaceholderTypes{"b": types.Bool},
+			tree.PlaceholderTypes{nil, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 12
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				tree.NewPlaceholder("a"),
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), boolType),
+				newPlaceholder(0),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), boolType),
 			},
-			tree.PlaceholderTypes{"b": types.Bool},
+			tree.PlaceholderTypes{nil, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 13
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("b"), boolType),
-				tree.NewPlaceholder("a"),
+				annot(newPlaceholder(0), intType),
+				annot(newPlaceholder(1), boolType),
+				newPlaceholder(0),
 			},
-			tree.PlaceholderTypes{"a": types.Int, "b": types.Bool},
+			tree.PlaceholderTypes{types.Int, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 14
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				tree.NewPlaceholder("a"),
-				annot(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("b"), boolType),
+				newPlaceholder(0),
+				annot(newPlaceholder(0), intType),
+				annot(newPlaceholder(1), boolType),
 			},
-			tree.PlaceholderTypes{"a": types.Int, "b": types.Bool},
+			tree.PlaceholderTypes{types.Int, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{"a": types.Float, "b": types.Bool},
+		{ // 15
+			tree.PlaceholderTypes{types.Float, types.Bool},
 			[]tree.Expr{
-				tree.NewPlaceholder("a"),
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), boolType),
+				newPlaceholder(0),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), boolType),
 			},
-			tree.PlaceholderTypes{"a": types.Float, "b": types.Bool},
+			tree.PlaceholderTypes{types.Float, types.Bool},
 		},
-		{
-			tree.PlaceholderTypes{"a": types.Float, "b": types.Float},
+		{ // 16
+			tree.PlaceholderTypes{types.Float, types.Float},
 			[]tree.Expr{
-				tree.NewPlaceholder("a"),
-				cast(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("b"), boolType),
+				newPlaceholder(0),
+				cast(newPlaceholder(0), intType),
+				cast(newPlaceholder(1), boolType),
 			},
-			tree.PlaceholderTypes{"a": types.Float, "b": types.Float},
+			tree.PlaceholderTypes{types.Float, types.Float},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 17
+			tree.PlaceholderTypes{nil},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("a"), intType),
-				cast(tree.NewPlaceholder("a"), intType),
+				cast(newPlaceholder(0), intType),
+				annot(newPlaceholder(0), intType),
+				cast(newPlaceholder(0), intType),
 			},
-			tree.PlaceholderTypes{"a": types.Int},
+			tree.PlaceholderTypes{types.Int},
 		},
-		{
-			tree.PlaceholderTypes{},
+		{ // 18
+			tree.PlaceholderTypes{nil},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("a"), boolType),
-				cast(tree.NewPlaceholder("a"), intType),
+				cast(newPlaceholder(0), intType),
+				annot(newPlaceholder(0), boolType),
+				cast(newPlaceholder(0), intType),
 			},
-			tree.PlaceholderTypes{"a": types.Bool},
+			tree.PlaceholderTypes{types.Bool},
 		},
 	}
 	for i, d := range testData {
-		args := d.initArgs
-		stmt := &tree.ValuesClause{Rows: []tree.Exprs{d.stmtExprs}}
-		if err := args.ProcessPlaceholderAnnotations(stmt); err != nil {
-			t.Errorf("%d: unexpected error returned from ProcessPlaceholderAnnotations: %v", i, err)
-		} else if !reflect.DeepEqual(args, d.desired) {
-			t.Errorf("%d: expected args %v after processing placeholder annotations for %v, found %v", i, d.desired, stmt, args)
-		}
+		t.Run(fmt.Sprintf("%d", i), func(t *testing.T) {
+			args := d.initArgs
+			stmt := &tree.ValuesClause{Rows: []tree.Exprs{d.stmtExprs}}
+			if err := tree.ProcessPlaceholderAnnotations(stmt, args); err != nil {
+				t.Errorf("%d: unexpected error returned from ProcessPlaceholderAnnotations: %v", i, err)
+			} else if !reflect.DeepEqual(args, d.desired) {
+				t.Errorf(
+					"%d: expected args %v after processing placeholder annotations for %v, found %v",
+					i, d.desired, stmt, args,
+				)
+			}
+		})
 	}
 }
 
 func TestProcessPlaceholderAnnotationsError(t *testing.T) {
-	intType := coltypes.Int
+	intType := coltypes.Int8
 	floatType := coltypes.Float8
 
 	testData := []struct {
@@ -518,60 +535,74 @@ func TestProcessPlaceholderAnnotationsError(t *testing.T) {
 		expected  string
 	}{
 		{
-			tree.PlaceholderTypes{},
+			tree.PlaceholderTypes{nil},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), floatType),
-				annot(tree.NewPlaceholder("a"), intType),
+				annot(newPlaceholder(0), floatType),
+				annot(newPlaceholder(0), intType),
 			},
-			"multiple conflicting type annotations around a",
+			"multiple conflicting type annotations around \\$1",
 		},
 		{
-			tree.PlaceholderTypes{},
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), floatType),
-				cast(tree.NewPlaceholder("a"), floatType),
-				cast(tree.NewPlaceholder("b"), floatType),
-				annot(tree.NewPlaceholder("a"), intType),
+				annot(newPlaceholder(0), floatType),
+				cast(newPlaceholder(0), floatType),
+				cast(newPlaceholder(1), floatType),
+				annot(newPlaceholder(0), intType),
 			},
-			"multiple conflicting type annotations around a",
+			"multiple conflicting type annotations around \\$1",
 		},
 		{
-			tree.PlaceholderTypes{"a": types.Float},
+			tree.PlaceholderTypes{nil, nil},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), intType),
+				annot(newPlaceholder(0), floatType),
+				annot(newPlaceholder(1), floatType),
+				annot(newPlaceholder(1), intType),
+				annot(newPlaceholder(0), intType),
 			},
-			"type annotation around a that conflicts with previously inferred type float",
+			"multiple conflicting type annotations around \\$1",
 		},
 		{
-			tree.PlaceholderTypes{"a": types.Float},
+			tree.PlaceholderTypes{types.Float},
 			[]tree.Expr{
-				cast(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("a"), intType),
+				annot(newPlaceholder(0), intType),
 			},
-			"type annotation around a that conflicts with previously inferred type float",
+			"type annotation around \\$1 conflicts with specified type float",
 		},
 		{
-			tree.PlaceholderTypes{"a": types.Float},
+			tree.PlaceholderTypes{types.Float},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), floatType),
-				annot(tree.NewPlaceholder("a"), intType),
+				cast(newPlaceholder(0), intType),
+				annot(newPlaceholder(0), intType),
 			},
-			"type annotation around a that conflicts with previously inferred type float",
+			"type annotation around \\$1 conflicts with specified type float",
 		},
 		{
-			tree.PlaceholderTypes{"a": types.Float},
+			tree.PlaceholderTypes{types.Float},
 			[]tree.Expr{
-				annot(tree.NewPlaceholder("a"), intType),
-				annot(tree.NewPlaceholder("a"), floatType),
+				annot(newPlaceholder(0), floatType),
+				annot(newPlaceholder(0), intType),
 			},
-			"type annotation around a that conflicts with previously inferred type float",
+			"type annotation around \\$1 conflicts with specified type float",
+		},
+		{
+			tree.PlaceholderTypes{types.Float},
+			[]tree.Expr{
+				annot(newPlaceholder(0), intType),
+				annot(newPlaceholder(0), floatType),
+			},
+			"type annotation around \\$1 conflicts with specified type float",
 		},
 	}
 	for i, d := range testData {
 		args := d.initArgs
 		stmt := &tree.ValuesClause{Rows: []tree.Exprs{d.stmtExprs}}
-		if err := args.ProcessPlaceholderAnnotations(stmt); !testutils.IsError(err, d.expected) {
-			t.Errorf("%d: expected %s, but found %v", i, d.expected, err)
+		if err := tree.ProcessPlaceholderAnnotations(stmt, args); !testutils.IsError(err, d.expected) {
+			t.Errorf("%d: expected '%s', got '%v'", i, d.expected, err)
 		}
 	}
 }
+
+func newPlaceholder(id types.PlaceholderIdx) *tree.Placeholder {
+	return &tree.Placeholder{Idx: id}
+}
