@@ -17,6 +17,7 @@ package types
 import (
 	"bytes"
 	"fmt"
+	"math"
 
 	"github.com/lib/pq/oid"
 )
@@ -51,6 +52,8 @@ var (
 	Unknown T = tUnknown{}
 	// Bool is the type of a DBool. Can be compared with ==.
 	Bool T = tBool{}
+	// BitArray is the type of a DBitArray. Can be compared with ==.
+	BitArray T = tBitArray{}
 	// Int is the type of a DInt. Can be compared with ==.
 	Int T = tInt{}
 	// Float is the type of a DFloat. Can be compared with ==.
@@ -86,6 +89,7 @@ var (
 	// AnyNonArray contains all non-array types.
 	AnyNonArray = []T{
 		Bool,
+		BitArray,
 		Int,
 		Float,
 		Decimal,
@@ -143,6 +147,15 @@ func (tInt) Oid() oid.Oid             { return oid.T_int8 }
 func (tInt) SQLName() string          { return "bigint" }
 func (tInt) IsAmbiguous() bool        { return false }
 
+type tBitArray struct{}
+
+func (tBitArray) String() string           { return "varbit" }
+func (tBitArray) Equivalent(other T) bool  { return UnwrapType(other) == BitArray || other == Any }
+func (tBitArray) FamilyEqual(other T) bool { return UnwrapType(other) == BitArray }
+func (tBitArray) Oid() oid.Oid             { return oid.T_varbit }
+func (tBitArray) SQLName() string          { return "bit varying" }
+func (tBitArray) IsAmbiguous() bool        { return false }
+
 type tFloat struct{}
 
 func (tFloat) String() string           { return "float" }
@@ -180,6 +193,10 @@ type TCollatedString struct {
 
 // String implements the fmt.Stringer interface.
 func (t TCollatedString) String() string {
+	if t.Locale == "" {
+		// Used in telemetry.
+		return "collatedstring{*}"
+	}
 	return fmt.Sprintf("collatedstring{%s}", t.Locale)
 }
 
@@ -380,13 +397,27 @@ func (t TTuple) IsAmbiguous() bool {
 	return len(t.Types) == 0
 }
 
+// PlaceholderIdx is the 0-based index of a placeholder. Placeholder "$1"
+// has PlaceholderIdx=0.
+type PlaceholderIdx uint16
+
+// MaxPlaceholderIdx is the maximum allowed value of a PlaceholderIdx.
+// The pgwire protocol is limited to 2^16 placeholders, so we limit the IDs to
+// this range as well.
+const MaxPlaceholderIdx = math.MaxUint16
+
+// String returns the index as a placeholder string representation ($1, $2 etc).
+func (idx PlaceholderIdx) String() string {
+	return fmt.Sprintf("$%d", idx+1)
+}
+
 // TPlaceholder is the type of a placeholder.
 type TPlaceholder struct {
-	Name string
+	Idx PlaceholderIdx
 }
 
 // String implements the fmt.Stringer interface.
-func (t TPlaceholder) String() string { return fmt.Sprintf("placeholder{%s}", t.Name) }
+func (t TPlaceholder) String() string { return fmt.Sprintf("placeholder{%d}", t.Idx+1) }
 
 // Equivalent implements the T interface.
 func (t TPlaceholder) Equivalent(other T) bool {
@@ -397,7 +428,7 @@ func (t TPlaceholder) Equivalent(other T) bool {
 		return true
 	}
 	u, ok := UnwrapType(other).(TPlaceholder)
-	return ok && t.Name == u.Name
+	return ok && t.Idx == u.Idx
 }
 
 // FamilyEqual implements the T interface.
@@ -418,7 +449,13 @@ func (TPlaceholder) IsAmbiguous() bool { panic("TPlaceholder.IsAmbiguous() is un
 // TArray is the type of a DArray.
 type TArray struct{ Typ T }
 
-func (a TArray) String() string { return a.Typ.String() + "[]" }
+func (a TArray) String() string {
+	if a.Typ == nil {
+		// Used in telemetry.
+		return "*[]"
+	}
+	return a.Typ.String() + "[]"
+}
 
 // Equivalent implements the T interface.
 func (a TArray) Equivalent(other T) bool {
@@ -488,12 +525,14 @@ func IsStringType(t T) bool {
 
 // IsValidArrayElementType returns true if the T
 // can be used in TArray.
-func IsValidArrayElementType(t T) bool {
+// If the valid return is false, the issue number should
+// be included in the error report to inform the user.
+func IsValidArrayElementType(t T) (valid bool, issueNum int) {
 	switch t {
 	case JSON:
-		return false
+		return false, 23468
 	default:
-		return true
+		return true, 0
 	}
 }
 
