@@ -20,9 +20,6 @@ import (
 	"sort"
 	"strings"
 
-	"github.com/pkg/errors"
-	yaml "gopkg.in/yaml.v2"
-
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/keys"
@@ -33,12 +30,19 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
+	"github.com/gogo/protobuf/proto"
+	yaml "gopkg.in/yaml.v2"
 )
 
+type optionValue struct {
+	inheritValue  bool
+	explicitValue tree.TypedExpr
+}
+
 type setZoneConfigNode struct {
 	zoneSpecifier tree.ZoneSpecifier
 	yamlConfig    tree.TypedExpr
-	options       map[tree.Name]tree.TypedExpr
+	options       map[tree.Name]optionValue
 	setDefault    bool
 
 	run setZoneConfigRun
@@ -51,14 +55,25 @@ var supportedZoneConfigOptions = map[tree.Name]struct {
 	requiredType types.T
 	setter       func(*config.ZoneConfig, tree.Datum)
 }{
-	"range_min_bytes": {types.Int, func(c *config.ZoneConfig, d tree.Datum) { c.RangeMinBytes = int64(tree.MustBeDInt(d)) }},
-	"range_max_bytes": {types.Int, func(c *config.ZoneConfig, d tree.Datum) { c.RangeMaxBytes = int64(tree.MustBeDInt(d)) }},
-	"num_replicas":    {types.Int, func(c *config.ZoneConfig, d tree.Datum) { c.NumReplicas = int32(tree.MustBeDInt(d)) }},
-	"gc.ttlseconds":   {types.Int, func(c *config.ZoneConfig, d tree.Datum) { c.GC.TTLSeconds = int32(tree.MustBeDInt(d)) }},
+	"range_min_bytes": {types.Int, func(c *config.ZoneConfig, d tree.Datum) { c.RangeMinBytes = proto.Int64(int64(tree.MustBeDInt(d))) }},
+	"range_max_bytes": {types.Int, func(c *config.ZoneConfig, d tree.Datum) { c.RangeMaxBytes = proto.Int64(int64(tree.MustBeDInt(d))) }},
+	"num_replicas":    {types.Int, func(c *config.ZoneConfig, d tree.Datum) { c.NumReplicas = proto.Int32(int32(tree.MustBeDInt(d))) }},
+	"gc.ttlseconds": {types.Int, func(c *config.ZoneConfig, d tree.Datum) {
+		c.GC = &config.GCPolicy{TTLSeconds: int32(tree.MustBeDInt(d))}
+	}},
 	"constraints": {types.String, func(c *config.ZoneConfig, d tree.Datum) {
-		loadYAML((*config.ConstraintsList)(&c.Constraints), string(tree.MustBeDString(d)))
+		constraintsList := config.ConstraintsList{
+			Constraints: c.Constraints,
+			Inherited:   c.InheritedConstraints,
+		}
+		loadYAML(&constraintsList, string(tree.MustBeDString(d)))
+		c.Constraints = constraintsList.Constraints
+		c.InheritedConstraints = false
+	}},
+	"lease_preferences": {types.String, func(c *config.ZoneConfig, d tree.Datum) {
+		loadYAML(&c.LeasePreferences, string(tree.MustBeDString(d)))
+		c.InheritedLeasePreferences = false
 	}},
-	"lease_preferences": {types.String, func(c *config.ZoneConfig, d tree.Datum) { loadYAML(&c.LeasePreferences, string(tree.MustBeDString(d))) }},
 }
 
 // zoneOptionKeys contains the keys from suportedZoneConfigOptions in
@@ -101,31 +116,38 @@ func (p *planner) SetZoneConfig(ctx context.Context, n *tree.SetZoneConfig) (pla
 		case types.String:
 		case types.Bytes:
 		default:
-			return nil, fmt.Errorf("zone config must be of type string or bytes, not %s", typ)
+			return nil, pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError,
+				"zone config must be of type string or bytes, not %s", typ)
 		}
 	}
 
-	var options map[tree.Name]tree.TypedExpr
+	var options map[tree.Name]optionValue
 	if n.Options != nil {
 		// We have a CONFIGURE ZONE USING ... assignment.
 		// Here we are constrained by the supported ZoneConfig fields,
 		// as described by supportedZoneConfigOptions above.
 
-		options = make(map[tree.Name]tree.TypedExpr)
+		options = make(map[tree.Name]optionValue)
 		for _, opt := range n.Options {
 			if _, alreadyExists := options[opt.Key]; alreadyExists {
-				return nil, fmt.Errorf("duplicate zone config parameter: %q", tree.ErrString(&opt.Key))
+				return nil, pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError,
+					"duplicate zone config parameter: %q", tree.ErrString(&opt.Key))
 			}
 			req, ok := supportedZoneConfigOptions[opt.Key]
 			if !ok {
-				return nil, fmt.Errorf("unsupported zone config parameter: %q", tree.ErrString(&opt.Key))
+				return nil, pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError,
+					"unsupported zone config parameter: %q", tree.ErrString(&opt.Key))
+			}
+			if opt.Value == nil {
+				options[opt.Key] = optionValue{inheritValue: true, explicitValue: nil}
+				continue
 			}
 			valExpr, err := p.analyzeExpr(
 				ctx, opt.Value, nil, tree.IndexedVarHelper{}, req.requiredType, true /*requireType*/, string(opt.Key))
 			if err != nil {
 				return nil, err
 			}
-			options[opt.Key] = valExpr
+			options[opt.Key] = optionValue{inheritValue: false, explicitValue: valExpr}
 		}
 	}
 
@@ -146,6 +168,7 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 	var yamlConfig string
 	var setters []func(c *config.ZoneConfig)
 	deleteZone := false
+	subzonePlaceholder := false
 
 	// Evaluate the configuration input.
 	if n.yamlConfig != nil {
@@ -167,6 +190,7 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 		yamlConfig = strings.TrimSpace(yamlConfig)
 	}
 	var optionStr strings.Builder
+	var copyFromParentList []tree.Name
 	if n.options != nil {
 		// Set from var = value attributes.
 		//
@@ -175,23 +199,39 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 		// the event log remains deterministic.
 		for i := range zoneOptionKeys {
 			name := (*tree.Name)(&zoneOptionKeys[i])
-			expr, ok := n.options[*name]
+			val, ok := n.options[*name]
 			if !ok {
 				continue
 			}
+			// We don't add the setters for the fields that will copy values
+			// from the parents. These fields will be set by taking what
+			// value would apply to the zone and setting that value explicitly.
+			// Instead we add the fields to a list that we use at a later time
+			// to copy values over.
+			inheritVal, expr := val.inheritValue, val.explicitValue
+			if inheritVal {
+				copyFromParentList = append(copyFromParentList, *name)
+				if optionStr.Len() > 0 {
+					optionStr.WriteString(", ")
+				}
+				fmt.Fprintf(&optionStr, "%s = COPY FROM PARENT", name)
+				continue
+			}
 			datum, err := expr.Eval(params.EvalContext())
 			if err != nil {
 				return err
 			}
 			if datum == tree.DNull {
-				return fmt.Errorf("unsupported NULL value for %q", tree.ErrString(name))
+				return pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError,
+					"unsupported NULL value for %q", tree.ErrString(name))
 			}
 			setter := supportedZoneConfigOptions[*name].setter
 			setters = append(setters, func(c *config.ZoneConfig) { setter(c, datum) })
 			if optionStr.Len() > 0 {
 				optionStr.WriteString(", ")
 			}
-			fmt.Fprintf(&optionStr, "%s = %s", string(*name), datum)
+			fmt.Fprintf(&optionStr, "%s = %s", name, datum)
+
 		}
 	}
 
@@ -230,6 +270,28 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 		return err
 	}
 
+	// Retrieve the partial zone configuration
+	partialZone, err := getZoneConfigRaw(params.ctx, params.p.txn, targetID)
+	if err != nil {
+		return err
+	}
+
+	// No zone was found. Possibly a SubzonePlaceholder dependig on the index.
+	if partialZone == nil {
+		partialZone = config.NewZoneConfig()
+		if index != nil {
+			subzonePlaceholder = true
+		}
+	}
+
+	var partialSubzone *config.Subzone
+	if index != nil {
+		partialSubzone = partialZone.GetSubzone(uint32(index.ID), partition)
+		if partialSubzone == nil {
+			partialSubzone = &config.Subzone{Config: *config.NewZoneConfig()}
+		}
+	}
+
 	// Retrieve the zone configuration.
 	//
 	// If the statement was USING DEFAULT, we want to ignore the zone
@@ -237,7 +299,9 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 	// default (whichever applies -- a database if targetID is a table,
 	// default if targetID is a database, etc.). For this, we use the last
 	// parameter getInheritedDefault to GetZoneConfigInTxn().
-	_, zone, subzone, err := GetZoneConfigInTxn(params.ctx, params.p.txn,
+	// These zones are only used for validations. The merged zone is will
+	// not be written.
+	_, completeZone, completeSubzone, err := GetZoneConfigInTxn(params.ctx, params.p.txn,
 		uint32(targetID), index, partition, n.setDefault)
 
 	if err == errNoZoneConfigApplies {
@@ -247,21 +311,27 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 		// the target ID is not a database object, i.e. one of the system
 		// ranges (liveness, meta, etc.), and did not have a zone config
 		// already.
-		zone = config.DefaultZoneConfig()
+		defZone := config.DefaultZoneConfig()
+		completeZone = &defZone
 	} else if err != nil {
 		return err
 	}
 
+	// Copy the fields set by the INHERIT field command.
+	partialZone.CopyFromZone(*completeZone, copyFromParentList)
+
 	if deleteZone {
 		if index != nil {
-			didDelete := zone.DeleteSubzone(uint32(index.ID), partition)
+			didDelete := completeZone.DeleteSubzone(uint32(index.ID), partition)
+			_ = partialZone.DeleteSubzone(uint32(index.ID), partition)
 			if !didDelete {
 				// If we didn't do any work, return early. We'd otherwise perform an
 				// update that would make it look like one row was affected.
 				return nil
 			}
 		} else {
-			zone.DeleteTableConfig()
+			completeZone.DeleteTableConfig()
+			partialZone.DeleteTableConfig()
 		}
 	} else {
 		// Validate the user input.
@@ -272,18 +342,42 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 		}
 
 		// Determine where to load the configuration.
-		newZone := zone
-		if subzone != nil {
-			newZone = subzone.Config
+		newZone := *completeZone
+		if completeSubzone != nil {
+			newZone = completeSubzone.Config
 		}
 
+		// Determine where to load the partial configuration.
+		// finalZone is where the new changes are unmarshalled onto.
+		// It must be a fresh ZoneConfig if a new subzone is being created.
+		// If an existing subzone is being modified, finalZone is overridden.
+		finalZone := *partialZone
+		if partialSubzone != nil {
+			finalZone = partialSubzone.Config
+		}
+
+		// ALTER RANGE default USING DEFAULT sets the default to the in
+		// memory default value.
+		if n.setDefault && keys.RootNamespaceID == uint32(targetID) {
+			finalZone = config.DefaultZoneConfig()
+		} else if n.setDefault {
+			finalZone = *config.NewZoneConfig()
+		}
 		// Load settings from YAML. If there was no YAML (e.g. because the
 		// query specified CONFIGURE ZONE USING), the YAML string will be
 		// empty, in which case the unmarshaling will be a no-op. This is
 		// innocuous.
 		if err := yaml.UnmarshalStrict([]byte(yamlConfig), &newZone); err != nil {
-			return fmt.Errorf("could not parse zone config: %v", err)
+			return pgerror.NewErrorf(pgerror.CodeCheckViolationError,
+				"could not parse zone config: %v", err)
 		}
+
+		// Load settings from YAML into the partial zone as well.
+		if err := yaml.UnmarshalStrict([]byte(yamlConfig), &finalZone); err != nil {
+			return pgerror.NewErrorf(pgerror.CodeCheckViolationError,
+				"could not parse zone config: %v", err)
+		}
+
 		// Load settings from var = val assignments. If there were no such
 		// settings, (e.g. because the query specified CONFIGURE ZONE = or
 		// USING DEFAULT), the setter slice will be empty and this will be
@@ -302,7 +396,9 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 						}
 					}
 				}()
+
 				setter(&newZone)
+				setter(&finalZone)
 				return nil
 			}(); err != nil {
 				return err
@@ -321,32 +417,73 @@ func (n *setZoneConfigNode) startExec(params runParams) error {
 		// Are we operating on an index?
 		if index == nil {
 			// No: the final zone config is the one we just processed.
-			zone = newZone
+			completeZone = &newZone
+			partialZone = &finalZone
 		} else {
 			// If the zone config for targetID was a subzone placeholder, it'll have
 			// been skipped over by GetZoneConfigInTxn. We need to load it regardless
 			// to avoid blowing away other subzones.
-			zone, err = getZoneConfigRaw(params.ctx, params.p.txn, targetID)
+
+			// TODO(ridwanmsharif): How is this supposed to change? getZoneConfigRaw
+			// gives no guarantees about completeness. Some work might need to happen
+			// here to complete the missing fields. The reason is because we don't know
+			// here if a zone is a placeholder or not. Can we do a GetConfigInTxn here?
+			// And if it is a placeholder, we use getZoneConfigRaw to create one.
+			completeZone, err = getZoneConfigRaw(params.ctx, params.p.txn, targetID)
 			if err != nil {
 				return err
+			} else if completeZone == nil {
+				completeZone = config.NewZoneConfig()
 			}
-			zone.SetSubzone(config.Subzone{
+			completeZone.SetSubzone(config.Subzone{
 				IndexID:       uint32(index.ID),
 				PartitionName: partition,
 				Config:        newZone,
 			})
+
+			// The partial zone might just be empty. If so,
+			// replace it with a SubzonePlaceholder.
+			if subzonePlaceholder {
+				partialZone.DeleteTableConfig()
+			}
+
+			partialZone.SetSubzone(config.Subzone{
+				IndexID:       uint32(index.ID),
+				PartitionName: partition,
+				Config:        finalZone,
+			})
 		}
 
-		// Finally revalidate everything.
-		if err := zone.Validate(); err != nil {
-			return fmt.Errorf("could not validate zone config: %s", err)
+		// Finally revalidate everything. Validate only the completeZone config.
+		if err := completeZone.Validate(); err != nil {
+			return pgerror.NewErrorf(pgerror.CodeCheckViolationError,
+				"could not validate zone config: %v", err)
 		}
 	}
 
-	// Write the new zone configuration.
+	// If cluster version is below 2.2, just write the complete zone
+	// config instead of the partial for backwards compatibility reasons.
+	// Otherwise write the partial zone configuration.
 	hasNewSubzones := !deleteZone && index != nil
+	execConfig := params.extendedEvalCtx.ExecCfg
+	zoneToWrite := partialZone
+	if !execConfig.Settings.Version.IsActive(cluster.VersionCascadingZoneConfigs) {
+		zoneToWrite = completeZone
+	}
+
+	// Finally check for the extra protection partial zone configs would
+	// require from changes made to parent zones. The extra protections are:
+	//
+	// RangeMinBytes and RangeMaxBytes must be set together
+	// LeasePreferences cannot be set unless Constraints are explicitly set
+	// Per-replica constraints cannot be set unless num_replicas is explicitly set
+	if err := zoneToWrite.ValidateTandemFields(); err != nil {
+		return pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError,
+			"could not validate zone config: %v", err).SetHintf(
+			"try ALTER ... CONFIGURE ZONE USING <field_name> = COPY FROM PARENT [, ...] so populate the field")
+	}
 	n.run.numAffected, err = writeZoneConfig(params.ctx, params.p.txn,
-		targetID, table, zone, params.extendedEvalCtx.ExecCfg, hasNewSubzones)
+		targetID, table, zoneToWrite, execConfig, hasNewSubzones)
 	if err != nil {
 		return err
 	}
@@ -454,7 +591,7 @@ func validateZoneAttrsAndLocalities(
 			}
 		}
 		if !found {
-			return fmt.Errorf(
+			return pgerror.NewErrorf(pgerror.CodeCheckViolationError,
 				"constraint %q matches no existing nodes within the cluster - did you enter it correctly?",
 				constraint)
 		}
@@ -468,32 +605,36 @@ func writeZoneConfig(
 	txn *client.Txn,
 	targetID sqlbase.ID,
 	table *sqlbase.TableDescriptor,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	execCfg *ExecutorConfig,
 	hasNewSubzones bool,
 ) (numAffected int, err error) {
 	if len(zone.Subzones) > 0 {
 		st := execCfg.Settings
-		if !st.Version.IsMinSupported(cluster.VersionPartitioning) {
-			return 0, errors.New("cluster version does not support zone configs on indexes or partitions")
+		if !st.Version.IsActive(cluster.VersionPartitioning) {
+			return 0, pgerror.NewError(pgerror.CodeCheckViolationError,
+				"cluster version does not support zone configs on indexes or partitions")
 		}
 		zone.SubzoneSpans, err = GenerateSubzoneSpans(
 			st, execCfg.ClusterID(), table, zone.Subzones, hasNewSubzones)
 		if err != nil {
 			return 0, err
 		}
+	} else {
+		// To keep the Subzone and SubzoneSpan arrays consistent
+		zone.SubzoneSpans = nil
 	}
 	if len(zone.Constraints) > 1 || (len(zone.Constraints) == 1 && zone.Constraints[0].NumReplicas != 0) {
 		st := execCfg.Settings
-		if !st.Version.IsMinSupported(cluster.VersionPerReplicaZoneConstraints) {
-			return 0, errors.New(
+		if !st.Version.IsActive(cluster.VersionPerReplicaZoneConstraints) {
+			return 0, pgerror.NewError(pgerror.CodeCheckViolationError,
 				"cluster version does not support zone configs with per-replica constraints")
 		}
 	}
 	if len(zone.LeasePreferences) > 0 {
 		st := execCfg.Settings
-		if !st.Version.IsMinSupported(cluster.VersionLeasePreferences) {
-			return 0, errors.New(
+		if !st.Version.IsActive(cluster.VersionLeasePreferences) {
+			return 0, pgerror.NewError(pgerror.CodeCheckViolationError,
 				"cluster version does not support zone configs with lease placement preferences")
 		}
 	}
@@ -503,9 +644,10 @@ func writeZoneConfig(
 			"DELETE FROM system.zones WHERE id = $1", targetID)
 	}
 
-	buf, err := protoutil.Marshal(&zone)
+	buf, err := protoutil.Marshal(zone)
 	if err != nil {
-		return 0, fmt.Errorf("could not marshal zone config: %s", err)
+		return 0, pgerror.NewErrorf(pgerror.CodeCheckViolationError,
+			"could not marshal zone config: %v", err)
 	}
 	return execCfg.InternalExecutor.Exec(ctx, "update-zone", txn,
 		"UPSERT INTO system.zones (id, config) VALUES ($1, $2)", targetID, buf)
@@ -513,22 +655,22 @@ func writeZoneConfig(
 
 // getZoneConfigRaw looks up the zone config with the given ID. Unlike
 // getZoneConfig, it does not attempt to ascend the zone config hierarchy. If no
-// zone config exists for the given ID, it returns an empty zone config.
+// zone config exists for the given ID, it returns nil.
 func getZoneConfigRaw(
 	ctx context.Context, txn *client.Txn, id sqlbase.ID,
-) (config.ZoneConfig, error) {
+) (*config.ZoneConfig, error) {
 	kv, err := txn.Get(ctx, config.MakeZoneKey(uint32(id)))
 	if err != nil {
-		return config.ZoneConfig{}, err
+		return nil, err
 	}
 	if kv.Value == nil {
-		return config.ZoneConfig{}, nil
+		return nil, nil
 	}
 	var zone config.ZoneConfig
 	if err := kv.ValueProto(&zone); err != nil {
-		return config.ZoneConfig{}, err
+		return nil, err
 	}
-	return zone, nil
+	return &zone, nil
 }
 
 func removeIndexZoneConfigs(
@@ -546,18 +688,18 @@ func removeIndexZoneConfigs(
 	zone, err := getZoneConfigRaw(ctx, txn, tableID)
 	if err != nil {
 		return err
+	} else if zone == nil {
+		zone = config.NewZoneConfig()
 	}
 
 	for _, indexDesc := range indexDescs {
 		zone.DeleteIndexSubzones(uint32(indexDesc.ID))
 	}
 
-	hasNewSubzones := false
-	_, err = writeZoneConfig(ctx, txn, tableID, tableDesc, zone, execCfg, hasNewSubzones)
-	if sqlbase.IsCCLRequiredError(err) {
-		return sqlbase.NewCCLRequiredError(fmt.Errorf("schema change requires a CCL binary "+
-			"because table %q has at least one remaining index or partition with a zone config",
-			tableDesc.Name))
+	// Ignore CCL required error to allow schema change to progress.
+	_, err = writeZoneConfig(ctx, txn, tableID, tableDesc, zone, execCfg, false /* hasNewSubzones */)
+	if err != nil && !sqlbase.IsCCLRequiredError(err) {
+		return err
 	}
-	return err
+	return nil
 }
