@@ -26,14 +26,15 @@ import (
 	"time"
 	"unicode/utf8"
 
-	"github.com/cockroachdb/cockroach/pkg/security"
-	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
-
 	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/jobs/jobspb"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/security"
 	"github.com/cockroachdb/cockroach/pkg/sql"
+	"github.com/cockroachdb/cockroach/pkg/sql/lex"
+	"github.com/cockroachdb/cockroach/pkg/sql/parser"
+	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/tests"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
@@ -42,6 +43,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/kr/pretty"
+	"github.com/pkg/errors"
 )
 
 func TestShowCreateTable(t *testing.T) {
@@ -55,9 +57,9 @@ func TestShowCreateTable(t *testing.T) {
 		CREATE DATABASE d;
 		SET DATABASE = d;
 		CREATE TABLE items (
-			a int,
-			b int,
-			c int unique,
+			a int8,
+			b int8,
+			c int8 unique,
 			primary key (a, b)
 		);
 		CREATE DATABASE o;
@@ -72,7 +74,7 @@ func TestShowCreateTable(t *testing.T) {
 	}{
 		{
 			stmt: `CREATE TABLE %s (
-	i INT,
+	i INT8,
 	s STRING NULL,
 	v FLOAT NOT NULL,
 	t TIMESTAMP DEFAULT now():::TIMESTAMP,
@@ -81,7 +83,7 @@ func TestShowCreateTable(t *testing.T) {
 	FAMILY fam_1_s (s)
 )`,
 			expect: `CREATE TABLE %s (
-	i INT NULL,
+	i INT8 NULL,
 	s STRING NULL,
 	v FLOAT8 NOT NULL,
 	t TIMESTAMP NULL DEFAULT now():::TIMESTAMP,
@@ -92,7 +94,7 @@ func TestShowCreateTable(t *testing.T) {
 		},
 		{
 			stmt: `CREATE TABLE %s (
-	i INT CHECK (i > 0),
+	i INT8 CHECK (i > 0),
 	s STRING NULL,
 	v FLOAT NOT NULL,
 	t TIMESTAMP DEFAULT now():::TIMESTAMP,
@@ -100,7 +102,7 @@ func TestShowCreateTable(t *testing.T) {
 	FAMILY fam_1_s (s)
 )`,
 			expect: `CREATE TABLE %s (
-	i INT NULL,
+	i INT8 NULL,
 	s STRING NULL,
 	v FLOAT8 NOT NULL,
 	t TIMESTAMP NULL DEFAULT now():::TIMESTAMP,
@@ -111,14 +113,14 @@ func TestShowCreateTable(t *testing.T) {
 		},
 		{
 			stmt: `CREATE TABLE %s (
-	i INT NULL,
+	i INT8 NULL,
 	s STRING NULL,
 	CONSTRAINT ck CHECK (i > 0),
 	FAMILY "primary" (i, rowid),
 	FAMILY fam_1_s (s)
 )`,
 			expect: `CREATE TABLE %s (
-	i INT NULL,
+	i INT8 NULL,
 	s STRING NULL,
 	FAMILY "primary" (i, rowid),
 	FAMILY fam_1_s (s),
@@ -127,24 +129,24 @@ func TestShowCreateTable(t *testing.T) {
 		},
 		{
 			stmt: `CREATE TABLE %s (
-	i INT PRIMARY KEY
+	i INT8 PRIMARY KEY
 )`,
 			expect: `CREATE TABLE %s (
-	i INT NOT NULL,
+	i INT8 NOT NULL,
 	CONSTRAINT "primary" PRIMARY KEY (i ASC),
 	FAMILY "primary" (i)
 )`,
 		},
 		{
 			stmt: `
-				CREATE TABLE %s (i INT, f FLOAT, s STRING, d DATE,
+				CREATE TABLE %s (i INT8, f FLOAT, s STRING, d DATE,
 				  FAMILY "primary" (i, f, d, rowid),
 				  FAMILY fam_1_s (s));
 				CREATE INDEX idx_if on %[1]s (f, i) STORING (s, d);
 				CREATE UNIQUE INDEX on %[1]s (d);
 			`,
 			expect: `CREATE TABLE %s (
-	i INT NULL,
+	i INT8 NULL,
 	f FLOAT8 NULL,
 	s STRING NULL,
 	d DATE NULL,
@@ -156,20 +158,20 @@ func TestShowCreateTable(t *testing.T) {
 		},
 		{
 			stmt: `CREATE TABLE %s (
-	"te""st" INT NOT NULL,
+	"te""st" INT8 NOT NULL,
 	CONSTRAINT "pri""mary" PRIMARY KEY ("te""st" ASC),
 	FAMILY "primary" ("te""st")
 )`,
 		},
 		{
 			stmt: `CREATE TABLE %s (
-	a int,
-	b int,
+	a int8,
+	b int8,
 	index c(a asc, b desc)
 )`,
 			expect: `CREATE TABLE %s (
-	a INT NULL,
-	b INT NULL,
+	a INT8 NULL,
+	b INT8 NULL,
 	INDEX c (a ASC, b DESC),
 	FAMILY "primary" (a, b, rowid)
 )`,
@@ -178,19 +180,39 @@ func TestShowCreateTable(t *testing.T) {
 		// have their db name omitted.
 		{
 			stmt: `CREATE TABLE %s (
-	i int,
-	j int,
+	i int8,
+	j int8,
 	FOREIGN KEY (i, j) REFERENCES items (a, b),
 	k int REFERENCES items (c)
 )`,
 			expect: `CREATE TABLE %s (
-	i INT NULL,
-	j INT NULL,
-	k INT NULL,
+	i INT8 NULL,
+	j INT8 NULL,
+	k INT8 NULL,
 	CONSTRAINT fk_i_ref_items FOREIGN KEY (i, j) REFERENCES items (a, b),
-	INDEX t7_auto_index_fk_i_ref_items (i ASC, j ASC),
+	INDEX %[1]s_auto_index_fk_i_ref_items (i ASC, j ASC),
 	CONSTRAINT fk_k_ref_items FOREIGN KEY (k) REFERENCES items (c),
-	INDEX t7_auto_index_fk_k_ref_items (k ASC),
+	INDEX %[1]s_auto_index_fk_k_ref_items (k ASC),
+	FAMILY "primary" (i, j, k, rowid)
+)`,
+		},
+		// Check that FK dependencies using MATCH FULL on a non-composite key still
+		// show
+		{
+			stmt: `CREATE TABLE %s (
+	i int8,
+	j int8,
+	k int REFERENCES items (c) MATCH FULL,
+	FOREIGN KEY (i, j) REFERENCES items (a, b) MATCH FULL
+)`,
+			expect: `CREATE TABLE %s (
+	i INT8 NULL,
+	j INT8 NULL,
+	k INT8 NULL,
+	CONSTRAINT fk_i_ref_items FOREIGN KEY (i, j) REFERENCES items (a, b) MATCH FULL,
+	INDEX %[1]s_auto_index_fk_i_ref_items (i ASC, j ASC),
+	CONSTRAINT fk_k_ref_items FOREIGN KEY (k) REFERENCES items (c) MATCH FULL,
+	INDEX %[1]s_auto_index_fk_k_ref_items (k ASC),
 	FAMILY "primary" (i, j, k, rowid)
 )`,
 		},
@@ -198,13 +220,13 @@ func TestShowCreateTable(t *testing.T) {
 		// have their db name prefixed.
 		{
 			stmt: `CREATE TABLE %s (
-	x INT,
+	x INT8,
 	CONSTRAINT fk_ref FOREIGN KEY (x) REFERENCES o.foo (x)
 )`,
 			expect: `CREATE TABLE %s (
-	x INT NULL,
+	x INT8 NULL,
 	CONSTRAINT fk_ref FOREIGN KEY (x) REFERENCES o.public.foo (x),
-	INDEX t8_auto_index_fk_ref (x ASC),
+	INDEX %[1]s_auto_index_fk_ref (x ASC),
 	FAMILY "primary" (x, rowid)
 )`,
 		},
@@ -212,19 +234,19 @@ func TestShowCreateTable(t *testing.T) {
 		// are pretty-printed properly. Regression test for #32529.
 		{
 			stmt: `CREATE TABLE %s (
-	i int DEFAULT 123,
-	j int DEFAULT 123,
+	i int8 DEFAULT 123,
+	j int8 DEFAULT 123,
 	FOREIGN KEY (i, j) REFERENCES items (a, b) ON DELETE SET DEFAULT,
-	k int REFERENCES items (c) ON DELETE SET NULL
+	k int8 REFERENCES items (c) ON DELETE SET NULL
 )`,
 			expect: `CREATE TABLE %s (
-	i INT NULL DEFAULT 123:::INT,
-	j INT NULL DEFAULT 123:::INT,
-	k INT NULL,
+	i INT8 NULL DEFAULT 123:::INT8,
+	j INT8 NULL DEFAULT 123:::INT8,
+	k INT8 NULL,
 	CONSTRAINT fk_i_ref_items FOREIGN KEY (i, j) REFERENCES items (a, b) ON DELETE SET DEFAULT,
-	INDEX t9_auto_index_fk_i_ref_items (i ASC, j ASC),
+	INDEX %[1]s_auto_index_fk_i_ref_items (i ASC, j ASC),
 	CONSTRAINT fk_k_ref_items FOREIGN KEY (k) REFERENCES items (c) ON DELETE SET NULL,
-	INDEX t9_auto_index_fk_k_ref_items (k ASC),
+	INDEX %[1]s_auto_index_fk_k_ref_items (k ASC),
 	FAMILY "primary" (i, j, k, rowid)
 )`,
 		},
@@ -232,13 +254,13 @@ func TestShowCreateTable(t *testing.T) {
 		// have their db name omitted.
 		{
 			stmt: `CREATE TABLE %s (
-	a INT,
-	b INT,
+	a INT8,
+	b INT8,
 	PRIMARY KEY (a, b)
 ) INTERLEAVE IN PARENT items (a, b)`,
 			expect: `CREATE TABLE %s (
-	a INT NOT NULL,
-	b INT NOT NULL,
+	a INT8 NOT NULL,
+	b INT8 NOT NULL,
 	CONSTRAINT "primary" PRIMARY KEY (a ASC, b ASC),
 	FAMILY "primary" (a, b)
 ) INTERLEAVE IN PARENT items (a, b)`,
@@ -247,18 +269,41 @@ func TestShowCreateTable(t *testing.T) {
 		// database are prefixed by their db name.
 		{
 			stmt: `CREATE TABLE %s (
-	x INT PRIMARY KEY
+	x INT8 PRIMARY KEY
 ) INTERLEAVE IN PARENT o.foo (x)`,
 			expect: `CREATE TABLE %s (
-	x INT NOT NULL,
+	x INT8 NOT NULL,
 	CONSTRAINT "primary" PRIMARY KEY (x ASC),
 	FAMILY "primary" (x)
 ) INTERLEAVE IN PARENT o.public.foo (x)`,
 		},
+		// Check that FK dependencies using MATCH FULL and MATCH SIMPLE are both
+		// pretty-printed properly.
+		{
+			stmt: `CREATE TABLE %s (
+	i int DEFAULT 1,
+	j int DEFAULT 2,
+	k int DEFAULT 3,
+	l int DEFAULT 4,
+	FOREIGN KEY (i, j) REFERENCES items (a, b) MATCH SIMPLE ON DELETE SET DEFAULT,
+	FOREIGN KEY (k, l) REFERENCES items (a, b) MATCH FULL ON UPDATE CASCADE
+)`,
+			expect: `CREATE TABLE %s (
+	i INT8 NULL DEFAULT 1:::INT8,
+	j INT8 NULL DEFAULT 2:::INT8,
+	k INT8 NULL DEFAULT 3:::INT8,
+	l INT8 NULL DEFAULT 4:::INT8,
+	CONSTRAINT fk_i_ref_items FOREIGN KEY (i, j) REFERENCES items (a, b) ON DELETE SET DEFAULT,
+	INDEX %[1]s_auto_index_fk_i_ref_items (i ASC, j ASC),
+	CONSTRAINT fk_k_ref_items FOREIGN KEY (k, l) REFERENCES items (a, b) MATCH FULL ON UPDATE CASCADE,
+	INDEX %[1]s_auto_index_fk_k_ref_items (k ASC, l ASC),
+	FAMILY "primary" (i, j, k, l, rowid)
+)`,
+		},
 	}
 	for i, test := range tests {
-		t.Run(fmt.Sprintf("%d/%s", i, strings.Replace(test.stmt, "\n", "", -1)), func(t *testing.T) {
-			name := fmt.Sprintf("t%d", i)
+		name := fmt.Sprintf("t%d", i)
+		t.Run(name, func(t *testing.T) {
 			if test.expect == "" {
 				test.expect = test.stmt
 			}
@@ -303,9 +348,6 @@ func TestShowCreateTable(t *testing.T) {
 
 func TestShowCreateView(t *testing.T) {
 	defer leaktest.AfterTest(t)()
-	if testing.Short() {
-		t.Skip("short #26969")
-	}
 
 	params, _ := tests.CreateTestServerParams()
 	s, sqlDB, _ := serverutils.StartServer(t, params)
@@ -344,8 +386,8 @@ func TestShowCreateView(t *testing.T) {
 			`CREATE VIEW %s (count) AS SELECT count(*) FROM d.public.t`,
 		},
 		{
-			`CREATE VIEW %s AS SELECT s, count(*) FROM t GROUP BY s HAVING count(*) > 3:::INT`,
-			`CREATE VIEW %s (s, count) AS SELECT s, count(*) FROM d.public.t GROUP BY s HAVING count(*) > 3:::INT`,
+			`CREATE VIEW %s AS SELECT s, count(*) FROM t GROUP BY s HAVING count(*) > 3:::INT8`,
+			`CREATE VIEW %s (s, count) AS SELECT s, count(*) FROM d.public.t GROUP BY s HAVING count(*) > 3:::INT8`,
 		},
 		{
 			`CREATE VIEW %s (a, b, c, d) AS SELECT i, s, v, t FROM t`,
@@ -514,7 +556,6 @@ func TestShowQueries(t *testing.T) {
 		if stmt == selectStmt {
 			found = true
 			const showQuery = "SELECT node_id, (now() - start)::FLOAT8, query FROM [SHOW CLUSTER QUERIES]"
-			const showVersionQuery = "SHOW CLUSTER SETTING version"
 
 			rows, err := conn1.Query(showQuery)
 			if err != nil {
@@ -522,27 +563,16 @@ func TestShowQueries(t *testing.T) {
 			}
 			defer rows.Close()
 
-			count := 0
+			var stmts []string
 			for rows.Next() {
-				count++
-
 				var nodeID int
-				var sql string
+				var stmt string
 				var delta float64
-				if err := rows.Scan(&nodeID, &delta, &sql); err != nil {
+				if err := rows.Scan(&nodeID, &delta, &stmt); err != nil {
 					failure = err
 					return
 				}
-				switch sql {
-				case showQuery, showVersionQuery, expectedSelectStmt:
-				default:
-					failure = fmt.Errorf(
-						"unexpected query in SHOW QUERIES: %+q, expected: %+q",
-						sql,
-						expectedSelectStmt,
-					)
-					return
-				}
+				stmts = append(stmts, stmt)
 				if nodeID < 1 || nodeID > 2 {
 					failure = fmt.Errorf("invalid node ID: %d", nodeID)
 					return
@@ -562,9 +592,14 @@ func TestShowQueries(t *testing.T) {
 				return
 			}
 
-			if expectedCount := 2; count != expectedCount {
-				failure = fmt.Errorf("unexpected number of running queries: %d, expected %d", count, expectedCount)
-				return
+			foundSelect := false
+			for _, stmt := range stmts {
+				if stmt == expectedSelectStmt {
+					foundSelect = true
+				}
+			}
+			if !foundSelect {
+				failure = fmt.Errorf("original query not found in SHOW QUERIES. expected: %s\nactual: %v", selectStmt, stmts)
 			}
 		}
 	}
@@ -600,7 +635,7 @@ func TestShowQueries(t *testing.T) {
 	// Now check the behavior on error.
 	tc.StopServer(1)
 
-	rows, err := conn1.Query(`SELECT node_id, query FROM [SHOW CLUSTER QUERIES]`)
+	rows, err := conn1.Query(`SELECT node_id, query FROM [SHOW ALL CLUSTER QUERIES]`)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -711,7 +746,7 @@ func TestShowSessions(t *testing.T) {
 	// Now check the behavior on error.
 	tc.StopServer(1)
 
-	rows, err = conn.Query(`SELECT node_id, active_queries FROM [SHOW CLUSTER SESSIONS]`)
+	rows, err = conn.Query(`SELECT node_id, active_queries FROM [SHOW ALL CLUSTER SESSIONS]`)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -958,6 +993,68 @@ func TestShowJobs(t *testing.T) {
 	}
 }
 
+func TestShowAutomaticJobs(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	params, _ := tests.CreateTestServerParams()
+	s, rawSQLDB, _ := serverutils.StartServer(t, params)
+	sqlDB := sqlutils.MakeSQLRunner(rawSQLDB)
+	defer s.Stopper().Stop(context.TODO())
+
+	// row represents a row returned from crdb_internal.jobs, but
+	// *not* a row in system.jobs.
+	type row struct {
+		id      int64
+		typ     string
+		status  string
+		details jobspb.Details
+	}
+
+	rows := []row{
+		{
+			id:      1,
+			typ:     "CREATE STATS",
+			status:  "running",
+			details: jobspb.CreateStatsDetails{Name: "my_stats"},
+		},
+		{
+			id:      2,
+			typ:     "AUTO CREATE STATS",
+			status:  "running",
+			details: jobspb.CreateStatsDetails{Name: "__auto__"},
+		},
+	}
+
+	for _, in := range rows {
+		// system.jobs is part proper SQL columns, part protobuf, so we can't use the
+		// row struct directly.
+		inPayload, err := protoutil.Marshal(&jobspb.Payload{
+			Details: jobspb.WrapPayloadDetails(in.details),
+		})
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		sqlDB.Exec(t,
+			`INSERT INTO system.jobs (id, status, payload) VALUES ($1, $2, $3)`,
+			in.id, in.status, inPayload,
+		)
+	}
+
+	var out row
+	sqlDB.QueryRow(t, `SELECT job_id, job_type FROM [SHOW JOBS]`).Scan(&out.id, &out.typ)
+	if out.id != 1 || out.typ != "CREATE STATS" {
+		t.Fatalf("Expected id:%d and type:%s but found id:%d and type:%s",
+			1, "CREATE STATS", out.id, out.typ)
+	}
+
+	sqlDB.QueryRow(t, `SELECT job_id, job_type FROM [SHOW AUTOMATIC JOBS]`).Scan(&out.id, &out.typ)
+	if out.id != 2 || out.typ != "AUTO CREATE STATS" {
+		t.Fatalf("Expected id:%d and type:%s but found id:%d and type:%s",
+			2, "AUTO CREATE STATS", out.id, out.typ)
+	}
+}
+
 func TestShowJobsWithError(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
@@ -1078,3 +1175,122 @@ func TestShowJobsWithError(t *testing.T) {
 	}
 	rowNum++
 }
+
+func TestLintClusterSettingNames(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	params, _ := tests.CreateTestServerParams()
+	s, sqlDB, _ := serverutils.StartServer(t, params)
+	defer s.Stopper().Stop(context.TODO())
+
+	rows, err := sqlDB.Query(`SELECT variable, setting_type, description FROM [SHOW ALL CLUSTER SETTINGS]`)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer rows.Close()
+
+	for rows.Next() {
+		var varName, sType, desc string
+		if err := rows.Scan(&varName, &sType, &desc); err != nil {
+			t.Fatal(err)
+		}
+
+		if strings.ToLower(varName) != varName {
+			t.Errorf("%s: variable name must be all lowercase", varName)
+		}
+
+		suffixSuggestions := map[string]string{
+			"_ttl":     ".ttl",
+			"_enabled": ".enabled",
+			"_timeout": ".timeout",
+		}
+
+		nameErr := func() error {
+			segments := strings.Split(varName, ".")
+			for _, segment := range segments {
+				if strings.TrimSpace(segment) != segment {
+					return errors.Errorf("%s: part %q has heading or trailing whitespace", varName, segment)
+				}
+				tokens, ok := parser.Tokens(segment)
+				if !ok {
+					return errors.Errorf("%s: part %q does not scan properly", varName, segment)
+				}
+				if len(tokens) == 0 || len(tokens) > 1 {
+					return errors.Errorf("%s: part %q has invalid structure", varName, segment)
+				}
+				if tokens[0].TokenID != parser.IDENT {
+					cat, ok := lex.KeywordsCategories[tokens[0].Str]
+					if !ok {
+						return errors.Errorf("%s: part %q has invalid structure", varName, segment)
+					}
+					if cat == "R" {
+						return errors.Errorf("%s: part %q is a reserved keyword", varName, segment)
+					}
+				}
+			}
+
+			for suffix, repl := range suffixSuggestions {
+				if strings.HasSuffix(varName, suffix) {
+					return errors.Errorf("%s: use %q instead of %q", varName, repl, suffix)
+				}
+			}
+
+			if sType == "b" && !strings.HasSuffix(varName, ".enabled") {
+				return errors.Errorf("%s: use .enabled for booleans", varName)
+			}
+
+			return nil
+		}()
+		if nameErr != nil {
+			var grandFathered = map[string]string{
+				"server.declined_reservation_timeout":                `server.declined_reservation_timeout: use ".timeout" instead of "_timeout"`,
+				"server.failed_reservation_timeout":                  `server.failed_reservation_timeout: use ".timeout" instead of "_timeout"`,
+				"server.web_session_timeout":                         `server.web_session_timeout: use ".timeout" instead of "_timeout"`,
+				"sql.distsql.flow_stream_timeout":                    `sql.distsql.flow_stream_timeout: use ".timeout" instead of "_timeout"`,
+				"debug.panic_on_failed_assertions":                   `debug.panic_on_failed_assertions: use .enabled for booleans`,
+				"diagnostics.reporting.send_crash_reports":           `diagnostics.reporting.send_crash_reports: use .enabled for booleans`,
+				"kv.closed_timestamp.follower_reads_enabled":         `kv.closed_timestamp.follower_reads_enabled: use ".enabled" instead of "_enabled"`,
+				"kv.raft_log.disable_synchronization_unsafe":         `kv.raft_log.disable_synchronization_unsafe: use .enabled for booleans`,
+				"kv.range_merge.queue_enabled":                       `kv.range_merge.queue_enabled: use ".enabled" instead of "_enabled"`,
+				"kv.range_split.by_load_enabled":                     `kv.range_split.by_load_enabled: use ".enabled" instead of "_enabled"`,
+				"kv.transaction.write_pipelining_enabled":            `kv.transaction.write_pipelining_enabled: use ".enabled" instead of "_enabled"`,
+				"server.clock.forward_jump_check_enabled":            `server.clock.forward_jump_check_enabled: use ".enabled" instead of "_enabled"`,
+				"sql.defaults.experimental_optimizer_mutations":      `sql.defaults.experimental_optimizer_mutations: use .enabled for booleans`,
+				"sql.distsql.distribute_index_joins":                 `sql.distsql.distribute_index_joins: use .enabled for booleans`,
+				"sql.distsql.temp_storage.joins":                     `sql.distsql.temp_storage.joins: use .enabled for booleans`,
+				"sql.distsql.temp_storage.sorts":                     `sql.distsql.temp_storage.sorts: use .enabled for booleans`,
+				"sql.metrics.statement_details.dump_to_logs":         `sql.metrics.statement_details.dump_to_logs: use .enabled for booleans`,
+				"sql.metrics.statement_details.sample_logical_plans": `sql.metrics.statement_details.sample_logical_plans: use .enabled for booleans`,
+				"sql.trace.log_statement_execute":                    `sql.trace.log_statement_execute: use .enabled for booleans`,
+				"trace.debug.enable":                                 `trace.debug.enable: use .enabled for booleans`,
+				// These two settings have been deprecated in favor of a new (better named) setting
+				// but the old name is still around to support migrations.
+				// TODO(knz): remove these cases when these settings are retired.
+				"timeseries.storage.10s_resolution_ttl": `timeseries.storage.10s_resolution_ttl: part "10s_resolution_ttl" has invalid structure`,
+				"timeseries.storage.30m_resolution_ttl": `timeseries.storage.30m_resolution_ttl: part "30m_resolution_ttl" has invalid structure`,
+			}
+			expectedErr, found := grandFathered[varName]
+			if !found || expectedErr != nameErr.Error() {
+				t.Error(nameErr)
+			}
+		}
+
+		if strings.TrimSpace(desc) != desc {
+			t.Errorf("%s: description %q has heading or trailing whitespace", varName, desc)
+		}
+
+		if len(desc) == 0 {
+			t.Errorf("%s: description is empty", varName)
+		}
+
+		if len(desc) > 0 {
+			if strings.ToLower(desc[0:1]) != desc[0:1] {
+				t.Errorf("%s: description %q must not start with capital", varName, desc)
+			}
+			if strings.Contains(desc, ". ") != (desc[len(desc)-1] == '.') {
+				t.Errorf("%s: description %q must end with period if and only if it contains a secondary sentence", varName, desc)
+			}
+		}
+	}
+
+}
