@@ -16,9 +16,11 @@ package sqlbase
 
 import (
 	"fmt"
+	"math"
 	"strings"
 	"unicode/utf8"
 
+	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/sql/coltypes"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
@@ -110,8 +112,23 @@ func DatumTypeToColumnType(ptyp types.T) (ColumnType, error) {
 // (e.g. via CastTargetToDatumType).
 func PopulateTypeAttrs(base ColumnType, typ coltypes.T) (ColumnType, error) {
 	switch t := typ.(type) {
-	case *coltypes.TInt:
+	case *coltypes.TBitArray:
+		if t.Width > math.MaxInt32 {
+			return ColumnType{}, fmt.Errorf("bit width too large: %d", t.Width)
+		}
 		base.Width = int32(t.Width)
+		if t.Variable {
+			base.VisibleType = ColumnType_VARBIT
+		}
+
+	case *coltypes.TInt:
+		// Ensure that "naked" INT types are promoted to INT8 to preserve
+		// compatibility with previous versions.
+		if t.Width == 0 {
+			base.Width = 64
+		} else {
+			base.Width = int32(t.Width)
+		}
 
 		// For 2.1 nodes only Width is sufficient, but we also populate
 		// VisibleType for compatibility with pre-2.1 nodes.
@@ -224,6 +241,16 @@ func (c *ColumnType) stringTypeName() string {
 // See also InformationSchemaVisibleType() below.
 func (c *ColumnType) SQLString() string {
 	switch c.SemanticType {
+	case ColumnType_BIT:
+		typName := "BIT"
+		if c.VisibleType == ColumnType_VARBIT {
+			typName = "VARBIT"
+		}
+		if (c.VisibleType != ColumnType_VARBIT && c.Width > 1) ||
+			(c.VisibleType == ColumnType_VARBIT && c.Width > 0) {
+			typName = fmt.Sprintf("%s(%d)", typName, c.Width)
+		}
+		return typName
 	case ColumnType_INT:
 		// Pre-2.1 BIT was using column type INT with arbitrary width We
 		// map this to INT now. See #34161.
@@ -362,7 +389,7 @@ func (c *ColumnType) InformationSchemaVisibleType() string {
 // expectations.
 func (c *ColumnType) MaxCharacterLength() (int32, bool) {
 	switch c.SemanticType {
-	case ColumnType_STRING, ColumnType_COLLATEDSTRING:
+	case ColumnType_STRING, ColumnType_COLLATEDSTRING, ColumnType_BIT:
 		if c.Width > 0 {
 			return c.Width, true
 		}
@@ -482,6 +509,8 @@ func (c *ColumnType) FloatProperties() (int32, int32) {
 // determination of DatumTypeHasCompositeKeyEncoding().
 func datumTypeToColumnSemanticType(ptyp types.T) (ColumnType_SemanticType, error) {
 	switch ptyp {
+	case types.BitArray:
+		return ColumnType_BIT, nil
 	case types.Bool:
 		return ColumnType_BOOL, nil
 	case types.Int:
@@ -539,6 +568,8 @@ func datumTypeToColumnSemanticType(ptyp types.T) (ColumnType_SemanticType, error
 // column type.
 func columnSemanticTypeToDatumType(c *ColumnType, k ColumnType_SemanticType) types.T {
 	switch k {
+	case ColumnType_BIT:
+		return types.BitArray
 	case ColumnType_BOOL:
 		return types.Bool
 	case ColumnType_INT:
@@ -620,26 +651,32 @@ func ColumnTypesToDatumTypes(colTypes []ColumnType) []types.T {
 	return res
 }
 
-// CheckValueWidth checks that the width (for strings, byte arrays,
-// and bit strings) and scale (for decimals) of the value fits the
-// specified column type. Used by INSERT and UPDATE.
-func CheckValueWidth(typ ColumnType, val tree.Datum, name *string) error {
+// LimitValueWidth checks that the width (for strings, byte arrays, and bit
+// strings) and scale (for decimals) of the value fits the specified column
+// type. In case of decimals, it can truncate fractional digits in the input
+// value in order to fit the target column. If the input value fits the target
+// column, it is returned unchanged. If the input value can be truncated to fit,
+// then a truncated copy is returned. Otherwise, an error is returned. This
+// method is used by INSERT and UPDATE.
+func LimitValueWidth(
+	typ ColumnType, inVal tree.Datum, name *string,
+) (outVal tree.Datum, err error) {
 	switch typ.SemanticType {
 	case ColumnType_STRING, ColumnType_COLLATEDSTRING:
 		var sv string
-		if v, ok := tree.AsDString(val); ok {
+		if v, ok := tree.AsDString(inVal); ok {
 			sv = string(v)
-		} else if v, ok := val.(*tree.DCollatedString); ok {
+		} else if v, ok := inVal.(*tree.DCollatedString); ok {
 			sv = v.Contents
 		}
 
 		if typ.Width > 0 && utf8.RuneCountInString(sv) > int(typ.Width) {
-			return pgerror.NewErrorf(pgerror.CodeStringDataRightTruncationError,
+			return nil, pgerror.NewErrorf(pgerror.CodeStringDataRightTruncationError,
 				"value too long for type %s (column %q)",
-				typ.SQLString(), tree.ErrNameString(name))
+				typ.SQLString(), tree.ErrNameStringP(name))
 		}
 	case ColumnType_INT:
-		if v, ok := tree.AsDInt(val); ok {
+		if v, ok := tree.AsDInt(inVal); ok {
 			if typ.Width == 32 || typ.Width == 64 || typ.Width == 16 {
 				// Width is defined in bits.
 				width := uint(typ.Width - 1)
@@ -647,36 +684,85 @@ func CheckValueWidth(typ ColumnType, val tree.Datum, name *string) error {
 				// We're performing bounds checks inline with Go's implementation of min and max ints in Math.go.
 				shifted := v >> width
 				if (v >= 0 && shifted > 0) || (v < 0 && shifted < -1) {
-					return pgerror.NewErrorf(pgerror.CodeNumericValueOutOfRangeError,
+					return nil, pgerror.NewErrorf(pgerror.CodeNumericValueOutOfRangeError,
 						"integer out of range for type %s (column %q)",
-						typ.VisibleType, tree.ErrNameString(name))
+						typ.VisibleType, tree.ErrNameStringP(name))
+				}
+			}
+		}
+	case ColumnType_BIT:
+		if v, ok := tree.AsDBitArray(inVal); ok {
+			if typ.Width > 0 {
+				bitLen := v.BitLen()
+				switch typ.VisibleType {
+				case ColumnType_VARBIT:
+					if bitLen > uint(typ.Width) {
+						return nil, pgerror.NewErrorf(pgerror.CodeStringDataRightTruncationError,
+							"bit string length %d too large for type %s", bitLen, typ.SQLString())
+					}
+				default:
+					if bitLen != uint(typ.Width) {
+						return nil, pgerror.NewErrorf(pgerror.CodeStringDataLengthMismatchError,
+							"bit string length %d does not match type %s", bitLen, typ.SQLString())
+					}
 				}
 			}
 		}
 	case ColumnType_DECIMAL:
-		if v, ok := val.(*tree.DDecimal); ok {
-			if err := tree.LimitDecimalWidth(&v.Decimal, int(typ.Precision), int(typ.Width)); err != nil {
-				return errors.Wrapf(err, "type %s (column %q)",
-					typ.SQLString(), tree.ErrNameString(name))
+		if inDec, ok := inVal.(*tree.DDecimal); ok {
+			if inDec.Form != apd.Finite || typ.Precision == 0 {
+				// Non-finite form or unlimited target precision, so no need to limit.
+				break
 			}
+			if int64(typ.Precision) >= inDec.NumDigits() && typ.Width == inDec.Exponent {
+				// Precision and scale of target column are sufficient.
+				break
+			}
+
+			var outDec tree.DDecimal
+			outDec.Set(&inDec.Decimal)
+			err := tree.LimitDecimalWidth(&outDec.Decimal, int(typ.Precision), int(typ.Width))
+			if err != nil {
+				return nil, errors.Wrapf(err, "type %s (column %q)",
+					typ.SQLString(), tree.ErrNameStringP(name))
+			}
+			return &outDec, nil
 		}
 	case ColumnType_ARRAY:
-		if v, ok := val.(*tree.DArray); ok {
+		if inArr, ok := inVal.(*tree.DArray); ok {
+			var outArr *tree.DArray
 			elementType := *typ.elementColumnType()
-			for i := range v.Array {
-				if err := CheckValueWidth(elementType, v.Array[i], name); err != nil {
-					return err
+			for i, inElem := range inArr.Array {
+				outElem, err := LimitValueWidth(elementType, inElem, name)
+				if err != nil {
+					return nil, err
 				}
+				if outElem != inElem {
+					if outArr == nil {
+						outArr = &tree.DArray{
+							ParamTyp: inArr.ParamTyp,
+							Array:    make(tree.Datums, len(inArr.Array)),
+							HasNulls: inArr.HasNulls,
+						}
+						copy(outArr.Array, inArr.Array[:i])
+					}
+				}
+				if outArr != nil {
+					outArr.Array[i] = inElem
+				}
+			}
+			if outArr != nil {
+				return outArr, nil
 			}
 		}
 	}
-	return nil
+	return inVal, nil
 }
 
 // elementColumnType works on a ColumnType with semantic type ARRAY
 // and retrieves the ColumnType of the elements of the array.
 //
-// This is used by CheckValueWidth() and SQLType().
+// This is used by LimitValueWidth() and SQLType().
 //
 // TODO(knz): make this return a bool and avoid a heap allocation.
 func (c *ColumnType) elementColumnType() *ColumnType {
@@ -709,16 +795,16 @@ func CheckDatumTypeFitsColumnType(
 	// populated 'colTyp' with a type to assign to it.
 	colTyp := col.Type.ToDatumType()
 	if p, pok := typ.(types.TPlaceholder); pok {
-		if err := pmap.SetType(p.Name, colTyp); err != nil {
+		if err := pmap.SetType(p.Idx, colTyp); err != nil {
 			return pgerror.NewErrorf(pgerror.CodeIndeterminateDatatypeError,
 				"cannot infer type for placeholder %s from column %q: %s",
-				p.Name, tree.ErrNameString(&col.Name), err)
+				p.Idx, tree.ErrNameString(col.Name), err)
 		}
 	} else if !typ.Equivalent(colTyp) {
 		// Not a placeholder; check that the value cast has succeeded.
 		return pgerror.NewErrorf(pgerror.CodeDatatypeMismatchError,
 			"value type %s doesn't match type %s of column %q",
-			typ, col.Type.SQLString(), tree.ErrNameString(&col.Name))
+			typ, col.Type.SQLString(), tree.ErrNameString(col.Name))
 	}
 	return nil
 }
