@@ -17,10 +17,8 @@ package sqlbase
 import (
 	"context"
 	"testing"
-
-	"unsafe"
-
 	"time"
+	"unsafe"
 
 	"github.com/cockroachdb/apd"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
@@ -300,7 +298,7 @@ func TestEncDatumFromBuffer(t *testing.T) {
 			}
 			buf, err = ed[i].Encode(&types[i], &alloc, enc[i], buf)
 			if err != nil {
-				t.Fatal(err)
+				t.Fatalf("Failed to encode type %v: %s", types[i], err)
 			}
 		}
 		// Decode the buffer.
@@ -312,14 +310,14 @@ func TestEncDatumFromBuffer(t *testing.T) {
 			var decoded EncDatum
 			decoded, b, err = EncDatumFromBuffer(&types[i], enc[i], b)
 			if err != nil {
-				t.Fatal(err)
+				t.Fatalf("%+v: encdatum from %+v: %+v (%+v)", ed[i].Datum, enc[i], err, &types[i])
 			}
 			err = decoded.EnsureDecoded(&types[i], &alloc)
 			if err != nil {
-				t.Fatal(err)
+				t.Fatalf("%+v: ensuredecoded: %v (%+v)", ed[i], err, &types[i])
 			}
 			if decoded.Datum.Compare(evalCtx, ed[i].Datum) != 0 {
-				t.Errorf("decoded datum %s doesn't equal original %s", decoded.Datum, ed[i].Datum)
+				t.Errorf("decoded datum %+v doesn't equal original %+v", decoded.Datum, ed[i].Datum)
 			}
 		}
 		if len(b) != 0 {
@@ -548,12 +546,14 @@ func TestEncDatumSize(t *testing.T) {
 		asc  = DatumEncoding_ASCENDING_KEY
 		desc = DatumEncoding_DESCENDING_KEY
 
-		DIntSize           = unsafe.Sizeof(tree.DInt(0))
-		DFloatSize         = unsafe.Sizeof(tree.DFloat(0))
-		DDecimalStructSize = unsafe.Sizeof(tree.DDecimal{Decimal: *apd.New(0, 0)})
-		DStringSize        = unsafe.Sizeof(*tree.NewDString(""))
+		DIntSize    = unsafe.Sizeof(tree.DInt(0))
+		DFloatSize  = unsafe.Sizeof(tree.DFloat(0))
+		DStringSize = unsafe.Sizeof(*tree.NewDString(""))
 	)
 
+	dec12300 := &tree.DDecimal{Decimal: *apd.New(123, 2)}
+	decimalSize := dec12300.Size()
+
 	testCases := []struct {
 		encDatum     EncDatum
 		expectedSize uintptr
@@ -611,16 +611,16 @@ func TestEncDatumSize(t *testing.T) {
 			expectedSize: EncDatumOverhead + 4, // 123.0 is encoded with length 4 byte array
 		},
 		{
-			encDatum:     DatumToEncDatum(ColumnType{SemanticType: ColumnType_DECIMAL}, &tree.DDecimal{Decimal: *apd.New(123, 2)}),
-			expectedSize: EncDatumOverhead + DDecimalStructSize + 40, // In this case, DDecimal references 40 bytes for internal things
+			encDatum:     DatumToEncDatum(ColumnType{SemanticType: ColumnType_DECIMAL}, dec12300),
+			expectedSize: EncDatumOverhead + decimalSize,
 		},
 		{
 			encDatum: EncDatum{
 				encoding: asc,
-				encoded:  encoding.EncodeDecimalAscending(nil, apd.New(123, 2)),
-				Datum:    &tree.DDecimal{Decimal: *apd.New(123, 2)},
+				encoded:  encoding.EncodeDecimalAscending(nil, &dec12300.Decimal),
+				Datum:    dec12300,
 			},
-			expectedSize: EncDatumOverhead + 4 + DDecimalStructSize + 40, // 12300.0 is encoded with length 4 byte array
+			expectedSize: EncDatumOverhead + 4 + decimalSize,
 		},
 		{
 			encDatum:     EncDatumFromEncoded(asc, encoding.EncodeStringAscending(nil, "")),
