@@ -18,8 +18,6 @@ import (
 	"fmt"
 	"sort"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
@@ -27,6 +25,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/util/encoding"
 	"github.com/cockroachdb/cockroach/pkg/util/json"
+	"github.com/pkg/errors"
 )
 
 // This file contains facilities to encode primary and secondary
@@ -348,6 +347,13 @@ func appendEncDatumsToKey(
 	return key, nil
 }
 
+// EncodeTableIDIndexID encodes a table id followed by an index id.
+func EncodeTableIDIndexID(key []byte, tableID ID, indexID IndexID) []byte {
+	key = encoding.EncodeUvarintAscending(key, uint64(tableID))
+	key = encoding.EncodeUvarintAscending(key, uint64(indexID))
+	return key
+}
+
 // DecodeTableIDIndexID decodes a table id followed by an index id.
 func DecodeTableIDIndexID(key []byte) ([]byte, ID, IndexID, error) {
 	var tableID uint64
@@ -518,63 +524,6 @@ func DecodeIndexKeyWithoutTableIDIndexIDPrefix(
 	return key, true, nil
 }
 
-// consumeIndexKeyWithoutTableIDIndexIDPrefix consumes an index key that's
-// already pre-stripped of its table ID index ID prefix, up to nCols columns,
-// returning the number of bytes consumed. For example, given an input key
-// with values (6,7,8,9) such as /Table/60/1/6/7/#/61/1/8/9, stripping 3 columns
-// from this key would eat all but the final, 4th column 9 in this example,
-// producing /Table/60/1/6/7/#/61/1/8. If nCols was 2, instead, the result
-// would include the trailing table ID index ID pair, since that's a more
-// precise key: /Table/60/1/6/7/#/61/1.
-func consumeIndexKeyWithoutTableIDIndexIDPrefix(
-	index *IndexDescriptor, nCols int, key []byte,
-) (int, error) {
-	origKeyLen := len(key)
-	consumedCols := 0
-	for _, ancestor := range index.Interleave.Ancestors {
-		length := int(ancestor.SharedPrefixLen)
-		// Skip up to length values.
-		for j := 0; j < length; j++ {
-			if consumedCols == nCols {
-				// We're done early, in the middle of an interleave.
-				return origKeyLen - len(key), nil
-			}
-			l, err := encoding.PeekLength(key)
-			if err != nil {
-				return 0, err
-			}
-			key = key[l:]
-			consumedCols++
-		}
-		var ok bool
-		key, ok = encoding.DecodeIfInterleavedSentinel(key)
-		if !ok {
-			return 0, errors.New("unexpected lack of sentinel key")
-		}
-
-		// Skip the TableID/IndexID pair for each ancestor except for the
-		// first, which has already been skipped in our input.
-		for j := 0; j < 2; j++ {
-			idLen, err := encoding.PeekLength(key)
-			if err != nil {
-				return 0, err
-			}
-			key = key[idLen:]
-		}
-	}
-
-	// Decode the remaining values in the key, in the final interleave.
-	for ; consumedCols < nCols; consumedCols++ {
-		l, err := encoding.PeekLength(key)
-		if err != nil {
-			return 0, err
-		}
-		key = key[l:]
-	}
-
-	return origKeyLen - len(key), nil
-}
-
 // DecodeKeyVals decodes the values that are part of the key. The decoded
 // values are stored in the vals. If this slice is nil, the direction
 // used will default to encoding.Ascending.
@@ -762,7 +711,7 @@ func EncodeInvertedIndexTableKeys(val tree.Datum, inKey []byte) (key [][]byte, e
 // EncodeSecondaryIndex encodes key/values for a secondary
 // index. colMap maps ColumnIDs to indices in `values`. This returns a
 // slice of IndexEntry. Forward indexes will return one value, while
-// inverted indicies can return multiple values.
+// inverted indices can return multiple values.
 func EncodeSecondaryIndex(
 	tableDesc *TableDescriptor,
 	secondaryIndex *IndexDescriptor,
