@@ -20,6 +20,8 @@ import (
 
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util"
@@ -31,18 +33,18 @@ import (
 type tableUpserterBase struct {
 	tableWriterBase
 
-	ri    sqlbase.RowInserter
+	ri    row.Inserter
 	alloc *sqlbase.DatumAlloc
 
 	// Should we collect the rows for a RETURNING clause?
 	collectRows bool
 
 	// Rows returned if collectRows is true.
-	rowsUpserted *sqlbase.RowContainer
+	rowsUpserted *rowcontainer.RowContainer
 
 	// A mapping of column IDs to the return index used to shape the resulting
 	// rows to those required by the returning clause. Only required if
-	// colllectRows is true.
+	// collectRows is true.
 	colIDToReturnIndex map[sqlbase.ColumnID]int
 
 	// Do the result rows have a different order than insert rows. Only set if
@@ -54,7 +56,11 @@ type tableUpserterBase struct {
 	resultCount int
 
 	// Contains all the rows to be inserted.
-	insertRows sqlbase.RowContainer
+	insertRows rowcontainer.RowContainer
+
+	// existingRows is used to store rows in a batch when checking for conflicts
+	// with rows earlier in the batch. Is is reused per batch.
+	existingRows *rowcontainer.RowContainer
 
 	// For allocation avoidance.
 	indexKeyPrefix []byte
@@ -70,7 +76,7 @@ func (tu *tableUpserterBase) init(txn *client.Txn, evalCtx *tree.EvalContext) er
 
 	// collectRows, set upon initialization, indicates whether or not we want rows returned from the operation.
 	if tu.collectRows {
-		tu.rowsUpserted = sqlbase.NewRowContainer(
+		tu.rowsUpserted = rowcontainer.NewRowContainer(
 			evalCtx.Mon.MakeBoundAccount(),
 			sqlbase.ColTypeInfoFromColDescs(tableDesc.Columns),
 			tu.insertRows.Len(),
@@ -102,21 +108,20 @@ func (tu *tableUpserterBase) init(txn *client.Txn, evalCtx *tree.EvalContext) er
 		evalCtx.Mon.MakeBoundAccount(), sqlbase.ColTypeInfoFromColDescs(tu.ri.InsertCols), 0,
 	)
 
-	tu.indexKeyPrefix = sqlbase.MakeIndexKeyPrefix(tableDesc, tableDesc.PrimaryIndex.ID)
+	tu.indexKeyPrefix = sqlbase.MakeIndexKeyPrefix(tableDesc.TableDesc(), tableDesc.PrimaryIndex.ID)
 
 	return nil
 }
 
-func (tu *tableUpserterBase) tableDesc() *sqlbase.TableDescriptor {
+func (tu *tableUpserterBase) tableDesc() *sqlbase.ImmutableTableDescriptor {
 	return tu.ri.Helper.TableDesc
 }
 
 // row is part of the tableWriter interface.
-func (tu *tableUpserterBase) row(
-	ctx context.Context, row tree.Datums, traceKV bool,
-) (tree.Datums, error) {
+func (tu *tableUpserterBase) row(ctx context.Context, row tree.Datums, traceKV bool) error {
 	tu.batchSize++
-	return tu.insertRows.AddRow(ctx, row)
+	_, err := tu.insertRows.AddRow(ctx, row)
+	return err
 }
 
 // flushAndStartNewBatch is part of the extendedTableWriter interface.
@@ -125,6 +130,9 @@ func (tu *tableUpserterBase) flushAndStartNewBatch(ctx context.Context) error {
 	if tu.collectRows {
 		tu.rowsUpserted.Clear(ctx)
 	}
+	if tu.existingRows != nil {
+		tu.existingRows.Clear(ctx)
+	}
 	return tu.tableWriterBase.flushAndStartNewBatch(ctx, tu.tableDesc())
 }
 
@@ -144,6 +152,9 @@ func (tu *tableUpserterBase) curBatchSize() int { return tu.insertRows.Len() }
 // close is part of the tableWriter interface.
 func (tu *tableUpserterBase) close(ctx context.Context) {
 	tu.insertRows.Close(ctx)
+	if tu.existingRows != nil {
+		tu.existingRows.Close(ctx)
+	}
 	if tu.rowsUpserted != nil {
 		tu.rowsUpserted.Close(ctx)
 	}
@@ -151,9 +162,9 @@ func (tu *tableUpserterBase) close(ctx context.Context) {
 
 // finalize is part of the tableWriter interface.
 func (tu *tableUpserterBase) finalize(
-	ctx context.Context, autoCommit autoCommitOpt, traceKV bool,
-) (*sqlbase.RowContainer, error) {
-	return nil, tu.tableWriterBase.finalize(ctx, autoCommit, tu.tableDesc())
+	ctx context.Context, traceKV bool,
+) (*rowcontainer.RowContainer, error) {
+	return nil, tu.tableWriterBase.finalize(ctx, tu.tableDesc())
 }
 
 // makeResultFromRow reshapes a row that was inserted or updated to a row
@@ -181,7 +192,7 @@ func (tu *tableUpserterBase) makeResultFromRow(
 }
 
 // fkSpanCollector is part of the tableWriter interface.
-func (tu *tableUpserterBase) fkSpanCollector() sqlbase.FkSpanCollector {
+func (tu *tableUpserterBase) fkSpanCollector() row.FkSpanCollector {
 	return tu.ri.Fks
 }
 
@@ -248,15 +259,22 @@ type tableUpserter struct {
 	// allocations.
 	updateValues tree.Datums
 
+	// cleanedRow is a temporary buffer reused from one row to the next in
+	// appendKnownConflictingRow to limit allocations.
+	cleanedRow tree.Datums
+
 	// Set by init.
-	fkTables              sqlbase.TableLookupsByID // for fk checks in update case
-	ru                    sqlbase.RowUpdater
+	fkTables              row.FkTableMetadata // for fk checks in update case
+	ru                    row.Updater
 	updateColIDtoRowIndex map[sqlbase.ColumnID]int
 	fetchCols             []sqlbase.ColumnDescriptor
 	fetchColIDtoRowIndex  map[sqlbase.ColumnID]int
-	fetcher               sqlbase.RowFetcher
+	fetcher               row.Fetcher
 }
 
+// desc is part of the tableWriter interface.
+func (*tableUpserter) desc() string { return "upserter" }
+
 // init is part of the tableWriter interface.
 func (tu *tableUpserter) init(txn *client.Txn, evalCtx *tree.EvalContext) error {
 	tu.tableWriterBase.init(txn)
@@ -274,15 +292,15 @@ func (tu *tableUpserter) init(txn *client.Txn, evalCtx *tree.EvalContext) error
 
 	if len(tu.updateCols) == 0 {
 		tu.fetchCols = requestedCols
-		tu.fetchColIDtoRowIndex = sqlbase.ColIDtoRowIndexFromCols(requestedCols)
+		tu.fetchColIDtoRowIndex = row.ColIDtoRowIndexFromCols(requestedCols)
 	} else {
-		tu.ru, err = sqlbase.MakeRowUpdater(
+		tu.ru, err = row.MakeUpdater(
 			txn,
 			tableDesc,
 			tu.fkTables,
 			tu.updateCols,
 			requestedCols,
-			sqlbase.RowUpdaterDefault,
+			row.UpdaterDefault,
 			evalCtx,
 			tu.alloc,
 		)
@@ -307,7 +325,7 @@ func (tu *tableUpserter) init(txn *client.Txn, evalCtx *tree.EvalContext) error
 		}
 	}
 
-	tableArgs := sqlbase.RowFetcherTableArgs{
+	tableArgs := row.FetcherTableArgs{
 		Desc:            tableDesc,
 		Index:           &tableDesc.PrimaryIndex,
 		ColIdxMap:       tu.fetchColIDtoRowIndex,
@@ -315,16 +333,28 @@ func (tu *tableUpserter) init(txn *client.Txn, evalCtx *tree.EvalContext) error
 		ValNeededForCol: valNeededForCol,
 	}
 
-	return tu.fetcher.Init(
+	if err := tu.fetcher.Init(
 		false /* reverse */, false /*returnRangeInfo*/, false /* isCheck */, tu.alloc, tableArgs,
+	); err != nil {
+		return err
+	}
+
+	tu.cleanedRow = make(tree.Datums, len(tu.fetchColIDtoRowIndex))
+	pkColTypeInfo, err := sqlbase.MakeColTypeInfo(tu.tableDesc(), tu.fetchColIDtoRowIndex)
+	if err != nil {
+		return err
+	}
+	tu.existingRows = rowcontainer.NewRowContainer(
+		tu.evalCtx.Mon.MakeBoundAccount(), pkColTypeInfo, tu.insertRows.Len(),
 	)
+	return nil
 }
 
 // atBatchEnd is part of the extendedTableWriter interface.
 func (tu *tableUpserter) atBatchEnd(ctx context.Context, traceKV bool) error {
 	// Fetch the information about which rows in tu.insertRows currently
 	// conflict with rows in-db.
-	existingRows, pkToRowIdx, conflictingPKs, err := tu.fetchExisting(ctx, traceKV)
+	pkToRowIdx, conflictingPKs, err := tu.fetchExisting(ctx, traceKV)
 	if err != nil {
 		return err
 	}
@@ -375,8 +405,9 @@ func (tu *tableUpserter) atBatchEnd(ctx context.Context, traceKV bool) error {
 		// Do we have a conflict?
 		if conflictingRowIdx == -1 {
 			// We don't have a conflict. This is a new row in KV. Create it.
-			resultRow, existingRows, err = tu.insertNonConflictingRow(
-				ctx, tu.b, insertRow, conflictingRowPK, existingRows, pkToRowIdx, tableDesc, traceKV)
+			resultRow, err = tu.insertNonConflictingRow(
+				ctx, tu.b, insertRow, conflictingRowPK, pkToRowIdx, tableDesc, traceKV,
+			)
 			if err != nil {
 				return err
 			}
@@ -412,7 +443,7 @@ func (tu *tableUpserter) atBatchEnd(ctx context.Context, traceKV bool) error {
 
 			// existingRow carries the values previously seen in
 			// KV or newly inserted earlier in this batch.
-			existingRow := existingRows[conflictingRowIdx]
+			existingRow := tu.existingRows.At(conflictingRowIdx)
 
 			// Check the ON CONFLICT DO UPDATE WHERE ... clause.
 			shouldUpdate, err := tu.evaler.shouldUpdate(insertRow, existingRow)
@@ -428,11 +459,11 @@ func (tu *tableUpserter) atBatchEnd(ctx context.Context, traceKV bool) error {
 			}
 
 			// We know there was a row already, and we know we need to update it. Do it.
-			resultRow, existingRows, err = tu.updateConflictingRow(
+			resultRow, err = tu.updateConflictingRow(
 				ctx, tu.b, insertRow,
 				conflictingRowPK, conflictingRowIdx, existingRow,
-				existingRows, pkToRowIdx,
-				tableDesc, traceKV)
+				pkToRowIdx, tableDesc, traceKV,
+			)
 			if err != nil {
 				return err
 			}
@@ -462,8 +493,9 @@ func (tu *tableUpserter) atBatchEnd(ctx context.Context, traceKV bool) error {
 	return nil
 }
 
-// updateConflictingRow updates the existing row
-// in the table, when there was a conflict.
+// updateConflictingRow updates the existing row in the table, when there was a
+// conflict. existingRows contains the previously seen rows, and is modified
+// or extended depending on how the PK columns are updated by the SET clauses.
 // Inputs:
 // - b is the KV batch to use for the insert.
 // - insertRow is the new row to upsert, containing the "excluded" values.
@@ -476,9 +508,6 @@ func (tu *tableUpserter) atBatchEnd(ctx context.Context, traceKV bool) error {
 //   descriptor. This may be different than the shape of insertRow if there are
 //   nullable columns. This is only returned if collectRows is true.
 // Input/Outputs:
-// - existingRows contains the previously seen rows, and is modified
-//   or extended depending on how the PK columns are updated by the SET
-//   clauses.
 // - pkToRowIdx is extended with the index of the new entry in existingRows.
 func (tu *tableUpserter) updateConflictingRow(
 	ctx context.Context,
@@ -487,23 +516,22 @@ func (tu *tableUpserter) updateConflictingRow(
 	conflictingRowPK roachpb.Key,
 	conflictingRowIdx int,
 	conflictingRowValues tree.Datums,
-	existingRows []tree.Datums,
 	pkToRowIdx map[string]int,
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.ImmutableTableDescriptor,
 	traceKV bool,
-) (resultRow tree.Datums, newExistingRows []tree.Datums, err error) {
+) (resultRow tree.Datums, err error) {
 	// First compute all the updates via SET (or the pseudo-SET generated
 	// for UPSERT statements).
 
 	updateValues, err := tu.evaler.eval(insertRow, conflictingRowValues, tu.updateValues)
 	if err != nil {
-		return nil, nil, err
+		return nil, err
 	}
 
 	checkHelper := tu.fkTables[tableDesc.ID].CheckHelper
 
 	// Do we need to (re-)compute computed columns or CHECK expressions?
-	if tu.anyComputed || len(checkHelper.Exprs) > 0 {
+	if tu.anyComputed || checkHelper != nil {
 		// For computed columns, the goal for the following code appends
 		// the computed columns at the end of updateValues.
 		// For CHECK constraints, the goal of the following code
@@ -534,10 +562,20 @@ func (tu *tableUpserter) updateConflictingRow(
 		// of updateValues.
 		updateValues, err = tu.evaler.evalComputedCols(newValues, updateValues)
 		if err != nil {
-			return nil, nil, err
+			return nil, err
 		}
 
-		if len(checkHelper.Exprs) > 0 {
+		// Ensure that all the values produced by SET comply with the schema constraints.
+		// We can assume that values *prior* to SET are OK because:
+		// - data source values have been validated by GenerateInsertRow() already.
+		// - values coming from the table are valid by construction.
+		// However the SET expression can be arbitrary and can introduce errors
+		// downstream, so we need to (re)validate here.
+		if err := enforceLocalColumnConstraints(updateValues, tu.ru.UpdateCols); err != nil {
+			return nil, err
+		}
+
+		if checkHelper != nil {
 			// If there are CHECK expressions, we must add the computed
 			// columns to the input row.
 
@@ -547,11 +585,11 @@ func (tu *tableUpserter) updateConflictingRow(
 			}
 
 			// Check CHECK constraints.
-			if err := checkHelper.LoadRow(tu.ru.FetchColIDtoRowIndex, newValues, false); err != nil {
-				return nil, nil, err
+			if err := checkHelper.LoadEvalRow(tu.ru.FetchColIDtoRowIndex, newValues, false); err != nil {
+				return nil, err
 			}
-			if err := checkHelper.Check(tu.evalCtx); err != nil {
-				return nil, nil, err
+			if err := checkHelper.CheckEval(tu.evalCtx); err != nil {
+				return nil, err
 			}
 		}
 	}
@@ -560,10 +598,10 @@ func (tu *tableUpserter) updateConflictingRow(
 	// containing the updated values for every column in the
 	// table. This is useful for RETURNING, which we collect below.
 	updatedRow, err := tu.ru.UpdateRow(
-		ctx, b, conflictingRowValues, updateValues, sqlbase.CheckFKs, traceKV,
+		ctx, b, conflictingRowValues, updateValues, row.CheckFKs, traceKV,
 	)
 	if err != nil {
-		return nil, nil, err
+		return nil, err
 	}
 
 	// Keep the slice for reuse.
@@ -574,9 +612,9 @@ func (tu *tableUpserter) updateConflictingRow(
 	// tu.evaler.ccIvarContainer.Mapping which contains the suitable
 	// mapping for the table columns already.
 	updatedConflictingRowPK, _, err := sqlbase.EncodeIndexKey(
-		tableDesc, &tableDesc.PrimaryIndex, tu.evaler.ccIvarContainer.Mapping, updatedRow, tu.indexKeyPrefix)
+		tableDesc.TableDesc(), &tableDesc.PrimaryIndex, tu.evaler.ccIvarContainer.Mapping, updatedRow, tu.indexKeyPrefix)
 	if err != nil {
-		return nil, nil, err
+		return nil, err
 	}
 
 	// It's possible that the PK for the updated values is different
@@ -594,7 +632,7 @@ func (tu *tableUpserter) updateConflictingRow(
 		//
 		// We need to update that known copy, so that subsequent
 		// iterations can find it.
-		copy(existingRows[updatedConflictingRowIdx], updatedRow)
+		copy(tu.existingRows.At(updatedConflictingRowIdx), updatedRow)
 
 		// The following line is meant to read:
 		//
@@ -609,7 +647,11 @@ func (tu *tableUpserter) updateConflictingRow(
 		pkChanged = true
 
 		// Now add the new one.
-		existingRows = tu.appendKnownConflictingRow(updatedRow, updatedConflictingRowPK, existingRows, pkToRowIdx)
+		if err := tu.appendKnownConflictingRow(
+			ctx, updatedRow, updatedConflictingRowPK, pkToRowIdx,
+		); err != nil {
+			return nil, err
+		}
 	}
 
 	if pkChanged {
@@ -619,11 +661,11 @@ func (tu *tableUpserter) updateConflictingRow(
 
 	// We only need a result row if we're collecting rows.
 	if !tu.collectRows {
-		return nil, existingRows, nil
+		return nil, nil
 	}
 
 	// We now need a row that has the shape of the result row.
-	return tu.makeResultFromRow(updatedRow, tu.evaler.ccIvarContainer.Mapping), existingRows, nil
+	return tu.makeResultFromRow(updatedRow, tu.evaler.ccIvarContainer.Mapping), nil
 }
 
 // insertNonConflictingRow inserts the source row insertRow
@@ -638,22 +680,20 @@ func (tu *tableUpserter) updateConflictingRow(
 //   descriptor. This may be different than the shape of insertRow if there are
 //   nullable columns. This is only returned if collectRows is true.
 // Input/Outputs:
-// - existingRows is extended with resultRow to produce newExistingRows.
 // - pkToRowIdx is extended with the index of the new entry in existingRows.
 func (tu *tableUpserter) insertNonConflictingRow(
 	ctx context.Context,
 	b *client.Batch,
 	insertRow tree.Datums,
 	conflictingRowPK roachpb.Key,
-	existingRows []tree.Datums,
 	pkToRowIdx map[string]int,
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.ImmutableTableDescriptor,
 	traceKV bool,
-) (resultRow tree.Datums, newExistingRows []tree.Datums, err error) {
+) (resultRow tree.Datums, err error) {
 	// Perform the insert proper.
 	if err := tu.ri.InsertRow(
-		ctx, b, insertRow, false /* ignoreConflicts */, sqlbase.CheckFKs, traceKV); err != nil {
-		return nil, nil, err
+		ctx, b, insertRow, false /* ignoreConflicts */, row.CheckFKs, traceKV); err != nil {
+		return nil, err
 	}
 
 	// We may not know the conflictingRowPK yet for the new row, for
@@ -661,44 +701,46 @@ func (tu *tableUpserter) insertNonConflictingRow(
 	// In that case, compute it now.
 	if conflictingRowPK == nil {
 		conflictingRowPK, _, err = sqlbase.EncodeIndexKey(
-			tableDesc, &tableDesc.PrimaryIndex, tu.ri.InsertColIDtoRowIndex, insertRow, tu.indexKeyPrefix)
+			tableDesc.TableDesc(), &tableDesc.PrimaryIndex, tu.ri.InsertColIDtoRowIndex, insertRow, tu.indexKeyPrefix)
 		if err != nil {
-			return nil, nil, err
+			return nil, err
 		}
 	}
 
 	// Then remember it for further upserts.
-	existingRows = tu.appendKnownConflictingRow(insertRow, conflictingRowPK, existingRows, pkToRowIdx)
+	if err := tu.appendKnownConflictingRow(ctx, insertRow, conflictingRowPK, pkToRowIdx); err != nil {
+		return nil, err
+	}
 
 	if !tu.collectRows {
-		return nil, existingRows, nil
+		return nil, nil
 	}
 
 	// Reshape the row if needed.
 	if tu.insertReorderingRequired {
-		return tu.makeResultFromRow(insertRow, tu.ri.InsertColIDtoRowIndex), existingRows, nil
+		return tu.makeResultFromRow(insertRow, tu.ri.InsertColIDtoRowIndex), nil
 	}
-	return insertRow, existingRows, nil
+	return insertRow, nil
 }
 
-// appendKnownConflictingRow adds a new row to existingRows and
-// remembers its position in pkToRowIdx.
+// appendKnownConflictingRow adds a new row to existingRows and remembers its
+// position in pkToRowIdx.
 func (tu *tableUpserter) appendKnownConflictingRow(
-	newRow tree.Datums, newRowPK roachpb.Key, existingRows []tree.Datums, pkToRowIdx map[string]int,
-) (newExistingRows []tree.Datums) {
-	pkToRowIdx[string(newRowPK)] = len(existingRows)
+	ctx context.Context, newRow tree.Datums, newRowPK roachpb.Key, pkToRowIdx map[string]int,
+) error {
+	pkToRowIdx[string(newRowPK)] = tu.existingRows.Len()
 	// We need to convert the new row to match the fetch columns required for
 	// checking if there is a conflict.
-	cleanedRow := make(tree.Datums, len(tu.fetchColIDtoRowIndex))
 	for fetchColID, fetchRowIndex := range tu.fetchColIDtoRowIndex {
 		insertRowIndex, ok := tu.ri.InsertColIDtoRowIndex[fetchColID]
 		if ok {
-			cleanedRow[fetchRowIndex] = newRow[insertRowIndex]
+			tu.cleanedRow[fetchRowIndex] = newRow[insertRowIndex]
 		} else {
-			cleanedRow[fetchRowIndex] = tree.DNull
+			tu.cleanedRow[fetchRowIndex] = tree.DNull
 		}
 	}
-	return append(existingRows, cleanedRow)
+	_, err := tu.existingRows.AddRow(ctx, tu.cleanedRow)
+	return err
 }
 
 // getConflictingRowPK returns the primary key of the row that may
@@ -715,7 +757,7 @@ func (tu *tableUpserter) getConflictingRowPK(
 	insertRow tree.Datums,
 	rowIdx int,
 	conflictingPKs map[int]roachpb.Key,
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.ImmutableTableDescriptor,
 ) (conflictingRowPK roachpb.Key, err error) {
 	if conflictingPKs != nil {
 		// If a secondary index helped us find the conflicting PK for this
@@ -727,7 +769,7 @@ func (tu *tableUpserter) getConflictingRowPK(
 
 	// Otherwise, encode the values to determine the primary key.
 	insertRowPK, _, err := sqlbase.EncodeIndexKey(
-		tableDesc, &tableDesc.PrimaryIndex, tu.ri.InsertColIDtoRowIndex, insertRow, tu.indexKeyPrefix)
+		tableDesc.TableDesc(), &tableDesc.PrimaryIndex, tu.ri.InsertColIDtoRowIndex, insertRow, tu.indexKeyPrefix)
 	return insertRowPK, err
 }
 
@@ -764,7 +806,7 @@ func (tu *tableUpserter) upsertRowPKSpans(
 
 			// Compute the PK span for the current row.
 			upsertRowPKSpan, _, err := sqlbase.EncodeIndexSpan(
-				tableDesc, &tu.conflictIndex, tu.ri.InsertColIDtoRowIndex, insertRow, tu.indexKeyPrefix)
+				tableDesc.TableDesc(), &tu.conflictIndex, tu.ri.InsertColIDtoRowIndex, insertRow, tu.indexKeyPrefix)
 			if err != nil {
 				return nil, nil, err
 			}
@@ -790,7 +832,7 @@ func (tu *tableUpserter) upsertRowPKSpans(
 	for i := 0; i < tu.insertRows.Len(); i++ {
 		insertRow := tu.insertRows.At(i)
 		entries, err := sqlbase.EncodeSecondaryIndex(
-			tableDesc, &tu.conflictIndex, tu.ri.InsertColIDtoRowIndex, insertRow)
+			tableDesc.TableDesc(), &tu.conflictIndex, tu.ri.InsertColIDtoRowIndex, insertRow)
 		if err != nil {
 			return nil, nil, err
 		}
@@ -810,7 +852,7 @@ func (tu *tableUpserter) upsertRowPKSpans(
 	for i, result := range b.Results {
 		if len(result.Rows) == 1 {
 			if result.Rows[0].Value != nil {
-				upsertRowPK, err := sqlbase.ExtractIndexKey(tu.alloc, tableDesc, result.Rows[0])
+				upsertRowPK, err := sqlbase.ExtractIndexKey(tu.alloc, tableDesc.TableDesc(), result.Rows[0])
 				if err != nil {
 					return nil, nil, err
 				}
@@ -836,7 +878,6 @@ func (tu *tableUpserter) upsertRowPKSpans(
 // fetchExisting returns any existing rows in the table that conflict with the
 // ones in tu.insertRows.
 // Outputs:
-// - existingRows contains data for conflicting rows.
 // - pkToRowIdx relates the primary key values in the
 //   data source to which entry in the returned slice contain data
 //   for that primary key.
@@ -845,18 +886,13 @@ func (tu *tableUpserter) upsertRowPKSpans(
 //   conflicts found and the conflict index was a secondary index.
 func (tu *tableUpserter) fetchExisting(
 	ctx context.Context, traceKV bool,
-) (
-	existingRows []tree.Datums,
-	pkToRowIdx map[string]int,
-	conflictingPKs map[int]roachpb.Key,
-	err error,
-) {
+) (pkToRowIdx map[string]int, conflictingPKs map[int]roachpb.Key, err error) {
 	tableDesc := tu.tableDesc()
 
 	// primaryKeySpans contains the PK values to check for conflicts.
 	primaryKeySpans, conflictingPKs, err := tu.upsertRowPKSpans(ctx, traceKV)
 	if err != nil {
-		return nil, nil, nil, err
+		return nil, nil, err
 	}
 
 	// pkToRowIdx maps the PK values to positions in existingRows.
@@ -864,7 +900,7 @@ func (tu *tableUpserter) fetchExisting(
 
 	if len(primaryKeySpans) == 0 {
 		// We know already there is no conflicting row, so there's nothing to fetch.
-		return existingRows, pkToRowIdx, conflictingPKs, nil
+		return pkToRowIdx, conflictingPKs, nil
 	}
 
 	// pkSpans will contain the spans for every entry in primaryKeySpans.
@@ -876,40 +912,36 @@ func (tu *tableUpserter) fetchExisting(
 	// Start retrieving the PKs.
 	// We don't limit batches here because the spans are unordered.
 	if err := tu.fetcher.StartScan(ctx, tu.txn, pkSpans, false /* no batch limits */, 0, traceKV); err != nil {
-		return nil, nil, nil, err
+		return nil, nil, err
 	}
 
 	// Populate existingRows and pkToRowIdx.
 	for {
 		row, _, _, err := tu.fetcher.NextRowDecoded(ctx)
 		if err != nil {
-			return nil, nil, nil, err
+			return nil, nil, err
 		}
 		if row == nil {
 			break // Done
 		}
 
 		rowPrimaryKey, _, err := sqlbase.EncodeIndexKey(
-			tableDesc, &tableDesc.PrimaryIndex, tu.fetchColIDtoRowIndex, row, tu.indexKeyPrefix)
+			tableDesc.TableDesc(), &tableDesc.PrimaryIndex, tu.fetchColIDtoRowIndex, row, tu.indexKeyPrefix)
 		if err != nil {
-			return nil, nil, nil, err
+			return nil, nil, err
 		}
-		// The rows returned by rowFetcher are invalidated after the call to
-		// NextRow, so we have to copy them to save them.
-		// TODO(knz/nathan): try to reuse a large slice instead
-		// of making many small slices.
-		rowCopy := make(tree.Datums, len(row))
-		copy(rowCopy, row)
 
-		pkToRowIdx[string(rowPrimaryKey)] = len(existingRows)
-		existingRows = append(existingRows, rowCopy)
+		pkToRowIdx[string(rowPrimaryKey)] = tu.existingRows.Len()
+		if _, err := tu.existingRows.AddRow(ctx, row); err != nil {
+			return nil, nil, err
+		}
 	}
 
-	return existingRows, pkToRowIdx, conflictingPKs, nil
+	return pkToRowIdx, conflictingPKs, nil
 }
 
 // tableDesc is part of the tableWriter interface.
-func (tu *tableUpserter) tableDesc() *sqlbase.TableDescriptor {
+func (tu *tableUpserter) tableDesc() *sqlbase.ImmutableTableDescriptor {
 	return tu.ri.Helper.TableDesc
 }
 
