@@ -17,7 +17,7 @@ package tests
 import (
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/storage"
-	"github.com/cockroachdb/cockroach/pkg/storage/batcheval"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
 )
 
 // CreateTestServerParams creates a set of params suitable for SQL tests.
@@ -30,7 +30,7 @@ func CreateTestServerParams() (base.TestServerArgs, *CommandFilters) {
 	cmdFilters.AppendFilter(CheckEndTransactionTrigger, true)
 	params := base.TestServerArgs{}
 	params.Knobs.Store = &storage.StoreTestingKnobs{
-		EvalKnobs: batcheval.TestingKnobs{
+		EvalKnobs: storagebase.BatchEvalTestingKnobs{
 			TestingEvalFilter: cmdFilters.RunFilters,
 		},
 	}
