@@ -21,15 +21,13 @@ import (
 	"testing"
 	"time"
 
-	opentracing "github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
+	// We dot-import fsm to use common names such as fsm.True/False.
+	. "github.com/cockroachdb/cockroach/pkg/util/fsm"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
@@ -37,9 +35,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/mon"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
-
-	// We dot-import fsm to use common names such as fsm.True/False.
-	. "github.com/cockroachdb/cockroach/pkg/util/fsm"
+	"github.com/gogo/protobuf/proto"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
 )
 
 var noRewindExpected = CmdPos(-1)
@@ -121,7 +119,6 @@ func (tc *testContext) createOpenState(
 		sp:            sp,
 		cancel:        cancel,
 		sqlTimestamp:  timeutil.Now(),
-		isolation:     enginepb.SERIALIZABLE,
 		priority:      roachpb.NormalUserPriority,
 		mon:           &txnStateMon,
 		txnAbortCount: metric.NewCounter(MetaTxnAbort),
@@ -202,7 +199,6 @@ func checkAdv(adv advanceInfo, expCode advanceCode, expRewPos CmdPos, expEv txnE
 // correspond to expectations. Any field left nil will not be checked.
 type expKVTxn struct {
 	debugName    *string
-	isolation    *enginepb.IsolationType
 	userPriority *roachpb.UserPriority
 	// For the timestamps we just check the physical part. The logical part is
 	// incremented every time the clock is read and so it's unpredictable.
@@ -219,10 +215,6 @@ func checkTxn(txn *client.Txn, exp expKVTxn) error {
 		return errors.Errorf("expected DebugName: %s, but got: %s",
 			*exp.debugName, txn.DebugName())
 	}
-	if exp.isolation != nil && *exp.isolation != txn.Isolation() {
-		return errors.Errorf("expected isolation: %s, but got: %s",
-			*exp.isolation, txn.Isolation())
-	}
 	if exp.userPriority != nil && *exp.userPriority != txn.UserPriority() {
 		return errors.Errorf("expected UserPriority: %s, but got: %s",
 			*exp.userPriority, txn.UserPriority())
@@ -267,7 +259,6 @@ func TestTransitions(t *testing.T) {
 
 	txnName := sqlTxnName
 	now := testCon.clock.Now()
-	iso := enginepb.SERIALIZABLE
 	pri := roachpb.NormalUserPriority
 	maxTS := testCon.clock.Now().Add(testCon.clock.MaxOffset().Nanoseconds(), 0 /* logical */)
 	type test struct {
@@ -307,9 +298,10 @@ func TestTransitions(t *testing.T) {
 				s, ts := testCon.createNoTxnState()
 				return s, ts, nil
 			},
-			ev:        eventTxnStart{ImplicitTxn: True},
-			evPayload: makeEventTxnStartPayload(iso, pri, tree.ReadWrite, timeutil.Now(), tranCtx),
-			expState:  stateOpen{ImplicitTxn: True, RetryIntent: False},
+			ev: eventTxnStart{ImplicitTxn: True},
+			evPayload: makeEventTxnStartPayload(pri, tree.ReadWrite, timeutil.Now(),
+				nil /* historicalTimestamp */, tranCtx),
+			expState: stateOpen{ImplicitTxn: True, RetryIntent: False},
 			expAdv: expAdvance{
 				// We expect to stayInPlace; upon starting a txn the statement is
 				// executed again, this time in state Open.
@@ -318,7 +310,6 @@ func TestTransitions(t *testing.T) {
 			},
 			expTxn: &expKVTxn{
 				debugName:    &txnName,
-				isolation:    &iso,
 				userPriority: &pri,
 				tsNanos:      &now.WallTime,
 				origTSNanos:  &now.WallTime,
@@ -332,16 +323,16 @@ func TestTransitions(t *testing.T) {
 				s, ts := testCon.createNoTxnState()
 				return s, ts, nil
 			},
-			ev:        eventTxnStart{ImplicitTxn: False},
-			evPayload: makeEventTxnStartPayload(iso, pri, tree.ReadWrite, timeutil.Now(), tranCtx),
-			expState:  stateOpen{ImplicitTxn: False, RetryIntent: False},
+			ev: eventTxnStart{ImplicitTxn: False},
+			evPayload: makeEventTxnStartPayload(pri, tree.ReadWrite, timeutil.Now(),
+				nil /* historicalTimestamp */, tranCtx),
+			expState: stateOpen{ImplicitTxn: False, RetryIntent: False},
 			expAdv: expAdvance{
 				expCode: advanceOne,
 				expEv:   txnStart,
 			},
 			expTxn: &expKVTxn{
 				debugName:    &txnName,
-				isolation:    &iso,
 				userPriority: &pri,
 				tsNanos:      &now.WallTime,
 				origTSNanos:  &now.WallTime,
@@ -416,7 +407,7 @@ func TestTransitions(t *testing.T) {
 			},
 			evFun: func(ts *txnState) (Event, EventPayload) {
 				b := eventRetriableErrPayload{
-					err: roachpb.NewHandledRetryableTxnError(
+					err: roachpb.NewTransactionRetryWithProtoRefreshError(
 						"test retriable err",
 						ts.mu.txn.ID(),
 						*ts.mu.txn.Serialize()),
@@ -443,7 +434,7 @@ func TestTransitions(t *testing.T) {
 			},
 			evFun: func(ts *txnState) (Event, EventPayload) {
 				b := eventRetriableErrPayload{
-					err: roachpb.NewHandledRetryableTxnError(
+					err: roachpb.NewTransactionRetryWithProtoRefreshError(
 						"test retriable err",
 						ts.mu.txn.ID(),
 						*ts.mu.txn.Serialize()),
@@ -469,7 +460,7 @@ func TestTransitions(t *testing.T) {
 			},
 			evFun: func(ts *txnState) (Event, EventPayload) {
 				b := eventRetriableErrPayload{
-					err: roachpb.NewHandledRetryableTxnError(
+					err: roachpb.NewTransactionRetryWithProtoRefreshError(
 						"test retriable err",
 						ts.mu.txn.ID(),
 						*ts.mu.txn.Serialize()),
@@ -499,7 +490,7 @@ func TestTransitions(t *testing.T) {
 			},
 			evFun: func(ts *txnState) (Event, EventPayload) {
 				b := eventRetriableErrPayload{
-					err: roachpb.NewHandledRetryableTxnError(
+					err: roachpb.NewTransactionRetryWithProtoRefreshError(
 						"test retriable err",
 						ts.mu.txn.ID(),
 						*ts.mu.txn.Serialize()),
@@ -542,7 +533,7 @@ func TestTransitions(t *testing.T) {
 			},
 			evFun: func(ts *txnState) (Event, EventPayload) {
 				b := eventRetriableErrPayload{
-					err: roachpb.NewHandledRetryableTxnError(
+					err: roachpb.NewTransactionRetryWithProtoRefreshError(
 						"test retriable err",
 						ts.mu.txn.ID(),
 						*ts.mu.txn.Serialize()),
@@ -567,7 +558,7 @@ func TestTransitions(t *testing.T) {
 			},
 			evFun: func(ts *txnState) (Event, EventPayload) {
 				b := eventRetriableErrPayload{
-					err: roachpb.NewHandledRetryableTxnError(
+					err: roachpb.NewTransactionRetryWithProtoRefreshError(
 						"test retriable err",
 						ts.mu.txn.ID(),
 						*ts.mu.txn.Serialize()),
@@ -659,15 +650,37 @@ func TestTransitions(t *testing.T) {
 				s, ts := testCon.createAbortedState(retryIntentSet)
 				return s, ts, nil
 			},
-			ev:        eventTxnStart{ImplicitTxn: False},
-			evPayload: makeEventTxnStartPayload(iso, pri, tree.ReadWrite, timeutil.Now(), tranCtx),
-			expState:  stateOpen{ImplicitTxn: False, RetryIntent: True},
+			ev: eventTxnStart{ImplicitTxn: False},
+			evPayload: makeEventTxnStartPayload(pri, tree.ReadWrite, timeutil.Now(),
+				nil /* historicalTimestamp */, tranCtx),
+			expState: stateOpen{ImplicitTxn: False, RetryIntent: True},
 			expAdv: expAdvance{
 				expCode: advanceOne,
 				expEv:   noEvent,
 			},
 			expTxn: &expKVTxn{},
 		},
+		{
+			// The txn is starting again (e.g. ROLLBACK TO SAVEPOINT while in Aborted).
+			// Verify that the historical timestamp from the evPayload is propagated
+			// to the expTxn.
+			name: "Aborted->Starting (historical)",
+			init: func() (State, *txnState, error) {
+				s, ts := testCon.createAbortedState(retryIntentSet)
+				return s, ts, nil
+			},
+			ev: eventTxnStart{ImplicitTxn: False},
+			evPayload: makeEventTxnStartPayload(pri, tree.ReadOnly, now.GoTime(),
+				&now, tranCtx),
+			expState: stateOpen{ImplicitTxn: False, RetryIntent: True},
+			expAdv: expAdvance{
+				expCode: advanceOne,
+				expEv:   noEvent,
+			},
+			expTxn: &expKVTxn{
+				tsNanos: proto.Int64(now.WallTime),
+			},
+		},
 		//
 		// Tests starting from the RestartWait state.
 		//
