@@ -32,6 +32,10 @@ type unaryRun struct {
 	consumed bool
 }
 
+func (*unaryNode) startExec(runParams) error {
+	return nil
+}
+
 func (*unaryNode) Values() tree.Datums { return nil }
 
 func (u *unaryNode) Next(runParams) (bool, error) {
