@@ -21,9 +21,12 @@ import (
 
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/privilege"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
+	"github.com/cockroachdb/cockroach/pkg/sql/rowcontainer"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
+	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/pkg/errors"
 )
@@ -97,59 +100,29 @@ func (p *planner) Update(
 		return nil, err
 	}
 
+	// Analyze any check constraints.
+	checkHelper, err := sqlbase.NewEvalCheckHelper(ctx, p.analyzeExpr, desc)
+	if err != nil {
+		return nil, err
+	}
+
 	// Determine what are the foreign key tables that are involved in the update.
-	fkTables, err := sqlbase.TablesNeededForFKs(
+	fkTables, err := row.MakeFkMetadata(
 		ctx,
-		*desc,
-		sqlbase.CheckUpdates,
-		p.lookupFKTable,
+		desc,
+		row.CheckUpdates,
+		p.LookupTableByID,
 		p.CheckPrivilege,
 		p.analyzeExpr,
+		checkHelper,
 	)
 	if err != nil {
 		return nil, err
 	}
 
-	// Pre-perform early subquery analysis and extraction. Usually
-	// this is done by analyzeExpr(), and, in fact, analyzeExpr() is indirectly
-	// called below as well. Why not call analyzeExpr() already?
-	//
-	// The obstacle is that there's a circular dependency here.
-	//
-	// - We can't call analyzeExpr() at this point because the RHS of
-	//   UPDATE SET can contain the special expression "DEFAULT", and
-	//   analyzeExpr() does not work with DEFAULT.
-	//
-	// - The substitution of DEFAULT by the actual default expression
-	//   occurs below (in addOrMergeExpr), but we can't do that yet here
-	//   because we first need to decompose a tuple in the LHS into
-	//   multiple assignments.
-	//
-	// - We can't decompose the tuple in the LHS without validating it
-	//   against the arity of the RHS (to properly reject mismatched
-	//   arities with an error) until subquery analysis has occurred.
-	//
-	// So we need the subquery analysis to be done early and we can't
-	// call analyzeExpr() to do so.
-	//
-	// TODO(knz): arguably we could do the tuple decomposition _before_
-	// the arity check, in this order: decompose tuples, substitute
-	// DEFAULT / run addOrMergeExpr which itself calls analyzeExpr, and
-	// then check the arity. This improvement is left as an exercise for
-	// the reader.
-	setExprs := make([]*tree.UpdateExpr, len(n.Exprs))
-	for i, expr := range n.Exprs {
-		// Analyze the sub-query nodes.
-		err := p.analyzeSubqueries(ctx, expr.Expr, len(expr.Names))
-		if err != nil {
-			return nil, err
-		}
-		setExprs[i] = &tree.UpdateExpr{Tuple: expr.Tuple, Expr: expr.Expr, Names: expr.Names}
-	}
-
 	// Extract all the LHS column names, and verify that the arity of
 	// the LHS and RHS match when assigning tuples.
-	names, err := p.namesForExprs(setExprs)
+	names, setExprs, err := p.namesForExprs(ctx, n.Exprs)
 	if err != nil {
 		return nil, err
 	}
@@ -197,25 +170,55 @@ func (p *planner) Update(
 	rowsNeeded := resultsNeeded(n.Returning)
 
 	var requestedCols []sqlbase.ColumnDescriptor
-	if rowsNeeded || len(desc.Checks) > 0 {
+	if rowsNeeded {
 		// TODO(dan): This could be made tighter, just the rows needed for RETURNING
 		// exprs.
-		// TODO(nvanbenschoten): This could be made tighter, just the rows needed for
-		// the CHECK exprs.
 		requestedCols = desc.Columns
+	} else if len(desc.ActiveChecks()) > 0 {
+		// Request any columns we'll need when validating check constraints. We
+		// could be smarter and only validate check constraints which depend on
+		// columns that are being modified in the UPDATE statement, in which
+		// case we'd only need to request the columns used by that subset of
+		// check constraints, but that doesn't seem worth the effort.
+		//
+		// TODO(nvanbenschoten): These conditions shouldn't be mutually
+		// exclusive, but since rowsNeeded implies that requestedCols =
+		// desc.Columns, there's no reason to enter this block if rowsNeeded is
+		// true. Remove this when the TODO above is addressed.
+		var requestedColSet util.FastIntSet
+		for _, col := range requestedCols {
+			requestedColSet.Add(int(col.ID))
+		}
+		for _, ck := range desc.ActiveChecks() {
+			cols, err := ck.ColumnsUsed(desc.TableDesc())
+			if err != nil {
+				return nil, err
+			}
+			for _, colID := range cols {
+				if !requestedColSet.Contains(int(colID)) {
+					col, err := desc.FindColumnByID(colID)
+					if err != nil {
+						return nil, errors.Wrapf(err, "error finding column %d in table %s",
+							colID, desc.Name)
+					}
+					requestedCols = append(requestedCols, *col)
+					requestedColSet.Add(int(colID))
+				}
+			}
+		}
 	}
 
 	// Create the table updater, which does the bulk of the work.
-	// As a result of MakeRowUpdater, ru.FetchCols include all the
+	// As a result of MakeUpdater, ru.FetchCols include all the
 	// columns in the table descriptor + any columns currently in the
 	// process of being added.
-	ru, err := sqlbase.MakeRowUpdater(
+	ru, err := row.MakeUpdater(
 		p.txn,
 		desc,
 		fkTables,
 		updateCols,
 		requestedCols,
-		sqlbase.RowUpdaterDefault,
+		row.UpdaterDefault,
 		p.EvalContext(),
 		&p.alloc,
 	)
@@ -280,7 +283,7 @@ func (p *planner) Update(
 		// visible.
 		// We do not want these to be available for RETURNING below.
 		//
-		// MakeRowUpdater guarantees that the first columns of the source
+		// MakeUpdater guarantees that the first columns of the source
 		// are those specified in requestedCols, which, in the case where
 		// rowsNeeded is true, is also desc.Columns. So we can truncate to
 		// the length of that to only see public columns.
@@ -315,6 +318,7 @@ func (p *planner) Update(
 
 					currentUpdateIdx++
 				}
+
 			case *tree.Subquery:
 				selectExpr := tree.SelectExpr{Expr: t}
 				desiredTupleType := types.TTuple{Types: make([]types.T, len(setExpr.Names))}
@@ -329,13 +333,19 @@ func (p *planner) Update(
 
 				colIdx := render.addOrReuseRender(col, expr, false)
 
-				sourceSlots = append(sourceSlots, tupleSlot{
+				tSlot := tupleSlot{
 					columns:     updateCols[currentUpdateIdx : currentUpdateIdx+len(setExpr.Names)],
+					emptySlot:   make(tree.Datums, len(setExpr.Names)),
 					sourceIndex: colIdx,
-				})
+				}
+				for i := range tSlot.emptySlot {
+					tSlot.emptySlot[i] = tree.DNull
+				}
+				sourceSlots = append(sourceSlots, tSlot)
 				currentUpdateIdx += len(setExpr.Names)
+
 			default:
-				panic(fmt.Sprintf("assigning to tuple with expression that is neither a tuple nor a subquery: %s", setExpr.Expr))
+				return nil, pgerror.NewAssertionErrorf("assigning to tuple with expression that is neither a tuple nor a subquery: %s", setExpr.Expr)
 			}
 
 		} else {
@@ -376,7 +386,7 @@ func (p *planner) Update(
 		columns: columns,
 		run: updateRun{
 			tu:           tableUpdater{ru: ru},
-			checkHelper:  fkTables[desc.ID].CheckHelper,
+			checkHelper:  checkHelper,
 			rowsNeeded:   rowsNeeded,
 			computedCols: computedCols,
 			computeExprs: computeExprs,
@@ -414,11 +424,7 @@ type updateRun struct {
 	done bool
 
 	// rows contains the accumulated result rows if rowsNeeded is set.
-	rows *sqlbase.RowContainer
-
-	// autoCommit indicates whether the last KV batch processed by
-	// this update will also commit the KV txn.
-	autoCommit autoCommitOpt
+	rows *rowcontainer.RowContainer
 
 	// traceKV caches the current KV tracing flag.
 	traceKV bool
@@ -489,7 +495,7 @@ func (u *updateNode) startExec(params runParams) error {
 	u.run.traceKV = params.p.ExtendedEvalContext().Tracing.KVTracingEnabled()
 
 	if u.run.rowsNeeded {
-		u.run.rows = sqlbase.NewRowContainer(
+		u.run.rows = rowcontainer.NewRowContainer(
 			params.EvalContext().Mon.MakeBoundAccount(),
 			sqlbase.ColTypeInfoFromResCols(u.columns), 0)
 	}
@@ -564,13 +570,19 @@ func (u *updateNode) BatchedNext(params runParams) (bool, error) {
 	}
 
 	if lastBatch {
-		if _, err := u.run.tu.finalize(params.ctx, u.run.autoCommit, u.run.traceKV); err != nil {
+		if _, err := u.run.tu.finalize(params.ctx, u.run.traceKV); err != nil {
 			return false, err
 		}
 		// Remember we're done for the next call to BatchedNext().
 		u.run.done = true
 	}
 
+	// Possibly initiate a run of CREATE STATISTICS.
+	params.ExecCfg().StatsRefresher.NotifyMutation(
+		u.run.tu.tableDesc().ID,
+		u.run.rowCount,
+	)
+
 	return u.run.rowCount > 0, nil
 }
 
@@ -642,35 +654,33 @@ func (u *updateNode) processSourceRow(params runParams, sourceVals tree.Datums)
 	// Verify the schema constraints. For consistency with INSERT/UPSERT
 	// and compatibility with PostgreSQL, we must do this before
 	// processing the CHECK constraints.
-	for i, val := range u.run.updateValues {
-		col := &u.run.tu.ru.UpdateCols[i]
-		if val == tree.DNull {
-			// Verify no NULL makes it to a nullable column.
-			if !col.Nullable {
-				return sqlbase.NewNonNullViolationError(col.Name)
-			}
-		} else {
-			// Verify that the data width matches the column constraint.
-			if err := sqlbase.CheckValueWidth(col.Type, val, &col.Name); err != nil {
-				return err
-			}
-		}
+	if err := enforceLocalColumnConstraints(u.run.updateValues, u.run.tu.ru.UpdateCols); err != nil {
+		return err
 	}
 
-	// Run the CHECK constraints, if any.
-	// TODO(justin): we have actually constructed the whole row at this point and
-	// thus should be able to avoid loading it separately like this now.
-	if len(u.run.checkHelper.Exprs) > 0 {
-		if err := u.run.checkHelper.LoadRow(
-			u.run.tu.ru.FetchColIDtoRowIndex, oldValues, false); err != nil {
-			return err
-		}
-		if err := u.run.checkHelper.LoadRow(
-			u.run.updateColsIdx, u.run.updateValues, true); err != nil {
-			return err
-		}
-		if err := u.run.checkHelper.Check(params.EvalContext()); err != nil {
-			return err
+	// Run the CHECK constraints, if any. CheckHelper will either evaluate the
+	// constraints itself, or else inspect boolean columns from the input that
+	// contain the results of evaluation.
+	if u.run.checkHelper != nil {
+		if u.run.checkHelper.NeedsEval() {
+			// TODO(justin): we have actually constructed the whole row at this point and
+			// thus should be able to avoid loading it separately like this now.
+			if err := u.run.checkHelper.LoadEvalRow(
+				u.run.tu.ru.FetchColIDtoRowIndex, oldValues, false); err != nil {
+				return err
+			}
+			if err := u.run.checkHelper.LoadEvalRow(
+				u.run.updateColsIdx, u.run.updateValues, true); err != nil {
+				return err
+			}
+			if err := u.run.checkHelper.CheckEval(params.EvalContext()); err != nil {
+				return err
+			}
+		} else {
+			checkVals := sourceVals[len(u.run.tu.ru.FetchCols)+len(u.run.tu.ru.UpdateCols):]
+			if err := u.run.checkHelper.CheckInput(checkVals); err != nil {
+				return err
+			}
 		}
 	}
 
@@ -717,7 +727,7 @@ func (u *updateNode) Close(ctx context.Context) {
 
 // enableAutoCommit implements the autoCommitNode interface.
 func (u *updateNode) enableAutoCommit() {
-	u.run.autoCommit = autoCommitEnabled
+	u.run.tu.enableAutoCommit()
 }
 
 // sourceSlot abstracts the idea that our update sources can either be tuples
@@ -740,9 +750,15 @@ type sourceSlot interface {
 type tupleSlot struct {
 	columns     []sqlbase.ColumnDescriptor
 	sourceIndex int
+	// emptySlot is to be returned when the source subquery
+	// returns no rows.
+	emptySlot tree.Datums
 }
 
 func (ts tupleSlot) extractValues(row tree.Datums) tree.Datums {
+	if row[ts.sourceIndex] == tree.DNull {
+		return ts.emptySlot
+	}
 	return row[ts.sourceIndex].(*tree.DTuple).D
 }
 
@@ -794,13 +810,52 @@ func (p *planner) addOrMergeExpr(
 	return render.addOrReuseRender(col, expr, true), nil
 }
 
-// namesForExprs collects all the names mentioned in the LHS of the
-// UpdateExprs.  That is, it will transform SET (a,b) = (1,2), b = 3,
-// (a,c) = 4 into [a,b,b,a,c].
+// namesForExprs both preperforms early subquery analysis and extraction and
+// then collects all the names mentioned in the LHS of the UpdateExprs.  That
+// is, it will transform SET (a,b) = (1,2), b = 3, (a,c) = 4 into [a,b,b,a,c].
 //
 // It also checks that the arity of the LHS and RHS match when
 // assigning tuples.
-func (p *planner) namesForExprs(exprs tree.UpdateExprs) (tree.NameList, error) {
+func (p *planner) namesForExprs(
+	ctx context.Context, exprs tree.UpdateExprs,
+) (tree.NameList, tree.UpdateExprs, error) {
+	// Pre-perform early subquery analysis and extraction. Usually
+	// this is done by analyzeExpr(), and, in fact, analyzeExpr() is indirectly
+	// called below as well. Why not call analyzeExpr() already?
+	//
+	// The obstacle is that there's a circular dependency here.
+	//
+	// - We can't call analyzeExpr() at this point because the RHS of
+	//   UPDATE SET can contain the special expression "DEFAULT", and
+	//   analyzeExpr() does not work with DEFAULT.
+	//
+	// - The substitution of DEFAULT by the actual default expression
+	//   occurs below (in addOrMergeExpr), but we can't do that yet here
+	//   because we first need to decompose a tuple in the LHS into
+	//   multiple assignments.
+	//
+	// - We can't decompose the tuple in the LHS without validating it
+	//   against the arity of the RHS (to properly reject mismatched
+	//   arities with an error) until subquery analysis has occurred.
+	//
+	// So we need the subquery analysis to be done early and we can't
+	// call analyzeExpr() to do so.
+	//
+	// TODO(knz): arguably we could do the tuple decomposition _before_
+	// the arity check, in this order: decompose tuples, substitute
+	// DEFAULT / run addOrMergeExpr which itself calls analyzeExpr, and
+	// then check the arity. This improvement is left as an exercise for
+	// the reader.
+	setExprs := make(tree.UpdateExprs, len(exprs))
+	for i, expr := range exprs {
+		// Analyze the sub-query nodes.
+		err := p.analyzeSubqueries(ctx, expr.Expr, len(expr.Names))
+		if err != nil {
+			return nil, nil, err
+		}
+		setExprs[i] = &tree.UpdateExpr{Tuple: expr.Tuple, Expr: expr.Expr, Names: expr.Names}
+	}
+
 	var names tree.NameList
 	for _, expr := range exprs {
 		if expr.Tuple {
@@ -816,16 +871,18 @@ func (p *planner) namesForExprs(exprs tree.UpdateExprs) (tree.NameList, error) {
 				n = len(t.D)
 			}
 			if n < 0 {
-				return nil, errors.Errorf("unsupported tuple assignment: %T", expr.Expr)
+				return nil, nil, pgerror.UnimplementedWithIssueDetailErrorf(35713,
+					fmt.Sprintf("%T", expr.Expr),
+					"source for a multiple-column UPDATE item must be a sub-SELECT or ROW() expression; not supported: %T", expr.Expr)
 			}
 			if len(expr.Names) != n {
-				return nil, fmt.Errorf("number of columns (%d) does not match number of values (%d)",
-					len(expr.Names), n)
+				return nil, nil, pgerror.NewErrorf(pgerror.CodeSyntaxError,
+					"number of columns (%d) does not match number of values (%d)", len(expr.Names), n)
 			}
 		}
 		names = append(names, expr.Names...)
 	}
-	return names, nil
+	return names, setExprs, nil
 }
 
 func fillDefault(expr tree.Expr, index int, defaultExprs []tree.TypedExpr) tree.Expr {
@@ -842,8 +899,35 @@ func fillDefault(expr tree.Expr, index int, defaultExprs []tree.TypedExpr) tree.
 func checkHasNoComputedCols(cols []sqlbase.ColumnDescriptor) error {
 	for i := range cols {
 		if cols[i].IsComputed() {
-			return sqlbase.CannotWriteToComputedColError(&cols[i])
+			return sqlbase.CannotWriteToComputedColError(cols[i].Name)
 		}
 	}
 	return nil
 }
+
+// enforceLocalColumnConstraints asserts the column constraints that
+// do not require data validation from other sources than the row data
+// itself. This includes:
+// - rejecting null values in non-nullable columns;
+// - checking width constraints from the column type;
+// - truncating results to the requested precision (not width).
+// Note: the second point is what distinguishes this operation
+// from a regular SQL cast -- here widths are checked, not
+// used to truncate the value silently.
+//
+// The row buffer is modified in-place with the result of the
+// checks.
+func enforceLocalColumnConstraints(row tree.Datums, cols []sqlbase.ColumnDescriptor) error {
+	for i := range cols {
+		col := &cols[i]
+		if !col.Nullable && row[i] == tree.DNull {
+			return sqlbase.NewNonNullViolationError(col.Name)
+		}
+		outVal, err := sqlbase.LimitValueWidth(col.Type, row[i], &col.Name)
+		if err != nil {
+			return err
+		}
+		row[i] = outVal
+	}
+	return nil
+}
