@@ -16,9 +16,10 @@ package sql
 
 import (
 	"context"
-	"fmt"
 	"sync"
 
+	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
+	"github.com/cockroachdb/cockroach/pkg/sql/row"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/types"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
@@ -49,8 +50,8 @@ var _ autoCommitNode = &upsertNode{}
 func (p *planner) newUpsertNode(
 	ctx context.Context,
 	n *tree.Insert,
-	desc *sqlbase.TableDescriptor,
-	ri sqlbase.RowInserter,
+	desc *sqlbase.ImmutableTableDescriptor,
+	ri row.Inserter,
 	tn, alias *tree.TableName,
 	sourceRows planNode,
 	needRows bool,
@@ -58,7 +59,7 @@ func (p *planner) newUpsertNode(
 	defaultExprs []tree.TypedExpr,
 	computeExprs []tree.TypedExpr,
 	computedCols []sqlbase.ColumnDescriptor,
-	fkTables sqlbase.TableLookupsByID,
+	fkTables row.FkTableMetadata,
 	desiredTypes []types.T,
 ) (res batchedPlanNode, err error) {
 	// Extract the index that will detect upsert conflicts
@@ -125,10 +126,14 @@ func (p *planner) newUpsertNode(
 		// Determine which columns are updated by the RHS of INSERT
 		// ... ON CONFLICT DO UPDATE, or the non-PK columns in an
 		// UPSERT.
-		names, err := p.namesForExprs(updateExprs)
+		names, newUpdateExprs, err := p.namesForExprs(ctx, updateExprs)
 		if err != nil {
 			return nil, err
 		}
+		// namesForExprs is responsible for re-shaping the SET RHS, we
+		// need to use its result as new input for newUpsertHelper()
+		// below.
+		updateExprs = newUpdateExprs
 
 		// We use ensureColumns = false in processColumns, because
 		// updateCols may be legitimately empty (when there is no DO
@@ -188,7 +193,8 @@ func (p *planner) newUpsertNode(
 			// can't specify it, which means we need to do a lookup and so we can't use
 			// the fast path. When adding or removing an index, same result, so the fast
 			// path is disabled during all mutations.
-			len(desc.Mutations) == 0 &&
+			len(desc.MutationColumns()) == 0 &&
+			len(desc.MutationIndexes()) == 0 &&
 			// For the fast path, all columns must be specified in the insert.
 			len(ri.InsertCols) == len(desc.Columns) &&
 			// We cannot use the fast path if we also have a RETURNING clause, because
@@ -253,10 +259,6 @@ type upsertRun struct {
 	// BatchedNext() has completed the work already.
 	done bool
 
-	// autoCommit indicates whether the last KV batch processed by
-	// this update will also commit the KV txn.
-	autoCommit autoCommitOpt
-
 	// traceKV caches the current KV tracing flag.
 	traceKV bool
 }
@@ -342,13 +344,19 @@ func (n *upsertNode) BatchedNext(params runParams) (bool, error) {
 	}
 
 	if lastBatch {
-		if _, err := n.run.tw.finalize(params.ctx, n.run.autoCommit, n.run.traceKV); err != nil {
+		if _, err := n.run.tw.finalize(params.ctx, n.run.traceKV); err != nil {
 			return false, err
 		}
 		// Remember we're done for the next call to BatchedNext().
 		n.run.done = true
 	}
 
+	// Possibly initiate a run of CREATE STATISTICS.
+	params.ExecCfg().StatsRefresher.NotifyMutation(
+		n.run.tw.tableDesc().ID,
+		n.run.tw.batchedCount(),
+	)
+
 	return n.run.tw.batchedCount() > 0, nil
 }
 
@@ -363,7 +371,7 @@ func (n *upsertNode) processSourceRow(params runParams, sourceVals tree.Datums)
 		n.run.computeExprs,
 		n.run.insertCols,
 		n.run.computedCols,
-		*params.EvalContext(),
+		*params.EvalContext().Copy(),
 		n.run.tw.tableDesc(),
 		sourceVals,
 		&n.run.iVarContainerForComputedCols,
@@ -372,21 +380,29 @@ func (n *upsertNode) processSourceRow(params runParams, sourceVals tree.Datums)
 		return err
 	}
 
-	// Run the CHECK constraints, if any.
-	if len(n.run.checkHelper.Exprs) > 0 {
-		insertColIDtoRowIndex := n.run.iVarContainerForComputedCols.Mapping
-		if err := n.run.checkHelper.LoadRow(insertColIDtoRowIndex, rowVals, false); err != nil {
-			return err
-		}
-		if err := n.run.checkHelper.Check(params.EvalContext()); err != nil {
-			return err
+	// Run the CHECK constraints, if any. CheckHelper will either evaluate the
+	// constraints itself, or else inspect boolean columns from the input that
+	// contain the results of evaluation.
+	if n.run.checkHelper != nil {
+		if n.run.checkHelper.NeedsEval() {
+			insertColIDtoRowIndex := n.run.iVarContainerForComputedCols.Mapping
+			if err := n.run.checkHelper.LoadEvalRow(insertColIDtoRowIndex, rowVals, false); err != nil {
+				return err
+			}
+			if err := n.run.checkHelper.CheckEval(params.EvalContext()); err != nil {
+				return err
+			}
+		} else {
+			checkVals := sourceVals[len(sourceVals)-n.run.checkHelper.Count():]
+			if err := n.run.checkHelper.CheckInput(checkVals); err != nil {
+				return err
+			}
 		}
 	}
 
 	// Process the row. This is also where the tableWriter will accumulate
 	// the row for later.
-	_, err = n.run.tw.row(params.ctx, rowVals, n.run.traceKV)
-	return err
+	return n.run.tw.row(params.ctx, rowVals, n.run.traceKV)
 }
 
 // BatchedCount implements the batchedPlanNode interface.
@@ -406,7 +422,7 @@ func (n *upsertNode) Close(ctx context.Context) {
 
 // enableAutoCommit is part of the autoCommitNode interface.
 func (n *upsertNode) enableAutoCommit() {
-	n.run.autoCommit = autoCommitEnabled
+	n.run.tw.enableAutoCommit()
 }
 
 // upsertExcludedTable is the name of a synthetic table used in an upsert's set
@@ -481,7 +497,7 @@ type upsertHelper struct {
 	// IndexedVarContainer and the IndexedVar objects in sub-expressions
 	// will link to it by reference after checkRenderStar / analyzeExpr.
 	// Enforce this using NoCopy.
-	noCopy util.NoCopy
+	_ util.NoCopy
 }
 
 var _ tableUpsertEvaler = (*upsertHelper)(nil)
@@ -522,7 +538,7 @@ func (uh *upsertHelper) IndexedVarNodeFormatter(idx int) tree.NodeFormatter {
 func (p *planner) newUpsertHelper(
 	ctx context.Context,
 	tn *tree.TableName,
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.ImmutableTableDescriptor,
 	insertCols []sqlbase.ColumnDescriptor,
 	updateCols []sqlbase.ColumnDescriptor,
 	updateExprs tree.UpdateExprs,
@@ -553,6 +569,10 @@ func (p *planner) newUpsertHelper(
 	// contains the metadata for [a,b,c,d] already, so there's always
 	// exactly one entry in defaultExprs for the untuplification to
 	// find.
+	//
+	// However we need to be careful when the RHS is not a tuple
+	// constructor, for example it is a tuple-returning unary expression
+	// or a subquery.
 	untupledExprs := make(tree.Exprs, 0, len(updateExprs))
 	i := 0
 	for _, updateExpr := range updateExprs {
@@ -563,6 +583,9 @@ func (p *planner) newUpsertHelper(
 					untupledExprs = append(untupledExprs, e)
 					i++
 				}
+			} else {
+				return nil, pgerror.UnimplementedWithIssueErrorf(8330,
+					"cannot use this type of expression on the right of UPDATE SET: %s", updateExpr.Expr)
 			}
 		} else {
 			e := fillDefault(updateExpr.Expr, i, defaultExprs)
@@ -657,7 +680,7 @@ func (p *planner) newUpsertHelper(
 	// column IDs to row datum positions is straightforward.
 	helper.ccIvarContainer = sqlbase.RowIndexedVarContainer{
 		Cols:    tableDesc.Columns,
-		Mapping: sqlbase.ColIDtoRowIndexFromCols(tableDesc.Columns),
+		Mapping: row.ColIDtoRowIndexFromCols(tableDesc.Columns),
 	}
 
 	return helper, nil
@@ -735,7 +758,7 @@ func (uh *upsertHelper) shouldUpdate(insertRow tree.Datums, existingRow tree.Dat
 //   or auto-generated assignments for UPSERT.
 // - conflictIdx: the conflicting index, if specified.
 func upsertExprsAndIndex(
-	tableDesc *sqlbase.TableDescriptor,
+	tableDesc *sqlbase.ImmutableTableDescriptor,
 	onConflict tree.OnConflict,
 	insertCols []sqlbase.ColumnDescriptor,
 ) (
@@ -807,5 +830,6 @@ func upsertExprsAndIndex(
 			return false, onConflict.Exprs, &tableDesc.Indexes[i], nil
 		}
 	}
-	return false, nil, nil, fmt.Errorf("there is no unique or exclusion constraint matching the ON CONFLICT specification")
+	return false, nil, nil, pgerror.NewErrorf(pgerror.CodeInvalidColumnReferenceError,
+		"there is no unique or exclusion constraint matching the ON CONFLICT specification")
 }
