@@ -24,13 +24,14 @@ import (
 	"time"
 
 	"github.com/cockroachdb/cockroach/pkg/build"
+	"github.com/cockroachdb/cockroach/pkg/server/telemetry"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/sql/pgwire/pgerror"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/builtins"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sessiondata"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
+	"github.com/cockroachdb/cockroach/pkg/sql/sqltelemetry"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 )
@@ -49,6 +50,9 @@ type getStringValFn = func(
 // sessionVar provides a unified interface for performing operations on
 // variables such as the selected database, or desired syntax.
 type sessionVar struct {
+	// Hidden indicates that the variable should not show up in the output of SHOW ALL.
+	Hidden bool
+
 	// Get returns a string representation of a given variable to be used
 	// either by SHOW or in the pg_catalog table.
 	Get func(evalCtx *extendedEvalContext) string
@@ -180,8 +184,8 @@ var varGen = map[string]sessionVar{
 
 			if len(dbName) != 0 {
 				// Verify database descriptor exists.
-				if _, err := evalCtx.schemaAccessors.logical.GetDatabaseDesc(dbName,
-					DatabaseLookupFlags{ctx: ctx, txn: evalCtx.Txn, required: true}); err != nil {
+				if _, err := evalCtx.schemaAccessors.logical.GetDatabaseDesc(ctx, evalCtx.Txn, dbName,
+					DatabaseLookupFlags{required: true}); err != nil {
 					return "", err
 				}
 			}
@@ -219,13 +223,45 @@ var varGen = map[string]sessionVar{
 		Get:           func(evalCtx *extendedEvalContext) string { return "ISO, MDY" },
 		GlobalDefault: func(_ *settings.Values) string { return "ISO, MDY" },
 	},
-
+	// Controls the subsequent parsing of a "naked" INT type.
+	// TODO(bob): Remove or no-op this in v2.4: https://github.com/cockroachdb/cockroach/issues/32844
+	`default_int_size`: {
+		Get: func(evalCtx *extendedEvalContext) string {
+			return strconv.FormatInt(int64(evalCtx.SessionData.DefaultIntSize), 10)
+		},
+		GetStringVal: makeIntGetStringValFn("default_int_size"),
+		Set: func(ctx context.Context, m *sessionDataMutator, val string) error {
+			i, err := strconv.ParseInt(val, 10, 64)
+			if err != nil {
+				return wrapSetVarError("default_int_size", val, "%v", err)
+			}
+			if i != 4 && i != 8 {
+				return pgerror.NewError(pgerror.CodeInvalidParameterValueError,
+					`only 4 or 8 are supported by default_int_size`)
+			}
+			// Only record when the value has been changed to a non-default
+			// value, since we really just want to know how useful int4-mode
+			// is. If we were to record counts for size.4 and size.8
+			// variables, we'd have to distinguish cases in which a session
+			// was opened in int8 mode and switched to int4 mode, versus ones
+			// set to int4 by a connection string.
+			// TODO(bob): Change to 8 in v2.3: https://github.com/cockroachdb/cockroach/issues/32534
+			if i == 4 {
+				telemetry.Inc(sqltelemetry.DefaultIntSize4Counter)
+			}
+			m.SetDefaultIntSize(int(i))
+			return nil
+		},
+		GlobalDefault: func(sv *settings.Values) string {
+			return strconv.FormatInt(defaultIntSize.Get(sv), 10)
+		},
+	},
 	// See https://www.postgresql.org/docs/10/static/runtime-config-client.html#GUC-DEFAULT-TRANSACTION-ISOLATION
 	`default_transaction_isolation`: {
 		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
 			switch strings.ToUpper(s) {
 			case `READ UNCOMMITTED`, `READ COMMITTED`, `SNAPSHOT`, `REPEATABLE READ`, `SERIALIZABLE`, `DEFAULT`:
-				m.SetDefaultIsolationLevel(enginepb.SERIALIZABLE)
+				// Do nothing. All transactions execute with serializable isolation.
 			default:
 				return newVarValueError(`default_transaction_isolation`, s, "serializable")
 			}
@@ -233,11 +269,10 @@ var varGen = map[string]sessionVar{
 			return nil
 		},
 		Get: func(evalCtx *extendedEvalContext) string {
-			return evalCtx.SessionData.DefaultIsolationLevel.ToLowerCaseString()
+			return "serializable"
 		},
 		GlobalDefault: func(sv *settings.Values) string { return "default" },
 	},
-
 	// See https://www.postgresql.org/docs/9.3/static/runtime-config-client.html#GUC-DEFAULT-TRANSACTION-READ-ONLY
 	`default_transaction_read_only`: {
 		GetStringVal: makeBoolGetStringValFn("default_transaction_read_only"),
@@ -273,23 +308,6 @@ var varGen = map[string]sessionVar{
 		},
 	},
 
-	// CockroachDB extension.
-	`experimental_force_lookup_join`: {
-		GetStringVal: makeBoolGetStringValFn(`experimental_force_lookup_join`),
-		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
-			b, err := parsePostgresBool(s)
-			if err != nil {
-				return err
-			}
-			m.SetLookupJoinEnabled(b)
-			return nil
-		},
-		Get: func(evalCtx *extendedEvalContext) string {
-			return formatBoolAsPostgresSetting(evalCtx.SessionData.LookupJoinEnabled)
-		},
-		GlobalDefault: globalFalse,
-	},
-
 	// CockroachDB extension.
 	`experimental_force_split_at`: {
 		GetStringVal: makeBoolGetStringValFn(`experimental_force_split_at`),
@@ -308,8 +326,8 @@ var varGen = map[string]sessionVar{
 	},
 
 	// CockroachDB extension.
-	`experimental_force_zigzag_join`: {
-		GetStringVal: makeBoolGetStringValFn(`experimental_force_zigzag_join`),
+	`experimental_enable_zigzag_join`: {
+		GetStringVal: makeBoolGetStringValFn(`experimental_enable_zigzag_join`),
 		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
 			b, err := parsePostgresBool(s)
 			if err != nil {
@@ -324,6 +342,48 @@ var varGen = map[string]sessionVar{
 		GlobalDefault: globalFalse,
 	},
 
+	// CockroachDB extension.
+	`reorder_joins_limit`: {
+		GetStringVal: makeIntGetStringValFn(`reorder_joins_limit`),
+		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
+			b, err := strconv.ParseInt(s, 10, 64)
+			if err != nil {
+				return err
+			}
+			if b < 0 {
+				return pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError,
+					"cannot set reorder_joins_limit to a negative value: %d", b)
+			}
+			m.SetReorderJoinsLimit(int(b))
+			return nil
+		},
+		Get: func(evalCtx *extendedEvalContext) string {
+			return strconv.FormatInt(int64(evalCtx.SessionData.ReorderJoinsLimit), 10)
+		},
+		GlobalDefault: func(sv *settings.Values) string {
+			return strconv.FormatInt(ReorderJoinsLimitClusterValue.Get(sv), 10)
+		},
+	},
+
+	// CockroachDB extension.
+	`experimental_vectorize`: {
+		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
+			mode, ok := sessiondata.VectorizeExecModeFromString(s)
+			if !ok {
+				return newVarValueError(`experimental_vectorize`, s, "off", "on", "always")
+			}
+			m.SetVectorize(mode)
+			return nil
+		},
+		Get: func(evalCtx *extendedEvalContext) string {
+			return evalCtx.SessionData.Vectorize.String()
+		},
+		GlobalDefault: func(sv *settings.Values) string {
+			return sessiondata.VectorizeExecMode(
+				VectorizeClusterMode.Get(sv)).String()
+		},
+	},
+
 	// CockroachDB extension.
 	`optimizer`: {
 		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
@@ -365,17 +425,7 @@ var varGen = map[string]sessionVar{
 
 	// See https://www.postgresql.org/docs/10/static/runtime-config-client.html
 	`extra_float_digits`: {
-		GetStringVal: func(
-			ctx context.Context, evalCtx *extendedEvalContext, values []tree.TypedExpr,
-		) (string, error) {
-			// TODO(knz): extract this code into a common makeIntGetStringValFn when
-			// there are multiple integer parameters.
-			s, err := getIntVal(&evalCtx.EvalContext, `extra_float_digits`, values)
-			if err != nil {
-				return "", err
-			}
-			return strconv.FormatInt(s, 10), nil
-		},
+		GetStringVal: makeIntGetStringValFn(`extra_float_digits`),
 		Set: func(
 			_ context.Context, m *sessionDataMutator, s string,
 		) error {
@@ -398,13 +448,39 @@ var varGen = map[string]sessionVar{
 		},
 		GlobalDefault: func(sv *settings.Values) string { return "0" },
 	},
-
+	// CockroachDB extension. See docs on SessionData.ForceSavepointRestart.
+	// https://github.com/cockroachdb/cockroach/issues/30588
+	`force_savepoint_restart`: {
+		Get: func(evalCtx *extendedEvalContext) string {
+			return formatBoolAsPostgresSetting(evalCtx.SessionData.ForceSavepointRestart)
+		},
+		GetStringVal: makeBoolGetStringValFn("force_savepoint_restart"),
+		Set: func(_ context.Context, m *sessionDataMutator, val string) error {
+			b, err := parsePostgresBool(val)
+			if err != nil {
+				return err
+			}
+			if b {
+				telemetry.Inc(sqltelemetry.ForceSavepointRestartCounter)
+			}
+			m.SetForceSavepointRestart(b)
+			return nil
+		},
+		GlobalDefault: globalFalse,
+	},
 	// See https://www.postgresql.org/docs/10/static/runtime-config-preset.html
 	`integer_datetimes`: makeReadOnlyVar("on"),
 
 	// See https://www.postgresql.org/docs/10/static/runtime-config-client.html#GUC-INTERVALSTYLE
 	`intervalstyle`: makeCompatStringVar(`IntervalStyle`, "postgres"),
 
+	// See https://www.postgresql.org/docs/10/static/runtime-config-client.html#GUC-LOC-TIMEOUT
+	`lock_timeout`: makeCompatIntVar(`lock_timeout`, 0),
+
+	// See https://www.postgresql.org/docs/10/static/runtime-config-client.html#GUC-IDLE-IN-TRANSACTION-SESSION-TIMEOUT
+	// See also issue #5924.
+	`idle_in_transaction_session_timeout`: makeCompatIntVar(`idle_in_transaction_session_timeout`, 0),
+
 	// See https://www.postgresql.org/docs/10/static/runtime-config-preset.html#GUC-MAX-INDEX-KEYS
 	`max_index_keys`: makeReadOnlyVar("32"),
 
@@ -415,6 +491,14 @@ var varGen = map[string]sessionVar{
 		},
 	},
 
+	// CockroachDB extension.
+	// TODO(dan): This should also work with SET.
+	`results_buffer_size`: {
+		Get: func(evalCtx *extendedEvalContext) string {
+			return strconv.FormatInt(evalCtx.SessionData.ResultsBufferSize, 10)
+		},
+	},
+
 	// CockroachDB extension (inspired by MySQL).
 	// See https://dev.mysql.com/doc/refman/5.7/en/server-system-variables.html#sysvar_sql_safe_updates
 	`sql_safe_updates`: {
@@ -478,6 +562,25 @@ var varGen = map[string]sessionVar{
 	// See https://www.postgresql.org/docs/10/static/runtime-config-preset.html#GUC-SERVER-VERSION-NUM
 	`server_version_num`: makeReadOnlyVar(PgServerVersionNum),
 
+	// See https://www.postgresql.org/docs/9.4/runtime-config-connection.html
+	`ssl_renegotiation_limit`: {
+		Hidden:        true,
+		GetStringVal:  makeIntGetStringValFn(`ssl_renegotiation_limit`),
+		Get:           func(_ *extendedEvalContext) string { return "0" },
+		GlobalDefault: func(_ *settings.Values) string { return "0" },
+		Set: func(_ context.Context, _ *sessionDataMutator, s string) error {
+			i, err := strconv.ParseInt(s, 10, 64)
+			if err != nil {
+				return wrapSetVarError("ssl_renegotiation_limit", s, "%v", err)
+			}
+			if i != 0 {
+				// See pg src/backend/utils/misc/guc.c: non-zero values are not to be supported.
+				return newVarValueError("ssl_renegotiation_limit", s, "0")
+			}
+			return nil
+		},
+	},
+
 	// CockroachDB extension.
 	`crdb_version`: makeReadOnlyVar(build.GetInfo().Short()),
 
@@ -488,9 +591,31 @@ var varGen = map[string]sessionVar{
 		Get: func(evalCtx *extendedEvalContext) string { return evalCtx.SessionData.User },
 	},
 
+	// See pg sources src/backend/utils/misc/guc.c. The variable is defined
+	// but is hidden from SHOW ALL.
+	`session_authorization`: {
+		Hidden: true,
+		Get:    func(evalCtx *extendedEvalContext) string { return evalCtx.SessionData.User },
+	},
+
 	// Supported for PG compatibility only.
 	// See https://www.postgresql.org/docs/10/static/runtime-config-compatible.html#GUC-STANDARD-CONFORMING-STRINGS
-	`standard_conforming_strings`: makeCompatStringVar(`standard_conforming_strings`, "on"),
+	`standard_conforming_strings`: makeCompatBoolVar(`standard_conforming_strings`, true, false /* anyAllowed */),
+
+	// See https://www.postgresql.org/docs/10/static/runtime-config-compatible.html#GUC-SYNCHRONIZE-SEQSCANS
+	// The default in pg is "on" but the behavior in CockroachDB is "off". As this does not affect
+	// results received by clients, we accept both values.
+	`synchronize_seqscans`: makeCompatBoolVar(`synchronize_seqscans`, true, true /* anyAllowed */),
+
+	// See https://www.postgresql.org/docs/10/static/runtime-config-client.html#GUC-ROW-SECURITY
+	// The default in pg is "on" but row security is not supported in CockroachDB.
+	// We blindly accept both values because as long as there are now row security policies defined,
+	// either value produces the same query results in PostgreSQL. That is, as long as CockroachDB
+	// does not support row security, accepting either "on" and "off" but ignoring the result
+	// is postgres-compatible.
+	// If/when CockroachDB is extended to support row security, the default and allowed values
+	// should be modified accordingly.
+	`row_security`: makeCompatBoolVar(`row_security`, false, true /* anyAllowed */),
 
 	`statement_timeout`: {
 		GetStringVal: stmtTimeoutVarGetStringVal,
@@ -525,17 +650,14 @@ var varGen = map[string]sessionVar{
 	// See https://github.com/postgres/postgres/blob/REL_10_STABLE/src/backend/utils/misc/guc.c#L3401-L3409
 	`transaction_isolation`: {
 		Get: func(evalCtx *extendedEvalContext) string {
-			return evalCtx.Txn.Isolation().ToLowerCaseString()
+			return "serializable"
 		},
 		RuntimeSet: func(_ context.Context, evalCtx *extendedEvalContext, s string) error {
-			isolationLevel, ok := tree.IsolationLevelMap[s]
+			_, ok := tree.IsolationLevelMap[s]
 			if !ok {
 				return newVarValueError(`transaction_isolation`, s, "serializable")
 			}
-			return evalCtx.TxnModesSetter.setTransactionModes(
-				tree.TransactionModes{
-					Isolation: isolationLevel,
-				})
+			return nil
 		},
 		GlobalDefault: func(_ *settings.Values) string { return "serializable" },
 	},
@@ -589,6 +711,23 @@ var varGen = map[string]sessionVar{
 		},
 		// Setting is done by the SetTracing statement.
 	},
+
+	// CockroachDB extension.
+	`allow_prepare_as_opt_plan`: {
+		Hidden: true,
+		Get: func(evalCtx *extendedEvalContext) string {
+			return formatBoolAsPostgresSetting(evalCtx.SessionData.AllowPrepareAsOptPlan)
+		},
+		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
+			b, err := parsePostgresBool(s)
+			if err != nil {
+				return err
+			}
+			m.SetAllowPrepareAsOptPlan(b)
+			return nil
+		},
+		GlobalDefault: globalFalse,
+	},
 }
 
 func makeBoolGetStringValFn(varName string) getStringValFn {
@@ -610,7 +749,47 @@ func makeReadOnlyVar(value string) sessionVar {
 	}
 }
 
-func globalFalse(_ *settings.Values) string { return formatBoolAsPostgresSetting(false) }
+func displayPgBool(val bool) func(_ *settings.Values) string {
+	strVal := formatBoolAsPostgresSetting(val)
+	return func(_ *settings.Values) string { return strVal }
+}
+
+var globalFalse = displayPgBool(false)
+
+func makeCompatBoolVar(varName string, displayValue, anyValAllowed bool) sessionVar {
+	displayValStr := formatBoolAsPostgresSetting(displayValue)
+	return sessionVar{
+		Get: func(_ *extendedEvalContext) string { return displayValStr },
+		Set: func(_ context.Context, m *sessionDataMutator, s string) error {
+			b, err := parsePostgresBool(s)
+			if err != nil {
+				return pgerror.NewErrorf(pgerror.CodeInvalidParameterValueError, "%v", err)
+			}
+			if anyValAllowed || b == displayValue {
+				return nil
+			}
+			telemetry.Inc(sqltelemetry.UnimplementedSessionVarValueCounter(varName, s))
+			allowedVals := []string{displayValStr}
+			if anyValAllowed {
+				allowedVals = append(allowedVals, formatBoolAsPostgresSetting(!displayValue))
+			}
+			return newVarValueError(varName, s, allowedVals...).SetDetailf(
+				"this parameter is currently recognized only for compatibility and has no effect in CockroachDB.")
+		},
+		GlobalDefault: func(sv *settings.Values) string { return displayValStr },
+	}
+}
+
+func makeCompatIntVar(varName string, displayValue int, extraAllowed ...int) sessionVar {
+	displayValueStr := strconv.Itoa(displayValue)
+	extraAllowedStr := make([]string, len(extraAllowed))
+	for i, v := range extraAllowed {
+		extraAllowedStr[i] = strconv.Itoa(v)
+	}
+	varObj := makeCompatStringVar(varName, displayValueStr, extraAllowedStr...)
+	varObj.GetStringVal = makeIntGetStringValFn(varName)
+	return varObj
+}
 
 func makeCompatStringVar(varName, displayValue string, extraAllowed ...string) sessionVar {
 	allowedVals := append(extraAllowed, strings.ToLower(displayValue))
@@ -625,6 +804,7 @@ func makeCompatStringVar(varName, displayValue string, extraAllowed ...string) s
 					return nil
 				}
 			}
+			telemetry.Inc(sqltelemetry.UnimplementedSessionVarValueCounter(varName, s))
 			return newVarValueError(varName, s, allowedVals...).SetDetailf(
 				"this parameter is currently recognized only for compatibility and has no effect in CockroachDB.")
 		},
@@ -632,6 +812,18 @@ func makeCompatStringVar(varName, displayValue string, extraAllowed ...string) s
 	}
 }
 
+// makeIntGetStringValFn returns a getStringValFn which allows
+// the user to provide plain integer values to a SET variable.
+func makeIntGetStringValFn(name string) getStringValFn {
+	return func(ctx context.Context, evalCtx *extendedEvalContext, values []tree.TypedExpr) (string, error) {
+		s, err := getIntVal(&evalCtx.EvalContext, name, values)
+		if err != nil {
+			return "", err
+		}
+		return strconv.FormatInt(s, 10), nil
+	}
+}
+
 // IsSessionVariableConfigurable returns true iff there is a session
 // variable with the given name and it is settable by a client
 // (e.g. in pgwire).
@@ -667,3 +859,51 @@ func getSingleBool(
 	}
 	return b, nil
 }
+
+func getSessionVar(name string, missingOk bool) (bool, sessionVar, error) {
+	if _, ok := UnsupportedVars[name]; ok {
+		return false, sessionVar{}, pgerror.Unimplemented("set."+name,
+			"the configuration setting %q is not supported", name)
+	}
+
+	v, ok := varGen[name]
+	if !ok {
+		if missingOk {
+			return false, sessionVar{}, nil
+		}
+		return false, sessionVar{}, pgerror.NewErrorf(pgerror.CodeUndefinedObjectError,
+			"unrecognized configuration parameter %q", name)
+	}
+
+	return true, v, nil
+}
+
+// GetSessionVar implements the EvalSessionAccessor interface.
+func (p *planner) GetSessionVar(
+	_ context.Context, varName string, missingOk bool,
+) (bool, string, error) {
+	name := strings.ToLower(varName)
+	ok, v, err := getSessionVar(name, missingOk)
+	if err != nil || !ok {
+		return ok, "", err
+	}
+
+	return true, v.Get(&p.extendedEvalCtx), nil
+}
+
+// SetSessionVar implements the EvalSessionAccessor interface.
+func (p *planner) SetSessionVar(ctx context.Context, varName, newVal string) error {
+	name := strings.ToLower(varName)
+	_, v, err := getSessionVar(name, false /* missingOk */)
+	if err != nil {
+		return err
+	}
+
+	if v.Set == nil && v.RuntimeSet == nil {
+		return newCannotChangeParameterError(name)
+	}
+	if v.RuntimeSet != nil {
+		return v.RuntimeSet(ctx, &p.extendedEvalCtx, newVal)
+	}
+	return v.Set(ctx, p.sessionDataMutator, newVal)
+}
