@@ -21,6 +21,7 @@ import (
 	"reflect"
 	"strconv"
 
+	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/sql/sem/tree"
 	"github.com/cockroachdb/cockroach/pkg/sql/sqlbase"
 	"github.com/cockroachdb/cockroach/pkg/util"
@@ -50,11 +51,21 @@ type planObserver struct {
 	// expr is invoked for each expression field in each node.
 	expr func(verbosity observeVerbosity, nodeName, fieldName string, n int, expr tree.Expr)
 
+	// spans is invoked for spans embbeded in each node.
+	spans func(nodeName, fieldName string, index *sqlbase.IndexDescriptor, spans []roachpb.Span)
+
 	// attr is invoked for non-expression metadata in each node.
 	attr func(nodeName, fieldName, attr string)
 
 	// leaveNode is invoked upon leaving a tree node.
 	leaveNode func(nodeName string, plan planNode) error
+
+	// followRowSourceToPlanNode controls whether the tree walker continues
+	// walking when it encounters a rowSourceToPlanNode, which indicates that the
+	// logical plan has been mutated for distribution. This should normally be
+	// set to false, as normally the planNodeToRowSource on the other end will
+	// take care of propagating signals via its own walker.
+	followRowSourceToPlanNode bool
 }
 
 // walkPlan performs a depth-first traversal of the plan given as
@@ -179,12 +190,17 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 			if n.specifiedIndex != nil {
 				v.observer.attr(name, "hint", fmt.Sprintf("force index @%s", n.specifiedIndex.Name))
 			}
-			spans := sqlbase.PrettySpans(n.index, n.spans, 2)
-			if spans != "" {
-				if spans == "-" {
-					spans = "ALL"
-				}
-				v.observer.attr(name, "spans", spans)
+		}
+		if v.observer.spans != nil {
+			v.observer.spans(name, "spans", n.index, n.spans)
+		}
+		if v.observer.attr != nil {
+			// Only print out "parallel" when it makes sense. i.e. don't print if
+			// we know we will get only one result from the scan. There are cases
+			// in which "parallel" will be printed out even though the spans cover
+			// a single range, but there is nothing we can do about that.
+			if n.canParallelize() && (len(n.spans) > 1 || n.maxResults > 1) {
+				v.observer.attr(name, "parallel", "")
 			}
 			if n.hardLimit > 0 && isFilterTrue(n.filter) {
 				v.observer.attr(name, "limit", fmt.Sprintf("%d", n.hardLimit))
@@ -209,18 +225,49 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 		n.source.plan = v.visit(n.source.plan)
 
 	case *indexJoinNode:
+		if v.observer.attr != nil {
+			v.observer.attr(name, "table", fmt.Sprintf("%s@%s", n.table.desc.Name, n.table.index.Name))
+			v.expr(name, "filter", -1, n.table.filter)
+		}
 		v.visitConcrete(n.index)
-		v.visitConcrete(n.table)
 
 	case *lookupJoinNode:
 		if v.observer.attr != nil {
+			v.observer.attr(name, "table", fmt.Sprintf("%s@%s", n.table.desc.Name, n.table.index.Name))
 			v.observer.attr(name, "type", joinTypeStr(n.joinType))
 		}
 		if v.observer.expr != nil && n.onCond != nil && n.onCond != tree.DBoolTrue {
 			v.expr(name, "pred", -1, n.onCond)
 		}
 		n.input = v.visit(n.input)
-		v.visitConcrete(n.table)
+
+	case *zigzagJoinNode:
+		if v.observer.attr != nil {
+			v.observer.attr(name, "type", joinTypeStr(sqlbase.InnerJoin))
+			if v.observer.expr != nil && n.onCond != nil && n.onCond != tree.DBoolTrue {
+				v.expr(name, "pred", -1, n.onCond)
+			}
+			for _, side := range n.sides {
+				v.visitConcrete(side.scan)
+				if side.fixedVals != nil {
+					description := fmt.Sprintf(
+						"%d column%s",
+						len(side.fixedVals.columns),
+						util.Pluralize(int64(len(side.fixedVals.columns))),
+					)
+					v.observer.attr(name, "fixedvals", description)
+				}
+			}
+		}
+
+	case *applyJoinNode:
+		if v.observer.attr != nil {
+			v.observer.attr(name, "type", joinTypeStr(n.joinType))
+		}
+		if v.observer.expr != nil {
+			v.expr(name, "pred", -1, n.pred.onCond)
+		}
+		n.input.plan = v.visit(n.input.plan)
 
 	case *joinNode:
 		if v.observer.attr != nil {
@@ -231,13 +278,19 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 			v.observer.attr(name, "type", jType)
 
 			if len(n.pred.leftColNames) > 0 {
-				f := tree.NewFmtCtxWithBuf(tree.FmtSimple)
+				f := tree.NewFmtCtx(tree.FmtSimple)
 				f.WriteByte('(')
 				f.FormatNode(&n.pred.leftColNames)
 				f.WriteString(") = (")
 				f.FormatNode(&n.pred.rightColNames)
 				f.WriteByte(')')
 				v.observer.attr(name, "equality", f.CloseAndGetString())
+				if n.pred.leftEqKey {
+					v.observer.attr(name, "left cols are key", "")
+				}
+				if n.pred.rightEqKey {
+					v.observer.attr(name, "right cols are key", "")
+				}
 			}
 			if len(n.mergeJoinOrdering) > 0 {
 				// The ordering refers to equality columns
@@ -265,6 +318,9 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 		}
 		n.plan = v.visit(n.plan)
 
+	case *max1RowNode:
+		n.plan = v.visit(n.plan)
+
 	case *distinctNode:
 		if v.observer.attr == nil {
 			n.plan = v.visit(n.plan)
@@ -311,12 +367,6 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 				order.addOrderColumn(o.ColIdx, o.Direction)
 			}
 			v.observer.attr(name, "order", order.AsString(columns))
-			switch ss := n.run.sortStrategy.(type) {
-			case *iterativeSortStrategy:
-				v.observer.attr(name, "strategy", "iterative")
-			case *sortTopKStrategy:
-				v.observer.attr(name, "strategy", fmt.Sprintf("top %d", ss.topK))
-			}
 		}
 		n.plan = v.visit(n.plan)
 
@@ -389,14 +439,17 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 			}
 			buf.WriteByte(')')
 			v.observer.attr(name, "into", buf.String())
+			v.observer.attr(name, "strategy", n.run.ti.desc())
 		}
 
 		if v.observer.expr != nil {
 			for i, dexpr := range n.run.defaultExprs {
 				v.metadataExpr(name, "default", i, dexpr)
 			}
-			for i, cexpr := range n.run.checkHelper.Exprs {
-				v.metadataExpr(name, "check", i, cexpr)
+			if n.run.checkHelper != nil {
+				for i, cexpr := range n.run.checkHelper.Exprs {
+					v.metadataExpr(name, "check", i, cexpr)
+				}
 			}
 		}
 		n.source = v.visit(n.source)
@@ -414,14 +467,17 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 			}
 			buf.WriteByte(')')
 			v.observer.attr(name, "into", buf.String())
+			v.observer.attr(name, "strategy", n.run.tw.desc())
 		}
 
 		if v.observer.expr != nil {
 			for i, dexpr := range n.run.defaultExprs {
 				v.metadataExpr(name, "default", i, dexpr)
 			}
-			for i, cexpr := range n.run.checkHelper.Exprs {
-				v.metadataExpr(name, "check", i, cexpr)
+			if n.run.checkHelper != nil {
+				for i, cexpr := range n.run.checkHelper.Exprs {
+					v.metadataExpr(name, "check", i, cexpr)
+				}
 			}
 			n.run.tw.walkExprs(func(d string, i int, e tree.TypedExpr) {
 				v.metadataExpr(name, d, i, e)
@@ -442,13 +498,16 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 				}
 				v.observer.attr(name, "set", buf.String())
 			}
+			v.observer.attr(name, "strategy", n.run.tu.desc())
 		}
 		if v.observer.expr != nil {
 			for i, cexpr := range n.run.computeExprs {
 				v.metadataExpr(name, "computed", i, cexpr)
 			}
-			for i, cexpr := range n.run.checkHelper.Exprs {
-				v.metadataExpr(name, "check", i, cexpr)
+			if n.run.checkHelper != nil {
+				for i, cexpr := range n.run.checkHelper.Exprs {
+					v.metadataExpr(name, "check", i, cexpr)
+				}
 			}
 		}
 		// An updater has no sub-expressions, so nothing special to do here.
@@ -457,10 +516,19 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 	case *deleteNode:
 		if v.observer.attr != nil {
 			v.observer.attr(name, "from", n.run.td.tableDesc().Name)
+			v.observer.attr(name, "strategy", n.run.td.desc())
 		}
 		// A deleter has no sub-expressions, so nothing special to do here.
 		n.source = v.visit(n.source)
 
+	case *deleteRangeNode:
+		if v.observer.attr != nil {
+			v.observer.attr(name, "from", n.desc.Name)
+		}
+		if v.observer.spans != nil {
+			v.observer.spans(name, "spans", &n.desc.PrimaryIndex, n.spans)
+		}
+
 	case *serializeNode:
 		v.visitConcrete(n.source)
 
@@ -539,6 +607,11 @@ func (v *planVisitor) visitInternal(plan planNode, name string) {
 			}
 		}
 		n.source = v.visit(n.source)
+
+	case *rowSourceToPlanNode:
+		if v.observer.followRowSourceToPlanNode && n.originalPlanNode != nil {
+			v.visit(n.originalPlanNode)
+		}
 	}
 }
 
@@ -579,6 +652,12 @@ func nodeName(plan planNode) string {
 		if n.emitAll {
 			return "append"
 		}
+
+	case *joinNode:
+		if len(n.mergeJoinOrdering) > 0 {
+			return "merge-join"
+		}
+		return "hash-join"
 	}
 
 	name, ok := planNodeNames[reflect.TypeOf(plan)]
@@ -638,6 +717,10 @@ var planNodeNames = map[reflect.Type]string{
 	reflect.TypeOf(&alterSequenceNode{}):        "alter sequence",
 	reflect.TypeOf(&alterTableNode{}):           "alter table",
 	reflect.TypeOf(&alterUserSetPasswordNode{}): "alter user",
+	reflect.TypeOf(&applyJoinNode{}):            "apply-join",
+	reflect.TypeOf(&commentOnColumnNode{}):      "comment on column",
+	reflect.TypeOf(&commentOnDatabaseNode{}):    "comment on database",
+	reflect.TypeOf(&commentOnTableNode{}):       "comment on table",
 	reflect.TypeOf(&cancelQueriesNode{}):        "cancel queries",
 	reflect.TypeOf(&cancelSessionsNode{}):       "cancel sessions",
 	reflect.TypeOf(&controlJobsNode{}):          "control jobs",
@@ -650,6 +733,7 @@ var planNodeNames = map[reflect.Type]string{
 	reflect.TypeOf(&createViewNode{}):           "create view",
 	reflect.TypeOf(&delayedNode{}):              "virtual table",
 	reflect.TypeOf(&deleteNode{}):               "delete",
+	reflect.TypeOf(&deleteRangeNode{}):          "delete range",
 	reflect.TypeOf(&distinctNode{}):             "distinct",
 	reflect.TypeOf(&dropDatabaseNode{}):         "drop database",
 	reflect.TypeOf(&dropIndexNode{}):            "drop index",
@@ -667,6 +751,7 @@ var planNodeNames = map[reflect.Type]string{
 	reflect.TypeOf(&joinNode{}):                 "join",
 	reflect.TypeOf(&limitNode{}):                "limit",
 	reflect.TypeOf(&lookupJoinNode{}):           "lookup-join",
+	reflect.TypeOf(&max1RowNode{}):              "max1row",
 	reflect.TypeOf(&ordinalityNode{}):           "ordinality",
 	reflect.TypeOf(&projectSetNode{}):           "project set",
 	reflect.TypeOf(&relocateNode{}):             "relocate",
@@ -686,7 +771,6 @@ var planNodeNames = map[reflect.Type]string{
 	reflect.TypeOf(&setVarNode{}):               "set",
 	reflect.TypeOf(&setZoneConfigNode{}):        "configure zone",
 	reflect.TypeOf(&showFingerprintsNode{}):     "showFingerprints",
-	reflect.TypeOf(&showRangesNode{}):           "showRanges",
 	reflect.TypeOf(&showTraceNode{}):            "show trace for",
 	reflect.TypeOf(&showTraceReplicaNode{}):     "replica trace",
 	reflect.TypeOf(&showZoneConfigNode{}):       "show zone configuration",
@@ -699,6 +783,8 @@ var planNodeNames = map[reflect.Type]string{
 	reflect.TypeOf(&updateNode{}):               "update",
 	reflect.TypeOf(&upsertNode{}):               "upsert",
 	reflect.TypeOf(&valuesNode{}):               "values",
+	reflect.TypeOf(&virtualTableNode{}):         "virtual table values",
 	reflect.TypeOf(&windowNode{}):               "window",
 	reflect.TypeOf(&zeroNode{}):                 "norows",
+	reflect.TypeOf(&zigzagJoinNode{}):           "zigzag-join",
 }
