@@ -20,8 +20,6 @@ import (
 	"sort"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/keys"
@@ -39,6 +37,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/pkg/errors"
 )
 
 var (
@@ -64,8 +63,13 @@ var _ base.ModuleTestingKnobs = &MigrationManagerTestingKnobs{}
 // startup. They will always be run from top-to-bottom, and because they are
 // assumed to be backward-compatible, they will be run regardless of what other
 // node versions are currently running within the cluster.
-// Migrations must be idempotent: a migration may run successfully but not be recorded
-// as completed, causing a second run.
+// Migrations must be idempotent: a migration may run successfully but not be
+// recorded as completed, causing a second run.
+//
+// Attention: If a migration is creating new tables, it should also be added to
+// the metadata schema written by bootstrap (see addSystemDatabaseToSchema())
+// and it should have the includedInBootstrap field set (see comments on that
+// field too).
 var backwardCompatibleMigrations = []migrationDescriptor{
 	{
 		// Introduced in v1.0. Baked into v2.0.
@@ -190,6 +194,20 @@ var backwardCompatibleMigrations = []migrationDescriptor{
 		name:   "add progress to system.jobs",
 		workFn: addJobsProgress,
 	},
+	{
+		// Introduced in v2.2.
+		// TODO(knz): bake this migration into v2.3.
+		name:                "create system.comment table",
+		workFn:              createCommentTable,
+		includedInBootstrap: true,
+		newDescriptorIDs:    staticIDs(keys.CommentsTableID),
+	},
+	{
+		// Introduced in v2.2.
+		// TODO(knz): bake this migration into v2.3.
+		name:   "propagate the ts purge interval to the new setting names",
+		workFn: retireOldTsPurgeIntervalSettings,
+	},
 }
 
 func staticIDs(ids ...sqlbase.ID) func(ctx context.Context, db db) ([]sqlbase.ID, error) {
@@ -217,8 +235,18 @@ type migrationDescriptor struct {
 	// name must be unique amongst all hard-coded migrations.
 	name string
 	// workFn must be idempotent so that we can safely re-run it if a node failed
-	// while running it.
+	// while running it. nil if the migration has been "backed in" and is no
+	// longer to be performed at cluster startup.
 	workFn func(context.Context, runner) error
+	// includedInBootstrap is set for migrations that need to be performed for
+	// updating old clusters, but are also covered by the MetadataSchema that gets
+	// created by hand for a new cluster when it bootstraps itself. This kind of
+	// duplication between a migration and the MetadataSchema is useful for
+	// migrations that create system descriptor - for new clusters (particularly
+	// for tests) we want to create these tables by hand so that a corresponding
+	// range is created at bootstrap time. Otherwise, we'd have the split queue
+	// asynchronously creating some ranges which is annoying for tests.
+	includedInBootstrap bool
 	// doesBackfill should be set to true if the migration triggers a backfill.
 	doesBackfill bool
 	// newDescriptorIDs is a function that returns the IDs of any additional
@@ -309,7 +337,10 @@ func ExpectedDescriptorIDs(ctx context.Context, db db) (sqlbase.IDs, error) {
 	}
 	descriptorIDs := sqlbase.MakeMetadataSchema().DescriptorIDs()
 	for _, migration := range backwardCompatibleMigrations {
-		if migration.newDescriptorIDs == nil {
+		// Is the migration not creating descriptors?
+		if migration.newDescriptorIDs == nil ||
+			// Is the migration included in the metadata schema considered above?
+			migration.includedInBootstrap {
 			continue
 		}
 		if _, ok := completedMigrations[string(migrationKey(migration))]; ok {
@@ -324,10 +355,21 @@ func ExpectedDescriptorIDs(ctx context.Context, db db) (sqlbase.IDs, error) {
 	return descriptorIDs, nil
 }
 
+// MigrationFilter is used to indicate to EnsureMigrations what to run.
+type MigrationFilter bool
+
+const (
+	// AllMigrations means run all the migrations.
+	AllMigrations MigrationFilter = true
+	// ExcludeMigrationsIncludedInBootstrap means don't run migrations with the
+	// includedInBootstrap field set.
+	ExcludeMigrationsIncludedInBootstrap MigrationFilter = false
+)
+
 // EnsureMigrations should be run during node startup to ensure that all
 // required migrations have been run (and running all those that are definitely
 // safe to run).
-func (m *Manager) EnsureMigrations(ctx context.Context) error {
+func (m *Manager) EnsureMigrations(ctx context.Context, filter MigrationFilter) error {
 	// First, check whether there are any migrations that need to be run.
 	completedMigrations, err := getCompletedMigrations(ctx, m.db)
 	if err != nil {
@@ -335,8 +377,10 @@ func (m *Manager) EnsureMigrations(ctx context.Context) error {
 	}
 	allMigrationsCompleted := true
 	for _, migration := range backwardCompatibleMigrations {
-		if migration.workFn == nil {
-			// Migration has been baked in. Ignore it.
+		if migration.workFn == nil || // has the migration been baked in?
+			// is the migration unnecessary?
+			(migration.includedInBootstrap &&
+				filter == ExcludeMigrationsIncludedInBootstrap) {
 			continue
 		}
 		if m.testingKnobs.DisableBackfillMigrations && migration.doesBackfill {
@@ -450,9 +494,7 @@ func (m *Manager) EnsureMigrations(ctx context.Context) error {
 			return errors.Wrapf(err, "failed to run migration %q", migration.name)
 		}
 
-		if log.V(1) {
-			log.Infof(ctx, "trying to persist record of completing migration %s", migration.name)
-		}
+		log.VEventf(ctx, 1, "persisting record of completing migration %s", migration.name)
 		if err := m.db.Put(ctx, key, startTime); err != nil {
 			return errors.Wrapf(err, "failed to persist record of completing migration %q",
 				migration.name)
@@ -503,6 +545,10 @@ func createSystemTable(ctx context.Context, r runner, desc sqlbase.TableDescript
 	return err
 }
 
+func createCommentTable(ctx context.Context, r runner) error {
+	return createSystemTable(ctx, r, sqlbase.CommentsTable)
+}
+
 var reportingOptOut = envutil.EnvOrDefaultBool("COCKROACH_SKIP_ENABLING_DIAGNOSTIC_REPORTING", false)
 
 func runStmtAsRootWithRetry(
@@ -565,10 +611,7 @@ func populateVersionSetting(ctx context.Context, r runner) error {
 		v = cluster.VersionByKey(cluster.VersionBase)
 	}
 
-	b, err := protoutil.Marshal(&cluster.ClusterVersion{
-		MinimumVersion: v,
-		UseVersion:     v,
-	})
+	b, err := protoutil.Marshal(&cluster.ClusterVersion{Version: v})
 	if err != nil {
 		return errors.Wrap(err, "while marshaling version")
 	}
@@ -683,7 +726,7 @@ func upgradeDescsWithFn(
 							return err
 						} else if upgraded {
 							// It's safe to ignore the DROP state here and
-							// unconditionally set UpVersion. For proof, see
+							// unconditionally increment the version. For proof, see
 							// TestDropTableWhileUpgradingFormat.
 							//
 							// In fact, it's of the utmost importance that this migration
@@ -695,12 +738,9 @@ func upgradeDescsWithFn(
 							// concern: consider that dropping a large table can take several
 							// days, while upgrading to a new version can take as little as a
 							// few minutes.
-							table.Version++
 							now = txn.CommitTimestamp()
-							idVersions = append(idVersions,
-								sql.NewIDVersion(
-									table.Name, table.ID, table.Version-2,
-								))
+							idVersions = append(idVersions, sql.NewIDVersionPrev(table))
+							table.Version++
 							// Use ValidateTable() instead of Validate()
 							// because of #26422. We still do not know why
 							// a table can reference a dropped database.
@@ -820,7 +860,7 @@ func addJobsProgress(ctx context.Context, r runner) error {
 		if err := txn.SetSystemConfigTrigger(); err != nil {
 			return err
 		}
-		desc, err := sqlbase.GetTableDescFromID(ctx, txn, keys.JobsTableID)
+		desc, err := sqlbase.GetMutableTableDescFromID(ctx, txn, keys.JobsTableID)
 		if err != nil {
 			return err
 		}
@@ -841,3 +881,39 @@ func addJobsProgress(ctx context.Context, r runner) error {
 		return txn.Put(ctx, sqlbase.MakeDescMetadataKey(desc.ID), sqlbase.WrapDescriptor(desc))
 	})
 }
+
+func retireOldTsPurgeIntervalSettings(ctx context.Context, r runner) error {
+	// We are going to deprecate `timeseries.storage.10s_resolution_ttl`
+	// into `timeseries.storage.resolution_10s.ttl` if the latter is not
+	// defined.
+	//
+	// Ditto for the `30m` resolution.
+
+	// Copy 'timeseries.storage.10s_resolution_ttl' into
+	// 'timeseries.storage.resolution_10s.ttl' if the former is defined
+	// and the latter is not defined yet.
+	//
+	// We rely on the SELECT returning no row if the original setting
+	// was not defined, and INSERT ON CONFLICT DO NOTHING to ignore the
+	// insert if the new name was already set.
+	if _, err := r.sqlExecutor.Exec(ctx, "copy-setting", nil /* txn */, `
+INSERT INTO system.settings (name, value, "lastUpdated", "valueType")
+   SELECT 'timeseries.storage.resolution_10s.ttl', value, "lastUpdated", "valueType"
+     FROM system.settings WHERE name = 'timeseries.storage.10s_resolution_ttl'
+ON CONFLICT (name) DO NOTHING`,
+	); err != nil {
+		return err
+	}
+
+	// Ditto 30m.
+	if _, err := r.sqlExecutor.Exec(ctx, "copy-setting", nil /* txn */, `
+INSERT INTO system.settings (name, value, "lastUpdated", "valueType")
+   SELECT 'timeseries.storage.resolution_30m.ttl', value, "lastUpdated", "valueType"
+     FROM system.settings WHERE name = 'timeseries.storage.30m_resolution_ttl'
+ON CONFLICT (name) DO NOTHING`,
+	); err != nil {
+		return err
+	}
+
+	return nil
+}
