@@ -20,17 +20,17 @@ import (
 	"fmt"
 	"math"
 	"math/rand"
+	"strings"
 	"time"
 
-	"github.com/pkg/errors"
-	"go.etcd.io/etcd/raft"
-
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/pkg/errors"
+	"go.etcd.io/etcd/raft"
 )
 
 const (
@@ -163,7 +163,19 @@ func (ae *allocatorError) Error() string {
 		}
 		return baseMsg + "; likely not enough nodes in cluster"
 	}
-	return fmt.Sprintf("%s; must match constraints %v", baseMsg, ae.constraints)
+	var b strings.Builder
+	b.WriteString(baseMsg)
+	b.WriteString("; must match constraints [")
+	for i := range ae.constraints {
+		if i > 0 {
+			b.WriteByte(' ')
+		}
+		b.WriteByte('{')
+		b.WriteString(ae.constraints[i].String())
+		b.WriteByte('}')
+	}
+	b.WriteString("]")
+	return b.String()
 }
 
 func (*allocatorError) purgatoryErrorMarker() {}
@@ -280,7 +292,7 @@ func GetNeededReplicas(zoneConfigReplicaCount int32, clusterNodes int) int {
 // supplied range, as governed by the supplied zone configuration. It
 // returns the required action that should be taken and a priority.
 func (a *Allocator) ComputeAction(
-	ctx context.Context, zone config.ZoneConfig, rangeInfo RangeInfo,
+	ctx context.Context, zone *config.ZoneConfig, rangeInfo RangeInfo,
 ) (AllocatorAction, float64) {
 	if a.storePool == nil {
 		// Do nothing if storePool is nil for some unittests.
@@ -291,7 +303,7 @@ func (a *Allocator) ComputeAction(
 	have := len(rangeInfo.Desc.Replicas)
 	decommissioningReplicas := a.storePool.decommissioningReplicas(rangeInfo.Desc.RangeID, rangeInfo.Desc.Replicas)
 	clusterNodes := a.storePool.ClusterNodeCount()
-	need := GetNeededReplicas(zone.NumReplicas, clusterNodes)
+	need := GetNeededReplicas(*zone.NumReplicas, clusterNodes)
 	desiredQuorum := computeQuorum(need)
 	quorum := computeQuorum(have)
 
@@ -384,11 +396,11 @@ type decisionDetails struct {
 // a store.
 func (a *Allocator) AllocateTarget(
 	ctx context.Context,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	existing []roachpb.ReplicaDescriptor,
 	rangeInfo RangeInfo,
 ) (*roachpb.StoreDescriptor, string, error) {
-	sl, aliveStoreCount, throttledStoreCount := a.storePool.getStoreList(rangeInfo.Desc.RangeID, storeFilterThrottled)
+	sl, aliveStoreCount, throttled := a.storePool.getStoreList(rangeInfo.Desc.RangeID, storeFilterThrottled)
 
 	target, details := a.allocateTargetFromList(
 		ctx, sl, zone, existing, rangeInfo, a.scorerOptions())
@@ -399,21 +411,23 @@ func (a *Allocator) AllocateTarget(
 
 	// When there are throttled stores that do match, we shouldn't send
 	// the replica to purgatory.
-	if throttledStoreCount > 0 {
-		return nil, "", errors.Errorf("%d matching stores are currently throttled", throttledStoreCount)
+	if len(throttled) > 0 {
+		return nil, "", errors.Errorf(
+			"%d matching stores are currently throttled: %v", len(throttled), throttled,
+		)
 	}
 	return nil, "", &allocatorError{
 		constraints:      zone.Constraints,
 		existingReplicas: len(existing),
 		aliveStores:      aliveStoreCount,
-		throttledStores:  throttledStoreCount,
+		throttledStores:  len(throttled),
 	}
 }
 
 func (a *Allocator) allocateTargetFromList(
 	ctx context.Context,
 	sl StoreList,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	existing []roachpb.ReplicaDescriptor,
 	rangeInfo RangeInfo,
 	options scorerOptions,
@@ -440,7 +454,7 @@ func (a *Allocator) allocateTargetFromList(
 func (a Allocator) simulateRemoveTarget(
 	ctx context.Context,
 	targetStore roachpb.StoreID,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	candidates []roachpb.ReplicaDescriptor,
 	rangeInfo RangeInfo,
 ) (roachpb.ReplicaDescriptor, string, error) {
@@ -464,7 +478,7 @@ func (a Allocator) simulateRemoveTarget(
 // replicas.
 func (a Allocator) RemoveTarget(
 	ctx context.Context,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	candidates []roachpb.ReplicaDescriptor,
 	rangeInfo RangeInfo,
 ) (roachpb.ReplicaDescriptor, string, error) {
@@ -527,7 +541,7 @@ func (a Allocator) RemoveTarget(
 // under-utilized store.
 func (a Allocator) RebalanceTarget(
 	ctx context.Context,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	raftStatus *raft.Status,
 	rangeInfo RangeInfo,
 	filter storeFilter,
@@ -660,7 +674,7 @@ func (a *Allocator) scorerOptions() scorerOptions {
 // unless asked to do otherwise by the checkTransferLeaseSource parameter.
 func (a *Allocator) TransferLeaseTarget(
 	ctx context.Context,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	existing []roachpb.ReplicaDescriptor,
 	leaseStoreID roachpb.StoreID,
 	rangeID roachpb.RangeID,
@@ -807,7 +821,7 @@ func (a *Allocator) TransferLeaseTarget(
 // attributes.
 func (a *Allocator) ShouldTransferLease(
 	ctx context.Context,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	existing []roachpb.ReplicaDescriptor,
 	leaseStoreID roachpb.StoreID,
 	rangeID roachpb.RangeID,
@@ -1112,7 +1126,7 @@ func (a Allocator) shouldTransferLeaseWithoutStats(
 }
 
 func (a Allocator) preferredLeaseholders(
-	zone config.ZoneConfig, existing []roachpb.ReplicaDescriptor,
+	zone *config.ZoneConfig, existing []roachpb.ReplicaDescriptor,
 ) []roachpb.ReplicaDescriptor {
 	// Go one preference at a time. As soon as we've found replicas that match a
 	// preference, we don't need to look at the later preferences, because
