@@ -594,7 +594,7 @@ func rebalanceCandidates(
 			// ConvergesOnMean calculations below. This is subtle but important.
 			if nodeHasReplica(store.Node.NodeID, rangeInfo.Desc.Replicas) &&
 				!storeHasReplica(store.StoreID, rangeInfo.Desc.Replicas) {
-				log.Infof(ctx, "nodeHasReplica(n%d, %v)=true", store.Node.NodeID, rangeInfo.Desc.Replicas)
+				log.VEventf(ctx, 2, "nodeHasReplica(n%d, %v)=true", store.Node.NodeID, rangeInfo.Desc.Replicas)
 				continue
 			}
 			constraintsOK, necessary := rebalanceFromConstraintsCheck(
@@ -907,7 +907,7 @@ func analyzeConstraints(
 	ctx context.Context,
 	getStoreDescFn func(roachpb.StoreID) (roachpb.StoreDescriptor, bool),
 	existing []roachpb.ReplicaDescriptor,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 ) analyzedConstraints {
 	result := analyzedConstraints{
 		constraints: zone.Constraints,
@@ -933,7 +933,7 @@ func analyzeConstraints(
 			}
 		}
 	}
-	if constrainedReplicas > 0 && constrainedReplicas < zone.NumReplicas {
+	if constrainedReplicas > 0 && constrainedReplicas < *zone.NumReplicas {
 		result.unconstrainedReplicas = true
 	}
 	return result
