@@ -23,11 +23,11 @@ import (
 	"sort"
 	"testing"
 
-	"github.com/kr/pretty"
-
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
+	"github.com/gogo/protobuf/proto"
+	"github.com/kr/pretty"
 )
 
 type storeScore struct {
@@ -902,9 +902,9 @@ func TestAllocateConstraintsCheck(t *testing.T) {
 
 	for _, tc := range testCases {
 		t.Run(tc.name, func(t *testing.T) {
-			zone := config.ZoneConfig{
+			zone := &config.ZoneConfig{
 				Constraints: tc.constraints,
-				NumReplicas: tc.zoneNumReplicas,
+				NumReplicas: proto.Int32(tc.zoneNumReplicas),
 			}
 			analyzed := analyzeConstraints(
 				context.Background(), getTestStoreDesc, testStoreReplicas(tc.existing), zone)
@@ -1035,9 +1035,9 @@ func TestRemoveConstraintsCheck(t *testing.T) {
 					StoreID: storeID,
 				})
 			}
-			zone := config.ZoneConfig{
+			zone := &config.ZoneConfig{
 				Constraints: tc.constraints,
-				NumReplicas: tc.zoneNumReplicas,
+				NumReplicas: proto.Int32(tc.zoneNumReplicas),
 			}
 			analyzed := analyzeConstraints(context.Background(), getTestStoreDesc, existing, zone)
 			for storeID, expected := range tc.expected {
