@@ -32,8 +32,10 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/storage/spanset"
 	"github.com/cockroachdb/cockroach/pkg/storage/stateloader"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/cockroachdb/cockroach/pkg/util/log/logtags"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/pkg/errors"
 )
@@ -54,7 +56,7 @@ func declareKeysEndTransaction(
 	declareKeysWriteTransaction(desc, header, req, spans)
 	et := req.(*roachpb.EndTransactionRequest)
 	// The spans may extend beyond this Range, but it's ok for the
-	// purpose of the command queue. The parts in our Range will
+	// purpose of acquiring latches. The parts in our Range will
 	// be resolved eagerly.
 	for _, span := range et.IntentSpans {
 		spans.Add(spanset.SpanReadWrite, span)
@@ -171,118 +173,105 @@ func evalEndTransaction(
 	// Fetch existing transaction.
 	var existingTxn roachpb.Transaction
 	if ok, err := engine.MVCCGetProto(
-		ctx, batch, key, hlc.Timestamp{}, true /* consistent */, nil /* txn */, &existingTxn,
+		ctx, batch, key, hlc.Timestamp{}, &existingTxn, engine.MVCCGetOptions{},
 	); err != nil {
 		return result.Result{}, err
 	} else if !ok {
+		// No existing transaction record was found - create one by writing it
+		// below in updateTxnWithExternalIntents.
+		reply.Txn = h.Txn.Clone()
+
+		// Verify that it is safe to create the transaction record. We only need
+		// to perform this verification for commits. Rollbacks can always write
+		// an aborted txn record.
 		if args.Commit {
-			return result.Result{}, roachpb.NewTransactionNotFoundStatusError()
-		}
-		// For rollbacks, we don't consider not finding the txn record an error;
-		// we accept without fuss rollbacks for transactions where the txn record
-		// was never written (because the BeginTransaction's batch failed, or
-		// even where the BeginTransaction was never sent to server).
-		txn := h.Txn.Clone()
-		reply.Txn = &txn
-		return result.Result{}, nil
-	}
-	// We're using existingTxn on the reply, although it can be stale
-	// compared to the Transaction in the request (e.g. the Sequence,
-	// and various timestamps). We must be careful to update it with the
-	// supplied ba.Txn if we return it with an error which might be
-	// retried, as for example to avoid client-side serializable restart.
-	reply.Txn = &existingTxn
-
-	// Verify that we can either commit it or abort it (according
-	// to args.Commit), and also that the Timestamp and Epoch have
-	// not suffered regression.
-	switch reply.Txn.Status {
-	case roachpb.COMMITTED:
-		return result.Result{}, roachpb.NewTransactionStatusError("already committed")
-
-	case roachpb.ABORTED:
-		if !args.Commit {
-			// The transaction has already been aborted by other.
-			// Do not return TransactionAbortedError since the client anyway
-			// wanted to abort the transaction.
-			desc := cArgs.EvalCtx.Desc()
-			externalIntents, err := resolveLocalIntents(ctx, desc, batch, ms, *args, reply.Txn, cArgs.EvalCtx)
-			if err != nil {
+			if err := CanCreateTxnRecord(cArgs.EvalCtx, reply.Txn); err != nil {
 				return result.Result{}, err
 			}
-			if err := updateTxnWithExternalIntents(
-				ctx, batch, ms, *args, reply.Txn, externalIntents,
-			); err != nil {
-				return result.Result{}, err
-			}
-			// Use alwaysReturn==true because the transaction is definitely
-			// aborted, no matter what happens to this command.
-			return result.FromEndTxn(reply.Txn, true /* alwaysReturn */, args.Poison), nil
 		}
-		// If the transaction was previously aborted by a concurrent writer's
-		// push, any intents written are still open. It's only now that we know
-		// them, so we return them all for asynchronous resolution (we're
-		// currently not able to write on error, but see #1989).
-		//
-		// Similarly to above, use alwaysReturn==true. The caller isn't trying
-		// to abort, but the transaction is definitely aborted and its intents
-		// can go.
-		reply.Txn.Intents = args.IntentSpans
-		return result.FromEndTxn(reply.Txn, true /* alwaysReturn */, args.Poison),
-			roachpb.NewTransactionAbortedError(roachpb.ABORT_REASON_ABORTED_RECORD_FOUND)
+	} else {
+		// We're using existingTxn on the reply, although it can be stale
+		// compared to the Transaction in the request (e.g. the Sequence,
+		// and various timestamps). We must be careful to update it with the
+		// supplied ba.Txn if we return it with an error which might be
+		// retried, as for example to avoid client-side serializable restart.
+		reply.Txn = &existingTxn
 
-	case roachpb.PENDING:
-		if h.Txn.Epoch < reply.Txn.Epoch {
-			// TODO(tschottdorf): this leaves the Txn record (and more
-			// importantly, intents) dangling; we can't currently write on
-			// error. Would panic, but that makes TestEndTransactionWithErrors
-			// awkward.
-			return result.Result{}, roachpb.NewTransactionStatusError(
-				fmt.Sprintf("epoch regression: %d", h.Txn.Epoch),
-			)
-		} else if h.Txn.Epoch == reply.Txn.Epoch && reply.Txn.Timestamp.Less(h.Txn.OrigTimestamp) {
-			// The transaction record can only ever be pushed forward, so it's an
-			// error if somehow the transaction record has an earlier timestamp
-			// than the original transaction timestamp.
+		// Verify that we can either commit it or abort it (according
+		// to args.Commit), and also that the Timestamp and Epoch have
+		// not suffered regression.
+		switch reply.Txn.Status {
+		case roachpb.COMMITTED:
+			return result.Result{}, roachpb.NewTransactionCommittedStatusError()
 
-			// TODO(tschottdorf): see above comment on epoch regression.
+		case roachpb.ABORTED:
+			if !args.Commit {
+				// The transaction has already been aborted by other.
+				// Do not return TransactionAbortedError since the client anyway
+				// wanted to abort the transaction.
+				desc := cArgs.EvalCtx.Desc()
+				externalIntents, err := resolveLocalIntents(ctx, desc, batch, ms, *args, reply.Txn, cArgs.EvalCtx)
+				if err != nil {
+					return result.Result{}, err
+				}
+				if err := updateTxnWithExternalIntents(
+					ctx, batch, ms, *args, reply.Txn, externalIntents,
+				); err != nil {
+					return result.Result{}, err
+				}
+				// Use alwaysReturn==true because the transaction is definitely
+				// aborted, no matter what happens to this command.
+				return result.FromEndTxn(reply.Txn, true /* alwaysReturn */, args.Poison), nil
+			}
+			// If the transaction was previously aborted by a concurrent writer's
+			// push, any intents written are still open. It's only now that we know
+			// them, so we return them all for asynchronous resolution (we're
+			// currently not able to write on error, but see #1989).
+			//
+			// Similarly to above, use alwaysReturn==true. The caller isn't trying
+			// to abort, but the transaction is definitely aborted and its intents
+			// can go.
+			reply.Txn.Intents = args.IntentSpans
+			return result.FromEndTxn(reply.Txn, true /* alwaysReturn */, args.Poison),
+				roachpb.NewTransactionAbortedError(roachpb.ABORT_REASON_ABORTED_RECORD_FOUND)
+
+		case roachpb.PENDING:
+			if h.Txn.Epoch < reply.Txn.Epoch {
+				// TODO(tschottdorf): this leaves the Txn record (and more
+				// importantly, intents) dangling; we can't currently write on
+				// error. Would panic, but that makes TestEndTransactionWithErrors
+				// awkward.
+				return result.Result{}, roachpb.NewTransactionStatusError(
+					fmt.Sprintf("epoch regression: %d", h.Txn.Epoch),
+				)
+			} else if h.Txn.Epoch == reply.Txn.Epoch && reply.Txn.Timestamp.Less(h.Txn.OrigTimestamp) {
+				// The transaction record can only ever be pushed forward, so it's an
+				// error if somehow the transaction record has an earlier timestamp
+				// than the original transaction timestamp.
+
+				// TODO(tschottdorf): see above comment on epoch regression.
+				return result.Result{}, roachpb.NewTransactionStatusError(
+					fmt.Sprintf("timestamp regression: %s", h.Txn.OrigTimestamp),
+				)
+			}
+
+		default:
 			return result.Result{}, roachpb.NewTransactionStatusError(
-				fmt.Sprintf("timestamp regression: %s", h.Txn.OrigTimestamp),
+				fmt.Sprintf("bad txn status: %s", reply.Txn),
 			)
 		}
 
-	default:
-		return result.Result{}, roachpb.NewTransactionStatusError(
-			fmt.Sprintf("bad txn status: %s", reply.Txn),
-		)
+		// Update the existing txn with the supplied txn.
+		reply.Txn.Update(h.Txn)
 	}
 
-	// Update the existing txn with the supplied txn.
-	reply.Txn.Update(h.Txn)
-
 	var pd result.Result
 
 	// Set transaction status to COMMITTED or ABORTED as per the
 	// args.Commit parameter.
 	if args.Commit {
-		if retry, reason := IsEndTransactionTriggeringRetryError(reply.Txn, *args); retry {
-			return result.Result{}, roachpb.NewTransactionRetryError(reason)
-		}
-
-		if IsEndTransactionExceedingDeadline(reply.Txn.Timestamp, *args) {
-			// If the deadline has lapsed return an error and rely on the client
-			// issuing a Rollback() that aborts the transaction and cleans up
-			// intents. Unfortunately, we're returning an error and unable to
-			// write on error (see #1989): we can't write ABORTED into the master
-			// transaction record which remains PENDING, and thus rely on the
-			// client to issue a Rollback() for cleanup.
-			//
-			// N.B. This deadline test is expected to be a Noop for Serializable
-			// transactions; unless the client misconfigured the txn, the deadline can
-			// only be expired if the txn has been pushed, and pushed Serializable
-			// transactions are detected above.
-			return result.Result{}, roachpb.NewTransactionStatusError(
-				"transaction deadline exceeded")
+		if retry, reason, extraMsg := IsEndTransactionTriggeringRetryError(reply.Txn, *args); retry {
+			return result.Result{}, roachpb.NewTransactionRetryError(reason, extraMsg)
 		}
 
 		reply.Txn.Status = roachpb.COMMITTED
@@ -332,9 +321,9 @@ func evalEndTransaction(
 		}
 	}
 
-	// Note: there's no need to clear the AbortSpan state if we've
-	// successfully finalized a transaction, as there's no way in which an abort
-	// cache entry could have been written (the txn would already have been in
+	// Note: there's no need to clear the AbortSpan state if we've successfully
+	// finalized a transaction, as there's no way in which an abort cache entry
+	// could have been written (the txn would already have been in
 	// state=ABORTED).
 	//
 	// Summary of transaction replay protection after EndTransaction: When a
@@ -342,12 +331,14 @@ func evalEndTransaction(
 	// write will succeed but only as an intent with a newer timestamp (with a
 	// WriteTooOldError). However, the replayed intent cannot be resolved by a
 	// subsequent replay of this EndTransaction call because the txn timestamp
-	// will be too old. Replays which include a BeginTransaction never succeed
-	// because EndTransaction inserts in the write timestamp cache, forcing the
-	// BeginTransaction to fail with a transaction retry error. If the replay
-	// didn't include a BeginTransaction, any push will immediately succeed as a
-	// missing txn record on push sets the transaction to aborted. In both
-	// cases, the txn will be GC'd on the slow path.
+	// will be too old. Replays of requests which attempt to create a new txn
+	// record (BeginTransaction, HeartbeatTxn, or EndTransaction) never succeed
+	// because EndTransaction inserts in the write timestamp cache in Replica's
+	// updateTimestampCache method, forcing the call to CanCreateTxnRecord to
+	// return false, resulting in a transaction retry error. If the replay
+	// didn't attempt to create a txn record, any push will immediately succeed
+	// as a missing txn record on push where CanCreateTxnRecord returns false
+	// succeeds. In both cases, the txn will be GC'd on the slow path.
 	//
 	// We specify alwaysReturn==false because if the commit fails below Raft, we
 	// don't want the intents to be up for resolution. That should happen only
@@ -363,15 +354,16 @@ func evalEndTransaction(
 // IsEndTransactionExceedingDeadline returns true if the transaction
 // exceeded its deadline.
 func IsEndTransactionExceedingDeadline(t hlc.Timestamp, args roachpb.EndTransactionRequest) bool {
-	return args.Deadline != nil && args.Deadline.Less(t)
+	return args.Deadline != nil && !t.Less(*args.Deadline)
 }
 
 // IsEndTransactionTriggeringRetryError returns true if the
 // EndTransactionRequest cannot be committed and needs to return a
-// TransactionRetryError.
+// TransactionRetryError. It also returns the reason and possibly an extra
+// message to be used for the error.
 func IsEndTransactionTriggeringRetryError(
 	txn *roachpb.Transaction, args roachpb.EndTransactionRequest,
-) (retry bool, reason roachpb.TransactionRetryReason) {
+) (retry bool, reason roachpb.TransactionRetryReason, extraMsg string) {
 	// If we saw any WriteTooOldErrors, we must restart to avoid lost
 	// update anomalies.
 	if txn.WriteTooOld {
@@ -381,25 +373,31 @@ func IsEndTransactionTriggeringRetryError(
 		origTimestamp.Forward(txn.RefreshedTimestamp)
 		isTxnPushed := txn.Timestamp != origTimestamp
 
-		// If the isolation level is SERIALIZABLE, return a transaction
-		// retry error if the commit timestamp isn't equal to the txn
-		// timestamp.
+		// Return a transaction retry error if the commit timestamp isn't equal to
+		// the txn timestamp.
 		if isTxnPushed {
-			if txn.Isolation == enginepb.SERIALIZABLE {
-				retry, reason = true, roachpb.RETRY_SERIALIZABLE
-			} else if txn.RetryOnPush {
-				// If pushing requires a retry and the transaction was pushed, retry.
-				retry, reason = true, roachpb.RETRY_DELETE_RANGE
-			}
+			retry, reason = true, roachpb.RETRY_SERIALIZABLE
 		}
 	}
 
-	// A serializable transaction can still avoid a retry under certain conditions.
-	if retry && txn.IsSerializable() && canForwardSerializableTimestamp(txn, args.NoRefreshSpans) {
+	// A transaction can still avoid a retry under certain conditions.
+	if retry && canForwardSerializableTimestamp(txn, args.NoRefreshSpans) {
 		retry, reason = false, 0
 	}
 
-	return retry, reason
+	if !retry {
+		if IsEndTransactionExceedingDeadline(txn.Timestamp, args) {
+			exceededBy := txn.Timestamp.GoTime().Sub(args.Deadline.GoTime())
+			fromStart := txn.Timestamp.GoTime().Sub(txn.OrigTimestamp.GoTime())
+			extraMsg = fmt.Sprintf(
+				"txn timestamp pushed too much; deadline exceeded by %s (%s > %s), "+
+					"original timestamp %s ago (%s)",
+				exceededBy, txn.Timestamp, args.Deadline, fromStart, txn.OrigTimestamp)
+			retry, reason = true, roachpb.RETRY_COMMIT_DEADLINE_EXCEEDED
+		}
+	}
+
+	return retry, reason, extraMsg
 }
 
 // canForwardSerializableTimestamp returns whether a serializable txn can
@@ -494,7 +492,7 @@ func resolveLocalIntents(
 			}
 			return nil
 		}(); err != nil {
-			log.Fatalf(ctx, "error resolving intent at %s on end transaction [%s]: %s", span, txn.Status, err)
+			return nil, errors.Wrapf(err, "resolving intent at %s on end transaction [%s]", span, txn.Status)
 		}
 	}
 	// If the poison arg is set, make sure to set the abort span entry.
@@ -527,7 +525,8 @@ func updateTxnWithExternalIntents(
 		return engine.MVCCDelete(ctx, batch, ms, key, hlc.Timestamp{}, nil /* txn */)
 	}
 	txn.Intents = externalIntents
-	return engine.MVCCPutProto(ctx, batch, ms, key, hlc.Timestamp{}, nil /* txn */, txn)
+	txnRecord := txn.AsRecord()
+	return engine.MVCCPutProto(ctx, batch, ms, key, hlc.Timestamp{}, nil /* txn */, &txnRecord)
 }
 
 // RunCommitTrigger runs the commit trigger from an end transaction request.
@@ -763,9 +762,10 @@ func splitTrigger(
 	split *roachpb.SplitTrigger,
 	ts hlc.Timestamp,
 ) (enginepb.MVCCStats, result.Result, error) {
-	// TODO(tschottdorf): should have an incoming context from the corresponding
-	// EndTransaction, but the plumbing has not been done yet.
-	sp := rec.Tracer().StartSpan("split", tracing.LogTagsFromCtx(ctx))
+	// TODO(andrei): should this span be a child of the ctx's (if any)?
+	sp := rec.ClusterSettings().Tracer.StartRootSpan(
+		"split", logtags.FromContext(ctx), tracing.NonRecordableSpan,
+	)
 	defer sp.Finish()
 	desc := rec.Desc()
 	if !bytes.Equal(desc.StartKey, split.LeftDesc.StartKey) ||
@@ -907,6 +907,29 @@ func splitTrigger(
 			log.VEventf(ctx, 1, "LHS's TxnSpanGCThreshold of split is not set")
 		}
 
+		// We're about to write the initial state for the replica. We migrated
+		// the formerly replicated truncated state into unreplicated keyspace
+		// in 2.2., but this range may still be using the replicated version
+		// and we need to make a decision about what to use for the RHS that
+		// is consistent across the followers: do for the RHS what the LHS
+		// does: if the LHS has the legacy key, initialize the RHS with a
+		// legacy key as well.
+		//
+		// See VersionUnreplicatedRaftTruncatedState.
+		truncStateType := stateloader.TruncatedStateUnreplicated
+		if found, err := engine.MVCCGetProto(
+			ctx,
+			batch,
+			keys.RaftTruncatedStateLegacyKey(rec.GetRangeID()),
+			hlc.Timestamp{},
+			nil,
+			engine.MVCCGetOptions{},
+		); err != nil {
+			return enginepb.MVCCStats{}, result.Result{}, errors.Wrap(err, "unable to load legacy truncated state")
+		} else if found {
+			truncStateType = stateloader.TruncatedStateLegacyReplicated
+		}
+
 		// Writing the initial state is subtle since this also seeds the Raft
 		// group. It becomes more subtle due to proposer-evaluated Raft.
 		//
@@ -937,8 +960,10 @@ func splitTrigger(
 		// writeInitialReplicaState which essentially writes a ReplicaState
 		// only.
 		rightMS, err = stateloader.WriteInitialReplicaState(
-			ctx, rec.ClusterSettings(), batch, rightMS, split.RightDesc,
+			ctx, batch, rightMS, split.RightDesc,
 			rightLease, *gcThreshold, *txnSpanGCThreshold,
+			rec.ClusterSettings().Version.Version().Version,
+			truncStateType,
 		)
 		if err != nil {
 			return enginepb.MVCCStats{}, result.Result{}, errors.Wrap(err, "unable to write initial Replica state")
@@ -949,7 +974,7 @@ func splitTrigger(
 			// clobber downstream simply because that's what 1.0 does and if we
 			// don't write it here, then a 1.0 version applying it as a follower
 			// won't write a HardState at all and is guaranteed to crash.
-			rsl := stateloader.Make(rec.ClusterSettings(), split.RightDesc.RangeID)
+			rsl := stateloader.Make(split.RightDesc.RangeID)
 			if err := rsl.SynthesizeRaftState(ctx, batch); err != nil {
 				return enginepb.MVCCStats{}, result.Result{}, errors.Wrap(err, "unable to synthesize initial Raft state")
 			}
@@ -976,7 +1001,7 @@ func splitTrigger(
 	var pd result.Result
 	// This makes sure that no reads are happening in parallel; see #3148.
 	pd.Replicated.BlockReads = true
-	pd.Replicated.Split = &storagebase.Split{
+	pd.Replicated.Split = &storagepb.Split{
 		SplitTrigger: *split,
 		RHSDelta:     rightDeltaMS,
 	}
@@ -1032,7 +1057,7 @@ func mergeTrigger(
 
 	var pd result.Result
 	pd.Replicated.BlockReads = true
-	pd.Replicated.Merge = &storagebase.Merge{
+	pd.Replicated.Merge = &storagepb.Merge{
 		MergeTrigger: *merge,
 	}
 	return pd, nil
@@ -1067,10 +1092,10 @@ func changeReplicasTrigger(
 	cpy.NextReplicaID = change.NextReplicaID
 	// TODO(tschottdorf): duplication of Desc with the trigger below, should
 	// likely remove it from the trigger.
-	pd.Replicated.State = &storagebase.ReplicaState{
+	pd.Replicated.State = &storagepb.ReplicaState{
 		Desc: &cpy,
 	}
-	pd.Replicated.ChangeReplicas = &storagebase.ChangeReplicas{
+	pd.Replicated.ChangeReplicas = &storagepb.ChangeReplicas{
 		ChangeReplicasTrigger: *change,
 	}
 
