@@ -19,14 +19,13 @@ import (
 	"context"
 	"fmt"
 	"math"
+	"math/rand"
 	"reflect"
 	"sync"
 	"sync/atomic"
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
@@ -36,8 +35,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/storage/batcheval"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
@@ -49,6 +48,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/randutil"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/assert"
+	"github.com/stretchr/testify/require"
 )
 
 // TestRangeCommandClockUpdate verifies that followers update their
@@ -64,7 +66,13 @@ func TestRangeCommandClockUpdate(t *testing.T) {
 		manuals = append(manuals, hlc.NewManualClock(1))
 		clocks = append(clocks, hlc.NewClock(manuals[i].UnixNano, 100*time.Millisecond))
 	}
-	mtc := &multiTestContext{clocks: clocks}
+	mtc := &multiTestContext{
+		clocks: clocks,
+		// This test was written before the multiTestContext started creating many
+		// system ranges at startup, and hasn't been update to take that into
+		// account.
+		startWithSingleRange: true,
+	}
 	defer mtc.Stop()
 	mtc.Start(t, numNodes)
 	mtc.replicateRange(1, 1, 2)
@@ -82,7 +90,8 @@ func TestRangeCommandClockUpdate(t *testing.T) {
 	testutils.SucceedsSoon(t, func() error {
 		values := []int64{}
 		for _, eng := range mtc.engines {
-			val, _, err := engine.MVCCGet(context.Background(), eng, roachpb.Key("a"), clocks[0].Now(), true, nil)
+			val, _, err := engine.MVCCGet(context.Background(), eng, roachpb.Key("a"), clocks[0].Now(),
+				engine.MVCCGetOptions{})
 			if err != nil {
 				return err
 			}
@@ -148,7 +157,8 @@ func TestRejectFutureCommand(t *testing.T) {
 	if advance := ts3.GoTime().Sub(ts2.GoTime()); advance != 0 {
 		t.Fatalf("expected clock not to advance, but it advanced by %s", advance)
 	}
-	val, _, err := engine.MVCCGet(context.Background(), mtc.engines[0], key, ts3, true, nil)
+	val, _, err := engine.MVCCGet(context.Background(), mtc.engines[0], key, ts3,
+		engine.MVCCGetOptions{})
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -221,10 +231,8 @@ func TestTxnPutOutOfOrder(t *testing.T) {
 		}
 	eng := engine.NewInMem(roachpb.Attributes{}, 10<<20)
 	stopper.AddCloser(eng)
-	store := createTestStoreWithEngine(t,
-		eng,
-		true,
-		cfg,
+	store := createTestStoreWithOpts(t,
+		testStoreOpts{eng: eng, cfg: &cfg},
 		stopper,
 	)
 
@@ -362,7 +370,15 @@ func TestRangeLookupUseReverse(t *testing.T) {
 	storeCfg.TestingKnobs.DisableMergeQueue = true
 	stopper := stop.NewStopper()
 	defer stopper.Stop(context.TODO())
-	store := createTestStoreWithConfig(t, stopper, storeCfg)
+	store := createTestStoreWithOpts(
+		t,
+		testStoreOpts{
+			// This test was written before the test stores were able to start with
+			// more than one range and is not prepared to handle many ranges.
+			dontCreateSystemRanges: true,
+			cfg:                    &storeCfg,
+		},
+		stopper)
 
 	// Init test ranges:
 	// ["","a"), ["a","c"), ["c","e"), ["e","g") and ["g","\xff\xff").
@@ -529,6 +545,9 @@ func setupLeaseTransferTest(t *testing.T) *leaseTransferTest {
 	}
 
 	l.mtc = &multiTestContext{}
+	// This test was written before the multiTestContext started creating many
+	// system ranges at startup, and hasn't been update to take that into account.
+	l.mtc.startWithSingleRange = true
 	l.mtc.storeConfig = &cfg
 	l.mtc.Start(t, 2)
 	l.mtc.initGossipNetwork()
@@ -895,7 +914,7 @@ func TestRangeLimitTxnMaxTimestamp(t *testing.T) {
 	mtc.clocks = []*hlc.Clock{clock1, clock2}
 
 	// Start a transaction using node2 as a gateway.
-	txn := roachpb.MakeTransaction("test", keyA, 1, enginepb.SERIALIZABLE, clock2.Now(), 250 /* maxOffsetNs */)
+	txn := roachpb.MakeTransaction("test", keyA, 1, clock2.Now(), 250 /* maxOffsetNs */)
 	// Simulate a read to another range on node2 by setting the observed timestamp.
 	txn.UpdateObservedTimestamp(2, clock2.Now())
 
@@ -967,7 +986,13 @@ func TestLeaseMetricsOnSplitAndTransfer(t *testing.T) {
 			}
 			return nil
 		}
-	mtc := &multiTestContext{storeConfig: &sc}
+	mtc := &multiTestContext{
+		storeConfig: &sc,
+		// This test was written before the multiTestContext started creating many
+		// system ranges at startup, and hasn't been update to take that into
+		// account.
+		startWithSingleRange: true,
+	}
 	defer mtc.Stop()
 	mtc.Start(t, 2)
 
@@ -1053,6 +1078,9 @@ func TestLeaseMetricsOnSplitAndTransfer(t *testing.T) {
 // See replica.mu.minLeaseProposedTS for the reasons why this isn't allowed.
 func TestLeaseNotUsedAfterRestart(t *testing.T) {
 	defer leaktest.AfterTest(t)()
+
+	t.Skip("https://github.com/cockroachdb/cockroach/issues/34111")
+
 	ctx := context.Background()
 
 	sc := storage.TestStoreConfig(nil)
@@ -1134,10 +1162,10 @@ func TestLeaseNotUsedAfterRestart(t *testing.T) {
 // lease holder) is not blocked by ongoing reads.
 // The test relies on two things:
 // 1) Lease extensions, unlike lease transfers, are not blocked by reads through
-// their ReplicatedEvalResult.BlockReads.
+//    their ReplicatedEvalResult.BlockReads.
 // 2) Requests such as RequestLeaseRequest don't declare to touch the whole key
-// span of the range, and thus don't conflict through the command queue with
-// other reads.
+//    span of the range, and thus don't acquire latches that conflict with other
+//    reads.
 func TestLeaseExtensionNotBlockedByRead(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	readBlocked := make(chan struct{})
@@ -1154,7 +1182,7 @@ func TestLeaseExtensionNotBlockedByRead(t *testing.T) {
 		base.TestServerArgs{
 			Knobs: base.TestingKnobs{
 				Store: &storage.StoreTestingKnobs{
-					EvalKnobs: batcheval.TestingKnobs{
+					EvalKnobs: storagebase.BatchEvalTestingKnobs{
 						TestingEvalFilter: cmdFilter,
 					},
 				},
@@ -1202,11 +1230,6 @@ func TestLeaseExtensionNotBlockedByRead(t *testing.T) {
 			t.Fatalf("replica descriptor for key %s not found", rKey)
 		}
 
-		curLease, _, err := s.GetRangeLease(context.TODO(), key)
-		if err != nil {
-			t.Fatal(err)
-		}
-
 		leaseReq := roachpb.RequestLeaseRequest{
 			RequestHeader: roachpb.RequestHeader{
 				Key: key,
@@ -1216,10 +1239,30 @@ func TestLeaseExtensionNotBlockedByRead(t *testing.T) {
 				Expiration: s.Clock().Now().Add(time.Second.Nanoseconds(), 0).Clone(),
 				Replica:    replDesc,
 			},
-			PrevLease: curLease,
 		}
-		if _, pErr := client.SendWrapped(context.Background(), s.DB().NonTransactionalSender(), &leaseReq); pErr != nil {
-			t.Fatal(pErr)
+
+		for {
+			curLease, _, err := s.GetRangeLease(context.TODO(), key)
+			if err != nil {
+				t.Fatal(err)
+			}
+			leaseReq.PrevLease = curLease
+
+			_, pErr := client.SendWrapped(context.Background(), s.DB().NonTransactionalSender(), &leaseReq)
+			if _, ok := pErr.GetDetail().(*roachpb.AmbiguousResultError); ok {
+				log.Infof(context.Background(), "retrying lease after %s", pErr)
+				continue
+			}
+			if _, ok := pErr.GetDetail().(*roachpb.LeaseRejectedError); ok {
+				// Lease rejected? Try again. The extension should work because
+				// extending is idempotent (assuming the PrevLease matches).
+				log.Infof(context.Background(), "retrying lease after %s", pErr)
+				continue
+			}
+			if pErr != nil {
+				t.Errorf("%T %s", pErr.GetDetail(), pErr) // NB: don't fatal or shutdown hangs
+			}
+			break
 		}
 		// Unblock the read.
 		readBlocked <- struct{}{}
@@ -1249,10 +1292,7 @@ func LeaseInfo(
 
 func TestLeaseInfoRequest(t *testing.T) {
 	defer leaktest.AfterTest(t)()
-	tc := testcluster.StartTestCluster(t, 3,
-		base.TestClusterArgs{
-			ReplicationMode: base.ReplicationManual,
-		})
+	tc := testcluster.StartTestCluster(t, 3, base.TestClusterArgs{})
 	defer tc.Stopper().Stop(context.TODO())
 
 	kvDB0 := tc.Servers[0].DB()
@@ -1263,16 +1303,6 @@ func TestLeaseInfoRequest(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-
-	rangeDesc, err = tc.AddReplicas(
-		rangeDesc.StartKey.AsRawKey(), tc.Target(1), tc.Target(2),
-	)
-	if err != nil {
-		t.Fatal(err)
-	}
-	if len(rangeDesc.Replicas) != 3 {
-		t.Fatalf("expected 3 replicas, got %+v", rangeDesc.Replicas)
-	}
 	replicas := make([]roachpb.ReplicaDescriptor, 3)
 	for i := 0; i < 3; i++ {
 		var ok bool
@@ -1282,12 +1312,24 @@ func TestLeaseInfoRequest(t *testing.T) {
 		}
 	}
 
-	// Lease should start on Server 0, since nobody told it to move.
-	leaseHolderReplica := LeaseInfo(t, kvDB0, rangeDesc, roachpb.INCONSISTENT).Lease.Replica
-	if leaseHolderReplica != replicas[0] {
-		t.Fatalf("lease holder should be replica %+v, but is: %+v", replicas[0], leaseHolderReplica)
+	// Transfer the lease to Servers[0] so we start in a known state. Otherwise,
+	// there might be already a lease owned by a random node.
+	err = tc.TransferRangeLease(rangeDesc, tc.Target(0))
+	if err != nil {
+		t.Fatal(err)
 	}
 
+	// Now test the LeaseInfo. We might need to loop until the node we query has
+	// applied the lease.
+	testutils.SucceedsSoon(t, func() error {
+		leaseHolderReplica := LeaseInfo(t, kvDB0, rangeDesc, roachpb.INCONSISTENT).Lease.Replica
+		if leaseHolderReplica != replicas[0] {
+			return fmt.Errorf("lease holder should be replica %+v, but is: %+v",
+				replicas[0], leaseHolderReplica)
+		}
+		return nil
+	})
+
 	// Transfer the lease to Server 1 and check that LeaseInfoRequest gets the
 	// right answer.
 	err = tc.TransferRangeLease(rangeDesc, tc.Target(1))
@@ -1297,7 +1339,7 @@ func TestLeaseInfoRequest(t *testing.T) {
 	// An inconsistent LeaseInfoReqeust on the old lease holder should give us the
 	// right answer immediately, since the old holder has definitely applied the
 	// transfer before TransferRangeLease returned.
-	leaseHolderReplica = LeaseInfo(t, kvDB0, rangeDesc, roachpb.INCONSISTENT).Lease.Replica
+	leaseHolderReplica := LeaseInfo(t, kvDB0, rangeDesc, roachpb.INCONSISTENT).Lease.Replica
 	if leaseHolderReplica != replicas[1] {
 		t.Fatalf("lease holder should be replica %+v, but is: %+v",
 			replicas[1], leaseHolderReplica)
@@ -1324,9 +1366,34 @@ func TestLeaseInfoRequest(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	leaseHolderReplica = LeaseInfo(t, kvDB1, rangeDesc, roachpb.INCONSISTENT).Lease.Replica
+
+	// We're now going to ask servers[1] for the lease info. We don't use kvDB1;
+	// instead we go directly to the store because otherwise the DistSender might
+	// use an old, cached, version of the range descriptor that doesn't have the
+	// local replica in it (and so the request would be routed away).
+	// TODO(andrei): Add a batch option to not use the range cache.
+	s, err := tc.Servers[1].Stores().GetStore(tc.Servers[1].GetFirstStoreID())
+	if err != nil {
+		t.Fatal(err)
+	}
+	leaseInfoReq := &roachpb.LeaseInfoRequest{
+		RequestHeader: roachpb.RequestHeader{
+			Key: rangeDesc.StartKey.AsRawKey(),
+		},
+	}
+	reply, pErr := client.SendWrappedWith(
+		context.Background(), s, roachpb.Header{
+			RangeID:         rangeDesc.RangeID,
+			ReadConsistency: roachpb.INCONSISTENT,
+		}, leaseInfoReq)
+	if pErr != nil {
+		t.Fatal(pErr)
+	}
+	resp := *(reply.(*roachpb.LeaseInfoResponse))
+	leaseHolderReplica = resp.Lease.Replica
+
 	if leaseHolderReplica != replicas[2] {
-		t.Fatalf("lease holder should be replica %+v, but is: %+v", replicas[2], leaseHolderReplica)
+		t.Fatalf("lease holder should be replica %s, but is: %s", replicas[2], leaseHolderReplica)
 	}
 
 	// TODO(andrei): test the side-effect of LeaseInfoRequest when there's no
@@ -1352,7 +1419,7 @@ func TestErrorHandlingForNonKVCommand(t *testing.T) {
 		base.TestServerArgs{
 			Knobs: base.TestingKnobs{
 				Store: &storage.StoreTestingKnobs{
-					EvalKnobs: batcheval.TestingKnobs{
+					EvalKnobs: storagebase.BatchEvalTestingKnobs{
 						TestingEvalFilter: cmdFilter,
 					},
 				},
@@ -1385,6 +1452,10 @@ func TestRangeInfo(t *testing.T) {
 	storeCfg.TestingKnobs.DisableMergeQueue = true
 	mtc := &multiTestContext{
 		storeConfig: &storeCfg,
+		// This test was written before the multiTestContext started creating many
+		// system ranges at startup, and hasn't been update to take that into
+		// account.
+		startWithSingleRange: true,
 	}
 	defer mtc.Stop()
 	mtc.Start(t, 2)
@@ -1478,7 +1549,7 @@ func TestRangeInfo(t *testing.T) {
 			EndKey: roachpb.KeyMax,
 		},
 	}
-	txn := roachpb.MakeTransaction("test", roachpb.KeyMin, 1, enginepb.SERIALIZABLE, mtc.clock.Now(), 0)
+	txn := roachpb.MakeTransaction("test", roachpb.KeyMin, 1, mtc.clock.Now(), 0)
 	h.Txn = &txn
 	reply, pErr = client.SendWrappedWith(context.Background(), mtc.distSenders[0], h, &scanArgs)
 	if pErr != nil {
@@ -1538,8 +1609,11 @@ func TestRangeInfo(t *testing.T) {
 	if pErr != nil {
 		t.Fatal(pErr)
 	}
-	lhsLease, _ = lhsReplica1.GetLease()
-	rhsLease, _ = rhsReplica1.GetLease()
+	// Read the expected lease from replica0 rather than replica1 as it may serve
+	// a follower read which will contain the new lease information before
+	// replica1 has applied the lease transfer.
+	lhsLease, _ = lhsReplica0.GetLease()
+	rhsLease, _ = rhsReplica0.GetLease()
 	expRangeInfos = []roachpb.RangeInfo{
 		{
 			Desc:  *lhsReplica1.Desc(),
@@ -1570,7 +1644,7 @@ func TestDrainRangeRejection(t *testing.T) {
 
 	drainingIdx := 1
 	mtc.stores[drainingIdx].SetDraining(true)
-	if err := repl.ChangeReplicas(
+	if _, err := repl.ChangeReplicas(
 		context.Background(),
 		roachpb.ADD_REPLICA,
 		roachpb.ReplicationTarget{
@@ -1578,7 +1652,7 @@ func TestDrainRangeRejection(t *testing.T) {
 			StoreID: mtc.idents[drainingIdx].StoreID,
 		},
 		repl.Desc(),
-		storage.ReasonRangeUnderReplicated,
+		storagepb.ReasonRangeUnderReplicated,
 		"",
 	); !testutils.IsError(err, "store is draining") {
 		t.Fatalf("unexpected error: %v", err)
@@ -1595,29 +1669,37 @@ func TestSystemZoneConfigs(t *testing.T) {
 		t.Skip()
 	}
 
+	ctx := context.Background()
 	tc := testcluster.StartTestCluster(t, 7, base.TestClusterArgs{
 		ServerArgs: base.TestServerArgs{
+			Knobs: base.TestingKnobs{
+				Store: &storage.StoreTestingKnobs{
+					// Disable LBS because when the scan is happening at the rate it's happening
+					// below, it's possible that one of the system ranges trigger a split.
+					DisableLoadBasedSplitting: true,
+				},
+			},
 			// Scan like a bat out of hell to ensure replication and replica GC
 			// happen in a timely manner.
 			ScanInterval: 50 * time.Millisecond,
 		},
 	})
-	ctx := context.TODO()
 	defer tc.Stopper().Stop(ctx)
 	log.Info(ctx, "TestSystemZoneConfig: test cluster started")
+
 	expectedSystemRanges, err := tc.Servers[0].ExpectedInitialRangeCount()
 	if err != nil {
 		t.Fatal(err)
 	}
-
-	expectedUserRanges, err := tc.Servers[0].ExpectedInitialUserRangeCount()
-	if err != nil {
-		t.Fatal(err)
-	}
-
+	expectedUserRanges := 1
 	expectedSystemRanges -= expectedUserRanges
-	expectedReplicas := expectedSystemRanges*int(config.DefaultSystemZoneConfig().NumReplicas) +
-		expectedUserRanges*int(config.DefaultZoneConfig().NumReplicas)
+	systemNumReplicas := int(*config.DefaultSystemZoneConfig().NumReplicas)
+	userNumReplicas := int(*config.DefaultZoneConfig().NumReplicas)
+	expectedReplicas := expectedSystemRanges*systemNumReplicas + expectedUserRanges*userNumReplicas
+	log.Infof(ctx, "TestSystemZoneConfig: expecting %d system ranges and %d user ranges",
+		expectedSystemRanges, expectedUserRanges)
+	log.Infof(ctx, "TestSystemZoneConfig: expected (%dx%d) + (%dx%d) = %d replicas total",
+		expectedSystemRanges, systemNumReplicas, expectedUserRanges, userNumReplicas, expectedReplicas)
 
 	waitForReplicas := func() error {
 		var conflictingID roachpb.RangeID
@@ -1672,7 +1754,7 @@ func TestSystemZoneConfigs(t *testing.T) {
 	// Finally, verify the system ranges. Note that in a new cluster there are
 	// two system ranges, which we have to take into account here.
 	sqlutils.SetZoneConfig(t, sqlDB, "RANGE system", "num_replicas: 7")
-	expectedReplicas += 6
+	expectedReplicas += 4
 	testutils.SucceedsSoon(t, waitForReplicas)
 	log.Info(ctx, "TestSystemZoneConfig: up-replication of system ranges succeeded")
 }
@@ -1749,3 +1831,415 @@ func TestClearRange(t *testing.T) {
 	verifyKeysWithPrefix(keys.LocalStoreSuggestedCompactionsMin,
 		[]roachpb.Key{keys.StoreSuggestedCompactionKey(lg1, lg3)})
 }
+
+// TestLeaseTransferInSnapshotUpdatesTimestampCache prevents a regression of
+// #34025. A Replica is targeted for a lease transfer target when it needs a
+// Raft snapshot to catch up. Normally we try to prevent this case, but it is
+// possible and hard to prevent entirely. The Replica will only learn that it is
+// the new leaseholder when it applies the snapshot. When doing so, it should
+// make sure to apply the lease-related side-effects to its in-memory state.
+func TestLeaseTransferInSnapshotUpdatesTimestampCache(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	ctx := context.Background()
+	sc := storage.TestStoreConfig(nil)
+	// We'll control replication by hand.
+	sc.TestingKnobs.DisableReplicateQueue = true
+	// Avoid fighting with the merge queue while trying to reproduce this race.
+	sc.TestingKnobs.DisableMergeQueue = true
+	mtc := &multiTestContext{storeConfig: &sc}
+	defer mtc.Stop()
+	mtc.Start(t, 3)
+	store2 := mtc.Store(2)
+
+	keyA := roachpb.Key("a")
+	keyB := roachpb.Key("b")
+	keyC := roachpb.Key("c")
+
+	// First, do a couple of writes; we'll use these to determine when
+	// the dust has settled.
+	incA := incrementArgs(keyA, 1)
+	if _, pErr := client.SendWrapped(ctx, mtc.stores[0].TestSender(), incA); pErr != nil {
+		t.Fatal(pErr)
+	}
+	incC := incrementArgs(keyC, 2)
+	if _, pErr := client.SendWrapped(ctx, mtc.stores[0].TestSender(), incC); pErr != nil {
+		t.Fatal(pErr)
+	}
+
+	// Split the system range from the rest of the keyspace.
+	splitArgs := adminSplitArgs(keys.SystemMax)
+	if _, pErr := client.SendWrapped(ctx, mtc.stores[0].TestSender(), splitArgs); pErr != nil {
+		t.Fatal(pErr)
+	}
+
+	// Get the range's ID.
+	repl0 := mtc.stores[0].LookupReplica(roachpb.RKey(keyA))
+	rangeID := repl0.RangeID
+
+	// Replicate the range onto nodes 1 and 2.
+	// Wait for all replicas to be caught up.
+	mtc.replicateRange(rangeID, 1, 2)
+	mtc.waitForValues(keyA, []int64{1, 1, 1})
+	mtc.waitForValues(keyC, []int64{2, 2, 2})
+
+	// Create a transaction that will try to write "under" a served read.
+	// The read will have been served by the original leaseholder (node 0)
+	// and the write will be attempted on the new leaseholder (node 2).
+	// It should not succeed because it should run into the timestamp cache.
+	db := mtc.dbs[0]
+	txnOld := client.NewTxn(ctx, db, 0 /* gatewayNodeID */, client.RootTxn)
+
+	// Perform a write with txnOld so that its timestamp gets set.
+	if _, err := txnOld.Inc(ctx, keyB, 3); err != nil {
+		t.Fatal(err)
+	}
+
+	// Read keyC with txnOld, which is updated below. This prevents the
+	// transaction from refreshing when it hits the serializable error.
+	if _, err := txnOld.Get(ctx, keyC); err != nil {
+		t.Fatal(err)
+	}
+
+	// Ensure that the transaction sends its first hearbeat so that it creates
+	// its transaction record and doesn't run into trouble with the low water
+	// mark of the new leaseholder's timestamp cache. Amusingly, if the bug
+	// we're regression testing against here still existed, we would not have
+	// to do this.
+	hb, hbH := heartbeatArgs(txnOld.Serialize(), mtc.clock.Now())
+	if _, pErr := client.SendWrappedWith(ctx, mtc.stores[0].TestSender(), hbH, hb); pErr != nil {
+		t.Fatal(pErr)
+	}
+
+	// Another client comes along at a higher timestamp and reads. We should
+	// never be able to write under this time or we would be rewriting history.
+	if _, err := db.Get(ctx, keyA); err != nil {
+		t.Fatal(err)
+	}
+
+	// Partition node 2 from the rest of its range. Once partitioned, perform
+	// another write and truncate the Raft log on the two connected nodes. This
+	// ensures that that when node 2 comes back up it will require a snapshot
+	// from Raft.
+	mtc.transport.Listen(store2.Ident.StoreID, &unreliableRaftHandler{
+		rangeID:            rangeID,
+		RaftMessageHandler: store2,
+	})
+
+	if _, pErr := client.SendWrapped(ctx, mtc.stores[0].TestSender(), incC); pErr != nil {
+		t.Fatal(pErr)
+	}
+	mtc.waitForValues(keyC, []int64{4, 4, 2})
+
+	// Truncate the log at index+1 (log entries < N are removed, so this
+	// includes the increment). This necessitates a snapshot when the
+	// partitioned replica rejoins the rest of the range.
+	index, err := repl0.GetLastIndex()
+	if err != nil {
+		t.Fatal(err)
+	}
+	truncArgs := truncateLogArgs(index+1, rangeID)
+	truncArgs.Key = keyA
+	if _, err := client.SendWrapped(ctx, mtc.stores[0].TestSender(), truncArgs); err != nil {
+		t.Fatal(err)
+	}
+
+	// Finally, transfer the lease to node 2 while it is still unavailable and
+	// behind. We try to avoid this case when picking new leaseholders in practice,
+	// but we're never 100% successful.
+	if err := repl0.AdminTransferLease(ctx, store2.Ident.StoreID); err != nil {
+		t.Fatal(err)
+	}
+
+	// Remove the partition. A snapshot to node 2 should follow. This snapshot
+	// will inform node 2 that it is the new leaseholder for the range. Node 2
+	// should act accordingly and update its internal state to reflect this.
+	mtc.transport.Listen(store2.Ident.StoreID, store2)
+	mtc.waitForValues(keyC, []int64{4, 4, 4})
+
+	// Perform a write on the new leaseholder underneath the previously served
+	// read. This write should hit the timestamp cache and flag the txn for a
+	// restart when we try to commit it below. With the bug in #34025, the new
+	// leaseholder who heard about the lease transfer from a snapshot had an
+	// empty timestamp cache and would simply let us write under the previous
+	// read.
+	if _, err := txnOld.Inc(ctx, keyA, 4); err != nil {
+		t.Fatal(err)
+	}
+	const exp = `TransactionRetryError: retry txn \(RETRY_SERIALIZABLE\)`
+	if err := txnOld.Commit(ctx); !testutils.IsError(err, exp) {
+		t.Fatalf("expected retry error, got: %v; did we write under a read?", err)
+	}
+}
+
+// TestConcurrentAdminChangeReplicasRequests ensures that when two attempts to
+// change replicas for a range race, only one will succeed.
+func TestConcurrentAdminChangeReplicasRequests(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	// With 5 nodes the test is set up to have 2 actors trying to change the
+	// replication concurrently. The first one attempts to change the replication
+	// from [1] to [1, 2, 3, 4] and the second one starts by assuming that the
+	// first actor succeeded on its first request and expected [1, 2] and tries
+	// to move the replication to [1, 2, 4, 5]. One of these actors should
+	// succeed.
+	const numNodes = 5
+	tc := testcluster.StartTestCluster(t, numNodes, base.TestClusterArgs{
+		ReplicationMode: base.ReplicationManual,
+	})
+	ctx := context.Background()
+	defer tc.Stopper().Stop(ctx)
+	key := roachpb.Key("a")
+	db := tc.Servers[0].DB()
+	rangeInfo, err := getRangeInfo(ctx, db, key)
+	require.Nil(t, err)
+	require.Len(t, rangeInfo.Desc.Replicas, 1)
+	targets1, targets2 := makeReplicationTargets(2, 3, 4), makeReplicationTargets(4, 5)
+	expects1 := rangeInfo.Desc
+	expects2 := rangeInfo.Desc
+	expects2.Replicas = append(expects2.Replicas, roachpb.ReplicaDescriptor{
+		NodeID:    2,
+		StoreID:   2,
+		ReplicaID: expects2.NextReplicaID,
+	})
+	expects2.NextReplicaID++
+	var err1, err2 error
+	var res1, res2 *roachpb.RangeDescriptor
+	var wg sync.WaitGroup
+	wg.Add(2)
+	go func() {
+		res1, err1 = db.AdminChangeReplicas(
+			ctx, key, roachpb.ADD_REPLICA, targets1, expects1)
+		wg.Done()
+	}()
+	go func() {
+		res2, err2 = db.AdminChangeReplicas(
+			ctx, key, roachpb.ADD_REPLICA, targets2, expects2)
+		wg.Done()
+	}()
+	wg.Wait()
+
+	infoAfter, err := getRangeInfo(ctx, db, key)
+	require.Nil(t, err)
+
+	assert.Falsef(t, err1 == nil && err2 == nil,
+		"expected one of racing AdminChangeReplicasRequests to fail but neither did")
+	// It is possible that an error can occur due to a rejected preemptive
+	// snapshot from the target range. We don't want to fail the test if we got
+	// one of those.
+	isSnapshotErr := func(err error) bool {
+		return testutils.IsError(err, "snapshot failed:")
+	}
+	atLeastOneIsSnapshotErr := isSnapshotErr(err1) || isSnapshotErr(err2)
+	assert.Falsef(t, err1 != nil && err2 != nil && !atLeastOneIsSnapshotErr,
+		"expected only one of racing AdminChangeReplicasRequests to fail but both "+
+			"had errors and neither were snapshot: %v %v", err1, err2)
+	replicaNodeIDs := func(desc roachpb.RangeDescriptor) (ids []int) {
+		for _, r := range desc.Replicas {
+			ids = append(ids, int(r.NodeID))
+		}
+		return ids
+	}
+	if err1 == nil {
+		assert.ElementsMatch(t, replicaNodeIDs(infoAfter.Desc), []int{1, 2, 3, 4})
+		assert.EqualValues(t, infoAfter.Desc, *res1)
+	} else if err2 == nil {
+		assert.ElementsMatch(t, replicaNodeIDs(infoAfter.Desc), []int{1, 2, 4, 5})
+		assert.EqualValues(t, infoAfter.Desc, *res2)
+	}
+}
+
+// TestRandomConcurrentAdminChangeReplicasRequests ensures that when multiple
+// AdminChangeReplicasRequests are issued concurrently, so long as requests
+// provide the the value of the RangeDescriptor they will not accidentally
+// perform replication changes. In particular this test runs a number of
+// concurrent actors which all use the same expectations of the RangeDescriptor
+// and verifies that at most one actor succeeds in making all of its changes.
+func TestRandomConcurrentAdminChangeReplicasRequests(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	const numNodes = 6
+	tc := testcluster.StartTestCluster(t, numNodes, base.TestClusterArgs{
+		ReplicationMode: base.ReplicationManual,
+	})
+	ctx := context.Background()
+	defer tc.Stopper().Stop(ctx)
+	const actors = 10
+	errors := make([]error, actors)
+	var wg sync.WaitGroup
+	key := roachpb.Key("a")
+	db := tc.Servers[0].DB()
+	require.Nil(t, db.AdminRelocateRange(ctx, key, makeReplicationTargets(1, 2, 3)))
+	// Random targets consisting of a random number of nodes from the set of nodes
+	// in the cluster which currently do not have a replica.
+	pickTargets := func() []roachpb.ReplicationTarget {
+		availableIDs := make([]int, 0, numNodes-3)
+		for id := 4; id <= numNodes; id++ {
+			availableIDs = append(availableIDs, id)
+		}
+		rand.Shuffle(len(availableIDs), func(i, j int) {
+			availableIDs[i], availableIDs[j] = availableIDs[j], availableIDs[i]
+		})
+		n := rand.Intn(len(availableIDs)) + 1
+		return makeReplicationTargets(availableIDs[:n]...)
+	}
+	// TODO(ajwerner): consider doing this read inside the addReplicas function
+	// and then allowing multiple writes to overlap and validate that the state
+	// corresponds to a valid history of events.
+	rangeInfo, err := getRangeInfo(ctx, db, key)
+	require.Nil(t, err)
+	addReplicas := func() error {
+		_, err := db.AdminChangeReplicas(
+			ctx, key, roachpb.ADD_REPLICA, pickTargets(), rangeInfo.Desc)
+		return err
+	}
+	wg.Add(actors)
+	for i := 0; i < actors; i++ {
+		go func(i int) { errors[i] = addReplicas(); wg.Done() }(i)
+	}
+	wg.Wait()
+	var gotSuccess bool
+	for _, err := range errors {
+		if err != nil {
+			const exp = "change replicas of .* failed: descriptor changed" +
+				"|snapshot failed:"
+			assert.True(t, testutils.IsError(err, exp), err)
+		} else if gotSuccess {
+			t.Error("expected only one success")
+		} else {
+			gotSuccess = true
+		}
+	}
+}
+
+// TestAdminRelocateRangeSafety exercises a situation where calls to
+// AdminRelocateRange can race with calls to ChangeReplicas and verifies
+// that such races do not leave the range in an under-replicated state.
+func TestAdminRelocateRangeSafety(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	// The test is going to verify that when a replica removal due to a
+	// Replica.ChangeReplicas call coincides with the removal phase of an
+	// AdminRelocateRangeRequest that one of the removals will fail.
+	// In order to ensure that the AdminChangeReplicas command coincides with
+	// the remove phase of the AdminRelocateReplicas the test injects a response
+	// filter which, when useSeenAdd holds true, signals on seenAdd when it sees
+	// an AdminChangeReplicasRequest which added a replica.
+	const numNodes = 4
+	var useSeenAdd atomic.Value
+	useSeenAdd.Store(false)
+	seenAdd := make(chan struct{}, 1)
+	responseFilter := func(ba roachpb.BatchRequest, _ *roachpb.BatchResponse) *roachpb.Error {
+		if ba.IsSingleRequest() {
+			changeReplicas, ok := ba.Requests[0].GetInner().(*roachpb.AdminChangeReplicasRequest)
+			if ok && changeReplicas.ChangeType == roachpb.ADD_REPLICA && useSeenAdd.Load().(bool) {
+				seenAdd <- struct{}{}
+			}
+		}
+		return nil
+	}
+	tc := testcluster.StartTestCluster(t, numNodes, base.TestClusterArgs{
+		ReplicationMode: base.ReplicationManual,
+		ServerArgs: base.TestServerArgs{
+			Knobs: base.TestingKnobs{
+				Store: &storage.StoreTestingKnobs{
+					TestingResponseFilter: responseFilter,
+				},
+			},
+		},
+	})
+	ctx := context.Background()
+	defer tc.Stopper().Stop(ctx)
+	db := tc.Servers[rand.Intn(numNodes)].DB()
+
+	// The test assumes from the way that the range gets set up that the lease
+	// holder is node 1 and from the above relocate call that the range in
+	// question has replicas on nodes 1-3. Make the call to AdminRelocate range
+	// to set up the replication and then verify the assumed state.
+
+	key := roachpb.Key("a")
+	assert.Nil(t, db.AdminRelocateRange(ctx, key, makeReplicationTargets(1, 2, 3)))
+	rangeInfo, err := getRangeInfo(ctx, db, key)
+	assert.Nil(t, err)
+	assert.Len(t, rangeInfo.Desc.Replicas, 3)
+	assert.Equal(t, rangeInfo.Lease.Replica.NodeID, roachpb.NodeID(1))
+	for id := roachpb.StoreID(1); id <= 3; id++ {
+		_, hasReplica := rangeInfo.Desc.GetReplicaDescriptor(id)
+		assert.Truef(t, hasReplica, "missing replica descriptor for store %d", id)
+	}
+
+	// The test now proceeds to use AdminRelocateRange to move a replica from node
+	// 3 to node 4. The call should first which will first add 4 and then
+	// remove 3. Concurrently a separate goroutine will attempt to remove the
+	// replica on node 2. The ResponseFilter passed in the TestingKnobs will
+	// prevent the remove call from proceeding until after the Add of 4 has
+	// completed.
+
+	// Code above verified r1 is the leaseholder, so use it to ChangeReplicas.
+	r1, err := tc.Servers[0].Stores().GetReplicaForRangeID(rangeInfo.Desc.RangeID)
+	assert.Nil(t, err)
+	expDescAfterAdd := rangeInfo.Desc // for use with ChangeReplicas
+	expDescAfterAdd.NextReplicaID++
+	expDescAfterAdd.Replicas = append(expDescAfterAdd.Replicas, roachpb.ReplicaDescriptor{
+		NodeID:    4,
+		StoreID:   4,
+		ReplicaID: 4,
+	})
+	var relocateErr, changeErr error
+	var changedDesc *roachpb.RangeDescriptor // only populated if changeErr == nil
+	change := func() {
+		<-seenAdd
+		changedDesc, changeErr = r1.ChangeReplicas(ctx, roachpb.REMOVE_REPLICA,
+			makeReplicationTargets(2)[0], &expDescAfterAdd, "replicate", "testing")
+	}
+	relocate := func() {
+		relocateErr = db.AdminRelocateRange(ctx, key, makeReplicationTargets(1, 2, 4))
+	}
+	useSeenAdd.Store(true)
+	var wg sync.WaitGroup
+	wg.Add(2)
+	go func() { relocate(); wg.Done() }()
+	go func() { change(); wg.Done() }()
+	wg.Wait()
+	rangeInfo, err = getRangeInfo(ctx, db, key)
+	assert.Nil(t, err)
+	assert.True(t, len(rangeInfo.Desc.Replicas) >= 3)
+	assert.Falsef(t, relocateErr == nil && changeErr == nil,
+		"expected one of racing AdminRelocateReplicas and ChangeReplicas "+
+			"to fail but neither did")
+	assert.Falsef(t, relocateErr != nil && changeErr != nil,
+		"expected only one of racing AdminRelocateReplicas and ChangeReplicas "+
+			"to fail but both did")
+	if changeErr == nil {
+		assert.EqualValues(t, *changedDesc, rangeInfo.Desc)
+	}
+}
+
+// getRangeInfo retreives range info by performing a get against the provided
+// key and setting the ReturnRangeInfo flag to true.
+func getRangeInfo(
+	ctx context.Context, db *client.DB, key roachpb.Key,
+) (ri *roachpb.RangeInfo, err error) {
+	err = db.Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
+		b := txn.NewBatch()
+		b.Header.ReturnRangeInfo = true
+		b.AddRawRequest(roachpb.NewGet(key))
+		if err = db.Run(ctx, b); err != nil {
+			return err
+		}
+		resp := b.RawResponse()
+		ri = &resp.Responses[0].GetInner().Header().RangeInfos[0]
+		return nil
+	})
+	return ri, err
+}
+
+// makeReplicationTargets creates a slice of replication targets where each
+// target has a NodeID and StoreID with a value corresponding to an id in ids.
+func makeReplicationTargets(ids ...int) (targets []roachpb.ReplicationTarget) {
+	for _, id := range ids {
+		targets = append(targets, roachpb.ReplicationTarget{
+			NodeID:  roachpb.NodeID(id),
+			StoreID: roachpb.StoreID(id),
+		})
+	}
+	return targets
+}
