@@ -33,14 +33,10 @@ import (
 	"testing"
 	"time"
 
-	"github.com/cenk/backoff"
-	"github.com/kr/pretty"
-	"github.com/pkg/errors"
-	circuit "github.com/rubyist/circuitbreaker"
-	"go.etcd.io/etcd/raft"
-	"google.golang.org/grpc"
-
+	"github.com/cenkalti/backoff"
+	circuit "github.com/cockroachdb/circuitbreaker"
 	"github.com/cockroachdb/cockroach/pkg/base"
+	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/gossip/resolver"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
@@ -66,6 +62,10 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/kr/pretty"
+	"github.com/pkg/errors"
+	"go.etcd.io/etcd/raft"
+	"google.golang.org/grpc"
 )
 
 // createTestStore creates a test store using an in-memory
@@ -73,34 +73,55 @@ import (
 func createTestStore(t testing.TB, stopper *stop.Stopper) (*storage.Store, *hlc.ManualClock) {
 	manual := hlc.NewManualClock(123)
 	cfg := storage.TestStoreConfig(hlc.NewClock(manual.UnixNano, time.Nanosecond))
-	store := createTestStoreWithConfig(t, stopper, cfg)
+	store := createTestStoreWithOpts(t, testStoreOpts{cfg: &cfg}, stopper)
 	return store, manual
 }
 
+// DEPRECATED. Use createTestStoreWithOpts().
 func createTestStoreWithConfig(
 	t testing.TB, stopper *stop.Stopper, storeCfg storage.StoreConfig,
 ) *storage.Store {
-	eng := engine.NewInMem(roachpb.Attributes{}, 10<<20)
-	stopper.AddCloser(eng)
-	store := createTestStoreWithEngine(t,
-		eng,
-		true,
-		storeCfg,
+	store := createTestStoreWithOpts(t,
+		testStoreOpts{
+			cfg: &storeCfg,
+		},
 		stopper,
 	)
 	return store
 }
 
+// testStoreOpts affords control over aspects of store creation.
+type testStoreOpts struct {
+	// dontBootstrap, if set, means that the engine will not be bootstrapped.
+	dontBootstrap bool
+	// dontCreateSystemRanges is relevant only if dontBootstrap is not set.
+	// If set, the store will have a single range. If not set, the store will have
+	// all the system ranges that are generally created for a cluster at boostrap.
+	dontCreateSystemRanges bool
+
+	cfg *storage.StoreConfig
+	eng engine.Engine
+}
+
 // createTestStoreWithEngine creates a test store using the given engine and clock.
 // TestStoreConfig() can be used for creating a config suitable for most
 // tests.
-func createTestStoreWithEngine(
-	t testing.TB,
-	eng engine.Engine,
-	bootstrap bool,
-	storeCfg storage.StoreConfig,
-	stopper *stop.Stopper,
+func createTestStoreWithOpts(
+	t testing.TB, opts testStoreOpts, stopper *stop.Stopper,
 ) *storage.Store {
+	var storeCfg storage.StoreConfig
+	if opts.cfg == nil {
+		manual := hlc.NewManualClock(123)
+		storeCfg = storage.TestStoreConfig(hlc.NewClock(manual.UnixNano, time.Nanosecond))
+	} else {
+		storeCfg = *opts.cfg
+	}
+	eng := opts.eng
+	if eng == nil {
+		eng = engine.NewInMem(roachpb.Attributes{}, 10<<20)
+		stopper.AddCloser(eng)
+	}
+
 	tracer := storeCfg.Settings.Tracer
 	ac := log.AmbientContext{Tracer: tracer}
 	storeCfg.AmbientCtx = ac
@@ -147,14 +168,31 @@ func createTestStoreWithEngine(
 	storeCfg.Transport = storage.NewDummyRaftTransport(storeCfg.Settings)
 	// TODO(bdarnell): arrange to have the transport closed.
 	ctx := context.Background()
-	if bootstrap {
-		if err := storage.Bootstrap(ctx, eng, roachpb.StoreIdent{NodeID: 1, StoreID: 1}, storeCfg.Settings.Version.BootstrapVersion()); err != nil {
+	if !opts.dontBootstrap {
+		if err := storage.Bootstrap(
+			ctx, eng, roachpb.StoreIdent{NodeID: 1, StoreID: 1},
+			storeCfg.Settings.Version.BootstrapVersion(),
+		); err != nil {
 			t.Fatal(err)
 		}
 	}
 	store := storage.NewStore(storeCfg, eng, nodeDesc)
-	if bootstrap {
-		err := store.BootstrapRange(sqlbase.MakeMetadataSchema().GetInitialValues(), storeCfg.Settings.Version.ServerVersion)
+	if !opts.dontBootstrap {
+		var kvs []roachpb.KeyValue
+		var splits []roachpb.RKey
+		kvs, tableSplits := sqlbase.MakeMetadataSchema().GetInitialValues()
+		if !opts.dontCreateSystemRanges {
+			splits = config.StaticSplits()
+			splits = append(splits, tableSplits...)
+			sort.Slice(splits, func(i, j int) bool {
+				return splits[i].Less(splits[j])
+			})
+		}
+		err := store.WriteInitialData(
+			ctx,
+			kvs,
+			storeCfg.Settings.Version.BootstrapVersion().Version,
+			1 /* numStores */, splits)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -178,6 +216,22 @@ func createTestStoreWithEngine(
 		return nil
 	})
 
+	// Wait for the system config to be available in gossip. All sorts of things
+	// might not work properly while the system config is not available.
+	testutils.SucceedsSoon(t, func() error {
+		if cfg := store.Gossip().GetSystemConfig(); cfg == nil {
+			return errors.Errorf("system config not available in gossip yet")
+		}
+		return nil
+	})
+
+	// Make all the initial ranges part of replication queue purgatory. This is
+	// similar to what a real cluster does after bootstrap - we want the initial
+	// ranges to up-replicate as soon as other nodes join.
+	if err := store.ForceReplicationScanAndProcess(); err != nil {
+		t.Fatal(err)
+	}
+
 	return store
 }
 
@@ -187,7 +241,11 @@ type multiTestContext struct {
 	manualClock *hlc.ManualClock
 	clock       *hlc.Clock
 	rpcContext  *rpc.Context
-	injEngines  bool
+	// By default, a multiTestContext starts with a bunch of system ranges, just
+	// like a regular Server after bootstrap. If startWithSingleRange is set,
+	// we'll start with a single range spanning all the key space. The split
+	// queue, if not disabled, might then create other range system ranges.
+	startWithSingleRange bool
 
 	nodeIDtoAddrMu struct {
 		*syncutil.RWMutex
@@ -210,9 +268,8 @@ type multiTestContext struct {
 	// We use multiple stoppers so we can restart different parts of the
 	// test individually. transportStopper is for 'transport', and the
 	// 'stoppers' slice corresponds to the 'stores'.
-	transportStopper   *stop.Stopper
-	engineStoppers     []*stop.Stopper
-	timeUntilStoreDead time.Duration
+	transportStopper *stop.Stopper
+	engineStoppers   []*stop.Stopper
 
 	// The fields below may mutate at runtime so the pointers they contain are
 	// protected by 'mu'.
@@ -245,8 +302,7 @@ func (m *multiTestContext) Start(t testing.TB, numStores int) {
 		mCopy.clock = nil
 		mCopy.engines = nil
 		mCopy.engineStoppers = nil
-		mCopy.injEngines = false
-		mCopy.timeUntilStoreDead = 0
+		mCopy.startWithSingleRange = false
 		var empty multiTestContext
 		if !reflect.DeepEqual(empty, mCopy) {
 			t.Fatalf("illegal fields set in multiTestContext:\n%s", pretty.Diff(empty, mCopy))
@@ -302,7 +358,7 @@ func (m *multiTestContext) Start(t testing.TB, numStores int) {
 	// Wait for gossip to startup.
 	testutils.SucceedsSoon(t, func() error {
 		for i, g := range m.gossips {
-			if _, ok := g.GetSystemConfig(); !ok {
+			if cfg := g.GetSystemConfig(); cfg == nil {
 				return errors.Errorf("system config not available at index %d", i)
 			}
 		}
@@ -477,22 +533,26 @@ func (t *multiTestContextKVTransport) SendNext(
 	// would happen with real RPCs.
 	t.mtc.mu.RLock()
 	s := t.mtc.stoppers[nodeIndex]
+	sender := t.mtc.senders[nodeIndex]
 	t.mtc.mu.RUnlock()
+
 	if s == nil {
 		t.setPending(rep.ReplicaID, false)
 		return nil, roachpb.NewSendError("store is stopped")
 	}
 
-	t.mtc.mu.RLock()
-	sender := t.mtc.senders[nodeIndex]
-	t.mtc.mu.RUnlock()
 	// Clone txn of ba args for sending.
 	ba.Replica = rep.ReplicaDescriptor
 	if txn := ba.Txn; txn != nil {
-		txnClone := ba.Txn.Clone()
-		ba.Txn = &txnClone
+		ba.Txn = ba.Txn.Clone()
+	}
+	var br *roachpb.BatchResponse
+	var pErr *roachpb.Error
+	if err := s.RunTask(ctx, "mtc send", func(ctx context.Context) {
+		br, pErr = sender.Send(ctx, ba)
+	}); err != nil {
+		pErr = roachpb.NewError(err)
 	}
-	br, pErr := sender.Send(ctx, ba)
 	if br == nil {
 		br = &roachpb.BatchResponse{}
 	}
@@ -681,11 +741,12 @@ func (m *multiTestContext) populateDB(idx int, stopper *stop.Stopper) {
 	m.dbs[idx] = client.NewDB(ambient, tcsFactory, m.clocks[idx])
 }
 
-func (m *multiTestContext) populateStorePool(idx int, nodeLiveness *storage.NodeLiveness) {
-	storage.TimeUntilStoreDead.Override(&m.storeConfig.Settings.SV, m.timeUntilStoreDead)
+func (m *multiTestContext) populateStorePool(
+	idx int, cfg storage.StoreConfig, nodeLiveness *storage.NodeLiveness,
+) {
 	m.storePools[idx] = storage.NewStorePool(
-		m.storeConfig.AmbientCtx,
-		m.storeConfig.Settings,
+		cfg.AmbientCtx,
+		cfg.Settings,
 		m.gossips[idx],
 		m.clocks[idx],
 		nodeLiveness.GetNodeCount,
@@ -707,7 +768,12 @@ func (m *multiTestContext) addStore(idx int) {
 	var needBootstrap bool
 	if len(m.engines) > idx {
 		eng = m.engines[idx]
-		needBootstrap = m.injEngines
+		_, err := storage.ReadStoreIdent(context.Background(), eng)
+		if _, notBootstrapped := err.(*storage.NotBootstrappedError); notBootstrapped {
+			needBootstrap = true
+		} else if err != nil {
+			m.t.Fatal(err)
+		}
 	} else {
 		engineStopper := stop.NewStopper()
 		m.engineStoppers = append(m.engineStoppers, engineStopper)
@@ -745,9 +811,6 @@ func (m *multiTestContext) addStore(idx int) {
 		m.transportStopper,
 		metric.NewRegistry(),
 	)
-	if m.timeUntilStoreDead == 0 {
-		m.timeUntilStoreDead = storage.TestTimeUntilStoreDeadOff
-	}
 
 	nodeID := roachpb.NodeID(idx + 1)
 	cfg := m.makeStoreConfig(idx)
@@ -758,7 +821,7 @@ func (m *multiTestContext) addStore(idx int) {
 		ambient, m.clocks[idx], m.dbs[idx], m.engines, m.gossips[idx],
 		nlActive, nlRenewal, cfg.Settings, metric.TestSampleInterval,
 	)
-	m.populateStorePool(idx, m.nodeLivenesses[idx])
+	m.populateStorePool(idx, cfg, m.nodeLivenesses[idx])
 	cfg.DB = m.dbs[idx]
 	cfg.NodeLiveness = m.nodeLivenesses[idx]
 	cfg.StorePool = m.storePools[idx]
@@ -774,8 +837,21 @@ func (m *multiTestContext) addStore(idx int) {
 	}
 	store := storage.NewStore(cfg, eng, &roachpb.NodeDescriptor{NodeID: nodeID})
 	if needBootstrap && idx == 0 {
-		// Bootstrap the initial range on the first store
-		err := store.BootstrapRange(sqlbase.MakeMetadataSchema().GetInitialValues(), cfg.Settings.Version.ServerVersion)
+		// Bootstrap the initial range on the first store.
+		var splits []roachpb.RKey
+		kvs, tableSplits := sqlbase.MakeMetadataSchema().GetInitialValues()
+		if !m.startWithSingleRange {
+			splits = config.StaticSplits()
+			splits = append(splits, tableSplits...)
+			sort.Slice(splits, func(i, j int) bool {
+				return splits[i].Less(splits[j])
+			})
+		}
+		err := store.WriteInitialData(
+			ctx,
+			kvs,
+			cfg.Settings.Version.BootstrapVersion().Version,
+			len(m.engines), splits)
 		if err != nil {
 			m.t.Fatal(err)
 		}
@@ -832,7 +908,6 @@ func (m *multiTestContext) addStore(idx int) {
 	if err := m.gossipNodeDesc(m.gossips[idx], nodeID); err != nil {
 		m.t.Fatal(err)
 	}
-	store.WaitForInit()
 
 	ran := struct {
 		sync.Once
@@ -849,6 +924,9 @@ func (m *multiTestContext) addStore(idx int) {
 			close(ran.ch)
 		})
 	})
+
+	store.WaitForInit()
+
 	// Wait until we see the first heartbeat by waiting for the callback (which
 	// fires *after* the node becomes live).
 	<-ran.ch
@@ -917,7 +995,7 @@ func (m *multiTestContext) restartStoreWithoutHeartbeat(i int) {
 		log.AmbientContext{Tracer: m.storeConfig.Settings.Tracer}, m.clocks[i], m.dbs[i], m.engines,
 		m.gossips[i], nlActive, nlRenewal, cfg.Settings, metric.TestSampleInterval,
 	)
-	m.populateStorePool(i, m.nodeLivenesses[i])
+	m.populateStorePool(i, cfg, m.nodeLivenesses[i])
 	cfg.DB = m.dbs[i]
 	cfg.NodeLiveness = m.nodeLivenesses[i]
 	cfg.StorePool = m.storePools[i]
@@ -1012,16 +1090,6 @@ func (m *multiTestContext) changeReplicas(
 ) (roachpb.ReplicaID, error) {
 	ctx := context.Background()
 
-	// Perform a consistent read to get the updated range descriptor (as
-	// opposed to just going to one of the stores), to make sure we have
-	// the effects of any previous ChangeReplicas call. By the time
-	// ChangeReplicas returns the raft leader is guaranteed to have the
-	// updated version, but followers are not.
-	var desc roachpb.RangeDescriptor
-	if err := m.dbs[0].GetProto(ctx, keys.RangeDescriptorKey(startKey), &desc); err != nil {
-		return 0, err
-	}
-
 	var alreadyDoneErr string
 	switch changeType {
 	case roachpb.ADD_REPLICA:
@@ -1034,13 +1102,25 @@ func (m *multiTestContext) changeReplicas(
 		InitialBackoff: time.Millisecond,
 		MaxBackoff:     50 * time.Millisecond,
 	}
+	var desc roachpb.RangeDescriptor
 	for r := retry.Start(retryOpts); r.Next(); {
-		err := m.dbs[0].AdminChangeReplicas(
+
+		// Perform a consistent read to get the updated range descriptor (as
+		// opposed to just going to one of the stores), to make sure we have
+		// the effects of any previous ChangeReplicas call. By the time
+		// ChangeReplicas returns the raft leader is guaranteed to have the
+		// updated version, but followers are not.
+		if err := m.dbs[0].GetProto(ctx, keys.RangeDescriptorKey(startKey), &desc); err != nil {
+			return 0, err
+		}
+
+		_, err := m.dbs[0].AdminChangeReplicas(
 			ctx, startKey.AsRawKey(), changeType,
 			[]roachpb.ReplicationTarget{{
 				NodeID:  m.idents[dest].NodeID,
 				StoreID: m.idents[dest].StoreID,
 			}},
+			desc,
 		)
 
 		if err == nil || testutils.IsError(err, alreadyDoneErr) {
@@ -1058,6 +1138,7 @@ func (m *multiTestContext) changeReplicas(
 		// is lost. We could make a this into a roachpb.Error but it seems overkill
 		// for this one usage.
 		if testutils.IsError(err, "snapshot failed: .*") {
+			log.Info(ctx, err)
 			continue
 		}
 		return 0, err
@@ -1136,7 +1217,8 @@ func (m *multiTestContext) unreplicateRangeNonFatal(rangeID roachpb.RangeID, des
 func (m *multiTestContext) readIntFromEngines(key roachpb.Key) []int64 {
 	results := make([]int64, len(m.engines))
 	for i, eng := range m.engines {
-		val, _, err := engine.MVCCGet(context.Background(), eng, key, m.clocks[i].Now(), true, nil)
+		val, _, err := engine.MVCCGet(context.Background(), eng, key, m.clocks[i].Now(),
+			engine.MVCCGetOptions{})
 		if err != nil {
 			log.VEventf(context.TODO(), 1, "engine %d: error reading from key %s: %s", i, key, err)
 		} else if val == nil {
@@ -1322,6 +1404,31 @@ func truncateLogArgs(index uint64, rangeID roachpb.RangeID) *roachpb.TruncateLog
 	}
 }
 
+func heartbeatArgs(
+	txn *roachpb.Transaction, now hlc.Timestamp,
+) (*roachpb.HeartbeatTxnRequest, roachpb.Header) {
+	return &roachpb.HeartbeatTxnRequest{
+		RequestHeader: roachpb.RequestHeader{
+			Key: txn.Key,
+		},
+		Now: now,
+	}, roachpb.Header{Txn: txn}
+}
+
+func pushTxnArgs(
+	pusher, pushee *roachpb.Transaction, pushType roachpb.PushTxnType,
+) *roachpb.PushTxnRequest {
+	return &roachpb.PushTxnRequest{
+		RequestHeader: roachpb.RequestHeader{
+			Key: pushee.Key,
+		},
+		PushTo:    pusher.Timestamp.Next(),
+		PusherTxn: *pusher,
+		PusheeTxn: pushee.TxnMeta,
+		PushType:  pushType,
+	}
+}
+
 func TestSortRangeDescByAge(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	var replicaDescs []roachpb.ReplicaDescriptor
@@ -1383,7 +1490,7 @@ func TestSortRangeDescByAge(t *testing.T) {
 }
 
 func verifyRangeStats(eng engine.Reader, rangeID roachpb.RangeID, expMS enginepb.MVCCStats) error {
-	ms, err := stateloader.Make(nil /* st */, rangeID).LoadMVCCStats(context.Background(), eng)
+	ms, err := stateloader.Make(rangeID).LoadMVCCStats(context.Background(), eng)
 	if err != nil {
 		return err
 	}
