@@ -20,15 +20,14 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
-
-	"golang.org/x/sync/errgroup"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/storage/closedts/ctpb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/randutil"
+	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/pkg/errors"
+	"golang.org/x/sync/errgroup"
 )
 
 func ExampleSingleStorage() {
@@ -124,6 +123,10 @@ func ExampleSingleStorage() {
 	fmt.Println("This would resolve itself if reasonably spaced updates kept coming in.")
 	fmt.Println(s)
 
+	fmt.Println("Finally, when the storage is cleared, all buckets are reset.")
+	s.Clear()
+	fmt.Println(s)
+
 	// Output:
 	// The empty storage renders as below:
 	// +--+---------------------+----------------------+----------------------+----------------------+
@@ -242,6 +245,14 @@ func ExampleSingleStorage() {
 	//   r8                   711                    711                       711
 	//   r9                  2020                   2020                      2020                    2020
 	// +----+---------------------+----------------------+-------------------------+-----------------------+
+	//
+	// Finally, when the storage is cleared, all buckets are reset.
+	// +--+---------------------+----------------------+----------------------+----------------------+
+	//         0.000000000,0         0.000000000,0          0.000000000,0          0.000000000,0
+	//      age=0s (target ≤0s)   age=0s (target ≤10s)   age=0s (target ≤20s)   age=0s (target ≤40s)
+	//            epoch=0               epoch=0                epoch=0                epoch=0
+	// +--+---------------------+----------------------+----------------------+----------------------+
+	// +--+---------------------+----------------------+----------------------+----------------------+
 }
 
 func ExampleMultiStorage_epoch() {
@@ -335,6 +346,39 @@ func ExampleMultiStorage_epoch() {
 	// +-----+---------------------+----------------------+
 }
 
+func TestZeroValueGetsStored(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	// This test ensures that a zero values MLAI is stored for an epoch especially
+	// after we've already stored a non-zero MLAI for a different range in the
+	// same epoch. See #32904.
+	ms := NewMultiStorage(func() SingleStorage {
+		return NewMemStorage(time.Millisecond, 10)
+	})
+	e := ctpb.Entry{
+		Epoch:           1,
+		ClosedTimestamp: hlc.Timestamp{WallTime: timeutil.Now().UnixNano()},
+		MLAI:            map[roachpb.RangeID]ctpb.LAI{1: 1},
+	}
+	ms.Add(1, e)
+	e.ClosedTimestamp.WallTime++
+	r := roachpb.RangeID(2)
+	e.MLAI = map[roachpb.RangeID]ctpb.LAI{r: 0}
+	ms.Add(1, e)
+	var seen bool
+	ms.VisitDescending(1, func(e ctpb.Entry) (done bool) {
+		for rr, mlai := range e.MLAI {
+			if rr == r && mlai == 0 {
+				seen = true
+				return true
+			}
+		}
+		return false
+	})
+	if !seen {
+		t.Fatalf("Failed to see added zero value MLAI for range %v", r)
+	}
+}
+
 // TestConcurrent runs a very basic sanity check against a Storage, verifiying
 // that the bucketed Entries don't regress in obvious ways.
 func TestConcurrent(t *testing.T) {
