@@ -18,6 +18,7 @@ import (
 	"context"
 
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/storage/diskmap"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
@@ -101,24 +102,24 @@ type Iterator interface {
 	// chosen from the key ranges listed in keys.NoSplitSpans and will always
 	// sort equal to or after minSplitKey.
 	FindSplitKey(start, end, minSplitKey MVCCKey, targetSize int64) (MVCCKey, error)
-	// MVCCGet retrieves the value for the key at the specified timestamp. The
-	// value is returned in batch repr format with the key being present as the
-	// empty string. If an intent exists at the specified key, it will be
-	// returned in batch repr format in the separate intent return value.
-	// Specify true for tombstones to return a value if the key has been
-	// deleted (Value.RawBytes will be empty).
-	MVCCGet(key roachpb.Key, timestamp hlc.Timestamp,
-		txn *roachpb.Transaction, consistent, tombstones bool,
-	) (*roachpb.Value, []roachpb.Intent, error)
-	// MVCCScan scans the underlying engine from start to end keys and returns
-	// key/value pairs which have a timestamp less than or equal to the supplied
-	// timestamp, up to a max rows. The key/value pairs are returned as a buffer
-	// of varint-prefixed slices, alternating from key to value, numKvs pairs.
-	// Specify true for tombstones to return deleted values (the value portion
-	// will be empty).
-	MVCCScan(start, end roachpb.Key, max int64, timestamp hlc.Timestamp,
-		txn *roachpb.Transaction, consistent, reverse, tombstone bool,
-	) (kvs []byte, numKvs int64, intents []byte, err error)
+	// MVCCGet is the internal implementation of the family of package-level
+	// MVCCGet functions.
+	//
+	// There is little reason to use this function directly. Use the package-level
+	// MVCCGet, or one of its variants, instead.
+	MVCCGet(
+		key roachpb.Key, timestamp hlc.Timestamp, opts MVCCGetOptions,
+	) (*roachpb.Value, *roachpb.Intent, error)
+	// MVCCScan is the internal implementation of the family of package-level
+	// MVCCScan functions. The notable difference is that key/value pairs are
+	// returned raw, as a buffer of varint-prefixed slices, alternating from key
+	// to value, where numKVs specifies the number of pairs in the buffer.
+	//
+	// There is little reason to use this function directly. Use the package-level
+	// MVCCScan, or one of its variants, instead.
+	MVCCScan(
+		start, end roachpb.Key, max int64, timestamp hlc.Timestamp, opts MVCCScanOptions,
+	) (kvData []byte, numKVs int64, resumeSpan *roachpb.Span, intents []roachpb.Intent, err error)
 	// SetUpperBound installs a new upper bound for this iterator.
 	SetUpperBound(roachpb.Key)
 
@@ -134,10 +135,14 @@ type IterOptions struct {
 	// but iteration (using Next) over keys without the same user-key
 	// prefix will not work correctly (keys may be skipped).
 	Prefix bool
-	// UpperBound gives this iterator an upper bound. Attempts to Seek or Next
-	// past this point will invalidate the iterator. UpperBound must be provided
-	// unless Prefix is true, in which case the end of the prefix will be used as
-	// the upper bound.
+	// LowerBound gives this iterator an inclusive lower bound. Attempts to
+	// SeekReverse or Prev to a key that is strictly less than the bound will
+	// invalidate the iterator.
+	LowerBound roachpb.Key
+	// UpperBound gives this iterator an exclusive upper bound. Attempts to Seek
+	// or Next to a key that is greater than or equal to the bound will invalidate
+	// the iterator. UpperBound must be provided unless Prefix is true, in which
+	// case the end of the prefix will be used as the upper bound.
 	UpperBound roachpb.Key
 	// If WithStats is true, the iterator accumulates RocksDB performance
 	// counters over its lifetime which can be queried via `Stats()`.
@@ -202,6 +207,14 @@ type Writer interface {
 	// Note that clear actually removes entries from the storage
 	// engine, rather than inserting tombstones.
 	Clear(key MVCCKey) error
+	// SingleClear removes the most recent write to the item from the db with
+	// the given key. Whether older version of the item will come back to life
+	// if not also removed with SingleClear is undefined. See the following:
+	//   https://github.com/facebook/rocksdb/wiki/Single-Delete
+	// for details on the SingleDelete operation that this method invokes. Note
+	// that clear actually removes entries from the storage engine, rather than
+	// inserting tombstones.
+	SingleClear(key MVCCKey) error
 	// ClearRange removes a set of entries, from start (inclusive) to end
 	// (exclusive). Similar to Clear, this method actually removes entries from
 	// the storage engine.
@@ -229,9 +242,7 @@ type Writer interface {
 	// Put sets the given key to the value provided.
 	Put(key MVCCKey, value []byte) error
 	// LogData adds the specified data to the RocksDB WAL. The data is
-	// uninterpreted by RocksDB (i.e. not added to the memtable or
-	// sstables). Currently only used for performance testing of appending to the
-	// RocksDB WAL.
+	// uninterpreted by RocksDB (i.e. not added to the memtable or sstables).
 	LogData(data []byte) error
 	// LogLogicalOp logs the specified logical mvcc operation with the provided
 	// details to the writer, if it has logical op logging enabled. For most
@@ -257,6 +268,9 @@ type Engine interface {
 	Flush() error
 	// GetStats retrieves stats from the engine.
 	GetStats() (*Stats, error)
+	// GetEnvStats retrieves stats about the engine's environment
+	// For RocksDB, this includes details of at-rest encryption.
+	GetEnvStats() (*EnvStats, error)
 	// GetAuxiliaryDir returns a path under which files can be stored
 	// persistently, and from which data can be ingested by the engine.
 	//
@@ -287,11 +301,16 @@ type Engine interface {
 	// original engine has been stopped.
 	NewSnapshot() Reader
 	// IngestExternalFiles atomically links a slice of files into the RocksDB
-	// log-structured merge-tree. May modify the files (including the underlying
-	// file in the case of hard-links) if allowFileModifications is passed as
-	// well. See additional comments in db.cc's IngestExternalFile explaining
-	// modification behavior.
-	IngestExternalFiles(ctx context.Context, paths []string, allowFileModifications bool) error
+	// log-structured merge-tree. skipWritingSeqNo = true may be passed iff this
+	// rocksdb will never be read by versions prior to 5.16. Otherwise, if it is
+	// false, ingestion may modify the files (including the underlying file in the
+	// case of hard-links) when allowFileModifications true. See additional
+	// comments in db.cc's IngestExternalFile explaining modification behavior.
+	IngestExternalFiles(ctx context.Context, paths []string, skipWritingSeqNo, allowFileModifications bool) error
+	// PreIngestDelay offers an engine the chance to backpressure ingestions.
+	// When called, it may choose to block if the engine determines that it is in
+	// or approaching a state where further ingestions may risk its health.
+	PreIngestDelay(ctx context.Context)
 	// ApproximateDiskBytes returns an approximation of the on-disk size for the given key span.
 	ApproximateDiskBytes(from, to roachpb.Key) (uint64, error)
 	// CompactRange ensures that the specified range of key value pairs is
@@ -314,6 +333,17 @@ type Engine interface {
 	// the engine implementation. For RocksDB, this means using the Env responsible for the file
 	// which may handle extra logic (eg: copy encryption settings for EncryptedEnv).
 	LinkFile(oldname, newname string) error
+	// CreateCheckpoint creates a checkpoint of the engine in the given directory,
+	// which must not exist. The directory should be on the same file system so
+	// that hard links can be used.
+	CreateCheckpoint(dir string) error
+}
+
+// MapProvidingEngine is an Engine that also provides facilities for making a
+// sorted map that's persisted by the Engine.
+type MapProvidingEngine interface {
+	Engine
+	diskmap.Factory
 }
 
 // WithSSTables extends the Engine interface with a method to get info
@@ -328,20 +358,33 @@ type WithSSTables interface {
 type Batch interface {
 	ReadWriter
 	// Commit atomically applies any batched updates to the underlying
-	// engine. This is a noop unless the engine was created via NewBatch(). If
+	// engine. This is a noop unless the batch was created via NewBatch(). If
 	// sync is true, the batch is synchronously committed to disk.
 	Commit(sync bool) error
 	// Distinct returns a view of the existing batch which only sees writes that
 	// were performed before the Distinct batch was created. That is, the
 	// returned batch will not read its own writes, but it will read writes to
-	// the parent batch performed before the call to Distinct(). The returned
+	// the parent batch performed before the call to Distinct(), except if the
+	// parent batch is a WriteOnlyBatch, in which case the Distinct() batch will
+	// read from the underlying engine.
+	//
+	// The returned
 	// batch needs to be closed before using the parent batch again. This is used
 	// as an optimization to avoid flushing mutations buffered by the batch in
 	// situations where we know all of the batched operations are for distinct
 	// keys.
+	//
+	// TODO(tbg): it seems insane that you cannot read from a WriteOnlyBatch but
+	// you can read from a Distinct on top of a WriteOnlyBatch but randomly don't
+	// see the batch at all. I was personally just bitten by this.
 	Distinct() ReadWriter
-	// Empty returns whether the batch is empty or not.
+	// Empty returns whether the batch has been written to or not.
 	Empty() bool
+	// Len returns the size of the underlying representation of the batch.
+	// Because of the batch header, the size of the batch is never 0 and should
+	// not be used interchangeably with Empty. The method avoids the memory copy
+	// that Repr imposes, but it still may require flushing the batch's mutations.
+	Len() int
 	// Repr returns the underlying representation of the batch and can be used to
 	// reconstitute the batch on a remote node using Writer.ApplyBatchRepr().
 	Repr() []byte
@@ -369,6 +412,7 @@ type Stats struct {
 	Compactions                    int64
 	TableReadersMemEstimate        int64
 	PendingCompactionBytesEstimate int64
+	L0FileCount                    int64
 }
 
 // EnvStats is a set of RocksDB env stats, including encryption status.
@@ -381,6 +425,9 @@ type EnvStats struct {
 	ActiveKeyFiles uint64
 	// ActiveKeyBytes is the size of files using the active data key.
 	ActiveKeyBytes uint64
+	// EncryptionType is an enum describing the active encryption algorithm.
+	// See: ccl/storageccl/engineccl/enginepbccl/key_registry.proto
+	EncryptionType int32
 	// EncryptionStatus is a serialized enginepbccl/stats.proto::EncryptionStatus protobuf.
 	EncryptionStatus []byte
 }
@@ -430,3 +477,18 @@ func Scan(engine Reader, start, end MVCCKey, max int64) ([]MVCCKeyValue, error)
 	})
 	return kvs, err
 }
+
+// WriteSyncNoop carries out a synchronous no-op write to the engine.
+func WriteSyncNoop(ctx context.Context, eng Engine) error {
+	batch := eng.NewBatch()
+	defer batch.Close()
+
+	if err := batch.LogData(nil); err != nil {
+		return err
+	}
+
+	if err := batch.Commit(true /* sync */); err != nil {
+		return err
+	}
+	return nil
+}
