@@ -18,19 +18,22 @@ import (
 	"bytes"
 	"context"
 	"math/rand"
+	"path/filepath"
 	"reflect"
 	"sort"
 	"strconv"
 	"testing"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
+	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/assert"
 )
 
 func ensureRangeEqual(
@@ -187,7 +190,7 @@ func TestEngineBatchStaleCachedIterator(t *testing.T) {
 			// result back, we'll see the (newly deleted) value (due to the
 			// failure mode above).
 			if v, _, err := MVCCGet(context.Background(), batch, key,
-				hlc.Timestamp{}, true, nil); err != nil {
+				hlc.Timestamp{}, MVCCGetOptions{}); err != nil {
 				t.Fatal(err)
 			} else if v != nil {
 				t.Fatalf("expected no value, got %+v", v)
@@ -791,3 +794,33 @@ func insertKeysAndValues(keys []MVCCKey, values [][]byte, engine Engine, t *test
 		}
 	}
 }
+
+func TestCreateCheckpoint(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	dir, cleanup := testutils.TempDir(t)
+	defer cleanup()
+
+	rocksDB, err := NewRocksDB(
+		RocksDBConfig{
+			Settings: cluster.MakeTestingClusterSettings(),
+			Dir:      dir,
+		},
+		RocksDBCache{},
+	)
+
+	db := Engine(rocksDB) // be impl neutral from now on
+	defer db.Close()
+
+	dir = filepath.Join(dir, "checkpoint")
+
+	assert.NoError(t, err)
+	assert.NoError(t, db.CreateCheckpoint(dir))
+	assert.DirExists(t, dir)
+	m, err := filepath.Glob(dir + "/*")
+	assert.NoError(t, err)
+	assert.True(t, len(m) > 0)
+	if err := db.CreateCheckpoint(dir); !testutils.IsError(err, "exists") {
+		t.Fatal(err)
+	}
+}
