@@ -14,11 +14,7 @@
 
 package enginepb
 
-import (
-	"fmt"
-
-	proto "github.com/gogo/protobuf/proto"
-)
+import "fmt"
 
 // ToStats converts the receiver to an MVCCStats.
 func (ms *MVCCStatsDelta) ToStats() MVCCStats {
@@ -40,18 +36,6 @@ func (ms *MVCCStats) ToPersistentStats() MVCCPersistentStats {
 	return MVCCPersistentStats(*ms)
 }
 
-var isolationTypeLowerCase = map[int32]string{
-	0: "serializable",
-	1: "snapshot",
-}
-
-// ToLowerCaseString returns the lower case version of String().
-// Asking for lowercase is common enough (pg_setting / SHOW in SQL)
-// that we don't want to call strings.ToLower(x.String()) all the time.
-func (x IsolationType) ToLowerCaseString() string {
-	return proto.EnumName(isolationTypeLowerCase, int32(x))
-}
-
 // MustSetValue is like SetValue, except it resets the enum and panics if the
 // provided value is not a valid variant type.
 func (op *MVCCLogicalOp) MustSetValue(value interface{}) {
