@@ -17,7 +17,6 @@ package engine
 import (
 	"bytes"
 	"context"
-	"encoding/binary"
 	"fmt"
 	"io/ioutil"
 	"math"
@@ -31,13 +30,10 @@ import (
 	"time"
 	"unsafe"
 
-	"github.com/dustin/go-humanize"
-	"github.com/elastic/gosigar"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/storage/diskmap"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
@@ -48,6 +44,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	humanize "github.com/dustin/go-humanize"
+	"github.com/elastic/gosigar"
+	"github.com/pkg/errors"
 )
 
 // TODO(tamird): why does rocksdb not link jemalloc,snappy statically?
@@ -81,16 +80,56 @@ var rocksdbConcurrency = envutil.EnvOrDefaultInt(
 		return max
 	}())
 
+var ingestDelayL0Threshold = settings.RegisterIntSetting(
+	"rocksdb.ingest_backpressure.l0_file_count_threshold",
+	"number of L0 files after which to backpressure SST ingestions",
+	20,
+)
+
+var ingestDelayPerFile = settings.RegisterDurationSetting(
+	"rocksdb.ingest_backpressure.delay_l0_file",
+	"delay to add to SST ingestions per file in L0 over the configured limit",
+	time.Millisecond*200,
+)
+
+var ingestDelayPendingLimit = settings.RegisterByteSizeSetting(
+	"rocksdb.ingest_backpressure.pending_compaction_threshold",
+	"pending compaction estimate above which to backpressure SST ingestions",
+	64<<30,
+)
+
+var ingestDelayTime = settings.RegisterDurationSetting(
+	"rocksdb.ingest_backpressure.max_delay",
+	"maximum amount of time to backpressure a single SST ingestion",
+	time.Second*5,
+)
+
 // Set to true to perform expensive iterator debug leak checking. In normal
 // operation, we perform inexpensive iterator leak checking but those checks do
 // not indicate where the leak arose. The expensive checking tracks the stack
 // traces of every iterator allocated. DO NOT ENABLE in production code.
 const debugIteratorLeak = false
 
+//export rocksDBV
+func rocksDBV(sevLvl C.int, infoVerbosity C.int) bool {
+	sev := log.Severity(sevLvl)
+	return sev == log.Severity_INFO && log.V(int32(infoVerbosity)) ||
+		sev == log.Severity_WARNING ||
+		sev == log.Severity_ERROR ||
+		sev == log.Severity_FATAL
+}
+
 //export rocksDBLog
-func rocksDBLog(logLevel C.int, s *C.char, n C.int) {
-	if log.V(int32(logLevel)) {
-		ctx := logtags.AddTag(context.Background(), "rocksdb", nil)
+func rocksDBLog(sevLvl C.int, s *C.char, n C.int) {
+	ctx := logtags.AddTag(context.Background(), "rocksdb", nil)
+	switch log.Severity(sevLvl) {
+	case log.Severity_WARNING:
+		log.Warning(ctx, C.GoStringN(s, n))
+	case log.Severity_ERROR:
+		log.Error(ctx, C.GoStringN(s, n))
+	case log.Severity_FATAL:
+		log.Fatal(ctx, C.GoStringN(s, n))
+	default:
 		log.Info(ctx, C.GoStringN(s, n))
 	}
 }
@@ -649,6 +688,7 @@ func (r *RocksDB) syncLoop() {
 	s.Lock()
 
 	var lastSync time.Time
+	var err error
 
 	for {
 		for len(s.pending) == 0 && !s.closed {
@@ -674,8 +714,14 @@ func (r *RocksDB) syncLoop() {
 
 		s.Unlock()
 
-		var err error
-		if r.cfg.Dir != "" {
+		// Linux only guarantees we'll be notified of a writeback error once
+		// during a sync call. After sync fails once, we cannot rely on any
+		// future data written to WAL being crash-recoverable. That's because
+		// any future writes will be appended after a potential corruption in
+		// the WAL, and RocksDB's recovery terminates upon encountering any
+		// corruption. So, we must not call `DBSyncWAL` again after it has
+		// failed once.
+		if r.cfg.Dir != "" && err == nil {
 			err = statusToError(C.DBSyncWAL(r.rdb))
 			lastSync = timeutil.Now()
 		}
@@ -726,6 +772,14 @@ func (r *RocksDB) Close() {
 	r.syncer.Unlock()
 }
 
+// CreateCheckpoint creates a RocksDB checkpoint in the given directory (which
+// must not exist). This directory should be located on the same file system, or
+// copies of all data are used instead of hard links, which is very expensive.
+func (r *RocksDB) CreateCheckpoint(dir string) error {
+	status := C.DBCreateCheckpoint(r.rdb, goToCSlice([]byte(dir)))
+	return errors.Wrap(statusToError(status), "unable to take RocksDB checkpoint")
+}
+
 // Closed returns true if the engine is closed.
 func (r *RocksDB) Closed() bool {
 	return r.rdb == nil
@@ -793,6 +847,11 @@ func (r *RocksDB) Clear(key MVCCKey) error {
 	return dbClear(r.rdb, key)
 }
 
+// SingleClear removes the most recent item from the db with the given key.
+func (r *RocksDB) SingleClear(key MVCCKey) error {
+	return dbSingleClear(r.rdb, key)
+}
+
 // ClearRange removes a set of entries, from start (inclusive) to end
 // (exclusive).
 func (r *RocksDB) ClearRange(start, end MVCCKey) error {
@@ -902,6 +961,11 @@ func (r *RocksDB) CompactRange(start, end roachpb.Key, forceBottommost bool) err
 	return statusToError(C.DBCompactRange(r.rdb, goToCSlice(start), goToCSlice(end), C.bool(forceBottommost)))
 }
 
+// disableAutoCompaction disables automatic compactions. For testing use only.
+func (r *RocksDB) disableAutoCompaction() error {
+	return statusToError(C.DBDisableAutoCompaction(r.rdb))
+}
+
 // ApproximateDiskBytes returns the approximate on-disk size of the specified key range.
 func (r *RocksDB) ApproximateDiskBytes(from, to roachpb.Key) (uint64, error) {
 	start := MVCCKey{Key: from}
@@ -911,11 +975,6 @@ func (r *RocksDB) ApproximateDiskBytes(from, to roachpb.Key) (uint64, error) {
 	return uint64(result), err
 }
 
-// Destroy destroys the underlying filesystem data associated with the database.
-func (r *RocksDB) Destroy() error {
-	return statusToError(C.DBDestroy(goToCSlice([]byte(r.cfg.Dir))))
-}
-
 // Flush causes RocksDB to write all in-memory data to disk immediately.
 func (r *RocksDB) Flush() error {
 	return statusToError(C.DBFlush(r.rdb))
@@ -1034,6 +1093,10 @@ func (r *rocksDBReadOnly) Clear(key MVCCKey) error {
 	panic("not implemented")
 }
 
+func (r *rocksDBReadOnly) SingleClear(key MVCCKey) error {
+	panic("not implemented")
+}
+
 func (r *rocksDBReadOnly) ClearRange(start, end MVCCKey) error {
 	panic("not implemented")
 }
@@ -1171,9 +1234,50 @@ func (r *RocksDB) GetStats() (*Stats, error) {
 		Compactions:                    int64(s.compactions),
 		TableReadersMemEstimate:        int64(s.table_readers_mem_estimate),
 		PendingCompactionBytesEstimate: int64(s.pending_compaction_bytes_estimate),
+		L0FileCount:                    int64(s.l0_file_count),
 	}, nil
 }
 
+// GetTickersAndHistograms retrieves maps of all RocksDB tickers and histograms.
+// It differs from `GetStats` by getting _every_ ticker and histogram, and by not
+// getting anything else (DB properties, for example).
+func (r *RocksDB) GetTickersAndHistograms() (*enginepb.TickersAndHistograms, error) {
+	res := new(enginepb.TickersAndHistograms)
+	var s C.DBTickersAndHistogramsResult
+	if err := statusToError(C.DBGetTickersAndHistograms(r.rdb, &s)); err != nil {
+		return nil, err
+	}
+
+	tickers := (*[maxArrayLen / C.sizeof_TickerInfo]C.TickerInfo)(
+		unsafe.Pointer(s.tickers))[:s.tickers_len:s.tickers_len]
+	res.Tickers = make(map[string]uint64)
+	for _, ticker := range tickers {
+		name := cStringToGoString(ticker.name)
+		value := uint64(ticker.value)
+		res.Tickers[name] = value
+	}
+	C.free(unsafe.Pointer(s.tickers))
+
+	res.Histograms = make(map[string]enginepb.HistogramData)
+	histograms := (*[maxArrayLen / C.sizeof_HistogramInfo]C.HistogramInfo)(
+		unsafe.Pointer(s.histograms))[:s.histograms_len:s.histograms_len]
+	for _, histogram := range histograms {
+		name := cStringToGoString(histogram.name)
+		value := enginepb.HistogramData{
+			Mean:  float64(histogram.mean),
+			P50:   float64(histogram.p50),
+			P95:   float64(histogram.p95),
+			P99:   float64(histogram.p99),
+			Max:   float64(histogram.max),
+			Count: uint64(histogram.count),
+			Sum:   uint64(histogram.sum),
+		}
+		res.Histograms[name] = value
+	}
+	C.free(unsafe.Pointer(s.histograms))
+	return res, nil
+}
+
 // GetCompactionStats returns the internal RocksDB compaction stats. See
 // https://github.com/facebook/rocksdb/wiki/RocksDB-Tuning-Guide#rocksdb-statistics.
 func (r *RocksDB) GetCompactionStats() string {
@@ -1192,6 +1296,7 @@ func (r *RocksDB) GetEnvStats() (*EnvStats, error) {
 		TotalBytes:       uint64(s.total_bytes),
 		ActiveKeyFiles:   uint64(s.active_key_files),
 		ActiveKeyBytes:   uint64(s.active_key_bytes),
+		EncryptionType:   int32(s.encryption_type),
 		EncryptionStatus: cStringToGoBytes(s.encryption_status),
 	}, nil
 }
@@ -1359,6 +1464,11 @@ func (r *distinctBatch) Clear(key MVCCKey) error {
 	return nil
 }
 
+func (r *distinctBatch) SingleClear(key MVCCKey) error {
+	r.builder.SingleClear(key)
+	return nil
+}
+
 func (r *distinctBatch) ClearRange(start, end MVCCKey) error {
 	if !r.writeOnly {
 		panic("readable batch")
@@ -1457,21 +1567,17 @@ func (r *batchIterator) FindSplitKey(
 }
 
 func (r *batchIterator) MVCCGet(
-	key roachpb.Key, timestamp hlc.Timestamp, txn *roachpb.Transaction, consistent, tombstones bool,
-) (*roachpb.Value, []roachpb.Intent, error) {
+	key roachpb.Key, timestamp hlc.Timestamp, opts MVCCGetOptions,
+) (*roachpb.Value, *roachpb.Intent, error) {
 	r.batch.flushMutations()
-	return r.iter.MVCCGet(key, timestamp, txn, consistent, tombstones)
+	return r.iter.MVCCGet(key, timestamp, opts)
 }
 
 func (r *batchIterator) MVCCScan(
-	start, end roachpb.Key,
-	max int64,
-	timestamp hlc.Timestamp,
-	txn *roachpb.Transaction,
-	consistent, reverse, tombstones bool,
-) (kvs []byte, numKvs int64, intents []byte, err error) {
+	start, end roachpb.Key, max int64, timestamp hlc.Timestamp, opts MVCCScanOptions,
+) (kvData []byte, numKVs int64, resumeSpan *roachpb.Span, intents []roachpb.Intent, err error) {
 	r.batch.flushMutations()
-	return r.iter.MVCCScan(start, end, max, timestamp, txn, consistent, reverse, tombstones)
+	return r.iter.MVCCScan(start, end, max, timestamp, opts)
 }
 
 func (r *batchIterator) SetUpperBound(key roachpb.Key) {
@@ -1538,11 +1644,20 @@ type rocksDBBatch struct {
 	commitWG           sync.WaitGroup
 }
 
+var batchPool = sync.Pool{
+	New: func() interface{} {
+		return &rocksDBBatch{}
+	},
+}
+
 func newRocksDBBatch(parent *RocksDB, writeOnly bool) *rocksDBBatch {
-	r := &rocksDBBatch{
-		parent:    parent,
-		writeOnly: writeOnly,
-	}
+	// Get a new batch from the pool. Batches in the pool may have their closed
+	// fields set to true to facilitate some sanity check assertions. Reset this
+	// field and set others.
+	r := batchPool.Get().(*rocksDBBatch)
+	r.closed = false
+	r.parent = parent
+	r.writeOnly = writeOnly
 	r.distinct.rocksDBBatch = r
 	return r
 }
@@ -1568,7 +1683,12 @@ func (r *rocksDBBatch) Close() {
 		C.DBClose(r.batch)
 		r.batch = nil
 	}
-	r.closed = true
+	r.builder.reset()
+	*r = rocksDBBatch{
+		builder: r.builder,
+		closed:  true,
+	}
+	batchPool.Put(r)
 }
 
 // Closed returns true if the engine is closed.
@@ -1609,6 +1729,7 @@ func (r *rocksDBBatch) ApplyBatchRepr(repr []byte, sync bool) error {
 	if r.distinctOpen {
 		panic("distinct batch open")
 	}
+	r.distinctNeedsFlush = true
 	return r.builder.ApplyRepr(repr)
 }
 
@@ -1659,6 +1780,15 @@ func (r *rocksDBBatch) Clear(key MVCCKey) error {
 	return nil
 }
 
+func (r *rocksDBBatch) SingleClear(key MVCCKey) error {
+	if r.distinctOpen {
+		panic("distinct batch open")
+	}
+	r.distinctNeedsFlush = true
+	r.builder.SingleClear(key)
+	return nil
+}
+
 func (r *rocksDBBatch) ClearRange(start, end MVCCKey) error {
 	if r.distinctOpen {
 		panic("distinct batch open")
@@ -1907,24 +2037,19 @@ func (r *rocksDBBatch) commitInternal(sync bool) error {
 
 	warnLargeBatches := r.parent.cfg.WarnLargeBatchThreshold > 0
 	if elapsed := timeutil.Since(start); warnLargeBatches && (elapsed >= r.parent.cfg.WarnLargeBatchThreshold) {
-		log.Warningf(context.TODO(), "batch [%d/%d/%d] commit took %s (>%s):\n%s",
-			count, size, r.flushes, elapsed, r.parent.cfg.WarnLargeBatchThreshold, debug.Stack())
+		log.Warningf(context.TODO(), "batch [%d/%d/%d] commit took %s (>= warning threshold %s)",
+			count, size, r.flushes, elapsed, r.parent.cfg.WarnLargeBatchThreshold)
 	}
 
 	return nil
 }
 
 func (r *rocksDBBatch) Empty() bool {
-	return r.flushes == 0 && r.builder.Empty()
+	return r.flushes == 0 && r.builder.count == 0 && !r.builder.logData
 }
 
-func (r *rocksDBBatch) Repr() []byte {
-	if r.flushes == 0 {
-		// We've never flushed to C++. Return the mutations only.
-		return r.builder.getRepr()
-	}
-	r.flushMutations()
-	return cSliceToGoBytes(C.DBBatchRepr(r.batch))
+func (r *rocksDBBatch) Len() int {
+	return len(r.unsafeRepr())
 }
 
 func (r *rocksDBBatch) unsafeRepr() []byte {
@@ -1936,6 +2061,20 @@ func (r *rocksDBBatch) unsafeRepr() []byte {
 	return cSliceToUnsafeGoBytes(C.DBBatchRepr(r.batch))
 }
 
+func (r *rocksDBBatch) Repr() []byte {
+	if r.flushes == 0 {
+		// We've never flushed to C++. Return the mutations only. We make a copy
+		// of the builder's byte slice so that the return []byte is valid even
+		// if the builder is reset or finished.
+		repr := r.builder.getRepr()
+		cpy := make([]byte, len(repr))
+		copy(cpy, repr)
+		return cpy
+	}
+	r.flushMutations()
+	return cSliceToGoBytes(C.DBBatchRepr(r.batch))
+}
+
 func (r *rocksDBBatch) Distinct() ReadWriter {
 	if r.distinctNeedsFlush {
 		r.flushMutations()
@@ -2015,8 +2154,8 @@ func (r *rocksDBIterator) init(rdb *C.DBEngine, opts IterOptions, engine Reader,
 		r.parent.iters.Unlock()
 	}
 
-	if !opts.Prefix && len(opts.UpperBound) == 0 {
-		panic("iterator must set prefix or upper bound")
+	if !opts.Prefix && len(opts.UpperBound) == 0 && len(opts.LowerBound) == 0 {
+		panic("iterator must set prefix or upper bound or lower bound")
 	}
 
 	r.iter = C.DBNewIter(rdb, goToCIterOptions(opts))
@@ -2030,9 +2169,10 @@ func (r *rocksDBIterator) setOptions(opts IterOptions) {
 	if opts.MinTimestampHint != (hlc.Timestamp{}) || opts.MaxTimestampHint != (hlc.Timestamp{}) {
 		panic("iterator with timestamp hints cannot be reused")
 	}
-	if !opts.Prefix && len(opts.UpperBound) == 0 {
-		panic("iterator must set prefix or upper bound")
+	if !opts.Prefix && len(opts.UpperBound) == 0 && len(opts.LowerBound) == 0 {
+		panic("iterator must set prefix or upper bound or lower bound")
 	}
+	C.DBIterSetLowerBound(r.iter, goToCKey(MakeMVCCMetadataKey(opts.LowerBound)))
 	C.DBIterSetUpperBound(r.iter, goToCKey(MakeMVCCMetadataKey(opts.UpperBound)))
 }
 
@@ -2112,22 +2252,22 @@ func (r *rocksDBIterator) Valid() (bool, error) {
 
 func (r *rocksDBIterator) Next() {
 	r.checkEngineOpen()
-	r.setState(C.DBIterNext(r.iter, false /* skip_current_key_versions */))
+	r.setState(C.DBIterNext(r.iter, C.bool(false) /* skip_current_key_versions */))
 }
 
 func (r *rocksDBIterator) Prev() {
 	r.checkEngineOpen()
-	r.setState(C.DBIterPrev(r.iter, false /* skip_current_key_versions */))
+	r.setState(C.DBIterPrev(r.iter, C.bool(false) /* skip_current_key_versions */))
 }
 
 func (r *rocksDBIterator) NextKey() {
 	r.checkEngineOpen()
-	r.setState(C.DBIterNext(r.iter, true /* skip_current_key_versions */))
+	r.setState(C.DBIterNext(r.iter, C.bool(true) /* skip_current_key_versions */))
 }
 
 func (r *rocksDBIterator) PrevKey() {
 	r.checkEngineOpen()
-	r.setState(C.DBIterPrev(r.iter, true /* skip_current_key_versions */))
+	r.setState(C.DBIterPrev(r.iter, C.bool(true) /* skip_current_key_versions */))
 }
 
 func (r *rocksDBIterator) Key() MVCCKey {
@@ -2212,9 +2352,9 @@ func (r *rocksDBIterator) FindSplitKey(
 }
 
 func (r *rocksDBIterator) MVCCGet(
-	key roachpb.Key, timestamp hlc.Timestamp, txn *roachpb.Transaction, consistent, tombstones bool,
-) (*roachpb.Value, []roachpb.Intent, error) {
-	if !consistent && txn != nil {
+	key roachpb.Key, timestamp hlc.Timestamp, opts MVCCGetOptions,
+) (*roachpb.Value, *roachpb.Intent, error) {
+	if opts.Inconsistent && opts.Txn != nil {
 		return nil, nil, errors.Errorf("cannot allow inconsistent reads within a transaction")
 	}
 	if len(key) == 0 {
@@ -2223,14 +2363,14 @@ func (r *rocksDBIterator) MVCCGet(
 
 	r.clearState()
 	state := C.MVCCGet(
-		r.iter, goToCSlice(key), goToCTimestamp(timestamp),
-		goToCTxn(txn), C.bool(consistent), C.bool(tombstones),
+		r.iter, goToCSlice(key), goToCTimestamp(timestamp), goToCTxn(opts.Txn),
+		C.bool(opts.Inconsistent), C.bool(opts.Tombstones), C.bool(opts.IgnoreSequence),
 	)
 
 	if err := statusToError(state.status); err != nil {
 		return nil, nil, err
 	}
-	if err := uncertaintyToError(timestamp, state.uncertainty_timestamp, txn); err != nil {
+	if err := uncertaintyToError(timestamp, state.uncertainty_timestamp, opts.Txn); err != nil {
 		return nil, nil, err
 	}
 
@@ -2238,11 +2378,18 @@ func (r *rocksDBIterator) MVCCGet(
 	if err != nil {
 		return nil, nil, err
 	}
-	if consistent && len(intents) > 0 {
+	if !opts.Inconsistent && len(intents) > 0 {
 		return nil, nil, &roachpb.WriteIntentError{Intents: intents}
 	}
+
+	var intent *roachpb.Intent
+	if len(intents) > 1 {
+		return nil, nil, errors.Errorf("expected 0 or 1 intents, got %d", len(intents))
+	} else if len(intents) == 1 {
+		intent = &intents[0]
+	}
 	if state.data.len == 0 {
-		return nil, intents, nil
+		return nil, intent, nil
 	}
 
 	count := state.data.count
@@ -2250,7 +2397,7 @@ func (r *rocksDBIterator) MVCCGet(
 		return nil, nil, errors.Errorf("expected 0 or 1 result, found %d", count)
 	}
 	if count == 0 {
-		return nil, intents, nil
+		return nil, intent, nil
 	}
 
 	// Extract the value from the batch data.
@@ -2263,38 +2410,61 @@ func (r *rocksDBIterator) MVCCGet(
 		RawBytes:  rawValue,
 		Timestamp: mvccKey.Timestamp,
 	}
-	return value, intents, nil
+	return value, intent, nil
 }
 
 func (r *rocksDBIterator) MVCCScan(
-	start, end roachpb.Key,
-	max int64,
-	timestamp hlc.Timestamp,
-	txn *roachpb.Transaction,
-	consistent, reverse, tombstones bool,
-) (kvs []byte, numKvs int64, intents []byte, err error) {
-	if !consistent && txn != nil {
-		return nil, 0, nil, errors.Errorf("cannot allow inconsistent reads within a transaction")
+	start, end roachpb.Key, max int64, timestamp hlc.Timestamp, opts MVCCScanOptions,
+) (kvData []byte, numKVs int64, resumeSpan *roachpb.Span, intents []roachpb.Intent, err error) {
+	if opts.Inconsistent && opts.Txn != nil {
+		return nil, 0, nil, nil, errors.Errorf("cannot allow inconsistent reads within a transaction")
 	}
 	if len(end) == 0 {
-		return nil, 0, nil, emptyKeyError()
+		return nil, 0, nil, nil, emptyKeyError()
+	}
+	if max == 0 {
+		resumeSpan = &roachpb.Span{Key: start, EndKey: end}
+		return nil, 0, resumeSpan, nil, nil
 	}
 
 	r.clearState()
 	state := C.MVCCScan(
 		r.iter, goToCSlice(start), goToCSlice(end),
 		goToCTimestamp(timestamp), C.int64_t(max),
-		goToCTxn(txn), C.bool(consistent), C.bool(reverse), C.bool(tombstones),
+		goToCTxn(opts.Txn), C.bool(opts.Inconsistent),
+		C.bool(opts.Reverse), C.bool(opts.Tombstones),
+		C.bool(opts.IgnoreSequence),
 	)
 
 	if err := statusToError(state.status); err != nil {
-		return nil, 0, nil, err
+		return nil, 0, nil, nil, err
 	}
-	if err := uncertaintyToError(timestamp, state.uncertainty_timestamp, txn); err != nil {
-		return nil, 0, nil, err
+	if err := uncertaintyToError(timestamp, state.uncertainty_timestamp, opts.Txn); err != nil {
+		return nil, 0, nil, nil, err
 	}
-	kvs = copyFromSliceVector(state.data.bufs, state.data.len)
-	return kvs, int64(state.data.count), cSliceToGoBytes(state.intents), nil
+
+	kvData = copyFromSliceVector(state.data.bufs, state.data.len)
+	numKVs = int64(state.data.count)
+
+	if resumeKey := cSliceToGoBytes(state.resume_key); resumeKey != nil {
+		if opts.Reverse {
+			resumeSpan = &roachpb.Span{Key: start, EndKey: roachpb.Key(resumeKey).Next()}
+		} else {
+			resumeSpan = &roachpb.Span{Key: resumeKey, EndKey: end}
+		}
+	}
+
+	intents, err = buildScanIntents(cSliceToGoBytes(state.intents))
+	if err != nil {
+		return nil, 0, nil, nil, err
+	}
+	if !opts.Inconsistent && len(intents) > 0 {
+		// When encountering intents during a consistent scan we still need to
+		// return the resume key.
+		return nil, 0, resumeSpan, nil, &roachpb.WriteIntentError{Intents: intents}
+	}
+
+	return kvData, numKVs, resumeSpan, intents, nil
 }
 
 func (r *rocksDBIterator) SetUpperBound(key roachpb.Key) {
@@ -2441,6 +2611,7 @@ func goToCTxn(txn *roachpb.Transaction) C.DBTxn {
 	if txn != nil {
 		r.id = goToCSlice(txn.ID.GetBytes())
 		r.epoch = C.uint32_t(txn.Epoch)
+		r.sequence = C.int32_t(txn.Sequence)
 		r.max_timestamp = goToCTimestamp(txn.MaxTimestamp)
 	}
 	return r
@@ -2449,6 +2620,7 @@ func goToCTxn(txn *roachpb.Transaction) C.DBTxn {
 func goToCIterOptions(opts IterOptions) C.DBIterOptions {
 	return C.DBIterOptions{
 		prefix:             C.bool(opts.Prefix),
+		lower_bound:        goToCKey(MakeMVCCMetadataKey(opts.LowerBound)),
 		upper_bound:        goToCKey(MakeMVCCMetadataKey(opts.UpperBound)),
 		min_timestamp_hint: goToCTimestamp(opts.MinTimestampHint),
 		max_timestamp_hint: goToCTimestamp(opts.MaxTimestampHint),
@@ -2582,6 +2754,13 @@ func dbClear(rdb *C.DBEngine, key MVCCKey) error {
 	return statusToError(C.DBDelete(rdb, goToCKey(key)))
 }
 
+func dbSingleClear(rdb *C.DBEngine, key MVCCKey) error {
+	if len(key.Key) == 0 {
+		return emptyKeyError()
+	}
+	return statusToError(C.DBSingleDelete(rdb, goToCKey(key)))
+}
+
 func dbClearRange(rdb *C.DBEngine, start, end MVCCKey) error {
 	if err := statusToError(C.DBDeleteRange(rdb, goToCKey(start), goToCKey(end))); err != nil {
 		return err
@@ -2686,9 +2865,9 @@ func (fr *RocksDBSstFileReader) IngestExternalFile(data []byte) error {
 	cPathLen := C.size_t(len(cPaths))
 	defer C.free(unsafe.Pointer(cPaths[0]))
 
-	const noMove, modify = false, true
+	const noMove, writeSeqNo, modify = false, false, true
 	return statusToError(C.DBIngestExternalFiles(
-		fr.rocksDB.rdb, &cPaths[0], cPathLen, noMove, modify,
+		fr.rocksDB.rdb, &cPaths[0], cPathLen, noMove, writeSeqNo, modify,
 	))
 }
 
@@ -2754,6 +2933,8 @@ func (fw *RocksDBSstFileWriter) Delete(k MVCCKey) error {
 	return statusToError(C.DBSstFileWriterDelete(fw.fw, goToCKey(k)))
 }
 
+var _ = (*RocksDBSstFileWriter).Delete
+
 // Finish finalizes the writer and returns the constructed file's contents. At
 // least one kv entry must have been added.
 func (fw *RocksDBSstFileWriter) Finish() ([]byte, error) {
@@ -2825,10 +3006,61 @@ func (r *RocksDB) setAuxiliaryDir(d string) error {
 	return nil
 }
 
+// PreIngestDelay may choose to block for some duration if L0 has an excessive
+// number of files in it or if PendingCompactionBytesEstimate is elevated. This
+// it is intended to be called before ingesting a new SST, since we'd rather
+// backpressure the bulk operation adding SSTs than slow down the whole RocksDB
+// instance and impact all forground traffic by adding too many files to it.
+// After the number of L0 files exceeds the configured limit, it gradually
+// begins delaying more for each additional file in L0 over the limit until
+// hitting its configured (via settings) maximum delay. If the pending
+// compaction limit is exceeded, it waits for the maximum delay.
+func (r *RocksDB) PreIngestDelay(ctx context.Context) {
+	if r.cfg.Settings == nil {
+		return
+	}
+	stats, err := r.GetStats()
+	if err != nil {
+		log.Warningf(ctx, "failed to read stats: %+v", err)
+		return
+	}
+	targetDelay := calculatePreIngestDelay(r.cfg, stats)
+
+	if targetDelay == 0 {
+		return
+	}
+	log.VEventf(ctx, 2, "delaying SST ingestion %s. %d L0 files, %db pending compaction", targetDelay, stats.L0FileCount, stats.PendingCompactionBytesEstimate)
+
+	select {
+	case <-time.After(targetDelay):
+	case <-ctx.Done():
+	}
+}
+
+func calculatePreIngestDelay(cfg RocksDBConfig, stats *Stats) time.Duration {
+	maxDelay := ingestDelayTime.Get(&cfg.Settings.SV)
+	l0Filelimit := ingestDelayL0Threshold.Get(&cfg.Settings.SV)
+	compactionLimit := ingestDelayPendingLimit.Get(&cfg.Settings.SV)
+
+	if stats.PendingCompactionBytesEstimate >= compactionLimit {
+		return maxDelay
+	}
+	const ramp = 10
+	if stats.L0FileCount > l0Filelimit {
+		delayPerFile := maxDelay / time.Duration(ramp)
+		targetDelay := time.Duration(stats.L0FileCount-l0Filelimit) * delayPerFile
+		if targetDelay > maxDelay {
+			return maxDelay
+		}
+		return targetDelay
+	}
+	return 0
+}
+
 // IngestExternalFiles atomically links a slice of files into the RocksDB
 // log-structured merge-tree.
 func (r *RocksDB) IngestExternalFiles(
-	ctx context.Context, paths []string, allowFileModifications bool,
+	ctx context.Context, paths []string, skipWritingSeqNo, allowFileModifications bool,
 ) error {
 	cPaths := make([]*C.char, len(paths))
 	for i := range paths {
@@ -2845,6 +3077,7 @@ func (r *RocksDB) IngestExternalFiles(
 		&cPaths[0],
 		C.size_t(len(cPaths)),
 		C._Bool(true), // move_files
+		C._Bool(!skipWritingSeqNo),
 		C._Bool(allowFileModifications),
 	))
 }
@@ -2909,6 +3142,16 @@ func (r *RocksDB) LinkFile(oldname, newname string) error {
 	return nil
 }
 
+// NewSortedDiskMap implements the MapProvidingEngine interface.
+func (r *RocksDB) NewSortedDiskMap() diskmap.SortedDiskMap {
+	return NewRocksDBMap(r)
+}
+
+// NewSortedDiskMultiMap implements the MapProvidingEngine interface.
+func (r *RocksDB) NewSortedDiskMultiMap() diskmap.SortedDiskMap {
+	return NewRocksDBMultiMap(r)
+}
+
 // IsValidSplitKey returns whether the key is a valid split key. Certain key
 // ranges cannot be split (the meta1 span and the system DB span); split keys
 // chosen within any of these ranges are considered invalid. And a split key
@@ -2933,44 +3176,12 @@ func unlockFile(lock C.DBFileLock) error {
 	return statusToError(C.DBUnlockFile(lock))
 }
 
-// mvccScanSkipKeyValue is like MVCCScanDecodeKeyValue, but doesn't bother
-// actually decoding the kvs. Instead, it skips the kv and returns the rest of
-// the byte buffer.
-func mvccScanSkipKeyValue(repr []byte) ([]byte, error) {
-	if len(repr) < 8 {
-		return repr, errors.Errorf("unexpected batch EOF")
-	}
-	v := binary.LittleEndian.Uint64(repr)
-	keySize := v >> 32
-	valSize := v & ((1 << 32) - 1)
-	if (keySize + valSize) > uint64(len(repr)) {
-		return nil, fmt.Errorf("expected %d bytes, but only %d remaining",
-			keySize+valSize, len(repr))
-	}
-	repr = repr[8+keySize+valSize:]
-	return repr, nil
-}
-
 // MVCCScanDecodeKeyValue decodes a key/value pair returned in an MVCCScan
 // "batch" (this is not the RocksDB batch repr format), returning both the
 // key/value and the suffix of data remaining in the batch.
 func MVCCScanDecodeKeyValue(repr []byte) (key MVCCKey, value []byte, orepr []byte, err error) {
-	if len(repr) < 8 {
-		return key, nil, repr, errors.Errorf("unexpected batch EOF")
-	}
-	v := binary.LittleEndian.Uint64(repr)
-	keySize := v >> 32
-	valSize := v & ((1 << 32) - 1)
-	if (keySize + valSize) > uint64(len(repr)) {
-		return key, nil, nil, fmt.Errorf("expected %d bytes, but only %d remaining",
-			keySize+valSize, len(repr))
-	}
-	repr = repr[8:]
-	rawKey := repr[:keySize]
-	value = repr[keySize : keySize+valSize]
-	repr = repr[keySize+valSize:]
-	key, err = DecodeKey(rawKey)
-	return key, value, repr, err
+	k, ts, value, orepr, err := enginepb.ScanDecodeKeyValue(repr)
+	return MVCCKey{k, ts}, value, orepr, err
 }
 
 func notFoundErrOrDefault(err error) error {
