@@ -41,6 +41,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/randutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/stretchr/testify/require"
 )
 
 const testCacheSize = 1 << 30 // 1 GB
@@ -71,7 +72,7 @@ func TestBatchReadLaterWrite(t *testing.T) {
 	// Read from a batch that was opened before the value was written to the
 	// underlying engine. The batch will see the write.
 	{
-		rv, _, err := MVCCGet(ctx, batch, key, hlc.Timestamp{}, true, nil)
+		rv, _, err := MVCCGet(ctx, batch, key, hlc.Timestamp{}, MVCCGetOptions{})
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -87,7 +88,7 @@ func TestBatchReadLaterWrite(t *testing.T) {
 	// Read from a snapshot opened prior to the write. The snapshot won't see the
 	// write.
 	{
-		rv, _, err := MVCCGet(ctx, snap, key, hlc.Timestamp{}, true, nil)
+		rv, _, err := MVCCGet(ctx, snap, key, hlc.Timestamp{}, MVCCGetOptions{})
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -208,12 +209,15 @@ func TestBatchPrefixIter(t *testing.T) {
 	}
 }
 
-func TestIterUpperBound(t *testing.T) {
+func TestIterBounds(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	db := setupMVCCInMemRocksDB(t, "iter_upper_bound")
+	db := setupMVCCInMemRocksDB(t, "iter_bounds")
 	defer db.Close()
 
+	if err := db.Put(mvccKey("0"), []byte("val")); err != nil {
+		t.Fatal(err)
+	}
 	if err := db.Put(mvccKey("a"), []byte("val")); err != nil {
 		t.Fatal(err)
 	}
@@ -235,6 +239,47 @@ func TestIterUpperBound(t *testing.T) {
 			e := tc.createEngine()
 			defer e.Close()
 
+			if _, ok := e.(*rocksDBBatch); !ok { // batches do not support reverse iteration
+				// Test that a new iterator's lower bound is applied.
+				func() {
+					iter := e.NewIterator(IterOptions{LowerBound: roachpb.Key("b")})
+					defer iter.Close()
+					iter.SeekReverse(mvccKey("b"))
+					if ok, err := iter.Valid(); err != nil {
+						t.Fatal(err)
+					} else if !ok {
+						t.Fatalf("expected iterator to be valid, but was invalid")
+					}
+					iter.SeekReverse(mvccKey("a"))
+					if ok, err := iter.Valid(); err != nil {
+						t.Fatal(err)
+					} else if ok {
+						t.Fatalf("expected iterator to be invalid, but was valid")
+					}
+				}()
+
+				// Test that the cached iterator, if the underlying engine implementation
+				// caches iterators, can take on a new lower bound.
+				func() {
+					iter := e.NewIterator(IterOptions{LowerBound: roachpb.Key("a")})
+					defer iter.Close()
+
+					iter.SeekReverse(mvccKey("a"))
+					if ok, err := iter.Valid(); !ok {
+						t.Fatal(err)
+					}
+					if !mvccKey("a").Equal(iter.Key()) {
+						t.Fatalf("expected key a, but got %q", iter.Key())
+					}
+					iter.Prev()
+					if ok, err := iter.Valid(); err != nil {
+						t.Fatal(err)
+					} else if ok {
+						t.Fatalf("expected iterator to be invalid, but was valid")
+					}
+				}()
+			}
+
 			// Test that a new iterator's upper bound is applied.
 			func() {
 				iter := e.NewIterator(IterOptions{UpperBound: roachpb.Key("a")})
@@ -268,9 +313,7 @@ func TestIterUpperBound(t *testing.T) {
 				}
 			}()
 
-			// If the engine supports writes, test that the upper bound is applied to
-			// newly written keys. This notably tests the logic in BaseDeltaIterator
-			// on batches.
+			// Perform additional tests if the engine supports writes.
 			w, isReadWriter := e.(ReadWriter)
 			if _, isSecretlyReadOnly := e.(*rocksDBReadOnly); !isReadWriter || isSecretlyReadOnly {
 				return
@@ -296,7 +339,7 @@ func makeKey(i int) MVCCKey {
 	return MakeMVCCMetadataKey(roachpb.Key(strconv.Itoa(i)))
 }
 
-func benchmarkIterOnBatch(b *testing.B, writes int) {
+func benchmarkIterOnBatch(ctx context.Context, b *testing.B, writes int) {
 	engine := createTestEngine()
 	defer engine.Close()
 
@@ -327,7 +370,7 @@ func benchmarkIterOnBatch(b *testing.B, writes int) {
 }
 
 func benchmarkIterOnReadWriter(
-	b *testing.B, writes int, f func(Engine) ReadWriter, closeReadWriter bool,
+	ctx context.Context, b *testing.B, writes int, f func(Engine) ReadWriter, closeReadWriter bool,
 ) {
 	engine := createTestEngine()
 	defer engine.Close()
@@ -634,7 +677,7 @@ func TestConcurrentBatch(t *testing.T) {
 				t.Fatalf("target size (%d) should be larger than the max batch group size (%d)",
 					targetSize, maxBatchGroupSize)
 			}
-			if len(batch.Repr()) >= targetSize {
+			if batch.Len() >= targetSize {
 				break
 			}
 		}
@@ -1270,6 +1313,12 @@ func TestRocksDBDeleteRangeCompaction(t *testing.T) {
 	db := setupMVCCInMemRocksDB(t, "delrange").(InMem)
 	defer db.Close()
 
+	// Disable automatic compactions which interfere with test expectations
+	// below.
+	if err := db.disableAutoCompaction(); err != nil {
+		t.Fatal(err)
+	}
+
 	makeKey := func(prefix string, i int) roachpb.Key {
 		return roachpb.Key(fmt.Sprintf("%s%09d", prefix, i))
 	}
@@ -1312,7 +1361,7 @@ func TestRocksDBDeleteRangeCompaction(t *testing.T) {
 			t.Fatal(err)
 		}
 
-		if err := db.IngestExternalFiles(context.Background(), []string{filename}, true); err != nil {
+		if err := db.IngestExternalFiles(context.Background(), []string{filename}, true, true); err != nil {
 			t.Fatal(err)
 		}
 		if testing.Verbose() {
@@ -1439,7 +1488,7 @@ func BenchmarkRocksDBDeleteRangeIterate(b *testing.B) {
 							b.Fatal(err)
 						}
 
-						err = db.IngestExternalFiles(context.Background(), []string{filename}, true)
+						err = db.IngestExternalFiles(context.Background(), []string{filename}, true, true)
 						if err != nil {
 							b.Fatal(err)
 						}
@@ -1556,7 +1605,7 @@ func TestSstFileWriterTimeBound(t *testing.T) {
 		if err := db.WriteFile(`ingest`, sstContents); err != nil {
 			t.Fatal(err)
 		}
-		if err := db.IngestExternalFiles(ctx, []string{`ingest`}, true); err != nil {
+		if err := db.IngestExternalFiles(ctx, []string{`ingest`}, true, true); err != nil {
 			t.Fatal(err)
 		}
 	}
@@ -1587,7 +1636,21 @@ func TestSstFileWriterTimeBound(t *testing.T) {
 func TestRocksDBWALFileEmptyBatch(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	e := NewInMem(roachpb.Attributes{}, 1<<20)
+	dir, cleanup := testutils.TempDir(t)
+	defer cleanup()
+
+	// NB: The in-mem RocksDB instance doesn't support syncing the WAL which is
+	// necessary for this test.
+	e, err := NewRocksDB(
+		RocksDBConfig{
+			Settings: cluster.MakeTestingClusterSettings(),
+			Dir:      dir,
+		},
+		RocksDBCache{},
+	)
+	if err != nil {
+		t.Fatal(err)
+	}
 	defer e.Close()
 
 	// Commit a batch with one key.
@@ -1596,7 +1659,7 @@ func TestRocksDBWALFileEmptyBatch(t *testing.T) {
 	if err := b.Put(mvccKey("foo"), []byte{'b', 'a', 'r'}); err != nil {
 		t.Fatal(err)
 	}
-	if err := b.Commit(true); err != nil {
+	if err := b.Commit(true /* sync */); err != nil {
 		t.Fatal(err)
 	}
 
@@ -1615,7 +1678,7 @@ func TestRocksDBWALFileEmptyBatch(t *testing.T) {
 	// Commit an empty batch.
 	b = e.NewBatch()
 	defer b.Close()
-	if err := b.Commit(true); err != nil {
+	if err := b.Commit(true /* sync */); err != nil {
 		t.Fatal(err)
 	}
 
@@ -1628,4 +1691,75 @@ func TestRocksDBWALFileEmptyBatch(t *testing.T) {
 		t.Fatalf("expected wal files %#v after committing empty batch, but got %#v",
 			walsBefore, walsAfter)
 	}
+
+	// Regression test a bug that would accidentally make Commit a no-op (via an
+	// errant fast-path) when a batch contained only LogData.
+	testutils.RunTrueAndFalse(t, "distinct", func(t *testing.T, distinct bool) {
+		walsBefore, err := e.GetSortedWALFiles()
+		if err != nil {
+			t.Fatal(err)
+		}
+		if len(walsBefore) != 1 {
+			t.Fatalf("expected one WAL file, got %d", len(walsBefore))
+		}
+
+		batch := e.NewBatch()
+		defer batch.Close()
+
+		var writer ReadWriter = batch
+		if distinct {
+			// NB: we can't actually close this distinct batch because it auto-
+			// closes when the batch commits.
+			writer = batch.Distinct()
+		}
+
+		if err := writer.LogData([]byte("foo")); err != nil {
+			t.Fatal(err)
+		}
+		if batch.Empty() {
+			t.Error("batch is not empty")
+		}
+
+		if err := batch.Commit(true /* sync */); err != nil {
+			t.Fatal(err)
+		}
+
+		// Verify that the WAL has grown.
+		walsAfter, err := e.GetSortedWALFiles()
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		if len(walsAfter) != 1 {
+			t.Fatalf("expected one WAL file, got %+v", walsAfter)
+		}
+
+		if after, before := walsAfter[0].Size, walsBefore[0].Size; after <= before {
+			t.Fatalf("wal size was expected to increase, got %d -> %d", before, after)
+		}
+	})
+}
+
+func TestIngestDelayLimit(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	cfg := RocksDBConfig{Settings: cluster.MakeTestingClusterSettings()}
+
+	max, ramp := time.Second*5, time.Second*5/10
+
+	for _, tc := range []struct {
+		exp   time.Duration
+		stats Stats
+	}{
+		{0, Stats{}},
+		{0, Stats{L0FileCount: 19}},
+		{0, Stats{L0FileCount: 20}},
+		{ramp, Stats{L0FileCount: 21}},
+		{ramp * 2, Stats{L0FileCount: 22}},
+		{max, Stats{L0FileCount: 55}},
+		{0, Stats{PendingCompactionBytesEstimate: 20 << 30}},
+		{max, Stats{L0FileCount: 25, PendingCompactionBytesEstimate: 80 << 30}},
+		{max, Stats{L0FileCount: 35, PendingCompactionBytesEstimate: 20 << 30}},
+	} {
+		require.Equal(t, tc.exp, calculatePreIngestDelay(cfg, &tc.stats))
+	}
 }
