@@ -16,9 +16,7 @@
 
 package engine
 
-import (
-	"unsafe"
-)
+import "unsafe"
 
 //extern runtime.mallocgc
 func mallocgc(size uintptr, typ unsafe.Pointer, needzero bool) unsafe.Pointer
