@@ -27,6 +27,7 @@ import (
 	"time"
 	"unsafe"
 
+	circuit "github.com/cockroachdb/circuitbreaker"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
@@ -36,13 +37,13 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/storage/rditer"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/randutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/pkg/errors"
-	"github.com/rubyist/circuitbreaker"
 )
 
 // AddReplica adds the replica to the store's replica map and to the sorted
@@ -77,72 +78,14 @@ func (s *Store) ComputeMVCCStats() (enginepb.MVCCStats, error) {
 	return totalStats, err
 }
 
-func forceScanAndProcess(s *Store, q *baseQueue) {
-	newStoreReplicaVisitor(s).Visit(func(repl *Replica) bool {
-		q.MaybeAdd(repl, s.cfg.Clock.Now())
-		return true
-	})
-
-	q.DrainQueue(s.stopper)
-}
-
-// ForceReplicationScanAndProcess iterates over all ranges and
-// enqueues any that need to be replicated.
-func (s *Store) ForceReplicationScanAndProcess() {
-	forceScanAndProcess(s, s.replicateQueue.baseQueue)
-}
-
-// ForceReplicaGCScanAndProcess iterates over all ranges and enqueues any that
-// may need to be GC'd.
-func (s *Store) ForceReplicaGCScanAndProcess() {
-	forceScanAndProcess(s, s.replicaGCQueue.baseQueue)
-}
-
-// ForceMergeScanAndProcess iterates over all ranges and enqueues any that
-// may need to be merged.
-func (s *Store) ForceMergeScanAndProcess() {
-	forceScanAndProcess(s, s.mergeQueue.baseQueue)
-}
-
-// ForceSplitScanAndProcess iterates over all ranges and enqueues any that
-// may need to be split.
-func (s *Store) ForceSplitScanAndProcess() {
-	forceScanAndProcess(s, s.splitQueue.baseQueue)
-}
-
-// ForceRaftLogScanAndProcess iterates over all ranges and enqueues any that
-// need their raft logs truncated and then process each of them.
-func (s *Store) ForceRaftLogScanAndProcess() {
-	forceScanAndProcess(s, s.raftLogQueue.baseQueue)
-}
-
-// ForceTimeSeriesMaintenanceQueueProcess iterates over all ranges, enqueuing
-// any that need time series maintenance, then processes the time series
-// maintenance queue.
-func (s *Store) ForceTimeSeriesMaintenanceQueueProcess() {
-	forceScanAndProcess(s, s.tsMaintenanceQueue.baseQueue)
-}
-
-// ForceRaftSnapshotQueueProcess iterates over all ranges, enqueuing
-// any that need raft snapshots, then processes the raft snapshot
-// queue.
-func (s *Store) ForceRaftSnapshotQueueProcess() {
-	forceScanAndProcess(s, s.raftSnapshotQueue.baseQueue)
-}
-
 // ConsistencyQueueShouldQueue invokes the shouldQueue method on the
 // store's consistency queue.
 func (s *Store) ConsistencyQueueShouldQueue(
-	ctx context.Context, now hlc.Timestamp, r *Replica, cfg config.SystemConfig,
+	ctx context.Context, now hlc.Timestamp, r *Replica, cfg *config.SystemConfig,
 ) (bool, float64) {
 	return s.consistencyQueue.shouldQueue(ctx, now, r, cfg)
 }
 
-// GetDeadReplicas exports s.deadReplicas for tests.
-func (s *Store) GetDeadReplicas() roachpb.StoreDeadReplicas {
-	return s.deadReplicas()
-}
-
 // LogReplicaChangeTest adds a fake replica change event to the log for the
 // range which contains the given key.
 func (s *Store) LogReplicaChangeTest(
@@ -151,7 +94,7 @@ func (s *Store) LogReplicaChangeTest(
 	changeType roachpb.ReplicaChangeType,
 	replica roachpb.ReplicaDescriptor,
 	desc roachpb.RangeDescriptor,
-	reason RangeLogEventReason,
+	reason storagepb.RangeLogEventReason,
 	details string,
 ) error {
 	return s.logChange(ctx, txn, changeType, replica, desc, reason, details)
@@ -200,16 +143,6 @@ func (s *Store) SetReplicaScannerActive(active bool) {
 	s.setScannerActive(active)
 }
 
-// GetOrCreateReplica passes through to its lowercase sibling.
-func (s *Store) GetOrCreateReplica(
-	ctx context.Context,
-	rangeID roachpb.RangeID,
-	replicaID roachpb.ReplicaID,
-	creatingReplica *roachpb.ReplicaDescriptor,
-) (*Replica, bool, error) {
-	return s.getOrCreateReplica(ctx, rangeID, replicaID, creatingReplica)
-}
-
 // EnqueueRaftUpdateCheck enqueues the replica for a Raft update check, forcing
 // the replica's Raft group into existence.
 func (s *Store) EnqueueRaftUpdateCheck(rangeID roachpb.RangeID) {
@@ -217,8 +150,8 @@ func (s *Store) EnqueueRaftUpdateCheck(rangeID roachpb.RangeID) {
 }
 
 func manualQueue(s *Store, q queueImpl, repl *Replica) error {
-	cfg, ok := s.Gossip().GetSystemConfig()
-	if !ok {
+	cfg := s.Gossip().GetSystemConfig()
+	if cfg == nil {
 		return fmt.Errorf("%s: system config not yet available", s)
 	}
 	ctx := repl.AnnotateCtx(context.TODO())
@@ -240,6 +173,12 @@ func (s *Store) ReservationCount() int {
 	return len(s.snapshotApplySem)
 }
 
+// ClearClosedTimestampStorage clears the closed timestamp storage of all
+// knowledge about closed timestamps.
+func (s *Store) ClearClosedTimestampStorage() {
+	s.cfg.ClosedTimestamp.Storage.Clear()
+}
+
 // AssertInvariants verifies that the store's bookkeping is self-consistent. It
 // is only valid to call this method when there is no in-flight traffic to the
 // store (e.g., after the store is shut down).
@@ -277,8 +216,8 @@ func NewTestStorePool(cfg StoreConfig) *StorePool {
 		func() int {
 			return 1
 		},
-		func(roachpb.NodeID, time.Time, time.Duration) NodeLivenessStatus {
-			return NodeLivenessStatus_LIVE
+		func(roachpb.NodeID, time.Time, time.Duration) storagepb.NodeLivenessStatus {
+			return storagepb.NodeLivenessStatus_LIVE
 		},
 		/* deterministic */ false,
 	)
@@ -389,11 +328,12 @@ func (r *Replica) ShouldBackpressureWrites() bool {
 	return r.shouldBackpressureWrites()
 }
 
-// GetRaftLogSize returns the raft log size.
-func (r *Replica) GetRaftLogSize() int64 {
+// GetRaftLogSize returns the approximate raft log size and whether it is
+// trustworthy.. See r.mu.raftLogSize for details.
+func (r *Replica) GetRaftLogSize() (int64, bool) {
 	r.mu.RLock()
 	defer r.mu.RUnlock()
-	return r.mu.raftLogSize
+	return r.mu.raftLogSize, r.mu.raftLogSizeTrusted
 }
 
 // GetCachedLastTerm returns the cached last term value. May return
@@ -422,7 +362,8 @@ func (r *Replica) HasQuorum() bool {
 // GetStoreList exposes getStoreList for testing only, but with a hardcoded
 // storeFilter of storeFilterNone.
 func (sp *StorePool) GetStoreList(rangeID roachpb.RangeID) (StoreList, int, int) {
-	return sp.getStoreList(rangeID, storeFilterNone)
+	list, available, throttled := sp.getStoreList(rangeID, storeFilterNone)
+	return list, available, len(throttled)
 }
 
 // Stores returns a copy of sl.stores.
@@ -432,28 +373,10 @@ func (sl *StoreList) Stores() []roachpb.StoreDescriptor {
 	return stores
 }
 
-const (
-	sideloadBogusIndex = 12345
-	sideloadBogusTerm  = 67890
-)
-
-func (r *Replica) PutBogusSideloadedData() {
-	r.raftMu.Lock()
-	defer r.raftMu.Unlock()
-	if err := r.raftMu.sideloaded.Put(context.Background(), sideloadBogusIndex, sideloadBogusTerm, []byte("bogus")); err != nil {
-		panic(err)
-	}
-}
-
-func (r *Replica) HasBogusSideloadedData() bool {
-	r.raftMu.Lock()
-	defer r.raftMu.Unlock()
-	if _, err := r.raftMu.sideloaded.Get(context.Background(), sideloadBogusIndex, sideloadBogusTerm); err == errSideloadedFileNotFound {
-		return false
-	} else if err != nil {
-		panic(err)
-	}
-	return true
+// SideloadedRaftMuLocked returns r.raftMu.sideloaded. Requires a previous call
+// to RaftLock() or some other guarantee that r.raftMu is held.
+func (r *Replica) SideloadedRaftMuLocked() SideloadStorage {
+	return r.raftMu.sideloaded
 }
 
 func MakeSSTable(key, value string, ts hlc.Timestamp) ([]byte, engine.MVCCKeyValue) {
@@ -513,8 +436,8 @@ func SetMockAddSSTable() (undo func()) {
 		args := cArgs.Args.(*roachpb.AddSSTableRequest)
 
 		return result.Result{
-			Replicated: storagebase.ReplicatedEvalResult{
-				AddSSTable: &storagebase.ReplicatedEvalResult_AddSSTable{
+			Replicated: storagepb.ReplicatedEvalResult{
+				AddSSTable: &storagepb.ReplicatedEvalResult_AddSSTable{
 					Data:  args.Data,
 					CRC32: util.CRC32(args.Data),
 				},
@@ -554,13 +477,13 @@ func (r *Replica) UnquiesceAndWakeLeader() {
 }
 
 func (nl *NodeLiveness) SetDrainingInternal(
-	ctx context.Context, liveness *Liveness, drain bool,
+	ctx context.Context, liveness *storagepb.Liveness, drain bool,
 ) error {
 	return nl.setDrainingInternal(ctx, liveness, drain)
 }
 
 func (nl *NodeLiveness) SetDecommissioningInternal(
-	ctx context.Context, nodeID roachpb.NodeID, liveness *Liveness, decommission bool,
+	ctx context.Context, nodeID roachpb.NodeID, liveness *storagepb.Liveness, decommission bool,
 ) (changeCommitted bool, err error) {
 	return nl.setDecommissioningInternal(ctx, nodeID, liveness, decommission)
 }
