@@ -21,6 +21,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/storage/batcheval/result"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
+	"github.com/cockroachdb/cockroach/pkg/storage/rangefeed"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
@@ -66,57 +67,6 @@ var (
 		Unit:        metric.Unit_COUNT,
 	}
 
-	// Replica CommandQueue metrics. Max size metrics track the maximum value
-	// seen for all replicas during a single replica scan.
-	metaMaxCommandQueueSize = metric.Metadata{
-		Name:        "replicas.commandqueue.maxsize",
-		Help:        "Largest number of commands in any CommandQueue",
-		Measurement: "Commands",
-		Unit:        metric.Unit_COUNT,
-	}
-	metaMaxCommandQueueWriteCount = metric.Metadata{
-		Name:        "replicas.commandqueue.maxwritecount",
-		Help:        "Largest number of read-write commands in any CommandQueue",
-		Measurement: "Commands",
-		Unit:        metric.Unit_COUNT,
-	}
-	metaMaxCommandQueueReadCount = metric.Metadata{
-		Name:        "replicas.commandqueue.maxreadcount",
-		Help:        "Largest number of read-only commands in any CommandQueue",
-		Measurement: "Commands",
-		Unit:        metric.Unit_COUNT,
-	}
-	metaMaxCommandQueueTreeSize = metric.Metadata{
-		Name:        "replicas.commandqueue.maxtreesize",
-		Help:        "Largest number of intervals in any CommandQueue's interval tree",
-		Measurement: "Intervals",
-		Unit:        metric.Unit_COUNT,
-	}
-	metaMaxCommandQueueOverlaps = metric.Metadata{
-		Name:        "replicas.commandqueue.maxoverlaps",
-		Help:        "Largest number of overlapping commands seen when adding to any CommandQueue",
-		Measurement: "Commands",
-		Unit:        metric.Unit_COUNT,
-	}
-	metaCombinedCommandQueueSize = metric.Metadata{
-		Name:        "replicas.commandqueue.combinedqueuesize",
-		Help:        "Number of commands in all CommandQueues combined",
-		Measurement: "Commands",
-		Unit:        metric.Unit_COUNT,
-	}
-	metaCombinedCommandWriteCount = metric.Metadata{
-		Name:        "replicas.commandqueue.combinedwritecount",
-		Help:        "Number of read-write commands in all CommandQueues combined",
-		Measurement: "Commands",
-		Unit:        metric.Unit_COUNT,
-	}
-	metaCombinedCommandReadCount = metric.Metadata{
-		Name:        "replicas.commandqueue.combinedreadcount",
-		Help:        "Number of read-only commands in all CommandQueues combined",
-		Measurement: "Commands",
-		Unit:        metric.Unit_COUNT,
-	}
-
 	// Range metrics.
 	metaRangeCount = metric.Metadata{
 		Name:        "ranges",
@@ -136,6 +86,12 @@ var (
 		Measurement: "Ranges",
 		Unit:        metric.Unit_COUNT,
 	}
+	metaOverReplicatedRangeCount = metric.Metadata{
+		Name:        "ranges.overreplicated",
+		Help:        "Number of ranges with more live replicas than the replication target",
+		Measurement: "Ranges",
+		Unit:        metric.Unit_COUNT,
+	}
 
 	// Lease request metrics.
 	metaLeaseRequestSuccessCount = metric.Metadata{
@@ -330,6 +286,14 @@ var (
 		Unit:        metric.Unit_COUNT,
 	}
 
+	// Metric for tracking follower reads.
+	metaFollowerReadsCount = metric.Metadata{
+		Name:        "follower_reads.success_count",
+		Help:        "Number of reads successfully processed by any replica",
+		Measurement: "Read Ops",
+		Unit:        metric.Unit_COUNT,
+	}
+
 	// RocksDB metrics.
 	metaRdbBlockCacheHits = metric.Metadata{
 		Name:        "rocksdb.block.cache.hits",
@@ -491,6 +455,18 @@ var (
 		Measurement: "Latency",
 		Unit:        metric.Unit_NANOSECONDS,
 	}
+	metaRaftHandleReadyLatency = metric.Metadata{
+		Name:        "raft.process.handleready.latency",
+		Help:        "Latency histogram for handling a Raft ready",
+		Measurement: "Latency",
+		Unit:        metric.Unit_NANOSECONDS,
+	}
+	metaRaftApplyCommittedLatency = metric.Metadata{
+		Name:        "raft.process.applycommitted.latency",
+		Help:        "Latency histogram for applying all committed Raft commands in a Raft ready",
+		Measurement: "Latency",
+		Unit:        metric.Unit_NANOSECONDS,
+	}
 
 	// Raft message metrics.
 	metaRaftRcvdProp = metric.Metadata{
@@ -914,18 +890,10 @@ var (
 		Unit:        metric.Unit_COUNT,
 	}
 
-	// Intent resolver metrics.
-	metaIntentResolverAsyncThrottled = metric.Metadata{
-		Name:        "intentresolver.async.throttled",
-		Help:        "Number of intent resolution attempts not run asynchronously due to throttling",
-		Measurement: "Intent Resolutions",
-		Unit:        metric.Unit_COUNT,
-	}
-
 	// Slow request metrics.
-	metaSlowCommandQueueRequests = metric.Metadata{
-		Name:        "requests.slow.commandqueue",
-		Help:        "Number of requests that have been stuck for a long time in the command queue",
+	metaLatchRequests = metric.Metadata{
+		Name:        "requests.slow.latch",
+		Help:        "Number of requests that have been stuck for a long time acquiring latches",
 		Measurement: "Requests",
 		Unit:        metric.Unit_COUNT,
 	}
@@ -969,6 +937,23 @@ var (
 		Measurement: "Ingestions",
 		Unit:        metric.Unit_COUNT,
 	}
+
+	// Encryption-at-rest metrics.
+	// TODO(mberhault): metrics for key age, per-key file/bytes counts.
+	metaEncryptionAlgorithm = metric.Metadata{
+		Name:        "rocksdb.encryption.algorithm",
+		Help:        "algorithm in use for encryption-at-rest, see ccl/storageccl/engineccl/enginepbccl/key_registry.proto",
+		Measurement: "Encryption At Rest",
+		Unit:        metric.Unit_CONST,
+	}
+
+	// Closed timestamp metrics.
+	metaClosedTimestampMaxBehindNanos = metric.Metadata{
+		Name:        "kv.closed_timestamp.max_behind_nanos",
+		Help:        "Largest latency between realtime and replica max closed timestamp",
+		Measurement: "Nanoseconds",
+		Unit:        metric.Unit_NANOSECONDS,
+	}
 )
 
 // StoreMetrics is the set of metrics for a given store.
@@ -983,20 +968,11 @@ type StoreMetrics struct {
 	LeaseHolderCount              *metric.Gauge
 	QuiescentCount                *metric.Gauge
 
-	// Replica CommandQueue metrics.
-	MaxCommandQueueSize       *metric.Gauge
-	MaxCommandQueueWriteCount *metric.Gauge
-	MaxCommandQueueReadCount  *metric.Gauge
-	MaxCommandQueueTreeSize   *metric.Gauge
-	MaxCommandQueueOverlaps   *metric.Gauge
-	CombinedCommandQueueSize  *metric.Gauge
-	CombinedCommandWriteCount *metric.Gauge
-	CombinedCommandReadCount  *metric.Gauge
-
 	// Range metrics.
 	RangeCount                *metric.Gauge
 	UnavailableRangeCount     *metric.Gauge
 	UnderReplicatedRangeCount *metric.Gauge
+	OverReplicatedRangeCount  *metric.Gauge
 
 	// Lease request metrics for successful and failed lease requests. These
 	// count proposals (i.e. it does not matter how many replicas apply the
@@ -1035,6 +1011,9 @@ type StoreMetrics struct {
 	AverageQueriesPerSecond *metric.GaugeFloat64
 	AverageWritesPerSecond  *metric.GaugeFloat64
 
+	// Follower read metrics.
+	FollowerReadsCount *metric.Counter
+
 	// RocksDB metrics.
 	RdbBlockCacheHits           *metric.Gauge
 	RdbBlockCacheMisses         *metric.Gauge
@@ -1065,12 +1044,14 @@ type StoreMetrics struct {
 	RangeRaftLeaderTransfers        *metric.Counter
 
 	// Raft processing metrics.
-	RaftTicks                *metric.Counter
-	RaftWorkingDurationNanos *metric.Counter
-	RaftTickingDurationNanos *metric.Counter
-	RaftCommandsApplied      *metric.Counter
-	RaftLogCommitLatency     *metric.Histogram
-	RaftCommandCommitLatency *metric.Histogram
+	RaftTicks                 *metric.Counter
+	RaftWorkingDurationNanos  *metric.Counter
+	RaftTickingDurationNanos  *metric.Counter
+	RaftCommandsApplied       *metric.Counter
+	RaftLogCommitLatency      *metric.Histogram
+	RaftCommandCommitLatency  *metric.Histogram
+	RaftHandleReadyLatency    *metric.Histogram
+	RaftApplyCommittedLatency *metric.Histogram
 
 	// Raft message metrics.
 	RaftRcvdMsgProp           *metric.Counter
@@ -1156,13 +1137,10 @@ type StoreMetrics struct {
 	GCResolveTotal               *metric.Counter
 	GCResolveSuccess             *metric.Counter
 
-	// Intent resolver metrics.
-	IntentResolverAsyncThrottled *metric.Counter
-
 	// Slow request counts.
-	SlowCommandQueueRequests *metric.Gauge
-	SlowLeaseRequests        *metric.Gauge
-	SlowRaftRequests         *metric.Gauge
+	SlowLatchRequests *metric.Gauge
+	SlowLeaseRequests *metric.Gauge
+	SlowRaftRequests  *metric.Gauge
 
 	// Backpressure counts.
 	BackpressuredOnSplitRequests *metric.Gauge
@@ -1173,6 +1151,16 @@ type StoreMetrics struct {
 	AddSSTableApplications      *metric.Counter
 	AddSSTableApplicationCopies *metric.Counter
 
+	// Encryption-at-rest stats.
+	// EncryptionAlgorithm is an enum representing the cipher in use, so we use a gauge.
+	EncryptionAlgorithm *metric.Gauge
+
+	// RangeFeed counts.
+	RangeFeedMetrics *rangefeed.Metrics
+
+	// Closed timestamp metrics.
+	ClosedTimestampMaxBehindNanos *metric.Gauge
+
 	// Stats for efficient merges.
 	mu struct {
 		syncutil.Mutex
@@ -1193,20 +1181,11 @@ func newStoreMetrics(histogramWindow time.Duration) *StoreMetrics {
 		LeaseHolderCount:              metric.NewGauge(metaLeaseHolderCount),
 		QuiescentCount:                metric.NewGauge(metaQuiescentCount),
 
-		// Replica CommandQueue metrics.
-		MaxCommandQueueSize:       metric.NewGauge(metaMaxCommandQueueSize),
-		MaxCommandQueueWriteCount: metric.NewGauge(metaMaxCommandQueueWriteCount),
-		MaxCommandQueueReadCount:  metric.NewGauge(metaMaxCommandQueueReadCount),
-		MaxCommandQueueTreeSize:   metric.NewGauge(metaMaxCommandQueueTreeSize),
-		MaxCommandQueueOverlaps:   metric.NewGauge(metaMaxCommandQueueOverlaps),
-		CombinedCommandQueueSize:  metric.NewGauge(metaCombinedCommandQueueSize),
-		CombinedCommandWriteCount: metric.NewGauge(metaCombinedCommandWriteCount),
-		CombinedCommandReadCount:  metric.NewGauge(metaCombinedCommandReadCount),
-
 		// Range metrics.
 		RangeCount:                metric.NewGauge(metaRangeCount),
 		UnavailableRangeCount:     metric.NewGauge(metaUnavailableRangeCount),
 		UnderReplicatedRangeCount: metric.NewGauge(metaUnderReplicatedRangeCount),
+		OverReplicatedRangeCount:  metric.NewGauge(metaOverReplicatedRangeCount),
 
 		// Lease request metrics.
 		LeaseRequestSuccessCount:  metric.NewCounter(metaLeaseRequestSuccessCount),
@@ -1245,6 +1224,9 @@ func newStoreMetrics(histogramWindow time.Duration) *StoreMetrics {
 		AverageQueriesPerSecond: metric.NewGaugeFloat64(metaAverageQueriesPerSecond),
 		AverageWritesPerSecond:  metric.NewGaugeFloat64(metaAverageWritesPerSecond),
 
+		// Follower reads metrics.
+		FollowerReadsCount: metric.NewCounter(metaFollowerReadsCount),
+
 		// RocksDB metrics.
 		RdbBlockCacheHits:           metric.NewGauge(metaRdbBlockCacheHits),
 		RdbBlockCacheMisses:         metric.NewGauge(metaRdbBlockCacheMisses),
@@ -1270,12 +1252,14 @@ func newStoreMetrics(histogramWindow time.Duration) *StoreMetrics {
 		RangeRaftLeaderTransfers:        metric.NewCounter(metaRangeRaftLeaderTransfers),
 
 		// Raft processing metrics.
-		RaftTicks:                metric.NewCounter(metaRaftTicks),
-		RaftWorkingDurationNanos: metric.NewCounter(metaRaftWorkingDurationNanos),
-		RaftTickingDurationNanos: metric.NewCounter(metaRaftTickingDurationNanos),
-		RaftCommandsApplied:      metric.NewCounter(metaRaftCommandsApplied),
-		RaftLogCommitLatency:     metric.NewLatency(metaRaftLogCommitLatency, histogramWindow),
-		RaftCommandCommitLatency: metric.NewLatency(metaRaftCommandCommitLatency, histogramWindow),
+		RaftTicks:                 metric.NewCounter(metaRaftTicks),
+		RaftWorkingDurationNanos:  metric.NewCounter(metaRaftWorkingDurationNanos),
+		RaftTickingDurationNanos:  metric.NewCounter(metaRaftTickingDurationNanos),
+		RaftCommandsApplied:       metric.NewCounter(metaRaftCommandsApplied),
+		RaftLogCommitLatency:      metric.NewLatency(metaRaftLogCommitLatency, histogramWindow),
+		RaftCommandCommitLatency:  metric.NewLatency(metaRaftCommandCommitLatency, histogramWindow),
+		RaftHandleReadyLatency:    metric.NewLatency(metaRaftHandleReadyLatency, histogramWindow),
+		RaftApplyCommittedLatency: metric.NewLatency(metaRaftApplyCommittedLatency, histogramWindow),
 
 		// Raft message metrics.
 		RaftRcvdMsgProp:           metric.NewCounter(metaRaftRcvdProp),
@@ -1359,13 +1343,10 @@ func newStoreMetrics(histogramWindow time.Duration) *StoreMetrics {
 		GCResolveTotal:               metric.NewCounter(metaGCResolveTotal),
 		GCResolveSuccess:             metric.NewCounter(metaGCResolveSuccess),
 
-		// Intent resolver metrics.
-		IntentResolverAsyncThrottled: metric.NewCounter(metaIntentResolverAsyncThrottled),
-
 		// Wedge request counters.
-		SlowCommandQueueRequests: metric.NewGauge(metaSlowCommandQueueRequests),
-		SlowLeaseRequests:        metric.NewGauge(metaSlowLeaseRequests),
-		SlowRaftRequests:         metric.NewGauge(metaSlowRaftRequests),
+		SlowLatchRequests: metric.NewGauge(metaLatchRequests),
+		SlowLeaseRequests: metric.NewGauge(metaSlowLeaseRequests),
+		SlowRaftRequests:  metric.NewGauge(metaSlowRaftRequests),
 
 		// Backpressure counters.
 		BackpressuredOnSplitRequests: metric.NewGauge(metaBackpressuredOnSplitRequests),
@@ -1374,6 +1355,15 @@ func newStoreMetrics(histogramWindow time.Duration) *StoreMetrics {
 		AddSSTableProposals:         metric.NewCounter(metaAddSSTableProposals),
 		AddSSTableApplications:      metric.NewCounter(metaAddSSTableApplications),
 		AddSSTableApplicationCopies: metric.NewCounter(metaAddSSTableApplicationCopies),
+
+		// Encryption-at-rest.
+		EncryptionAlgorithm: metric.NewGauge(metaEncryptionAlgorithm),
+
+		// RangeFeed counters.
+		RangeFeedMetrics: rangefeed.NewMetrics(),
+
+		// Closed timestamp metrics.
+		ClosedTimestampMaxBehindNanos: metric.NewGauge(metaClosedTimestampMaxBehindNanos),
 	}
 
 	sm.raftRcvdMessages[raftpb.MsgProp] = sm.RaftRcvdMsgProp
@@ -1447,6 +1437,10 @@ func (sm *StoreMetrics) updateRocksDBStats(stats engine.Stats) {
 	sm.RdbTableReadersMemEstimate.Update(stats.TableReadersMemEstimate)
 }
 
+func (sm *StoreMetrics) updateEnvStats(stats engine.EnvStats) {
+	sm.EncryptionAlgorithm.Update(int64(stats.EncryptionType))
+}
+
 func (sm *StoreMetrics) handleMetricsResult(ctx context.Context, metric result.Metrics) {
 	sm.LeaseRequestSuccessCount.Inc(int64(metric.LeaseRequestSuccess))
 	metric.LeaseRequestSuccess = 0
