@@ -20,8 +20,6 @@ import (
 	"sync/atomic"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
@@ -31,6 +29,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/storage/closedts"
 	"github.com/cockroachdb/cockroach/pkg/storage/closedts/ctpb"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
@@ -39,6 +39,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/pkg/errors"
 )
 
 var (
@@ -52,6 +53,11 @@ var (
 	// the underlying liveness record has had its epoch incremented.
 	ErrEpochIncremented = errors.New("heartbeat failed on epoch increment")
 
+	// ErrEpochAlreadyIncremented is returned by IncrementEpoch when
+	// someone else has already incremented the epoch to the desired
+	// value.
+	ErrEpochAlreadyIncremented = errors.New("epoch already incremented")
+
 	errLiveClockNotLive = errors.New("not live")
 )
 
@@ -101,48 +107,6 @@ var (
 	}
 )
 
-// IsLive returns whether the node is considered live at the given time with the
-// given clock offset.
-func (l *Liveness) IsLive(now hlc.Timestamp, maxOffset time.Duration) bool {
-	if maxOffset == timeutil.ClocklessMaxOffset {
-		// When using clockless reads, we're live without a buffer period.
-		maxOffset = 0
-	}
-	expiration := hlc.Timestamp(l.Expiration).Add(-maxOffset.Nanoseconds(), 0)
-	return now.Less(expiration)
-}
-
-// IsDead returns whether the node is considered dead at the given time with the
-// given threshold.
-func (l *Liveness) IsDead(now hlc.Timestamp, threshold time.Duration) bool {
-	deadAsOf := hlc.Timestamp(l.Expiration).GoTime().Add(threshold)
-	return !now.GoTime().Before(deadAsOf)
-}
-
-// LivenessStatus returns a NodeLivenessStatus enumeration value for this liveness
-// based on the provided timestamp, threshold, and clock max offset.
-func (l *Liveness) LivenessStatus(
-	now time.Time, threshold, maxOffset time.Duration,
-) NodeLivenessStatus {
-	nowHlc := hlc.Timestamp{WallTime: now.UnixNano()}
-	if l.IsDead(nowHlc, threshold) {
-		if l.Decommissioning {
-			return NodeLivenessStatus_DECOMMISSIONED
-		}
-		return NodeLivenessStatus_DEAD
-	}
-	if l.Decommissioning {
-		return NodeLivenessStatus_DECOMMISSIONING
-	}
-	if l.Draining {
-		return NodeLivenessStatus_UNAVAILABLE
-	}
-	if l.IsLive(nowHlc, maxOffset) {
-		return NodeLivenessStatus_LIVE
-	}
-	return NodeLivenessStatus_UNAVAILABLE
-}
-
 // LivenessMetrics holds metrics for use with node liveness activity.
 type LivenessMetrics struct {
 	LiveNodes          *metric.Gauge
@@ -188,7 +152,7 @@ type NodeLiveness struct {
 	mu struct {
 		syncutil.Mutex
 		callbacks         []IsLiveCallback
-		nodes             map[roachpb.NodeID]Liveness
+		nodes             map[roachpb.NodeID]storagepb.Liveness
 		heartbeatCallback HeartbeatCallback
 	}
 }
@@ -226,7 +190,7 @@ func NewNodeLiveness(
 		EpochIncrements:    metric.NewCounter(metaEpochIncrements),
 		HeartbeatLatency:   metric.NewLatency(metaHeartbeatLatency, histogramWindow),
 	}
-	nl.mu.nodes = map[roachpb.NodeID]Liveness{}
+	nl.mu.nodes = map[roachpb.NodeID]storagepb.Liveness{}
 	nl.heartbeatToken <- struct{}{}
 
 	livenessRegex := gossip.MakePrefixPattern(gossip.KeyNodeLivenessPrefix)
@@ -304,11 +268,11 @@ func (nl *NodeLiveness) SetDecommissioning(
 		// decommissioning commands in a tight loop on different nodes sometimes
 		// results in unintentional no-ops (due to the Gossip lag); this could be
 		// observed by users in principle, too.
-		var oldLiveness Liveness
+		var oldLiveness storagepb.Liveness
 		if err := nl.db.GetProto(ctx, keys.NodeLivenessKey(nodeID), &oldLiveness); err != nil {
 			return false, errors.Wrap(err, "unable to get liveness")
 		}
-		if (oldLiveness == Liveness{}) {
+		if (oldLiveness == storagepb.Liveness{}) {
 			return false, ErrNoLivenessRecord
 		}
 
@@ -331,7 +295,7 @@ func (nl *NodeLiveness) SetDecommissioning(
 }
 
 func (nl *NodeLiveness) setDrainingInternal(
-	ctx context.Context, liveness *Liveness, drain bool,
+	ctx context.Context, liveness *storagepb.Liveness, drain bool,
 ) error {
 	nodeID := nl.gossip.NodeID.Get()
 	sem := nl.sem(nodeID)
@@ -346,7 +310,7 @@ func (nl *NodeLiveness) setDrainingInternal(
 	}()
 
 	update := livenessUpdate{
-		Liveness: Liveness{
+		Liveness: storagepb.Liveness{
 			NodeID: nodeID,
 			Epoch:  1,
 		},
@@ -357,7 +321,7 @@ func (nl *NodeLiveness) setDrainingInternal(
 	update.Draining = drain
 	update.ignoreCache = true
 
-	if err := nl.updateLiveness(ctx, update, liveness, func(actual Liveness) error {
+	if err := nl.updateLiveness(ctx, update, liveness, func(actual storagepb.Liveness) error {
 		nl.maybeUpdate(actual)
 		if actual.Draining == update.Draining {
 			return errNodeDrainingSet
@@ -374,7 +338,7 @@ func (nl *NodeLiveness) setDrainingInternal(
 }
 
 type livenessUpdate struct {
-	Liveness
+	storagepb.Liveness
 	// When ignoreCache is set, we won't assume that our in-memory cached version
 	// of the liveness record is accurate and will use a CPut on the liveness
 	// table with whatever the client supplied. This is used for operations that
@@ -383,10 +347,10 @@ type livenessUpdate struct {
 }
 
 func (nl *NodeLiveness) setDecommissioningInternal(
-	ctx context.Context, nodeID roachpb.NodeID, liveness *Liveness, decommission bool,
+	ctx context.Context, nodeID roachpb.NodeID, liveness *storagepb.Liveness, decommission bool,
 ) (changeCommitted bool, err error) {
 	update := livenessUpdate{
-		Liveness: Liveness{
+		Liveness: storagepb.Liveness{
 			NodeID: nodeID,
 			Epoch:  1,
 		},
@@ -398,7 +362,7 @@ func (nl *NodeLiveness) setDecommissioningInternal(
 	update.ignoreCache = true
 
 	var conditionFailed bool
-	if err := nl.updateLiveness(ctx, update, liveness, func(actual Liveness) error {
+	if err := nl.updateLiveness(ctx, update, liveness, func(actual storagepb.Liveness) error {
 		conditionFailed = true
 		if actual.Decommissioning == update.Decommissioning {
 			return nil
@@ -440,7 +404,7 @@ func (nl *NodeLiveness) IsHealthy(nodeID roachpb.NodeID) (bool, error) {
 		nl.GetLivenessThreshold(),
 		nl.clock.MaxOffset(),
 	)
-	return ls == NodeLivenessStatus_LIVE, nil
+	return ls == storagepb.NodeLivenessStatus_LIVE, nil
 }
 
 // StartHeartbeat starts a periodic heartbeat to refresh this node's
@@ -474,30 +438,31 @@ func (nl *NodeLiveness) StartHeartbeat(
 			case <-stopper.ShouldStop():
 				return
 			}
-			func(ctx context.Context) {
-				// Give the context a timeout approximately as long as the time we
-				// have left before our liveness entry expires.
-				ctx, cancel := context.WithTimeout(ctx, nl.livenessThreshold-nl.heartbeatInterval)
-				defer cancel()
-
-				// Retry heartbeat in the event the conditional put fails.
-				for r := retry.StartWithCtx(ctx, retryOpts); r.Next(); {
-					liveness, err := nl.Self()
-					if err != nil && err != ErrNoLivenessRecord {
-						log.Errorf(ctx, "unexpected error getting liveness: %v", err)
-					}
-					if err := nl.heartbeatInternal(ctx, liveness, incrementEpoch); err != nil {
-						if err == ErrEpochIncremented {
-							log.Infof(ctx, "%s; retrying", err)
-							continue
+			// Give the context a timeout approximately as long as the time we
+			// have left before our liveness entry expires.
+			if err := contextutil.RunWithTimeout(ctx, "node liveness heartbeat", nl.livenessThreshold-nl.heartbeatInterval,
+				func(ctx context.Context) error {
+					// Retry heartbeat in the event the conditional put fails.
+					for r := retry.StartWithCtx(ctx, retryOpts); r.Next(); {
+						liveness, err := nl.Self()
+						if err != nil && err != ErrNoLivenessRecord {
+							log.Errorf(ctx, "unexpected error getting liveness: %v", err)
+						}
+						if err := nl.heartbeatInternal(ctx, liveness, incrementEpoch); err != nil {
+							if err == ErrEpochIncremented {
+								log.Infof(ctx, "%s; retrying", err)
+								continue
+							}
+							return err
 						}
-						log.Warningf(ctx, "failed node liveness heartbeat: %v", err)
-					} else {
 						incrementEpoch = false // don't increment epoch after first heartbeat
+						break
 					}
-					break
-				}
-			}(ctx)
+					return nil
+				}); err != nil {
+				log.Warningf(ctx, "failed node liveness heartbeat: %v", err)
+			}
+
 			nl.heartbeatToken <- struct{}{}
 			select {
 			case <-ticker.C:
@@ -542,15 +507,15 @@ var errNodeAlreadyLive = errors.New("node already live")
 // Heartbeat is called to update a node's expiration timestamp. This
 // method does a conditional put on the node liveness record, and if
 // successful, stores the updated liveness record in the nodes map.
-func (nl *NodeLiveness) Heartbeat(ctx context.Context, liveness *Liveness) error {
+func (nl *NodeLiveness) Heartbeat(ctx context.Context, liveness *storagepb.Liveness) error {
 	return nl.heartbeatInternal(ctx, liveness, false /* increment epoch */)
 }
 
 func (nl *NodeLiveness) heartbeatInternal(
-	ctx context.Context, liveness *Liveness, incrementEpoch bool,
+	ctx context.Context, liveness *storagepb.Liveness, incrementEpoch bool,
 ) error {
-	ctx, finish := tracing.EnsureChildSpan(ctx, nl.ambientCtx.Tracer, "liveness heartbeat")
-	defer finish()
+	ctx, sp := tracing.EnsureChildSpan(ctx, nl.ambientCtx.Tracer, "liveness heartbeat")
+	defer sp.Finish()
 	defer func(start time.Time) {
 		dur := timeutil.Now().Sub(start)
 		nl.metrics.HeartbeatLatency.RecordValue(dur.Nanoseconds())
@@ -572,7 +537,7 @@ func (nl *NodeLiveness) heartbeatInternal(
 	}()
 
 	update := livenessUpdate{
-		Liveness: Liveness{
+		Liveness: storagepb.Liveness{
 			NodeID: nodeID,
 			Epoch:  1,
 		},
@@ -603,7 +568,7 @@ func (nl *NodeLiveness) heartbeatInternal(
 			return errors.Errorf("proposed liveness update expires earlier than previous record")
 		}
 	}
-	if err := nl.updateLiveness(ctx, update, liveness, func(actual Liveness) error {
+	if err := nl.updateLiveness(ctx, update, liveness, func(actual storagepb.Liveness) error {
 		// Update liveness to actual value on mismatch.
 		nl.maybeUpdate(actual)
 		// If the actual liveness is different than expected, but is
@@ -634,19 +599,29 @@ func (nl *NodeLiveness) heartbeatInternal(
 // is returned in the event that the node has neither heartbeat its
 // liveness record successfully, nor received a gossip message containing
 // a former liveness update on restart.
-func (nl *NodeLiveness) Self() (*Liveness, error) {
+func (nl *NodeLiveness) Self() (*storagepb.Liveness, error) {
 	nl.mu.Lock()
 	defer nl.mu.Unlock()
 	return nl.getLivenessLocked(nl.gossip.NodeID.Get())
 }
 
+// IsLiveMapEntry encapsulates data about current liveness for a
+// node.
+type IsLiveMapEntry struct {
+	IsLive bool
+	Epoch  int64
+}
+
+// IsLiveMap is a type alias for a map from NodeID to IsLiveMapEntry.
+type IsLiveMap map[roachpb.NodeID]IsLiveMapEntry
+
 // GetIsLiveMap returns a map of nodeID to boolean liveness status of
 // each node. This excludes nodes that were removed completely (dead +
 // decommissioning).
-func (nl *NodeLiveness) GetIsLiveMap() map[roachpb.NodeID]bool {
+func (nl *NodeLiveness) GetIsLiveMap() IsLiveMap {
 	nl.mu.Lock()
 	defer nl.mu.Unlock()
-	lMap := map[roachpb.NodeID]bool{}
+	lMap := IsLiveMap{}
 	now := nl.clock.Now()
 	maxOffset := nl.clock.MaxOffset()
 	for nID, l := range nl.mu.nodes {
@@ -655,7 +630,10 @@ func (nl *NodeLiveness) GetIsLiveMap() map[roachpb.NodeID]bool {
 			// This is a node that was completely removed. Skip over it.
 			continue
 		}
-		lMap[nID] = isLive
+		lMap[nID] = IsLiveMapEntry{
+			IsLive: isLive,
+			Epoch:  l.Epoch,
+		}
 	}
 	return lMap
 }
@@ -663,10 +641,10 @@ func (nl *NodeLiveness) GetIsLiveMap() map[roachpb.NodeID]bool {
 // GetLivenesses returns a slice containing the liveness status of
 // every node on the cluster known to gossip. Callers should consider
 // calling (statusServer).NodesWithLiveness() instead where possible.
-func (nl *NodeLiveness) GetLivenesses() []Liveness {
+func (nl *NodeLiveness) GetLivenesses() []storagepb.Liveness {
 	nl.mu.Lock()
 	defer nl.mu.Unlock()
-	livenesses := make([]Liveness, 0, len(nl.mu.nodes))
+	livenesses := make([]storagepb.Liveness, 0, len(nl.mu.nodes))
 	for _, l := range nl.mu.nodes {
 		livenesses = append(livenesses, l)
 	}
@@ -676,7 +654,7 @@ func (nl *NodeLiveness) GetLivenesses() []Liveness {
 // GetLiveness returns the liveness record for the specified nodeID.
 // ErrNoLivenessRecord is returned in the event that nothing is yet
 // known about nodeID via liveness gossip.
-func (nl *NodeLiveness) GetLiveness(nodeID roachpb.NodeID) (*Liveness, error) {
+func (nl *NodeLiveness) GetLiveness(nodeID roachpb.NodeID) (*storagepb.Liveness, error) {
 	nl.mu.Lock()
 	defer nl.mu.Unlock()
 	return nl.getLivenessLocked(nodeID)
@@ -689,13 +667,13 @@ func (nl *NodeLiveness) GetLiveness(nodeID roachpb.NodeID) (*Liveness, error) {
 //
 // GetLivenessStatusMap() includes removed nodes (dead +
 // decommissioned).
-func (nl *NodeLiveness) GetLivenessStatusMap() map[roachpb.NodeID]NodeLivenessStatus {
+func (nl *NodeLiveness) GetLivenessStatusMap() map[roachpb.NodeID]storagepb.NodeLivenessStatus {
 	now := nl.clock.PhysicalTime()
 	livenesses := nl.GetLivenesses()
 	threshold := TimeUntilStoreDead.Get(&nl.st.SV)
 	maxOffset := nl.clock.MaxOffset()
 
-	statusMap := make(map[roachpb.NodeID]NodeLivenessStatus, len(livenesses))
+	statusMap := make(map[roachpb.NodeID]storagepb.NodeLivenessStatus, len(livenesses))
 	for _, liveness := range livenesses {
 		status := liveness.LivenessStatus(
 			now, threshold, maxOffset,
@@ -705,15 +683,13 @@ func (nl *NodeLiveness) GetLivenessStatusMap() map[roachpb.NodeID]NodeLivenessSt
 	return statusMap
 }
 
-func (nl *NodeLiveness) getLivenessLocked(nodeID roachpb.NodeID) (*Liveness, error) {
+func (nl *NodeLiveness) getLivenessLocked(nodeID roachpb.NodeID) (*storagepb.Liveness, error) {
 	if l, ok := nl.mu.nodes[nodeID]; ok {
 		return &l, nil
 	}
 	return nil, ErrNoLivenessRecord
 }
 
-var errEpochAlreadyIncremented = errors.New("epoch already incremented")
-
 // IncrementEpoch is called to increment the current liveness epoch,
 // thereby invalidating anything relying on the liveness of the
 // previous epoch. This method does a conditional put on the node
@@ -721,7 +697,7 @@ var errEpochAlreadyIncremented = errors.New("epoch already incremented")
 // record in the nodes map. If this method is called on a node ID
 // which is considered live according to the most recent information
 // gathered through gossip, an error is returned.
-func (nl *NodeLiveness) IncrementEpoch(ctx context.Context, liveness *Liveness) error {
+func (nl *NodeLiveness) IncrementEpoch(ctx context.Context, liveness *storagepb.Liveness) error {
 	// Allow only one increment at a time.
 	sem := nl.sem(liveness.NodeID)
 	select {
@@ -738,18 +714,15 @@ func (nl *NodeLiveness) IncrementEpoch(ctx context.Context, liveness *Liveness)
 	}
 	update := livenessUpdate{Liveness: *liveness}
 	update.Epoch++
-	if err := nl.updateLiveness(ctx, update, liveness, func(actual Liveness) error {
+	if err := nl.updateLiveness(ctx, update, liveness, func(actual storagepb.Liveness) error {
 		defer nl.maybeUpdate(actual)
 		if actual.Epoch > liveness.Epoch {
-			return errEpochAlreadyIncremented
+			return ErrEpochAlreadyIncremented
 		} else if actual.Epoch < liveness.Epoch {
 			return errors.Errorf("unexpected liveness epoch %d; expected >= %d", actual.Epoch, liveness.Epoch)
 		}
 		return errors.Errorf("mismatch incrementing epoch for %+v; actual is %+v", *liveness, actual)
 	}); err != nil {
-		if err == errEpochAlreadyIncremented {
-			return nil
-		}
 		return err
 	}
 
@@ -788,29 +761,14 @@ func (nl *NodeLiveness) RegisterCallback(cb IsLiveCallback) {
 func (nl *NodeLiveness) updateLiveness(
 	ctx context.Context,
 	update livenessUpdate,
-	oldLiveness *Liveness,
-	handleCondFailed func(actual Liveness) error,
+	oldLiveness *storagepb.Liveness,
+	handleCondFailed func(actual storagepb.Liveness) error,
 ) error {
 	for {
 		// Before each attempt, ensure that the context has not expired.
 		if err := ctx.Err(); err != nil {
 			return err
 		}
-		for _, eng := range nl.engines {
-			// Synchronously writing to all disks before updating node liveness because
-			// we don't want any excessively slow disks to prevent the lease from
-			// shifting to other nodes. If the disk is slow, batch.Commit() will block.
-			batch := eng.NewBatch()
-			defer batch.Close()
-
-			if err := batch.LogData(nil); err != nil {
-				return errors.Wrapf(err, "couldn't update node liveness because LogData to disk fails")
-			}
-
-			if err := batch.Commit(true /* sync */); err != nil {
-				return errors.Wrapf(err, "couldn't update node liveness because Commit to disk fails")
-			}
-		}
 		if err := nl.updateLivenessAttempt(ctx, update, oldLiveness, handleCondFailed); err != nil {
 			// Intentionally don't errors.Cause() the error, or we'd hop past errRetryLiveness.
 			if _, ok := err.(*errRetryLiveness); ok {
@@ -826,8 +784,8 @@ func (nl *NodeLiveness) updateLiveness(
 func (nl *NodeLiveness) updateLivenessAttempt(
 	ctx context.Context,
 	update livenessUpdate,
-	oldLiveness *Liveness,
-	handleCondFailed func(actual Liveness) error,
+	oldLiveness *storagepb.Liveness,
+	handleCondFailed func(actual storagepb.Liveness) error,
 ) error {
 	// First check the existing liveness map to avoid known conditional
 	// put failures.
@@ -868,9 +826,9 @@ func (nl *NodeLiveness) updateLivenessAttempt(
 		case *roachpb.ConditionFailedError:
 			if handleCondFailed != nil {
 				if tErr.ActualValue == nil {
-					return handleCondFailed(Liveness{})
+					return handleCondFailed(storagepb.Liveness{})
 				}
-				var actualLiveness Liveness
+				var actualLiveness storagepb.Liveness
 				if err := tErr.ActualValue.GetProto(&actualLiveness); err != nil {
 					return errors.Wrapf(err, "couldn't update node liveness from CPut actual value")
 				}
@@ -895,7 +853,7 @@ func (nl *NodeLiveness) updateLivenessAttempt(
 
 // maybeUpdate replaces the liveness (if it appears newer) and invokes the
 // registered callbacks if the node became live in the process.
-func (nl *NodeLiveness) maybeUpdate(new Liveness) {
+func (nl *NodeLiveness) maybeUpdate(new storagepb.Liveness) {
 	nl.mu.Lock()
 	// Note that this works fine even if `old` is empty.
 	old := nl.mu.nodes[new.NodeID]
@@ -919,8 +877,8 @@ func (nl *NodeLiveness) maybeUpdate(new Liveness) {
 	}
 }
 
-func shouldReplaceLiveness(old, new Liveness) bool {
-	if (old == Liveness{}) {
+func shouldReplaceLiveness(old, new storagepb.Liveness) bool {
+	if (old == storagepb.Liveness{}) {
 		return true
 	}
 
@@ -946,7 +904,7 @@ func shouldReplaceLiveness(old, new Liveness) bool {
 // livenessGossipUpdate is the gossip callback used to keep the
 // in-memory liveness info up to date.
 func (nl *NodeLiveness) livenessGossipUpdate(key string, content roachpb.Value) {
-	var liveness Liveness
+	var liveness storagepb.Liveness
 	if err := content.GetProto(&liveness); err != nil {
 		log.Error(context.TODO(), err)
 		return
