@@ -25,24 +25,23 @@ import (
 
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
-	"github.com/cockroachdb/cockroach/pkg/keys"
-	"github.com/cockroachdb/cockroach/pkg/server"
-	"github.com/cockroachdb/cockroach/pkg/testutils/testcluster"
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/gossip"
+	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/server"
 	"github.com/cockroachdb/cockroach/pkg/storage"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
+	"github.com/cockroachdb/cockroach/pkg/testutils/testcluster"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/log/logtags"
-	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/gogo/protobuf/proto"
+	"github.com/pkg/errors"
 )
 
 func verifyLiveness(t *testing.T, mtc *multiTestContext) {
@@ -317,8 +316,8 @@ func TestNodeLivenessEpochIncrement(t *testing.T) {
 		t.Errorf("expected epoch increment == 1; got %d", c)
 	}
 
-	// Verify noop on incrementing an already-incremented epoch.
-	if err := mtc.nodeLivenesses[0].IncrementEpoch(context.Background(), oldLiveness); err != nil {
+	// Verify error on incrementing an already-incremented epoch.
+	if err := mtc.nodeLivenesses[0].IncrementEpoch(context.Background(), oldLiveness); err != storage.ErrEpochAlreadyIncremented {
 		t.Fatalf("unexpected error incrementing a non-live node: %s", err)
 	}
 
@@ -348,7 +347,7 @@ func TestNodeLivenessRestart(t *testing.T) {
 	for _, g := range mtc.gossips {
 		key := gossip.MakeNodeLivenessKey(g.NodeID.Get())
 		expKeys = append(expKeys, key)
-		if err := g.AddInfoProto(key, &storage.Liveness{}, 0); err != nil {
+		if err := g.AddInfoProto(key, &storagepb.Liveness{}, 0); err != nil {
 			t.Fatal(err)
 		}
 	}
@@ -405,7 +404,7 @@ func TestNodeLivenessSelf(t *testing.T) {
 	// Verify liveness is properly initialized. This needs to be wrapped in a
 	// SucceedsSoon because node liveness gets initialized via an async gossip
 	// callback.
-	var liveness *storage.Liveness
+	var liveness *storagepb.Liveness
 	testutils.SucceedsSoon(t, func() error {
 		var err error
 		liveness, err = mtc.nodeLivenesses[0].GetLiveness(g.NodeID.Get())
@@ -462,7 +461,11 @@ func TestNodeLivenessGetIsLiveMap(t *testing.T) {
 	verifyLiveness(t, mtc)
 	pauseNodeLivenessHeartbeats(mtc, true)
 	lMap := mtc.nodeLivenesses[0].GetIsLiveMap()
-	expectedLMap := map[roachpb.NodeID]bool{1: true, 2: true, 3: true}
+	expectedLMap := storage.IsLiveMap{
+		1: {IsLive: true, Epoch: 1},
+		2: {IsLive: true, Epoch: 1},
+		3: {IsLive: true, Epoch: 1},
+	}
 	if !reflect.DeepEqual(expectedLMap, lMap) {
 		t.Errorf("expected liveness map %+v; got %+v", expectedLMap, lMap)
 	}
@@ -483,7 +486,11 @@ func TestNodeLivenessGetIsLiveMap(t *testing.T) {
 
 	// Now verify only node 0 is live.
 	lMap = mtc.nodeLivenesses[0].GetIsLiveMap()
-	expectedLMap = map[roachpb.NodeID]bool{1: true, 2: false, 3: false}
+	expectedLMap = storage.IsLiveMap{
+		1: {IsLive: true, Epoch: 1},
+		2: {IsLive: false, Epoch: 1},
+		3: {IsLive: false, Epoch: 1},
+	}
 	if !reflect.DeepEqual(expectedLMap, lMap) {
 		t.Errorf("expected liveness map %+v; got %+v", expectedLMap, lMap)
 	}
@@ -603,7 +610,7 @@ func TestNodeLivenessConcurrentIncrementEpochs(t *testing.T) {
 		}()
 	}
 	for i := 0; i < concurrency; i++ {
-		if err := <-errCh; err != nil {
+		if err := <-errCh; err != nil && err != storage.ErrEpochAlreadyIncremented {
 			t.Fatalf("concurrent increment epoch %d failed: %s", i, err)
 		}
 	}
@@ -636,7 +643,7 @@ func TestNodeLivenessSetDraining(t *testing.T) {
 
 	// Verify success on failed update of a liveness record that already has the
 	// given draining setting.
-	if err := mtc.nodeLivenesses[drainingNodeIdx].SetDrainingInternal(ctx, &storage.Liveness{}, false); err != nil {
+	if err := mtc.nodeLivenesses[drainingNodeIdx].SetDrainingInternal(ctx, &storagepb.Liveness{}, false); err != nil {
 		t.Fatal(err)
 	}
 
@@ -774,6 +781,9 @@ func TestNodeLivenessStatusMap(t *testing.T) {
 				// Disable replica rebalancing to ensure that the liveness range
 				// does not get out of the first node (we'll be shutting down nodes).
 				DisableReplicaRebalancing: true,
+				// Disable LBS because when the scan is happening at the rate it's happening
+				// below, it's possible that one of the system ranges trigger a split.
+				DisableLoadBasedSplitting: true,
 			},
 		},
 		RaftConfig: base.RaftConfig{
@@ -803,7 +813,7 @@ func TestNodeLivenessStatusMap(t *testing.T) {
 	zoneConfig := config.DefaultZoneConfig()
 	// Force just one replica per range to ensure that we can shut down
 	// nodes without endangering the liveness range.
-	zoneConfig.NumReplicas = 1
+	zoneConfig.NumReplicas = proto.Int32(1)
 	config.TestingSetZoneConfig(keys.MetaRangesID, zoneConfig)
 
 	log.Infof(ctx, "starting 3 more nodes")
@@ -813,6 +823,7 @@ func TestNodeLivenessStatusMap(t *testing.T) {
 
 	log.Infof(ctx, "waiting for node statuses")
 	tc.WaitForNodeStatuses(t)
+	tc.WaitForNodeLiveness(t)
 	log.Infof(ctx, "waiting done")
 
 	firstServer := tc.Server(0).(*server.TestServer)
@@ -844,21 +855,23 @@ func TestNodeLivenessStatusMap(t *testing.T) {
 	// See what comes up in the status.
 	callerNodeLiveness := firstServer.GetNodeLiveness()
 
-	type expectedStatus struct {
+	type testCase struct {
 		nodeID         roachpb.NodeID
-		expectedStatus storage.NodeLivenessStatus
+		expectedStatus storagepb.NodeLivenessStatus
 	}
-	testData := []expectedStatus{
-		{liveNodeID, storage.NodeLivenessStatus_LIVE},
-		{deadNodeID, storage.NodeLivenessStatus_DEAD},
-		{decommissioningNodeID, storage.NodeLivenessStatus_DECOMMISSIONING},
-		{removedNodeID, storage.NodeLivenessStatus_DECOMMISSIONED},
+
+	// Below we're going to check that all statuses converge and stabilize
+	// to a known situation.
+	testData := []testCase{
+		{liveNodeID, storagepb.NodeLivenessStatus_LIVE},
+		{deadNodeID, storagepb.NodeLivenessStatus_DEAD},
+		{decommissioningNodeID, storagepb.NodeLivenessStatus_DECOMMISSIONING},
+		{removedNodeID, storagepb.NodeLivenessStatus_DECOMMISSIONED},
 	}
 
 	for _, test := range testData {
-		t.Run(test.expectedStatus.String(), func(t *testing.T) {
+		t.Run(fmt.Sprintf("n%d->%s", test.nodeID, test.expectedStatus), func(t *testing.T) {
 			nodeID, expectedStatus := test.nodeID, test.expectedStatus
-			t.Parallel()
 
 			testutils.SucceedsSoon(t, func() error {
 				// Ensure that dead nodes are quickly recognized as dead by
@@ -871,21 +884,17 @@ func TestNodeLivenessStatusMap(t *testing.T) {
 				storage.TimeUntilStoreDead.Override(&firstServer.ClusterSettings().SV,
 					storage.TestTimeUntilStoreDead)
 
-				log.Infof(ctx, "checking expected status for node %d", nodeID)
+				log.Infof(ctx, "checking expected status (%s) for node %d", expectedStatus, nodeID)
 				nodeStatuses := callerNodeLiveness.GetLivenessStatusMap()
-				if st, ok := nodeStatuses[nodeID]; !ok {
-					return fmt.Errorf("%s node not in statuses", expectedStatus)
-				} else {
-					if st != expectedStatus {
-						if expectedStatus == storage.NodeLivenessStatus_DECOMMISSIONING && st == storage.NodeLivenessStatus_DECOMMISSIONED {
-							// Server somehow shut down super-fast. Tolerating the mismatch.
-							return nil
-						}
-						return fmt.Errorf("unexpected status: got %s, expected %s",
-							st, expectedStatus)
-					}
+				st, ok := nodeStatuses[nodeID]
+				if !ok {
+					return errors.Errorf("node %d: not in statuses\n", nodeID)
+				}
+				if st != expectedStatus {
+					return errors.Errorf("node %d: unexpected status: got %s, expected %s\n",
+						nodeID, st, expectedStatus,
+					)
 				}
-				log.Infof(ctx, "node %d status ok", nodeID)
 				return nil
 			})
 		})
@@ -906,7 +915,7 @@ func testNodeLivenessSetDecommissioning(t *testing.T, decommissionNodeIdx int) {
 
 	// Verify success on failed update of a liveness record that already has the
 	// given decommissioning setting.
-	if _, err := callerNodeLiveness.SetDecommissioningInternal(ctx, nodeID, &storage.Liveness{}, false); err != nil {
+	if _, err := callerNodeLiveness.SetDecommissioningInternal(ctx, nodeID, &storagepb.Liveness{}, false); err != nil {
 		t.Fatal(err)
 	}
 
@@ -964,7 +973,7 @@ func TestNodeLivenessDecommissionAbsent(t *testing.T) {
 	}
 
 	// Pretend the node was once there but isn't gossiped anywhere.
-	if err := mtc.dbs[0].CPut(ctx, keys.NodeLivenessKey(goneNodeID), &storage.Liveness{
+	if err := mtc.dbs[0].CPut(ctx, keys.NodeLivenessKey(goneNodeID), &storagepb.Liveness{
 		NodeID:     goneNodeID,
 		Epoch:      1,
 		Expiration: hlc.LegacyTimestamp(mtc.clock.Now()),
@@ -1011,12 +1020,12 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 	threshold := 5 * time.Minute
 
 	for _, tc := range []struct {
-		liveness storage.Liveness
-		expected storage.NodeLivenessStatus
+		liveness storagepb.Liveness
+		expected storagepb.NodeLivenessStatus
 	}{
 		// Valid status.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1025,11 +1034,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: false,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_LIVE,
+			expected: storagepb.NodeLivenessStatus_LIVE,
 		},
 		// Minimum bound of liveness is now + max offset.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1038,11 +1047,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: false,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_LIVE,
+			expected: storagepb.NodeLivenessStatus_LIVE,
 		},
 		// Expired status.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1051,11 +1060,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: false,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_UNAVAILABLE,
+			expected: storagepb.NodeLivenessStatus_UNAVAILABLE,
 		},
 		// Expired status.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1064,11 +1073,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: false,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_UNAVAILABLE,
+			expected: storagepb.NodeLivenessStatus_UNAVAILABLE,
 		},
 		// Max bound of expired.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1077,11 +1086,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: false,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_UNAVAILABLE,
+			expected: storagepb.NodeLivenessStatus_UNAVAILABLE,
 		},
 		// Dead status.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1090,11 +1099,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: false,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_DEAD,
+			expected: storagepb.NodeLivenessStatus_DEAD,
 		},
 		// Decommissioning.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1103,11 +1112,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: true,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_DECOMMISSIONING,
+			expected: storagepb.NodeLivenessStatus_DECOMMISSIONING,
 		},
 		// Decommissioned.
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1116,11 +1125,11 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: true,
 				Draining:        false,
 			},
-			expected: storage.NodeLivenessStatus_DECOMMISSIONED,
+			expected: storagepb.NodeLivenessStatus_DECOMMISSIONED,
 		},
 		// Draining (reports as unavailable).
 		{
-			liveness: storage.Liveness{
+			liveness: storagepb.Liveness{
 				NodeID: 1,
 				Epoch:  1,
 				Expiration: hlc.LegacyTimestamp{
@@ -1129,7 +1138,7 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 				Decommissioning: false,
 				Draining:        true,
 			},
-			expected: storage.NodeLivenessStatus_UNAVAILABLE,
+			expected: storagepb.NodeLivenessStatus_UNAVAILABLE,
 		},
 	} {
 		t.Run("", func(t *testing.T) {
@@ -1139,171 +1148,3 @@ func TestNodeLivenessLivenessStatus(t *testing.T) {
 		})
 	}
 }
-
-// TestUpdateLiveness verifies that a node always writes to its local RocksDB
-// before updating its liveness. If write is blocking, updateLiveness will also
-// block. It also tests the updateLiveness will fail if we inject the errors to
-// make write to disk fail.
-func TestUpdateLiveness(t *testing.T) {
-	defer leaktest.AfterTest(t)()
-	mtc := &multiTestContext{}
-	defer mtc.Stop()
-
-	engineStopper := stop.NewStopper()
-	mtc.engineStoppers = append(mtc.engineStoppers, engineStopper)
-	eng := &MockEngine{InMem: engine.NewInMem(roachpb.Attributes{}, 1<<20)}
-	engineStopper.AddCloser(eng)
-	mtc.engines = append(mtc.engines, eng)
-	mtc.injEngines = true
-	mtc.Start(t, 1)
-
-	// Verify liveness of all nodes for all nodes.
-	verifyLiveness(t, mtc)
-	pauseNodeLivenessHeartbeats(mtc, true)
-
-	// Advance clock past the liveness threshold to verify IsLive becomes false.
-	mtc.manualClock.Increment(mtc.nodeLivenesses[0].GetLivenessThreshold().Nanoseconds() + 1)
-	for idx, nl := range mtc.nodeLivenesses {
-		nodeID := mtc.gossips[idx].NodeID.Get()
-		live, err := nl.IsLive(nodeID)
-		if err != nil {
-			t.Error(err)
-		} else if live {
-			t.Errorf("expected node %d to be considered not-live after advancing node clock", nodeID)
-		}
-		testutils.SucceedsSoon(t, func() error {
-			if a, e := nl.Metrics().LiveNodes.Value(), int64(0); a != e {
-				return errors.Errorf("expected node %d's LiveNodes metric to be %d; got %d",
-					nodeID, e, a)
-			}
-			return nil
-		})
-	}
-
-	// Inject a customized error into engine and check the MockBatch's Commit()
-	// will fail with the exact same error.
-	const commitErr = "update liveness failed on commit"
-	errUpdateLiveness := errors.New(commitErr)
-	eng.mu.Lock()
-	eng.err = errUpdateLiveness
-	eng.mu.Unlock()
-	for idx, nl := range mtc.nodeLivenesses {
-		l, err := nl.Self()
-		if err != nil {
-			t.Fatal(err)
-		}
-		testutils.SucceedsSoon(t, func() error {
-			for {
-				nodeID := mtc.gossips[idx].NodeID.Get()
-				err := nl.Heartbeat(context.Background(), l)
-
-				if err == nil {
-					return errors.Errorf("expected node %d to fail to update its liveness because Commit() should fail", nodeID)
-				}
-
-				if testutils.IsError(err, commitErr) {
-					live, err := nl.IsLive(nodeID)
-					if err != nil {
-						return err
-					} else if live {
-						return errors.Errorf("expected node %d to be considered not live because updateLiveness should fail", nodeID)
-					}
-					break
-				}
-
-				if err == storage.ErrEpochIncremented {
-					log.Warningf(context.Background(), "retrying after %s", err)
-					continue
-				}
-
-				return err
-			}
-			return nil
-		})
-	}
-
-	// Trigger a manual heartbeat and check node is not alive within 5 ms because
-	// MockEngine's write will block for 5 ms before updating liveness.
-	// Then verify liveness is reestablished after 5 ms.
-	ch := make(chan bool)
-	eng.mu.Lock()
-	eng.ch = ch
-	eng.err = nil
-	eng.mu.Unlock()
-	for idx, nl := range mtc.nodeLivenesses {
-		l, err := nl.Self()
-		if err != nil {
-			t.Fatal(err)
-		}
-		testutils.SucceedsSoon(t, func() error {
-			for {
-				errCh := make(chan error)
-				go func() {
-					err := nl.Heartbeat(context.Background(), l)
-					errCh <- err
-				}()
-
-				nodeID := mtc.gossips[idx].NodeID.Get()
-				select {
-				case <-errCh:
-					return errors.Errorf("expected node %d to be considered not live within 5 milliseconds", nodeID)
-				case <-time.After(5 * time.Millisecond):
-					ch <- true
-					err = <-errCh
-				}
-
-				if err == nil {
-					live, err := nl.IsLive(nodeID)
-					if err != nil {
-						return err
-					} else if !live {
-						return errors.Errorf("expected node %d to be considered live after waiting 5 milliseconds", nodeID)
-					}
-					break
-				}
-
-				if err == storage.ErrEpochIncremented {
-					log.Warningf(context.Background(), "retrying after %s", err)
-					continue
-				}
-
-				return err
-			}
-			return nil
-		})
-	}
-}
-
-// *MockEngine implements engine.Engine interface.
-type MockEngine struct {
-	engine.InMem
-	mu  syncutil.Mutex
-	err error
-	ch  chan bool
-}
-
-func (m *MockEngine) NewBatch() engine.Batch {
-	return MockBatch{Batch: m.InMem.NewBatch(), engine: m}
-}
-
-// MockBatch implements engine.Batch interface.
-type MockBatch struct {
-	engine.Batch
-	engine *MockEngine
-}
-
-func (mb MockBatch) Commit(syncCommit bool) error {
-	mb.engine.mu.Lock()
-	ch := mb.engine.ch
-	err := mb.engine.err
-	mb.engine.mu.Unlock()
-
-	if ch != nil {
-		<-ch
-	}
-	if err != nil {
-		return err
-	}
-	err = mb.Batch.Commit(syncCommit)
-	return err
-}
