@@ -25,6 +25,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/util/causer"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
@@ -175,12 +176,12 @@ type queueImpl interface {
 	// and returns whether it should be queued and if so, at what priority.
 	// The Replica is guaranteed to be initialized.
 	shouldQueue(
-		context.Context, hlc.Timestamp, *Replica, config.SystemConfig,
+		context.Context, hlc.Timestamp, *Replica, *config.SystemConfig,
 	) (shouldQueue bool, priority float64)
 
 	// process accepts a replica, and the system config and executes
 	// queue-specific work on it. The Replica is guaranteed to be initialized.
-	process(context.Context, *Replica, config.SystemConfig) error
+	process(context.Context, *Replica, *config.SystemConfig) error
 
 	// timer returns a duration to wait between processing the next item
 	// from the queue. The duration of the last processing of a replica
@@ -414,11 +415,10 @@ func (bq *baseQueue) MaybeAdd(repl *Replica, now hlc.Timestamp) {
 
 func (bq *baseQueue) maybeAddLocked(ctx context.Context, repl *Replica, now hlc.Timestamp) {
 	// Load the system config if it's needed.
-	var cfg config.SystemConfig
-	var cfgOk bool
+	var cfg *config.SystemConfig
 	if bq.needsSystemConfig {
-		cfg, cfgOk = bq.gossip.GetSystemConfig()
-		if !cfgOk {
+		cfg = bq.gossip.GetSystemConfig()
+		if cfg == nil {
 			if log.V(1) {
 				log.Infof(ctx, "no system config available. skipping")
 			}
@@ -438,7 +438,7 @@ func (bq *baseQueue) maybeAddLocked(ctx context.Context, repl *Replica, now hlc.
 		repl.maybeInitializeRaftGroup(ctx)
 	}
 
-	if cfgOk && bq.requiresSplit(cfg, repl) {
+	if cfg != nil && bq.requiresSplit(cfg, repl) {
 		// Range needs to be split due to zone configs, but queue does
 		// not accept unsplit ranges.
 		if log.V(1) {
@@ -465,7 +465,7 @@ func (bq *baseQueue) maybeAddLocked(ctx context.Context, repl *Replica, now hlc.
 	}
 }
 
-func (bq *baseQueue) requiresSplit(cfg config.SystemConfig, repl *Replica) bool {
+func (bq *baseQueue) requiresSplit(cfg *config.SystemConfig, repl *Replica) bool {
 	if bq.acceptsUnsplitRanges {
 		return false
 	}
@@ -638,17 +638,17 @@ func (bq *baseQueue) processLoop(stopper *stop.Stopper) {
 					annotatedCtx := repl.AnnotateCtx(ctx)
 					if stopper.RunAsyncTask(
 						annotatedCtx, fmt.Sprintf("storage.%s: processing replica", bq.name),
-						func(annotatedCtx context.Context) {
+						func(ctx context.Context) {
 							// Release semaphore when finished processing.
 							defer func() { <-bq.processSem }()
 
 							start := timeutil.Now()
-							err := bq.processReplica(annotatedCtx, repl)
+							err := bq.processReplica(ctx, repl)
 
 							duration := timeutil.Since(start)
-							bq.recordProcessDuration(annotatedCtx, duration)
+							bq.recordProcessDuration(ctx, duration)
 
-							bq.finishProcessingReplica(annotatedCtx, stopper, repl, err)
+							bq.finishProcessingReplica(ctx, stopper, repl, err)
 						}) != nil {
 						// Release semaphore on task failure.
 						<-bq.processSem
@@ -693,86 +693,70 @@ func (bq *baseQueue) recordProcessDuration(ctx context.Context, dur time.Duratio
 // processReplica processes a single replica. This should not be
 // called externally to the queue. bq.mu.Lock must not be held
 // while calling this method.
-func (bq *baseQueue) processReplica(queueCtx context.Context, repl *Replica) error {
+//
+// ctx should already be annotated by repl.AnnotateCtx().
+func (bq *baseQueue) processReplica(ctx context.Context, repl *Replica) error {
 	// Load the system config if it's needed.
-	var cfg config.SystemConfig
-	var cfgOk bool
+	var cfg *config.SystemConfig
 	if bq.needsSystemConfig {
-		cfg, cfgOk = bq.gossip.GetSystemConfig()
-		if !cfgOk {
-			if log.V(1) {
-				log.Infof(queueCtx, "no system config available. skipping")
-			}
+		cfg = bq.gossip.GetSystemConfig()
+		if cfg == nil {
+			log.VEventf(ctx, 1, "no system config available. skipping")
 			return nil
 		}
 	}
 
-	if cfgOk && bq.requiresSplit(cfg, repl) {
+	if cfg != nil && bq.requiresSplit(cfg, repl) {
 		// Range needs to be split due to zone configs, but queue does
 		// not accept unsplit ranges.
-		if log.V(3) {
-			log.Infof(queueCtx, "split needed; skipping")
-		}
+		log.VEventf(ctx, 3, "split needed; skipping")
 		return nil
 	}
 
-	// Putting a span in a context means that events will no longer go to the
-	// event log. Use queueCtx for events that are intended for the event log.
-	ctx, span := bq.AnnotateCtxWithSpan(queueCtx, bq.name)
+	ctx, span := bq.AnnotateCtxWithSpan(ctx, bq.name)
 	defer span.Finish()
-	// Also add the Replica annotations to ctx.
-	ctx = repl.AnnotateCtx(ctx)
-	ctx, cancel := context.WithTimeout(ctx, bq.processTimeout)
-	defer cancel()
-	if log.V(1) {
-		log.Infof(ctx, "processing replica")
-	}
 
-	if !repl.IsInitialized() {
-		// We checked this when adding the replica, but we need to check it again
-		// in case this is a different replica with the same range ID (see #14193).
-		return errors.New("cannot process uninitialized replica")
-	}
+	return contextutil.RunWithTimeout(ctx, fmt.Sprintf("%s queue process replica %d", bq.name, repl.RangeID),
+		bq.processTimeout, func(ctx context.Context) error {
+			log.VEventf(ctx, 1, "processing replica")
 
-	if reason, err := repl.IsDestroyed(); err != nil {
-		if !bq.queueConfig.processDestroyedReplicas || reason == destroyReasonRemoved {
-			if log.V(3) {
-				log.Infof(queueCtx, "replica destroyed (%s); skipping", err)
+			if !repl.IsInitialized() {
+				// We checked this when adding the replica, but we need to check it again
+				// in case this is a different replica with the same range ID (see #14193).
+				return errors.New("cannot process uninitialized replica")
 			}
-			return nil
-		}
-	}
 
-	// If the queue requires a replica to have the range lease in
-	// order to be processed, check whether this replica has range lease
-	// and renew or acquire if necessary.
-	if bq.needsLease {
-		if _, pErr := repl.redirectOnOrAcquireLease(ctx); pErr != nil {
-			switch v := pErr.GetDetail().(type) {
-			case *roachpb.NotLeaseHolderError, *roachpb.RangeNotFoundError:
-				if log.V(3) {
-					log.Infof(queueCtx, "%s; skipping", v)
+			if reason, err := repl.IsDestroyed(); err != nil {
+				if !bq.queueConfig.processDestroyedReplicas || reason == destroyReasonRemoved {
+					log.VEventf(ctx, 3, "replica destroyed (%s); skipping", err)
+					return nil
 				}
-				log.Eventf(ctx, "%s; skipping", v)
-				return nil
-			default:
-				log.VErrEventf(ctx, 2, "could not obtain lease: %s", pErr)
-				return errors.Wrapf(pErr.GoError(), "%s: could not obtain lease", repl)
 			}
-		}
-	}
 
-	if log.V(3) {
-		log.Infof(queueCtx, "processing")
-	}
-	if err := bq.impl.process(ctx, repl, cfg); err != nil {
-		return err
-	}
-	if log.V(3) {
-		log.Infof(ctx, "done")
-	}
-	bq.successes.Inc(1)
-	return nil
+			// If the queue requires a replica to have the range lease in
+			// order to be processed, check whether this replica has range lease
+			// and renew or acquire if necessary.
+			if bq.needsLease {
+				if _, pErr := repl.redirectOnOrAcquireLease(ctx); pErr != nil {
+					switch v := pErr.GetDetail().(type) {
+					case *roachpb.NotLeaseHolderError, *roachpb.RangeNotFoundError:
+						log.VEventf(ctx, 3, "%s; skipping", v)
+						return nil
+					default:
+						log.VErrEventf(ctx, 2, "could not obtain lease: %s", pErr)
+						return errors.Wrapf(pErr.GoError(), "%s: could not obtain lease", repl)
+					}
+				}
+			}
+
+			log.VEventf(ctx, 3, "processing...")
+			if err := bq.impl.process(ctx, repl, cfg); err != nil {
+				return err
+			}
+			log.VEventf(ctx, 3, "processing... done")
+			bq.successes.Inc(1)
+			return nil
+		})
 }
 
 type benignError struct {
@@ -916,9 +900,9 @@ func (bq *baseQueue) addToPurgatoryLocked(
 						annotatedCtx := repl.AnnotateCtx(ctx)
 						if stopper.RunTask(
 							annotatedCtx, fmt.Sprintf("storage.%s: purgatory processing replica", bq.name),
-							func(annotatedCtx context.Context) {
-								err := bq.processReplica(annotatedCtx, repl)
-								bq.finishProcessingReplica(annotatedCtx, stopper, repl, err)
+							func(ctx context.Context) {
+								err := bq.processReplica(ctx, repl)
+								bq.finishProcessingReplica(ctx, stopper, repl, err)
 							}) != nil {
 							return
 						}
