@@ -66,7 +66,13 @@ type raftLogQueue struct {
 	logSnapshots util.EveryN
 }
 
-// newRaftLogQueue returns a new instance of raftLogQueue.
+// newRaftLogQueue returns a new instance of raftLogQueue. Replicas are passed
+// to the queue both proactively (triggered by write load) and periodically
+// (via the scanner). When processing a replica, the queue decides whether the
+// Raft log can be truncated, which is a tradeoff between wanting to keep the
+// log short overall and allowing slower followers to catch up before they get
+// cut off by a truncation and need a snapshot. See newTruncateDecision for
+// details on this decision making process.
 func newRaftLogQueue(store *Store, db *client.DB, gossip *gossip.Gossip) *raftLogQueue {
 	rlq := &raftLogQueue{
 		db:           db,
@@ -92,6 +98,59 @@ func newRaftLogQueue(store *Store, db *client.DB, gossip *gossip.Gossip) *raftLo
 // newTruncateDecision returns a truncateDecision for the given Replica if no
 // error occurs. If input data to establish a truncateDecision is missing, a
 // zero decision is returned.
+//
+// At a high level, a truncate decision operates based on the Raft log size, the
+// number of entries in the log, and the Raft status of the followers. In an
+// ideal world and most of the time, followers are reasonably up to date, and a
+// decision to truncate to the index acked on all replicas will be made whenever
+// there is at least a little bit of log to truncate (think a hundred records or
+// ~100kb of data). If followers fall behind, are offline, or are waiting for a
+// snapshot, a second strategy is needed to make sure that the Raft log is
+// eventually truncated: when the raft log size exceeds a limit (4mb at time of
+// writing), truncations become willing and able to cut off followers as long as
+// a quorum has acked the truncation index. The quota pool ensures that the delta
+// between "acked by quorum" and "acked by all" is bounded, while Raft limits the
+// size of the uncommitted, i.e. not "acked by quorum", part of the log; thus
+// the "quorum" truncation strategy bounds the absolute size of the log on all
+// followers.
+//
+// Exceptions are made for replicas for which information is missing ("probing
+// state") as long as they are known to have been online recently, and for
+// in-flight snapshots (in particular preemptive snapshots) which are not
+// adequately reflected in the Raft status and would otherwise be cut off with
+// regularity. Probing live followers should only remain in this state for a
+// short moment and so we deny a log truncation outright (as there's no safe
+// index to truncate to); for snapshots, we can still truncate, but not past
+// the snapshot's index.
+//
+// A challenge for log truncation is to deal with sideloaded log entries, that
+// is, entries which contain SSTables for direct ingestion into the storage
+// engine. Such log entries are very large, and failing to account for them in
+// the heuristics can trigger overly aggressive truncations.
+//
+// The raft log size used in the decision making process is principally updated
+// in the main Raft command apply loop, and adds a Replica to this queue
+// whenever the log size has increased by a non-negligible amount that would be
+// worth truncating (~100kb).
+//
+// Unfortunately, the size tracking is not very robust as it suffers from two
+// limitations at the time of writing:
+// 1. it may undercount as it is in-memory and incremented only as proposals
+//    are handled; that is, a freshly started node will believe its Raft log to be
+//    zero-sized independent of its actual size, and
+// 2. the addition and corresponding subtraction happen in very different places
+//    and are difficult to keep bug-free, meaning that there is low confidence that
+//    we maintain the delta in a completely accurate manner over time. One example
+//    of potential errors are sideloaded proposals, for which the subtraction needs
+//    to load the size of the file on-disk (i.e. supplied by the fs), whereas
+//    the addition uses the in-memory representation of the file.
+//
+// Ideally, a Raft log that grows large for whichever reason (for instance the
+// queue being stuck on another replica) wouldn't be more than a nuisance on
+// nodes with sufficient disk space. Unfortunately, at the time of writing, the
+// Raft log is included in Raft snapshots. On the other hand, IMPORT/RESTORE's
+// split/scatter phase interacts poorly with overly aggressive truncations and
+// can DDOS the Raft snapshot queue.
 func newTruncateDecision(ctx context.Context, r *Replica) (truncateDecision, error) {
 	rangeID := r.RangeID
 	now := timeutil.Now()
@@ -112,14 +171,15 @@ func newTruncateDecision(ctx context.Context, r *Replica) (truncateDecision, err
 	// efficient to catch up via a snapshot than via applying a long tail of log
 	// entries.
 	targetSize := r.store.cfg.RaftLogTruncationThreshold
-	if targetSize > r.mu.maxBytes {
-		targetSize = r.mu.maxBytes
+	if targetSize > *r.mu.zone.RangeMaxBytes {
+		targetSize = *r.mu.zone.RangeMaxBytes
 	}
 	raftStatus := r.raftStatusRLocked()
 
 	firstIndex, err := r.raftFirstIndexLocked()
 	pendingSnapshotIndex := r.getAndGCSnapshotLogTruncationConstraintsLocked(now)
 	lastIndex := r.mu.lastIndex
+	logSizeTrusted := r.mu.raftLogSizeTrusted
 	r.mu.Unlock()
 
 	if err != nil {
@@ -161,6 +221,7 @@ func newTruncateDecision(ctx context.Context, r *Replica) (truncateDecision, err
 		RaftStatus:                     *raftStatus,
 		LogSize:                        raftLogSize,
 		MaxLogSize:                     targetSize,
+		LogSizeTrusted:                 logSizeTrusted,
 		FirstIndex:                     firstIndex,
 		LastIndex:                      lastIndex,
 		PendingPreemptiveSnapshotIndex: pendingSnapshotIndex,
@@ -208,6 +269,7 @@ const (
 type truncateDecisionInput struct {
 	RaftStatus                     raft.Status
 	LogSize, MaxLogSize            int64
+	LogSizeTrusted                 bool // false when LogSize might be off
 	FirstIndex, LastIndex          uint64
 	PendingPreemptiveSnapshotIndex uint64
 }
@@ -276,6 +338,9 @@ func (td *truncateDecision) String() string {
 	if n := td.NumNewRaftSnapshots(); n > 0 {
 		_, _ = fmt.Fprintf(&buf, "; implies %d Raft snapshot%s", n, util.Pluralize(int64(n)))
 	}
+	if !td.Input.LogSizeTrusted {
+		_, _ = fmt.Fprintf(&buf, "; log size untrusted")
+	}
 	buf.WriteRune(']')
 
 	return buf.String()
@@ -283,12 +348,7 @@ func (td *truncateDecision) String() string {
 
 func (td *truncateDecision) NumTruncatableIndexes() int {
 	if td.NewFirstIndex < td.Input.FirstIndex {
-		log.Fatalf(
-			context.Background(),
-			"invalid truncate decision: first index would move from %d to %d",
-			td.Input.FirstIndex,
-			td.NewFirstIndex,
-		)
+		return 0
 	}
 	return int(td.NewFirstIndex - td.Input.FirstIndex)
 }
@@ -409,7 +469,11 @@ func computeTruncateDecision(input truncateDecisionInput) truncateDecision {
 func getQuorumIndex(raftStatus *raft.Status) uint64 {
 	match := make([]uint64, 0, len(raftStatus.Progress))
 	for _, progress := range raftStatus.Progress {
-		match = append(match, progress.Match)
+		if progress.State == raft.ProgressStateReplicate {
+			match = append(match, progress.Match)
+		} else {
+			match = append(match, 0)
+		}
 	}
 	sort.Sort(uint64Slice(match))
 	quorum := computeQuorum(len(match))
@@ -420,25 +484,83 @@ func getQuorumIndex(raftStatus *raft.Status) uint64 {
 // is true only if the replica is the raft leader and if the total number of
 // the range's raft log's stale entries exceeds RaftLogQueueStaleThreshold.
 func (rlq *raftLogQueue) shouldQueue(
-	ctx context.Context, now hlc.Timestamp, r *Replica, _ config.SystemConfig,
+	ctx context.Context, now hlc.Timestamp, r *Replica, _ *config.SystemConfig,
 ) (shouldQ bool, priority float64) {
 	decision, err := newTruncateDecision(ctx, r)
 	if err != nil {
 		log.Warning(ctx, err)
 		return false, 0
 	}
-	return decision.ShouldTruncate(), float64(decision.Input.LogSize)
+
+	shouldQ, _, prio := rlq.shouldQueueImpl(ctx, decision)
+	return shouldQ, prio
+}
+
+// shouldQueueImpl returns whether the given truncate decision should lead to
+// a log truncation. This is either the case if the decision says so or of
+// we want to recompute the log size (in which case `recomputeRaftLogSize` and
+// `shouldQ` are both true and a reasonable priority is returned).
+func (rlq *raftLogQueue) shouldQueueImpl(
+	ctx context.Context, decision truncateDecision,
+) (shouldQ bool, recomputeRaftLogSize bool, priority float64) {
+	if decision.ShouldTruncate() {
+		return true, !decision.Input.LogSizeTrusted, float64(decision.Input.LogSize)
+	}
+	if decision.Input.LogSizeTrusted ||
+		decision.Input.LastIndex == decision.Input.FirstIndex {
+
+		return false, false, 0
+	}
+	// We have a nonempty log (first index != last index) and can't vouch that
+	// the bytes in the log are known. Queue the replica; processing it will
+	// force a recomputation. For the priority, we have to pick one as we
+	// usually use the log size which is not available here. Going half-way
+	// between zero and the MaxLogSize should give a good tradeoff between
+	// processing the recomputation quickly, and not starving replicas which see
+	// a significant amount of write traffic until they run over and truncate
+	// more aggressively than they need to.
+	return true, true, 1.0 + float64(decision.Input.MaxLogSize)/2.0
 }
 
 // process truncates the raft log of the range if the replica is the raft
 // leader and if the total number of the range's raft log's stale entries
 // exceeds RaftLogQueueStaleThreshold.
-func (rlq *raftLogQueue) process(ctx context.Context, r *Replica, _ config.SystemConfig) error {
+func (rlq *raftLogQueue) process(ctx context.Context, r *Replica, _ *config.SystemConfig) error {
 	decision, err := newTruncateDecision(ctx, r)
 	if err != nil {
 		return err
 	}
 
+	if _, recompute, _ := rlq.shouldQueueImpl(ctx, decision); recompute {
+		log.VEventf(ctx, 2, "recomputing raft log based on decision %+v", decision)
+
+		// We need to hold raftMu both to access the sideloaded storage and to
+		// make sure concurrent Raft activity doesn't foul up our update to the
+		// cached in-memory values.
+		r.raftMu.Lock()
+		n, err := ComputeRaftLogSize(ctx, r.RangeID, r.Engine(), r.raftMu.sideloaded)
+		if err == nil {
+			r.mu.Lock()
+			r.mu.raftLogSize = n
+			r.mu.raftLogLastCheckSize = n
+			r.mu.raftLogSizeTrusted = true
+			r.mu.Unlock()
+		}
+		r.raftMu.Unlock()
+
+		if err != nil {
+			return errors.Wrap(err, "recomputing raft log size")
+		}
+
+		log.VEventf(ctx, 2, "recomputed raft log size to %s", humanizeutil.IBytes(n))
+
+		// Override the decision, now that an accurate log size is available.
+		decision, err = newTruncateDecision(ctx, r)
+		if err != nil {
+			return err
+		}
+	}
+
 	// Can and should the raft logs be truncated?
 	if decision.ShouldTruncate() {
 		if n := decision.NumNewRaftSnapshots(); log.V(1) || n > 0 && rlq.logSnapshots.ShouldProcess(timeutil.Now()) {
