@@ -15,6 +15,7 @@
 package storage
 
 import (
+	"bytes"
 	"context"
 	"fmt"
 	"math"
@@ -26,8 +27,8 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
+	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
@@ -91,17 +92,29 @@ func TestGetQuorumIndex(t *testing.T) {
 			Progress: make(map[uint64]raft.Progress),
 		}
 		for j, v := range c.progress {
-			status.Progress[uint64(j)] = raft.Progress{Match: v}
+			status.Progress[uint64(j)] = raft.Progress{State: raft.ProgressStateReplicate, Match: v}
 		}
 		quorumMatchedIndex := getQuorumIndex(status)
 		if c.expected != quorumMatchedIndex {
 			t.Fatalf("%d: expected %d, but got %d", i, c.expected, quorumMatchedIndex)
 		}
 	}
+
+	// Verify that only replicating followers are taken into account (i.e. others
+	// are treated as Match == 0).
+	status := &raft.Status{
+		Progress: map[uint64]raft.Progress{
+			1: {State: raft.ProgressStateReplicate, Match: 100},
+			2: {State: raft.ProgressStateSnapshot, Match: 100},
+			3: {State: raft.ProgressStateReplicate, Match: 90},
+		},
+	}
+	assert.Equal(t, uint64(90), getQuorumIndex(status))
 }
 
 func TestComputeTruncateDecision(t *testing.T) {
 	defer leaktest.AfterTest(t)()
+	ctx := context.Background()
 
 	const targetSize = 1000
 
@@ -200,23 +213,48 @@ func TestComputeTruncateDecision(t *testing.T) {
 			"should truncate: false [truncate 0 entries to first index 2 (chosen via: first index)]",
 		}}
 	for i, c := range testCases {
-		status := raft.Status{
-			Progress: make(map[uint64]raft.Progress),
-		}
-		for j, v := range c.progress {
-			status.Progress[uint64(j)] = raft.Progress{RecentActive: true, State: raft.ProgressStateReplicate, Match: v, Next: v + 1}
-		}
-		decision := computeTruncateDecision(truncateDecisionInput{
-			RaftStatus:                     status,
-			LogSize:                        c.raftLogSize,
-			MaxLogSize:                     targetSize,
-			FirstIndex:                     c.firstIndex,
-			LastIndex:                      c.lastIndex,
-			PendingPreemptiveSnapshotIndex: c.pendingSnapshot,
+		t.Run(fmt.Sprintf("%+v", c), func(t *testing.T) {
+			status := raft.Status{
+				Progress: make(map[uint64]raft.Progress),
+			}
+			for j, v := range c.progress {
+				status.Progress[uint64(j)] = raft.Progress{RecentActive: true, State: raft.ProgressStateReplicate, Match: v, Next: v + 1}
+			}
+			input := truncateDecisionInput{
+				RaftStatus:                     status,
+				LogSize:                        c.raftLogSize,
+				MaxLogSize:                     targetSize,
+				LogSizeTrusted:                 true,
+				FirstIndex:                     c.firstIndex,
+				LastIndex:                      c.lastIndex,
+				PendingPreemptiveSnapshotIndex: c.pendingSnapshot,
+			}
+			decision := computeTruncateDecision(input)
+			if act, exp := decision.String(), c.exp; act != exp {
+				t.Errorf("%d: got:\n%s\nwanted:\n%s", i, act, exp)
+			}
+
+			// Verify the triggers that queue a range for recomputation. In
+			// essence, when the raft log size is not trusted we want to suggest
+			// a truncation and also a recomputation. If the size *is* trusted,
+			// we'll just see the decision play out as before.
+			// The real tests for this are in TestRaftLogQueueShouldQueue, but this is
+			// some nice extra coverage.
+			should, recompute, prio := (*raftLogQueue)(nil).shouldQueueImpl(ctx, decision)
+			assert.Equal(t, decision.ShouldTruncate(), should)
+			assert.False(t, recompute)
+			assert.Equal(t, decision.ShouldTruncate(), prio != 0)
+			input.LogSizeTrusted = false
+			input.RaftStatus.RaftState = raft.StateLeader
+			if input.LastIndex <= input.FirstIndex {
+				input.LastIndex = input.FirstIndex + 1
+			}
+			decision = computeTruncateDecision(input)
+			should, recompute, prio = (*raftLogQueue)(nil).shouldQueueImpl(ctx, decision)
+			assert.True(t, should)
+			assert.True(t, prio > 0)
+			assert.True(t, recompute)
 		})
-		if act, exp := decision.String(), c.exp; act != exp {
-			t.Errorf("%d: got:\n%s\nwanted:\n%s", i, act, exp)
-		}
 	}
 }
 
@@ -269,10 +307,11 @@ func TestComputeTruncateDecisionProgressStatusProbe(t *testing.T) {
 			}
 
 			input := truncateDecisionInput{
-				RaftStatus: status,
-				MaxLogSize: 1024,
-				FirstIndex: 10,
-				LastIndex:  500,
+				RaftStatus:     status,
+				MaxLogSize:     1024,
+				FirstIndex:     10,
+				LastIndex:      500,
+				LogSizeTrusted: true,
 			}
 			if tooLarge {
 				input.LogSize += 2 * input.MaxLogSize
@@ -293,7 +332,7 @@ func TestTruncateDecisionZeroValue(t *testing.T) {
 	assert.False(t, decision.ShouldTruncate())
 	assert.Zero(t, decision.NumNewRaftSnapshots())
 	assert.Zero(t, decision.NumTruncatableIndexes())
-	assert.Equal(t, "should truncate: false [truncate 0 entries to first index 0 (chosen via: )]", decision.String())
+	assert.Equal(t, "should truncate: false [truncate 0 entries to first index 0 (chosen via: ); log size untrusted]", decision.String())
 }
 
 func TestTruncateDecisionNumSnapshots(t *testing.T) {
@@ -320,22 +359,16 @@ func TestTruncateDecisionNumSnapshots(t *testing.T) {
 }
 
 func verifyLogSizeInSync(t *testing.T, r *Replica) {
+	t.Helper()
 	r.raftMu.Lock()
 	defer r.raftMu.Unlock()
 	r.mu.Lock()
 	raftLogSize := r.mu.raftLogSize
 	r.mu.Unlock()
-	start := engine.MakeMVCCMetadataKey(keys.RaftLogKey(r.RangeID, 1))
-	end := engine.MakeMVCCMetadataKey(keys.RaftLogKey(r.RangeID, math.MaxUint64))
-
-	var ms enginepb.MVCCStats
-	iter := r.store.engine.NewIterator(engine.IterOptions{UpperBound: end.Key})
-	defer iter.Close()
-	ms, err := iter.ComputeStats(start, end, 0 /* nowNanos */)
+	actualRaftLogSize, err := ComputeRaftLogSize(context.Background(), r.RangeID, r.Engine(), r.SideloadedRaftMuLocked())
 	if err != nil {
 		t.Fatal(err)
 	}
-	actualRaftLogSize := ms.SysBytes
 	if actualRaftLogSize != raftLogSize {
 		t.Fatalf("replica claims raft log size %d, but computed %d", raftLogSize, actualRaftLogSize)
 	}
@@ -402,13 +435,47 @@ func TestUpdateRaftStatusActivity(t *testing.T) {
 	}
 }
 
+func TestNewTruncateDecisionMaxSize(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	stopper := stop.NewStopper()
+	defer stopper.Stop(context.TODO())
+
+	cfg := TestStoreConfig(hlc.NewClock(hlc.NewManualClock(123).UnixNano, time.Nanosecond))
+	const exp = 1881
+	cfg.RaftLogTruncationThreshold = exp
+	store := createTestStoreWithConfig(
+		t, stopper, testStoreOpts{createSystemRanges: true}, &cfg,
+	)
+
+	repl, err := store.GetReplica(1)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	ctx := context.Background()
+	td, err := newTruncateDecision(ctx, repl)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	if td.Input.MaxLogSize != exp {
+		t.Fatalf("MaxLogSize %d is unexpected, wanted %d", td.Input.MaxLogSize, exp)
+	}
+}
+
 // TestNewTruncateDecision verifies that old raft log entries are correctly
 // removed.
 func TestNewTruncateDecision(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	stopper := stop.NewStopper()
 	defer stopper.Stop(context.TODO())
-	store, _ := createTestStore(t, stopper)
+	store, _ := createTestStore(t,
+		testStoreOpts{
+			// This test was written before test stores could start with more than one
+			// range and was not adapted.
+			createSystemRanges: false,
+		},
+		stopper)
 	store.SetRaftLogQueueActive(false)
 
 	r, err := store.GetReplica(1)
@@ -457,7 +524,7 @@ func TestNewTruncateDecision(t *testing.T) {
 
 	// Enable the raft log scanner and and force a truncation.
 	store.SetRaftLogQueueActive(true)
-	store.ForceRaftLogScanAndProcess()
+	store.MustForceRaftLogScanAndProcess()
 	store.SetRaftLogQueueActive(false)
 
 	// There can be a delay from when the truncation command is issued and the
@@ -487,7 +554,7 @@ func TestNewTruncateDecision(t *testing.T) {
 	// Again, enable the raft log scanner and and force a truncation. This time
 	// we expect no truncation to occur.
 	store.SetRaftLogQueueActive(true)
-	store.ForceRaftLogScanAndProcess()
+	store.MustForceRaftLogScanAndProcess()
 	store.SetRaftLogQueueActive(false)
 
 	// Unlike the last iteration, where we expect a truncation and can wait on
@@ -531,7 +598,13 @@ func TestProactiveRaftLogTruncate(t *testing.T) {
 		t.Run("", func(t *testing.T) {
 			stopper := stop.NewStopper()
 			defer stopper.Stop(ctx)
-			store, _ := createTestStore(t, stopper)
+			store, _ := createTestStore(t,
+				testStoreOpts{
+					// This test was written before test stores could start with more than one
+					// range and was not adapted.
+					createSystemRanges: false,
+				},
+				stopper)
 
 			// Note that turning off the replica scanner does not prevent the queues
 			// from processing entries (in this case specifically the raftLogQueue),
@@ -634,3 +707,228 @@ func TestSnapshotLogTruncationConstraints(t *testing.T) {
 
 	assert.Equal(t, r.mu.snapshotLogTruncationConstraints, map[uuid.UUID]snapTruncationInfo(nil))
 }
+
+// TestTruncateLog verifies that the TruncateLog command removes a
+// prefix of the raft logs (modifying FirstIndex() and making them
+// inaccessible via Entries()).
+func TestTruncateLog(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+	tc := testContext{}
+	stopper := stop.NewStopper()
+	defer stopper.Stop(context.TODO())
+	tc.Start(t, stopper)
+	tc.repl.store.SetRaftLogQueueActive(false)
+
+	// Populate the log with 10 entries. Save the LastIndex after each write.
+	var indexes []uint64
+	for i := 0; i < 10; i++ {
+		args := incrementArgs([]byte("a"), int64(i))
+
+		if _, pErr := tc.SendWrapped(&args); pErr != nil {
+			t.Fatal(pErr)
+		}
+		idx, err := tc.repl.GetLastIndex()
+		if err != nil {
+			t.Fatal(err)
+		}
+		indexes = append(indexes, idx)
+	}
+
+	rangeID := tc.repl.RangeID
+
+	// Discard the first half of the log.
+	truncateArgs := truncateLogArgs(indexes[5], rangeID)
+	if _, pErr := tc.SendWrappedWith(roachpb.Header{RangeID: 1}, &truncateArgs); pErr != nil {
+		t.Fatal(pErr)
+	}
+
+	// FirstIndex has changed.
+	firstIndex, err := tc.repl.GetFirstIndex()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if firstIndex != indexes[5] {
+		t.Errorf("expected firstIndex == %d, got %d", indexes[5], firstIndex)
+	}
+
+	// We can still get what remains of the log.
+	tc.repl.mu.Lock()
+	entries, err := tc.repl.raftEntriesLocked(indexes[5], indexes[9], math.MaxUint64)
+	tc.repl.mu.Unlock()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if len(entries) != int(indexes[9]-indexes[5]) {
+		t.Errorf("expected %d entries, got %d", indexes[9]-indexes[5], len(entries))
+	}
+
+	// But any range that includes the truncated entries returns an error.
+	tc.repl.mu.Lock()
+	_, err = tc.repl.raftEntriesLocked(indexes[4], indexes[9], math.MaxUint64)
+	tc.repl.mu.Unlock()
+	if err != raft.ErrCompacted {
+		t.Errorf("expected ErrCompacted, got %s", err)
+	}
+
+	// The term of the last truncated entry is still available.
+	tc.repl.mu.Lock()
+	term, err := tc.repl.raftTermRLocked(indexes[4])
+	tc.repl.mu.Unlock()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if term == 0 {
+		t.Errorf("invalid term 0 for truncated entry")
+	}
+
+	// The terms of older entries are gone.
+	tc.repl.mu.Lock()
+	_, err = tc.repl.raftTermRLocked(indexes[3])
+	tc.repl.mu.Unlock()
+	if err != raft.ErrCompacted {
+		t.Errorf("expected ErrCompacted, got %s", err)
+	}
+
+	// Truncating logs that have already been truncated should not return an
+	// error.
+	truncateArgs = truncateLogArgs(indexes[3], rangeID)
+	if _, pErr := tc.SendWrapped(&truncateArgs); pErr != nil {
+		t.Fatal(pErr)
+	}
+
+	// Truncating logs that have the wrong rangeID included should not return
+	// an error but should not truncate any logs.
+	truncateArgs = truncateLogArgs(indexes[9], rangeID+1)
+	if _, pErr := tc.SendWrapped(&truncateArgs); pErr != nil {
+		t.Fatal(pErr)
+	}
+
+	tc.repl.mu.Lock()
+	// The term of the last truncated entry is still available.
+	term, err = tc.repl.raftTermRLocked(indexes[4])
+	tc.repl.mu.Unlock()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if term == 0 {
+		t.Errorf("invalid term 0 for truncated entry")
+	}
+}
+
+func TestRaftLogQueueShouldQueueRecompute(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	ctx := context.Background()
+	var rlq *raftLogQueue
+
+	_ = ctx
+	_ = rlq
+
+	// NB: Cases for which decision.ShouldTruncate() is true are tested in
+	// TestComputeTruncateDecision, so here the decision itself is never
+	// positive.
+	var decision truncateDecision
+	decision.Input.LogSizeTrusted = true
+	decision.Input.LogSize = 12
+	decision.Input.MaxLogSize = 1000
+
+	verify := func(shouldQ bool, recompute bool, prio float64) {
+		t.Helper()
+		isQ, isR, isP := rlq.shouldQueueImpl(ctx, decision)
+		assert.Equal(t, shouldQ, isQ)
+		assert.Equal(t, recompute, isR)
+		assert.Equal(t, prio, isP)
+	}
+
+	verify(false, false, 0)
+
+	// Check all the boxes: unknown log size, leader, and non-empty log.
+	decision.Input.LogSize = 123
+	decision.Input.LogSizeTrusted = false
+	decision.Input.FirstIndex = 10
+	decision.Input.LastIndex = 20
+
+	verify(true, true, 1+float64(decision.Input.MaxLogSize)/2)
+
+	golden := decision
+
+	// Check all boxes except that log is empty.
+	decision = golden
+	decision.Input.LastIndex = decision.Input.FirstIndex
+	verify(false, false, 0)
+}
+
+// TestTruncateLogRecompute checks that if raftLogSize is not trusted, the raft
+// log queue picks up the replica, recomputes the log size, and considers a
+// truncation.
+func TestTruncateLogRecompute(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	ctx := context.Background()
+	dir, cleanup := testutils.TempDir(t)
+	defer cleanup()
+
+	cache := engine.NewRocksDBCache(1 << 20)
+	defer cache.Release()
+	eng, err := engine.NewRocksDB(engine.RocksDBConfig{Dir: dir}, cache)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer eng.Close()
+
+	tc := testContext{
+		engine: eng,
+	}
+	stopper := stop.NewStopper()
+	defer stopper.Stop(context.TODO())
+	tc.Start(t, stopper)
+	tc.repl.store.SetRaftLogQueueActive(false)
+
+	key := roachpb.Key("a")
+	repl := tc.store.LookupReplica(keys.MustAddr(key))
+
+	trusted := func() bool {
+		repl.mu.Lock()
+		defer repl.mu.Unlock()
+		return repl.mu.raftLogSizeTrusted
+	}
+
+	put := func() {
+		var v roachpb.Value
+		v.SetBytes(bytes.Repeat([]byte("x"), RaftLogQueueStaleSize*5))
+		put := roachpb.NewPut(key, v)
+		var ba roachpb.BatchRequest
+		ba.Add(put)
+		ba.RangeID = repl.RangeID
+
+		if _, pErr := tc.store.Send(ctx, ba); pErr != nil {
+			t.Fatal(pErr)
+		}
+	}
+
+	put()
+
+	decision, err := newTruncateDecision(ctx, repl)
+	assert.NoError(t, err)
+	assert.True(t, decision.ShouldTruncate())
+	// Should never trust initially, until recomputed at least once.
+	assert.False(t, trusted())
+
+	repl.mu.Lock()
+	repl.mu.raftLogSizeTrusted = false
+	repl.mu.raftLogSize += 12          // garbage
+	repl.mu.raftLogLastCheckSize += 12 // garbage
+	repl.mu.Unlock()
+
+	// Force a raft log queue run. The result should be a nonzero Raft log of
+	// size below the threshold (though we won't check that since it could have
+	// grown over threshold again; we compute instead that its size is correct).
+	tc.store.SetRaftLogQueueActive(true)
+	tc.store.MustForceRaftLogScanAndProcess()
+
+	for i := 0; i < 2; i++ {
+		verifyLogSizeInSync(t, repl)
+		assert.True(t, trusted())
+		put() // make sure we remain trusted and in sync
+	}
+}
