@@ -28,13 +28,13 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/rpc/nodedialer"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
-	"github.com/cockroachdb/cockroach/pkg/storage/batcheval"
-	"github.com/cockroachdb/cockroach/pkg/storage/batcheval/result"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/util/causer"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/ctxgroup"
+	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
@@ -43,105 +43,9 @@ import (
 	"go.etcd.io/etcd/raft/raftpb"
 )
 
-// evaluateCommand delegates to the eval method for the given
-// roachpb.Request. The returned Result may be partially valid
-// even if an error is returned. maxKeys is the number of scan results
-// remaining for this batch (MaxInt64 for no limit).
-func evaluateCommand(
-	ctx context.Context,
-	raftCmdID storagebase.CmdIDKey,
-	index int,
-	batch engine.ReadWriter,
-	rec batcheval.EvalContext,
-	ms *enginepb.MVCCStats,
-	h roachpb.Header,
-	maxKeys int64,
-	args roachpb.Request,
-	reply roachpb.Response,
-) (result.Result, *roachpb.Error) {
-
-	// If a unittest filter was installed, check for an injected error; otherwise, continue.
-	if filter := rec.EvalKnobs().TestingEvalFilter; filter != nil {
-		filterArgs := storagebase.FilterArgs{
-			Ctx:   ctx,
-			CmdID: raftCmdID,
-			Index: index,
-			Sid:   rec.StoreID(),
-			Req:   args,
-			Hdr:   h,
-		}
-		if pErr := filter(filterArgs); pErr != nil {
-			log.Infof(ctx, "test injecting error: %s", pErr)
-			return result.Result{}, pErr
-		}
-	}
-
-	var err error
-	var pd result.Result
-
-	if cmd, ok := batcheval.LookupCommand(args.Method()); ok {
-		cArgs := batcheval.CommandArgs{
-			EvalCtx: rec,
-			Header:  h,
-			// Some commands mutate their arguments, so give each invocation
-			// its own copy (shallow to mimic earlier versions of this code
-			// in which args were passed by value instead of pointer).
-			Args:    args.ShallowCopy(),
-			MaxKeys: maxKeys,
-			Stats:   ms,
-		}
-		pd, err = cmd.Eval(ctx, batch, cArgs, reply)
-	} else {
-		err = errors.Errorf("unrecognized command %s", args.Method())
-	}
-
-	if h.ReturnRangeInfo {
-		returnRangeInfo(reply, rec)
-	}
-
-	// TODO(peter): We'd like to assert that the hlc clock is always updated
-	// correctly, but various tests insert versioned data without going through
-	// the proper channels. See TestPushTxnUpgradeExistingTxn for an example.
-	//
-	// if header.Txn != nil && !header.Txn.Timestamp.Less(h.Timestamp) {
-	// 	if now := r.store.Clock().Now(); now.Less(header.Txn.Timestamp) {
-	// 		log.Fatalf(ctx, "hlc clock not updated: %s < %s", now, header.Txn.Timestamp)
-	// 	}
-	// }
-
-	if log.V(2) {
-		log.Infof(ctx, "evaluated %s command %+v: %+v, err=%v", args.Method(), args, reply, err)
-	}
-
-	// Create a roachpb.Error by initializing txn from the request/response header.
-	var pErr *roachpb.Error
-	if err != nil {
-		txn := reply.Header().Txn
-		if txn == nil {
-			txn = h.Txn
-		}
-		pErr = roachpb.NewErrorWithTxn(err, txn)
-	}
-
-	return pd, pErr
-}
-
-func returnRangeInfo(reply roachpb.Response, rec batcheval.EvalContext) {
-	header := reply.Header()
-	lease, _ := rec.GetLease()
-	desc := rec.Desc()
-	header.RangeInfos = []roachpb.RangeInfo{
-		{
-			Desc:  *desc,
-			Lease: lease,
-		},
-	}
-	reply.SetHeader(header)
-}
-
 // AdminSplit divides the range into into two ranges using args.SplitKey.
 func (r *Replica) AdminSplit(
-	ctx context.Context, args roachpb.AdminSplitRequest,
+	ctx context.Context, args roachpb.AdminSplitRequest, reason string,
 ) (reply roachpb.AdminSplitResponse, _ *roachpb.Error) {
 	if len(args.SplitKey) == 0 {
 		return roachpb.AdminSplitResponse{}, roachpb.NewErrorf("cannot split range with no key provided")
@@ -169,7 +73,7 @@ func (r *Replica) AdminSplit(
 			return roachpb.AdminSplitResponse{}, pErr
 		}
 
-		reply, lastErr = r.adminSplitWithDescriptor(ctx, args, r.Desc(), true /* delayable */)
+		reply, lastErr = r.adminSplitWithDescriptor(ctx, args, r.Desc(), true /* delayable */, reason)
 		// On seeing a ConditionFailedError or an AmbiguousResultError, retry
 		// the command with the updated descriptor.
 		if retry := causer.Visit(lastErr, func(err error) bool {
@@ -194,11 +98,11 @@ func maybeDescriptorChangedError(desc *roachpb.RangeDescriptor, err error) (stri
 		// Provide a better message in the common case that the range being changed
 		// was already changed by a concurrent transaction.
 		var actualDesc roachpb.RangeDescriptor
-		if err := detail.ActualValue.GetProto(&actualDesc); err == nil {
-			if desc.RangeID == actualDesc.RangeID && !desc.Equal(actualDesc) {
-				return fmt.Sprintf("descriptor changed: [expected] %s != [actual] %s",
-					desc, actualDesc), true
-			}
+		if !detail.ActualValue.IsPresent() {
+			return fmt.Sprintf("descriptor changed: expected %s != [actual] nil (range subsumed)", desc), true
+		} else if err := detail.ActualValue.GetProto(&actualDesc); err == nil &&
+			desc.RangeID == actualDesc.RangeID && !desc.Equal(actualDesc) {
+			return fmt.Sprintf("descriptor changed: [expected] %s != [actual] %s", desc, actualDesc), true
 		}
 	}
 	return "", false
@@ -261,6 +165,7 @@ func (r *Replica) adminSplitWithDescriptor(
 	args roachpb.AdminSplitRequest,
 	desc *roachpb.RangeDescriptor,
 	delayable bool,
+	reason string,
 ) (roachpb.AdminSplitResponse, error) {
 	var reply roachpb.AdminSplitResponse
 
@@ -322,7 +227,7 @@ func (r *Replica) adminSplitWithDescriptor(
 	}
 	log.Event(ctx, "found split key")
 
-	// Create right hand side range descriptor with the newly-allocated Range ID.
+	// Create right hand side range descriptor.
 	rightDesc, err := r.store.NewRangeDescriptor(ctx, splitKey, desc.EndKey, desc.Replicas)
 	if err != nil {
 		return reply, errors.Errorf("unable to allocate right hand side range descriptor: %s", err)
@@ -330,13 +235,7 @@ func (r *Replica) adminSplitWithDescriptor(
 
 	// Init updated version of existing range descriptor.
 	leftDesc := *desc
-	if r.ClusterSettings().Version.IsMinSupported(cluster.VersionRangeMerges) {
-		// To be safe, don't increment the generation counter unless all nodes are
-		// known to be aware of its existence. Since encoded range descriptors are
-		// used in CPut operations, it is theorized that non-nil generation counters
-		// in mixed-version clusters could cause subtle breakage.
-		leftDesc.IncrementGeneration()
-	}
+	leftDesc.IncrementGeneration()
 	leftDesc.EndKey = splitKey
 
 	var extra string
@@ -345,8 +244,8 @@ func (r *Replica) adminSplitWithDescriptor(
 	}
 	extra += splitSnapshotWarningStr(r.RangeID, r.RaftStatus())
 
-	log.Infof(ctx, "initiating a split of this range at key %s [r%d]%s",
-		splitKey, rightDesc.RangeID, extra)
+	log.Infof(ctx, "initiating a split of this range at key %s [r%d] (%s)%s",
+		splitKey.StringWithDirs(nil /* valDirs */, 50 /* maxLen */), rightDesc.RangeID, reason, extra)
 
 	if err := r.store.DB().Txn(ctx, func(ctx context.Context, txn *client.Txn) error {
 		log.Event(ctx, "split closure begins")
@@ -447,14 +346,10 @@ func (r *Replica) adminSplitWithDescriptor(
 // The supplied RangeDescriptor is used as a form of optimistic lock. See the
 // comment of "AdminSplit" for more information on this pattern.
 func (r *Replica) AdminMerge(
-	ctx context.Context, args roachpb.AdminMergeRequest,
+	ctx context.Context, args roachpb.AdminMergeRequest, reason string,
 ) (roachpb.AdminMergeResponse, *roachpb.Error) {
 	var reply roachpb.AdminMergeResponse
 
-	if err := r.ClusterSettings().Version.CheckVersion(cluster.VersionRangeMerges, "range merges"); err != nil {
-		return reply, roachpb.NewError(err)
-	}
-
 	origLeftDesc := r.Desc()
 	if origLeftDesc.EndKey.Equal(roachpb.RKeyMax) {
 		// Merging the final range doesn't make sense.
@@ -488,7 +383,7 @@ func (r *Replica) AdminMerge(
 
 		updatedLeftDesc.IncrementGeneration()
 		updatedLeftDesc.EndKey = rightDesc.EndKey
-		log.Infof(ctx, "initiating a merge of %s into this range", rightDesc)
+		log.Infof(ctx, "initiating a merge of %s into this range (%s)", rightDesc, reason)
 	}
 
 	runMergeTxn := func(txn *client.Txn) error {
@@ -632,7 +527,7 @@ func (r *Replica) AdminMerge(
 		if err != nil {
 			txn.CleanupOnError(ctx, err)
 		}
-		if _, canRetry := errors.Cause(err).(*roachpb.HandledRetryableTxnError); !canRetry {
+		if _, canRetry := errors.Cause(err).(*roachpb.TransactionRetryWithProtoRefreshError); !canRetry {
 			if err != nil {
 				return reply, roachpb.NewErrorf("merge of range into %d failed: %s",
 					origLeftDesc.RangeID, err)
@@ -645,27 +540,25 @@ func (r *Replica) AdminMerge(
 func waitForApplication(
 	ctx context.Context, dialer *nodedialer.Dialer, desc roachpb.RangeDescriptor, leaseIndex uint64,
 ) error {
-	var cancel context.CancelFunc
-	ctx, cancel = context.WithTimeout(ctx, 5*time.Second)
-	defer cancel()
-
-	g := ctxgroup.WithContext(ctx)
-	for _, repl := range desc.Replicas {
-		repl := repl // copy for goroutine
-		g.GoCtx(func(ctx context.Context) error {
-			conn, err := dialer.Dial(ctx, repl.NodeID)
-			if err != nil {
-				return errors.Wrapf(err, "could not dial n%d", repl.NodeID)
-			}
-			_, err = NewPerReplicaClient(conn).WaitForApplication(ctx, &WaitForApplicationRequest{
-				StoreRequestHeader: StoreRequestHeader{NodeID: repl.NodeID, StoreID: repl.StoreID},
-				RangeID:            desc.RangeID,
-				LeaseIndex:         leaseIndex,
+	return contextutil.RunWithTimeout(ctx, "wait for application", 5*time.Second, func(ctx context.Context) error {
+		g := ctxgroup.WithContext(ctx)
+		for _, repl := range desc.Replicas {
+			repl := repl // copy for goroutine
+			g.GoCtx(func(ctx context.Context) error {
+				conn, err := dialer.Dial(ctx, repl.NodeID)
+				if err != nil {
+					return errors.Wrapf(err, "could not dial n%d", repl.NodeID)
+				}
+				_, err = NewPerReplicaClient(conn).WaitForApplication(ctx, &WaitForApplicationRequest{
+					StoreRequestHeader: StoreRequestHeader{NodeID: repl.NodeID, StoreID: repl.StoreID},
+					RangeID:            desc.RangeID,
+					LeaseIndex:         leaseIndex,
+				})
+				return err
 			})
-			return err
-		})
-	}
-	return g.Wait()
+		}
+		return g.Wait()
+	})
 }
 
 // waitForReplicasInit blocks until it has proof that the replicas listed in
@@ -675,26 +568,24 @@ func waitForApplication(
 func waitForReplicasInit(
 	ctx context.Context, dialer *nodedialer.Dialer, desc roachpb.RangeDescriptor,
 ) error {
-	var cancel context.CancelFunc
-	ctx, cancel = context.WithTimeout(ctx, 5*time.Second)
-	defer cancel()
-
-	g := ctxgroup.WithContext(ctx)
-	for _, repl := range desc.Replicas {
-		repl := repl // copy for goroutine
-		g.GoCtx(func(ctx context.Context) error {
-			conn, err := dialer.Dial(ctx, repl.NodeID)
-			if err != nil {
-				return errors.Wrapf(err, "could not dial n%d", repl.NodeID)
-			}
-			_, err = NewPerReplicaClient(conn).WaitForReplicaInit(ctx, &WaitForReplicaInitRequest{
-				StoreRequestHeader: StoreRequestHeader{NodeID: repl.NodeID, StoreID: repl.StoreID},
-				RangeID:            desc.RangeID,
+	return contextutil.RunWithTimeout(ctx, "wait for replicas init", 5*time.Second, func(ctx context.Context) error {
+		g := ctxgroup.WithContext(ctx)
+		for _, repl := range desc.Replicas {
+			repl := repl // copy for goroutine
+			g.GoCtx(func(ctx context.Context) error {
+				conn, err := dialer.Dial(ctx, repl.NodeID)
+				if err != nil {
+					return errors.Wrapf(err, "could not dial n%d", repl.NodeID)
+				}
+				_, err = NewPerReplicaClient(conn).WaitForReplicaInit(ctx, &WaitForReplicaInitRequest{
+					StoreRequestHeader: StoreRequestHeader{NodeID: repl.NodeID, StoreID: repl.StoreID},
+					RangeID:            desc.RangeID,
+				})
+				return err
 			})
-			return err
-		})
-	}
-	return g.Wait()
+		}
+		return g.Wait()
+	})
 }
 
 type snapshotError struct {
@@ -721,6 +612,10 @@ func IsSnapshotError(err error) bool {
 //
 // The supplied RangeDescriptor is used as a form of optimistic lock. See the
 // comment of "adminSplitWithDescriptor" for more information on this pattern.
+// The returned RangeDescriptor is the new value of the range's descriptor
+// following the successful commit of the transaction. It can be used when
+// making a series of changes to detect and prevent races between concurrent
+// actors.
 //
 // Changing the replicas for a range is complicated. A change is initiated by
 // the "replicate" queue when it encounters a range which has too many
@@ -773,10 +668,10 @@ func IsSnapshotError(err error) bool {
 // Raft log entries. This is important for the replicate queue to be aware
 // of. A node can process hundreds or thousands of ChangeReplicas operations
 // per second even though the actual replication of data proceeds at a much
-// slower base. In order to avoid having this background replication overwhelm
-// the system, replication is throttled via a reservation system. When
-// allocating a new replica for a range, the replicate queue reserves space for
-// that replica on the target store via a ReservationRequest. (See
+// slower base. In order to avoid having this background replication and
+// overwhelming the system, replication is throttled via a reservation system.
+// When allocating a new replica for a range, the replicate queue reserves space
+// for that replica on the target store via a ReservationRequest. (See
 // StorePool.reserve). The reservation is fulfilled when the snapshot is
 // applied.
 //
@@ -794,21 +689,25 @@ func (r *Replica) ChangeReplicas(
 	changeType roachpb.ReplicaChangeType,
 	target roachpb.ReplicationTarget,
 	desc *roachpb.RangeDescriptor,
-	reason RangeLogEventReason,
+	reason storagepb.RangeLogEventReason,
 	details string,
-) error {
+) (updatedDesc *roachpb.RangeDescriptor, _ error) {
 	return r.changeReplicas(ctx, changeType, target, desc, SnapshotRequest_REBALANCE, reason, details)
 }
 
+// changeReplicas reutrn
 func (r *Replica) changeReplicas(
 	ctx context.Context,
 	changeType roachpb.ReplicaChangeType,
 	target roachpb.ReplicationTarget,
 	desc *roachpb.RangeDescriptor,
 	priority SnapshotRequest_Priority,
-	reason RangeLogEventReason,
+	reason storagepb.RangeLogEventReason,
 	details string,
-) error {
+) (_ *roachpb.RangeDescriptor, _ error) {
+	if desc == nil {
+		return nil, errors.Errorf("%s: the current RangeDescriptor must not be nil", r)
+	}
 	repDesc := roachpb.ReplicaDescriptor{
 		NodeID:  target.NodeID,
 		StoreID: target.StoreID,
@@ -836,9 +735,9 @@ func (r *Replica) changeReplicas(
 		// abort the replica add.
 		if nodeUsed {
 			if repDescIdx != -1 {
-				return errors.Errorf("%s: unable to add replica %v which is already present", r, repDesc)
+				return nil, errors.Errorf("%s: unable to add replica %v which is already present", r, repDesc)
 			}
-			return errors.Errorf("%s: unable to add replica %v; node already has a replica", r, repDesc)
+			return nil, errors.Errorf("%s: unable to add replica %v; node already has a replica", r, repDesc)
 		}
 
 		// Send a pre-emptive snapshot. Note that the replica to which this
@@ -864,7 +763,7 @@ func (r *Replica) changeReplicas(
 		// progress which might be required for this ChangeReplicas operation to
 		// complete. See #10409.
 		if err := r.sendSnapshot(ctx, repDesc, snapTypePreemptive, priority); err != nil {
-			return err
+			return nil, err
 		}
 
 		repDesc.ReplicaID = updatedDesc.NextReplicaID
@@ -875,7 +774,7 @@ func (r *Replica) changeReplicas(
 		// If that exact node-store combination does not have the replica,
 		// abort the removal.
 		if repDescIdx == -1 {
-			return errors.Errorf("%s: unable to remove replica %v which is not present", r, repDesc)
+			return nil, errors.Errorf("%s: unable to remove replica %v which is not present", r, repDesc)
 		}
 		updatedDesc.Replicas[repDescIdx] = updatedDesc.Replicas[len(updatedDesc.Replicas)-1]
 		updatedDesc.Replicas = updatedDesc.Replicas[:len(updatedDesc.Replicas)-1]
@@ -960,10 +859,10 @@ func (r *Replica) changeReplicas(
 		if msg, ok := maybeDescriptorChangedError(desc, err); ok {
 			err = &benignError{errors.New(msg)}
 		}
-		return errors.Wrapf(err, "change replicas of r%d failed", rangeID)
+		return nil, errors.Wrapf(err, "change replicas of r%d failed", rangeID)
 	}
 	log.Event(ctx, "txn complete")
-	return nil
+	return &updatedDesc, nil
 }
 
 // sendSnapshot sends a snapshot of the replica state to the specified
@@ -998,8 +897,43 @@ func (r *Replica) sendSnapshot(
 		return &benignError{errors.New("raft status not initialized")}
 	}
 
+	usesReplicatedTruncatedState, err := engine.MVCCGetProto(
+		ctx, snap.EngineSnap, keys.RaftTruncatedStateLegacyKey(r.RangeID), hlc.Timestamp{}, nil, engine.MVCCGetOptions{},
+	)
+	if err != nil {
+		return errors.Wrap(err, "loading legacy truncated state")
+	}
+
+	canAvoidSendingLog := !usesReplicatedTruncatedState &&
+		snap.State.TruncatedState.Index < snap.State.RaftAppliedIndex &&
+		r.store.ClusterSettings().Version.IsActive(cluster.VersionSnapshotsWithoutLog)
+
+	if canAvoidSendingLog {
+		// If we're not using a legacy (replicated) truncated state, we avoid
+		// sending the (past) Raft log in the snapshot in the first place and
+		// send only those entries that are actually useful to the follower.
+		// This is done by changing the truncated state, which we're allowed
+		// to do since it is not a replicated key (and thus not subject to
+		// matching across replicas). The actual sending happens here:
+		_ = (*kvBatchSnapshotStrategy)(nil).Send
+		// and results in no log entries being sent at all. Note that
+		// Metadata.Index is really the applied index of the replica.
+		snap.State.TruncatedState = &roachpb.RaftTruncatedState{
+			Index: snap.RaftSnap.Metadata.Index,
+			Term:  snap.RaftSnap.Metadata.Term,
+		}
+	}
+
 	req := SnapshotRequest_Header{
 		State: snap.State,
+		// Tell the recipient whether it needs to synthesize the new
+		// unreplicated TruncatedState. It could tell by itself by peeking into
+		// the data, but it uses a write only batch for performance which
+		// doesn't support that; this is easier. Notably, this is true if the
+		// snap index itself is the one at which the migration happens.
+		//
+		// See VersionUnreplicatedRaftTruncatedState.
+		UnreplicatedTruncatedState: !usesReplicatedTruncatedState,
 		RaftMessageRequest: RaftMessageRequest{
 			RangeID:     r.RangeID,
 			FromReplica: fromRepDesc,
@@ -1110,8 +1044,6 @@ func updateRangeDescriptor(
 //
 // This is best-effort; it's possible that the replicate queue on the
 // leaseholder could take action at the same time, causing errors.
-//
-// TODO: How does this being a method on the leaseholder replica affect things?
 func (s *Store) AdminRelocateRange(
 	ctx context.Context, rangeDesc roachpb.RangeDescriptor, targets []roachpb.ReplicationTarget,
 ) error {
@@ -1183,8 +1115,33 @@ func (s *Store) AdminRelocateRange(
 		}
 	}
 
-	sysCfg, ok := s.cfg.Gossip.GetSystemConfig()
-	if !ok {
+	// updateRangeDesc updates the passed RangeDescriptor following the successful
+	// completion of an AdminChangeReplicasRequest with the single provided target
+	// and changeType.
+	// TODO(ajwerner): Remove this for 19.2 after AdminChangeReplicas always
+	// returns a non-nil Desc.
+	updateRangeDesc := func(
+		desc *roachpb.RangeDescriptor,
+		changeType roachpb.ReplicaChangeType,
+		target roachpb.ReplicationTarget,
+	) {
+		switch changeType {
+		case roachpb.ADD_REPLICA:
+			desc.Replicas = append(desc.Replicas, roachpb.ReplicaDescriptor{
+				NodeID:    target.NodeID,
+				StoreID:   target.StoreID,
+				ReplicaID: desc.NextReplicaID,
+			})
+			desc.NextReplicaID++
+		case roachpb.REMOVE_REPLICA:
+			desc.Replicas = removeTargetFromSlice(desc.Replicas, target)
+		default:
+			panic(errors.Errorf("unknown ReplicaChangeType %v", changeType))
+		}
+	}
+
+	sysCfg := s.cfg.Gossip.GetSystemConfig()
+	if sysCfg == nil {
 		return fmt.Errorf("no system config available, unable to perform RelocateRange")
 	}
 	zone, err := sysCfg.GetZoneConfigForKey(rangeDesc.StartKey)
@@ -1248,9 +1205,9 @@ func (s *Store) AdminRelocateRange(
 				NodeID:  targetStore.Node.NodeID,
 				StoreID: targetStore.StoreID,
 			}
-			if err := s.DB().AdminChangeReplicas(
-				ctx, startKey, roachpb.ADD_REPLICA, []roachpb.ReplicationTarget{target},
-			); err != nil {
+			newDesc, err := s.DB().AdminChangeReplicas(ctx, startKey, roachpb.ADD_REPLICA,
+				[]roachpb.ReplicationTarget{target}, *rangeInfo.Desc)
+			if err != nil {
 				returnErr := errors.Wrapf(err, "while adding target %v", target)
 				if !canRetry(err) {
 					return returnErr
@@ -1266,10 +1223,12 @@ func (s *Store) AdminRelocateRange(
 			// local copy of the range descriptor such that future allocator
 			// decisions take it into account.
 			addTargets = removeTargetFromSlice(addTargets, target)
-			rangeInfo.Desc.Replicas = append(rangeInfo.Desc.Replicas, roachpb.ReplicaDescriptor{
-				NodeID:  target.NodeID,
-				StoreID: target.StoreID,
-			})
+			if newDesc != nil {
+				rangeInfo.Desc = newDesc
+			} else {
+				// TODO(ajwerner): Remove this case for 19.2.
+				updateRangeDesc(rangeInfo.Desc, roachpb.ADD_REPLICA, target)
+			}
 		}
 
 		if len(removeTargets) > 0 && len(removeTargets) > len(addTargets) {
@@ -1286,9 +1245,9 @@ func (s *Store) AdminRelocateRange(
 			// the lease first in such scenarios. The first specified target should be
 			// the leaseholder now, so we can always transfer the lease there.
 			transferLease()
-			if err := s.DB().AdminChangeReplicas(
-				ctx, startKey, roachpb.REMOVE_REPLICA, []roachpb.ReplicationTarget{target},
-			); err != nil {
+			newDesc, err := s.DB().AdminChangeReplicas(ctx, startKey, roachpb.REMOVE_REPLICA,
+				[]roachpb.ReplicationTarget{target}, *rangeInfo.Desc)
+			if err != nil {
 				log.Warningf(ctx, "while removing target %v: %s", target, err)
 				if !canRetry(err) {
 					return err
@@ -1301,7 +1260,12 @@ func (s *Store) AdminRelocateRange(
 			// copy of the range descriptor such that future allocator decisions take
 			// its absence into account.
 			removeTargets = removeTargetFromSlice(removeTargets, target)
-			rangeInfo.Desc.Replicas = removeTargetFromSlice(rangeInfo.Desc.Replicas, target)
+			if newDesc != nil {
+				rangeInfo.Desc = newDesc
+			} else {
+				// TODO(ajwerner): Remove this case for 19.2.
+				updateRangeDesc(rangeInfo.Desc, roachpb.REMOVE_REPLICA, target)
+			}
 		}
 	}
 
@@ -1331,8 +1295,8 @@ func removeTargetFromSlice(
 func (r *Replica) adminScatter(
 	ctx context.Context, args roachpb.AdminScatterRequest,
 ) (roachpb.AdminScatterResponse, error) {
-	sysCfg, ok := r.store.cfg.Gossip.GetSystemConfig()
-	if !ok {
+	sysCfg := r.store.cfg.Gossip.GetSystemConfig()
+	if sysCfg == nil {
 		log.Infof(ctx, "scatter failed (system config not yet available)")
 		return roachpb.AdminScatterResponse{}, errors.New("system config not yet available")
 	}
