@@ -20,13 +20,14 @@ import (
 	"crypto/sha512"
 	"encoding/binary"
 	"fmt"
+	"math"
 	"sort"
 	"sync"
+	"sync/atomic"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
+	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/storage/batcheval"
@@ -35,12 +36,14 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/storage/rditer"
 	"github.com/cockroachdb/cockroach/pkg/storage/stateloader"
 	"github.com/cockroachdb/cockroach/pkg/util/bufalloc"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
 )
 
 var testingFatalOnStatsMismatch = envutil.EnvOrDefaultBool("COCKROACH_FATAL_ON_STATS_MISMATCH", false)
@@ -51,14 +54,30 @@ const (
 	// because the checksum might never be computed for a replica if that replica
 	// is caught up via a snapshot and never performs the ComputeChecksum
 	// operation.
-	collectChecksumTimeout = 5 * time.Second
+	collectChecksumTimeout = 15 * time.Second
 )
 
+// ReplicaChecksum contains progress on a replica checksum computation.
+type ReplicaChecksum struct {
+	CollectChecksumResponse
+	// started is true if the checksum computation has started.
+	started bool
+	// If gcTimestamp is nonzero, GC this checksum after gcTimestamp. gcTimestamp
+	// is zero if and only if the checksum computation is in progress.
+	gcTimestamp time.Time
+	// This channel is closed after the checksum is computed, and is used
+	// as a notification.
+	notify chan struct{}
+}
+
 // CheckConsistency runs a consistency check on the range. It first applies a
 // ComputeChecksum through Raft and then issues CollectChecksum commands to the
-// other replicas. When an inconsistency is detected and no diff was requested,
-// the consistency check will be re-run to collect a diff, which is then printed
-// before calling `log.Fatal`.
+// other replicas. These are inspected and a CheckConsistencyResponse is assembled.
+//
+// When args.Mode is CHECK_VIA_QUEUE and an inconsistency is detected and no
+// diff was requested, the consistency check will be re-run to collect a diff,
+// which is then printed before calling `log.Fatal`. This behavior should be
+// lifted to the consistency checker queue in the future.
 func (r *Replica) CheckConsistency(
 	ctx context.Context, args roachpb.CheckConsistencyRequest,
 ) (roachpb.CheckConsistencyResponse, *roachpb.Error) {
@@ -68,33 +87,86 @@ func (r *Replica) CheckConsistency(
 		RequestHeader: roachpb.RequestHeader{Key: startKey},
 		Version:       batcheval.ReplicaChecksumVersion,
 		Snapshot:      args.WithDiff,
+		Mode:          args.Mode,
+		Checkpoint:    args.Checkpoint,
 	}
 
+	isQueue := args.Mode == roachpb.ChecksumMode_CHECK_VIA_QUEUE
+
 	results, err := r.RunConsistencyCheck(ctx, checkArgs)
 	if err != nil {
 		return roachpb.CheckConsistencyResponse{}, roachpb.NewError(err)
 	}
 
 	var inconsistencyCount int
+	var missingCount int
 
-	for _, result := range results {
+	res := roachpb.CheckConsistencyResponse_Result{}
+	res.RangeID = r.RangeID
+	for i, result := range results {
 		expResponse := results[0].Response
-		if result.Err != nil || bytes.Equal(expResponse.Checksum, result.Response.Checksum) {
+		if result.Err != nil {
+			res.Detail += fmt.Sprintf("%s: %v\n", result.Replica, result.Err)
+			missingCount++
+			continue
+		}
+		if bytes.Equal(expResponse.Checksum, result.Response.Checksum) {
+			// Replica is consistent (or rather, agrees with the local result).
+			if i == 0 {
+				res.Detail += fmt.Sprintf("stats: %+v\n", expResponse.Persisted)
+			}
 			continue
 		}
 		inconsistencyCount++
 		var buf bytes.Buffer
-		_, _ = fmt.Fprintf(&buf, "replica %s is inconsistent: expected checksum %x, got %x",
-			result.Replica, expResponse.Checksum, result.Response.Checksum)
+		_, _ = fmt.Fprintf(&buf, "replica %s is inconsistent: expected checksum %x, got %x\n"+
+			"persisted stats: exp %+v, got %+v\n",
+			result.Replica,
+			expResponse.Checksum, result.Response.Checksum,
+			expResponse.Persisted, result.Response.Persisted,
+		)
 		if expResponse.Snapshot != nil && result.Response.Snapshot != nil {
 			diff := diffRange(expResponse.Snapshot, result.Response.Snapshot)
-			if report := r.store.cfg.TestingKnobs.BadChecksumReportDiff; report != nil {
+			if report := r.store.cfg.TestingKnobs.ConsistencyTestingKnobs.BadChecksumReportDiff; report != nil {
 				report(*r.store.Ident, diff)
 			}
-			buf.WriteByte('\n')
 			_, _ = diff.WriteTo(&buf)
 		}
-		log.Error(ctx, buf.String())
+		if isQueue {
+			log.Error(ctx, buf.String())
+		}
+		res.Detail += buf.String()
+	}
+
+	delta := enginepb.MVCCStats(results[0].Response.Delta)
+	delta.LastUpdateNanos = 0
+
+	res.StartKey = []byte(startKey)
+	res.Status = roachpb.CheckConsistencyResponse_RANGE_CONSISTENT
+	if inconsistencyCount != 0 {
+		res.Status = roachpb.CheckConsistencyResponse_RANGE_INCONSISTENT
+	} else if args.Mode != roachpb.ChecksumMode_CHECK_STATS && delta != (enginepb.MVCCStats{}) {
+		if delta.ContainsEstimates {
+			// When ContainsEstimates is set, it's generally expected that we'll get a different
+			// result when we recompute from scratch.
+			res.Status = roachpb.CheckConsistencyResponse_RANGE_CONSISTENT_STATS_ESTIMATED
+		} else {
+			// When ContainsEstimates is set, it's generally expected that we'll get a different
+			// result when we recompute from scratch.
+			res.Status = roachpb.CheckConsistencyResponse_RANGE_CONSISTENT_STATS_INCORRECT
+		}
+		res.Detail += fmt.Sprintf("stats delta: %+v\n", enginepb.MVCCStats(results[0].Response.Delta))
+	} else if missingCount > 0 {
+		res.Status = roachpb.CheckConsistencyResponse_RANGE_INDETERMINATE
+	}
+	var resp roachpb.CheckConsistencyResponse
+	resp.Result = append(resp.Result, res)
+
+	// Bail out at this point except if the queue is the caller. All of the stuff
+	// below should really happen in the consistency queue to keep CheckConsistency
+	// itself self-contained.
+	if !isQueue {
+		return resp, nil
 	}
 
 	if inconsistencyCount == 0 {
@@ -103,13 +175,12 @@ func (r *Replica) CheckConsistency(
 		// were many bugs in our stats computations. These are being fixed, but it
 		// is through this mechanism that existing ranges are updated. Hence, the
 		// logging below is relatively timid.
-		delta := enginepb.MVCCStats(results[0].Response.Delta)
-		delta.LastUpdateNanos = 0
+
 		// If there's no delta (or some nodes in the cluster may not know
 		// RecomputeStats, in which case sending it to them could crash them),
 		// there's nothing else to do.
-		if delta == (enginepb.MVCCStats{}) || !r.ClusterSettings().Version.IsMinSupported(cluster.VersionRecomputeStats) {
-			return roachpb.CheckConsistencyResponse{}, nil
+		if delta == (enginepb.MVCCStats{}) || !r.ClusterSettings().Version.IsActive(cluster.VersionRecomputeStats) {
+			return resp, nil
 		}
 
 		if !delta.ContainsEstimates && testingFatalOnStatsMismatch {
@@ -134,11 +205,11 @@ func (r *Replica) CheckConsistency(
 		b.AddRawRequest(&req)
 
 		err := r.store.db.Run(ctx, &b)
-		return roachpb.CheckConsistencyResponse{}, roachpb.NewError(err)
+		return resp, roachpb.NewError(err)
 	}
 
 	logFunc := log.Fatalf
-	if p := r.store.TestingKnobs().BadChecksumPanic; p != nil {
+	if p := r.store.cfg.TestingKnobs.ConsistencyTestingKnobs.BadChecksumPanic; p != nil {
 		if !args.WithDiff {
 			// We'll call this recursively with WithDiff==true; let's let that call
 			// be the one to trigger the handler.
@@ -150,7 +221,7 @@ func (r *Replica) CheckConsistency(
 	// Diff was printed above, so call logFunc with a short message only.
 	if args.WithDiff {
 		logFunc(ctx, "consistency check failed with %d inconsistent replicas", inconsistencyCount)
-		return roachpb.CheckConsistencyResponse{}, nil
+		return resp, nil
 	}
 
 	// No diff was printed, so we want to re-run with diff.
@@ -158,12 +229,25 @@ func (r *Replica) CheckConsistency(
 	log.Errorf(ctx, "consistency check failed with %d inconsistent replicas; fetching details",
 		inconsistencyCount)
 	args.WithDiff = true
-	if _, pErr := r.CheckConsistency(ctx, args); pErr != nil {
-		logFunc(ctx, "replica inconsistency detected; could not obtain actual diff: %s", pErr)
-	}
+	args.Checkpoint = true
 
+	// We've noticed in practice that if the diff is large, the log file in it
+	// is promptly rotated away. We already know we're going to fatal, and so we
+	// can afford disabling the log size limit altogether so that the diff can
+	// stick around. For reasons of cleanliness we try to reset things to normal
+	// in tests but morally speaking we're really just disabling it for good
+	// until the process crashes.
+	//
+	// See:
+	// https://github.com/cockroachdb/cockroach/issues/36861
+	oldLogLimit := atomic.LoadInt64(&log.LogFilesCombinedMaxSize)
+	atomic.CompareAndSwapInt64(&log.LogFilesCombinedMaxSize, oldLogLimit, math.MaxInt64)
+	if _, pErr := r.CheckConsistency(ctx, args); pErr != nil {
+		log.Fatalf(ctx, "replica inconsistency detected; could not obtain actual diff: %s", pErr)
+	}
 	// Not reached except in tests.
-	return roachpb.CheckConsistencyResponse{}, nil
+	atomic.CompareAndSwapInt64(&log.LogFilesCombinedMaxSize, math.MaxInt64, oldLogLimit)
+	return resp, nil
 }
 
 // A ConsistencyCheckResult contains the outcome of a CollectChecksum call.
@@ -183,10 +267,10 @@ func (r *Replica) collectChecksumFromReplica(
 	}
 	client := NewPerReplicaClient(conn)
 	req := &CollectChecksumRequest{
-		StoreRequestHeader{NodeID: replica.NodeID, StoreID: replica.StoreID},
-		r.RangeID,
-		id,
-		checksum,
+		StoreRequestHeader: StoreRequestHeader{NodeID: replica.NodeID, StoreID: replica.StoreID},
+		RangeID:            r.RangeID,
+		ChecksumID:         id,
+		Checksum:           checksum,
 	}
 	resp, err := client.CollectChecksum(ctx, req)
 	if err != nil {
@@ -238,14 +322,17 @@ func (r *Replica) RunConsistencyCheck(
 			func(ctx context.Context) {
 				defer wg.Done()
 
-				ctx, cancel := context.WithTimeout(ctx, collectChecksumTimeout)
-				defer cancel()
-
-				var masterChecksum []byte
-				if len(results) > 0 {
-					masterChecksum = results[0].Response.Checksum
-				}
-				resp, err := r.collectChecksumFromReplica(ctx, replica, ccRes.ChecksumID, masterChecksum)
+				var resp CollectChecksumResponse
+				err := contextutil.RunWithTimeout(ctx, "collect checksum", collectChecksumTimeout,
+					func(ctx context.Context) error {
+						var masterChecksum []byte
+						if len(results) > 0 {
+							masterChecksum = results[0].Response.Checksum
+						}
+						var err error
+						resp, err = r.collectChecksumFromReplica(ctx, replica, ccRes.ChecksumID, masterChecksum)
+						return err
+					})
 				resultCh <- ConsistencyCheckResult{
 					Replica:  replica,
 					Response: resp,
@@ -338,6 +425,7 @@ func (r *Replica) computeChecksumDone(
 			delta := result.PersistedMS
 			delta.Subtract(result.RecomputedMS)
 			c.Delta = enginepb.MVCCStatsDelta(delta)
+			c.Persisted = result.PersistedMS
 		}
 		c.gcTimestamp = timeutil.Now().Add(batcheval.ReplicaChecksumGCInterval)
 		c.Snapshot = snapshot
@@ -364,7 +452,10 @@ func (r *Replica) sha512(
 	desc roachpb.RangeDescriptor,
 	snap engine.Reader,
 	snapshot *roachpb.RaftSnapshotData,
+	mode roachpb.ChecksumMode,
 ) (*replicaHash, error) {
+	statsOnly := mode == roachpb.ChecksumMode_CHECK_STATS
+
 	// Iterate over all the data in the range.
 	iter := snap.NewIterator(engine.IterOptions{UpperBound: desc.EndKey.AsRawKey()})
 	defer iter.Close()
@@ -415,25 +506,59 @@ func (r *Replica) sha512(
 	}
 
 	var ms enginepb.MVCCStats
-	for _, span := range rditer.MakeReplicatedKeyRanges(&desc) {
-		spanMS, err := engine.ComputeStatsGo(
-			iter, span.Start, span.End, 0 /* nowNanos */, visitor,
-		)
-		if err != nil {
-			return nil, err
+	// In statsOnly mode, we hash only the RangeAppliedState. In regular mode, hash
+	// all of the replicated key space.
+	if !statsOnly {
+		for _, span := range rditer.MakeReplicatedKeyRanges(&desc) {
+			spanMS, err := engine.ComputeStatsGo(
+				iter, span.Start, span.End, 0 /* nowNanos */, visitor,
+			)
+			if err != nil {
+				return nil, err
+			}
+			ms.Add(spanMS)
 		}
-		ms.Add(spanMS)
 	}
 
 	var result replicaHash
 	result.RecomputedMS = ms
-	hasher.Sum(result.SHA512[:0])
 
-	curMS, err := stateloader.Make(r.store.cfg.Settings, desc.RangeID).LoadMVCCStats(ctx, snap)
+	rangeAppliedState, err := stateloader.Make(desc.RangeID).LoadRangeAppliedState(ctx, snap)
 	if err != nil {
 		return nil, err
 	}
-	result.PersistedMS = curMS
+	if rangeAppliedState == nil {
+		// This error is transient: the range applied state is used in v2.1 already
+		// but is migrated into on a per-range basis for clusters bootstrapped before
+		// v2.1. Clusters bootstrapped at v2.1 or higher will never hit this path since
+		// there's always an applied state.
+		return nil, errors.New("no range applied state found")
+	}
+	result.PersistedMS = rangeAppliedState.RangeStats.ToStats()
+
+	if statsOnly {
+		b, err := protoutil.Marshal(rangeAppliedState)
+		if err != nil {
+			return nil, err
+		}
+		if snapshot != nil {
+			// Add LeaseAppliedState to the diff.
+			kv := roachpb.RaftSnapshotData_KeyValue{
+				Timestamp: hlc.Timestamp{},
+			}
+			kv.Key = keys.RangeAppliedStateKey(desc.RangeID)
+			var v roachpb.Value
+			if err := v.SetProto(rangeAppliedState); err != nil {
+				return nil, err
+			}
+			kv.Value = v.RawBytes
+		}
+		if _, err := hasher.Write(b); err != nil {
+			return nil, err
+		}
+	}
+
+	hasher.Sum(result.SHA512[:0])
 
 	// We're not required to do so, but it looks nicer if both stats are aged to
 	// the same timestamp.
