@@ -23,24 +23,26 @@ import (
 	"time"
 	"unsafe"
 
-	"github.com/kr/pretty"
-	opentracing "github.com/opentracing/opentracing-go"
-	"github.com/pkg/errors"
-	"golang.org/x/time/rate"
-
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/storage/batcheval/result"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
-	"github.com/cockroachdb/cockroach/pkg/storage/rditer"
+	"github.com/cockroachdb/cockroach/pkg/storage/spanset"
+	"github.com/cockroachdb/cockroach/pkg/storage/stateloader"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
+	"github.com/cockroachdb/cockroach/pkg/util/sysutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
+	"github.com/kr/pretty"
+	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/pkg/errors"
+	"golang.org/x/time/rate"
 )
 
 // ProposalData is data about a command which allows it to be
@@ -67,10 +69,10 @@ type ProposalData struct {
 
 	// command is serialized and proposed to raft. In the event of
 	// reproposals its MaxLeaseIndex field is mutated.
-	command *storagebase.RaftCommand
+	command *storagepb.RaftCommand
 
-	// endCmds.finish is called after command execution to update the timestamp cache &
-	// command queue.
+	// endCmds.finish is called after command execution to update the
+	// timestamp cache & release latches.
 	endCmds *endCmds
 
 	// doneCh is used to signal the waiting RPC handler (the contents of
@@ -109,7 +111,7 @@ type ProposalData struct {
 // counted on to invoke endCmds itself.)
 func (proposal *ProposalData) finishApplication(pr proposalResult) {
 	if proposal.endCmds != nil {
-		proposal.endCmds.done(pr.Reply, pr.Err, pr.ProposalRetry)
+		proposal.endCmds.done(pr.Reply, pr.Err)
 		proposal.endCmds = nil
 	}
 	if proposal.sp != nil {
@@ -121,7 +123,7 @@ func (proposal *ProposalData) finishApplication(pr proposalResult) {
 // returnProposalResult signals proposal.doneCh with the proposal result if it
 // has not already been signaled. The method can be called even before the
 // proposal has finished replication and command application, and does not
-// release the command from the command queue.
+// release the request's latches.
 func (proposal *ProposalData) signalProposalResult(pr proposalResult) {
 	if proposal.doneCh != nil {
 		proposal.doneCh <- pr
@@ -142,7 +144,7 @@ func (r *Replica) gcOldChecksumEntriesLocked(now time.Time) {
 	}
 }
 
-func (r *Replica) computeChecksumPostApply(ctx context.Context, cc storagebase.ComputeChecksum) {
+func (r *Replica) computeChecksumPostApply(ctx context.Context, cc storagepb.ComputeChecksum) {
 	stopper := r.store.Stopper()
 	now := timeutil.Now()
 	r.mu.Lock()
@@ -167,6 +169,22 @@ func (r *Replica) computeChecksumPostApply(ctx context.Context, cc storagebase.C
 	// Caller is holding raftMu, so an engine snapshot is automatically
 	// Raft-consistent (i.e. not in the middle of an AddSSTable).
 	snap := r.store.engine.NewSnapshot()
+	if cc.Checkpoint {
+		checkpointBase := filepath.Join(r.store.engine.GetAuxiliaryDir(), "checkpoints")
+		_ = os.MkdirAll(checkpointBase, 0700)
+		sl := stateloader.Make(r.RangeID)
+		rai, _, err := sl.LoadAppliedIndex(ctx, snap)
+		if err != nil {
+			log.Warningf(ctx, "unable to load applied index, continuing anyway")
+		}
+		// NB: the names here will match on all nodes, which is nice for debugging.
+		checkpointDir := filepath.Join(checkpointBase, fmt.Sprintf("r%d_at_%d", r.RangeID, rai))
+		if err := r.store.engine.CreateCheckpoint(checkpointDir); err != nil {
+			log.Warningf(ctx, "unable to create checkpoint %s: %s", checkpointDir, err)
+		} else {
+			log.Infof(ctx, "created checkpoint %s", checkpointDir)
+		}
+	}
 
 	// Compute SHA asynchronously and store it in a map by UUID.
 	if err := stopper.RunAsyncTask(ctx, "storage.Replica: computing checksum", func(ctx context.Context) {
@@ -175,7 +193,7 @@ func (r *Replica) computeChecksumPostApply(ctx context.Context, cc storagebase.C
 		if cc.SaveSnapshot {
 			snapshot = &roachpb.RaftSnapshotData{}
 		}
-		result, err := r.sha512(ctx, desc, snap, snapshot)
+		result, err := r.sha512(ctx, desc, snap, snapshot, cc.Mode)
 		if err != nil {
 			log.Errorf(ctx, "%v", err)
 			result = nil
@@ -189,11 +207,22 @@ func (r *Replica) computeChecksumPostApply(ctx context.Context, cc storagebase.C
 	}
 }
 
-// leasePostApply is called when a RequestLease or TransferLease
-// request is executed for a range.
-func (r *Replica) leasePostApply(ctx context.Context, newLease roachpb.Lease) {
+// leasePostApply updates the Replica's internal state to reflect the
+// application of a new Range lease. The method is idempotent, so it can be
+// called repeatedly for the same lease safely. However, the method will panic
+// if passed a lease with a lower sequence number than the current lease. By
+// default, the method will also panic if passed a lease that indicates a
+// forward sequence number jump (i.e. a skipped lease). This behavior can
+// be disabled by passing permitJump as true.
+func (r *Replica) leasePostApply(ctx context.Context, newLease roachpb.Lease, permitJump bool) {
 	r.mu.Lock()
 	replicaID := r.mu.replicaID
+	// Pull out the last lease known to this Replica. It's possible that this is
+	// not actually the last lease in the Range's lease sequence because the
+	// Replica may have missed the application of a lease between prevLease and
+	// newLease. However, this should only be possible if a snapshot includes a
+	// lease update. All other forms of lease updates should be continuous
+	// without jumps (see permitJump).
 	prevLease := *r.mu.state.Lease
 	r.mu.Unlock()
 
@@ -207,8 +236,8 @@ func (r *Replica) leasePostApply(ctx context.Context, newLease roachpb.Lease) {
 	if iAmTheLeaseHolder {
 		// Log lease acquisition whenever an Epoch-based lease changes hands (or verbose
 		// logging is enabled).
-		if (newLease.Type() == roachpb.LeaseEpoch && leaseChangingHands) || log.V(1) {
-			log.Infof(ctx, "new range lease %s following %s", newLease, prevLease)
+		if newLease.Type() == roachpb.LeaseEpoch && leaseChangingHands || log.V(1) {
+			log.VEventf(ctx, 1, "new range lease %s following %s", newLease, prevLease)
 		}
 	}
 
@@ -235,10 +264,7 @@ func (r *Replica) leasePostApply(ctx context.Context, newLease roachpb.Lease) {
 		// requests, this is kosher). This means that we don't use the old
 		// lease's expiration but instead use the new lease's start to initialize
 		// the timestamp cache low water.
-		desc := r.Desc()
-		for _, keyRange := range rditer.MakeReplicatedKeyRanges(desc) {
-			r.store.tsCache.SetLowWater(keyRange.Start.Key, keyRange.End.Key, newLease.Start)
-		}
+		setTimestampCacheLowWaterMark(r.store.tsCache, r.Desc(), newLease.Start)
 
 		// Reset the request counts used to make lease placement decisions whenever
 		// starting a new lease.
@@ -266,9 +292,7 @@ func (r *Replica) leasePostApply(ctx context.Context, newLease roachpb.Lease) {
 			}
 		case s2 == s1+1:
 			// Lease sequence incremented by 1. Expected case.
-		case s2 > s1+1:
-			// Snapshots will never call leasePostApply, so we always expect
-			// leases to increment one at a time here.
+		case s2 > s1+1 && !permitJump:
 			log.Fatalf(ctx, "lease sequence jump, prevLease=%s, newLease=%s",
 				log.Safe(prevLease), log.Safe(newLease))
 		}
@@ -361,9 +385,9 @@ func addSSTablePreApply(
 	ctx context.Context,
 	st *cluster.Settings,
 	eng engine.Engine,
-	sideloaded sideloadStorage,
+	sideloaded SideloadStorage,
 	term, index uint64,
-	sst storagebase.ReplicatedEvalResult_AddSSTable,
+	sst storagepb.ReplicatedEvalResult_AddSSTable,
 	limiter *rate.Limiter,
 ) bool {
 	checksum := util.CRC32(sst.Data)
@@ -383,6 +407,14 @@ func addSSTablePreApply(
 		log.Fatalf(ctx, "sideloaded SSTable at term %d, index %d is missing", term, index)
 	}
 
+	eng.PreIngestDelay(ctx)
+
+	// as of VersionUnreplicatedRaftTruncatedState we were on rocksdb 5.17 so this
+	// cluster version should indicate that we will never use rocksdb < 5.16 to
+	// read these SSTs, so it is safe to use https://github.com/facebook/rocksdb/pull/4172
+	// to avoid needing the global seq_no edits and the copies they required.
+	canSkipSeqNo := st.Version.IsActive(cluster.VersionUnreplicatedRaftTruncatedState)
+
 	copied := false
 	if inmem, ok := eng.(engine.InMem); ok {
 		path = fmt.Sprintf("%x", checksum)
@@ -392,6 +424,7 @@ func addSSTablePreApply(
 	} else {
 		ingestPath := path + ".ingested"
 
+		canLinkToRaftFile := false
 		// The SST may already be on disk, thanks to the sideloading mechanism.  If
 		// so we can try to add that file directly, via a new hardlink if the file-
 		// system support it, rather than writing a new copy of it. However, this is
@@ -401,12 +434,26 @@ func addSSTablePreApply(
 		// tell Rocks that it is not allowed to modify the file, in which case it
 		// will return and error if it would have tried to do so, at which point we
 		// can fall back to writing a new copy for Rocks to ingest.
-		if _, err := os.Stat(path); err == nil {
+		if _, links, err := sysutil.StatAndLinkCount(path); err == nil {
+			// HACK: RocksDB does not like ingesting the same file (by inode) twice.
+			// See facebook/rocksdb#5133. We can tell that we have tried to ingest
+			// this file already if it has more than one link – one from the file raft
+			// wrote and one from rocks. In that case, we should not try to give
+			// rocks a link to the same file again.
+			if links == 1 {
+				canLinkToRaftFile = true
+			} else {
+				log.Warningf(ctx, "SSTable at index %d term %d may have already been ingested (link count %d) -- falling back to ingesting a copy",
+					index, term, links)
+			}
+		}
+
+		if canLinkToRaftFile {
 			// If the fs supports it, make a hard-link for rocks to ingest. We cannot
 			// pass it the path in the sideload store as it deletes the passed path on
 			// success.
 			if linkErr := eng.LinkFile(path, ingestPath); linkErr == nil {
-				ingestErr := eng.IngestExternalFiles(ctx, []string{ingestPath}, noModify)
+				ingestErr := eng.IngestExternalFiles(ctx, []string{ingestPath}, canSkipSeqNo, noModify)
 				if ingestErr == nil {
 					// Adding without modification succeeded, no copy necessary.
 					log.Eventf(ctx, "ingested SSTable at index %d, term %d: %s", index, term, ingestPath)
@@ -416,8 +463,18 @@ func addSSTablePreApply(
 					log.Fatalf(ctx, "failed to move ingest sst: %v", rmErr)
 				}
 				const seqNoMsg = "Global seqno is required, but disabled"
-				if err, ok := ingestErr.(*engine.RocksDBError); ok && !strings.Contains(ingestErr.Error(), seqNoMsg) {
-					log.Fatalf(ctx, "while ingesting %s: %s", ingestPath, err)
+				const seqNoOnReIngest = "external file have non zero sequence number"
+				// Repeated ingestion is still possible even with the link count checked
+				// above, since rocks might have already compacted away the file.
+				// However it does not flush compacted files from its cache, so it can
+				// still react poorly to attempting to ingest again. If we get an error
+				// that indicates we can't ingest, we'll make a copy and try again. That
+				// attempt must succeed or we'll fatal, so any persistent error is still
+				// going to be surfaced.
+				ingestErrMsg := ingestErr.Error()
+				isSeqNoErr := strings.Contains(ingestErrMsg, seqNoMsg) || strings.Contains(ingestErrMsg, seqNoOnReIngest)
+				if _, ok := ingestErr.(*engine.RocksDBError); !ok || !isSeqNoErr {
+					log.Fatalf(ctx, "while ingesting %s: %s", ingestPath, ingestErr)
 				}
 			}
 		}
@@ -447,7 +504,7 @@ func addSSTablePreApply(
 		copied = true
 	}
 
-	if err := eng.IngestExternalFiles(ctx, []string{path}, modify); err != nil {
+	if err := eng.IngestExternalFiles(ctx, []string{path}, canSkipSeqNo, modify); err != nil {
 		log.Fatalf(ctx, "while ingesting %s: %s", path, err)
 	}
 	log.Eventf(ctx, "ingested SSTable at index %d, term %d: %s", index, term, path)
@@ -456,7 +513,7 @@ func addSSTablePreApply(
 
 func (r *Replica) handleReplicatedEvalResult(
 	ctx context.Context,
-	rResult storagebase.ReplicatedEvalResult,
+	rResult storagepb.ReplicatedEvalResult,
 	raftAppliedIndex, leaseAppliedIndex uint64,
 ) (shouldAssert bool) {
 	// Fields for which no action is taken in this method are zeroed so that
@@ -465,8 +522,6 @@ func (r *Replica) handleReplicatedEvalResult(
 	{
 		rResult.IsLeaseRequest = false
 		rResult.Timestamp = hlc.Timestamp{}
-		rResult.DeprecatedStartKey = nil
-		rResult.DeprecatedEndKey = nil
 		rResult.PrevLeaseProposal = nil
 	}
 
@@ -493,10 +548,10 @@ func (r *Replica) handleReplicatedEvalResult(
 	r.store.metrics.addMVCCStats(deltaStats)
 	rResult.Delta = enginepb.MVCCStatsDelta{}
 
-	if r.store.splitQueue != nil && needsSplitBySize { // the boostrap store has a nil split queue
+	if r.store.splitQueue != nil && needsSplitBySize { // the bootstrap store has a nil split queue
 		r.store.splitQueue.MaybeAdd(r, r.store.Clock().Now())
 	}
-	if r.store.mergeQueue != nil && needsMergeBySize { // the boostrap store has a nil merge queue
+	if r.store.mergeQueue != nil && needsMergeBySize { // the bootstrap store has a nil merge queue
 		r.store.mergeQueue.MaybeAdd(r, r.store.Clock().Now())
 	}
 
@@ -516,14 +571,14 @@ func (r *Replica) handleReplicatedEvalResult(
 
 			// Clear any entries in the Raft log entry cache for this range up
 			// to and including the most recently truncated index.
-			r.store.raftEntryCache.clearTo(r.RangeID, newTruncState.Index+1)
+			r.store.raftEntryCache.Clear(r.RangeID, newTruncState.Index+1)
 
 			// Truncate the sideloaded storage. Note that this is safe only if the new truncated state
 			// is durably on disk (i.e.) synced. This is true at the time of writing but unfortunately
 			// could rot.
 			{
 				log.Eventf(ctx, "truncating sideloaded storage up to (and including) index %d", newTruncState.Index)
-				if size, err := r.raftMu.sideloaded.TruncateTo(ctx, newTruncState.Index+1); err != nil {
+				if size, _, err := r.raftMu.sideloaded.TruncateTo(ctx, newTruncState.Index+1); err != nil {
 					// We don't *have* to remove these entries for correctness. Log a
 					// loud error, but keep humming along.
 					log.Errorf(ctx, "while removing sideloaded files during log truncation: %s", err)
@@ -555,7 +610,7 @@ func (r *Replica) handleReplicatedEvalResult(
 			r.mu.Unlock()
 		}
 
-		if (*rResult.State == storagebase.ReplicaState{}) {
+		if (*rResult.State == storagepb.ReplicaState{}) {
 			rResult.State = nil
 		}
 	}
@@ -601,7 +656,7 @@ func (r *Replica) handleReplicatedEvalResult(
 	// The rest of the actions are "nontrivial" and may have large effects on the
 	// in-memory and on-disk ReplicaStates. If any of these actions are present,
 	// we want to assert that these two states do not diverge.
-	shouldAssert = !rResult.Equal(storagebase.ReplicatedEvalResult{})
+	shouldAssert = !rResult.Equal(storagepb.ReplicatedEvalResult{})
 
 	// Process Split or Merge. This needs to happen after stats update because
 	// of the ContainsEstimates hack.
@@ -635,7 +690,7 @@ func (r *Replica) handleReplicatedEvalResult(
 		}
 
 		if newLease := rResult.State.Lease; newLease != nil {
-			r.leasePostApply(ctx, *newLease)
+			r.leasePostApply(ctx, *newLease, false /* permitJump */)
 			rResult.State.Lease = nil
 		}
 
@@ -664,7 +719,7 @@ func (r *Replica) handleReplicatedEvalResult(
 			rResult.State.UsingAppliedStateKey = false
 		}
 
-		if (*rResult.State == storagebase.ReplicaState{}) {
+		if (*rResult.State == storagepb.ReplicaState{}) {
 			rResult.State = nil
 		}
 	}
@@ -676,10 +731,14 @@ func (r *Replica) handleReplicatedEvalResult(
 			// that the other nodes have finished this command as well (since
 			// processing the removal from the queue looks up the Range at the
 			// lease holder, being too early here turns this into a no-op).
-			if _, err := r.store.replicaGCQueue.Add(r, replicaGCPriorityRemoved); err != nil {
-				// Log the error; the range should still be GC'd eventually.
-				log.Errorf(ctx, "unable to add to replica GC queue: %s", err)
-			}
+			// Lock ordering dictates that we don't hold any mutexes when adding,
+			// so we fire it off in a task.
+			_ = r.store.stopper.RunAsyncTask(ctx, "add-to-replicag-queue", func(ctx context.Context) {
+				if _, err := r.store.replicaGCQueue.Add(r, replicaGCPriorityRemoved); err != nil {
+					// Log the error; the range should still be GC'd eventually.
+					log.Errorf(ctx, "unable to add to replica GC queue: %s", err)
+				}
+			})
 		}
 		rResult.ChangeReplicas = nil
 	}
@@ -689,8 +748,8 @@ func (r *Replica) handleReplicatedEvalResult(
 		rResult.ComputeChecksum = nil
 	}
 
-	if !rResult.Equal(storagebase.ReplicatedEvalResult{}) {
-		log.Fatalf(ctx, "unhandled field in ReplicatedEvalResult: %s", pretty.Diff(rResult, storagebase.ReplicatedEvalResult{}))
+	if !rResult.Equal(storagepb.ReplicatedEvalResult{}) {
+		log.Fatalf(ctx, "unhandled field in ReplicatedEvalResult: %s", pretty.Diff(rResult, storagepb.ReplicatedEvalResult{}))
 	}
 	return shouldAssert
 }
@@ -777,7 +836,7 @@ func (r *Replica) handleLocalEvalResult(ctx context.Context, lResult result.Loca
 func (r *Replica) handleEvalResultRaftMuLocked(
 	ctx context.Context,
 	lResult *result.LocalResult,
-	rResult storagebase.ReplicatedEvalResult,
+	rResult storagepb.ReplicatedEvalResult,
 	raftAppliedIndex, leaseAppliedIndex uint64,
 ) {
 	shouldAssert := r.handleReplicatedEvalResult(ctx, rResult, raftAppliedIndex, leaseAppliedIndex)
@@ -792,3 +851,157 @@ func (r *Replica) handleEvalResultRaftMuLocked(
 		r.mu.Unlock()
 	}
 }
+
+// proposalResult indicates the result of a proposal. Exactly one of
+// Reply and Err is set, and it represents the result of the proposal.
+type proposalResult struct {
+	Reply   *roachpb.BatchResponse
+	Err     *roachpb.Error
+	Intents []result.IntentsWithArg
+	EndTxns []result.EndTxnIntents
+}
+
+// evaluateProposal generates a Result from the given request by
+// evaluating it, returning both state which is held only on the
+// proposer and that which is to be replicated through Raft. The
+// return value is ready to be inserted into Replica's proposal map
+// and subsequently passed to submitProposalLocked.
+//
+// The method also returns a flag indicating if the request needs to
+// be proposed through Raft and replicated. This flag will be false
+// either if the request was a no-op or if it hit an error. In this
+// case, the result can be sent directly back to the client without
+// going through Raft, but while still handling LocalEvalResult.
+//
+// Replica.mu must not be held.
+func (r *Replica) evaluateProposal(
+	ctx context.Context, idKey storagebase.CmdIDKey, ba roachpb.BatchRequest, spans *spanset.SpanSet,
+) (*result.Result, bool, *roachpb.Error) {
+	if ba.Timestamp == (hlc.Timestamp{}) {
+		return nil, false, roachpb.NewErrorf("can't propose Raft command with zero timestamp")
+	}
+
+	// Evaluate the commands. If this returns without an error, the batch should
+	// be committed. Note that we don't hold any locks at this point. This is
+	// important since evaluating a proposal is expensive.
+	// TODO(tschottdorf): absorb all returned values in `res` below this point
+	// in the call stack as well.
+	batch, ms, br, res, pErr := r.evaluateWriteBatch(ctx, idKey, ba, spans)
+
+	// Note: reusing the proposer's batch when applying the command on the
+	// proposer was explored as an optimization but resulted in no performance
+	// benefit.
+	defer batch.Close()
+
+	if pErr != nil {
+		pErr = r.maybeSetCorrupt(ctx, pErr)
+
+		txn := pErr.GetTxn()
+		if txn != nil && ba.Txn == nil {
+			log.Fatalf(ctx, "error had a txn but batch is non-transactional. Err txn: %s", txn)
+		}
+
+		// Failed proposals can't have any Result except for what's
+		// whitelisted here.
+		intents := res.Local.DetachIntents()
+		endTxns := res.Local.DetachEndTxns(true /* alwaysOnly */)
+		res.Local = result.LocalResult{
+			Intents: &intents,
+			EndTxns: &endTxns,
+			Metrics: res.Local.Metrics,
+		}
+		res.Replicated.Reset()
+		return &res, false /* needConsensus */, pErr
+	}
+
+	// Set the local reply, which is held only on the proposing replica and is
+	// returned to the client after the proposal completes, or immediately if
+	// replication is not necessary.
+	res.Local.Reply = br
+
+	// needConsensus determines if the result needs to be replicated and
+	// proposed through Raft. This is necessary if at least one of the
+	// following conditions is true:
+	// 1. the request created a non-empty write batch.
+	// 2. the request had an impact on the MVCCStats. NB: this is possible
+	//    even with an empty write batch when stats are recomputed.
+	// 3. the request has replicated side-effects.
+	// 4. the cluster is in "clockless" mode, in which case consensus is
+	//    used to enforce a linearization of all reads and writes.
+	needConsensus := !batch.Empty() ||
+		ms != (enginepb.MVCCStats{}) ||
+		!res.Replicated.Equal(storagepb.ReplicatedEvalResult{}) ||
+		r.store.Clock().MaxOffset() == timeutil.ClocklessMaxOffset
+
+	if needConsensus {
+		// Set the proposal's WriteBatch, which is the serialized representation of
+		// the proposals effect on RocksDB.
+		res.WriteBatch = &storagepb.WriteBatch{
+			Data: batch.Repr(),
+		}
+
+		// Set the proposal's replicated result, which contains metadata and
+		// side-effects that are to be replicated to all replicas.
+		res.Replicated.IsLeaseRequest = ba.IsLeaseRequest()
+		res.Replicated.Timestamp = ba.Timestamp
+		if r.store.cfg.Settings.Version.IsActive(cluster.VersionMVCCNetworkStats) {
+			res.Replicated.Delta = ms.ToStatsDelta()
+		} else {
+			res.Replicated.DeprecatedDelta = &ms
+		}
+		// If the RangeAppliedState key is not being used and the cluster version is
+		// high enough to guarantee that all current and future binaries will
+		// understand the key, we send the migration flag through Raft. Because
+		// there is a delay between command proposal and application, we may end up
+		// setting this migration flag multiple times. This is ok, because the
+		// migration is idempotent.
+		// TODO(nvanbenschoten): This will be baked in to 2.1, so it can be removed
+		// in the 2.2 release.
+		r.mu.RLock()
+		usingAppliedStateKey := r.mu.state.UsingAppliedStateKey
+		r.mu.RUnlock()
+		if !usingAppliedStateKey &&
+			r.ClusterSettings().Version.IsActive(cluster.VersionRangeAppliedStateKey) {
+			if res.Replicated.State == nil {
+				res.Replicated.State = &storagepb.ReplicaState{}
+			}
+			res.Replicated.State.UsingAppliedStateKey = true
+		}
+	}
+
+	return &res, needConsensus, nil
+}
+
+// requestToProposal converts a BatchRequest into a ProposalData, by
+// evaluating it. The returned ProposalData is partially valid even
+// on a non-nil *roachpb.Error and should be proposed through Raft
+// if ProposalData.command is non-nil.
+func (r *Replica) requestToProposal(
+	ctx context.Context,
+	idKey storagebase.CmdIDKey,
+	ba roachpb.BatchRequest,
+	endCmds *endCmds,
+	spans *spanset.SpanSet,
+) (*ProposalData, *roachpb.Error) {
+	res, needConsensus, pErr := r.evaluateProposal(ctx, idKey, ba, spans)
+
+	// Fill out the results even if pErr != nil; we'll return the error below.
+	proposal := &ProposalData{
+		ctx:     ctx,
+		idKey:   idKey,
+		endCmds: endCmds,
+		doneCh:  make(chan proposalResult, 1),
+		Local:   &res.Local,
+		Request: &ba,
+	}
+
+	if needConsensus {
+		proposal.command = &storagepb.RaftCommand{
+			ReplicatedEvalResult: res.Replicated,
+			WriteBatch:           res.WriteBatch,
+			LogicalOpLog:         res.LogicalOpLog,
+		}
+	}
+
+	return proposal, pErr
+}
