@@ -18,17 +18,21 @@ import (
 	"context"
 	"reflect"
 	"testing"
-
-	"github.com/pkg/errors"
-	"google.golang.org/grpc/metadata"
+	"time"
 
 	"github.com/cockroachdb/cockroach/pkg/internal/client"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/storage"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
+	"github.com/cockroachdb/cockroach/pkg/testutils/sqlutils"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/require"
+	"go.etcd.io/etcd/raft"
+	"go.etcd.io/etcd/raft/raftpb"
+	"google.golang.org/grpc/metadata"
 )
 
 // testStream is a mock implementation of roachpb.Internal_RangeFeedServer.
@@ -79,7 +83,13 @@ func TestReplicaRangefeed(t *testing.T) {
 	ctx := context.Background()
 	sc := storage.TestStoreConfig(nil)
 	storage.RangefeedEnabled.Override(&sc.Settings.SV, true)
-	mtc := &multiTestContext{storeConfig: &sc}
+	mtc := &multiTestContext{
+		storeConfig: &sc,
+		// This test was written before the multiTestContext started creating many
+		// system ranges at startup, and hasn't been update to take that into
+		// account.
+		startWithSingleRange: true,
+	}
 	defer mtc.Stop()
 	mtc.Start(t, 3)
 	mtc.replicateRange(1, 1, 2)
@@ -119,7 +129,7 @@ func TestReplicaRangefeed(t *testing.T) {
 				Span: roachpb.Span{Key: roachpb.Key("a"), EndKey: roachpb.Key("z")},
 			}
 
-			pErr := mtc.Store(i).RangeFeed(ctx, &req, stream)
+			pErr := mtc.Store(i).RangeFeed(&req, stream)
 			streamErrC <- pErr
 		}(i)
 	}
@@ -207,15 +217,65 @@ func TestReplicaRangefeed(t *testing.T) {
 			t.Fatalf("got error for RangeFeed: %v", pErr)
 		}
 	}
+
+	// Bump the GC threshold and assert that RangeFeed below the timestamp will
+	// catch an error.
+	gcReq := &roachpb.GCRequest{
+		Threshold: initTime.Add(0, 1),
+	}
+	gcReq.Key = roachpb.Key(startKey)
+	gcReq.EndKey = mtc.Store(0).LookupReplica(startKey).Desc().EndKey.AsRawKey()
+	var ba roachpb.BatchRequest
+	ba.RangeID = rangeID
+	ba.Add(gcReq)
+	if _, pErr := mtc.Store(0).Send(ctx, ba); pErr != nil {
+		t.Fatal(pErr)
+	}
+
+	req := roachpb.RangeFeedRequest{
+		Header: roachpb.Header{
+			Timestamp: initTime,
+			RangeID:   rangeID,
+		},
+		Span: roachpb.Span{Key: roachpb.Key("a"), EndKey: roachpb.Key("z")},
+	}
+
+	testutils.SucceedsSoon(t, func() error {
+		for i := 0; i < replNum; i++ {
+			repl := mtc.Store(i).LookupReplica(startKey)
+			if repl == nil {
+				return errors.Errorf("replica not found on node #%d", i+1)
+			}
+			if cur := repl.GetGCThreshold(); cur.Less(gcReq.Threshold) {
+				return errors.Errorf("%s has GCThreshold %s < %s; hasn't applied the bump yet", repl, cur, gcReq.Threshold)
+			}
+			stream := newTestStream()
+			timer := time.AfterFunc(10*time.Second, stream.Cancel)
+			defer timer.Stop()
+			defer stream.Cancel()
+
+			if pErr := mtc.Store(i).RangeFeed(&req, stream); !testutils.IsPError(
+				pErr, `must be after replica GC threshold`,
+			) {
+				return pErr.GoError()
+			}
+		}
+		return nil
+	})
 }
 
 func TestReplicaRangefeedExpiringLeaseError(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	ctx := context.Background()
 	sc := storage.TestStoreConfig(nil)
 	storage.RangefeedEnabled.Override(&sc.Settings.SV, true)
-	mtc := &multiTestContext{storeConfig: &sc}
+	mtc := &multiTestContext{
+		storeConfig: &sc,
+		// This test was written before the multiTestContext started creating many
+		// system ranges at startup, and hasn't been update to take that into
+		// account.
+		startWithSingleRange: true,
+	}
 	defer mtc.Stop()
 	mtc.Start(t, 1)
 
@@ -232,7 +292,7 @@ func TestReplicaRangefeedExpiringLeaseError(t *testing.T) {
 	// immediately even if it didn't return the correct error.
 	stream.Cancel()
 
-	pErr := mtc.Store(0).RangeFeed(ctx, &req, stream)
+	pErr := mtc.Store(0).RangeFeed(&req, stream)
 	const exp = "expiration-based leases are incompatible with rangefeeds"
 	if !testutils.IsPError(pErr, exp) {
 		t.Errorf("expected error %q, found %v", exp, pErr)
@@ -248,7 +308,13 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 		subT.Helper()
 		sc := storage.TestStoreConfig(nil)
 		storage.RangefeedEnabled.Override(&sc.Settings.SV, true)
-		mtc := &multiTestContext{storeConfig: &sc}
+		mtc := &multiTestContext{
+			storeConfig: &sc,
+			// This test was written before the multiTestContext started creating many
+			// system ranges at startup, and hasn't been update to take that into
+			// account.
+			startWithSingleRange: true,
+		}
 		mtc.Start(subT, 3)
 		mtc.replicateRange(1, 1, 2)
 
@@ -341,7 +407,7 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 				Span: roachpb.Span{Key: roachpb.Key("a"), EndKey: roachpb.Key("z")},
 			}
 
-			pErr := mtc.Store(removeStore).RangeFeed(ctx, &req, stream)
+			pErr := mtc.Store(removeStore).RangeFeed(&req, stream)
 			streamErrC <- pErr
 		}()
 
@@ -370,7 +436,7 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 				Span: roachpb.Span{Key: roachpb.Key("a"), EndKey: roachpb.Key("z")},
 			}
 
-			pErr := mtc.Store(0).RangeFeed(ctx, &req, stream)
+			pErr := mtc.Store(0).RangeFeed(&req, stream)
 			streamErrC <- pErr
 		}()
 
@@ -418,7 +484,7 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 				Span: roachpb.Span{Key: roachpb.Key("a"), EndKey: splitKey},
 			}
 
-			pErr := mtc.Store(0).RangeFeed(ctx, &req, streamLeft)
+			pErr := mtc.Store(0).RangeFeed(&req, streamLeft)
 			streamLeftErrC <- pErr
 		}()
 
@@ -433,7 +499,7 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 				Span: roachpb.Span{Key: splitKey, EndKey: roachpb.Key("z")},
 			}
 
-			pErr := mtc.Store(0).RangeFeed(ctx, &req, streamRight)
+			pErr := mtc.Store(0).RangeFeed(&req, streamRight)
 			streamRightErrC <- pErr
 		}()
 
@@ -457,9 +523,9 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 		assertRangefeedRetryErr(t, pErrRight, roachpb.RangeFeedRetryError_REASON_RANGE_MERGED)
 	})
 	t.Run(roachpb.RangeFeedRetryError_REASON_RAFT_SNAPSHOT.String(), func(t *testing.T) {
-		const partitionStore = 2
 		mtc, rangeID := setup(t)
 		defer mtc.Stop()
+		partitionStore := mtc.Store(2)
 
 		mtc.stores[0].SetReplicaGCQueueActive(false)
 		mtc.stores[1].SetReplicaGCQueueActive(false)
@@ -476,15 +542,37 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 				Span: roachpb.Span{Key: roachpb.Key("a"), EndKey: roachpb.Key("z")},
 			}
 
-			pErr := mtc.Store(partitionStore).RangeFeed(ctx, &req, stream)
+			timer := time.AfterFunc(10*time.Second, stream.Cancel)
+			defer timer.Stop()
+
+			pErr := partitionStore.RangeFeed(&req, stream)
 			streamErrC <- pErr
 		}()
 
 		// Wait for the first checkpoint event.
 		waitForInitialCheckpoint(t, stream, streamErrC)
 
+		// Force the leader off the replica on partitionedStore. If it's the
+		// leader, this test will fall over when it cuts the replica off from
+		// Raft traffic.
+		testutils.SucceedsSoon(t, func() error {
+			repl, err := partitionStore.GetReplica(rangeID)
+			if err != nil {
+				return err
+			}
+			raftStatus := repl.RaftStatus()
+			if raftStatus != nil && raftStatus.RaftState == raft.StateFollower {
+				return nil
+			}
+			err = repl.AdminTransferLease(ctx, roachpb.StoreID(1))
+			return errors.Errorf("not raft follower: %+v, transferred lease: %v", raftStatus, err)
+		})
+
 		// Partition the replica from the rest of its range.
-		mtc.transport.GetCircuitBreaker(mtc.idents[partitionStore].NodeID).Break()
+		mtc.transport.Listen(partitionStore.Ident.StoreID, &unreliableRaftHandler{
+			rangeID:            rangeID,
+			RaftMessageHandler: partitionStore,
+		})
 
 		// Perform a write on the range.
 		pArgs := putArgs(roachpb.Key("c"), []byte("val2"))
@@ -512,7 +600,20 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 		}
 
 		// Remove the partition. Snapshot should follow.
-		mtc.transport.GetCircuitBreaker(mtc.idents[partitionStore].NodeID).Reset()
+		mtc.transport.Listen(partitionStore.Ident.StoreID, &unreliableRaftHandler{
+			rangeID:            rangeID,
+			RaftMessageHandler: partitionStore,
+			dropReq: func(req *storage.RaftMessageRequest) bool {
+				// Make sure that even going forward no MsgApp for what we just truncated can
+				// make it through. The Raft transport is asynchronous so this is necessary
+				// to make the test pass reliably.
+				// NB: the Index on the message is the log index that _precedes_ any of the
+				// entries in the MsgApp, so filter where msg.Index < index, not <= index.
+				return req.Message.Type == raftpb.MsgApp && req.Message.Index < index
+			},
+			dropHB:   func(*storage.RaftHeartbeat) bool { return false },
+			dropResp: func(*storage.RaftMessageResponse) bool { return false },
+		})
 
 		// Check the error.
 		pErr := <-streamErrC
@@ -533,7 +634,7 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 				Span: roachpb.Span{Key: roachpb.Key("a"), EndKey: roachpb.Key("z")},
 			}
 
-			pErr := mtc.Store(0).RangeFeed(ctx, &req, stream)
+			pErr := mtc.Store(0).RangeFeed(&req, stream)
 			streamErrC <- pErr
 		}()
 
@@ -555,3 +656,111 @@ func TestReplicaRangefeedRetryErrors(t *testing.T) {
 		assertRangefeedRetryErr(t, pErr, roachpb.RangeFeedRetryError_REASON_LOGICAL_OPS_MISSING)
 	})
 }
+
+// TestReplicaRangefeedNudgeSlowClosedTimestamp tests that rangefeed detects
+// that its closed timestamp updates have stalled and requests new information
+// from its Range's leaseholder. This is a regression test for #35142.
+func TestReplicaRangefeedNudgeSlowClosedTimestamp(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	ctx := context.Background()
+	tc, db, desc, repls := setupTestClusterForClosedTimestampTesting(ctx, t, testingTargetDuration)
+	defer tc.Stopper().Stop(ctx)
+
+	sqlDB := sqlutils.MakeSQLRunner(db)
+	sqlDB.Exec(t, `SET CLUSTER SETTING kv.rangefeed.enabled = true`)
+	// While we're here, drop the target duration. This was set to
+	// testingTargetDuration above, but this is higher then it needs to be now
+	// that cluster and schema setup is complete.
+	sqlDB.Exec(t, `SET CLUSTER SETTING kv.closed_timestamp.target_duration = '10ms'`)
+
+	// Make sure all the nodes have gotten the rangefeed enabled setting from
+	// gossip, so that they will immediately be able to accept RangeFeeds. The
+	// target_duration one is just to speed up the test, we don't care if it has
+	// propagated everywhere yet.
+	testutils.SucceedsSoon(t, func() error {
+		for i := 0; i < tc.NumServers(); i++ {
+			var enabled bool
+			if err := tc.ServerConn(i).QueryRow(
+				`SHOW CLUSTER SETTING kv.rangefeed.enabled`,
+			).Scan(&enabled); err != nil {
+				return err
+			}
+			if !enabled {
+				return errors.Errorf(`waiting for rangefeed to be enabled on node %d`, i)
+			}
+		}
+		return nil
+	})
+
+	ts1 := tc.Server(0).Clock().Now()
+	rangeFeedCtx, rangeFeedCancel := context.WithCancel(ctx)
+	defer rangeFeedCancel()
+	rangeFeedChs := make([]chan *roachpb.RangeFeedEvent, len(repls))
+	rangeFeedErrC := make(chan error, len(repls))
+	for i := range repls {
+		ds := tc.Server(i).DistSender()
+		rangeFeedCh := make(chan *roachpb.RangeFeedEvent)
+		rangeFeedChs[i] = rangeFeedCh
+		go func() {
+			req := roachpb.RangeFeedRequest{
+				Header: roachpb.Header{Timestamp: ts1},
+				Span: roachpb.Span{
+					Key: desc.StartKey.AsRawKey(), EndKey: desc.EndKey.AsRawKey(),
+				},
+			}
+			rangeFeedErrC <- ds.RangeFeed(rangeFeedCtx, &req, rangeFeedCh)
+		}()
+	}
+
+	// Wait for a RangeFeed checkpoint on each RangeFeed after the RangeFeed
+	// initial scan time (which is the timestamp passed in the request) to make
+	// sure everything is set up. We intentionally don't care about the spans in
+	// the checkpoints, just verifying that something has made it past the
+	// initial scan and is running.
+	waitForCheckpoint := func(ts hlc.Timestamp) {
+		t.Helper()
+		for _, rangeFeedCh := range rangeFeedChs {
+			checkpointed := false
+			for !checkpointed {
+				select {
+				case event := <-rangeFeedCh:
+					if c := event.Checkpoint; c != nil && ts.Less(c.ResolvedTS) {
+						checkpointed = true
+					}
+				case err := <-rangeFeedErrC:
+					t.Fatal(err)
+				}
+			}
+		}
+	}
+	waitForCheckpoint(ts1)
+
+	// Clear the closed timestamp storage on each server. This simulates the case
+	// where a closed timestamp message is lost or a node restarts. To recover,
+	// the servers will need to request an update from the leaseholder.
+	for i := 0; i < tc.NumServers(); i++ {
+		stores := tc.Server(i).GetStores().(*storage.Stores)
+		err := stores.VisitStores(func(s *storage.Store) error {
+			s.ClearClosedTimestampStorage()
+			return nil
+		})
+		require.NoError(t, err)
+	}
+
+	// Wait for another RangeFeed checkpoint after the store was cleared. Without
+	// RangeFeed nudging closed timestamps, this doesn't happen on its own. Again,
+	// we intentionally don't care about the spans in the checkpoints, just
+	// verifying that something has made it past the cleared time.
+	ts2 := tc.Server(0).Clock().Now()
+	waitForCheckpoint(ts2)
+
+	// Make sure the RangeFeed hasn't errored yet.
+	select {
+	case err := <-rangeFeedErrC:
+		t.Fatal(err)
+	default:
+	}
+	// Now cancel it and wait for it to shut down.
+	rangeFeedCancel()
+}
