@@ -33,8 +33,10 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
+	"github.com/cockroachdb/cockroach/pkg/storage/raftentry"
 	"github.com/cockroachdb/cockroach/pkg/storage/stateloader"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
@@ -44,6 +46,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/tracing"
 	"github.com/kr/pretty"
 	"github.com/pkg/errors"
+	"github.com/stretchr/testify/assert"
 	"go.etcd.io/etcd/raft/raftpb"
 	"golang.org/x/time/rate"
 )
@@ -54,7 +57,7 @@ func entryEq(l, r raftpb.Entry) error {
 	}
 	_, lData := DecodeRaftCommand(l.Data)
 	_, rData := DecodeRaftCommand(r.Data)
-	var lc, rc storagebase.RaftCommand
+	var lc, rc storagepb.RaftCommand
 	if err := protoutil.Unmarshal(lData, &lc); err != nil {
 		return errors.Wrap(err, "unmarshalling LHS")
 	}
@@ -68,10 +71,10 @@ func entryEq(l, r raftpb.Entry) error {
 }
 
 func mkEnt(
-	v raftCommandEncodingVersion, index, term uint64, as *storagebase.ReplicatedEvalResult_AddSSTable,
+	v raftCommandEncodingVersion, index, term uint64, as *storagepb.ReplicatedEvalResult_AddSSTable,
 ) raftpb.Entry {
 	cmdIDKey := strings.Repeat("x", raftCommandIDLen)
-	var cmd storagebase.RaftCommand
+	var cmd storagepb.RaftCommand
 	cmd.ReplicatedEvalResult.AddSSTable = as
 	b, err := protoutil.Marshal(&cmd)
 	if err != nil {
@@ -91,7 +94,7 @@ func TestSideloadingSideloadedStorage(t *testing.T) {
 	t.Run("Disk", func(t *testing.T) {
 		maker := func(
 			s *cluster.Settings, rangeID roachpb.RangeID, rep roachpb.ReplicaID, name string, eng engine.Engine,
-		) (sideloadStorage, error) {
+		) (SideloadStorage, error) {
 			return newDiskSideloadStorage(s, rangeID, rep, name, rate.NewLimiter(rate.Inf, math.MaxInt64), eng)
 		}
 		testSideloadingSideloadedStorage(t, maker)
@@ -100,7 +103,7 @@ func TestSideloadingSideloadedStorage(t *testing.T) {
 
 func testSideloadingSideloadedStorage(
 	t *testing.T,
-	maker func(*cluster.Settings, roachpb.RangeID, roachpb.ReplicaID, string, engine.Engine) (sideloadStorage, error),
+	maker func(*cluster.Settings, roachpb.RangeID, roachpb.ReplicaID, string, engine.Engine) (SideloadStorage, error),
 ) {
 	dir, cleanup := testutils.TempDir(t)
 	defer cleanup()
@@ -190,7 +193,7 @@ func testSideloadingSideloadedStorage(
 		{
 			err: nil,
 			fun: func() error {
-				_, err := ss.TruncateTo(ctx, 123)
+				_, _, err := ss.TruncateTo(ctx, 123)
 				return err
 			},
 		},
@@ -231,13 +234,6 @@ func testSideloadingSideloadedStorage(
 		}
 	}
 
-	// Verify a sideloaded storage for another ReplicaID doesn't see the files.
-	if otherSS, err := maker(st, 1, 999 /* ReplicaID */, dir, eng); err != nil {
-		t.Fatal(err)
-	} else if _, err = otherSS.Get(ctx, payloads[0], highTerm); err != errSideloadedFileNotFound {
-		t.Fatal("expected not found")
-	}
-
 	// Just for fun, recreate the original storage (unless it's the in-memory
 	// one), which shouldn't change anything about its state.
 	if !isInMem {
@@ -262,7 +258,7 @@ func testSideloadingSideloadedStorage(
 
 	for n := range payloads {
 		// Truncate indexes <= payloads[n] (payloads is sorted in increasing order).
-		if _, err := ss.TruncateTo(ctx, payloads[n]); err != nil {
+		if _, _, err := ss.TruncateTo(ctx, payloads[n]); err != nil {
 			t.Fatalf("%d: %s", n, err)
 		}
 		// Index payloads[n] and above are still there (truncation is exclusive)
@@ -295,7 +291,7 @@ func testSideloadingSideloadedStorage(
 		}
 		defer f.Close()
 
-		_, err = ss.TruncateTo(ctx, math.MaxUint64)
+		_, _, err = ss.TruncateTo(ctx, math.MaxUint64)
 		if err == nil {
 			t.Fatalf("sideloaded directory should not have been removable due to extra file %s", nonRemovableFile)
 		}
@@ -310,7 +306,7 @@ func testSideloadingSideloadedStorage(
 		}
 
 		// Test that directory is removed when filepath.Glob returns 0 matches.
-		if _, err := ss.TruncateTo(ctx, math.MaxUint64); err != nil {
+		if _, _, err := ss.TruncateTo(ctx, math.MaxUint64); err != nil {
 			t.Fatal(err)
 		}
 		// Ensure directory is removed, now that all files should be gone.
@@ -334,7 +330,7 @@ func testSideloadingSideloadedStorage(
 			}
 		}
 		assertCreated(true)
-		if _, err := ss.TruncateTo(ctx, math.MaxUint64); err != nil {
+		if _, _, err := ss.TruncateTo(ctx, math.MaxUint64); err != nil {
 			t.Fatal(err)
 		}
 		// Ensure directory is removed when all records are removed.
@@ -356,7 +352,7 @@ func testSideloadingSideloadedStorage(
 	assertCreated(false)
 
 	// Sanity check that we can call TruncateTo without the directory existing.
-	if _, err := ss.TruncateTo(ctx, 1); err != nil {
+	if _, _, err := ss.TruncateTo(ctx, 1); err != nil {
 		t.Fatal(err)
 	}
 
@@ -394,6 +390,72 @@ func testSideloadingSideloadedStorage(
 	}
 }
 
+func TestSideloadedStorageReplicaIDMigration(t *testing.T) {
+	defer leaktest.AfterTest(t)()
+
+	dir, cleanup := testutils.TempDir(t)
+	defer cleanup()
+
+	ctx := context.Background()
+	preV := cluster.VersionByKey(cluster.VersionSideloadedStorageNoReplicaID - 1)
+	stPre := cluster.MakeTestingClusterSettingsWithVersion(preV, preV)
+
+	postV := cluster.VersionByKey(cluster.VersionSideloadedStorageNoReplicaID)
+	stPost := cluster.MakeTestingClusterSettingsWithVersion(postV, postV)
+
+	const rangeID = roachpb.RangeID(1)
+
+	cleanup, cache, eng := newRocksDB(t)
+	defer cleanup()
+	defer cache.Release()
+	defer eng.Close()
+	limiter := rate.NewLimiter(rate.Inf, 1<<9)
+
+	var ss SideloadStorage
+	create := func(st *cluster.Settings, replicaID roachpb.ReplicaID) *diskSideloadStorage {
+		t.Helper()
+		if err := moveSideloadedData(ss, dir, rangeID, replicaID); err != nil {
+			t.Fatal(err)
+		}
+		ss, err := newDiskSideloadStorage(st, rangeID, replicaID, dir, limiter, eng)
+		if err != nil {
+			t.Fatal(err)
+		}
+		return ss
+	}
+
+	ss = create(stPre, 120)
+	foo := []byte("foo")
+	if err := ss.Put(ctx, 1, 10, foo); err != nil {
+		t.Fatal(err)
+	}
+
+	// Jog moveSideloadedData (via create) prior to the migration.
+	for _, replicaID := range []roachpb.ReplicaID{123, 124} {
+		ss = create(stPre, replicaID)
+		v, err := ss.Get(ctx, 1, 10)
+		if err != nil {
+			t.Fatal(err)
+		}
+		assert.Equal(t, foo, v)
+	}
+
+	// Migrate.
+	for _, replicaID := range []roachpb.ReplicaID{125, 126} {
+		ss = create(stPost, replicaID)
+		v, err := ss.Get(ctx, 1, 10)
+		if err != nil {
+			t.Fatal(err)
+		}
+		assert.Equal(t, foo, v)
+	}
+
+	// Check that we really did migrate.
+	prevDir := ss.Dir()
+	ss = create(stPost, 129)
+	assert.Equal(t, prevDir, ss.Dir())
+}
+
 func TestRaftSSTableSideloadingInline(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
@@ -405,22 +467,22 @@ func TestRaftSSTableSideloadingInline(t *testing.T) {
 		// after having (perhaps) been modified.
 		thin, fat raftpb.Entry
 		// Populate the raft entry cache and sideload storage before running the test.
-		setup func(*raftEntryCache, sideloadStorage)
+		setup func(*raftentry.Cache, SideloadStorage)
 		// If nonempty, the error expected from maybeInlineSideloadedRaftCommand.
 		expErr string
 		// If nonempty, a regex that the recorded trace span must match.
 		expTrace string
 	}
 
-	sstFat := storagebase.ReplicatedEvalResult_AddSSTable{
+	sstFat := storagepb.ReplicatedEvalResult_AddSSTable{
 		Data:  []byte("foo"),
 		CRC32: 0, // not checked
 	}
-	sstThin := storagebase.ReplicatedEvalResult_AddSSTable{
+	sstThin := storagepb.ReplicatedEvalResult_AddSSTable{
 		CRC32: 0, // not checked
 	}
 
-	putOnDisk := func(ec *raftEntryCache, ss sideloadStorage) {
+	putOnDisk := func(ec *raftentry.Cache, ss SideloadStorage) {
 		if err := ss.Put(context.Background(), 5, 6, sstFat.Data); err != nil {
 			t.Fatal(err)
 		}
@@ -447,14 +509,14 @@ func TestRaftSSTableSideloadingInline(t *testing.T) {
 		},
 		"v2-with-payload-with-file-with-cache": {
 			thin: mkEnt(v2, 5, 6, &sstThin), fat: mkEnt(v2, 5, 6, &sstFat),
-			setup: func(ec *raftEntryCache, ss sideloadStorage) {
+			setup: func(ec *raftentry.Cache, ss SideloadStorage) {
 				putOnDisk(ec, ss)
-				ec.addEntries(rangeID, []raftpb.Entry{mkEnt(v2, 5, 6, &sstFat)})
+				ec.Add(rangeID, []raftpb.Entry{mkEnt(v2, 5, 6, &sstFat)}, true)
 			}, expTrace: "using cache hit",
 		},
 		"v2-fat-without-file": {
 			thin: mkEnt(v2, 5, 6, &sstFat), fat: mkEnt(v2, 5, 6, &sstFat),
-			setup:    func(ec *raftEntryCache, ss sideloadStorage) {},
+			setup:    func(ec *raftentry.Cache, ss SideloadStorage) {},
 			expTrace: "already inlined",
 		},
 	}
@@ -463,7 +525,7 @@ func TestRaftSSTableSideloadingInline(t *testing.T) {
 		ctx, collect, cancel := tracing.ContextWithRecordingSpan(context.Background(), "test-recording")
 		defer cancel()
 
-		ec := newRaftEntryCache(1024) // large enough
+		ec := raftentry.NewCache(1024) // large enough
 		ss := mustNewInMemSideloadStorage(rangeID, roachpb.ReplicaID(1), ".")
 		if test.setup != nil {
 			test.setup(ec, ss)
@@ -507,68 +569,10 @@ func TestRaftSSTableSideloadingInline(t *testing.T) {
 	}
 }
 
-func TestRaftSSTableSideloadingInflight(t *testing.T) {
-	defer leaktest.AfterTest(t)()
-
-	ctx, collect, cancel := tracing.ContextWithRecordingSpan(context.Background(), "test-recording")
-	defer cancel()
-
-	sideloaded := mustNewInMemSideloadStorage(roachpb.RangeID(5), roachpb.ReplicaID(7), ".")
-
-	// We'll set things up so that while sideloading this entry, there
-	// unmarshaled one is already in memory (so the payload here won't even be
-	// looked at).
-	preEnts := []raftpb.Entry{mkEnt(raftVersionSideloaded, 7, 1, &storagebase.ReplicatedEvalResult_AddSSTable{
-		Data:  []byte("not the payload you're looking for"),
-		CRC32: 0, // not checked
-	})}
-
-	origBytes := []byte("compare me")
-
-	// Pretend there's an inflight command that actually has an SSTable in it.
-	var pendingCmd storagebase.RaftCommand
-	pendingCmd.ReplicatedEvalResult.AddSSTable = &storagebase.ReplicatedEvalResult_AddSSTable{
-		Data: origBytes, CRC32: 0, // not checked
-	}
-	maybeCmd := func(cmdID storagebase.CmdIDKey) (storagebase.RaftCommand, bool) {
-		return pendingCmd, true
-	}
-
-	// The entry should be recognized as "to be sideloaded", then maybeCmd is
-	// invoked and supplies the RaftCommand, whose SSTable is then persisted.
-	postEnts, size, err := maybeSideloadEntriesImpl(ctx, preEnts, sideloaded, maybeCmd)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	if len(postEnts) != 1 {
-		t.Fatalf("expected exactly one entry: %+v", postEnts)
-	}
-	if size != int64(len(origBytes)) {
-		t.Fatalf("expected %d sideloadedSize, but found %d", len(origBytes), size)
-	}
-
-	if b, err := sideloaded.Get(ctx, preEnts[0].Index, preEnts[0].Term); err != nil {
-		t.Fatal(err)
-	} else if !bytes.Equal(b, origBytes) {
-		t.Fatalf("expected payload %s, got %s", origBytes, b)
-	}
-
-	re := regexp.MustCompile(`(?ms)copying entries slice of length 1.*command already in memory.*writing payload`)
-	if trace := tracing.FormatRecordedSpans(collect()); !re.MatchString(trace) {
-		t.Fatalf("trace did not match %s:\n%s", re, trace)
-	}
-}
-
 func TestRaftSSTableSideloadingSideload(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 
-	ctx := context.Background()
-	noCmd := func(storagebase.CmdIDKey) (cmd storagebase.RaftCommand, ok bool) {
-		return
-	}
-
-	addSST := storagebase.ReplicatedEvalResult_AddSSTable{
+	addSST := storagepb.ReplicatedEvalResult_AddSSTable{
 		Data: []byte("foo"), CRC32: 0, // not checked
 	}
 
@@ -622,8 +626,9 @@ func TestRaftSSTableSideloadingSideload(t *testing.T) {
 
 	for _, test := range testCases {
 		t.Run(test.name, func(t *testing.T) {
+			ctx := context.Background()
 			sideloaded := mustNewInMemSideloadStorage(roachpb.RangeID(3), roachpb.ReplicaID(17), ".")
-			postEnts, size, err := maybeSideloadEntriesImpl(ctx, test.preEnts, sideloaded, noCmd)
+			postEnts, size, err := maybeSideloadEntriesImpl(ctx, test.preEnts, sideloaded)
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -781,7 +786,7 @@ func testRaftSSTableSideloadingProposal(t *testing.T, engineInMem, mockSideloade
 	if _, err := tc.store.raftLogQueue.Add(tc.repl, 99.99 /* priority */); err != nil {
 		t.Fatal(err)
 	}
-	tc.store.ForceRaftLogScanAndProcess()
+	tc.store.MustForceRaftLogScanAndProcess()
 	// SST is definitely truncated now, so recomputing the Raft log keys should match up with
 	// the tracked size.
 	verifyLogSizeInSync(t, tc.repl)
@@ -895,7 +900,7 @@ func TestRaftSSTableSideloadingSnapshot(t *testing.T) {
 		}
 
 		var ent raftpb.Entry
-		var cmd storagebase.RaftCommand
+		var cmd storagepb.RaftCommand
 		var finalEnt raftpb.Entry
 		for _, entryBytes := range mockSender.logEntries {
 			if err := protoutil.Unmarshal(entryBytes, &ent); err != nil {
@@ -933,13 +938,13 @@ func TestRaftSSTableSideloadingSnapshot(t *testing.T) {
 		tc.repl.mu.Lock()
 		defer tc.repl.mu.Unlock()
 		for _, withSS := range []bool{false, true} {
-			tc.store.raftEntryCache.clearTo(tc.repl.RangeID, sideloadedIndex+1)
+			tc.store.raftEntryCache.Clear(tc.repl.RangeID, sideloadedIndex+1)
 
-			var ss sideloadStorage
+			var ss SideloadStorage
 			if withSS {
 				ss = tc.repl.raftMu.sideloaded
 			}
-			rsl := stateloader.Make(tc.store.ClusterSettings(), tc.repl.RangeID)
+			rsl := stateloader.Make(tc.repl.RangeID)
 			entries, err := entries(
 				ctx, rsl, tc.store.Engine(), tc.repl.RangeID, tc.store.raftEntryCache,
 				ss, sideloadedIndex, sideloadedIndex+1, 1<<20,
@@ -950,7 +955,7 @@ func TestRaftSSTableSideloadingSnapshot(t *testing.T) {
 			if len(entries) != 1 {
 				t.Fatalf("no or too many entries returned from cache: %+v", entries)
 			}
-			ents, _, _, _ := tc.store.raftEntryCache.getEntries(nil, tc.repl.RangeID, sideloadedIndex, sideloadedIndex+1, 1<<20)
+			ents, _, _, _ := tc.store.raftEntryCache.Scan(nil, tc.repl.RangeID, sideloadedIndex, sideloadedIndex+1, 1<<20)
 			if withSS {
 				// We passed the sideload storage, so we expect to get our
 				// inlined index back from the cache.
@@ -999,7 +1004,7 @@ func TestRaftSSTableSideloadingSnapshot(t *testing.T) {
 		tc.repl.raftMu.Unlock()
 		// Additionally we need to clear out the entry from the cache because
 		// that would still save the day.
-		tc.store.raftEntryCache.clearTo(tc.repl.RangeID, sideloadedIndex+1)
+		tc.store.raftEntryCache.Clear(tc.repl.RangeID, sideloadedIndex+1)
 
 		mockSender := &mockSender{}
 		err = sendSnapshot(
@@ -1089,62 +1094,3 @@ func TestRaftSSTableSideloadingTruncation(t *testing.T) {
 	}
 
 }
-
-func TestRaftSSTableSideloadingUpdatedReplicaID(t *testing.T) {
-	defer leaktest.AfterTest(t)()
-
-	cleanup, cache, eng := newRocksDB(t)
-	defer cleanup()
-	defer cache.Release()
-	defer eng.Close()
-
-	tc := testContext{}
-	stopper := stop.NewStopper()
-	defer stopper.Stop(context.TODO())
-	tc.engine = eng
-
-	tc.Start(t, stopper)
-	repl := tc.repl
-	ctx := context.Background()
-
-	const (
-		index = 123
-		term  = 456
-	)
-
-	val := []byte("foo")
-
-	repl.raftMu.Lock()
-	oldDir := repl.raftMu.sideloaded.Dir()
-	err := repl.raftMu.sideloaded.Put(ctx, index, term, val)
-	repl.raftMu.Unlock()
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	// Set the ReplicaID on the replica.
-	if err := repl.setReplicaID(2); err != nil {
-		t.Fatal(err)
-	}
-
-	newDir := repl.raftMu.sideloaded.Dir()
-
-	if oldDir == newDir {
-		t.Fatalf("old and new sideloaded directory are equal: %s", oldDir)
-	}
-
-	// We assert below that oldDir moved to newDir.
-
-	repl.raftMu.Lock()
-	_, err = repl.raftMu.sideloaded.Get(ctx, index, term)
-	repl.raftMu.Unlock()
-
-	log.Infof(ctx, "olddir is %s, newdir is %s", oldDir, newDir)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	if _, err := os.Stat(oldDir); !os.IsNotExist(err) {
-		t.Fatal(err)
-	}
-}
