@@ -21,17 +21,18 @@ import (
 	"sync/atomic"
 	"time"
 
-	"github.com/pkg/errors"
-	"go.etcd.io/etcd/raft"
-
+	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/pkg/errors"
+	"go.etcd.io/etcd/raft"
 )
 
 const (
@@ -167,7 +168,16 @@ func newReplicateQueue(store *Store, g *gossip.Gossip, allocator Allocator) *rep
 	// Register gossip and node liveness callbacks to signal that
 	// replicas in purgatory might be retried.
 	if g != nil { // gossip is nil for some unittests
-		g.RegisterCallback(gossip.MakePrefixPattern(gossip.KeyStorePrefix), func(_ string, _ roachpb.Value) {
+		g.RegisterCallback(gossip.MakePrefixPattern(gossip.KeyStorePrefix), func(key string, _ roachpb.Value) {
+			if !rq.store.IsStarted() {
+				return
+			}
+			// Because updates to our store's own descriptor won't affect
+			// replicas in purgatory, skip updating the purgatory channel
+			// in this case.
+			if storeID, err := gossip.StoreIDFromKey(key); err == nil && storeID == rq.store.StoreID() {
+				return
+			}
 			updateFn()
 		})
 	}
@@ -181,7 +191,7 @@ func newReplicateQueue(store *Store, g *gossip.Gossip, allocator Allocator) *rep
 }
 
 func (rq *replicateQueue) shouldQueue(
-	ctx context.Context, now hlc.Timestamp, repl *Replica, sysCfg config.SystemConfig,
+	ctx context.Context, now hlc.Timestamp, repl *Replica, sysCfg *config.SystemConfig,
 ) (shouldQ bool, priority float64) {
 	if !repl.store.splitQueue.Disabled() && repl.needsSplitBySize() {
 		// If the range exceeds the split threshold, let that finish first.
@@ -195,13 +205,8 @@ func (rq *replicateQueue) shouldQueue(
 		return
 	}
 
-	// Find the zone config for this range.
-	desc := repl.Desc()
-	zone, err := sysCfg.GetZoneConfigForKey(desc.StartKey)
-	if err != nil {
-		log.Error(ctx, err)
-		return
-	}
+	// Get the descriptor and zone config for this range.
+	desc, zone := repl.DescAndZone()
 
 	rangeInfo := rangeInfoForRepl(repl, desc)
 	action, priority := rq.allocator.ComputeAction(ctx, zone, rangeInfo)
@@ -236,7 +241,7 @@ func (rq *replicateQueue) shouldQueue(
 }
 
 func (rq *replicateQueue) process(
-	ctx context.Context, repl *Replica, sysCfg config.SystemConfig,
+	ctx context.Context, repl *Replica, sysCfg *config.SystemConfig,
 ) error {
 	retryOpts := retry.Options{
 		InitialBackoff: 50 * time.Millisecond,
@@ -278,11 +283,11 @@ func (rq *replicateQueue) process(
 func (rq *replicateQueue) processOneChange(
 	ctx context.Context,
 	repl *Replica,
-	sysCfg config.SystemConfig,
+	sysCfg *config.SystemConfig,
 	canTransferLease func() bool,
 	dryRun bool,
 ) (requeue bool, _ error) {
-	desc := repl.Desc()
+	desc, zone := repl.DescAndZone()
 
 	// Avoid taking action if the range has too many dead replicas to make
 	// quorum.
@@ -295,11 +300,6 @@ func (rq *replicateQueue) processOneChange(
 		}
 	}
 
-	zone, err := sysCfg.GetZoneConfigForKey(desc.StartKey)
-	if err != nil {
-		return false, err
-	}
-
 	rangeInfo := rangeInfoForRepl(repl, desc)
 	switch action, _ := rq.allocator.ComputeAction(ctx, zone, rangeInfo); action {
 	case AllocatorNoop:
@@ -321,7 +321,7 @@ func (rq *replicateQueue) processOneChange(
 		}
 
 		clusterNodes := rq.allocator.storePool.ClusterNodeCount()
-		need := GetNeededReplicas(zone.NumReplicas, clusterNodes)
+		need := GetNeededReplicas(*zone.NumReplicas, clusterNodes)
 		willHave := len(desc.Replicas) + 1
 
 		// Only up-replicate if there are suitable allocation targets such
@@ -363,7 +363,7 @@ func (rq *replicateQueue) processOneChange(
 			newReplica,
 			desc,
 			SnapshotRequest_RECOVERY,
-			ReasonRangeUnderReplicated,
+			storagepb.ReasonRangeUnderReplicated,
 			details,
 			dryRun,
 		); err != nil {
@@ -371,19 +371,65 @@ func (rq *replicateQueue) processOneChange(
 		}
 	case AllocatorRemove:
 		log.VEventf(ctx, 1, "removing a replica")
-		lastReplAdded, lastAddedTime := repl.LastReplicaAdded()
-		if timeutil.Since(lastAddedTime) > newReplicaGracePeriod {
-			lastReplAdded = 0
+
+		// This retry loop involves quick operations on local state, so a
+		// small MaxBackoff is good (but those local variables change on
+		// network time scales as raft receives responses).
+		//
+		// TODO(bdarnell): There's another retry loop at process(). It
+		// would be nice to combine these, but I'm keeping them separate
+		// for now so we can tune the options separately.
+		retryOpts := retry.Options{
+			InitialBackoff: time.Millisecond,
+			MaxBackoff:     200 * time.Millisecond,
+			Multiplier:     2,
+		}
+
+		var candidates []roachpb.ReplicaDescriptor
+		deadline := timeutil.Now().Add(2 * base.NetworkTimeout)
+		for r := retry.StartWithCtx(ctx, retryOpts); r.Next() && timeutil.Now().Before(deadline); {
+			lastReplAdded, lastAddedTime := repl.LastReplicaAdded()
+			if timeutil.Since(lastAddedTime) > newReplicaGracePeriod {
+				lastReplAdded = 0
+			}
+			raftStatus := repl.RaftStatus()
+			if raftStatus == nil || raftStatus.RaftState != raft.StateLeader {
+				// If we've lost raft leadership, we're unlikely to regain it so give up immediately.
+				return false, &benignError{errors.Errorf("not raft leader while range needs removal")}
+			}
+			candidates = filterUnremovableReplicas(raftStatus, desc.Replicas, lastReplAdded)
+			log.VEventf(ctx, 3, "filtered unremovable replicas from %v to get %v as candidates for removal: %s",
+				desc.Replicas, candidates, rangeRaftProgress(raftStatus, desc.Replicas))
+			if len(candidates) > 0 {
+				break
+			}
+			if len(raftStatus.Progress) <= 2 {
+				// HACK(bdarnell): Downreplicating to a single node from
+				// multiple nodes is not really supported. There are edge
+				// cases in which the two peers stop communicating with each
+				// other too soon and we don't reach a satisfactory
+				// resolution. However, some tests (notably
+				// TestRepartitioning) get into this state, and if the
+				// replication queue spends its entire timeout waiting for the
+				// downreplication to finish the test will time out. As a
+				// hack, just fail-fast when we're trying to go down to a
+				// single replica.
+				break
+			}
+			// After upreplication, the candidates for removal could still
+			// be catching up. The allocator determined that the range was
+			// over-replicated, and it's important to clear that state as
+			// quickly as we can (because over-replicated ranges may be
+			// under-diversified). If we return an error here, this range
+			// probably won't be processed again until the next scanner
+			// cycle, which is too long, so we retry here.
 		}
-		candidates := filterUnremovableReplicas(repl.RaftStatus(), desc.Replicas, lastReplAdded)
-		log.VEventf(ctx, 3, "filtered unremovable replicas from %v to get %v as candidates for removal: %s",
-			desc.Replicas, candidates, rangeRaftProgress(repl.RaftStatus(), desc.Replicas))
 		if len(candidates) == 0 {
-			// After rapid upreplication, the candidates for removal could still be catching up.
-			// Mark this error as benign so it doesn't create confusion in the logs.
-			return false, &benignError{errors.Errorf("no removable replicas from range that needs a removal: %s",
-				rangeRaftProgress(repl.RaftStatus(), desc.Replicas))}
+			// If we timed out and still don't have any valid candidates, give up.
+			return false, errors.Errorf("no removable replicas from range that needs a removal: %s",
+				rangeRaftProgress(repl.RaftStatus(), desc.Replicas))
 		}
+
 		removeReplica, details, err := rq.allocator.RemoveTarget(ctx, zone, candidates, rangeInfo)
 		if err != nil {
 			return false, err
@@ -424,7 +470,7 @@ func (rq *replicateQueue) processOneChange(
 				StoreID: removeReplica.StoreID,
 			}
 			if err := rq.removeReplica(
-				ctx, repl, target, desc, ReasonRangeOverReplicated, details, dryRun,
+				ctx, repl, target, desc, storagepb.ReasonRangeOverReplicated, details, dryRun,
 			); err != nil {
 				return false, err
 			}
@@ -468,7 +514,7 @@ func (rq *replicateQueue) processOneChange(
 				StoreID: decommissioningReplica.StoreID,
 			}
 			if err := rq.removeReplica(
-				ctx, repl, target, desc, ReasonStoreDecommissioning, "", dryRun,
+				ctx, repl, target, desc, storagepb.ReasonStoreDecommissioning, "", dryRun,
 			); err != nil {
 				return false, err
 			}
@@ -487,7 +533,7 @@ func (rq *replicateQueue) processOneChange(
 			StoreID: deadReplica.StoreID,
 		}
 		if err := rq.removeReplica(
-			ctx, repl, target, desc, ReasonStoreDead, "", dryRun,
+			ctx, repl, target, desc, storagepb.ReasonStoreDead, "", dryRun,
 		); err != nil {
 			return false, err
 		}
@@ -515,7 +561,7 @@ func (rq *replicateQueue) processOneChange(
 					rebalanceReplica,
 					desc,
 					SnapshotRequest_REBALANCE,
-					ReasonRebalance,
+					storagepb.ReasonRebalance,
 					details,
 					dryRun,
 				); err != nil {
@@ -566,7 +612,7 @@ func (rq *replicateQueue) findTargetAndTransferLease(
 	ctx context.Context,
 	repl *Replica,
 	desc *roachpb.RangeDescriptor,
-	zone config.ZoneConfig,
+	zone *config.ZoneConfig,
 	opts transferLeaseOptions,
 ) (bool, error) {
 	candidates := filterBehindReplicas(repl.RaftStatus(), desc.Replicas)
@@ -618,14 +664,14 @@ func (rq *replicateQueue) addReplica(
 	target roachpb.ReplicationTarget,
 	desc *roachpb.RangeDescriptor,
 	priority SnapshotRequest_Priority,
-	reason RangeLogEventReason,
+	reason storagepb.RangeLogEventReason,
 	details string,
 	dryRun bool,
 ) error {
 	if dryRun {
 		return nil
 	}
-	if err := repl.changeReplicas(ctx, roachpb.ADD_REPLICA, target, desc, priority, reason, details); err != nil {
+	if _, err := repl.changeReplicas(ctx, roachpb.ADD_REPLICA, target, desc, priority, reason, details); err != nil {
 		return err
 	}
 	rangeInfo := rangeInfoForRepl(repl, desc)
@@ -638,14 +684,14 @@ func (rq *replicateQueue) removeReplica(
 	repl *Replica,
 	target roachpb.ReplicationTarget,
 	desc *roachpb.RangeDescriptor,
-	reason RangeLogEventReason,
+	reason storagepb.RangeLogEventReason,
 	details string,
 	dryRun bool,
 ) error {
 	if dryRun {
 		return nil
 	}
-	if err := repl.ChangeReplicas(ctx, roachpb.REMOVE_REPLICA, target, desc, reason, details); err != nil {
+	if _, err := repl.ChangeReplicas(ctx, roachpb.REMOVE_REPLICA, target, desc, reason, details); err != nil {
 		return err
 	}
 	rangeInfo := rangeInfoForRepl(repl, desc)
