@@ -26,6 +26,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/humanizeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
@@ -44,15 +45,19 @@ const (
 	TestTimeUntilStoreDeadOff = 24 * time.Hour
 )
 
-// declinedReservationsTimeout needs to be non-zero to prevent useless retries
-// in the replicateQueue.process() retry loop.
-var declinedReservationsTimeout = settings.RegisterNonNegativeDurationSetting(
+// DeclinedReservationsTimeout specifies a duration during which the local
+// replicate queue will not consider stores which have rejected a reservation a
+// viable target.
+var DeclinedReservationsTimeout = settings.RegisterNonNegativeDurationSetting(
 	"server.declined_reservation_timeout",
 	"the amount of time to consider the store throttled for up-replication after a reservation was declined",
 	1*time.Second,
 )
 
-var failedReservationsTimeout = settings.RegisterNonNegativeDurationSetting(
+// FailedReservationsTimeout specifies a duration during which the local
+// replicate queue will not consider stores which have failed a reservation a
+// viable target.
+var FailedReservationsTimeout = settings.RegisterNonNegativeDurationSetting(
 	"server.failed_reservation_timeout",
 	"the amount of time to consider the store throttled for up-replication after a failed reservation call",
 	5*time.Second,
@@ -86,16 +91,16 @@ type NodeCountFunc func() int
 
 // A NodeLivenessFunc accepts a node ID, current time and threshold before
 // a node is considered dead and returns whether or not the node is live.
-type NodeLivenessFunc func(roachpb.NodeID, time.Time, time.Duration) NodeLivenessStatus
+type NodeLivenessFunc func(roachpb.NodeID, time.Time, time.Duration) storagepb.NodeLivenessStatus
 
 // MakeStorePoolNodeLivenessFunc returns a function which determines
 // the status of a node based on information provided by the specified
 // NodeLiveness.
 func MakeStorePoolNodeLivenessFunc(nodeLiveness *NodeLiveness) NodeLivenessFunc {
-	return func(nodeID roachpb.NodeID, now time.Time, threshold time.Duration) NodeLivenessStatus {
+	return func(nodeID roachpb.NodeID, now time.Time, threshold time.Duration) storagepb.NodeLivenessStatus {
 		liveness, err := nodeLiveness.GetLiveness(nodeID)
 		if err != nil {
-			return NodeLivenessStatus_UNAVAILABLE
+			return storagepb.NodeLivenessStatus_UNAVAILABLE
 		}
 		return liveness.LivenessStatus(now, threshold, nodeLiveness.clock.MaxOffset())
 	}
@@ -106,10 +111,12 @@ type storeDetail struct {
 	// throttledUntil is when a throttled store can be considered available again
 	// due to a failed or declined snapshot.
 	throttledUntil time.Time
+	// throttledBecause is set to the most recent reason for which a store was
+	// marked as throttled.
+	throttledBecause string
 	// lastUpdatedTime is set when a store is first consulted and every time
 	// gossip arrives for a store.
 	lastUpdatedTime time.Time
-	deadReplicas    map[roachpb.RangeID][]roachpb.ReplicaDescriptor
 }
 
 // isThrottled returns whether the store is currently throttled.
@@ -164,20 +171,17 @@ func (sd *storeDetail) status(
 	// Even if the store has been updated via gossip, we still rely on
 	// the node liveness to determine whether it is considered live.
 	switch nl(sd.desc.Node.NodeID, now, threshold) {
-	case NodeLivenessStatus_DEAD, NodeLivenessStatus_DECOMMISSIONED:
+	case storagepb.NodeLivenessStatus_DEAD, storagepb.NodeLivenessStatus_DECOMMISSIONED:
 		return storeStatusDead
-	case NodeLivenessStatus_DECOMMISSIONING:
+	case storagepb.NodeLivenessStatus_DECOMMISSIONING:
 		return storeStatusDecommissioning
-	case NodeLivenessStatus_UNKNOWN, NodeLivenessStatus_UNAVAILABLE:
+	case storagepb.NodeLivenessStatus_UNKNOWN, storagepb.NodeLivenessStatus_UNAVAILABLE:
 		return storeStatusUnknown
 	}
 
 	if sd.isThrottled(now) {
 		return storeStatusThrottled
 	}
-	if len(sd.deadReplicas[rangeID]) > 0 {
-		return storeStatusReplicaCorrupted
-	}
 
 	return storeStatusAvailable
 }
@@ -247,8 +251,6 @@ func NewStorePool(
 	// hasn't otherwise changed.
 	storeRegex := gossip.MakePrefixPattern(gossip.KeyStorePrefix)
 	g.RegisterCallback(storeRegex, sp.storeGossipUpdate, gossip.Redundant)
-	deadReplicasRegex := gossip.MakePrefixPattern(gossip.KeyDeadReplicasPrefix)
-	g.RegisterCallback(deadReplicasRegex, sp.deadReplicasGossipUpdate)
 
 	return sp
 }
@@ -265,11 +267,12 @@ func (sp *StorePool) String() string {
 
 	var buf bytes.Buffer
 	now := sp.clock.PhysicalTime()
+	timeUntilStoreDead := TimeUntilStoreDead.Get(&sp.st.SV)
 
 	for _, id := range ids {
 		detail := sp.detailsMu.storeDetails[id]
 		fmt.Fprintf(&buf, "%d", id)
-		status := detail.status(now, TimeUntilStoreDead.Get(&sp.st.SV), 0, sp.nodeLivenessFn)
+		status := detail.status(now, timeUntilStoreDead, 0, sp.nodeLivenessFn)
 		if status != storeStatusAvailable {
 			fmt.Fprintf(&buf, " (status=%d)", status)
 		}
@@ -307,25 +310,6 @@ func (sp *StorePool) storeGossipUpdate(_ string, content roachpb.Value) {
 	sp.localitiesMu.Unlock()
 }
 
-// deadReplicasGossipUpdate is the gossip callback used to keep the StorePool up to date.
-func (sp *StorePool) deadReplicasGossipUpdate(_ string, content roachpb.Value) {
-	var replicas roachpb.StoreDeadReplicas
-	if err := content.GetProto(&replicas); err != nil {
-		ctx := sp.AnnotateCtx(context.TODO())
-		log.Error(ctx, err)
-		return
-	}
-
-	sp.detailsMu.Lock()
-	defer sp.detailsMu.Unlock()
-	detail := sp.getStoreDetailLocked(replicas.StoreID)
-	deadReplicas := make(map[roachpb.RangeID][]roachpb.ReplicaDescriptor)
-	for _, r := range replicas.Replicas {
-		deadReplicas[r.RangeID] = append(deadReplicas[r.RangeID], r.Replica)
-	}
-	detail.deadReplicas = deadReplicas
-}
-
 // updateLocalStoreAfterRebalance is used to update the local copy of the
 // target store immediately after a replica addition or removal.
 func (sp *StorePool) updateLocalStoreAfterRebalance(
@@ -391,9 +375,7 @@ func (sp *StorePool) updateLocalStoresAfterLeaseTransfer(
 // newStoreDetail makes a new storeDetail struct. It sets index to be -1 to
 // ensure that it will be processed by a queue immediately.
 func newStoreDetail() *storeDetail {
-	return &storeDetail{
-		deadReplicas: make(map[roachpb.RangeID][]roachpb.ReplicaDescriptor),
-	}
+	return &storeDetail{}
 }
 
 // getStoreDetailLocked returns the store detail for the given storeID.
@@ -435,9 +417,11 @@ func (sp *StorePool) decommissioningReplicas(
 	defer sp.detailsMu.Unlock()
 
 	now := sp.clock.PhysicalTime()
+	timeUntilStoreDead := TimeUntilStoreDead.Get(&sp.st.SV)
+
 	for _, repl := range repls {
 		detail := sp.getStoreDetailLocked(repl.StoreID)
-		switch detail.status(now, TimeUntilStoreDead.Get(&sp.st.SV), rangeID, sp.nodeLivenessFn) {
+		switch detail.status(now, timeUntilStoreDead, rangeID, sp.nodeLivenessFn) {
 		case storeStatusDecommissioning:
 			decommissioningReplicas = append(decommissioningReplicas, repl)
 		}
@@ -464,29 +448,15 @@ func (sp *StorePool) liveAndDeadReplicas(
 	defer sp.detailsMu.Unlock()
 
 	now := sp.clock.PhysicalTime()
+	timeUntilStoreDead := TimeUntilStoreDead.Get(&sp.st.SV)
+
 	for _, repl := range repls {
 		detail := sp.getStoreDetailLocked(repl.StoreID)
 		// Mark replica as dead if store is dead.
-		status := detail.status(now, TimeUntilStoreDead.Get(&sp.st.SV), rangeID, sp.nodeLivenessFn)
+		status := detail.status(now, timeUntilStoreDead, rangeID, sp.nodeLivenessFn)
 		switch status {
 		case storeStatusDead:
 			deadReplicas = append(deadReplicas, repl)
-		case storeStatusReplicaCorrupted:
-			// Check whether the replica we're examining has been marked as dead.
-			var corrupt bool
-			for _, deadRepl := range detail.deadReplicas[rangeID] {
-				if deadRepl.ReplicaID == repl.ReplicaID {
-					corrupt = true
-					break
-				}
-			}
-			// This replica is the corrupt replica, so consider the store dead.
-			if corrupt {
-				deadReplicas = append(deadReplicas, repl)
-			} else {
-				// Otherwise, consider the store live.
-				liveReplicas = append(liveReplicas, repl)
-			}
 		case storeStatusAvailable, storeStatusThrottled, storeStatusDecommissioning:
 			// We count both available and throttled stores to be live for the
 			// purpose of computing quorum.
@@ -608,6 +578,8 @@ const (
 	storeFilterThrottled
 )
 
+type throttledStoreReasons []string
+
 // getStoreList returns a storeList that contains all active stores that contain
 // the required attributes and their associated stats. The storeList is filtered
 // according to the provided storeFilter. It also returns the total number of
@@ -615,7 +587,7 @@ const (
 // corrupted replicas.
 func (sp *StorePool) getStoreList(
 	rangeID roachpb.RangeID, filter storeFilter,
-) (StoreList, int, int) {
+) (StoreList, int, throttledStoreReasons) {
 	sp.detailsMu.RLock()
 	defer sp.detailsMu.RUnlock()
 
@@ -630,7 +602,7 @@ func (sp *StorePool) getStoreList(
 // from the subset of passed in store IDs.
 func (sp *StorePool) getStoreListFromIDs(
 	storeIDs roachpb.StoreIDSlice, rangeID roachpb.RangeID, filter storeFilter,
-) (StoreList, int, int) {
+) (StoreList, int, throttledStoreReasons) {
 	sp.detailsMu.RLock()
 	defer sp.detailsMu.RUnlock()
 	return sp.getStoreListFromIDsRLocked(storeIDs, rangeID, filter)
@@ -640,7 +612,7 @@ func (sp *StorePool) getStoreListFromIDs(
 // that the detailsMU read lock is held.
 func (sp *StorePool) getStoreListFromIDsRLocked(
 	storeIDs roachpb.StoreIDSlice, rangeID roachpb.RangeID, filter storeFilter,
-) (StoreList, int, int) {
+) (StoreList, int, throttledStoreReasons) {
 	if sp.deterministic {
 		sort.Sort(storeIDs)
 	} else {
@@ -648,16 +620,18 @@ func (sp *StorePool) getStoreListFromIDsRLocked(
 	}
 
 	var aliveStoreCount int
-	var throttledStoreCount int
+	var throttled throttledStoreReasons
 	var storeDescriptors []roachpb.StoreDescriptor
 
 	now := sp.clock.PhysicalTime()
+	timeUntilStoreDead := TimeUntilStoreDead.Get(&sp.st.SV)
+
 	for _, storeID := range storeIDs {
 		detail := sp.detailsMu.storeDetails[storeID]
-		switch s := detail.status(now, TimeUntilStoreDead.Get(&sp.st.SV), rangeID, sp.nodeLivenessFn); s {
+		switch s := detail.status(now, timeUntilStoreDead, rangeID, sp.nodeLivenessFn); s {
 		case storeStatusThrottled:
 			aliveStoreCount++
-			throttledStoreCount++
+			throttled = append(throttled, detail.throttledBecause)
 			if filter != storeFilterThrottled {
 				storeDescriptors = append(storeDescriptors, *detail.desc)
 			}
@@ -672,7 +646,7 @@ func (sp *StorePool) getStoreListFromIDsRLocked(
 			panic(fmt.Sprintf("unknown store status: %d", s))
 		}
 	}
-	return makeStoreList(storeDescriptors), aliveStoreCount, throttledStoreCount
+	return makeStoreList(storeDescriptors), aliveStoreCount, throttled
 }
 
 type throttleReason int
@@ -688,10 +662,11 @@ const (
 // for up-replication or rebalancing until after the configured timeout period
 // has elapsed. Declined being true indicates that the remote store explicitly
 // declined a snapshot.
-func (sp *StorePool) throttle(reason throttleReason, storeID roachpb.StoreID) {
+func (sp *StorePool) throttle(reason throttleReason, why string, storeID roachpb.StoreID) {
 	sp.detailsMu.Lock()
 	defer sp.detailsMu.Unlock()
 	detail := sp.getStoreDetailLocked(storeID)
+	detail.throttledBecause = why
 
 	// If a snapshot is declined, be it due to an error or because it was
 	// rejected, we mark the store detail as having been declined so it won't
@@ -699,20 +674,20 @@ func (sp *StorePool) throttle(reason throttleReason, storeID roachpb.StoreID) {
 	// timeout period has passed.
 	switch reason {
 	case throttleDeclined:
-		timeout := declinedReservationsTimeout.Get(&sp.st.SV)
+		timeout := DeclinedReservationsTimeout.Get(&sp.st.SV)
 		detail.throttledUntil = sp.clock.PhysicalTime().Add(timeout)
 		if log.V(2) {
 			ctx := sp.AnnotateCtx(context.TODO())
-			log.Infof(ctx, "snapshot declined, s%d will be throttled for %s until %s",
-				storeID, timeout, detail.throttledUntil)
+			log.Infof(ctx, "snapshot declined (%s), s%d will be throttled for %s until %s",
+				why, storeID, timeout, detail.throttledUntil)
 		}
 	case throttleFailed:
-		timeout := failedReservationsTimeout.Get(&sp.st.SV)
+		timeout := FailedReservationsTimeout.Get(&sp.st.SV)
 		detail.throttledUntil = sp.clock.PhysicalTime().Add(timeout)
 		if log.V(2) {
 			ctx := sp.AnnotateCtx(context.TODO())
-			log.Infof(ctx, "snapshot failed, s%d will be throttled for %s until %s",
-				storeID, timeout, detail.throttledUntil)
+			log.Infof(ctx, "snapshot failed (%s), s%d will be throttled for %s until %s",
+				why, storeID, timeout, detail.throttledUntil)
 		}
 	}
 }
