@@ -21,10 +21,10 @@ import (
 	"sort"
 	"time"
 
-	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings"
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
+	"github.com/cockroachdb/cockroach/pkg/util/contextutil"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
@@ -191,7 +191,7 @@ func (sr *StoreRebalancer) Start(ctx context.Context, stopper *stop.Stopper) {
 				timer.Reset(jitteredInterval(storeRebalancerTimerDuration))
 			}
 
-			if !sr.st.Version.IsMinSupported(cluster.VersionLoadBasedRebalancing) {
+			if !sr.st.Version.IsActive(cluster.VersionLoadBasedRebalancing) {
 				continue
 			}
 
@@ -236,11 +236,6 @@ func (sr *StoreRebalancer) rebalanceStore(
 
 	var replicasToMaybeRebalance []replicaWithStats
 	storeMap := storeListToMap(storeList)
-	sysCfg, cfgOk := sr.rq.allocator.storePool.gossip.GetSystemConfig()
-	if !cfgOk {
-		log.VEventf(ctx, 1, "no system config available, unable to choose lease transfer targets")
-		return
-	}
 
 	log.Infof(ctx,
 		"considering load-based lease transfers for s%d with %.2f qps (mean=%.2f, upperThreshold=%.2f)",
@@ -249,7 +244,7 @@ func (sr *StoreRebalancer) rebalanceStore(
 	hottestRanges := sr.replRankings.topQPS()
 	for localDesc.Capacity.QueriesPerSecond > qpsMaxThreshold {
 		replWithStats, target, considerForRebalance := sr.chooseLeaseToTransfer(
-			ctx, sysCfg, &hottestRanges, localDesc, storeList, storeMap, qpsMinThreshold, qpsMaxThreshold)
+			ctx, &hottestRanges, localDesc, storeList, storeMap, qpsMinThreshold, qpsMaxThreshold)
 		replicasToMaybeRebalance = append(replicasToMaybeRebalance, considerForRebalance...)
 		if replWithStats.repl == nil {
 			break
@@ -257,15 +252,12 @@ func (sr *StoreRebalancer) rebalanceStore(
 
 		log.VEventf(ctx, 1, "transferring r%d (%.2f qps) to s%d to better balance load",
 			replWithStats.repl.RangeID, replWithStats.qps, target.StoreID)
-		replCtx, cancel := context.WithTimeout(replWithStats.repl.AnnotateCtx(ctx), sr.rq.processTimeout)
-		if err := sr.rq.transferLease(
-			replCtx, replWithStats.repl, target, replWithStats.qps,
-		); err != nil {
-			cancel()
-			log.Errorf(replCtx, "unable to transfer lease to s%d: %v", target.StoreID, err)
+		if err := contextutil.RunWithTimeout(ctx, "transfer lease", sr.rq.processTimeout, func(ctx context.Context) error {
+			return sr.rq.transferLease(ctx, replWithStats.repl, target, replWithStats.qps)
+		}); err != nil {
+			log.Errorf(ctx, "unable to transfer lease to s%d: %v", target.StoreID, err)
 			continue
 		}
-		cancel()
 		sr.metrics.LeaseTransferCount.Inc(1)
 
 		// Finally, update our local copies of the descriptors so that if
@@ -303,7 +295,6 @@ func (sr *StoreRebalancer) rebalanceStore(
 	for localDesc.Capacity.QueriesPerSecond > qpsMaxThreshold {
 		replWithStats, targets := sr.chooseReplicaToRebalance(
 			ctx,
-			sysCfg,
 			&replicasToMaybeRebalance,
 			localDesc,
 			storeList,
@@ -320,13 +311,12 @@ func (sr *StoreRebalancer) rebalanceStore(
 		descBeforeRebalance := replWithStats.repl.Desc()
 		log.VEventf(ctx, 1, "rebalancing r%d (%.2f qps) from %v to %v to better balance load",
 			replWithStats.repl.RangeID, replWithStats.qps, descBeforeRebalance.Replicas, targets)
-		replCtx, cancel := context.WithTimeout(replWithStats.repl.AnnotateCtx(ctx), sr.rq.processTimeout)
-		if err := sr.rq.store.AdminRelocateRange(replCtx, *descBeforeRebalance, targets); err != nil {
-			cancel()
-			log.Errorf(replCtx, "unable to relocate range to %v: %v", targets, err)
+		if err := contextutil.RunWithTimeout(ctx, "relocate range", sr.rq.processTimeout, func(ctx context.Context) error {
+			return sr.rq.store.AdminRelocateRange(ctx, *descBeforeRebalance, targets)
+		}); err != nil {
+			log.Errorf(ctx, "unable to relocate range to %v: %v", targets, err)
 			continue
 		}
-		cancel()
 		sr.metrics.RangeRebalanceCount.Inc(1)
 
 		// Finally, update our local copies of the descriptors so that if
@@ -363,7 +353,6 @@ func (sr *StoreRebalancer) rebalanceStore(
 // account here or just continue to let that happen in allocator.go?
 func (sr *StoreRebalancer) chooseLeaseToTransfer(
 	ctx context.Context,
-	sysCfg config.SystemConfig,
 	hottestRanges *[]replicaWithStats,
 	localDesc *roachpb.StoreDescriptor,
 	storeList StoreList,
@@ -401,7 +390,7 @@ func (sr *StoreRebalancer) chooseLeaseToTransfer(
 			continue
 		}
 
-		desc := replWithStats.repl.Desc()
+		desc, zone := replWithStats.repl.DescAndZone()
 		log.VEventf(ctx, 3, "considering lease transfer for r%d with %.2f qps",
 			desc.RangeID, replWithStats.qps)
 
@@ -435,20 +424,15 @@ func (sr *StoreRebalancer) chooseLeaseToTransfer(
 				raftStatus = sr.getRaftStatusFn(replWithStats.repl)
 			}
 			if replicaIsBehind(raftStatus, candidate.ReplicaID) {
-				log.VEventf(ctx, 3, "%v is behind or this store isn't the raft leader; raftStatus: %v",
-					candidate, raftStatus)
+				log.VEventf(ctx, 3, "%v is behind or this store isn't the raft leader for r%d; raftStatus: %v",
+					candidate, desc.RangeID, raftStatus)
 				continue
 			}
 
-			zone, err := sysCfg.GetZoneConfigForKey(desc.StartKey)
-			if err != nil {
-				log.Error(ctx, err)
-				return replicaWithStats{}, roachpb.ReplicaDescriptor{}, considerForRebalance
-			}
-
 			preferred := sr.rq.allocator.preferredLeaseholders(zone, desc.Replicas)
 			if len(preferred) > 0 && !storeHasReplica(candidate.StoreID, preferred) {
-				log.VEventf(ctx, 3, "s%d not a preferred leaseholder; preferred: %v", candidate.StoreID, preferred)
+				log.VEventf(ctx, 3, "s%d not a preferred leaseholder for r%d; preferred: %v",
+					candidate.StoreID, desc.RangeID, preferred)
 				continue
 			}
 
@@ -477,7 +461,6 @@ func (sr *StoreRebalancer) chooseLeaseToTransfer(
 
 func (sr *StoreRebalancer) chooseReplicaToRebalance(
 	ctx context.Context,
-	sysCfg config.SystemConfig,
 	hottestRanges *[]replicaWithStats,
 	localDesc *roachpb.StoreDescriptor,
 	storeList StoreList,
@@ -513,20 +496,12 @@ func (sr *StoreRebalancer) chooseReplicaToRebalance(
 			continue
 		}
 
-		desc := replWithStats.repl.Desc()
+		desc, zone := replWithStats.repl.DescAndZone()
 		log.VEventf(ctx, 3, "considering replica rebalance for r%d with %.2f qps",
 			desc.RangeID, replWithStats.qps)
 
-		// Pick out the stores that we want the range on, keeping existing replicas
-		// around if they aren't on overfull stores.
-		zone, err := sysCfg.GetZoneConfigForKey(desc.StartKey)
-		if err != nil {
-			log.Error(ctx, err)
-			return replicaWithStats{}, nil
-		}
 		clusterNodes := sr.rq.allocator.storePool.ClusterNodeCount()
-		desiredReplicas := GetNeededReplicas(zone.NumReplicas, clusterNodes)
-
+		desiredReplicas := GetNeededReplicas(*zone.NumReplicas, clusterNodes)
 		targets := make([]roachpb.ReplicationTarget, 0, desiredReplicas)
 		targetReplicas := make([]roachpb.ReplicaDescriptor, 0, desiredReplicas)
 
