@@ -21,10 +21,6 @@ import (
 	"math"
 	"time"
 
-	"github.com/pkg/errors"
-	"go.etcd.io/etcd/raft/raftpb"
-	"golang.org/x/time/rate"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/settings"
@@ -36,6 +32,9 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/protoutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/pkg/errors"
+	"go.etcd.io/etcd/raft/raftpb"
+	"golang.org/x/time/rate"
 )
 
 const (
@@ -141,11 +140,12 @@ func (kvSS *kvBatchSnapshotStrategy) Receive(
 			}
 
 			inSnap := IncomingSnapshot{
-				SnapUUID:   snapUUID,
-				Batches:    batches,
-				LogEntries: logEntries,
-				State:      &header.State,
-				snapType:   snapTypeRaft,
+				UsesUnreplicatedTruncatedState: header.UnreplicatedTruncatedState,
+				SnapUUID:                       snapUUID,
+				Batches:                        batches,
+				LogEntries:                     logEntries,
+				State:                          &header.State,
+				snapType:                       snapTypeRaft,
 			}
 			if header.RaftMessageRequest.ToReplica.ReplicaID == 0 {
 				inSnap.snapType = snapTypePreemptive
@@ -186,7 +186,7 @@ func (kvSS *kvBatchSnapshotStrategy) Send(
 			return err
 		}
 
-		if int64(len(b.Repr())) >= kvSS.batchSize {
+		if int64(b.Len()) >= kvSS.batchSize {
 			if err := kvSS.limiter.WaitN(ctx, 1); err != nil {
 				return err
 			}
@@ -281,7 +281,7 @@ func (kvSS *kvBatchSnapshotStrategy) Send(
 			if !sniffSideloadedRaftCommand(ent.Data) {
 				continue
 			}
-			if err := snap.WithSideloaded(func(ss sideloadStorage) error {
+			if err := snap.WithSideloaded(func(ss SideloadStorage) error {
 				newEnt, err := maybeInlineSideloadedRaftCommand(
 					ctx, rangeID, ent, ss, snap.RaftEntryCache,
 				)
@@ -405,37 +405,145 @@ func (s *Store) reserveSnapshot(
 // this store's replica (i.e. the snapshot is not from an older incarnation of
 // the replica) and a placeholder can be added to the replicasByKey map (if
 // necessary). If a placeholder is required, it is returned as the first value.
+// The authoritative bool determines whether the check is carried out with the
+// intention of actually applying the snapshot (in which case an existing replica
+// must exist and have its raftMu locked) or as a preliminary check.
 func (s *Store) canApplySnapshot(
-	ctx context.Context, snapHeader *SnapshotRequest_Header,
+	ctx context.Context, snapHeader *SnapshotRequest_Header, authoritative bool,
 ) (*ReplicaPlaceholder, error) {
 	s.mu.Lock()
 	defer s.mu.Unlock()
-	return s.canApplySnapshotLocked(ctx, snapHeader)
+	return s.canApplySnapshotLocked(ctx, snapHeader, authoritative)
 }
 
 func (s *Store) canApplySnapshotLocked(
-	ctx context.Context, snapHeader *SnapshotRequest_Header,
+	ctx context.Context, snapHeader *SnapshotRequest_Header, authoritative bool,
 ) (*ReplicaPlaceholder, error) {
 	desc := *snapHeader.State.Desc
-	if v, ok := s.mu.replicas.Load(int64(desc.RangeID)); ok && (*Replica)(v).IsInitialized() {
-		// We have an initialized replica. Preemptive snapshots can be applied with
-		// no further checks if they do not widen the existing replica. Raft
-		// snapshots can be applied with no further checks even if they widen the
-		// existing replica—we can't reject them at this point—but see the comments
-		// in Replica.maybeAcquireSnapshotMergeLock for how this is made safe.
-		existingDesc := (*Replica)(v).Desc()
-		if !snapHeader.IsPreemptive() || !existingDesc.EndKey.Less(desc.EndKey) {
-			return nil, nil
+
+	// First, check for an existing Replica.
+	//
+	// We call canApplySnapshotLocked twice for each snapshot application. In
+	// the first case, it's an optimization early before having received any
+	// data (and we don't use the placeholder if one is returned), and the
+	// replica may or may not be present.
+	//
+	// The second call happens right before we actually plan to apply the
+	// snapshot (and a Replica is always in place at that point). This means
+	// that without a Replica, we can have false positives, but if we have a
+	// replica it needs to take everything into account.
+	//
+	// TODO(tbg): untangle these two use cases.
+	if v, ok := s.mu.replicas.Load(
+		int64(desc.RangeID),
+	); !ok {
+		if authoritative {
+			return nil, errors.Errorf("authoritative call requires a replica present")
+		}
+	} else {
+		existingRepl := (*Replica)(v)
+		// The raftMu is held which allows us to use the existing replica as a
+		// placeholder when we decide that the snapshot can be applied. As long
+		// as the caller releases the raftMu only after feeding the snapshot
+		// into the replica, this is safe.
+		if authoritative {
+			existingRepl.raftMu.AssertHeld()
 		}
 
-		// We have a preemptive snapshot that widens an existing replica. Proceed
-		// by checking the keyspace covered by the snapshot but not the existing
-		// replica.
-		desc.StartKey = existingDesc.EndKey
+		existingRepl.mu.RLock()
+		existingDesc := existingRepl.descRLocked()
+		existingIsInitialized := existingRepl.isInitializedRLocked()
+		existingIsPreemptive := existingRepl.mu.replicaID == 0
+		existingRepl.mu.RUnlock()
+
+		if existingIsInitialized {
+			if !snapHeader.IsPreemptive() {
+				// Regular Raft snapshots can't be refused at this point,
+				// even if they widen the existing replica. See the comments
+				// in Replica.maybeAcquireSnapshotMergeLock for how this is
+				// made safe.
+				//
+				// NB: we expect the replica to know its replicaID at this point
+				// (i.e. !existingIsPreemptive), though perhaps it's possible
+				// that this isn't true if the leader initiates a Raft snapshot
+				// (that would provide a range descriptor with this replica in
+				// it) but this node reboots (temporarily forgetting its
+				// replicaID) before the snapshot arrives.
+				return nil, nil
+			}
+
+			if existingIsPreemptive {
+				// Allow applying a preemptive snapshot on top of another
+				// preemptive snapshot. We only need to acquire a placeholder
+				// for the part (if any) of the new snapshot that extends past
+				// the old one. If there's no such overlap, return early; if
+				// there is, "forward" the descriptor's StartKey so that the
+				// later code will only check the overlap.
+				//
+				// NB: morally it would be cleaner to ask for the existing
+				// replica to be GC'ed first, but consider that the preemptive
+				// snapshot was likely left behind by a failed attempt to
+				// up-replicate. This is a relatively common scenario and not
+				// worth discarding and resending another snapshot for. Let the
+				// snapshot through, which means "pretending that it doesn't
+				// intersect the existing replica".
+				if !existingDesc.EndKey.Less(desc.EndKey) {
+					return nil, nil
+				}
+				desc.StartKey = existingDesc.EndKey
+			}
+			// NB: If the existing snapshot is *not* preemptive (i.e. the above
+			// branch wasn't taken), the overlap check below will hit an error.
+			// This path is hit after a rapid up-down-upreplication to the same
+			// store and will resolve as the replicaGCQueue removes the existing
+			// replica. We are pretty sure that the existing replica is gc'able,
+			// because a preemptive snapshot implies that someone is trying to
+			// add this replica to the group at the moment. (We are not however,
+			// sure enough that this couldn't happen by accident to GC the
+			// replica ourselves - the replica GC queue will perform the proper
+			// check).
+		} else if snapHeader.IsPreemptive() {
+			// Morally, the existing replica now has a nonzero replica ID
+			// because we already know that it is not initialized (i.e. has no
+			// data). Interestingly, the case in which it has a zero replica ID
+			// is also possible and should see the snapshot accepted as it
+			// occurs when a preemptive snapshot is handled: we first create a
+			// Replica in this state, run this check, and then apply the
+			// preemptive snapshot.
+			if !existingIsPreemptive {
+				// This is similar to the case of a preemptive snapshot hitting
+				// a fully initialized replica (i.e. not a preemptive snapshot)
+				// at the end of the last branch (which we don't allow), so we
+				// want to reject the snapshot. There is a tricky problem to
+				// to solve here, though: existingRepl doesn't know anything
+				// about its key bounds, and so to check whether it is actually
+				// gc'able would require a full scan of the meta2 entries (and
+				// we would also need to teach the queues how to deal with un-
+				// initialized replicas).
+				//
+				// So we let the snapshot through (by falling through to the
+				// overlap check, where it either picks up placeholders or
+				// fails). This is safe (or at least we assume so) because we
+				// carry out all snapshot decisions through Raft (though it
+				// still is an odd path that we would be wise to avoid if it
+				// weren't so difficult).
+				//
+				// A consequence of letting this snapshot through is opening this
+				// replica up to the possibility of erroneous replicaGC. This is
+				// because it will retain the replicaID of the current replica,
+				// which is going to be initialized after the snapshot (and thus
+				// gc'able).
+				_ = 0 // avoid staticcheck failure
+			}
+		}
 	}
 
-	// We don't have the range, or we have an uninitialized placeholder, or the
-	// existing range is less wide. Will we be able to create/initialize it?
+	// We have a key range [desc.StartKey,desc.EndKey) which we want to apply a
+	// snapshot for. Is there a conflicting existing placeholder or an
+	// overlapping range?
+
+	// NB: this check seems redundant since placeholders are also represented in
+	// replicasByKey (and thus returned in getOverlappingKeyRangeLocked).
 	if exRng, ok := s.mu.replicaPlaceholders[desc.RangeID]; ok {
 		return nil, errors.Errorf("%s: canApplySnapshotLocked: cannot add placeholder, have an existing placeholder %s", s, exRng)
 	}
@@ -481,13 +589,18 @@ func (s *Store) canApplySnapshotLocked(
 			if inactive(exReplica) {
 				gcPriority = replicaGCPriorityCandidate
 			}
-			if _, err := s.replicaGCQueue.Add(exReplica, gcPriority); err != nil {
-				log.Errorf(ctx, "%s: unable to add replica to GC queue: %s", exReplica, err)
-			} else {
-				msg += "; initiated GC:"
-			}
+
+			msg += "; initiated GC:"
+			_ = s.stopper.RunAsyncTask(ctx, "add-to-replicagc-queue", func(ctx context.Context) {
+				// We can't hold any mutexes while adding to queues, so we have to do it in a task. This
+				// is especially important since we hold the store mutex in the surrounding method, which
+				// we wouldn't want to block even if it were legal.
+				if _, err := s.replicaGCQueue.Add(exReplica, gcPriority); err != nil {
+					log.Errorf(ctx, "%s: unable to add replica to GC queue: %s", exReplica, err)
+				}
+			})
 		}
-		return nil, errors.Errorf("%s %v", msg, exReplica) // exReplica can be nil
+		return nil, errors.Errorf("%s %v (incoming %v)", msg, exReplica, snapHeader.State.Desc.RSpan()) // exReplica can be nil
 	}
 
 	placeholder := &ReplicaPlaceholder{
@@ -517,7 +630,7 @@ func (s *Store) receiveSnapshot(
 	// We'll perform this check again later after receiving the rest of the
 	// snapshot data - this is purely an optimization to prevent downloading
 	// a snapshot that we know we won't be able to apply.
-	if _, err := s.canApplySnapshot(ctx, header); err != nil {
+	if _, err := s.canApplySnapshot(ctx, header, false /* authoritative */); err != nil {
 		return sendSnapshotError(stream,
 			errors.Wrapf(err, "%s,r%d: cannot apply snapshot", s, header.State.Desc.RangeID),
 		)
@@ -566,14 +679,23 @@ func sendSnapshotError(stream incomingSnapshotStream, err error) error {
 
 // SnapshotStorePool narrows StorePool to make sendSnapshot easier to test.
 type SnapshotStorePool interface {
-	throttle(reason throttleReason, toStoreID roachpb.StoreID)
+	throttle(reason throttleReason, why string, toStoreID roachpb.StoreID)
 }
 
+// rebalanceSnapshotRate is the rate at which preemptive snapshots can be sent.
+// This includes snapshots generated for upreplication or for rebalancing.
 var rebalanceSnapshotRate = settings.RegisterByteSizeSetting(
 	"kv.snapshot_rebalance.max_rate",
-	"the rate limit (bytes/sec) to use for rebalance snapshots",
-	envutil.EnvOrDefaultBytes("COCKROACH_PREEMPTIVE_SNAPSHOT_RATE", 2<<20),
+	"the rate limit (bytes/sec) to use for rebalance and upreplication snapshots",
+	envutil.EnvOrDefaultBytes("COCKROACH_PREEMPTIVE_SNAPSHOT_RATE", 8<<20),
 )
+
+// recoverySnapshotRate is the rate at which Raft-initiated spanshots can be
+// sent. Ideally, one would never see a Raft-initiated snapshot; we'd like all
+// the snapshots to be preemptive. However, it has proved unfeasible to
+// completely get rid of them.
+// TODO(tbg): The existence of this rate, separate from rebalanceSnapshotRate,
+// does not make a whole lot of sense.
 var recoverySnapshotRate = settings.RegisterByteSizeSetting(
 	"kv.snapshot_recovery.max_rate",
 	"the rate limit (bytes/sec) to use for recovery snapshots",
@@ -624,32 +746,35 @@ func sendSnapshot(
 	// Wait until we get a response from the server.
 	resp, err := stream.Recv()
 	if err != nil {
-		storePool.throttle(throttleFailed, to.StoreID)
+		storePool.throttle(throttleFailed, err.Error(), to.StoreID)
 		return err
 	}
 	switch resp.Status {
 	case SnapshotResponse_DECLINED:
 		if header.CanDecline {
-			storePool.throttle(throttleDeclined, to.StoreID)
 			declinedMsg := "reservation rejected"
 			if len(resp.Message) > 0 {
 				declinedMsg = resp.Message
 			}
-			return &benignError{errors.Errorf("%s: remote declined %s: %s", to, snap, declinedMsg)}
+			err := &benignError{errors.Errorf("%s: remote declined %s: %s", to, snap, declinedMsg)}
+			storePool.throttle(throttleDeclined, err.Error(), to.StoreID)
+			return err
 		}
-		storePool.throttle(throttleFailed, to.StoreID)
-		return errors.Errorf("%s: programming error: remote declined required %s: %s",
+		err := errors.Errorf("%s: programming error: remote declined required %s: %s",
 			to, snap, resp.Message)
+		storePool.throttle(throttleFailed, err.Error(), to.StoreID)
+		return err
 	case SnapshotResponse_ERROR:
-		storePool.throttle(throttleFailed, to.StoreID)
+		storePool.throttle(throttleFailed, resp.Message, to.StoreID)
 		return errors.Errorf("%s: remote couldn't accept %s with error: %s",
 			to, snap, resp.Message)
 	case SnapshotResponse_ACCEPTED:
 	// This is the response we're expecting. Continue with snapshot sending.
 	default:
-		storePool.throttle(throttleFailed, to.StoreID)
-		return errors.Errorf("%s: server sent an invalid status while negotiating %s: %s",
+		err := errors.Errorf("%s: server sent an invalid status while negotiating %s: %s",
 			to, snap, resp.Status)
+		storePool.throttle(throttleFailed, err.Error(), to.StoreID)
+		return err
 	}
 
 	log.Infof(ctx, "sending %s", snap)
