@@ -18,6 +18,7 @@ package tscache
 import (
 	"bytes"
 	"container/list"
+	"context"
 	"encoding/binary"
 	"fmt"
 	"sync/atomic"
@@ -25,10 +26,9 @@ import (
 	"unsafe"
 
 	"github.com/andy-kimball/arenaskl"
-
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
-	"github.com/cockroachdb/cockroach/pkg/util/interval"
+	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
 )
@@ -239,8 +239,19 @@ func (s *intervalSkl) AddRange(from, to []byte, opt rangeOptions, val cacheValue
 
 		switch {
 		case cmp > 0:
-			// Starting key is after ending key. This shouldn't happen.
-			panic(interval.ErrInvertedRange)
+			// Starting key is after ending key. This shouldn't happen. Determine
+			// the index where the keys diverged and panic.
+			d := 0
+			for d < len(from) && d < len(to) {
+				if from[d] != to[d] {
+					break
+				}
+				d++
+			}
+			msg := fmt.Sprintf("inverted range (issue #32149): key lens = [%d,%d), diff @ index %d",
+				len(from), len(to), d)
+			log.Errorf(context.Background(), "%s, [%s,%s)", msg, from, to)
+			panic(log.Safe(msg))
 		case cmp == 0:
 			// Starting key is same as ending key, so just add single node.
 			if opt == (excludeFrom | excludeTo) {
@@ -363,7 +374,7 @@ func (s *intervalSkl) frontPage() *sklPage {
 // pushNewPage prepends a new empty page to the front of the pages list. It
 // accepts an optional arena argument to facilitate re-use.
 func (s *intervalSkl) pushNewPage(maxWallTime int64, arena *arenaskl.Arena) {
-	if arena != nil && arena.Size() == s.pageSize {
+	if arena != nil {
 		// Re-use the provided arena, if possible.
 		arena.Reset()
 	} else {
@@ -826,6 +837,18 @@ func (p *sklPage) ratchetValueSet(
 			// attempt then we must have raced with node initialization before.
 			return nil
 		}
+		if (meta & cantInit) != 0 {
+			// If the meta has the cantInit flag set to true, we fail with an
+			// ErrArenaFull error to force the current goroutine to retry on a
+			// new page.
+			return arenaskl.ErrArenaFull
+		}
+
+		newMeta := meta
+		updateInit := setInit && !inited
+		if updateInit {
+			newMeta |= initialized
+		}
 
 		var keyValUpdate, gapValUpdate bool
 		oldKeyVal, oldGapVal := decodeValueSet(it.Value(), meta)
@@ -833,19 +856,6 @@ func (p *sklPage) ratchetValueSet(
 		gapVal, gapValUpdate = ratchetValue(oldGapVal, gapVal)
 		updateVals := keyValUpdate || gapValUpdate
 
-		newMeta := meta
-		updateInit := setInit && !inited
-		if updateInit {
-			// If the meta has the cantInit flag set to true, we fail with an
-			// ErrArenaFull error to force the current goroutine to retry on a
-			// new page.
-			if (meta & cantInit) != 0 {
-				return arenaskl.ErrArenaFull
-			}
-
-			newMeta |= initialized
-		}
-
 		if updateVals {
 			// If we're updating the values (and maybe the init flag) then we
 			// need to call it.Set. This can return an ErrArenaFull, which we
