@@ -25,14 +25,13 @@ import (
 	"testing"
 	"time"
 
-	"github.com/stretchr/testify/require"
-
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
 	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
+	"github.com/stretchr/testify/require"
 )
 
 var (
@@ -1046,7 +1045,8 @@ func TestIntervalSklConcurrentVsSequential(t *testing.T) {
 			rounds /= 2
 		}
 		for j := 0; j < rounds; j++ {
-			t.Logf("round %d", j)
+			// This is a lot of log output so only un-comment to debug.
+			// t.Logf("round %d", j)
 
 			// Create a set of actions to perform.
 			type action struct {
@@ -1066,18 +1066,19 @@ func TestIntervalSklConcurrentVsSequential(t *testing.T) {
 				}
 				a.val = cacheValue{ts: ts, txnID: txnIDs[i]}
 
-				t.Logf("action (%s,%s)[%d] = %s", string(a.from), string(a.to), a.opt, a.val)
+				// This is a lot of log output so only un-comment to debug.
+				// t.Logf("action (%s,%s)[%d] = %s", string(a.from), string(a.to), a.opt, a.val)
 				actions[i] = a
 			}
 
 			// Perform each action, first in order on the "sequential"
 			// intervalSkl, then in parallel on the "concurrent" intervalSkl.
-			t.Log("sequential actions")
+			// t.Log("sequential actions")
 			for _, a := range actions {
 				sequentialS.AddRange(a.from, a.to, a.opt, a.val)
 			}
 
-			t.Log("concurrent actions")
+			// t.Log("concurrent actions")
 			var wg sync.WaitGroup
 			for _, a := range actions {
 				wg.Add(1)
@@ -1217,10 +1218,10 @@ func BenchmarkIntervalSklAddAndLookup(b *testing.B) {
 					rng := rand.New(rand.NewSource(timeutil.Now().UnixNano()))
 
 					for n := 0; n < b.N/parallel; n++ {
-						readFrac := rng.Int31()
+						readFrac := rng.Int31n(10)
 						keyNum := rng.Int31n(max)
 
-						if (readFrac % 10) < int32(i) {
+						if readFrac < int32(i) {
 							key := []byte(fmt.Sprintf("%020d", keyNum))
 							s.LookupTimestamp(key)
 						} else {
