@@ -26,7 +26,6 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/keys"
 	"github.com/cockroachdb/cockroach/pkg/roachpb"
 	"github.com/cockroachdb/cockroach/pkg/storage/engine"
-	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
 	"github.com/cockroachdb/cockroach/pkg/storage/storagebase"
 	"github.com/cockroachdb/cockroach/pkg/storage/txnwait"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
@@ -35,6 +34,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/uuid"
 	"github.com/pkg/errors"
+	"github.com/stretchr/testify/assert"
 )
 
 func writeTxnRecord(ctx context.Context, tc *testContext, txn *roachpb.Transaction) error {
@@ -45,7 +45,7 @@ func writeTxnRecord(ctx context.Context, tc *testContext, txn *roachpb.Transacti
 // createTxnForPushQueue creates a txn struct and writes a "fake"
 // transaction record for it to the underlying engine.
 func createTxnForPushQueue(ctx context.Context, tc *testContext) (*roachpb.Transaction, error) {
-	txn := newTransaction("txn", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+	txn := newTransaction("txn", roachpb.Key("a"), 1, tc.Clock())
 	return txn, writeTxnRecord(ctx, tc, txn)
 }
 
@@ -54,6 +54,23 @@ type RespWithErr struct {
 	pErr *roachpb.Error
 }
 
+func checkAllGaugesZero(tc testContext) error {
+	m := tc.store.GetTxnWaitMetrics()
+	if act := m.PusheeWaiting.Value(); act != 0 {
+		return errors.Errorf("expected PusheeWaiting to be 0, got %d instead", act)
+	}
+	if act := m.PusherWaiting.Value(); act != 0 {
+		return errors.Errorf("expected PusherWaiting to be 0, got %d instead", act)
+	}
+	if act := m.QueryWaiting.Value(); act != 0 {
+		return errors.Errorf("expected QueryWaiting to be 0, got %d instead", act)
+	}
+	if act := m.PusherSlow.Value(); act != 0 {
+		return errors.Errorf("expected PusherSlow to be 0, got %d instead", act)
+	}
+	return nil
+}
+
 func TestTxnWaitQueueEnableDisable(t *testing.T) {
 	defer leaktest.AfterTest(t)()
 	tc := testContext{}
@@ -71,13 +88,18 @@ func TestTxnWaitQueueEnableDisable(t *testing.T) {
 	if !q.IsEnabled() {
 		t.Errorf("expected push txn queue is enabled")
 	}
+	if err := checkAllGaugesZero(tc); err != nil {
+		t.Fatal(err.Error())
+	}
 
 	q.Enqueue(txn)
 	if _, ok := q.TrackedTxns()[txn.ID]; !ok {
 		t.Fatalf("expected pendingTxn to be in txns map after enqueue")
 	}
+	m := tc.store.GetTxnWaitMetrics()
+	assert.EqualValues(tc, 1, m.PusheeWaiting.Value())
 
-	pusher := newTransaction("pusher", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+	pusher := newTransaction("pusher", roachpb.Key("a"), 1, tc.Clock())
 	req := roachpb.PushTxnRequest{
 		PushType:  roachpb.PUSH_ABORT,
 		PusherTxn: *pusher,
@@ -95,6 +117,13 @@ func TestTxnWaitQueueEnableDisable(t *testing.T) {
 		if deps := q.GetDependents(txn.ID); !reflect.DeepEqual(deps, expDeps) {
 			return errors.Errorf("expected GetDependents %+v; got %+v", expDeps, deps)
 		}
+		if act, exp := m.PusherWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushers, but want %d", act, exp)
+		}
+		if act, exp := m.PusheeWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushees, but want %d", act, exp)
+		}
+
 		return nil
 	})
 
@@ -103,6 +132,9 @@ func TestTxnWaitQueueEnableDisable(t *testing.T) {
 	if q.IsEnabled() {
 		t.Errorf("expected queue to be disabled")
 	}
+	if err := checkAllGaugesZero(tc); err != nil {
+		t.Fatal(err.Error())
+	}
 
 	respWithErr := <-retCh
 	if respWithErr.resp != nil {
@@ -120,6 +152,9 @@ func TestTxnWaitQueueEnableDisable(t *testing.T) {
 	if q.IsEnabled() {
 		t.Errorf("expected enqueue to silently fail since queue is disabled")
 	}
+	if err := checkAllGaugesZero(tc); err != nil {
+		t.Fatal(err.Error())
+	}
 
 	q.UpdateTxn(context.Background(), txn)
 	if len(q.TrackedTxns()) != 0 {
@@ -129,6 +164,9 @@ func TestTxnWaitQueueEnableDisable(t *testing.T) {
 	if resp, pErr := q.MaybeWaitForPush(context.TODO(), tc.repl, &req); resp != nil || pErr != nil {
 		t.Errorf("expected nil resp and err as queue is disabled; got %+v, %s", resp, pErr)
 	}
+	if err := checkAllGaugesZero(tc); err != nil {
+		t.Fatal(err.Error())
+	}
 }
 
 func TestTxnWaitQueueCancel(t *testing.T) {
@@ -142,7 +180,7 @@ func TestTxnWaitQueueCancel(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	pusher := newTransaction("pusher", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+	pusher := newTransaction("pusher", roachpb.Key("a"), 1, tc.Clock())
 	req := roachpb.PushTxnRequest{
 		PushType:  roachpb.PUSH_ABORT,
 		PusherTxn: *pusher,
@@ -151,7 +189,13 @@ func TestTxnWaitQueueCancel(t *testing.T) {
 
 	q := tc.repl.txnWaitQueue
 	q.Enable()
+	if err := checkAllGaugesZero(tc); err != nil {
+		t.Fatal(err.Error())
+	}
 	q.Enqueue(txn)
+	m := tc.store.GetTxnWaitMetrics()
+	assert.EqualValues(tc, 1, m.PusheeWaiting.Value())
+	assert.EqualValues(tc, 0, m.PusherWaiting.Value())
 
 	ctx, cancel := context.WithCancel(context.Background())
 	retCh := make(chan RespWithErr, 1)
@@ -170,6 +214,9 @@ func TestTxnWaitQueueCancel(t *testing.T) {
 		if deps := q.GetDependents(txn.ID); !reflect.DeepEqual(deps, expDeps) {
 			return errors.Errorf("expected GetDependents %+v; got %+v", expDeps, deps)
 		}
+		if act, exp := m.PusherWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushers, but want %d", act, exp)
+		}
 		return nil
 	})
 	cancel()
@@ -196,8 +243,8 @@ func TestTxnWaitQueueUpdateTxn(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	pusher1 := newTransaction("pusher1", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
-	pusher2 := newTransaction("pusher2", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+	pusher1 := newTransaction("pusher1", roachpb.Key("a"), 1, tc.Clock())
+	pusher2 := newTransaction("pusher2", roachpb.Key("a"), 1, tc.Clock())
 	req1 := roachpb.PushTxnRequest{
 		PushType:  roachpb.PUSH_ABORT,
 		PusherTxn: *pusher1,
@@ -209,6 +256,8 @@ func TestTxnWaitQueueUpdateTxn(t *testing.T) {
 	q := tc.repl.txnWaitQueue
 	q.Enable()
 	q.Enqueue(txn)
+	m := tc.store.GetTxnWaitMetrics()
+	assert.EqualValues(tc, 1, m.PusheeWaiting.Value())
 
 	retCh := make(chan RespWithErr, 2)
 	go func() {
@@ -222,6 +271,18 @@ func TestTxnWaitQueueUpdateTxn(t *testing.T) {
 		}
 		return nil
 	})
+	testutils.SucceedsSoon(t, func() error {
+		if act, exp := m.PusherWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushers, but want %d", act, exp)
+		}
+		if act, exp := m.PusheeWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushees, but want %d", act, exp)
+		}
+		if act, exp := m.QueryWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d queries, but want %d", act, exp)
+		}
+		return nil
+	})
 
 	go func() {
 		resp, pErr := q.MaybeWaitForPush(context.Background(), tc.repl, &req2)
@@ -232,12 +293,24 @@ func TestTxnWaitQueueUpdateTxn(t *testing.T) {
 		if deps := q.GetDependents(txn.ID); !reflect.DeepEqual(deps, expDeps) {
 			return errors.Errorf("expected GetDependents %+v; got %+v", expDeps, deps)
 		}
+		if act, exp := m.PusherWaiting.Value(), int64(2); act != exp {
+			return errors.Errorf("%d pushers, but want %d", act, exp)
+		}
+		if act, exp := m.PusheeWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushees, but want %d", act, exp)
+		}
+		if act, exp := m.QueryWaiting.Value(), int64(2); act != exp {
+			return errors.Errorf("%d queries, but want %d", act, exp)
+		}
 		return nil
 	})
 
 	updatedTxn := *txn
 	updatedTxn.Status = roachpb.COMMITTED
 	q.UpdateTxn(context.Background(), &updatedTxn)
+	testutils.SucceedsSoon(tc.TB, func() error {
+		return checkAllGaugesZero(tc)
+	})
 
 	for i := 0; i < 2; i++ {
 		respWithErr := <-retCh
@@ -281,7 +354,7 @@ func TestTxnWaitQueueTxnSilentlyCompletes(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	pusher := newTransaction("pusher", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+	pusher := newTransaction("pusher", roachpb.Key("a"), 1, tc.Clock())
 	req := &roachpb.PushTxnRequest{
 		PushType:  roachpb.PUSH_ABORT,
 		PusherTxn: *pusher,
@@ -297,11 +370,22 @@ func TestTxnWaitQueueTxnSilentlyCompletes(t *testing.T) {
 		resp, pErr := q.MaybeWaitForPush(context.Background(), tc.repl, req)
 		retCh <- RespWithErr{resp, pErr}
 	}()
+
+	m := tc.store.GetTxnWaitMetrics()
 	testutils.SucceedsSoon(t, func() error {
 		expDeps := []uuid.UUID{pusher.ID}
 		if deps := q.GetDependents(txn.ID); !reflect.DeepEqual(deps, expDeps) {
 			return errors.Errorf("expected GetDependents %+v; got %+v", expDeps, deps)
 		}
+		if act, exp := m.PusherWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushers, but want %d", act, exp)
+		}
+		if act, exp := m.PusheeWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushees, but want %d", act, exp)
+		}
+		if act, exp := m.QueryWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d queries, but want %d", act, exp)
+		}
 		return nil
 	})
 
@@ -317,6 +401,18 @@ func TestTxnWaitQueueTxnSilentlyCompletes(t *testing.T) {
 	if respWithErr.resp == nil || respWithErr.resp.PusheeTxn.Status != roachpb.COMMITTED {
 		t.Errorf("expected committed txn response; got %+v, err=%v", respWithErr.resp, respWithErr.pErr)
 	}
+	testutils.SucceedsSoon(t, func() error {
+		if act, exp := m.PusherWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushers, but want %d", act, exp)
+		}
+		if act, exp := m.PusheeWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushees, but want %d", act, exp)
+		}
+		if act, exp := m.QueryWaiting.Value(), int64(0); act != exp {
+			return errors.Errorf("%d queries, but want %d", act, exp)
+		}
+		return nil
+	})
 }
 
 // TestTxnWaitQueueUpdateNotPushedTxn verifies that no PushTxnResponse
@@ -333,7 +429,7 @@ func TestTxnWaitQueueUpdateNotPushedTxn(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	pusher := newTransaction("pusher", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+	pusher := newTransaction("pusher", roachpb.Key("a"), 1, tc.Clock())
 	req := roachpb.PushTxnRequest{
 		PushType:  roachpb.PUSH_ABORT,
 		PusherTxn: *pusher,
@@ -369,6 +465,9 @@ func TestTxnWaitQueueUpdateNotPushedTxn(t *testing.T) {
 	if respWithErr.pErr != nil {
 		t.Errorf("expected nil error; got %s", respWithErr.pErr)
 	}
+	testutils.SucceedsSoon(tc.TB, func() error {
+		return checkAllGaugesZero(tc)
+	})
 }
 
 // TestTxnWaitQueuePusheeExpires verifies that just one pusher is
@@ -379,7 +478,7 @@ func TestTxnWaitQueuePusheeExpires(t *testing.T) {
 
 	manual := hlc.NewManualClock(123)
 	clock := hlc.NewClock(manual.UnixNano, time.Nanosecond)
-	txn := newTransaction("txn", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, clock)
+	txn := newTransaction("txn", roachpb.Key("a"), 1, clock)
 	// Move the clock forward so that when the PushTxn is sent, the txn appears
 	// expired.
 	manual.Set(txnwait.TxnExpiration(txn).WallTime)
@@ -397,8 +496,8 @@ func TestTxnWaitQueuePusheeExpires(t *testing.T) {
 	defer stopper.Stop(context.TODO())
 	tc.StartWithStoreConfig(t, stopper, tsc)
 
-	pusher1 := newTransaction("pusher1", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
-	pusher2 := newTransaction("pusher2", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+	pusher1 := newTransaction("pusher1", roachpb.Key("a"), 1, tc.Clock())
+	pusher2 := newTransaction("pusher2", roachpb.Key("a"), 1, tc.Clock())
 	req1 := roachpb.PushTxnRequest{
 		PushType:  roachpb.PUSH_ABORT,
 		PusherTxn: *pusher1,
@@ -451,6 +550,19 @@ func TestTxnWaitQueuePusheeExpires(t *testing.T) {
 		}
 	}
 
+	m := tc.store.GetTxnWaitMetrics()
+	testutils.SucceedsSoon(t, func() error {
+		if act, exp := m.PusherWaiting.Value(), int64(2); act != exp {
+			return errors.Errorf("%d pushers, but want %d", act, exp)
+		}
+		if act, exp := m.PusheeWaiting.Value(), int64(1); act != exp {
+			return errors.Errorf("%d pushees, but want %d", act, exp)
+		}
+		if act, exp := m.QueryWaiting.Value(), int64(0); act != exp {
+			return errors.Errorf("%d queries, but want %d", act, exp)
+		}
+		return nil
+	})
 	if a, minExpected := atomic.LoadInt32(&queryTxnCount), int32(2); a < minExpected {
 		t.Errorf("expected no fewer than %d query txns; got %d", minExpected, a)
 	}
@@ -478,7 +590,7 @@ func TestTxnWaitQueuePusherUpdate(t *testing.T) {
 				t.Fatal(err)
 			}
 		} else {
-			pusher = newTransaction("pusher", roachpb.Key("a"), 1, enginepb.SERIALIZABLE, tc.Clock())
+			pusher = newTransaction("pusher", roachpb.Key("a"), 1, tc.Clock())
 		}
 
 		req := roachpb.PushTxnRequest{
@@ -527,6 +639,20 @@ func TestTxnWaitQueuePusherUpdate(t *testing.T) {
 		if !testutils.IsPError(respWithErr.pErr, regexp.QuoteMeta(expErr)) {
 			t.Errorf("expected %s; got %v", expErr, respWithErr.pErr)
 		}
+
+		m := tc.store.GetTxnWaitMetrics()
+		testutils.SucceedsSoon(t, func() error {
+			if act, exp := m.PusherWaiting.Value(), int64(1); act != exp {
+				return errors.Errorf("%d pushers, but want %d", act, exp)
+			}
+			if act, exp := m.PusheeWaiting.Value(), int64(1); act != exp {
+				return errors.Errorf("%d pushees, but want %d", act, exp)
+			}
+			if act, exp := m.QueryWaiting.Value(), int64(0); act != exp {
+				return errors.Errorf("%d queries, but want %d", act, exp)
+			}
+			return nil
+		})
 	})
 }
 
@@ -576,6 +702,8 @@ func TestTxnWaitQueueDependencyCycle(t *testing.T) {
 	for _, txn := range []*roachpb.Transaction{txnA, txnB, txnC} {
 		q.Enqueue(txn)
 	}
+	m := tc.store.GetTxnWaitMetrics()
+	assert.EqualValues(tc, 0, m.DeadlocksTotal.Count())
 
 	retCh := make(chan RespWithErr, 3)
 	for _, req := range []*roachpb.PushTxnRequest{reqA, reqB, reqC} {
@@ -594,6 +722,13 @@ func TestTxnWaitQueueDependencyCycle(t *testing.T) {
 	if respWithErr.resp != nil {
 		t.Errorf("expected nil response; got %+v", respWithErr.resp)
 	}
+
+	testutils.SucceedsSoon(t, func() error {
+		if act, exp := m.DeadlocksTotal.Count(), int64(0); act <= exp {
+			return errors.Errorf("%d deadlocks, but want more than %d", act, exp)
+		}
+		return nil
+	})
 	cancel()
 	for i := 0; i < 2; i++ {
 		<-retCh
@@ -617,7 +752,7 @@ func TestTxnWaitQueueDependencyCycleWithPriorityInversion(t *testing.T) {
 
 	// Create txnA with a lower priority so it won't think it could push
 	// txnB without updating its priority.
-	txnA := newTransaction("txn", roachpb.Key("a"), -1, enginepb.SERIALIZABLE, tc.Clock())
+	txnA := newTransaction("txn", roachpb.Key("a"), -1, tc.Clock())
 	// However, write an "updated" txnA with higher priority, which it
 	// will need to read via a QueryTxn request in order to realize it
 	// can in fact break the deadlock.
@@ -629,7 +764,7 @@ func TestTxnWaitQueueDependencyCycleWithPriorityInversion(t *testing.T) {
 	// Create txnB with priority=2, so txnA won't think it can push, but
 	// when we set up txnB as the pusher, the request will include txnA's
 	// updated priority, making txnB think it can't break a deadlock.
-	txnB := newTransaction("txn", roachpb.Key("a"), -2, enginepb.SERIALIZABLE, tc.Clock())
+	txnB := newTransaction("txn", roachpb.Key("a"), -2, tc.Clock())
 	if err := writeTxnRecord(context.Background(), &tc, txnB); err != nil {
 		t.Fatal(err)
 	}
@@ -652,6 +787,8 @@ func TestTxnWaitQueueDependencyCycleWithPriorityInversion(t *testing.T) {
 	for _, txn := range []*roachpb.Transaction{txnA, txnB} {
 		q.Enqueue(txn)
 	}
+	m := tc.store.GetTxnWaitMetrics()
+	assert.EqualValues(tc, 0, m.DeadlocksTotal.Count())
 
 	retCh := make(chan ReqWithErr, 2)
 	for _, req := range []*roachpb.PushTxnRequest{reqA, reqB} {
@@ -671,6 +808,13 @@ func TestTxnWaitQueueDependencyCycleWithPriorityInversion(t *testing.T) {
 	if reqWithErr.pErr != txnwait.ErrDeadlock {
 		t.Errorf("expected errDeadlock; got %v", reqWithErr.pErr)
 	}
+
+	testutils.SucceedsSoon(t, func() error {
+		if act, exp := m.DeadlocksTotal.Count(), int64(0); act <= exp {
+			return errors.Errorf("%d deadlocks, but want more than %d", act, exp)
+		}
+		return nil
+	})
 	cancel()
 	<-retCh
 }
