@@ -60,7 +60,7 @@ func ShouldPushImmediately(req *roachpb.PushTxnRequest) bool {
 // for transactions with pushed timestamps.
 func isPushed(req *roachpb.PushTxnRequest, txn *roachpb.Transaction) bool {
 	return (txn.Status != roachpb.PENDING ||
-		(req.PushType == roachpb.PUSH_TIMESTAMP && req.PushTo.Less(txn.Timestamp)))
+		(req.PushType == roachpb.PUSH_TIMESTAMP && !txn.Timestamp.Less(req.PushTo)))
 }
 
 // TxnExpiration computes the timestamp after which the transaction will be
@@ -78,7 +78,7 @@ func IsExpired(now hlc.Timestamp, txn *roachpb.Transaction) bool {
 // copy of the supplied transaction. It is necessary to fully copy
 // each field in the transaction to avoid race conditions.
 func createPushTxnResponse(txn *roachpb.Transaction) *roachpb.PushTxnResponse {
-	return &roachpb.PushTxnResponse{PusheeTxn: txn.Clone()}
+	return &roachpb.PushTxnResponse{PusheeTxn: *txn}
 }
 
 // A waitingPush represents a PushTxn command that is waiting on the
@@ -136,6 +136,8 @@ type StoreInterface interface {
 	Clock() *hlc.Clock
 	Stopper() *stop.Stopper
 	DB() *client.DB
+	GetTxnWaitKnobs() TestingKnobs
+	GetTxnWaitMetrics() *Metrics
 }
 
 // ReplicaInterface provides some parts of a Replica without incurring a dependency.
@@ -143,6 +145,14 @@ type ReplicaInterface interface {
 	ContainsKey(roachpb.Key) bool
 }
 
+// TestingKnobs represents testing knobs for a Queue.
+type TestingKnobs struct {
+	// OnTxnWaitEnqueue is called when a would-be pusher joins a wait queue.
+	OnPusherBlocked func(ctx context.Context, push *roachpb.PushTxnRequest)
+	// OnTxnUpdate is called by Queue.UpdateTxn.
+	OnTxnUpdate func(ctx context.Context, txn *roachpb.Transaction)
+}
+
 // Queue enqueues PushTxn requests which are waiting on extant txns
 // with conflicting intents to abort or commit.
 //
@@ -207,6 +217,11 @@ func (q *Queue) Clear(disable bool) {
 		queryWaitersCount += waitingQueries.count
 	}
 
+	metrics := q.store.GetTxnWaitMetrics()
+	metrics.PusheeWaiting.Dec(int64(len(q.mu.txns)))
+	metrics.PusherWaiting.Dec(int64(len(pushWaiters)))
+	metrics.QueryWaiting.Dec(int64(queryWaitersCount))
+
 	if log.V(1) {
 		log.Infof(
 			context.Background(),
@@ -257,6 +272,7 @@ func (q *Queue) Enqueue(txn *roachpb.Transaction) {
 	if pt, ok := q.mu.txns[txn.ID]; ok {
 		pt.txn.Store(txn)
 	} else {
+		q.store.GetTxnWaitMetrics().PusheeWaiting.Inc(1)
 		pt = &pendingTxn{}
 		pt.txn.Store(txn)
 		q.mu.txns[txn.ID] = pt
@@ -269,6 +285,9 @@ func (q *Queue) Enqueue(txn *roachpb.Transaction) {
 func (q *Queue) UpdateTxn(ctx context.Context, txn *roachpb.Transaction) {
 	txn.AssertInitialized(ctx)
 	q.mu.Lock()
+	if f := q.store.GetTxnWaitKnobs().OnTxnUpdate; f != nil {
+		f(ctx, txn)
+	}
 
 	q.releaseWaitingQueriesLocked(ctx, txn.ID)
 
@@ -289,6 +308,9 @@ func (q *Queue) UpdateTxn(ctx context.Context, txn *roachpb.Transaction) {
 	pending.txn.Store(txn)
 	q.mu.Unlock()
 
+	q.store.GetTxnWaitMetrics().PusheeWaiting.Dec(1)
+	q.store.GetTxnWaitMetrics().PusherWaiting.Dec(int64(len(waitingPushes)))
+
 	if log.V(1) && len(waitingPushes) > 0 {
 		log.Infof(context.Background(), "updating %d push waiters for %s", len(waitingPushes), txn.ID.Short())
 	}
@@ -343,6 +365,8 @@ func (q *Queue) isTxnUpdated(pending *pendingTxn, req *roachpb.QueryTxnRequest)
 
 func (q *Queue) releaseWaitingQueriesLocked(ctx context.Context, txnID uuid.UUID) {
 	if w, ok := q.mu.queries[txnID]; ok {
+		metrics := q.store.GetTxnWaitMetrics()
+		metrics.QueryWaiting.Dec(int64(w.count))
 		log.VEventf(ctx, 2, "releasing %d waiting queries for %s", w.count, txnID.Short())
 		close(w.pending)
 		delete(q.mu.queries, txnID)
@@ -399,6 +423,9 @@ func (q *Queue) MaybeWaitForPush(
 		pending: make(chan *roachpb.Transaction, 1),
 	}
 	pending.waitingPushes = append(pending.waitingPushes, push)
+	if f := q.store.GetTxnWaitKnobs().OnPusherBlocked; f != nil {
+		f(ctx, req)
+	}
 	// Because we're adding another dependent on the pending
 	// transaction, send on the waiting queries' channel to
 	// indicate there is a new dependent and they should proceed
@@ -426,10 +453,8 @@ func (q *Queue) MaybeWaitForPush(
 	var queryPusherErrCh <-chan *roachpb.Error    // accepts errors querying the pusher txn
 	var readyCh chan struct{}                     // signaled when pusher txn should be queried
 
-	// Query the pusher if it's a valid transaction which already [may]
-	// have a transaction record (note that key being non-nil does not
-	// guarantee a txn record has been successfully written yet).
-	if req.PusherTxn.ID != uuid.Nil && req.PusherTxn.Key != nil {
+	// Query the pusher if it's a valid read-write transaction.
+	if req.PusherTxn.ID != uuid.Nil && req.PusherTxn.IsWriting() {
 		// Create a context which will be canceled once this call completes.
 		// This ensures that the goroutine created to query the pusher txn
 		// is properly cleaned up.
@@ -445,33 +470,51 @@ func (q *Queue) MaybeWaitForPush(
 			}
 		}()
 	}
-	var pusheeTxnTimer timeutil.Timer
-	defer pusheeTxnTimer.Stop()
 	pusherPriority := req.PusherTxn.Priority
 	pusheePriority := req.PusheeTxn.Priority
 
-	first := true
-	for {
-		// Set the timer to check for the pushee txn's expiration.
-		if !first {
-			expiration := TxnExpiration(pending.txn.Load().(*roachpb.Transaction)).GoTime()
-			now := q.store.Clock().Now().GoTime()
-			pusheeTxnTimer.Reset(expiration.Sub(now))
-		} else {
-			// The first time we want to check the pushee's txn record immediately:
-			// the pushee might be gone by the time the pusher gets here if it cleaned
-			// itself up after the pusher saw an intent but before it entered this
-			// queue.
-			pusheeTxnTimer.Reset(0)
-			first = false
-		}
+	metrics := q.store.GetTxnWaitMetrics()
+	metrics.PusherWaiting.Inc(1)
+	tBegin := timeutil.Now()
+	defer func() { metrics.PusherWaitTime.RecordValue(timeutil.Since(tBegin).Nanoseconds()) }()
 
+	slowTimerThreshold := time.Minute
+	slowTimer := timeutil.NewTimer()
+	defer slowTimer.Stop()
+	slowTimer.Reset(slowTimerThreshold)
+
+	var pusheeTxnTimer timeutil.Timer
+	defer pusheeTxnTimer.Stop()
+	// The first time we want to check the pushee's txn record immediately:
+	// the pushee might be gone by the time the pusher gets here if it cleaned
+	// itself up after the pusher saw an intent but before it entered this
+	// queue.
+	pusheeTxnTimer.Reset(0)
+	for {
 		select {
+		case <-slowTimer.C:
+			slowTimer.Read = true
+			metrics.PusherSlow.Inc(1)
+			log.Warningf(ctx, "pusher %s: have been waiting %.2fs for pushee %s",
+				req.PusherTxn.ID.Short(),
+				timeutil.Since(tBegin).Seconds(),
+				req.PusheeTxn.ID.Short(),
+			)
+			defer func() {
+				metrics.PusherSlow.Dec(1)
+				log.Warningf(ctx, "pusher %s: finished waiting after %.2fs for pushee %s",
+					req.PusherTxn.ID.Short(),
+					timeutil.Since(tBegin).Seconds(),
+					req.PusheeTxn.ID.Short(),
+				)
+			}()
 		case <-ctx.Done():
 			// Caller has given up.
 			log.VEvent(ctx, 2, "pusher giving up due to context cancellation")
 			return nil, roachpb.NewError(ctx.Err())
-
+		case <-q.store.Stopper().ShouldQuiesce():
+			// Let the push out so that they can be sent looking elsewhere.
+			return nil, nil
 		case txn := <-push.pending:
 			log.VEventf(ctx, 2, "result of pending push: %v", txn)
 			// If txn is nil, the queue was cleared, presumably because the
@@ -515,12 +558,16 @@ func (q *Queue) MaybeWaitForPush(
 				log.VEventf(ctx, 1, "pushing expired txn %s", req.PusheeTxn.ID.Short())
 				return nil, nil
 			}
+			// Set the timer to check for the pushee txn's expiration.
+			expiration := TxnExpiration(updatedPushee).GoTime()
+			now := q.store.Clock().Now().GoTime()
+			pusheeTxnTimer.Reset(expiration.Sub(now))
 
 		case updatedPusher := <-queryPusherCh:
 			switch updatedPusher.Status {
 			case roachpb.COMMITTED:
 				log.VEventf(ctx, 1, "pusher committed: %v", updatedPusher)
-				return nil, roachpb.NewErrorWithTxn(roachpb.NewTransactionStatusError("already committed"), updatedPusher)
+				return nil, roachpb.NewErrorWithTxn(roachpb.NewTransactionCommittedStatusError(), updatedPusher)
 			case roachpb.ABORTED:
 				log.VEventf(ctx, 1, "pusher aborted: %v", updatedPusher)
 				return nil, roachpb.NewErrorWithTxn(
@@ -568,6 +615,7 @@ func (q *Queue) MaybeWaitForPush(
 						req.PusheeTxn.ID.Short(),
 						dependents,
 					)
+					metrics.DeadlocksTotal.Inc(1)
 					return nil, ErrDeadlock
 
 				}
@@ -635,6 +683,11 @@ func (q *Queue) MaybeWaitForQuery(
 	}
 	q.mu.Unlock()
 
+	metrics := q.store.GetTxnWaitMetrics()
+	metrics.QueryWaiting.Inc(1)
+	tBegin := timeutil.Now()
+	defer func() { metrics.QueryWaitTime.RecordValue(timeutil.Since(tBegin).Nanoseconds()) }()
+
 	// When we return, make sure to unregister the query so that it doesn't
 	// leak. If query.pending if closed, the query will have already been
 	// cleaned up, so this will be a no-op.
@@ -642,6 +695,7 @@ func (q *Queue) MaybeWaitForQuery(
 		q.mu.Lock()
 		if query == q.mu.queries[req.Txn.ID] {
 			query.count--
+			metrics.QueryWaiting.Dec(1)
 			if query.count == 0 {
 				delete(q.mu.queries, req.Txn.ID)
 			}
@@ -708,6 +762,11 @@ func (q *Queue) startQueryPusherTxn(
 				} else if updatedPusher == nil {
 					// No pusher to query; the BeginTransaction hasn't yet created the
 					// pusher's record. Continue in order to backoff and retry.
+					// TODO(nvanbenschoten): we shouldn't hit this case in a 2.2
+					// cluster now that QueryTxn requests synthesize
+					// transactions from their provided TxnMeta. However, we
+					// need to keep the logic while we want to support
+					// compatibility with 2.1 nodes. Remove this in 2.3.
 					log.Event(ctx, "no pusher found; backing off")
 					continue
 				}
@@ -726,7 +785,7 @@ func (q *Queue) startQueryPusherTxn(
 
 				// Send an update of the pusher txn.
 				pusher.Update(updatedPusher)
-				ch <- &pusher
+				ch <- pusher
 
 				// Wait for context cancellation or indication on readyCh that the
 				// push waiter requires another query of the pusher txn.
@@ -797,7 +856,9 @@ func (q *Queue) queryTxnStatus(
 	}
 	br := b.RawResponse()
 	resp := br.Responses[0].GetInner().(*roachpb.QueryTxnResponse)
-	// ID can be nil if no BeginTransaction has been sent yet.
+	// ID can be nil if no BeginTransaction has been sent yet and we're talking
+	// to a 2.1 node.
+	// TODO(nvanbenschoten): Remove this in 2.3.
 	if updatedTxn := &resp.QueriedTxn; updatedTxn.ID != (uuid.UUID{}) {
 		return updatedTxn, resp.WaitingTxns, nil
 	}
