@@ -22,8 +22,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/base"
 	"github.com/cockroachdb/cockroach/pkg/config"
 	"github.com/cockroachdb/cockroach/pkg/gossip"
@@ -32,6 +30,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/server"
 	"github.com/cockroachdb/cockroach/pkg/server/serverpb"
 	"github.com/cockroachdb/cockroach/pkg/storage"
+	"github.com/cockroachdb/cockroach/pkg/storage/storagepb"
 	"github.com/cockroachdb/cockroach/pkg/testutils"
 	"github.com/cockroachdb/cockroach/pkg/testutils/serverutils"
 	"github.com/cockroachdb/cockroach/pkg/util/hlc"
@@ -39,6 +38,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/retry"
 	"github.com/cockroachdb/cockroach/pkg/util/stop"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/pkg/errors"
 )
 
 // TestCluster represents a set of TestServers. The hope is that it can be used
@@ -151,6 +151,17 @@ func StartTestCluster(t testing.TB, nodes int, args base.TestClusterArgs) *TestC
 		} else {
 			serverArgs = args.ServerArgs
 		}
+
+		// If there are multiple nodes, place them in different localities by
+		// default.
+		if nodes > 0 {
+			tiers := []roachpb.Tier{
+				{Key: "region", Value: "test"},
+				{Key: "dc", Value: fmt.Sprintf("dc%d", i+1)},
+			}
+			serverArgs.Locality = roachpb.Locality{Tiers: tiers}
+		}
+
 		if i > 0 {
 			serverArgs.JoinAddr = tc.Servers[0].ServingAddr()
 		}
@@ -172,6 +183,9 @@ func StartTestCluster(t testing.TB, nodes int, args base.TestClusterArgs) *TestC
 			t.Fatal(err)
 		}
 	}
+
+	// Wait until a NodeStatus is persisted for every node (see #25488, #25649, #31574).
+	tc.WaitForNodeStatuses(t)
 	return tc
 }
 
@@ -263,6 +277,13 @@ func (tc *TestCluster) doAddServer(t testing.TB, serverArgs base.TestServerArgs)
 		}
 	}
 
+	// Disable LBS if the server being added has a scan interval arg this low.
+	if serverArgs.ScanInterval > 0 && serverArgs.ScanInterval <= 100*time.Millisecond {
+		if _, err := conn.Exec(`SET CLUSTER SETTING kv.range_split.by_load_enabled = false`); err != nil {
+			t.Fatal(err)
+		}
+	}
+
 	tc.Servers = append(tc.Servers, s.(*server.TestServer))
 	tc.Conns = append(tc.Conns, conn)
 	tc.mu.Lock()
@@ -342,18 +363,19 @@ func (tc *TestCluster) changeReplicas(
 	changeType roachpb.ReplicaChangeType, startKey roachpb.RKey, targets ...roachpb.ReplicationTarget,
 ) (roachpb.RangeDescriptor, error) {
 	ctx := context.TODO()
-	if err := tc.Servers[0].DB().AdminChangeReplicas(
-		ctx, startKey.AsRawKey(), changeType, targets,
-	); err != nil {
-		return roachpb.RangeDescriptor{}, errors.Wrap(err, "AdminChangeReplicas error")
-	}
-	var rangeDesc roachpb.RangeDescriptor
+	var beforeDesc roachpb.RangeDescriptor
 	if err := tc.Servers[0].DB().GetProto(
-		ctx, keys.RangeDescriptorKey(startKey), &rangeDesc,
+		ctx, keys.RangeDescriptorKey(startKey), &beforeDesc,
 	); err != nil {
 		return roachpb.RangeDescriptor{}, errors.Wrap(err, "range descriptor lookup error")
 	}
-	return rangeDesc, nil
+	desc, err := tc.Servers[0].DB().AdminChangeReplicas(
+		ctx, startKey.AsRawKey(), changeType, targets, beforeDesc,
+	)
+	if err != nil {
+		return roachpb.RangeDescriptor{}, errors.Wrap(err, "AdminChangeReplicas error")
+	}
+	return *desc, nil
 }
 
 // AddReplicas is part of TestClusterInterface.
@@ -530,7 +552,7 @@ func (tc *TestCluster) findMemberStore(storeID roachpb.StoreID) (*storage.Store,
 // WaitForFullReplication waits until all stores in the cluster
 // have no ranges with replication pending.
 func (tc *TestCluster) WaitForFullReplication() error {
-	if int32(len(tc.Servers)) < config.DefaultZoneConfig().NumReplicas {
+	if int32(len(tc.Servers)) < *config.DefaultZoneConfig().NumReplicas {
 		return nil
 	}
 
@@ -543,8 +565,20 @@ func (tc *TestCluster) WaitForFullReplication() error {
 	notReplicated := true
 	for r := retry.Start(opts); r.Next() && notReplicated; {
 		notReplicated = false
-		for _, s := range tc.Servers {
+		for i, s := range tc.Servers {
 			err := s.Stores().VisitStores(func(s *storage.Store) error {
+				if n := s.ClusterNodeCount(); n != len(tc.Servers) {
+					log.Infof(context.TODO(), "%s only sees %d/%d available nodes", s, n, len(tc.Servers))
+					notReplicated = true
+					return nil
+				}
+				// Force the first node to upreplicate everything. Otherwise, if we rely
+				// on the scanner to do it, it'll take a while.
+				if i == 0 {
+					if err := s.ForceReplicationScanAndProcess(); err != nil {
+						return err
+					}
+				}
 				if err := s.ComputeMetrics(context.TODO(), 0); err != nil {
 					// This can sometimes fail since ComputeMetrics calls
 					// updateReplicationGauges which needs the system config gossiped.
@@ -584,13 +618,44 @@ func (tc *TestCluster) WaitForNodeStatuses(t testing.TB) {
 			return err
 		}
 
-		if len(response.Nodes) != tc.NumServers() {
+		if len(response.Nodes) < tc.NumServers() {
 			return fmt.Errorf("expected %d nodes registered, got %+v", tc.NumServers(), response)
 		}
+
+		// Check that all the nodes in the testcluster have a status. We tolerate
+		// other nodes having statuses (in some tests the cluster is configured with
+		// a pre-existing store).
+		nodeIDs := make(map[roachpb.NodeID]bool)
 		for _, node := range response.Nodes {
 			if len(node.StoreStatuses) == 0 {
 				return fmt.Errorf("missing StoreStatuses in NodeStatus: %+v", node)
 			}
+			nodeIDs[node.Desc.NodeID] = true
+		}
+		for _, s := range tc.Servers {
+			if id := s.GetNode().Descriptor.NodeID; !nodeIDs[id] {
+				return fmt.Errorf("missing n%d in NodeStatus: %+v", id, response)
+			}
+		}
+		return nil
+	})
+}
+
+// WaitForNodeLiveness waits until a liveness record is persisted for every
+// node in the cluster.
+func (tc *TestCluster) WaitForNodeLiveness(t testing.TB) {
+	testutils.SucceedsSoon(t, func() error {
+		db := tc.Servers[0].DB()
+		for _, s := range tc.Servers {
+			key := keys.NodeLivenessKey(s.NodeID())
+			var liveness storagepb.Liveness
+			if err := db.GetProto(context.Background(), key, &liveness); err != nil {
+				return err
+			}
+			if (liveness == storagepb.Liveness{}) {
+				return fmt.Errorf("no liveness record")
+			}
+			fmt.Printf("n%d: found liveness\n", s.NodeID())
 		}
 		return nil
 	})
