@@ -11,8 +11,7 @@
 import React from "react";
 import * as d3 from "d3";
 
-import worldShapes from "./world.json";
-import usStateShapes from "./us-states.json";
+import shapes from "./world.json";
 
 interface WorldMapProps {
   projection: d3.geo.Projection;
@@ -24,7 +23,7 @@ export class WorldMap extends React.Component<WorldMapProps> {
     return (
       <g>
         <g>
-          {worldShapes.features.map((feature: any, i: number) =>
+          {shapes.features.map((feature: any, i: number) =>
             <path
               key={i}
               className="geopath"
@@ -33,15 +32,6 @@ export class WorldMap extends React.Component<WorldMapProps> {
             />,
           )}
         </g>
-        <g>
-          {usStateShapes.features.map((feature: any, i: number) =>
-            <path
-              key={i}
-              className="geopath"
-              d={pathGen(feature)}
-            />,
-          )}
-        </g>
       </g>
     );
   }
