@@ -1,6 +1,6 @@
 // TODO(vilterp): pull this out into CSS...
 
-export const MAIN_BLUE = "#3A7DE1";
+export const MAIN_BLUE = "#3A7DE1";  // This should match $main-blue-color in palette.styl
 export const BACKGROUND_BLUE = "#B8CCEC";
 export const LIGHT_TEXT_BLUE = "#85A7E3";
 export const DARK_BLUE = "#152849";
