@@ -16,7 +16,6 @@ import classNames from "classnames";
 import _ from "lodash";
 import Long from "long";
 import moment from "moment";
-import { Link } from "react-router";
 import React from "react";
 
 import * as protos from "src/js/protos";
@@ -71,6 +70,7 @@ const rangeTableDisplayList: RangeTableRow[] = [
   { variable: "commit", display: "Commit", compareToLeader: true },
   { variable: "lastIndex", display: "Last Index", compareToLeader: true },
   { variable: "logSize", display: "Log Size", compareToLeader: false },
+  { variable: "logSizeTrusted", display: "Log Size Trusted?", compareToLeader: false },
   { variable: "leaseHolderQPS", display: "Lease Holder QPS", compareToLeader: false },
   { variable: "keysWrittenPS", display: "Average Keys Written Per Second", compareToLeader: false },
   { variable: "approxProposalQuota", display: "Approx Proposal Quota", compareToLeader: false },
@@ -87,10 +87,8 @@ const rangeTableDisplayList: RangeTableRow[] = [
   { variable: "mvccIntentBytesCount", display: "MVCC Intent Bytes/Count", compareToLeader: true },
   { variable: "mvccSystemBytesCount", display: "MVCC System Bytes/Count", compareToLeader: true },
   { variable: "rangeMaxBytes", display: "Max Range Size Before Split", compareToLeader: true },
-  { variable: "cmdQWrites", display: "CmdQ Writes Local/Global", compareToLeader: false },
-  { variable: "cmdQReads", display: "CmdQ Reads Local/Global", compareToLeader: false },
-  { variable: "cmdQMaxOverlapsSeen", display: "CmdQ Max Overlaps Local/Global", compareToLeader: false },
-  { variable: "cmdQTreeSize", display: "CmdQ Tree Size Local/Global", compareToLeader: false },
+  { variable: "writeLatches", display: "Write Latches Local/Global", compareToLeader: false },
+  { variable: "readLatches", display: "Read Latches Local/Global", compareToLeader: false },
 ];
 
 const rangeTableEmptyContent: RangeTableCellContent = {
@@ -152,7 +150,7 @@ export default class RangeTable extends React.Component<RangeTableProps, {}> {
     return {
       value: [`${humanizedBytes} / ${count.toString()} count`],
       title: [`${humanizedBytes} / ${count.toString()} count`,
-              `${bytes.toString()} bytes / ${count.toString()} count`],
+      `${bytes.toString()} bytes / ${count.toString()} count`],
     };
   }
 
@@ -176,7 +174,7 @@ export default class RangeTable extends React.Component<RangeTableProps, {}> {
     };
   }
 
-  contentCommandQueue(
+  contentLatchInfo(
     local: Long | number, global: Long | number, isRaftLeader: boolean,
   ): RangeTableCellContent {
     if (isRaftLeader) {
@@ -219,6 +217,9 @@ export default class RangeTable extends React.Component<RangeTableProps, {}> {
     if (problems.underreplicated) {
       results = _.concat(results, "Underreplicated (or slow)");
     }
+    if (problems.overreplicated) {
+      results = _.concat(results, "Overreplicated");
+    }
     if (problems.no_raft_leader) {
       results = _.concat(results, "No Raft Leader");
     }
@@ -398,33 +399,6 @@ export default class RangeTable extends React.Component<RangeTableProps, {}> {
     );
   }
 
-  renderCommandQueueVizRow(
-    sortedStoreIDs: number[],
-    rangeID: Long,
-    leader: protos.cockroach.server.serverpb.IRangeInfo,
-  ) {
-    const vizLink = (
-      <Link
-        to={`/reports/range/${rangeID}/cmdqueue`}
-        className="debug-link">
-        Visualize
-      </Link>
-    );
-
-    return (
-      <tr className="range-table__row">
-        <th className="range-table__cell range-table__cell--header">
-          CmdQ State
-        </th>
-        {sortedStoreIDs.map((storeId) => (
-          <td className="range-table__cell" key={storeId}>
-            {storeId === leader.source_store_id ? vizLink : "-"}
-          </td>
-        ))}
-      </tr>
-    );
-  }
-
   render() {
     const { infos, replicas } = this.props;
     const leader = _.head(infos);
@@ -501,6 +475,7 @@ export default class RangeTable extends React.Component<RangeTableProps, {}> {
         commit: this.contentIf(!dormant, () => this.createContent(FixLong(info.raft_state.hard_state.commit))),
         lastIndex: this.createContent(FixLong(info.state.last_index)),
         logSize: this.contentBytes(FixLong(info.state.raft_log_size)),
+        logSizeTrusted: this.createContent(info.state.raft_log_size_trusted.toString()),
         leaseHolderQPS: leaseHolder ? this.createContent(info.stats.queries_per_second.toFixed(4)) : rangeTableEmptyContent,
         keysWrittenPS: this.createContent(info.stats.writes_per_second.toFixed(4)),
         approxProposalQuota: raftLeader ? this.createContent(FixLong(info.state.approximate_proposal_quota)) : rangeTableEmptyContent,
@@ -520,24 +495,14 @@ export default class RangeTable extends React.Component<RangeTableProps, {}> {
         mvccIntentBytesCount: this.contentMVCC(FixLong(mvcc.intent_bytes), FixLong(mvcc.intent_count)),
         mvccSystemBytesCount: this.contentMVCC(FixLong(mvcc.sys_bytes), FixLong(mvcc.sys_count)),
         rangeMaxBytes: this.contentBytes(FixLong(info.state.range_max_bytes)),
-        cmdQWrites: this.contentCommandQueue(
-          FixLong(info.cmd_q_local.write_commands),
-          FixLong(info.cmd_q_global.write_commands),
+        writeLatches: this.contentLatchInfo(
+          FixLong(info.latches_local.write_count),
+          FixLong(info.latches_global.write_count),
           raftLeader,
         ),
-        cmdQReads: this.contentCommandQueue(
-          FixLong(info.cmd_q_local.read_commands),
-          FixLong(info.cmd_q_global.read_commands),
-          raftLeader,
-        ),
-        cmdQMaxOverlapsSeen: this.contentCommandQueue(
-          FixLong(info.cmd_q_local.max_overlaps_seen),
-          FixLong(info.cmd_q_global.max_overlaps_seen),
-          raftLeader,
-        ),
-        cmdQTreeSize: this.contentCommandQueue(
-          info.cmd_q_local.tree_size,
-          info.cmd_q_global.tree_size,
+        readLatches: this.contentLatchInfo(
+          FixLong(info.latches_local.read_count),
+          FixLong(info.latches_global.read_count),
           raftLeader,
         ),
       });
@@ -572,7 +537,6 @@ export default class RangeTable extends React.Component<RangeTableProps, {}> {
                 )
               ))
             }
-            {this.renderCommandQueueVizRow(sortedStoreIDs, rangeID, leader)}
             {
               _.map(replicas, (replica, key) => (
                 this.renderRangeReplicaRow(
