@@ -23,26 +23,45 @@ type Chart<T> = (sel: d3.Selection<T>) => void;
  */
 export default function createChartComponent<T>(containerTy: string, chart: Chart<T>) {
   return class WrappedChart extends React.Component<T> {
-    containerEl: Element;
+    containerEl: React.RefObject<Element> = React.createRef();
 
     componentDidMount() {
-      d3.select(this.containerEl)
-        .datum(this.props)
-        .call(chart);
+      this.redraw();
+      this.addResizeHandler();
+    }
+
+    componentWillUnmount() {
+      this.removeResizeHandler();
     }
 
     shouldComponentUpdate(props: T) {
-      d3.select(this.containerEl)
-        .datum(props)
-        .call(chart);
+      this.redraw(props);
 
       return false;
     }
 
+    redraw(props: T = this.props) {
+      d3.select(this.containerEl.current)
+        .datum(props)
+        .call(chart);
+    }
+
+    handleResize = () => {
+      this.redraw();
+    }
+
+    addResizeHandler() {
+      window.addEventListener("resize", this.handleResize);
+    }
+
+    removeResizeHandler() {
+      window.removeEventListener("resize", this.handleResize);
+    }
+
     render() {
       return React.createElement(
         containerTy,
-        { ref: (el) => this.containerEl = el },
+        { ref: this.containerEl },
       );
     }
   };
