@@ -20,6 +20,7 @@ import (
 	"fmt"
 	"math"
 	"math/big"
+	"strings"
 	"time"
 
 	"github.com/cockroachdb/cockroach/pkg/util/arith"
@@ -30,6 +31,7 @@ const (
 	nanosInDay    = 24 * int64(time.Hour) // Try as I might, couldn't do this without the cast.
 	nanosInMonth  = daysInMonth * nanosInDay
 	nanosInSecond = 1000 * 1000 * 1000
+	nanosInMicro  = 1000
 
 	// Used in overflow calculations.
 	maxYearsInDuration = math.MaxInt64 / nanosInMonth
@@ -57,12 +59,70 @@ var ErrEncodeOverflow = errors.New("overflow during Encode")
 // For the purposes of Compare and Encode, 1 month is considered equivalent to
 // 30 days and 1 day is equivalent to 24 * 60 * 60 * 1E9 nanoseconds.
 //
+// Although the Nanos field is a number of nanoseconds, all operations
+// round to the nearest microsecond. Any setting of this field should avoid
+// setting with precision below microseconds. The only exceptions are the
+// encode/decode operations.
+//
 // TODO(dan): Until the overflow and underflow handling is fixed, this is only
 // useful for durations of < 292 years.
 type Duration struct {
 	Months int64
 	Days   int64
-	Nanos  int64
+	// nanos is an unexported field so that it cannot be misused by other
+	// packages. It should almost always be rounded to the nearest microsecond.
+	nanos int64
+}
+
+// MakeDuration returns a Duration rounded to the nearest microsecond.
+func MakeDuration(nanos, days, months int64) Duration {
+	return Duration{
+		Months: months,
+		Days:   days,
+		nanos:  rounded(nanos),
+	}
+}
+
+// DecodeDuration returns a Duration without rounding nanos.
+func DecodeDuration(months, days, nanos int64) Duration {
+	return Duration{
+		Months: months,
+		Days:   days,
+		nanos:  nanos,
+	}
+}
+
+// Nanos returns the nanos of d.
+func (d Duration) Nanos() int64 {
+	return d.nanos
+}
+
+// SetNanos rounds and sets nanos.
+func (d *Duration) SetNanos(nanos int64) {
+	d.nanos = rounded(nanos)
+}
+
+// round rounds nanos to the nearest microsecond.
+func (d Duration) round() Duration {
+	d.nanos = rounded(d.nanos)
+	return d
+}
+
+// rounded returns nanos rounded to the nearest microsecond.
+func (d Duration) rounded() int64 {
+	return rounded(d.nanos)
+}
+
+// rounded returns nanos rounded to the nearest microsecond.
+func rounded(nanos int64) int64 {
+	dur := time.Duration(nanos) * time.Nanosecond
+	v := dur.Round(time.Microsecond).Nanoseconds()
+	// Near the boundaries of int64 will return the argument unchanged. Check
+	// for those cases and truncate instead of round so that we never have nanos.
+	if m := v % nanosInMicro; m != 0 {
+		v -= m
+	}
+	return v
 }
 
 // Compare returns an integer representing the relative length of two Durations.
@@ -78,9 +138,9 @@ func (d Duration) Compare(x Duration) int {
 		return -1
 	} else if normD.Days > normX.Days {
 		return 1
-	} else if normD.Nanos < normX.Nanos {
+	} else if normD.nanos < normX.nanos {
 		return -1
-	} else if normD.Nanos > normX.Nanos {
+	} else if normD.nanos > normX.nanos {
 		return 1
 	}
 	return 0
@@ -91,24 +151,23 @@ func (d Duration) Compare(x Duration) int {
 func FromInt64(x int64) Duration {
 	days := x / (nanosInDay / nanosInSecond)
 	seconds := x % (nanosInDay / nanosInSecond)
-	d := Duration{Days: days, Nanos: seconds * nanosInSecond}
+	d := Duration{Days: days, nanos: seconds * nanosInSecond}
 	return d.normalize()
 }
 
-// FromFloat64 converts a float64 number of seconds to a
-// duration. Inverse conversion of AsFloat64.
+// FromFloat64 converts a float64 number of seconds to a duration. Inverse
+// conversion of AsFloat64.
 func FromFloat64(x float64) Duration {
 	months := int64(x / float64(nanosInMonth/nanosInSecond))
 	secDays := math.Mod(x, float64(nanosInMonth/nanosInSecond))
 	days := int64(secDays / float64(nanosInDay/nanosInSecond))
 	secsRem := math.Mod(secDays, float64(nanosInDay/nanosInSecond))
-	d := Duration{Months: months, Days: days, Nanos: int64(secsRem * 1e9)}
-	return d.normalize()
+	d := Duration{Months: months, Days: days, nanos: int64(secsRem * 1e9)}
+	return d.normalize().round()
 }
 
-// FromBigInt converts a big.Int number of nanoseconds to a
-// duration. Inverse conversion of AsBigInt. Boolean false
-// if the result overflows.
+// FromBigInt converts a big.Int number of nanoseconds to a duration. Inverse
+// conversion of AsBigInt. Boolean false if the result overflows.
 func FromBigInt(src *big.Int) (Duration, bool) {
 	var rem big.Int
 	var monthsDec big.Int
@@ -123,8 +182,8 @@ func FromBigInt(src *big.Int) (Duration, bool) {
 	// Note: we do not need to check for overflow of daysDec because any
 	// excess bits were spilled into months above already.
 
-	d := Duration{Months: monthsDec.Int64(), Days: daysDec.Int64(), Nanos: nanosRem.Int64()}
-	return d.normalize(), true
+	d := Duration{Months: monthsDec.Int64(), Days: daysDec.Int64(), nanos: nanosRem.Int64()}
+	return d.normalize().round(), true
 }
 
 // AsInt64 converts a duration to an int64 number of seconds.
@@ -142,70 +201,105 @@ func (d Duration) AsInt64() (int64, bool) {
 	if dSecs, ok = arith.AddWithOverflow(mSecs, dSecs); !ok {
 		return 0, ok
 	}
-	return arith.AddWithOverflow(dSecs, d.Nanos/nanosInSecond)
+	return arith.AddWithOverflow(dSecs, d.nanos/nanosInSecond)
 }
 
 // AsFloat64 converts a duration to a float64 number of seconds.
 func (d Duration) AsFloat64() float64 {
 	mSecs := float64(d.Months) * float64(nanosInMonth/nanosInSecond)
 	dSecs := float64(d.Days) * float64(nanosInDay/nanosInSecond)
-	return float64(d.Nanos)/float64(nanosInSecond) + mSecs + dSecs
+	// Uses rounded instead of nanos here to remove any on-disk nanos.
+	return float64(d.rounded())/float64(nanosInSecond) + mSecs + dSecs
 }
 
-// AsBigInt converts a duration to a big.Int with the number of
-// nanoseconds.
+// AsBigInt converts a duration to a big.Int with the number of nanoseconds.
 func (d Duration) AsBigInt(dst *big.Int) {
 	dst.SetInt64(d.Months)
 	dst.Mul(dst, bigDaysInMonth)
 	dst.Add(dst, big.NewInt(d.Days))
 	dst.Mul(dst, bigNanosInDay)
-	dst.Add(dst, big.NewInt(d.Nanos))
+	// Uses rounded instead of nanos here to remove any on-disk nanos.
+	dst.Add(dst, big.NewInt(d.rounded()))
 }
 
-// Format emits a string representation of a Duration to a Buffer.
+const (
+	hourNanos   = uint64(time.Hour / time.Nanosecond)
+	minuteNanos = uint64(time.Minute / time.Nanosecond)
+	secondNanos = uint64(time.Second / time.Nanosecond)
+)
+
+// Format emits a string representation of a Duration to a Buffer truncated to microseconds.
 func (d Duration) Format(buf *bytes.Buffer) {
-	if d.Nanos == 0 && d.Days == 0 && d.Months == 0 {
-		buf.WriteString("0s")
+	if d.nanos == 0 && d.Days == 0 && d.Months == 0 {
+		buf.WriteString("00:00:00")
 		return
 	}
 
-	if absGE(d.Months, 11) {
-		fmt.Fprintf(buf, "%dy", d.Months/12)
+	wrote := false
+	// Defining both arguments in the signature gives a 10% speedup.
+	wrotePrev := func(wrote bool, buf *bytes.Buffer) bool {
+		if wrote {
+			buf.WriteString(" ")
+		}
+		return true
+	}
+
+	negDays := d.Months < 0 || d.Days < 0
+	if absGE(d.Months, 12) {
+		years := d.Months / 12
+		wrote = wrotePrev(wrote, buf)
+		fmt.Fprintf(buf, "%d year%s", years, isPlural(years))
 		d.Months %= 12
 	}
 	if d.Months != 0 {
-		fmt.Fprintf(buf, "%dmon", d.Months)
+		wrote = wrotePrev(wrote, buf)
+		fmt.Fprintf(buf, "%d mon%s", d.Months, isPlural(d.Months))
 	}
 	if d.Days != 0 {
-		fmt.Fprintf(buf, "%dd", d.Days)
+		wrote = wrotePrev(wrote, buf)
+		fmt.Fprintf(buf, "%d day%s", d.Days, isPlural(d.Days))
 	}
 
-	// The following comparisons are careful to preserve the sign in
-	// case the value is MinInt64, and thus cannot be made positive lest
-	// an overflow occur.
-	if absGE(d.Nanos, time.Hour.Nanoseconds()) {
-		fmt.Fprintf(buf, "%dh", d.Nanos/time.Hour.Nanoseconds())
-		d.Nanos %= time.Hour.Nanoseconds()
+	if d.nanos == 0 {
+		return
 	}
-	if absGE(d.Nanos, time.Minute.Nanoseconds()) {
-		fmt.Fprintf(buf, "%dm", d.Nanos/time.Minute.Nanoseconds())
-		d.Nanos %= time.Minute.Nanoseconds()
+
+	wrotePrev(wrote, buf)
+
+	if d.nanos/nanosInMicro < 0 {
+		buf.WriteString("-")
+	} else if negDays {
+		buf.WriteString("+")
 	}
-	if absGE(d.Nanos, time.Second.Nanoseconds()) {
-		fmt.Fprintf(buf, "%ds", d.Nanos/time.Second.Nanoseconds())
-		d.Nanos %= time.Second.Nanoseconds()
+
+	// Extract abs(d.nanos). See https://play.golang.org/p/U3_gNMpyUew.
+	var nanos uint64
+	if d.nanos >= 0 {
+		nanos = uint64(d.nanos)
+	} else {
+		nanos = uint64(-d.nanos)
 	}
-	if absGE(d.Nanos, time.Millisecond.Nanoseconds()) {
-		fmt.Fprintf(buf, "%dms", d.Nanos/time.Millisecond.Nanoseconds())
-		d.Nanos %= time.Millisecond.Nanoseconds()
+
+	hn := nanos / hourNanos
+	nanos %= hourNanos
+	mn := nanos / minuteNanos
+	nanos %= minuteNanos
+	sn := nanos / secondNanos
+	nanos %= secondNanos
+	fmt.Fprintf(buf, "%02d:%02d:%02d", hn, mn, sn)
+
+	micros := nanos / nanosInMicro
+	if micros != 0 {
+		s := fmt.Sprintf(".%06d", micros)
+		buf.WriteString(strings.TrimRight(s, "0"))
 	}
-	if absGE(d.Nanos, time.Microsecond.Nanoseconds()) {
-		fmt.Fprintf(buf, "%dµs", d.Nanos/time.Microsecond.Nanoseconds())
-		d.Nanos %= time.Microsecond.Nanoseconds()
-	}
-	if d.Nanos != 0 {
-		fmt.Fprintf(buf, "%dns", d.Nanos)
+}
+
+func isPlural(i int64) string {
+	if i == 1 {
+		return ""
 	}
+	return "s"
 }
 
 // absGE returns whether x is greater than or equal to y in magnitude.
@@ -224,6 +318,19 @@ func (d Duration) String() string {
 	return buf.String()
 }
 
+// StringNanos returns a string representation of a Duration including
+// its hidden nanoseconds value. To be used only by the encoding/decoding
+// packages for pretty printing of on-disk values.
+func (d Duration) StringNanos() string {
+	var buf bytes.Buffer
+	d.Format(&buf)
+	nanos := d.nanos % nanosInMicro
+	if nanos != 0 {
+		fmt.Fprintf(&buf, "%+dns", nanos)
+	}
+	return buf.String()
+}
+
 // Encode returns three integers such that the original Duration is recoverable
 // (using Decode) and the first int will approximately sort a collection of
 // encoded Durations.
@@ -233,12 +340,12 @@ func (d Duration) Encode() (sortNanos int64, months int64, days int64, err error
 	//
 	// TODO(dan): Compute overflow exactly, then document that EncodeBigInt can be
 	// used in overflow cases.
-	years := d.Months/12 + d.Days/daysInMonth/12 + d.Nanos/nanosInMonth/12
+	years := d.Months/12 + d.Days/daysInMonth/12 + d.nanos/nanosInMonth/12
 	if years > maxYearsInDuration || years < minYearsInDuration {
 		return 0, 0, 0, ErrEncodeOverflow
 	}
 
-	totalNanos := d.Months*nanosInMonth + d.Days*nanosInDay + d.Nanos
+	totalNanos := d.Months*nanosInMonth + d.Days*nanosInDay + d.nanos
 	return totalNanos, d.Months, d.Days, nil
 }
 
@@ -249,7 +356,7 @@ func (d Duration) EncodeBigInt() (sortNanos *big.Int, months int64, days int64)
 	bigMonths.Mul(bigMonths, big.NewInt(nanosInMonth))
 	bigDays := big.NewInt(d.Days)
 	bigDays.Mul(bigDays, big.NewInt(nanosInDay))
-	totalNanos := big.NewInt(d.Nanos)
+	totalNanos := big.NewInt(d.nanos)
 	totalNanos.Add(totalNanos, bigMonths).Add(totalNanos, bigDays)
 	return totalNanos, d.Months, d.Days
 }
@@ -260,53 +367,138 @@ func Decode(sortNanos int64, months int64, days int64) (Duration, error) {
 	nanos := sortNanos - months*nanosInMonth - days*nanosInDay
 	// TODO(dan): Handle underflow, then document that DecodeBigInt can be used
 	// in underflow cases.
-	return Duration{Months: months, Days: days, Nanos: nanos}, nil
+	return Duration{Months: months, Days: days, nanos: nanos}, nil
 }
 
 // TODO(dan): Write DecodeBigInt.
 
-// Add returns the time t+d.
-func Add(t time.Time, d Duration) time.Time {
-	// TODO(dan): Overflow handling.
-	return t.AddDate(0, int(d.Months), int(d.Days)).Add(time.Duration(d.Nanos) * time.Nanosecond)
+// AdditionMode controls date normalization behaviors in Add().
+type AdditionMode bool
+
+const (
+	// AdditionModeCompatible applies a date-normalization strategy
+	// which produces results which are more compatible with
+	// PostgreSQL in which adding a month "rounds down" to the last
+	// day of a month instead of producing a value in the following
+	// month.
+	//
+	// See PostgreSQL 10.5: src/backend/utils/adt/timestamp.c timestamp_pl_interval().
+	AdditionModeCompatible AdditionMode = false
+	// AdditionModeLegacy delegates to time.Time.AddDate() for
+	// performing Time+Duration math.
+	AdditionModeLegacy AdditionMode = true
+)
+
+func (m AdditionMode) String() string {
+	switch m {
+	case AdditionModeLegacy:
+		return "legacy"
+	default:
+		return "compatible"
+	}
+}
+
+// Context is used to prevent a package-dependency cycle via tree.EvalContext.
+type Context interface {
+	GetAdditionMode() AdditionMode
+}
+
+// GetAdditionMode allows an AdditionMode to be used as its own context.
+func (m AdditionMode) GetAdditionMode() AdditionMode {
+	return m
+}
+
+// Add returns the time t+d, using a configurable mode.
+func Add(ctx Context, t time.Time, d Duration) time.Time {
+	var mode AdditionMode
+	if ctx != nil {
+		mode = ctx.GetAdditionMode()
+	}
+
+	// We can fast-path if the duration is always a fixed amount of time,
+	// or if the day number that we're starting from can never result
+	// in normalization.
+	if mode == AdditionModeLegacy || d.Months == 0 || t.Day() <= 28 {
+		return t.AddDate(0, int(d.Months), int(d.Days)).Add(time.Duration(d.nanos))
+	}
+
+	// Adjustments for 1-based math.
+	expectedMonth := time.Month((int(t.Month())-1+int(d.Months))%12 + 1)
+
+	// Use AddDate() to get a rough value.  This might overshoot the
+	// end of the expected month by multiple days.  We could iteratively
+	// subtract a day until we jump a month backwards, but that's
+	// at least twice as slow as computing the correct value ourselves.
+	res := t.AddDate(0 /* years */, int(d.Months), 0 /* days */)
+
+	// Unpack fields as little as possible.
+	year, month, _ := res.Date()
+	hour, min, sec := res.Clock()
+
+	if month != expectedMonth {
+		// Pro-tip: Count across your knuckles and the divots between
+		// them, wrapping around when you hit July. Knuckle == 31 days.
+		var lastDayOfMonth int
+		switch expectedMonth {
+		case time.February:
+			// Leap year if divisible by 4, but not centuries unless also divisible by 400.
+			// Adjust the earth's orbital parameters?
+			if year%4 == 0 && (year%100 != 0 || year%400 == 0) {
+				lastDayOfMonth = 29
+			} else {
+				lastDayOfMonth = 28
+			}
+		case time.January, time.March, time.May, time.July, time.August, time.October, time.December:
+			lastDayOfMonth = 31
+		default:
+			lastDayOfMonth = 30
+		}
+
+		res = time.Date(
+			year, expectedMonth, lastDayOfMonth,
+			hour, min, sec,
+			res.Nanosecond(), res.Location())
+	}
+
+	return res.AddDate(0, 0, int(d.Days)).Add(time.Duration(d.nanos))
 }
 
 // Add returns a Duration representing a time length of d+x.
 func (d Duration) Add(x Duration) Duration {
-	return Duration{d.Months + x.Months, d.Days + x.Days, d.Nanos + x.Nanos}
+	return MakeDuration(d.nanos+x.nanos, d.Days+x.Days, d.Months+x.Months)
 }
 
 // Sub returns a Duration representing a time length of d-x.
 func (d Duration) Sub(x Duration) Duration {
-	return Duration{d.Months - x.Months, d.Days - x.Days, d.Nanos - x.Nanos}
+	return MakeDuration(d.nanos-x.nanos, d.Days-x.Days, d.Months-x.Months)
 }
 
 // Mul returns a Duration representing a time length of d*x.
 func (d Duration) Mul(x int64) Duration {
-	return Duration{d.Months * x, d.Days * x, d.Nanos * x}
+	return MakeDuration(d.nanos*x, d.Days*x, d.Months*x)
 }
 
 // Div returns a Duration representing a time length of d/x.
 func (d Duration) Div(x int64) Duration {
-	return Duration{d.Months / x, d.Days / x, d.Nanos / x}
+	return MakeDuration(d.nanos/x, d.Days/x, d.Months/x)
 }
 
 // MulFloat returns a Duration representing a time length of d*x.
 func (d Duration) MulFloat(x float64) Duration {
-	return Duration{
-		int64(float64(d.Months) * x),
-		int64(float64(d.Days) * x),
-		int64(float64(d.Nanos) * x),
-	}
+	return MakeDuration(
+		int64(float64(d.nanos)*x),
+		int64(float64(d.Days)*x),
+		int64(float64(d.Months)*x),
+	)
 }
 
 // DivFloat returns a Duration representing a time length of d/x.
 func (d Duration) DivFloat(x float64) Duration {
-	return Duration{
-		int64(float64(d.Months) / x),
-		int64(float64(d.Days) / x),
-		int64(float64(d.Nanos) / x),
-	}
+	return MakeDuration(
+		int64(float64(d.nanos)/x),
+		int64(float64(d.Days)/x),
+		int64(float64(d.Months)/x),
+	)
 }
 
 // normalized returns a new Duration transformed using the equivalence rules.
@@ -324,10 +516,10 @@ func (d Duration) normalize() Duration {
 	//   made a full month.
 	// - Months did hit MaxInt64 or MinInt64, in which case there can be no more
 	//   months. We only need to shift nanos.
-	if d.Nanos > 0 {
+	if d.nanos > 0 {
 		d = d.shiftPosNanosToDays()
 		d = d.shiftPosDaysToMonths()
-	} else if d.Nanos < 0 {
+	} else if d.nanos < 0 {
 		d = d.shiftNegNanosToDays()
 		d = d.shiftNegDaysToMonths()
 	}
@@ -354,9 +546,9 @@ func (d Duration) shiftPosNanosToDays() Duration {
 		// rate, we can never transfer more than math.MaxInt64 anyway.
 		maxDays = math.MaxInt64 - d.Days
 	}
-	daysFromNanos := int64Min(d.Nanos/nanosInDay, maxDays)
+	daysFromNanos := int64Min(d.nanos/nanosInDay, maxDays)
 	d.Days += daysFromNanos
-	d.Nanos -= daysFromNanos * nanosInDay
+	d.nanos -= daysFromNanos * nanosInDay
 	return d
 }
 
@@ -380,9 +572,9 @@ func (d Duration) shiftNegNanosToDays() Duration {
 		// rate, we can never transfer more than math.MaxInt64 anyway.
 		minDays = math.MinInt64 - d.Days
 	}
-	daysFromNanos := int64Max(d.Nanos/nanosInDay, minDays)
+	daysFromNanos := int64Max(d.nanos/nanosInDay, minDays)
 	d.Days += daysFromNanos
-	d.Nanos -= daysFromNanos * nanosInDay
+	d.nanos -= daysFromNanos * nanosInDay
 	return d
 }
 
