@@ -25,9 +25,8 @@ import (
 	"math/big"
 	"unsafe"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/apd"
+	"github.com/pkg/errors"
 )
 
 // EncodeDecimalAscending returns the resulting byte slice with the encoded decimal
@@ -356,17 +355,16 @@ func getDecimalLen(buf []byte) (int, error) {
 // makeDecimalFromMandE reconstructs the decimal from the mantissa M and
 // exponent E.
 func makeDecimalFromMandE(negative bool, e int, m []byte, tmp []byte) (apd.Decimal, error) {
+	if len(m) == 0 {
+		return apd.Decimal{}, errors.New("expected mantissa, got zero bytes")
+	}
 	// ±dddd.
 	b := tmp[:0]
 	if n := len(m)*2 + 1; cap(b) < n {
 		b = make([]byte, 0, n)
 	}
-	for i, v := range m {
-		t := int(v)
-		if i == len(m) {
-			t--
-		}
-		t /= 2
+	for _, v := range m {
+		t := int(v) / 2
 		if t < 0 || t > 99 {
 			return apd.Decimal{}, errors.Errorf("base-100 encoded digit %d out of range [0,99]", t)
 		}
@@ -623,55 +621,61 @@ func encodeNonsortingDecimalValueWithoutExp(digits []big.Word, buf []byte) []byt
 // of the encoded value is.
 func DecodeNonsortingDecimal(buf []byte, tmp []byte) (apd.Decimal, error) {
 	var dec apd.Decimal
+	err := DecodeIntoNonsortingDecimal(&dec, buf, tmp)
+	return dec, err
+}
 
+// DecodeIntoNonsortingDecimal is like DecodeNonsortingDecimal, but it operates
+// on the passed-in *apd.Decimal instead of producing a new one.
+func DecodeIntoNonsortingDecimal(dec *apd.Decimal, buf []byte, tmp []byte) error {
 	switch buf[0] {
 	case decimalNaN:
 		dec.Form = apd.NaN
-		return dec, nil
+		return nil
 	case decimalNegativeInfinity:
 		dec.Form = apd.Infinite
 		dec.Negative = true
-		return dec, nil
+		return nil
 	case decimalInfinity:
 		dec.Form = apd.Infinite
-		return dec, nil
+		return nil
 	case decimalZero:
-		return dec, nil
+		return nil
 	}
 
 	dec.Form = apd.Finite
 	switch {
 	case buf[0] == decimalNegLarge:
-		if err := decodeNonsortingDecimalValue(&dec, false, buf[1:], tmp); err != nil {
-			return apd.Decimal{}, err
+		if err := decodeNonsortingDecimalValue(dec, false, buf[1:], tmp); err != nil {
+			return err
 		}
 		dec.Negative = true
-		return dec, nil
+		return nil
 	case buf[0] == decimalNegMedium:
-		decodeNonsortingDecimalValueWithoutExp(&dec, buf[1:], tmp)
+		decodeNonsortingDecimalValueWithoutExp(dec, buf[1:], tmp)
 		dec.Negative = true
-		return dec, nil
+		return nil
 	case buf[0] == decimalNegSmall:
-		if err := decodeNonsortingDecimalValue(&dec, true, buf[1:], tmp); err != nil {
-			return apd.Decimal{}, err
+		if err := decodeNonsortingDecimalValue(dec, true, buf[1:], tmp); err != nil {
+			return err
 		}
 		dec.Negative = true
-		return dec, nil
+		return nil
 	case buf[0] == decimalPosSmall:
-		if err := decodeNonsortingDecimalValue(&dec, true, buf[1:], tmp); err != nil {
-			return apd.Decimal{}, err
+		if err := decodeNonsortingDecimalValue(dec, true, buf[1:], tmp); err != nil {
+			return err
 		}
-		return dec, nil
+		return nil
 	case buf[0] == decimalPosMedium:
-		decodeNonsortingDecimalValueWithoutExp(&dec, buf[1:], tmp)
-		return dec, nil
+		decodeNonsortingDecimalValueWithoutExp(dec, buf[1:], tmp)
+		return nil
 	case buf[0] == decimalPosLarge:
-		if err := decodeNonsortingDecimalValue(&dec, false, buf[1:], tmp); err != nil {
-			return apd.Decimal{}, err
+		if err := decodeNonsortingDecimalValue(dec, false, buf[1:], tmp); err != nil {
+			return err
 		}
-		return dec, nil
+		return nil
 	default:
-		return apd.Decimal{}, errors.Errorf("unknown prefix of the encoded byte slice: %q", buf)
+		return errors.Errorf("unknown decimal prefix of the encoded byte slice: %q", buf)
 	}
 }
 
