@@ -197,21 +197,15 @@ func NewArrayBuilder(numAddsHint int) *ArrayBuilder {
 
 // Add appends JSON to the sequence.
 func (b *ArrayBuilder) Add(j JSON) {
-	if b.jsons == nil {
-		panic(msgModifyAfterBuild)
-	}
 	b.jsons = append(b.jsons, j)
 }
 
-// Build returns a JSON array built from a JSON sequence. After that, it should
-// not be modified any longer.
+// Build returns the constructed JSON array. A caller may not modify the array,
+// and the ArrayBuilder reserves the right to re-use the array returned (though
+// the data will not be modified).  This is important in the case of a window
+// function, which might want to incrementally update an aggregation.
 func (b *ArrayBuilder) Build() JSON {
-	if b.jsons == nil {
-		panic(msgModifyAfterBuild)
-	}
-	arr := jsonArray(b.jsons)
-	b.jsons = nil
-	return arr
+	return jsonArray(b.jsons)
 }
 
 // ArrayBuilderWithCounter builds JSON Array by a JSON sequence with a size counter.
@@ -749,6 +743,30 @@ func (j jsonObject) encodeInvertedIndexKeys(b []byte) ([][]byte, error) {
 	return outKeys, nil
 }
 
+// NumInvertedIndexEntries returns the number of inverted index entries that
+// would be created for the given JSON value. Since identical elements of an
+// array are encoded identically in the inverted index, the total number of
+// distinct index entries may be less than the total number of paths.
+func NumInvertedIndexEntries(j JSON) (int, error) {
+	// TODO (lucy): Figure out how to avoid allocating every path
+	keys, err := EncodeInvertedIndexKeys(make([]byte, 0), j)
+	if err != nil {
+		return 0, err
+	}
+
+	// Count distinct keys
+	sort.Slice(keys, func(i int, j int) bool {
+		return bytes.Compare(keys[i], keys[j]) < 0
+	})
+	n := 0
+	for i := 0; i < len(keys); i++ {
+		if i == 0 || bytes.Compare(keys[i-1], keys[i]) < 0 {
+			n++
+		}
+	}
+	return n, nil
+}
+
 // AllPaths returns a slice of new JSON documents, each a path to a leaf
 // through the input. Note that leaves include the empty object and array
 // in addition to scalars.
@@ -875,14 +893,14 @@ func fromMap(v map[string]interface{}) (JSON, error) {
 // FromInt returns a JSON value given a int.
 func FromInt(v int) JSON {
 	dec := apd.Decimal{}
-	dec.SetCoefficient(int64(v))
+	dec.SetFinite(int64(v), 0)
 	return jsonNumber(dec)
 }
 
 // FromInt64 returns a JSON value given a int64.
 func FromInt64(v int64) JSON {
 	dec := apd.Decimal{}
-	dec.SetCoefficient(v)
+	dec.SetFinite(v, 0)
 	return jsonNumber(dec)
 }
 
@@ -1026,27 +1044,7 @@ func setValKeyOrIdx(j JSON, key string, to JSON, createMissing bool) (JSON, erro
 		}
 		return setValKeyOrIdx(n, key, to, createMissing)
 	case jsonObject:
-		result := make(jsonObject, 0, len(v)+1)
-		curIdx := 0
-		for curIdx < len(v) && string(v[curIdx].k) < key {
-			result = append(result, v[curIdx])
-			curIdx++
-		}
-		keyAlreadyExists := curIdx < len(v) && string(v[curIdx].k) == key
-		if createMissing || keyAlreadyExists {
-			result = append(result, jsonKeyValuePair{
-				k: jsonString(key),
-				v: to,
-			})
-		}
-		if keyAlreadyExists {
-			curIdx++
-		}
-		for curIdx < len(v) {
-			result = append(result, v[curIdx])
-			curIdx++
-		}
-		return result, nil
+		return v.SetKey(key, to, createMissing)
 	case jsonArray:
 		idx, err := strconv.Atoi(key)
 		if err != nil {
@@ -1118,6 +1116,94 @@ func deepSet(j JSON, path []string, to JSON, createMissing bool) (JSON, error) {
 	}
 }
 
+var errCannotReplaceExistingKey = pgerror.NewError(pgerror.CodeInvalidParameterValueError, "cannot replace existing key")
+
+func insertValKeyOrIdx(j JSON, key string, newVal JSON, insertAfter bool) (JSON, error) {
+	switch v := j.(type) {
+	case *jsonEncoded:
+		n, err := v.shallowDecode()
+		if err != nil {
+			return nil, err
+		}
+		return insertValKeyOrIdx(n, key, newVal, insertAfter)
+	case jsonObject:
+		result, err := v.SetKey(key, newVal, true)
+		if err != nil {
+			return nil, err
+		}
+		if len(result) == len(v) {
+			return nil, errCannotReplaceExistingKey
+		}
+		return result, nil
+	case jsonArray:
+		idx, err := strconv.Atoi(key)
+		if err != nil {
+			return nil, err
+		}
+		if idx < 0 {
+			idx = len(v) + idx
+		}
+		if insertAfter {
+			idx++
+		}
+
+		var result = make(jsonArray, len(v)+1)
+		if idx <= 0 {
+			copy(result[1:], v)
+			result[0] = newVal
+		} else if idx >= len(v) {
+			copy(result, v)
+			result[len(result)-1] = newVal
+		} else {
+			copy(result[:idx], v[:idx])
+			copy(result[idx+1:], v[idx:])
+			result[idx] = newVal
+		}
+		return result, nil
+	}
+	return j, nil
+}
+
+// DeepInsert inserts a value at a path in a JSON document.
+// Implements the jsonb_insert builtin.
+func DeepInsert(j JSON, path []string, to JSON, insertAfter bool) (JSON, error) {
+	if j.isScalar() {
+		return nil, errCannotSetPathInScalar
+	}
+	return deepInsert(j, path, to, insertAfter)
+}
+
+func deepInsert(j JSON, path []string, to JSON, insertAfter bool) (JSON, error) {
+	switch len(path) {
+	case 0:
+		return j, nil
+	case 1:
+		return insertValKeyOrIdx(j, path[0], to, insertAfter)
+	default:
+		switch v := j.(type) {
+		case *jsonEncoded:
+			n, err := v.shallowDecode()
+			if err != nil {
+				return nil, err
+			}
+			return deepInsert(n, path, to, insertAfter)
+		default:
+			fetched, err := j.FetchValKeyOrIdx(path[0])
+			if err != nil {
+				return nil, err
+			}
+			if fetched == nil {
+				return j, nil
+			}
+			sub, err := deepInsert(fetched, path[1:], to, insertAfter)
+			if err != nil {
+				return nil, err
+			}
+			return setValKeyOrIdx(j, path[0], sub, true)
+		}
+	}
+}
+
 func (j jsonObject) FetchValKeyOrIdx(key string) (JSON, error) {
 	return j.FetchValKey(key)
 }
@@ -1142,6 +1228,34 @@ func (jsonNumber) FetchValKeyOrIdx(string) (JSON, error) { return nil, nil }
 var errCannotDeleteFromScalar = pgerror.NewError(pgerror.CodeInvalidParameterValueError, "cannot delete from scalar")
 var errCannotDeleteFromObject = pgerror.NewError(pgerror.CodeInvalidParameterValueError, "cannot delete from object using integer index")
 
+func (j jsonObject) SetKey(key string, to JSON, createMissing bool) (jsonObject, error) {
+	result := make(jsonObject, 0, len(j)+1)
+	curIdx := 0
+
+	for curIdx < len(j) && string(j[curIdx].k) < key {
+		result = append(result, j[curIdx])
+		curIdx++
+	}
+
+	keyAlreadyExists := curIdx < len(j) && string(j[curIdx].k) == key
+	if createMissing || keyAlreadyExists {
+		result = append(result, jsonKeyValuePair{
+			k: jsonString(key),
+			v: to,
+		})
+	}
+	if keyAlreadyExists {
+		curIdx++
+	}
+
+	for curIdx < len(j) {
+		result = append(result, j[curIdx])
+		curIdx++
+	}
+
+	return result, nil
+}
+
 func (j jsonArray) RemoveString(s string) (JSON, bool, error) {
 	b := NewArrayBuilder(j.Len())
 	removed := false
@@ -1344,7 +1458,11 @@ func (jsonNull) Exists(string) (bool, error)   { return false, nil }
 func (jsonTrue) Exists(string) (bool, error)   { return false, nil }
 func (jsonFalse) Exists(string) (bool, error)  { return false, nil }
 func (jsonNumber) Exists(string) (bool, error) { return false, nil }
-func (jsonString) Exists(string) (bool, error) { return false, nil }
+
+func (j jsonString) Exists(s string) (bool, error) {
+	return string(j) == s, nil
+}
+
 func (j jsonArray) Exists(s string) (bool, error) {
 	for i := 0; i < len(j); i++ {
 		if elem, ok := j[i].(jsonString); ok && string(elem) == s {
