@@ -39,6 +39,7 @@ import (
 
 	"github.com/cockroachdb/cockroach/pkg/build"
 	"github.com/cockroachdb/cockroach/pkg/util/caller"
+	"github.com/cockroachdb/cockroach/pkg/util/envutil"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/sysutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
@@ -46,6 +47,10 @@ import (
 	"github.com/petermattis/goid"
 )
 
+// maxSyncDuration is set to a conservative value since this is a new mechanism.
+// In practice, even a fraction of that would indicate a problem.
+var maxSyncDuration = envutil.EnvOrDefaultDuration("COCKROACH_LOG_MAX_SYNC_DURATION", 30*time.Second)
+
 const fatalErrorPostamble = `
 
 ****************************************************************************
@@ -349,17 +354,22 @@ var entryRE = regexp.MustCompile(
 // buffer. Each entry is preceded by a single big-ending uint32
 // describing the next entry's length.
 type EntryDecoder struct {
+	re                 *regexp.Regexp
 	scanner            *bufio.Scanner
 	truncatedLastEntry bool
 }
 
 // NewEntryDecoder creates a new instance of EntryDecoder.
 func NewEntryDecoder(in io.Reader) *EntryDecoder {
-	d := &EntryDecoder{scanner: bufio.NewScanner(in)}
+	d := &EntryDecoder{scanner: bufio.NewScanner(in), re: entryRE.Copy()}
 	d.scanner.Split(d.split)
 	return d
 }
 
+// MessageTimeFormat is the format of the timestamp in log message headers as
+// used in time.Parse and time.Format.
+const MessageTimeFormat = "060102 15:04:05.999999"
+
 // Decode decodes the next log entry into the provided protobuf message.
 func (d *EntryDecoder) Decode(entry *Entry) error {
 	for {
@@ -370,12 +380,12 @@ func (d *EntryDecoder) Decode(entry *Entry) error {
 			return io.EOF
 		}
 		b := d.scanner.Bytes()
-		m := entryRE.FindSubmatch(b)
+		m := d.re.FindSubmatch(b)
 		if m == nil {
 			continue
 		}
 		entry.Severity = Severity(strings.IndexByte(severityChar, m[1][0]) + 1)
-		t, err := time.Parse("060102 15:04:05.999999", string(m[2]))
+		t, err := time.Parse(MessageTimeFormat, string(m[2]))
 		if err != nil {
 			return err
 		}
@@ -403,7 +413,7 @@ func (d *EntryDecoder) split(data []byte, atEOF bool) (advance int, token []byte
 		return 0, nil, nil
 	}
 	if d.truncatedLastEntry {
-		i := entryRE.FindIndex(data)
+		i := d.re.FindIndex(data)
 		if i == nil {
 			// If there's no entry that starts in this chunk, advance past it, since
 			// we've truncated the entry it was originally part of.
@@ -421,7 +431,7 @@ func (d *EntryDecoder) split(data []byte, atEOF bool) (advance int, token []byte
 	}
 	// From this point on, we assume we're currently positioned at a log entry.
 	// We want to find the next one so we start our search at data[1].
-	i := entryRE.FindIndex(data[1:])
+	i := d.re.FindIndex(data[1:])
 	if i == nil {
 		if atEOF {
 			return len(data), data, nil
@@ -638,8 +648,10 @@ func LoggingToStderr(s Severity) bool {
 // StartGCDaemon starts the log file GC -- this must be called after
 // command-line parsing has completed so that no data is lost when the
 // user configures larger max sizes than the defaults.
-func StartGCDaemon() {
-	go logging.gcDaemon()
+//
+// The logger's GC daemon stops when the provided context is canceled.
+func StartGCDaemon(ctx context.Context) {
+	go logging.gcDaemon(ctx)
 }
 
 // Flush flushes all pending log I/O.
@@ -891,6 +903,8 @@ func (l *loggingT) outputLogEntry(s Severity, file string, line int, msg string)
 		}
 		exitCalled := make(chan struct{})
 
+		// This defer prevents outputLogEntry() from returning until the
+		// exit function has been called.
 		defer func() {
 			<-exitCalled
 		}()
@@ -937,6 +951,15 @@ func (l *loggingT) outputLogEntry(s Severity, file string, line int, msg string)
 	if s == Severity_FATAL {
 		l.flushAndSync(true /*doSync*/)
 		close(fatalTrigger)
+		// Note: although it seems like the function is allowed to return
+		// below when s == Severity_FATAL, this is not so, because the
+		// anonymous function func() { <-exitCalled } is deferred
+		// above. That function ensures that outputLogEntry() will wait
+		// until the exit function has been called. If the exit function
+		// is os.Exit, it will never return, outputLogEntry()'s defer will
+		// never complete and all is well. If the exit function was
+		// overridden, then the client that has overridden the exit
+		// function is expecting log.Fatal to return and all is well too.
 	}
 	l.mu.Unlock()
 }
@@ -1242,17 +1265,33 @@ func (l *loggingT) lockAndSetSync(sync bool) {
 // attempts to sync its data to disk.
 // l.mu is held.
 func (l *loggingT) flushAndSync(doSync bool) {
-	if l.file != nil {
-		_ = l.file.Flush() // ignore error
-		if doSync {
-			_ = l.file.Sync() // ignore error
-		}
+	if l.file == nil {
+		return
+	}
+
+	// If we can't sync within this duration, exit the process.
+	t := time.AfterFunc(maxSyncDuration, func() {
+		// NB: the disk-stall-detected roachtest matches on this message.
+		Shout(context.Background(), Severity_FATAL, fmt.Sprintf(
+			"disk stall detected: unable to sync log files within %s", maxSyncDuration,
+		))
+	})
+	defer t.Stop()
+
+	_ = l.file.Flush() // ignore error
+	if doSync {
+		_ = l.file.Sync() // ignore error
 	}
 }
 
-func (l *loggingT) gcDaemon() {
+func (l *loggingT) gcDaemon(ctx context.Context) {
 	l.gcOldFiles()
-	for range l.gcNotify {
+	for {
+		select {
+		case <-ctx.Done():
+			return
+		case <-l.gcNotify:
+		}
 		l.mu.Lock()
 		if !l.disableDaemons {
 			l.gcOldFiles()
@@ -1354,9 +1393,9 @@ func NewStdLogger(severity Severity) *stdLog.Logger {
 // of its .go suffix, and uses filepath.Match, which is a little more
 // general than the *? matching used in C++.
 // l.mu is held.
-func (l *loggingT) setV(pc uintptr) level {
-	fn := runtime.FuncForPC(pc)
-	file, _ := fn.FileLine(pc)
+func (l *loggingT) setV(pc [1]uintptr) level {
+	frame, _ := runtime.CallersFrames(pc[:]).Next()
+	file := frame.File
 	// The file is something like /a/b/c/d.go. We want just the d.
 	if strings.HasSuffix(file, ".go") {
 		file = file[:len(file)-3]
@@ -1366,11 +1405,11 @@ func (l *loggingT) setV(pc uintptr) level {
 	}
 	for _, filter := range l.vmodule.filter {
 		if filter.match(file) {
-			l.vmap[pc] = filter.level
+			l.vmap[pc[0]] = filter.level
 			return filter.level
 		}
 	}
-	l.vmap[pc] = 0
+	l.vmap[pc[0]] = 0
 	return 0
 }
 
@@ -1440,7 +1479,7 @@ func VDepth(l int32, depth int) bool {
 		logging.mu.Lock()
 		v, ok := logging.vmap[pcs[0]]
 		if !ok {
-			v = logging.setV(pcs[0])
+			v = logging.setV(pcs)
 		}
 		logging.mu.Unlock()
 		logging.pcsPool.Put(poolObj)
