@@ -20,8 +20,6 @@ import (
 	"math"
 	"math/bits"
 
-	"github.com/pkg/errors"
-
 	"github.com/cockroachdb/cockroach/pkg/settings/cluster"
 	"github.com/cockroachdb/cockroach/pkg/util"
 	"github.com/cockroachdb/cockroach/pkg/util/envutil"
@@ -29,6 +27,7 @@ import (
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/metric"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
+	"github.com/pkg/errors"
 )
 
 // BoundAccount and BytesMonitor together form the mechanism by which
@@ -394,11 +393,9 @@ func (mm *BytesMonitor) doStop(ctx context.Context, check bool) {
 	}
 
 	if check && mm.mu.curAllocated != 0 {
-		var reportables []interface{}
 		log.ReportOrPanic(
 			ctx, &mm.settings.SV,
-			fmt.Sprintf("%s: unexpected %d leftover bytes", mm.name, mm.mu.curAllocated),
-			reportables)
+			fmt.Sprintf("%s: unexpected %d leftover bytes", mm.name, mm.mu.curAllocated))
 		mm.releaseBytes(ctx, mm.mu.curAllocated)
 	}
 
@@ -600,14 +597,16 @@ func (mm *BytesMonitor) reserveBytes(ctx context.Context, x int64) error {
 	}
 
 	// Report "large" queries to the log for further investigation.
-	if mm.mu.curAllocated > mm.noteworthyUsageBytes {
-		// We only report changes in binary magnitude of the size. This is to
-		// limit the amount of log messages when a size blowup is caused by
-		// many small allocations.
-		if bits.Len64(uint64(mm.mu.curAllocated)) != bits.Len64(uint64(mm.mu.curAllocated-x)) {
-			log.Infof(ctx, "%s: bytes usage increases to %s (+%d)",
-				mm.name,
-				humanizeutil.IBytes(mm.mu.curAllocated), x)
+	if log.V(1) {
+		if mm.mu.curAllocated > mm.noteworthyUsageBytes {
+			// We only report changes in binary magnitude of the size. This is to
+			// limit the amount of log messages when a size blowup is caused by
+			// many small allocations.
+			if bits.Len64(uint64(mm.mu.curAllocated)) != bits.Len64(uint64(mm.mu.curAllocated-x)) {
+				log.Infof(ctx, "%s: bytes usage increases to %s (+%d)",
+					mm.name,
+					humanizeutil.IBytes(mm.mu.curAllocated), x)
+			}
 		}
 	}
 
@@ -626,8 +625,10 @@ func (mm *BytesMonitor) releaseBytes(ctx context.Context, sz int64) {
 	mm.mu.Lock()
 	defer mm.mu.Unlock()
 	if mm.mu.curAllocated < sz {
-		panic(fmt.Sprintf("%s: no bytes to release, current %d, free %d",
-			mm.name, mm.mu.curAllocated, sz))
+		sz = mm.mu.curAllocated
+		log.ReportOrPanic(ctx, &mm.settings.SV,
+			"%s: no bytes to release, current %d, free %d",
+			mm.name, mm.mu.curAllocated, sz)
 	}
 	mm.mu.curAllocated -= sz
 	if mm.curBytesCount != nil {
