@@ -16,9 +16,7 @@
 
 package timeutil
 
-import (
-	"time"
-)
+import "time"
 
 // Now returns the current UTC time.
 func Now() time.Time {
