@@ -14,9 +14,7 @@
 
 package examples
 
-import (
-	"github.com/cockroachdb/cockroach/pkg/workload"
-)
+import "github.com/cockroachdb/cockroach/pkg/workload"
 
 const (
 	episodesSchema = `(id INT PRIMARY KEY, season INT, num INT, title TEXT, stardate DECIMAL)`
@@ -33,10 +31,11 @@ func init() {
 }
 
 var startrekMeta = workload.Meta{
-	Name:        `startrek`,
-	Description: `Star Trek models episodes and quotes from the tv show`,
-	Version:     `1.0.0`,
-	New:         func() workload.Generator { return startrek{} },
+	Name:         `startrek`,
+	Description:  `Star Trek models episodes and quotes from the tv show`,
+	Version:      `1.0.0`,
+	PublicFacing: true,
+	New:          func() workload.Generator { return startrek{} },
 }
 
 // Meta implements the Generator interface.
