@@ -18,30 +18,36 @@ package tpcc
 import (
 	"context"
 	gosql "database/sql"
-	"math/rand"
+	"fmt"
 	"net/url"
 	"strings"
 	"sync"
-
-	"fmt"
 	"time"
 
 	"github.com/cockroachdb/cockroach/pkg/util/log"
 	"github.com/cockroachdb/cockroach/pkg/util/syncutil"
 	"github.com/cockroachdb/cockroach/pkg/util/timeutil"
+	"github.com/cockroachdb/cockroach/pkg/util/uint128"
+	"github.com/cockroachdb/cockroach/pkg/util/uuid"
 	"github.com/cockroachdb/cockroach/pkg/workload"
+	"github.com/cockroachdb/cockroach/pkg/workload/histogram"
+	"github.com/jackc/pgx"
 	"github.com/pkg/errors"
 	"github.com/spf13/pflag"
+	"golang.org/x/exp/rand"
+	"golang.org/x/sync/errgroup"
 )
 
 type tpcc struct {
-	flags workload.Flags
+	flags     workload.Flags
+	connFlags *workload.ConnFlags
 
-	seed             int64
+	seed             uint64
 	warehouses       int
 	activeWarehouses int
 	interleaved      bool
 	nowString        string
+	numConns         int
 
 	mix        string
 	doWaits    bool
@@ -49,13 +55,13 @@ type tpcc struct {
 	fks        bool
 	dbOverride string
 
-	txs []tx
-	// deck contains indexes into the txs slice.
+	txInfos []txInfo
+	// deck contains indexes into the txInfos slice.
 	deck []int
 
 	auditor *auditor
 
-	reg *workload.HistogramRegistry
+	reg *histogram.Registry
 
 	split   bool
 	scatter bool
@@ -63,10 +69,11 @@ type tpcc struct {
 	partitions        int
 	affinityPartition int
 	zones             []string
+	wPart             *partitioner
 
 	usePostgres  bool
 	serializable bool
-	txOpts       *gosql.TxOptions
+	txOpts       *pgx.TxOptions
 
 	expensiveChecks bool
 
@@ -74,7 +81,7 @@ type tpcc struct {
 		syncutil.Mutex
 		values [][]int
 	}
-	rngPool *sync.Pool
+	localsPool *sync.Pool
 }
 
 func init() {
@@ -84,16 +91,17 @@ func init() {
 // FromWarehouses returns a tpcc generator pre-configured with the specified
 // number of warehouses.
 func FromWarehouses(warehouses int) workload.Generator {
-	gen := tpccMeta.New().(*tpcc)
-	gen.warehouses = warehouses
-	return gen
+	return workload.FromFlags(tpccMeta, fmt.Sprintf(`--warehouses=%d`, warehouses))
 }
 
 var tpccMeta = workload.Meta{
 	Name: `tpcc`,
 	Description: `TPC-C simulates a transaction processing workload` +
 		` using a rich schema of multiple tables`,
-	Version: `2.0.1`,
+	// TODO(anyone): when bumping this version and regenerating fixtures, please
+	// address the TODO in PostLoad.
+	Version:      `2.0.1`,
+	PublicFacing: true,
 	New: func() workload.Generator {
 		g := &tpcc{}
 		g.flags.FlagSet = pflag.NewFlagSet(`tpcc`, pflag.ContinueOnError)
@@ -108,12 +116,13 @@ var tpccMeta = workload.Meta{
 			`split`:              {RuntimeOnly: true},
 			`wait`:               {RuntimeOnly: true},
 			`workers`:            {RuntimeOnly: true},
+			`conns`:              {RuntimeOnly: true},
 			`zones`:              {RuntimeOnly: true},
 			`active-warehouses`:  {RuntimeOnly: true},
 			`expensive-checks`:   {RuntimeOnly: true, CheckConsistencyOnly: true},
 		}
 
-		g.flags.Int64Var(&g.seed, `seed`, 1, `Random number generator seed`)
+		g.flags.Uint64Var(&g.seed, `seed`, 1, `Random number generator seed`)
 		g.flags.IntVar(&g.warehouses, `warehouses`, 1, `Number of warehouses for loading`)
 		g.flags.BoolVar(&g.interleaved, `interleaved`, false, `Use interleaved tables`)
 		// Hardcode this since it doesn't seem like anyone will want to change
@@ -123,13 +132,21 @@ var tpccMeta = workload.Meta{
 		g.flags.StringVar(&g.mix, `mix`,
 			`newOrder=10,payment=10,orderStatus=1,delivery=1,stockLevel=1`,
 			`Weights for the transaction mix. The default matches the TPCC spec.`)
+
 		g.flags.BoolVar(&g.doWaits, `wait`, true, `Run in wait mode (include think/keying sleeps)`)
 		g.flags.StringVar(&g.dbOverride, `db`, ``,
 			`Override for the SQL database to use. If empty, defaults to the generator name`)
-		g.flags.IntVar(&g.workers, `workers`, 0,
-			`Number of concurrent workers. Defaults to --warehouses * 10`)
+
+		g.flags.IntVar(&g.workers, `workers`, 0, fmt.Sprintf(
+			`Number of concurrent workers. Defaults to --warehouses * %d`, numWorkersPerWarehouse,
+		))
+		g.flags.IntVar(&g.numConns, `conns`, 0, fmt.Sprintf(
+			`Number of connections. Defaults to --warehouses * %d (except in nowait mode, where it defaults to --workers`,
+			numConnsPerWarehouse,
+		))
+
 		g.flags.BoolVar(&g.fks, `fks`, true, `Add the foreign keys`)
-		g.flags.IntVar(&g.partitions, `partitions`, 0, `Partition tables (requires split)`)
+		g.flags.IntVar(&g.partitions, `partitions`, 1, `Partition tables (requires split)`)
 		g.flags.IntVar(&g.affinityPartition, `partition-affinity`, -1, `Run load generator against specific partition (requires partitions)`)
 		g.flags.IntVar(&g.activeWarehouses, `active-warehouses`, 0, `Run the load generator against a specific number of warehouses. Defaults to --warehouses'`)
 		g.flags.BoolVar(&g.scatter, `scatter`, false, `Scatter ranges`)
@@ -138,6 +155,7 @@ var tpccMeta = workload.Meta{
 		g.flags.StringSliceVar(&g.zones, "zones", []string{}, "Zones for partitioning, the number of zones should match the number of partitions and the zones used to start cockroach.")
 
 		g.flags.BoolVar(&g.expensiveChecks, `expensive-checks`, false, `Run expensive checks`)
+		g.connFlags = workload.NewConnFlags(&g.flags)
 		return g
 	},
 }
@@ -152,44 +170,55 @@ func (w *tpcc) Flags() workload.Flags { return w.flags }
 func (w *tpcc) Hooks() workload.Hooks {
 	return workload.Hooks{
 		Validate: func() error {
-			if w.activeWarehouses > w.warehouses {
-				return errors.Errorf(`--active-warehouses needs to be less than or equal to warehouses`)
+			if w.warehouses < 1 {
+				return errors.Errorf(`--warehouses must be positive`)
 			}
 
-			if w.activeWarehouses == 0 {
+			if w.activeWarehouses > w.warehouses {
+				return errors.Errorf(`--active-warehouses needs to be less than or equal to warehouses`)
+			} else if w.activeWarehouses == 0 {
 				w.activeWarehouses = w.warehouses
 			}
 
-			if w.workers == 0 {
-				w.workers = w.activeWarehouses * numWorkersPerWarehouse
-			}
-			if w.doWaits && w.workers != w.activeWarehouses*numWorkersPerWarehouse {
-				return errors.Errorf(`--waits=true and --warehouses=%d requires --workers=%d`,
-					w.activeWarehouses, w.warehouses*numWorkersPerWarehouse)
-			}
-
-			if w.partitions != 0 && !w.split {
-				return errors.Errorf(`--partitions requires --split`)
-			}
-
-			if w.affinityPartition != -1 && w.partitions == 0 {
-				return errors.Errorf(`--partition-affinity requires --partitions`)
+			if w.partitions < 1 {
+				return errors.Errorf(`--partitions must be positive`)
+			} else if w.partitions > 1 && !w.split {
+				return errors.Errorf(`multiple partitions requires --split`)
 			}
 
 			if w.affinityPartition < -1 {
-				return errors.Errorf(`--partition-affinity should be greater than or equal to 0`)
-			}
-
-			if w.affinityPartition > w.partitions-1 {
-				return errors.Errorf(`--partition-affinity should be less than the total number of partitions.`)
+				return errors.Errorf(`if specified, --partition-affinity should be greater than or equal to 0`)
+			} else if w.affinityPartition >= w.partitions {
+				return errors.Errorf(`--partition-affinity out of bounds of --partitions`)
 			}
 
 			if len(w.zones) > 0 && (len(w.zones) != w.partitions) {
 				return errors.Errorf(`--zones should have the sames length as --partitions.`)
 			}
 
+			if w.workers == 0 {
+				w.workers = w.activeWarehouses * numWorkersPerWarehouse
+			}
+
+			if w.numConns == 0 {
+				// If we're not waiting, open up a connection for each worker. If we are
+				// waiting, we only use up to a set number of connections per warehouse.
+				// This isn't mandated by the spec, but opening a connection per worker
+				// when they each spend most of their time waiting is wasteful.
+				if !w.doWaits {
+					w.numConns = w.workers
+				} else {
+					w.numConns = w.activeWarehouses * numConnsPerWarehouse
+				}
+			}
+
+			if w.doWaits && w.workers != w.activeWarehouses*numWorkersPerWarehouse {
+				return errors.Errorf(`--wait=true and --warehouses=%d requires --workers=%d`,
+					w.activeWarehouses, w.warehouses*numWorkersPerWarehouse)
+			}
+
 			if w.serializable {
-				w.txOpts = &gosql.TxOptions{Isolation: gosql.LevelSerializable}
+				w.txOpts = &pgx.TxOptions{IsoLevel: pgx.Serializable}
 			}
 
 			w.auditor = newAuditor(w.warehouses)
@@ -197,25 +226,55 @@ func (w *tpcc) Hooks() workload.Hooks {
 			return initializeMix(w)
 		},
 		PostLoad: func(sqlDB *gosql.DB) error {
-			fkStmts := []string{
-				`alter table district add foreign key (d_w_id) references warehouse (w_id)`,
-				`alter table customer add foreign key (c_w_id, c_d_id) references district (d_w_id, d_id)`,
-				`alter table history add foreign key (h_c_w_id, h_c_d_id, h_c_id) references customer (c_w_id, c_d_id, c_id)`,
-				`alter table history add foreign key (h_w_id, h_d_id) references district (d_w_id, d_id)`,
-				`alter table "order" add foreign key (o_w_id, o_d_id, o_c_id) references customer (c_w_id, c_d_id, c_id)`,
-				`alter table stock add foreign key (s_w_id) references warehouse (w_id)`,
-				`alter table stock add foreign key (s_i_id) references item (i_id)`,
-				`alter table order_line add foreign key (ol_w_id, ol_d_id, ol_o_id) references "order" (o_w_id, o_d_id, o_id)`,
-				`alter table order_line add foreign key (ol_supply_w_id, ol_d_id) references stock (s_w_id, s_i_id)`,
-			}
-			for _, fkStmt := range fkStmts {
-				if _, err := sqlDB.Exec(fkStmt); err != nil {
-					// If the statement failed because the fk already exists,
-					// ignore it. Return the error for any other reason.
-					const duplFKErr = "columns cannot be used by multiple foreign key constraints"
-					if !strings.Contains(err.Error(), duplFKErr) {
+			if w.fks {
+				fkStmts := []string{
+					`alter table district add foreign key (d_w_id) references warehouse (w_id)`,
+					`alter table customer add foreign key (c_w_id, c_d_id) references district (d_w_id, d_id)`,
+					`alter table history add foreign key (h_c_w_id, h_c_d_id, h_c_id) references customer (c_w_id, c_d_id, c_id)`,
+					`alter table history add foreign key (h_w_id, h_d_id) references district (d_w_id, d_id)`,
+					`alter table "order" add foreign key (o_w_id, o_d_id, o_c_id) references customer (c_w_id, c_d_id, c_id)`,
+					`alter table stock add foreign key (s_w_id) references warehouse (w_id)`,
+					`alter table stock add foreign key (s_i_id) references item (i_id)`,
+					`alter table order_line add foreign key (ol_w_id, ol_d_id, ol_o_id) references "order" (o_w_id, o_d_id, o_id)`,
+				}
+
+				// TODO(anyone): Remove this check. Once fixtures are
+				// regenerated and the meta version is bumped on this workload,
+				// we won't need it anymore.
+				{
+					const q = `SELECT column_name
+						       FROM information_schema.statistics
+						       WHERE index_name = 'order_line_fk'
+						         AND seq_in_index = 2`
+					var fkCol string
+					if err := sqlDB.QueryRow(q).Scan(&fkCol); err != nil {
 						return err
 					}
+					var fkStmt string
+					switch fkCol {
+					case "ol_i_id":
+						// The corrected column. When the TODO above is addressed,
+						// this should be moved into fkStmts.
+						fkStmt = `alter table order_line add foreign key (ol_supply_w_id, ol_i_id) references stock (s_w_id, s_i_id)`
+					case "ol_d_id":
+						// The old, incorrect column. When the TODO above is addressed,
+						// this should be removed entirely.
+						fkStmt = `alter table order_line add foreign key (ol_supply_w_id, ol_d_id) references stock (s_w_id, s_i_id)`
+					default:
+						return errors.Errorf("unexpected column %q in order_line_fk", fkCol)
+					}
+					fkStmts = append(fkStmts, fkStmt)
+				}
+
+				for _, fkStmt := range fkStmts {
+					if _, err := sqlDB.Exec(fkStmt); err != nil {
+						// If the statement failed because the fk already exists,
+						// ignore it. Return the error for any other reason.
+						const duplFKErr = "columns cannot be used by multiple foreign key constraints"
+						if !strings.Contains(err.Error(), duplFKErr) {
+							return err
+						}
+					}
 				}
 			}
 			return nil
@@ -226,13 +285,13 @@ func (w *tpcc) Hooks() workload.Hooks {
 			fmt.Println(totalHeader)
 
 			const newOrderName = `newOrder`
-			w.reg.Tick(func(t workload.HistogramTick) {
+			w.reg.Tick(func(t histogram.Tick) {
 				if newOrderName == t.Name {
 					tpmC := float64(t.Cumulative.TotalCount()) / startElapsed.Seconds() * 60
 					fmt.Printf("%7.1fs %10.1f %5.1f%% %8.1f %8.1f %8.1f %8.1f %8.1f %8.1f\n",
 						startElapsed.Seconds(),
 						tpmC,
-						100*tpmC/(12.86*float64(w.activeWarehouses)),
+						100*tpmC/(SpecWarehouseFactor*float64(w.activeWarehouses)),
 						time.Duration(t.Cumulative.Mean()).Seconds()*1000,
 						time.Duration(t.Cumulative.ValueAtQuantile(50)).Seconds()*1000,
 						time.Duration(t.Cumulative.ValueAtQuantile(90)).Seconds()*1000,
@@ -265,12 +324,34 @@ func (w *tpcc) Hooks() workload.Hooks {
 
 // Tables implements the Generator interface.
 func (w *tpcc) Tables() []workload.Table {
-	if w.rngPool == nil {
-		w.rngPool = &sync.Pool{
-			New: func() interface{} { return rand.New(rand.NewSource(timeutil.Now().UnixNano())) },
+	if w.localsPool == nil {
+		w.localsPool = &sync.Pool{
+			New: func() interface{} {
+				return &generateLocals{
+					rng: rand.New(rand.NewSource(uint64(timeutil.Now().UnixNano()))),
+				}
+			},
 		}
 	}
 
+	// splits is a convenience method for constructing table splits that returns
+	// a zero value if the workload does not have splits enabled.
+	splits := func(t workload.BatchedTuples) workload.BatchedTuples {
+		if w.split {
+			return t
+		}
+		return workload.BatchedTuples{}
+	}
+
+	// numBatches is a helper to calculate how many split batches exist exist given
+	// the total number of rows and the desired number of rows per split.
+	numBatches := func(total, per int) int {
+		batches := total / per
+		if total%per == 0 {
+			batches--
+		}
+		return batches
+	}
 	warehouse := workload.Table{
 		Name:   `warehouse`,
 		Schema: tpccWarehouseSchema,
@@ -278,6 +359,14 @@ func (w *tpcc) Tables() []workload.Table {
 			w.warehouses,
 			w.tpccWarehouseInitialRow,
 		),
+		Splits: splits(workload.BatchedTuples{
+			NumBatches: numBatches(w.warehouses, numWarehousesPerRange),
+			NumTotal:   w.warehouses,
+			Batch: func(i int) [][]interface{} {
+				return [][]interface{}{{(i + 1) * numWarehousesPerRange}}
+			},
+		}),
+		Stats: w.tpccWarehouseStats(),
 	}
 	district := workload.Table{
 		Name:   `district`,
@@ -286,6 +375,14 @@ func (w *tpcc) Tables() []workload.Table {
 			numDistrictsPerWarehouse*w.warehouses,
 			w.tpccDistrictInitialRow,
 		),
+		Splits: splits(workload.BatchedTuples{
+			NumBatches: numBatches(w.warehouses, numWarehousesPerRange),
+			NumTotal:   w.warehouses,
+			Batch: func(i int) [][]interface{} {
+				return [][]interface{}{{(i + 1) * numWarehousesPerRange, 0}}
+			},
+		}),
+		Stats: w.tpccDistrictStats(),
 	}
 	customer := workload.Table{
 		Name:   `customer`,
@@ -294,6 +391,7 @@ func (w *tpcc) Tables() []workload.Table {
 			numCustomersPerWarehouse*w.warehouses,
 			w.tpccCustomerInitialRow,
 		),
+		Stats: w.tpccCustomerStats(),
 	}
 	history := workload.Table{
 		Name:   `history`,
@@ -302,6 +400,16 @@ func (w *tpcc) Tables() []workload.Table {
 			numCustomersPerWarehouse*w.warehouses,
 			w.tpccHistoryInitialRow,
 		),
+		Splits: splits(workload.BatchedTuples{
+			NumBatches: historyRanges - 1,
+			Batch: func(i int) [][]interface{} {
+				at := uint128.FromInts(uint64(i+1)*numHistoryValsPerRange, 0)
+				return [][]interface{}{
+					{uuid.FromUint128(at).String()},
+				}
+			},
+		}),
+		Stats: w.tpccHistoryStats(),
 	}
 	order := workload.Table{
 		Name:   `order`,
@@ -310,6 +418,7 @@ func (w *tpcc) Tables() []workload.Table {
 			numOrdersPerWarehouse*w.warehouses,
 			w.tpccOrderInitialRow,
 		),
+		Stats: w.tpccOrderStats(),
 	}
 	newOrder := workload.Table{
 		Name:   `new_order`,
@@ -318,6 +427,7 @@ func (w *tpcc) Tables() []workload.Table {
 			numNewOrdersPerWarehouse*w.warehouses,
 			w.tpccNewOrderInitialRow,
 		),
+		Stats: w.tpccNewOrderStats(),
 	}
 	item := workload.Table{
 		Name:   `item`,
@@ -326,6 +436,14 @@ func (w *tpcc) Tables() []workload.Table {
 			numItems,
 			w.tpccItemInitialRow,
 		),
+		Splits: splits(workload.BatchedTuples{
+			NumBatches: numBatches(numItems, numItemsPerRange),
+			NumTotal:   numItems,
+			Batch: func(i int) [][]interface{} {
+				return [][]interface{}{{numItemsPerRange * (i + 1)}}
+			},
+		}),
+		Stats: w.tpccItemStats(),
 	}
 	stock := workload.Table{
 		Name:   `stock`,
@@ -334,6 +452,7 @@ func (w *tpcc) Tables() []workload.Table {
 			numStockPerWarehouse*w.warehouses,
 			w.tpccStockInitialRow,
 		),
+		Stats: w.tpccStockStats(),
 	}
 	orderLine := workload.Table{
 		Name:   `order_line`,
@@ -342,6 +461,7 @@ func (w *tpcc) Tables() []workload.Table {
 			NumBatches: numOrdersPerWarehouse * w.warehouses,
 			Batch:      w.tpccOrderLineInitialRowBatch,
 		},
+		Stats: w.tpccOrderLineStats(),
 	}
 	if w.interleaved {
 		district.Schema += tpccDistrictSchemaInterleave
@@ -360,7 +480,7 @@ func (w *tpcc) Tables() []workload.Table {
 }
 
 // Ops implements the Opser interface.
-func (w *tpcc) Ops(urls []string, reg *workload.HistogramRegistry) (workload.QueryLoad, error) {
+func (w *tpcc) Ops(urls []string, reg *histogram.Registry) (workload.QueryLoad, error) {
 	sqlDatabase, err := workload.SanitizeUrls(w, w.dbOverride, urls)
 	if err != nil {
 		return workload.QueryLoad{}, err
@@ -370,90 +490,108 @@ func (w *tpcc) Ops(urls []string, reg *workload.HistogramRegistry) (workload.Que
 		return workload.QueryLoad{}, err
 	}
 
+	w.reg = reg
 	w.usePostgres = parsedURL.Port() == "5432"
 
-	nConns := w.activeWarehouses / len(urls)
-	dbs := make([]*gosql.DB, len(urls))
-	for i, url := range urls {
-		dbs[i], err = gosql.Open(`postgres`, url)
-		if err != nil {
-			return workload.QueryLoad{}, err
-		}
-		// Allow a maximum of concurrency+1 connections to the database.
-		dbs[i].SetMaxOpenConns(nConns)
-		dbs[i].SetMaxIdleConns(nConns)
+	// We can't use a single MultiConnPool because we want to implement partition
+	// affinity. Instead we have one MultiConnPool per server.
+	cfg := workload.MultiConnPoolCfg{
+		MaxTotalConnections: (w.numConns + len(urls) - 1) / len(urls), // round up
+		// Limit the number of connections per pool (otherwise preparing statements
+		// at startup can be slow).
+		MaxConnsPerPool: 50,
+	}
+	fmt.Printf("Initializing %d connections...\n", w.numConns)
+	dbs := make([]*workload.MultiConnPool, len(urls))
+	var g errgroup.Group
+	for i := range urls {
+		i := i
+		g.Go(func() error {
+			var err error
+			dbs[i], err = workload.NewMultiConnPool(cfg, urls[i])
+			return err
+		})
+	}
+	if err := g.Wait(); err != nil {
+		return workload.QueryLoad{}, err
+	}
+
+	// Create a partitioner to help us partition the warehouses. The base-case is
+	// where w.warehouses == w.activeWarehouses and w.partitions == 1.
+	w.wPart, err = makePartitioner(w.warehouses, w.activeWarehouses, w.partitions)
+	if err != nil {
+		return workload.QueryLoad{}, err
 	}
 
 	// We're adding this check here because repartitioning a table can take
 	// upwards of 10 minutes so if a cluster is already set up correctly we won't
 	// do this operation again.
-	alreadyPartitioned, err := isTableAlreadyPartitioned(dbs[0])
+	alreadyPartitioned, err := isTableAlreadyPartitioned(dbs[0].Get())
 	if err != nil {
 		return workload.QueryLoad{}, err
 	}
 
-	if !alreadyPartitioned {
-		if w.split {
-			splitTables(dbs[0], w.warehouses)
-
-			if w.partitions > 0 {
-				partitionTables(dbs[0], w.warehouses, w.partitions, w.zones)
-			}
-		}
-	} else {
-		fmt.Println("Tables are not being parititioned because they've been previously partitioned.")
+	if shouldPartition := w.split && w.partitions > 1; shouldPartition && !alreadyPartitioned {
+		partitionTables(dbs[0].Get(), w.wPart, w.zones)
+	} else if shouldPartition {
+		fmt.Println("Tables are not being partitioned because they've been previously partitioned.")
 	}
 
 	if w.scatter {
-		scatterRanges(dbs[0])
+		scatterRanges(dbs[0].Get())
 	}
 
-	// If no partitions were specified, pretend there is a single partition
-	// containing all warehouses.
-	if w.partitions == 0 {
-		w.partitions = 1
-	}
-	// Assign each DB connection pool to a partition. This assumes that dbs[i] is
-	// a machine that holds partition "i % *partitions".
-	partitionDBs := make([][]*gosql.DB, w.partitions)
-	if w.affinityPartition == -1 {
+	// Assign each DB connection pool to a local partition. This assumes that
+	// dbs[i] is a machine that holds partition "i % *partitions". If we have an
+	// affinity partition, all connections will be for the same partition.
+	partitionDBs := make([][]*workload.MultiConnPool, w.partitions)
+	if w.affinityPartition >= 0 {
+		// All connections are for our local partition.
+		partitionDBs[w.affinityPartition] = dbs
+	} else {
 		for i, db := range dbs {
 			p := i % w.partitions
 			partitionDBs[p] = append(partitionDBs[p], db)
 		}
-	}
-	for i := range partitionDBs {
-		if partitionDBs[i] == nil {
-			partitionDBs[i] = dbs
+		for i := range partitionDBs {
+			// Possible if we have more partitions than DB connections.
+			if partitionDBs[i] == nil {
+				partitionDBs[i] = dbs
+			}
 		}
 	}
 
-	w.reg = reg
-
-	startWarehouse := 0
-	if w.affinityPartition != -1 {
-		startWarehouse = w.affinityPartition * (w.warehouses / w.partitions)
-	}
+	fmt.Printf("Initializing %d workers and preparing statements...\n", w.workers)
 	ql := workload.QueryLoad{SQLDatabase: sqlDatabase}
-	for workerIdx := 0; workerIdx < w.workers; workerIdx++ {
-		warehouse := workerIdx % w.activeWarehouses
-		// NB: Each partition contains "warehouses / partitions" warehouses. See
-		// partitionTables().
-		p := (warehouse * w.partitions) / w.activeWarehouses
+	ql.WorkerFns = make([]func(context.Context) error, w.workers)
+	var group errgroup.Group
+	// Limit the amount of workers we initialize in parallel, to avoid running out
+	// of memory (#36897).
+	sem := make(chan struct{}, 100)
+	for workerIdx := range ql.WorkerFns {
+		workerIdx := workerIdx
+		warehouse := w.wPart.totalElems[workerIdx%len(w.wPart.totalElems)]
+
+		p := w.wPart.partElemsMap[warehouse]
+		if w.affinityPartition >= 0 && w.affinityPartition != p {
+			// This isn't part of our local partition.
+			continue
+		}
 		dbs := partitionDBs[p]
 		db := dbs[warehouse%len(dbs)]
-		worker := &worker{
-			config:    w,
-			hists:     reg.GetHandle(),
-			idx:       workerIdx,
-			db:        db,
-			warehouse: warehouse + startWarehouse,
-			deckPerm:  make([]int, len(w.deck)),
-			permIdx:   len(w.deck),
-		}
-		copy(worker.deckPerm, w.deck)
 
-		ql.WorkerFns = append(ql.WorkerFns, worker.run)
+		sem <- struct{}{}
+		group.Go(func() error {
+			worker, err := newWorker(context.TODO(), w, db, reg.GetHandle(), warehouse)
+			if err == nil {
+				ql.WorkerFns[workerIdx] = worker.run
+			}
+			<-sem
+			return err
+		})
+	}
+	if err := group.Wait(); err != nil {
+		return workload.QueryLoad{}, err
 	}
 	// Preregister all of the histograms so they always print.
 	for _, tx := range allTxs {
